# (c) Copyright 2011-2012 Cloudera, Inc.
import os
import simplejson
import logging
import codecs
import resource

logger = logging.getLogger()

class MachineCollectionFsDao(object):
    """DAO for accessing machine collection data from a cluster statistics
    collection extraction."""

    def __init__(self, machine_root_dir):
        self.__root_dir = machine_root_dir

    def get_root_dir(self):
        return self.__root_dir

    def get_run_log(self):
        return self.__read_and_close(os.path.join(self.__root_dir, "run.log"))

    ### Sys Stat Methods ###

    def get_sys_stat(self, key, std_stream):
        assert std_stream in ("stdout", "stderr", "retcode")
        path = os.path.join(self.__get_sys_stat_dir(), key + "_" + std_stream)
        return self.__read_and_close(path)

    def get_sys_stats_run(self):
        path = os.path.join(self.__get_sys_stat_dir(), "stats.run")
        return self.__get_stats_run(path)

    def put_sys_stat(self, key, stdout_data, stderr_data, retcode=None):
        self.__create_dir_if_not_exist(self.__get_sys_stat_dir())
        self.__write_and_close(
            os.path.join(self.__get_sys_stat_dir(), key + "_stdout"), stdout_data)
        self.__write_and_close(
            os.path.join(self.__get_sys_stat_dir(), key + "_stderr"), stderr_data)
        if retcode is not None:
            self.__write_and_close(
                os.path.join(self.__get_sys_stat_dir(), key + "_retcode"), str(retcode))

    def put_sys_stats_run(self, stats):
        self.__create_dir_if_not_exist(self.__get_sys_stat_dir())
        path = os.path.join(self.__get_sys_stat_dir(), "stats.run")
        self.__put_stats_run(path, stats)

    ### Performance Measuring Methods ###
    def initialize_performance(self):
        self.__performance_data = []
        self.update_performance("start")

    def update_performance(self, stat_key):
        if self.__performance_data is not None:
            self_perf = resource.getrusage(resource.RUSAGE_SELF)
            child_perf = resource.getrusage(resource.RUSAGE_CHILDREN)
            self.__performance_data.append({"command": stat_key,
                                            "self_usage": self.__convert_resource(self_perf),
                                            "child_usage": self.__convert_resource(child_perf)})
        else:
            logger.error("Cannot call update_performance before initializing.")

    def __convert_resource(self, r):
        """Converts an object of type resource.struct_rusage to a dictionary from fields we
        are interested in to values. These are all named properties starting with 'ru_'."""
        return dict( (a, getattr(r, a)) for a in dir(r) if a.startswith("ru_") )

    def complete_performance(self):
        if self.__performance_data is not None:
            self.update_performance("end")
            self.__write_and_close(
                os.path.join(self.__get_sys_stat_dir(), "performance.json"),
                simplejson.dumps(self.__performance_data))
            self.__performance_data = []
        else:
            logger.error("Cannot call complete_performance before initializing.")

    ### Private Helper Methods ###

    def __get_sys_stat_dir(self):
        return os.path.join(self.__root_dir, "sysstats")

    def __create_dir_if_not_exist(self, dir_path):
        if not os.path.isdir(dir_path):
            os.mkdir(dir_path)

    def __read_and_close(self, path):
        try:
            fd = codecs.open(path, "r", encoding="utf-8", errors="replace")
            try:
                return fd.read()
            finally:
                fd.close()
        except IOError:
            return None

    def __write_and_close(self, path, data):
        fd = codecs.open(path, "w", encoding="utf-8", errors="replace")
        try:
            if data is not None:
                fd.write(data)
        finally:
            fd.close()

    def __get_stats_run(self, path):
        stats_run_data = self.__read_and_close(path)
        if stats_run_data is not None:
            return simplejson.loads(stats_run_data)
        else:
            return None

    def __put_stats_run(self, path, stats):
        handle = open(path, "w")
        try:
            simplejson.dump(stats, handle, indent = 4)
        finally:
            handle.close()
