# (c) Copyright 2011-2012 Cloudera, Inc.

"""This module contains the stat collectors, which when run, will collect stats.
Each stat collector can also report on the location of each stat it collects on
the file system, relative to a root dir."""

import fcntl
import os
import signal
import logging
import platform
import re
import select
import simplejson
import supervisor.options
import sys
import traceback
from subprocess import Popen, PIPE
from threading import Timer
from datetime import datetime

logger = logging.getLogger()

DEFAULT_TIMEOUT=60
HOST_COMMAND_TERMINATED="Host Command Terminated"


class StatResult(object):
    """The result of a stat collection. Encapsulates the result data collected,
    and potentially error_data."""

    def __init__(self, result_data, error_data=None, other=None):
        self.__result_data = result_data
        self.__error_data = error_data
        self.__other = other

    def get_result_data(self):
        return self.__result_data

    def get_error_data(self):
        return self.__error_data

    def get_other(self):
        return self.__other

    def apply_processor(self, processor):
        self.__result_data = processor(self.__result_data)
        self.__error_data = processor(self.__error_data)


class ProcWrapper(object):
    """Wraps a POPEN PROC so we can provide a kill method that
    can be called to kill the execution if command does not complete before a timeout"""
    
    def __init__(self, command_list, new_env, timeout):
        # Set process group of command. Required because some host stats invoke bash.
        self.__proc = Popen(command_list, bufsize=-1, stdout=PIPE, stderr=PIPE, stdin=None, env=new_env, close_fds=True, preexec_fn=os.setpgrp)
        if timeout is None:
            timeout = DEFAULT_TIMEOUT
        self.__timer = Timer(timeout, self.kill);
        self.__timer.start()
        self.__killed = False
        self.__command_list = command_list
    
    def get_proc(self):
        return self.__proc    
        
    def get_timer(self):
        return self.__timer    
        
    def was_killed(self):
        return self.__killed
    
    def done(self):
	logger.info("Done %s at %s" % (str(self.__command_list), datetime.now()))

    def kill(self):
        logger.info("Kill (%s) at %s" % (str(self.__command_list), datetime.now()))
        self.__killed = True
        # Kill process group to make sure all sub-processes are killed too.
        os.killpg(os.getpgid(self.__proc.pid), signal.SIGTERM)
        
class Stat(object):
    """The base class for a stat that can be collected. The collect member
    function should be overridden in subclasses."""

    def collect(self, context, dependent_results, timeout):
        """Should be overridden by subclasses. Context is a dictionary, which
        Stats have the ability to add to or read from. It is useful if there
        are certain pieces of data a Stat needs that should be fetched while
        collecting the stat. The dependent_results parameter is a dictionary of
        StatResults of other Stats that were collected."""
        raise NotImplementedError


class CommandStat(Stat):
    """A stat that runs a process by launching the process and its arguments
    that are passed into its constructor.
    """

    def __init__(self, command_list, extra_env_path=None):
        """Construct the CommandStat, where the command_list parameter is the
        process and its arguments to launch, and extra_env_path are extra env
        variables to include in the environment of the launching process. The
        first item in the list of command_list is the executable, and the rest
        of the elements are the arguments."""
        self.__command_list = command_list
        if extra_env_path is None:
            extra_env_path = []
        self.__extra_env_path = extra_env_path

    def __str__(self):
        """Returns the command list as a string."""
        return " ".join(self.__command_list)

    def collect(self, context, dependent_results, timeout):
        """Collect the standard output and standard error from the command, and
        return those results wrapped in a StatResult instance."""
        try:
            wrapper = self._get_command_proc(timeout)
            proc = wrapper.get_proc()
            try:
                stdout_data, stderr_data = [self._decode_string(stream)
                                            for stream in proc.communicate()]
                retcode = proc.wait()
                wrapper.done()
            finally:
                wrapper.get_timer().cancel()
            if wrapper.was_killed() is True:
                stdout_data += self._decode_string(HOST_COMMAND_TERMINATED)
            return StatResult(stdout_data, stderr_data, retcode)
        except OSError, exc:
            command = " ".join(self.__command_list)
            logger.error("Error running command %s. %s" % (command, str(exc)))
            raise

    def _decode_string(self, string, encoding="utf-8", errors="replace"):
        """Decodes the string the the given unicode encoding and error replacement
        strategy. You should call this on strings pulled from sysstat commands or
        else non-unicode characters can cause errors when results are written."""
        if string is not None:
            return unicode(string, encoding, errors=errors)
        return None

    def _get_command_proc(self, timeout):
        """Runs a command in a process, and returns the Proc instance, which can
        be used to grab the stdout and stderr streams, as well as the return
        code."""
        new_env = os.environ.copy()
        new_env["PATH"] = "%s:/sbin:/usr/sbin" % new_env["PATH"]
        for p in self.__extra_env_path:
            new_env["PATH"] = "%s:%s" % (new_env["PATH"], p)

        # When the CM agent is running as non-root it will set the SUDO
        # environment variable with the configured sudo command. If that
        # is the case we use sudo to run the stat command.
        command_list = list(self.__command_list)
        if "SUDO" in new_env:
            command_list.insert(0, new_env["SUDO"])

        logger.info("Running %s at %s" % (str(command_list), datetime.now()))
        wrapper = ProcWrapper(command_list, new_env, timeout)
        return wrapper


class StreamingCommandStat(CommandStat):
    """CommandStat that allows command stdout and stderr streams to be read
    and processed one line at a time, rather than reading them entirely into
    memory before they can be processed.

    As each line of stdout or stderr is read, it is presented to a method for
    processing. Subclasses can override these methods to implement filters or
    other transformations on the output. This allows us, for example, to filter
    the output of commands that produce large amounts of output without using
    an excessive amount of memory.
    """

    class StreamData(object):
        """Helper class for accumulating stream data. Stores the most recent
        partial input line.
        """
        def __init__(self):
            self.line_fragment = ""
            self.output_lines = []

    def collect(self, context, dependent_results, timeout):
        """Collects the standard output and standard error from the command,
        and returns those results wrapped in a StatResult instance. Rather
        than reading the output streams entirely into memory, buffers them
        in, and presents them one line at a time for processing by subclasses.
        """

        try:
            wrapper = self._get_command_proc(timeout)
            proc = wrapper.get_proc()
            
            try:
                # Set stdout/stderr streams non-blocking.
                for fd in [proc.stdout.fileno(), proc.stderr.fileno()]:
                    flags = fcntl.fcntl(fd, fcntl.F_GETFL)
                    fcntl.fcntl(fd, fcntl.F_SETFL, flags | os.O_NONBLOCK)
    
                eof = False
                exit_code = None
                stdout_data = self.StreamData()
                stderr_data = self.StreamData()
    
                # Read and process stdout and stderr of the subprocess until both
                # are exhausted. Note that we can't block on input from a given
                # stream, since the subprocess may be blocked on output on the
                # other stream, leading to deadlock.
                while (exit_code is None) or not eof:
                    exit_code = proc.poll()
                    ready = select.select([proc.stdout, proc.stderr], [], [])
    
                    stdout_eof = False
                    if proc.stdout in ready[0]:
                        stdout_eof = self._process_stream_data(proc.stdout,
                                                               stdout_data,
                                                               self._process_stdout_line)
                    stderr_eof = False
                    if proc.stderr in ready[0]:
                        stderr_eof = self._process_stream_data(proc.stderr,
                                                               stderr_data,
                                                               self._process_stderr_line)
                    eof = stdout_eof and stderr_eof
                wrapper.done()
            finally:
                wrapper.get_timer().cancel()
            if wrapper.was_killed() is True:
                stdout_data.output_lines.append(HOST_COMMAND_TERMINATED)
            return StatResult("".join(stdout_data.output_lines),
                              "".join(stderr_data.output_lines), exit_code)
        except OSError, exc:
            command = " ".join(self.__command_list)
            logger.error("Error running command %s. %s" % (command, str(exc)))
            raise

    def _process_stream_data(self, stream, stream_data, process_line):
        """Does a nonblocking read on the given stream, and invokes the
        specified method to process all complete lines in the input.
        """
        eof = False
        buf = stream.read(8192)
        if buf:
            buf = stream_data.line_fragment + self._decode_string(buf)
            stream_data.line_fragment = ""
            for line in buf.splitlines(True):
                if line.endswith("\n"):
                    processed_line = process_line(line)
                    if processed_line:
                        stream_data.output_lines.append(processed_line)
                else:
                    stream_data.line_fragment = line
        else:
            if stream_data.line_fragment:
                processed_line = process_line(stream_data.line_fragment)
                if processed_line:
                    stream_data.output_lines.append(processed_line)
                stream_data.line_fragment = ""
            eof = True

        return eof

    def _process_stdout_line(self, line):
        return line

    def _process_stderr_line(self, line):
        return line


class StatChainRunner(object):
    """Runs a chain of Stat instances. An instance of this should never be
    created directly, but instances should be created using the
    StatChainRunnerBuilder.

    An instance of the builder can be obtained from the static get_builder
    instance of StatChainRunner. The builder can set the persister, and the
    Stat instances to run, as well as determine the dependency among stats,
    and verify that they be run in the right order. An example of using the
    builder:

    my_stat_chain_runner = StatChainRunner.get_builder(). \
        set_persister(some_persister). \
        add("stat_key1", some_stat_instance1). \
        add("stat_key2", some_stat_instance2, ["stat_key1"]). \
        build()
    """

    class StatChainRunnerBuilderException(Exception):
        """An exception type for exceptions that occur during construction
        of a StatChainRunner."""
        pass

    class StatChainRunnerBuilder(object):
        """A builder that is responsible for constructing StatChainRunner
        instances. Stat instances can be added using the add method, and
        a persister can be set using the set_persister method.

        When build is called, the builder determines the order the Stat
        instances should be run in, verifies that there are no cyclical
        dependencies, and verifies that a persister was provided. If there
        are errors, build will throw StatChainRunnerBuilderException,
        otherwise, it will pass that info along to a StatChainRunner
        instance that can be used to run and collect the stats.
        """

        def __init__(self):
            self.__persister = None
            self.__chain = {}
            self.__processors = []
            self.__timeout = None

        def set_persister(self, persister):
            """Sets a persister that will be called for each stat collected.
            Returns self so method calls can be chained."""
            self.__persister = persister
            return self

        def set_timeout(self, timeout):
            """Sets timeout for command execution.
            Commands will be killed if not completed before timeout."""
            self.__timeout = timeout
            return self
            
        def remove_persister(self):
            """Removes a persister from the builder. Returns self so method
            calls can be chained."""
            self.__persister = None
            return self

        def set_processors(self, processors):
            """Adds processors that will be called for each stat collected.
            Returns self so method calls can be chained.

            A processor is a callable that takes a string and returns the
            processed string. Interface:

            def process(data):
                # Perform operations on data
                return data"""
            self.__processors = processors or []
            return self

        def add(self, key, stat, dependency_keys=None):
            """Adds a stat to the chain. Each stat must have a key so that it
            can be referenced as a dependency. The dependency_keys parameter
            is a list of keys this stat has a dependency on, which means those
            dependent stats will run before this stat, and their results will
            be passed to this stat.

            Returns self so that method calls can be chained.
            """
            if dependency_keys is None:
                dependency_keys = []
            if key in self.__chain:
                msg = "Key %s already in chain." % (key,)
                raise StatChainRunner.StatChainRunnerBuilderException(msg)
            self.__chain[key] = (stat, dependency_keys)
            return self

        def add_all(self, stat_tuples):
            """Given a list of stat_tuples, which are of the form (key, stat,
            dependency_keys) (the same parameters sent to the add member func,
            it will add all those stat_tuples to the chain."""
            for key, stat, dependency_keys in stat_tuples:
                self.add(key, stat, dependency_keys)

        def remove(self, key):
            """Removes stat with key represented by parameter key from the chain
            Returns self so that method calls can be chained."""
            del self.__chain[key]
            return self

        def clear(self):
            """Clears all stats from the chain. Returns self so that method
            calls can be chained."""
            self.__chain.clear()
            return self

        def get_stats_dict(self):
            stats_dict = {}
            for stat_key, (stat, _) in self.__chain.iteritems():
                stats_dict[stat_key] = stat
            return stats_dict

        def build(self):
            """Will build a StatChainRunner instance from the Stat instances
            that should be run, and the persister. This will throw an exception
            if a persister has not been provided. It will throw an exception if
            there are cycles in the dependency chain. Finally, it will throw an
            exception if a dependency was specified that does not exist.

            This function will return the StatChainRunner instance that is
            configured to run the Stat instances in a valid order.
            """
            if self.__persister is None:
                msg = "Must provide a persister to build a StatChainRunner"
                raise StatChainRunner.StatChainRunnerBuilderException(msg)
            run_order = self.__get_run_order()
            return StatChainRunner(self.__chain, run_order, self.__persister, self.__processors, self.__timeout)

        def __get_run_order(self):
            run_order_keys = []
            for stat_key, (stat, dependency_keys) in self.__chain.iteritems():
                if len(dependency_keys) == 0 and stat_key not in run_order_keys:
                    run_order_keys.insert(0, stat_key)
                elif stat_key not in run_order_keys:
                    if self.__are_there_cycles(set([stat_key]), dependency_keys):
                        msg = "Cycle in the stat dependencies: %s" % (str(self.__chain))
                        raise StatChainRunner.StatChainRunnerBuilderException(msg)
                    self.__append_dependencies(run_order_keys, dependency_keys)
                    run_order_keys.append(stat_key)
                    
            return run_order_keys

        def __append_dependencies(self, dep_list, dependency_keys):
            for dependency_key in dependency_keys:
                if dependency_key not in self.__chain:
                    msg = "Dependency key %s not in chain." % (dependency_key)
                    raise StatChainRunner.StatChainRunnerBuilderException(msg)
                self.__append_dependencies(dep_list, self.__chain[dependency_key][1])
                if dependency_key not in dep_list:
                    dep_list.append(dependency_key)

        def __are_there_cycles(self, ancestor_keys, dependency_keys):
            for ancestor_key in ancestor_keys:
                if ancestor_key in dependency_keys:
                    return True
            for dependency_key in dependency_keys:
                if dependency_key not in self.__chain:
                    msg = "Dependency key %s not in chain." % (dependency_key)
                    raise StatChainRunner.StatChainRunnerBuilderException(msg)
                new_ancestor_keys = set(ancestor_keys)
                new_ancestor_keys.add(dependency_key)
                if self.__are_there_cycles(
                    new_ancestor_keys, self.__chain[dependency_key][1]):
                    return True
            return False

    @staticmethod
    def get_builder():
        """Get a builder instance that knows how to construct a StatChainRunner
        instance."""
        return StatChainRunner.StatChainRunnerBuilder()

    def __init__(self, chain, run_order, persister, processors, timeout):
        """This constructor should never be called by user code, but should be
        called by a builder."""
        self.__chain = chain
        self.__run_order = run_order
        self.__persister = persister
        self.__processors = processors
        self.__dependency_keys = set()
        for _, dependencies in chain.itervalues():
            for dependency in dependencies:
                self.__dependency_keys.add(dependency)
        self.__timeout = timeout

    def run(self, context=None):
        """Runs the chain of Stat instances. The results of any dependencies
        will be passed to any stat which has a dependency on those. After
        each stat collect is run, the results will be sent to the configured
        persister. Context can be a dictionary that will be passed to each
        Stat, which can read and modify it, when that Stat's collect method
        is invoked."""
        dependency_results = {}
        if context is None:
            context = {}
        for stat_key in self.__run_order:
            stat, dependencies = self.__chain[stat_key]
            dependent_results = {}
            for dependency in dependencies:
                dependent_results[dependency] = dependency_results[dependency]
            try:
                timeout = self.__timeout
                stat_result = stat.collect(context, dependent_results, timeout)
                for processor in self.__processors:
                    stat_result.apply_processor(processor)
            except Exception, exc:
                logger.error("Error collecting stat: %s" % (str(exc),))
                logger.error(traceback.format_exc())
                continue
            if stat_key in self.__dependency_keys:
                dependency_results[stat_key] = stat_result
            self.__persister(stat_key, stat_result)

    def to_builder(self):
        """Constructs a builder initialized with the current StatChainRunner.
        This builder can then be modified in any way to construct a new
        StatChainRunner. This is very usefule when testing, for example, if
        you want to take a StatChainRunner, and stub out its persister, or
        remove or modify any stats to be run.
        """
        builder = StatChainRunner.StatChainRunnerBuilder()
        builder.set_persister(self.__persister)
        for stat_key, stat_deps_pair in self.__chain.iteritems():
            stat = stat_deps_pair[0]
            dependency_keys = stat_deps_pair[1]
            builder.add(stat_key, stat, dependency_keys)
        return builder

    def get_run_order(self):
        """Return the run order of the keys."""
        # Return a clone so the run order can't be modified
        return list(self.__run_order)

    def get_stats_dict(self):
        return self.to_builder().get_stats_dict()


class ProcStdoutStderrStat(Stat):
    """Collects the output of running 'tail' on the stdout and stderr
    logs in the run directories of processes managed by CM.

    Note that there is a separate implementation that takes the most
    recent processes and gathers more of their stderr/stdout using CM
    machinery proper. This supplements that implementation by capturing
    the logs of older processes, which have typically been purged from
    CM's database, and so are not picked up by the other implementation.
    The run directories of these older processes are kept for 60 days
    after the processes have been cleared from the DB, and often contain
    important debugging information. In particular, one-off processes are
    purged aggressively from the DB, but their logs can be essential to
    debugging a CM workflow failure.

    Accordingly, this stat excludes running processes (these are handled
    by the other mechanism), and prioritizes failed processes.
    """

    # Some process logs are more important than others when it comes to
    # debugging a support case. If we were to collect the logs in their
    # entirety, we might consume a lot of space in the support bundle;
    # the log files can run to several MB each. So we categorize the
    # processes as high, medium, or low priority, and tail more of the
    # higher-priority processes. The constants below give the max bytes
    # we tail for logs in each category, and the maximum number of
    # processes allowed in each category.
    __BIG_TAIL_BYTES = 2 * 1024 * 1024
    __BIG_TAIL_MAX = 4
    __BIG_TAIL_PROCS = set(["oozie-OOZIE-SERVER-upload-sharelib"])

    __MEDIUM_TAIL_BYTES = 20 * 1024
    __MEDIUM_TAIL_MAX = 20

    __SMALL_TAIL_BYTES = 2 * 1024
    __SMALL_TAIL_MAX = 80

    def collect(self, context, dependent_results, timeout):
        # Get list of process run dirs. Only accept directories whose names
        # match the pattern we know the agent uses for run dirs. This also
        # frees us from having to deal with spaces, newlines, and shell
        # metacharacters in directory names.
        parent_dir = os.path.dirname(os.getcwd())
        proc_dirs = []
        dirname_pattern = re.compile("^[a-zA-Z0-9_.-]+$")

        for file_name in os.listdir(parent_dir):
            abs_path = os.path.join(parent_dir, file_name)
            if os.path.isdir(abs_path) and dirname_pattern.match(file_name):
                proc_dirs.append(file_name)

        # Sort processs run dirs by modification time (most recent first).
        def mod_time(dir):
            return os.path.getmtime(os.path.join(parent_dir, dir))

        proc_dirs.sort(key=mod_time, reverse=True)

        # Get map of run dir name to supervisor state.
        supervisor_proc_info = self.__get_supervisor_proc_info()

        big_tails = []
        medium_tails = []
        small_tails = []

        # Partition the list of process dirs into three lists, corresponding
        # to high, medium, and low priority. The higher the priority, the
        # more log data we will collect. Enforce a maximum size for all three
        # lists, so that the size of the data collected is bounded.
        for proc_dir in proc_dirs:
            # Exclude running processes. These are covered by the main process
            # stdout/stderr log archiver.
            if self.__get_proc_state(supervisor_proc_info, proc_dir) == "RUNNING":
                continue

            # Failed processes have higher priority than other processes, and
            # failed Upload Oozie Sharelib processes have the highest priority
            # of all, since that is the top cause of CDH 5.x upgrade failure.
            if self.__proc_failed(proc_dir, parent_dir, supervisor_proc_info):
                tokens = proc_dir.split("-", 1)
                if len(tokens) == 2 and tokens[1] in ProcStdoutStderrStat.__BIG_TAIL_PROCS:
                    big_tails.append(proc_dir)
                else:
                    medium_tails.append(proc_dir)
            else:
                # The most recent deploy cc process for each service is collected
                # elsewhere in the support bundle, and older successful deploy
                # ccs aren't very interesting.
                if proc_dir.startswith("ccdeploy_"):
                    continue
                small_tails.append(proc_dir)

        # Impose size limits on the lists of process dirs. Prepend the overflow
        # of each list to the next-lower-priority list, if any.
        big_tails, medium_tails = self._handle_overflow(big_tails, medium_tails,
            ProcStdoutStderrStat.__BIG_TAIL_MAX)
        medium_tails, small_tails = self._handle_overflow(medium_tails, small_tails,
            ProcStdoutStderrStat.__MEDIUM_TAIL_MAX)
        small_tails, _ = self._handle_overflow(small_tails, None,
            ProcStdoutStderrStat.__SMALL_TAIL_MAX)

        # Build up to three tail commands, corresponding to the three lists.
        # The only difference between the three is the limit imposed on the
        # output of the tail command.
        tail_cmd = self.__append_tail_cmd("cd ..", big_tails,
            ProcStdoutStderrStat.__BIG_TAIL_BYTES)
        tail_cmd = self.__append_tail_cmd(tail_cmd, medium_tails,
            ProcStdoutStderrStat.__MEDIUM_TAIL_BYTES)
        tail_cmd = self.__append_tail_cmd(tail_cmd, small_tails,
            ProcStdoutStderrStat.__SMALL_TAIL_BYTES)

        return CommandStat(["bash", "-c", tail_cmd]).collect(context, dependent_results)

    def __get_supervisor_proc_info(self):
        """Returns a dict associating process run dir name to process information
        as reported by supervisor. If we are unable to get the process info from
        supervisor for any reason, returns an empty dict; we can still produce
        useful output in this case.
        """

        try:
            supervisor_opts = supervisor.options.ClientOptions()
            supervisor_opts.realize(args=["-c", "../../supervisor/supervisord.conf"])
            supervisor_client = supervisor_opts.getServerProxy()
            info_list = supervisor_client.supervisor.getAllProcessInfo()
        except:
            return {}

        info_dict = dict()
        for info in info_list:
            info_dict[info["name"]] = info

        return info_dict

    def __get_proc_state(self, supervisor_proc_info, proc_name):
        """Returns the supervisor process state name for a process, or None if
        not known."""

        info = supervisor_proc_info.get(proc_name)
        if info is not None:
            return info.get("statename")

        return None

    def __proc_failed(self, proc_dir, parent_dir, supervisor_proc_info):
        """Returns True if we know for certain that a given process failed,
        or exited with error status.
        """

        # Check the supervisor process state first, since that's cheap.
        proc_state = self.__get_proc_state(supervisor_proc_info, proc_dir)
        if proc_state == "FATAL":
            return True
        if proc_state == "RUNNING":
            return False

        # Look for a file called "exit_code" in the run dir, which (in CM 5.10+)
        # contains the exit status of one-off processes. If that's not found,
        # look for "__cloudera_metadata__", which contains a JSON document
        # giving the exit status of ccdeploy processes.
        proc_dir_path = os.path.join(parent_dir, proc_dir)
        if os.path.exists(os.path.join(proc_dir_path, "exit_code")):
            exit_code_str = self.__read_file(
                os.path.join(proc_dir_path, "exit_code")).strip()
            return len(exit_code_str) > 0 and exit_code_str != "0"
        elif os.path.exists(os.path.join(proc_dir_path, "__cloudera_metadata__")):
            json = self.__read_file(os.path.join(proc_dir_path, "__cloudera_metadata__"))
            try:
                metadata = simplejson.loads(json)
                return "exit_code" in metadata and metadata["exit_code"] != 0
            except:
                return False
        else:
            return False

    def __read_file(self, path):
        """Helper method that returns the contents of a file as a string.
        If an exception occurs, swallows it and returns an empty string.
        Thus, there is no way to tell if an empty string result means that
        the file was empty, or that an exception occurred. This is by intention.
        """

        # We can't use 'with' because we have to be compatible with Python 2.4.
        file = None
        try:
            file = open(path)
            return file.read()
        except:
            return ""
        finally:
            if file:
                file.close()

    def _handle_overflow(self, main_list, overflow_list, max_elem):
        """If main_list has more than max_elem elements, returns a tuple
        containing a list of the first max_elem elements of main_list, and
        a list formed by prepending the excess elements of max_list to
        overflow_list, or None if the latter is None. Otherwise, returns
        a tuple of (main_list, overflow_list).
        """
        if len(main_list) > max_elem:
            if overflow_list is not None:
                return main_list[:max_elem], main_list[max_elem:] + overflow_list
            else:
                return main_list[:max_elem], None
        else:
            return main_list, overflow_list

    def __append_tail_cmd(self, tail_cmd, proc_dirs, tail_bytes):
        """Constructs a bash command for tailing the stdout/stderr log files
        in proc_dirs, adds it to tail_cmd, and returns the resulting string.
        """
        new_cmd = ""
        if len(proc_dirs) > 0:
            if len(tail_cmd) > 0:
                new_cmd += "; "
            # The "tr" command is needed because the "-I" option to xargs
            # causes it not to recognize blanks as argument separators;
            # instead, arguments must be separated by newlines.
            new_cmd += ("echo '%s' | tr ' ' '\n' | xargs -I XXX "
                        "tail -c %d XXX/logs/stderr.log XXX/logs/stdout.log") % \
                       (" ".join(proc_dirs), tail_bytes)

        return tail_cmd + new_cmd


class DedupingLsofCommandStat(StreamingCommandStat):
    """On more recent distros (rhel 7+, ubuntu 14+, debian 7+), lsof shows
    the files open by each thread in the process. Since, in general, threads
    inherit the open files of the caller, this leads to there being many
    entries for a given file, differing only by the associated TID. This adds
    a lot of noise to the lsof output, making it harder to read. And the output
    can be as large as 250MB+, which bloats the support bundle, and can cause
    the stats collection process to OOM.

    This class reads the output of lsof one line at a time, de-duping lines
    that differ only by TID.
    """

    def __init__(self):
        super(DedupingLsofCommandStat, self).__init__(["lsof", "-n", "-P"])
        self.__pid_to_files = dict()

    def _process_stdout_line(self, line):

        # "lsof -n -P" output format on distros where thread reporting is supported (e.g., rhel7):
        # COMMAND     PID   TID         USER   FD      TYPE             DEVICE  SIZE/OFF       NODE NAME
        # systemd       1               root  cwd       DIR              202,1      4096        128 /

        # "lsof -n -P" output format on distros where thread reporting not is supported (e.g., rhel6):
        # COMMAND     PID         USER   FD      TYPE             DEVICE  SIZE/OFF       NODE NAME
        # init          1         root  cwd       DIR              202,1      4096          2 /

        fields = line.split()
        if fields[0] != "COMMAND" or fields[1] != "PID":
            pid = int(fields[1])
            filename = fields[-1]

            if pid in self.__pid_to_files:
                pid_filenames = self.__pid_to_files[pid]
                if filename in pid_filenames:
                    return ""
            else:
                pid_filenames = set()
                self.__pid_to_files[pid] = pid_filenames

            pid_filenames.add(filename)

        return line


# NOTE:  169.254.169.254 is the special link-local address that Amazon uses to
# pass data into an EC2 instance. Should not be bound to anything in any other
# normal domain.
EC2_PROBE_CMD = "curl -m 1 http://169.254.169.254/2011-01-01/meta-data/instance-type"

def create_sys_stat_runner(machine_collection_dao=None, processors=None, timeout=None):
    """Returns a system stats runner that collects the output of various commands"""
    def persist_sys_stat(stat_key, stat_result):
        if machine_collection_dao is None:
            return
        machine_collection_dao.put_sys_stat(
            stat_key, stat_result.get_result_data(),
            stat_result.get_error_data(), stat_result.get_other())
        machine_collection_dao.update_performance(stat_key)

    distro, version_number, version_name = platform.dist()
    major_version = int(version_number.split('.')[0])
    agent_bin = os.path.join(sys.prefix, "bin", "cm")

    stat_chain_runner_builder = StatChainRunner.get_builder(). \
        set_persister(persist_sys_stat). \
        set_timeout(timeout). \
        add("df",              CommandStat(["df", "-k"])). \
        add("df-i",            CommandStat(["df", "-i"])). \
        add("cpuinfo",         CommandStat(["cat", "/proc/cpuinfo"])). \
        add("meminfo",         CommandStat(["cat", "/proc/meminfo"])). \
        add("interrupts",      CommandStat(["cat", "/proc/interrupts"])). \
        add("mounts",          CommandStat(["cat", "/proc/mounts"])). \
        add("swaps",           CommandStat(["cat", "/proc/swaps"])). \
        add("diskstats",       CommandStat(["cat", "/proc/diskstats"])). \
        add("partitions",      CommandStat(["cat", "/proc/partitions"])). \
        add("procvmstat",      CommandStat(["cat", "/proc/vmstat"])). \
        add("entropy_avail",   CommandStat(["cat", "/proc/sys/kernel/random/entropy_avail"])). \
        add("mdstat",          CommandStat(["cat", "/proc/mdstat"])). \
        add("uptime_cmd",      CommandStat(["uptime"])). \
        add("fstab",           CommandStat(["cat", "/etc/fstab"])). \
        add("mtab",            CommandStat(["cat", "/etc/mtab"])). \
        add("security_limits", CommandStat(["cat", "/etc/security/limits.conf"])). \
        add("iostat",          CommandStat(["iostat", "4", "10"])). \
        add("ip",              CommandStat(["ip", "-s", "-s", "a"])). \
        add("ethtool",         CommandStat(["bash", "-c",
            """
            for n in $(ip --oneline l | cut -d ':' -f 2 | cut -d '@' -f 1); do
              ethtool $n
              echo ' '
            done
            """])). \
        add("ethtool-s",       CommandStat(["bash", "-c",
            """
            for n in $(ip --oneline l | cut -d ' ' -f 2 | cut -d '@' -f 1); do
              echo -n $n ' '
              ethtool -S $n
              echo ' '
            done
            """])). \
        add("lsmod",           CommandStat(["lsmod"])). \
        add("lspci",           CommandStat(["lspci"])). \
        add("lscpu",           CommandStat(["lscpu"])). \
        add("lshw",            CommandStat(["lshw"])). \
        add("lsb_release",     CommandStat(["lsb_release", "-a"])). \
        add("java_version",    CommandStat(["java", "-version"])). \
        add("uname-a",         CommandStat(["uname", "-a"])). \
        add("uname-r",         CommandStat(["uname", "-r"])). \
        add("issue",           CommandStat(["cat", "/etc/issue"])). \
        add("redhat-release",  CommandStat(["cat", "/etc/redhat-release"])). \
        add("oracle-release",  CommandStat(["cat", "/etc/oracle-release"])). \
        add("python-platform-dist",   CommandStat(["python", "-c", "import platform;print platform.dist()"])). \
        add("hosts",           CommandStat(["cat", "/etc/hosts"])). \
        add("resolv",          CommandStat(["cat", "/etc/resolv.conf"])). \
        add("nsswitch",        CommandStat(["cat", "/etc/nsswitch.conf"])). \
        add("sysctl",          CommandStat(["/sbin/sysctl", "-A"])). \
        add("service",         CommandStat(["service",  "--status-all"])). \
        add("dmidecode",       CommandStat(["dmidecode"])). \
        add("rpm-qa",          CommandStat(["rpm", "-qa"])). \
        add("dpkg-l",          CommandStat(["dpkg", "-l"])). \
        add("vmstat",          CommandStat(["vmstat"])). \
        add("ec2inst_type",    CommandStat(EC2_PROBE_CMD.split(" "))). \
        add("ps_aux",          CommandStat(["ps", "aux"])). \
        add("host_conf",       CommandStat(["cat", "/etc/host.conf"])). \
        add("yum_repos",       CommandStat(["ls", "/etc/yum.repos.d"])). \
        add("parcels",         CommandStat(["ls", "-la", "/opt/cloudera/parcels"])). \
        add("sqoop",           CommandStat(["ls", "-la", "/var/lib/sqoop"])). \
        add("apt_sources",     CommandStat(["cat", "/etc/apt/sources.list"])). \
        add("zypper_repos",    CommandStat(["zypper", "repos", "-d"])). \
        add("lvdisplay",       CommandStat(["lvdisplay"])). \
        add("vgdisplay",       CommandStat(["vgdisplay"])). \
        add("pvdisplay",       CommandStat(["pvdisplay"])). \
        add("dmsetup",         CommandStat(["dmsetup", "ls", "--tree"])). \
        add("lsblk",           CommandStat(["lsblk"])). \
        add("blkid",           CommandStat(["blkid"])). \
        add("dmesg",           CommandStat(["bash", "-c", "dmesg -T 2>/dev/null || dmesg"])). \
        add("selinux",         CommandStat(["cat", "/etc/selinux/config"])). \
        add("sestatus",        CommandStat(["sestatus"])). \
        add("suse-release",    CommandStat(["cat", "/etc/SuSE-release", "/etc/suse-release"])). \
        add("sar",             CommandStat(["sar", "-A"])). \
        add("swappiness",      CommandStat(["cat", "/proc/sys/vm/swappiness"])). \
        add("iptables",        CommandStat(["iptables", "-L", "-v", "-n"])). \
        add("top",             CommandStat(["top", "-b", "-n 1"])). \
        add("ifcfg-rhel",      CommandStat(["cat", "/etc/sysconfig/network-scripts/ifcfg-eth0"])). \
        add("ifcfg-sles",      CommandStat(["cat", "/etc/sysconfig/network/ifcfg-eth0"])). \
        add("network",         CommandStat(["cat", "/etc/sysconfig/network"])). \
        add("etc_hostname",    CommandStat(["cat", "/etc/hostname"])). \
        add("hostname",        CommandStat(["hostname", "--fqdn"])). \
        add("ss-a",            CommandStat(["ss", "-a"])). \
        add("nstat",           CommandStat(["nstat", "-asz"])) .\
        add("sys_kernel_mm",   CommandStat(["grep", "-r", ".", "/sys/kernel/mm"])). \
        add("sys_rotational",  CommandStat(["bash", "-c", "grep -r . /sys/block/*/queue/rotational"])). \
        add("limits_content",  CommandStat(["bash", "-c", "for x in /etc/security/limits.d/*; do echo \"file $x:\"; cat $x; echo; echo; done"])). \
        add("alternatives_content",  CommandStat(["bash", "-c",
           "PATH=/usr/bin:/usr/sbin:\"$PATH\"; " +
           "for x in /etc/alternatives/*; do " +
           "  echo $x; " +
           "  update-alternatives --display $(basename $x); " +
           "  echo -------------; " +
           "done"])). \
        add("messages",        CommandStat(["cat", "/var/log/messages"])). \
        add("kern_log",        CommandStat(["cat", "/var/log/kern.log"])). \
        add("date",            CommandStat(["date"])). \
        add("ntpstat",         CommandStat(["ntpstat"])). \
        add("ntpq-pn",         CommandStat(["ntpq", "-pn"])). \
        add("krb5",            CommandStat(["cat", "/etc/krb5.conf"])). \
        add("kdc",             CommandStat(["cat", "/var/kerberos/krb5kdc/kdc.conf"])). \
        add("kadm5",           CommandStat(["cat", "/var/kerberos/krb5kdc/kadm5.acl"])). \
        add("python-socket-fqdn", CommandStat(["python", "-c", "import socket; print socket.getfqdn(); print socket.gethostbyname(socket.getfqdn())"])). \
        add("nslookup-cm-hostname", CommandStat(["nslookup", "-query=any", os.environ.get("CM_HOST_NAME", "")])). \
        add("dig-cm-hostname", CommandStat(["dig", "any", os.environ.get("CM_HOST_NAME", "")])). \
        add("host-cm-hostname", CommandStat(["host", "-v", "-t", "A", os.environ.get("CM_HOST_NAME", "")])). \
        add("journalctl",      CommandStat(["journalctl", "--lines=2000"])). \
        add("process-stdout-stderrs", ProcStdoutStderrStat()). \
        add("du-process-logs",
            # List all existing process log dirs along with their sizes. This is
            # primarily useful in debugging the "process-stdout-stderrs" stat,
            # which has not always captured the logs we need. It can also help
            # us determine which files to ask for when requesting additional
            # information from a customer.
            CommandStat(["bash", "-c",
                 "cd ..; find -maxdepth 2 -mindepth 2 -type d -name logs "
                 "-printf '%T@ %p\n' | sort -k 1nr | awk '{ print $2 }' | xargs du -sh"])). \
        add("agent_cert",      CommandStat([agent_bin, "diag", "dump_cert"])). \
        add("lsof",            DedupingLsofCommandStat())

    # Collect numa information in the diagnostics bundle (OPSAPS-33570)
    if os.path.exists("/var/run/numad.pid"):
        stat_chain_runner_builder.add("numad_log", CommandStat(["cat", "/var/log/numad.log"]))

    stat_chain_runner_builder.add("numactl--harware", CommandStat(["numactl", "--hardware"]))

    if distro == 'Ubuntu':
        stat_chain_runner_builder.add("initctl", CommandStat(["initctl", "list"]))
    else:
        # Run systemctl instead of chkconfig for redhat family 7+ (OPSAPS-33214)
        if distro in set(['redhat', 'centos', 'oracle']) and major_version >= 7:
            stat_chain_runner_builder.add("systemctl", CommandStat(["systemctl", "list-unit-files"]))
        else:
            stat_chain_runner_builder.add("chkconfig", CommandStat(["chkconfig", "--list"]))

    if distro in set(['redhat', 'centos', 'oracle']) and major_version >= 7:
        # Capture firewalld information for redhat family 7+ (OPSAPS-33217)
        stat_chain_runner_builder.add("systemctl_firewalld", CommandStat(["systemctl", "-l status firewalld.service"]))
        stat_chain_runner_builder.add("firewall-cmd", CommandStat(["firewall-cmd", "--list-all-zones"]))

    if processors:
        stat_chain_runner_builder.set_processors(processors)

    return stat_chain_runner_builder.build()
