# (c) Copyright 2010 Cloudera, Inc.

import ConfigParser
import getopt
import sys
import os
import StringIO

class Config(object):
    """Config class that encapsulates a configuration for a component.
    It will read a standard python configuration file, and look at
    one particular section, which is initialized in the constructor."""

    def __init__(self, name, main_path):
        """Initialize the Config object by discovering the
        configuration files, which include a default, and
        a site configuraiton if the site configuration exists"""

        self.__name = name
        self.__config = ConfigParser.ConfigParser()
        
        # initialize config with the default configuration file by
        # searching the main_path for a file [name]-default.cfg
        default_conf_file = self.__find_file_on_path(self.__name + "-default.cfg", 
                                                     main_path)
        if default_conf_file is not None:
            self.__config.readfp(open(default_conf_file))

        # initialize config with the site configuration file by
        # searching the syspath for a file [name]-site.cfg, and adding
        # it to the config if it exists
        site_conf_file = self.__find_file_on_syspath(self.__name + "-site.cfg")
        if site_conf_file is not None:
            self.__config.read([site_conf_file])

    def __find_file_on_syspath(self, file_name):
        """Search the sys.path directories for a file with name file_name,
        and return the fully qualified file, else return None"""
        fully_qualified_file=None
        for path in sys.path:
            test_fully_qualified_file = self.__find_file_on_path(file_name, path)
            if test_fully_qualified_file is not None:
                fully_qualified_file=test_fully_qualified_file
                break
        return fully_qualified_file

    def __find_file_on_path(self, file_name, path):
        """Search the path directory for a file with file_name, and return the
        fully qualified file, else return None"""
        fully_qualified_file = None
        test_fully_qualified_file = os.path.join(path, file_name)
        if os.path.isfile(test_fully_qualified_file):
            fully_qualified_file = test_fully_qualified_file
        return fully_qualified_file
                    
    def get(self, prop, section=None):
        """Get a prop property from the config. Returns None if the property
        does not exist."""
        ret = None
        if section is None:
            section = self.__name
        if self.__config.has_option(section, prop):
            ret = self.__config.get(section, prop)
        return ret

    def get_name(self):
        """Get the name of the config."""
        return self.__name

    def get_config_as_file_handle(self):
        """Return an open file handle to the merged site and default 
        configuration files"""
        file_handle = StringIO.StringIO()
        self.__config.write(file_handle)
        return StringIO.StringIO(file_handle.getvalue())

configs = {}

def initialize_config(name, main_file):
    """Initialize a config object for a specific configuration name. This will
    look for a configuration file called [name]-default.xml, and
    [name]-site.xml, and work within the [name] section within that
    configuration file. Returns the Config instance initialized."""
    if name in configs:
        raise ValueError, "Config with name %s already initialized" % (name)
    configs[name] = Config(name, os.path.dirname(main_file))
    return configs[name]

def get_config(name):
    """Get an already initialized Config instance."""
    if name not in configs:
        raise ValueError, "No Config exists with name %s" % (name)
    return configs[name]
