# (c) Copyright 2010 Cloudera, Inc.

"""This module contains ssh command utilities, which include the capability to scp files
to and from remote hosts, and execute commands on remote hosts"""

import os
import subprocess
import re
import logging
import signal
import threading

logger = logging.getLogger()

class SshCommands(object):

    class SshCommandResult(object):
        def __init__(self, retcode, stdout, stderr):
            self.__retcode = retcode
            self.__stdout = stdout
            self.__stderr = stderr

        def get_retcode(self):
            return self.__retcode

        def get_success(self):
            if self.__retcode == 0:
                return True
            else:
                return False

        def get_stdout(self):
            return self.__stdout

        def get_stderr(self):
            return self.__stderr

        def __str__(self):
            return "RETCODE: %s\nSTDOUT:\n%s\nSTDERR:\n%s\n" % (
                str(self.__retcode), self.__stdout, self.__stderr)

    def __init__(self, user, host, ssh_connect_timeout=3, ssh_agent_env_file=None):
        """Constructor which takes three arguments:
        user: the user to ssh as
        host: the host to ssh to
        ssh_agent_env_file: a file containing the output of a run of ssh-agent. This
        file contains environment variables which we read that allow the ssh client to
        use the agent for login credentials.
        """
        self.__user = user
        self.__host = host
        self.__ssh_connect_timeout = ssh_connect_timeout
        self.__process_env = os.environ.copy()

        if ssh_agent_env_file is not None:
            handle = open(ssh_agent_env_file, "r")
            try:
                for line in handle:
                    m = re.match("SSH_AUTH_SOCK=(?P<ssh_auth_sock>[^;]+);", line)
                    if m is not None:
                        self.__process_env["SSH_AUTH_SOCK"] = m.group("ssh_auth_sock")
                        logger.info("SSH_AUTH_SOCK=%s", m.group("ssh_auth_sock"))
                    m = re.match("SSH_AGENT_PID=(?P<ssh_agent_pid>[^;]+);", line)
                    if m is not None:
                        self.__process_env["SSH_AGENT_PID"] = m.group("ssh_agent_pid")
                        logger.info("SSH_AGENT_PID=%s", m.group("ssh_agent_pid"))
            finally:
                handle.close()

    def __ssh_command_result_from_proc(self, proc):
        stdout, stderr = proc.communicate()
        ret_code = proc.wait()
        logger.info("Process %d returned code: %d" % (proc.pid, ret_code))
        return SshCommands.SshCommandResult(ret_code, stdout, stderr)

    def __open_proc(self, cmd):
        return subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, 
                                stderr=subprocess.PIPE, env=self.__process_env)

    def __kill_proc(self, proc_id):
        logger.info("Killing process %d with SIGTERM." % (proc_id,))
        try:
            os.kill(proc_id, signal.SIGTERM)
        except OSError, exc:
            logger.warn("Killing process %d failed with %s" % (proc_id, str(exc)))
        
    def __execute_command(self, cmd, timeout):
        proc = self.__open_proc(cmd)
        logger.info("Running command %s with process id: %d" % (cmd, proc.pid))
        if timeout is not None:
            timer = threading.Timer(timeout, self.__kill_proc, [proc.pid])
            timer.start()
            result = self.__ssh_command_result_from_proc(proc)
            timer.cancel()
            return result
        else:
            return self.__ssh_command_result_from_proc(proc)

    def __get_ssh_options(self):
        return ["-o", "ConnectTimeout=%d" % (self.__ssh_connect_timeout,), "-o",
                "UserKnownHostsFile=/dev/null", "-o", "StrictHostKeyChecking=no"]

    def get_host(self):
        return self.__host

    def get_user(self):
        return self.__user

    def get_ssh_connect_timeout(self):
        return self.__ssh_connect_timeout

    def scp_files(self, directory, files, extra_options=None, timeout=None):
        """scp files to user@host:directory. On success, return True. Otherwise,
        return false."""
        if extra_options is None:
            extra_options = []
        
        # Double escape spaces in remote paths, single escape local paths.
        directory = self.escape_spaces(directory, double_escape=True)
        files = [self.escape_spaces(filename) for filename in files]

        cmd_list = ["scp"] + self.__get_ssh_options() + extra_options
        cmd_list.extend(files)
        cmd_list.append("%s@%s:\"%s\"" % (self.__user, self.__host, directory))
        return self.__execute_command(" ".join(cmd_list), timeout).get_success()

    def scp_file_from_host(self, path, output_path, extra_options=None, timeout=None):
        """scp file at location path from user@host to the local output_path.
        On success, return True, otherwise, return False."""
        if extra_options is None:
            extra_options = []

        # Double escape spaces in remote paths, single escape local paths.
        path = self.escape_spaces(path, double_escape=True)
        output_path = self.escape_spaces(output_path)

        cmd_list = ["scp"] + self.__get_ssh_options() + extra_options + \
            ["%s@%s:\"%s\"" % (self.__user, self.__host, path), output_path]
        return self.__execute_command(" ".join(cmd_list), timeout).get_success()

    def execute_remote_command(self, command, extra_options=None, timeout=None):
        """Execute a remote command on user@host. Return an SshCommandResult object, which
        can be used to inspect stdout and return code."""
        if extra_options is None:
            extra_options = []
        cmd_list = ["ssh"] + self.__get_ssh_options() + extra_options + \
            ["%s@%s" % (self.__user, self.__host), command]
        return self.__execute_command(" ".join(cmd_list), timeout)

    def escape_spaces(self, path, double_escape=False):
        """Escapes spaces in a path, optionally double-escaping (escaping the 
        '\' as well). In cases where an escaped space needs to be sent to a
        remote machine, we need to double-escape. A single escape would escape
        the space locally, and the remote machine would see a space, so we need
        to escape the escape character as well."""
        escaped_space = "\ "
        if double_escape:
            escaped_space = "\\ "
        
        return path.replace(" ", escaped_space)
