# (c) Copyright 2011-2012 Cloudera, Inc.

import time
from file_line_reader import FileLineReader

class LogFileEventReader(object):
    """Reads log events out of a file. Log events can span multiple lines, so
    some of the functionality of this class deals with reading in multiple lines
    from a file while constructing each log event.

    The public interface supports getting the first and last events from a file,
    iterating over events with the get_next_event and get_prev_event functions,
    as well as doing a binary search on the file for log events based on their
    datetime."""

    def __init__(self, event_parser, file_handle=None, file_path=None, start_offset=0,
                 max_newline_seek_distance=0, max_event_seek_distance=0, deadline_ms=None):
        """Initialize the event reader with an event parser, an optional file handle, path,
        and offset. If file_handle is not passed in, one must call initialze before
        attempting to read log events.

        max_newline_seek_distance is the max distance we will look for newlines. If set to 0,
        the default, it will have no limit.

        max_event_seek_distance is the max number of lines to look for a new log event. If
        set to 0, the default will have no limit.

        deadline_ms is the max time-limit (in millis since epoch) for performing
        any further execution in event_reader.
        functions may throw an exception if the deadline_ms is crossed

        In the worst case, The memory usage of this class should at most be around:
        max_newline_seek_distance * max_event_seek_distance."""
        if file_handle is None:
            self.__line_reader = None
        else:
            self.__line_reader = FileLineReader(file_handle, file_path, start_offset,
                                                max_newline_seek_distance)

        self.__max_newline_seek_distance = max_newline_seek_distance
        self.__max_event_seek_distance = max_event_seek_distance
        self.__event_parser = event_parser
        self.__lookahead_event = None
        self.__deadline_ms = deadline_ms

    def initialize(self, file_handle, file_path, start_offset=0):
        """Initialize a new file handle, path, and offset to start reading from."""
        self.__line_reader = FileLineReader(file_handle, file_path, start_offset,
                                            self.__max_newline_seek_distance)
        self.__lookahead_event = None

    def get_next_event(self):
        """Get the next event in the currently owned file. One must either pass a file
        handle to the constructor or call initialize before calling this function.
        """
        if self.__lookahead_event is not None:
            event = self.__lookahead_event
            self.__lookahead_event = None
        else:
            event, _ = self.__find_next_event_start()
        if event is not None:
            self.__lookahead_event, prior_data = self.__find_next_event_start()
            event.append_to_message(prior_data)
        return event

    def get_first_event(self):
        """Get the first event in the currently owned file."""
        self.__line_reader.goto_beginning()
        self.__lookahead_event = None
        return self.get_next_event()

    def get_prev_event(self):
        """Get the previous event in the currently owned file from the current location.
        One must either pass a file handle to the constructor or call initialize before
        calling this function.
        """
        if self.__lookahead_event is not None:
            self.__lookahead_event = None
        event = self.__find_prev_event_start()
        if event is not None:
            _, prior_data = self.__find_next_event_start()
            event.append_to_message(prior_data)
            self.__find_prev_event_start()
        return event

    def get_last_event(self):
        """Get the last event in the currently owned file."""
        self.__line_reader.goto_end()
        self.__lookahead_event = None
        return self.get_prev_event()

    def find_nearest_event_datetime(self, dt):
        """Find the event nearest to the datetime, either equal to or greater than
        the passed in dt."""
        low = 0
        high = self.__line_reader.get_file_size()
        event = None
        while low < high:
            if self.__deadline_ms != None and time.time() > self.__deadline_ms:
                raise Exception("Timeout Error: Deadline crossed")
            mid = (low + high)/2
            event = self.find_event_containing_pos(mid)

            if event is None:
                # Ran off the beginning of the file. Get the first
                # event in the file.
                event = self.get_next_event()
                break

            if event.get_datetime() < dt:
                low = event.get_end_offset()+1
            else:
                high = min(event.get_offset()-1, high-1)

        while event is not None and event.get_datetime() < dt:
            if self.__deadline_ms != None and time.time() > self.__deadline_ms:
                raise Exception("Timeout Error: Deadline crossed")
            event = self.get_next_event()

        return event

    def find_event_containing_pos(self, pos):
        """Get the log event that contains the offset "pos" in the file."""
        self.__line_reader.goto_location(pos)
        event = self.get_prev_event()
        if event is not None and event.get_end_offset() > pos:
            self.__line_reader.goto_location(event.get_end_offset())
            return event
        # The pos might be early enough in the event that getting the previous
        # event goes back to the event before the one that contains pos. So
        # if fetching the previous event returns None, then get the next event.
        else:
            event = self.get_next_event()
            if event is not None:
                self.__line_reader.goto_location(min(event.get_end_offset(), self.__line_reader.get_file_size()))
            return event

    def __find_next_event_start(self):
        """Find the next event's start line, returning the event instance,
        as well as the data read prior to hitting that event. Return the
        event, data read pair."""
        prior_data = []
        event = None
        cnt = 0
        while True:
            # If we have searched more lines than the max number of lines we are supposed
            # to search for the next event, return (None, ""), and reset the buffer to the
            # end of the file. This is to protect us from large files that don't contain
            # log data from causing us to run out of memory looking for the next log event.
            if self.__deadline_ms != None and time.time() > self.__deadline_ms:
                raise Exception("Timeout Error: Deadline crossed")
            if self.__max_event_seek_distance != 0 and cnt > self.__max_event_seek_distance:
                prior_data = []
                self.__line_reader.goto_end()
                break
            line = self.__line_reader.get_next_line()
            if line is None:
                break
            event = self.__event_parser.parse_event(line)
            if event is not None:
                break
            else:
                prior_data.append(line.get_content())
                cnt += 1
        return event, ''.join(prior_data)

    def __find_prev_event_start(self):
        """Find the previous event's start line, returning the event instance,
        as well as the content read before finding the line. Return the event,
        data read pair."""
        event = None
        cnt = 0
        while True:
            # If we have searched more lines than the max number of lines we are supposed
            # to search for the prev event, return None, and reset the buffer to the
            # start of the file. This is to protect us from large files that don't contain
            # log data from causing us to run out of memory looking for the prev log event.
            if self.__deadline_ms != None and time.time() > self.__deadline_ms:
                raise Exception("Timeout Error: Deadline crossed")
            if self.__max_event_seek_distance != 0 and cnt > self.__max_event_seek_distance:
                self.__line_reader.goto_beginning()
                break
            line = self.__line_reader.get_prev_line()
            if line is None:
                break
            event = self.__event_parser.parse_event(line)
            if event is not None:
                break
            else:
                cnt += 1
        return event
