# (c) Copyright 2011 Cloudera, Inc.

import os
from file import LogFile

class StreamStrategy(object):
    """A StreamStrategy is used to implement the stream algorithm for a
    LogEventStreamer. This class should be extended, and the stream
    member function should be overridden. The stream function should
    return an Iterator or a Generator, which yields the events.

    Classes that inherit from this class should ALWAYS call this the
    StreamStrategy constructor explicitly: StreamStrategy.__init__(self)
    """

    def __init__(self):
        self.__total_events_read = 0
        self.__events_filtered_out = 0
        self.__files_read = []

    def get_total_events_read(self):
        """Get the total number of events read by this strategy."""
        return self.__total_events_read

    def get_events_filtered_out(self):
        """Get the total number of events filtered out of this strategy."""
        return self.__events_filtered_out

    def get_files_read(self):
        """Get the files that were read with this strtegy."""
        return self.__files_read

    def stream(self, event_reader, event_file, event_filter=None, ascending=True):
        """Override this method. It should return an Iterator or a Generator,
        which yields the events. The event_reader is the reader events will be
        read from. The event_file is the file that we will stream from.
        event_filter is a callabale that will return True if the event should
        be filtered. ascending is a boolean indicating whether the order of  
        events streamed should be in ascending or descending order."""
        raise NotImplementedError

    def _incr_total_events_read(self, amt):
        self.__total_events_read += amt

    def _incr_events_filtered_out(self, amt):
        self.__events_filtered_out += amt

    def _add_file_read(self, f):
        self.__files_read.append(f)


class DefaultStreamStrategy(StreamStrategy):
    """The default StreamStrategy. This strategy simply loops through, fetching
    all events returned by the event_reader until we hit the end_offset of the
    file we are fetching from.

    Before events are yielded, this strategy will see if the event should be
    filtered out, and if so, will continue reading from the event_reader.
    """

    def __init__(self):
        StreamStrategy.__init__(self)

    def stream(self, event_reader, log_file, event_filter=None, ascending=True):
        total_events_read_from_file = 0
        events_filtered_out_from_file = 0
        while True:
            if ascending:
                event = event_reader.get_next_event()
            else:
                event = event_reader.get_prev_event()

            if event is not None and event.get_offset() < log_file.get_end_offset():
                total_events_read_from_file += 1
                self._incr_total_events_read(1)
                if event_filter is not None:
                    if not event_filter(event):
                        events_filtered_out_from_file += 1
                        self._incr_events_filtered_out(1)
                        continue
                yield event
            else:
                self._add_file_read(
                    (log_file.get_path(),
                     log_file.get_start_offset(),
                     total_events_read_from_file,
                     events_filtered_out_from_file))
                break


class LogEventStreamer(object):
    """This class offers streaming of events from log files in a directory. It
    allows you to specify file and event filters to filter files in the directory,
    as well as events read out of those files."""

    def __init__(self, log_dir, event_reader, stream_strategy=None,
                 file_filter=None, event_filter=None):
        """Initialize the LogEventStreamer with a log directory, and an event reader.
        The optional file_filter and event_filter can be specified to support filtering
        of files and events. 
        
        NOTE: when an instance of this class is created, the list of files that 
        will be read is computed, so any new files added to the log_dir will not
        be picked up."""
        self.__log_dir = log_dir
        self.__event_reader = event_reader
        if stream_strategy is None:
            self.__stream_strategy = DefaultStreamStrategy()
        else:
            self.__stream_strategy = stream_strategy
        self.__file_filter = file_filter
        self.__event_filter = event_filter
        self.__files_in_log_dir = os.listdir(self.__log_dir)
        self.__filtered_file_list = self.__apply_file_filter()

    def get_log_dir(self):
        return self.__log_dir

    def get_total_events_read(self):
        return self.__stream_strategy.get_total_events_read()

    def get_events_filtered_out(self):
        return self.__stream_strategy.get_events_filtered_out()

    def get_files_in_log_dir(self):
        return self.__files_in_log_dir

    def get_files_read(self):
        return self.__stream_strategy.get_files_read()

    def stream(self, ascending=True):
        """Loop through each file, and stream log events from each of those files. This
        function is a generator."""
        file_list = self.__sort_file_list(self.__filtered_file_list, ascending)
        for f in file_list:
            handle = f.open_log_file()
            if ascending:
                offset = f.get_start_offset()
            else:
                offset = f.get_end_offset()

            self.__event_reader.initialize(handle, f.get_path(), offset)
            try:
                for event in self.__stream_strategy.stream(
                    self.__event_reader, f, self.__event_filter, ascending):
                    yield event
            except:
                # python 2.4 doesn't allow yield in a try finally block, so instead
                # we catch all exceptions, close the handle, and re-raise the exception.
                # in addition, we close the handle after the except clause in case no 
                # exception was raised.
                handle.close()
                raise
            handle.close()

    def __sort_file_list(self, file_list, ascending=True):
        """Sorts the file list by their first events' datetimes. Sort order is
        ascending by default."""
        first_event_datetimes = []
        files_without_events = []
        for f in file_list:
            handle = f.open_log_file()
            try:
                self.__event_reader.initialize(handle, f.get_path())
                first_event = self.__event_reader.get_first_event()
                if first_event is not None:
                    first_event_datetimes.append(first_event.get_datetime())
                else:
                    files_without_events.append(f)
            finally:
                handle.close()

        file_datetime_tuples = zip(file_list, first_event_datetimes)
        file_datetime_tuples = sorted(
            file_datetime_tuples, key=lambda x: x[1], reverse=not ascending)
        return [f for f, dt in file_datetime_tuples] + files_without_events

    def __apply_file_filter(self):
        """Apply the file filter, and return the filtered file list"""
        filter_context = {
            "file_list": [
                LogFile(os.path.join(self.__log_dir, f)) for f in self.__files_in_log_dir
                ]
            }
        if self.__file_filter is not None:
            self.__file_filter(filter_context)
        return filter_context["file_list"]
