# (c) Copyright 2011 Cloudera, Inc.

"""This module contains file filters that can be chained together to
filter out files in a context according to filter specific rules."""

import os
import re

class DirectoryFilter(object):
    """Filter out files in the context that are actually directories."""

    def __call__(self, context):
        context["file_list"] = [
            f for f in context["file_list"] if not os.path.isdir(f.get_path())
            ]
        return True


class FilePatternFilter(object):
    """Filter out files in the context that don't start with a regex. Construct
    using a regex object or a string which will then be compiled into a
    regular expression."""

    def __init__(self, reg_exp):
        if isinstance(reg_exp, basestring):
            self.__reg_exp = re.compile(reg_exp)
        else:
            self.__reg_exp = reg_exp

    def __call__(self, context):
        file_list = []
        for f in context["file_list"]:
            base_file_name = os.path.basename(f.get_path())
            if self.__reg_exp.match(base_file_name) is not None:
                file_list.append(f)
        context["file_list"] = file_list
        return True


class TimeRangeFilter(object):
    """Filter out log files that don't have log entries within the given
    time range (represented by the start_date_time and end_date_time passed
    to the constructor). It does this by peaking at the first and last log
    event in the file, and assuming that log events are in time increasing
    order."""

    def __init__(self, start_date_time, end_date_time, event_reader):
        self.__start_date_time = start_date_time
        self.__end_date_time = end_date_time
        self.__event_reader = event_reader

    def __call__(self, context):
        new_file_list = []
        for f in context["file_list"]:
            file_handle = f.open_log_file()
            try:
                self.__event_reader.initialize(file_handle, f.get_path())
                evt1 = self.__event_reader.get_first_event()
                if evt1 is not None and evt1.get_datetime() <= self.__end_date_time:
                    evt2 = self.__event_reader.get_last_event()
                    if evt2 is not None and evt2.get_datetime() >= self.__start_date_time:
                        self.__set_start_offset(f, evt1.get_datetime())
                        self.__set_end_offset(f, evt2.get_datetime())
                        new_file_list.append(f)
            finally:
                file_handle.close()

        context["file_list"] = new_file_list
        return True

    def __set_start_offset(self, f, first_event_datetime):
        if first_event_datetime < self.__start_date_time:
            event = self.__event_reader.find_nearest_event_datetime(self.__start_date_time)
            f.set_start_offset(event.get_offset())

    def __set_end_offset(self, f, last_event_datetime):
        if last_event_datetime > self.__end_date_time:
            event = self.__event_reader.find_nearest_event_datetime(self.__end_date_time)
            f.set_end_offset(event.get_offset())
