# (c) Copyright 2011 Cloudera, Inc.

class Line(object):
    """Line class, which encapsulates the line content, the offset within the file,
    as well as the file path the line was found in."""

    def __init__(self, content, file_path, offset):
        self.__content = content
        self.__file_path = file_path
        self.__offset = offset

    def get_content(self):
        return self.__content

    def get_file_path(self):
        return self.__file_path

    def get_offset(self):
        return self.__offset


class FileLineReader(object):
    """FileLineReader allows one to iterate over lines in a file, returning Line instances,
    which contain the data as well as the offset within the file that the line ocurred.
    FileLineReader supports iterating both forward and backwards over lines in the file,
    as well as jumping to the beginning or end of the file. It also supports jumping
    to a point in the middle of the file, and will itself from that point to the first
    character of the line that point is on.
    """

    # size of blocks to read from disk
    BLOCK_SIZE = 4096

    # number of bytes to read at a time when searching for a newline
    NEWLINE_READ_BUFFER = 256

    # enumeration of read directions
    FORWARD, REVERSE, NONE = range(3)

    def __init__(self, file_handle, file_path, start_offset=0, max_seek_distance=0):
        """Initialize a FileLineReader. This will reset the position in the file_handle
        to that of start_offset, which defaults to 0.

        If specifying a non default start_offset, one must make sure that the start_offset
        falls on the beginning of a line, as this constructor doesn't adjust the position
        to that point.

        max_seek_distance is the maximum distance to look for a newline. This ensures
        that we will not keep more than BLOCK_SIZE + max_seek_distance in memory dedicated
        to finding the newline. If set to 0 (the default), there is no limit.
        """
        self.__file_handle = file_handle
        self.__file_handle.seek(0, 2)
        self.__file_size = self.__file_handle.tell()
        self.__file_handle.seek(start_offset)
        self.__file_path = file_path
        self.__max_seek_distance = max_seek_distance

        self.__line_buffer = []
        self.__curr_pos = 0
        self.__last_block_size_read = 0
        self.__last_block_direction = FileLineReader.NONE
        self.__last_line_direction = FileLineReader.NONE

    def get_next_line(self):
        """Get the next line in the file"""
        # if the last read direction was reverse, we want to increment the current postion
        # so we don't re-read what was just read by the get_prev_line call.
        if self.__last_line_direction == FileLineReader.REVERSE:
            self.__curr_pos += 1

        # if we've run past the end of our current line buffer, read in another block
        # of data that we'll split into lines and set as our next line buffer.
        if self.__curr_pos >= len(self.__line_buffer):
            data, block_offset = self.__read_next_block()
            # if reading the next block failed, we are at the end of the file.
            # reset the line buffer and state associated with that, and return None.
            if data is None:
                self.__reset_buffer()
                return None
            self.__line_buffer = self.__get_lines_from_block(data, block_offset)
            self.__curr_pos = 0
        line = self.__line_buffer[self.__curr_pos]
        self.__curr_pos += 1
        self.__last_line_direction = FileLineReader.FORWARD
        return line

    def get_prev_line(self):
        """Get the previous line in the file from the current location"""
        # if the last read direction was forward, we want to decrement the current position
        # so we don't re-read what was just read by the get_next_line call.
        if self.__last_line_direction == FileLineReader.FORWARD:
            self.__curr_pos -= 1

        # if our current position in the line buffer is at the beginning of it, attempt to
        # read in another block of data that we'll split into lines and set as our next
        # line buffer.
        if self.__curr_pos == 0:
            data, block_offset = self.__read_prev_block()
            # if reading the next block failed, we are at the beginning of the file. reset
            # the line buffer and state associated with that, and return None
            if data is None:
                self.__reset_buffer()
                return None
            self.__line_buffer = self.__get_lines_from_block(data, block_offset)
            self.__curr_pos = len(self.__line_buffer)
        self.__curr_pos -= 1
        self.__last_line_direction = FileLineReader.REVERSE
        return self.__line_buffer[self.__curr_pos]

    def get_file_size(self):
        """Get the size of the file this line reader is currently reading from."""
        return self.__file_size

    def goto_location(self, location):
        """Go to location in the file. This call will adjust the location to align with a
        newline before the location (or beginning of the file if none is found before that.
        The functionality of this class would break if the current offset in the file didn't
        fall on a newline or at one of the ends of the file."""
        if location < 0 or location > self.__file_size:
            raise ValueError("Location is %d and file size is %s" % (location, self.__file_size))
        self.__file_handle.seek(location)
        self.__reset_buffer()
        self.__read_data_till_prev_newline()

    def goto_end(self):
        """Go to the end of the file"""
        self.__file_handle.seek(0, 2)
        self.__reset_buffer()

    def goto_beginning(self):
        """Go to the beginning of the file."""
        self.__file_handle.seek(0)
        self.__reset_buffer()

    def __reset_buffer(self):
        """Reset state related to the line buffer."""
        self.__line_buffer = []
        self.__curr_pos = 0
        self.__last_block_size_read = 0
        self.__last_block_direction = FileLineReader.NONE
        self.__last_line_direction = FileLineReader.NONE

    def __read_data_till_prev_newline(self):
        """Read backwards until a newline is found, or the beginning of the file,  and
        set the current position to one past that newline. Return the data read while
        searching for the newline. If we seek more than self.__max_seek_distance for
        a newline and don't find one, this will reset the file pointer to the beginning
        of the file, and return None."""
        data = ""
        seek_cnt = 0
        while True:
            if self.__max_seek_distance > 0 and seek_cnt >= self.__max_seek_distance:
                # we have reached the max_seek_distance allowed, reset the file pointer
                # to the beginning of the file and return None.
                self.goto_beginning()
                return None

            curr_file_pos = self.__file_handle.tell()
            if curr_file_pos == 0:
                break

            # don't seek more than max distance
            seek_len = FileLineReader.NEWLINE_READ_BUFFER
            if self.__max_seek_distance > 0 and seek_len > self.__max_seek_distance - seek_cnt:
                seek_len = self.__max_seek_distance - seek_cnt

            # don't seek more than the # of bytes to the beginning of the file
            if seek_len > curr_file_pos:
                seek_len = curr_file_pos

            self.__file_handle.seek(-seek_len, 1)
            datum = self.__file_handle.read(seek_len)

            newline_index = datum.rfind("\n")
            if newline_index == -1:
                data = datum + data
                seek_cnt += seek_len
                self.__file_handle.seek(-seek_len, 1)
            else:
                data = datum[newline_index+1:] + data
                # reposition at the byte immediately after the newline
                self.__file_handle.seek(-1 * (seek_len - newline_index) + 1, 1)
                break

        return data

    def __read_data_till_next_newline(self):
        """Read forward until the next newline is found, or the end of the file, and
        set the current position to one past that newline. Return the data read while
        searching for that newline. If we seek more than self.__max_seek_distance for
        a newline and don't find one, this will reset the file pointer to the end of
        the file, and return None."""
        if self.__max_seek_distance == 0:
            # readline() is fast, and there's no need to do the below if we do
            # not have to adhere to a max seek distance.
            line = self.__file_handle.readline()
            # jython gets into an inconsistent state after the readline in a way that
            # relative seeks work incorrectly.
            # We might be hitting http://bugs.jython.org/issue1793
            # Resetting the filepointer using absolute seek is a workaround
            self.__file_handle.seek(self.__file_handle.tell())
            return line

        data = ""
        seek_cnt = 0
        while True:
            if self.__max_seek_distance > 0 and seek_cnt >= self.__max_seek_distance:
                # we have reached the max_seek_distance allowed, reset the file pointer
                # to the end of the file and return None.
                self.goto_end()
                return None

            seek_len = FileLineReader.NEWLINE_READ_BUFFER
            if self.__max_seek_distance > 0 and seek_len > self.__max_seek_distance - seek_cnt:
                seek_len = self.__max_seek_distance - seek_cnt

            datum = self.__file_handle.read(seek_len)
            if not datum:
                break

            newline_index = datum.find("\n")
            if newline_index == -1:
                data += datum
                seek_cnt += seek_len
            else:
                data += datum[:newline_index + 1]
                # reposition at the byte immediately after the newline
                self.__file_handle.seek(-1 * (seek_len - newline_index) + 1, 1)
                break

        return data

    def __read_next_block(self):
        """Read the next block of data, returning a data, block offset pair."""
        # If we last read a block in the reverse direction, the position in the file
        # is on the wrong side of the block, so adjust the position forward to the
        # end of the block.
        if self.__last_block_direction == FileLineReader.REVERSE:
            self.__file_handle.seek(self.__last_block_size_read, 1)

        # If we are at the end of the file, reset line buffer state, and return None, None
        if self.__file_handle.tell() >= self.__file_size:
            self.__reset_buffer()
            return None, None
        else:
            # record the block position, and read the block size, as well as any additional
            # data until a newline is reached. Since lines are our records, blocks should
            # align with our records.
            block_offset = self.__file_handle.tell()
            data = self.__file_handle.read(FileLineReader.BLOCK_SIZE)
            if not data.endswith("\n"):
                data_till_newline = self.__read_data_till_next_newline()
                if data_till_newline is None:
                    # if self.__read_data_till_next_newline returned None, that means
                    # we weren't able to find a newline within the max_seek_distance,
                    # so the file pointer was reset to the end of the file. Return
                    # None for both data and block_offset.
                    return None, None
                else:
                    data += data_till_newline

            self.__last_block_size_read = len(data)
            self.__last_block_direction = FileLineReader.FORWARD
            return data, block_offset

    def __read_prev_block(self):
        """Read the previous block of data, returning a data, block offset pair."""
        # If we last read a block in the forward direction, the position in the file is
        # on the wrong side of the block, so adjust the position backwards to the
        # beginning of the block.
        if self.__last_block_direction == FileLineReader.FORWARD:
            self.__file_handle.seek(-self.__last_block_size_read, 1)

        # If we are at the beginning of the file, reset the line buffer state, and return None
        if self.__file_handle.tell() == 0:
            self.__reset_buffer()
            return None, None
        else:
            # seek to the start of the block to read, and read the block. Then seek
            # back to teh start, and read data backwards until a newline is found.
            # Prepend this data to the block. This is because blocks shoudl align with
            # the record boundaries.
            bytes_to_read = min(self.__file_handle.tell(), FileLineReader.BLOCK_SIZE)
            self.__file_handle.seek(-bytes_to_read, 1)
            block_offset = self.__file_handle.tell()
            data = self.__file_handle.read(bytes_to_read)
            self.__file_handle.seek(-bytes_to_read, 1)

            data_till_newline = self.__read_data_till_prev_newline()
            if data_till_newline is None:
                # if read_data_till_prev_newline returned None, that means we weren't
                # able to find a newline within the max_seek_distance, so the file
                # pointer was reset to the beginning of the file. return None for both
                # data and block_offset.
                data = None
                block_offset = None
            else:
                data = data_till_newline + data
                block_offset -= len(data_till_newline)
                self.__last_block_size_read = len(data)
                self.__last_block_direction = FileLineReader.REVERSE
            return data, block_offset

    def __get_lines_from_block(self, data, block_offset):
        """Given a block a data, split it into a list of Line instances"""
        line_content_list = data.splitlines(True)
        offset_in_block = 0

        lines = []
        for line_content in line_content_list:
            lines.append(Line(line_content, self.__file_path, block_offset + offset_in_block))
            offset_in_block += len(line_content)

        return lines
