# Copyright (c) 2011-2012 Cloudera, Inc. All rights reserved.
import os
import time
from threading import Event, Timer

from event_streamer import StreamStrategy
from file_line_reader import FileLineReader

"""
Same as default stream strat but for this difference
- does not parse lines which don't contain desired text (parsing is expensive)
"""
class GrepStreamStrategy(StreamStrategy):
  CHUNK_SIZE = 1024 * 1024 * 8

  def __init__(self, regex, deadline=None):
    """Pass in a compiled regex to initialize and a streaming deadline in
    seconds since the epoch."""
    StreamStrategy.__init__(self)
    self.__regex = regex
    self.__deadline = deadline

    self.__handle = None
    self.__timeout_event = None
    self.__timer = None

  def stream(self, event_reader, event_file, event_filter=None, ascending=True):
    total_events_read_from_file = 0
    events_filtered_out_from_file = 0

    self.__handle = open(event_file.get_path())
    start_offset = event_file.get_start_offset()
    end_offset = event_file.get_end_offset()


    try:
      self.__line_reader = FileLineReader(self.__handle,
                                          event_file.get_path())
      if ascending:
        curr_offset = start_offset
      else:
        curr_offset = end_offset
        

      if self.__deadline:
        if time.time() > self.__deadline:
          raise TimeoutException("Stream strategy timed out!")
        self.__timeout_event = Event()
        self.__timer = Timer(self.__deadline - time.time(), self.__timeout_event.set)
        self.__timer.start()

      read_size = self.CHUNK_SIZE
      last_event_offset = -1

      while True:
        if ascending:
          # decide how much to read
          if curr_offset + self.CHUNK_SIZE > end_offset:
            read_size = end_offset - curr_offset

          # read chunk and update chunk pointer
          self.__handle.seek(curr_offset)
          chunk = self.__handle.read(read_size)
          current_chunk_start = curr_offset
         
          # only process the chunk up to the last line break or until the end offset
          if curr_offset + read_size == end_offset:
            last_line_index = read_size
          else:
            last_line_index = chunk.rfind('\n')
          
          curr_offset = curr_offset + last_line_index
          
          # find all matches
          match_itr = self.__regex.finditer(chunk, 0, last_line_index)

        else:
          # decide how much to read
          if curr_offset - self.CHUNK_SIZE < start_offset:
            read_size = curr_offset - start_offset

          
          # read chunk and update chunk pointer
          current_chunk_start = curr_offset - read_size 
          self.__handle.seek(current_chunk_start)
          chunk = self.__handle.read(read_size)
          
          # process the chunk starting from the first line terminator
          if curr_offset - read_size == start_offset:
            first_line_index = 0
          else:
            first_line_index = chunk.find('\n')
 
          # get a reverse list of all the matches 
          match_itr = reversed(list(self.__regex.finditer(chunk, first_line_index, curr_offset)))
          
          curr_offset = current_chunk_start + first_line_index 
          
        if chunk == "":
          break 
        
        for match in match_itr:
          # find the begining of the line where the match is found
          match_line_start = chunk.rfind('\n', 0, match.start())
          match_line_start += 1

          event = event_reader.find_event_containing_pos(current_chunk_start + match_line_start)
          if event is not None and event.get_offset() != last_event_offset:
            # We found the event, log some stats
            total_events_read_from_file += 1
            self._incr_total_events_read(1)
            # Now, if we have a filter, and this event doesn't
            # pass the filter, don't yield it.
            if event_filter is not None and not event_filter(event):
              events_filtered_out_from_file += 1
              self._incr_events_filtered_out(1)
              continue
            last_event_offset = event.get_offset()
            yield event
        
        if self.__timeout_event is not None and self.__timeout_event.isSet():
          raise TimeoutException("Stream strategy timed out!")

        if ascending and curr_offset == end_offset:
          break
        elif not ascending and current_chunk_start == start_offset:
          break

      # Record stats about the file read.
      self._add_file_read(
        (event_file.get_path(),
         event_file.get_start_offset(),
         total_events_read_from_file,
         events_filtered_out_from_file))
    except:
      # python 2.4 doesn't allow yield in a try finally block, so instead
      # we catch all exceptions, and execute our finally clause. in addition,
      # we execute our finally clause after the except clause in case no
      # exception was raised.
      self.__stream_finally()
      raise
    self.__stream_finally()

  def __stream_finally(self):
    """The finally clause that executes in stream()."""
    self.__handle.close()
    if self.__timer is not None:
      self.__timer.cancel()

class TimeoutException(Exception):
  """Thrown when we exceed the alloted time for streaming"""
  def __init__(self, value):
    self.value = value

  def __str__(self):
    return repr(self.value)
