# (c) Copyright 2011 Cloudera, Inc.

""""A program which, given a log directory and a time range, scans the
log files for log events, and prints them to standard out. It does this
efficiently by not scanning files whose log events don't fall in the
specified time range, and doing a binary search into log files to the
log events at the time you are interested in."""

import sys
import optparse
import time
from datetime import datetime

from clusterstats.log.streaming.event_streamer import LogEventStreamer
from clusterstats.log.streaming.event_reader import LogFileEventReader
from clusterstats.log.streaming.event_parser import Log4JEventParser
from clusterstats.log.streaming.file_filter import *
from clusterstats.log.streaming.event_filter import *
import clusterstats.common.chain as chain

valid_components = ["namenode", "datanode", "jobtracker", "tasktracker"]
valid_log_levels = ["TRACE", "DEBUG", "INFO", "WARN", "ERROR", "FATAL"]

def get_datetimes(parser, options):
    """Get the timerange datetimes from the options parser."""
    if options.start_datetime is None or options.end_datetime is None:
        sys.stderr.write("ERROR: You must provide a start and an end datetime.\n\n")
        parser.print_help(sys.stderr)
        sys.exit(1)
    try:
        format = "%Y-%m-%d_%H:%M:%S"
        start_datetime = datetime(*(time.strptime(options.start_datetime, format)[0:6]))
        end_datetime = datetime(*(time.strptime(options.end_datetime, format)[0:6]))
    except ValueError:
        error_msg = "ERROR: datetimes must be of the format YYYY-mm-dd_hh:mm:ss.\n\n"
        sys.stderr.write(error_msg)
        parser.print_help(sys.stderr)
        sys.exit(1)
    return start_datetime, end_datetime

def get_components(parser, options):
    """Get the hadoop components from the options parser."""
    components = options.components.split(",")
    components = [component.strip() for component in components]
    for component in components:
        try:
            valid_components.index(component)
        except ValueError:
            error_msg = "ERROR: components must be a comma separated list of any or all " + \
                "of the following: namenode, datanode, jobtracker, tasktracker.\n\n"
            sys.stderr.write(error_msg)
            parser.print_help(sys.stderr)
            sys.exit(1)
    return components

def get_log_level(parser, options):
    """Get the log level from the options parser."""
    log_level = options.log_level
    try:
        valid_log_levels.index(log_level)
    except ValueError:
        sys.stderr.write("ERROR: Log level must be one of the following: TRACE, DEBUG, " + \
                             "INFO, WARN, ERROR, FATAL.\n\n")
        parser.print_help(sys.stderr)
        sys.exit(1)
    return log_level

def handle_args():
    """Handle the command line arguments"""

    description = "Given a log directory and a time range, scans the log files for log " + \
        "events, and prints them to standard out. It does this efficiently by not " + \
        "scanning files whose log events don't fall in the specified time range, " + \
        "and doing a binary search into log files to the log events at the time " + \
        "you are interested in."

    usage = "log_stream_cli.py " + \
        "--start-datetime= " + \
        "--end-datetime= " + \
        "[--log-dir=] " + \
        "[--log-level=] " + \
        "[--components=]"

    log_dir_help = "Location of the log directory which will be searched for logs. " + \
        "The default value is /var/log/hadoop."
    log_level_help = "The lowest log level you want to appear in the stream. " + \
        "The default value is WARN"
    start_datetime_help = "The lower end of the timerange you want to scan for logs in. " + \
        "The expected datetime format is YYYY-MM-DD_HH:MM:SS"
    end_datetime_help = "The upper end of the timerange you want to scan for logs in. " + \
        "The expected datetime format is YYYY-MM-DD_HH:MM:SS"
    components_help = "Comma separated list of components whose logfiles you want " + \
        "to scan. Valid options are namenode, datanode, jobtracker, and tasktracker. " + \
        "The default value is namenode,datanode,jobtracker,tasktracker."

    parser = optparse.OptionParser(description=description, usage=usage)
    parser.add_option("-d", "--log-dir", default="/var/log/hadoop", help=log_dir_help)
    parser.add_option("-l", "--log-level", default="WARN", help=log_level_help)
    parser.add_option("-s", "--start-datetime", help=start_datetime_help)
    parser.add_option("-e", "--end-datetime", help=end_datetime_help)
    parser.add_option("-c", "--components", default="namenode,datanode,jobtracker,tasktracker",
                      help=components_help)

    (options, _) = parser.parse_args()
    log_dir = options.log_dir
    start_datetime, end_datetime = get_datetimes(parser, options)
    components = get_components(parser, options)
    log_level = get_log_level(parser, options)

    return log_dir, log_level, start_datetime, end_datetime, components

def build_logfile_regex(components):
    """Build the logfile regex that will be passed to the filename filter."""
    return "^.*-(" + "|".join(components) + ")-.*\.log(\..*){0,1}$"

def main():
    """Main entry point to drive the log stream cli"""

    log_dir, log_level, start_datetime, end_datetime, components = handle_args()
    event_reader = LogFileEventReader(Log4JEventParser())

    # set up the file filters
    directory_filter = DirectoryFilter()
    file_pattern_filter = FilePatternFilter(build_logfile_regex(components))
    time_range_filter = TimeRangeFilter(start_datetime, end_datetime, event_reader)
    filters = chain.build_chain([directory_filter, file_pattern_filter, time_range_filter])

    # set up the event filter
    log_level_filter = LogLevelFilter(log_level)

    # initialize the streamer
    streamer = LogEventStreamer(log_dir, event_reader, filters, log_level_filter)

    current_file_path = None
    for event in streamer.stream():
        # new file, output the file name to stdout
        if current_file_path != event.get_file_path():
            print "#### %s ####\n\n" % (event.get_file_path(),)
            current_file_path = event.get_file_path()
        print str(event)

if __name__ == "__main__":
    main()
