# (c) Copyright 2011 Cloudera, Inc.

import unittest
import mox
import os
import time
import tempfile
import subprocess
from clusterstats.common.ssh_commands import SshCommands

class TestSshCommands(unittest.TestCase):

    def setUp(self):
        self.mocker = mox.Mox()
        self.command_runner = SshCommands("user", "host", 5)

    def tearDown(self):
        self.mocker.UnsetStubs()

    def test_scp_files(self):
        self.mocker.StubOutWithMock(self.command_runner, "_SshCommands__execute_command")
        self.command_runner._SshCommands__execute_command(
          'scp -o ConnectTimeout=5 -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no '
          'one two three user@host:"/tmp"', None).AndReturn(
          SshCommands.SshCommandResult(0, "output", ""))
        self.mocker.ReplayAll()
        self.assertTrue(self.command_runner.scp_files("/tmp", ["one", "two", "three"]))
        self.mocker.VerifyAll()

    def test_scp_file_from_host(self):
        self.mocker.StubOutWithMock(self.command_runner, "_SshCommands__execute_command")
        self.command_runner._SshCommands__execute_command(
          'scp -o ConnectTimeout=5 -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no '
          'user@host:"/tmp/foo.txt" /tmp/bar', None).AndReturn(SshCommands.SshCommandResult(0, "output", ""))
        self.mocker.ReplayAll()
        self.assertTrue(self.command_runner.scp_file_from_host("/tmp/foo.txt", "/tmp/bar"))
        self.mocker.VerifyAll()

    def test_execute_remote_command(self):
        self.mocker.StubOutWithMock(self.command_runner, "_SshCommands__execute_command")
        self.command_runner._SshCommands__execute_command(
                    "ssh -o ConnectTimeout=5 -o UserKnownHostsFile=/dev/null -o "
                    "StrictHostKeyChecking=no user@host ls -la /tmp", None
                    ).AndReturn(SshCommands.SshCommandResult(0, "output", ""))
        self.mocker.ReplayAll()
        self.assertTrue(self.command_runner.execute_remote_command("ls -la /tmp"))
        self.mocker.VerifyAll()

    def test_ssh_command_result(self):
        res = SshCommands.SshCommandResult(0, "hello", "")
        self.assertTrue(res.get_success())
        self.assertEquals(res.get_retcode(), 0)
        self.assertEquals(res.get_stdout(), "hello")

        res2 = SshCommands.SshCommandResult(1, "bye", "")
        self.assertFalse(res2.get_success())
        self.assertEquals(res2.get_retcode(), 1)
        self.assertEquals(res2.get_stdout(), "bye")

    def test_execute_command(self):
        tempdir = tempfile.mkdtemp()
        file_path = os.path.join(tempdir, "file")
        handle = open(file_path, "w")
        try:
            handle.write("content")
        finally:
            handle.close()
        res = self.command_runner._SshCommands__execute_command(["ls %s" % (tempdir)], None)
        self.assertTrue(res.get_success())
        self.assertEquals(res.get_retcode(), 0)
        self.assertEquals(res.get_stdout(), "file\n")

    def test_ssh_agent_path_construction(self):
        tempdir = tempfile.mkdtemp()
        ssh_agent_path = os.path.join(tempdir, ".ssh_agent")
        handle = open(ssh_agent_path, "w")
        try:
            handle.write("""SSH_AUTH_SOCK=/tmp/ssh-OwXjXt6368/agent.6368; export SSH_AUTH_SOCK;
SSH_AGENT_PID=6369; export SSH_AGENT_PID;
echo Agent pid 6369;""")
        finally:
            handle.close()
        command_runner = SshCommands("user", "host", 3, ssh_agent_path)
        self.assertEquals(
            command_runner._SshCommands__process_env["SSH_AUTH_SOCK"],
            "/tmp/ssh-OwXjXt6368/agent.6368")
        self.assertEquals(
            command_runner._SshCommands__process_env["SSH_AGENT_PID"], "6369")

    def test_ssh_timeout(self):
        self.mocker.StubOutWithMock(self.command_runner, "_SshCommands__open_proc")
        popen_mock = self.mocker.CreateMock(subprocess.Popen)

        # When the timeout occurs, it will attempt to kill the ssh process.
        # Launch a simple sleep process, and pass it's pid to be killed so
        # we also get a test of the __kill_proc function.
        sleep_proc = sleep_process = subprocess.Popen("sleep 5", shell=True)
        popen_mock.pid = sleep_proc.pid

        self.command_runner._SshCommands__open_proc("ssh -o ConnectTimeout=5 -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no user@host ls -la /tmp").AndReturn(popen_mock)
        popen_mock.communicate().AndReturn(("stdout", "stderr"))
        # wait on the popen instance will take 3 seconds because of this sleep
        def sleep_function():
            time.sleep(3)
        popen_mock.wait().WithSideEffects(sleep_function).AndReturn(-255)
        self.mocker.ReplayAll()
        # timeout in two seconds
        ret = self.command_runner.execute_remote_command("ls -la /tmp", timeout=2)
        self.assertFalse(ret.get_success())
        self.mocker.VerifyAll()
