#!/usr/bin/env python
"""
Copyright (c) 2011-2016 Cloudera, Inc. All rights reserved.

The CMF Agent starts a "status" web server, starts "supervisord",
and heartbeats regularly to the "master" server.
"""
from __future__ import absolute_import

# pylint: disable=too-many-lines,too-many-public-methods,too-many-statements
# pylint: disable=too-many-instance-attributes,import-error
# pylint: disable=line-too-long,bad-continuation,bare-except,broad-except,invalid-name,
# pylint: disable=missing-docstring,too-many-branches,too-many-lines

import asyncore
import atexit
import errno
import glob
import grp
import logging.handlers
import multiprocessing
import pickle
import platform
import pwd
import random
import resource
import shutil
import signal
import socket
import sys
import threading
import time

try:
  from hashlib import sha1 as _sha1
except ImportError:
  from sha import new as _sha1

try:
  import simplejson as json
except ImportError:
  import json

import avro.datafile
import avro.io
import avro.ipc
import avro.protocol
import os
import psutil
import stat
from daemonize import Daemonize
from guppy import hpy

import cmf.logging
from cmf.logging import setup_rolling_logging, get_logging_level
from cmf.process import Process, ReconstructedProcess

import cmf.audit.util
import cmf.config
import cmf.cgroups
import cmf.client_configs
import cmf.monitor.daemon
import cmf.downloader
import cmf.https
import cmf.kt_renewer
import cmf.log_tailer
import cmf.monitor
import cmf.parcel
import cmf.parcel_cache
import cmf.public_cloud
import cmf.status_server
import cmf.version

import cmf.heartbeat_tracker
import cmf.audit.navigator_thread
import cmf.audit.navigator_plugin

from cmf.monitor.host.filesystem_map import FilesystemMap
from cmf.redaction import REDACTOR
from cmf.safety_valve import SafetyValve
from cmf.supervisor_listener import SupervisordEventMonitor
from cmf.template_util import render_text
from cmf.throttling_logger import ThrottlingLogger
from cmf.tmpfs import mount_tmpfs
from cmf.util import (bytes_to_hex, hex_to_bytes, user_group_to_ids, write_conf_file)
from cmf.supervisor import SupervisorWrapper

from cmf.version_detector import VersionDetector, COMPONENT_VERSION_KEY, CDH_VERSION_KEY

from cmf.util.os_ops import OSOps
from cmf.util.avro import generate_avro_sanitizer

LOG = logging.getLogger(__name__)
THROTTLED_LOG = ThrottlingLogger(LOG, cmf.monitor.constants.DEFAULT_MONITOR_LOG_RATE)

DEFAULT_HEARTBEAT_INTERVAL_SECONDS = 5.0

# This will be used by xmlrpc.
DEFAULT_SOCKET_TIMEOUT = 45.0

# Number of heartbeats at which to compute heavier-duty
# heartbeats (like mountpoints).
RARE_HEARTBEAT_MULTIPLIER = 60

MASTER_AGENT_PROTOCOL = avro.protocol.parse(
  open(os.path.join(os.path.dirname(__file__), "master_agent.avpr")).read())
HEARTBEAT_REQUEST_SCHEMA = MASTER_AGENT_PROTOCOL.types_dict.get("HeartbeatRequest")
HEARTBEAT_RESPONSE_SCHEMA = MASTER_AGENT_PROTOCOL.types_dict.get("HeartbeatResponse")

RESPONSE_CACHE_FILE = "response.avro"
CM_GUID_FILE = "cm_guid"

STATUS_CODES_PROTOCOL = avro.protocol.parse(
  open(os.path.join(os.path.dirname(__file__), "status_codes.avpr")).read())
CM_STATUS_CODE_SCHEMA = STATUS_CODES_PROTOCOL.types_dict.get("CMStatusCode")
CM_STATUS_CODES_ENV_VAR = "CM_STATUS_CODES"

MAX_RECENT_SUPERVISOR_FAILURES = 5

# Number of seconds within which to log connection refused error at most once
CONN_REFUSED_THROTTLE_DURATION = 60

# Number of seconds to wait between supervisor connect attempts
SUPERVISOR_CONNECT_WAIT_TIME_SECS = 1

# Number of seconds to wait between version detection
VERSION_DETECTION_INTERVAL = 120

# Process name for the supervisord listener.
SUPERVISOR_LISTENER_PROC = "cmflistener"

# Process name for the flood daemon
FLOOD_DAEMON_PROC = "flood"

# Process name for the status server
STATUS_SERVER_PROC = "status_server"

# Default user/group for the agent
AGENT_USER = "cloudera-scm"
AGENT_GROUP = "cloudera-scm"

# User/group for flood-related FS entries
FLOOD_FS_USER = AGENT_USER
FLOOD_FS_GROUP = AGENT_GROUP

# The Avro schema specific unicode sanitizer
UNICODE_SANITIZER = generate_avro_sanitizer(HEARTBEAT_REQUEST_SCHEMA)

class SupervisorError(Exception):
  pass

# Be sure to update comments in config.ini when changing this. These can be
# overridden by ini file, environment variables, or command line. They are
# always overridden by parcels.
# DB drivers are handled separately
def make_default_env(cfg):
  return {
    # CM properties in the [Cloudera] section of ini
    'MGMT_HOME'           : os.path.join(cfg.base_dir, 'cm'),
    'CMF_SERVER_ROOT'     : os.path.join(cfg.base_dir, 'cm'),
    # CDH properties in the [Hadoop] section of ini
    'CDH_CRUNCH_HOME': '/usr/lib/crunch',
    'CDH_FLUME_HOME': '/usr/lib/flume-ng',
    'CDH_HADOOP_BIN': '/usr/bin/hadoop',
    'CDH_HADOOP_HOME': '/usr/lib/hadoop',
    'CDH_HBASE_HOME': '/usr/lib/hbase',
    'CDH_HBASE_INDEXER_HOME': '/usr/lib/hbase-solr',
    'CDH_HCAT_HOME': '/usr/lib/hive-hcatalog', # CDH5 location
    'CDH_HDFS_HOME': '/usr/lib/hadoop-hdfs',
    'CDH_HIVE_HOME': '/usr/lib/hive',
    'CDH_HTTPFS_HOME': '/usr/lib/hadoop-httpfs',
    'CDH_HUE_HOME': '/usr/lib/hue', # CDH5 location
    'CDH_HUE_PLUGINS_HOME': '/usr/lib/hadoop',
    'CDH_IMPALA_HOME': '/usr/lib/impala',
    'CDH_KUDU_HOME' : '/usr/lib/kudu',
    'CDH_LLAMA_HOME': '/usr/lib/llama/',
    'CDH_MR1_HOME': '/usr/lib/hadoop-0.20-mapreduce',
    'CDH_MR2_HOME': '/usr/lib/hadoop-mapreduce',
    'CDH_OOZIE_HOME': '/usr/lib/oozie',
    'CDH_PARQUET_HOME': '/usr/lib/parquet',
    'CDH_PIG_HOME': '/usr/lib/pig',
    'CDH_SENTRY_HOME': '/usr/lib/sentry',
    'CDH_SOLR_HOME': '/usr/lib/solr',
    'CDH_SPARK_HOME' : '/usr/lib/spark',
    'CDH_SQOOP_HOME': '/usr/lib/sqoop',
    'CDH_SQOOP2_HOME': '/usr/lib/sqoop2',
    'CDH_YARN_HOME': '/usr/lib/hadoop-yarn',
    'CDH_ZOOKEEPER_HOME': '/usr/lib/zookeeper',
    'HIVE_DEFAULT_XML': '/etc/hive/conf.dist/hive-default.xml',
    'WEBHCAT_DEFAULT_XML': '/etc/hive-webhcat/conf.dist/webhcat-default.xml', # see special handling in hive.sh
    'JSVC_HOME': '/usr/libexec/bigtop-utils',
    'TOMCAT_HOME': '/usr/lib/bigtop-tomcat',
    'CDH_KMS_HOME' : '/usr/lib/hadoop-kms',
    'KEYTRUSTEE_KP_HOME' : '/usr/share/keytrustee-keyprovider',
    'KEYTRUSTEE_SERVER_HOME' : '/usr/lib/keytrustee-server',
    'CDH_KAFKA_HOME' : '/usr/lib/kafka',
    'ORACLE_HOME' : '/usr/share/oracle/instantclient',
  }

def add_vars_in_config_ini_to_supervisee_environment(env, value_map, ini_group):
  """ Add variables in config.ini's ini_group to supervisee's environment.
      Override already existing variables.
  """
  if ini_group not in value_map:
    return
  for env_key, env_value in value_map[ini_group].iteritems():
    env_key = env_key.upper()
    action_taken = 'adding'
    if env_key in env:
      action_taken = 'overriding'
    LOG.info("config.ini %s env var to supervisee: %s: %s", action_taken, env_key, env_value)
    env[env_key] = env_value

def add_db_driver_jars(env):
  """ If missing, adds env vars for JDBC driver jars. Checks os env for
      driver variables, which don't get automatically picked up because they
      don't start with CMF_AGENT_
  """
  mysql_jar = env.get('CLOUDERA_MYSQL_CONNECTOR_JAR') or \
              os.getenv('CLOUDERA_MYSQL_CONNECTOR_JAR',
                        "/usr/share/java/mysql-connector-java.jar")
  oracle_jar = env.get('CLOUDERA_ORACLE_CONNECTOR_JAR') or \
               os.getenv('CLOUDERA_ORACLE_CONNECTOR_JAR',
                         "/usr/share/java/oracle-connector-java.jar")
  postgresql_jar = env.get('CLOUDERA_POSTGRESQL_JDBC_JAR') or \
                   os.getenv('CLOUDERA_POSTGRESQL_JDBC_JAR')
  if postgresql_jar and len(postgresql_jar) > 0:
    if not os.path.exists(postgresql_jar):
      # not a legal jar, guard against potentially buggy cmf-agent.default override
      LOG.warn("Ignoring override for CLOUDERA_POSTGRESQL_JDBC_JAR=%s "
               "because file does not exist. Using default.", postgresql_jar)
      postgresql_jar = None
  if not postgresql_jar:
    cmf_mgmt_home = env['MGMT_HOME']
    queryPath = os.path.join(cmf_mgmt_home, "lib/postgresql-*jre*.jar")
    postgresql_candidates = glob.glob(queryPath)
    if postgresql_candidates:
      postgresql_jar = postgresql_candidates[0]
    else:
      postgresql_jar = os.path.join(cmf_mgmt_home, "lib/postgresql-connector.jar")

  for jar in (mysql_jar, oracle_jar, postgresql_jar):
    if not os.path.exists(jar):
      LOG.info("Missing database jar: %s (normal, if you're not using this database type)", jar)
    else:
      LOG.info("Found database jar: %s", jar)

  env['CLOUDERA_MYSQL_CONNECTOR_JAR'] = mysql_jar
  env['CLOUDERA_ORACLE_CONNECTOR_JAR'] = oracle_jar
  env['CLOUDERA_POSTGRESQL_JDBC_JAR'] = postgresql_jar

def add_path_to_supervisee_environment(env):
  """
  Explicitly configures PATH of supervised processes to be
  the agent's path.  In OPSAPS-17930, we ran into the issue
  of a supervised processes' path (the tasktracker, in this case),
  being "PATH=/usr/lib64/cmf/agent/build/env/bin:/sbin:/usr/sbin:/bin:/usr/bin".
  /usr/lib64/cmf/agent/build/env/bin was added in /usr/lib64/cmf/agent/build/env/bin/supervisord,
  as part of "activate_this".  This snuck up on us because whether or not this
  happens depends on the version of virtualenv in play.

  When the agent is executed via $CMF_PATH/build/env/bin/cmf-agent, that directory enters PATH,
  and so must be removed.

  By adding the path here, supervised processes explicitly inherit
  the agent's path (which typically gets it from however it got started
  at boot time), and it also gets logged.
  """
  path = os.environ.get('PATH')
  if path:
    env['PATH'] = cmf.util.remove_agent_path(path)

def add_sudo_cmd_to_supervisee_environment(cfg, env):
  if cfg.sudo:
    env["SUDO"] = cfg.sudo

def add_cm_status_codes_to_env_dict(env):
  assert not env.has_key(CM_STATUS_CODES_ENV_VAR)
  env[CM_STATUS_CODES_ENV_VAR] = str.join(" ", CM_STATUS_CODE_SCHEMA.symbols)

def build_supervisee_environment(cfg):
  env = make_default_env(cfg)
  cfg_values = cfg.value_map()
  # Add the environment variables that start with 'CMF_AGENT_' to child process
  # environment. Then add variables from config.ini, overriding if already present.
  add_path_to_supervisee_environment(env)
  add_vars_in_config_ini_to_supervisee_environment(env, cfg_values, cfg.HADOOP)
  add_vars_in_config_ini_to_supervisee_environment(env, cfg_values, cfg.CLOUDERA)
  add_vars_in_config_ini_to_supervisee_environment(env, cfg_values, cfg.JDBC)
  add_cm_status_codes_to_env_dict(env)
  add_sudo_cmd_to_supervisee_environment(cfg, env)
  add_db_driver_jars(env)
  env['CMF_PACKAGE_DIR'] = cfg.package_dir
  return env

# Flood is enabled only on python >= 2.6
FLOOD_ENABLED = sys.version_info[0] >= 2 and sys.version_info[1] >= 6

FLOOD_LOG_FILE_NAME = "cloudera-flood.log"

class HeartbeatRequestMonitor(asyncore.file_dispatcher):
  """
  Dispatcher used to monitor the pipe notified by the status server when an
  explicit heartbeat request arrives.
  """

  def __init__(self, fd):
    self.pipe = fd
    self.heartbeat = False
    asyncore.file_dispatcher.__init__(self, fd)

  def writable(self):
    return False

  def handle_read(self):
    while True:
      try:
        os.read(self.pipe, 32)
      except OSError, ose:
        if ose.errno == errno.EINTR:
          continue
        if ose.errno == errno.EAGAIN:
          break
        raise
    self.heartbeat = True

  def is_hearbeat_requested(self):
    """
    Returns whether a heartbeat has been requested. Resets the internal state
    so that a subsequent call returns False (unless a new request arrives).
    """
    ret = self.heartbeat
    self.heartbeat = False
    return ret



class Agent(object):
  """
  The Agent checks in with the master to keep
  the host machine in line with the processes the server
  asks it to be running.

  start() is the main loop.
  """
  ACTIVE_PARCELS_FILE = "active_parcels.json"

  def __init__(self, cfg=None, exec_args=None):
    if exec_args is None:
      exec_args = []
    self.exec_args = exec_args
    if cfg is None:
      def default_config():
        # This occurs in various unit tests
        return cmf.config.AgentConfig()
      cfg = default_config()

    self.cfg = cfg

    self.cache = None
    self.cg_manager = None
    self.cloud_info = None
    self.event_mon = None
    self.fs_map = None
    self.hb_req_monitor = None
    self.requestor = None
    self.rm_enabled = None
    self.server_managed_parcels = None
    self.should_exit = None
    self.signal_caught = None
    self.supervisee_environment = None
    self.supervisor_client = None
    self.supports_dssd = None
    self.tailer = None
    self.url = None
    self.cm_guid = None
    self.token = None

    # Initialize State

    self.daemon = None

    # Identifier is used to confirm that the agent is talkign to the supervisor
    # that it itself launched.  We've seen instances where the supervisor
    # doesn't die after its parent agent dies, and this causes considerable
    # consternation.
    self.identifier = "agent-%d-%d" % (os.getpid(), long(time.time()))

    # Establish the Agent Protocol version from the schema
    self.agent_protocol_version = None
    for field in HEARTBEAT_REQUEST_SCHEMA.fields:
      if field.name == "version":
        self.agent_protocol_version = field.default
        break
    if self.agent_protocol_version is None:
      raise Exception("Unable to establish agent protocol version")

    self.shutdown_registry = []
    # Because the web server is multi-threaded, we need
    # to have some thread-safety.
    self.lock = threading.Lock()

    # Initialized by argument parsing
    self.dir = None

    # Initial Heartbeat Interval
    self.heartbeat_interval = DEFAULT_HEARTBEAT_INTERVAL_SECONDS

    self.is_late_init_done = False
    self.last_successful_heartbeat = None
    self.last_heartbeat_request = None
    self.last_heartbeat_response = None
    self.last_failed_heartbeat_request = None
    self.last_available_active_parcels = None

    self.parcel_dir_is_set = False
    self.parcel_dir_from_config = False

    self.started = False

    self.last_logged_connection_refused = 0

    # For counting the number of consecutive supervisor failures
    self.recent_supervisor_failures = 0

    # For rarely changing information that we don't want to compute every
    # heartbeat
    self.last_rare_heartbeat = 0
    self.cached_filesystem_info = None

    # For version detection; initialized to zero
    # to trigger detection on agent start.
    self.last_version_detection = 0

    # Used to determine when component lists need re-visiting
    self.previous_active_parcels = None
    # Initialize to [] so that it's valid as part of a heartbeat
    self.cached_component_info = []

    # A simple safety valve object that knows how to parse the safety valve
    # host configuration parameter.
    self.safety_valve = SafetyValve()
    self.monitor_daemon = None

    # Conditional so we don't blow up on OSX or whatever.
    self.dist = None
    if "dist" in dir(platform):
      dist = dict()
      dist["name"], dist["version"], dist["id"] = platform.dist()
      self.dist = dist

    self.repo = None
    # Stores any error that would occur during parcel repo initialization.
    self.parcel_init_error = None
    self.downloader = None
    self.inotify = None
    self.client_configs = None


    # The last time of refreshing process config directory
    self.last_config_directory_refresh_time = time.time()

    # Flag variable to indicate whether heap is to be dumped
    self.dump_heap_flag = False

    # Thread responsible for dumping the heap if the agent is aborted
    # Thread is None or not alive unless an active heap dump is happening.
    self.dump_heap_thread = None

    # The last recorded time for a process exit due to an OutOfMemory error.
    # The time is set by reading the contents of a file produced by killparent.sh.
    self.last_oom_exit_time = None

    # Store the number of physical and logical CPU cores. They are reported to
    # the CM server via heartbeat.
    self.num_logical_cpu_cores = self._get_logical_cpu_core_number()
    self.num_physical_cpu_cores = \
        self._get_physical_cpu_core_number(self.num_logical_cpu_cores)

    # Store DSSD support status
    self.cached_dssd_device_dir = None
    self._update_dssd_support_status()

    # Store the extra configs
    self.extra_configs = None

    # Store the custom JAVA_HOME config carried in the last response
    self.java_home_config = None

    # Supervisor Info
    self.supervisor_info = None
    self.supervisor_procs = None

    self.heartbeat_tracker = cmf.heartbeat_tracker.HeartbeatTracker()

    self.flood_opener = None
    self._process = None
    self.hb_callbacks = []


    uid, gid = os.getuid(), os.getgid()
    self.agent_user, self.agent_group = pwd.getpwuid(uid)[0], grp.getgrgid(gid)[0]
    if uid != 0:
      self.honor_users_and_groups = False
    else:
      self.honor_users_and_groups = True

    self.os_ops = OSOps(self.honor_users_and_groups, sudo=self.cfg.sudo)

    # Cache helps prevent updating the external status server config repeatedly
    self.diag_bundle_redaction_policy = None
    self.status_server_update_count = 0

  @property
  def process(self):
    # This is sort of like @lazy, but we want to be able to reset the property
    # when daemonizing, so there's no need to introduce @lazy for the sake of this property.
    if self._process is None:
      self._process = psutil.Process(os.getpid())
    return self._process

  def configure_process(self, daemon=False, pidfile=None, comm_name=None):
    """
    Configure various aspects of the agent process such as:
    - limits
    - daemonization
    - pid file
    - signal handlers
    """

    atexit.register(self.stop)
    os.environ['PATH'] = os.environ['PATH'] + ":/usr/kerberos/bin"
    os.umask(022)
    if daemon:
      self.daemon = Daemonize(comm_name, pidfile,
                              lambda *p: None,
                              logger=logging.getLogger("daemon"),
                              verbose=True)
      self.daemon.start()
      # We have forked but self._process refers to the original pid.  Must be reset.
      self._process = None

    elif pidfile:
      try:
        pidfile = open(pidfile, "w")
        pidfile.write("%s\n" % os.getpid())
        os.chmod(pidfile, 0666)
        pidfile.close()
      except Exception, e:
        LOG.info("Failed to write pid file: %s", pidfile)
        print "Failed to write pid file: %s" % pidfile
        sys.exit(-1)


    if comm_name:
      try:
        # OPSAPS-31457 Write name of process (cloudera-scm-agent) to /proc/self/comm -
        # making it easier to identify this process in ps & /proc
        commFile = None
        if os.path.exists("/proc/self/comm"):
          commFile = open("/proc/self/comm", "wb") # open without newline mode
          commFile.write(comm_name)
          commFile.close()
      except Exception, e:
        LOG.info("Failed to write to comm file: %s", e)
        if commFile:
          commFile.close()

    pid = os.getpid()
    if pid == os.getpgid(pid):
      # We're the process group leader already, which is ok.
      # Typically happens when running from the command line directly.
      LOG.info("Not starting a new session.")
    else:
      os.setsid()

    self.should_exit = False
    self.signal_caught = None
    signal.signal(signal.SIGINT, self.ready_exit_on_signal)
    signal.signal(signal.SIGHUP, self.ready_exit_on_signal)
    signal.signal(signal.SIGTERM, self.ready_exit_on_signal)
    signal.signal(signal.SIGUSR1, self.restart_signal)

  def configure_directories(self, reset=False):
    # Ensure "self.dir" exists.

    if not self.honor_users_and_groups:
      self.os_ops.sudo_command("mkdir", "-p", self.dir)
      self.os_ops.sudo_command("chown", "-R", "%s.%s" % (self.agent_user, self.agent_group), self.dir)

    self.mkworkdir()
    self.mkworkdir(self.cfg.paths.supervisor_dir)
    self.mkworkdir(self.cfg.paths.flood_dir, user=FLOOD_FS_USER, group=FLOOD_FS_GROUP)
    self.mkworkdir(self.cfg.paths.supervisor_include_path)

    self.mkworkdir(self.cfg.paths.cgroups_dir)
    self.mkworkdir(self.cfg.paths.process_dir)
    if not mount_tmpfs(self.cfg.paths.process_dir, reset, self.cfg.sudo):
      LOG.warning("Couldn't mount tmpfs for process directory. Generated "
                  "process scripts may be prevented from executing.")


  def configure_service(self):

    if self.cfg.ip_address == "127.0.0.1":
      LOG.warning("Agent is running on 127.0.0.1 (localhost). This is a misconfiguration for multi-machine clusters. Check your hostname settings.")
    if "_" in self.cfg.listening_hostname:
      raise Exception("Hostname is invalid; it contains an underscore character.")

    if self.cfg.logfile != '-':
      setup_rolling_logging(self.cfg.logfile)


    # Logging to the agent log file is now setup. Let's log a few things about
    # the agent.
    LOG.info("================================================================================")
    LOG.info("SCM Agent Version: %s", cmf.version.VERSION)
    LOG.info("Agent Protocol Version: %s", self.agent_protocol_version)
    LOG.info("Using Host ID: %s", self.cfg.host_id)
    LOG.info("Using directory: %s", self.dir)
    LOG.info("Using supervisor binary path: %s", self.cfg.supervisord_path)
    level = get_logging_level()
    LOG.info("Agent Logging Level: %s", logging.getLevelName(level))
    LOG.info("Agent config:")
    for (sec, sec_map) in self.cfg.value_map().iteritems():
      for (k, v) in sec_map.iteritems():
        LOG.info("\t%s.%s\t= %s", sec, k, v)
    if self.cfg.verify_cert_file is None and self.cfg.verify_cert_dir is None:
      LOG.info("Neither verify_cert_file nor verify_cert_dir are configured. "
               "Not performing validation of server certificates in HTTPS "
               "communication. These options can be configured in this agent's "
               "config.ini file to enable certificate validation.")

    if self.cfg.master_host is None:
      raise Exception("No server_host line was found in agent's configuration.")

    uid, gid = os.getuid(), os.getgid()
    LOG.info("Agent starting as pid %s user %s(%d) group %s(%d).",
             os.getpid(), self.agent_user, uid, self.agent_group, gid)

    if uid != 0:
      LOG.info("Because agent not running as root, all processes will run with "
               "current user.")

    Process.ClassInit(self.os_ops, self.cfg)

    # Token that can be used to verify http requests to the agent status server
    # first use of this will be for the download log file functionality
    self.token = bytes_to_hex(generate_random_bytes(16))
    LOG.debug("Agent token is: %s", self.token)
    with self.os_ops.open_write_file(self.cfg.paths.auth_token, mode=0700) as f:
      f.write(self.token)

    ssl_context = cmf.https.make_ssl_context(self.cfg)
    cmf.https.ssl_url_opener = cmf.https.SSLContextOpener(ssl_context, self.os_ops)

    self.cloud_info = cmf.public_cloud.get_public_cloud_info()
    if self.cloud_info is not None:
      LOG.info("Agent is running on a public cloud. Public hostname: %s", self.cloud_info["public_host_name"])
      LOG.debug("Public cloud info: %s", self.cloud_info)

    # Initialize cgroups state.

    cgroup_paths = dict()
    cgroup_paths["cpu"] = self.cfg.cpu_cgroup_path
    cgroup_paths["cpuacct"] = self.cfg.cpuacct_cgroup_path
    cgroup_paths["blkio"] = self.cfg.blkio_cgroup_path
    cgroup_paths["memory"] = self.cfg.memory_cgroup_path

    self.cg_manager = cmf.cgroups.CGroupManager(self.cfg.paths.cgroups_dir, cgroup_paths, self.cfg.sudo)
    if self.cg_manager.capabilities:
      LOG.info("Found cgroups capabilities: %s", self.cg_manager.capabilities)
      self.register_for_shutdown(self._cgroups_conditional_unmount)
    else:
      LOG.info("No cgroups capabilities found")
    # Tracks whether resource management is enabled; independent of whether
    # cgroups are actually mounted.
    self.rm_enabled = False

    # Tracks whether parcels are to be distributed
    self.server_managed_parcels = None

    self.tailer = cmf.log_tailer.LogTailer(self.cfg.host_id, self.cfg.listening_hostname)

    known_files = {}

    # Auto-TLS configuration:  all files in this directory will become valid
    # resources for use by processes, when referred to with a "{{CMF_CONF_DIR}}" path prefix.
    # These files will then be copied on request into process directories.
    if self.cfg.cm_auto_cert_dir:
      for f in os.listdir(self.cfg.cm_auto_cert_dir):
        full_path = os.path.join(self.cfg.cm_auto_cert_dir, f)
        if os.path.exists(full_path):
          known_files["%s/%s" % (Process.CONF_DIR_REF, f)] = full_path

    Process.known_resources_files.update(known_files)

    self.fs_map = FilesystemMap(self.cfg.monitored_nodev_filesystem_types,
                                poll_timeout=self.cfg.filesystem_map_poll_timeout_seconds,
                                join_timeout=self.cfg.filesystem_map_join_timeout_seconds,
                                tolerance=self.cfg.filesystem_map_cache_tolerance,
                                local_filesystem_whitelist=self.cfg.local_filesystem_whitelist)

    # dummy URL - here for completeness; start() will set this up on status_server init.
    self.url = "http://dummy"

    # Initialize the credential cache to be used to point to the agent directory
    cmf.kt_renewer.set_agent_wide_credential_cache_directory(
      self.cfg.agent_wide_credential_cache_location)


    self.register_for_shutdown(self.tailer.shutdown)

    # Create the named pipe that will be used to read events from supervisord,
    # if supervisord_events are enabled. Eager heartbeats and detection of
    # unexpected exits rely on supervisord_events being enabled.
    # The event fifo is kept in the agent's
    # working directory, and is written to the supervisord configuration file.
    # This means that if the agent needs to be re-started, it should use the
    # same log directory, or supervisord will have to be re-started too. (Note
    # that sending SIGHUP to supervisord makes it re-start all processes being
    # monitored, which is not what we want.)

    if self.cfg.supervisord_events:
      LOG.info("Setting up supervisord event monitor.")
      if not os.path.exists(self.cfg.paths.event_pipe_path):
        os.mkfifo(self.cfg.paths.event_pipe_path, 0600)
      self.event_mon = SupervisordEventMonitor(self.cfg.paths.event_pipe_path)
      self.register_for_shutdown(self.event_mon.close)

    def kill_active_children():
      '''
      Kill any children that may have been spawned by "multiprocessing" and thus have
      open file-descriptors to agent files/sockets. (These children may be hung on
      unresponsive NFS mounts.)  (See OPSAPS-32988.)
      :return:
      '''
      for child in multiprocessing.active_children():
        try:
          os.kill(child.pid, signal.SIGKILL)
        except:
          pass # bad pid, pid exited, we don't care

    self.register_for_shutdown(kill_active_children)

  def __issue_heartbeat(self):
    """
    Issue the actual heartbeat and handle the response.
    :return:
    """

    try:
      # The heartbeat mechanism is broken down into
      # three methods.
      heartbeat = self.prepare_heartbeat()
      heartbeat_response = self.send_heartbeat(heartbeat)
      if heartbeat_response is not None:
        # Payload content of HB may be empty, meaning "don't do anything" - OPSAPS-35294
        resp_data = heartbeat_response["data"]
        if resp_data:
          if not self.is_late_init_done:
            self._init_after_first_heartbeat_response(resp_data)
            self.is_late_init_done = True
          assert self.cm_guid is not None

          if self.cm_guid != resp_data.get("cm_guid"):
            LOG.error("Error, CM server guid updated, expected %s, received %s",
                      self.cm_guid, resp_data.get("cm_guid"))
            resp_data = None  # Now it's a no-op HB

        if resp_data:
          self.handle_heartbeat_response(resp_data)
          self.last_heartbeat_response = heartbeat_response

        # We may not have processed the response correctly, but we did HB to the server, so record that.
        self.last_successful_heartbeat = time.time()
        self.last_heartbeat_request = heartbeat

    except SystemExit:
      # Must explicitly re-raise to avoid getting caught and swallowed
      # below.
      raise
    except:
      LOG.exception("Caught unexpected exception in main loop.")

  def pre_startup(self):
    """
    Miscellaneous actions that need to be performed once at agent start up.
    """

    # OPSAPS-40536: enforce permissions of proc.json/config.zip files.
    # Simpler to just always do it than to check if it's needed.
    for fname in self.os_ops.file_tree_walk(self.cfg.paths.process_dir):
      if os.path.split(fname)[1] in [Process.PROC_JSON, Process.CONFIG_ZIP]:
        try:
          self.os_ops.set_path_metadata(fname, mode=0600)
        except Exception, exc:
          LOG.error("Failed to change permissions on: %s %s", fname, exc)

  def start(self, enable_supervisor_start=False):
    """
      Starts daemons and waits for exit
    """
    self.register_sigabrt_handler()
    self.started = True

    cmf.audit.util.start_init(self)
    cmf.audit.navigator_thread.NavigatorThread.start_threads()

    # Create the monitoring daemon so that any supervised process can create
    # its monitor.
    LOG.info("Using metrics_url_timeout_seconds of %f", self.cfg.metrics_url_timeout_seconds)
    LOG.info("Using task_metrics_timeout_seconds of %f", self.cfg.task_metrics_timeout_seconds)
    LOG.info("Using max_collection_wait_seconds of %f", self.cfg.max_collection_wait_seconds)

    self.monitor_daemon = cmf.monitor.daemon.MonitorDaemon(self.cfg.reported_hostname,
                                                           self.cfg.host_id,
                                                           self.fs_map,
                                                           self)
    self._read_cached_heartbeat()

    LOG.info("Supervised processes will add the following to their environment "
             "(in addition to the supervisor's env): %s", self.supervisee_environment)
    LOG.info("To override these variables, use /etc/cloudera-scm-agent/config.ini."
             " Environment variables for CDH locations are not used when CDH is"
             " installed from parcels.")

    self.find_or_start_supervisor(enable_supervisor_start)

    # To use eager heartbeats, create a pipe that the status server will use
    # to notify the main thread when an explicit heartbeat request arrives.
    wakeup_fd = os.pipe()
    self.hb_req_monitor = HeartbeatRequestMonitor(wakeup_fd[0])
    self.register_for_shutdown(self.hb_req_monitor.close)
    wakeup_fd = wakeup_fd[1]

    self.gen_supervisor_conf()

    self.url = cmf.status_server.start_status_server(self, self.cfg, wakeup_fd)


    self.supervisor_client.start_process(STATUS_SERVER_PROC)
    self.register_for_shutdown(lambda:
                               self.supervisor_client.stop_process(STATUS_SERVER_PROC))

    # Start the monitoring daemon, after we have the supervisor info
    self.monitor_daemon.start()

    LOG.info("Setting default socket timeout to %d", DEFAULT_SOCKET_TIMEOUT)
    socket.setdefaulttimeout(DEFAULT_SOCKET_TIMEOUT)

    # read the cm_guid if present
    self._read_cm_guid()

    # This is the main loop
    should_heartbeat = True

    self.read_active_parcels()
    self.read_saved_heartbeat_response()

    while not self.should_exit:
      if should_heartbeat:

        last = time.time()
        self.__issue_heartbeat()
        completed = time.time()
        if completed - last > 5:
          LOG.warn("Long HB processing time: %s", completed-last)
      try:
        self.tailer.process_events()
      except:
        LOG.exception("Error processing tailer events.")

      # Check for inotify events
      if self.inotify != None and self.inotify.check_events():
        self.inotify.read_events()
        self.inotify.process_events()

      diff = time.time() - last
      if self.cfg.supervisord_events:
        asyncore.loop(max(0, self.heartbeat_interval - diff),
                      use_poll=True,
                      count=1)
      elif diff < self.heartbeat_interval:
        time.sleep(max(0, self.heartbeat_interval - diff))

      if self.cfg.supervisord_events:
        events = self.event_mon.consume_events()
      else:
        events = []

      hb_requested = self.hb_req_monitor.is_hearbeat_requested()
      diff = time.time() - last
      if events or hb_requested:
        should_heartbeat = True
        LOG.debug("Sending eager heartbeat (%d events, explicit request = %s).",
                  len(events), hb_requested and "true" or "false")
      else:
        should_heartbeat = self.heartbeat_interval <= diff

      if diff >= 1.2 * self.heartbeat_interval:
        LOG.warn("Delayed HB: %ds since last", diff - self.heartbeat_interval)

      if self.cfg.supervisord_events:
        self.handle_unexpected_process_exits(events)

      # Check the staleness of the orphan process config directory every
      # orphan_process_directory_refresh_interval seconds and remove the
      # obsolete ones.
      now = time.time()
      duration = now - self.last_config_directory_refresh_time
      if duration > self.cfg.orphan_process_directory_refresh_interval:
        self.remove_obsolete_orphan_conf_dirs(now)
        self.last_config_directory_refresh_time = now

      if self.dump_heap_flag and \
          (self.dump_heap_thread is None or not self.dump_heap_thread.isAlive()):
        LOG.info("Starting thread to dump heap")
        self.dump_heap_thread = threading.Thread(target=self.dumpheap)
        self.dump_heap_thread.daemon = True
        self.dump_heap_thread.start()

    self.stop()
    LOG.info("Agent exiting; caught signal %s", self.signal_caught)
    sys.exit(0)

  def _read_cm_guid(self):
    # if cm_guid file is present then read the guid from it
    cm_guid_path = os.path.join(self.cfg.lib_dir, CM_GUID_FILE)
    if os.path.exists(cm_guid_path):
      cmGuidFile = None
      try:
        try:
          cmGuidFile = open(cm_guid_path, "r")
          self.cm_guid = cmGuidFile.read()
          assert self.cm_guid is not None
        except Exception, e:
          LOG.error("Failed to read cm_guid file: %s", e)
      finally:
        if cmGuidFile:
          cmGuidFile.close()

  def _init_after_first_heartbeat_response(self, response_data):
    """
    Separate init method that runs after the first heartbeat response is
    received. This allows us to delay initialization of components that
    require heartbeat response state
    """
    LOG.debug("Completing init after first heartbeat response")
    uid, gid = os.getuid(), os.getgid()
    user, group = pwd.getpwuid(uid)[0], grp.getgrgid(gid)[0]

    # if cm_guid file is not present then create one
    if self.cm_guid is None:
      cm_guid_path = os.path.join(self.cfg.lib_dir, CM_GUID_FILE)
      cmGuidFile = None
      assert response_data.get("cm_guid") is not None
      self.cm_guid = response_data.get("cm_guid")
      try:
        try:
          cmGuidFile = open(cm_guid_path, "w")
          cmGuidFile.write(self.cm_guid)
        except Exception, e:
          LOG.error("Failed to write cm_guid file: %s", e)
      finally:
        if cmGuidFile:
          cmGuidFile.close()

    LOG.info("CM server guid: %s", self.cm_guid)

    # The server provided parcels directory can be overridden by the config file.
    # We allow this for backwards compatibility - the config file used to be the
    # only way to control this location.
    if self.cfg.parcel_dir is not None:
      LOG.info("Using parcels directory from config.ini value: %s", self.cfg.parcel_dir)

    if self.cfg.parcel_dir is None:
      self.cfg.parcel_dir = response_data.get("parcels_directory")
      LOG.info("Using parcels directory from server provided value: %s", self.cfg.parcel_dir)
    if self.cfg.parcel_dir is None:
      LOG.info("Using default parcel directory: %s", cmf.config.DEFAULT_PARCEL_DIR)
      self.cfg.parcel_dir = cmf.config.DEFAULT_PARCEL_DIR

    # We get the real path (which eliminates any symlinks in the path) so that
    # later when we serve files from the agent directories we can easily
    # enforce that no symlinks are followed when restricting serving of files
    # to the process directory.
    self.dir = self.cfg.paths.base_dir
    #if args.clear_agent_dir:
    #  LOG.info("Clearing agent dir %s", self.dir)
    #  shutil.rmtree(self.dir, ignore_errors=True) # may contain a tmpfs mount

    self.parcel_dir_is_set = True
    parcel_cache_dir = None
    try:
      self.mkabsdir(self.cfg.parcel_dir, user=user, group=group, mode=0755)
      parcel_cache_dir = os.path.join(os.path.dirname(os.path.normpath(self.cfg.parcel_dir)), "parcel-cache")
      self.mkabsdir(parcel_cache_dir, user=user, group=group, mode=0755)
    except Exception, e:
      msg = "Failed to create parcel directories at %s: %s" % (self.cfg.parcel_dir, str(e))
      LOG.error(msg)
      self.parcel_init_error = msg
      return

    self.repo = cmf.parcel.Repository(self.cfg.parcel_dir, self.cfg.package_dir, self.cfg.sudo, self.os_ops)
    self.repo.set_ensure_users_groups_permissions(response_data.get("apply_parcel_users_groups_permissions"))
    self.repo.refresh()
    try:
      if self.inotify is None:
        self.inotify = self.repo.configure_inotify()
        self.register_for_shutdown(self.inotify.stop)
    except:
      LOG.exception("Failed to configure inotify. Parcel repository will not auto-refresh.")
      self.inotify = None


    if self.downloader is None:
      # pylint: disable=redefined-variable-type
      opener = self.flood_opener
      if not FLOOD_ENABLED or self.flood_opener is None:
        opener = cmf.downloader.URLopener()

      self.downloader = cmf.downloader.Downloader(parcel_cache_dir, 1, opener)

    self.register_for_shutdown(self.downloader.cleanUp)
    self.cache = cmf.parcel_cache.ParcelCache(parcel_cache_dir, self.repo, self.downloader)

    # Client config directories will be created in
    # self.cfg.paths.process_dir, which means they'll be subject to
    # process garbage collection (see handle_orphan_processes()).
    self.client_configs = cmf.client_configs.ClientConfigs(
      self,
      package_dir=self.cfg.package_dir,
      parcel_repo=self.repo,
      deploy_dir=self.cfg.paths.process_dir,
      sudo=self.cfg.sudo,
      deploy_env=self.supervisee_environment)
    self.client_configs.load()

  def _cgroups_conditional_unmount(self):
    """Unmount mounted cgroups roots only if there are no existing cgroups."""
    if self.cg_manager.has_at_least_one_group():
      LOG.info("At least one outstanding cgroup; retaining cgroup mounts")
      return
    else:
      LOG.info("No extant cgroups; unmounting any cgroup roots")
      self.cg_manager.disable()

  @staticmethod
  def _get_logical_cpu_core_number():
    num_logical_cpu_core = 0
    try:
      try:
        f = open('/proc/cpuinfo')
        for line in f:
          if line.startswith('processor'):
            num_logical_cpu_core += 1
      finally:
        f.close()
    except:
      LOG.exception("Failed to open /proc/cpuinfo")

    return num_logical_cpu_core

  # Get the physical CPU core number
  # It is possible that the host is running on a VM which doesn't have a
  # physical CPU. So in this case, we need to set it to the logical CPU core
  # number.
  @staticmethod
  def _get_physical_cpu_core_number(logic_cpu_core_num):
    # Count the set bit in a mapping file.
    def _path_sibling(path):
      result = 0
      try:
        try:
          f = open(path)
          for character in iter(lambda: f.read(1), ''):
            if character.isdigit():
              value = int(character)
              while value > 0:
                value &= (value - 1)
                result += 1
        finally:
          f.close()
      except:
        LOG.exception("Failed to open path: %s", path)

      return result


    physical_cpu_core_num = 0
    num_hyperthreads_per_physcial_cpu =\
      _path_sibling("/sys/devices/system/cpu/cpu0/topology/thread_siblings")
    if num_hyperthreads_per_physcial_cpu:
      physical_cpu_core_num =\
        logic_cpu_core_num / int(num_hyperthreads_per_physcial_cpu)
    else:
      physical_cpu_core_num = logic_cpu_core_num
    return physical_cpu_core_num

  def _update_dssd_support_status(self):
    """Sets supports_dssd to True if host is configured to connect to a DSSD appliance"""
    dssd_device_dir = self.safety_valve.get("dssd_vpci_device_dir",
                                            "/dev/dssd/vpci/client").strip()
    if dssd_device_dir != self.cached_dssd_device_dir:
      self.cached_dssd_device_dir = dssd_device_dir
      self.supports_dssd = False
      if os.path.isdir(dssd_device_dir):
        for f in os.listdir(dssd_device_dir):
          mode = os.stat(os.path.join(dssd_device_dir, f)).st_mode
          if stat.S_ISCHR(mode):
            self.supports_dssd = True
            break

      if self.supports_dssd:
        LOG.debug("DSSD vpci device detected in %s", dssd_device_dir)
      else:
        LOG.debug("No DSSD vpci device detected in %s", dssd_device_dir)

  def _set_memory_stats(self, stats):

    if self.cfg.memory_cgroup_path is not None:
      mem_limit, mem_usage = self.cg_manager.get_cgroup_mem_info()
      if mem_limit is not None and mem_usage is not None:
        stats["physical_memory_total"] = mem_limit
        stats["physical_memory_used"] = mem_usage
        return

    vmem = psutil.virtual_memory()
    stats["physical_memory_total"] = vmem.total
    # To determine used memory, we should not use vmem.used, since, according to the
    # psutil documentation, this is provided for informational purposes only. Furthermore
    # we found that it may report an incorrect value when significant memory is reserved as
    # shared memory. Shared memory is not available to new processes, but vmem.used
    # does not take it into account.
    stats["physical_memory_used"] = vmem.total - vmem.available

  def _host_heartbeat(self):
    """Populates dict data with host heartbeat info."""
    status = dict()
    status["host_name"] = self.cfg.reported_hostname
    status["ip_address"] = self.cfg.ip_address
    status["agent_url"] = self.url
    status["agent_token"] = hex_to_bytes(self.token)
    status["agent_system_user"] = self.agent_user
    status["agent_system_group"] = self.agent_group

    # Agent needs to report the number of physical and logical CPU cores.
    # 1. status["num_cores"] saves the logical CPU core number
    # 2. status["num_physical_cores"] saves the physical CPU core number
    if self.cfg.cpu_cgroup_path is None:
      status["num_cores"] = self.num_logical_cpu_cores
      status["num_physical_cores"] = self.num_physical_cpu_cores
    else:
      # Agent is running inside a container i.e logical_cores == physical_cores
      cpu_shares = self.cg_manager.get_cpu_shares()
      num_of_cpus = int(cpu_shares / 1024)
      status["num_cores"] = num_of_cpus
      status["num_physical_cores"] = num_of_cpus

    status["supports_dssd"] = self.supports_dssd

    now = time.time()
    if self.cached_filesystem_info is None or \
       (now - self.last_rare_heartbeat >=
        RARE_HEARTBEAT_MULTIPLIER * self.heartbeat_interval):
      self.fs_map.refresh()
      self.cached_filesystem_info = self.fs_map.get_info_for_scm_heartbeat()
      self.last_rare_heartbeat = now

    status["filesystem_info"] = self.cached_filesystem_info["filesystem_info"]
    status["component_info"] = self.cached_component_info
    if self.repo is None:
      status["parcel_info"] = {}
      status["parcel_err"] = {}
    else:
      status["parcel_info"] = self.repo.products
      status["parcel_err"] = self.repo.errors

    if self.last_available_active_parcels is None:
      status["active_parcels"] = {}
    else:
      status["active_parcels"] = self.last_available_active_parcels

    # We cannot send the parcel directory value until we know what it is, which
    # won't happen until the first heartbeat response is evaluated.
    if self.parcel_dir_is_set:
      status["parcels_directory"] = self.cfg.parcel_dir
    else:
      # We must explicitly set this, even though it's just the default value, or
      # avro will fail to validate it.
      status["parcels_directory"] = ''
    status["parcels_directory_overridden"] = self.parcel_dir_from_config
    status["distro_info"] = self.dist
    status["cgroup_capabilities"] = self.cg_manager.capabilities

    status["cloud_status"] = self.cloud_info

    status["ephemeral_port_range"] = self.os_ops.get_ephemeral_port_range()

    stats = dict()
    stats["load_avg"] = list(os.getloadavg())

    self._set_memory_stats(stats)

    virtmem_usage = psutil.swap_memory()
    stats["virtual_memory_total"] = virtmem_usage.total
    stats["virtual_memory_used"] = virtmem_usage.used
    stats["mountpoint_available_space_bytes"] = \
        self.cached_filesystem_info["mountpoint_available_space_bytes"]
    parcel_downloads = []
    parcel_downloads.extend(self.report_parcel_init_errors())
    if self.downloader is not None:
      parcel_downloads.extend(self.downloader.get_current_downloads())
    stats["parcel_downloads"] = parcel_downloads

    return status, stats

  def report_parcel_init_errors(self):
    # Report back the parcel initialization error for all managed parcels
    # so that it surfaces in the parcel UI.
    download_errors = []
    if self.parcel_init_error and self.server_managed_parcels:
      for url in self.server_managed_parcels:
        d = dict()
        d[cmf.downloader.URL_FIELD] = url
        d[cmf.downloader.STATE_FIELD] = cmf.downloader.EXCEPTION
        d[cmf.downloader.ACTIVE_FIELD] = False
        err = dict()
        err[cmf.downloader.ERROR_CODE_FIELD] = str(cmf.downloader.UNEXPECTED_ERROR_CODE)
        err[cmf.downloader.ERROR_MSG_FIELD] = self.parcel_init_error
        d[cmf.downloader.ERROR_FIELD] = err
        download_errors.append(d)
    return download_errors

  def parcel_directory_in_use(self):
    """Check if the parcel directory is in use. The parcel directory is in use
    if there is at least one parcel in 'distributed' state. The 'distributed'
    state is tracked by the agent's 'server_managed_parcel' variable"""
    return (self.server_managed_parcels is not None and
            len(self.server_managed_parcels) > 0)

  def monitoring_supervisord_events(self):
    return self.cfg.supervisord_events

  def supervisor_check(self):
    """
    Check if supervisord process is still alive.  If not, restart the agent; which will
    attempt to restart supervisord via our normal startup path.
    """
    if self.supervisor_client.is_running():
      return

    LOG.warning("Supervisor failed (pid %s).  Restarting agent.", self.supervisor_client.pid)
    try:
      os.execv(self.exec_args[0], self.exec_args)
    except Exception, e:
      LOG.exception("Failed agent re-exec: %s", e)
      LOG.error("Agent exiting.")
      sys.exit(-1)


  def get_supervisor_data(self, count_failures=False):
    data = self.supervisor_client.get_supervisor_data(count_failures)

    return data

  def prepare_heartbeat(self):
    # AVRO objects are dict-based
    request = dict()
    request["version"] = self.agent_protocol_version
    request["host_id"] = self.cfg.host_id

    # Get host-level data
    request["status"] = dict()
    request["status"]["host"], request["host_stats"] = self._host_heartbeat()

    # Let's work through active processes
    process_statuses = []
    process_stats = []

    # Gather supervisor data
    self.supervisor_check()

    supervisor_data = self.get_supervisor_data(True)

    # We combine both lists since we want to report back to the server
    # about currently active processes, but also processes that failed
    # to activate.
    ids_seen = set()
    for process in Process.GetProcesses():

      if process.id in ids_seen:
        # This should never happen.
        LOG.warning("Duplicate process id: %s", process.id)
      ids_seen.add(process.id)

      process.update_supervisor_data(supervisor_data.get(process.name, None))
      status, stats = process.make_status()
      if (status, stats) != (None, None):
        process_statuses.append(status)
        process_stats.append(stats)

    request["status"]["processes"] = process_statuses
    request["process_stats"] = process_stats

    # The client config module is only initialized after the first heartbeat
    # due to its dependency on the parcel repo. It's safe to report nothing
    # in the first heartbeat; the server treats statuses as purely additive.
    client_config_statuses = []
    if self.client_configs:
      client_config_statuses = self.client_configs.get_status()
    request["status"]["client_configs"] = client_config_statuses

    md = _sha1(pickle.dumps(request["status"]))
    request["status_hash"] = md.digest()

    cpu_total = 0.0
    try:
      # Shouldn't fail, but, if it does, let's not
      # worry about it.
      cpu_times = self.process.cpu_times()
      cpu_total += cpu_times.user + cpu_times.system
    except:
      LOG.exception("Exception getting agent cpu times.")

    try:
      # If the supervisor process is gone, this
      # won't work, and we don't want that to prevent
      # heartbeating.
      cpu_times = self.supervisor_client.psutil_process.cpu_times()
      cpu_total += cpu_times.user + cpu_times.system
    except:
      pid = "Unknown"
      # There is no supervisor_client during unit tests
      if self.supervisor_client is not None:
        pid = self.supervisor_client.pid
      LOG.exception("Exception getting supervisor cpu times (supervisor pid: %s).", pid)

    request["total_cpu"] = cpu_total

    if self.last_heartbeat_response is not None:
      request["last_response_hash"] = self.last_heartbeat_response["data_hash"]

      if request["status_hash"] == self.last_heartbeat_response["last_request_hash"]:
        LOG.debug("Matched request hashes. Not sending status")
        del request["status"]
      else:
        LOG.debug("Unmatched request hashes. Sending status")
    else:
      request["last_response_hash"] = ""

    # send the last known CM server guid
    if self.cm_guid is not None:
      request["cm_guid"] = self.cm_guid

    # Because heartbeating sends log information to the master,
    # including the heartbeat in the log message triggers
    # recursively larger and larger heartbeats.
    return request

  def send_heartbeat(self, heartbeat):
    '''
    Wrapper around _send_heartbeat to track RPC time;
    trying to avoid another level of indentation in _send_heartbeat().
    This would be much nicer with "with".... but RHEL 5.5.

    :param heartbeat:
    :return:
    '''
    start = time.time()
    response = self._send_heartbeat(heartbeat)
    end = time.time()
    self.heartbeat_tracker.record(start, end)
    return response


  def _send_heartbeat(self, heartbeat):
    try:
      if self.requestor is None:
        if self.cfg.use_tls:
          transceiver = cmf.https.HTTPSTransceiver(self.cfg.master_host,
                                                   self.cfg.master_port,
                                                   self.cfg.verify_cert_file,
                                                   self.cfg.verify_cert_dir,
                                                   self.cfg.client_key_file,
                                                   self.cfg.client_cert_file,
                                                   self.cfg.client_keypw_file,
                                                   self.cfg.client_keypw_cmd,
                                                   self.cfg.max_cert_depth)
        else:
          transceiver = avro.ipc.HTTPTransceiver(self.cfg.master_host,
                                                 self.cfg.master_port)
        self.requestor = avro.ipc.Requestor(MASTER_AGENT_PROTOCOL,
                                            transceiver)
      LOG.debug("Sending heartbeat to master.")
      assert self.requestor is not None
      heartbeat_data = dict(request=UNICODE_SANITIZER(heartbeat))
      response = self.requestor.request('heartbeat', heartbeat_data)
      LOG.debug("Received response from master: %s", response)
      if (self.last_heartbeat_response is not None and
          (response["data_hash"] ==
           self.last_heartbeat_response["data_hash"])):
        LOG.debug("Matched response hashes. Using previous response")
        self.last_heartbeat_response["last_request_hash"] = \
          response["last_request_hash"]
        response = self.last_heartbeat_response
      else:
        LOG.debug("Unmatched response hashes. Using and caching new response")
        response_path = os.path.join(self.cfg.lib_dir, RESPONSE_CACHE_FILE)
        try:
          new_path = response_path + ".new"
          with self.os_ops.open_write_file(new_path, mode=stat.S_IRUSR | stat.S_IWUSR) as new_file:
            avro_writer = avro.datafile.DataFileWriter(new_file,
                                                       avro.io.DatumWriter(),
                                                       HEARTBEAT_RESPONSE_SCHEMA)
            avro_writer.append(response)
            avro_writer.close()
          self.os_ops.move_file(response_path, new_path)

        except:
          LOG.exception("Failed to cache last heartbeat response: %s", response_path)
    except:
      self.log_heartbeat_exception("Heartbeating to %s:%s failed." % (self.cfg.master_host, self.cfg.master_port))
      self.last_failed_heartbeat_request = heartbeat
      self.requestor = None
      response = self.read_saved_heartbeat_response()


    return response

  def read_saved_heartbeat_response(self):
    # If we failed to heartbeat and have no previous response, attempt to
    # load the last saved heartbeat response, if there is one. This will
    # allow us to complete initialisation and move forward as best we can.
    if self.last_heartbeat_response is not None:
      return None
    response = None
    try:
      response_path = os.path.join(self.cfg.lib_dir, RESPONSE_CACHE_FILE)
      if os.path.exists(response_path):
        LOG.info("Loading last saved hb response to complete initialization: %s", response_path)
        avro_reader = avro.datafile.DataFileReader(open(response_path),
                                                   avro.io.DatumReader())
        response = avro_reader.next()
        avro_reader.close()
    except:
      LOG.exception("Failed to read saved heartbeat response.")
    self.last_heartbeat_response = response
    return response

  def log_heartbeat_exception(self, message):
    ex = sys.exc_info()[1]
    if ex is not None and hasattr(ex, 'errno') and ex.errno == errno.ECONNREFUSED:
      elapsed_time = time.time() - self.last_logged_connection_refused
      if elapsed_time > 0 and elapsed_time <= CONN_REFUSED_THROTTLE_DURATION:
        return
      else:
        self.last_logged_connection_refused = time.time()
    LOG.exception(message)


  def update_component_info(self, active_parcels):
    """
    Updates self.cached_component_info.
    """
    recalculate = False
    # active_parcels is a map (product -> version);
    # recalculate if it's changed.
    if active_parcels != self.last_available_active_parcels:
      LOG.info("Active parcel list updated; recalculating component info.")
      recalculate = True

    now = time.time()
    if now - self.last_version_detection >= VERSION_DETECTION_INTERVAL:
      recalculate = True

    if recalculate:
      # Extract the component versions of all active parcels
      active_parcel_components = []
      for k, v in active_parcels.iteritems():
        active_parcel_components += self.repo.get_components_for_active_parcel(k, v)

      component_info = \
        VersionDetector.get_component_versions(
          self.supervisee_environment,
          self.cfg.package_dir,
          java_home_config=self.java_home_config,
          supervisor_info=self.supervisor_info)
      self.last_version_detection = now

      # VersionDetecter looks at component versions largely based on package
      # locations.  If parcels are in use, most of the pkg_component_info
      # components need to be marked inactive, and we need to restrict the list
      # to installed components.
      if len(active_parcel_components) > 0:
        # Remove uninstalled components
        component_info = [
          p for p in component_info
          if p[COMPONENT_VERSION_KEY] != "NO_SCRIPT"
        ]
        # Mark CDH components inactive
        for p in component_info:
          if p[CDH_VERSION_KEY] != "NOT_APPLICABLE":
            p["active"] = False

      self.cached_component_info = active_parcel_components + component_info

      # pylint: disable=using-constant-test
      if False:
        # This disabled block may be useful for debugging.

        # Check that active and source are populated.
        for p in self.cached_component_info:
          if "active" not in p or "source" not in p:
            LOG.error("Code bug. Unexpectedly missing active/source " +
                      "attribute in component: " + str(p))

        # Print active and inactive components:
        LOG.info("Active: " + ",".join([p["name"] + "/" + p["source"] for p in self.cached_component_info if p["active"]]))
        LOG.info("Inactive: " + ",".join([p["name"] + "/" + p["source"] for p in self.cached_component_info if not p["active"]]))

  def handle_heartbeat_response(self, response):
    self.lock.acquire()
    LOG.debug("Begin Heartbeat Response handling")
    try:
      try:
        self._handle_heartbeat_response(response)
      except:
        # Worst case exception handler
        LOG.exception("Failed to handle Heartbeat Response: %s", response)
    finally:
      self.lock.release()
      LOG.debug("Completed Heartbeat Response handling")


  def save_active_parcels(self, active_parcels):
    """
    Write the active parcels file.  This is needed to allow the agent to restart and remember
    active parcels so as to not de-activate parcels unnecessarily.
    """
    path = os.path.join(self.cfg.lib_dir, self.ACTIVE_PARCELS_FILE)

    with self.os_ops.open_write_file(path) as f:
      f.write(json.dumps(active_parcels))

  def read_active_parcels(self):
    """
    Read the active parcels file.
    """
    path = os.path.join(self.cfg.lib_dir, self.ACTIVE_PARCELS_FILE)
    try:
      with open(path) as f:
        self.last_available_active_parcels = json.load(f)
        LOG.info("Previously active parcels: %s", self.last_available_active_parcels)
    except Exception as exc:
      LOG.info("Failed to read available parcel file: %s", exc)

  def _update_parcel_activation_state(self, response):
    self.repo.set_ensure_users_groups_permissions(response.get("apply_parcel_users_groups_permissions"))
    active_parcels = response.get("active_parcels")

    # (De)Activate system symlinks for parcels
    available_active_parcels = active_parcels.copy()
    manage_new_parcels = response.get("create_parcel_symlinks")

    old_available_active_parcels = dict()
    manage_old_parcels = False
    if self.last_heartbeat_response is not None:
      old_response = self.last_heartbeat_response["data"]
      old_available_active_parcels = self.last_available_active_parcels
      manage_old_parcels = old_response.get("create_parcel_symlinks")
    else:
      # Remove all alternatives in case we're immediately told to discard some
      # parcels
      for (name, parcels) in self.repo.products.iteritems():
        for (version, parcel) in parcels.iteritems():
          self.repo.ensure_alternatives(parcel, False)
          self.repo.ensure_active_symlink(parcel, False)

    # Filter out any parcels from symlinks consideration that are not yet
    # available. We do this to avoid trying to activate any parcels too early
    # By also removing these entries from the old symlink map we avoid any
    # scenario where deactivate might accidentally get called
    LOG.debug("active_parcels: " + str(active_parcels))
    for name, version in active_parcels.items():
      if not (name in self.repo.products and version in self.repo.products[name]):
        # We just delete the name as the active_parcels map never contains
        # multiple versions of the same product
        if name in available_active_parcels:
          del available_active_parcels[name]
        if old_available_active_parcels is not None:
          if name in old_available_active_parcels:
            del old_available_active_parcels[name]
    LOG.debug("available_active_parcels: " + str(available_active_parcels))
    LOG.debug("old_available_active_parcels: " + str(old_available_active_parcels))

    self.repo.configure_all_symlinks(old_available_active_parcels,
                                     available_active_parcels,
                                     manage_old_parcels,
                                     manage_new_parcels)

    # Calculate the component versions
    self.update_component_info(available_active_parcels)

    self.last_available_active_parcels = available_active_parcels

    self.save_active_parcels(available_active_parcels)


    if response.get("server_manages_parcels"):
      self.server_managed_parcels = response.get("server_managed_parcels")
      retain_parcels_in_cache = response.get("retain_parcels_in_cache")
      LOG.debug("Server Managed Parcels: %s | Retain Parcels in Cache: %s",
                str(self.server_managed_parcels),
                str(retain_parcels_in_cache))
      self.cache.update(self.server_managed_parcels, retain_parcels_in_cache)

  def _update_diag_bundle_redaction_policy(self, diag_bundle_redaction_policy):
    # Not certain whether REDACTOR object is used within agent proper process.
    # Updating it regardless to avoid any regressions.
    REDACTOR.set_policy_from_string(diag_bundle_redaction_policy)

    # In case there were errors in updating config in status server it's
    # good to send an update once a day even if config hasn't changed.
    self.status_server_update_count += 1
    num_heartbeats_in_a_day = (24 * 60 * 60) / 15

    if self.diag_bundle_redaction_policy != diag_bundle_redaction_policy or \
      self.status_server_update_count % num_heartbeats_in_a_day == 0:
      # Send updated config to external status server
      try:
        with self.os_ops.open_write_file(self.cfg.paths.redaction_policy_string, mode=0700) as f:
          f.write(diag_bundle_redaction_policy)
        status_server_pid = get_single_proc_pid(STATUS_SERVER_PROC)
        if status_server_pid is not None:
          LOG.info("Reloading status server config, sending SIGHUP")
          os.kill(status_server_pid, signal.SIGHUP)
        else:
          raise Exception("External status server not found")
        self.diag_bundle_redaction_policy = diag_bundle_redaction_policy
      except:
        LOG.exception("Failed updating diag bundle redaction policy. Redaction may not work.")

  def _handle_heartbeat_response(self, response):
    # pylint: disable=too-many-locals
    self.heartbeat_interval = response["heartbeat_interval"]

    self.extra_configs = response['extra_configs']
    if self.extra_configs is not None:
      self.java_home_config = self.extra_configs['JAVA_HOME']
    else:
      self.java_home_config = None

    # This must run before any parcel-related processing of HB data.
    # This requirement ensures that any new download attempts will
    # not occur until after any reconfiguration attempts are made.
    self.handle_heartbeat_flood(response["flood_rack_peers"],
                                response["flood_seed_timeout"],
                                response["flood_torrent_port"])

    # Update our monitor daemon with the latest firehose information. Note
    # that if no information is specified, this will turn off reporting to
    # the firehose.
    firehose_connection_info = response.get("firehoses", [])
    self.monitor_daemon.update_firehose_connection_info(
      firehose_connection_info)

    host_config_data = response.get("host_collection_config_data")
    if not host_config_data is None:
      host_config = dict()
      for c in host_config_data:
        if c.has_key('config_name') and c.has_key('config_value'):
          host_config[c['config_name']] = c['config_value']
      self._update_safety_valve(host_config)
      if host_config['diag_bundle_redaction_policy'] is not None:
        self._update_diag_bundle_redaction_policy(host_config['diag_bundle_redaction_policy'])
      self.monitor_daemon.update_host_config(host_config)
    else:
      LOG.debug("Heartbeat response did not contain host collection config data.")

    if self.repo is None:
      # Something went wrong with repo initialization
      # Just record managed parcels to report errors on.
      if response.get("server_managed_parcels"):
        self.server_managed_parcels = response.get("server_managed_parcels")
    else:
      self._update_parcel_activation_state(response)

    logs_and_rules_config = response.get("log_tailing_config")
    eventserver_host = response.get("eventserver_host")
    eventserver_port = response.get("eventserver_port")
    if logs_and_rules_config is not None:
      try:
        self.tailer.set_config(logs_and_rules_config)
        self.tailer.set_eventserver_config(eventserver_host, eventserver_port)
      except:
        LOG.exception("Error setting tailer configuration.")
    else:
      LOG.debug("Heartbeat response did not contain log tailing configuration.")

    # Mount/unmount cgroup subsystems. Cgroup state is taken into consideration:
    # 1. If there are any cgroups, we mount so that we can properly
    #    destroy them.
    # 2. Likewise, we only unmount when all cgroups are gone, even if
    #    the server has disabled RM in the interim.
    #
    # In the interest of performance, this won't remount subsystems that
    # were unmounted by an outside party.
    self.rm_enabled = response.get("rm_enabled")
    if self.rm_enabled or self.cg_manager.has_at_least_one_group():
      self.cg_manager.enable()
      self.cg_manager.destroy_orphaned_groups()
    elif not self.cg_manager.has_at_least_one_group():
      self.cg_manager.disable()

    # Handle client configs.
    #
    # Done after parcels, because client config deployment may depend on
    # parcel-specific environment variables.
    #
    # Done before processes in case a process depends on a client config
    # (it shouldn't, but we're being extra safe).
    try:
      cc = response.get("client_configs")
      if self.client_configs is not None:
        self.client_configs.rectify(
          client_configs=cc,
          custom_java_home_config=self.java_home_config)
    except:
      LOG.exception("Failed to handle client configs: %s",
                    response.get("client_configs", None))

    self.handle_heartbeat_processes(response)

    for cb in self.hb_callbacks:
      try:
        cb(self, response)
      except Exception as exc:
        THROTTLED_LOG.exception("Failed HB callback: %s %s", cb, exc)

  def register_hb_callback(self, cb):
    self.hb_callbacks.append(cb)

  def handle_heartbeat_processes(self, response):
    '''
    Process-related handling of heartbeats.
    - Clean-up old/dead processes.
    - Start new processes.

    :param response: The heartbeat response to be processed.
    :return:
    '''

    # Remove past processes.  Ignore already present orders.  Add new ones.
    new_processes = dict((pr["id"], pr) for pr in response["processes"])
    assert len(new_processes) == len(set(new_processes.keys())), "Processes must have unique ids."
    # Clear failed processes that the server no longer cares about,
    # and don't reactivate failed processes.

    procs = []
    for proc in Process.GetProcesses():
      # Subtract from new_process immediately, in case an exception is
      # raised. If we don't do this now, we may end up orphaning the
      # existing process without actually deactivating it.
      try:
        new_raw = new_processes.pop(proc.id)
      except KeyError:
        new_raw = None

      if proc.is_failed():
        # This process failed in a fatal way already;
        # we don't want to activate it again
        continue

      procs.append((proc, new_raw, False))
    Process.issue_updates(self, procs)

    # We already subtracted the pre-existing ones from new_processes.
    # These are the truly new processes.
    procs = [(None, raw, True) for raw in new_processes.values()]

    Process.issue_updates(self, procs)

  def handle_unexpected_process_exits(self, events):
    ''' Looks through the received events for unexpected process exits
        and updates the corresponding monitors unexpected exit state.'''

    processes = Process.GetProcesses()
    unexpected_process_exits = dict()

    for event in events:
      if event.properties["eventname"] == "PROCESS_STATE_EXITED":
        process_pid = event.payload_properties["pid"]
        if process_pid is None:
          LOG.warn("Process exited event payload '%s' does not have a "\
                     "'pid' field - ignoring event", event.payload)
          continue

        process_name = event.payload_properties["processname"]
        if process_name is None:
          LOG.warn("Process exited event payload '%s' for PID '%s' " \
                     "does not have a 'processname' field - " \
                     "ignoring event", event.payload, process_pid)
          continue

        expected = event.payload_properties["expected"]
        if expected is None:
          LOG.warn("Process exited event payload '%s' for PID '%s' " \
                     "associated with process '%s' does not have a " \
                     "'expected' field - ignoring event",
                   event.payload, process_pid, process_name)
          continue

        # An 'expected' value of 0 indicates that the process has
        # exited unexpectedly
        if expected == str(0):
          LOG.info("PID '%s' associated with process '%s' " \
                     "with payload '%s' exited unexpectedly",
                    process_pid, process_name, event.payload)
          unexpected_process_exits[process_name] = process_pid

    for process in processes:
      if unexpected_process_exits.has_key(process.name):
        if process.monitor is None:
          LOG.warn("Process exited event for process '%s' does not have a " \
                     "monitor to handle event - the event will be ignored.",
                   process.name)
          continue

        is_oom_exit = False
        try:
          oom_kill_path = os.path.join(process.dir, 'killed_by_killparent_on_oom')
          if os.path.isfile(oom_kill_path):
            oom_exit_time = open(oom_kill_path).read()
            if self.last_oom_exit_time != oom_exit_time:
              is_oom_exit = True
              self.last_oom_exit_time = oom_exit_time
        except:
          THROTTLED_LOG.warn("Error trying to determine if process was killed due to OOM")
        LOG.info("[%s] Recording unexpected exit.", process)
        process.monitor.record_unexpected_exit(is_oom_exit)

  def gen_supervisor_conf(self):
    '''
    Supervisor related configuration of flood and the status server,
    '''

    sup_data = {
      'cmf_python' : sys.executable,
      'cmf_python_path' : os.path.join(sys.prefix, "bin"),
      'flood_enabled' : FLOOD_ENABLED,
      'flood_dir' : self.cfg.paths.flood_dir,
      'agent_dir' : self.cfg.paths.base_dir,
      'log_dir': self.cfg.log_dir,
    }

    if self.honor_users_and_groups:
      sup_data["flood_user"] = FLOOD_FS_USER
      sup_data["flood_group"] = FLOOD_FS_GROUP

    if self.honor_users_and_groups:
      sup_data["agent_user"] = AGENT_USER
      sup_data["agent_group"] = AGENT_GROUP

    sup_config = render_text("supervisor-cm.conf", data=sup_data)
    write_conf_file(os.path.join(self.cfg.paths.supervisor_include_path, "flood.conf"), sup_config)

    self.supervisor_client.reload()

  def handle_heartbeat_flood(self, rack_peers, seed_timeout, torrent_port):
    '''
    Handle Flood-related HB data.

    seed_timeout and rack_peers are expected to change freely and don't
    imply any requirement to restart flood.  Changes to torrent_port should
    not be made while there are active downloads (this includes setting
    it to 0, which disables flood).

    :param rack_peers: Corresponding data from HB response.
    :param seed_timeout: Corresponding data from HB response.
    :param torrent_port: Corresnponding data from HB response.
    :return:dow
    '''
    if not FLOOD_ENABLED:
      return
    flood_dir = os.path.join(self.cfg.parcel_dir, ".flood")
    flood_server = "localhost:%s" % self.cfg.flood_rpc_port
    old = self.flood_opener

    if torrent_port is None or torrent_port == 0:
      # Won't be using peer-assisted torrent downloads.
      opener = None

    elif old is None:
      opener = cmf.downloader.FloodOpener(flood_server,
                                          seed_timeout,
                                          torrent_port,
                                          self.cfg.flood_rpc_port,
                                          flood_dir)
    else:
      opener = old.refresh(flood_server,
                           seed_timeout,
                           torrent_port,
                           self.cfg.flood_rpc_port,
                           flood_dir)

    # A new opener object means that flood is being reconfigured in a manner
    # requiring it to be restarted. This is not allowed while there are active
    # downloads, since it blows up the potential test matrix.
    # (This is effectively applies to just changes in flood ports.)
    # All this code below is ugly.... but hopefully it will never execute.
    if self.downloader is not None and self.downloader.is_active() and \
            opener != self.downloader.opener:
      #pylint: disable=redefined-variable-type
      LOG.warn("Ignoring Flood reconfiguration attempt due to active downloader")
      opener = self.downloader.opener

    # Note that downloader can't become active from this point on, because
    # the call to handle_heartbeat_flood() occurs before and in sequence with
    # the logic that triggers parcel downloads in HB processing.

    if opener is None:
      # We don't stop flood - we jut don't use it.
      self.flood_opener = None
    elif opener is not old:
      LOG.info("Flood daemon (re)start attempt")
      try:
        self.supervisor_client.stop_process(FLOOD_DAEMON_PROC)
      except:
        LOG.exception("Failed stopping flood")

      self.mkabsdir(flood_dir, user=FLOOD_FS_USER, group=FLOOD_FS_GROUP, mode=0755)

      try:
        logfile = None
        if self.cfg.logfile != "-":
          logfile = os.path.join(self.cfg.log_dir, FLOOD_LOG_FILE_NAME)
        opener.write_flood_config(self.cfg.paths.flood_dir, logfile,
                                  self.cfg.flood_expiration_time_sec)

        # Fault exception can only come from the line below.
        self.supervisor_client.start_process(FLOOD_DAEMON_PROC)

        self.flood_opener = opener

      except Exception, e:
        LOG.error("Failed writing flood config: %s", e)
        opener = self.downloader.opener  # Don't bother making changes to downloader.

    if self.flood_opener:
      self.flood_opener.set_seed_timeout(seed_timeout)
      if rack_peers:
        self.flood_opener.set_rackpeers(rack_peers)


    if self.flood_opener is None:
      opener = cmf.downloader.URLopener()
    else:
      opener = self.flood_opener
    self.downloader.update_opener(opener)

  def mkworkdir(self, *args, **kwargs):
    """
    Creates a directory that's a subdirectory of the agent
    directory.

    Args are paths to be concatenated with the
    agent working directory.

    The only valid keyword args are user and group.
    """
    assert set(["user", "group"]).issuperset(kwargs.keys())
    user = kwargs.get("user")
    group = kwargs.get("group")

    d = os.path.join(self.dir, *args)
    created = self.mkabsdir(d, user=user, group=group, mode=0751)
    if not created:
      LOG.info("Re-using pre-existing directory: %s", d)
    return d

  def mkabsdir(self, path, user, group, mode=None):
    """
    Creates a directory; path is absolute.

    Returns true if a directory was created.
    """

    # Note that makedirs is hierarchical, but we only change permissions
    # on the leaf directory.
    try:
      os.makedirs(path)
    except OSError as e:
      if e.errno != errno.EEXIST:
        raise
      stat_info = os.stat(path)
      # Log warnings if user/group/mode are different than what's expected
      if self.honor_users_and_groups:
        path_user = pwd.getpwuid(stat_info.st_uid)[0]
        if user is not None and user != path_user:
          THROTTLED_LOG.warning('Expected user %s for %s but was %s', user, path, path_user)
        path_grp = grp.getgrgid(stat_info.st_gid)[0]
        if group is not None and group != path_grp:
          THROTTLED_LOG.warning('Expected group %s for %s but was %s', group, path, path_grp)
      path_mode = oct(stat.S_IMODE(stat_info.st_mode))
      if mode is not None and oct(mode) != path_mode:
        THROTTLED_LOG.warning('Expected mode %s for %s but was %s', oct(mode), path, path_mode)
      return False

    LOG.info("Created %s", path)
    if self.honor_users_and_groups and user is not None and group is not None:
      uid, gid = user_group_to_ids(user, group)
      LOG.info("Chowning %s to %s (%s) %s (%s)", path, user, uid, group, gid)
      os.chown(path, uid, gid)
    if mode is not None:
      LOG.info("Chmod'ing %s to 0%o", path, mode)
      try:
        os.chmod(path, mode)
      except:
        LOG.warning("Chmod failed (%s)", path)
    return True

  def handle_existing_process(self, full_id, existing_process):

    # Our event listener/flood don't conform to the CM process naming
    # convention. They're also not processes monitored by CM itself,
    # so it's best just to ignore them.
    if full_id in [SUPERVISOR_LISTENER_PROC, FLOOD_DAEMON_PROC, STATUS_SERVER_PROC]:
      return

    LOG.info("Pre-existing process: %s", full_id)
    ReconstructedProcess(full_id, existing_process, self.cfg, self)


  def find_or_start_supervisor(self, enable_supervisor_start=False):
    # Because I am wary of multi-threading issues,
    # we spin off a subprocess with supervisor.

    # Explicitly create the working directories to ensure they
    # exist and have sufficient permissions to be usable.

    # Check for an already running supervisor. We may have deliberately left
    # it running to avoid shutting down processes, or the agent may have
    # crashed. In either case, we want to reconnect and reuse it.
    try:
      # We need the running supervisor's id, and this call has the nice
      # side-effect that if it fails, we don't have a reusable supervisor.
      self.connect_to_supervisor()
      LOG.info("Connecting to previous supervisor: %s.", self.supervisor_client.identifier)

      # Restore the CM version in which the previous supervisor was started.
      # For backward compatibility in cases where the supervisor marker file
      # doesn't exist for a running supervisor.
      self.supervisor_info = self.supervisor_client.load_supervisord_marker_file(False)

      # Find the existing processes and re-construct the bare minimum Agent
      # state for them. This allows us to be aware of them and avoid treating
      # them as brand-new processes when the first heatbeat response arrives.
      #
      # Note that these 'fake' process objects are not complete enough to
      # use for a heartbeat.
      existing_processes = None
      while existing_processes is None:
        existing_processes = self.get_supervisor_data(True)

      ReconstructedProcess.rebuild_process_set(self.cfg, existing_processes, self)

      if enable_supervisor_start:
        self.register_supervisor_for_shutdown()
      # We've found the previous supervisor, so we're done here.
      return
    except:
      LOG.exception("Failed to connect to previous supervisor.")
      if not enable_supervisor_start:
        raise

    SupervisorWrapper.start_daemon(self.cfg, self.identifier, self.os_ops)
    try:
      self.connect_to_supervisor()
      LOG.info("Connecting to newly launched supervisor: %s.", self.supervisor_client.identifier)
      #
      # Rebuild process info for any CDH process(s) left running during
      # an upgrade of agent
      #
      # On systems with systemd (like el7), supervisord exists
      # as a separate service, independent of cloudera-agent and
      # is specified as required startup dependency for agent. On
      # non-systemd systems (like el6), cloudera-agent spawns
      # supervisord during it's own startup.
      #
      # In either case, the agent should fetch information about
      # existing CDH processes. The first try block above works
      # for the systemd scenario. This second try block works
      # for the non-systemd scenario.
      #
      # The scenarios to support are:
      # 1) Upgrade from CM5x to CM 6x on non-systemd -> Does not
      #    need to use code path below as on proc_monitor(s) are
      #    used during the upgrade from 5x to 6x cloudera-agent.
      # 2) Upgrade from CM5x to CM6x on systemd -> Uses above code
      #    to recover the existing process information.
      # 3) Upgrade from CM6x to CM6x on non-systemd -> The code
      #    below ensures that existing CDH process information is
      #    recovered.
      # 4) Upgrade from CM6x to CM6x on systemd -> The code above
      #    performs the existing process information recovery.
      LOG.debug("Fetching existing processes from new supervisor")
      existing_processes_post_upgrade = self.get_supervisor_data(False)
      if existing_processes_post_upgrade is not None:
        ReconstructedProcess.rebuild_process_set(self.cfg, existing_processes_post_upgrade, self)

    except SupervisorError:
      LOG.error("Failed to connect to newly launched supervisor. Agent will exit")
      self.stop()
      sys.exit(-1)

    self.register_supervisor_for_shutdown()

    # Generate a marker file to store the CM version in which the supervisor
    # is started.
    self.supervisor_info = self.supervisor_client.load_supervisord_marker_file(True)

  def connect_to_supervisor(self):
    """
    Connect to the newly launched supervisor and try to get its process info
    If we fail to connect to the supervisor the agent exits.
    """
    self.supervisor_client = SupervisorWrapper.connect(self.cfg, self.os_ops)
    self.get_supervisor_process_info()


  def get_supervisor_process_info(self):
    """
    Retrieves the supervisor ID and PID. Throws an exception if it fails to
    connect to the supervisor.
    """
    LOG.info("Supervisor version: %s, pid: %s",
             self.supervisor_client.version,
             self.supervisor_client.pid)

  def register_supervisor_for_shutdown(self):
    """
    Register the supervisor for shutdown using kill_supervisor which only kills
    it if it is not managing any processes.
    """
    # Helper for killing. Popen.kill(), alas, doesn't exist until 2.6
    def kill_supervisor(kill_function):
      num_running = len(Process.GetProcesses(lambda proc: not proc.is_failed() and proc.activated))
      if num_running > 0:
        LOG.info("%u processes are being managed; Supervisor will continue to run.", num_running)
        return
      kill_function()

    self.register_for_shutdown(lambda: kill_supervisor(self.supervisor_client.kill))

  def eval_config(self):
    # We get the real path (which eliminates any symlinks in the path) so that
    # later when we serve files from the agent directories we can easily
    # enforce that no symlinks are followed when restricting serving of files
    # to the process directory.
    self.dir = os.path.realpath(self.cfg.agent_dir)
    #if args.clear_agent_dir:
    #  LOG.info("Clearing agent dir %s", self.dir)
    #  shutil.rmtree(self.dir, ignore_errors=True) # may contain a tmpfs mount


    if not os.path.exists(self.cfg.supervisord_path):
      raise Exception("Supervisord path does not exist: %s", self.cfg.supervisord_path)
    if not os.path.exists(self.cfg.package_dir):
      raise Exception("Package directory does not exist: %s", self.cfg.package_dir)

    self.parcel_dir_from_config = self.cfg.parcel_dir is not None

    self.supervisee_environment = build_supervisee_environment(self.cfg)

  def register_for_shutdown(self, f):
    """
    Calls provided callback when agent is being stopped."
    """
    self.shutdown_registry.append(f)

  def stop(self):
    if not self.started:
      return
    LOG.info("Stopping agent...")
    for f in self.shutdown_registry:
      try:
        f()
      except:
        LOG.exception("Shutdown callback failed.")
    self.started = False

  def restart_signal(self, sig_num, stack_frame):
    '''
    Re-exec the agent.
    '''
    # pylint: disable=unused-argument
    LOG.info("Restarting agent.")
    os.execv(self.exec_args[0], self.exec_args)

  def ready_exit_on_signal(self, sig_num, stack_frame):
    """Marks loop to exit."""
    # pylint: disable=unused-argument
    self.signal_caught = sig_num
    self.should_exit = True

  def sigabrt_handler(self, sig_num, stack_frame):
    """Set main thread to dump heap"""
    # pylint: disable=unused-argument
    self.dump_heap_flag = True

  def register_sigabrt_handler(self):
    signal.signal(signal.SIGABRT, self.sigabrt_handler)

  def dumpheap(self):
    """
    Dump heap to the LOG.
    """
    h = hpy()
    heap_list = h.heap()
    LOG.info("Top heap consumers:\n" + str(heap_list))
    for i in xrange(min(3, len(heap_list))):
      by_type = str(heap_list[i].bytype)
      by_referrer = str(heap_list[i].byrcs)
      LOG.info(by_type+ "\n" + by_referrer)
    self.dump_heap_flag = False


  # Check the staleness of the given process config directory.
  # If marker doesn't exist, return False. Otherwise, compare its timestamp
  # with the staleness threshold to make the decision.
  def check_process_config_dir_staleness(self, now, directory):
    marker = os.path.join(directory, "process_timestamp")
    if not os.path.isfile(marker):
      return False
    else:
      f = open(marker, 'r')
      staleness = now - float(f.readline())
      f.close()
      return staleness > self.cfg.orphan_process_directory_staleness_threshold

  # Remove the obsolete orphan config directories.
  # Among the list of subdirectories under the config directory, we need to
  # remove the ones which are older than
  # orphan_process_directory_staleness_threshold seconds by checking the
  # marker file timestamp in each directory if exists.
  def remove_obsolete_orphan_conf_dirs(self, now):
    try:
      dirs = [d for d in os.listdir(self.cfg.paths.process_dir)\
              if os.path.isdir(\
                os.path.join(self.cfg.paths.process_dir, d)\
              )]
      for dir_name in dirs:
        dir_path = os.path.join(self.cfg.paths.process_dir, dir_name)
        if self.check_process_config_dir_staleness(now, dir_path):
          try:
            LOG.info("Removing the orphan process config dir %s", dir_path)
            shutil.rmtree(dir_path)
          except:
            LOG.exception("Failed to remove the orphan process config dir %s",
                          dir_path)
    except:
      LOG.exception("Caught exception in refreshing config directories")

  def _read_cached_heartbeat(self):
    """
    This function tries to read the cached heartbeat and update the MonitorDaemon
    with the cached host config. This is needed as there are values in the host
    config that affect monitors creation and we need the information before
    reading in the supervisord data and instantiating monitors for managed
    processes already running on the host.
    """
    try:
      response_path = os.path.join(self.cfg.lib_dir, RESPONSE_CACHE_FILE)
      if os.path.exists(response_path):
        avro_reader = avro.datafile.DataFileReader(open(response_path),
                                                   avro.io.DatumReader())
        response = avro_reader.next()
        avro_reader.close()
        response = response.get("data", None)
        if response is None:
          return
        host_config_data = response.get("host_collection_config_data", None)
        if host_config_data is None:
          return
        host_config = dict()
        for c in host_config_data:
          if c.has_key('config_name') and c.has_key('config_value'):
            host_config[c['config_name']] = c['config_value']
        self._update_safety_valve(host_config)
        if host_config.get('diag_bundle_redaction_policy', None) is not None:
          REDACTOR.set_policy_from_string(host_config['diag_bundle_redaction_policy'])
    except:
      LOG.exception("Failed to read saved heartbeat response.")

  def _update_safety_valve(self, host_config):
    """
    Updates the cached safety valve based on the supplied host configuration,
    and performs other actions needed when the safety valve is updated.
    """
    self.safety_valve.update_safety_valve(host_config)
    self._update_dssd_support_status()


def close_most_fds():
  """
  Closes most file descriptors (i.e. all but stdin, stdout, and stderr).
  They may have been inherited from the parent process and, if they're pipes,
  will prevent the parent process from exiting.

  See the following URL for more details:
  http://code.activestate.com/recipes/278731-creating-a-daemon-the-python-way
  """
  if os.uname()[0] == "Darwin":
    # OS X Snow Leopard throws Illegal Instruction here.
    # It's the same bug probably as referred to by
    # http://lists.supervisord.org/pipermail/supervisor-users/2010-January/000525.html
    # Note: we don't really support Darwin as a target platform,
    # but sometimes it's handy for development.
    return

  maxfd = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
  if maxfd == resource.RLIM_INFINITY:
    maxfd = 1024

  # Close the fds. We'll get an OSError if a particular fd was already closed.
  for fd in range(3, maxfd):
    try:
      os.close(fd)
    except OSError:
      pass


def generate_random_bytes(num):
  return "".join(chr(random.randint(0, 255)) for _ in xrange(num))


def get_single_proc_pid(process_name):
  """
  Return pid of the first matching process_name.
  :param process_name:
  :return: pid of the process_name, None if not found
  """

  for proc in psutil.process_iter():
    c = proc.cmdline()
    if len(c) > 2 and os.path.normpath(c[1]) == os.path.normpath(sys.argv[0]) and \
      c[2] == process_name:
      return proc.pid

  return None
