# Copyright (c) 2013-2015 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import
import logging
import os
import StringIO
import ConfigParser

from cmf.audit.navigator_conf import (AuditConfig, MetadataConfig, ProfileConfig, DEFAULT_AUDIT_BATCH_TIMEOUT,
    AUDIT_CONFIG_FILENAME, DEFAULT_METADATA_BATCH_TIMEOUT, METADATA_CONFIG_FILENAME,
    DEFAULT_PROFILE_BATCH_TIMEOUT, PROFILE_CONFIG_FILENAME)
from cmf.audit.navigator_thread import AuditThread, MetadataThread, ProfileThread
from cmf.audit.constants import CONFIG_HEADER
from cmf.throttling_logger import ThrottlingLogger

LOG = logging.getLogger( "Audit" )
THROTTLED_LOG = ThrottlingLogger(LOG)

audit_thread = AuditThread(DEFAULT_AUDIT_BATCH_TIMEOUT)
metadata_thread = MetadataThread(DEFAULT_METADATA_BATCH_TIMEOUT)
profile_thread = ProfileThread(DEFAULT_PROFILE_BATCH_TIMEOUT)

def parse_audit_configuration(conf_dir):
  """
  Parses an audit configuration file and returns an AuditConfig object
  with parsed key and values. If the file is not found (means auditing
  server isn't present), or if the configuration file is empty (means auditing
  isn't enabled), None is returned.
  """
  config = parse_navigator_configuration(conf_dir, AUDIT_CONFIG_FILENAME,
                                         CONFIG_HEADER)
  if config is None:
    return None

  conf = AuditConfig(config)
  THROTTLED_LOG.info("Read auditing config for role %s. Audit server url: %s",
                     conf.role_type, conf.server_url)
  if conf.is_service_configured():
    return conf
  else:
    return None

def parse_metadata_configuration(conf_dir):
  """
  Parses a metadata configuration file and returns a MetadataConfig object
  with parsed key and values. If the file is not found (means auditing
  server isn't present), or if the configuration file is empty (means auditing
  isn't enabled), None is returned.
  """
  config = parse_navigator_configuration(conf_dir, METADATA_CONFIG_FILENAME,
                                         CONFIG_HEADER)
  if config is None:
    return None

  conf = MetadataConfig(config)
  THROTTLED_LOG.info("Read metadata config for role %s. Metadata server url: %s",
                     conf.role_type, conf.server_url)
  if conf.is_service_configured():
    return conf
  else:
    return None

def parse_profile_configuration(conf_dir):
  """
  Parses a profile configuration file and returns a ProfileConfig object
  with parsed key and values. If the file is not found (means auditing
  server isn't present), or if the configuration file is empty (means auditing
  isn't enabled), None is returned.
  """
  config = parse_navigator_configuration(conf_dir, PROFILE_CONFIG_FILENAME,
                                         CONFIG_HEADER)
  if config is None:
    return None

  conf = ProfileConfig(config)
  THROTTLED_LOG.info("Read profile config for role %s. Profile server url: %s",
                     conf.role_type, conf.server_url)
  if conf.is_service_configured():
    return conf
  else:
    return None

def parse_navigator_configuration(conf_dir, config_filename, config_header):
  """
  Parses a configuration file and returns a dictionary with parsed key values.
  If the file is not found (means navigator server isn't present), or if the
  configuration file is empty (means navigator service isn't enabled), None
  is returned.
  """
  conf_file = os.path.join(conf_dir, config_filename)
  if os.path.isfile(conf_file) and os.path.getsize(conf_file) > 0:
    try:
      # convert properties file to add a dummy section so that we can use
      # ConfigParser to parse it
      conf_file_fh = open(conf_file, 'r')
      conf_ini_string = "[{0}]\n{1}".format(config_header, conf_file_fh.read())
      conf_file_fh.close()
      conf_fd = StringIO.StringIO(conf_ini_string)
      config = ConfigParser.SafeConfigParser()
      config.readfp(conf_fd)

      # read all options in a dict so that it can be used by cmf.util.config class
      options = {}
      for key_value in config.items(config_header):
        options[config_header + "." + key_value[0]] = key_value[1]

      return options
    except IOError, ex:
      LOG.exception("Error when accessing %s file.", conf_file)
      return None
  else:
    LOG.debug("Navigator service is not enabled as config file %s is "
              "missing or empty. audit dir: %s", config_filename, conf_dir)
    return None
