# Copyright (c) 2014-2015 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import
import logging

from cmf.monitor import schema
from cmf.util import synchronized
from threading import RLock

LOG = logging.getLogger( __name__ )

class AgentNavigatorMetrics(object):
  """
  This class stores information about the log entries processed by a Navigator
  plugin (note: this class is acting as an abstract base class, so you must use
  a subclass). The information is then conveyed to SMON to make available to
  end users. The following metrics are stored in this class:
  1. Num of cumulative failures when sending log entry to navigator server.
  2. Bytes successfully sent to navigator server.
  3. Number of events successfully sent to navigator server.
  4. Bytes still left to be sent to navigator server.
  """

  def __init__(self):
    """
    Stores navigator metrics from the agent code.
    """
    self._failures = 0L
    self._bytes_sent = 0L
    self._events_sent = 0L
    self._agent_nav_metrics_lock = RLock()

  def get_lock(self):
    """
    Returns the lock that is used to synchronize methods in this class.
    """
    return self._agent_nav_metrics_lock

  @synchronized
  def add_metrics(self, update, nav_pipelines):
    """
    Add current metrics to the "update" object which will then be propagated
    to SMON.
    """
    raise NotImplementedError('This method must be implemented in a subclass.')

  @synchronized
  def increment_send_failures(self):
    """
    Add 1 to the number of failures we have alreday seen.
    """
    self._failures += 1

  @synchronized
  def increment_events_sent(self, num_events):
    """
    Add "num_events" to the number of events we have successfully sent
    for the plugin.
    """
    self._events_sent += num_events

  @synchronized
  def increment_bytes_sent(self, num_bytes):
    """
    Update the total number of bytes we have successfully sent to the
    navigator server.
    """
    self._bytes_sent += num_bytes


class AgentAuditMetrics(AgentNavigatorMetrics):
  @synchronized
  def add_metrics(self, update, nav_pipelines):
    """
    Add current metrics to the "update" object which will then be propagated
    to SMON.
    """
    update.add_metric(schema.AUDIT_AGENT_SEND_FAILURES, self._failures)
    update.add_metric(schema.AUDIT_AGENT_BYTES_SENT, self._bytes_sent)
    update.add_metric(schema.AUDIT_AGENT_EVENTS_SENT, self._events_sent)

    bytes_to_send = 0L
    for pipeline in nav_pipelines:
      log_dir = pipeline.get_nav_logs_dir()
      if log_dir is not None:
        current_nav_log = pipeline.get_current_nav_log()
        bytes_to_send += log_dir.get_bytes_to_send(current_nav_log)

    update.add_metric(schema.AUDIT_AGENT_BYTES_LEFT, bytes_to_send)
