# Copyright (c) 2013 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import
import copy
import time

def parse_impalad_audit_event(service_type, service_name, audit_log_json,
    log_parser_model):
  """
  Method to parse a JSON audit log and create one or multiple audit events
  from it.
  """
  # audit_log_json is a dict with only one entry:
  #    key is timestamp of event, value is the audit event
  timestamp = audit_log_json.keys()[0]
  audit_json = audit_log_json[timestamp]

  # audit event start time is like 2013-08-05 14:10:05.324341000. Convert it
  # to seconds since epoch (python 2.4 limitation, we need to drop fractional
  # second part)
  start_time = audit_json['start_time'].strip().split(".")[0]
  start_time_struct = time.strptime(start_time, "%Y-%m-%d %H:%M:%S")

  # keep a record of catalog objects already seen so we don't generate more than
  # one audit entry for the same catalog object
  processed_objects = set()

  audit_event = dict()
  audit_event['serviceName'] = service_name
  audit_event['username'] = audit_json['user']
  audit_event['impersonator'] = audit_json['impersonator']
  audit_event['ipAddress'], port = _get_ipaddress_port(audit_json['network_address'])
  audit_event['operation'] = audit_json['statement_type']
  # server needs time in milliseconds
  audit_event['eventTime'] = int(time.mktime(start_time_struct) * 1000)
  audit_event['operationText'] = audit_json['sql_statement']
  audit_event['allowed'] = not audit_json['authorization_failure']
  audit_event['queryId'] = audit_json['query_id']
  audit_event['sessionId'] = audit_json['session_id']
  audit_event['status'] = audit_json['status']

  if audit_json['catalog_objects']:
    for catalog_object in audit_json['catalog_objects']:
      object_type = catalog_object['object_type']
      # fully qualified catalog object name is used to deduplicate
      obj_name = catalog_object['name']

      # Don't generate multiple events for the same catalog object
      if obj_name in processed_objects:
        continue
      processed_objects.add(obj_name)

      # CDH-23975 : _impala_builtins is a internal table, skip it
      # Impala will independently fix this bug
      if "_impala_builtins" == obj_name:
        continue

      # split up the catalog object name into <db_name>.<table_name> for the
      # right object types (tables and views)
      if object_type in ["TABLE", "VIEW"]:
        db_name, table_name = obj_name.split(".")
      elif object_type == "DATABASE":
        db_name, table_name = obj_name, None
      else:
        msg = "Expecting table/view/database object type, got %s" % object_type
        raise NotImplementedError(msg)

      audit_event['databaseName'] = db_name
      audit_event['tableName'] = table_name
      audit_event['objectType'] = object_type
      audit_event['privilege'] = catalog_object['privilege']
      yield copy.copy(audit_event)
  else:
    yield copy.copy(audit_event)

def _get_ipaddress_port(network_address):
  """
  Method to parse IP Address provided in Impala log and get IP Address
  and port from it.
  """
  ip_address = None
  port = None
  if network_address is not None:
    ip_address, port = network_address.rsplit(":", 1)

  return ip_address, port
