# Copyright (c) 2013 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging
import os
import sys

from cmf.util import catch_all
from pyinotify import ProcessEvent
from cmf.throttling_logger import ThrottlingLogger

LOG = logging.getLogger( __name__ )
THROTTLED_LOG = ThrottlingLogger(LOG)

class INotifyEventProcessor(ProcessEvent):
  """
  This class extends pyinotify ProcessEvent class to receive filesystem events,
  allowing us to track rollover of navigator log files.
  """

  def my_init(self, nav_plugins):
    """
    ProcessEvent C'tor will call this method with optional parameter. Here is
    the doc verbatim:

    Note: you should not override __init__ in your subclass instead define
    a my_init() method, this method will be called automatically from the
    constructor of this class with its optional parameters.
    """
    self._nav_plugins = nav_plugins
    self._nav_log_dirs = {}
    self._csd_log_dirs = {}

    for plugin in self._nav_plugins:
      log_dir = plugin.nav_conf.log_dir
      if plugin.is_csd_process():
        self._csd_log_dirs.setdefault(log_dir, plugin)
      for pipeline in plugin.get_nav_pipelines():
        pipeline_log_dir = pipeline.get_nav_logs_dir()
        self._nav_log_dirs.setdefault(log_dir, []).append(pipeline_log_dir)

    all_logs = ",".join(self._nav_log_dirs.keys())
    THROTTLED_LOG.info("Setting watch for directories: %s", all_logs)

  def _get_nav_log_dirs(self, log_dir):
    return self._nav_log_dirs[log_dir]

  def _update_nav_logs_dirs(self, event):
    """
    For CSD plugins, the log and wal files are removed only when an end-marker
    is received. One a pipeline is removed, its corresponding nav_logs_dir object
    would continue to live in the self._nav_logs_dirs mapping, hence when we receive
    an event saying a wal file was deleted for a given CSD plugin pipeline, we remove
    the corresponding nav_logs_dir object from the self._nav_log_dirs mapping
    """
    csd_plugin = self._csd_log_dirs.get(event.path)
    if csd_plugin is not None and event.name.endswith("_wal"):
      # Wal filenames are filed as such
      # <ApplicationExeId>_metadata_wal e.g. SampleSparkApp_metadata_wal
      pipeline_name = event.name.split('_metadata_wal')[0]
      log_dirs = self._nav_log_dirs[event.path]
      for log_dir in log_dirs:
        if log_dir.get_pipeline_name() == pipeline_name:
          log_dirs.remove(log_dir)
          break

  @catch_all()
  def process_IN_CREATE(self, event):
    """
    Called when a new file is created in the directory we are monitoring. We
    check that if this is because of log rotation and then start monitoring
    the new file.
    """
    LOG.debug("Got a new file in %s directory: %s" % (event.path, event.name))

    if event.path in self._csd_log_dirs and not event.name.endswith("_wal") and not event.name.endswith("tmp"):
      # For csd processes we want to ignore wal files and tmp files, Note that on checkpointing a temporary
      # file audit_wal.tmp is created.
      csd_plugin = self._csd_log_dirs.get(event.path)
      new_pipeline = csd_plugin.add_csd_pipeline(event)
      if new_pipeline:
        self._nav_log_dirs.setdefault(event.path, []).append(new_pipeline.get_nav_logs_dir())

    for log_dir in self._get_nav_log_dirs(event.path):
      log_dir.append_new_log_to_list(event.name)

  @catch_all()
  def process_IN_DELETE(self, event):
    LOG.debug("%s file was deleted in navigator directory %s." % (event.name, event.path))
    for log_dir in self._get_nav_log_dirs(event.path):
      log_dir.remove_log_from_list(event.name)
    self._update_nav_logs_dirs(event)

  @catch_all()
  def process_IN_MOVED_FROM(self, event):
    LOG.debug("%s was moved out of navigator directory %s." % (event.name, event.path))
    for log_dir in self._get_nav_log_dirs(event.path):
      log_dir.remove_log_from_list(event.name)
    self._update_nav_logs_dirs(event)
