# Copyright (c) 2013-2015 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging

from cmf.throttling_logger import ThrottlingLogger
from cmf.util.config import Config, ConfigItem, ConfigIntItem
from cmf.audit.constants import CONFIG_HEADER

AUDIT_CONFIG_FILENAME = "navigator.client.properties"
DEFAULT_AUDIT_BATCH_TIMEOUT = 5

METADATA_CONFIG_FILENAME = "navigator.lineage.client.properties"
DEFAULT_METADATA_BATCH_TIMEOUT = 5

PROFILE_CONFIG_FILENAME = "telepub.client.properties"
DEFAULT_PROFILE_BATCH_TIMEOUT = 5

LOG = logging.getLogger(__name__)
THROTTLED_LOG = ThrottlingLogger(LOG)

# Config file defaults. Used in this class and in test code.
DEFAULT_MIN_BATCH_SIZE = 10000
DEFAULT_MAX_REQUEST_SIZE = 2 * 1024 * 1024  # 2 MB
DEFAULT_MAX_NUM_LOG = 10
DEFAULT_CONNECTION_TIMEOUT = 120

# Max nnumber of files read/send by a pipeline, to avoid starvation
MAX_READ_PROCESSED_FILES = 10

class NavigatorConfig(Config):
  role_type = ConfigItem(CONFIG_HEADER, name="navigator.client.roleType")
  role_name = ConfigItem(CONFIG_HEADER, name="navigator.client.roleName")
  service_name = ConfigItem(CONFIG_HEADER, name="navigator.client.serviceName")
  service_type = ConfigItem(CONFIG_HEADER, name="navigator.client.serviceType")
  cluster_name = ConfigItem(CONFIG_HEADER, name="navigator.client.clusterName")
  batch_min_size = ConfigIntItem(CONFIG_HEADER, name="navigator.batch.min_size", default=DEFAULT_MIN_BATCH_SIZE)
  max_request_size = ConfigIntItem(CONFIG_HEADER, name="navigator.client.max_request_size", default=DEFAULT_MAX_REQUEST_SIZE)
  max_num_navigator_logs = ConfigIntItem(CONFIG_HEADER, name="navigator.client.max_num_audit_log", default=DEFAULT_MAX_NUM_LOG)
  max_read_processed_files = ConfigIntItem(CONFIG_HEADER, name="navigator.client.max_read_processed_files",
                                           default=MAX_READ_PROCESSED_FILES)
  connection_timeout = ConfigIntItem(CONFIG_HEADER, name="navigator.server.timeout_seconds",
                                     default=DEFAULT_CONNECTION_TIMEOUT)
  tp_url = ConfigItem(CONFIG_HEADER, name="telemetry_publisher.server.url")
  nms_url = ConfigItem(CONFIG_HEADER, name="navigator.server.url")
  #nas_url and nms_url both point to navigator audit server
  nas_url = nms_url
  server_url = ConfigItem(CONFIG_HEADER, name="telemetry_publisher.server.url",
                          default=lambda cfg_item, cfg: cfg.nms_url)

  def is_service_configured(self):
    """
    Check if navigator logging is configured for the role. The conf file should
    contain at least navigator server url and navigator log directory to be
    considered navigator ready.
    """
    return self.server_url and self.log_dir

  def has_telemetry_publisher_url(self):
    """
    Returns a boolean indicating that telemetry publisher url is present in config.
    """
    return self.tp_url is not None

  def has_navigator_url(self):
    """
    Returns a boolean indicating that navigator url is present in config.
    """
    return self.nas_url is not None

def make_config_class(log_dir_param, default_batch_timeout):
  """
  Creates a audit server's or metadata server's config class and
  returns that.
  """
  class DerivedClass(NavigatorConfig):
    log_dir = ConfigItem(CONFIG_HEADER, name=log_dir_param)
    batch_timeout = ConfigIntItem(CONFIG_HEADER, name="navigator.batch.timeout", default=default_batch_timeout)

  return DerivedClass

AuditConfig = make_config_class("audit_event_log_dir", DEFAULT_AUDIT_BATCH_TIMEOUT)
MetadataConfig = make_config_class("lineage_event_log_dir", DEFAULT_METADATA_BATCH_TIMEOUT)
ProfileConfig = make_config_class("profile_event_log_dir", DEFAULT_PROFILE_BATCH_TIMEOUT)
