# Copyright (c) 2013 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import
import bisect
import logging
import os
import sys
import simplejson

from cmf.throttling_logger import ThrottlingLogger
from cmf.util import synchronized
from threading import RLock

LOG = logging.getLogger( __name__ )
THROTTLED_LOG = ThrottlingLogger(LOG)
SEEK_CUR = 1
PROFILE_PREFIX = "impala_profile_log"

def _get_nav_log_timestamp(nav_log_filename):
  """
  Returns the timestamp embedded in the navigator log filename. The format
  of the Impala audit filename is 'impala_audit_event_log_<version>-<timestamp>,
  for Sentry it's 'role_name.<timestamp>' (or 'role_name' for the first log
  file), and for others it is 'role_name-<timestamp>'.
  """
  index = nav_log_filename.rfind('-')
  try:
    return long(nav_log_filename[index+1:])
  except ValueError:
    index = nav_log_filename.rfind('.')
  try:
    return long(nav_log_filename[index+1:])
  except ValueError:
    # If the log filename does not have a timestamp, return 0 so we can sort it
    # before the other log files (for Sentry, the first log filename does not
    # have a timestamp, and all log files created after the first one have a
    # timestamp in the name, so this file must be earlier than the log files
    # that have timestamps).
    return 0

class NavigatorLogNotFoundError(Exception):
  """
  Error raised when a log file name is looked into the list of file names in the
  log directory and its not found in the list. Indicates that the file is
  deleted since we read the file list in the directory.
  """
  def __init__(self, file_name, next_file_name):
    """
    file_name is the file that is not found in the list of log files we are
    tracking.
    next_file_name is the name of file in the list whose timestamp comes
    immediately after the timestamp of file_name
    """
    self._file_name = file_name
    self._next_file_name = next_file_name

  def __str__(self):
    return "File %s not found in list of navigator log files." % (self._file_name, )

  def get_next_file_name(self):
    return self._next_file_name

class NavigatorLogDirectory(object):
  """
  This class maintains list of log files in the order they were created and
  provides helper methods to iterate over them. This is used to make sure that
  no log entries are skipped, even if this process stops running. When this
  code is executed it picks up from the log where processing was last done.
  """
  def __init__(self, role_name, log_dir, log_prefix, is_csd_process=False, pipeline_name=None):
    self._role_name = role_name # used for logging, to disambiguate multiple roles
    self._log_dir = log_dir
    self._log_prefix = log_prefix
    self._is_csd_process = is_csd_process
    self._pipeline_name = pipeline_name
    self._nav_logs = []
    self._nav_logs_timestamp = []
    self._nav_logs_lock = RLock()
    self._get_all_nav_logs()

  def get_lock(self):
    """
    Returns the lock that is used to synchronize methods in this class.
    """
    return self._nav_logs_lock

  def __str__(self):
    return "Navigator Log Directory " + self._log_dir

  def _get_prefix(self):
    prefix = self._log_prefix
    if self._is_csd_process:
      prefix = self._log_prefix + self._pipeline_name
    return prefix

  def get_pipeline_name(self):
    return self._pipeline_name

  @synchronized
  def _get_all_nav_logs(self):
    """
    Reads all log files in the navigator log directory and sorts them
    based on their timestamp. The resulting list is stored as a member of this
    class.
    """
    if not os.path.exists(self._log_dir):
      return

    self._nav_logs = [name for name in os.listdir(self._log_dir)
      if name.startswith(self._get_prefix())]
    self._nav_logs.sort(key=_get_nav_log_timestamp)
    self._nav_logs_timestamp = [_get_nav_log_timestamp(file_name)
                                  for file_name in self._nav_logs]
    THROTTLED_LOG.info("%d navigator log files with prefix %s for role %s exist in "
                       "navigator log directory %s",
                       len(self._nav_logs), self._log_prefix, self._role_name, self._log_dir)

  @synchronized
  def refresh_nav_logs(self):
    """
    Refresh our view of the log directory. This is called when we temporarily
    disable inotify notifications (to update it) and there is chance that a
    rollover occurs at same time.
    """
    self._get_all_nav_logs()

  @synchronized
  def _get_log_index(self, file_name):
    """
    Returns index of the log specified by 'file_name' in the list of navigator
    logs. If the file_name is not present in the list, throws
    NavigatorLogNotFoundError.
    Bisect modules returns index within list, so zero is considered valid index.
    """
    num_logs = len(self._nav_logs_timestamp)
    if num_logs == 0:
      raise NavigatorLogNotFoundError(file_name, None)

    timestamp = _get_nav_log_timestamp(file_name)
    idx = bisect.bisect_left(self._nav_logs_timestamp, timestamp)
    if idx == num_logs:
      raise NavigatorLogNotFoundError(file_name, None)

    # Also, the file should be in our list. If timestamps don't match
    # the file doesn't exist
    if self._nav_logs_timestamp[idx] != timestamp:
      raise NavigatorLogNotFoundError(file_name, self._nav_logs[idx])

    return idx

  @synchronized
  def get_log(self, file_name):
    """
    Returns a NavigatorLogFile object for the log file specified by file_name.
    If file_name is None, then the oldest timestamp file in the directory
    is returned. If there is no log file in the directory, a NoneNavigatorLogFile
    object is returned.
    """
    if file_name is None:
      if self._nav_logs:
        return NavigatorLogFile(self._log_dir, self._nav_logs[0])
      else:
        return NoneNavigatorLogFile()
    else:
      log_path = os.path.join(self._log_dir, file_name)
      if os.path.exists(log_path):
        return NavigatorLogFile(self._log_dir, file_name)
      else:
        return NonexistentNavigatorLogFile(file_name)

  @synchronized
  def get_next_log_file_name(self, prev_file_name):
    """
    Checks if a log file following the prev_file_name specified is present in
    the list. If yes, the name of the file is returned, else None is returned.
    """
    ret = None
    if prev_file_name is None:
      if self._nav_logs:
        ret = self._nav_logs[0]
    elif prev_file_name.startswith(self._get_prefix()):
      try:
        idx = self._get_log_index(prev_file_name) + 1
        if idx < len(self._nav_logs):
          ret = self._nav_logs[idx]
      except NavigatorLogNotFoundError, ex:
        ret = ex.get_next_file_name()

    THROTTLED_LOG.debug("Prev log: %s, next log: %s" % (prev_file_name, ret))
    return ret

  @synchronized
  def append_new_log_to_list(self, file_name):
    """Adds a newly created nav log file to the list of nav log files"""
    # First validate if the filename timestamp follows all existing files,
    # otherwise log a warning and ignore the file
    if file_name.startswith(self._get_prefix()):
      timestamp = _get_nav_log_timestamp(file_name)
      last_timestamp = -1
      if self._nav_logs:
        last_timestamp = _get_nav_log_timestamp(self._nav_logs[-1])

      if timestamp > last_timestamp:
        self._nav_logs.append(file_name)
        self._nav_logs_timestamp.append(timestamp)
      elif timestamp == last_timestamp:
        LOG.info("Log file %s already known. Ignoring notification.", (file_name,))
      else:
        LOG.warning(
          "A new navigator log file (%s) has timestamp older than latest log (%s)." %
          (file_name, self._nav_logs[-1]))

  @synchronized
  def remove_log_from_list(self, file_name):
    """
    Updates the listing of the log directory by removing the file that got
    deleted or moved out of the log directory.
    """
    if file_name.startswith(self._get_prefix()):
      try:
        idx = self._get_log_index(file_name)
        del self._nav_logs[idx]
        del self._nav_logs_timestamp[idx]
      except NavigatorLogNotFoundError, ex:
        # Not able to find the file in our list of logs. Log a message
        LOG.info("Untracked %s navigator log file is removed from filesystem." %
                 (file_name, ))

  @synchronized
  def get_number_of_purge_files(self, current_filename, max_num_logs):
    """
    Function that returns ending idx of files that,
    can be purged.
    If return idx is less than 0, then current file idx is less than
    max number of logs, in this case caller should ignore return end idx
    """
    current_file_idx = self._get_log_index(current_filename)
    end_idx = current_file_idx - max_num_logs
    return end_idx

  @synchronized
  def purge_files_by_idx(self, purge_idx):
    """
    Function that pruges files upto specified purge_idx value.
    """
    if purge_idx < 0:
      #ideally this should not happen
      THROTTLED_LOG.warning("Got negative purge index value: %s" % (purge_idx, ))
      return

    idx = 0
    while idx <= purge_idx:
      log_file_name = self._nav_logs[idx]
      log_path = os.path.join(self._log_dir, log_file_name)
      if os.path.exists(log_path):
        os.remove(log_path) # inotify will call remove_log_from_list
      else:
        LOG.warning("Log %s exists in list but not on filesystem." % (log_path, ))
        self.remove_log_from_list(log_file_name)
      idx = idx + 1

  @synchronized
  def purge_old_nav_logs(self, current_filename, max_num_logs):
    """
    Checks if we have more than max_num_logs before current log, if yes then
    get rid of the older logs.
    """
    current_file_idx = self._get_log_index(current_filename)
    self.purge_files_by_idx(current_file_idx - max_num_logs)

  @synchronized
  def purge_all_logs(self):
    """
    Removes all log files for a given pipeline
    """
    for log_file_name in self._nav_logs:
      log_path = os.path.join(self._log_dir, log_file_name)
      if os.path.exists(log_path):
        os.remove(log_path)
      else:
        LOG.warning("Log %s exists in list but not on filesystem." % (log_path, ))
        self.remove_log_from_list(log_file_name)

  @synchronized
  def get_bytes_to_send(self, first_navigator_log):
    """
    Returns number of bytes in the log files that need to be transferred to the
    navigator server. First parameter is the current log that is being
    processed.
    """
    already_processed = first_navigator_log.tell()
    bytes_count = first_navigator_log.size() - already_processed

    next_nav_log_name = first_navigator_log.get_file_name()
    while True:
      next_nav_log_name = self.get_next_log_file_name(next_nav_log_name)
      if next_nav_log_name is None:
        break

      full_path = os.path.join(self._log_dir, next_nav_log_name)
      bytes_count += os.stat(full_path).st_size

    return bytes_count

class NavigatorLogFile(object):
  """
  This class wraps up a navigator log file and provides methods to seek/read
  from this file.
  """
  def __init__(self, log_dir, file_name):
    self._file_name = file_name
    self._fd = open(os.path.join(log_dir, file_name))

  def get_file_name(self):
    return self._file_name

  def close(self):
    self._fd.close()

  def exists(self):
    return True

  def seek(self, offset):
    """
    Perform an absolute seek based on the offset provided. os.SEEK_SET is the
    default optional argument to seek function.
    The function is a no-op if the offset is None.
    """
    if offset is not None:
      self._fd.seek(offset)

  def tell(self):
    """
    Returns the current offset (file position) in the file.
    """
    return self._fd.tell()

  def size(self):
    """
    Return the size of file in bytes.
    """
    return os.fstat(self._fd.fileno()).st_size

  def _clear_eof_log_file(self):
    """
    Clears the eof flag by seeking to current location.
    """
    # seek offset of zero from current location to clear the EOF flag
    self._fd.seek(0, SEEK_CUR)

  def read_nav_log(self):
    """
    Called when there is content to be read from the nav log file. The method
    reads new lines in the file and returns a tuple (filename, offset, json dict).
    The filename is the name of log file and offset is the offset of the
    start of next line. If the log entry can be sent successfully, then this offset
    is recorded in the WAL file.
    """
    # We will use while loop to read the lines, the better way to do this:
    #   for line in self._fd
    # doesn't fit our case as to make it efficient it has a read-ahead buffer,
    # which means that the "tell" method returns offset that is ahead of the
    # log entries we are getting.
    while True:
      line = self._fd.readline()
      if line:
        if line[-1] == "\n":
          try:
            if self._file_name.startswith(PROFILE_PREFIX):
                log_entry = line
            else:
                log_entry = simplejson.loads(line)
            yield self._file_name, self._fd.tell(), len(line), log_entry
          except simplejson.JSONDecodeError:
            # A \n terminated navigator log line that cannot be parsed is corrupt
            LOG.error("Stopping processing of %s because of corrupt navigator "
                      "log line: %s" % (self._file_name, line))
            self._fd.seek(-len(line), SEEK_CUR)
            return
        else:
          # We might read half a line as the role didn't flush after each write
          # Seek back to the start of the line (can also use distutils.text_file
          # for all this if needed, it exposes unreadline method)
          self._fd.seek(-len(line), SEEK_CUR)
          break
      else:
        break # eof

    # At the end clear EOF file so that we can read again
    self._clear_eof_log_file()

class NonexistentNavigatorLogFile(object):
  """
  This class represents a log file that used to exist but is no longer present.
  We can run into this case when WAL file contains the name of the log file,
  but the file is deleted. Alternatively, the log directory may have been
  changed and wal file got copied but the logs weren't.
  """
  def __init__(self, file_name):
    self._file_name = file_name;

  def close(self):
    pass

  def seek(self, offset):
    pass

  def tell(self):
    return 0

  def size(self):
    return 0

  def get_file_name(self):
    return self._file_name

  def read_nav_log(self):
    pass

  def exists(self):
    return False

class NoneNavigatorLogFile(object):
  """
  An empty class to act as placeholder for the case where there is no log file.
  This occurs during first start of the role when there is no entry in
  the log file.
  """
  def close(self):
    pass

  def seek(self, offset):
    pass

  def tell(self):
    return 0

  def size(self):
    return 0

  def get_file_name(self):
    return None

  def exists(self):
    return False

  def read_nav_log(self):
    pass
