# Copyright (c) 2013-2015 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import
import errno
import logging
from threading import RLock

import os
import simplejson as json

import cmf.audit
from cmf.audit.agent_navigator_metrics import AgentAuditMetrics
from cmf.audit.navigator_plugin_pipeline import AuditPluginPipeline, MetadataPluginPipeline, ProfilePluginPipeline
from cmf.audit.navigator_sender import AuditSender, MetadataSender, ProfileSender
from cmf.audit.generic_profile_parser import parse_generic_profile_event
from cmf.monitor import schema
from cmf.throttling_logger import ThrottlingLogger
from cmf.util import catch_all
from cmf.util import synchronized

LOG = logging.getLogger( __name__ )
THROTTLED_LOG = ThrottlingLogger(LOG)

# max number of pipeline names printed in any log message
MAX_DEBUG_PIPELINES = 6

class NavigatorPlugin(object):
  """
  Each instance of this plugin class will store information for each role (with
  audit/metadata logging) that we are tracking. This class contains information
  about the roles configuration including its navigator log directory etc.
  """

  def __init__(self, log_entry_parser, ipc_request_name, pipelines, nav_sender,
      process_name, is_csd_process=False, log_prefix=None, config_dictionary=None):
    """
    Stores configuration information about the role we are polling for log data.
    """
    self.nav_conf = None
    self._metadata_conf = None
    self._log_prefix = log_prefix
    self._pipelines = pipelines
    self._log_entry_parser = log_entry_parser
    #flag that governs if a plugin should delete file or not
    self._isPurgeAble = True
    # mark plugin as profile plugin or not
    if log_entry_parser == parse_generic_profile_event:
      self._is_profile = True
      #in case of impala profiles, we let impala handle purging
      self._isPurgeAble = False
    else:
      self._is_profile = False
    self._nav_sender = nav_sender
    self._ipc_request_name = ipc_request_name
    self._last_metrics_timestamp = 0
    self._log_parser_model = None
    self._refresh_plugin_args = None
    self._refresh_plugin_args_lock = RLock()
    self._pipelines_lock = RLock()
    self._stop_nav_monitoring = False
    self._process_name = process_name
    self._is_csd_process = is_csd_process
    self._config_dictionary = config_dictionary

  def get_lock(self):
    """
    Returns the lock that is used to synchronize methods in this class.
    """
    return self._pipelines_lock

  def get_process_name(self):
    return self._process_name

  def is_profile(self):
    """
    Check if plugin is for profile
    """
    return self._is_profile

  def isPurgeAble(self):
    """
    Check if plugin is allowed to purge files
    """
    return self._isPurgeAble

  def _safe_debug_string(self, plugin_type):
    """
    function that safe debug string, that goes into log message
    """
    # print only MAX_DEBUG_PIPELINES pipelines names
    pipelines = [pipeline.get_name() for count, pipeline in enumerate(self._pipelines)
                 if count <= MAX_DEBUG_PIPELINES]
    if len(self._pipelines) > MAX_DEBUG_PIPELINES:
      # add elipses to record there are more pipelines name not captured
      pipelines.append("..")
    return ("%s Plugin for %s with count %s pipelines names %s."
            % (plugin_type, self._process_name,
            len(self._pipelines), pipelines))

  def __str__(self):
    return self._safe_debug_string("Navigator")

  @catch_all()
  def stop_nav_monitoring(self):
    """
    Set a boolean to indicate that this plugin need to be stopped. Called
    by Process.py when process associated with this role is "deactivated".
    """
    self._stop_nav_monitoring = True

  @catch_all()
  @synchronized
  def immediate_stop(self):
    """
    This method will stop the plugin and its pipeline to stop monitoring this
    role's navigator log.
    """
    LOG.info("stopping %s" % (self,))
    self.nav_conf = None
    self._metadata_conf = None
    for pipeline in self._pipelines:
      pipeline.stop_nav_monitoring()

    # ensure all the old pipelines are reset
    self._pipelines = []
    # The plugin may get reused. Reset the flag that the plugin was stopped
    self._stop_nav_monitoring = False

  def is_stop_called(self):
    """
    Returns a boolean indicating if stop_nav_monitoring was called for this
    plugin.
    """
    return self._stop_nav_monitoring

  @catch_all()
  def _refresh(self,
      conf_dir,
      config_parser,
      refresh_thread = None):
    """
    Saves the request to a list that is then processed by navigator threads by
    calling "immediate_refresh" method. "refresh_called" method can be called
    to figure out if there is a pending audit/metadata plugin "refresh".
    """
    self._refresh_plugin_args_lock.acquire()
    try:
      THROTTLED_LOG.info("Scheduling a refresh for %s", self)
      self._refresh_plugin_args = (conf_dir, config_parser, refresh_thread)

      if refresh_thread is None:
        self._get_thread().add_navigator_plugin(self)
    finally:
      self._refresh_plugin_args_lock.release()

  def _recreate_pipelines_for_csd(self):
    if self._is_csd_process:
      existing_logs = [name for name in os.listdir(self.nav_conf.log_dir)
                     if name.startswith(self._log_prefix)]

      # For CSD processes, currently of which we only have spark, it is assumed
      # the format of the log files would follow the following naming convention:
      # <log_prefix>-<unique_yarn_application_id>-<timestamp-in-milliseconds>
      # For Spark it should look like this spark_lineage_log_Applicationid-Timestamp
      # e.g. spark_lineage_log_mySparkApp-1478759200610

      for existing_log in existing_logs:
        pipeline_name = self.extract_csd_pipeline_name(existing_log)
        self.create_new_lineage_csd_pipeline(pipeline_name)

  def update_pipelines_based_on_config(self):
    """
    This method should be implemented by all child classes.
    """
    raise NotImplementedError('This method should be implemented by all child classes')

  @catch_all()
  @synchronized
  def immediate_refresh(self):
    """
    Parses the nav configuration file to get the location of navigator server
    and navigator log file. If nav logging is enabled it notifies relevant
    thread that then reads from the nav log and passes log entries to the
    navigator server.
    """
    self._refresh_plugin_args_lock.acquire()
    try:
      if self._refresh_plugin_args is None:
        return

      conf_dir, config_parser, refresh_thread = self._refresh_plugin_args
      self._refresh_plugin_args = None
    finally:
      self._refresh_plugin_args_lock.release()

    try:
      nav_was_previously_enabled = self.is_nav_enabled()
      self.nav_conf = config_parser(conf_dir)
      self._recreate_pipelines_for_csd()

      #check for latest pipelines based on current config
      self.update_pipelines_based_on_config()

      THROTTLED_LOG.info("Refreshing %s", self)
      for plugin_pipeline in self._pipelines:
        plugin_pipeline.stop_nav_monitoring()

      if self.is_nav_enabled():
        for plugin_pipeline in self._pipelines:
          plugin_pipeline.start_nav_monitoring(self.nav_conf)

      if self.is_nav_enabled() or nav_was_previously_enabled:
        if refresh_thread is None:
          self._get_thread().refresh_nav_plugin(self)
        else:
          refresh_thread(self)
    except:
      LOG.exception("Exception caught when trying to refresh %s" % (self,))

  def is_csd_process(self):
    return self._is_csd_process

  def get_api_url(self, cluster_name, service_name, pipeline=None):
    """
    Returns api url endpoint that should be used to communicate when sending
    lineage information.
    """
    raise NotImplementedError('This method only applies to lineage generating processes.')

  def get_log_prefix(self):
    return self._log_prefix

  def is_nav_enabled(self):
    return self.nav_conf is not None

  def get_nav_sender(self):
    """
    This method should be implemented by all child classes.
    """
    raise NotImplementedError('This method should be implemented by all child classes')

  @synchronized
  def get_nav_pipelines(self):
    """
    Return list of pipelines for this plugin.
    """
    return self._pipelines

  def get_role_type(self):
    """
    Returns the role type for this plugin.
    """
    return self.nav_conf.role_type

  def get_role_name(self):
    """
    Return the name of the role to which this plugin is associated with.
    """
    return self.nav_conf.role_name

  def get_service_name(self):
    """
    Return the name of the service associated with this plugin.
    """
    return self.nav_conf.service_name

  def get_service_type(self):
    """
    Return the service type associated with this plugin.
    """
    try:
      return self.nav_conf.service_type
    except:
      return None

  def get_cluster_name(self):
    """
    Return the name of the service associated with this plugin.
    """
    return self.nav_conf.cluster_name

  def get_nav_connection_timeout(self):
    """
    Return the name of the service associated with this plugin.
    """
    return self.nav_conf.connection_timeout

  def get_log_entry_parser(self):
    """
    Returns the parser that can be used to parse the log entry data
    from the navigator log.
    """
    return self._log_entry_parser

  def get_avro_request(self, entries):
    """
    Returns the request map to send to the navigator server. This corresponds
    to an avro record.
    """
    raise NotImplementedError('This method must be implemented in a subclass.')

  def get_ipc_request_name(self):
    """
    Return the name of the IPC method that can be used to send log entries
    to the navigator server. This name is defined in navigator.avdl.
    """
    return self._ipc_request_name

  def get_agent_nav_metrics(self):
    """
    Returns the object that can be used to collect audit/metadata metrics for
    the code in the CM agent.
    """
    return self._agent_nav_metrics

  def get_log_parser_model(self):
    """
    Returns the fields with any special information, such as default value
    or renaming.
    """
    if self._log_parser_model is not None:
      return self._log_parser_model

    # Get the logParserModel file path for this service
    json_path = os.path.join(os.path.dirname(os.path.realpath(__file__)),
        "log_parser_models", self.get_service_type() + "LogParserModel.json")
    try :
      json_file = open(json_path, 'rU')
      json_data = json.load(json_file)

      LOG.info("Using log parser model %s" % (json_path,))
      self._log_parser_model = json_data
    except Exception, ex:
      if ex.errno != errno.ENOENT:
        LOG.exception("Error opening log parser model %s" % (json_path,))
      self._log_parser_model = {}

    return self._log_parser_model

  def extract_csd_pipeline_name(self, filename):
    # From the event.path determine the pipeline name
    # spark_lineage_log_YarnApplicationID-Timestamp.log
    name_and_timestamp = filename.split("spark_lineage_log_")[1]
    last_dash_loc = name_and_timestamp.rfind('-')
    name = name_and_timestamp[:last_dash_loc]
    return name

  @synchronized
  def add_csd_pipeline(self, event):
    """
    This function is specific to CSD plugins. Whenever, a new lineage log file gets
    written to the log directory, an inotify event is generated, which when processed
    creates a pipeline for the corresponding lineage application run, if the pipeline
    already doesn't exist.
    """
    assert self._is_csd_process is True

    # From the event.path determine the pipeline name
    # spark_lineage_log_YarnApplicationID-Timestamp.log
    pipeline_name = self.extract_csd_pipeline_name(event.name)
    create_new_pipeline = True
    new_pipeline = None

    for pipeline in self._pipelines:
      if pipeline.get_name() == pipeline_name:
        create_new_pipeline = False
        break

    if create_new_pipeline:
      new_pipeline = self.create_new_lineage_csd_pipeline(pipeline_name)
      new_pipeline.start_nav_monitoring(self.nav_conf, pipeline_name)

    return new_pipeline

  @synchronized
  def create_new_lineage_csd_pipeline(self, pipeline_name):
    assert self._is_csd_process is True

    # don't add if the pipeline already exists
    absent = True
    for pipeline in self._pipelines:
      if pipeline.get_name() == pipeline_name:
        absent = False
        new_pipeline = pipeline
        break

    if absent:
      new_pipeline = MetadataPluginPipeline(pipeline_name, self._log_prefix,
                                            self._process_name, True, self.nav_conf.nms_url)
      self._pipelines.append(new_pipeline)

    return new_pipeline

  @synchronized
  def remove_csd_pipeline(self, pipeline):
    assert self._is_csd_process is True
    self._pipelines.remove(pipeline)
    pipeline.stop_nav_monitoring()

  @catch_all()
  @synchronized
  def add_metrics(self, update):
    """
    Adds audit/metadata metrics for navigator plugins and for the log entries
    processed by the agent. Navigator plugins write a "metrics" file in the
    audit/metadata log directory. We read the file and get metrics from there.
    The code in CM Agent (i.e. this code) also maintains its metrics, which we
    also add to the "update" object.
    """
    self._add_nav_plugin_metrics(update)

    if self._agent_nav_metrics:
      self._agent_nav_metrics.add_metrics(update, self._pipelines)

  def _add_nav_plugin_metrics(self, update):
    """
    Read the "metrics" file from audit/metadata log directory and return metrics
    from that file.
    """
    raise NotImplementedError('This method must be implemented in a subclass.')

  def _add_double_metric(self, update, metrics_dict, metric):
    """
    Adds a single metrics to the "update" object. Logs a warning and
    does nothing if any error occurs (metric not found, value not
    correct etc.).
    """
    try:
      metric_id = schema.__dict__[metric.upper()]
      update.add_metric(metric_id, float(metrics_dict[metric]))
    except Exception, ex:
      THROTTLED_LOG.exception("Unknown exception occurred.")


class AuditPlugin(NavigatorPlugin):

  def __init__(self, log_entry_parser, ipc_request_name, pipeline_names,
      log_prefix=None, process_name=None, config_dictionary=None):
    """
    Check if TeleMetry publisher was enabled, update pipeline names to
    create new pipeline.
    """
    nav_conf = None
    new_pipeline_names = []
    if config_dictionary and 'conf_dir' in config_dictionary:
      config_parser = cmf.audit.parse_audit_configuration
      nav_conf = config_parser(config_dictionary['conf_dir'])
    if nav_conf and nav_conf.has_telemetry_publisher_url():
      for pipeline_name in pipeline_names:
        new_pipeline_names.append(pipeline_name + "TP")

    if nav_conf and not nav_conf.has_navigator_url():
      pipeline_names = []
    pipeline_names = pipeline_names + new_pipeline_names
    audit_pipelines = [
        AuditPluginPipeline(pipeline_name, log_prefix, process_name)
        for pipeline_name in pipeline_names]
    super(AuditPlugin, self).__init__(log_entry_parser, ipc_request_name,
        audit_pipelines, AuditSender, process_name, False, log_prefix)

    self._config_dictionary = config_dictionary
    self._agent_nav_metrics = AgentAuditMetrics()

  def __str__(self):
    return self._safe_debug_string("Audit")

  def get_avro_request(self, entries):
    """
    Returns the request map to send to the navigator server. This corresponds
    to the WAERequest record.
    """
    return {
        'serviceName': self.get_service_name(),
        'events': entries
    }

  @catch_all()
  def refresh(self,
      conf_dir,
      config_parser = cmf.audit.parse_audit_configuration,
      refresh_thread = None):
    """
    Wrapper that provides a default config_parser value for the _refresh method.
    """
    self._refresh(conf_dir, config_parser, refresh_thread)

  def _get_thread(self):
    return cmf.audit.audit_thread

  def _add_nav_plugin_metrics(self, update):
    """
    Read the "metrics" file from audit log directory and return metrics
    from that file.
    """
    if not self.is_nav_enabled():
      # This means navigator logging isn't setup
      return

    log_dir = self.nav_conf.log_dir
    metrics_file = os.path.join(log_dir, "metrics")
    if os.path.exists(metrics_file):
      ctime = os.path.getctime(metrics_file)
      if ctime <= self._last_metrics_timestamp:
        return

      metrics_fh = None
      try:
        metrics_fh = open(metrics_file)
        metrics = dict(line.strip().split("=") for line in metrics_fh)

        # Now add all the metrics we know about.
        self._add_plugin_metrics(update, metrics)
        self._last_metrics_timestamp = ctime
      finally:
        if metrics_fh is not None:
          metrics_fh.close()

  def _add_plugin_metrics(self, update, metrics):
    # Add all the metrics we know about. Currently all the metrics
    # we care about is of type double (float in python).
    self._add_double_metric(update, metrics, "audit_plugin_parse_errors")
    self._add_double_metric(update, metrics, "audit_plugin_filtered")
    self._add_double_metric(update, metrics, "audit_plugin_coalesced")
    self._add_double_metric(update, metrics, "audit_plugin_exceptions")
    self._add_double_metric(update, metrics, "audit_plugin_events")

  def get_api_url(self, cluster_name, service_name, pipeline=None):
    """
    Returns api url endpoint that should be used to communicate when sending
    audit information to telemetry publisher.
    """
    assert self.nav_conf.has_telemetry_publisher_url(), \
      "Invalid configuration. Api url is only present with telemetry publisher"

    return '/api/v1/audits/%s/%s' % (cluster_name, service_name)

  def get_nav_sender(self):
    """
    If telemetry role is enabled tp_url is set to telemetry url and
    nas_url is set to navigator server url
    """
    return self._nav_sender(self.nav_conf.tp_url, self.nav_conf.nas_url,
                            self._config_dictionary)

  def _get_audit_pipelines(self):
    """
    Check for this process what are the pipelines, in case of generic
    get those. Depedning on nav, TP or both set them accordingly
    """
    new_pipeline_info = set()
    pipelines = cmf.audit.util.get_audit_pipelines_names(self._process_name)
    for pipeline_name in pipelines:
      if self.nav_conf and self.nav_conf.tp_url:
        new_pipeline_info.add((pipeline_name + "TP", self.nav_conf.tp_url))
      if self.nav_conf and self.nav_conf.nas_url:
        new_pipeline_info.add((pipeline_name, self.nav_conf.nas_url))
    return new_pipeline_info

  def update_pipelines_based_on_config(self):
    """
    Check Current Plugin to see what pipelines are present,
    Depending on configuration update number of pipelines
    """
    new_pipeline_info = self._get_audit_pipelines()
    old_pipelines = set([p.get_name() for p in self._pipelines])

    # set log perfix in case of audit for impalad process
    self._pipelines = [
          AuditPluginPipeline(pipeline_name, self._log_prefix, self._process_name, url=url)
          for pipeline_name, url in new_pipeline_info]
    new_pipelines = set([p.get_name() for p in self._pipelines])
    if new_pipelines != old_pipelines:
      LOG.info("Pipelines updated for Audit Plugin: %s", list(new_pipelines))

class MetadataPlugin(NavigatorPlugin):

  def __init__(self, log_entry_parser, pipeline_names, log_prefix=None,
               process_name=None, config_dictionary=None, is_csd_process=False):
    if is_csd_process:
      # For CSDs, pipelines are only created on iNotify Events
      metadata_pipelines = []
    else:
      nav_conf = None
      new_pipeline_names = []
      if config_dictionary and 'conf_dir' in config_dictionary:
        config_parser = cmf.audit.parse_metadata_configuration
        nav_conf = config_parser(config_dictionary['conf_dir'])
      if nav_conf and nav_conf.has_telemetry_publisher_url():
        for pipeline_name in pipeline_names:
          new_pipeline_names.append(pipeline_name + "TP")
      if nav_conf and not nav_conf.has_navigator_url():
        pipeline_names = []
      pipeline_names = pipeline_names + new_pipeline_names
      metadata_pipelines = [MetadataPluginPipeline(pipeline_name, log_prefix, process_name, False)
                            for pipeline_name in pipeline_names]

    super(MetadataPlugin, self).__init__(log_entry_parser, None,
      metadata_pipelines, MetadataSender, process_name, is_csd_process, log_prefix)

    self._agent_nav_metrics = None
    self._config_dictionary = config_dictionary

  def __str__(self):
    return self._safe_debug_string("Metadata")

  def get_avro_request(self, entries):
    """
    Returns the request map to send to the navigator server. This corresponds
    to the WAERequest record.
    """
    return {
      'serviceName': self.get_service_name(),
      'entries': entries
    }

  @catch_all()
  def refresh(self,
      conf_dir,
      config_parser = cmf.audit.parse_metadata_configuration,
      refresh_thread = None):
    """
    Wrapper that provides a default config_parser value for the _refresh method.
    """
    self._refresh(conf_dir, config_parser, refresh_thread)

  def _get_thread(self):
    return cmf.audit.metadata_thread

  def _add_nav_plugin_metrics(self, update):
    """
    We don't have monitoring for metadata.
    """
    pass

  def get_api_url(self, cluster_name, service_name, pipeline=None, is_tp_test=False):
    """
    Returns api url endpoint that should be used to communicate when sending
    lineage information.
    """
    api_url = '/api/v5/metadata/%s/%s' % (cluster_name, service_name)
    if self.is_csd_process():
      # Currently, we only have spark lineage as the CSD process.
      api_url = '/api/v10/metadata/spark/%s/%s' % (cluster_name, service_name)
      if self.nav_conf.has_telemetry_publisher_url():
        api_url = '/api/v1/operations/spark/%s/%s' % (cluster_name, service_name)
      return api_url

    if ((pipeline and pipeline.isTP() and self.nav_conf.has_telemetry_publisher_url()) or
       is_tp_test):
      api_url = '/api/v1/operations/%s/%s' % (cluster_name, service_name)

    return api_url

  def get_nav_sender(self):
    """
    If telemetry role is enabled server url is set to tp_url and
    nms_url is set to navigator server url.
    """
    return self._nav_sender(self.nav_conf.tp_url, self.nav_conf.nms_url,
                            self._config_dictionary)

  def _get_metadata_pipelines(self):
    """
    Check for this process what are the pipelines, in case of generic
    get those. Depedning on nav, TP or both set them accordingly
    """
    new_pipeline_info = set()
    pipelines = cmf.audit.util.get_metadata_pipelines_names(self._process_name)
    for pipeline_name in pipelines:
      if self.nav_conf and self.nav_conf.tp_url:
        new_pipeline_info.add((pipeline_name + "TP", self.nav_conf.tp_url))
      if self.nav_conf and self.nav_conf.nms_url:
        new_pipeline_info.add((pipeline_name, self.nav_conf.nms_url))
    return new_pipeline_info

  def update_pipelines_based_on_config(self):
    """
    Check Current Plugin to see what pipelines are present,
    Depending on configuration update number of pipelines
    """
    new_pipeline_info = self._get_metadata_pipelines()
    old_pipelines = set([p.get_name() for p in self._pipelines])
    # skip csd process
    if not self._is_csd_process:
      self._pipelines = [
          MetadataPluginPipeline(pipeline_name, self._log_prefix, self._process_name, False, url=url)
          for pipeline_name, url in new_pipeline_info]

      new_pipelines = set([p.get_name() for p in self._pipelines])
      if new_pipelines != old_pipelines:
        LOG.info("Pipelines updated for Metadata Plugin: %s", list(new_pipelines))

class ProfilePlugin(NavigatorPlugin):

  def __init__(self, log_entry_parser, ipc_request_name, pipeline_names,
      log_prefix=None, process_name=None):
    profile_pipelines = [
        ProfilePluginPipeline(pipeline_name, log_prefix, process_name)
        for pipeline_name in pipeline_names]
    super(ProfilePlugin, self).__init__(log_entry_parser, ipc_request_name,
        profile_pipelines, ProfileSender, process_name, False, log_prefix)
    self._agent_nav_metrics = None
    self._config_dictionary = None

  def __str__(self):
    return self._safe_debug_string("Profile")

  def get_avro_request(self, entries):
    """
    Returns the request map to send to the navigator server. This corresponds
    to the WAERequest record.
    """
    return {
        'serviceName': self.get_service_name(),
        'events': entries
    }

  @catch_all()
  def refresh(self,
      conf_dir,
      config_parser = cmf.audit.parse_profile_configuration,
      refresh_thread = None):
    """
    Wrapper that provides a default config_parser value for the _refresh method.
    """
    self._refresh(conf_dir, config_parser, refresh_thread)

  def _get_thread(self):
    return cmf.audit.profile_thread

  def get_api_url(self, cluster_name, service_name, pipeline=None):
    """
    Returns api url endpoint that should be used to communicate when sending
    profile information to telemetry publisher.
    """
    assert self.nav_conf.has_telemetry_publisher_url(), \
      "Invalid configuration. Api url is only present with telemetry publisher"

    return '/api/v1/impalaProfiles/%s/%s' % (cluster_name, service_name)

  def get_nav_sender(self):
    """
    If telemetry role is enabled tp url is set to telemetry url
    """
    return self._nav_sender(self.nav_conf.tp_url, self._config_dictionary)

  def _get_profile_pipelines(self):
    """
    Check for this process what are the pipelines, in case of generic
    get those. Depedning on nav, TP or both set them accordingly
    """
    new_pipeline_info = set()
    pipelines = cmf.audit.util.get_profile_pipelines_names(self._process_name)
    for pipeline_name in pipelines:
      if self.nav_conf and self.nav_conf.tp_url:
        new_pipeline_info.add((pipeline_name, self.nav_conf.tp_url))
    return new_pipeline_info

  def update_pipelines_based_on_config(self):
    """
    Check Current Plugin to see what pipelines are present,
    Depending on configuration update number of pipelines
    """
    new_pipeline_info = self._get_profile_pipelines()
    old_pipelines = set([p.get_name() for p in self._pipelines])

    self._pipelines = [
      ProfilePluginPipeline(pipeline_name, self._log_prefix, self._process_name, url=url)
      for pipeline_name, url in new_pipeline_info]
    new_pipelines = set([p.get_name() for p in self._pipelines])
    if new_pipelines != old_pipelines:
      LOG.info("Pipelines updated for Profile Plugin: %s", new_pipelines)
