# Copyright (c) 2013-2015 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import
import logging

from cmf.audit.navigator_logs import NavigatorLogDirectory
from cmf.audit.wal import Wal

LOG = logging.getLogger( __name__ )

class NavigatorPluginPipeline(object):
  """
  Class to track navigator log files generated by one navigator pipeline for a
  role. A role can have multiple pipelines active at same time, both for
  reading audit logs and metadata logs.
  """

  def __init__(self, pipeline_name, log_prefix, process_name, is_csd_process=False, url=None):
    """
    Setup the pipeline name for the pipeline
    """
    self._pipeline_name = pipeline_name
    self._nav_logs_dir = None
    self._wal = None
    self._wal_file_name = None # This gets set in the derived classes
    self._current_nav_log = None
    self._log_prefix = log_prefix
    """
    This used to keep track of pipeline name, used as log prefix
    in case when log_prefix was not explicitly passed.
    """
    self._pipeline_log_prefix = self._pipeline_name.rstrip("TP")
    self._isTP = self._pipeline_name.endswith("TP")
    self._process_name = process_name
    self._is_csd_process = is_csd_process
    # url where data needs to be sent by this pipeline
    self._url = url

  def __str__(self):
    return "Navigator Plugin Pipeline '%s' for %s" % (self.get_name(),
        self._process_name)

  def get_name(self):
    """
    Returns the name of the pipeline.
    """
    return self._pipeline_name

  def get_process_name(self):
    return self._process_name

  def isTP(self):
    """
    Check if pipeline is for telemetry publisher
    """
    return self._isTP

  def get_url(self):
    """
    Get url for given pipeline, where data needs to be sent
    """
    return self._url

  def _get_log_prefix(self, role_name):
    """
    Returns the prefix of navigator logs. If a log prefix is defined (such as
    for Impala), the log prefix itself is returned. Otherwise, if a pipeline
    name is defined (such as for HDFS, HBase, Hive, and Hue),
    <role name>-<pipeline name> is returned. Generic services (such as Sentry)
    fall into the remaining category of having no log prefix and no pipeline
    name, and in this case <role name> is returned.
    """
    if self._log_prefix is None:
      if self._pipeline_log_prefix:
        return role_name + "-" + self._pipeline_log_prefix
      else:
        return role_name
    else:
      return self._log_prefix

  def _set_wal_filename(self, role_type):
    raise NotImplementedError('This method must be implemented in a subclass.')

  def get_nav_logs_dir(self):
    """
    Return a NavigatorLogDirectory object that tracks nav log files for this role
    in the nav log directory.
    """
    return self._nav_logs_dir

  def get_current_nav_log(self):
    """
    Return the current nav log file
    """
    return self._current_nav_log

  def get_wal(self):
    """
    Return the WAL file used to track the nav entry location.
    """
    return self._wal

  def switch_log(self, next_file_name):
    """
    Closes current nav log and switches to next log.
    """
    if self._current_nav_log is not None:
      self._current_nav_log.close()
    self._current_nav_log = self._nav_logs_dir.get_log(next_file_name)

  def stop_nav_monitoring(self):
    """
    Close all resources and reset.
    """
    LOG.info("Stopping %s (log dir: %s)" % (self, self._nav_logs_dir))
    self._nav_logs_dir = None
    if self._wal is not None:
      self._wal.close()
      self._wal = None

    if self._current_nav_log is not None:
      self._current_nav_log.close()
      self._current_nav_log = None

  def start_nav_monitoring(self, nav_conf, override_role_name=None):
    """
    Start plugin pipeline log monitoring based on navigator configuration.
    """
    log_dir = nav_conf.log_dir
    role_name = nav_conf.role_name

    # For csd plugins, whenever a pipeline gets created, the role-name is
    # set to the pipeline name itself
    if override_role_name:
      role_name = override_role_name

    role_type = nav_conf.role_type
    self._nav_logs_dir = NavigatorLogDirectory(role_name,
                                             log_dir,
                                             self._get_log_prefix(role_name),
                                             self._is_csd_process,
                                             self._pipeline_name)
    wal_file_name = self._set_wal_filename(role_type)
    self._wal = Wal(log_dir, wal_file_name,
        self._process_name)
    file_name, offset = self._wal.get_file_info()
    LOG.info("Got file %s from wal %s." % (file_name, self._wal._wal_file))
    self._current_nav_log = self._nav_logs_dir.get_log(file_name)
    self._current_nav_log.seek(offset)
    LOG.info("Starting with navigator log %s for role %s and pipeline %s" %
              (self._current_nav_log.get_file_name(), role_type, self._pipeline_name))

  def get_generated_wal_file_name(self):
    return self._wal_file_name


class AuditPluginPipeline(NavigatorPluginPipeline):
  def _set_wal_filename(self, role_type):
    """
    Return the wal filename to use. As Impala Wal file was named differently
    and we released it in CM4.7, we need to stick with that name for upgrade
    scenario.
    """
    wal_filename = role_type.lower()
    if self._pipeline_name:
      wal_filename = wal_filename + "_" + self._pipeline_name.lower()
    self._wal_file_name = wal_filename + "_audit_wal"

    return self._wal_file_name


class MetadataPluginPipeline(NavigatorPluginPipeline):
  def _set_wal_filename(self, role_type):
    """
    Return the wal filename to use. For CSD plugins the wal filename
    includes the pipeline name since there could be more than one wal files in the same directory.
    """
    wal_filename = ""
    if self._is_csd_process:
      wal_filename = self._pipeline_name
    else:
      wal_filename = role_type.lower()
    self._wal_file_name = wal_filename + "_metadata_wal"

    return self._wal_file_name

class ProfilePluginPipeline(NavigatorPluginPipeline):
  def _set_wal_filename(self, role_type):
    """
    Return the wal filename to use.
    """
    wal_filename = role_type.lower()
    if self._pipeline_name:
      wal_filename = wal_filename + "_" + self._pipeline_name.lower()
    self._wal_file_name = wal_filename + "_profile_wal"

    return self._wal_file_name
