# Copyright (c) 2013-2015 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import avro.protocol
import avro.ipc
import cmf.https
import logging
import socket
import urllib
import urllib2
from httplib import BadStatusLine
from urlparse import urlparse

import avro.ipc
import avro.protocol
import os
import simplejson

from cmf.throttling_logger import ThrottlingLogger
from cmf.util import enum
from cmf.util.url import urlopen_with_timeout

LOG = logging.getLogger( __name__ )
THROTTLED_LOG = ThrottlingLogger(LOG)

Result = enum(SUCCESS=1,
              FAILED=2,
              TIMED_OUT=3,
              CONNECTION_LOST=4)

class NavigatorSender(object):
  def close(self):
    raise NotImplementedError('This method must be implemented in a subclass.')

  def send_nav_log_entries(self, plugin, pipeline, log_entry_batch):
    raise NotImplementedError('This method must be implemented in a subclass.')

  def _get_parsed_entries(self, plugin, log_entry_batch):
    log_entry_parser = plugin.get_log_entry_parser()
    log_parser_model = plugin.get_log_parser_model()
    service_name = plugin.get_service_name()
    service_type = plugin.get_service_type()

    entries = []
    ended_entry = None
    for _, _, log_entry in log_entry_batch:
      try:
        parsed_entries = log_entry_parser(service_type, service_name,
          log_entry, log_parser_model)
      except:
        THROTTLED_LOG.exception("Unable to parse entry: %s" % (log_entry, ))
        raise Exception('Failure while parsing log entry.')

      for parsed_entry in parsed_entries:
        entries.append(parsed_entry)
        if isinstance(parsed_entry, dict) and parsed_entry.has_key('ended'):
          ended_entry = parsed_entry

    return entries, ended_entry

  def _send_avro_data(self, plugin, log_entry_batch, log_entry_removal):
    """
    Method to parse the audit log line and send one or multiple entries
    to the audit server or telemetry publisher.
    Returns True if the entry was successfully sent to the server, False
    otherwise.
    """
    assert len(log_entry_batch) > 0

    ret = Result.SUCCESS
    ipc_request_name = plugin.get_ipc_request_name()

    try:
      if self._transceiver is None:
        if self._nas_use_tls:
          self._transceiver = cmf.https.HTTPSTransceiver(self._nas_host_name,
              int(self._nas_port),
              self.verify_cert_file,
              self.verify_cert_dir,
              self.client_key_file,
              self.client_cert_file,
              self.client_keypw_file,
              self.client_keypw_cmd,
              self.max_cert_depth)
        else:
          self._transceiver = avro.ipc.HTTPTransceiver(self._nas_host_name,
              self._nas_port)
        self._requestor = avro.ipc.Requestor(self._NAVIGATOR_AUDIT_PROTOCOL,
            self._transceiver)
        LOG.info("Successfully connected to audit server at "
            "url: %s, port: %s" % (self._nas_host_name, self._nas_port))
    except socket.error, ex:
      # Unable to connect to server. wait for it to come up
      THROTTLED_LOG.error("Unable to communicate with audit server "
          "(host: %s, port: %s): %s" % (self._nas_host_name, self._nas_port, ex))
      return Result.FAILED, None

    try:
      entries, ended_entry = self._get_parsed_entries(plugin, log_entry_batch)

      # in case of Hive plugins and pipeline is not telemetry publisher,
      # audit data should skip field 'queryConfig'
      if plugin.get_service_type() == "HIVE":
        entries = log_entry_removal(entries)
      LOG.debug("Sending %d navigator entries to audit server." % (len(entries), ))
      request = plugin.get_avro_request(entries)
      response = self._requestor.request(ipc_request_name, dict(request=request))
      if response['success']:
        LOG.debug("Successfully sent %d entries to audit server" %
                  (len(entries), ))
      else:
        ret = Result.FAILED
    except socket.timeout, ex:
      ret = Result.TIMED_OUT
    except BadStatusLine, ex:
      # this means we lost connection to nav server and need to reestablish it
      # Return False and let caller retry
      ret = Result.CONNECTION_LOST
    except Exception, ex:
      ret = Result.FAILED

    if ( ret != Result.SUCCESS ):
      LOG.error("Unable to send message to audit server for %s service, (host: %s, port: %s), Error: %d" % (plugin.get_service_name(), self._nas_host_name, self._nas_port, ret) )
      self.close()

    return ret, ended_entry

  def _send_json_data(self, plugin, pipeline, log_entry_batch, log_entry_formatter=lambda entries: entries):
    """
    Method to parse the metadata log line and send one or multiple entries
    to the metadata server.
    Returns True if the entry was successfully sent to the server, False
    otherwise.
    """
    assert len(log_entry_batch) > 0

    ret = Result.SUCCESS
    ended_entry = None
    try:
      entries, ended_entry  = self._get_parsed_entries(plugin, log_entry_batch)

      cluster_name = urllib.quote(plugin.get_cluster_name())
      service_name = urllib.quote(plugin.get_service_name())
      """
      since url are different for telemetry publisher and navigator
      metadata/audit server
      """
      url = pipeline.get_url() + plugin.get_api_url(cluster_name, service_name, pipeline)

      LOG.debug("Sending %d navigator entries to url %s." % (len(entries), url))

      """
      For profile parser data is different, hence dump entries to json
      """
      if plugin.is_profile():
        request_data = simplejson.dumps(entries)
      else:
        request_data = simplejson.dumps(log_entry_formatter(entries))
      request = urllib2.Request(str(url))

      request.add_header('Content-Type', 'application/json')
      request.add_data(request_data)

      try:
        timeout = plugin.get_nav_connection_timeout()
        LOG.debug("Sending %d entries to server with timeout %d" %
                  (len(entries), timeout))

        cmf.https.ssl_url_opener.open(request, timeout=timeout)

        LOG.debug("Successfully sent %d entries to server" %
                  (len(entries), ))

      except urllib2.URLError, e:
        if hasattr(e, 'code') and str(e.code)[0] == '2':
          # Older versions of python raise an exception when the response code
          # is anything other than 200, even though the other 2xx responses
          # represent a success. If this happens, we will not treat this as an
          # error.
          pass
        elif isinstance(getattr(e, 'reason'), socket.timeout):
          ret = Result.TIMED_OUT
        else:
          THROTTLED_LOG.error("Error occurred when sending entry to server: %s, url: %s" % (e, pipeline.get_url()))
          ret = Result.FAILED
      except socket.timeout, e:
        ret = Result.TIMED_OUT

    except Exception, ex:
      THROTTLED_LOG.exception("Unable to send message to server url: %s" % (pipeline.get_url(),))
      ret = Result.FAILED

    return ret, ended_entry

class AuditSender(NavigatorSender):
  """
  This class parses a line from audit log and then creates one or more audit
  entries from it. The events are then sent to the audit server.
  """

  # A map from Service Type to field in AvroAuditEvent2
  AUDIT_EVENT2_TYPE_MAP = {"HDFS": "hdfs",
                           "HBASE": "hbase",
                           "HIVE": "hive",
                           "IMPALA": "impala"}

  def __init__(self, tp_url, nas_url, config_dictionary=None):
    self._tp_url = tp_url
    self._nas_url = nas_url
    self._nas_use_tls = None
    self._nas_host_name = None
    self._nas_port = None
    if tp_url:
      #url data for teleMetry publisher
      parsed_url = urlparse(tp_url)
      netloc = parsed_url[1]
      self._host_name, self._port = netloc.split(':', 1)
      self._use_tls = ("https" == parsed_url[0])
    if nas_url:
      #url data for Navigator
      parsed_url = urlparse(nas_url)
      netloc = parsed_url[1]
      self._nas_host_name, self._nas_port = netloc.split(':', 1)
      self._nas_use_tls = ("https" == parsed_url[0])

    self._transceiver = None
    self._requestor = None
    self._NAVIGATOR_AUDIT_PROTOCOL = avro.protocol.parse(
      file(os.path.join(os.path.dirname(__file__), "navigator.avpr")).read())
    if config_dictionary is None:
      self.verify_cert_file = None
      self.verify_cert_dir = None
      self.max_cert_depth = None
      self.client_key_file = None
      self.client_cert_file = None
      self.client_keypw_file = None
      self.client_keypw_cmd = None
    else:
      self.verify_cert_file = config_dictionary["verify_cert_file"]
      self.verify_cert_dir = config_dictionary["verify_cert_dir"]
      self.max_cert_depth = config_dictionary["max_cert_depth"]
      self.client_key_file = config_dictionary["client_key_file"]
      self.client_cert_file = config_dictionary["client_cert_file"]
      self.client_keypw_file = config_dictionary["client_keypw_file"]
      self.client_keypw_cmd = config_dictionary["client_keypw_cmd"]


  def close(self):
    if self._transceiver is not None:
      try:
        self._transceiver.close()
      except:
        pass # ignore

      self._requestor = None
      self._transceiver = None

  def send_nav_log_entries(self, plugin, pipeline, log_entry_batch):
    """
    Send audit data to either audit server in avro format or to Telemetry
    publisher in json format.
    """
    if pipeline.isTP():
      def log_entry_formatter(entries):
        updated_entries = []
        for entry in entries:
          updated_entry = dict()
          entry_type = AuditSender.AUDIT_EVENT2_TYPE_MAP.get(plugin.nav_conf.service_type, "generic")
          updated_entry[entry_type] = entry
          updated_entries.append(updated_entry)
        return updated_entries

      return self._send_json_data(plugin, pipeline, log_entry_batch, log_entry_formatter)
    else:
      def log_entry_removal(entries):
        updated_entries = []
        for entry in entries:
          if entry.has_key('queryConfig'):
            entry.pop('queryConfig', None)
          updated_entries.append(entry)
        return updated_entries

      return self._send_avro_data(plugin, log_entry_batch, log_entry_removal)

class MetadataSender(NavigatorSender):
  """
  This class parses a line from metadata log and then creates one or more
  metadata entries from it. The events are then sent to the metadata server.
  """
  def __init__(self, tp_url, nms_url, config_dictionary=None):
    self._tp_url = tp_url
    self._nms_url = nms_url

  def close(self):
    pass

  def send_nav_log_entries(self, plugin, pipeline, log_entry_batch):
    """
    Send lineage entries as JSON to the metadata server.
    """
    return self._send_json_data(plugin, pipeline, log_entry_batch)

class ProfileSender(NavigatorSender):
  """
  This class parses a line from impala profile log and then creates one or more
  entries from it. The events are then sent to the telemetry publisher.
  """
  def __init__(self, tp_url, config_dictionary=None):
    self._tp_url = tp_url

  def close(self):
    pass

  def send_nav_log_entries(self, plugin, pipeline, log_entry_batch):
    """
    Send lineage entries as JSON to the metadata server.
    """
    return self._send_json_data(plugin, pipeline, log_entry_batch)
