# Copyright (c) 2013-2015 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import
import copy
import itertools
import logging
import pyinotify
import time
import os

from cmf.util import catch_all
from cmf.audit.navigator_sender import Result
from cmf.audit.inotify_event_processor import INotifyEventProcessor
from cmf.throttling_logger import ThrottlingLogger
from pyinotify import WatchManager, ThreadedNotifier
from threading import Thread, RLock

LOG = logging.getLogger( __name__ )
THROTTLED_LOG = ThrottlingLogger(LOG)


class RequestTimedOut(Exception):
  """
  Error raised when a request to navigator server took long time to respond and
  the socket timed out. We need to wait for some time before resending the
  request.
  """
  def __str__(self):
    return "Request to Navigator server timed out."

class NavigatorThread(object):
  """
  Class for Navigator thread to tail log files for all the roles with audit
  data or metadata and then send those events to navigator server.
  """

  THREADS = [] # All thread objects

  def __init__(self, default_batch_timeout, name):
    """
    Initialize and start a daemon thread to start monitoring navigator events.
    """
    self._name = name
    self._nav_plugins = set()
    self._nav_plugins_lock = RLock()
    self._batch_timeout = default_batch_timeout
    self._watch_manager = None
    self._event_processor = None

    # Used to send event to navigator server
    self._log_entry_sender = None
    # inotify related variables
    self._notifier = None

    # Create and start main thread for processing navigator log files
    self._stop_nav_thread = False
    self._nav_thread = Thread(target = self._monitor_logs,
                                name = name)
    self._nav_thread.setDaemon(True)
    self.THREADS.append(self)

  @classmethod
  def start_threads(cls):
    for thr in cls.THREADS:
      thr._nav_thread.start()

  #@VisibleForTesting
  def get_watch_manager(self):
    """
    This method exposes the watch manager which is used only for test purposes
    """
    return self._watch_manager

  # @VisibleForTesting
  def get_event_processor(self):
    return self._event_processor

  def _purge_files_per_plugin(self, purge_file_list, num_pipelines):
    """
    Function to find min range of files that needs to be purged, across all
    pipelines for a plugin.
    For e.g. for four pipeline, content of 'purge_file_list' could look like
    this [4, 3, 1, 7], min of this list is the index till which files will be
    purged, zero is a valid index.
    """
    #check if all pipelines have move forward.
    if len(purge_file_list) != num_pipelines:
      """
      Since size of purge file list is not same as number of pipelines,
      this means some pipelines are still working on those log files.
      """
      return

    min_so_far = -1
    min_so_far_pipeline = None
    for index, pipeline in purge_file_list:
      if index >= 0 and min_so_far < index:
        min_so_far = index
        min_so_far_pipeline = pipeline

    if min_so_far >= 0:
      log_dir = min_so_far_pipeline.get_nav_logs_dir()
      LOG.info("Start Purging File till index: %d log_dir: %s" % (min_so_far, log_dir))
      log_dir.purge_files_by_idx(min_so_far)

  @catch_all()
  def _monitor_logs(self):
    """
    Main method that runs in a separate thread and monitors logs for all
    roles with audit data or metadata and sends events to navigator server.
    """
    event_poll_interval = self._batch_timeout
    # we wait longer if server is busy. Set the max wait time to be number of
    # minutes of event_poll_interval (which is in seconds)
    max_event_poll_interval = 60 * event_poll_interval
    num_success = 0
    while not self._stop_nav_thread:
      success = True
      self.stop_plugins()
      self.refresh_plugins()
      plugins = [plugin for plugin in self._nav_plugins
                   if plugin.is_nav_enabled()]
      continueProcessingPlugins = True
      """
      The outer loop will run as long as some pipeline within a plugin,
      return value conf.max_read_processed_files, this means there is lot of data
      generated by the process plugin
      """
      while continueProcessingPlugins:
        # if there are more files process then this will get updated
        continueProcessingPlugins = False
        for plugin in plugins:
          conf = plugin.nav_conf
          log_dir = None
          try:
            purge_file_list = []
            for pipeline in plugin.get_nav_pipelines():
              purge_files = self._read_and_send_entries(plugin, pipeline)
              purge_file_list.append((purge_files, pipeline))

            self._purge_files_per_plugin(purge_file_list,
                                         len(plugin.get_nav_pipelines()))
            # check if any of the pipeline is very busy in that case let's loop
            # again
            if conf.max_read_processed_files  in [idx for idx, pipeline in purge_file_list]:
              continueProcessingPlugins = True
              LOG.info("Done Processing %d files restart plugin loop " % (conf.max_read_processed_files, ))
          except RequestTimedOut, ex:
            # Exponentially delay time we send request to server
            event_poll_interval = min(max_event_poll_interval, event_poll_interval * 2)
            LOG.info("New Event Poll interval %d" % (event_poll_interval))
            success = False
            break
          except:
            # any other error log and continue normally
            LOG.exception("Exception caught when trying to read and send events.")

      if success and event_poll_interval > self._batch_timeout:
        # linearly decrease the timeout, takes about half an hour to get
        # back to original default value of 5 seconds
        num_success += 1
        new_interval = event_poll_interval - self._batch_timeout * num_success
        event_poll_interval = max(self._batch_timeout, new_interval)
        LOG.info("New Event Poll interval %d" % (event_poll_interval))
      else:
        num_success = 0

      # Sleep for batch timeout seconds, means that we will send events every
      # _batch_timeout seconds, thus batching them up.
      time.sleep(event_poll_interval)

  def refresh_csd_plugin(self, process_name, conf_dir):
    for plugin in self._nav_plugins:
      if plugin.get_process_name() == process_name:
        plugin.refresh(conf_dir)
        break

  def stop_plugin(self, process_name):
    """
    Calls stop() method on plugin, if the passed in process_name
    matches the process_name of the plugin
    """
    for plugin in self._nav_plugins:
      if plugin.get_process_name() == process_name:
        plugin.stop_nav_monitoring()
        break

  def stop_plugins(self):
    """
    Calls stop() method on plugins that are marked for stopping.
    """
    plugins = [plugin for plugin in self._nav_plugins if plugin.is_stop_called()]
    for plugin in plugins:
      plugin.immediate_stop()
      self._remove_navigator_plugin(plugin)

  def refresh_plugins(self):
    """
    Calls refresh_plugin() method on plugins that are marked for refresh.
    """
    for plugin in self._nav_plugins:
      plugin.immediate_refresh()

  def stop(self):
    """
    Signals thread to stop processing and exit cleanly. Waits for
    the thread to terminate and then returns.
    """
    self._stop_nav_thread = True
    self._nav_thread.join()
    LOG.info("Stopped navigator thread " + self._name)

  def is_plugin_already_added(self, process_name):
    """
    Returns true if the plugin is already added
    """
    try:
      is_added = False
      self._nav_plugins_lock.acquire()
      for plugin in self._nav_plugins:
        if plugin.get_process_name() == process_name:
          is_added = True
          break
    finally:
      self._nav_plugins_lock.release()

    return is_added

  def add_navigator_plugin(self, nav_plugin):
    """
    Adds a new navigator plugin to the set of plugins. If the plugin is just
    getting refreshed, it may already be present in the set and then this is
    a nop.
    """
    try:
      self._nav_plugins_lock.acquire()
      nav_plugins_copy = copy.copy(self._nav_plugins)
      nav_plugins_copy.add(nav_plugin)
      self._nav_plugins = nav_plugins_copy
    finally:
      self._nav_plugins_lock.release()

  def _remove_navigator_plugin(self, plugin):
    """
    Removes an existing plugin from the set of plugins.
    """
    try:
      self._nav_plugins_lock.acquire()
      nav_plugins_copy = copy.copy(self._nav_plugins)
      nav_plugins_copy.remove(plugin)
      self._nav_plugins = nav_plugins_copy
    finally:
      self._nav_plugins_lock.release()

  def refresh_nav_plugin(self, plugin, event_sender=None):
    """
    This method is called when state of a nav-enabled process changes, like:
    1. Process is started
    2. There was a change in value of Navigator client configuration properties
    """
    if not plugin.is_nav_enabled():
      assert plugin in self._nav_plugins
      LOG.info("Removing %s" % (plugin,))
      self._remove_navigator_plugin(plugin)
    else:
      LOG.debug("Initializing configuration for %s" % (plugin,))
      self._init_conf(plugin, event_sender)

    self._update_notifier()

  def _init_conf(self, plugin, log_entry_sender):
    """
    This method is called to set configurations that are supposed to apply to
    code in agent and should be same for all plugins. But as we only get these
    configurations as part of navigator.client.properties (instead of putting
    them in CM agent's config.ini), we have to wait till first of our nav-enabled
    process is started to initialize them.
    """
    if not plugin.is_nav_enabled():
      return

    conf = plugin.nav_conf
    self._batch_timeout = conf.batch_timeout
    if log_entry_sender is None:
      self._log_entry_sender = plugin.get_nav_sender()
    else:
      self._log_entry_sender = log_entry_sender

    LOG.debug("Log entry poll interval: %d", self._batch_timeout)
    LOG.debug("Navigator server url: %s", conf.server_url)

  def _update_notifier(self):
    """
    Updates the inotify notifier as per new navigator plugin information.
    """
    if self._notifier is not None:
      self._notifier.stop()
      self._notifier = None

    active_plugins = [plugin for plugin in self._nav_plugins
                      if plugin.is_nav_enabled()]
    audit_event_processor = INotifyEventProcessor(nav_plugins=active_plugins)
    self._event_processor = audit_event_processor
    wm = WatchManager()
    self._watch_manager = wm
    self._notifier = ThreadedNotifier(wm, audit_event_processor)
    self._notifier.setDaemon(True)
    self._notifier.start()

    # Get list of logs to monitor
    log_dirs = [plugin.nav_conf.log_dir for plugin in active_plugins]
    wm.add_watch(log_dirs, pyinotify.IN_CREATE | pyinotify.IN_DELETE | pyinotify.IN_MOVED_FROM)

    # Now we have notification setup, refresh log list
    for plugin in active_plugins:
      for pipeline in plugin.get_nav_pipelines():
        pipeline.get_nav_logs_dir().refresh_nav_logs()

  def purge_wal_file(self, plugin, pipeline):
    wal_file_path = os.path.join(plugin.nav_conf.log_dir, pipeline.get_generated_wal_file_name())
    if os.path.exists(wal_file_path):
      os.remove(wal_file_path)
    else:
      LOG.warning("Wal file %s exists but not on filesystem." % (wal_file_path))

  def handle_csd_plugin(self, plugin, pipeline):
    if plugin.is_csd_process():
      LOG.info("Remvoing %s pipeline for %s plugin", pipeline.get_name(), plugin.get_process_name())
      log_dir = pipeline.get_nav_logs_dir()
      log_dir.purge_all_logs()
      self.purge_wal_file(plugin, pipeline)
      plugin.remove_csd_pipeline(pipeline)

  def _get_purge_file_idx(self, log_dir, max_num_logs, file_name):
    # purge is based on array index, so updating max_num_logs by
    # subtracting 1 from it to make it array based index.
    return log_dir.get_number_of_purge_files(file_name, max_num_logs - 1)

  def _read_and_send_entries(self, plugin, plugin_pipeline):
    """
    This method reads a "batch_size" number of log entries and then tries to
    send them to the navigator server. In case of a failure the method keeps on
    retrying forever till it succeeds.
    If no new entries are found in the log, the method checks to see if the
    log file has rolled over, in which case it opens the new log file and starts
    reading (and sending) entries from that file.
    It returns idx of the file that needs to be purged, since its based on array/list
    indexing zero is valid index. Anything less than zero is invalid index.
    """
    num_files_to_purge = -1
    if self._log_entry_sender is None:
      return num_files_to_purge

    processed_file_count  = 0
    conf = plugin.nav_conf
    while True:
      current_log = plugin_pipeline.get_current_nav_log()
      current_file_name = current_log.get_file_name()
      current_log_offset = current_log.tell()
      log_dir = plugin_pipeline.get_nav_logs_dir()
      next_file_name = log_dir.get_next_log_file_name(current_file_name)

      # On successful transmission to navigator server, filename and offset are
      # saved to the WAL file
      nav_queue, bytes_read = self._read_entries(current_log, conf)

      if nav_queue:
        result = None
        try:
          result, ended_entry = self._send_entries(plugin, plugin_pipeline, nav_queue, bytes_read)
        finally:
          if result != Result.SUCCESS: # We need to reread events as we failed to send them
            current_log.seek(current_log_offset)

          if result == Result.SUCCESS:
            if ended_entry and ended_entry["ended"]:
              self.handle_csd_plugin(plugin, plugin_pipeline)
              return num_files_to_purge

        if result == Result.FAILED:
          return num_files_to_purge
      elif next_file_name is not None:
        processed_file_count += 1
        plugin_pipeline.switch_log(next_file_name)
        plugin_pipeline.get_wal().append(next_file_name, 0)
        LOG.info("Done processing navigator log %s, switching to next log %s, count %s" %
                 (current_file_name, next_file_name, processed_file_count))
        if processed_file_count == conf.max_read_processed_files and plugin.isPurgeAble():
          return self._get_purge_file_idx(log_dir, conf.max_num_navigator_logs, next_file_name)
      else:
        """
        check to make sure next_file_name is not None, before attempting to purge.
        For Impala Query Profile Plugin we cannot delete profiles, since its handle
        by impala service itself.
        """
        if current_file_name and plugin.isPurgeAble():
          num_files_to_purge = self._get_purge_file_idx(log_dir, conf.max_num_navigator_logs,
                                                        current_file_name)
        return num_files_to_purge

  def _read_entries(self, current_nav_log, nav_conf):
    """
    Reads entries from navigator log and returns "batch_size" number of
    entries. Also returns the number of bytes read for metrics purpose.
    """
    batch_size = nav_conf.batch_min_size
    max_request_size = nav_conf.max_request_size
    nav_queue = list()
    bytes_read = 0

    if current_nav_log.exists():
      # process events in batches. read_nav_log method returns a tuple:
      # filename, offset, bytes, event
      nav_file_iterator = itertools.islice(current_nav_log.read_nav_log(),
                                             0, batch_size)
      for file_name, offset, bytes, event in nav_file_iterator:
        bytes_read += bytes
        nav_queue.append((file_name, offset, event))
        if bytes_read > max_request_size:
          # If we just read more than max request size, stop here. For simplicity,
          # We don't care few extra bytes we have read. Trying to avoid unreading it,
          # or storing different offset in the WAL file than file pointer. All of
          # this extra code and its not super critical if we go few bytes over.
          break

    return nav_queue, bytes_read

  def _send_entries(self, nav_plugin, nav_plugin_pipeline, nav_queue, bytes_read):
    """
    Send data read from navigator log file to the navigator server. On
    failure return the code needed to handle the error.
    """
    LOG.debug("Fetched %d rows from navigator log." % (len(nav_queue),))
    nav_metrics = nav_plugin.get_agent_nav_metrics()
    result, ended_entry = self._log_entry_sender.send_nav_log_entries(nav_plugin,
                                                                      nav_plugin_pipeline,
                                                                      nav_queue)

    if result == Result.SUCCESS:
      filename, offset, _ = nav_queue.pop()

      if nav_metrics:
        nav_metrics.increment_events_sent(len(nav_queue))
        nav_metrics.increment_bytes_sent(bytes_read)

      nav_plugin_pipeline.get_wal().append(filename, offset)
    elif result == Result.FAILED:
      THROTTLED_LOG.info("Unable to send data to nav server. Will try again.")
      if nav_metrics:
        nav_metrics.increment_send_failures()
    elif result == Result.CONNECTION_LOST:
      pass  #NOP, we need to retry
    elif result == Result.TIMED_OUT:
      # Server is too busy, record the event so that we slow down and then
      # try to resend the events after some time
      raise RequestTimedOut()
    else:
      raise NotImplementedError("Don't know what to do for result %d" % (result,))

    return result, ended_entry


class AuditThread(NavigatorThread):
  """
  Class for Audit thread to tail log files for all the auditable roles and then
  send those events to navigator audit server.
  """

  def __init__(self, default_batch_timeout, name = "Audit-Plugin"):
    LOG.info("Creating the Audit Thread")
    super(AuditThread, self).__init__(default_batch_timeout, name)


class MetadataThread(NavigatorThread):
  """
  Class for Metadata thread to tail log files for all the roles with logged
  metadata and then send those events to navigator metadata server.
  """

  def __init__(self, default_batch_timeout, name = "Metadata-Plugin"):
    LOG.info("Creating the Metadata Thread")
    super(MetadataThread, self).__init__(default_batch_timeout, name)

class ProfileThread(NavigatorThread):
  """
  Class for Profile thread to tail log files for all the roles with logged
  metadata and then send those events to navigator profile server.
  """

  def __init__(self, default_batch_timeout, name = "Profile-Plugin"):
    LOG.info("Creating the Profile Thread")
    super(ProfileThread, self).__init__(default_batch_timeout, name)
