# Copyright (c) 2013-2015 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import
import logging
from StringIO import StringIO
from zipfile import ZipFile

import os
import hashlib

import cmf.audit
from cmf.audit import METADATA_CONFIG_FILENAME

from cmf.audit.impala.impalad_audit_parser import parse_impalad_audit_event
from cmf.audit.navigator_plugin import AuditPlugin, MetadataPlugin, ProfilePlugin
from cmf.audit.generic_audit_parser import parse_generic_audit_event
from cmf.audit.generic_metadata_parser import parse_generic_metadata_event
from cmf.audit.generic_profile_parser import parse_generic_profile_event
from cmf.audit.nop_audit_parser import parse_audit_event
from cmf.audit.spark.spark_lineage_parser import parse_spark_lineage_event
from cmf.util import catch_all

LOG = logging.getLogger( __name__ )

audit_impala_pipelines = [""]
audit_hdfs_pipelines = ["NavigatorLog4jAppender", "HdfsAuditLogger"]
audit_hbase_master_pipelines = ["HBaseAuditAppender", "MasterAuditCoProcessor"]
audit_hbase_region_pipelines = ["HBaseAuditAppender", "RegionAuditCoProcessor"]
audit_hive_pipelines = ["HiveExecHookContext", "HiveMetaStoreEventListener", "HiveSentryOnFailureHook"]
audit_hue_pipelines = ["HiveExecHookContext", "HiveMetaStoreEventListener", "HiveSentryOnFailureHook"]
audit_generic_pipelines = [""]

audit_process_to_pipeline_dict = {'impala-IMPALAD': audit_impala_pipelines,
                                  'hdfs-NAMENODE': audit_hdfs_pipelines,
                                  'hbase-MASTER': audit_hbase_master_pipelines,
                                  'hbase-REGIONSERVER': audit_hbase_region_pipelines,
                                  'hive-HIVESERVER2': audit_hive_pipelines,
                                  'hue-BEESWAX_SERVER': audit_hue_pipelines}
_audit_plugins = [
  ('impala-IMPALAD', lambda process_name, config_dictionary: AuditPlugin(parse_impalad_audit_event,
      "writeImpalaAuditEvents", audit_impala_pipelines, "impala_audit_event_log_",
      process_name=process_name, config_dictionary = config_dictionary)),
  ('hdfs-NAMENODE', lambda process_name, config_dictionary: AuditPlugin(parse_audit_event,
      "writeHdfsAuditEvents", audit_hdfs_pipelines,
      process_name=process_name, config_dictionary = config_dictionary)),
  ('hbase-MASTER', lambda process_name, config_dictionary: AuditPlugin(parse_audit_event,
      "writeHBaseAuditEvents", audit_hbase_master_pipelines,
      process_name=process_name, config_dictionary = config_dictionary)),
  ('hbase-REGIONSERVER', lambda process_name, config_dictionary: AuditPlugin(parse_audit_event,
      "writeHBaseAuditEvents", audit_hbase_region_pipelines,
      process_name=process_name)),
  ('hive-HIVESERVER2', lambda process_name, config_dictionary: AuditPlugin(parse_audit_event,
      "writeHiveAuditEvents", audit_hive_pipelines,
      process_name=process_name, config_dictionary = config_dictionary)),
  ('hue-BEESWAX_SERVER', lambda process_name, config_dictionary: AuditPlugin(parse_audit_event,
      "writeHiveAuditEvents", audit_hue_pipelines,
      process_name=process_name, config_dictionary = config_dictionary)),
  ('hive_exec-HIVESERVER2', lambda process_name, config_dictionary: AuditPlugin(parse_audit_event,
      "writeHiveAuditEvents", audit_hive_pipelines,
      process_name=process_name, config_dictionary = config_dictionary))
]

metadata_impala_pipelines = [""]
metadata_generic_pipelines = [""]
metadata_process_to_pipeline_dict = {'impala-IMPALAD': metadata_impala_pipelines}
_lineage_plugin_dict = {
  'impala-IMPALAD': lambda process_name, config_dictionary: MetadataPlugin(
      parse_generic_metadata_event, metadata_impala_pipelines, "impala_lineage_log_",
      process_name=process_name, config_dictionary=config_dictionary, is_csd_process=False)
}

profile_impala_pipelines = ["TP"]
profile_generic_pipelines = ["TP"]
profile_process_to_pipeline_dict = {'impala-IMPALAD': profile_impala_pipelines}
"""
This is only enabled for telemetry publisher
"""
_profile_plugins = [
  ('impala-IMPALAD', lambda process_name: ProfilePlugin(parse_generic_profile_event,
     "writeImpalaProfileEvents", profile_impala_pipelines, "impala_profile_log_",
     process_name=process_name))
]

class EmptyNavigatorPlugin(object):
  def refresh(self, conf_dir):
    pass

  def stop_nav_monitoring(self):
    pass

  def add_metrics(self, update):
    pass

def get_audit_pipelines_names(process_name):
  """
  Returns pipeline names for a given process, in case of process name being none
  that's generic process return generic pipelines
  """
  if process_name in audit_process_to_pipeline_dict:
    return audit_process_to_pipeline_dict[process_name]
  else:
    return audit_generic_pipelines

@catch_all(EmptyNavigatorPlugin())
def make_audit_plugin(process_name, config_dictionary=None):
  """
  Creates an audit plugin that will collect audit events and send them to
  the audit server.
  """
  for process_name_prefix, plugin in _audit_plugins:
    if process_name.startswith(process_name_prefix):
      LOG.info("Using specific audit plugin for process %s" % (process_name, ))
      return plugin(process_name = process_name, config_dictionary = config_dictionary)

  LOG.info("Using generic audit plugin for process %s" % (process_name, ))
  return AuditPlugin(parse_generic_audit_event,
      "writeGenericAuditEvents", audit_generic_pipelines, process_name=process_name, config_dictionary = config_dictionary)

def get_metadata_pipelines_names(process_name):
  """
  Returns pipeline names for a given process, in case of process name being none
  that's generic process return generic pipelines
  """
  if process_name in metadata_process_to_pipeline_dict:
    return metadata_process_to_pipeline_dict[process_name]
  else:
    return metadata_generic_pipelines

@catch_all(EmptyNavigatorPlugin())
def make_metadata_plugin(process_name, config_dictionary=None):
  """
  Creates a metadata plugin that will collect metadata entries and send them to
  the metadata server.
  """
  LOG.info("Creating metadata plugin for process %s" % (process_name, ))
  try:
    LOG.info("Using specific metadata plugin for process %s" % (process_name, ))
    return _lineage_plugin_dict[process_name](process_name, config_dictionary)
  except KeyError, ex:
    LOG.info("Using generic metadata plugin for process %s" % (process_name, ))
    return MetadataPlugin(parse_generic_metadata_event, metadata_generic_pipelines,
        process_name=process_name, config_dictionary = config_dictionary)

def get_profile_pipelines_names(process_name):
  """
  Returns pipeline names for a given process, in case of process name being none
  that's generic process return generic pipelines
  """
  if process_name in profile_process_to_pipeline_dict:
    return profile_process_to_pipeline_dict[process_name]
  else:
    return profile_generic_pipelines

@catch_all(EmptyNavigatorPlugin())
def make_profile_plugin(process_name):
  """
  Creates a profile plugin that will collect profile entries and send them to
  the telemetry publisher only.
  """
  LOG.info("Creating profile plugin for process %s" % (process_name, ))
  for process_name_prefix, plugin in _profile_plugins:
    if process_name.startswith(process_name_prefix):
      LOG.info("Using specific profile plugin for process %s" % (process_name, ))
      return plugin(process_name)

  LOG.info("Using generic profile plugin for process %s" % (process_name, ))
  return ProfilePlugin(parse_generic_profile_event,
        "writeGenericProfileEvents", profile_generic_pipelines, process_name=process_name)


lineage_checksums = {}

def watch_gateway_roles(agent, response):
  """
  This function is responsible for enabling lineage collection for Gateway
  roles. Currently this function should only be used for Spark Lineage.

  If a "alt_name/navigator.lineage.client.properties" file (for ex:
  "spark-conf/navigator.lineage.client.properties") is received as part of
  the config files, we assume it would only be for Spark for 5.11+. Then,
  we validate properties file from "/etc/", which is the host's persistent
  location.

  If the lineage file is available under "/etc/alt_name/path/" (for ex:
  "/etc/spark/conf.cloudera.SPARK_ON_YARN-1/") and is of size > 0, then a Spark
  lineage plugin is created. If any of these two conditions evaluate to false,
  we would assume lineage is turned off and the plugin is stopped.
  """
  configs = response["client_configs"]
  for config in configs:
    conf_data = config["data"]
    alt_name = config["alt_name"]
    unzipped_files = ZipFile(StringIO(conf_data))
    lineage_file_name = os.path.join(alt_name, METADATA_CONFIG_FILENAME)
    try:
      lineage_file_info = unzipped_files.getinfo(lineage_file_name)
    except KeyError:
      # Skip further evaluation if the configs doesn't contain the
      # ALT_NAME/METADATA_CONFIG_FILENAME file.
      continue

    conf_dir = config["path"]
    conf_file = os.path.join(conf_dir, METADATA_CONFIG_FILENAME)
    service_name = os.path.basename(conf_dir)
    is_already_added = cmf.audit.metadata_thread.is_plugin_already_added(service_name)
    is_prop_file_exists = os.path.exists(conf_file)
    # Theoretically, if "navigator.lineage.client.properties" file is present in
    # the configs["data"], CM should always be copying this file under
    # "/etc/alt_name/conf...". Here, we conservatively check for the file's existence
    # before getting to its size. Logging an error to note that lineage will not be
    # generated, if the file went missing for some reason.
    if not is_prop_file_exists:
      LOG.error("Lineage file not found, skipping Spark plugin creation: %s", conf_file)
    is_prop_file_valid = is_prop_file_exists and os.path.getsize(conf_file) > 0

    # If there are changes to navigator.lineage.client.properties file, then we'll need
    # to re-read it.
    if is_already_added and is_prop_file_valid:
      digest = hashlib.sha256()
      digest.update(unzipped_files.open(lineage_file_info).read())

      # Look-up old checksums using the FS path (conf_file).  We don't care which client
      # config may be responsible for this file (thus we don't hash by lineage_file_name), What
      # matters is whether this data bundle would modify existing FS content.
      old_digest = lineage_checksums.get(conf_file, None)
      if old_digest != digest.hexdigest():
        cmf.audit.metadata_thread.refresh_csd_plugin(service_name, conf_dir)
        lineage_checksums[conf_file] = digest.hexdigest()

    if not is_already_added and is_prop_file_valid:
      LOG.info("Creating Spark plugin using: %s", conf_file)
      # Currently only Spark gateway should have {@code METADATA_CONFIG_FILENAME} file
      spark_plugin = MetadataPlugin(parse_spark_lineage_event, [""], "spark_lineage_log_",
                                  service_name, is_csd_process=True)
      spark_plugin.refresh(conf_dir)
    if is_already_added and not is_prop_file_valid:
      # If the plugin is already added and the nav property file is of size zero, then
      # stop the plugin
      cmf.audit.metadata_thread.stop_plugin(service_name)


def start_init(agent):
  """
  Called at agent start time to allow for HB callback registration.
  """
  agent.register_hb_callback(watch_gateway_roles)
