# Copyright (c) 2013 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import
import logging
import os
import psutil

from cmf.util import synchronized
from threading import RLock
from time import time

AUDIT_WAL_TMP_FILE_NAME = "audit_wal.tmp"
AUDIT_WAL_CHECKPOINT_SIZE = 524288  # 512 KB
CHECKPOINT_INTERVAL_SECONDS = 5 * 60
SEEK_END = 2
SEPARATOR = ","

LOG = logging.getLogger( __name__ )

class Wal(object):
  """
  A class to manage Write After Log that is used to save the last audit record
  that was successfully sent to audit server. The file is used to resume
  operation after agent is restarted. Each line the file has format:
    _offset,_file_name
  The _file_name is used instead of full path as it helps in moving audit log
  directory along with all the audit logs in it.
  """

  def __init__(self, log_dir, wal_filename, process_name):
    self._offset = None
    self._file_name = None
    self._walfd = None
    self._last_checkpoint_time = time()
    self._process_name = process_name
    self._nav_wal_lock = RLock()

    self._log_dir = log_dir
    self._init_wal_file(wal_filename)

  def get_lock(self):
    """
    Returns the lock that is used to synchronize methods in this class.
    """
    return self._nav_wal_lock

  @synchronized
  def _init_wal_file(self, wal_filename):
    self._wal_file = os.path.join(self._log_dir, wal_filename)
    if os.path.exists(self._wal_file):
      self._walfd = open(self._wal_file, 'a+')
      self._file_name, self._offset = self._get_init_file_info()
      # checkpoint in case we are recovering from a crash
      if self._file_name is None:
        self._walfd.truncate(0)
      else:
        self._checkpoint(self._file_name, self._offset)

  @synchronized
  def close(self):
    """
    Close the open fd to the Wal file if it is not None.
    """
    if self._walfd is not None:
      self._walfd.close()

  @synchronized
  def _get_init_file_info(self):
    """
    Read the WAL file for a valid wal line and get the audit log file and
    its offset.
    A valid line in WAL file is terminated by "\n". If the line isn't newline
    terminated we assume that something crashed resulting in partial line
    written to the file.
    """
    file_name, offset = None, None

    self._walfd.seek(0)
    lines = self._walfd.readlines()
    while lines:
      last_line = lines.pop()
      if self._is_valid_wal_line(last_line):
        file_name, offset = self._get_filename_offset_from_line(last_line)
        break

    return file_name, offset

  def _get_filename_offset_from_line(self, line):
    offset, file_name = line.strip().split(SEPARATOR)
    return file_name, int(offset)

  def _is_valid_wal_line(self, line):
    """
    Checks validity of a line read from wal file and if the format isn't correct
    returns False. If the line is correctly formatted, returns True.
    """
    return line and line[-1] == "\n"

  def get_file_info(self):
    """
    Return the information about the current audit log file and _offset in it
    from where we should start collecting audit events.
    """
    file_name = None
    if self._file_name is not None:
      file_name = self._file_name

    offset = None
    if self._offset is not None:
      offset = int(self._offset)
    return file_name, offset

  @synchronized
  def append(self, filename, offset):
    """
    Add information about audit event that was just sent to the audit server.
    Offset is the next audit line that needs to be sent to server.
    """
    if self._walfd is None:
      self._walfd = open(self._wal_file, 'a+')
      # We flush the wal file, to make sure it is written out. In case of csd plugins
      # when an end-marker is read, the corresponding wal file gets deleted. However if
      # it hasn't be explicity flushed out, it is possible that we try to delete a
      # non-existent file.
      self._walfd.flush()
      os.fsync(self._walfd.fileno())

    if not self._checkpoint_if_needed(filename, offset):
      self._append(filename, offset)

    self._file_name = filename
    self._offset = offset

  @synchronized
  def _append(self, filename, offset):
    print >> self._walfd, "%d%s%s" % (offset, SEPARATOR, filename)
    self._walfd.flush()
    os.fsync(self._walfd)

  def _checkpoint_if_needed(self, filename, offset):
    """
    Method to checkpoint current state in the audit log periodically. The
    checkpointing is only performed if the WAL file is more than
    AUDIT_WAL_CHECKPOINT_SIZE in size.
    Returns a boolean indicating if checkpointing condition was met and if
    the checkpointing was performed.
    """
    checkpoint_done = False
    current_time = time()
    if current_time - self._last_checkpoint_time > CHECKPOINT_INTERVAL_SECONDS:
      size_in_bytes = os.path.getsize(self._wal_file)
      if size_in_bytes > AUDIT_WAL_CHECKPOINT_SIZE:
        checkpoint_done = self._checkpoint(filename, offset)
        if checkpoint_done:
          self._last_checkpoint_time = current_time

    return checkpoint_done

  @synchronized
  def _checkpoint(self, filename, offset):
    """
    Method to checkpoint current state in the audit log. If this file isn't
    already open elsewhere by this process, this method creates a new file and
    writes the current state in it. It then renames this file to the wal file
    (relies on posix guarantee of atomic rename).
    Returns a boolean indicating if the checkpointing was performed.
    """
    if self._is_wal_file_open_elsewhere():
      LOG.info("Not checkpointing %s because it is open elsewhere" % (
          self._wal_file,))
      return False

    LOG.info("checkpointing %s (making new file %s for process %s)" % (filename,
        self._wal_file, self._process_name))

    tmp_wal_file = os.path.join(self._log_dir, AUDIT_WAL_TMP_FILE_NAME)

    # write current file offset to the file and then rename
    tmp_wal = open(tmp_wal_file, 'w')
    print >> tmp_wal, "%d%s%s" % (offset, SEPARATOR, filename)
    tmp_wal.close()

    # rename the file to _wal_file
    self._walfd.close()
    os.rename(tmp_wal_file, self._wal_file)
    self._walfd = open(self._wal_file, 'a+')

    return True

  def _is_wal_file_open_elsewhere(self):
    """
    Determine whether or not this process has this file open elsewhere (i.e. not
    just by this Wal class). We need this to prevent overwriting a file that
    another Wal class is writing to, because if that happens the file will
    remain semi-open (the fd will exist, but it will be marked "deleted" and
    will no longer points to a legit file), and it will continue to "write" to
    it without errors, but when that file is later closed its content will not
    be persisted.
    """
    for open_file in psutil.Process().open_files():
      if open_file.path == os.path.realpath(self._wal_file) and \
          open_file.fd != self._walfd.fileno():
        return True
    return False
