#!/usr/bin/env python
from __future__ import absolute_import
# Copyright (c) 2013 Cloudera, Inc. All rights reserved.
"""
Functionality for manipulating Linux cgroups.
"""

import grp
import logging
import os
import pwd
import re
import resource
import subprocess
import sys
import tempfile
import time

from cmf.util import user_group_to_ids

LOG = logging.getLogger(__name__)
PROC_CGROUPS = "/proc/cgroups"
PROC_MOUNTS = "/proc/mounts"
PROC_SELF_CGROUPS = "/proc/self/cgroup"

# Memory limit values that logically represent -1.
SPECIAL_MEMORY_VALUES = [
  0x7fffffffffffffffL, # max llong (2^63 - 1)
  0xffffffffffffffffL, # max ullong (2^64 - 1)
  ]

# TODO: Need to model better relationships between subsystems, resources,
# and resource conversions (see Process.evaluate_resources()).

# Maps between subsystems and maps of resources to cgroups pseudofiles.
# TODO: Consider pseodufile-level capabilities (e.g. Debian 6 has the cpu
# subsystem but is missing cpu.rt_runtime_us).
SUBSYSTEMS = {"blkio"   : {"weight"      : "blkio.weight"},
              "cpu"     : {"shares"      : "cpu.shares",
                           "rt_runtime"  : "cpu.rt_runtime_us"},
              "cpuacct" : {},
              "memory"  : {"hard_limit"  : "memory.limit_in_bytes",
                           "soft_limit"  : "memory.soft_limit_in_bytes"},
              "devices" : {}}
# Max cgroup memory limit.
# The value comes from the cgroup setup in the memory management layer; by default,
# it's set to PAGE_COUNTER_MAX, which is LONG_MAX / PAGE_SIZE on 64-bit platforms,
# and multiplied by PAGE_SIZE again when read.
CGROUP_MAX_MEM_LIMIT = (sys.maxsize / long(resource.getpagesize())) * long(resource.getpagesize())


def rchown(path, uid, gid):
  """
  chown -R uid:gid path
  """
  try:
    os.chown(path, uid, gid)
  except OSError, e:
    LOG.warn("Chown error on %s: %s", path, e)

  if not os.path.isdir(path):
    return

  dirs = [path]
  files = []
  while dirs:
    path = dirs.pop(0)
    for base, d, f in os.walk(path):
      for d1 in d:
        dirs.append(os.path.join(base, d1))
      for f1 in f:
        files.append(os.path.join(base, f1))
  for f in files + dirs:
    try:
      os.chown(f, uid, gid)
    except OSError, e:
      LOG.warn("Chown error on %s: %s", f, e)


def init_capabilities():
  """
  Returns an initialized cgroups capabilities dictionary. Should match the
  layout of the Avro heartbeat.
  """
  caps = dict()
  caps["has_blkio"] = False
  caps["has_cpu"] = False
  caps["has_cpuacct"] = False
  caps["has_memory"] = False
  caps["has_devices"] = False
  caps["default_blkio_weight"] = -1
  caps["default_cpu_shares"] = -1
  caps["default_cpu_rt_runtime_us"] = -1
  caps["default_memory_limit_in_bytes"] = -1      # in CM representation
  caps["default_memory_soft_limit_in_bytes"] = -1 # in CM representation
  caps["writable_cgroup_dot_procs"] = False
  return caps

class CGroupManager(object):
  """
  Encapsulates cgroups management functionality.
  """
  ORPHANED_CGROUPS_REMOVAL_WAIT_TIME_SECS = 15
  def __init__(self, path, cgroup_paths=dict(), sudo=None):
    # Root of cgroup mounts.
    self._path = path

    # List of mounted subsystems.
    self._mounted_subsystems = []

    # Map of cgroup paths to their reference counts.
    self._ref_counts = {}

    # List of cgroups we failed to delete the first time around.
    self._orphaned_groups = []

    # Used to ensure there is some wait period before we attempt to
    # remove orphaned cgroups.
    self._last_orphaned_cgroup_deferral_time = 0

    self._sudo = sudo

    # List of relevant cgroup subsystems that have already been mounted.
    self._existing_subsystems = self._find_existing_subsystems(SUBSYSTEMS.keys())

    # Map of subsystems to Agent cgroups
    self._agent_cgroups = self._find_agent_cgroups(SUBSYSTEMS.keys(), cgroup_paths)

    # Probe capabilities.
    self._empty_capabilities = init_capabilities()
    self._capable_subsystems = []
    self.capabilities = self._probe()
    if self.capabilities:
      for cap, value in self.capabilities.items():
        subsys = self._avro_cap_to_subsystem(cap)
        if subsys and value:
          self._capable_subsystems.append(subsys)

    # Make subsystem directories.
    for subsys in [c for c in self._capable_subsystems if c not in self._existing_subsystems]:
      subsys_path = os.path.join(path, subsys)
      if os.path.isdir(subsys_path):
        LOG.info("Reusing %s" % (subsys_path,))
      else:
        LOG.info("Created %s" % (subsys_path,))
        os.makedirs(subsys_path)

  def _do_mount(self, subsys, path):
    """
    Mount the desired cgroups subsystem at the desired path.

    Returns True if successful, False otherwise.
    """
    LOG.debug("Mounting cgroups subsystem %s to %s" % (subsys, path))
    args = ["mount", "-t", "cgroup", "-o", subsys, "cm_cgroups", path]
    if self._sudo:
      args.insert(0, self._sudo)

    popen = subprocess.Popen(args, stderr=subprocess.PIPE)
    _, stderr = popen.communicate()
    if popen.returncode:
      LOG.info("Failed to mount cgroups subsystem %s to %s, rc: %s stderr: %s"
               % (subsys, path, popen.returncode, stderr))
      return False

    # Unlike regular file systems, cgroups doesn't support using the uid/gid
    # options to the mount command so we have to chown after mounting in
    # non-root mode.
    if self._sudo:
      uid, gid = os.getuid(), os.getgid()
      user, group = pwd.getpwuid(uid)[0], grp.getgrgid(gid)[0]
      popen = subprocess.Popen([self._sudo, "chown", "-R", "%s:%s" % (user, group), path], stderr=subprocess.PIPE)
      _, stderr = popen.communicate()
      if popen.returncode:
        LOG.info("Failed to change ownership of cgroups subsystem %s on %s to %s:%s, rc: %s stderr: %s"
                 % (subsys, path, user, group, popen.returncode, stderr))
        self._do_umount(path)

    return popen.returncode == 0

  def _do_umount(self, path):
    """
    Unmount the cgroups subsystem found at the desired path.

    Returns True if successful, False otherwise.
    """
    LOG.debug("Unmounting cgroups subsystem from %s" % (path,))
    args = ["umount", path]
    if self._sudo:
      args.insert(0, self._sudo)
    popen = subprocess.Popen(args, stderr=subprocess.PIPE)
    _, stderr = popen.communicate()
    if popen.returncode:
      LOG.info("Failed to unmount cgroups subsystem from %s, rc: %s stderr: %s"
               % (path, popen.returncode, stderr))
    return popen.returncode == 0

  def _subsystem_to_avro_cap(self, subsystem):
    """
    Returns the converted result of a subsystem name into an avro capability,
    or None if the subsystem isn't represented by a capability.
    """
    assert subsystem in SUBSYSTEMS
    cap_key = "has_" + subsystem
    if cap_key in self._empty_capabilities:
      return cap_key
    LOG.warning("Ignoring cgroup subsystem: %s" % subsystem)
    return None

  def _avro_cap_to_subsystem(self, avro_cap):
    """
    Returns the converted result of an avro capability into a subsystem name,
    or None if the capability does not represent a subsystem.
    """
    assert avro_cap in self._empty_capabilities
    subsys = re.sub("^has_", "", avro_cap)
    if subsys in SUBSYSTEMS:
      return subsys
    LOG.info("Ignoring cgroup avro_cap: %s" % avro_cap)
    return None

  def _memory_cm_to_cgroup(self, value):
    """
    Transforms memory resource values from their CM representation to what
    the cgroups memory subsystem expects.

    Returns the converted value.
    """
    if value != -1:
      # Convert from megabytes to bytes.
      value *= (1024 * 1024)
    return value

  def _memory_cgroup_to_cm(self, value):
    """
    Transforms memory resource values from their cgroup memory subsystem
    representation to what CM expects.

    Returns the converted value.
    """
    assert value >= 0
    if value in SPECIAL_MEMORY_VALUES:
      value = -1
    else:
      # Convert from bytes to megabytes.
      value /= (1024 * 1024)
    return value

  def _read_cgroup_value(self, subsys, pf):
    """
    Returns a value read from an opened cgroup pseudofile, converting
    into CM representation if necessary.
    """
    assert not pf.closed
    value = int(pf.read().strip())
    if subsys == "memory":
      value = self._memory_cgroup_to_cm(value)
    return value

  def _write_cgroup_value(self, subsys, pf, value):
    """
    Writes a value to an opened cgroup pseudofile, converting into
    cgroup representation if necessary.
    """
    assert not pf.closed
    if subsys == "memory":
      value = self._memory_cm_to_cgroup(value)
    pf.write(str(value))

  def _check_path(self, path):
    """
    Checks that the given path is within a mounted subsystem.

    We don't bother resolving aliases and links because it is assumed that
    the path came from get_group_paths.
    """
    for subsys in self._mounted_subsystems:
      subsys_root = os.path.join(self._path, subsys)
      if path.startswith(subsys_root):
        return
    for existing in self._existing_subsystems.values():
      if path.startswith(existing):
        return
    raise Exception("cgroup path %s not in mounted or existing subsystems" % (path,))

  def _create_group(self, path, user=None, group=None, mode=None, name=None, type=None):
    """
    Actually create a cgroup by path.
    """
    if not os.path.isdir(path):
      LOG.info("Creating cgroup %s" % (path,))
      os.mkdir(path)
    else:
      LOG.info("Reusing cgroup %s" % (path,))

    # Update permissions. Not needed in non-root mode (when using sudo) since
    # all processes run as the same user as the agent.
    if not self._sudo and user is not None and group is not None:
      uid, gid = user_group_to_ids(user, group)
      rchown(path, uid, gid)
    if mode is not None:
      os.chmod(path, mode)

    try:
      self._orphaned_groups.remove(path)
      LOG.info("Cancelled deferred destruction of cgroup %s", path)
    except ValueError:
      pass

  def _defer_group_removal(self, path):
    """
    Add a cgroup by path to be removed later.
    """
    LOG.info("Deferring destruction of cgroup %s", path)
    self._orphaned_groups.append(path)
    self._last_orphaned_cgroup_deferral_time = time.time()

  def _destroy_group(self, path, defer=False):
    """
    Actually destroy a cgroup by path.

    If defer is True, failure to destroy the group will move it to the
    orphaned groups list.
    """
    if os.path.isdir(path):
      LOG.info("Destroying cgroup %s" % (path,))
      try:
        # Destroy any leftover sub-cgroups first.
        #
        # Normally we wouldn't expect any, but processes may have forgotten
        # to clean up after themselves, or may have been shutdown abruptly.
        for root, dirs, _ in os.walk(path, topdown=False):
          for d in dirs:
            os.rmdir(os.path.join(root, d))
        os.rmdir(path)
      except:
        LOG.exception("Could not destroy cgroup %s" % path)
        if defer:
          self._defer_group_removal(path)
    else:
      LOG.info("The cgroup %s is already destroyed" % (path,))

  def _probe_subsystem(self, subsystem, probe_writable_cgroup_dot_procs,
                       pseudofiles):
    """
    Probe a cgroup subsystem to establish its capabilities.

    If the subsystem is not already mounted, this will mount it in a temporary
    location.

    :param subsystem: The name of the subsystem to probe
    :param probe_writable_cgroup_dot_procs: Whether to check if cgroup.procs is
                                            writable or not.
    :param pseudofiles: List of pseudofiles in the subsystem to check for.
    """

    caps = dict()
    if subsystem in self._existing_subsystems:
      d = self._existing_subsystems[subsystem]
      clean = False
    else:
      d = tempfile.mkdtemp()
      clean = True
    try:
      if clean and not self._do_mount(subsystem, d):
        return
      try:
        cap_key = self._subsystem_to_avro_cap(subsystem)
        caps[cap_key] = True

        # Probe the pseudofiles.
        for pf in pseudofiles:
          # Should match the Avro heartbeat.
          cap_key = "default_" + pf.replace(".", "_")

          pf_path = os.path.join(d, pf)
          if not os.path.exists(pf_path):
            LOG.info("cgroup pseudofile %s does not exist, skipping" %
                     (pf_path,))
            continue
          LOG.debug("Opening cgroups pseudofile %s" % (pf_path))
          opened_pf = open(pf_path)
          try:
            caps[cap_key] = self._read_cgroup_value(subsystem, opened_pf)
          finally:
            opened_pf.close()
        if probe_writable_cgroup_dot_procs:
          cg_dot_procs_path = os.path.join(d, "cgroup.procs")
          LOG.debug("Testing cgroup.procs %s" % (cg_dot_procs_path))
          caps["writable_cgroup_dot_procs"] = os.access(cg_dot_procs_path,
                                                        os.W_OK)
      finally:
        if clean:
          self._do_umount(d)
    finally:
      if clean:
        os.rmdir(d)

    return caps

  def _probe(self):
    """
    Probe the system for cgroups capabilities.

    Temporarily mounts one subsystem at a time to fully probe capabilities
    (note: we can mount them all together in one mount, but that fails if
    cgroups are already mounted somewhere on the system.

    Returns a dictionary with the results, or None if no support is found.
    """

    # Look for /proc/cgroups
    if not os.path.exists(PROC_CGROUPS):
      LOG.info("Cannot find %s; no cgroups capabilities found" %
                (PROC_CGROUPS,))
      return None

    # We need to see whether cgroup.procs is writable. We can do this on
    # any mounted subsystem, so we'll try on the first one we find.
    probe_writable_cgroup_dot_procs = True

    # Prepare the /proc/cgroups subsystem regex matchers.
    subsys_matchers = dict()
    for subsys in SUBSYSTEMS:
      matcher = re.compile("%s\s+" % (subsys,))
      subsys_matchers[matcher] = subsys

    caps = init_capabilities()
    args = ["cat", PROC_CGROUPS]
    if self._sudo:
      args.insert(0, self._sudo)
    try:
      popen = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
      stdout, stderr = popen.communicate()
      if popen.returncode:
        LOG.warn("Could not read capabilities from %s. error: %s" % (PROC_CGROUPS, stderr))
        return None
    except Exception, e:
      LOG.warn("Failed probing cgroup capabilities: %s", e)
      return None

    # Look for relevant subsystems in /proc/cgroups. When we find one,
    # mount it temporarily to extract other capabilities.
    for line in stdout.splitlines():
      more_caps = None
      for matcher, subsys in subsys_matchers.items():
        if matcher.match(line):
          LOG.info("Found cgroups subsystem: %s" % (subsys,))
          more_caps = self._probe_subsystem(subsys,
                                            probe_writable_cgroup_dot_procs,
                                            SUBSYSTEMS[subsys].values())
          probe_writable_cgroup_dot_procs = False
          if more_caps:
            LOG.info("Adding more caps: %s" % str(more_caps))
            caps = dict(caps.items() + more_caps.items())

    return caps

  def _find_existing_subsystems(self, relevant_subsystems):
    """
    Identify which cgroup subsystems are already mounted on the host.
    :param relevant_subsystems: Which subsystems do we care about
    :returns: Dictionary of subsystems to mount points.
    """

    ret = dict()
    try:
      fd = open(PROC_MOUNTS, 'r')
      try:
        for line in fd:
          elements = line.split(' ')
          if len(elements) < 4:
            LOG.debug("Ignoring malformed mount entry: '%s'" % (line, ))
            continue
          device = elements[0]
          mount = elements[1]
          fstype = elements[2]
          options = elements[3].split(',')
          if device in ['cgroup', 'cm_cgroups']:
            for o in options:
              if o in relevant_subsystems:
                LOG.info("Found existing subsystem %s at %s" % (o, mount))
                ret[o] = mount
      finally:
        fd.close()
    except IOError:
      LOG.exception("Failed to check for existing mount cgroup subsystems")
    return ret

  def _find_agent_cgroups(self, relevant_subsystems, cgroup_paths=dict()):
    """
    Identify the cgroups that the Agent has been placed in.
    :param relevant_subsystems: Which subsystems do we care about
    :returns: Dictionary of subsystems to cgroups.
    """
    ret = dict()
    if not os.path.exists(PROC_SELF_CGROUPS):
      LOG.info("System does not support cgroups")
      return ret
    try:
      fd = open(PROC_SELF_CGROUPS, 'r')
      try:
        for line in fd:
          elements = line.split(':', 2)
          if len(elements) != 3:
            LOG.debug("Ignoring malformed cgroup entry: '%s" % (line, ))
            continue

          subsystems = elements[1].split(',')
          # Strip off leading '/' as it breaks os.path.join() and trailing '\n'
          path = elements[2][1:].rstrip('\n')

          for s in subsystems:
            if s in relevant_subsystems:
              if s in cgroup_paths and cgroup_paths[s]:
                # Strip off leading '/'
                path = cgroup_paths[s]
                if path is not None and path != "":
                  path = path[1:]
              if path is None:
                path = ""
              ret[s] = path
      finally:
        fd.close()
    except IOError:
      LOG.exception("Failed to read Agent's current cgroups")
    return ret

  def _read_long_from_file(self, path):
    return long(self._read_from_file(path))

  def _read_from_file(self, path):
    data = "0"
    with open(path, 'r') as myfile:
      data = myfile.read().replace('\n', '')
    return data

  def enable(self):
    """
    Mount all supported cgroups subsystems that are not already mounted.
    """

    # Do we need to mount anything?
    subsys_needing_mount = ((set(self._capable_subsystems) -
                             set(self._mounted_subsystems)) -
                            set(self._existing_subsystems.keys()))
    if len(subsys_needing_mount) == 0:
      return

    LOG.debug("Need to mount %s" % (subsys_needing_mount,))

    # Mount whatever is left.
    for subsys in subsys_needing_mount:
      subsys_path = os.path.join(self._path, subsys)
      LOG.info("Mounting subsystem %s at %s" % (subsys, subsys_path))
      if not self._do_mount(subsys, subsys_path):
        continue
      self._mounted_subsystems.append(subsys)

  def disable(self):
    """
    Unmount all mounted subsystems.
    """

    # Do we need to unmount anything?
    if len(self._mounted_subsystems) == 0:
      return

    LOG.debug("Need to unmount %s" % (self._mounted_subsystems,))

    # We think we need to unmount something, but let's double-check in case
    # someone else unmounted something.
    subsys_needing_unmount = list()
    existing = self._find_existing_subsystems(SUBSYSTEMS.keys()).keys()
    for e in existing:
      if e in self._mounted_subsystems:
        LOG.debug("Found %s mounted" % (e, ))
        subsys_needing_unmount.append(e)

    # Remove the mount from our list even if unmount fails.
    for subsys in subsys_needing_unmount:
      subsys_path = os.path.join(self._path, subsys)
      LOG.info("Unmounting subsystem %s at %s" % (subsys, subsys_path))
      self._do_umount(subsys_path)
      self._mounted_subsystems.remove(subsys)

  def create_group(self, path, name=None, type=None, user=None, group=None,
                   mode=None, do_create=True):
    """
    Creates a new cgroup using the given path and, optionally, user, group, and mode.

    If do_create is False, only the reference counts are updated.
    """
    if do_create:
      self._check_path(path)

    count = self._ref_counts.get(path)
    if count is not None:
      self._ref_counts[path] = count + 1
      return
    if do_create:
      self._create_group(path, user, group, mode, name=name, type=type)
    self._ref_counts[path] = 1

  def destroy_group(self, path):
    """
    Destroys a cgroup with the given path. If there's a problem in cleanup,
    the cgroup is moved to the orphaned list (cleaned up via
    destroy_orphaned_groups).
    """
    self._check_path(path)

    count = self._ref_counts.get(path)
    assert count and count > 0

    if count > 1:
      self._ref_counts[path] = count - 1
      return

    del self._ref_counts[path]
    self._destroy_group(path, defer=True)

  def configure_group(self, path, subsys, resource):
    """
    Configures a cgroup with the given path using the given subsystem and
    resource dictionary (name to value).

    TODO: Dynamic resources will incur a pseudofile open() + read() with
    every heartbeat. This can be avoided by caching the current value and
    rereading it at startup.

    Returns True if the group was reconfigured, False otherwise.
    """
    self._check_path(path)

    assert path in self._ref_counts
    assert len(resource) > 0

    for resource_name, resource_value in resource.items():
      pf = SUBSYSTEMS[subsys][resource_name]
      pf_path = os.path.join(path, pf)
      if not os.path.exists(pf_path):
        LOG.debug("cgroup pseodofile %s does not exist, skipping" %
                  (pf_path,))
        continue
      f = open(pf_path, "r+")
      try:
        # Only update if the old value doesn't match the new.
        oldval = self._read_cgroup_value(subsys, f)
        if (oldval == resource_value):
          LOG.debug("Values for cgroup pseudofile %s match (%d)" %
                    (pf_path, resource_value))
        else:
          LOG.info("Reconfiguring cgroup pseudofile %s with value %d" %
                   (pf_path, resource_value))
          self._write_cgroup_value(subsys, f, resource_value)
      finally:
        f.close()

    return True

  def destroy_orphaned_groups(self):
    """
    Destroys any orphaned cgroups. Only one attempt is made, though failures
    are non-fatal.
    """
    num_orphaned_groups = len(self._orphaned_groups)
    if num_orphaned_groups == 0:
      return

    now = time.time()
    if (self._last_orphaned_cgroup_deferral_time + self.ORPHANED_CGROUPS_REMOVAL_WAIT_TIME_SECS) <\
            now:
      LOG.info("Removing orphaned cgroups, total %d last_orphaned_cgroup_deferral_time %f "
               "now %f", num_orphaned_groups, self._last_orphaned_cgroup_deferral_time, now)
      for path in self._orphaned_groups:
        self._destroy_group(path)
      self._orphaned_groups = []
    else:
      LOG.info("Skipping removal of orphaned cgroups; total %d last_orphaned_cgroup_deferral_time %f "
               "now %f", num_orphaned_groups, self._last_orphaned_cgroup_deferral_time, now)

  def has_at_least_one_group(self):
    """
    Returns True if there are any extant cgroups, False otherwise.

    Includes orphaned cgroups, as they are still considered to be under
    active management (at least until the next call to
    destroy_orphaned_groups).
    """
    return len(self._ref_counts) > 0 or len(self._orphaned_groups) > 0

  def get_group_paths(self, name, subsystem=None, mounted=True):
    """
    Returns a list of cgroups that are applicable for the given name. These
    cgroups need not yet exist.

    If subsystem is provided, the list is constrained to only that subsystem.

    If mounted is True, the list is constrained to only mounted subsystems.
    """
    assert not name.startswith("/")
    groups = []

    if subsystem:
      candidates = [subsystem]
      LOG.info("Get group path for %s / %s" % (name, subsystem))
    else:
      candidates = self._capable_subsystems

    # Identify pre-existing subsystems
    for c in candidates:
      if c in self._existing_subsystems:
        groups.append(os.path.join(self._existing_subsystems[c],
                                  self._agent_cgroups.get(c, ""), name))
      else:
        LOG.info("Candidate subsytem not found in existing subsystems: %s" % c)

    # Identify remaining mounted/mountable subsystems
    if mounted:
      all_subs = [c for c in candidates if c in self._mounted_subsystems]
    else:
      all_subs = [c for c in candidates if c not in self._existing_subsystems]
    for subsys in all_subs:
      groups.append(os.path.join(self._path, subsys,
                                self._agent_cgroups.get(subsys, ""), name))
      LOG.info("Group path is %s" % groups[-1])
    return groups

  def get_cgroup_mem_info(self):
    """
    Returns the memory limit and usage from cgroup
    """
    memory_subsystem = self.get_group_paths("", "memory")[0]
    mem_limit = self._read_long_from_file(memory_subsystem + "/memory.limit_in_bytes")
    # Check if mem_limit is set to unlimited. If yes, use the regular method
    if mem_limit == CGROUP_MAX_MEM_LIMIT or mem_limit <= 0:
      return None, None
    return mem_limit, self.get_memory_used()

  def get_memory_used(self):
    """
    Returns the memory used from memory subsystem.
    """
    memory_subsystem = self.get_group_paths("", "memory")[0]
    return self._read_long_from_file(memory_subsystem + "/memory.usage_in_bytes")

  def get_cpu_shares(self):
    """
    Returns the set cpu shares from cpu subsystem.
    """
    memory_subsystem = self.get_group_paths("", "cpu")[0]
    return self._read_long_from_file(memory_subsystem + "/cpu.shares")

  def find_group_path(self, paths, subsys, name):
    """
    If the cgroup path with the given name and subsystem exists in paths,
    returns it. Otherwise, returns None.
    """

    agent_cgroup = self._agent_cgroups.get(subsys, "")

    if subsys in self._existing_subsystems:
      candidate = os.path.join(self._existing_subsystems[subsys], agent_cgroup,
                               name)
    else:
      candidate = os.path.join(self._path, subsys, agent_cgroup, name)

    if candidate in paths:
      return candidate
    else:
      return None

  def get_mounts(self):
    """
    Returns a map (subsystem -> path) of the mount points.
    """
    ret = dict()
    for subsys in self._mounted_subsystems:
      ret[subsys] = os.path.join(self._path, subsys)
    ret.update(self._existing_subsystems)
    return ret

  def get_agent_cgroups(self):
    """
    Returns the Agent's cgroups
    """
    return self._agent_cgroups.copy()
