# Copyright (c) 2014 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import
#
# Code for deploying client configurations and interacting with the system
# alternatives database (/etc/alternatives).
#
# BACKGROUND
#
# A client config is a bundle of configuration files that, when placed in a
# well-known location on the host, provide needed configuration for client
# applications. Some client configs are explicitly surfaced to users via the
# host's alternatives database. Any number of configs may be registered as
# "alternatives" for a single generic link. For example, /etc/hadoop/conf1
# and /etc/hadoop/conf2 are both alternatives for /etc/hadoop/conf, and the
# alternatives database ensures that an /etc/hadoop/conf symlink exists and
# points to whichever alternative has the higher priority.
#
# OVERALL FLOW
#
# Client configs are managed by the CM server and pushed to the agent during
# a heartbeat request/response cycle. The heartbeat response contains a list
# of client configs that the server wishes the agent to deploy. The agent is
# responsible for comparing that list to the current state of the world and
# effecting necessary changes (i.e. deploying new client configs or deleting
# old ones). These changes are assumed to be small in scope and thus are
# performed synchronously. In the next heartbeat request the agent includes
# basic status information so that the server knows what happened.
#
# IMPLEMENTATION NOTES
#
# When necessary, the "state of the world" is reconstructed via the system
# alternatives database. This allows us to piggyback on an existing, robust
# system for persistence and simplifies the overall implementation. It's not
# without its drawbacks, however. There is no record of alternatives-less
# client configs (i.e. cluster client configs). They get redeployed every
# time the agent starts up and receives its first heartbeat response. This
# isn't ideal, but as deployment is supposed to be idempotent and low
# impact, it isn't fatal.
#
# For better performance the state of the world is read from disk only when
# the agent starts up, and is updated whenever the agent has made changes to
# the system. Any changes made outside the agent won't be noticed until the
# agent is restarted.
#
# Besides omitting cluster client configs altogether, entries in the
# alternatives database insufficiently describe a client config. For
# example, the outcome of deploying a client config (i.e. exit status) isn't
# recorded as part of the alternative. As such, on-disk client config is
# augmented somewhat:
# 1. A generation file is included in the deployed state, containing the
#    generation of the client config. Together with the alternative name and
#    path, the pre-deployed client config state can be precisely located
#    after the fact.
# 2. Log files and additional metadata are kept in the pre-deployed state.
#
# In-memory state consists of two sets: deployed client configs and
# deployment attempts.
#
# 1. The deployed set includes only those client configs that were
#    successfully deployed. It is reconstructed from system alternatives at
#    agent start time (see above) and used to decide what to deploy and what
#    to remove during a heartbeat response.
# 2. The attempts set also includes client configs that failed to deploy.
#    Like the deployed set, it is reconstructed from system alternatives,
#    but is used to provide client config status information to the server
#    during a heartbeat request.
#
# Cluster client configs present a further challenge: they are deployed into
# shared directories (like /etc), often times overwriting existing files.
# This makes it very difficult to safely delete deployed client configs; how
# are the overwritten files to be restored? To be safe, deletion is reduced
# to deregistering alternatives.
#
# CM-managed alternatives are distinguished from unmanaged alternatives via
# "conf.cloudera.", a magic string that prefixes every managed alternative's
# path name. Unmanaged alternatives are never deregistered.
#
# UPGRADE PROVISIONS
#
# A special provision is made to ease upgrades from older versions of CM:
# deployed client configs are deleted iff they include valid generations.
#
# After a CM upgrade but before an admin has redeployed client configs, the
# server-side client config state may be incomplete or missing entirely. If
# the agent were to immediately rectify using this state, some or all
# deployed client configs would be deleted, a most unpleasant outcome. By
# restricting deletion to client configs with generations, we only delete
# client configs that were created after the upgrade, not before.
#
# Note that a generation-less client config will be included in the "state
# of the world" when the agent restarts until 1) it is overwritten by a new
# generation-ful client config deployment with the same identity, or 2) it
# is deleted out-of-band by the admin.
#
# This provision isn't without its risks: any client configs deleted on the
# CM server following the upgrade prior to redeployment will be orphaned and
# will remain on the agent host until deleted out-of-band by the admin. This
# edge case is a reasonable tradeoff because it allows all upgrades to
# maintain client config function.

import logging
import os
import shutil
import simplejson as json
import tempfile

from cmf.subprocess_timeout import subprocess_with_timeout
from cmf.throttling_logger import ThrottlingLogger
from cmf.util import fix_hcat_home
from cmf.util import fix_hue_home
from cmf.util import try_tail
from cmf.util import unzip
from cmf.util.lazy import lazy
import cmf.process


LOG = logging.getLogger(__name__)
THROTTLED_LOG = ThrottlingLogger(LOG)

# The magic file that contains the current generation of a particular
# deployed client config.
GENERATION_FILENAME = "__cloudera_generation__"

# The magic generation value that indicates an old deployed client config
# that predates generations
NO_GENERATION_FOUND = -1

# The file that contains additional metadata about a deployment attempt.
METADATA_FILENAME = "__cloudera_metadata__"

# Client config directories that begin with this prefix are assumed to have
# been deployed by CM.
CM_MAGIC_PREFIX = "conf.cloudera."

# Maps Avro dictionary keys (from ClientConfigData in master_agent.avdl) to
# the environment variable names expected by deploy-cc.sh.
AVRO_TO_ENV_GENERIC = {
  "alt_link" : "ALT_LINK",
  "alt_name" : "ALT_NAME",
  "cdh_version" : "CDH_VERSION",
  "directory_name" : "DIRECTORY_NAME",
  "path" : "DEST_PATH",
  "priority" : "PRIORITY",
  "runner_program" : "RUNNER_PROGRAM",
}

# Maximum time (in seconds) to allow the deployment script to run before
# timing out.
DEPLOY_SCRIPT_TIMEOUT_SEC = 120

class ClientConfigKey(object):
  """
  A unique identifier for a client config.

  The name is the client config's identifier in the system alternatives
  database. It may be empty if there is no relationship with alternatives
  (i.e. if this is a cluster client config).

  The path is the location on disk where the client config is deployed.

  Together with ClientConfigValue, comprises a complete client config.
  """
  def __init__(self, name, path):
    self.name = name
    self.path = path

  def as_tuple(self):
    return (self.name, self.path)

  def __hash__(self):
    return hash(self.as_tuple())

  def __eq__(self, other):
    return (type(self) == type(other) and
            self.as_tuple() == other.as_tuple())

  def __repr__(self):
    return "<%s,%s>" % self.as_tuple()

class ClientConfigValue(object):
  """
  Additional metadata for a client config.

  The priority is the alternative priority for this client config in the
  system alternatives database. It is -1 if there is no relationship with
  alternatives (i.e. if this is a cluster client config).

  The generation is an opaque value that is updated by the CM server
  when a client config is changed. It reflects the contents of the client
  config, and wil only change if the client config materially changes.

  The remaining fields describe the result of deploying the client config
  and are always loaded from disk, never set directly.

  Together with ClientConfigKey, comprises a complete client config (though
  ideally we'd also include the alternative link, but
  "update-alternatives --display" doesn't provide that and
  "update-alternatives --query" isn't available on all distros).
  """

  # Object identity is interesting. We don't want values to be hashed;
  # that's what ClientConfigKey is for. Additionally, only generation should
  # factor into equality; that is the indicator for a material change.
  __hash__ = None

  def __init__(self, priority, generation):
    self.priority = priority
    self.generation = generation

    # These post-deployment fields are loaded from disk and are optional.
    self.stdout_tail = None
    self.stderr_tail = None
    self.exit_code = None
    self.parcels_in_use = None

  def __eq__(self, other):
    return (type(self) == type(other) and
            self.generation == other.generation)

  def __repr__(self):
    return "<%s,%s>" % (self.priority, self.generation)

  def load(self, deploy_path, conf_path):
    """
    Loads deployment results from various files.
    The deploy client config additional metadata are loaded from two different
    locations. The 'parcels_in_use' is from the config directory , for example
    /etc/sqoop/conf.cloudera.SQOOP_CLIENT-1/__cloudera_metadata__.
    The 'exit_code' is from the deployment directory such as:
    /var/run/cloudera-scm-agent/process/ccdeploy_xxxx/__cloudera_metadata__.
    """

    # Load log tails. It is assumed that the log contents are within
    # <path>/logs/stdout and <path>/logs/stderr.
    self.stdout_tail = try_tail(os.path.join(deploy_path, "logs", "stdout.log"))
    self.stderr_tail = try_tail(os.path.join(deploy_path, "logs", "stderr.log"))

    try:
      # Load the deploy result.
      pm = open(os.path.join(deploy_path, METADATA_FILENAME), "r")
      try:
        persistent_metadata = json.load(pm)
        self.exit_code = persistent_metadata["exit_code"]
        # This is mainly for upgrade scenario where the agent is restarted and
        # tries to load the parcels_in_use in the former metadata file. If there
        # is no deployment CC before the loading, this metadata will still be
        # in the original state (having exit_code and parcels_in_use). We should
        # get it from here instead of retrieving from the new metadata in the
        # config dir under /etc because the new metadata is still not created.
        # Once deploys client config, this metadata data will be overwritten to
        # only have exit_code. And the parcels_in_use will be written to the new
        # metadata file under conf dir. The load() will get the parcels_in_use
        # in the new metadata from then on.
        if "parcels_in_use" in persistent_metadata:
          self.parcels_in_use = persistent_metadata["parcels_in_use"]
          return
      finally:
        pm.close()
    except:
      THROTTLED_LOG.exception("Could not load post-deployment data from %s",
                              deploy_path)

    try:
      # Load the additional metadata.
      deploy_pm = open(os.path.join(conf_path, METADATA_FILENAME), "r")
      try:
        persistent_metadata = json.load(deploy_pm)
        self.parcels_in_use = persistent_metadata["parcels_in_use"]
      finally:
        deploy_pm.close()
    except:
      THROTTLED_LOG.exception("Could not load post-deployment data from %s",
                              conf_path)

class ClientConfigs(object):

  def __init__(self, agent, package_dir=None, deploy_dir=None,
               parcel_repo=None, sudo=None, deploy_env=None):
    """
    Creates an in-memory client-config handling database.

    'package_dir': path to where the deploy script can be found.
    'deploy_dir': path to where deployment runs should dump any temporary
                  output and logs.
    'sudo': path to the sudo binary that should be used before
            subcommands, if necessary.
    'deploy_env': environment variables to inject into deployment runs.
    """
    self.agent = agent
    self.osops = agent.os_ops

    # Static data.

    if not package_dir:
      package_dir = os.path.dirname(__file__)
    self.deploy_script = os.path.join(package_dir, "client", "deploy-cc.sh")
    if deploy_dir:
      self.deploy_dir = deploy_dir
    else:
      self.deploy_dir = os.path.dirname(__file__)
    self.parcel_repo = parcel_repo
    self.sudo = sudo
    if deploy_env:
      self.deploy_env = deploy_env
    else:
      self.deploy_env = {}
    self.java_home_config = None

    # Mutable data.

    # Deployed client configs (ClientConfigKey -> ClientConfigValue).
    #
    # First populated by load() and subsequently is a subset of the
    # client configs last passed to rectify() (namely, all those that were
    # successfully deployed).
    self.deployed = {}

    # Deployment attempts (ClientConfigKey -> ClientConfigValue).
    #
    # First populated by load() and subsequently contains an entry for each
    # client config last passed to rectify().
    self.results = {}

  @lazy
  def bin_path(self):
    return self.osops.alternatives.alt_bin

  def _read_generation(self, alt_path):
    """
    Reads and returns the generation from the generation file in 'alt_path'.

    Returns NO_GENERATION_FOUND if there was an error reading from the file.
    """
    filename = os.path.join(alt_path, GENERATION_FILENAME)
    try:
      f = open(filename, "r")
      try:
        return int(f.read().strip())
      finally:
        f.close()
    except:
      THROTTLED_LOG.exception("Could not read client config generation in %s",
                              alt_path)
      return NO_GENERATION_FOUND

  def _write_generation(self, alt_path, generation):
    """
    Writes 'generation' into the generation file in 'alt_path', overwriting
    any existing file.

    Failures are non-fatal because generations are merely an optimization
    to prevent excessive deployment.
    """
    filename = os.path.join(alt_path, GENERATION_FILENAME)
    try:
      f = open(filename, "w")
      try:
        f.write(str(generation) + "\n")
      finally:
        f.close()
    except:
      THROTTLED_LOG.exception("Could not write client config generation to %s",
                              alt_path)


  def _lookup_alternatives(self, alt_name):
    """
    Uses update-alternatives to find all alternatives for a particular name.
    Returns a dictionary mapping ClientConfigKey to ClientConfigValue.

    Alternatives not managed by CM are ignored.
    """
    alternatives = self.osops.alternatives.lookup(alt_name)
    ret = {}
    for (path, priority) in alternatives:
      # Ignore the alternative if it's not managed by CM.
      if CM_MAGIC_PREFIX not in os.path.basename(path):
        continue

      key = ClientConfigKey(alt_name, path)
      value = ClientConfigValue(priority, self._read_generation(path))
      ret[key] = value

    return ret

  def _deploy_client_config(self, cc):
    """
    Deploys the given client config.

    Returns the deployment path created.

    May throw an exception in the event of an error.
    """

    # Create a new deployment directory and its logging subdirectory.
    #
    # The deploy script will be executed in this directory.
    new_cwd, log_subdir = self._create_and_backup_deploy_dirs(cc["alt_name"],
                                                              cc["path"],
                                                              cc["generation"])

    # Go through the (laborious) task of setting up a proper environment for
    # the deploy script. It should look just like a process environment.
    env, parcels_in_use = self._adapt_cc_to_env(self.deploy_env, cc)
    env["CONF_DIR"] = new_cwd

    # If there is a custom JAVA_HOME setting, inject into the deployCC
    # environment
    if self.java_home_config is not None:
      env["JAVA_HOME"] = self.java_home_config

    fix_hcat_home(env)
    fix_hue_home(env)

    # Make the configuration data available.
    #
    # We also broaden the permissions somewhat as unzip() runs with a
    # restrictive umask. We only really need directories to be +x because
    # deploy-cc.sh will make all files world-readable, but this logic is
    # more consistent with process permission management.
    unzip(cc["data"], new_cwd)
    for root, dirs, files in os.walk(new_cwd):
      for f in files:
        os.chmod(os.path.join(root, f), 0640)
      for d in dirs:
        os.chmod(os.path.join(root, d), 0751)

    # Write out the generation file. It needs to go in the configuration
    # data subdirectory so that the script will copy it to the deployed
    # directory.
    self._write_generation(os.path.join(new_cwd, cc["directory_name"]),
                           cc["generation"])

    # Create resources
    if "resources" in cc:
      self.evaluate_resources(cc["resources"])

    def editor_fn(data):
      # Each key is something like "{{CMF_CONF_DIR}}/truststore.jks"
      # Each value is a real file path corresponding to the "logical" name represented by the key.
      for (k, v) in cmf.process.Process.known_resources_files.iteritems():
        data = data.replace(k, v)
      return data

    # Perform any required substitutions - the files we've received are incomplete.
    # They may contain invalid paths that require substitution.
    for path in self.agent.os_ops.file_tree_walk(new_cwd):
      if not os.path.isfile(path):
        continue
      self.agent.os_ops.edit_file(path, editor_fn)

    stdout = None
    stderr = None
    deploy_result = None
    pm = None
    try:
      # Create the log files.
      stdout = open(os.path.join(log_subdir, "stdout.log"), "w")
      stderr = open(os.path.join(log_subdir, "stderr.log"), "w")

      # Write out additional persistent metadata
      persistent_metadata = {}
      persistent_metadata["parcels_in_use"] = parcels_in_use
      pm = open(os.path.join(new_cwd, cc["directory_name"],
                             METADATA_FILENAME), "w")
      json.dump(persistent_metadata, pm, indent=2, sort_keys=True)
      # Ensure the file is flushed so the copy-during-deploy works on it
      pm.close()
      pm = None

      # Kick off the script and wait for it to finish.
      retcode, _, _ = subprocess_with_timeout(
        args=[self.deploy_script],
        timeout=DEPLOY_SCRIPT_TIMEOUT_SEC,
        stdout=stdout,
        stderr=stderr,
        cwd=new_cwd,
        env=env)

      deploy_result_metadata = {}
      deploy_result_metadata["exit_code"] = retcode
      deploy_result = open(os.path.join(new_cwd, METADATA_FILENAME), "w")
      json.dump(deploy_result_metadata, deploy_result, indent=2, sort_keys=True)

      return new_cwd
    finally:
      if stderr:
        stderr.close()
      if stdout:
        stdout.close()
      if deploy_result:
        deploy_result.close()
      if pm:
        pm.close()

  def _deregister_alternative(self, alt_name, alt_path):
    self.osops.alternatives.deactivate(alt_name, None, alt_path)

  def evaluate_resources(self, resources):
    """
    Evaluate client configs resources and make system changes as
    needed. Unknown resources are ignored.
    """
    directories = []
    for res in resources:
      try:
        if res["directory"]:
          directories.append(res)
      except:
        # Resource evaluation failures are non-fatal.
        LOG.exception("Could not evaluate resource %s", res)

    # Directory creation is sorted so that if a process
    # requests both /a/b and /a, then /a gets created with
    # the right permissions, rather than root/root.
    for res in sorted(directories, key=lambda x: x["directory"]["path"]):
      try:
        d = res["directory"]
        self.agent.mkabsdir(path=d["path"], user=d["user"], group=d["group"], mode=d["mode"])
      except:
        # Resource evaluation failures are non-fatal.
        LOG.exception("Could not evaluate resource %s", res)

  def _generate_deploy_subdir(self, alt_name, path, generation):
    """
    Returns a deployment subdirectory name using the alternative's name, path,
    and generation.
    """

    # e.g. "ccdeploy_hadoop-conf_etchadoopconf.cloudera.YARN1_1"
    base_parts = ["ccdeploy"]
    if alt_name:
      base_parts.append(alt_name)
    base_parts.append(path.replace("/", ""))
    base_parts.append(str(generation))
    return os.path.join(self.deploy_dir, "_".join(base_parts))

  def _create_and_backup_deploy_dirs(self, alt_name, path, generation):
    """
    If the target deployment directory doesn't exist, creates a brand
    new directory for intermediate deployment output, as well as
    a subdirectory for logs.
    If the intermediate directory already exists, back up the existing logs.
    Then wipe off the directory completely.

    Returns a 2-tuple of (deploy path, logging path).
    """
    deploy_subdir = self._generate_deploy_subdir(alt_name, path, generation)
    logs_subdir = os.path.join(deploy_subdir, "logs")

    tmp_log_dir = None
    try:
      # Before deploy the client config, backup the existing log file.
      existing_stdout_log = os.path.join(logs_subdir, "stdout.log")
      existing_stderr_log = os.path.join(logs_subdir, "stderr.log")
      if os.path.exists(existing_stdout_log):
        try:
          tmp_log_dir = tempfile.mkdtemp()
          shutil.move(
            existing_stdout_log,
            os.path.join(tmp_log_dir, "stdout.log"))
          shutil.move(
            existing_stderr_log,
            os.path.join(tmp_log_dir, "stderr.log"))
        except Exception, e:
          LOG.exception("Failed to move the existing log to temporary directory."\
                        " Error: %s" % e)

      if os.path.exists(deploy_subdir):
        shutil.rmtree(deploy_subdir, ignore_errors=True)

      os.makedirs(deploy_subdir)
      os.mkdir(logs_subdir)

      # After created the deployment directory, we need to move back the
      # previous log files as a backup. So in case of repeated failure, we'll
      # have at most two sets of log files, one of the current round of
      # deployment, another for the previous round. This will help save log
      # space if the deployment fails forever.
      if tmp_log_dir is not None:
        shutil.move(
          os.path.join(tmp_log_dir, "stdout.log"),
          existing_stdout_log + ".bak")
        shutil.move(
          os.path.join(tmp_log_dir, "stderr.log"),
          existing_stderr_log + ".bak")
    finally:
      if tmp_log_dir is not None:
        shutil.rmtree(tmp_log_dir, ignore_errors=True)

    return (deploy_subdir, logs_subdir)

  def _adapt_cc_to_env(self, env_template, cc):
    """
    Returns a 2-tuple of (new environment, parcels in use).

    'new environment' is a copy of 'env_template' with additional
    environment variables injected, based on information from 'cc' and from
    the parcel repo.

    'parcels in use' is a dict of product name to version describing the
    parcels that will actually be used in deployment, if any.
    """
    env = env_template.copy()

    # First adapt the generic stuff.
    for avro_version, env_version in AVRO_TO_ENV_GENERIC.iteritems():
      if avro_version in cc and cc[avro_version]:
        env[env_version] = str(cc[avro_version])

    # Set path to "update-alternatives" for deploy_script so that we
    # can be consistent between the logic in this file and the logic
    # in the deploy_script. See OPSAPS-24253 for further information.
    env["UPDATE_ALTERNATIVES"] = self.osops.alternatives.alt_bin

    # And now the stuff that needs special handling.
    if "runner_args" in cc and cc["runner_args"]:
      env["RUNNER_ARGS"] = " ".join(cc["runner_args"])
    if "additional_env" in cc:
      for key, val in cc["additional_env"].iteritems():
        env[key] = str(val)

    # And lastly, the parcels.
    parcels_in_use = {}
    if self.parcel_repo:
      parcels_in_use = self.parcel_repo.prepare_environment(cc["parcels"],
                                                            env,
                                                            cc["required_tags"],
                                                            cc["optional_tags"])

    return (env, parcels_in_use)

  def load(self):
    """
    Reads the system's alternatives database and loads all CM-managed entries
    into memory such that they can be used in rectify().
    """
    new_deployed = {}
    for _, _, fnames in os.walk(self.osops.alternatives.db_dir):
      for fname in fnames:
        new_deployed.update(self._lookup_alternatives(fname))

    # Load additional client config metadata from the on-disk deployment
    # subdirectories, if they exist.
    for key, value in new_deployed.iteritems():
      value.load(self._generate_deploy_subdir(key.name,
                                              key.path,
                                              value.generation),
                 key.path)

    self.deployed = new_deployed
    self.results = new_deployed.copy()

    for key, value in self.deployed.iteritems():
      LOG.info("Loaded client config: %s %s", key, value)

  def rectify(self, client_configs, custom_java_home_config=None):
    """
    Compares the client configs from 'client_configs' to those found in
    self.deployed and takes appropriate action based on the differences:

    1. New client configs are deployed.
    2. Missing client configs' alternative entries are deregistered.
    3. Changed client configs are deregistered and redeployed.

    Upon returning, self.results reflects all entries from 'client_configs'
    while self.deployed contains only those that were succesfully deployed.
    """
    self.java_home_config = custom_java_home_config

    if client_configs is None:
      return

    # We need several dictionaries:
    # 1. ClientConfigValue keyed by ClientConfigKey. Compared with
    #    self.deployed to determine which client configs have changed since
    #    the last call to rectify().
    # 2. Client configs keyed by ClientConfigKey. Stores all the information
    #    needed to deploy new client configs.
    new_deployed = {}
    new_ccs = {}
    for cc in client_configs:
      key = ClientConfigKey(cc["alt_name"], cc["path"])
      value = ClientConfigValue(cc["priority"], cc["generation"])
      new_deployed[key] = value
      new_ccs[key] = cc

    # This will test for new or missing ClientConfigKeys (i.e. client
    # configs added or removed) as well as generation changes to
    # ClientConfigValues.
    if self.deployed == new_deployed:
      return

    # Let's figure out what's changed.
    previous_keys = set(self.deployed.keys())
    new_keys = set(new_deployed.keys())
    to_delete = previous_keys.difference(new_keys)
    to_deploy = new_keys.difference(previous_keys)
    to_update = set()
    for key in new_keys.intersection(previous_keys):
      if new_deployed[key].generation != self.deployed[key].generation:
        to_update.add(key)
    LOG.info("Client configs to delete: %s", to_delete)
    LOG.info("Client configs to deploy: %s", to_deploy)
    LOG.info("Client configs to update: %s", to_update)

    # Perform all changes.
    #
    # For simplicity, updates are implemented via delete then deploy.
    for key in to_delete.union(to_update):
      old_value = self.deployed.pop(key)
      del self.results[key]
      # Note that we can only prune the client config from the alternatives
      # database; we don't always have enough information to remove its
      # files. As such, there's nothing to do for configs without
      # alternatives (i.e. cluster client configs).
      if key.name:
        # See UPGRADE PROVISIONS at the top of the file to understand why
        # only client configs with valid generations are pruned.
        if old_value.generation == NO_GENERATION_FOUND:
          LOG.info("Skipping deletion of generation-less client config %s",
                   key)
          continue
        self._deregister_alternative(key.name, key.path)
    for key in to_deploy.union(to_update):
      value = new_deployed[key]
      try:
        deploy_path = self._deploy_client_config(new_ccs[key])
        value.load(deploy_path, key.path)
        if value.exit_code == 0:
          self.deployed[key] = value

      except Exception, e:
        msg = "Failed to deploy client config %s: %s" % (key, e)
        value.exit_code = -1
        value.stdout_tail = msg
        value.stderr_tail = msg
        LOG.exception(msg)
      self.results[key] = value

  def get_status(self):
    """
    Returns a list of ClientConfigStatus entries (see master_agent.avdl)
    that reflect all recent deployment attempts; see rectify() for more
    details.
    If the deployment directories are pruned and the alternatives state
    reloaded, there may exist ClientConfigStatus entries without
    exit_codes. Using a value of 0 is safe because the mere existence of
    these entries implies that valid alternatives exist. In other words,
    the entries imply that the corresponding client config deployments
    were successful.
    """
    statuses = []
    for key, value in self.results.iteritems():
      status = {}
      status["alt_name"] = key.name
      status["path"] = key.path
      status["generation"] = value.generation

      stdout_tail = ""
      if value.stdout_tail:
        stdout_tail = value.stdout_tail
      status["stdout_tail"] = stdout_tail

      stderr_tail = ""
      if value.stderr_tail:
        stderr_tail = value.stderr_tail
      status["stderr_tail"] = stderr_tail

      exit_code = 0
      if value.exit_code is not None:
        exit_code = value.exit_code
      status["exit_code"] = exit_code

      parcels_in_use = {}
      if exit_code == 0 and value.parcels_in_use is not None:
        parcels_in_use = value.parcels_in_use
      status["parcels_in_use"] = parcels_in_use

      statuses.append(status)

    return statuses

  def get_active_deployments(self):
    """
    Returns a list of deployment directories (relative to self.deploy_dir)
    that are in use and should not be deleted.
    """

    deployments = []
    for key, value in self.results.iteritems():
      deploy_path = self._generate_deploy_subdir(key.name, key.path,
                                                 value.generation)
      deployments.append(os.path.basename(deploy_path))
    return deployments
