"""
Copyright 2016, Cloudera Inc.

Definition of Agent configuration parser/schema.
"""

from __future__ import absolute_import

import sys
import tempfile
import os

import cmf.util
from cmf.util.config import (Config, ConfigItem, ConfigIntItem,
                             ConfigBoolItem, ConfigFloatItem, ConfigListItem)
from cmf.util.lazy import lazy

PROG = "cloudera-scm-agent"
DEFAULT_BASE_DIR = "/opt/cloudera"
DEFAULT_RUN_DIR = '/var'
DEFAULT_PARCEL_DIR = os.path.join(DEFAULT_BASE_DIR, 'parcels')
DEFAULT_CONF_DIR = '/etc/cloudera-scm-agent'
DEFAULT_CONFIG_FILE = os.path.join(DEFAULT_CONF_DIR, 'config.ini')


class AgentConfig(Config):
  """
  Configuration definition for the CM agent.
  """
  # pylint: disable=attribute-defined-outside-init, too-few-public-methods, invalid-name

  GENERAL = "General"
  SECURITY = "Security"
  HADOOP = "Hadoop"
  CLOUDERA = "Cloudera"
  JDBC = "JDBC"
  CGROUP_PATHS = "Cgroup_Paths"

  scm_debug = ConfigItem(GENERAL, False)
  master_host = ConfigItem(GENERAL, name="server_host")
  master_port = ConfigIntItem(GENERAL, name="server_port")
  agent_httpd_port = ConfigIntItem(GENERAL, 9000, name='listening_port')
  agent_internal_port = ConfigIntItem(GENERAL, lambda cfg_item, cfg: cfg.agent_httpd_port + 1)
  clear_agent_dir = ConfigBoolItem(GENERAL, False)
  ip_address = ConfigItem(GENERAL, cmf.util.default_hostname()[1], name='listening_ip')
  reported_hostname = ConfigItem(GENERAL,
                                 lambda cfg_item, cfg: cfg.listening_hostname)
  listening_hostname = ConfigItem(GENERAL,
                                  cmf.util.default_hostname()[0])
  supervisord_httpd_port = ConfigIntItem(GENERAL, name='supervisord_port', default=19001)
  supervisord_path = ConfigItem(GENERAL,
                                os.path.join(sys.prefix, "bin", "supervisord"))

  host_id = ConfigItem(GENERAL,
                       lambda cfg_item, cfg: cmf.util.get_host_uuid(cfg.lib_dir))

  base_dir = ConfigItem(GENERAL, "/opt/cloudera")
  install_dir = ConfigItem(GENERAL,
                           lambda cfg_item, cfg: os.path.join(cfg.base_dir, 'cm-agent'))
  parcel_dir = ConfigItem(GENERAL)
  lib_dir = ConfigItem(GENERAL, os.path.join('/var/lib', PROG))
  log_dir = ConfigItem(GENERAL, os.path.join('/var/log', PROG))
  package_dir = ConfigItem(GENERAL,
                           lambda cfg_item, cfg: os.path.join(cfg.install_dir, "service"))
  agent_dir = ConfigItem(GENERAL, os.path.join("/var", "run", PROG))
  logfile = ConfigItem(GENERAL, name='log_file',
                       default=lambda cfg_item, cfg: os.path.join(cfg.log_dir, '%s.log' % PROG))

  max_collection_wait_seconds = ConfigFloatItem(GENERAL, 10.0)
  metrics_url_timeout_seconds = ConfigFloatItem(GENERAL, 30.0)
  task_metrics_timeout_seconds = ConfigFloatItem(GENERAL, 5.0)
  monitored_nodev_filesystem_types = ConfigListItem(GENERAL, ["nfs", "nfs4"])
  local_filesystem_whitelist = ConfigListItem(GENERAL, ["ext2", "ext3", "ext4", "xfs"])
  filesystem_map_poll_timeout_seconds = ConfigFloatItem(GENERAL, 2.0)
  filesystem_map_join_timeout_seconds = ConfigFloatItem(GENERAL, 0.1)
  filesystem_map_cache_tolerance = ConfigFloatItem(GENERAL, 60.0)
  impala_profile_bundle_max_bytes = ConfigIntItem(GENERAL, 1 <<30) #1GB
  stacks_log_bundle_max_bytes = ConfigIntItem(GENERAL, 1<<30) #1GB
  stacks_log_max_uncompressed_file_size_bytes = \
    ConfigIntItem(GENERAL, 5<<20) #5MB
  supervisord_events = ConfigBoolItem(GENERAL, name='enable_supervisord_events', default=True)
  sudo = ConfigItem(GENERAL, name='sudo_command',
                    default=lambda cfgItem, cfg: "sudo" if os.getuid() != 0 else None)
  dns_resolution_collection_timeout_seconds = ConfigIntItem(GENERAL, 30)
  agent_wide_credential_cache_location = ConfigItem(GENERAL, lambda cfg_item, cfg: cfg.agent_dir)

  flood_rpc_port = ConfigIntItem(GENERAL, 7190)
  flood_expiration_time_sec = ConfigIntItem(GENERAL, 24 * 60 * 60)
  orphan_process_directory_refresh_interval = \
    ConfigIntItem(GENERAL, name='orphan_process_dir_refresh_interval', default=60*60)
  orphan_process_directory_staleness_threshold = \
  ConfigIntItem(GENERAL, name='orphan_process_dir_staleness_threshold', default=60*24*60*60)

  use_tls = ConfigBoolItem(SECURITY, False)
  max_cert_depth = ConfigIntItem(SECURITY, 9)
  client_key_file = ConfigItem(SECURITY)
  client_cert_file = ConfigItem(SECURITY)
  verify_cert_file = ConfigItem(SECURITY)
  verify_cert_dir = ConfigItem(SECURITY)
  client_keypw_file = ConfigItem(SECURITY)
  client_keypw_cmd = ConfigItem(SECURITY)
  optional_token = ConfigBoolItem(SECURITY)
  cm_auto_cert_dir = ConfigItem(SECURITY)

  # Cgroup paths
  cpu_cgroup_path = ConfigItem(CGROUP_PATHS)
  cpuacct_cgroup_path = ConfigItem(CGROUP_PATHS)
  blkio_cgroup_path = ConfigItem(CGROUP_PATHS)
  memory_cgroup_path = ConfigItem(CGROUP_PATHS)

  @lazy
  def paths(self):
    """
    Provide an object representing paths implicitly derived by the configuration.
    """
    return Paths(self)

class Paths(object):
  """
  This is a distinct namespace object for paths that are derived from the agent configuration,
  but not directly configurable.
  """
  # pylint: disable=too-few-public-methods,too-many-instance-attributes

  def __init__(self, cfg):
    if cfg.agent_dir is None:
      self.base_dir = os.path.realpath(tempfile.mkdtemp(prefix="cmf_agent_"))
    else:
      self.base_dir = cfg.agent_dir
    self.event_pipe_path = os.path.join(self.base_dir, "events")

    self.supervisor_dir = os.path.join(self.base_dir, "supervisor")
    self.flood_dir = os.path.join(self.base_dir, "flood")
    self.supervisor_include_path = os.path.join(self.supervisor_dir, "include")
    self.cgroups_dir = os.path.join(self.base_dir, "cgroups")
    self.process_dir = os.path.join(self.base_dir, "process")
    self.auth_token = os.path.join(self.base_dir, "auth_token")
    self.redaction_policy_string = os.path.join(self.base_dir, "redaction_policy_string")
