# Copyright (c) 2012 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import
import avro
import calendar
import cStringIO as StringIO
import httplib
import logging
import os
import sys
import tarfile
import tempfile
import threading
import threadpool
import time
import urllib
import urllib2
import urlparse
import cmf.https

from datetime import datetime
from cmf.throttling_logger import ThrottlingLogger
try:
  from flood.client import TorrentRPCClient
  from flood.util.rpc import resolveHostPort
except (ImportError, SyntaxError):
  TorrentRPCClient = None
  resolveHostPort = None

from avro.ipc import AvroRemoteException
from cmf.util import synchronized

LOG = logging.getLogger(__name__)
THROTTLED_LOG = ThrottlingLogger(LOG)

# Download requested but not yet begun.
PENDING = 'pending'
# Actively downloading.
DOWNLOADING = 'downloading'
# Download complete, not necessarily finished callback.
DOWNLOADED = 'downloaded'
# Downloaded and finished processing callback, if any.
COMPLETE = 'complete'
# Exception occurred -- could be during download or callback.
EXCEPTION = 'exception'
# Aborted
ABORTED = 'aborted'
ALL_STATES = set([PENDING, DOWNLOADING, DOWNLOADED, COMPLETE, EXCEPTION, ABORTED])
ACTIVE_STATES = set([PENDING, DOWNLOADING, DOWNLOADED])
TERMINAL_STATES = ALL_STATES - ACTIVE_STATES

TEMP_FILE_SUFFIX = '.part'

# Keep in sync with master_agent.avdl
URL_FIELD = 'url'
STATE_FIELD = 'state'
ACTIVE_FIELD = 'active'
TOTAL_BYTES_FIELD = 'total_bytes'
DOWNLOADED_BYTES_FIELD = 'downloaded_bytes'
START_TIME_FIELD = 'start_time'
END_TIME_FIELD = 'end_time'
ERROR_FIELD = 'parcel_err'
ERROR_CODE_FIELD = 'code'
ERROR_MSG_FIELD = 'msg'

UNEXPECTED_ERROR_CODE = 600
CALLBACK_ERROR_CODE = 601

DOWNLOAD_BLOCK_SIZE = 8192

class URLopener(object):
  """
  Fetches URLs via HTTP and reports ongoing progress.
  """
  def rename(self, src, dest):
    '''This is here just to simplify mocking for tests'''
    os.rename(src, dest)

  # wrapper function for testing purpose
  def fetch_url(self, request):
    return cmf.https.ssl_url_opener.open(request)

  def download(self, op, destPathBase, progressFn):
    '''Download file from any url'''
    url = op.url

    o = urlparse.urlparse(url)
    # o[2] should ideally be o.path, but in 2.4 it's just a tuple
    filename = os.path.join(destPathBase, os.path.basename(o[2]))
    tempname = filename + TEMP_FILE_SUFFIX

    op.state = DOWNLOADING
    LOG.info("Starting download of: %s", url)

    exist_size = 0
    try:
      try:
        request = urllib2.Request(str(url))

        if os.path.exists(tempname):
          # If the file exists, only download the remainder
          output_file = open(tempname, "ab")
          exist_size = os.path.getsize(tempname)
          LOG.info("%s exists, add Range header: bytes=%s-" % \
                   (tempname, exist_size))
          request.headers["Range"] = "bytes=%s-" % (exist_size)
        else:
          output_file = open(tempname, "wb")

        response = self.fetch_url(request)
        op.headers = response.info()

        # If the whole file exists, don't download it again
        total_size = int(response.headers['content-length'])
        if total_size == exist_size:
          progressFn(total_size, exist_size, url)
        else:
          if exist_size > total_size:
            LOG.warn("The existing partially downloaded file [%s] size [%d] " \
                      "is larger than the expected total size [%d].",
                      tempname, exist_size, total_size)
            output_file.close()
            output_file = open(tempname, "wb")
            exist_size = 0

          current_size = exist_size
          while True:
            if op.state is ABORTED:
              LOG.info("Aborting the download for file [%s]." % filename)
              self.remove(tempname)
              return
            data = response.read(DOWNLOAD_BLOCK_SIZE)
            if not data:
              break
            output_file.write(data)
            current_size += len(data)
            progressFn(current_size, total_size, url)
      except urllib2.HTTPError, e:
        LOG.exception("HTTP error during download")
        op.set_exception(e.code, e)
        return
      except Exception, e:
        LOG.exception("Unexpected exception during download")
        op.set_exception(UNEXPECTED_ERROR_CODE, e)
        return
    finally:
      output_file.close()

    op.state = DOWNLOADED
    op.code = 200

    try:
      self.rename(tempname, filename)
      op.path = filename
    except Exception, e:
      op.set_exception(None, e)

    LOG.info("Completed download of %s code=%d state=%s", url, op.code, str(op.state))
    return op

class FloodOpener(URLopener):
  FLOOD_CONF = "flood.conf"
  STALL_TIME = 90

  def __init__(self,
               server,
               seeding_timeout,
               torrent_port,
               rpc_port,
               depot_path):
    '''
    Constructs new FloodOpener.

    :param server:  String host:port indicating how to connect to the local flood daemon.
    :param seeding_timeout: Seconds wait on inactivity before attempting HTTP download.
    :param torrent_port: Port for flood to use for torrent traffic.
    :param rpc_port: Port for flood control RPCs
    :param depot_path: FS location where to save torrent data.
    :return:
    '''
    self.client = TorrentRPCClient(server)
    self.seeding_timeout = seeding_timeout
    self.nodeList = []
    self.torrent_port = torrent_port
    self.depot_path = depot_path
    self.rpc_port = rpc_port
    self.rackpeers = set()

  def refresh(self, server, seeding_timeout, torrent_port, rpc_port, depot_path):
    '''
    Determines if a new FloodOpener is required (if so agent is expected to
    restart the flood daemon).

    :param server:  See __init__.
    :param seeding_timeout: See __init__.
    :param torrent_port: See __init__.
    :param rpc_port: See __init__.
    :param depot_path: See __init__.
    :return: New or existing FloodOpener object.
    '''
    # Currently only check for torrent_port changes to trigger restart.
    if (torrent_port == self.torrent_port and self.depot_path == depot_path and
      rpc_port == self.rpc_port):
      return self

    return self.__class__(server, seeding_timeout, torrent_port, rpc_port, depot_path)

  def write_flood_config(self, flood_dir, log_file, expiration_time_sec):
    '''
    Writes the config file for flood.   This is here in an attempt to isolate as much
    flood code out of agent.py as possible.
    :param flood_dir: Directory for flood work files.
    :param log_file: Name of primary log file to log to.
    :param expiration_time_sec: Time for expiration of idle torrents.
    :return:
    '''
    f = None
    try:
      f = open(os.path.join(flood_dir, self.FLOOD_CONF),"w")
      f.write("torrentPort=%s\n" % self.torrent_port)
      f.write("depotPath=%s\n" % self.depot_path)
      f.write("rpcPort=%s\n" % self.rpc_port)
      f.write("depotExpiration=%s\n" % expiration_time_sec)
      if log_file is not None:
        f.write("logFileName=%s\n" % log_file)
    finally:
      if f is not None:
        f.close()

  def set_seed_timeout(self, seed_timeout):
    self.seeding_timeout = seed_timeout

  def set_rackpeers(self, rackpeers):
    '''
    Inform Flood of a new peer set.
    Note: Flood can't "forget" prior peers, so we don't
    worry about removing stale peers here.
    :param rackpeers:
    :return:
    '''
    try:
      rackpeers = set(rackpeers)
      for peer in set(rackpeers):
        hp = resolveHostPort(peer)
        if hp not in self.rackpeers:
          self.client.AddDHTNode(hp.host, hp.port)
          self.rackpeers.add(hp)
    except Exception, e:
      LOG.error("Failed rack peer update: %s", e)

  def download(self, op, destPathBase, progressFn):
    '''Download file from any url'''
    url = op.url
    url_parts = urlparse.urlparse(url)

    torrent_name = os.path.split(url_parts[2])[1]
    filename = os.path.join(destPathBase, os.path.basename(url_parts[2]))

    # Fetch the .torrent url using the provided SSL context and
    # pass the file to flood, since flood doesn't have access to our SSL config.
    torrent_url = url + ".torrent"
    LOG.info("Fetching torrent: %s", torrent_url)
    try:
      torrent_file = os.path.join(destPathBase, os.path.basename(url_parts[2])+ ".torrent")
      cmf.https.ssl_url_opener.fetch_to_file(torrent_url, torrent_file)
      torrent_url = "file://" + torrent_file
    except Exception, e:
      op.state = ABORTED
      LOG.exception("Failed fetching torrent: %s", e)
      raise
    op.state = DOWNLOADING

    seed_time = time.time() + self.seeding_timeout
    fail_time = None
    do_http_fetch = False
    http_progress = 0
    stat = None
    LOG.info("Starting download of: %s", url)
    try:
      try:
        self.client.AddTorrent(torrent_url)
      except AvroRemoteException, e:
        LOG.info("Failed adding torrent: %s %s", torrent_url, e)

      while True:
        stat = self.client.stat(torrent_name)

        LOG.info("Current state: %s %s", torrent_name, stat)
        if op.state is ABORTED:
          LOG.info("Aborting the download for file [%s]." % filename)
          self.client.RemoveTorrent(torrent_name)
          return

        progressFn(stat.totalDownloaded, stat.totalSize, url)

        if self.seeding_timeout + time.time() < seed_time:
          seed_time = time.time() + self.seeding_timeout

        if seed_time <= time.time():
          do_http_fetch = True

        # This indicates there's no pieces left to download
        if stat.state == 'seeding':
          break

        if do_http_fetch and url not in stat.seed:
          LOG.info("Setting url seed: %s %s (%s)", torrent_name, url, stat.seed)
          self.client.AddTorrentSeed(torrent_name, url)
          if fail_time is None:
            fail_time = time.time() + self.STALL_TIME


        # If fetching via HTTP, then we expect progress to be made
        if fail_time and fail_time < time.time():
          if stat.totalDownloaded == http_progress:
            LOG.error("Failure due to stall on seeded torrent.")
            op.set_exception(UNEXPECTED_ERROR_CODE,
                             Exception("Failure due to stall on seeded torrent."))
            return
          http_progress = stat.totalDownloaded
          fail_time = time.time() + self.seeding_timeout
        time.sleep(2)

    except AvroRemoteException, e:
      LOG.error("Failed to query torrent: %s %s", torrent_name, e)
      op.set_exception(UNEXPECTED_ERROR_CODE, e)
      return
    except Exception, e:
      LOG.exception("Unexpected exception during download")
      op.set_exception(UNEXPECTED_ERROR_CODE, e)
      return

    op.path = os.path.join(stat.location, torrent_name)
    op.state = DOWNLOADED
    op.code = 200

    LOG.info("Completed download of %s code=%d state=%s", url, op.code, str(op.state))
    return op





class Download(object):
  """
  Track status of a download. The headers may not be available if file already existed on disk
  """
  def __init__(self, url):
    self.url = url
    self._state = PENDING
    self.total_bytes = 0
    self.downloaded_bytes = 0
    self.start_time = self.get_now_ms()
    self.download_end_time = None
    self.end_time = None
    self.code = 0
    self.exception_msg = None
    self.path = None
    self.headers = None

  def get_now_ms(self):
    return long(time.time() * 1000)

  def get_timems_str(self, time_in_ms):
    if time_in_ms is None:
      return ""
    else:
      return datetime.fromtimestamp(time_in_ms / 1000).strftime('%Y-%m-%d %H:%M:%S')

  def set_exception(self, code, exception):
    self.state = EXCEPTION
    self.exception_msg = str(exception)
    self.code = code

  def get_state(self):
    return self._state

  def set_state(self, state):
    if self._state in TERMINAL_STATES:
      return
    else:
      self._state = state
      if state in TERMINAL_STATES:
        self.end_time = self.get_now_ms()
      elif state is DOWNLOADED:
        self.download_end_time = self.get_now_ms()

  state = property(get_state, set_state)
  @property
  def progress(self):
    try:
      return "%2.0f%%" % (100*self.downloaded_bytes/self.total_bytes)
    except ZeroDivisionError:
      return "N/A"

  def __repr__(self):
    return "[ url: %s, state: %s, total_bytes: %d, downloaded_bytes: %d, start_time: %s, download_end_time: %s, end_time: %s, code: %d, exception_msg: %s, path: %s ]" % (self.url, self.state, self.total_bytes, self.downloaded_bytes, self.get_timems_str(self.start_time), self.get_timems_str(self.download_end_time), self.get_timems_str(self.end_time), self.code, self.exception_msg, self.path)

class Downloader(object):
  """
  Class that implements our parcel downloader.
  """

  def __init__(self, path, limit, opener=URLopener()):
    LOG.info("Downloader path: %s" % (path))
    self.path = path
    self.limit = limit
    self.threadpool = threadpool.ThreadPool(limit)
    self.ops = dict()
    self.ops_lock = threading.RLock()
    self.opener = opener
    self.remove = self._remove


  def get_lock(self):
    '''
    Conform to requirements of @synchronized decorator
    :return:
    '''
    return self.ops_lock

  @synchronized
  def update_opener(self, opener):
    '''
    Dynamically changing the opener is not a good idea and the effects are unknown/undefined.
    It is expected that this will only occur while the system is quiesced with no pending
    downloads.  If there are pending downloads, then the old "opener" object will still
    be live until it finishes. But the set of possibile system behaviors is sufficiently
    large so as to make it cost prohibitive to support such reconfiguration with pending
    downloads.

    :param opener: New "opener" object to use.
    :return:
    '''
    if opener is self.opener:
      return
    if self.is_active():
      raise Exception("Downloader is active.")
    self.opener = opener

  def cleanUp(self):
    self.threadpool.dismissWorkers(self.limit)

  def wait(self):
    return self.threadpool.wait()

  def poll(self):
    return self.threadpool.poll()

  def queue_download(self, url, callback=None, overwrite=False):
    """Queue a download for the url.  If there is already a download for the
       url in active state, the request will be ignored. Otherwise a new
       download will be enqueued. The 'overwrite' flag can be used to first
       delete any existing downloaded file. True is returned as long as the
       request is enqueued.

       The optional callback function takes as argument the url and a Download
       object that contains information about the download. The function will
       be called regardless of whether the download succeeded.
    """
    downloaded_path = self._get_downloaded_path(url)
    already_exist = os.path.exists(downloaded_path)
    if already_exist and overwrite:
      LOG.info("Removing existing file %s" % (downloaded_path))
      os.remove(downloaded_path)
      already_exist = False

    try:
      self.ops_lock.acquire()
      if url in self.ops:
        curr_op = self.ops[url]
        if curr_op.state in ACTIVE_STATES:
          THROTTLED_LOG.info("There is already an active download for %s" % (url))
          return False
        elif curr_op.state in TERMINAL_STATES:
          # overwrite history from prior downloads
          curr_op = Download(url)
        else:
          raise Exception("Unexpected state %s" % (curr_op.state))
      else:
        curr_op = Download(url)
      self.ops[url] = curr_op
    finally:
      self.ops_lock.release()

    def callable():
      LOG.debug("Executing request for op %s" % (str(curr_op)))
      if already_exist:
        # fake as if downloaded (ops state may be lost due to agent restart)
        # response headers are not available so client code such as checksum
        # should accommodate for this
        curr_op.state = DOWNLOADED
        curr_op.code = 200
        curr_op.path = downloaded_path
      elif curr_op.state is PENDING:
        self.opener.download(curr_op, self.path, self.progress)
      # directly process callback using the threadpool's threads. we choose
      # this over threadpool's callback mechanism because the latter relies on
      # calling poll() with our own thread, and callback can potentially take
      # a long time (e.g. untar huge parcel).
      if callback:
        try:
          callback(url, curr_op)
        except Exception, e:
          LOG.exception("Failed op: %s", e)
          curr_op.set_exception(CALLBACK_ERROR_CODE, e)
          if hasattr(e, 'keep_file') and e.keep_file:
            pass
          else:
            self.remove(curr_op.path)
      if curr_op.state is DOWNLOADED:
        curr_op.state = COMPLETE
      LOG.info("Finished download %s" % (curr_op))
      return curr_op

    request = threadpool.WorkRequest(callable)
    self.threadpool.putRequest(request)
    LOG.debug("Queued download for %s into %s" % (url, downloaded_path))
    return True

  def abort(self, url):
    try:
      self.ops_lock.acquire()
      if url not in self.ops:
        LOG.info("No download for url %s" % (url))
        return False
      op = self.ops[url]
      if op.state in ACTIVE_STATES:
        LOG.info("Aborting download request for %s" % (url))
        op.state = ABORTED
        return True
      else:
        LOG.debug("Ignoring abort for %s because its state %s is terminal" % (url, op.state))
        return False
    finally:
      self.ops_lock.release()

  @synchronized
  def is_active(self):
    """
    Return true/false to indicate if any downloads are in progress.

    :return:
    """
    for op in self.ops.itervalues():
      if op.state in ACTIVE_STATES:
        return True
    return False

  def prune_history(self, filter):
    """Prune download history of urls that satisfy the provided filter
       function. The function will be invoked with one argument the
       url string, and should return True if it is to be pruned. Active
       downloads cannot be pruned regardless.
    """
    try:
      self.ops_lock.acquire()
      pruned = dict()
      for op in self.ops.itervalues():
        if op.state in ACTIVE_STATES or not filter(op.url):
          pruned[op.url] = op
      self.ops = pruned
    finally:
      self.ops_lock.release()

  def _get_downloaded_path(self, url):
    o = urlparse.urlparse(url)
    # o[2] should ideally be o.path, but in 2.4 it's just a tuple
    return os.path.join(self.path, os.path.basename(o[2]))



  def _remove(self, file):
    if os.path.exists(file):
      try:
        os.remove(file)
      except OSError, e:
        LOG.exception("Error removing file %s", (file))
        raise e

  def set_remove_callout(self, func):
    self.remove = func

  def progress(self, current_size, total_size, url):
    op = self.ops[url]
    op.downloaded_bytes = current_size
    op.total_bytes = total_size

  def get_current_downloads(self):
    dws = []
    for op in self.ops.values():
      dw = dict()
      dw[URL_FIELD] = op.url
      dw[STATE_FIELD] = op.state
      dw[ACTIVE_FIELD] = op.state in ACTIVE_STATES
      dw[TOTAL_BYTES_FIELD] = op.total_bytes
      dw[DOWNLOADED_BYTES_FIELD] = op.downloaded_bytes
      dw[START_TIME_FIELD] = op.start_time
      dw[END_TIME_FIELD] = op.download_end_time
      if (op.state == EXCEPTION):
        err = dict()
        err[ERROR_CODE_FIELD] = str(op.code)
        err[ERROR_MSG_FIELD] = op.exception_msg
        dw[ERROR_FIELD] = err
      else:
        dw[ERROR_FIELD] = None
      dws.append(dw)
    LOG.debug("Current downloads: %s", dws)
    return dws

if __name__ == '__main__':
  if len(sys.argv) < 4:
    print "downloader.py <http|flood> <download dir> <max simultaneous downloads> [urls...]"
    sys.exit(1)

  opener = None
  if sys.argv[1].lower() == "flood":
    print "Using flood downloader."
    nodeList = os.environ.get('FLOOD_NODES', "").split(",")
    opener = FloodOpener(os.environ.get('FLOOD_SERVER', "127.0.0.1:7190"),
                         30, nodeList)
  else:
    opener = URLopener()
    print "Using http downloader."

  LOG_FORMAT = '[%(asctime)s] %(process)d %(threadName)s %(module)-12s %(levelname)-8s %(message)s'
  DATE_FORMAT = '%d/%b/%Y %H:%M:%S %z'
  logging.basicConfig(level=logging.DEBUG,
                      format=LOG_FORMAT, datefmt=DATE_FORMAT)

  downloader = Downloader(sys.argv[2], int(sys.argv[3]), opener)

  def callback(url, op):
    print downloader.ops[url]

  for url in sys.argv[4:]:
    downloader.queue_download(url, callback=callback)

  try:
    while True:
      for url, ops in downloader.ops.items():
        print "%s:%s" % (url, ops.progress)

      downloader.poll()
      time.sleep(5)
  except:
    # downloader.poll() throws an exception when there's no work left.
    pass

  downloader.cleanUp()
