#!/usr/bin/env python
from __future__ import absolute_import
# Copyright (c) 2012 Cloudera, Inc. All rights reserved.

import cherrypy
import subprocess
import socket
import logging
import os
import re
import signal

LOG = logging.getLogger(__name__)

KEY_FILE = "cloudera_manager_agent_fencer_secret_key.txt"

WHITESPACE = re.compile("\s+")

def log_and_append(data, message):
  LOG.info(message)
  data.append(message)

class FencingServer(object):
  """
  Exposes /fence URL to kill, and keep killed
  certain processes/pids.
  """

  def __init__(self, agent):
    self.agent = agent

  def find_processes(self, messages, name, secret_key):
    """Finds processes matching name."""
    try:
      self.agent.lock.acquire()

      # Step 2a: Find the processes
      found_processes = []
      for process in self.agent.active_processes.values():
        if name == process.raw_name:
          if process.dir is None:
            continue
          try:
            p_secret_key = file(os.path.join(process.dir, KEY_FILE)).read().rstrip()
          except:
            p_secret_key = None
          if secret_key == p_secret_key:
            log_and_append(messages, "Confirmed secret key for process %s" % (process.name))
            found_processes.append(process)
          else:
            log_and_append(messages, "Invalid secret key for process %s" % (process.name))

      log_and_append(messages, "Found %d managed processes for name: %s" % (len(found_processes), name))
      return found_processes
    finally:
      self.agent.lock.release()

  def stop_via_supervisor(self, messages, found_processes):
    # TODO: It's technically possible for there to be two
    # processes with the same name.  We should check the
    # role as well.
    try:
      self.agent.lock.acquire()

      # Try to stop the process gracefully,
      # by asking supervisord.
      for process in found_processes:
        try:
          log_and_append(messages, "Asked supervisord to stop process: %s" % (process.name, ))
          x = self.agent.supervisor_client.stop_process(process.name)
          log_and_append(messages, "Supervisord returned: %s" % (x, ))
        except Exception, e:
          LOG.exception("Failed to stop process: " + process.name)
          messages.append("Exception trying to stop process via supervisord: %s: %s" % (process.name, str(e)))
    finally:
      self.agent.lock.release()

  def fuser_kill(self, host, port, messages, secret_key):
    """
    Runs fuser to determine if there are processes listening
    to port.  If there are, kills them after checking the
    secret key.  Finally, to double-check, tries to talk to
    the socket.

    Returns True if we have good confidence that there are
    no processes listening on port.
    """
    log_and_append(messages, "Calling fuser.")

    fencing_succesful = True

    try:
      fuser_cmd = "fuser -v -n tcp %d" % (port, )
      log_and_append(messages, "Running: %s" % fuser_cmd)
      fuser = subprocess.Popen(
        fuser_cmd,
        shell=True,
        close_fds=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE)
      fuser_stdout, fuser_stderr = fuser.communicate()
      fuser_ret = fuser.returncode
      log_and_append(messages, "fuser output: retcode: %d, stdout: %s, stderr: %s" % (fuser_ret, fuser_stdout, fuser_stderr))
      if fuser_ret == 0:
        log_and_append(messages, "fuser returned 0")
	# fuser returns pids on stdout; everything else is not on stdout.
        pids = WHITESPACE.split(fuser_stdout.strip())
        pids = [ int(p) for p in pids ]

        if not pids:
          # Unexpected.
          log_and_append(messages, "couldn't parse fuser output")
          return False
        
        for pid in pids:
          p_secret_key = file(os.path.join("/proc", str(pid), "cwd", KEY_FILE)).read().rstrip()
          if secret_key != p_secret_key:
            log_and_append(messages, "mismatched secret keys; process not killed")
            return False
          os.kill(pid, signal.SIGKILL)
          
      elif fuser_ret == 1:
        log_and_append(messages, "fuser didn't kill anyone")
      else:
        log_and_append(messages, "fuser returned unexpected error code: %d" % (fuser_ret,) )
    except Exception, e:
      log_and_append(messages, "fuser failed: " + str(e))
      logging.exception("fuser failed")
      return False

    # Reimplementing netcat -z is easy, so we do:
    try:
      log_and_append(messages, "Connecting to socket: %s:%d" % (host, port))
      s = socket.socket()
      s.connect( (host, port) )
      s.close()
      log_and_append(messages, "Was able to connect to socket; target not dead.")
      fencing_succesful = False
    except socket.error, e:
      # don't use e.errno: it's not in py2.4
      if e.args[0] != 111: # "Connection refused"
        log_and_append(messages, "Socket connection failed with unexpected error: %s" % (e.args, ))
        fencing_succesful = False
      else:
        log_and_append(messages, "Socket connection failed (good): %s" % (e.args,) )
    
    return fencing_succesful

  def fence(self, name, host, port, secret_key):
    """
    Main entry point into agent-based fencing.  This
    is exposed via the fence/fence URL.

    name is the name of a DbProcess instance, which is
    typically "hdfs-NAMENODE".  host and port are
    taken from the arguments to the fencing command.

    secret_key is currently unimplemented
    """
    # Note that CherryPy's default socket timeout is 10
    # seconds.

    port = int(port)

    messages = []
    log_and_append(messages, "Received fencing request for %s:%d" % (host, port))

    # Step 1: Find relevant processes
    processes = self.find_processes(messages, name, secret_key)
    
    # Step 2: Stop processes via supervisor
    self.stop_via_supervisor(messages, processes)

    # Step 3: Confirm that nobody else is on the 'fencing' port.
    # If they are, kill them, but only if secret key matches up.
    ret = self.fuser_kill(host, port, messages, secret_key)

    # Step 4: Return the results
    cherrypy.response.headers["Content-Type"] = "text/plain"
    out = ""
    if ret:
      out += "OK\n"
    else:
      out += "FAIL\n"
    out += "Messages:\n"
    out += "\n".join(messages)
    return out

  fence.exposed = True

def run_manually():
  """
  Runs the fencing code manually.  Supervisord is not in the
  picture in this case, since that's not running in this
  mock environment.  This method is useful for testing.
  """
  class NonAgent(object):
    pass
  import threading
  import sys
  import logging

  logging.basicConfig(level=logging.DEBUG)

  agent = NonAgent()
  agent.lock = threading.Lock()
  agent.active_processes = {}

  f = FencingServer(agent)
  LOG.info(str(sys))
  print f.fence(name=sys.argv[1], host=sys.argv[2], port=sys.argv[3], secret_key=None)

if __name__ == "__main__":
  run_manually()
