# Copyright (c) 2016 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import
#  coding=utf-8
import logging
import time

LOG = logging.getLogger("hb_tracker")

class HeartbeatTracker(object):
  '''
  Object to track stats of how long it takes to issue
  heartbeats to CM server.
  Currently we just log the stats once per 10 minutes.
  What is tracked/logged is:
  - min/max of heartbeat times (lifetime)
  - min/max of heartbeat times in the past 10 minutes.
  - mean of heartbeat times in the past 10 minutes.
  - number of heartbeat RPCs in the past 10 minutes.
  '''
  REPORT_PERIOD = 10*60 # 10 minutes
  def __init__(self):
    self.data = []
    self.life_min = 2^32
    self.life_max = 0
    self.last_log = 0
    self.last_hb_duration_secs = 0

  def get_last_hb_latency_secs(self):
    return self.last_hb_duration_secs

  def record(self, start, stop):
    '''
    Record an RPC that started at time "start" and finished at time "stop".
    Log results if the REPORT_PERIOD has passed since the last logging.
    :param start: Start time of RPC
    :param stop:  End time of RPC.
    :return:
    '''
    delta = stop - start
    self.last_hb_duration_secs = delta
    self.data.append(delta)

    if time.time() < self.REPORT_PERIOD + self.last_log:
      return

    self.last_log = time.time()

    if delta < self.life_min:
      self.life_min = delta

    if delta > self.life_max:
      self.life_max = delta

    sum = 0
    min = stop
    max = 0
    for x in self.data:
      if x < min:
        min = x
      if x > max:
        max = x
      sum += x

    mean = sum / len(self.data)
    LOG.info("HB stats (seconds): num:%s LIFE_MIN:%.2f min:%.2f mean:%.2f max:%.2f LIFE_MAX:%.2f",
             len(self.data),
             self.life_min, min, mean, max,self.life_max)
    self.data = []
