#!/usr/bin/env python
from __future__ import absolute_import
# Copyright (c) 2011 Cloudera, Inc. All rights reserved.

import logging
import avro.ipc
import subprocess
import httplib
from M2Crypto import httpslib
from M2Crypto import m2
from M2Crypto import SSL
import urllib2

LOG = logging.getLogger(__name__)




def _make_ssl_context(ca_certs_file,
                      ca_certs_dir,
                      key_file,
                      cert_file,
                      key_pw_file,
                      key_pw_cmd,
                      max_cert_depth=9):
  '''
  Utility function for creating HTTPS connections to CM server.
  It uses M2Crypto's HTTPSConnection in all cases,
  though this has not always been true.

  1) No client or server validation
  M2Crypto's HTTPSConnection will be used.

  2) Only Server validation
  In this case, pass a ca_certs_file only.
  M2Crypto's HTTPSConnection will be used.

  3) Client and Server validation
  In this case, pass all the files.
  M2Crypto's HTTPSConnection will be used.

  4) Only Client validation (kinda)
  In this case, pass all the files except the ca_certs_file.
  M2Crypto's HTTPSConnection will be used.

  Note, however, that M2Crypto does some basic sanity checking
  of the server's certificate - most notably that the CN matches
  the FQDN. An arbitrary toy certificate may not pass this test,
  so some effort is still required.

  Historical note: for a long time, we used httplib's
  HTTPSConnection for case (1), where no validation
  was required.  HTTPSConnection, in python 2.4,
  has an unfortunate interaction with global
  socket timeouts.  We already work around this
  timeout issue in m2crypto, so it was easier
  to use m2crypto in all cases.
  https://twiki.cern.ch/twiki/bin/view/CDS/PythonGotchas#3_2_Incompatibility_between_SSL
  Here's the likely py2.4 bug:
  http://bugs.python.org/issue1380952
  '''
  key_password = None
  if key_pw_cmd != None:
    process = subprocess.Popen(key_pw_cmd, stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE, shell=True)
    proc_out, proc_err = process.communicate()
    if process.returncode:
      # Don't log stdout because maybe in a failure situation it contains
      # the password. Stderr seems more likely to be OK.
      LOG.error("Failure while running \"%s\"; return value %d; stderr :\"%s\"" %
                (key_pw_cmd, process.returncode, proc_err.strip('\n')))
    else:
      LOG.info("Obtained password for private key from \"%s\"" % key_pw_cmd)
      key_password = proc_out.strip('\n')

  elif key_pw_file != None:
    file = None
    try:
      try:
          file = open(key_pw_file, 'r')
          key_password = file.readline().strip('\n')
      except Exception, e:
          LOG.exception('Failed to open key password file: %s', key_pw_file)
    finally:
      if file:
        file.close()

  # Despite the name, 'sslv23' means to dynamically negotiate the
  # protocol version. The NO_SSL options ensure that old versions
  # will not be accepted. The end result is that the Agent will
  # use some version of TLS - hopefully the latest, but that depends
  # on how the server is configured.
  ctx = SSL.Context('sslv23')
  # Since there is no named variable in m2 crypto library for disallowing TLSv1_1
  # using the constant from OpenSSL source.
  # https://github.com/openssl/openssl/blob/OpenSSL_1_0_2k/ssl/ssl.h#L648
  ctx.set_options(m2.SSL_OP_NO_TLSv1 | 0x10000000L | m2.SSL_OP_NO_SSLv2 | m2.SSL_OP_NO_SSLv3)
  ctx.set_session_id_ctx("cm-agent")
  try:
    if key_file != None and cert_file != None:
      # Do client validation
      ctx.load_cert_chain(cert_file, key_file,
                          lambda * arg, **kw: key_password)
    if ca_certs_file == None and ca_certs_dir == None:
      # No server validation
      ctx.set_verify(SSL.verify_none, 0)
    else:
      # Do server validation
      ctx.set_verify(SSL.verify_peer | SSL.verify_fail_if_no_peer_cert,
                     max_cert_depth)
      ctx.load_verify_info(cafile=ca_certs_file, capath=ca_certs_dir)
  except Exception, ssl_error:
    LOG.exception("Error while setting up SSL context")
    raise ssl_error
  return ctx



class HTTPSTransceiver(avro.ipc.HTTPTransceiver):
  '''
  A subclass of HTTPTransceiver that supports TLS
  secured traffic.

  Note that HTTPSTransceiver does not have any
  timeouts.
  '''
  def __init__(self,
               host,
               port,
               ca_certs_file,
               ca_certs_dir,
               key_file,
               cert_file,
               key_pw_file,
               key_pw_cmd,
               max_cert_depth=9):
    ctx = _make_ssl_context(ca_certs_file,
                            ca_certs_dir, key_file, cert_file,
                            key_pw_file, key_pw_cmd, max_cert_depth)
    self.conn = httpslib.HTTPSConnection(host, port,
                                         ssl_context=ctx)
    self.conn.connect()


class SSLContextOpener(object):
  '''
  An instance of this type has an "open" method that can be used in place of urllib2.urlopen.
  This alternative will perform the actions of urllib2.urlopen, but ensure that the SSL configuration
  used for the request uses the SSL context embedded in this object (if any).

  This entire thing is a hack to work around the fact that we can't depend on the presence of Python >=2.7.9,
  where urllib2 has a "context" parameter.
  '''
  def __init__(self, ssl_context, osops):
    self.ssl_context = ssl_context
    self.osops = osops
    self.handler = None
    if ssl_context:
      self.handler = self.__make_https_handler()
      self.opener = urllib2.build_opener(self.handler).open
    else:
      self.opener = urllib2.urlopen

  def __make_https_handler(self):
    ssl_context = self.ssl_context
    '''
    This replicates a bunch of logic implied by the "ssl_context" argument to urllib2.urlopen in
    Python >= 2.7.9.
    :return:
    '''
    class HTTPSHandler(urllib2.HTTPSHandler):
      def https_open(self, req):
        def opener(*pargs, **kwargs):
          kwargs['ssl_context'] = ssl_context
          try:
            # httpslib/ M2Crypto barf if given a timeout argument.
            # We ignore timeouts for SSL sockets due to OPSAPS-4306
            del kwargs['timeout']
          except:
            pass
          return httpslib.HTTPSConnection(*pargs, **kwargs)
        return self.do_open(opener, req)

    return HTTPSHandler


  def open(self, url, *pargs, **kwargs):
    # OPSAPS-41908 - Guard against passing unicode to M2Crypto
    if isinstance(url, unicode):
      url = str(url)
    return self.opener(url, *pargs, **kwargs)


  def fetch_to_file(self, req_url, target_file):
    try:
      resp = self.open(req_url)
      with self.osops.open_write_file(target_file, mode=0644) as f:
        f.write(resp.read())
    except Exception, e:
      LOG.exception("Failed to retrieve/store URL: %s -> %s %s", req_url, target_file, e)
      raise

ssl_url_opener = None

def make_ssl_context(cfg):
  if cfg.use_tls:
    return _make_ssl_context(cfg.verify_cert_file,
                             cfg.verify_cert_dir,
                             cfg.client_key_file,
                             cfg.client_cert_file,
                             cfg.client_keypw_file,
                             cfg.client_keypw_cmd,
                             cfg.max_cert_depth)
  return None


def monkey_patch_34952():
  """
  Exceptions in "urlopen" due to 404 errors leak file descriptors.
  This monkey-patching ensures the file descriptor is closed.
  :return:
  """
  old = urllib2.HTTPDefaultErrorHandler.http_error_default
  def http_error_default(self, req, fp, code, msg, hdrs):
    try:
      return old(self, req, fp, code, msg, hdrs)
    except Exception, e:
      try:
        fp.close()
      except Exception:
        pass
      raise e

  urllib2.HTTPDefaultErrorHandler.http_error_default = http_error_default


monkey_patch_34952_done = False
if not monkey_patch_34952_done:
  monkey_patch_34952()
  monkey_patch_34952_done = True
