# Copyright (c) 2013 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging
import time
import os
import subprocess
import tempfile
import threading

from cmf.monitor import constants
from cmf.throttling_logger import ThrottlingLogger

KINIT_BIN = os.environ.get('KINIT_BIN', 'kinit')
KDESTROY_BIN = os.environ.get('KDESTROY_BIN', 'kdestroy')
LOG = logging.getLogger(__name__)
THROTTLED_LOG = ThrottlingLogger(LOG, constants.DEFAULT_MONITOR_LOG_RATE)

#
# A pathological KDC can renew every 10 minutes. But we don't
# want to renew on the dot. So give it some slack if it's almost
# at the limit.
#
_DEFAULT_RENEWAL_PERIOD = 10 * 60  # 10 minutes
_RENEWAL_SLACK = 10                # 10 seconds
_RENEWAL_PERIOD = \
  float(os.environ.get('KRB_RENEWAL_PERIOD', _DEFAULT_RENEWAL_PERIOD))
_CANDIDATES_LAST_KINIT_TIMESTAMP_KEY = "__candidates__"

_cred_manager = None
_cred_manager_lock = threading.Lock()  # protects singleton creation

#
# We suffix the cache name with the effective uid as the directory that is
# created is only accessible to the user who created it. See OPSAPS-23588 for
# more details.
#
_AGENT_CCACHE_FILE_NAME = 'krb5cc_cm_agent_' + str(os.geteuid())

class CredentialManager(threading.Thread):
  """
  A thread to periodically kinit to make sure that the agent has a valid ticket.
  (It doesn't matter who the principal is.) There are two challenges to
  consider.

  First, a valid (keytab, principal) combo could become invalid at any time. The
  CM user could've done a regenerate credential. When this happens, we want to
  throw away this cred and use something else.

  Second, the kinit could fail on a valid (keytab, principal) combo due to
  transient errors, like KDC going offline. We shouldn't throw away that valid
  credential.

  So we keep a map of candidate creds. When a Process activates and attempt to
  kinit, we steal and stash its creds (stored in `_candidates'). The thread main
  loop periodically tries to reinit using the first candidate. On failure, it'll
  move to the next one. And so on. When a Process deactivates its credentials
  are removed.
  """

  def __init__(self):
    threading.Thread.__init__(self, name='CredentialManager')
    self.setDaemon(True)

    self._should_stop = False

    # A map of (keytab, princ) that could potentially work
    self._candidates = {}
    # A map between a keytab and the last time it was kinited. We use a special
    # key for the candidates and the keytab as the key for the specifics.
    self._last_kinit = {}
    # a tuple (keytab, princ) of the last candidate we successfully called kinit
    # with.
    self._last_successful_candidate = None
    # A list of (keytab, principal, credential_cache_name) to renew.
    # These will be renewed every RENEWAL_PERIOD minutes.
    self._specifics = []
    self._lock = threading.Lock()
    self._trigger = threading.Condition(self._lock)
    self._numRuns = 0
    self._agent_ccache_name = None
    # Initialize our credential cache to use the system wide temp directory
    self.set_agent_wide_credential_cache_directory(tempfile.gettempdir(),
                                                   log=False)

  @property
  def num_runs(self):
    """Used for testing"""
    return self._numRuns

  @property
  def last_candidate_used(self):
    """
    The last candidate used. Only called by tests
    """
    return self._last_successful_candidate

  def add_candidate(self, keytab, princ):
    """
    Adds a keytab and a principal to the map of candidates we will attempt
    to kinit from. Kinit-int is done asynchronously.
    """
    self._lock.acquire()
    try:
      self._candidates[keytab] = princ
      self._trigger.notify()
    finally:
      self._lock.release()

  def remove_candidate(self, keytab):
    """
    Removes a candidate from the list of candidates we attempt to kinit from.
    """
    self._lock.acquire()
    try:
      if keytab in self._candidates:
        del self._candidates[keytab]
    finally:
      self._lock.release()

  def run(self):
    self._lock.acquire()
    while not self._should_stop:
      try:
        if LOG.isEnabledFor(logging.DEBUG):
          LOG.debug("Running with candidates: \n  %s" %
                    ("\n  ".join([str(c) for c in self._candidates.items()])))

        if self._should_kinit_key(_CANDIDATES_LAST_KINIT_TIMESTAMP_KEY):
          # Walk through all candidates retry until we hopefully find a valid
          # one.
          candidates = self._candidates.items()
          if self._last_successful_candidate is not None:
            # try our last successful candidate first
            candidates.insert(0, self._last_successful_candidate)

          for keytab, princ in candidates:
            cc_name = self._agent_ccache_name
            self._lock.release()
            try:
              res = _do_actual_kinit(keytab, princ, cc_name)
            finally:
              self._lock.acquire()
            # Note that we do not reset the last successful candidate on
            # failure: We assume that failures are temporary and as long as the
            # keytab belongs to a valid process and we could not kinit from any
            # other candidate the last successful candidate is a good bet to try
            # with first.
            if 0 == res:
              self._last_kinit[_CANDIDATES_LAST_KINIT_TIMESTAMP_KEY] = \
                time.time()
              if keytab in self._candidates:
                self._last_successful_candidate = (keytab, princ)
              else:
                self._last_successful_candidate = None
              break

        for keytab, princ, cc_name in self._specifics:
          if not self._should_kinit_key(keytab):
            continue
          self._lock.release()
          # we need this awkward construct for python 2.4
          try:
            try:
              res = _do_actual_kinit(keytab, princ, cc_name)
            except Exception, ex:
              LOG.exception(
                "Could not kinit specific principal %s with keytab %s. %s" %
                (princ, keytab, ex,))
          finally:
            self._lock.acquire()
            if res == 0:
                self._last_kinit[keytab] = time.time()

      except Exception, ex:
        LOG.exception(
          "Caught unexpected error. Will try to recover. %s" % (ex,))

      self._numRuns += 1

      # Sleep for the renewal period
      self._trigger.wait(_RENEWAL_PERIOD)

    self._lock.release()

  def stop(self):
    self._lock.acquire()
    try:
      self._should_stop = True
      self._trigger.notify()
    finally:
      self._lock.release()

  def _async_kinit_multi_for_test(self, *credentials):
    """
    credentials is a bunch of (keytab, princ) tuples.
    """
    self._lock.acquire()
    try:
      # Push onto the map and trigger
      cred_list = list(credentials)
      cred_list.reverse()
      for keytab, princ in cred_list:
        self._candidates[keytab] = princ
      self._trigger.notify()
    finally:
      self._lock.release()

  def add_specific(self, keytab, principal, cc_name):
    LOG.info("Adding principal %s to kinit using %s keytab, using %s for "
             "credential cache" % (principal, keytab, cc_name))

    self._lock.acquire()
    try:
      # append to the list and trigger
      self._specifics.append((keytab, principal, cc_name))
      self._last_kinit[keytab] = 0
      self._trigger.notify()
    finally:
      self._lock.release()

  def remove_specific(self, princ_to_remove, keytab_path,
                      destroy_credentials=True):
    self._lock.acquire()
    try:
      self._specifics = [record for record in self._specifics if
                         self._should_keep(record,
                                           princ_to_remove,
                                           keytab_path,
                                           destroy_credentials)]
    finally:
      self._lock.release()

  def _should_keep(self,
                   record,
                   princ_to_remove,
                   keytab_path,
                   destroy_credentials):
    # record is of the form of (keytab, principal, cc_name)
    if record[1] == princ_to_remove and record[0] == keytab_path:
      LOG.info("Removing principal %s using keytab at %s from the list of "
               "principal to kinit. destroy_credentials is: %s" %
               (princ_to_remove, keytab_path, destroy_credentials))
      if destroy_credentials:
        _do_actual_kdestroy(record[2])
        del self._last_kinit[keytab_path]
      return False
    return True

  def _should_kinit_key(self, keytab):
    if keytab not in self._last_kinit:
      return True
    now = time.time()
    last_renewal = self._last_kinit[keytab]
    return now - last_renewal >= _RENEWAL_PERIOD - _RENEWAL_SLACK

  def set_agent_wide_credential_cache_directory(self, directory, log=True):
    if directory is None:
        raise Exception("Directory cannot be None")

    self._lock.acquire()
    try:
      self._agent_ccache_name = os.path.join(directory, _AGENT_CCACHE_FILE_NAME)
      # This will ensure that next time we run we will recreate the credential
      # cache. Note that we do not check to see if the previous value is identical
      # to the new value as one spurious kinit is inconsequential.
      self._last_kinit.pop(_CANDIDATES_LAST_KINIT_TIMESTAMP_KEY, None)
      # Setting KRB5CCNAME will switch gssapi_krb5 library to use a FILE credential
      # cache with cache_name instead of the default <tmp dir>/krb5cc_<uid>
      # credential cache. To use a thread specific cache name see kerberos package
      # setGSSCredentialCacheName function.
      os.environ['KRB5CCNAME'] = self._agent_ccache_name
      self._trigger.notify()
      if log:
        LOG.info("Agent wide credential cache set to %s" % self._agent_ccache_name)
    finally:
      self._lock.release()

def add_keytab_to_kinit_from(keytab_path, principal):
  """
  add a keytab to a list of keytabs the agent will attempt to use to kinit
  the 'principal'. The actual kinit-ing is done in a separate thread. When the
  keytab path supplied here is no longer valid (e.g., the process is finished)
  it is the responsibility of the caller to notify the kt_renewer by calling
  remove_keytab_to_kinit_from.
  """
  _get_cred_manager().add_candidate(keytab_path, principal)
  THROTTLED_LOG.info("Added principal %s with keytab %s as a candidate to "
                     "kinit" % (principal, keytab_path,))

def remove_keytab_to_kinit_from(keytab):
  _get_cred_manager().remove_candidate(keytab)
  THROTTLED_LOG.info("Removed keytab %s as a candidate to kinit from" %
                     (keytab,))

def add_specific_principal_to_kinit(keytab_path, principal, cc_name):
  """
  Add a principal to kinit using the keytab at 'keytab_path' and storing
  the obtained ticket at the credential cache at 'cc_name'. The principal
  will be re-kinit every 10 minutes.
  """
  _get_cred_manager().add_specific(keytab_path, principal, cc_name)

def remove_specific_principal_to_kinit(principal,
                                       keytab,
                                       destroy_credentials=True):
  """
  Remove a principal from the list of principal to kinit. If 
  'destroy_credentials' is True kdestroy will be called.
  """
  _get_cred_manager().remove_specific(principal, keytab, destroy_credentials)

def find_keytab_for_principal(directory, principal):
  """
  Find a keytab containing a record for the principal in 'directory'. Returns
  the full keytab path if found None otherwise. Note that 'principal' should
  be the principal prefix and not the full, including-hostname, principal name
  (e.g., hdfs and not hdfs/foobar.com)
  """
  for root, _, files in os.walk(directory):
    for f in files:
      if f.endswith(".keytab"):
        f_path = os.path.join(root, f)

        # Is our principal potentially in there?
        if principal not in file(f_path).read():
          continue
        return f_path
  return None


def get_full_principal_name(principal_prefix, hostname):
  """
  Returns the full principal name that includes the prefix and the hostname.
  For example, for (hdfs, foobar.com) the full name hdfs/foobar.com will be
  returned.
  """
  return "%s/%s" % (principal_prefix, hostname)


def set_agent_wide_credential_cache_directory(directory):
  """
  Set the credential cache directory shared by all agent component that do not
  use a thread-specific credential cache. If kinit was already called and used a
  different credential cache there may be a delay until the new credential cache
  is created and authentication may fail in the meantime. Make sure to call this
  function very early (before any kinit has happened) or ensure that a kinit
  happens afterwards.
  """
  _get_cred_manager().set_agent_wide_credential_cache_directory(directory)

def _get_cred_manager():
  global _cred_manager
  if _cred_manager is not None:
    return _cred_manager

  _cred_manager_lock.acquire()
  try:
    if _cred_manager is None:
      cmgr = CredentialManager()
      cmgr.start()
      _cred_manager = cmgr
  finally:
    _cred_manager_lock.release()

  return _cred_manager


def stop_cred_manager():
  """Used for testing"""
  global _cred_manager
  _cred_manager_lock.acquire()
  try:
    if _cred_manager is not None:
      _cred_manager.stop()
    _cred_manager = None
    return True
  finally:
    _cred_manager_lock.release()


def _do_actual_kinit(keytab_path, principal, cc_name):
  """
  Do the actual kinit work. Returns:
    0:          success
    -1:        on failure
  """
  cmdv = [KINIT_BIN,
          '-k',
          '-t', str(keytab_path),
          '-c', str(cc_name),
          str(principal)]
  LOG.debug("kinit as '%s' using %s, cc name %s" %
            (principal, keytab_path, cc_name))
  popen = subprocess.Popen(cmdv,
                           stdout=subprocess.PIPE,
                           stderr=subprocess.STDOUT)
  stdout = popen.communicate()[0]
  if popen.returncode != 0:
    LOG.warn("Couldn't kinit as '%s' using %s --- %s" %
             (principal, keytab_path, stdout))
    return -1

  return 0


def _do_actual_kdestroy(cc_name):
  """
  Call kdestroy and destroy the credential in cc_name.
  """
  cmdv = [KDESTROY_BIN,
          '-c', str(cc_name)]
  LOG.debug("%s '%s'" % (KDESTROY_BIN, cc_name))
  popen = subprocess.Popen(cmdv,
                           stdout=subprocess.PIPE,
                           stderr=subprocess.STDOUT)
  stdout = popen.communicate()[0]
  if popen.returncode != 0:
    LOG.warn("Couldn't kdestroy '%s' --- %s" %
             (cc_name, stdout))
