# Copyright (c) 2014 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import os
import tempfile

from cmf.zip_util import ZipFile, ZIP_DEFLATED

def get_compressed_logs(path_list,
                        max_bytes_to_collect,
                        top_level_dir,
                        redactor,
                        file_open_function=lambda f: open(f, "rb")):
  """This function returns at most max_bytes_to_collect uncompressed files in
     a compressed zip file. It takes a list of file paths and writes those
     files to the given top level directory in the zip file.

     This function takes a limit on the number of bytes to collect (the limit is
     on the size before the files are zipped). If that limit is hit, a portion
     of a file may be included. That portion will be at the end of the file not
     at the beginning. This is based on the assumptiont that the list of files
     is an ordered list of log files, ordered most recent to least recent. It is
     also assumed that the log files contents are ordered old messages to new
     messages. As such, by taking the end of the least recent file we take, we
     guarantee that we're taking the log messages from that file contiguous with
     the more recent messages in the other included files.

     The archive resides in a temporary directory, and it is the caller's
     responsibility to delete the archive after using it.

     An optional file_open_function may be passed in to customize how the files
     in path_list are opened. This may be useful to perform certain checks at
     open, see util.py validate_and_open for an example."""

  tmp_os_file_descriptor, tmp_file_path = tempfile.mkstemp()
  tmp_file_handle = None
  zip_fd = None
  fd = None

  try:
    tmp_file_handle = os.fdopen(tmp_os_file_descriptor, "w")
    zip_fd = ZipFile(tmp_file_handle, "w", ZIP_DEFLATED)

    bytes_collected = 0
    for curr_file_path in path_list:
      if bytes_collected >= max_bytes_to_collect:
        break
      if not os.path.exists(curr_file_path):
        continue
      fd = file_open_function(curr_file_path)
      st = os.stat(curr_file_path)
      curr_file_length = st.st_size
      curr_file_name = os.path.basename(curr_file_path)
      arcname = top_level_dir + "/" + curr_file_name
      if bytes_collected + curr_file_length > max_bytes_to_collect:
        remaining_bytes = max_bytes_to_collect - bytes_collected
        fd.seek(curr_file_length - remaining_bytes)
        bytes_collected += remaining_bytes
      else:
        bytes_collected += curr_file_length
      zip_fd.write_fp(fd, st, arcname, redactor=redactor)
  finally:
    if fd:
      fd.close()
    if zip_fd:
      zip_fd.close()
    if tmp_file_handle:
      tmp_file_handle.close()

  return tmp_file_path
