#!/usr/bin/env python
from __future__ import absolute_import
# Copyright (c) 2012 Cloudera, Inc. All rights reserved.
import logging
import re
import os
import time

LOG = logging.getLogger(__name__)

ALWAYS = -1
NEVER = 0
ATTRIBUTE_KEYWORD = "attribute:"

# This class is a python implmentation of LogEventFilterer.java and is responsible
# for determining whether the new line should trigger an event
# This is missing two elements from FilterConfiguration because they don't apply
# to all logs:
#   - match_in_exception
#   - threshold
class FilterConfiguration(object):
  def __init__(self, logs_and_rules_json):
     self.file_rules = {}
     for filepath in logs_and_rules_json:
        rules = logs_and_rules_json[filepath]
        for rule in rules["rules"]:
          # Check to make sure the rule is valid
          # Rate and content are required, but period
          # has a default, as in the log4j implementation
          rate = self._get_rate(rule)
          if rate is None:
            LOG.warn("Rate not an integer for file: " + filepath)
            continue      
          period = self._get_period(rule)
          if period <= 0 and rate != ALWAYS and rate != NEVER:
            LOG.warn("Period should be a positive integer if a rate other than "\
                "always or never is specified: " + filepath)
            continue 
          content = rule.get('content')
          if content is None:
            LOG.warn("Content cannot be empty: " + filepath)
            continue
          groups = self._get_groups(rule)
          attributes = self._get_attributes(rule)
          self.file_rules.setdefault(filepath, []).append(CategoryRule
              (content, rate, period, groups, attributes))

  def _get_rate(self, rule):
    try:
      return int(rule.get('rate'))
    except ValueError:
      return None

  def _get_period(self, rule):
    try:
      return int(rule.get('period'))
    except ValueError:
      return 1 # Default is one minute

  # The group attribute configuration is "groupX:name" where X is the
  # number of the group in the regex and name is the key of the
  # corresponding event attribute you want to generate
  def _get_groups(self, rule):
     groups = []
     index = 0
     while True:
       group_attr = rule.get("group" + str(index))
       if group_attr is None:
         break
       else:
         groups.append(group_attr)
         index = index + 1
     return groups

  def _get_attributes(self, rule):
    attributes = {}
    for key in rule:
      if key.startswith(ATTRIBUTE_KEYWORD):
        attributes[key[len(ATTRIBUTE_KEYWORD):]] = rule[key]
    return attributes

  # Returns None if the line does not match and a dict of attributes
  # to add to the event if it does match
  def match(self, pathname, line):
    for rule in self.file_rules[pathname]:
      attributes = rule.check_line(line)
      if attributes is not None:
        return attributes
    return None

# This class contains the rule information and state
class CategoryRule(object):

  def __init__(self, content, rate, period, groups, attributes):
    self.regex = re.compile(content, re.MULTILINE | re.IGNORECASE)
    self.rate = rate
    self.period_secs = period * 60 # Convert minutes to seconds
    self.groups = groups
    self.attributes = attributes
    self.count_in_period = 0
    self.last_period = 0

  # Returns a list of attributes if the event should be sent, None
  # if the event should not be sent
  def check_line(self, line):
    match = re.search(self.regex, line)
    if not match:
      return None
 
    if not self._check_rate_limit():
      return None

    # Gather the group and default attributes
    group_attributes = {}
    group_count = 1
    for group in self.groups:
      group_attributes[group] = match.group(group_count)
      group_count = group_count + 1
    return dict(group_attributes.items() + self.attributes.items())

  def _check_rate_limit(self):
    if self.rate == NEVER:
      return False

    if self.rate == ALWAYS:
      return True

    ts = int(time.time())
    current_period = ts / self.period_secs
    if self.last_period == current_period:
      if self.count_in_period >= self.rate:
        return False
      else:
        self.count_in_period = self.count_in_period + 1
        return True
    else:
      self.count_in_period = 1
      self.last_period = current_period
      return True
