# Copyright (c) 2012 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import
import time
import os
import simplejson as json
import gzip
import traceback
import cmf.util as util
import logging
try:
  from cStringIO import StringIO
except:
  from StringIO import StringIO

from clusterstats.log.streaming.file import LogFile
from clusterstats.log.streaming.event_reader import LogFileEventReader
from clusterstats.log.streaming.event_parser import Log4JEventParser, HueEventParser, GLogEventParser, ScmAgentEventParser

from clusterstats.log.streaming.file_filter import DirectoryFilter
from clusterstats.log.streaming.file_filter import FilePatternFilter
from clusterstats.log.streaming.file_filter import TimeRangeFilter
import clusterstats.common.chain as chain

import cmf.log_search as log_search

TEN_SECONDS_IN_MILLIS=10000
LOG = logging.getLogger(__name__)

HUE = "HUE"
LOG4J = "LOG4J"
GLOG = "GLOG"
SCMAGENT = "SCMAGENT"

PYTHON_LOG_TYPES = set([HUE, SCMAGENT])
VALID_LOG_TYPES = set([LOG4J, GLOG]) | PYTHON_LOG_TYPES

EVENT_PARSER_CLASSES = {
  LOG4J: Log4JEventParser,
  HUE: HueEventParser,
  GLOG: GLogEventParser,
  SCMAGENT: ScmAgentEventParser
}

def retrieve_log_compressed(end_time,
                           log_path,
                           log_type,
                           total_byte_limit,
                           redactor,
                           search_timeout_millis=TEN_SECONDS_IN_MILLIS,
                           start_time=None):
  """
  Retrieves a compressed portions of a plaintext log file. If available,
  total_byte_limit bytes will be collected from this file and other rolled
  files, with the latest event being at end_time. However, if start_time is
  available, all the bytes between start_time and end_time will be collected
  (max_bytes will not be used as the limiting factor).
  A log type (valid values in VALID_LOG_TYPES) must also be passed.
  An optional search timeout can be used to stop collecting log data.

  This will return a generator that outputs a gzip file of the log events.
  """
  compressor = LogCompressor(log_path, log_type, start_time, end_time, total_byte_limit)
  log_gen = compressor.compress_log(search_timeout_millis, redactor)

  return log_gen

def build_gzip_generator_single(log_file_handle, start_offset, end_offset, redactor, deadline=None):
  """
  Gets a generator for gzipped content from the given log file handle, start offset, and end offset.
  The file handle should already be open. This method will close it.
  The generator will return gzipped content of the file from start offset
  to end offset.
  If a deadline is passed, compression will stop immediately if the deadline time
  is hit.
  """

  CHUNK_SIZE = 128*1024
  if start_offset == end_offset:
    return

  log_file_handle.seek(start_offset)

  bytes_to_read = end_offset - start_offset
  while bytes_to_read > 0:
    if deadline != None and time.time() > deadline:
      log_file_handle.close()
      return

    buf = StringIO()
    events_gzip_fd = gzip.GzipFile(mode="w", fileobj=buf)

    read_size = min(CHUNK_SIZE, bytes_to_read)
    redact_buf = log_file_handle.read(read_size).decode('UTF-8')

    # We don't want to start/end our redaction in the middle of a
    # line, so we back up to the end of the last line.
    retidx=redact_buf.rfind('\n')
    if retidx is not -1:
        retidx += 1  # include the \n
        # Move backwards in file to retidx (before redaction)
        log_file_handle.seek(retidx - len(redact_buf), 1)
        # And pretend we didn't really read data after retidx
        read_size = retidx
        # Note that buffer may change in size due to redaction
        redact_buf = redact_buf[:retidx]

    redact_buf = redactor.redact(redact_buf)

    events_gzip_fd.write(redact_buf.encode('UTF-8'))
    bytes_to_read -= read_size
    events_gzip_fd.close()
    yield buf.getvalue()

  # Done with this file.
  log_file_handle.close()

def build_gzip_generator_many(file_start_end_tuples, redactor, deadline=None):
  """
  For a list of LogFile, start offset, end offset tuples, creates a
  composite generator by calling build_gzip_generator on each file.
  """
  for (log_file, start_offset, end_offset) in file_start_end_tuples:
    log_file_handle = log_file.get_handle()
    for val in build_gzip_generator_single(log_file_handle, start_offset, end_offset, redactor, deadline):
      yield val

def calculate_log_estimate(start_time,
                           end_time,
                           log_paths,
                           log_types,
                           estimate_timeout_millis):
  """
  Calculates the available bytes between the start_time and end_time for
  all files matching the log_paths (log_paths can be in a regex format to match all
  rolling logs for a log path)

  :param start_time: Time at which the log estimate should start
  :param end_time: Time at which the log estimate should end
  :param log_paths: List of paths to logs for which the estimate should be calculated
  :param log_types: List of types of logs, like LOG4J or GLOG, we parse each log type
   differently
  :param estimate_timeout_millis: Time to wait until we cause a timeout for estimating
  a log path
  """
  estimator = LogEstimator(log_paths, log_types, start_time, end_time)
  estimated_bytes = estimator.estimate_log_size_in_bytes(estimate_timeout_millis)
  LOG.info("Estimated bytes between start time=%s and end time=%s: %s",
           start_time, end_time, estimated_bytes)
  result = dict(
    success=True,
    estimate=estimated_bytes)
  data_string = json.dumps(result, ensure_ascii=False)

  return {
    "data": unicode(data_string)
  }

"""
Wraps the functionality related to retrieving a set of log files.
"""
class LogRetriever(object):
  def __init__(self, log_paths, log_types, start_time, end_time):
    self._log_paths = log_paths
    self._log_types = log_types
    self._start_time = start_time
    self._end_time = end_time

  def get_byte_offsets(self, log_file, log_type, start_time, end_time, max_size, deadline_ms=None):
    """If start_time is not specified, this returns the byte offsets of the LogFile object log_file ending
    at end time and taking up at most max_size bytes. However if specified, the byte offsets of start_time
    and end_time are returned ignoring the max_size bytes.
    Returns start_offset, end_offset"""
    if max_size is not None and max_size <= 0:
      return 0, 0
    if end_time is None:
      return 0, 0
    if start_time is not None and start_time > end_time:
      return 0, 0
    if deadline_ms != None and time.time() > deadline_ms:
      raise Exception("Timeout Error: Deadline crossed")

    end_offset = log_file.get_end_offset()
    end_offset = log_search.offset_from_timestamp(end_time, end_offset, log_file.get_path(), log_type, deadline_ms)
    if start_time is not None:
      # As start_time is available, just directly use it to get the start offset.
      # It is not limited by max_size
      # This handles the requests coming in with a start_time, for example from
      # LogEstimator.estimate_log_size_in_bytes(...)
      start_offset = log_search.offset_from_timestamp(start_time, end_offset, log_file.get_path(),
                                                      log_type, deadline_ms)
    else:
      # This may collect very old data up to max_bytes. If end_time is long after
      # the data in the log path ends, we will just collect max_size from the end
      # of the file.
      start_offset = end_offset - max_size
      if start_offset <= 0:
        return 0, end_offset
      start_offset = self.get_offset_after(log_file.get_path(), log_type, start_offset, deadline_ms)
      if start_offset == end_offset:
        # This means that the log event is larger than the max_size
        # We should get the max_size form the top of the current log event
        start_offset = self.get_offset_before(log_file.get_path(), log_type, start_offset, deadline_ms)
        end_offset = start_offset + max_size
    return start_offset, end_offset

  def get_offset_after(self, log_path, log_type, offset, deadline_ms=None):
    """Gets the offset of a log line after a given offset. This is used
    to move to the next entity after a given offset to make sure we don't
    overrun a maximum size or cut a log line in half.
    If an error occurs or this fails, offset will be returned."""
    f = LogFile(log_path)
    handle = f.open_log_file()

    try:
      event_parser = EVENT_PARSER_CLASSES[log_type]()
      try:
        reader = LogFileEventReader(
          event_parser=event_parser,
          file_handle=handle,
          file_path=log_path,
          start_offset=offset,
          deadline_ms=deadline_ms)
        prev_event = reader.find_event_containing_pos(offset)
        event = None
        if prev_event is not None:
          event = reader.get_next_event()
        if event is not None:
          return event.get_offset()
        return offset
      except:
        return offset
    finally:
      handle.close()

  def get_offset_before(self, log_path, log_type, offset, deadline_ms=None):
    """Gets the starting offset of the log event containing the specified offset.
    If an error occurs or this fails, offset will be returned."""
    f = LogFile(log_path)
    handle = f.open_log_file()

    try:
      event_parser = EVENT_PARSER_CLASSES[log_type]()
      try:
        reader = LogFileEventReader(
          event_parser=event_parser,
          file_handle=handle,
          file_path=log_path,
          start_offset=offset,
          deadline_ms=deadline_ms)
        event = reader.find_event_containing_pos(offset)
        if event is not None:
          return event.get_offset()
        return offset
      except:
        return offset
    finally:
      handle.close()

  def get_log_files(self, log_path):
    """Gets the log files that match log_path* for the log_path this object
    was created with.
    """
    log_dir, log_name = os.path.split(log_path)
    file_pattern_filter = FilePatternFilter(log_name)

    filters = chain.build_chain([DirectoryFilter(), file_pattern_filter])
    file_list = []

    if not os.path.exists(log_dir):
      return file_list

    for f in os.listdir(log_dir):
      file_path = os.path.join(log_dir, f)
      if os.path.exists(file_path):
        file_list.append(LogFile(file_path))
      else:
        LOG.info("Skipping. File not found '%s'." % file_path)

    filter_context = { "file_list": file_list }
    filters(filter_context)

    return filter_context["file_list"]

  def sort_files(self, file_list):
    """
    Returns a list of files sorted by time where the newest file is
    returned first.
    """
    return sorted(file_list, key=lambda x: self._get_log_file_index(x))

  def _get_log_file_index(self, log_file):
    return self.get_log_index(log_file.get_path())

  def get_log_index(self, file_name):
    """Gets an appropriate index for log files where newer files have a smaller
    index. For files that are named
    x.log, x.log.1, x.log.2... this will be 0, 1, 2... For impala style
    logs like:
    impalad.nightly-5.ent.cloudera.com.impala.log.INFO.20121130-061408.14555
    impalad.nightly-5.ent.cloudera.com.impala.log.INFO.20121130-155821.24443
    ...
    This will be MAX_LONG - time(20121130-061408), MAX_LONG - time(20121130-155821), ..."""
    IMPALA_TIME_FORMAT = "%Y%m%d-%H%M%S"
    MAX_LONG = 2**63 - 1

    parts = file_name.split(".")
    if len(parts) > 2:
      try:
        t = time.strptime(parts[-2], IMPALA_TIME_FORMAT)
        return MAX_LONG - time.mktime(t)
      except:
        # Didn't parse. We expect this for non-impala names.
        pass
    try:
      return int(parts[-1])
    except:
      # No number on end of file. Default to 0
      return 0

"""
Wraps the functionality of compressing a set of log files.
"""
class LogCompressor(LogRetriever):
  def __init__(self, log_paths, log_types, start_time, end_time, total_byte_limit):
    super(LogCompressor, self).__init__(log_paths, log_types, start_time, end_time)
    self._total_byte_limit = total_byte_limit

  def compress_log(self, compress_timeout_millis, redactor):
    """
    Compresses the log files associated with this object when it was initialized into
    the passed directory. This may need to handle rolled over log files.
    If total compress time exceeds compress_timeout_millis then incomplete log files
    will be created.
    This method is a generator for the gzipped data
    """
    deadline = time.time() + int(compress_timeout_millis)/1000.0

    log_files = self.sort_files(self.get_log_files(self._log_paths))
    files_to_stream = self._prepare_files_for_streaming(log_files, self._log_types, self._start_time,
                                                        self._end_time, self._total_byte_limit, deadline)
    # Scan these from oldest to newest
    return build_gzip_generator_many(files_to_stream, redactor, deadline)

  def _prepare_files_for_streaming(self, log_files, log_types, start_time, end_time, byte_limit,
                                   deadline_ms=None):
    """Returns a list of LogFile, start offset, end offset tuples to stream. This
    is intended to be used to generate a list of log files that is ready for streaming.
    log_files should be a list of LogFile objects sorted from newest to oldest.
    The returned list will be ordered from oldest to newest.
    If start_time is not specified, then total offsets will not exceed total byte limit.
    Else, the total offsets will be directly calculated from start_time and end_time
    respectively; and will not be limited by max_bytes.
    The file handles will be open for each log file that should be streamed.
    We need to open the files here in case a rollover occurs during our processing.
    deadline_ms is the max time-limit (in millis since epoch) for performing
    any further execution.
    In case of an exception or the deadline_ms being passed, it returns the list of
    files processed till that stage."""
    result = []
    try:
      for f in log_files:
        start, end = self.get_byte_offsets(f, log_types, start_time, end_time, byte_limit, deadline_ms)
        if start != end:
          f.open_log_file()
        else:
          return reversed(result)
        result.append((f, start, end))

        # Change the number of bytes that remain
        byte_limit -= end - start
    except Exception:
      print(traceback.format_exc())
    return reversed(result)

"""
Wraps the functionality of calculating the size of a set of log files.
"""
class LogEstimator(LogRetriever):
  def __init__(self, log_paths, log_types, start_time, end_time):
    super(LogEstimator, self).__init__(log_paths, log_types, start_time, end_time)

  def estimate_log_size_in_bytes(self, estimate_timeout_millis):
    """
    Estimates the log files associated with this object when it was initialized into the
    passed directory. This may need to handle rolled over log files.
    :param estimate_timeout_millis: time in ms until a time out is triggered
    """
    total_estimated_size_in_bytes = 0
    deadline = None
    if estimate_timeout_millis is not None:
      deadline = time.time() + int(estimate_timeout_millis) / 1000.0
    for (log_path, log_type) in zip(self._log_paths, self._log_types):
      log_files = self.get_log_files(log_path)
      for log_file in log_files:
        start_offset, end_offset = self.get_byte_offsets(log_file, log_type, self._start_time,
                                                         self._end_time, None, deadline)
        estimated_size_in_bytes = end_offset - start_offset
        if estimated_size_in_bytes < 0:
          # This should never happen
          LOG.error("End time offset %s is less than the start time offset %s."
                    "Requested log file %s with estimate timeout %s has total offset of %s",
                    end_offset, start_offset, log_file.get_path(), deadline, log_file.get_end_offset())
          raise Exception("Offset Error: End time offset is less than the start time offset")
        total_estimated_size_in_bytes += estimated_size_in_bytes

    return total_estimated_size_in_bytes



