#!/usr/bin/env python
from __future__ import absolute_import
# Copyright (c) 2012 Cloudera, Inc. All rights reserved.
import simplejson as json
import logging
import cmf.log_filterer as log_filterer
import os
import pyinotify
import sys
from time import sleep, time

import avro.ipc as ipc
import avro.protocol as protocol

EVENTS_PROTOCOL = protocol.parse(
    file(os.path.join(os.path.dirname(__file__), "events.avpr")).read())

LOG = logging.getLogger(__name__)

DEFAULT_PYINOTIFY_TIMEOUT_IN_MILLIS = 10

# This class allows the agent to monitor log files. It takes in a
# configuration from the agent, watches the specified files using pyinotify,
# and generates events if the file changes match any of the configured
# rules.
#
# It has two monitoring components:
# 1. Monitors file modifications. This reads to the end of the file when
#    a file change happens and runs each line against the FilterConfiguration
# 2. Monitors changes on the directories. This looks for file creations,
#    deletions, moved to/from. We use it to watch for events like log
#    rotation and ensure that we're monitoring the correct paths
#
# Note that currently we don't monitor for directory creation / deletion.
# We believe that this should be a pretty rare event and when it does
# happen we refresh the directories we're monitoring when we receive
# a heartbeat
class LogTailer(pyinotify.ProcessEvent):

  # Only unit tests should actually pass in an event_proxy implementation
  def __init__(self, host_id, host_name, event_proxy=None):
    if event_proxy is None:
      event_proxy = EventServerProxy()
    self.event_proxy = event_proxy
    self.inotify = None
    self.directory_inotify = None
    self.files = {}
    self.paths_to_watch = []
    self.potential_paths = []
    self.host_id = host_id
    self.host_name = host_name

  def set_eventserver_config(self, eventserver_host, eventserver_port):
    self.event_proxy.set_event_server_config(eventserver_host, eventserver_port)

  def set_config(self, rules_config):
    # If need be we can hash the config and only run this code if it has changed from
    # the last time - note that the directories should be refreshed even if the
    # hash has not changed because we don't monitor directory creation
    self.config_json = rules_config
    try:
      self.logs_and_rules_json = json.loads(self.config_json)
    except:
      LOG.warn("Invalid logs and rules json")
      return
    self.filter_config = log_filterer.FilterConfiguration(self.logs_and_rules_json)

    # Go through each of the paths
    # If we're already watching them then do nothing
    # If we're not watching them, watch them
    new_potential_paths = [obj for obj in self.logs_and_rules_json]
    for potential_path in new_potential_paths:
      if potential_path not in self.potential_paths:
        self.potential_paths.append(potential_path)
      if potential_path not in self.paths_to_watch:
        file_info = self._watch_path(potential_path)
        if file_info is None:
          LOG.debug("Could not find file: " + potential_path)

    # Go through the list of files we were watching and if any of them aren't valid anymore
    # then remove them
    for potential_path in self.potential_paths:
      if potential_path not in new_potential_paths:
        self.potential_paths.remove(potential_path)
        self._stop_watching_path(potential_path)

    if self.inotify is None:
      self._init_files_inotify()
    else:
      self._update_file_watches()

    # Also monitor the directories so that we can be notified on file renames
    # and creations. We need to monitor these events so that we can handle
    # log rotation
    self.directories_to_watch = []
    for potential_path in self.potential_paths:
      log_dir, log_file = os.path.split(potential_path)
      if os.path.exists(log_dir):
        self.directories_to_watch.append(log_dir)
      else:
        LOG.debug("Could not find directory: " + log_dir)

    if self.directory_inotify is None:
      self._init_directories_inotify()
    else:
      self._update_directory_watches()

  def process_events(self):
    self._process_inotify_events(self.inotify)
    self._process_inotify_events(self.directory_inotify)

  def shutdown(self):
    for file_info in self.files.values():
      file_info.close()
    self.event_proxy.shutdown()
    if self.inotify is not None:
      self._stop_inotify(self.inotify, False)
    if self.directory_inotify is not None:
      self._stop_inotify(self.directory_inotify, True)

  def _process_inotify_events(self, inotify):
    if inotify is not None:
      if (inotify.check_events()):
        inotify.read_events()
        inotify.process_events()

  # Remove the file path from the paths to watch and the files dict.
  # We can stop watching files in a couple cases:
  # 1. The file was both created and deleted before we re-checked events
  # 2. The file was removed from the log_tailer_config, but never
  #    existed in the first place, so we never could open it
  def _stop_watching_path(self, pathname):
   file_info = self.files.get(pathname)
   if file_info is not None:
     file_info.close()
     del self.files[pathname]
   if pathname in self.paths_to_watch:
     self.paths_to_watch.remove(pathname)

  def process_file_modifications(self, event):
    file_info = self.files[event.pathname]
    # File info can be none if we're processing a file that was just removed
    # from the configuration
    if file_info is not None:
      self._read_to_end_of_file(file_info)

  def _read_to_end_of_file(self, file_info):
    for line in file_info.fd.readlines():
      if '\n' in line:
        # If we already had a partial line then add it to the beginning
        if file_info.line_buffer is not None:
          line = file_info.line_buffer + line
        file_info.line_buffer = None

        # Call the filter config to see if the line matches a regex
        match_results = self.filter_config.match(file_info.path, line)
        if match_results is not None:
          self._send_to_proxy(match_results, file_info.path, line[:-1]) # Trim \n
      else:
        # This is a partial line write
        if file_info.line_buffer is None:
          file_info.line_buffer = line
        else:
          file_info.line_buffer = file_info.line_buffer + line

  def _send_to_proxy(self, match_results, path, line):
    match_results['CATEGORY'] = 'LOG_MESSAGE'
    match_results['EVENTCODE'] = 'EV_LOG_EVENT'
    match_results['HOSTS'] = self.host_name
    match_results['HOST_IDS'] = self.host_id
    match_results['PATH'] = path
    succeeded = self.event_proxy.send_event(line, match_results)
    # Retry once if it failed. The retry will re-establish the connection
    # which we need to do because Avro sometimes loses connections
    if not succeeded:
      self.event_proxy.send_event(line, match_results)

  def process_IN_DELETE(self, event):
    pathname = os.path.join(event.path, event.name)
    if pathname not in self.potential_paths:
      return
    self._stop_watching_path(pathname)

  # This is called whenever we receive a MOVED_FROM event on a watched directory
  def process_IN_MOVED_FROM(self, event):
    pathname = os.path.join(event.path, event.name)
    if pathname not in self.potential_paths:
      return
    file_info = self.files[pathname]
    # Process any last writes for the moved file. Note that processes holding handles
    # to the renamed file can still write to it until they close their handle. Right
    # now I'm not dealing with this because I don't know an easy way to handle it
    # completely correctly and reading to the end of the file should handle almost
    # all the cases. One way to handle this would be to watch for the close event,
    # but this might not occur (if the file was already closed) and there could
    # be multiple processes holding handles.
    self._read_to_end_of_file(file_info)
    self._stop_watching_path(pathname)
    self._update_file_watches()

  # This is called whenever we receive a CREATE event on a watched directory
  def process_IN_CREATE(self, event):
    pathname = os.path.join(event.path, event.name)
    if pathname not in self.potential_paths:
      return
    file_info = self._watch_path(pathname, False)
    if file_info is not None:
      # We didn't have a watch on this file between the time it was created
      # and now so read to the end of the file
      self._read_to_end_of_file(self.files[pathname])
      self._update_file_watches()

  # This is called whenver we receieve a MOVED_TO event on a watched directory
  def process_IN_MOVED_TO(self, event):
    pathname = os.path.join(event.path, event.name)
    if pathname not in self.potential_paths:
      return
    self._watch_path(pathname)
    self._update_file_watches()

  # We need to create a new watch manager here because there's no way
  # I'm aware of to change the paths that inotify tracks. As a result we have to
  # keep the old reference around until we've started watching the new one because
  # otherwise we could miss events. Since we're using the same file descriptor we
  # don't have to worry about double counting.
  def _update_file_watches(self):
     old_inotify = self.inotify
     self.inotify = None
     self._init_files_inotify()
     self._process_inotify_events(old_inotify)
     self._stop_inotify(old_inotify, False)

  # See the comment above to _update_file_watches
  def _update_directory_watches(self):
     old_inotify = self.directory_inotify
     self.directory_inotify = None
     self._init_directories_inotify()
     self._process_inotify_events(old_inotify)
     self._stop_inotify(old_inotify, True)

  # Create a FileInfo object, taking into account the possibility
  # that the file doesn't exist
  def _watch_path(self, pathname, read_to_end=True):
    try:
      file_info = FileInfo(pathname, read_to_end)
      self.paths_to_watch.append(pathname)
      self.files[pathname] = file_info
      LOG.info("Monitoring file: " + pathname)
      return file_info
    except:
      # The file doesn't exist
      return None

  def _stop_inotify(self, inotify, directory):
    if inotify is None:
      return
    try:
      inotify.stop()
    except Exception, e:
      log_message = "Error stopping "
      if directory:
        log_message += "directory"
      else:
        log_message += "file"
      log_message += " inotify " + str(e)
      LOG.warn(log_message)

  def _init_files_inotify(self):
    if len(self.paths_to_watch) > 0:
      try:
        wm = pyinotify.WatchManager()
        self.inotify = pyinotify.Notifier(wm,
            default_proc_fun=self.process_file_modifications,
            timeout=DEFAULT_PYINOTIFY_TIMEOUT_IN_MILLIS)
        wm.add_watch(self.paths_to_watch, pyinotify.IN_MODIFY);
      except:
        LOG.exception("Error initializing files inotify.")

  def _init_directories_inotify(self):
    if len(self.directories_to_watch) > 0:
      try:
        wm = pyinotify.WatchManager()
        self.directory_inotify = pyinotify.Notifier(wm, self,
            timeout=DEFAULT_PYINOTIFY_TIMEOUT_IN_MILLIS)
        # This watch notifies the class whenver a file gets created in, deleted
        # from, or moved to/from the specified directories. When process_events is
        # called this triggers calls to process_IN_CREATE, process_IN_DELETE, etc...
        wm.add_watch(self.directories_to_watch, pyinotify.IN_CREATE |
            pyinotify.IN_MOVED_FROM | pyinotify.IN_MOVED_TO | pyinotify.IN_DELETE)
      except:
        LOG.exception("Error initializing directories inotify.")

# This class holds the file description and line_buffer (for partial line writes)
# for a pathname
class FileInfo(object):

  def __init__(self, path, seek_to_end = True):
    self.path = path
    self.fd = open(path, 'r')
    if seek_to_end:
      self.fd.seek(0, 2) # 2==os.SEEK_END, which isn't available on py2.4
    self.line_buffer = None

  def close(self):
    self.fd.close()

# This class sends events to the Event Server
class EventServerProxy(object):

  def __init__(self):
    self.host = None
    self.port = None
    self.client = None
    self.requestor = None

  def send_event(self, line, attributes):
    """Sends events to the event server. Returns True on success."""
    event = dict()
    event['content'] = line
    event['timestamp'] = long(time()) * 1000
    event_attributes = {}
    for key in attributes:
      event_attributes[key] = [attributes[key]]
    event['attributes'] = event_attributes

    request = dict()
    request['event'] = event

    try:
      if self.host is not None and self.port is not None:
        if self.requestor is None:
          LOG.info("Connecting to: " + self.host + ":" + str(self.port))
          self.client = ipc.HTTPTransceiver(self.host, self.port)
          self.requestor = ipc.Requestor(EVENTS_PROTOCOL, self.client)

        result = self.requestor.request('reportEvent', request)
        return True
      else:
        return False
    except:
      LOG.exception("Error sending event to the event server: " + line + ".")
      self.requestor = None
      try:
        self.client.close()
      except:
        pass # We can ignore errors in this case
      return False

  def set_event_server_config(self, host, port):
    if self.host is None:
      self.host = host
    if self.port is None:
      self.port = port

  def shutdown(self):
    if self.client is not None:
      self.client.close()

# This is some dummy test for testing that we can remove once we integrate this
# with the agent. The agent will ultimately do something very similar to this.
def main():

  jsonString =  "{ \"logsandrules\" :"
  jsonString +=  "[{ "
  jsonString +=    "\"filepath\" : \"/tmp/test.txt\", "
  jsonString +=    "\"rules\" : [ {\"content\" : \"test4*\"}, {\"content\" : \"ATest\"}]"
  jsonString +=   "}, {"
  jsonString +=    "\"filepath\" : \"/tmp/test2.txt\", "
  jsonString +=   "\"rules\" : [ {\"content\" : \"test5\"} ]"
  jsonString +=   "}]"
  jsonString +=  "}"

  print jsonString
  tailer = LogTailer()

  config = {}
  config["eventserver_host"] = ""
  config["eventserver_port"] = 0
  config["logs_and_rules_config"] = jsonString
  tailer.set_config(config)

  while True:
    tailer.process_events()
    sleep(1)

if __name__ == "__main__":
  main()
