from __future__ import absolute_import

import logging
import logging.handlers

import os

import cmf.version

from cmf.log_buffer import FixedBufferHandler
from cmf.monitor.constants import DATE_FORMAT

LOG=logging.getLogger(__name__)
MAX_LOG_SIZE_BYTES = 10*1024*1024 # 10MB
LOG_BACKUP_COUNT = 5
# Constant value for date format.
DATE_FORMAT = '%d/%b/%Y %H:%M:%S %z'


# A knob to control the agent logging level. By default, it is set to 'None' to
# make the logging level to 'INFO'. Set it to not-empty to make the logging
# level to 'DEBUG'
scm_agent_debug = 'INFO'
active_log_files = set() # Log files that are known to exist


BUFFER_SIZE = 10*1024*1024 # This is the size in characters, not bytes


def setup_rolling_logging(filename):

  if filename == "-":
    return

  # Create the directory, if it doesn't already exist.
  filename_dir = os.path.dirname(filename)
  if not os.path.exists(filename_dir):
    try:
      os.makedirs(filename_dir)
      LOG.info("Created log directory: %s", filename_dir)
    except:
      LOG.exception("Failed to create log directory: %s", filename_dir)

  # Now that we have more information, we setup
  # rolling logging.
  root_logger = logging.getLogger()
  original_handler = None
  if len(root_logger.handlers) == 1:
    original_handler = root_logger.handlers[0]

  # Buffer to keep recent log messages
  buffer_handler = FixedBufferHandler(BUFFER_SIZE)
  formatter = logging.Formatter(LOG_FORMAT, DATE_FORMAT)
  # We always want to catch all messages in our error report buffer
  buffer_handler.setLevel(logging.DEBUG)
  buffer_handler.setFormatter(formatter)
  root_logger = logging.getLogger()
  root_logger.addHandler(buffer_handler)

  # Rotating file handler, unless filename is "-"
  if filename != "-":
    try:
      rh = logging.handlers.RotatingFileHandler(filename,
        maxBytes=MAX_LOG_SIZE_BYTES, backupCount=LOG_BACKUP_COUNT)
      rh.setFormatter(formatter)
      rh.setLevel(get_logging_level())
      logging.info("Logging to " + filename)
      root_logger.addHandler(rh)
      if original_handler:
        root_logger.removeHandler(original_handler)
    except:
      logging.exception("Could not open log file; writing to stderr.")


LOG_FORMAT = '[%(asctime)s] %(process)d %(threadName)s %(module)-12s %(levelname)-8s %(message)s'


# The control comes from two sources. The first is from the agent global
# variable scm_agent_debug which is by default 'INFO' and can be changed by the
# 'scm_debug' switch in /etc/cloudera-scm-agent/config.ini. The second source is
# coming from the the agent process environment variable 'SCM_DEBUG'.
# Note:
#   Currently the process environment variable 'SCM_DEBUG' is not hooked up to
#   control the agent logging knob. By default, it is 'None'.

def get_logging_level():
  if os.environ.get('SCM_DEBUG') is not None \
    or scm_agent_debug == 'DEBUG':
    return logging.DEBUG
  else:
    return logging.INFO


def setup_initial_logging(name):
  logging.basicConfig(level=get_logging_level(),
                      format=cmf.logging.LOG_FORMAT, datefmt=DATE_FORMAT)
  LOG.info("SCM Agent (%s) Version: %s", name, cmf.version.VERSION)


def get_log_buffer():
  l = logging.getLogger()
  for h in l.handlers:
    if isinstance(h, cmf.log_buffer.FixedBufferHandler):
      return h.buf
  return None


def get_log_rotating_file_handler():
  l = logging.getLogger()
  for h in l.handlers:
    if isinstance(h, logging.handlers.RotatingFileHandler):
      return h
  return None


