#!/usr/bin/env python
"""
Copyright (c) 2016 Cloudera, Inc. All rights reserved.

CLI entry point for various CM Agent functions
"""
from __future__ import absolute_import
# pylint: disable=import-error, missing-docstring, too-many-arguments
# pylint: disable=bare-except,broad-except,invalid-name, fixme

import argparse
import sys
import time
import click
import os
import resource
import signal
import psutil

import logging

from cmf.config import DEFAULT_CONF_DIR, AgentConfig
from cmf.supervisor import SupervisorWrapper
from cmf.util import register_sigquit_stack_dump

import cmf.logging
import cmf.util
import cmf.util.os_ops
import cmf.util.agent_cert
import cmf.util.remote_console
from cmf.monitor.constants import DATE_FORMAT
import cmf.upgrade
import cmf.redactor

LOG = logging.getLogger(__name__)

def set_process_limits():
  if os.getuid() == 0:
    for limit in [(resource.RLIMIT_NOFILE, "file descriptor", 32*1024),
                  (resource.RLIMIT_NPROC, "process", 64*1024),
                  (resource.RLIMIT_MEMLOCK, "memory lock", resource.RLIM_INFINITY)]:
      curr = resource.getrlimit(limit[0])
      if ((curr[0] != resource.RLIM_INFINITY and limit[2] != resource.RLIM_INFINITY
           and curr[0] > limit[2]) or
          (curr[0] == resource.RLIM_INFINITY)):
        LOG.info("Using existing %s limit: %s", limit[1], curr)
      else:
        try:
          resource.setrlimit(limit[0], (limit[2], limit[2]))
        except Exception as e:
          LOG.error("Failed setting %s limit: %s", limit[1], e)

@click.group()
@click.pass_context
@click.option("--config", default=DEFAULT_CONF_DIR,
              help="Configuration to use. (If a directory, read and use all .ini files within it.")
@click.option('--override', help="Override config file setting (e.g. --override SECTION.KEY=VALUE)",
              multiple=True)
@click.option("--debug", help="Enter pdb", is_flag=True, default=False)
def root(ctx, config, override, debug):
  ctx.obj.config = config
  ctx.obj.override = override

  cfg = AgentConfig.make_config(config, override, AgentConfig.GENERAL)
  cmf.logging.scm_agent_debug = cfg.scm_debug
  logging.basicConfig(level=cmf.logging.get_logging_level(),
                      format=cmf.logging.LOG_FORMAT, datefmt=DATE_FORMAT)

  if cfg.logfile != '-':
    cfg.log_dir = os.path.dirname(cfg.logfile)

  ctx.obj.cfg = cfg
  if debug:
    import pdb
    pdb.set_trace()


@root.group(invoke_without_command=True)
@click.pass_context
@click.option('--daemon', help="Run as a daemon.", is_flag=True, default=False)
@click.option('--pidfile', help="File to write PID to.", default=None)
@click.option('--comm_name',
              help="Name of command to report in /proc/<pid>/comm", default="cmagent")
@click.option('--legacy-supervisor', help="Legacy supervisor launching mode (non-systemd)",
              is_flag=True, default=False)
def agent(ctx, daemon, pidfile, comm_name, legacy_supervisor):
  if ctx.invoked_subcommand:
    return

  # pylint: disable=unused-argument
  cmf.logging.setup_initial_logging("agent")
  LOG.info("Executing with arguments: %s", " ".join(sys.argv))

  def main_impl(args):
    # Need a function so that we can pass the reference by string name to profile.runctx()
    from cmf.agent import Agent
    try:

      register_sigquit_stack_dump()
      set_process_limits()
      ag = Agent(args.cfg, sys.argv[:])
      ag.eval_config()
      ag.configure_process(daemon, pidfile, comm_name)
      ag.configure_directories()
      ag.configure_service()
      ag.pre_startup()
      ag.start(legacy_supervisor)
    except Exception as e:
      LOG.exception("Top-level exception: %s", e)


  # An environment variable can enable profiling.
  if os.environ.get("CLOUDERA_SCM_AGENT_PROFILE") is not None:
    # pylint: disable=wrong-import-position, wrong-import-order
    try:
      # python >= 2.5
      import cProfile as profile
    except:
      import profile as profile
    profiling_filename = "/tmp/agent_profile." + str(time.time())
    # Logging is not set up yet, so we just print to stdout.
    print "Profiling to: " + profiling_filename
    print "Read (after killing process) with " \
          "'import pstats; pstats.Stats('%s').sort_stats('time').print_stats(60)'" \
          % (profiling_filename,)
    profile.runctx('main_impl(ctx.obj)', globals(), locals(),
                   '/tmp/agent_profile.' + str(time.time()))
  else:
    main_impl(ctx.obj)

@agent.command()
@click.pass_context
def restart(ctx):
  """
  Restart the running agent.
  """
  # pylint: disable=unused-argument
  pid = os.getpid()
  exe_path = os.path.normpath(sys.argv[0])
  for proc in psutil.process_iter():
    c = proc.cmdline()
    if (len(c) > 2 and os.path.normpath(c[1]) == exe_path and
        c[2] in ["agent", "status_server"] and proc.pid != pid):
      LOG.info("Restarting %s", proc.pid)
      os.kill(proc.pid, signal.SIGUSR1)

@agent.command()
@click.pass_context
@click.argument("uuid")
@click.option("--delay", type=int, default=0,
              help="Wait this many seconds before performing operation")
@click.option("--restart_agent", is_flag=True, default=False,
              help="Restart agent after reseting id.")
def reset_uuid(ctx, uuid, delay=0, restart_agent=False):
  """
  Reset the agent uuid file.  Requires agent restart afterwards to take effect.
  """
  if delay:
    time.sleep(delay)
  path = os.path.join(ctx.obj.cfg.lib_dir, "uuid")
  os_ops = cmf.util.os_ops.OSOps(True)
  with os_ops.open_write_file(path) as f:
    f.write(uuid)

  if restart_agent:
    restart.callback()

@root.command()
@click.pass_context
def dump_config(ctx):
  """
  Print the agent configuration.
  """
  obj = ctx.obj
  cfg = obj.cfg
  for (section, values) in cfg.value_map().iteritems():
    for (k, v) in values.iteritems():
      print "%s.%s = %s" % (section, k, v)

@click.pass_context
def pre_upgrade(ctx):
  """
  Prepare the system for an upgrade from CM5 to CM6
  """
  cmf.upgrade.pre_upgrade(ctx.obj.cfg)

@click.pass_context
@click.argument("pid", type=int)
@click.argument("command_args", nargs=-1)
def proc_monitor(ctx, pid, command_args):
  """
  Proxy process monitor; launched by supervisord to monitor existing CDH processes that ran before
  an upgrade.
  """
  cmf.upgrade.monitor_process(ctx.obj.cfg, pid, command_args)

@click.pass_context
@click.option("--one-off", is_flag=True, default=False, help="Run as a one-off process")
@click.argument("command_args", nargs=-1)
def launch(ctx, one_off, command_args):
  """
  Process to redact sensitive data out of stdin/stdout streams.
  """
  # pylint: disable=unused-argument
  cmf.redactor.launch(one_off, command_args)

@click.pass_context
@click.argument("pid", type=int)
def proc_watcher(ctx, pid):
  # pylint: disable=unused-argument
  cmf.redactor.watch(pid)

@click.pass_context
@click.option("--fds", type=(int, int),
              default=(None, None), help="File-descriptors to redact (on re-exec)")
def redactor(ctx, fds):
  # pylint: disable=unused-argument
  cmf.redactor.redactor(fds[0], fds[1])


@click.pass_context
@click.option("--term", is_flag=True, default=False, help="Send SIGTERM signal")
def restart_watcher(ctx, term):
  """
  Restart all redactor processes with a SIGUSR1 signal.
  (Optionally send a SIGTERM instead if --term is specified.)
  """
  from cmf.process import ReconstructedProcess
  ReconstructedProcess.signal_watchers(ctx.obj.cfg, term)

@root.command()
@click.pass_context
def status_server(ctx):
  """
  Launch the agent status server.
  """
  cmf.logging.setup_initial_logging("status_server")
  from cmf.status_server import start_external_server
  register_sigquit_stack_dump()
  if ctx.obj.cfg.log_dir not in ["-", None]:
    cmf.logging.setup_rolling_logging(os.path.join(ctx.obj.cfg.log_dir, "scm-status.log"))
  start_external_server(ctx.obj.cfg)

@root.command()
@click.pass_context
def supervisor(ctx):
  """
  Launch the supervisor (in the foreground). For use by systemd unit files.
  """
  cmf.logging.setup_initial_logging("supervisord")
  from cmf.agent import Agent
  set_process_limits()
  ag = Agent(ctx.obj.cfg)
  ag.eval_config()
  ag.configure_directories()
  SupervisorWrapper.execute(ag.cfg, ag.identifier)

@root.command()
@click.pass_context
def ps(ctx):
  """
  Show processes managed by the CM agent.
  """
  from cmf.process import ReconstructedProcess

  ReconstructedProcess.list_processes(ctx.obj.cfg)

@root.command()
@click.argument("certtar")
@click.pass_context
def install_certs(ctx, certtar):
  """
  Installs Auto-TLS certificates on this host and configures agent config to
  enable TLS if needed. This script takes a single argument, which is the tar
  file containing the key and certificate files generated by certmanager.
  """
  cmf.util.agent_cert.install_certs(ctx, certtar)
  restart.callback()
  LOG.info("Agent now running with updated TLS configuration")

@root.group()
def diag():
  """
  Commands related to diagnostic bundles
  """
  pass

@diag.command()
@click.pass_context
def dump_cert(ctx):
  """
  Dumps the agent cert for collection by diag bundles
  """
  cmf.util.agent_cert.dump_cert(ctx)


def main():
  # pylint: disable=unexpected-keyword-arg, no-value-for-parameter

  # Add these sub-commands only if we're not going to be generating usage information for users.
  # "proc_monitor" and "pre_upgrade" are special purpose commands only for packaging scripts.
  for arg in sys.argv[1:]:
    if arg[0] != "-":
      sub_command = True
      break
  else:
    sub_command = False
  if sub_command and "--help" not in sys.argv:
    root.command()(proc_monitor)
    root.command()(pre_upgrade)
    root.command()(redactor)
    root.command()(launch)
    root.command()(proc_watcher)
    root.command()(restart_watcher)

  root.add_command(cmf.util.remote_console.main, "console")
  root(obj=argparse.Namespace())

def agent_main():
  sys.argv = ["cm", "agent"] + sys.argv[1:]
  main()

if __name__ == "__main__":
  main()
