# Copyright (c) 2015 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import simplejson as json
import threading


class _Metric(object):

  def __init__(self, label, description, numerator, denominator=None):
    if label is None or description is None or numerator is None:
      raise Exception("name, description, and numerator cannot be None")
    self._label = label
    self._description = description
    self._numerator = numerator
    self._denominator = denominator

  @property
  def label(self):
    return self._label

  @property
  def description(self):
    return self._description

  @property
  def numerator(self):
    return self._numerator

  @property
  def denominator(self):
    return self._denominator

class Counter(_Metric):

  """
  A simple counter class.
  """
  def __init__(self,
               label,
               description,
               numerator,
               initial_value=0):
    super(Counter, self).__init__(label, description, numerator)
    self._count = initial_value
    self._lock = threading.Lock()

  def increment(self, delta=1):
    """
    increment the counter by delta.
    """
    try:
      self._lock.acquire()
      self._count += delta
    finally:
      self._lock.release()

  def decrement(self, delta=1):
    """
    decrement the counter by delta
    """
    self.increment(-delta)

  def set(self, value):
    """
    set the counter value to 'value'
    """
    try:
      self._lock.acquire()
      self._count = value
    finally:
      self._lock.release()

  def get_count(self):
    """
    returns the current value.
    """
    return self._count

class Gauge(_Metric):

  """
  A class representing a instantaneous sample.
  """

  def get_value(self):
    """
    returns the value of the gauge.
    """
    raise Exception("Subclasses must implement this")

class SimpleGauge(Gauge):
  """
  A simple Gauge that allows setting and reading its value.
  """
  def __init__(self,
               label,
               description,
               numerator,
               denominator=None,
               initial_value=0):
    super(SimpleGauge, self).__init__(label, description, numerator, denominator)
    self._value = initial_value

  def get_value(self):
    return self._value

  def set_value(self, value):
    self._value = value

class SamplingGauge(Gauge):
  """
  A sampling gauge that will call a callback each time a sample is required.
  """

  def __init__(self, label, description, numerator, callback, denominator=None):
    super(SamplingGauge, self).__init__(label,
                                        description,
                                        numerator,
                                        denominator)
    if callback is None:
      raise Exception("callback cannot be None")
    if not callable(callback):
      raise Exception("callback must be a callable")
    self._callback = callback

  def get_value(self):
     return self._callback()

class MetricRegistry(object):
  """
  A metric registry that serializes itself to json. The registry is thead safe.
  """
  def __init__(self):
    self._counters = dict()
    self._gauges = dict()
    self._lock = threading.Lock()

  def add_counter(self, key, counter):
    """
    adds a counter to the counters registry keyed on 'key'. If a counter with
    that key already exists an Exception is raised.
    """
    if key is None:
      raise Exception("key cannot be None")
    if not isinstance(counter, Counter):
      raise Exception("'counter' is not a Counter")

    try:
      self._lock.acquire()
      if key in self._counters:
        raise Exception("Counter with key '%s' already exists" % key)
      self._counters[key] = counter
    finally:
      self._lock.release()

  def get_or_add_counter(self, key, counter):
    """
    adds a counter to the counters registry keyed on 'key'. If a counter with
    that key already exists that counter is returned if not, the new counter
    is returned.
    """
    if key is None:
      raise Exception("key cannot be None")
    if not isinstance(counter, Counter):
      raise Exception("gauge must be a Guage")

    try:
      self._lock.acquire()
      if key in self._counters:
        return self._counters[key]
      self._counters[key] = counter
      return counter
    finally:
      self._lock.release()

  def get_counter(self, key):
    """
    returns a counter keyed on 'key', None if no such counter exists.
    """
    if key is None:
      raise Exception("key cannot be None")
    return self._counters[key]

  def add_gauge(self, key, gauge):
    """
    adds a gauge to the gauges registry keyed on 'key'. If a gauge with that key
    already exists an Exception is raised.
    """
    if key is None:
      raise Exception("key cannot be None")
    if not isinstance(gauge, Gauge):
      raise Exception("gauge must be a Guage")

    try:
      self._lock.acquire()
      if key in self._gauges:
        raise Exception("Gauge with key '%s' already exists" % key)
      self._gauges[key] = gauge
    finally:
      self._lock.release()

  def get_or_add_gauge(self, key, gauge):
    """
    adds a gauge to the gauges registry keyed on 'key'. If a gauge with that key
    already exists that gauge is returned if not the new gauge is returned.
    """
    if key is None:
      raise Exception("key cannot be None")
    if not isinstance(gauge, Gauge):
      raise Exception("gauge must be a Guage")

    try:
      self._lock.acquire()
      if key in self._gauges:
        return self._gauges[key]
      self._gauges[key] = gauge
      return gauge
    finally:
      self._lock.release()

  def get_gauge(self, key):
    """
    returns a gauge keyed on 'key', None if no such gauge exists.
    """
    if key is None:
      raise Exception("key cannot be None")
    return self._gauges[key]

  def dump_metrics(self, indent=None):
    """
    Returns all the metrics, json formatted, ordered by keys.
    """
    try:
      self._lock.acquire()
      counters_dict = self._dump_counters()
      gauges_dict = self._dump_gauges()
    finally:
      self._lock.release()

    return json.dumps(dict(counters=counters_dict,
                           gauges=gauges_dict),
                      indent=indent,
                      sort_keys=True)

  def reset(self):
    """
    Reset the metric registry removing any metrics registered.
    """
    try:
      self._lock.acquire()
      for metric_type in (self._counters, self._gauges):
        metric_type.clear()
      self._counters.clear()
    finally:
      self._lock.release()

  def _dump_counters(self):
    """
    Returns a formatted dictionary of all counters.
    """
    ret = {}
    for key, counter in self._counters.iteritems():
      ret[key] = {
        "label": counter.label,
        "description": counter.description,
        "numerator": counter.numerator,
        "denominator":counter.denominator,
        "count": counter.get_count()}
    return ret

  def _dump_gauges(self):
    """
    Returns a formatted dictionary of all gauges.
    """
    ret = {}
    for key, gauge in self._gauges.iteritems():
      ret[key] = {
        "label": gauge.label,
        "description": gauge.description,
        "numerator": gauge.numerator,
        "denominator":gauge.denominator,
        "value": gauge.get_value()}
    return ret

METRIC_REGISTRY = MetricRegistry()
