# Copyright (c) 2011 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging
import threading
import time
import traceback

import kerberos as k
import os
import psutil
import simplejson

import cmf.kt_renewer
from cmf.monitor import conf, schema, constants
from cmf.monitor.base import MonitorInterface
from cmf.monitor.dir_monitor import DirMonitor
from cmf.monitor.json_metrics_utils import prefer_first_attribute
from cmf.monitor.monitor_properties import ConfParser
from cmf.monitor.wakeable_thread import WakeableThread
from cmf.safety_valve_helpers import get_urlopen_retries_on_auth_errors
from cmf.safety_valve_helpers import get_urlopen_sleep_seconds_on_auth_errors
from cmf.throttling_logger import ThrottlingLogger
from cmf.util.url import urlopen_with_retry_on_authentication_errors
from cmf.util.url import urlopen_with_timeout

LOG = logging.getLogger(__name__)
THROTTLED_LOG = ThrottlingLogger(LOG, constants.DEFAULT_MONITOR_LOG_RATE)

_COLLECT_INTERVAL = 'collect_interval'

class CollectionError(Exception):
  def __init__(self, code):
    self.code = code;

class AbstractMonitor(MonitorInterface):
  """
  A common implementation of the MonitorInterface. Expectation on a derived
  class:
    * It uses a separate thread to do metrics collection, and implements
      the `collect()' method to perform the collection work.
    * It checks for configuration update (captured in `conf_parser') before
      each collection.
    * The first config section should contain a `collect_interval'.
  """

  def __init__(self, daemon, registry, cgroup_root, cgroup_name, role_type):
    MonitorInterface.__init__(self)
    self._cgroup_root = cgroup_root
    self._cgroup_name = cgroup_name
    self._daemon = daemon
    self._agent = daemon.get_agent()
    self._registry = registry
    self._role_type = role_type
    self._collection_thread = None
    self._conf_parser = None
    self._stacks_conf_parser = None
    self._last_error = None
    self._last_reported_timestamp = None
    self._conf_dir = None
    self._kerberos_principal = None
    self._keytab_path = None
    self._krb5cc_name = None

    # If the _pid is None then it means that the associated process is not
    # running, and that metric collection dependent on that process should
    # be skipped.
    self._pid = None

    # We throttle our missing pid log messages. See OPSAPS-5138.
    self._last_missing_pid = None

    # A counter tracking the number of unexpected exits encountered by a process.
    # Access to this field is serialized with self._lock.
    self._unexpected_exits = 0

    # A counter tracking the number of exits due to an OutOfMemory error.
    # Access to this field is serialized with self._lock.
    self._oom_exits = 0

    # Whether we ought to be refreshing our stacks collector or not. Access to
    # this field is serialized with self._lock.
    self._refresh_stacks = True

    # We use a lock to serialize access to the certain fields. Fields that need
    # serialized access will say so in their comments.
    self._lock = threading.RLock()

    # To make it possible to test our urlopen retry logic.
    self._urlopen_with_timeout = urlopen_with_timeout

  def collect(self):
    """
    The loop content in the collection thread.
    This method is responsible for loading any new update in `conf_parser'.
    """
    # Should be overridden by derived classes
    raise NotImplementedError("%s needs to override collect()" %
                              (self.__class__,))

  @property
  def conf_parser(self):
    return self._conf_parser

  @property
  def last_error(self):
    return self._last_error

  @property
  def daemon(self):
    return self._daemon

  @property
  def conf_dir(self):
    return self._conf_dir

  @property
  def kerberos_principal(self):
    # returns the full principal name (including hostname)
    return self._kerberos_principal

  @property
  def keytab_path(self):
    return self._keytab_path

  @property
  def krb5cc_name(self):
    return self._krb5cc_name

  def set_kerberos_principal_to_use(self, principal_prefix, krb5cc_name):
    """
    This method is intended to be called from derived classes that need to
    override the default principal in use when talked to a secured end-point.
    By default the HTTP/<hostname> principal is used.
    The function will construct the full principal name and set gssapi lib to
    use the credential cache named 'krb5cc_name'. It will then use the
    kt_renewer to kinit (and re-kinit periodically) from the keytab storing
    the credentials in the supplied credential cache. When the monitor is
    stopped the principal record from the kt_renewer will be removed and
    the credential cache deleted (see stop() below)
    @param principal_prefix: the principal name prefix, e.g., hdfs, solr. If
                             'None' the function is a no-op.
    @param krb5cc_name: the krb5 credential cache name to use. The credential
                        cache will be created in the configuration directory.
    @return: 'True' if successfully set the principal to use, 'False'
             otherwise.
    """
    if principal_prefix is None:
      return True
    keytab = cmf.kt_renewer.find_keytab_for_principal(
        self.conf_dir, principal_prefix)
    if keytab is None:
      THROTTLED_LOG.warn("Could not find keytab for principal %s in %s" %
                         (principal_prefix, self.conf_dir))
      return False
    credential_cache_name = "%s/%s" % (self.conf_dir, krb5cc_name)
    result, old_name = k.setGSSCredentialCacheName(credential_cache_name)
    # The only possible success value is '1' aka AUTH_GSS_COMPLETE.
    if result < 1:
      THROTTLED_LOG.warn("Could not set credential cache name: %d" % result)
      return False
    full_principal_name = cmf.kt_renewer.get_full_principal_name(
        principal_prefix, self._agent.cfg.reported_hostname)
    cmf.kt_renewer.add_specific_principal_to_kinit(
        keytab, full_principal_name, credential_cache_name)

    self._kerberos_principal = full_principal_name
    self._keytab_path = keytab
    self._krb5cc_name = credential_cache_name
    return True

  def record_error(self, exc):
    self._last_error = "%s\n%s" % (exc, traceback.format_exc())

  def refresh(self, conf_dir):
    """
    Refresh the monitor context and ensure the monitor is collecting. The monitor
    configuration is loaded into `conf_parser' after each refresh.
    """
    LOG.info("Refreshing %s" % (self,))
    try:
      self._conf_parser = ConfParser(conf_dir,
                                     conf.CLOUDERA_MONITOR_PROPERTIES)
      self._conf_dir = conf_dir
      if os.path.exists(os.path.join(conf_dir,
                                     conf.CLOUDERA_STACK_MONITOR_PROPERTIES)):
        self._stacks_conf_parser = ConfParser(
          conf_dir, conf.CLOUDERA_STACK_MONITOR_PROPERTIES)
      else:
        self._stacks_conf_parser = None
    except IOError, ex:
      self.record_error(ex)
      LOG.error("%s: %s" % (self, ex))
      return

    self.ensure_collecting()

  def refresh_pid(self, pid, skip_pid_checks=False):
    # If we have not been given a pid to monitor, or we've been instructed to
    # skip our pid checks (for testing) we're done.
    if pid is None or skip_pid_checks:
      self._set_pid(pid)
      return

    # If we have been given the same pid we're already monitoring, then we have
    # nothing to do.
    if pid == self._pid:
      return

    # First we check if the pid we've been given is for a process that looks
    # like what we're supposed to be monitoring. It's possible that it is not,
    # for example it might be an su process that is the parent of the actual
    # process we're interested in monitoring. See OPSAPS-4451.
    process = psutil.Process(pid)
    if self.is_process_to_monitor(process):
      self._set_pid(pid)
      return

    # If the pid we were given was not the process we're looking for, we'll
    # look through each of it's children looking for an appropriate process.
    for child in process.children():
      if self.is_process_to_monitor(child):
        self._set_pid(child.pid)
        return

    # If we found no match, then we turn off the monitor.
    self._pid = None
    if pid != self._last_missing_pid:
      LOG.error("%s unable to find appropriate process to monitor for pid %d." %
                (self, pid))
      self._last_missing_pid = pid

  def record_unexpected_exit(self, oom_exit=False):
    try:
      self._lock.acquire()
      self._unexpected_exits += 1
      if oom_exit:
        self._oom_exits += 1
    finally:
      self._lock.release()

  def add_unexpected_exit_metrics(self, update):
    """
    Adds the current unexpected exit count as a metric to the (role) update
    and then resets the count.
    """
    try:
      self._lock.acquire()
      update.add_metric(schema.UNEXPECTED_EXITS_COUNT, self._unexpected_exits)
      update.add_metric(schema.OOM_EXITS_COUNT, self._oom_exits)
    finally:
      self._lock.release()

  def _set_pid(self, pid):
    self._pid = pid
    self._last_missing_pid = None

  def is_process_to_monitor(self, process):
    if process is None:
      return False
    # We never want to monitor the priveleged 'jsvc.exec' process. We usually
    # want to monitor its child 'jsvc.exec' process which will not be running
    # as root.
    if process.name() in ('jsvc', 'jsvc.exec') and process.uids()[0] == 0:
      return False
    # As of this moment, the only processes we monitor are either 'java' or
    # 'jsvc.exec'. This check is sufficient to correctly identify our current
    # monitors, but may need to evolve in the future as we add more monitors.
    # With this check, we will not monitor 'su' processes, nor will we monitor
    # the wrong part of the 'jsvc.exec' process. Also, we will not end up
    # monitoring various 'tail' processes started by 'jsvc.exec'. We originally
    # tried to do more here -- matching particular arguments on the command
    # line -- but ran into issues when that command line regularly exceeded
    # the 4K characters exposed by /proc.
    return process.name() in ('java', 'jsvc', 'jsvc.exec')

  def ensure_collecting(self):
    if self._collection_thread is not None:
      return

    self._collection_thread = self.make_collection_thread()
    self._collection_thread.start()
    self._registry.register(self)
    self._daemon.new_monitor(self)

  def _refresh_stacks_collector(self):
    if (self._daemon is None or
        self._daemon.get_stacks_collection_manager() is None):
      return
    try:
      self._lock.acquire()
      if not self._refresh_stacks:
        return
      self._daemon.get_stacks_collection_manager().refresh_collector(
        self._stacks_conf_parser, self._pid, self._urlopen_from_worker_thread)
    finally:
      self._lock.release()

  def _stop_stacks_collector(self):
    if (self._daemon is None or
        self._daemon.get_stacks_collection_manager() is None):
      return
    try:
      self._lock.acquire()
      self._refresh_stacks = False
      self._daemon.get_stacks_collection_manager().stop_collector(
        self._stacks_conf_parser)
    finally:
      self._lock.release()

  def make_collection_thread(self):
    return WakeableThread(
        "Monitor-%s" % (self.__class__.__name__,),
        self.collect_wrapper)

  def collect_wrapper(self):
    self._refresh_stacks_collector()
    self.collect()

  def _get_collect_interval(self):
    val = self._conf_parser.getfloat(self._conf_parser.get_default_section(),
                                     _COLLECT_INTERVAL)
    if val is None:
      LOG.warn("%s is missing the '%s' config" %
               (self._conf_parser.path, _COLLECT_INTERVAL))
      return self._daemon.get_reporting_interval_seconds()
    return val

  def stop(self):
    """Stop the monitor"""
    self._registry.unregister(self)
    if self._collection_thread is not None:
      self._collection_thread.should_stop()
      self._collection_thread = None
    if self.kerberos_principal is not None and self.keytab_path is not None:
      cmf.kt_renewer.remove_specific_principal_to_kinit(
          self.kerberos_principal, self.keytab_path)
    self._stop_stacks_collector()

  def get_attempt_updates(self):
    """Returns a list of avro'd AttemptUpdate objects to be sent to the firehose"""
    return []

  def get_role_update(self):
    """Returns a RoleUpdate to be sent to the firehose"""
    return None

  def get_service_update(self):
    """Returns a ServiceUpdate to be sent to the firehose"""
    return None

  def start_collecting(self, latch):
    if self._collection_thread is not None:
      self._collection_thread.wake(latch)

  def urlopen(self, url, username=None, password=None):
    # We always pass the secure http service name because it has no effect on
    # unsecured connections.
    return urlopen_with_retry_on_authentication_errors(
      lambda: self._urlopen_with_timeout(
          url,
          timeout=self.get_metrics_url_timeout_seconds(),
          secure_http_service_name=self._get_secure_http_service_name(),
          username=username,
          password=password),
      retries = get_urlopen_retries_on_auth_errors(
        self._agent.safety_valve),
      sleeptime = get_urlopen_sleep_seconds_on_auth_errors(
        self._agent.safety_valve))

  def _urlopen_from_worker_thread(self, url, username=None, password=None):
    """
    Utility function used to perform a urlopen from a worker thread. This is
    needed to ensure that the worker thread has the appropriate thread local
    variable set if a particular kerberos principle is needed by this monitor.
    """
    if self._krb5cc_name is not None:
      result, old_name = k.setGSSCredentialCacheName(self._krb5cc_name)
      # The only possible success value is '1' aka AUTH_GSS_COMPLETE.
      if result < 1:
        THROTTLED_LOG.warn("Could not set credential cache name: %d" % result)
    return self.urlopen(url, username, password)

  def _get_secure_http_service_name(self):
    """
    Constructs a service name for a secure HTTP service on this host. This is based
    on the format of the urllib2_kerberos library which, by default, builds the
    service name with the host from the HTTP request. The service name is used
    by GSS to locate the secure service. While we know the service to be Kerberos-
    based, the GSS API is more generic and doesn't call these Kerberos principals,
    so we refer to them by the GSS term 'service name'.
    """
    return "HTTP@%s" % (self._agent.cfg.listening_hostname,)

  def collect_metrics_from_url(self,
                               url,
                               username=None,
                               password=None,
                               use_prefer_first_attribute_hook=False,
                               suppress_error_logging=False):
    result = None
    start = time.time()
    try:
      code = constants.WEB_METRICS_COLLECTION_COMMUNICATION_FAILURE
      openedUrl = self.urlopen(url, username=username, password=password)
      code = constants.WEB_METRICS_COLLECTION_PARSE_FAILURE
      if use_prefer_first_attribute_hook:
        # We use the 'prefer_first_attribute' JSON hook to ensure that the
        # first instance of a key is chosen.
        # This is to accommodate the JSON provided by Solr servers that
        # have duplicate entries for the 'name' key. See CDH-15751
        result = self.parse_metrics(simplejson.load(
            openedUrl, object_pairs_hook=prefer_first_attribute))
      else:
        result = self.parse_metrics(simplejson.load(openedUrl))
    except Exception, ex:
      self.record_error(ex)
      msg = "Error fetching metrics at '%s' for %s" % (url, self._role_type)
      if suppress_error_logging:
        THROTTLED_LOG.debug(msg, exc_info=True)
      else:
        THROTTLED_LOG.exception(msg)
      result = CollectionError(code)
    now = time.time()
    duration = int((now - start) * 1000.0)
    return result, now, duration

  def get_task_metrics_timeout_seconds(self):
    return self._daemon.get_task_metrics_timeout_seconds()

  def get_metrics_url_timeout_seconds(self):
    return self._daemon.get_metrics_url_timeout_seconds()

  # For testing.
  def clear_last_reported_timestamp(self):
    self._last_reported_timestamp = None

  # This method should be overridden by monitors to control when they are ready
  # to report metrics. This basic implementation says that the monitor is ready
  # as soon as the process pid is set. Examples of when this is not appropriate
  # are the host monitor, which has no PID, and something like a DataNode which
  # does have a pid, but also has a web UI from which it collects metrics that
  # it wants to ensure is up before claiming to be ready.
  #
  # Note that this is called by the overall monitoring daemon scheduling thread
  # so this needs to return quickly (calls should take no more than a second)
  # and it needs to be mindful of what members are accessed.
  def is_ready_to_report(self):
    return self._pid is not None

  # Returns true if the input URL responded with some JSON string within the
  # input timeout. False otherwise
  def is_metrics_url_responding(self, url, timeout):
    try:
      simplejson.load(urlopen_with_timeout(url, timeout=timeout))
      return True
    except Exception, ex:
      pass
    return False

  def create_dir_monitor(self, directory, monitors, dir_type=None):
    """
    Method creates a directory monitor for the directory specified
    by path 'dir'. If set, the 'dir_type' must correspond to the directory's
    template name if it is to be monitored as a role directory.
    """
    try:
      dir_monitor = DirMonitor(self._daemon.get_filesystem_map(),
                               directory, dir_type)
      if monitors is not None:
        monitors.append(dir_monitor)
      return dir_monitor
    except Exception:
      # This might happen if the directory for which we are trying to
      # instantiate a monitor for is mounted on a network share. The
      # DirMonitor uses psutil that only gives information for local file
      # systems. We should probably generate a health warning for this,
      # OPSAPS-4139 tracks this.
      LOG.exception("Could not create directory monitor for %s" %
                    (directory,))

  def add_metrics_from_webserver(self,
                                 metrics_tuple,
                                 last_reported_timestamp,
                                 update):
    metrics, timestamp, duration = metrics_tuple
    if metrics is None or timestamp == last_reported_timestamp:
      return False

    update.add_metric(schema.WEB_METRICS_COLLECTION_DURATION, duration)
    update.timestamp = timestamp
    if isinstance(metrics, CollectionError):
      update.add_metric(schema.WEB_METRICS_COLLECTION_STATUS, metrics.code)
    else:
      update.add_metric(schema.WEB_METRICS_COLLECTION_STATUS,
                        constants.WEB_METRICS_COLLECTION_GOOD)
      self.add_webserver_metrics_to_role_update(update, metrics)
    return True

  # This method should be overridden by monitors to add metrics from the input
  # JSON metrics structure, retrieved from the monitored role's webserver, to
  # the input role update
  def add_webserver_metrics_to_role_update(self, update, metrics):
    pass

  def _is_feature_enabled(self, role_update, feature):
    """
    Returns true if the feature in question is enabled for the role which
    the role_update is for.
    """
    safety_valve = self._agent.safety_valve
    role_key = "%s_%s_feature_%s_enabled" % \
               (role_update.service_name.lower(),
               role_update.role_name.lower(),
               feature)
    service_key = "%s_feature_%s_enabled" % \
                  (role_update.service_name.lower(),feature)
    return 'true'  == safety_valve.get(role_key, 'true').strip().lower() \
           and \
           'true'  == safety_valve.get(service_key, 'true').strip().lower()

