# Copyright (c) 2011 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import


class MonitorInterface(object):
  """
  The interface for a monitor. The monitor is responsible for collecting metrics
  and putting them into a report that will get sent to the firehose.
  """
  def refresh(self, conf_dir):
    """Refresh the monitor context and start the monitor"""
    raise NotImplementedError("%s needs to override refresh()" %
                              (self.__class__,))

  def refresh_pid(self, pid):
    """If the monitor is associated with an actual suprvised process, this will
       be called periodically with the pid of the associated supervised process.
       If called with None then the process is not running (or is in transition).
       In this case, the monitor should disable collection of the appropriate
       metrics."""
    raise NotImplementedError("%s needs to override refresh_pid()" %
                              (self.__class__,))

  def record_unexpected_exit(self, is_oom_exit):
    """If the monitor is associated with an actual suprvised process, this will
       be called when the process exits unexpectedly."""
    raise NotImplementedError("%s needs to override " \
                                "record_unexpected_exit()" %
                              (self.__class__,))

  def start_collecting(self, latch):
    """Start the collection process."""
    raise NotImplementedError("%s needs to override start_collecting()" %
                              (self.__class__,))

  def stop(self):
    """Stop the monitor"""
    raise NotImplementedError("%s needs to override stop()" %
                              (self.__class__,))

  def get_attempt_updates(self):
    """Returns a list of avro'd AttemptUpdate objects to be sent to the firehose"""
    raise NotImplementedError("%s needs to override get_attempts_update()" %
                              (self.__class__,))

  def get_service_update(self):
    """Returns a ServiceUpdate to be sent to the firehose"""
    raise NotImplementedError("%s needs to override get_service_update()" %
                              (self.__class__,))

  def get_role_update(self):
    """Returns a RoleUpdate to be sent to the firehose"""
    raise NotImplementedError("%s needs to override get_role_update()" %
                              (self.__class__,))

  def get_debug_status_report(self):
    """Returns a tuple of (config_dict, error_list)"""
    raise NotImplementedError("%s needs to override get_debug_status_report()" %
                              (self.__class__,))

  def parse_metrics(self, json):
    """Returns a metrics parser object built from the input json"""
    raise NotImplementedError("%s needs to override parse_metrics()" %
                              (self.__class__,))

class DebugStatusReport(object):
  """
  A structure to hold debugging info for the agent debug port
  """
  def __init__(self,
               config_items=None,
               errors=None,
               last_collection_ended=None,
               last_collection_length=None,
               stats=None):
    self.config_items = config_items or [ ]
    self.errors = errors or [ ]
    self.last_collection_ended = last_collection_ended
    self.last_collection_length = last_collection_length
    self.stats = stats or { }
