# Copyright (c) 2013 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import os
import cmf.monitor.schema as schema

jiffies_per_second = None
if 'SC_CLK_TCK' in os.sysconf_names:
  jiffies_per_second = float(os.sysconf(os.sysconf_names['SC_CLK_TCK']))

def _read_stat_file(path_to_stat):
  """
  The convention for cgroup stat files is that the stat file for a subsystem foo
  is named foo.stat. The format for these files is to have one metric per line.
  For example, this is what cpuacct.stat looks like:
  user 6248742
  system 1674817

  The output from this function is a dictionary mapping the keys on the left to
  their corresponding values. All values are returned as strings.

  Returns None if path_to_stat doesn't exist, is a directory, or an IOError occurs.
  """
  if os.path.isdir(path_to_stat) or not os.path.exists(path_to_stat):
    return None

  try:
    stat_dict = {}
    for line in open(path_to_stat):
      line_parts = line.strip().split()
      if len(line_parts) >= 2:
        stat_dict[line_parts[0]] = line_parts[1]

    return stat_dict
  except IOError, e:
    return None

def _add_metric_if_possible(update, key, stat_dict, metric, scaling_factor=1.0):
  if not key in stat_dict:
    return

  try:
    update.add_metric(metric, float(stat_dict[key]) / scaling_factor)
  except ValueError:
    pass

def _read_blkio_stat_file(path_to_stat):
  """
  The blkio subsystem has a slightly different mechanism for reporting disk
  utilization. A typical file is of the form:

  252:0 Read 41160  # Number of Read I/Os on /dev/vda1
  252:0 Write 3     # Number of Write I/Os on /dev/vda1
  252:0 Sync 41163  # Number of Synchronous I/Os on /dev/vda1
  252:0 Async 0     # Number of Asynchronous I/Os on /dev/vda1
  252:0 Total 41163 # Number of total I/Os on /dev/vda1
  Total 41163

  Here, the first entry on the first 5 lines corresponds to a disk device in
  major:minor format, the second entry corresponds to the type of operation on
  that disk device, and the last entry corresponds to the value for that
  metric for that disk.

  Because we don't currently support per-disk metrics for cgroups, we iterate
  over the lines in the stat file and compute the total number of IOs/bytes/time
  per operation, and then return that in a dictionary.

  We skip the 'Sync' and 'Async' attributes because we don't know what we can do
  with those metrics that would be useful.

  Returns None if path_to_stat doesn't exist, is a directory, or an IOError occurs.
  """
  if os.path.isdir(path_to_stat) or not os.path.exists(path_to_stat):
    return None

  try:
    result = {}
    for line in open(path_to_stat):
      line_parts = line.strip().split()
      if len(line_parts) >= 3:
        disk_device, category, value = line_parts[:3]
        value = float(value)
        result[category] = result.get(category, 0.0) + value

    return result
  except IOError, e:
    return None

def _add_blkio_metrics(update, cgroup_root, cgroup_name):
  stat_dict = _read_blkio_stat_file(os.path.join(cgroup_root, 'blkio', cgroup_name, 'blkio.io_serviced'))
  if stat_dict is not None:
    _add_metric_if_possible(update, 'Read', stat_dict, schema.CGROUP_READ_IOS)
    _add_metric_if_possible(update, 'Write', stat_dict, schema.CGROUP_WRITE_IOS)

  stat_dict = _read_blkio_stat_file(os.path.join(cgroup_root, 'blkio', cgroup_name, 'blkio.io_service_bytes'))
  if stat_dict is not None:
    _add_metric_if_possible(update, 'Read', stat_dict, schema.CGROUP_READ_BYTES)
    _add_metric_if_possible(update, 'Write', stat_dict, schema.CGROUP_WRITE_BYTES)

def _add_mem_metrics(update, cgroup_root, cgroup_name):
  # Look for memory metrics. We care about RSS, page cache, and swap used.
  mem_stat = _read_stat_file(os.path.join(cgroup_root, "memory", cgroup_name, "memory.stat"))
  if mem_stat is not None:
    _add_metric_if_possible(update, 'total_cache', mem_stat, schema.CGROUP_MEM_PAGE_CACHE)
    _add_metric_if_possible(update, 'total_swap', mem_stat, schema.CGROUP_MEM_SWAP)
    # From the memory subsystem documentation: to get the actual resident set
    # size, we have to add the total_mapped_file and the total_rss values.
    rss_str = mem_stat.get('total_rss')
    mapped_file_str = mem_stat.get('total_mapped_file')
    if rss_str is not None and mapped_file_str is not None:
      try:
        update.add_metric(schema.CGROUP_MEM_RSS, float(rss_str) + float(mapped_file_str))
      except ValueError:
        pass

def _add_cpu_metrics(update, cgroup_root, cgroup_name):
  # Look for CPU metrics
  cpu_stat = _read_stat_file(os.path.join(cgroup_root, "cpuacct", cgroup_name, "cpuacct.stat"))
  if cpu_stat is not None and jiffies_per_second is not None:
    # The cpuacct.stat file reports user and system time in terms of jiffies.
    # The number of jiffies in 1 second is specified by the USER_HZ constant in
    # the kernel.
    _add_metric_if_possible(update, 'user', cpu_stat,
       schema.CGROUP_TOTAL_CPU_USER, jiffies_per_second)
    _add_metric_if_possible(update, 'system', cpu_stat,
       schema.CGROUP_TOTAL_CPU_SYSTEM, jiffies_per_second)

def add_cgroup_metrics(update, cgroup_root, cgroup_name):
  # This means that cgroups are not enabled on this host
  if cgroup_root is None or not os.path.exists(cgroup_root):
    return

  # We don't collect metrics for the root cgroup
  if cgroup_name is None:
    return

  _add_mem_metrics(update, cgroup_root, cgroup_name)
  _add_cpu_metrics(update, cgroup_root, cgroup_name)
  _add_blkio_metrics(update, cgroup_root, cgroup_name)
