# Copyright (c) 2011 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

"""
A helper to monitor directories.
"""

import logging
import os
import pwd
import grp
from cmf.monitor import schema
from cmf.monitor.host_dir_update import HostDirUpdate
from cmf.monitor.role_dir_update import RoleDirUpdate
from cmf.throttling_logger import ThrottlingLogger

LOG = logging.getLogger(__name__)
THROTTLED_LOG = ThrottlingLogger(LOG)

class DirMonitor(object):
  def __init__(self, fs_map, path, dir_type=None):
    self._fs_map = fs_map
    self._path = path
    self._fs = None
    self._usage_info = None
    self._dir_type = dir_type
    self._permissions = None
    self._owner = None
    self._group = None

  def collect(self):
    # We re-resolve the filesystem on every collection in case a symlink has
    # been introduced. See OPSAPS-19355.
    self._fs = self._fs_map.get_filesystem_by_path(self._path)
    if self._fs is None:
      THROTTLED_LOG.error("Could not find local file system for %s" %
                          (self._path,))
      self._usage_info = None
      self._permissions = None
      self._owner = None
      self._group = None
      return
    self._usage_info = self._fs_map.disk_usage(self._fs.mountpoint)
    # On non-nodev filesystems, a call to os.stat() can hang, so avoid it. See OPSAPS-27354.
    if not self._fs.nodev:
      self._permissions, self._owner, self._group = self._get_directory_attributes()

  def get_role_dir_update(self):
    if self._fs is None:
      return None
    update = RoleDirUpdate(self._path,
                           self._fs.mountpoint,
                           self._fs.options,
                           self._fs.partition,
                           self._dir_type,
                           self._permissions,
                           self._owner,
                           self._group,
                           self._fs.fstype)
    self._add_metrics(update)
    return update

  def get_host_dir_update(self):
    if self._fs is None:
      return None
    update = HostDirUpdate(self._path,
                           self._fs.mountpoint,
                           self._fs.options,
                           self._fs.partition,
                           self._permissions,
                           self._owner,
                           self._group,
                           self._fs.fstype)
    self._add_metrics(update)
    return update

  def _add_metrics(self, update):
    if self._usage_info is not None:
      update.add_metric(schema.CAPACITY, self._usage_info['total'])
      update.add_metric(schema.CAPACITY_USED, self._usage_info['used'])
      update.add_metric(schema.CAPACITY_FREE, self._usage_info['free'])


  def _get_directory_attributes(self):
    """Returns a tuple of (permissions, owner, group) for the directory's path"""
    name, group, mode = None, None, None
    try:
      attrs = os.stat(self._path)
      mode = oct(attrs.st_mode)[-4:]
      uid = attrs.st_uid
      name = pwd.getpwuid(uid)[0]
      gid = attrs.st_gid
      group = grp.getgrgid(gid)[0]
    except:
      THROTTLED_LOG.exception("Error getting directory attributes for %s" % self._path)

    return mode, name, group

