# (c) Copyright 2011 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import avro.ipc
import avro.protocol
import os
import logging
import threading
import time

from httplib import BadStatusLine

from cmf.metric_registry import METRIC_REGISTRY, SimpleGauge, Counter
from cmf.monitor.base import DebugStatusReport
from cmf.throttling_logger import ThrottlingLogger
from cmf.util.avro import generate_avro_sanitizer

LOG = logging.getLogger(__name__)

_CONN_FAILED_LOG_DELAY = 10 * 60
THROTTLED_LOG = ThrottlingLogger(LOG, _CONN_FAILED_LOG_DELAY)

FIREHOSE_MESSAGE_PROTOCOL = avro.protocol.parse(
  file(os.path.join(os.path.dirname(__file__), "cmon_agent.avpr")).read())

# The Avro schema specific unicode sanitizer
UNICODE_SANITIZER = generate_avro_sanitizer(FIREHOSE_MESSAGE_PROTOCOL)

class Firehose(object):
  """This class is used for communication with a firehose. It is thread safe."""

  def __init__(self, name, roletype, address, port):
    self.name = name
    self.roletype = roletype
    self._address = address
    self._port = port
    self._requestor = None
    self._transceiver = None
    self._shutdown = False
    self._lock = threading.Lock()
    # Metrics
    #########
    self._last_message_transmit_duration_gauge = \
      METRIC_REGISTRY.get_or_add_gauge(
        self._get_name("last_message_transmit_duration"),
        SimpleGauge("Last Message Transmit Duration",
                    "The time (wall-clock) it took to transmit the last firehose "
                    "message",
                    "ms"))
    self._message_transmit_succeeded_counter = \
      METRIC_REGISTRY.get_or_add_counter(
        self._get_name("message_transmit_succeeded"),
        Counter("Messages Transmit Successes",
                "Number of messages successfully sent to the firehose",
                "messages"))
    self._message_transmit_failed_counter = \
      METRIC_REGISTRY.get_or_add_counter(
        self._get_name("message_transmit_failed"),
        Counter("Messages Transmit Failures",
                "Number of messages that the agent failed to send to the firehose",
                "messages"))
    self._requestor_bytes_sent = \
      METRIC_REGISTRY.get_or_add_counter(
        self._get_name("requestor_bytes_sent"),
        Counter("Bytes Sent",
                "Number of bytes sent in messages to the firehose",
                "bytes"))

  def get_last_message_transmit_duration(self):
    return self._last_message_transmit_duration_gauge.get_value()

  def get_message_transmit_succeeded(self):
    return self._message_transmit_succeeded_counter.get_count()

  def get_message_transmit_failed(self):
    return self._message_transmit_failed_counter.get_count()

  def get_message_bytes_sent(self):
    return self._requestor_bytes_sent.get_count()

  def _get_name(self, metric_name):
    return "%s::%s" % (self.name, metric_name)

  def set_connection_info(self, address, port):
   try:
      self._lock.acquire()
      self._set_connection_info(address, port)
   finally:
     self._lock.release()

  def _set_connection_info(self, address, port):    
    if self._address == address and self._port == port:
      return
    if self._shutdown:
      LOG.info("Ignoring connection update on shutdown firehose.")
      return
    if address is not None and port is not None:
      LOG.info("Updating firehose connection to %s:%s:" % (address, port))
    else:
      LOG.info("Updating firehose connection to be unconfigured.")
    self._address = address
    self._port = port
    self._reset()

  def send(self, messages):
    try:
      self._lock.acquire()
      self._send(messages)
    finally:
      self._lock.release()

  def _send(self, messages, retryOnBadStatusLine=True):
    """Sends messages to the firehose. Returns True on success."""
    if self._shutdown:
      LOG.info("Ignoring send request on shutdown firehose.")
      return False
    start = time.time()
    try:
      if self._requestor is None:
        self._transceiver = avro.ipc.HTTPTransceiver(self._address,
                                                     self._port)
        self._requestor = avro.ipc.Requestor(FIREHOSE_MESSAGE_PROTOCOL,
                                             self._transceiver)
      initial_requestor_bytes = self._requestor.get_requestor_bytes_sent()
      self._requestor.request('sendAgentMessages', dict(messages=UNICODE_SANITIZER(messages)))
      self._last_message_transmit_duration_gauge.set_value(
        (time.time() - start) * 1000)
      self._message_transmit_succeeded_counter.increment()
      self._requestor_bytes_sent.increment(
        self._requestor.get_requestor_bytes_sent() - initial_requestor_bytes)
      return True
    except BadStatusLine, ex:
      # We've lost our connection. In practice this usually means the server has
      # closed a connection that we expect to be open because of HTTP keep-alive.
      # We will do a single silent retry. If the problem persistest there, we'll
      # log.
      self._reset()
      if retryOnBadStatusLine:
        return self._send(messages, retryOnBadStatusLine=False)
      self._message_transmit_failed_counter.increment()
      THROTTLED_LOG.exception("Error sending messages to firehose (retry): " +
                              self.name)
      return False
    except Exception:
      THROTTLED_LOG.exception("Error sending messages to firehose: " + self.name)
      self._reset()
      self._message_transmit_failed_counter.increment()
      return False

  def _reset(self):
    if self._transceiver is not None:
      try:
        self._transceiver.close()
      except Exception:
        THROTTLED_LOG.exception("Error closing transciever: " + self.name)
    self._transceiver = None
    self._requestor = None

  def shutdown(self):
   try:
      self._lock.acquire()
      self._shutdown = True
      self._reset()
   finally:
     self._lock.release()

  def get_debug_status_report(self):
    return DebugStatusReport(config_items=[('address', self._address),
                                           ('port', self._port),
                                           ('shutdown', self._shutdown)])

  def __str__(self):
    return "Firehose(%s, %s, %s, %s)" % (self.name,
                                         self.roletype,
                                         self._address,
                                         self._port)
