# (c) Copyright 2013 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import cmf.monitor.constants as constants
import logging
import threading
import time

from cmf.monitor import schema
from cmf.monitor.firehose import Firehose

LOG = logging.getLogger(__name__)

_FIREHOSEN = { 'ACTIVITYMONITOR' : '_amon',
               'SERVICEMONITOR'  : '_smon',
               'HOSTMONITOR'     : '_hmon' }

# See normalize_metric_value below for why this is needed.
MAX_SIGNED_LONG = (1 << 63) - 1
MIN_SIGNED_LONG = -(1 << 63)

class Firehoses(object):
  """Coordinates communication with the various firehoses. It is thread safe."""

  def __init__(self, host_id, hostname):
    self._host_id = host_id
    self._hostname = hostname

    # The following are firehose objects connecting to the various firehoses.
    self._amon = None
    self._smon = None
    self._hmon = None

    # This member is only updated by calls from the SCM heartbeat thread. It is
    # a cache of information on how to connect to the firehoses from the SCM
    # heartbeat.
    self._firehose_connection_info = None

    # The following are accessed from the agent's debug server.
    self._last_amon_message = None
    self._last_smon_message = None
    self._last_hmon_message = None

    # This holds our reporting interval, which is specified in seconds. This
    # comes from the SCM heartbeat, so the initial value doesn't matter a whole
    # lot as it will quickly be updated to what is set on the server.
    self._reporting_interval_seconds = 5.0

    self._lock = threading.Lock()

  def update_firehose_connection_info(self, firehose_connection_info):
    """
    Updates connection information to the various firehoses. The input object is
    from the SCM heartbeat. See the avro schema for more information. This call
    simply updates the cached connection information. That information will be
    processed on the next call to a send function. We do this to avoid having
    to worry about blocking the SCM heartbeat thread.

    This function does update the reporting interval member. We do this without
    the lock to avoid blocking the heartbeat thread, but we asusme that because
    this is only called from the heartbeat thread this is serialized.
    """
    self._firehose_connection_info = firehose_connection_info
    if len(firehose_connection_info) == 0:
      return

    interval = self._reporting_interval_seconds
    self._reporting_interval_seconds = \
        min([fh["report_interval"] for fh in firehose_connection_info])
    if interval != self._reporting_interval_seconds:
      LOG.info("Reporting interval updated: %s -> %s" %
               (interval, self._reporting_interval_seconds))

  def _process_firehose_connection_for_roletype(self, roletype):
    """
    Actually processes the cached firehose connection information for the input
    roletype.

    We take the object's lock to serialize execution of the state mutating
    _process_firehose_update function.
    """
    firehose_connection_info = self._firehose_connection_info
    if firehose_connection_info is None:
      return
    connection_info = None
    for fh in firehose_connection_info:
      if fh["roletype"] == roletype:
        connection_info = fh
        break
    try:
      self._lock.acquire()
      self._process_firehose_update(roletype, connection_info)
    finally:
      self._lock.release()

  def _process_firehose_update(self, roletype, connection_info):
    member = _FIREHOSEN[roletype]
    firehose = getattr(self, member)
    if connection_info is not None:
      if firehose is None:
        LOG.info("Creating a connection to the %s." % roletype)
        firehose = Firehose(connection_info["rolename"],
                            connection_info["roletype"],
                            connection_info["address"],
                            connection_info["port"])
        setattr(self, member, firehose)
      else:
        firehose.set_connection_info(connection_info["address"],
                                     connection_info["port"])
    elif firehose is not None:
      LOG.info("Shutting down connection to the %s." % roletype)
      firehose.shutdown()
      setattr(self, member, None)

  def send_amon_update(self, attempt_updates):
    """Sends an agent message to AMON with the input information."""

    # We operate on a local copy of the firehose object so that we don't have to
    # worry about it being nulled out from under us by the heartbeat thread. The
    # object itself synchronizes access where necessary.
    self._process_firehose_connection_for_roletype('ACTIVITYMONITOR')
    amon = self._amon
    if (amon == None or
        len(attempt_updates) == 0):
      return
    self._last_amon_message = \
        self._send_agent_message(amon,
                                 attempt_updates=attempt_updates)

  def send_smon_update(self,
                       service_updates,
                       role_updates,
                       impala_query_updates):
    """Sends an agent message to SMON with the input information."""

    # We operate on a local copy of the firehose object so that we don't have to
    # worry about it being nulled out from under us by the heartbeat thread. The
    # object itself synchronizes access where necessary.
    self._process_firehose_connection_for_roletype('SERVICEMONITOR')
    smon = self._smon
    if (smon == None or
        (len(service_updates) == 0 and
         len(role_updates) == 0 and
         (impala_query_updates is None or len(impala_query_updates) == 0))):
      return
    self._last_smon_message = \
        self._send_agent_message(smon,
                                 service_updates=service_updates,
                                 role_updates=role_updates,
                                 impala_query_updates=impala_query_updates)

  def send_hmon_update(self, host_update):
    """Sends an agent message to HMON with the input information."""

    # We operate on a local copy of the firehose object so that we don't have to
    # worry about it being nulled out from under us by the heartbeat thread. The
    # object itself synchronizes access where necessary.
    self._process_firehose_connection_for_roletype('HOSTMONITOR')
    hmon = self._hmon
    if hmon == None or host_update is None:
      return
    self._last_hmon_message = \
        self._send_agent_message(hmon,
                                 host_update=host_update)

  def _send_agent_message(self,
                          firehose,
                          attempt_updates = [],
                          host_update = None,
                          impala_query_updates = None,
                          service_updates = [],
                          role_updates = []):
    agentmsg = dict(ts_secs=long(time.time()),
                    hostname=self._hostname,
                    attempt_metrics=attempt_updates,
                    impala_query_updates=impala_query_updates,
                    host_update=host_update,
                    service_updates=service_updates,
                    role_metrics=role_updates,
                    host_id=self._host_id,
                    version=constants.AGENT_VERSION)
    firehose.send(dict(agent_msgs=[agentmsg]))
    return agentmsg

  def get_reporting_interval_seconds(self):
    """Gets the firehose reporting interval, in seconds."""
    return self._reporting_interval_seconds

  def get_last_amon_message(self):
    """Returns the most recent message sent to AMON."""
    return self._last_amon_message

  def get_last_hmon_message(self):
    """Returns the most recent message sent to HMON."""
    return self._last_hmon_message

  def get_last_smon_message(self):
    """Returns the most recent message sent to SMON."""
    return self._last_smon_message

  def get_debug_status_reports(self):
    """Returns map of name -> debug status reports for every firehose."""
    ret = {}
    for member in _FIREHOSEN.itervalues():
      firehose = getattr(self, member)
      if firehose is not None:
        ret[firehose.name] = firehose.get_debug_status_report()
    return ret

  def get_firehose_metrics(self, roletype):
    """
    return the metrics for the firehose 'roletype'. Cannot return None.
    """
    metrics = {}
    if roletype == "HOSTMONITOR":
      firehose = self._hmon
      if firehose is None:
        return {}
      metrics[schema.HMON_MESSAGE_TRANSMIT_DURATION] = \
        firehose.get_last_message_transmit_duration()
      metrics[schema.HMON_MESSAGE_TRANSMIT_SUCCEEDED] = \
        firehose.get_message_transmit_succeeded()
      metrics[schema.HMON_MESSAGE_TRANSMIT_FAILED] = \
        firehose.get_message_transmit_failed()
      metrics[schema.HMON_MESSAGE_BYTES_SENT] = \
        firehose.get_message_bytes_sent()
    elif roletype == "SERVICEMONITOR":
      firehose = self._smon
      if firehose is None:
        return {}
      metrics[schema.SMON_MESSAGE_TRANSMIT_DURATION] = \
        firehose.get_last_message_transmit_duration()
      metrics[schema.SMON_MESSAGE_TRANSMIT_SUCCEEDED] = \
        firehose.get_message_transmit_succeeded()
      metrics[schema.SMON_MESSAGE_TRANSMIT_FAILED] = \
        firehose.get_message_transmit_failed()
      metrics[schema.SMON_MESSAGE_BYTES_SENT] = \
        firehose.get_message_bytes_sent()
    return metrics

  @staticmethod
  def normalize_metric_value(value):
    """
    The function ensures that metric values can be serialized (see MetricValue
    in cmon_agent.avdl for the avro schema). Any long that is bigger than
    MAX_SIGNED_LONG will fail to be serialized (see OPSAPS-24118 for more
    details). We will treat any long larger than that as a "counter reset"
    assuming that only counters can get to be this big. If there are any gauges
    that use the full range of longs we will probably solve this by converting
    their units.
    :param value:
    :return:
    """
    if isinstance(value, int) or isinstance(value, long):
      if (value > 0) :
        return value % MAX_SIGNED_LONG
      else:
        return value % MIN_SIGNED_LONG
    return value
