# Copyright (c) 2012 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging
import time

import cmf.monitor.flume.json_parser

from cmf.monitor import schema, constants
from cmf.monitor.flume.flume_update import FlumeUpdate
from cmf.monitor.abstract_monitor import AbstractMonitor
from cmf.monitor.base import DebugStatusReport
from cmf.monitor.role_update import RoleUpdate
from cmf.monitor.service_update import ServiceUpdate

from cmf.monitor.cgroup_metrics_utils import add_cgroup_metrics
from cmf.monitor.proc_metrics_utils import add_proc_metrics

LOG = logging.getLogger('FlumeAgentMonitor')

_ROLETYPE = "AGENT"
_SERVICETYPE = "FLUME"

# Section and key names in cloudera-monitor.properties
_SECTION = 'agent'
_SERVICE_NAME = 'service_name'
_SERVICE_VERSION = 'service_version'
_ROLE_NAME = 'role_name'
_HOST = 'host'
_LOG_DIR = 'log_dir'
_COLLECT_INTERVAL = 'collect_interval'
_FLUME_HTTP_PORT = 'flume_http_port'
# Added in CM 5.0 Beta 2
_AGENT_NAME = 'agent_name'
# Added in CM 5.4
_HEAP_DUMP_DIR = 'oom_heap_dump_dir'

class FlumeAgentMonitor(AbstractMonitor):
  def __init__(self, daemon, registry, cgroup_root=None, cgroup_name=None):
    AbstractMonitor.__init__(self, daemon, registry, cgroup_root, cgroup_name, _ROLETYPE)
    self._last_conf_path = None
    self._metrics_tuple = (None, None, None)
    self._last_agent_report_timestamp = None
    self._last_collection_ended = None
    self._last_collection_length = None
    self._role_dir_monitors = [ ]

    # Config data
    self._name = None
    self._service_name = None
    self._service_version = None
    self._log_dir = None
    self._metrics_url = None
    self._agent_name = None

  def __str__(self):
    return "FlumeAgentMonitor for %s" % (self._name,)

  def _reload_config(self, confparse):
    """Overrides AbstractMonitor. Returns True/False for success"""
    try:
      service = confparse.get(_SECTION, _SERVICE_NAME)
      name = confparse.get(_SECTION, _ROLE_NAME)
      version = confparse.getint(_SECTION, _SERVICE_VERSION)
      log_dir = confparse.get(_SECTION, _LOG_DIR)
      heap_dump_dir = confparse.get_with_default(_SECTION, _HEAP_DUMP_DIR, None)
      host = confparse.get(_SECTION, _HOST)
      http_port = confparse.getint(_SECTION, _FLUME_HTTP_PORT)
      agent_name = confparse.get_with_default(_SECTION, _AGENT_NAME, None)
    except ValueError, ex:
      self.record_error(ex)
      return False

    self._name = name
    self._service_name = service
    self._service_version = version
    self._log_dir = log_dir
    self._agent_name = agent_name
    self._metrics_url = "http://%s:%s/metrics?format=json" % (host, http_port)

    # Make monitors for the log directory and the data directories.
    self._role_dir_monitors = [ ]
    if heap_dump_dir is not None:
      self.create_dir_monitor(heap_dump_dir, self._role_dir_monitors)
    self.create_dir_monitor(self._log_dir, self._role_dir_monitors)

    return True

  def get_role_update(self):
    """Returns a RoleUpdate to be sent to the firehose"""
    update = RoleUpdate(name=self._name,
                        roletype=_ROLETYPE,
                        servicetype=_SERVICETYPE,
                        service_name=self._service_name)

    if self._agent_name is not None:
      update.set_additional_role_attributes(
        { constants.AGENT_NAME : self._agent_name })

    if self.add_metrics_from_webserver(self._metrics_tuple,
                                       self._last_reported_timestamp,
                                       update):
      self._last_reported_timestamp = update.timestamp

    add_proc_metrics(update, self._pid, LOG)
    add_cgroup_metrics(update, self._cgroup_root, self._cgroup_name)

    for mon in self._role_dir_monitors:
      update.add_dir_update(mon.get_role_dir_update())

    self.add_unexpected_exit_metrics(update)

    return update

  def add_webserver_metrics_to_role_update(self, update, metrics):
    if self._is_feature_enabled(update, "send_flume_update"):
      for (component_type, component_name), metricdict in metrics.updates.items():
        flume_update = FlumeUpdate(self._agent_name,
                                   component_type,
                                   component_name)
        for metric_id, metric_value in metricdict.items():
          flume_update.add_metric(metric_id, metric_value)
        update.add_flume_update(flume_update)

  def get_debug_status_report(self):
    return DebugStatusReport(config_items=self.conf_parser.items(_SECTION),
                             errors=(self.last_error,),
                             last_collection_ended=self._last_collection_ended,
                             last_collection_length=self._last_collection_length)

  def collect(self):
    """Overrides AbstractMonitor. Collect metrics. Invoked asynchronously."""
    start = time.time()

    # Reload config if necessary
    confparse = self.conf_parser
    if self._last_conf_path != confparse.path:
      if not self._reload_config(confparse):
        return
      self._last_conf_path = confparse.path

    # Collect directory info
    for mon in self._role_dir_monitors:
      mon.collect()

    # If our pid value is None, it means that the associated process is not
    # running. If this is the case, we skip metric collection.
    if self._pid is None:
      return

    LOG.debug("Fetching Flume metrics at '%s'" % (self._metrics_url,))
    self._metrics_tuple = self.collect_metrics_from_url(self._metrics_url)

    self._last_collection_ended = time.time()
    self._last_collection_length = self._last_collection_ended - start

  def parse_metrics(self, json):
    return cmf.monitor.flume.json_parser.JsonParser(json)
