# Copyright (c) 2013 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging
import os
import time

from cmf.monitor import constants
from cmf.monitor.conf import SERVICE_METRICS_PROPERTIES, CSD_MONITORING_PROPERTIES
from cmf.monitor.monitor_properties import ConfParser
from cmf.monitor.abstract_monitor import AbstractMonitor
from cmf.monitor.base import DebugStatusReport
from cmf.monitor.cgroup_metrics_utils import add_cgroup_metrics
from cmf.monitor.generic.metric_collectors import WebServerMetricCollector, \
  SampleFileMetricCollector, AbstractMetricCollector
from cmf.monitor.proc_metrics_utils import add_proc_metrics, \
    add_with_descendants_proc_metrics
from cmf.monitor.role_update import RoleUpdate
from cmf.throttling_logger import ThrottlingLogger

LOG = logging.getLogger('GenericMonitor')
THROTTLED_LOG = ThrottlingLogger(LOG, constants.DEFAULT_MONITOR_LOG_RATE)

ROLE_NAME = 'role_name'
SERVICE_NAME = 'service_name'
_SERVICE_VERSION = 'service_version'

class GenericMonitor(AbstractMonitor):

  def __init__(self,
               service_type,
               role_type,
               daemon,
               registry,
               adapter,
               cgroup_root=None,
               cgroup_name=None,
               audit_plugin=None):
    AbstractMonitor.__init__(self, daemon, registry, cgroup_root, cgroup_name, role_type)
    self._service_type = service_type
    self._section = role_type.lower()
    self._last_conf_path = None
    self._last_conf_timestamp = None
    self._url_collector = \
      WebServerMetricCollector(self._section, adapter, daemon.metric_accessors,
                               username_attr=adapter.get_metrics_url_username_attr(),
                               password_attr=adapter.get_metrics_url_password_attr())
    self._sample_file_collector = \
      SampleFileMetricCollector(adapter, daemon.metric_accessors)
    self._last_collection_ended = None
    self._last_collection_length = None
    self._role_dir_monitors = [ ]
    self._audit_plugin = audit_plugin
    self._adapter = adapter

    # Config data
    self._role_name = None
    self._service_name = None
    self._service_version = None
    self._previous_metrics = {}
    self._previous_cpu_user = None
    self._previous_cpu_system = None

  def __str__(self):
    return "GenericMonitor %s-%s for %s" % (self._service_type,
                                            self._role_type,
                                            self._role_name,)

  def _reload_config_if_necessary(self):
    """
    Returns True for success or no need to reload.
    Returns False if fail to reload the conf files.
    """
    conf = self.conf_parser
    if self._last_conf_path != conf.path or \
       self._last_conf_timestamp != conf.timestamp:
      if not self._reload_config(conf):
        return False
      self._last_conf_path = conf.path
      self._last_conf_timestamp = conf.timestamp

    return True

  def _reload_config(self, conf):
    """Returns True/False for success"""
    try:
      role_name = conf.get(self._section, ROLE_NAME)
      service_name = conf.get(self._section, SERVICE_NAME)
      service_version = conf.getint_with_default(
        self._section, _SERVICE_VERSION, 3)
      role_dir_monitors = self._adapter.get_monitored_directories(conf)
    except:
      LOG.exception("Error parsing %s" % os.path.abspath(conf.path))
      return False

    try:
      monitoring_properties_path = os.path.join(os.path.dirname(conf.path),
                                CSD_MONITORING_PROPERTIES)
      if os.path.exists(monitoring_properties_path):
        csd_conf_parser = ConfParser(os.path.dirname(conf.path),
                                CSD_MONITORING_PROPERTIES)
        self._adapter.read_csd_monitoring_properties(csd_conf_parser)
    except:
      LOG.exception("Error parsing %s" % os.path.abspath(monitoring_properties_path))
      return False

    for collector in self.get_all_collectors():
      try:
        collector.update_with_conf(conf)
      except:
        THROTTLED_LOG.exception(
          "Error updating collector [%s] with configurations from %s." %
          (collector.name, os.path.abspath(conf.path)))

    metrics_path = os.path.join(os.path.dirname(conf.path),
                                SERVICE_METRICS_PROPERTIES)
    try:
      if os.path.exists(metrics_path):
        self._adapter.read_service_defined_metrics(metrics_path)
    except:
      LOG.exception("Error parsing %s" % os.path.abspath(metrics_path))
      return False

    self._role_name = role_name
    self._service_name = service_name
    self._service_version = service_version
    self._role_dir_monitors = []
    for directory in role_dir_monitors:
      self.create_dir_monitor(directory['path'],
                              self._role_dir_monitors,
                              directory['type'])

    return True

  def ensure_collecting(self):
    AbstractMonitor.ensure_collecting(self)
    for collector in self.get_all_collectors():
      try:
        collector.on_start_collecting()
      except:
        LOG.exception("Error notifying collector [%s] on_start_collecting()" %
                      collector.name)

  def stop(self):
    AbstractMonitor.stop(self)
    for collector in self.get_all_collectors():
      try:
        collector.on_stop_collecting()
      except:
        LOG.exception("Error notifying collector [%s] on_stop_collecting()" %
                      collector.name)

  def collect(self):
    """Overrides AbstractMonitor. Collect standard metrics. Invoked
       asynchronously."""
    start = time.time()

    # Try to reload conf if necessary, abort if exception happens.
    success = self._reload_config_if_necessary()
    if not success:
      return

    # Collect directory info
    for mon in self._role_dir_monitors:
      mon.collect()

    # If our pid value is None, it means that the associated process is not
    # running. If this is the case, we skip metric collection.
    if self._pid is None:
      for collector in self.get_all_collectors():
        collector.clear_results()
      return

    for collector in self.get_all_collectors():
      if collector.is_enabled():
        LOG.debug("Fetching metrics for '%s' at '%s'" %
                  (self._role_name, collector.metrics_uri,))
        collector.collect_and_parse(self._conf_parser)

    self._last_collection_ended = time.time()
    self._last_collection_length = self._last_collection_ended - start

  def get_role_update(self):
    """Returns a RoleUpdate to be sent to the firehose"""
    update = RoleUpdate(name=self._role_name,
                        roletype=self._role_type,
                        servicetype=self._service_type,
                        service_name=self._service_name)

    self._adapter.role_update_init(update)

    for collector in self.get_all_collectors():
      if collector.is_enabled():
        collector.report_metrics(self._service_version, update)

    add_proc_metrics(update, self._pid, LOG)
    if self._adapter.descendants_proc_metrics_flag is True:
      self._previous_metrics, self._previous_cpu_user, self_previous_system =\
          add_with_descendants_proc_metrics(update,
                                            self._pid,
                                            self._previous_metrics,
                                            self._previous_cpu_user,
                                            self._previous_cpu_system,
                                            LOG)
    add_cgroup_metrics(update, self._cgroup_root, self._cgroup_name)

    if self._audit_plugin is not None:
      self._audit_plugin.add_metrics(update)

    for mon in self._role_dir_monitors:
      update.add_dir_update(mon.get_role_dir_update())

    self.add_unexpected_exit_metrics(update)
    return update

  def get_debug_status_report(self):
    return DebugStatusReport(config_items=self.conf_parser.items(self._section),
                             errors=(self.last_error,),
                             last_collection_ended=self._last_collection_ended,
                             last_collection_length=self._last_collection_length)

  def is_process_to_monitor(self, process):
    """
    For now, we almost always assume that the immediate child of supervisord is
    the process we are interested in monitoring. This may need to be extended in
    the future if we find that some of the things we are monitoring with the
    generic process monitor have intermediate processes in between supervisord
    and the process we actually want to monitor. The reason we override this at
    all is that the base functionality only works well for java processes.
    
    We make an exception here for roles for which we use the generic monitor we
    know also use jsvc, otherwise we end up with something like OPSAPS-20784.
    """
    if self._role_type == 'NFSGATEWAY' or self._role_type == 'DATANODE':
      return AbstractMonitor.is_process_to_monitor(self, process)
    return True

  def get_all_collectors(self):
    collectors = [self._url_collector, self._sample_file_collector]
    collectors.extend(self._adapter.get_adapter_specific_collectors())
    return collectors

  def is_ready_to_report(self):
    if self._pid is None:
      return False
    no_specific_collector_enabled = True

    # Try to reload conf if necessary, abort if exception happens.
    success = self._reload_config_if_necessary()
    if not success:
      return False

    for collector in self.get_all_collectors():
      collector_ready = collector.is_ready_to_report(self.conf_parser,
                                                     self._pid)
      if collector_ready is None:
        continue
      no_specific_collector_enabled = False
      if collector_ready:
        return True
    # if we are here it means that either no specific collector is enabled or
    # all enabled collectors returned False and are not ready to report.
    return no_specific_collector_enabled

  def clear_last_reported_timestamp(self):
    AbstractMonitor.clear_last_reported_timestamp(self)
    for collector in self.get_all_collectors():
      collector.clear_last_reported_timestamp()

