# Copyright (c) 2015 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging
import os
try:
  import json
except ImportError:
  import simplejson as json

from cmf.monitor.generic import ROLE_NAME, SERVICE_NAME

LOG = logging.getLogger('Adapter')

class Adapter(object):
  """
  Provides information for a specific monitored entity to the generic role
  monitor.
  """

  _HOST = 'host'
  _SERVICE_VERSION = 'service_version'
  # Added in CM 4.1
  _LOG_DIR = 'log_dir'
  # Added in CM 5.4
  _HEAP_DUMP_DIR = 'oom_heap_dump_dir'
  _SERVICE_RELEASE = 'service_release'
  _MONITORED_DIRECTORIES = "monitored_directories"
  # Added in CM 5.5
  _MONITORED_DIRECTORIES_WITH_TYPES = "monitored_directories_with_types"

  def __init__(self, service_type, role_type, safety_valve):
    if role_type is None:
      raise Exception("A role type is required!")
    if service_type is None:
      raise Exception("A service type is required!")
    if safety_valve is None:
      raise Exception("Safety valve is required!")
    self._descendants_proc_metrics_flag = False
    self._role_type = role_type
    self._service_type = service_type
    self._safety_valve = safety_valve

  @property
  def safety_valve(self):
    """
    The adapeter's safety valve.
    """
    return self._safety_valve

  @property
  def role_type(self):
    """
    The role type of the role being monitored.
    """
    return self._role_type

  @property
  def service_type(self):
    """
    The service type of the role being monitored.
    """
    return self._service_type

  @property
  def section(self):
    """
    The config section in the role's cloudera-monitor.properties from which
    config values should be read.
    """
    return self._role_type.lower()

  @property
  def descendants_proc_metrics_flag(self):
    """
    The flag indicates whether metrics with descendant process for node manager
    are collected.
    """
    return self._descendants_proc_metrics_flag

  def get_monitored_directories(self, conf):
    """
    Returns a list of (path, type) dicts of directories that should be monitored
    as role directories. Newly monitored roles should make use of the
    MONITORED_DIRECTORIES_WITH_TYPES path. A number of additional methods exist
    to support legacy role monitoring. See the MONITORED_DIRECTORIES path or
    the two helper methods below for more detail.
    """
    if conf is None:
      raise Exception("A configuration is required!")
    directories = []
    try:
      for key in self._get_monitored_directories_conf_keys():
        directory = conf.get_with_default(self.section, key, None)
        if directory is not None:
          directories.append(dict(path=directory, type=None))

      monitored_directories_lists_keys = [Adapter._MONITORED_DIRECTORIES]
      monitored_directories_lists_keys.extend(
        self._get_monitored_directories_lists_conf_keys())
      for key in monitored_directories_lists_keys:
        for directory in conf.getlist_with_default(self.section, key, []):
          directories.append(dict(path=directory, type=None))

      directories.extend(self._get_monitored_directories_with_types(conf))
    except:
      LOG.exception("Error parsing %s" % os.path.abspath(conf.path))
    return directories

  def _get_monitored_directories_with_types(self, conf):
    """
    Parse the json string value for "monitored_directories_with_types" property
    and return a list of (directory, type) dicts.
    Example of valid config:{[{"path":"/path/to/mon/dir1","type":"dfs_data_dirs"}]}
    """
    dir_type_json = conf.get_with_default(
      self.section, Adapter._MONITORED_DIRECTORIES_WITH_TYPES, "[]")
    dir_type_list = json.loads(dir_type_json)
    directories = [dict(path=d['path'],
                        type=d['type'])
                   for d in dir_type_list]
    return directories

  def _get_monitored_directories_conf_keys(self):
    """
    Used by the default implementation of get_monitored_directories. This
    returns a list of configuration keys to use in lookups into the role's
    cloudera-monitor.properties file. Each key should return a single directory
    to monitor as a string. It is okay if the configuration key does not exist
    in the cloudera-monitor.properties file.
    """
    return [Adapter._LOG_DIR,
            Adapter._HEAP_DUMP_DIR]

  def _get_monitored_directories_lists_conf_keys(self):
    """
    Used by the default implementation of get_monitored_directories. This
    returns a list of configuration keys to use in lookups into the role's
    cloudera-monitor.properties file. Each key should return a list of
    directories to monitor as a comma separated string. It is okay if the
    configuration key does not exist in the cloudera-monitor.properties file.
    """
    return []

  def get_metrics_url(self, conf):
    """
    Return the URL from which to fetch metrics. If None is returned, no attempt
    will be made to fetch metrics from the url.
    """
    return None

  def get_metrics_url_username_attr(self):
    """
    Return the username attribute to authenticate against metrics URL.
    """
    return None

  def get_metrics_url_password_attr(self):
    """
    Return the password attribute to authenticate against metrics URL.
    """
    return None

  def get_metrics_file(self, conf):
    """
    Return the full path to a metrics sample file from which to fetch metrics.
    If None is returned, no attempt will be made to fetch metrics from the file.
    """
    return None

  def get_metrics_sample(self, version):
    """
    Only used in testing, and only for adapters that support metrics. Return a
    metric sample fixture for the input version.
    """
    return None

  def read_service_defined_metrics(self, path):
    """
    Called during the role monitor's configuration refresh to trigger parsing of
    the service-metrics.properties file. By default this does nothing, and it
    should be implemented by adapters that use this file during collection.
    """
    pass

  def read_csd_monitoring_properties(self, path):
    '''
    Called during the role monitor's configuration refresh to trigger parsing of
    the csd-monitor.properties file. This file is generated from the
    MetricsSourceDescriptor in CSD.
    '''
    pass

  def parse_metrics_from_url(self, conf, json):
    """
    Parse the input metrics JSON and return an opaque object that will be passed
    back to the adapter in the call to add_webserver_metrics.
    """
    raise Exception(
        "%s-%s must implement parse_metrics!" % 
        (self.service_type, self.role_type))

  def parse_metrics_from_file(self, conf, json):
    """
    Parse the input metrics JSON retrieved from a sample file and return an
    opaque object that will be passed back to the adapter in the call to
    add_file_sample_collection_metrics. Note that if an adapter returned a non
    None value from the get_metrics_file it must implement this function.
    """
    raise Exception(
        "%s-%s must implement parse_metrics_from_file!" %
        (self.service_type, self.role_type))

  def add_webserver_metrics(self, version, update, metrics, accessors):
    """
    Populates the input role update object from the input url parsed metrics
    object. Note that if an adapter returned a non None value from the
    get_metrics_url it must implement this method.
    """
    raise Exception(
        "%s-%s must implement add_webserver_metrics!" % 
        (self.service_type, self.role_type))

  def add_sample_file_metrics(self, version, update, metrics, accessors):
    """
    Populates the input role update object from the input parsed metrics sample
    file object. Note that if an adapter returned a non None value from the
    get_metrics_file it must implement this method.
    """
    raise Exception(
        "%s-%s must implement add_sample_file_metrics!" %
        (self.service_type, self.role_type))

  def role_update_init(self, update):
    """
    Hook to initialize role update. Guaranteed to be called once per update,
    regardless of whether collection is successful or timestamp is valid etc.
    For example, Impalad adapter can use it to make sure RoleUpdate's
    ImpalaDaemonUpdate is always populated.
    """
    pass

  def _extract_role_and_service_name(self, conf):
    """
    Extract the role name and the service name from the configuration. Returns
    a tuple (role name, service name) or raises an exception if any of them
    cannot be extracted.
    """
    if conf is None:
      raise Exception("A configuration is required!")

    role_name = conf.get(self.section, ROLE_NAME)
    if role_name is None:
      raise Exception("%s entry missing from %s" %
                      (ROLE_NAME, os.path.abspath(conf.path)))

    service_name = conf.get(self.section, SERVICE_NAME)
    if service_name is None:
      raise Exception("%s entry missing from %s" %
                      (SERVICE_NAME, os.path.abspath(conf.path)))
    return role_name, service_name

  def get_version(self, conf):
    """
    A helper function to extract the service version from a configuration.
    """
    return conf.getint_with_default(self.section,
                                    self._SERVICE_VERSION,
                                    3)

  def _is_feature_enabled(self, role_update, feature, role_name = None, service_name = None):
    """
    Returns true if the feature in question is enabled for the role which
    the role_update is for. If role_update is provided, its role_name and
    service_name are used, otherwise the explicit arguments are used.
    """
    if role_update is None and (role_name is None or service_name is None):
      raise ValueError("Either role_update or explicit role/service names must be provided.")
    if role_update:
      role_name = role_update.role_name
      service_name = role_update.service_name
    role_key = "%s_%s_feature_%s_enabled" % \
               (service_name.lower(),
               role_name.lower(),
               feature)
    service_key = "%s_feature_%s_enabled" % \
                  (service_name.lower(),feature)
    return 'true'  == self._safety_valve.get(role_key, 'true').strip().lower() \
           and \
           'true'  == self._safety_valve.get(service_key, 'true').strip().lower()

  def get_release(self, conf):
    """
    A helper function to extract the service version from a configuration.
    """
    return conf.get_with_default(self.section,
                                 self._SERVICE_RELEASE,
                                 None)

  def get_adapter_specific_collectors(self):
    """
    Adapters may return a list of additional collectors, all must extend
    AbstractMetricCollector. These collectors will be used by the generic
    adapter to collect additional metrics, on top of file and web-based metrics.
    """
    return []
