# Copyright (c) 2017 Cloudera, Inc. All rights reserved.

import logging
import os

from cmf.monitor import constants
from cmf.monitor.generic.adapter import Adapter
from cmf.monitor.generic.service_defined_metrics import ServiceDefinedMetrics
from cmf.monitor.generic.service_defined_metrics import \
  ServiceDefinedMetricsJsonJmxBeans
from cmf.monitor.generic.utils import visit_json
from cmf.monitor.generic.utils import SimpleMetricsExtractor, JsonJmxBeanMetricsExtractor
from cmf.monitor.generic.utils import SplittingSourceProcessor
from cmf.throttling_logger import ThrottlingLogger

LOG = logging.getLogger('CommonAdapter')
THROTTLED_LOG = ThrottlingLogger(LOG, 60 * 60) # Once an hour.

class CommonAdapter(Adapter):
  """
  Class for simple role adapters. This adapter would try to enable fetching
  role level metrics from web url or metrics file.
  """
  _CONF_METRICS_AVAILABLE = 'metrics_available'
  _CONF_WEB_METRICS_ENABLED = 'web_metrics_enabled'
  _CONF_HTTP_URL_KEY = 'http_url'
  _CONF_HTTPS_URL_KEY = 'https_url'
  _CONF_FILE_METRICS_ENABLED = 'file_metrics_enabled'
  _CONF_METRICS_FILE_PATH_KEY = 'file_path'
  _CONF_SSL_ENABLED = 'ssl_enabled'
  _CONF_ROLE_ENTITY_TYPE_KEY = 'role_entity_type'
  _CONF_METRICS_FORMAT_KEY = 'metrics_format'
  _DEFAULT_SEPARATOR = '::'

  SIMPLE_METRICS_FORMAT = 'SIMPLE'
  JMX_METRICS_FORMAT = 'JMX'

  def __init__(self,
               service_type,
               role_type,
               safety_valve):
    Adapter.__init__(self, service_type, role_type, safety_valve)

    self._service_defined_metrics = None
    self._web_metrics_enabled = False
    self._web_server_url = None # HTTP or HTTPS
    self._ssl_enabled = None
    self._sample_file_enabled = False
    self._sample_file_path = None
    self._separator = CommonAdapter._DEFAULT_SEPARATOR
    self._role_entity_type = role_type
    self._metrics_format = CommonAdapter.SIMPLE_METRICS_FORMAT

  @property
  def role_entity_type(self):
    return self._role_entity_type

  def read_csd_monitoring_properties(self, csd_conf_parser):
    '''
    Read conf from csd-monitoring.properties. This function should be called
    every time cloudera-monitor.properties file is reloaded.
    '''
    if csd_conf_parser is None:
      raise Exception("Configuration file csd-monitoring.properties is required!")

    try:
      # Check if the service support metrics collection.
      metrics_available = csd_conf_parser.getboolean_with_default(self.section,
                                CommonAdapter._CONF_METRICS_AVAILABLE,
                                False)
      if not metrics_available:
        THROTTLED_LOG.warn(\
            "Metrics collection not supported by the service %s" % \
            self.service_type)
        return

      # In service-metrics.properties file, TimeSeriesEntityType's name is
      # used as the key to fetch all the metrics defined for the role. In
      # the past the role name is just used as the type name, but for CSD-
      # based roles, we have switched to SERVICE_NAME-ROLE_NAME. Get this
      # key from the CM server. By default, use the role type.
      role_entity_type = csd_conf_parser.get(self.section,
                            CommonAdapter._CONF_ROLE_ENTITY_TYPE_KEY)
      if role_entity_type:
        self._role_entity_type = role_entity_type

      # Check in which format are metrics exposed.
      metrics_format = csd_conf_parser.get(self.section,
                            CommonAdapter._CONF_METRICS_FORMAT_KEY)
      if metrics_format:
        if metrics_format != CommonAdapter.SIMPLE_METRICS_FORMAT and\
            metrics_format != CommonAdapter.JMX_METRICS_FORMAT:
          THROTTLED_LOG.exception("Unsupported metrics format %s" % metrics_format)
          return
        else:
          self._metrics_format = metrics_format

      # Check if metrics should be fetched through web url.
      web_metrics_enabled = csd_conf_parser.getboolean_with_default(self.section,
                                CommonAdapter._CONF_WEB_METRICS_ENABLED,
                                False)
      if web_metrics_enabled:
        web_server_url = None
        use_ssl = csd_conf_parser.get_with_default(self.section,
                                          CommonAdapter._CONF_SSL_ENABLED,
                                          False)
        if use_ssl:
          https_url = csd_conf_parser.get(self.section,
              CommonAdapter._CONF_HTTPS_URL_KEY)
          if https_url is None:
            THROTTLED_LOG.exception("%s entry missing from %s when web metrics collection is enabled" % \
                (CommonAdapter._CONF_HTTPS_URL_KEY, csd_conf_parser.path))
            return
          web_server_url = https_url
        else:
          http_url = csd_conf_parser.get(self.section,
              CommonAdapter._CONF_HTTP_URL_KEY)
          if http_url is None:
            THROTTLED_LOG.exception("%s entry missing from %s when web metrics collection is enabled" % \
                (CommonAdapter._CONF_HTTP_URL_KEY, csd_conf_parser.path))
            return
          web_server_url = http_url

        self._web_metrics_enabled = web_metrics_enabled
        self._web_server_url = web_server_url
      else:
        # Check if metrics should be fetched from sample file. If both web and file
        # metrics are enabled, we give preference to web metrics.
        sample_file_enabled = csd_conf_parser.getboolean_with_default(self.section,
                                     CommonAdapter._CONF_FILE_METRICS_ENABLED,
                                     False)
        if sample_file_enabled:
          sample_file_path = csd_conf_parser.get(self.section,
                                    CommonAdapter._CONF_METRICS_FILE_PATH_KEY)
          if sample_file_path is None:
            THROTTLED_LOG.exception("%s entry missing from %s" % \
                (CommonAdapter._CONF_METRICS_FILE_PATH_KEY, csd_conf_parser.path))
            return

          self._sample_file_enabled = sample_file_enabled
          self._sample_file_path = sample_file_path

    except:
      THROTTLED_LOG.exception("Failed to read conf file '%s'" % \
                              (os.path.abspath(csd_conf_parser.path)))

  def read_service_defined_metrics(self, path):
    if path is None:
      raise Exception("A path to service-metrics.properties is required!")
    if self._metrics_format == CommonAdapter.SIMPLE_METRICS_FORMAT:
      self._service_defined_metrics = ServiceDefinedMetrics(path,
                          SplittingSourceProcessor(self._separator))
    elif self._metrics_format == CommonAdapter.JMX_METRICS_FORMAT:
      self._service_defined_metrics = \
          ServiceDefinedMetricsJsonJmxBeans(path,
              SplittingSourceProcessor(self._separator))

  def _parse_metrics_from_json(self, conf, json):
    if json is None:
      raise Exception("A json sample is required!")
    if self._service_defined_metrics is None:
      raise Exception("No metrics have been loaded!")

    role_extractor = None
    if self._metrics_format == CommonAdapter.SIMPLE_METRICS_FORMAT:
      role_extractor = SimpleMetricsExtractor(
        self._service_defined_metrics.get_sources(self._role_entity_type))
    elif self._metrics_format == CommonAdapter.JMX_METRICS_FORMAT:
      role_extractor = JsonJmxBeanMetricsExtractor(
        self._service_defined_metrics.get_entity_beans(self._role_entity_type))
    else:
      THROTTLED_LOG.exception("Unsupported metrics format %s" % metrics_format)
      return None

    visit_json(json, [role_extractor])
    return role_extractor

  def _add_metrics(self, service_version, update, metrics, accessors):
    version = constants.VERSION_STRINGS.get(service_version, None)
    if version is None:
      LOG.warn("Unknown %s version: %s" % (self._role_type, version))
      return
    if metrics:
      for metricid, value in metrics.metrics.iteritems():
        update.add_metric(metricid, value)

  def get_metrics_url(self, conf):
    if self._web_metrics_enabled:
      return self._web_server_url
    return None

  def parse_metrics_from_url(self, conf, json):
    return self._parse_metrics_from_json(conf, json)

  def add_webserver_metrics(self, version, update, metrics, accessors):
    self._add_metrics(version, update, metrics, accessors)

  def get_metrics_file(self, version):
    if self._sample_file_enabled:
      return self._sample_file_path
    return None

  def parse_metrics_from_file(self, conf, json):
    return self._parse_metrics_from_json(conf, json)

  def add_sample_file_metrics(self, version, update, metrics, accessors):
    self._add_metrics(version, update, metrics, accessors)
