# Copyright (c) 2015 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

try:
  from cStringIO import StringIO
except:
  from StringIO import StringIO

import logging
import os
import re

from cmf.monitor import constants
from cmf.monitor.firehoses import Firehoses
from cmf.monitor.generic.adapter import Adapter
from cmf.monitor.jmx_beans import JsonParserJmx
from cmf.monitor.json_metrics_utils import add_derived_jvm_metrics_metrics, \
  add_derived_jvm_metrics_jmx, add_metrics_metrics, add_metrics_jmx

LOG = logging.getLogger('HBaseAdapters')

REGION_METRIC_PATTERN = re.compile('^[Nn]amespace_(.+)_table_(.+)_region_(.+)_metric_(.+)')
REGIONSERVER_REPLICATION_METRIC_PATTERN = re.compile('^source\.(.+)\.([^.]+)$')

class _HBaseAdapter(Adapter):
  """
  Base class for all HBase service role adapters. This exists to provide utility
  methods and to ensure consistency in the handling of common metrics. HBase
  is doing things differently enough between cdh4 and cdh5 that we do not derive
  from HadoopAdapter
  """
  # Added in CM 5.1
  _USE_SSL = 'use_ssl'

  def __init__(self, role_type, safety_valve, port_key=None):
    Adapter.__init__(self, "HBASE", role_type, safety_valve)
    self._port_key = port_key
    self._metrics = None

  def get_metrics_url(self, conf):
    try:
      host = conf.get(self.section, Adapter._HOST)
      if host is None:
        LOG.error("%s entry missing from %s" % (Adapter._HOST,
                                                os.path.abspath(conf.path)))
        return None
      port = conf.getint(self.section, self._port_key)
      if port is None:
        LOG.error("%s entry missing from %s" % (self._port_key,
                                                os.path.abspath(conf.path)))
        return None
      service_version = self.get_version(conf)
      if service_version is None:
        LOG.error("%s entry missing from %s" % (_HBaseAdapter._SERVICE_VERSION,
                                                os.path.abspath(conf.path)))
        return None
      use_ssl = conf.getboolean_with_default(self.section,
                                             _HBaseAdapter._USE_SSL,
                                             False)
    except:
      LOG.exception("Error parsing %s" % os.path.abspath(conf.path))
      return None

    protocol = 'http'
    if use_ssl:
      protocol = 'https'
    return "%s://%s:%s/jmx" % (protocol, host, port)

  def parse_metrics_from_url(self, conf, json):
    return JsonParserJmx(json, self.role_type)

  def add_webserver_metrics(self, version, update, metrics, accessors):
    if version == constants.SERVICE_VERSION_CDH5:
      self._add_webserver_metrics_to_role_update_cdh5(update,
                                                      metrics,
                                                      accessors)
      self._add_role_specific_webserver_metrics_cdh5(update, metrics, accessors)
    elif version == constants.SERVICE_VERSION_CDH6:
      self._add_webserver_metrics_to_role_update_cdh6(update,
                                                      metrics,
                                                      accessors)
      self._add_role_specific_webserver_metrics_cdh6(update, metrics, accessors)
    else:
      raise Exception("Unsupported operation.")

  def _add_webserver_metrics_to_role_update_cdh5(self,
                                                 update,
                                                 metrics,
                                                 accessors):
    add_derived_jvm_metrics_jmx(update, metrics)
    add_metrics_jmx(update, metrics,
                    accessors,
                    self._role_type,
                    'cdh5')

  def _add_webserver_metrics_to_role_update_cdh6(self,
                                                 update,
                                                 metrics,
                                                 accessors):
    add_derived_jvm_metrics_jmx(update, metrics)
    add_metrics_jmx(update, metrics,
                    accessors,
                    self._role_type,
                    'cdh6')

  def _add_role_specific_webserver_metrics_cdh5(self,
                                                update,
                                                metrics,
                                                accessors):
    """
    A method that role adapters can override to supply role specific metrics.
    """
    pass

  def _add_role_specific_webserver_metrics_cdh6(self,
                                                update,
                                                metrics,
                                                accessors):
    """
    A method that role adapters can override to supply role specific metrics.
    """
    pass

class MasterAdapter(_HBaseAdapter):

  def __init__(self, safety_valve):
    _HBaseAdapter.__init__(self,
                           "MASTER",
                           safety_valve,
                           'hbase_master_info_port')

  def get_metrics_sample(self, version):
    return open(os.path.join(os.path.dirname(__file__), "../samples/hbase_master_cdh5"), 'r').read()

class RegionServerAdapter(_HBaseAdapter):

  class RegionUpdate(object):
    """
    This models the RegionUpdate Avro record.
    """
    def __init__(self, region_id):
      namespace, table, region = region_id
      self._metrics = []
      self._namespace = namespace
      self._table = table
      self._region = region

    def add_metric(self, metricid, value):
      if metricid is None:
        raise ValueError()
      if value is None:
        raise ValueError()
      self._metrics.append((metricid, value))
      return self

    def __str__(self):
      sio = StringIO()
      sio.write("RegionUpdate(%s, %s, %s):" % \
          (self._namespace, self._table, self._region))
      for metricid, val in self._metrics:
        sio.write("\n   %-10d : %s" % (metricid, val))
      return sio.getvalue()

    def get_region_name(self):
      return self._region

    def to_avro_dict(self):
      """Return a Python dictionary that maps to the Avro struct"""
      return dict(namespaceName=self._namespace,
                  tableName=self._table,
                  regionName=self._region,
                  metrics=[ dict(id=k, value=Firehoses.normalize_metric_value(v))
                            for (k, v) in self._metrics ])

  class ReplicationPeerUpdate(object):
    """
    This models the RegionServerRepplicationPeerUpdate Avro record.
    """
    def __init__(self, peer_id):
      self._metrics = []
      self._peer_id = peer_id

    def add_metric(self, metricid, value):
      if metricid is None:
        raise ValueError()
      if value is None:
        raise ValueError()
      self._metrics.append((metricid, value))
      return self

    def __str__(self):
      sio = StringIO()
      sio.write("RegionServerReplicationPeerUpdate(%s):" % (self._peer_id,))
      for metricid, val in self._metrics:
        sio.write("\n   %-10d : %s" % (metricid, val))
      return sio.getvalue()

    def to_avro_dict(self):
      """Return a Python dictionary that maps to the Avro struct"""
      return dict(peerId=self._peer_id,
                  metrics=[ dict(id=k, value=Firehoses.normalize_metric_value(v))
                            for (k, v) in self._metrics ])

  def __init__(self, safety_valve):
    _HBaseAdapter.__init__(self,
                           "REGIONSERVER",
                           safety_valve,
                           'hbase_regionserver_info_port')

  def _add_role_specific_webserver_metrics_cdh5(self,
                                                update,
                                                metrics,
                                                accessors):
    if self._is_feature_enabled(update, 'send_region_updates'):
      self._add_region_metrics(update, metrics, accessors, 'cdh5')
    if self._is_feature_enabled(update, 'send_replication_metrics'):
      self._add_replication_metrics(update, metrics, accessors, 'cdh5')

  def _add_role_specific_webserver_metrics_cdh6(self,
                                                update,
                                                metrics,
                                                accessors):
    if self._is_feature_enabled(update, 'send_region_updates'):
      self._add_region_metrics(update, metrics, accessors, 'cdh6')
    if self._is_feature_enabled(update, 'send_replication_metrics'):
      self._add_replication_metrics(update, metrics, accessors, 'cdh6')

  def _add_region_metrics(self, update, metrics, accessors, cdh_version):
    metric_accessors = accessors.get("HREGION", {}).get(cdh_version, {})
    bean_dict = getattr(metrics, 'regions', None)
    if bean_dict is None:
      return
    regions_to_metrics = self._extract_regions(bean_dict)
    for region_id, metrics in regions_to_metrics.iteritems():
      region_update = RegionServerAdapter.RegionUpdate(region_id)
      for metric_id, accessor in metric_accessors.iteritems():
        if accessor.metric not in metrics:
          continue
        region_update.add_metric(metric_id, metrics[accessor.metric])
      update.add_region_update(region_update)

  # Extracts the per-region metrics from the flat map of the bean entries.
  #
  # The input looks like:
  # { ...
  #   "namespace_default_table_TestTable_region_a888f84e48a18e00d0d8dfb0f4314498_metric_get_99th_percentile" : 0.0,
  #   "namespace_default_table_TestTable_region_a888f84e48a18e00d0d8dfb0f4314498_metric_incrementCount" : 0,
  #   "namespace_default_table_TestTable_region_30ff6264bdae6ee0917a90701c0dfff0_metric_deleteCount" : 0,
  #   "namespace_default_table_TestTable_region_b7c4fd0cf3e9950ca284470869d5969d_metric_get_num_ops" : 0,
  #   ...
  # }
  # The output is a map of region identifiers to maps of metrics:
  # {
  #    (default, TestTable, a888f84e48a18e00d0d8dfb0f4314498) : ( 'get_99th_percentile' : 0.0,
  #                                                               'incrementCount'      : 0 ),
  #    (default, TestTable, 30ff6264bdae6ee0917a90701c0dfff0) : ( 'deleteCount'         : 0 )
  #    ...
  # }
  def _extract_regions(self, bean_dict):
    regions_to_metrics = {}
    for name, val in bean_dict.iteritems():
      m = REGION_METRIC_PATTERN.match(name)
      if m is None:
        continue
      namespace, table, region, metric = m.groups()
      region_identifier = (namespace, table, region)
      if region_identifier not in regions_to_metrics:
        regions_to_metrics[region_identifier] = {}
      regions_to_metrics[region_identifier][metric] = val
    return regions_to_metrics

  def _add_replication_metrics(self, update, metrics, accessors, cdh_version):
    metric_accessors = \
      accessors.get("HBASE_REGIONSERVER_REPLICATION_PEER", {}).get(cdh_version, {})
    bean_dict = getattr(metrics, 'replication', None)
    if bean_dict is None:
      return
    peers_to_metrics = self._extract_regionserver_replication_peer(bean_dict)
    for peer_id, metrics in peers_to_metrics.iteritems():
      peer_update = RegionServerAdapter.ReplicationPeerUpdate(peer_id)
      for metric_id, accessor in metric_accessors.iteritems():
        if accessor.metric not in metrics:
          continue
        peer_update.add_metric(metric_id, metrics[accessor.metric])
      update.add_regionserver_replication_peer_update(peer_update)

  def _extract_regionserver_replication_peer(self, bean_dict):
    peers_to_metrics = {}
    for name, val in bean_dict.iteritems():
      m = REGIONSERVER_REPLICATION_METRIC_PATTERN.match(name)
      if m is None:
        continue
      peer_id, metric = m.groups()
      if peer_id not in peers_to_metrics:
        peers_to_metrics[peer_id] = {}
      peers_to_metrics[peer_id][metric] = val
    return peers_to_metrics

  def get_metrics_sample(self, version):
    return open(os.path.join(os.path.dirname(__file__), "../samples/hbase_regionserver_cdh5"), 'r').read()
