# Copyright (c) 2015 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import os

from cmf.monitor import constants
from cmf.monitor.generic.hadoop_adapter import HadoopAdapter
from cmf.monitor.generic.service_defined_metrics import \
  ServiceDefinedMetricsJsonJmxBeansWithDefault
from cmf.monitor.generic.utils import SplittingSourceProcessor, \
  JsonJmxBeanMetricsExtractor, visit_json
from cmf.monitor.jmx_beans import JsonParserJmx
from cmf.monitor.jmx_beans import NAMENODE_CLIENT_RPC_PORT_NAME
from cmf.monitor.jmx_beans import NAMENODE_SERVICE_RPC_PORT_NAME

class NameNodeAdapter(HadoopAdapter):
  _DFS_NAME_DIRS = 'dfs_name_dirs'
  _DFS_HTTP_PORT = 'dfs_http_port'
  _NAMENODE_CLIENT_RPC_PORT = 'namenode_port'
  # Added in CM 4.6.2
  # This is the service RPC port -- the port to which HDFS service members talk.
  # This may not be set.
  _NAMENODE_SERVICE_RPC_PORT = 'dfs_namenode_servicerpc_address'
  # Added in CM 5.1
  _DFS_HTTPS_PORT = 'dfs_https_port'

  def __init__(self, safety_valve):
    HadoopAdapter.__init__(
        self,
        "HDFS",
        "NAMENODE",
        safety_valve,
        http_port_key=NameNodeAdapter._DFS_HTTP_PORT,
        https_port_key=NameNodeAdapter._DFS_HTTPS_PORT)

  def _get_monitored_directories_lists_conf_keys(self):
    keys = HadoopAdapter._get_monitored_directories_lists_conf_keys(self)
    keys.append(NameNodeAdapter._DFS_NAME_DIRS)
    return keys

  def parse_metrics_from_url(self, conf, json):
    # Build the metric lookup context. This always needs to contain the client
    # RPC port and may contain the service RPC port if it is set.
    metric_lookup_context = dict()

    client_rpc_port = conf.getint(
        self.section, NameNodeAdapter._NAMENODE_CLIENT_RPC_PORT)
    if client_rpc_port is None:
      raise Exception("%s entry missing from %s" %
                      (NameNodeAdapter._NAMENODE_CLIENT_RPC_PORT,
                       os.path.abspath(conf.path)))
    metric_lookup_context[NAMENODE_CLIENT_RPC_PORT_NAME] = client_rpc_port

    service_rpc_port = conf.getint_with_default(
        self.section, NameNodeAdapter._NAMENODE_SERVICE_RPC_PORT, None)
    if service_rpc_port is not None:
        metric_lookup_context[NAMENODE_SERVICE_RPC_PORT_NAME] = service_rpc_port

    return JsonParserJmx(json, self.role_type, metric_lookup_context)

  def get_metrics_sample(self, version):
    if version == constants.SERVICE_VERSION_CDH5:
      return open(os.path.join(os.path.dirname(__file__), "../samples/hdfs_nn_cdh5"), 'r').read()
    elif version == constants.SERVICE_VERSION_CDH6:
      return open(os.path.join(os.path.dirname(__file__), "../samples/hdfs_nn_cdh6"), 'r').read()
    else:
      return None

class SecondaryNameNodeAdapter(HadoopAdapter):
  _DFS_SECONDARY_HTTP_PORT = 'dfs_secondary_http_port'
  # Added in CM 4.1
  _FS_CHECKPOINT_DIR_LIST = 'fs_checkpoint_dir_list'
  # Added in CM 5.1
  _DFS_SECONDARY_HTTPS_PORT = 'dfs_secondary_https_port'

  def __init__(self, safety_valve):
    HadoopAdapter.__init__(
        self,
        "HDFS",
        "SECONDARYNAMENODE",
        safety_valve,
        http_port_key=SecondaryNameNodeAdapter._DFS_SECONDARY_HTTP_PORT,
        https_port_key=SecondaryNameNodeAdapter._DFS_SECONDARY_HTTPS_PORT)

  def _get_monitored_directories_lists_conf_keys(self):
    keys = HadoopAdapter._get_monitored_directories_lists_conf_keys(self)
    keys.append(SecondaryNameNodeAdapter._FS_CHECKPOINT_DIR_LIST)
    return keys

  def get_metrics_sample(self, version):
    if version == constants.SERVICE_VERSION_CDH5:
      return open(os.path.join(os.path.dirname(__file__), "../samples/hdfs_snn_cdh5"), 'r').read()
    elif version == constants.SERVICE_VERSION_CDH6:
      print "xiaobin"
      return open(os.path.join(os.path.dirname(__file__), "../samples/hdfs_snn_cdh6"), 'r').read()
    else:
      return None

class JournalNodeAdapter(HadoopAdapter):
  _DFS_JOURNALNODE_HTTP_PORT = 'dfs_journalnode_http_port'
  _DFS_JOURNALNODE_EDITS_DIR = 'dfs_journalnode_edits_dir'
  # Added in CM 5.1
  _DFS_JOURNALNODE_HTTPS_PORT = 'dfs_journalnode_https_port'

  def __init__(self, safety_valve):
    HadoopAdapter.__init__(
        self,
        "HDFS",
        "JOURNALNODE",
        safety_valve,
        http_port_key=JournalNodeAdapter._DFS_JOURNALNODE_HTTP_PORT,
        https_port_key=JournalNodeAdapter._DFS_JOURNALNODE_HTTPS_PORT)

  def _get_monitored_directories_conf_keys(self):
    keys = HadoopAdapter._get_monitored_directories_conf_keys(self)
    keys.append(JournalNodeAdapter._DFS_JOURNALNODE_EDITS_DIR)
    return keys

  def get_metrics_sample(self, version):
    if version == constants.SERVICE_VERSION_CDH5:
      return open(os.path.join(os.path.dirname(__file__), "../samples/hdfs_jn_cdh5"), 'r').read()
    elif version == constants.SERVICE_VERSION_CDH6:
      return open(os.path.join(os.path.dirname(__file__), "../samples/hdfs_jn_cdh6"), 'r').read()
    else:
      return None

class DataNodeAdapter(HadoopAdapter):
  _DFS_DATA_DIRS = 'dfs_data_dirs'
  _DFS_DATANODE_HTTP_PORT = 'dfs_datanode_http_port'
  # Added in CM 5.1
  _DFS_DATANODE_HTTPS_PORT = 'dfs_datanode_https_port'

  def __init__(self, safety_valve):
    HadoopAdapter.__init__(
        self,
        "HDFS",
        "DATANODE",
        safety_valve,
        http_port_key=DataNodeAdapter._DFS_DATANODE_HTTP_PORT,
        https_port_key=DataNodeAdapter._DFS_DATANODE_HTTPS_PORT)

  def _get_monitored_directories_lists_conf_keys(self):
    keys = HadoopAdapter._get_monitored_directories_lists_conf_keys(self)
    keys.append(DataNodeAdapter._DFS_DATA_DIRS)
    return keys

  def get_metrics_sample(self, version):
    if version == constants.SERVICE_VERSION_CDH5:
      return open(os.path.join(os.path.dirname(__file__), "../samples/hdfs_dn_cdh5"), 'r').read()
    elif version == constants.SERVICE_VERSION_CDH6:
      return open(os.path.join(os.path.dirname(__file__), "../samples/hdfs_dn_cdh6"), 'r').read()
    else:
      return None

class DssdDataNodeAdapter(HadoopAdapter):
  _DFS_DSSDDATANODE_HTTP_PORT = 'dfs_datanode_http_port'
  _DFS_DSSDDATANODE_HTTPS_PORT = 'dfs_datanode_https_port'

  def __init__(self, safety_valve):
    HadoopAdapter.__init__(
        self,
        "HDFS",
        "DSSDDATANODE",
        safety_valve,
        http_port_key=DssdDataNodeAdapter._DFS_DSSDDATANODE_HTTP_PORT,
        https_port_key=DssdDataNodeAdapter._DFS_DSSDDATANODE_HTTPS_PORT)
    self._metrics = None

  def parse_metrics_from_url(self, conf, json):
    if json is None:
      raise Exception("a json sample is required")
    if self._metrics is None:
      raise Exception("No metrics have been loaded!")

    role_extractor = JsonJmxBeanMetricsExtractor(
      self._metrics.get_entity_beans(self._role_type))
    visit_json(json, [role_extractor])
    return role_extractor.metrics

  def read_service_defined_metrics(self, path):
    if path is None:
      raise Exception("A path is required")
    self._metrics = \
      ServiceDefinedMetricsJsonJmxBeansWithDefault(path,
          SplittingSourceProcessor('::'))

  def get_metrics_sample(self, version):
    if version == constants.SERVICE_VERSION_CDH5:
      return open(os.path.join(os.path.dirname(__file__), "../samples/hdfs_dssddn_cdh5"), 'r').read()
    else:
      return None

  def add_webserver_metrics(self, version, update, metrics, accessors):
    for metricid, value in metrics.iteritems():
      update.add_metric(metricid, value)

DSSDDATANODE_SERVICE_METRICS = r"""
{
  "DSSDDATANODE" : {
    "70000" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::BlockChecksumOpAvgTime"
    },
    "70001" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::BlockChecksumOpNumOps"
    },
    "70002" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::BlockReportsAvgTime"
    },
    "70003" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::BlockReportsNumOps"
    },
    "70004" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::BlockVerificationFailures"
    },
    "70005" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::BlocksCached"
    },
    "70006" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::BlocksGetLocalPathInfo"
    },
    "70007" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::BlocksRead"
    },
    "70008" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::BlocksRemoved"
    },
    "70009" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::BlocksReplicated"
    },
    "70010" : {
      "source" : "Cloudera Manager Service Monitor"
    },
    "70011" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::BlocksUncached"
    },
    "70012" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::BlocksVerified"
    },
    "70013" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::BlocksWritten"
    },
    "70014" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::BytesRead"
    },
    "70015" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::BytesWritten"
    },
    "70016" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::CacheReportsAvgTime"
    },
    "70017" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::CacheReportsNumOps"
    },
    "70018" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::CopyBlockOpAvgTime"
    },
    "70019" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::CopyBlockOpNumOps"
    },
    "70020" : {
      "source" : "Cloudera Manager Service Monitor"
    },
    "70021" : {
      "source" : "Cloudera Manager Service Monitor"
    },
    "70022" : {
      "source" : "Cloudera Manager Service Monitor"
    },
    "70023" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=RpcDetailedActivityForPort\\d+::DeleteBlockPoolAvgTime::0"
    },
    "70024" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=RpcDetailedActivityForPort\\d+::DeleteBlockPoolNumOps::0"
    },
    "70025" : {
      "source" : "Cloudera Manager Service Monitor"
    },
    "70026" : {
      "source" : "Cloudera Manager Service Monitor"
    },
    "70027" : {
      "source" : "Cloudera Manager Service Monitor"
    },
    "70028" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=FloodObjectOpstats::CropOpTotalCount"
    },
    "70029" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=HeartbeatOpstats::UpdateHeartbeatInfoOpFailureCummulativeDuration"
    },
    "70030" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=HeartbeatOpstats::UpdateHeartbeatInfoOpFailureCount"
    },
    "70031" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=FloodObjectOpstats::ReadOpFailureCummulativeDuration"
    },
    "70032" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=FloodObjectOpstats::ReadOpFailureCount"
    },
    "70033" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=FloodObjectOpstats::WriteOpFailureCummulativeDuration"
    },
    "70034" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=FloodObjectOpstats::WriteOpFailureCount"
    },
    "70035" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=FloodObjectOpstats::ReadOpTotalCount"
    },
    "70036" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=FloodObjectOpstats::ReadOpSuccessCummulativeDuration"
    },
    "70037" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=FloodObjectOpstats::ReadOpSuccessCount"
    },
    "70038" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=FloodObjectOpstats::WriteOpSuccessCummulativeDuration"
    },
    "70039" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=FloodObjectOpstats::WriteOpSuccessCount"
    },
    "70040" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=FloodObjectOpstats::WriteOpTotalCount"
    },
    "70041" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=FloodDatasetOpstats::FlushBlockOpTotalCount"
    },
    "70042" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=FloodDatasetOpstats::GetAvailableOpTotalCount"
    },
    "70043" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=FloodDatasetOpstats::GetBlockOpTotalCount"
    },
    "70044" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=FloodDatasetOpstats::GetBlockPoolListOpTotalCount"
    },
    "70045" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=FloodDatasetOpstats::GetSizeOpTotalCount"
    },
    "70046" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=FloodDatasetOpstats::GetStorageReportOpTotalCount"
    },
    "70047" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=FloodDatasetOpstats::GetUsedOpTotalCount"
    },
    "70048" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=HeartbeatOpstats::UpdateHeartbeatInfoOpTotalCount"
    },
    "70049" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=PageCacheOpstats::ReadOpFailureCummulativeDuration"
    },
    "70050" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=PageCacheOpstats::ReadOpFailureCount"
    },
    "70051" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=PageCacheOpstats::WriteOpFailureCummulativeDuration"
    },
    "70052" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=PageCacheOpstats::WriteOpFailureCount"
    },
    "70053" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=PageCacheOpstats::PageEvictedTotalCount"
    },
    "70054" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=PageCacheOpstats::ObjReadHitCount"
    },
    "70055" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=PageCacheOpstats::ObjReadIssueCount"
    },
    "70056" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=PageCacheOpstats::ReadOpSuccessCummulativeDuration"
    },
    "70057" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=PageCacheOpstats::ReadOpSuccessCount"
    },
    "70058" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=PageCacheOpstats::WriteOpSuccessCummulativeDuration"
    },
    "70059" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=PageCacheOpstats::WriteOpSuccessCount"
    },
    "70060" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=PageCacheOpstats::ObjWriteDirtyCacheCount"
    },
    "70061" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=PageCacheOpstats::ObjWriteDirtyInitCount"
    },
    "70062" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=PageCacheOpstats::WriteOpTotalCount"
    },
    "70063" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=PageCacheOpstats::ObjWriteRMWCount"
    },
    "70064" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=FloodObjectOpstats::GetSizeOpTotalCount"
    },
    "70065" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=HeartbeatOpstats::UpdateHeartbeatInfoOpSuccessCummulativeDuration"
    },
    "70066" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=HeartbeatOpstats::UpdateHeartbeatInfoOpSuccessCount"
    },
    "70067" : {
      "counter" : true,
      "source" : "REGEX$$DSSD:service=FloodDS,name=FloodObjectOpstats::TrimOpTotalCount"
    },
    "70068" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::FlushNanosAvgTime"
    },
    "70069" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::FlushNanosNumOps"
    },
    "70070" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::FsyncNanosAvgTime"
    },
    "70071" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::FsyncNanosNumOps"
    },
    "70072" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::FsyncCount"
    },
    "70073" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=RpcDetailedActivityForPort\\d+::GetBlockLocalPathInfoAvgTime::0"
    },
    "70074" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=RpcDetailedActivityForPort\\d+::GetBlockLocalPathInfoNumOps::0"
    },
    "70075" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=RpcDetailedActivityForPort\\d+::GetHdfsBlocksMetadataAvgTime::0"
    },
    "70076" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=RpcDetailedActivityForPort\\d+::GetHdfsBlocksMetadataNumOps::0"
    },
    "70077" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=RpcDetailedActivityForPort\\d+::GetReplicaVisibleLengthAvgTime::0"
    },
    "70078" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=RpcDetailedActivityForPort\\d+::GetReplicaVisibleLengthNumOps::0"
    },
    "70079" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::HeartbeatsAvgTime"
    },
    "70080" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::HeartbeatsNumOps"
    },
    "70081" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=RpcDetailedActivityForPort\\d+::InitReplicaRecoveryAvgTime::0"
    },
    "70082" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=RpcDetailedActivityForPort\\d+::InitReplicaRecoveryNumOps::0"
    },
    "70083" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=JvmMetrics::ThreadsBlocked"
    },
    "70084" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=JvmMetrics::GcCount"
    },
    "70085" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=JvmMetrics::GcTimeMillis"
    },
    "70086" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=JvmMetrics::MemHeapCommittedM"
    },
    "70087" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=JvmMetrics::MemHeapUsedM"
    },
    "70088" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=JvmMetrics::MemMaxM"
    },
    "70089" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=JvmMetrics::ThreadsNew"
    },
    "70090" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=JvmMetrics::MemNonHeapCommittedM"
    },
    "70091" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=JvmMetrics::MemNonHeapUsedM"
    },
    "70092" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=JvmMetrics::ThreadsRunnable"
    },
    "70093" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=JvmMetrics::ThreadsTerminated"
    },
    "70094" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=JvmMetrics::ThreadsTimedWaiting"
    },
    "70095" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=JvmMetrics::ThreadsWaiting"
    },
    "70096" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=JvmMetrics::LogError"
    },
    "70097" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=JvmMetrics::LogFatal"
    },
    "70098" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=JvmMetrics::LogInfo"
    },
    "70099" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=JvmMetrics::LogWarn"
    },
    "70100" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=UgiMetrics::LoginFailureAvgTime"
    },
    "70101" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=UgiMetrics::LoginFailureNumOps"
    },
    "70102" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=UgiMetrics::LoginSuccessAvgTime"
    },
    "70103" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=UgiMetrics::LoginSuccessNumOps"
    },
    "70104" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=MetricsSystem,sub=Stats::DroppedPubAll"
    },
    "70105" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=MetricsSystem,sub=Stats::NumActiveSinks"
    },
    "70106" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=MetricsSystem,sub=Stats::NumActiveSources"
    },
    "70107" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=MetricsSystem,sub=Stats::NumAllSinks"
    },
    "70108" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=MetricsSystem,sub=Stats::NumAllSources"
    },
    "70109" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=MetricsSystem,sub=Stats::PublishAvgTime"
    },
    "70110" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=MetricsSystem,sub=Stats::PublishNumOps"
    },
    "70111" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=MetricsSystem,sub=Stats::SnapshotAvgTime"
    },
    "70112" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=MetricsSystem,sub=Stats::SnapshotNumOps"
    },
    "70113" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::PacketAckRoundTripTimeNanosAvgTime"
    },
    "70114" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::PacketAckRoundTripTimeNanosNumOps"
    },
    "70115" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::ReadBlockOpAvgTime"
    },
    "70116" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::ReadBlockOpNumOps"
    },
    "70117" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::ReadsFromLocalClient"
    },
    "70118" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::ReadsFromRemoteClient"
    },
    "70119" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=RpcDetailedActivityForPort\\d+::RefreshNamenodesAvgTime::0"
    },
    "70120" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=RpcDetailedActivityForPort\\d+::RefreshNamenodesNumOps::0"
    },
    "70121" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::ReplaceBlockOpAvgTime"
    },
    "70122" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::ReplaceBlockOpNumOps"
    },
    "70123" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=RpcActivityForPort\\d+::RpcAuthenticationFailures::0"
    },
    "70124" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=RpcActivityForPort\\d+::RpcAuthenticationSuccesses::0"
    },
    "70125" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=RpcActivityForPort\\d+::RpcAuthorizationFailures::0"
    },
    "70126" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=RpcActivityForPort\\d+::RpcAuthorizationSuccesses::0"
    },
    "70127" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=RpcActivityForPort\\d+::CallQueueLength::0"
    },
    "70128" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=RpcActivityForPort\\d+::NumOpenConnections::0"
    },
    "70129" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=RpcActivityForPort\\d+::RpcProcessingTimeAvgTime::0"
    },
    "70130" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=RpcActivityForPort\\d+::RpcProcessingTimeNumOps::0"
    },
    "70131" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=RpcActivityForPort\\d+::RpcQueueTimeAvgTime::0"
    },
    "70132" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=RpcActivityForPort\\d+::RpcQueueTimeNumOps::0"
    },
    "70133" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=RpcActivityForPort\\d+::ReceivedBytes::0"
    },
    "70134" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=RpcActivityForPort\\d+::SentBytes::0"
    },
    "70135" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::SendDataPacketBlockedOnNetworkNanosAvgTime"
    },
    "70136" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::SendDataPacketBlockedOnNetworkNanosNumOps"
    },
    "70137" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::SendDataPacketTransferNanosAvgTime"
    },
    "70138" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::SendDataPacketTransferNanosNumOps"
    },
    "70139" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=RpcDetailedActivityForPort\\d+::UpdateReplicaUnderRecoveryAvgTime::0"
    },
    "70140" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=RpcDetailedActivityForPort\\d+::UpdateReplicaUnderRecoveryNumOps::0"
    },
    "70141" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::WriteBlockOpAvgTime"
    },
    "70142" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::WriteBlockOpNumOps"
    },
    "70143" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::WritesFromLocalClient"
    },
    "70144" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeActivity::WritesFromRemoteClient"
    },
    "70145" : {
      "source" : "REGEX$$Hadoop:service=DataNode,name=DataNodeInfo::XceiverCount"
    },
    "70146" : {
      "source" : "REGEX$$DSSD:service=FloodDS,name=FloodObjectOpstats::AsyncReadOpTotalCount"
    },
    "70147" : {
      "source" : "REGEX$$DSSD:service=FloodDS,name=FloodObjectOpstats::AsyncWriteOpTotalCount"
    }
  }
}
"""
