# Copyright (c) 2015 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging
import os

from cmf.monitor.generic.adapter import Adapter
from cmf.monitor.generic.service_defined_metrics import SourceProcessor, \
  ServiceDefinedMetrics
from cmf.monitor.generic.utils import visit_json
from cmf.monitor.generic.utils import SimpleMetricsExtractorWithDefaults

LOG = logging.getLogger('HiveAdapters')

class AbstractHiveAdapter(Adapter):
  """
  Base class for all Hive role adapters. This exists to provide utility
  methods and to ensure consistency in the handling of common metrics. The hive
  adapters support collecting from a sample file as the different hive roles do
  not expose metrics through a web server.
  """

  class HiveMetricsProcessor(SourceProcessor):
    """
    A class that parses a hive metric source and returns a MetricSource object.
    This is needed as we want to add default values for the lazily initialized
    api metrics for which a default value of '0' makes sense.
    """
    def process_source(self, source):
      if source is None:
        raise Exception("source cannot be None")
      keys = source.split('::')
      default_value = self._determine_default_value(source)
      return SimpleMetricsExtractorWithDefaults.MetricSource(keys,
                                                             default_value)

    def _determine_default_value(self, source):
      if source is None:
        raise Exception("source cannot be None")
      if source.find('api_') >= 0:
        return 0.0
      return None

  _HIVE_ROLE_METRICS_SAMPLE_FILE_KEY = "hive.service.metrics.file.location"

  def __init__(self, role_type, safety_valve):
    Adapter.__init__(self, "HIVE", role_type, safety_valve)
    self._metrics = None

  def get_metrics_file(self, conf):
    if conf is None:
      raise Exception("A configuration is required!")

    try:
      sample_metric_file = conf.get(
        self.section,
        AbstractHiveAdapter._HIVE_ROLE_METRICS_SAMPLE_FILE_KEY)

      return sample_metric_file
    except:
      LOG.exception("Failed to retrieve %s from monitoring configuration file" %
                    AbstractHiveAdapter._HIVE_ROLE_METRICS_SAMPLE_FILE_KEY)

  def read_service_defined_metrics(self, path):
    if path is None:
      raise Exception("A path is required")
    self._metrics = ServiceDefinedMetrics(
      path,
      AbstractHiveAdapter.HiveMetricsProcessor())

  def parse_metrics_from_file(self, conf, json):
    if json is None:
      raise Exception("a json sample is required")
    if self._metrics is None:
      raise Exception("No metrics have been loaded!")

    role_extractor = SimpleMetricsExtractorWithDefaults(
      self._metrics.get_sources(self._role_type))
    visit_json(json, [role_extractor])
    return role_extractor.metrics

  def add_sample_file_metrics(self, version, update, metrics, accessors):
    for metric_id, value in metrics.iteritems():
      update.add_metric(metric_id, value)

class HiveMetastoreAdapter(AbstractHiveAdapter):
  def __init__(self, safety_valve):
    AbstractHiveAdapter.__init__(self, 'HIVEMETASTORE', safety_valve)

  def get_metrics_sample(self, version):
    return open(os.path.join(os.path.dirname(__file__), "../samples/hive_hms"), 'r').read()

class HiveServer2Adapter(AbstractHiveAdapter):
  def __init__(self, safety_valve):
    AbstractHiveAdapter.__init__(self, 'HIVESERVER2', safety_valve)

  def get_metrics_sample(self, version):
    return open(os.path.join(os.path.dirname(__file__), "../samples/hive_hs2"), 'r').read()
