# Copyright (c) 2015 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging
import os
import time

from cmf.monitor import schema
from cmf.monitor.abstract_monitor import CollectionError
from cmf.monitor.constants import WEB_METRICS_COLLECTION_GOOD, \
  DEFAULT_MONITOR_LOG_RATE, WEB_METRICS_COLLECTION_COMMUNICATION_FAILURE
from cmf.monitor.generic import AbstractMetricCollector
from cmf.monitor.generic.adapter import Adapter
from cmf.monitor.generic.service_defined_metrics import ServiceDefinedMetrics
from cmf.monitor.generic.utils import SplittingSourceProcessor, visit_json, \
  SimpleMetricsExtractor
from cmf.monitor.generic.utils import VersionUtils, VersionType
from cmf.throttling_logger import ThrottlingLogger
from cmf.util.url import head_request_with_timeout

LOG = logging.getLogger('HueServerAdapter')
THROTTLED_LOG = ThrottlingLogger(LOG, DEFAULT_MONITOR_LOG_RATE)

class _HueIsAliveCollector(AbstractMetricCollector):
  """
  A collector that generates web-metric collection metrics from the hue is alive
  end-point.
  """

  def __init__(self, adapter):
    AbstractMetricCollector.__init__(self, adapter)
    if not isinstance(adapter, HueServerAdapter):
      raise Exception("Unsupported adapter type")

  def update_with_conf(self, conf):
    self._metrics_uri = self._adapter._get_is_alive_url(conf)

  def collect_and_parse(self, conf):
    start = time.time()
    try:
      self._call_is_alive(self._metrics_uri)
      result = WEB_METRICS_COLLECTION_GOOD
    except Exception, ex:
      THROTTLED_LOG.exception("Error calling is alive at '%s'" %
                              (self._metrics_uri,))
      result = CollectionError(WEB_METRICS_COLLECTION_COMMUNICATION_FAILURE)
    now = time.time()
    self._metrics_tuple = result, now, int((now - start) * 1000)

  def _add_metrics(self, service_version, update, collected_metrics, duration):
    update.add_metric(schema.WEB_METRICS_COLLECTION_DURATION, duration)
    if isinstance(collected_metrics, CollectionError):
      update.add_metric(schema.WEB_METRICS_COLLECTION_STATUS,
                        collected_metrics.code)
    else:
      update.add_metric(schema.WEB_METRICS_COLLECTION_STATUS,
                        collected_metrics)
    return True

  def is_ready_to_report(self, conf, pid):
    if pid is None:
      raise Exception("Pid cannot be None")

    metrics_url = self._adapter._get_is_alive_url(conf)
    if metrics_url is None:
      return AbstractMetricCollector._COLLECTOR_NOT_SUPPORTED
    try:
      self._call_is_alive(metrics_url, 0.5)
      return True
    except Exception, ex:
      pass
    return False

  def _call_is_alive(self, is_alive_url, timeout=None):
    """
    We need this function to make testing eaiser.
    """
    head_request_with_timeout(is_alive_url, timeout=timeout)

class HueServerAdapter(Adapter):
  """
  An adapter that collects and generates metrics for the HUE_SERVER. The adapter
  collects metrics from a sample file generated by hue. It also generates
  web-metric-collection metrics using an is-alive endpoint exposed by hue.
  Note that the web-metric-collection metrics generation is done as part of the
  sample file metric collection. The reason is that the hue is-alive end-point
  accepts HTTP HEAD requests and we can't use the regular web metric collector
  to generate these metics.
  """

  _HUE_SERVER_METRICS_SAMPLE_FILE_KEY = "location"
  _HUE_SERVER_HTTP_HOST_KEY = "http_host"
  _HUE_SERVER_HTTP_PORT_KEY = "http_port"
  _HUE_SERVER_SSL_ENABLED_KEY = "ssl_enable"
  _SERVICE_RELEASE = 'service_release'

  def __init__(self, safety_valve):
    Adapter.__init__(self, "HUE", "HUE_SERVER", safety_valve)
    self._metrics = None
    self._is_alive_collector = _HueIsAliveCollector(self)

  def read_service_defined_metrics(self, path):
    if path is None:
      raise Exception("A path is required!")
    self._metrics = ServiceDefinedMetrics(path, SplittingSourceProcessor('::'))

  def get_metrics_file(self, conf):
    if conf is None:
      raise Exception("A configuration is required!")

    try:
      return conf.get(
        self.section,
        HueServerAdapter._HUE_SERVER_METRICS_SAMPLE_FILE_KEY)
    except:
      LOG.exception("Failed to retrieve %s from monitoring configuration file" %
                    HueServerAdapter._HUE_SERVER_METRICS_SAMPLE_FILE_KEY)

  def _get_is_alive_url(self, conf):
    if conf is None:
      raise Exception("A configuration is required!")

    if not self._is_alive_supported(conf):
      return None

    try:
      host = conf.get(self.section, HueServerAdapter._HUE_SERVER_HTTP_HOST_KEY)
      if host is None:
        LOG.error("%s entry missing from monitoring configuration file" %
                  (HueServerAdapter._HUE_SERVER_HTTP_HOST_KEY, ))
        return None
      port = conf.getint(self.section,
                         HueServerAdapter._HUE_SERVER_HTTP_PORT_KEY)
      if port is None:
        LOG.error("%s entry missing from monitoring configuration file" %
                  (HueServerAdapter._HUE_SERVER_HTTP_PORT_KEY, ))
        return None
      if conf.getboolean_with_default(
          self.section,
          HueServerAdapter._HUE_SERVER_SSL_ENABLED_KEY,
          False):
        url_format = "https://%s:%s/desktop/debug/is_alive"
      else:
        url_format = "http://%s:%s/desktop/debug/is_alive"
      return url_format % (host, port)
    except:
      LOG.exception("Failed to read monitoring configuration file")
      return None

  def parse_metrics_from_file(self, conf, json):
    if json is None:
      raise Exception("a json sample is required")
    if self._metrics is None:
      raise Exception("No metrics have been loaded!")

    role_extractor = SimpleMetricsExtractor(
      self._metrics.get_sources(self._role_type))
    visit_json(json, [role_extractor])
    return role_extractor.metrics

  def add_sample_file_metrics(self, version, update, metrics, accessors):
    for metric_id, value in metrics.iteritems():
      update.add_metric(metric_id, value)

  def get_metrics_sample(self, version):
    return open(os.path.join(os.path.dirname(__file__), "../samples/hue_server"), 'r').read()

  def _is_alive_supported(self, conf):
    release = self.get_release(conf)
    if release is None:
      return False
    # Version detection logic should be handled in the server (tracked in: OPSAPS-36429)
    if VersionUtils.less_than(release, "5.3", VersionType.MINOR):
      return False
    return True

  def get_adapter_specific_collectors(self):
    return [self._is_alive_collector]
