# Copyright (c) 2016 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

try:
  from cStringIO import StringIO
except:
  from StringIO import StringIO

import logging
import re

from cmf.monitor import constants
from cmf.monitor.entity_type_update import EntityTypeUpdate
from cmf.monitor.entity_type_update import EntityTypeUpdateEntry
from cmf.monitor.generic.adapter import Adapter
from cmf.monitor.generic.service_defined_metrics import ServiceDefinedMetrics
from cmf.monitor.generic.utils import extract_metric_from_json
from cmf.monitor.generic.utils import JsonVisitor, visit_json
from cmf.monitor.generic.utils import SimpleMetricsExtractor
from cmf.monitor.generic.utils import SplittingSourceProcessor
from cmf.monitor.generic.utils import VersionUtils, VersionType
from cmf.monitor.impala_daemon_update import ImpalaDaemonUpdate
from cmf.monitor.impalad.query_monitor import QueryMonitorMetricCollector
from cmf.throttling_logger import ThrottlingLogger

LOG = logging.getLogger('ImpalaAdapters')
THROTTLED_LOG = ThrottlingLogger(LOG, 60 * 60) # Once an hour.

_IMPALA_ALIASES = {
   "impala-server.io-mgr.num-buffers": "impala-server.io.mgr.num.buffers",
   "impala-server.io-mgr.num-open-files": "impala-server.io.mgr.num.open-files",
   "impala-server.mem-pool.total-bytes": "impala-server.mem.pool.total-bytes",
   "impala-server.num-fragments": "impala-server.num.fragments",
   "impala-server.num-queries": "impala-server.num.queries",
   "impala-server.scan-ranges.num-missing-volume-id": "impala-server.scan_ranges.num-missing-volume-id",
   "impala-server.scan-ranges.total": "impala-server.scan.ranges.total",
   "simple-scheduler.local-assignments.total": "simple-scheduler.local.assignments.total"
}

# This function uses the input accessors map and adds all available metrics from
# the input metrics object to the input update. The arguments are as follows:
#
# update: a RoleUpdate
# metrics: a JsonParser object built from a JSON object produced by Impala
# accessors: a map of metric accessors built by build_metric_accessors
# entity: an entity type string from a metric source, e.g. 'IMPALA'
# version: a version string from a metric source, e.g. 'cdh4'
def add_metrics_impala(update, metrics, accessors, entity, version):
  if metrics is None or metrics.json is None:
    return
  metric_accesors = accessors.get(entity, {}).get(version, {})
  for metric_id, accessor in metric_accesors.iteritems():
    value = None
    if accessor.metric in metrics.json:
      value = metrics.json.get(accessor.metric)
    # Impala adjusted the names of some of their metrics just before CM 4.5.
    # To be nice, we'll try looking up metrcs under their old names if they
    # are not present under their current names.
    if value == None and accessor.metric in _IMPALA_ALIASES:
      value = metrics.json.get(_IMPALA_ALIASES[accessor.metric])
    # If this is a stats metric, we have a further lookup to do. Value here is
    # a dictionary keyed on the individual stats. See the sample metrics for
    # an example.
    if value is not None and accessor.stat is not None:
      value = value.get(accessor.stat, None)
    if value is not None:
      # The Impala JSON exposes its numbers as strings, e.g. "1" instead of 1.
      # This means we have to coerce its metric values into number formats. We
      # use float below because all our metrics are ultimately just stored as
      # doubles anyway. We will undo this change when Impala adjusts their JSON
      # printing code, but we are waiting for an opportunity for a compatability
      # break to do so. See OPSAPS-11201 for more detail.
      update.add_metric(metric_id, float(value))

class _ImpalaAdapter(Adapter):
  """
  Base class for all Impala service role adapters. This exists to provide utility
  methods and to ensure consistency in the handling of common metrics. Role dir
  monitors rely on MONITORED_DIRS_WITH_TYPE driven by config.
  """
  _SERVICETYPE = "IMPALA"

  _COLLECT_INTERVAL = 'collect_interval'
  _USE_SSL = 'use_ssl'

  def __init__(self, role_type, safety_valve, port_key):
    Adapter.__init__(self, _ImpalaAdapter._SERVICETYPE, role_type,
                     safety_valve)
    self._port_key = port_key
    self._service_defined_metrics = None

  def _use_new_metrics_url(self, release):
    try:
      # Version detection logic should be handled in the server (tracked in: OPSAPS-36429)
      return release is not None and VersionUtils.at_least(release, "5.4", VersionType.MINOR)
    except:
      LOG.warn("Malformed or unknown release string: %s" % release)
      return False

  def _compute_metrics_url(self, host, port, use_ssl, release):
    if host is None or port is None:
      return None
    protocol = "http"
    if use_ssl:
      protocol = "https"
    url = "%s://%s:%s/jsonmetrics" % (protocol, host, port)
    if self._use_new_metrics_url(release):
      url += "?json"
    return url

  def get_metrics_url(self, conf):
    try:
      host = conf.get(self.section, _ImpalaAdapter._HOST)
      web_port = conf.get(self.section, self._port_key)
      use_ssl = conf.getboolean_with_default(self.section, _ImpalaAdapter._USE_SSL, False)
      release = conf.get_with_default(self.section, _ImpalaAdapter._SERVICE_RELEASE, None)
    except ValueError, ex:
      self.record_error(ex)
      return False

    return self._compute_metrics_url(host, web_port, use_ssl, release)

  def read_service_defined_metrics(self, path):
    if path is None:
      raise Exception("A path to service-metrics.properties is required for service defined metrics!")
    self._service_defined_metrics = ServiceDefinedMetrics(path,
                                        SplittingSourceProcessor('::'))

  def parse_metrics_from_url(self, conf, json):
    '''
    This function would extract the service defined metrics from the json map,
    fetch additional entity type updates and return them along with the json map
    for the add_webserver_metrics to further extract metrics defined in
    schema.json.

    Notice the result would be used by impalad/query_monitor.py. The code there
    might need to be changed according to the return format.
    '''
    json_parser = None
    role_extractor = None
    entity_type_updates = None

    if json is None:
      raise Exception("A json response is required")
    json_parser = JsonParser(json)

    if self._service_defined_metrics:
      role_extractor = SimpleMetricsExtractor(
          self._service_defined_metrics.get_sources(self.role_type))
      visit_json(json, [role_extractor])

    entity_type_updates = self._get_entity_type_updates(conf, json)

    return dict(jsonParser = json_parser,
                roleExtractor = role_extractor,
                entityTypeUpdates = entity_type_updates)

  def add_webserver_metrics(self, service_version, update, metrics, accessors):
    json_parser = metrics['jsonParser']
    role_extractor = metrics['roleExtractor']
    entity_type_updates = metrics['entityTypeUpdates']

    self._set_additional_role_updates(json_parser, update)

    version = constants.VERSION_STRINGS.get(service_version, None)
    if version is None:
      LOG.warn("Unknown %s version: %s" % (self._role_type, version))
      return

    # Add schema.json defined metrics
    if json_parser:
      add_metrics_impala(update,
                        json_parser,
                        accessors,
                        self._role_type,
                        version)

    # Add service defined metrics
    if role_extractor and role_extractor.metrics:
      for metric_id, value in role_extractor.metrics.iteritems():
        update.add_metric(metric_id, value)
    else:
      THROTTLED_LOG.exception('Role_extractor for %s is not initialized correctly.' % self.role_type)

    # Add entity_type_updates
    if entity_type_updates:
      self._add_entity_type_updates(service_version, update, entity_type_updates, accessors)

  def _set_additional_role_updates(self, json_parser, update):
    '''
    Interface to enable child classes to set additional role updates in the
    role_update object.
    '''
    pass

  def _get_entity_type_updates(self, conf, json):
    '''
    Interface for Roles who need to further add entity type updates. This should
    generate the necessary entity_type_updates;
    '''
    return None

  def _add_entity_type_updates(self, service_version, update, entity_type_updates, accessors):
    '''
    Interface for Roles who need to further add entity type updates. This should
    add entity type updates to role updates.
    '''
    pass

class JsonParser(object):
  def __init__(self, json_dict):
    self.json = json_dict

class CatalogServerAdapter(_ImpalaAdapter):

  _ROLETYPE = "CATALOGSERVER"

  _WEBSERVER_PORT = 'catalogserver_webserver_port'
  _WEBSERVER_USERNAME = 'webserver_username'
  _WEBSERVER_PASSWORD = 'webserver_password'

  def __init__(self, safety_valve):
    _ImpalaAdapter.__init__(self,
                           CatalogServerAdapter._ROLETYPE,
                           safety_valve,
                           CatalogServerAdapter._WEBSERVER_PORT)

  def get_metrics_url_username_attr(self):
    return CatalogServerAdapter._WEBSERVER_USERNAME

  def get_metrics_url_password_attr(self):
    return CatalogServerAdapter._WEBSERVER_PASSWORD

  #
  # See the version hash in the metrics for when the sample was generated.
  #
  SAMPLE_CDH5 = r"""
{
   "catalog.version":"catalogd version 1.2.2-INTERNAL RELEASE (build 4161d79734b48bcbff5b47a7c9e09d206cc4b848)",
   "impala.thrift-server.CatalogService.connections-in-use":0,
   "impala.thrift-server.CatalogService.total-connections":0,
   "statestore-subscriber.connected":1,
   "statestore-subscriber.heartbeat-duration":{
      "count":14437,
      "last":0.000111741,
      "min":3.4552e-05,
      "max":0.00209998,
      "mean":8.22269e-05,
      "stddev":5.79026e-05
   },
   "statestore-subscriber.heartbeat-interval-time":{
      "count":14437,
      "last":0.50049,
      "min":0.499322,
      "max":0.61974,
      "mean":0.500526,
      "stddev":0.00101264
   },
   "statestore-subscriber.last-recovery-duration":0,
   "statestore-subscriber.last-recovery-time":"N/A",
   "statestore-subscriber.statestore.client-cache.clients-in-use":0,
   "statestore-subscriber.statestore.client-cache.total-clients":1,
   "statestore-susbcriber.registration-id":"ab4d20fd9e964412:eba41d77590acfb0",
   "tcmalloc.bytes-in-use":23051464,
   "tcmalloc.pageheap-free-bytes":38633472,
   "tcmalloc.pageheap-unmapped-bytes":0,
   "tcmalloc.physical-bytes-reserved":71663616,
   "tcmalloc.total-bytes-reserved":71663616
}
"""

class StateStoreAdapter(_ImpalaAdapter):

  _ROLETYPE = "STATESTORE"

  _WEBSERVER_PORT = 'statestore_webserver_port'
  _WEBSERVER_USERNAME = 'webserver_username'
  _WEBSERVER_PASSWORD = 'webserver_password'

  def __init__(self, safety_valve):
    _ImpalaAdapter.__init__(self,
                           StateStoreAdapter._ROLETYPE,
                           safety_valve,
                           StateStoreAdapter._WEBSERVER_PORT)

  def get_metrics_url_username_attr(self):
    return StateStoreAdapter._WEBSERVER_USERNAME

  def get_metrics_url_password_attr(self):
    return StateStoreAdapter._WEBSERVER_PASSWORD

class _PoolMetricsExtractor(JsonVisitor):
  """
  Responsible for extracting a EntityTypeUpdate object for the daemon pool
  entity type.
  """
  # Regex used to extract topic names from metrics sample.
  _POOL_METRICS_REGEX = \
      re.compile('^admission-controller.([^.]+).(.+)')


  def __init__(self, role_name, service_name, sources, max_pool):
    self._role_name = role_name
    self._service_name = service_name
    self._sources = sources
    self._max_pool = max_pool
    # Maps daemon pool name to corresponding EntityTypeUpdateEntry.
    self.daemon_pools = dict()
    self.deleteOldEntities = True
    # Build map from source to metric id
    self._source_to_id = dict()
    for metric_id, source in self._sources.iteritems():
      # Source is defined as a list, while for Impala daemon pool, currently
      # it should only contain one string.
      if len(source) != 1:
        THROTTLED_LOG.exception('Impala daemon pool metric source should be simple string without any separator. Source: %s' %
                               str(source))
        continue
      self._source_to_id[source[0]] = metric_id

  def visit_node(self, key, value):
    match = _PoolMetricsExtractor._POOL_METRICS_REGEX.match(key)
    if not match:
      return True

    pool_name = match.group(2)

    # Add daemon pool update entry.
    daemon_pool_etue = self.daemon_pools.get(pool_name, None)

    # First time seeing the pool, create.
    if daemon_pool_etue is None:
      # Unless the cap has been reached.
      if len(self.daemon_pools) > self._max_pool:
        THROTTLED_LOG.info("Max pool update of %s reached. Ignoring unseen pool %s" %
                           (str(self._max_pool), pool_name))
        return True
      attributes = dict()
      attributes['poolName'] = pool_name
      attributes['roleName'] = self._role_name
      attributes['serviceName'] = self._service_name
      daemon_pool_etue = EntityTypeUpdateEntry(attributes)
      self.daemon_pools[pool_name] = daemon_pool_etue

    metric_name = match.group(1)
    if metric_name not in self._source_to_id:
      return True
    metric_id = self._source_to_id[metric_name]
    if metric_id is None:
      return True
    else:
      daemon_pool_etue.add_metric(metric_id, value)

    return True

  def get_entity_type_update(self):
    daemon_pool_etu = EntityTypeUpdate(ImpaladAdapter._DAEMON_POOL_ENTITY_TYPE, False, True)
    for etue in self.daemon_pools.values():
      daemon_pool_etu.add_entity(etue)
    return [daemon_pool_etu]

class ImpaladAdapter(_ImpalaAdapter):

  _ROLETYPE = 'IMPALAD'

  _SECTION = 'impalad'
  _WEBSERVER_PORT = 'impalad_webserver_port'
  # Added in CM 4.6.1
  _WEBSERVER_USERNAME = 'webserver_username'
  _WEBSERVER_PASSWORD = 'webserver_password'
  # Added in CM 5.2
  _SCRATCH_DIRS = 'scratch_dirs'
  _DAEMON_POOL_ENTITY_TYPE = 'IMPALA_DAEMON_POOL'

  _POOL_UPDATE_FEATURE = "send_pool_entity_update"
  _POOL_UPDATE_MAX = "max"
  _POOL_UPDATE_MAX_DEFAULT = 1000

  def __init__(self, daemon, safety_valve, query_monitor = None):
    _ImpalaAdapter.__init__(self,
                            ImpaladAdapter._ROLETYPE,
                            safety_valve,
                            ImpaladAdapter._WEBSERVER_PORT)
    self._query_monitor_collector = QueryMonitorMetricCollector(daemon, self, query_monitor)

  def get_metrics_url_username_attr(self):
    return ImpaladAdapter._WEBSERVER_USERNAME

  def get_metrics_url_password_attr(self):
    return ImpaladAdapter._WEBSERVER_PASSWORD

  def _get_ready(self, json_parser):
    if (json_parser is None or not hasattr(json_parser, 'json')):
      return constants.IMPALA_DAEMON_READY_STATE_UNKNOWN
    return int(json_parser.json.get('impala-server.ready',
                                constants.IMPALA_DAEMON_READY_STATE_UNKNOWN))

  def _get_entity_type_updates(self, conf, json):
    entity_type_updates = None
    role_name, service_name = self._extract_role_and_service_name(conf)
    if self._service_defined_metrics is not None and \
            self._is_feature_enabled(None,
                                     ImpaladAdapter._POOL_UPDATE_FEATURE,
                                     role_name,
                                     service_name):
      # Impala daemon pool
      pools_extractor = _PoolMetricsExtractor(
        role_name,
        service_name,
        self._service_defined_metrics.get_sources(ImpaladAdapter._DAEMON_POOL_ENTITY_TYPE),
        self._get_pool_update_max(conf))
      visit_json(json, [pools_extractor])
      entity_type_updates = pools_extractor.get_entity_type_update()
    return entity_type_updates

  def _add_entity_type_updates(self, service_version, update, entity_type_updates, accessors):
    if self._is_feature_enabled(update, ImpaladAdapter._POOL_UPDATE_FEATURE):
      update.set_entity_type_updates(entity_type_updates)

  def _get_pool_update_max(self, conf):
    """
    The cap on the number of (daemon) pool updates sent to firehose. The
    role-level safety valve has highest priority, followed by service-level,
    and neither are specified, defaults to _POOL_UPDATE_MAX.
    """
    role_name, service_name = self._extract_role_and_service_name(conf)
    role_key = "%s_%s_%s_%s" % \
               (service_name.lower(),
                role_name.lower(),
                ImpaladAdapter._POOL_UPDATE_FEATURE,
                ImpaladAdapter._POOL_UPDATE_MAX)
    service_key = "%s_%s_%s" % \
                  (service_name.lower(),
                   ImpaladAdapter._POOL_UPDATE_FEATURE,
                   ImpaladAdapter._POOL_UPDATE_MAX)
    max_pool = None
    max_pool = self._safety_valve.get(service_key)
    max_pool = self._safety_valve.get(role_key)
    if max_pool:
      return int(max_pool)
    else:
      return ImpaladAdapter._POOL_UPDATE_MAX_DEFAULT

  def role_update_init(self, update):
    impala_daemon_update = ImpalaDaemonUpdate()
    impala_daemon_update.set_ready(constants.IMPALA_DAEMON_READY_STATE_UNKNOWN)
    update.set_impala_daemon_update(impala_daemon_update)

  def _set_additional_role_updates(self, json_parser, update):
    impala_daemon_update = ImpalaDaemonUpdate()
    impala_daemon_update.set_ready(self._get_ready(json_parser))
    update.set_impala_daemon_update(impala_daemon_update)

  def get_monitored_directories(self, conf):
    # Need to override because scratch dir config is a list.
    directories = _ImpalaAdapter.get_monitored_directories(self, conf)
    for scratch_dir in conf.getlist_with_default(
            ImpaladAdapter._SECTION,
            ImpaladAdapter._SCRATCH_DIRS,
            []):
      directories.append(dict(path=scratch_dir, type=None))
    return directories

  def get_adapter_specific_collectors(self):
    return [self._query_monitor_collector]

  #
  # See the version hash in the metrics for when the sample was generated.
  #

  SAMPLE_CDH5 = r"""
  {
     "admission-controller.total-admitted.root.admin.child": 2,
     "admission-controller.total-released.root.admin.child": 13,
     "admission-controller.known-metric.root.admin": 123,
     "admission-controller.total-admitted.foo": 5,
     "admission-controller.total-released.foo": 7,
     "catalog.num-databases":1,
     "catalog.num-tables":0,
     "catalog.ready":1,
     "cgroups-mgr.active-cgroups":0,
     "impala-server.backends.client-cache.clients-in-use":0,
     "impala-server.backends.client-cache.total-clients":0,
     "impala-server.hash-table.total-bytes":0,
     "impala-server.io-mgr.num-buffers":0,
     "impala-server.io-mgr.num-open-files":0,
     "impala-server.io-mgr.num-unused-buffers":0,
     "impala-server.io-mgr.total-bytes":0,
     "impala-server.last-refresh-time":"n/a",
     "impala-server.mem-pool.total-bytes":0,
     "impala-server.num-files-open-for-insert":0,
     "impala-server.num-fragments":0,
     "impala-server.num-open-beeswax-sessions":0,
     "impala-server.num-open-hiveserver2-sessions":2,
     "impala-server.num-queries":0,
     "impala-server.num-queries-expired":0,
     "impala-server.num-sessions-expired":0,
     "impala-server.ready":1,
     "impala-server.scan-ranges.num-missing-volume-id":0,
     "impala-server.scan-ranges.total":0,
     "impala-server.start-time":"2013-12-10 06:54:33",
     "impala-server.version":"impalad version 1.2.2-INTERNAL RELEASE (build 4161d79734b48bcbff5b47a7c9e09d206cc4b848)",
     "impala.thrift-server.backend.connections-in-use":0,
     "impala.thrift-server.backend.connection-setup-queue-size":0,
     "impala.thrift-server.backend.total-connections":0,
     "impala.thrift-server.beeswax-frontend.connections-in-use":0,
     "impala.thrift-server.beeswax-frontend.total-connections":0,
     "impala.thrift-server.hiveserver2-frontend.connections-in-use":3,
     "impala.thrift-server.hiveserver2-frontend.total-connections":3,
     "mem-tracker.process.bytes-freed-by-last-gc":-1,
     "mem-tracker.process.bytes-over-limit":-1,
     "mem-tracker.process.num-gcs":0,
     "mem-tracker.process.limit":123738374,
     "impala.thrift-server.hiveserver2-frontend.connections-in-use":0,
     "impala.thrift-server.hiveserver2-frontend.total-connections":0,
     "impala.thrift-server.llama-callback.connections-in-use":1,
     "impala.thrift-server.llama-callback.total-connections":1,
     "resource-broker.request-response-time":{
        "count":97581,
        "last":0.500591,
        "min":0.242867,
        "max":0.869685,
        "mean":0.500618,
        "stddev":0.00168997
     },
     "resource-broker.request-rpc-time":{
        "count":97581,
        "last":0.500591,
        "min":0.242867,
        "max":0.869685,
        "mean":0.500618,
        "stddev":0.00168997
     },
     "resource-broker.requests-failed":0,
     "resource-broker.requests-fulfilled":0,
     "resource-broker.requests-rejected":0,
     "resource-broker.requests-released":0,
     "resource-broker.requests-timedout":0,
     "resource-broker.requests-total":0,
     "simple-scheduler.assignments.total":0,
     "simple-scheduler.initialized":1,
     "simple-scheduler.local-assignments.total":0,
     "simple-scheduler.num-backends":4,
     "statestore-subscriber.connected":1,
     "statestore-subscriber.heartbeat-duration":{
        "count":13283,
        "last":1.9549e-05,
        "min":8.874e-06,
        "max":0.505313,
        "mean":5.93936e-05,
        "stddev":0.00438408
     },
     "statestore-subscriber.heartbeat-interval-time":{
        "count":13283,
        "last":0.500402,
        "min":0.000650477,
        "max":1.00574,
        "mean":0.50043,
        "stddev":0.00616953
     },
     "statestore-subscriber.last-recovery-duration":0,
     "statestore-subscriber.last-recovery-time":"N/A",
     "statestore-subscriber.statestore.client-cache.clients-in-use":0,
     "statestore-subscriber.statestore.client-cache.total-clients":1,
     "statestore-susbcriber.registration-id":"674c7c93ccb367ac:6f8c055803d19698",
     "tcmalloc.bytes-in-use":25334552,
     "tcmalloc.pageheap-free-bytes":43589632,
     "tcmalloc.pageheap-unmapped-bytes":0,
     "tcmalloc.physical-bytes-reserved":94437376,
     "tcmalloc.total-bytes-reserved":94437376,
     "thread-manager.running-threads":142,
     "thread-manager.total-threads-created":142
  }
  """

