# Copyright (c) 2015 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging
import os
import re
import urllib2

import simplejson

from cmf.monitor.entity_type_update import EntityTypeUpdate
from cmf.monitor.entity_type_update import EntityTypeUpdateEntry
from cmf.monitor.generic.adapter import Adapter
from cmf.monitor.generic.service_defined_metrics import ServiceDefinedMetrics
from cmf.monitor.generic.utils import JsonVisitor, visit_json
from cmf.monitor.generic.utils import SimpleMetricsExtractor
from cmf.monitor.generic.utils import SplittingSourceProcessor
from cmf.monitor.generic.utils import VersionUtils, VersionType
from cmf.monitor.generic.utils import extract_metric_from_json
from cmf.monitor.generic.utils import parse_properties_file
from cmf.throttling_logger import ThrottlingLogger
from cmf.util.url import urlopen_with_timeout

LOG = logging.getLogger('KafkaAdapters')
THROTTLED_LOG = ThrottlingLogger(LOG, 60 * 60) # Once an hour.

class _KafkaAdapter(Adapter):
  """
  Base class for all Kafka service role adapters. This exists to provide utility
  methods and to ensure consistency in the handling of common metrics.
  """

  KAFKA_MONITORING_PROPERTIES = "kafka-monitoring.properties"

  def __init__(self,
               role_type,
               safety_valve,
               host_key=None,
               port_key=None,
               enabled_key=None):
    Adapter.__init__(self, "KAFKA", role_type, safety_valve)
    self._host_key = host_key
    self._port_key = port_key
    self._enabled_key = enabled_key
    self._metrics = None

  def get_metrics_url(self, conf):
    return self._get_api_url(conf, "metrics")

  def get_topicmetricnames_url(self, conf):
    return self._get_api_url(conf, "topicmetricnames")

  def _get_api_url(self, conf, endpoint):
    if conf is None:
      raise Exception("A configuration is required!")
    if endpoint is None:
      raise Exception("An endpoint is required!")

    if (self._host_key is None or
        self._port_key is None or
        self._enabled_key is None):
      return None

    path = os.path.join(os.path.dirname(conf.path),
                        _KafkaAdapter.KAFKA_MONITORING_PROPERTIES)
    try:
      parser = parse_properties_file(path, self.section)
      if parser is None:
        raise Exception("%s does not exist" % (os.path.abspath(path)))

      enabled = parser.getboolean(self.section, self._enabled_key)
      if not enabled:
        return None
      host = parser.get(self.section, self._host_key)
      if host is None:
        LOG.error("%s entry missing from %s" % (self._host_key, path))
        return None
      port = parser.getint(self.section, self._port_key)
      if port is None:
        LOG.error("%s entry missing from %s" % (self._port_key, path))
        return None
      return "http://%s:%s/api/%s" % (host, port, endpoint)
    except:
      LOG.exception("Failed to read conf file '%s'" % (os.path.abspath(path)))
      return None

  def read_service_defined_metrics(self, path):
    if path is None:
      raise Exception("A path is required!")
    self._metrics = ServiceDefinedMetrics(path, SplittingSourceProcessor('::'))

class _TopicMetricsExtractor(JsonVisitor):
  # Regex used to extract topic names from metrics sample.
  _TOPIC_METRICS_REGEX = \
      re.compile('^kafka.server.BrokerTopicMetrics.topic.(.+)$')

  def __init__(self, role_name, service_name, sources, topic_names):
    self._role_name = role_name
    self._service_name = service_name
    self._sources = sources
    self._topic_names = topic_names
    self.topics = dict()
    self.deleteOldEntities = True

  def visit_node(self, key, json):
    # If we don't have topic names, there's nothing to do. We will have already
    # logged errors during topic name collection.
    if self._topic_names is not None and len(self._topic_names) == 0:
      self.deleteOldEntities = False
      return True

    match = _TopicMetricsExtractor._TOPIC_METRICS_REGEX.match(key)
    if not match:
      return True

    name = match.group(1)

    if self._topic_names is None:
      topic_name = name
    else:
      topic_name = self._topic_names.get(name, None)
      if topic_name is None:
        # When a topic is deleted from Kakfa, the actual topic metrics on the
        # broker stick around for a while -- either because the broker is in
        # the process of cleaning up, or because of Kafka bugs. The topic will
        # not be in the topic metric names map, so we will not report on it. We
        # avoid warning here, or logging every time since we expect to see this
        # in practice at times.
        THROTTLED_LOG.info("No topic '%s' in topic names map. The topic may "
                           "have been deleted. Ignoring topic metrics." %
                           (name))
        # We have not collected cleanly, we'll leave old entiites in place.
        self.deleteOldEntities = False
        return True

    attributes = dict()
    attributes['roleName'] = self._role_name
    attributes['serviceName'] = self._service_name
    attributes['kafkaTopicName'] = topic_name
    topic = EntityTypeUpdateEntry(attributes)

    for metricid, source in self._sources.iteritems():
      value = extract_metric_from_json(json, source)
      if value is None:
        continue
      topic.add_metric(metricid, value)

    self.topics[topic_name] = topic

    return True


class _TopicPartitionMetricsExtractor(JsonVisitor):
  # Regex used to extract partitions and topic names from metrics sample.
  _TOPIC_PARTITION_METRICS_REGEX = \
    re.compile('^kafka.server.BrokerTopicMetrics.partition.(.+).topic.(.+)$')

  def __init__(self, role_name, service_name, sources, topic_names):
    self._role_name = role_name
    self._service_name = service_name
    self._sources = sources
    self._topic_names = topic_names
    self.topic_partitions = dict()
    self.deleteOldEntities = True

  def visit_node(self, key, json):
    # If we don't have topic names, there's nothing to do. We will have already
    # logged errors during topic name collection.
    if self._topic_names is not None and len(self._topic_names) == 0:
      self.deleteOldEntities = False
      return True

    match = _TopicPartitionMetricsExtractor._TOPIC_PARTITION_METRICS_REGEX.match(key)
    if not match:
      return True

    partition_id = match.group(1)
    name = match.group(2)

    if self._topic_names is None:
      topic_name = name
    else:
      topic_name = self._topic_names.get(name, None)
      if topic_name is None:
        # When a topic is deleted from Kakfa, the actual topic metrics on the
        # broker stick around for a while -- either because the broker is in
        # the process of cleaning up, or because of Kafka bugs. The topic will
        # not be in the topic metric names map, so we will not report on it. We
        # avoid warning here, or logging every time since we expect to see this
        # in practice at times.
        THROTTLED_LOG.info("No topic '%s' in topic names map. The topic may "
                           "have been deleted. Ignoring topic metrics." %
                           (name))
        # We have not collected cleanly, we'll leave old entiites in place.
        self.deleteOldEntities = False
        return True

    attributes = dict()
    attributes['roleName'] = self._role_name
    attributes['serviceName'] = self._service_name
    attributes['kafkaTopicName'] = topic_name
    attributes['kafkaPartitionId'] = partition_id
    topic_partition = EntityTypeUpdateEntry(attributes)

    for metricid, source in self._sources.iteritems():
      value = extract_metric_from_json(json, source)
      if value is None:
        continue
      topic_partition.add_metric(metricid, value)

    self.topic_partitions['%s:%s' % (topic_name, partition_id)] = topic_partition

    return True

class _ReplicaMetricsExtractor(JsonVisitor):
  # Regex used to extract partition and topic names from metrics sample.
  _REPLICA_METRICS_REGEX = \
      re.compile('^kafka.(log.Log|cluster.Partition).partition.(.+).topic.(.+)$')

  def __init__(self, role_name, service_name, sources, topic_names):
    self._role_name = role_name
    self._service_name = service_name
    self._sources = sources
    self._topic_names = topic_names
    # keyed on topic name : partition id
    self.replicas = dict()
    self.deleteOldEntities = True

  def visit_node(self, key, json):
    # If we don't have topic names, there's nothing to do. We will have already
    # logged errors during topic name collection.
    if self._topic_names is not None and len(self._topic_names) == 0:
      self.deleteOldEntities = False
      return True

    match = _ReplicaMetricsExtractor._REPLICA_METRICS_REGEX.match(key)
    if not match:
      return True

    partition_id = match.group(2)

    name = match.group(3)
    if self._topic_names is None:
      topic_name = name
    else:
      topic_name = self._topic_names.get(name, None)
      if topic_name is None:
        # When a topic is deleted from Kakfa, the actual topic metrics on the
        # broker stick around for a while -- either because the broker is in
        # the process of cleaning up, or because of Kafka bugs. The topic will
        # not be in the topic metric names map, so we will not report on it. We
        # avoid warning here, or logging every time since we expect to see this
        # in practice at times.
        THROTTLED_LOG.info("No topic '%s' in topic names map. The topic may "
                           "have been deleted. Ignoring topic metrics." %
                           (name))
        # We have not collected cleanly, we'll leave old entiites in place.
        self.deleteOldEntities = False
        return True

    attributes = dict()
    attributes['roleName'] = self._role_name
    attributes['serviceName'] = self._service_name
    attributes['kafkaTopicName'] = topic_name
    attributes['kafkaPartitionId'] = partition_id
    replica = EntityTypeUpdateEntry(attributes)

    for metricid, source in self._sources.iteritems():
      value = extract_metric_from_json(json, source)
      if value is None:
        continue
      replica.add_metric(metricid, value)

    self.replicas['%s:%s' % (topic_name, partition_id)] = replica

    return True

class _ProducerMetricsExtractor(JsonVisitor):
  # Regex used to extract client id, partition and topic names from metrics sample.
  _PRODUCER_METRICS_REGEX = \
    re.compile('^kafka.server.BrokerClientMetrics.clientId.(.+).partition.(.+).topic.(.+)$')

  def __init__(self, role_name, service_name, sources, topic_names):
    self._role_name = role_name
    self._service_name = service_name
    self._sources = sources
    self._topic_names = topic_names
    # keyed on topic name : partition id
    self.producer_metrics = dict()
    self.deleteOldEntities = False

  def visit_node(self, key, json):
    # If we don't have topic names, there's nothing to do. We will have already
    # logged errors during topic name collection.
    if self._topic_names is not None and len(self._topic_names) == 0:
      self.deleteOldEntities = False
      return True

    match = _ProducerMetricsExtractor._PRODUCER_METRICS_REGEX.match(key)
    if not match:
      return True

    client_id = match.group(1)
    partition_id = match.group(2)
    name = match.group(3)

    if self._topic_names is None:
      topic_name = name
    else:
      topic_name = self._topic_names.get(name, None)
      if topic_name is None:
        # When a topic is deleted from Kakfa, the actual topic metrics on the
        # broker stick around for a while -- either because the broker is in
        # the process of cleaning up, or because of Kafka bugs. The topic will
        # not be in the topic metric names map, so we will not report on it. We
        # avoid warning here, or logging every time since we expect to see this
        # in practice at times.
        THROTTLED_LOG.info("No topic '%s' in topic names map. The topic may "
                           "have been deleted. Ignoring topic metrics." %
                           (name))
        # We have not collected cleanly, we'll leave old entiites in place.
        self.deleteOldEntities = False
        return True

    attributes = dict()
    attributes['roleName'] = self._role_name
    attributes['serviceName'] = self._service_name
    attributes['kafkaClientId'] = client_id
    attributes['kafkaTopicName'] = topic_name
    attributes['kafkaPartitionId'] = partition_id
    producer_metric = EntityTypeUpdateEntry(attributes)

    for metricid, source in self._sources.iteritems():
      value = extract_metric_from_json(json, source)
      if value is None:
        continue
      producer_metric.add_metric(metricid, value)

    self.producer_metrics['%s:%s:%s' % (client_id, topic_name, partition_id)] = producer_metric

    return True

class BrokerAdapter(_KafkaAdapter):
  # Config keys used to read from kafka-monitor.properties.
  _KAFKA_BROKER_WEBSERVER_HOST = "kafka.http.metrics.host"
  _KAFKA_BROKER_WEBSERVER_PORT = "kafka.http.metrics.port"
  _KAFKA_BROKER_WEBSERVER_ENABLED = "monitoring.enabled"
  _KAFKA_BROKER_ID = "broker.id"
  _ROLE_ENTITY_TYPE = 'KAFKA-KAFKA_BROKER'
  _BROKER_TOPIC_ENTITY_TYPE = 'KAFKA_BROKER_TOPIC'
  _BROKER_TOPIC_PARTITION_ENTITY_TYPE = 'KAFKA_BROKER_TOPIC_PARTITION'
  _REPLICA_ENTITY_TYPE = 'KAFKA_REPLICA'
  _PRODUCER_ENTITY_TYPE = 'KAFKA_PRODUCER'
  _KAFKA_BROKER_ID_ATTRIBUTE = 'kafkaBrokerId'

  def __init__(self, safety_valve):
    _KafkaAdapter.__init__(self,
                           "KAFKA_BROKER",
                           safety_valve,
                           BrokerAdapter._KAFKA_BROKER_WEBSERVER_HOST,
                           BrokerAdapter._KAFKA_BROKER_WEBSERVER_PORT,
                           BrokerAdapter._KAFKA_BROKER_WEBSERVER_ENABLED)
    self._broker_id = None
    # This exists to ease testing.
    self._url_open_function = urlopen_with_timeout

  def _get_broker_id(self, conf):
    if conf is None:
      raise Exception("A configuration is required!")
    path = os.path.join(os.path.dirname(conf.path),
                        _KafkaAdapter.KAFKA_MONITORING_PROPERTIES)
    try:
      parser = parse_properties_file(path, self.section)
      if parser is None:
        raise Exception("%s does not exist" % (os.path.abspath(path)))
      return parser.get(self.section, BrokerAdapter._KAFKA_BROKER_ID)
    except:
      LOG.exception("Failed to udpate broker ID from '%s'" %
                    (os.path.abspath(path)))
      return None

  def _fetch_topic_names(self, conf):
    # See OPSAPS-27547 for context. This is definitely a little odd, both that
    # we have to do this, but also that we're making an additional urlopen call
    # from inside this adapter, which is somewhat gross. This call pulls down
    # the mapping of kafka topic name to kafka topic name in the metrics system
    # and produces a reversed map so that we can undo the mapping before we send
    # our metrics. See the sample below for an example of the output. If we can
    # not get the topic names, we'll return an empty dictionary which prevents
    # publication of topic and replica metrics. If we think we are interacting
    # with a Kafka broker that does not support the topics name servlet, we will
    # return None, which reuslts in the topic names map being ignored in metric
    # collection.
    if conf is None:
      raise Exception("A configuration is required!")

    key = 'kafka_topicmetricnames_policy'
    policy = self._safety_valve.get(key, 'default').strip().lower()
    if policy == 'skip':
      THROTTLED_LOG.debug("Skipping check for topicmetricnames because of '%s' "
                          "safety valve setting." % key)
      return None
    elif policy == 'force':
      THROTTLED_LOG.debug("Forcing check for topicmetricnames because of '%s' "
                          "safety valve setting." % key)
    elif policy == 'default':
      pass
    else:
      THROTTLED_LOG.info("Unknown value for '%s' safety valve setting: '%s'. "
                         "Using 'default' instead." % (key, policy))

    try:
      url = self.get_topicmetricnames_url(conf)
      if url is None:
        raise Exception("Could not determine topicmetricnames URL!")
      json = simplejson.load(self._url_open_function(url))
      topics = json.get("topicNames", None)
      if topics is None:
        raise Exception("Unexpected JSON from '%s': %s" % (url,
                                                           topic_names_json))
      metric_name_to_name = dict()
      for topic_name, metric_name in topics.iteritems():
        metric_name_to_name[metric_name] = topic_name
      return metric_name_to_name
    except urllib2.HTTPError, err:
      if policy == 'default' and err.code == 404:
        # We intentionally log nothing here, assuming this broker does not have
        # servlet support. Returning None means that collection will proceed
        # without trying to adjust topic-related metric names. Note that if the
        # policy is 'force', we will not reach here and instead log the error
        # and return an empty dict(), which will impact metric collection.
        return None
      else:
        LOG.exception("Failed to get topic metric names.")
        return dict()
    except:
      LOG.exception("Failed to get topic metric names.")
      return dict()

  def parse_metrics_from_url(self, conf, json):
    if conf is None:
      raise Exception("A configuration is required!")
    if json is None:
      raise Exception("A json sample is required!")
    if self._metrics is None:
      raise Exception("No metrics have been loaded!")

    release = self.get_release(conf)
    if release is None:
      raise Exception("Could not determine release!")

    # The kafka broker ID was exposed from the CSD in CM 5.5.
    # Version detection logic should be handled in the server (tracked in: OPSAPS-36429)
    if VersionUtils.at_least(release, "5.5", VersionType.MINOR):
      if self._broker_id is None:
        self._broker_id = self._get_broker_id(conf)

    role_name, service_name = self._extract_role_and_service_name(conf)

    broker_metrics_extractor = SimpleMetricsExtractor(
      self._metrics.get_sources(BrokerAdapter._ROLE_ENTITY_TYPE))

    topic_names = self._fetch_topic_names(conf)
    topic_metrics_extractor = _TopicMetricsExtractor(
      role_name,
      service_name,
      self._metrics.get_sources(BrokerAdapter._BROKER_TOPIC_ENTITY_TYPE),
      topic_names)
    topic_partition_metrics_extractor = _TopicPartitionMetricsExtractor(
      role_name,
      service_name,
      self._metrics.get_sources(BrokerAdapter._BROKER_TOPIC_PARTITION_ENTITY_TYPE),
      topic_names)
    replica_metrics_extractor = _ReplicaMetricsExtractor(
      role_name,
      service_name,
      self._metrics.get_sources(BrokerAdapter._REPLICA_ENTITY_TYPE),
      topic_names)
    producer_metrics_extractor = _ProducerMetricsExtractor(
      role_name,
      service_name,
      self._metrics.get_sources(BrokerAdapter._PRODUCER_ENTITY_TYPE),
      topic_names
    )

    visit_json(json, [broker_metrics_extractor,
                      topic_metrics_extractor,
                      topic_partition_metrics_extractor,
                      replica_metrics_extractor,
                      producer_metrics_extractor])

    return (broker_metrics_extractor,
            topic_metrics_extractor,
            topic_partition_metrics_extractor,
            replica_metrics_extractor,
            producer_metrics_extractor)

  def add_webserver_metrics(self, version, update, metrics, accessors):
    (broker_metrics_extractor,
     topic_metrics_extractor,
     topic_partition_metrics_extractor,
     replica_metrics_extractor,
     producer_metrics_extractor) = metrics

    if self._broker_id is not None:
      update.set_additional_role_attributes(
        { BrokerAdapter._KAFKA_BROKER_ID_ATTRIBUTE : self._broker_id })

    for metricid, value in broker_metrics_extractor.metrics.iteritems():
      update.add_metric(metricid, value)

    entity_type_updates = []

    if self._is_feature_enabled(update, "send_broker_topic_entity_update"):
      entity_type_update = EntityTypeUpdate(
        BrokerAdapter._BROKER_TOPIC_ENTITY_TYPE,
        deleteOldEntities=topic_metrics_extractor.deleteOldEntities,
        createNewEntities=True)
      for topic in topic_metrics_extractor.topics.itervalues():
        entity_type_update.add_entity(topic)
      entity_type_updates.append(entity_type_update)

    if self._is_feature_enabled(update, "send_replica_entity_update"):
      entity_type_update = EntityTypeUpdate(
        BrokerAdapter._REPLICA_ENTITY_TYPE,
        deleteOldEntities=replica_metrics_extractor.deleteOldEntities,
        createNewEntities=True)
      for replica in replica_metrics_extractor.replicas.itervalues():
        entity_type_update.add_entity(replica)
      entity_type_updates.append(entity_type_update)

    # not collecting KAFKA_BROKER_TOPIC_PARTITION by default,
    # only if <service name>_broker_topic_partition_for_smm_enabled=true is set
    if self._is_feature_enabled(update, "send_broker_topic_partition_entity_update") and self._is_smm_metrics_enabled(update):
      entity_type_update = EntityTypeUpdate(
        BrokerAdapter._BROKER_TOPIC_PARTITION_ENTITY_TYPE,
        deleteOldEntities=topic_partition_metrics_extractor.deleteOldEntities,
        createNewEntities=True)
      for topic_partition in topic_partition_metrics_extractor.topic_partitions.itervalues():
        entity_type_update.add_entity(topic_partition)
      entity_type_updates.append(entity_type_update)

    if self._is_feature_enabled(update, "send_producer_entity_update"):
      entity_type_update = EntityTypeUpdate(
        BrokerAdapter._PRODUCER_ENTITY_TYPE,
        deleteOldEntities=producer_metrics_extractor.deleteOldEntities,
        createNewEntities=True)
      for producer_metric in producer_metrics_extractor.producer_metrics.itervalues():
        entity_type_update.add_entity(producer_metric)
      entity_type_updates.append(entity_type_update)

    update.set_entity_type_updates(entity_type_updates)

  def _is_smm_metrics_enabled(self, role_update):
    service_name = role_update.service_name
    service_key = "%s_broker_topic_partition_metrics_for_smm_enabled" % service_name.lower()
    return 'true' == self.safety_valve.get(service_key, 'false').strip().lower()

  def get_metrics_sample(self, version):
    return open(os.path.join(os.path.dirname(__file__), "../samples/kafka_broker"), 'r').read()

_TOPICMETRICNAMES_SAMPLE = r"""
{
   "topicNames":{
      "tm":"tm",
      "TM":"TM",
      "a1":"a1",
      "a2":"a2",
      "t1":"t1",
      "t2":"t2",
      "meter1":"meter1",
      "shapira":"shapira",
      "hosts":"hosts",
      "channel1":"channel1",
      "channel2":"channel2",
      "smmtest":"smmtest"
   }
}
"""
