# Copyright (c) 2015 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging
import os

from cmf.monitor.entity_type_update import EntityTypeUpdate
from cmf.monitor.entity_type_update import EntityTypeUpdateEntry
from cmf.monitor.generic.adapter import Adapter
from cmf.monitor.generic.service_defined_metrics import ServiceDefinedMetrics
from cmf.monitor.generic.utils import extract_metric_from_json
from cmf.monitor.generic.utils import parse_properties_file
from cmf.monitor.generic.utils import SplittingSourceProcessor

LOG = logging.getLogger('KuduAdapters')

class _KuduMetricExtractor(object):
  def __init__(self, sources):
    self._sources = sources

  def extract_metrics(self, json):
    metrics_by_name = dict()
    for metric in json.get('metrics', []):
      name = metric.get('name', None)
      if name is None:
        raise Exception("Metric missing 'name'!")
      metrics_by_name[name] = metric

    metrics = dict()
    for metricid, source in self._sources.iteritems():
      value = extract_metric_from_json(metrics_by_name, source)
      if value is None:
        continue
      metrics[metricid] = value
    return metrics

class _RoleMetricsExtractor(_KuduMetricExtractor):
  def __init__(self, sources):
    _KuduMetricExtractor.__init__(self, sources)
    self.metrics = dict()

  def extract(self, json):
    for node in json:
      if node.get('type', None) == 'server':
        self.metrics.update(self.extract_metrics(node))
        return
    raise Exception("Could not find server metrics!")

class _ReplicaMetricsExtractor(_KuduMetricExtractor):
  def __init__(self, role_name, sources):
    _KuduMetricExtractor.__init__(self, sources)
    self._role_name = role_name
    self.replicas = dict()

  def extract(self, json):
    for node in json:
      if node.get('type', None) == 'tablet':
        self._extract_replica(node)

  def _extract_replica(self, json):
    kuduTabletId = json.get('id', None)
    if kuduTabletId is None:
      raise Exception("Tablet entry missing 'id'!")

    attributes = dict()
    attributes['roleName'] = self._role_name
    attributes['kuduTabletId'] = kuduTabletId
    replica = EntityTypeUpdateEntry(attributes)
    self.replicas[kuduTabletId] = replica

    replica.add_metrics(self.extract_metrics(json))

class _KuduAdapter(Adapter):
  """
  Base class for all Kudu service role adapters. This exists to provide utility
  methods and to ensure consistency in the handling of common metrics.
  """

  # Metric sources we care about in addition to the roles.
  _REPLICA_ENTITY_TYPE = 'KUDU_REPLICA'

  def __init__(self,
               role_type,
               role_entity_type,
               safety_valve,
               host_key=None,
               port_key=None):
    Adapter.__init__(self, "KUDU", role_type, safety_valve)
    self._role_entity_type = role_entity_type
    self._host_key = host_key
    self._port_key = port_key
    self._metrics = None

  def get_metrics_url(self, conf):
    if conf is None:
      raise Exception("A configuration is required!")

    if (self._host_key is None or
        self._port_key is None):
      return None

    path = os.path.join(os.path.dirname(conf.path),
                        "kudu-monitoring.properties")
    try:
      parser = parse_properties_file(path, self.section)
      if parser is None:
        raise Exception("%s does not exist" % (os.path.abspath(path)))

      scheme = "http"
      if parser.has_option(self.section, "ssl_enabled"):
        enabled = parser.get(self.section, "ssl_enabled", "false")
        if enabled == "true":
          scheme = "https"

      host = "0.0.0.0"
      if parser.has_option(self.section, self._host_key):
        host = parser.get(self.section, self._host_key, "0.0.0.0")
      if host == "" or host == "0.0.0.0":
        host = "localhost"
      port = parser.getint(self.section, self._port_key)
      if port is None:
        LOG.error("%s entry missing from %s" % (self._port_key, path))
        return None
      return "%s://%s:%s/jsonmetricz" % (scheme, host, port)
    except:
      LOG.exception("Failed to read conf file '%s'" % (os.path.abspath(path)))
      return None

  def read_service_defined_metrics(self, path):
    if path is None:
      raise Exception("A path is required!")
    self._metrics = ServiceDefinedMetrics(path, SplittingSourceProcessor('::'))

  def parse_metrics_from_url(self, conf, json):
    if json is None:
      raise Exception("A json sample is required!")
    if self._metrics is None:
      raise Exception("No metrics have been loaded!")

    role_name, _ = self._extract_role_and_service_name(conf)

    role_metrics_extractor = _RoleMetricsExtractor(
      self._metrics.get_sources(self._role_entity_type))
    role_metrics_extractor.extract(json)

    replica_metrics_extractor = _ReplicaMetricsExtractor(
        role_name,
        self._metrics.get_sources(_KuduAdapter._REPLICA_ENTITY_TYPE))
    replica_metrics_extractor.extract(json)

    return role_metrics_extractor, replica_metrics_extractor

  def add_webserver_metrics(self, version, update, metrics, accessors):
    role_metrics_extractor, replica_metrics_extractor = metrics

    for metricid, value in role_metrics_extractor.metrics.iteritems():
      update.add_metric(metricid, value)

    if self._is_feature_enabled(update, "send_replica_entity_update"):
      entity_type_update = EntityTypeUpdate(
        _KuduAdapter._REPLICA_ENTITY_TYPE,
        deleteOldEntities=False,
        createNewEntities=False)
      for replica in replica_metrics_extractor.replicas.itervalues():
        entity_type_update.add_entity(replica)
      update.set_entity_type_updates([entity_type_update])

class MasterAdapter(_KuduAdapter):
  def __init__(self, safety_valve):
    _KuduAdapter.__init__(self,
                          "KUDU_MASTER",
                          'KUDU-KUDU_MASTER',
                          safety_valve,
                          "webserver_interface",
                          "webserver_port")

  def get_metrics_sample(self, version):
    return open(os.path.join(os.path.dirname(__file__), "../samples/kudu_master"), 'r').read()

class TServerAdapter(_KuduAdapter):
  def __init__(self, safety_valve):
    _KuduAdapter.__init__(self,
                          "KUDU_TSERVER",
                          'KUDU-KUDU_TSERVER',
                          safety_valve,
                          "webserver_interface",
                          "webserver_port")

  def get_metrics_sample(self, version):
    return open(os.path.join(os.path.dirname(__file__), "../samples/kudu_tserver"), 'r').read()
