# Copyright (c) 2015 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging
import time

import psutil

try:
  from cStringIO import StringIO
except:
  from StringIO import StringIO
from cmf.monitor import constants, schema
from cmf.monitor.abstract_monitor import CollectionError
from cmf.safety_valve_helpers import get_urlopen_retries_on_auth_errors
from cmf.safety_valve_helpers import get_urlopen_sleep_seconds_on_auth_errors
from cmf.throttling_logger import ThrottlingLogger
from cmf.util.url import urlopen_with_timeout
from cmf.util.url import urlopen_with_retry_on_authentication_errors
import os
import simplejson

# We will use the same logger for all collectors.
LOG = logging.getLogger('AbstractMetricCollector')
THROTTLED_LOG = ThrottlingLogger(LOG, constants.DEFAULT_MONITOR_LOG_RATE)

class AbstractMetricCollector(object):
  """
  A helper class to keep track of metric collection from an end-point.
  """

  # Readability constant.
  _COLLECTOR_NOT_SUPPORTED = None

  def __init__(self, adapter, metric_accessors=None, name=None):
    if adapter is None:
      raise Exception("Adapter cannot be None")

    # The uri from which to retrieve the metrics. Each collector interprets this
    # differently. For example, this can point to a url or a file.
    self._metrics_uri = None
    # The results of the collection. The first in the tuple is the results. This
    # can be either the metrics opaque structure (understood by the adapter) or
    # a CollectionError object. The second member in the tuple is the timestamp
    # for the time the collection happened. The third member is the collection
    # duration, in seconds.
    self._metrics_tuple = (None, None, None)
    # The adapter that understand the metrics format.
    self._adapter = adapter
    # metrics metadata.
    self._metric_accessors = metric_accessors
    # Name of this collector.
    self._name = name
    # the last time report_metrics was called.
    self._last_reported_timestamp = None

  @property
  def metrics_uri(self):
    return self._metrics_uri

  @property
  def metric_accessors(self):
    """
    Return the metric accessors. Can be None for some collectors.
    """
    return self._metric_accessors

  @property
  def name(self):
    return self._name

  def is_enabled(self):
    """
    Returns 'true' if metric collection is enabled from this end-point, 'false'
    otherwise.
    :return:
    """
    return self._metrics_uri is not None

  def update_with_conf(self, conf):
    """
    Update the collector based on the supplied configurations, including but
    not limited to the metrics uri, such that subsequent collections are based
    on these configurations. Must be implemented by subclasses.
    """
    raise Exception("Endpoint for %s must implement update_with_conf!")

  @property
  def metrics_tuple(self):
    return self._metrics_tuple

  def on_start_collecting(self):
    """
    Invoked when collection should begin. The target for collection can be
    assumed to be active at this point, but is not guaranteed to, e.g. in case
    of unexpected failures. Raising exceptions neither prevent other functions
    from being called, nor guarantee any impact on the owning monitor.
    """
    pass

  def on_stop_collecting(self):
    """
    Invoked when collection should stop. The target for collection will become
    inactive afterwards. Raising exceptions does not guarantee any impact
    on the owning monitor.
    """
    pass

  def clear_results(self):
    """
    Reset the collected metrics tuple.
    """
    self._metrics_tuple = (None, None, None)

  def collect_and_parse(self, conf):
    """
    Collect and parse the metrics from the end-point. End-point handlers should
    implement this.
    """
    raise Exception(
        "Endpoint for %s must implement collect_and_parse!" %
        (self._metrics_uri,))

  def _add_metrics(self, service_version, update, collected_metrics, duration):
    """
    Add the metrics collected by this end-point. Returns 'True' if metrics were
    added, 'False' otherwise. Implementations must implement this.
    """
    raise Exception(
        "Endpoint for %s must implement get_collection_status_metric!" %
        (self._metrics_uri,))

  def report_metrics(self, service_version, update):
    """
    Add the metrics from this end-point to the update. Note that the update
    timestamp will be updated only if the end-point collection time is earlier
    than the update timestamp. This is only meaningful in case the update
    includes metrics from more than one end-point. Returns 'True' if metrics
    were added, 'False' otherwise.
    """
    metrics, timestamp, duration = self._metrics_tuple
    if metrics is None or timestamp == self._last_reported_timestamp:
      return False

    if update.timestamp > timestamp:
      update.timestamp = timestamp

    ret = self._add_metrics(service_version, update, metrics, duration)
    self._last_reported_timestamp = timestamp
    return ret

  def is_ready_to_report(self, conf, pid):
    """
    Returns 'true' if the collector is ready to report 'false' if the collector
    is not ready to report. Returns 'None' if the collector is not enabled for
    the configuration passed in 'conf'. 'pid' cannot be None.
    Derived classes must implement this.
    """
    raise Exception("Endpoint for %s must implement is_ready_to_report!")

  def clear_last_reported_timestamp(self):
    """
    For testing only.
    """
    self._last_reported_timestamp = None

class WebServerMetricCollector(AbstractMetricCollector):
  """
  A metric collector for a web server end-point exposing a json output.
  """

  def __init__(self, section, adapter, metric_accessors, username_attr=None,
               password_attr=None, urlopen_callout = urlopen_with_timeout,
               name="WebServerMetricCollector"):
    AbstractMetricCollector.__init__(self,
                                     adapter,
                                     metric_accessors,
                                     name)
    self._section = section
    self._username_attr = username_attr
    self._password_attr = password_attr
    self._username_value = None
    self._password_value = None
    self._urlopen_callout = urlopen_callout

  def update_with_conf(self, conf):
    self._metrics_uri = self._adapter.get_metrics_url(conf)
    self._username_value = conf.get_with_default(self._section, self._username_attr, None)
    self._password_value = conf.get_with_default(self._section, self._password_attr, None)

  def collect_and_parse(self, conf):
    self._metrics_tuple = self._collect_and_parse_and_return(conf)

  def _collect_and_parse_and_return(self, conf):
    code = constants.WEB_METRICS_COLLECTION_COMMUNICATION_FAILURE
    start = time.time()
    try:
      opened_url = urlopen_with_retry_on_authentication_errors(
        self._open_url,
        retries = get_urlopen_retries_on_auth_errors(
          self._adapter.safety_valve),
        sleeptime = get_urlopen_sleep_seconds_on_auth_errors(
          self._adapter.safety_valve))
      code = constants.WEB_METRICS_COLLECTION_PARSE_FAILURE
      result = self._adapter.parse_metrics_from_url(conf,
                                                    simplejson.load(opened_url))
    except Exception, ex:
      THROTTLED_LOG.exception("Error fetching metrics at '%s'" %
                              (self._metrics_uri,))
      result = CollectionError(code)
    now = time.time()
    return result, now, int((now - start) * 1000)

  def _open_url(self):
    """
    This function is to make testing easier.
    """
    return self._urlopen_callout(url=self._metrics_uri,
                                username=self._username_value,
                                password=self._password_value)

  def _add_metrics(self, service_version, update, collected_metrics, duration):
    update.add_metric(schema.WEB_METRICS_COLLECTION_DURATION, duration)
    if isinstance(collected_metrics, CollectionError):
      update.add_metric(schema.WEB_METRICS_COLLECTION_STATUS,
                        collected_metrics.code)
    else:
      update.add_metric(schema.WEB_METRICS_COLLECTION_STATUS,
                        constants.WEB_METRICS_COLLECTION_GOOD)
      self._adapter.add_webserver_metrics(service_version,
                                          update,
                                          collected_metrics,
                                          self.metric_accessors)
    return True

  def is_ready_to_report(self, conf, pid):
    if pid is None:
      raise Exception("Pid cannot be None")

    metrics_url = self._adapter.get_metrics_url(conf)
    if metrics_url is None:
      return AbstractMetricCollector._COLLECTOR_NOT_SUPPORTED
    try:
      simplejson.load(urlopen_with_timeout(metrics_url, timeout=0.5))
      return True
    except Exception, ex:
      pass
    return False

class SampleFileMetricCollector(AbstractMetricCollector):
  """
  A metric collector for a json file base end-point.
  """
  def __init__(self, adapter, metric_accessors):
    AbstractMetricCollector.__init__(self,
                                     adapter,
                                     metric_accessors,
                                     "SampleFileMetricCollector")
    # The monitored process start time. We will only collect sample files with
    # creation time older than this timestamp.
    self._process_start_time = None
    # The last time the sample file we read was modified. This timestamp will
    # be used to determine if we need to report any metrics. Be default we will
    # wait until a sample file newer than the collector start time is available.
    self._last_sample_file_collected_modified_time = time.time()


  def update_with_conf(self, conf):
    self._metrics_uri = self._adapter.get_metrics_file(conf)

  def collect_and_parse(self, conf):
    code = constants.FILE_METRICS_COLLECTION_UNKNOWN
    start = time.time()
    successful_collection = False
    sample_file = self._metrics_uri
    try:
      if os.path.exists(sample_file) and os.path.isfile(sample_file):
        sample = open(self._metrics_uri, 'r')
        mtime = os.fstat(sample.fileno()).st_mtime
        if mtime > self._last_sample_file_collected_modified_time:
          json = self._get_json(sample)
          code = constants.FILE_METRICS_COLLECTION_PARSE_FAILURE
          result = self._adapter.parse_metrics_from_file(conf,
                                                         simplejson.load(json))
          self._last_sample_file_collected_modified_time = mtime
          successful_collection = True
        else:
          result = \
            CollectionError(constants.FILE_METRICS_COLLECTION_STALE_SAMPLE)
      else:
        result = \
          CollectionError(constants.FILE_METRICS_COLLECTION_FILE_NOT_FOUND)
      if isinstance(result, CollectionError):
        THROTTLED_LOG.warn("Error collecting metrics from sample file '%s'. Result: %d" %
                           (self._metrics_uri, result.code))
    except Exception:
      THROTTLED_LOG.exception("Error fetching metrics at '%s'" % (sample_file,))
      result = CollectionError(code)
    now = time.time()
    if successful_collection:
      self._metrics_tuple = result, \
                            self._last_sample_file_collected_modified_time, \
                            int((now - start) * 1000)
    else:
      # We want to report the error so we will use the collection end-time
      # as the timestamp for the result.
      self._metrics_tuple = result, now, int((now - start) * 1000)

  def _add_metrics(self, service_version, update, collected_metrics, duration):
    if isinstance(collected_metrics, CollectionError):
      return False
    self._adapter.add_sample_file_metrics(service_version,
                                          update,
                                          collected_metrics,
                                          self.metric_accessors)
    return True

  def _get_json(self, file):
    """
    used for testing.
    :return:
    """
    sample = file.read()
    return StringIO(sample)

  def is_ready_to_report(self, conf, pid):
    if pid is None:
      raise Exception("Pid cannot be None")

    sample_file = self._adapter.get_metrics_file(conf)
    if sample_file is None:
      return AbstractMetricCollector._COLLECTOR_NOT_SUPPORTED

    if self._process_start_time is None:
      process = psutil.Process(pid)
      self._process_start_time = process.create_time()
      if self._process_start_time is None:
        return False

    try:
      if os.path.exists(sample_file) and os.path.isfile(sample_file):
        file_modified_time = os.path.getmtime(sample_file)
        if file_modified_time > self._process_start_time:
          # this ensures that we collect this or a newer sample.
          self._last_sample_file_collected_modified_time = \
            self._process_start_time
          return True
    except Exception, ex:
      pass
    return False
