# Copyright (c) 2015 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging
import os

from cmf.monitor import schema
from cmf.monitor.generic.adapter import Adapter
from cmf.monitor.generic.service_defined_metrics import ServiceDefinedMetrics
from cmf.monitor.generic.utils import SplittingSourceProcessor
from cmf.monitor.jmx_beans import JsonParserJmx
from cmf.monitor.json_metrics_utils import add_derived_jvm_metrics_jmx
from cmf.monitor.json_metrics_utils import add_derived_gc_metrics_jmx
from cmf.monitor.json_metrics_utils import add_metrics_jmx
from cmf.monitor.json_metrics_utils import add_pause_metrics

LOG = logging.getLogger('MgmtAdapters')

class _MgmtAdapter(Adapter):
  """
  Base class for all MGMT service role adapters. This exists to provide utility
  methods and to ensure consistency in the handling of common metrics.
  """
  _USE_SSL = 'use_ssl'

  def __init__(self, role_type, safety_valve, port_key=None):
    Adapter.__init__(self, "MGMT", role_type, safety_valve)
    self._port_key = port_key
    self._metrics = None

  def get_metrics_url(self, conf):
    if conf is None:
      raise Exception("A configuration is required!")
    if self._port_key is None:
      return None
    try:
      host = conf.get(self.section, Adapter._HOST)
      if host is None:
        LOG.error("%s entry missing from %s" % (Adapter._HOST,
                                                os.path.abspath(conf.path)))
        return None
      port = conf.getint(self.section, self._port_key)
      if port is None:
        LOG.error("%s entry missing from %s" % (self._port_key,
                                                os.path.abspath(conf.path)))
        return None
      ssl_enabled = conf.getboolean_with_default(self.section,
                                                 _MgmtAdapter._USE_SSL, False)
      return self.format_metrics_url(host, port, ssl_enabled)
    except:
      LOG.exception("Error parsing %s" % os.path.abspath(conf.path))
      return None

  def format_metrics_url(self, host, port, ssl_enabled):
      if ssl_enabled:
        return "https://%s:%s/jmx" % (host, port)
      else:
        return "http://%s:%s/jmx" % (host, port)

  def read_service_defined_metrics(self, path):
    if path is None:
      raise Exception("A path is required!")
    self._metrics = ServiceDefinedMetrics(path, SplittingSourceProcessor('::'))

    # To make metric extraction more efficient, we take the metric sources for
    # this role and create a map from <bean name> to a list of (<metric id>, <field>).
    self._sources_by_name = dict()
    sources = self._metrics.get_sources(self.role_type)
    for metricid, source in sources.iteritems():
      name, field = source
      if name not in self._sources_by_name:
          self._sources_by_name[name] = list()
      self._sources_by_name[name].append((metricid, field))

  def parse_metrics_from_url(self, conf, json):
    if json is None:
      raise Exception("A json sample is required!")

    # Parse out the role metrics. Note that we have to tolerate this file not
    # being present because of upgrade cases.
    extracted_metrics = dict()
    if self._metrics is not None:
      for bean in json.get('beans', []):
        name = bean.get('name', None)
        if name is None:
          continue
        if name not in self._sources_by_name:
          continue
        metrics_list = self._sources_by_name.get(name)
        for metricid, field in metrics_list:
          value = bean.get(field, None)
          if value is None:
            continue
          extracted_metrics[metricid] = float(value)

    return (JsonParserJmx(json, self.role_type), extracted_metrics)

  def add_webserver_metrics(self, version, update, metrics, accessors):
    json_jmx_parser, extracted_metrics = metrics

    add_derived_jvm_metrics_jmx(update, json_jmx_parser, True, False)
    add_derived_gc_metrics_jmx(update, json_jmx_parser)
    add_pause_metrics(update, json_jmx_parser)

    for metricid, value in extracted_metrics.iteritems():
      update.add_metric(metricid, value)

class _FirehoseAdapter(_MgmtAdapter):
  """
  Base class for the three firehose-based MGMT roles.
  """
  _FIREHOSE_DEBUG_PORT = "firehose_debug_port"
  # Added in CM 5.0
  _LDB_DIR = "ldb_dir"

  def __init__(self, role_type, safety_valve, uses_leveldb=True):
    _MgmtAdapter.__init__(self,
                          role_type,
                          safety_valve,
                          _FirehoseAdapter._FIREHOSE_DEBUG_PORT)
    self._uses_leveldb = uses_leveldb

  def _get_monitored_directories_conf_keys(self):
    keys = _MgmtAdapter._get_monitored_directories_conf_keys(self)
    if self._uses_leveldb:
      keys.append(_FirehoseAdapter._LDB_DIR)
    return keys

  def add_webserver_metrics(self, version, update, metrics, accessors):
    _MgmtAdapter.add_webserver_metrics(
      self, version, update, metrics, accessors)

    json_jmx_parser, _ = metrics

    if hasattr(json_jmx_parser, "event_stats"):
      update.add_metric(schema.EVENT_PUBLISHER_SEND_QUEUE_SIZE,
                        json_jmx_parser.event_stats.SendQueueSize)
      update.add_metric(schema.EVENT_PUBLISHER_DROP_COUNT,
                        json_jmx_parser.event_stats.DropCount)
      update.add_metric(schema.EVENT_PUBLISHER_EXCEPTION_COUNT,
                        json_jmx_parser.event_stats.ExceptionCount)
      update.add_metric(schema.EVENT_PUBLISHER_PUSHBACK_COUNT,
                        json_jmx_parser.event_stats.PushbackCount)
      update.add_metric(schema.EVENT_PUBLISHER_SENT_COUNT,
                        json_jmx_parser.event_stats.SentCount)


class ServiceMonitorAdapter(_FirehoseAdapter):
  def __init__(self, safety_valve):
    _FirehoseAdapter.__init__(self, "SERVICEMONITOR", safety_valve)

  def add_webserver_metrics(self, version, update, metrics, accessors):
    _FirehoseAdapter.add_webserver_metrics(
      self, version, update, metrics, accessors)

    json_jmx_parser, _ = metrics

    if json_jmx_parser.role_stage is not None:
      update.add_metric(schema.SMON_ROLE_DROPS,
                        json_jmx_parser.role_stage.EventsDropped +
                        json_jmx_parser.role_stage.EventsSkipped)
      update.add_metric(schema.SMON_ROLE_PROCESSED,
                        json_jmx_parser.role_stage.EventsProcessed)
      update.add_metric(schema.SMON_ROLE_QUEUE_SIZE,
                        json_jmx_parser.role_stage.InputQueueSize)

  def get_metrics_sample(self, version):
    return open(os.path.join(os.path.dirname(__file__), "../samples/mgmt_smon"), 'r').read()

class HostMonitorAdapter(_FirehoseAdapter):
  def __init__(self, safety_valve):
    _FirehoseAdapter.__init__(self, "HOSTMONITOR", safety_valve)

  def add_webserver_metrics(self, version, update, metrics, accessors):
    _FirehoseAdapter.add_webserver_metrics(
      self, version, update, metrics, accessors)

    json_jmx_parser, _ = metrics

    if json_jmx_parser.host_stage is not None:
      update.add_metric(schema.HMON_HOST_DROPS,
                        json_jmx_parser.host_stage.EventsDropped +
                        json_jmx_parser.host_stage.EventsSkipped)
      update.add_metric(schema.HMON_HOST_PROCESSED,
                        json_jmx_parser.host_stage.EventsProcessed)
      update.add_metric(schema.HMON_HOST_QUEUE_SIZE,
                        json_jmx_parser.host_stage.InputQueueSize)

  def get_metrics_sample(self, version):
    return open(os.path.join(os.path.dirname(__file__), "../samples/mgmt_hmon"), 'r').read()

class ActivityMonitorAdapter(_FirehoseAdapter):
  def __init__(self, safety_valve):
    _FirehoseAdapter.__init__(self,
                              "ACTIVITYMONITOR",
                              safety_valve,
                              uses_leveldb=False)

  def add_webserver_metrics(self, version, update, metrics, accessors):
    _FirehoseAdapter.add_webserver_metrics(
      self, version, update, metrics, accessors)

    json_jmx_parser, _ = metrics

    if json_jmx_parser.activity_tree_pipeline is not None:
      update.add_metric(schema.AMON_ACTIVITY_TREE_DROPS,
                        json_jmx_parser.activity_tree_pipeline.EventsDropped +
                        json_jmx_parser.activity_tree_pipeline.EventsSkipped)
      update.add_metric(schema.AMON_ACTIVITY_TREE_PROCESSED,
                        json_jmx_parser.activity_tree_pipeline.EventsProcessed)
      update.add_metric(schema.AMON_ACTIVITY_TREE_QUEUE_SIZE,
                        json_jmx_parser.activity_tree_pipeline.InputQueueSize)

    if json_jmx_parser.activity_monitor_pipeline is not None:
      update.add_metric(schema.AMON_ACTIVITY_MONITOR_DROPS,
                        json_jmx_parser.activity_monitor_pipeline.EventsDropped +
                        json_jmx_parser.activity_monitor_pipeline.EventsSkipped)
      update.add_metric(schema.AMON_ACTIVITY_MONITOR_PROCESSED,
                        json_jmx_parser.activity_monitor_pipeline.EventsProcessed)
      update.add_metric(schema.AMON_ACTIVITY_MONITOR_QUEUE_SIZE,
                        json_jmx_parser.activity_monitor_pipeline.InputQueueSize)

    if json_jmx_parser.activity_monitor_purger is not None:
      update.add_metric(schema.AMON_PURGER_SERVICE_RUNS,
                        json_jmx_parser.activity_monitor_purger.RunCount)
      update.add_metric(schema.AMON_PURGER_SERVICE_SLOW_RUNS,
                        json_jmx_parser.activity_monitor_purger.SlowRunCount)
      update.add_metric(schema.AMON_PURGER_SERVICE_RUNS_WITH_EXCEPTION,
                        json_jmx_parser.activity_monitor_purger.RunWithExceptionCount)

  def get_metrics_sample(self, version):
    return open(os.path.join(os.path.dirname(__file__), "../samples/mgmt_amon"), 'r').read()

class EventServerAdapter(_MgmtAdapter):
  _EVENTSERVER_DEBUG_PORT = "eventserver_debug_port"
   # Added in CM 4.1
  _INDEX_DIR = "index_dir"

  def __init__(self, safety_valve):
    _MgmtAdapter.__init__(self,
                          "EVENTSERVER",
                          safety_valve,
                          EventServerAdapter._EVENTSERVER_DEBUG_PORT)

  def _get_monitored_directories_conf_keys(self):
    keys = Adapter._get_monitored_directories_conf_keys(self)
    keys.append(EventServerAdapter._INDEX_DIR)
    return keys

  def add_webserver_metrics(self, version, update, metrics, accessors):
    _MgmtAdapter.add_webserver_metrics(
      self, version, update, metrics, accessors)

    json_jmx_parser, _ = metrics

    if json_jmx_parser.event_server is not None:
      update.add_metric(schema.EVENTS_STORED,
                        json_jmx_parser.event_server.EventCount)
      update.add_metric(schema.EVENT_CAPACITY,
                        json_jmx_parser.event_server.MaxEventCapacity)

    if json_jmx_parser.event_pipeline is not None:
      update.add_metric(schema.EVENT_WRITER_DROPS,
                        json_jmx_parser.event_pipeline.EventsDropped +
                        json_jmx_parser.event_pipeline.EventsSkipped)
      update.add_metric(schema.EVENT_WRITER_PROCESSED,
                        json_jmx_parser.event_pipeline.EventsProcessed)
      update.add_metric(schema.EVENT_WRITER_QUEUE_SIZE,
                        json_jmx_parser.event_pipeline.InputQueueSize)

  def get_metrics_sample(self, version):
    return open(os.path.join(os.path.dirname(__file__), "../samples/mgmt_events"), 'r').read()

class ReportsManagerAdapter(_MgmtAdapter):
   # Added in CM 4.1
  _SCRATCH_DIR = "scratch_dir"
   # Added in CM 5.5
  _HEADLAMP_DEBUG_PORT = "headlamp_debug_port"

  def __init__(self, safety_valve):
    _MgmtAdapter.__init__(self,
                          "REPORTSMANAGER",
                          safety_valve,
                          ReportsManagerAdapter._HEADLAMP_DEBUG_PORT)

  def _get_monitored_directories_conf_keys(self):
    keys = Adapter._get_monitored_directories_conf_keys(self)
    keys.append(ReportsManagerAdapter._SCRATCH_DIR)
    return keys

  def get_metrics_sample(self, version):
    return open(os.path.join(os.path.dirname(__file__), "../samples/mgmt_rman"), 'r').read()

class NavigatorAdapter(_MgmtAdapter):
  _NAVIGATOR_DEBUG_PORT = "navigator_debug_port"

  def __init__(self, safety_valve):
    _MgmtAdapter.__init__(self,
                          "NAVIGATOR",
                          safety_valve,
                          NavigatorAdapter._NAVIGATOR_DEBUG_PORT)

  def add_webserver_metrics(self, version, update, metrics, accessors):
    _MgmtAdapter.add_webserver_metrics(
      self, version, update, metrics, accessors)

    json_jmx_parser, _ = metrics

    if json_jmx_parser.navigator_counters is not None:
      update.add_metric(schema.READ_REQUESTS,
                        json_jmx_parser.navigator_counters.ReadRequests)
      update.add_metric(schema.WRITE_REQUESTS,
                        json_jmx_parser.navigator_counters.WriteRequests)
      update.add_metric(schema.ACTIVE_READ_REQUESTS,
                        json_jmx_parser.navigator_counters.ActiveReadRequests)
      update.add_metric(schema.ACTIVE_WRITE_REQUESTS,
                        json_jmx_parser.navigator_counters.ActiveWriteRequests)
      update.add_metric(schema.PERSISTED_REQUESTS,
                        json_jmx_parser.navigator_counters.EventsInsertedCount)

  def get_metrics_sample(self, version):
    return open(os.path.join(os.path.dirname(__file__), "../samples/mgmt_nav"), 'r').read()

class NavigatorMetaServerAdapter(_MgmtAdapter):
  # The main port of NMS server also serves debug page
  _NMS_DEBUG_PORT = "navigator_server_port"

  def __init__(self, safety_valve):
    _MgmtAdapter.__init__(self,
                          "NAVIGATORMETASERVER",
                          safety_valve,
                          NavigatorMetaServerAdapter._NMS_DEBUG_PORT)

  def add_webserver_metrics(self, version, update, metrics, accessors):
      _MgmtAdapter.add_webserver_metrics(
          self, version, update, metrics, accessors)

      json_jmx_parser, _ = metrics
      add_metrics_jmx(update, json_jmx_parser, accessors, "NAVIGATORMETASERVER", "enterprise")

  def format_metrics_url(self, host, port, ssl_enabled):
    """
    NMS server hosts jmx page at the main web server's 'jmx' endpoint. So we
    override this method to return correct url for NMS server.
    """
    if ssl_enabled:
      return "https://%s:%s/debug/jmx" % (host, port)
    else:
      return "http://%s:%s/debug/jmx" % (host, port)

  def get_metrics_sample(self, version):
    return open(os.path.join(os.path.dirname(__file__), "../samples/mgmt_nms"), 'r').read()

class TelemetryPublisherAdapter(_MgmtAdapter):
  _SCRATCH_DIR = "scratch_dir"
  _TP_DEBUG_PORT = "telemetry_publisher_debug_port"

  def __init__(self, safety_valve):
    _MgmtAdapter.__init__(self,
                          "TELEMETRYPUBLISHER",
                          safety_valve,
                          TelemetryPublisherAdapter._TP_DEBUG_PORT)

  def add_webserver_metrics(self, version, update, metrics, accessors):
    _MgmtAdapter.add_webserver_metrics(
      self, version, update, metrics, accessors)

    json_jmx_parser, _ = metrics
    add_metrics_jmx(update, json_jmx_parser, accessors, "TELEMETRYPUBLISHER", "enterprise")

  def _get_monitored_directories_conf_keys(self):
    keys = Adapter._get_monitored_directories_conf_keys(self)
    keys.append(TelemetryPublisherAdapter._SCRATCH_DIR)
    return keys

  def get_metrics_sample(self, version):
    return open(os.path.join(os.path.dirname(__file__), "../samples/mgmt_tp"), 'r').read()