# Copyright (c) 2011-2012 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

"""
The host monitor
"""

import logging
import os
import psutil
import simplejson as json
import time

from cmf.monitor import constants
from cmf.monitor import schema
from cmf.monitor.abstract_monitor import AbstractMonitor
from cmf.monitor.host.filesystem_map import FilesystemMap
from cmf.monitor.host_update import HostUpdate
from cmf.monitor.proc_metrics_utils import get_swap_out, get_host_fd_stats
from cmf.monitor.proc_metrics_utils import get_overcommit_values
from cmf.monitor.wakeable_thread import WakeableThread
from cmf.monitor.host.network_interfaces import NetworkInterfaces
from cmf.monitor.host.disk_devices import DiskDevices
from cmf.monitor.host.dns_names import DnsNames
from cmf.monitor.host.ntp_monitor import NTPMonitor
from cmf.monitor.host.tcp_metrics import TcpMetrics
from cmf.monitor.host.certificate_metrics import CertificateMetrics
from cmf.monitor.host.filesystems import Filesystems
from cmf.monitor.host.entropy_monitor import EntropyMonitor
from cmf.monitor.proc_metrics_utils import get_fd_count, get_fd_max
from cmf.throttling_logger import ThrottlingLogger

LOG = logging.getLogger(__name__)
THROTTLED_LOG = ThrottlingLogger(LOG)

class HostMonitor(AbstractMonitor):

  def __init__(self, daemon, registry):
    AbstractMonitor.__init__(self, daemon, registry, None, None, None)
    self._filesystems = Filesystems(daemon.get_filesystem_map())
    self._disk_devices = DiskDevices()
    self._tcp_metrics = TcpMetrics()
    self._certificate_metrics = CertificateMetrics(self._agent.cfg)
    self._entropy_monitor = EntropyMonitor()
    self._supervisor_successes = 0
    self._supervisor_latency_sec = 0.0
    try:
      self._network_interfaces = NetworkInterfaces()
    except IOError, ioex:
      LOG.warning("Unable to create network interface stats reader, "
                  "agent will not report NIC metrics. errno=%s" % (ioex.errno,))
      self._network_interfaces = None
    self._host_dir_monitors = [ ]
    self.create_dir_monitor(
      os.path.dirname(self._agent.cfg.logfile), self._host_dir_monitors)
    self.create_dir_monitor(
      self._agent.cfg.paths.process_dir, self._host_dir_monitors)
    # We can't know the parcel directory until after the first hearteat. So we
    # cannot create the monitor yet. See OPSAPS-26672.
    self._parcel_dir_monitor_created = False

    self._ntp_monitor = NTPMonitor()
    self._monitor_ntp = False

    self._dns_names = None
    try:
      agent_jar_path = "%s/lib/agent-%s.jar" % \
          (self._agent.supervisee_environment["MGMT_HOME"],
           constants.AGENT_VERSION)
      if os.path.exists(agent_jar_path):
        self._dns_names = DnsNames(
          self._agent.cfg.package_dir,
          agent_jar_path,
          poll_timeout=self._agent.cfg.dns_resolution_collection_timeout_seconds,
          polling_interval=self._agent.cfg.dns_resolution_collection_timeout_seconds)
        LOG.info("Created DNS monitor.")
      else:
        LOG.error("%s not found." % agent_jar_path)
    except:
      LOG.exception("Unable to create DNS monitor")

  def __str__(self):
    return "HostMonitor"

  def make_collection_thread(self):
    return WakeableThread(
          "Monitor-%s" % (self,),
          self.collect)

  def collect(self):
    """The loop content in the collection thread."""
    self._filesystems.collect()
    self._disk_devices.collect()
    if self._network_interfaces is not None:
      self._network_interfaces.collect()
    # Create the parcel directory monitor as soon as we notice that the parcel
    # directory has been set. This should happen after the first hearbeat.
    if (self._agent.parcel_dir_is_set and
        not self._parcel_dir_monitor_created):
      self.create_dir_monitor(
        self._agent.cfg.parcel_dir, self._host_dir_monitors)
      self._parcel_dir_monitor_created = True
    for mon in self._host_dir_monitors:
      mon.collect()
    if self._monitor_ntp and self._ntp_monitor is not None:
      self._ntp_monitor.collect()

  def ensure_collecting(self):
    AbstractMonitor.ensure_collecting(self)
    if self._dns_names is not None:
      self._dns_names.start()

  def stop(self):
    AbstractMonitor.stop(self)
    if self._dns_names is not None:
      self._dns_names.stop()

  def get_host_update(self):
    """Returns a host update to be sent to the firehose"""
    agent_logfile_dir = os.path.dirname(self._agent.cfg.logfile)
    agent_process_dir = self._agent.cfg.paths.process_dir
    # We include the parcel directory in the HostUpdate only if it is in use. If
    # this is not set, then the host's parcel directry free space checks will be
    # disabled.
    agent_parcel_dir = None
    if self._agent.parcel_directory_in_use():
      agent_parcel_dir = self._agent.cfg.parcel_dir
    update = HostUpdate(agent_logfile_dir, agent_process_dir, agent_parcel_dir)

    for fs_update in self._filesystems.get_fs_updates():
      update.add_fs_update(fs_update)

    for disk_update in self._disk_devices.get_disk_updates():
      update.add_disk_update(disk_update)

    if self._network_interfaces is not None:
      for nic_update in self._network_interfaces.get_nic_updates():
        update.add_nic_update(nic_update)

    for mon in self._host_dir_monitors:
      update.add_dir_update(mon.get_host_dir_update())

    update.add_metric(schema.CPU_PERCENT, psutil.cpu_percent(interval=None))
    # Record the numbre of cores. This includes hyperthreads
    update.add_metric(schema.CORES, psutil.cpu_count(logical=True))

    load_avg = os.getloadavg()
    update.add_metric(schema.LOAD_1, load_avg[0])
    update.add_metric(schema.LOAD_5, load_avg[1])
    update.add_metric(schema.LOAD_15, load_avg[2])

    cpu_time = psutil.cpu_times()
    update.add_metric(schema.TOTAL_CPU_USER, cpu_time.user)
    update.add_metric(schema.TOTAL_CPU_SYSTEM, cpu_time.system)
    update.add_metric(schema.TOTAL_CPU_IDLE, cpu_time.idle)
    update.add_metric(schema.TOTAL_CPU_NICE, cpu_time.nice)
    update.add_metric(schema.TOTAL_CPU_IOWAIT, cpu_time.iowait)
    update.add_metric(schema.TOTAL_CPU_IRQ, cpu_time.irq)
    update.add_metric(schema.TOTAL_CPU_SOFT_IRQ, cpu_time.softirq)
    update.add_metric(schema.TOTAL_CPU_STEAL, getattr(cpu_time, 'steal', None))
    update.add_metric(schema.TOTAL_CPU_GUEST, getattr(cpu_time, 'guest', None))
    update.add_metric(schema.TOTAL_CPU_GUEST_NICE,
                      getattr(cpu_time, 'guest_nice', None))

    agent_process = psutil.Process(os.getpid())
    agent_cpu_time = agent_process.cpu_times()
    if agent_cpu_time is not None:
      update.add_metric(schema.AGENT_CPU_USER, agent_cpu_time.user)
      update.add_metric(schema.AGENT_CPU_SYSTEM, agent_cpu_time.system)
    rss, vms = agent_process.memory_info()[0:2]
    update.add_metric(schema.AGENT_PHYSICAL_MEMORY_USED, rss)
    update.add_metric(schema.AGENT_VIRTUAL_MEMORY_USED, vms)
    update.add_metric(schema.AGENT_FD_OPEN, get_fd_count(os.getpid()))
    update.add_metric(schema.AGENT_FD_MAX, get_fd_max(os.getpid()))
    update.add_metric(schema.AGENT_HB_LATENCY_MILLIS,
                      self._agent.heartbeat_tracker.get_last_hb_latency_secs() * 1000)

    for firehose in ["HOSTMONITOR", "SERVICEMONITOR"]:
      metrics = self.daemon.firehoses.get_firehose_metrics(firehose)
      for metric_id, metric_value in metrics.iteritems():
        update.add_metric(metric_id, metric_value)

    uptime = time.time() - psutil.boot_time()
    if uptime >= 0:
      update.add_metric(schema.UPTIME, uptime)
    else:
      THROTTLED_LOG.warning("Negative host uptime observed. Not reporting "
                            "uptime to Host Monitor.")

    try:
      supervisord_process = self._agent.supervisor_client.psutil_process
      supervisord_pid = self._agent.supervisor_client.pid
      supervisord_cpu_time = supervisord_process.cpu_times()
      if supervisord_cpu_time is not None:
        update.add_metric(schema.SUPERVISORD_CPU_USER, supervisord_cpu_time.user)
        update.add_metric(schema.SUPERVISORD_CPU_SYSTEM, supervisord_cpu_time.system)
      rss, vms = supervisord_process.memory_info()[0:2]
      update.add_metric(schema.SUPERVISORD_PHYSICAL_MEMORY_USED, rss)
      update.add_metric(schema.SUPERVISORD_VIRTUAL_MEMORY_USED, vms)
      update.add_metric(schema.SUPERVISORD_FD_OPEN, get_fd_count(supervisord_pid))
      update.add_metric(schema.SUPERVISORD_FD_MAX, get_fd_max(supervisord_pid))
    except Exception, e:
      LOG.info("Couldn't get supervisord metrics: %s" % (e,))

    supervisor_statistics = self._agent.supervisor_client.stats
    supervisor_successes = supervisor_statistics["successes"]
    supervisor_successes_delta = supervisor_successes - self._supervisor_successes
    self._supervisor_successes = supervisor_successes
    supervisor_latency_sec = supervisor_statistics["latency_sec"]
    supervisor_latency_delta = supervisor_latency_sec - self._supervisor_latency_sec
    self._supervisor_latency_sec = supervisor_latency_sec
    if supervisor_successes_delta > 0:
      update.add_metric(schema.SUPERVISORD_LATENCY, supervisor_latency_delta / supervisor_successes_delta);
    update.add_metric(schema.SUPERVISORD_FAILURES, supervisor_statistics["failures"]);

    vmem = psutil.virtual_memory()

    update.add_metric(schema.PHYSICAL_MEMORY_USED, vmem.used)
    update.add_metric(schema.PHYSICAL_MEMORY_CACHED, vmem.cached)
    update.add_metric(schema.PHYSICAL_MEMORY_BUFFERS, vmem.buffers)
    update.add_metric(schema.PHYSICAL_MEMORY_TOTAL, vmem.total)
    update.add_metric(schema.PHYSICAL_MEMORY_DIRTY, vmem.dirty)
    update.add_metric(schema.PHYSICAL_MEMORY_WRITEBACK, vmem.writeback)
    update.add_metric(schema.PHYSICAL_MEMORY_MEMFREE, vmem.free)
    update.add_metric(schema.PHYSICAL_MEMORY_MAPPED, vmem.mapped)
    update.add_metric(schema.PHYSICAL_MEMORY_DIRTY_RATIO, vmem.dirty_ratio_percent)
    update.add_metric(schema.PHYSICAL_MEMORY_COMMIT_LIMIT, vmem.commit_limit)
    swap_usage = psutil.swap_memory()
    update.add_metric(schema.SWAP_TOTAL, swap_usage.total)
    update.add_metric(schema.SWAP_FREE, swap_usage.free)
    update.add_metric(schema.SWAP_USED, swap_usage.used)
    swap_out = get_swap_out()
    if swap_out is not None:
      update.add_metric(schema.SWAP_OUT, swap_out)
    overcommit_memory, overcommit_ratio = get_overcommit_values()
    update.set_overcommit_memory_mode(overcommit_memory)
    update.add_metric(schema.OVERCOMMIT_RATIO, overcommit_ratio)
    fd_open, fd_max = get_host_fd_stats(LOG)
    update.add_metric(schema.FD_OPEN, fd_open)
    update.add_metric(schema.FD_MAX, fd_max)

    self._tcp_metrics.add_tcp_metrics(update)
    self._certificate_metrics.add_certificate_metrics(update)
    self._entropy_monitor.add_entropy_metrics(update)

    if self._dns_names is not None:
      self._dns_names.add_dns_metrics(update)

    if self._monitor_ntp and self._ntp_monitor is not None:
      self._ntp_monitor.add_ntp_metrics(update)

    return update

  def start_collecting(self, latch):
    if self._collection_thread is not None:
      self._collection_thread.wake(latch)

  def update_host_config(self, host_config):
    if (host_config.has_key('host_network_interface_collection_filter')):
      if self._network_interfaces is not None:
        self._network_interfaces.set_iface_exclude_regex(
             host_config['host_network_interface_collection_filter'])
    if (host_config.has_key('host_disk_collection_filter')):
      self._disk_devices.set_device_exclude_regex(
           host_config['host_disk_collection_filter'])
    if (host_config.has_key('host_fs_collection_filter')):
      self._filesystems.set_fs_exclude_regex(
           host_config['host_fs_collection_filter'])
    self._monitor_ntp = self.update_monitor_ntp(host_config)
    if self._dns_names is not None:
      self._dns_names.update_monitor_dns(host_config)

  def is_ready_to_report(self):
    return True

  def update_monitor_ntp(self, host_config):
    clock_offset_enabled = False
    if host_config.has_key('host_clock_offset_thresholds'):
      try:
        thresholds = json.loads(host_config['host_clock_offset_thresholds'])
        clock_offset_enabled = (thresholds['warning'] != constants.NEVER or
                                thresholds['critical'] != constants.NEVER)
      except:
        THROTTLED_LOG.info("Unable to parse host_clock_offset_thresholds : %s" \
                           % host_config['host_clock_offset_thresholds'])
    return clock_offset_enabled
