# Copyright (c) 2012 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

'''
Block device stat file format
$ cat /sys/block/<dev>/stat
21000724  1219997 701701450 14862220 17769848 12443731 881114416 69933960        0  8640776 84785344

Device can be an entire disk (e.g. "/sys/block/sda") or
a single partition (e.g. "/sys/block/sda/sda1")

read I/Os       requests      number of read I/Os processed
read merges     requests      number of read I/Os merged with in-queue I/O
read sectors    sectors       number of sectors read
read ticks      milliseconds  total wait time for read requests
write I/Os      requests      number of write I/Os processed
write merges    requests      number of write I/Os merged with in-queue I/O
write sectors   sectors       number of sectors written
write ticks     milliseconds  total wait time for write requests
in_flight       requests      number of I/Os currently in flight
io_ticks        milliseconds  total time this block device has been active (had requests queued)
time_in_queue   milliseconds  total wait time for all requests
(from Documentation/block/stat.txt)
'''

''' As described in the kernel Documentation/block/stat.txt, "the "sectors" in question are
    the standard UNIX 512-byte sectors, not any device- or filesystem-specific block size."
'''
UNIX_SECTOR_SIZE = 512
LONG_MAX_VALUE = (1 << 63) - 1

class BlockDeviceStats(object):
  """ Represents the stats available for block devices in /sys/<device>/stat """
  def __init__(self, stats = None):
    if stats is None:
      stats = [0] * 11

    self._read_ios = stats[0]
    self._read_merges = stats[1]
    self._read_sectors = stats[2]
    self._read_ticks = stats[3]
    self._write_ios = stats[4]
    self._write_merges = stats[5]
    self._write_sectors = stats[6]
    self._write_ticks = stats[7]
    self._in_flight = stats[8]
    # io ticks represents the total time the device has had requests _queued_
    # but it is NOT the sum of read_ticks and write_ticks, which are the times
    # that read/write requests have _waited_ on the device
    self._io_ticks = stats[9]
    self._time_in_queue = stats[10]

  def get_read_ios(self):
    return self._read_ios

  def get_read_merges(self):
    return self._read_merges

  def get_read_sectors(self):
    return self._read_sectors

  def get_read_ticks(self):
    return self._read_ticks

  def get_write_ios(self):
    return self._write_ios

  def get_write_merges(self):
    return self._write_merges

  def get_write_sectors(self):
    return self._write_sectors

  def get_write_ticks(self):
    return self._write_ticks

  def get_in_flight(self):
    return self._in_flight

  def get_io_ticks(self):
    return self._io_ticks

  def get_time_in_queue(self):
    return self._time_in_queue

  def get_rw_wait_ticks(self):
    return self._read_ticks + self._write_ticks

  def get_nr_ios(self):
    return self._read_ios + self._write_ios

  def get_read_bytes(self):
    # "UNIX sector size" is always 512 bytes, see comment above
    num_bytes = self._read_sectors * UNIX_SECTOR_SIZE
    # Python longs are unbounded, so don't report a value if it will overflow in Java
    if num_bytes <= LONG_MAX_VALUE:
      return num_bytes
    else:
      return None

  def get_write_bytes(self):
    # "UNIX sector size" is always 512 bytes, see comment above
    num_bytes = self._write_sectors * UNIX_SECTOR_SIZE
    # Python longs are unbounded, so don't report a value if it will overflow in Java
    if num_bytes <= LONG_MAX_VALUE:
      return num_bytes
    else:
      return None

  # Need to override __eq__ and __ne__ to check equality, used in unit tests
  def __eq__(self, other):
    if type(other) is type(self):
      return self.__dict__ == other.__dict__
    else:
      return False

  def __ne__(self, other):
    return not self.__eq__(other)
