# Copyright (c) 2012-2013 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging
import os

from cmf.monitor import schema
from cmf.monitor.disk_update import DiskUpdate
from cmf.monitor.host.block_device_stats import BlockDeviceStats
from cmf.throttling_logger import ThrottlingLogger

LOG = logging.getLogger(__name__)
THROTTLED_LOG = ThrottlingLogger(LOG)

def get_disk_update(device_name, curr_stats, prev_stats, delta_millis):
  """Returns a disk update for the disk device"""
  await, await_r, await_w = calculate_await_times(curr_stats, prev_stats)
  service_time = calculate_service_time(curr_stats, prev_stats)
  average_request_size = calculate_average_request_size(curr_stats,
                                                        prev_stats)
  average_queue_length = calculate_average_queue_length(curr_stats,
                                                        prev_stats,
                                                        delta_millis)
  utilization = calculate_utilization(curr_stats, prev_stats, delta_millis)

  partition = False
  try:
    partition = not os.path.exists('/sys/block/%s' % device_name)
  except Exception, ex:
    THROTTLED_LOG.warning("Error testing '%s': %s"
                          % (device_name, ex))

  update = DiskUpdate(device_name, partition)
  update.add_metric(schema.AWAIT_TIME, await)
  update.add_metric(schema.AWAIT_READ_TIME, await_r)
  update.add_metric(schema.AWAIT_WRITE_TIME, await_w)
  update.add_metric(schema.SERVICE_TIME, service_time)
  update.add_metric(schema.READ_BYTES, curr_stats.get_read_bytes())
  update.add_metric(schema.WRITE_BYTES, curr_stats.get_write_bytes())
  update.add_metric(schema.READ_IOS, curr_stats.get_read_ios())
  update.add_metric(schema.WRITE_IOS, curr_stats.get_write_ios())
  update.add_metric(schema.AVERAGE_QUEUE_LENGTH, average_queue_length)
  update.add_metric(schema.AVERAGE_REQUEST_SIZE, average_request_size)
  update.add_metric(schema.UTILIZATION, utilization)
  return update

def calculate_await_times(curr_stats, prev_stats):
  """Returns a tuple containing the read and write await times"""

  await_r_time = None
  new_read_ios = curr_stats.get_read_ios() - prev_stats.get_read_ios()
  new_read_ticks = curr_stats.get_read_ticks() - prev_stats.get_read_ticks()
  if new_read_ios == 0:
    await_r_time = 0
  elif new_read_ios > 0 and new_read_ticks > 0:
    await_r_time = new_read_ticks / float(new_read_ios)

  await_w_time = None
  new_write_ios = curr_stats.get_write_ios() - prev_stats.get_write_ios()
  new_write_ticks = curr_stats.get_write_ticks() - prev_stats.get_write_ticks()
  if new_write_ios == 0:
    await_w_time = 0
  elif new_write_ios > 0 and new_write_ticks > 0:
    await_w_time = new_write_ticks / float(new_write_ios)

  await_time = None
  if (new_read_ios >= 0 and new_read_ticks >= 0 and
      new_write_ios >= 0 and new_write_ticks >= 0):
    total_ios = new_read_ios + new_write_ios
    total_ticks = new_read_ticks + new_write_ticks
    if total_ios == 0:
      await_time = 0
    elif total_ios > 0 and total_ticks > 0:
      await_time = total_ticks / float(total_ios)

  return (await_time, await_r_time, await_w_time)

def calculate_service_time(curr_stats, prev_stats):
  new_ios = curr_stats.get_nr_ios() - prev_stats.get_nr_ios()
  if new_ios < 0:
    return None
  elif new_ios == 0:
    return 0.0
  delta_io_ticks = curr_stats.get_io_ticks() - prev_stats.get_io_ticks()
  if delta_io_ticks < 0:
    return None
  return float(delta_io_ticks) / float(new_ios)

def calculate_average_request_size(curr_stats, prev_stats):
  new_ios = curr_stats.get_nr_ios() - prev_stats.get_nr_ios()
  if new_ios < 0:
    return None
  elif new_ios == 0:
    return 0.0
  new_sectors_of_io = \
      curr_stats.get_read_sectors() - prev_stats.get_read_sectors() + \
      curr_stats.get_write_sectors() - prev_stats.get_write_sectors()
  if new_sectors_of_io < 0:
    return None
  return float(new_sectors_of_io) / float(new_ios)

def calculate_average_queue_length(curr_stats, prev_stats, delta_millis):
  if delta_millis <= 0:
    return None
  delta_time_in_queue = \
      curr_stats.get_time_in_queue() - prev_stats.get_time_in_queue()
  if delta_time_in_queue < 0:
    return None
  return float(delta_time_in_queue) / float(delta_millis)

def calculate_utilization(curr_stats, prev_stats, delta_millis):
  if delta_millis <= 0:
    return None
  delta_io_ticks = curr_stats.get_io_ticks() - prev_stats.get_io_ticks()
  if delta_io_ticks < 0:
    return None
  return 100.0 * float(delta_io_ticks) / float(delta_millis)
