# Copyright (c) 2012-2013 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging
import time
import re

from psutil import cpu_times

from cmf.monitor.host.disk_device_utils import get_disk_update

from cmf.throttling_logger import ThrottlingLogger
from cmf.monitor.host.block_device_stats import BlockDeviceStats

LOG = logging.getLogger(__name__)
THROTTLED_LOG = ThrottlingLogger(LOG)

class DiskDevices(object):
  _DEVICE_EXCLUDE_PATTERN_DEFAULT = "^$"
  _DEVICE_EXCLUDE_REGEX_DEFAULT = re.compile(_DEVICE_EXCLUDE_PATTERN_DEFAULT)
  _PROC_PARTITIONS = "/proc/partitions"
  _PROC_DISKSTATS = "/proc/diskstats"

  """
  Provides access to all the mounted disk devices on the host
  """
  def __init__(self, partitions_file=_PROC_PARTITIONS, diskstats_file=_PROC_DISKSTATS):
    self._disks = []
    self._latest_updates_pair = ([], None)
    self._last_reported_timestamp = 0
    self._device_exclude_regex = self._DEVICE_EXCLUDE_REGEX_DEFAULT
    self._last_regex_pattern = self._DEVICE_EXCLUDE_PATTERN_DEFAULT

    try:
      self._disk_devices = self.get_disk_devices(partitions_file=partitions_file)
    except Exception, ex:
      self._disk_devices = []
      LOG.warning("Unable to read disk devices to monitor; disk metrics will not"
                  " be collected: %s" % (ex,))

    self._diskstats_file = diskstats_file
    try:
      self._prev_diskstats, self._prev_diskstats_time = \
          self.get_diskstats(self._disk_devices, diskstats_file=diskstats_file)
    except Exception, ex:
      self._prev_diskstats = {}
      self._prev_diskstats_time = self.get_cpu0_uptime()
      LOG.warning("Unable to read disk device statistics for disk monitoring: "
                  "%s" % (ex,))


  """
     Gets the list of disk devices from /proc/partitions. Note that the 'partitions'
     in this file are the names of both disks and partitions, e.g. sda and sda1.

     This code is from psutil.disk_io_counters(), but it is not exposed directly
     so it's provided here. The reason we don't use disk_io_counters() is because
     it doesn't return all stats. (See get_diskstats() below.)

     This method is different from psutil's disk_partitions() method in that
     it returns device names from /proc/partitions instead of /etc/mtab (or
     'mount'). The names are different for LVM-based disk devices and we need
     these names because they are the same device names used in /proc/diskstats.
  
     The optional parameter for the partitions file is exposed only for testing.
  """
  @classmethod
  def get_disk_devices(cls, partitions_file=_PROC_PARTITIONS):
    disk_devices = []
    f = open(partitions_file, "r")
    try:
        lines = f.readlines()[2:]
    finally:
        f.close()
    for line in lines:
        _, _, _, name = line.split()
        # psutil only returns partitions, not disks, so it has the condition:
        # if name[-1].isdigit():
        disk_devices.append(name)
    return disk_devices

  """
     Gets the uptime for CPU0. This is inspired by code in read_stat_cpu and in
     iostat. The uptime of CPU0 is used in various time-based calculations for
     iostat metrics.
     
     As mentioned in iostat source: guest and guest_nice are not included as user
     and nice already include them.
  """
  @classmethod
  def get_cpu0_uptime(self):
    cpu0_cpu_times = cpu_times(percpu=True)[0]._asdict()
    return cpu0_cpu_times.get('user', 0) + \
           cpu0_cpu_times.get('nice', 0) + \
           cpu0_cpu_times.get('system', 0) + \
           cpu0_cpu_times.get('idle', 0) + \
           cpu0_cpu_times.get('iowait', 0) + \
           cpu0_cpu_times.get('irq', 0) + \
           cpu0_cpu_times.get('softirq', 0) + \
           cpu0_cpu_times.get('steal', 0)

  """
     Gets the disk stats for the specified devices from /proc/diskstats.
  
     This code is based on code from psutil.disk_io_counters(), but it does not 
     return all stats and it's simple enough, so we can just re-implement it
     exactly as we want. (More specifically, we need the 'io_ticks' counter from
     diskstats in order to calculate the disk service time.)
  
     The optional parameter for the diskstats file is exposed only for testing.

     Return value is a tuple of map from device name to BlockDeviceStats object
     and the time at which the stats were sampled.
  """
  @classmethod
  def get_diskstats(self, devices, diskstats_file=_PROC_DISKSTATS):
    retdict = {}
    f = open(diskstats_file, "r")
    try:
      lines = f.readlines()
      readtime = self.get_cpu0_uptime()
    finally:
      f.close()

    for line in lines:
      line = line.split()
      name = line[2]
      if name in devices:
        raw_stats = [long(stat) for stat in line[3:14]]
        retdict[name] = BlockDeviceStats(raw_stats)

    return retdict, readtime

  def set_device_exclude_regex(self, regex_pattern):
    if self._last_regex_pattern != regex_pattern:
      self._last_regex_pattern = regex_pattern
      try:
        self._device_exclude_regex = re.compile(regex_pattern)
      except Exception:
        THROTTLED_LOG.warning("Unable to compile disk device exclusion regex "
                              "'%s', using default." % (regex_pattern,))
        self._device_exclude_regex = self._DEVICE_EXCLUDE_REGEX_DEFAULT

  def collect(self):
    # capture the regex as a local variable in case it changes
    exclude_regex = self._device_exclude_regex
    diskstats, diskstats_time = \
        self.get_diskstats(self._disk_devices, diskstats_file=self._diskstats_file)
    delta_millis = (diskstats_time - self._prev_diskstats_time) * 1000
    updates = DiskDevices.build_updates(diskstats,
                                        self._prev_diskstats,
                                        delta_millis,
                                        exclude_regex)
    self._prev_diskstats = diskstats
    self._prev_diskstats_time = diskstats_time
    self._latest_updates_pair = (updates, time.time())

  @classmethod
  def build_updates(cls, diskstats, prev_diskstats, delta_millis, exclude_regex):
    updates = []
    for device_name in diskstats:
      try:
        if not exclude_regex.match(device_name) and prev_diskstats[device_name]:
          curr_stats = diskstats[device_name]
          prev_stats = prev_diskstats[device_name]

          update = get_disk_update(device_name,
                                   curr_stats,
                                   prev_stats,
                                   delta_millis)
          if update is not None and update.contains_metrics():
            updates.append(update)
      except Exception, ex:
        THROTTLED_LOG.warning("Error collecting disk update for '%s': %s"
                              % (device_name, ex))
    return updates

  def get_disk_updates(self):
    (updates, timestamp) = self._latest_updates_pair

    if timestamp == self._last_reported_timestamp:
      return list()

    self._last_reported_timestamp = timestamp
    return updates
