# Copyright (c) 2013 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging
import simplejson as json
import threading
import time

from cmf.monitor import schema, constants
from cmf.monitor.stoppable_thread import StoppableThread
from cmf.subprocess_timeout import subprocess_with_timeout
from cmf.throttling_logger import ThrottlingLogger
import cmf.version_detector
import os.path

LOG = logging.getLogger(__name__)
THROTTLED_LOG = ThrottlingLogger(LOG)

NEVER = "-2.0"

class DnsNames(object):
  """
  The DnsNames gathers metrics about DNS name resolution.
  It checks if the Java host name and canonical name are the same.
  It measures the duration of the Java getLocalHost call.
  It reports the success / failure status of the Java checks.
  """

  # Standard location for resolv_conf on Linux
  DEFAULT_RESOLV_CONF = "/etc/resolv.conf"
  # Minimum seconds between resolv.conf refreshes
  RESOLV_CONF_REFRESH_SECS = 5 * 60

  def __init__(self,
               package_dir,
               agent_jar,
               poll_timeout=30.0,
               polling_interval=60.0):
    LOG.info("Using timeout of %f" % (poll_timeout,))
    self._dns_names = None
    self._package_dir = package_dir
    self._agent_jar = agent_jar
    self._poll_timeout = poll_timeout
    self._resolv_conf_valid_until = 0
    self._monitor_dns = False
    self._collection_thread = StoppableThread("DnsResolutionMonitor",
                                              polling_interval,
                                              self.collect_dns_metrics)
    # Acces to _dns_names should only be done holding this lock.
    self._metrics_lock = threading.Lock()

  def java_location(self):
    return cmf.version_detector.VersionDetector.get_java_location()

  def _agent_jar_path(self):
    return self._agent_jar

  def _subprocess_with_timeout(self, args, timeout):
    return subprocess_with_timeout(args, timeout)

  def start(self):
    self._collection_thread.start()

  def stop(self):
    self._collection_thread.should_stop()

  def collect_dns_metrics(self):
    if not self._monitor_dns:
      return

    self._read_resolv_conf()
    dns_names = dict()
    try:
      java_location = self.java_location()
      if java_location is None:
        THROTTLED_LOG.info("DnsTest not running. Java not located.")
        return
      # Use the exact path to the agent-*.jar intead of the more robust *
      # for the classpath to minimize the Java start up time for the test
      classpath = self._agent_jar_path()
      args = [os.path.join(java_location, "bin/java"),
              "-classpath",
              classpath,
              "com.cloudera.cmon.agent.DnsTest"]
      try:
        result, stdout, stderr = \
          self._subprocess_with_timeout(args, self._poll_timeout)
      except Exception, e:
        THROTTLED_LOG.exception("Failed to run DnsTest.")
        self._make_and_set_dns_error_result(
          constants.DNS_RESOLUTION_STATUS_SUBPROCESS_FAILED)
        return

      if result is not 0:
        THROTTLED_LOG.info("DnsTest return=%d\nstdout=%s\nstderr=%s",
                           result, stdout, stderr)
        self._make_and_set_dns_error_result(
          constants.DNS_RESOLUTION_STATUS_SUBPROCESS_FAILED)
        return

      dns = json.loads(stdout)
      # determine metrics
      if (dns['status'] != '0'):
        dns_names['status'] = constants.DNS_RESOLUTION_STATUS_ERROR
      else:
        dns_names['duration'] = int(dns['localhostDuration'])
        if dns['hostname'] and dns['canonicalname'] \
            and self._match(dns['hostname'].lower(),
              dns['canonicalname'].lower()):
          dns_names['status'] = constants.DNS_RESOLUTION_STATUS_MATCH
        else:
          dns_names['status'] = constants.DNS_RESOLUTION_STATUS_MISMATCH
          THROTTLED_LOG.warning("hostname %s differs from the canonical name %s" \
                                % (dns['hostname'], dns['canonicalname']))
      try:
        self._metrics_lock.acquire()
        self._dns_names = dns_names
      finally:
        self._metrics_lock.release()

    except Exception, e:
      THROTTLED_LOG.exception("Failed to collect java-based DNS names")

  def _make_and_set_dns_error_result(self, result):
    dns_names_result = dict()
    dns_names_result['status'] = result
    try:
      self._metrics_lock.acquire()
      self._dns_names = dns_names_result
    finally:
      self._metrics_lock.release()

  def add_dns_metrics(self, update):
    """Adds DNS metrics to the update"""
    if self._dns_names is not None:
      try:
        self._metrics_lock.acquire()
        dns_names = self._dns_names
        self._dns_names = None
      finally:
        self._metrics_lock.release()

      update.add_metric(schema.DNS_NAME_RESOLUTION_DURATION,
                        dns_names.get('duration', None))
      update.add_metric(schema.DNS_NAME_RESOLUTION_STATUS,
                        dns_names.get('status', None))


  def _read_resolv_conf(self,
                        resolv_conf=DEFAULT_RESOLV_CONF,
                        now=time.time()):
    """Read and set the domain and search information in the resolv_conf file
       Allows resolv_conf path to be passed as an argument for testing
       Please use DEFAULT_RESOLV_CONF if not testing
       Allows now (current time) to be passed as an argument for testing
       Please do not set now if not testing"""
    if self._resolv_conf_valid_until > now:
      return
    domain = None
    search = []
    list_of_lines = self._read_resolv_conf_contents(resolv_conf)
    for line in list_of_lines:
      if len(line) == 0 or line[0] == '#' or line[0] == ';':
        continue
      tokens = line.split()
      if len(tokens) == 0:
        continue
      if tokens[0] == 'domain':
        domain = tokens[1].lower()
      elif tokens[0] == 'search':
        for suffix in tokens[1:]:
          search.append(suffix.lower())
    self._domain = domain
    self._search = search
    self._resolv_conf_valid_until = now + DnsNames.RESOLV_CONF_REFRESH_SECS

  def _read_resolv_conf_contents(self, resolv_conf):
    """Read the contents of resolv_conf as a list of lines"""
    try:
      file_object = open(resolv_conf, 'r')
      try:
        return file_object.read().splitlines()
      finally:
        file_object.close()
    except:
      THROTTLED_LOG.exception("Unable to read %s" % (resolv_conf))
      return []

  def _match(self, hostname, canonicalname):
    """Match the hostname with canonicalname"""
    if hostname == canonicalname:
      return True
    if self._domain:
      name = hostname + "." + self._domain
      if name == canonicalname:
        return True
    for domain in self._search:
      name = hostname + "." + domain
      if name == canonicalname:
        return True
    return False

  def update_monitor_dns(self, host_config):
    dns_resolution_enabled = \
      host_config.has_key('host_dns_resolution_enabled') and \
      host_config['host_dns_resolution_enabled'] == 'true'
    dns_resolution_duration_enabled = False
    if host_config.has_key('host_dns_resolution_duration_thresholds'):
      try:
        duration_thresholds = \
          json.loads(host_config['host_dns_resolution_duration_thresholds'])
        dns_resolution_duration_enabled = \
          (duration_thresholds['warning'] != NEVER or
           duration_thresholds['critical'] != NEVER)
      except:
        THROTTLED_LOG.info(
          "Unable to parse host_dns_resolution_duration_thresholds : %s" \
          % host_config['host_dns_resolution_duration_thresholds'])
    self._monitor_dns = dns_resolution_enabled or dns_resolution_duration_enabled
