# Copyright (c) 2011 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging
from multiprocessing import Process, Pipe, active_children
import os
import os.path
import psutil
import threading
import time

import cmf.util
from cmf.monitor.host import filesystem
from cmf.throttling_logger import ThrottlingLogger

LOG = logging.getLogger(__name__)
THROTTLED_LOG = ThrottlingLogger(LOG)

# VisibleForTesting
def disk_usage_wrapper(pipe, paths):
  """
  Helper used by _disk_usage_for_non_local_device_filesystem. See that call for
  more details.
  """
  results = dict()
  for p in paths:
    usage = psutil.disk_usage(p)
    results[p] = dict(total=usage.total,
                      used=usage.used,
                      free=usage.free)
  pipe.send(results)
  pipe.close()

class FilesystemMap(object):
  """
  The FilesystemMap helps caller find out which filessytem a path belongs to. It
  also acts as a central spot for safely querying information about filesystems.
  We centralize this functionality since for remote filesystems, particularly
  NFS, we have to take care to avoid introducing instability.
  """
  def __init__(self,
               monitored_nodev_fstypes=[],
               poll_timeout=2.0,
               join_timeout=0.1,
               tolerance=60.0,
               local_filesystem_whitelist=[],
               fstypes_info_proc_file="/proc/filesystems"):
    self.lock = threading.RLock()
    # Mount point -> Filesystem
    self._fs_map = {}
    # A command separated list of the nodev filesystems, used for some log
    # messages.
    self._nodev_filesystems = ""
    # Mount point -> disk usage dictionary
    self._nodev_disk_usage = {}
    self._last_nodev_refresh = None
    # VisibleForTesting
    self.nodev_refresh_process = None

    LOG.info("Monitored nodev filesystem types: %s" % (monitored_nodev_fstypes,))
    LOG.info("Using timeout of %f" % (poll_timeout,))
    LOG.info("Using join timeout of %f" % (join_timeout,))
    LOG.info("Using tolerance of %f" % (tolerance,))
    LOG.info("Local filesystem types whitelist: %s" % (local_filesystem_whitelist,))
    self._monitored_nodev_fstypes = monitored_nodev_fstypes
    self._local_filesystem_whitelist = local_filesystem_whitelist
    self._poll_timeout = poll_timeout
    self._join_timeout = join_timeout
    # VisibleForTesting
    self.tolerance = tolerance
    self._fstypes_info_proc_file = fstypes_info_proc_file
    
    self.refresh()

  def refresh(self):
    """
    Construct the mount point info. We enumerate the filesystems in a fashion
    similar to psutil's diks_partition's call with the 'all' parameter set to
    false except that we allow in certain additional filesystem types that it
    would exclude.
    """
    try:
      monitored_fstypes = set(self._monitored_nodev_fstypes)
      monitored_fstypes.update(self._local_filesystem_whitelist)
      f = open(self._fstypes_info_proc_file, "r")
      try:
        for line in f:
          if not line.startswith("nodev"):
            monitored_fstypes.add(line.strip())
      finally:
        f.close()

      partition_list = []
      for p in self.get_all_mounted_partitions():
        if p.fstype not in monitored_fstypes:
          continue

        partition_list.append(p)

    except Exception:
      LOG.exception("Error reading partition info")
      return False

    fs_map = { }
    nodev_filesystems = []
    for p in partition_list:
      fs = filesystem.Filesystem(self,
                                 p.mountpoint,
                                 p.device,
                                 p.opts,
                                 p.fstype in self._monitored_nodev_fstypes,
                                 p.fstype)
      fs_map[p.mountpoint] = fs
      if fs.nodev:
        nodev_filesystems.append(p.mountpoint)
    try:
      self.lock.acquire()
      self._fs_map = fs_map
      self._nodev_filesystems = ",".join(nodev_filesystems)
    finally:
      self.lock.release()

    return True

  # Added for testing
  def get_all_mounted_partitions(self):
    return [p for p in psutil.disk_partitions(all=True) if os.path.exists(p.mountpoint) and os.path.isdir(p.mountpoint) ]

  def get_filesystem_by_path(self, path):
    """
    Given a path, return the local filesystem that it belongs to or None if not
    found.
    """
    prev = None
    # We must resolve any symlinks in the path before trying to determine the
    # mount information.
    path = os.path.realpath(path)
    while not os.path.ismount(path):
      prev = path
      path = os.path.dirname(path)
      if prev == path:
        LOG.error("Failed to determine the mount point of '%s'." % (path,))
        break

    try:
      self.lock.acquire()
      return self._fs_map.get(path)
    finally:
      self.lock.release()

  def get_map(self):
    """
    Grabs a copy of the current filesystem map.
    """
    try:
      self.lock.acquire()
      return self._fs_map.copy()
    finally:
      self.lock.release()

  def get_info_for_scm_heartbeat(self):
    """
    Returns a dictionary of file-system related information that wil be sent to
    SCM as part of the heartbeat. The key "filesystem_info" contains a list
    where each entry is a dict that corresponds to a filesystem in which we are
    interested (see above). The keys of this dict are name, total_space_bytes,
    nodev and mount_point. The key "mountpoint_available_space_bytes" contains a
    dictionary of mount points names to available free space on the mount point.
    """
    try:
      self.lock.acquire()

      fs_details = []
      fs_free_stat = dict()
      max_int = cmf.util.get_max_int()

      for mountpoint, fs in self._fs_map.iteritems():
        if 'ro' in fs.options.split(","):
            continue;

        usage = self.disk_usage(mountpoint)
        if usage is None:
          continue

        # Mount points on object storage/cloud return sentinel
        # value of LONG_MAX + 1 to indicate large capacity bytes. See OPSAPS-35742.
        # Since Avro doesn't support unsigned integers, values greater than LONG_MAX
        # lead to heartbeat failures.
        #
        # Longer term we should incorporate big unsigned integers.
        free_bytes = min(usage['free'], max_int)
        if free_bytes < usage['free']:
          THROTTLED_LOG.warn("Limiting free_bytes from %ld to %ld, nodev %r",
                             usage['free'], free_bytes, fs.nodev)

        total_bytes = min(usage['total'], max_int)
        if total_bytes < usage['total']:
          THROTTLED_LOG.warn("Limiting total_bytes from %ld to %ld nodev %r",
                             usage['total'], total_bytes, fs.nodev)

        fs_free_stat[mountpoint] = free_bytes

        name = fs.partition
        if name is None or name == '':
          name = fs.fstype

        fs_details.append(dict(name=name,
                               mount_point=mountpoint,
                               total_space_bytes=total_bytes,
                               nodev=fs.nodev))

      return dict(filesystem_info=fs_details,
                  mountpoint_available_space_bytes=fs_free_stat)

    finally:
      self.lock.release()

  def disk_usage(self, mountpoint):
    """
    Returns a dictionary of disk usage information for the input mountpoint. The
    dictionary is inspired by psutil's disk_usage call, but that is not returned
    directly because of implementation concerns. The dictionary contains 'free',
    'used' and 'total' entries. None is returned on an error. Note that for some
    filesystems this may actually involve waiting for a request to timeout.
    """
    try:
      self.lock.acquire()
      if mountpoint not in self._fs_map:
        LOG.error("Mountpoint %s not in filesystem map" % (mountpoint,))
        return None
      fs = self._fs_map[mountpoint]
      if fs.nodev:
        return self._disk_usage_for_non_local_device_filesystem(mountpoint)
      else:
        return self._disk_usage_for_local_device_filesystem(mountpoint)
    finally:
      self.lock.release()

  def _disk_usage_for_local_device_filesystem(self, mountpoint):
    """
    Simple call to psutil.disk_usage used for filesystems on local devices.
    """
    try:
      usage = psutil.disk_usage(mountpoint)
      return dict(total=usage.total,
                  used=usage.used,
                  free=usage.free)
    except Exception:
      LOG.error("Failed to collect filesystem usage on %s" % (mountpoint,))
      return None

  # VisibleForTesting
  def get_nodev_disk_usage_function(self):
    return disk_usage_wrapper

  def _disk_usage_for_non_local_device_filesystem(self, mountpoint):
    """
    Paranoid call to get the equivalent of psutil.disk_usage used for filesystems
    that are not on local devices. This effectively forks a child process that
    does the actual call.
    """
    try:
      self.lock.acquire()
      
      # If we have cached values that are still valid, just return those.
      if self._last_nodev_refresh is not None and \
         time.time() < self._last_nodev_refresh + self.tolerance:
         return self._nodev_disk_usage.get(mountpoint)

      # Next, check if we have a zombie child process still around. If so, make
      # another attempt to join it. If we can't get it cleaned up, do not start
      # another.
      if self.nodev_refresh_process is not None:
        if self.nodev_refresh_process in active_children():
          THROTTLED_LOG.error("Child collection process still around. Skipping "
                              "new collection. All nodev filesystems will have "
                              "unknown usage until collection resumes. Current "
                              "nodev filesystems: %s" % self._nodev_filesystems)
          return None
        self.nodev_refresh_process = None

      try:
        parent_pipe, child_pipe = Pipe(False)
        process = Process(target=self.get_nodev_disk_usage_function(),
                          args=(child_pipe, self._fs_map.keys()))
        process.daemon = True
        process.start()

        self._nodev_disk_usage = {}

        if parent_pipe.poll(self._poll_timeout):
          self._nodev_disk_usage = parent_pipe.recv()
        else:
          THROTTLED_LOG.error("Timed out waiting for worker process collecting "
                              "filesystem usage to complete. This may occur if "
                              "the host has an NFS or other remote filesystem "
                              "that is not responding to requests in a timely "
                              "fashion. Current nodev filesystems: %s" % 
                              self._nodev_filesystems)

        parent_pipe.close()
        process.terminate()
        process.join(self._join_timeout)

        if process in active_children():
          LOG.warn("Failed to join worker process collecting filesystem usage. "
                   "All nodev filesystems will have unknown usage until the "
                   "worker process is no longer active. Current nodev "
                   "filesystems: %s" % self._nodev_filesystems)
          self.nodev_refresh_process = process

        self._last_nodev_refresh = time.time()

        return self._nodev_disk_usage.get(mountpoint)

      except Exception:
        LOG.exception("Failed to collect filesystem usage on %s" % (mountpoint,))
        return None

    finally:
      self.lock.release()
