# Copyright (c) 2011 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging
import re
from time import time

from cmf.throttling_logger import ThrottlingLogger
from cmf.monitor.host.filesystem_map import FilesystemMap
import os.path

LOG = logging.getLogger(__name__)
THROTTLED_LOG = ThrottlingLogger(LOG)

class Filesystems(object):
  """Provides access to mounted filesystems"""
  _FS_EXCLUDE_PATTERN_DEFAULT = "^$"
  _FS_EXCLUDE_REGEX_DEFAULT = re.compile(_FS_EXCLUDE_PATTERN_DEFAULT)

  def __init__(self, fs_map=None):
    if fs_map is not None:
      self._fs_map = fs_map
    else:
      self._fs_map = FilesystemMap()

    self._fs_exclude_regex = self._FS_EXCLUDE_REGEX_DEFAULT
    self._last_regex_pattern = self._FS_EXCLUDE_PATTERN_DEFAULT
    self._latest_updates_pair = ([], None)
    self._last_reported_timestamp = 0

  def collect(self):
    updates = []
    # capture the regex as a local variable in case it changes
    exclude_regex = self._fs_exclude_regex

    for (mountpoint, fs) in self._fs_map.get_map().items():
      try:
        # Docker containers have various mount points that are files-only - ignore them.
        if os.path.isfile(mountpoint):
          continue
        if not exclude_regex.match(mountpoint):
          fs.collect()
          updates.append(fs.get_fs_update())
      except Exception:
        THROTTLED_LOG.exception("Error collecting FS update for %s (%s)" %
                                (mountpoint, fs.partition,))
    self._latest_updates_pair = (updates, time())

  def get_fs_updates(self):
    (updates, timestamp) = self._latest_updates_pair

    if timestamp == self._last_reported_timestamp:
      return list()

    self._last_reported_timestamp = timestamp
    return updates

  def set_fs_exclude_regex(self, regex_pattern):
    if self._last_regex_pattern != regex_pattern:
      self._last_regex_pattern = regex_pattern
      try:
        self._fs_exclude_regex = re.compile(regex_pattern)
      except re.error:
        THROTTLED_LOG.error("Unable to compile disk device exclusion regex '%s', using default.",
                            regex_pattern)
        self._fs_exclude_regex = self._FS_EXCLUDE_REGEX_DEFAULT
