# Copyright (c) 2012 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging
import re
import socket
import array
import struct
import fcntl
import errno
import time

from cmf.monitor.constants import DUPLEX_FULL, DUPLEX_HALF
from cmf.monitor.nic_update import NicUpdate
from cmf.throttling_logger import ThrottlingLogger

LOG = logging.getLogger(__name__)
THROTTLED_LOG = ThrottlingLogger(LOG)

'''
Kernel structure for the ethtool_cmd, from linux/ethtool.h

The speed, duplex, and speed_hi fields are used.

offset | struct ethtool_cmd {
00     |  __u32  cmd;
04     |  __u32  supported;  /* Features this interface supports */
08     |  __u32  advertising;  /* Features this interface advertises */
12     |  __u16  speed;
14     |  __u8  duplex;    /* Duplex, half or full */
15     |  __u8  port;    /* Which connector port */
16     |  __u8  phy_address;
17     |  __u8  transceiver;  /* Which transceiver to use */
18     |  __u8  autoneg;  /* Enable or disable autonegotiation */
19     |  __u8  mdio_support;
20     |  __u32  maxtxpkt;  /* Tx pkts before generating tx int */
24     |  __u32  maxrxpkt;  /* Rx pkts before generating rx int */
28     |  __u16  speed_hi;
30     |  __u8  eth_tp_mdix;
31     |  __u8  reserved2;
32     |  __u32  lp_advertising;  /* Features the link partner advertises */
36     |  __u32  reserved[2];
44     | };
'''

# From linux/ethtool.h
ETHTOOL_GSET = 0x00000001 # Get settings

# From linux/sockios.h
SIOCETHTOOL  = 0x8946     # Ethtool interface

class NetworkInterfaces(object):
  """Provides access to all the network interfaces on the host"""

  _PROC_NET_DEV = "/proc/net/dev"
  _IFACE_EXCLUDE_REGEX_DEFAULT = re.compile("^lo$")
  
  # /proc/net/dev line format (it is a single line - broken below to avoid long line):
  # <id>: <bytes_rx> <packets_rx> <errs_rx> <drop_rx> <fifo_rx> <frame_rx> <cmpr_rx> <mlti_rx> \
  #       <bytes_tx> <packets_tx> <errs_tx> <drop_tx> <fifo_tx> <frame_tx> <cmpr_tx> <mlti_tx>
  
  _IDX_BYTES_RECEIVE = 0
  _IDX_PACKETS_RECEIVE = 1
  _IDX_ERRS_RECEIVE = 2
  _IDX_DROP_RECEIVE = 3
  _IDX_FIFO_RECEIVE = 4 # unused
  _IDX_FRAME_RECEIVE = 5
  _IDX_COMPRESSED_RECEIVE = 6 # unused
  _IDX_MULTICAST_RECEIVE = 7 # unused
  _IDX_BYTES_TRANSMIT = 8
  _IDX_PACKETS_TRANSMIT = 9
  _IDX_ERRS_TRANSMIT = 10
  _IDX_DROP_TRANSMIT = 11
  _IDX_FIFO_TRANSMIT = 12 # unused
  _IDX_COLLS_TRANSMIT = 13
  _IDX_CARRIER_TRANSMIT = 14 # unused
  _IDX_COMPRESSED_TRANSMIT = 15 # unused

  def __init__(self, proc_net_dev = _PROC_NET_DEV):
    # Test if file exists
    open(proc_net_dev).close()
    self._proc_net_dev = proc_net_dev
    self._ifaces_skip_mode_check = set([])
    self._iface_exclude_regex = self._IFACE_EXCLUDE_REGEX_DEFAULT
    self._last_regex_pattern = None
    self._latest_update_pair = ([], None)
    self._last_reported_timestamp = 0

  def ioctl_ethtool(self, sockfd, iface, ecmd):
    ifreq = struct.pack('16sP', iface, ecmd.buffer_info()[0])
    fcntl.ioctl(sockfd, SIOCETHTOOL, ifreq)
    return ecmd

  def set_iface_exclude_regex(self, regex_pattern):
    if self._last_regex_pattern != regex_pattern:
      self._last_regex_pattern = regex_pattern
      try:
        self._iface_exclude_regex = re.compile(regex_pattern)
      except re.error:
        THROTTLED_LOG.exception("Unable to compile network interface exclusion regex '%s', using default.",
                                regex_pattern)
        self._iface_exclude_regex = self._IFACE_EXCLUDE_REGEX_DEFAULT

  def collect(self):
    dev_lines = []
    try:
      proc_file = open(self._proc_net_dev, 'r')
      try:
        # The filter(None,...) and strip will clean-up and remove empty lines
        # and we want to ignore the first two lines which are the header
        dev_lines = filter(None, [line.strip() for line in proc_file.readlines()[2:]])
      finally:
        proc_file.close()
    except Exception:
      THROTTLED_LOG.exception("Failed to read NIC data from /proc/net/dev")

    updates = []
    # capture the regex as a local variable in case it changes
    exclude_regex = self._iface_exclude_regex

    for dev_line in dev_lines:
      try:
        iface_name, metrics_string = dev_line.split(':')
        if not exclude_regex.match(iface_name):
          speed, duplex_mode = self._get_iface_mode(iface_name)

          raw_metrics = [ long(m) for m in metrics_string.split() ]

          updates.append(NicUpdate(
              iface_name,
              speed,
              duplex_mode,
              raw_metrics[NetworkInterfaces._IDX_BYTES_RECEIVE],
              raw_metrics[NetworkInterfaces._IDX_PACKETS_RECEIVE],
              raw_metrics[NetworkInterfaces._IDX_ERRS_RECEIVE],
              raw_metrics[NetworkInterfaces._IDX_DROP_RECEIVE],
              raw_metrics[NetworkInterfaces._IDX_FRAME_RECEIVE],
              raw_metrics[NetworkInterfaces._IDX_BYTES_TRANSMIT],
              raw_metrics[NetworkInterfaces._IDX_PACKETS_TRANSMIT],
              raw_metrics[NetworkInterfaces._IDX_ERRS_TRANSMIT],
              raw_metrics[NetworkInterfaces._IDX_DROP_TRANSMIT],
              raw_metrics[NetworkInterfaces._IDX_COLLS_TRANSMIT]))
      except Exception:
        THROTTLED_LOG.exception("Failed to collect metric for line='%s'", dev_line)
    self._latest_update_pair = (updates, time.time())

  def get_nic_updates(self):
    updates, timestamp = self._latest_update_pair

    if timestamp == self._last_reported_timestamp:
      return list()

    self._last_reported_timestamp = timestamp

    return updates

  def _get_iface_mode(self, iface):
    """Gets the speed and duplex mode of the interface as reported by the kernel

       Returns a tuple, the speed (Mb/s) and duplex-mode (0 half, 1 full)
    """
    speed = None
    duplex = None
    if not iface in self._ifaces_skip_mode_check:
      try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
          # The ethtool_cmd structure is 44 bytes- pack the command (ETHTOOL_GSET)
          # (int) as the first 4 bytes and fill the remaining 40 bytes with \x00
          ecmd = array.array('B', struct.pack('I40s', ETHTOOL_GSET, '\x00'*40))
          ecmd = self.ioctl_ethtool(sock.fileno(), iface, ecmd)
          res = ecmd.tostring()

          # The structure is packed and unpacked manually in the following way:
          # (X:12B)(speed:ushort:2B)(duplex:uchar:1B)(X:13B)(speed_hi:ushort:2B)(X:14B)
          # see the ethtool_cmd struct at the top of the file as a reference
          speed_lo, duplex, speed_hi = struct.unpack('12xHB13xH14x', res)
          # The speed is unknown if it is:
          #  * 0
          #  * (ushort)-1 = 2^16-1
          #  *   (uint)-1 = 2^32-1
          speed = (speed_hi << 16) | speed_lo
          if speed == 0 or speed == (pow(2,16)-1) or speed == (pow(2,32)-1):
            LOG.warn("Interface '%s' reported unknown speed '%s'" % (iface, speed))
            speed = None
            self._ifaces_skip_mode_check.add(iface)
          if duplex != DUPLEX_FULL and duplex != DUPLEX_HALF:
            LOG.warn("Interface '%s' reported unknown duplex mode '%s'" % (iface, duplex))
            duplex = None
            self._ifaces_skip_mode_check.add(iface)
        finally:
          sock.close()
      except IOError, ex:
        # Some older kernels will return EINVAL for SIOCETHTOOL with unsupported network
        # interfaces, e.g. sit0, though they should return EOPNOTSUPP. For any other errors,
        # log a warning with the error code, but we don't check this interface again.
        if ex.errno == errno.EOPNOTSUPP or ex.errno == errno.EINVAL:
          LOG.info("NIC iface %s doesn't support ETHTOOL (%s)" % (iface, ex.errno))
        else:
          LOG.warn("NIC iface %s returned unexpected error (%s): %s" %
                   (iface, ex.errno, ex.strerror))
        self._ifaces_skip_mode_check.add(iface)
      except Exception:
        THROTTLED_LOG.exception("Failed to access %s NIC mode due to unknown Exception. "
                                "This is not likely to cause other problems." % (iface,))
        self._ifaces_skip_mode_check.add(iface)
    return speed, duplex
