# Copyright (c) 2013 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging
import os
import re
import subprocess
import time

from cmf.monitor import schema, constants
from cmf.subprocess_timeout import subprocess_with_timeout
from cmf.throttling_logger import ThrottlingLogger

LOG = logging.getLogger(__name__)
THROTTLED_LOG = ThrottlingLogger(LOG)

class NTPMonitor(object):
  """
  The NTP monitor gathers metrics from the NTP service running on the system.
  Currently this uses 'ntpq -np' or 'chronyc sources' to get information.
  See their usage for more information.
  """

  def __init__(self, timeout=2.0):
    LOG.info("Using timeout of %f" % (timeout,))
    self._ntp_status = constants.NTP_STATUS_UNKNOWN
    self._clock_offset = None
    self._timeout = timeout

    self._is_chronyd_running = False
    try:
      result, stdout, stderr = self._subprocess_with_timeout(["pidof", "chronyd"], self._timeout)
      if result is 0:
        self._is_chronyd_running = True
    except Exception, e:
      THROTTLED_LOG.exception("Failed to run `pidof chronyd` command within timeout=%d", self._timeout)

  def _subprocess_with_timeout(self, args, timeout):
    return subprocess_with_timeout(args, timeout)

  def collect(self):
    self._ntp_status = constants.NTP_STATUS_UNKNOWN
    self._clock_offset = None
    try:
      if self._is_chronyd_running:
        self.collect_chronyd()
        return
      else:
        self.collect_ntpd()
        return
    except Exception, e:
      THROTTLED_LOG.exception("Failed to collect NTP metrics")

  def collect_chronyd(self):
    args = ["chronyc", "-n", "sources"]
    result, stdout, stderr = self._subprocess_with_timeout(args, self._timeout)
    if result is not 0:
      THROTTLED_LOG.info("chronyc return=%d\nstdout=%s\nstderr=%s",
                         result, stdout, stderr)
      self._ntp_status = constants.NTP_STATUS_UNREACHEABLE
      self._clock_offset = None
      return
    self._ntp_status, self._clock_offset = self._parse_chronyd_clock_offset(stdout)

  def collect_ntpd(self):
    args = ["ntpq", "-np"]
    result, stdout, stderr = self._subprocess_with_timeout(args, self._timeout)
    if result is not 0:
      THROTTLED_LOG.info("ntpq return=%d\nstdout=%s\nstderr=%s",
                         result, stdout, stderr)
      return
    if (stderr.startswith('ntpq: read: Connection refused') or
          stderr.startswith('Name or service not known')):
      self._ntp_status = constants.NTP_STATUS_UNREACHEABLE
      self._clock_offset = None
      return
    self._ntp_status, self._clock_offset = self._parse_ntpd_clock_offset(stdout)

  # This returns a tuple of NTP status, offset in milliseconds, or None if no
  # offset could be determined.
  def _parse_ntpd_clock_offset(self, output):
    for line in output.splitlines():
      if not line.startswith('*'):
        continue
      offset = abs(float(line.split()[8]))
      return constants.NTP_STATUS_RUNNING, offset
    THROTTLED_LOG.error("ntpq: ntpq -np: not synchronized to any server")
    return constants.NTP_STATUS_NOT_SYNCHRONIZED, None

  def _parse_chronyd_clock_offset(self, output):
    for line in output.splitlines():
      if not line.startswith(('^*', '=*', '#*')):
        continue
      value = line.split('[')[1].split(']')[0]
      if value.endswith('ms'):
        offset = abs(float(value[:-2]))
      elif value.endswith('us'):
        offset = abs(0.001 * float(value[:-2]))
      elif value.endswith('ns'):
        offset = abs(0.000001 * float(value[:-2]))
      elif value.endswith('s'):
        offset = abs(1000 * float(value[:-1]))
      else:
        offset = abs(1000 * float(value))
      return constants.NTP_STATUS_RUNNING, offset
    THROTTLED_LOG.error("chronyc: chronyc sources: not synchronized to any server")
    return constants.NTP_STATUS_NOT_SYNCHRONIZED, None


  def add_ntp_metrics(self, update):
    update.set_ntp_status(self._ntp_status)
    if self._clock_offset is not None:
      update.add_metric(schema.CLOCK_OFFSET, self._clock_offset)
