# Copyright (c) 2012 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging
import re

from cmf.monitor import schema
from cmf.throttling_logger import ThrottlingLogger

LOG = logging.getLogger(__name__)
THROTTLED_LOG = ThrottlingLogger(LOG)

'''
TCP States
  TCP_ESTABLISHED = 1
  TCP_SYN_SENT    = 2
  TCP_SYN_RECV    = 3
  TCP_FIN_WAIT1   = 4
  TCP_FIN_WAIT2   = 5
  TCP_TIME_WAIT   = 6
  TCP_CLOSE       = 7
  TCP_CLOSE_WAIT  = 8
  TCP_LAST_ACK    = 9
  TCP_LISTEN      = 10
  TCP_CLOSING     = 11
'''

_PROC_NET_TCP = "/proc/net/tcp"
_PROC_NET_TCP6 = "/proc/net/tcp6"

'''
/proc/net/tcp format:
  sl  local_address rem_address   st tx_queue rx_queue tr tm->when retrnsmt   uid  timeout inode
   0: 00000000:C36E 00000000:0000 0A 00000000:00000000 00:00000000 00000000   119        0 288886 1 0000000000000000 300 0 0 2 -1

/proc/net/tcp6 format:
  sl  local_address                         remote_address                        st tx_queue rx_queue tr tm->when retrnsmt   uid  timeout inode
   0: 00000000000000000000000000000000:1C0E 00000000000000000000000000000000:0000 0A 00000000:00000000 00:00000000 00000000  1000        0 25410663 1 0000000000000000 300 0 0 2 -1
'''

# Regex for the TCP line format.  Extracts the status (4th column) which is
# 2 hex digits following 3 (alphanumeric and colon) groups.
_TCP_RE = re.compile("^(?:\s*[\w:]+)(?:\s+[\w:]+){2}\s+(\w{2}).*$")

class TcpMetrics(object):
  def __init__(self,
               proc_net_tcp = _PROC_NET_TCP,
               proc_net_tcp6 = _PROC_NET_TCP6):
    self._tcp_files = []
    for proc_file_name in [proc_net_tcp, proc_net_tcp6]:
      try:
        proc_tcp_file = open(proc_file_name)
        self._tcp_files.append(proc_file_name)
        proc_tcp_file.close()
      except IOError, ioex:
        LOG.warning("File '%s' couldn't be opened for tcp statistic "
                    "collection, error=%s" % (proc_file_name, ioex.errno))
      except Exception:
        LOG.exception("Unknown error opening file '%s'" % (proc_file_name))

  def add_tcp_metrics(self, update):
    """Adds TCP connection count metrics to the update"""
    try:
      tcp_lines = []
      for proc_file_name in self._tcp_files:
        try:
          proc_tcp_file = open(proc_file_name)
        except IOError, ioex:
          THROTTLED_LOG.warn("File '%s' couldn't be opened, error=%s"
                             % (proc_file_name, ioex.errno))
          continue

        try:
          tcp_lines.extend(proc_tcp_file.readlines()[1:])
        finally:
          proc_tcp_file.close()

      if len(tcp_lines) == 0:
        return

      # Extracts the statuses from each line and converts the string to an int
      # The state is a base16 integer, so it is converted to base10 via int(STATE,16)
      statuses = [ int(_TCP_RE.match(line).groups()[0],16) for line in tcp_lines ]
      # counts the number of statuses in the array for numbers [1..12) (12 excluded)
      # corresponding to the 11 possible TCP states.
      status_counts = dict((st, statuses.count(st)) for st in range(1,12))

      update.add_metric(schema.TCP_CONNECTION_COUNT_ESTABLISHED, status_counts[1])
      update.add_metric(schema.TCP_CONNECTION_COUNT_SYN_SENT, status_counts[2])
      update.add_metric(schema.TCP_CONNECTION_COUNT_SYN_RECV, status_counts[3])
      update.add_metric(schema.TCP_CONNECTION_COUNT_FIN_WAIT1, status_counts[4])
      update.add_metric(schema.TCP_CONNECTION_COUNT_FIN_WAIT2, status_counts[5])
      update.add_metric(schema.TCP_CONNECTION_COUNT_TIME_WAIT, status_counts[6])
      update.add_metric(schema.TCP_CONNECTION_COUNT_CLOSE, status_counts[7])
      update.add_metric(schema.TCP_CONNECTION_COUNT_CLOSE_WAIT, status_counts[8])
      update.add_metric(schema.TCP_CONNECTION_COUNT_LAST_ACK, status_counts[9])
      update.add_metric(schema.TCP_CONNECTION_COUNT_LISTEN, status_counts[10])
      update.add_metric(schema.TCP_CONNECTION_COUNT_CLOSING, status_counts[11])
    except Exception:
      THROTTLED_LOG.exception("Failed to collect TCP metrics")