# Copyright (c) 2011-2012 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

try:
  from cStringIO import StringIO
except:
  from StringIO import StringIO

import time
from cmf.monitor import constants
from cmf.monitor.firehoses import Firehoses

class HostUpdate(object):
  """
  This models the HostUpdate Avro record and gives it a bit more structure.
  """
  def __init__(self,
               agent_log_directory,
               agent_process_directory,
               agent_parcel_directory,
               timestamp=None):
    self._agent_log_directory = agent_log_directory
    self._agent_process_directory = agent_process_directory
    if agent_parcel_directory is None:
      # When we do not have a directory path, we use the Avro default '""' as
      # per the HostUpdate schema in 'cmon_agent.avdl'.
      # We expect the 'agent_parcel_directory' to be None when the parcel
      # directory is not in use. See OPSAPS-14127 for additional context.
      self._agent_parcel_directory = ""
    else:
      self._agent_parcel_directory = agent_parcel_directory
    if timestamp is None:
      timestamp = time.time()
    self._timestamp = timestamp
    self._metrics = [ ]
    self._fs_updates = [ ]
    self._nic_updates = [ ]
    self._disk_updates = [ ]
    self._dir_updates = [ ]
    self._ntp_status = constants.NTP_STATUS_UNKNOWN
    self._overcommit_memory_mode = constants.OVERCOMMIT_UNKNOWN

  def add_metric(self, metricid, value):
    # A value of None causes avro to blow up when we encode the message, so to
    # be safe, we protect against that here.
    if value is not None:
      self._metrics.append((metricid, value))
    return self

  def add_fs_update(self, fs_update):
    self._fs_updates.append(fs_update)
    return self

  def add_disk_update(self, disk_update):
    self._disk_updates.append(disk_update)
    return self

  def add_nic_update(self, nic_update):
    self._nic_updates.append(nic_update)
    return self

  def add_dir_update(self, host_dir_update):
    if host_dir_update is not None:
      self._dir_updates.append(host_dir_update)
    return self

  def set_ntp_status(self, status):
    self._ntp_status = status
    return self

  def set_overcommit_memory_mode(self, mode):
    self._overcommit_memory_mode = mode
    return self

  def merge(self, other):
    """Merge the other host_update into us"""
    for m in other._metrics:
      self._metrics.append(m)
    for fs in other._fs_updates:
      self._fs_updates.append(fs)
    for disk in other._disk_updates:
      self._disk_updates.append(disk)
    for nic in other._nic_updates:
      self._nic_updates.append(nic)
    for d in other._dir_updates:
      self._dir_updates.append(d)
    return self

  def __str__(self):
    sio = StringIO()
    sio.write(">> %s:" %
              time.strftime("%x %X", time.localtime(self._timestamp)))
    for metricid, val in self._metrics:
      sio.write("\n   %-10d : %s" % (metricid, val))
    for fs_update in self._fs_updates:
      sio.write("\n%s" % (fs_update,))
    for nic_update in self._nic_updates:
      sio.write("\n%s" % (nic_update,))
    for disk_update in self._disk_updates:
      sio.write("\n%s" % (disk_update,))
    for dir_update in self._dir_updates:
      sio.write("\n%s" % (dir_update,))
    return sio.getvalue()

  def to_avro_dict(self):
    """Return a Python dictionary that maps to the Avro struct"""
    return dict(ts_secs=long(self._timestamp),
                agentLogDirectory=self._agent_log_directory,
                agentProcessDirectory=self._agent_process_directory,
                agentParcelDirectory=self._agent_parcel_directory,
                metrics=[ dict(id=k, value=Firehoses.normalize_metric_value(v))
                          for (k, v) in self._metrics ],
                fsUpdates=[ fs.to_avro_dict() for fs in self._fs_updates ],
                nicUpdates=[ nic.to_avro_dict() for nic in self._nic_updates ],
                diskUpdates=[ disk.to_avro_dict() for disk in self._disk_updates ],
                dirUpdates=[ d.to_avro_dict() for d in self._dir_updates ],
                ntpStatus=self._ntp_status,
                overcommitMemory=self._overcommit_memory_mode)
