# Copyright (c) 2013 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import datetime
import logging
import time
from base64 import b64decode

import clusterstats.common.chain as chain
import os
import psutil
import re
from clusterstats.log.streaming.event import LogEvent
from clusterstats.log.streaming.event_reader import LogFileEventReader
from clusterstats.log.streaming.event_streamer import LogEventStreamer
from clusterstats.log.streaming.file_filter import DirectoryFilter

from cmf.monitor import constants, schema
from cmf.monitor.abstract_monitor import CollectionError
from cmf.monitor.generic.metric_collectors import WebServerMetricCollector
from cmf.monitor.impala_query_update import ImpalaQueryUpdate, ImpalaRuntimeProfile
from cmf.monitor.stoppable_thread import StoppableThread
from cmf.throttling_logger import ThrottlingLogger
from cmf.util.url import urlopen_with_timeout

LOG = logging.getLogger('ImpaladQueryMonitor')
THROTTLED_LOG = ThrottlingLogger(LOG)

def _get_cur_process():
  return psutil.Process(os.getpid())

def _safe_get_io_counters():
  try:
    return _get_cur_process().io_counters()
  except psutil.NoSuchProcess:
    # We don't support I/O metrics on linux machines with kernels that don't
    # support I/O Stats (that don't have /proc/<pid>/io)
    return psutil.ntuple_io(0,0,0,0)

# We implement our own TimeRangeFilter that's similar in how it filters to the
# clusterstats one. The differences are that we keep track of the time of the
# latest event time within the requested range and latest file processed.
# For each file passed to this Filter, we extract 'timestamp' from file name
# using given regex pattern, compare it with passed in timestamp and skip
# processing any file with an older timestamp.
class _TimeRangeFilter(object):
  """Filter out log files based on start_date_time,end_date_time and timestamp
  passed to the constructor. For each file passed to this Filter, we extract
  'timestamp' from file name using given regex pattern, compare it with passed
  in timestamp and skip processing any file with an older timestamp. For any
  non-filtered files, we find first and last log event in the file that falls
  in the given start/end datetime. We assume that log events are in time
  increasing order."""

  def __init__(self,
               start_date_time,
               end_date_time,
               event_reader,
               file_pattern,
               last_accessed_file_timestamp = None):
    self.__start_date_time = start_date_time
    self.__end_date_time = end_date_time
    self.__event_reader = event_reader
    self.last_accessed_file_timestamp = last_accessed_file_timestamp
    self.latest_event_datetime = None
    self.file_regex = file_pattern

  def __call__(self, context):
    new_file_list = []
    latest_file_timestamp = self.last_accessed_file_timestamp
    for f in context["file_list"]:
      file_name = os.path.basename(f.get_path())
      m = self.file_regex.match(file_name)
      if m is None or m.group("timestamp") is None:
        continue
      timestamp = long(m.group("timestamp"))

      if self.last_accessed_file_timestamp is not None and \
         timestamp < self.last_accessed_file_timestamp:
        continue

      file_handle = f.open_log_file()
      try:
        self.__event_reader.initialize(file_handle, f.get_path())
        evt1 = self.__event_reader.get_first_event()
        if evt1 is not None and evt1.get_datetime() <= self.__end_date_time:
          evt2 = self.__event_reader.get_last_event()

          if latest_file_timestamp is None or \
             timestamp > latest_file_timestamp:
            latest_file_timestamp = timestamp

          if evt2 is not None and evt2.get_datetime() >= self.__start_date_time:
            self.__set_start_offset(f, evt1.get_datetime())
            self.__set_end_offset(f, evt2.get_datetime())
            new_file_list.append(f)
      finally:
        file_handle.close()

    self.last_accessed_file_timestamp = latest_file_timestamp
    context["file_list"] = new_file_list
    return True

  def __set_start_offset(self, f, first_event_datetime):
    if first_event_datetime < self.__start_date_time:
      event = self.__event_reader.find_nearest_event_datetime(self.__start_date_time)
      f.set_start_offset(event.get_offset())

  # This function contains the only changes to the clusterstats TimeRangeFilter.
  def __set_end_offset(self, f, last_event_datetime):
    end_datetime = last_event_datetime

    if last_event_datetime > self.__end_date_time:
      event = self.__event_reader.find_nearest_event_datetime(self.__end_date_time)
      f.set_end_offset(event.get_offset())
      end_datetime = event.get_datetime()

    if (self.latest_event_datetime is None or
        end_datetime > self.latest_event_datetime):
      self.latest_event_datetime = end_datetime

# We extend LogEvent to explicitly track the query ID and profile we parse out
# below. We also simply hard code some fields that do not make sense for our
# file.
class _QueryLogEvent(LogEvent):
  def __init__(self,
               file_path,
               offset,
               end_offset,
               date_time,
               query_id,
               profile):
    LogEvent.__init__(self,
                      file_path,
                      offset,
                      end_offset,
                      date_time,
                      "INFO",
                      "QUERYLOG",
                      "<query>")
    self.query_id = query_id
    self.profile = profile

# Every line in the file is as follows:
#
# 136924029461 63488804fc4451a8:ab2d3f8e4134a18e eJyFlU1...
#
# The format here is the timestamp in millis since the epoch, then the query ID,
# then a base64 encoded, zlib compressed compact, binary thrift profile.
class _QueryLogEventParser(object):

  def __init__(self):
    timestamp_format = "\d+"
    query_id_format = "\S+"
    profile_format = ".+\\n"
    regex_string = "(?P<timestamp>" + timestamp_format + \
                   ")\s+(?P<query_id>" + query_id_format + \
                   ")\s+(?P<profile>" + profile_format + ")"
    file_name_regex_string = ".*_(?P<version>\d+\.\d+)-(?P<timestamp>\d+)"
    self.__regex = re.compile(regex_string)
    self.__file_name_regex = re.compile(file_name_regex_string)

  def parse_event(self, line):
    m = self.__regex.match(line.get_content())
    if m is not None:
      return self.__construct_event(
        line.get_file_path(),
        line.get_offset(),
        line.get_offset() + len(line.get_content()),
        m)
    else:
      return None

  def __construct_event(self, file_path, offset, end_offset, m):
    file_name = os.path.basename(file_path)
    m2 = self.__file_name_regex.match(file_name)
    if m2 is None or m2.group("version") is None:
      return None

    if m2.group("version") == "1.0":
      dt = _utc_from_local_timestamp(long(m.group("timestamp")) / 1000.0)
    else:
      dt = datetime.datetime.utcfromtimestamp(long(m.group("timestamp")) / 1000.0)
    return _QueryLogEvent(file_path,
                          offset,
                          end_offset,
                          dt,
                          m.group("query_id"),
                          m.group("profile"))

# Converts query timestamp in impalad version-1.0 query profile to naive utc
# datetime object. timestamp in impalad version-1.0 query profile is generated
# from localtime.
def _utc_from_local_timestamp(timestamp):
  time_delta = _time_delta_between_utc_and_local()
  dt = datetime.datetime.utcfromtimestamp(timestamp)
  utc_dt = dt + time_delta
  return utc_dt

# Helper to get timeDelta obj between utc datetime and local datetime.
def _time_delta_between_utc_and_local():
  timestamp = time.time()
  dt = datetime.datetime.fromtimestamp(timestamp)
  utc_dt = datetime.datetime.utcfromtimestamp(timestamp)
  time_delta = utc_dt - dt
  return time_delta

# If we request a profile for which there is no information we get a message
# like this: "Query id 323c59bd14a544f5:a8329df97b197e9a not found."
_PROFILE_NOT_FOUND_PREFIX_OLD = "Query id"
# In CDH5.2, Impala changed the message to start with this new prefix.
_PROFILE_NOT_FOUND_PREFIX = "Could not obtain runtime profile"

# Scans the logical log file at the input path for completed query entries
# between the start and end times, starting with the input query ID . This is
# based on our log search code, so it will handle multiple log files, rotation,
# etc. This function returns a tuple of the following:
#
# (next start datetime to use to continue the search,
#  timestamp of last file where we left,
#  list of (query ID, thrift profile) tuples.)
#
# log_path: path to the log file to search
# start_datetime: start of the search time window
# end_datetime: end of the search time window
# last_reported_query_id: should be specified when streaming to avoid double
#  reporting of queries, this is the query id of the last full query seen by
#  the previous stream searches
# limit: no more than this number of completed queries will be returned
# stats: statistics object for tracking various things
# last_accessed_file_timestamp: timestamp portion of last log file from which
#  we published an event.
#
# Profiles are returned in zlib compressed, binary compact thrift format.
def _get_completed_queries(log_path,
                           start_datetime,
                           end_datetime,
                           last_reported_query_id,
                           limit,
                           stats,
                           last_accessed_file_timestamp):
  log_directory, log_filename = os.path.split(log_path)
  if not os.path.isdir(log_directory):
    return None, None, []
  reader = LogFileEventReader(event_parser=_QueryLogEventParser())
  file_regex = re.compile(log_filename + '.*-(?P<timestamp>\d+)')
  time_range_filter = _TimeRangeFilter(start_datetime,
                                       end_datetime,
                                       reader,
                                       file_regex,
                                       last_accessed_file_timestamp)
  filters = chain.build_chain([DirectoryFilter(),
                               time_range_filter])
  streamer = LogEventStreamer(log_directory,
                              reader,
                              file_filter=filters)

  completed_queries = []

  # If we hit the limit on the number of queries we'll report, then we want to
  # start at that event next time.
  limit_event_datetime = None
  limit_file_timestamp = None
  for event in streamer.stream():
    if len(completed_queries) >= limit:
      stats.increment('total_completed_query_limit_reached')
      LOG.debug("Hit completed query limit.")
      limit_event_datetime = event.get_datetime()
      m = file_regex.match(os.path.basename(event.get_file_path()))
      limit_file_timestamp = long(m.group("timestamp"))
      break

    query_id = event.query_id
    profile = event.profile

    # If we run into the last reported query ID, then clear any accumulated
    # completed queries.
    if query_id == last_reported_query_id:
      completed_queries = []
      continue

    try:
      completed_queries.append((query_id, b64decode(profile)))
    except Exception, ex:
      LOG.exception("Error parsing query with id '%s': '%s'" %
                    (query_id, profile))
      stats.increment('total_completed_query_parse_exceptions')

  # We need to do a little bookkeeping now. We will return the start time that
  # the next query should use, as well as the name of file that was last
  # accessed, which the next query should use as well.
  # Note that this is not determined by the end time we used above. It is
  # determined by the time of the last event our log search actually saw, or the
  # time of the event on which we hit our return limit. If we didn't stream any
  # events, then we'll say to just keep using the passed in start time.
  if limit_event_datetime is not None:
    next_start_datetime = limit_event_datetime
    latest_file_timestamp = limit_file_timestamp
  elif time_range_filter.latest_event_datetime is not None:
    next_start_datetime = time_range_filter.latest_event_datetime
    latest_file_timestamp = time_range_filter.last_accessed_file_timestamp
  else:
    next_start_datetime = start_datetime
    # We use the latest file timestamp for the following use case:
    # When None is passed in as the input last_accessed_file_timestamp, and
    # there are many existing log files, we will have to open all the files,
    # if we find no new queries, we will reach here. We do not want to open
    # all those files in the subsequent calls. So we return the timestamp
    # of the latest file we saw. We will open that file in the next call.
    latest_file_timestamp = time_range_filter.last_accessed_file_timestamp

  return next_start_datetime, latest_file_timestamp, completed_queries

# Parses the data returned by the executing query ID servlet, returning a list of
# query IDs. We expect the servlet to return plain text with query IDs separated
# by a single space.
def _parse_executing_query_ids(opened_url):
  query_ids_string = opened_url.read()
  if query_ids_string is None:
    return []
  query_ids_string = query_ids_string.strip()
  if len(query_ids_string) == 0:
    return []
  return query_ids_string.split('\n')

# Parses the data returned by the executing query profile servlet, returning an
# encoded zlib compressed, binary compact thrift query profile or None if one
# could not be parsed.
def _parse_executing_query_profile(opened_url):
  profile = opened_url.read()
  if len(profile) == 0 or \
     profile.startswith(_PROFILE_NOT_FOUND_PREFIX) or \
     profile.startswith(_PROFILE_NOT_FOUND_PREFIX_OLD):
    return None
  try:
    return b64decode(profile)
  except TypeError, ex:
    THROTTLED_LOG.exception("Error decoding executing query profile %s" %
                            (profile, ))
    raise ex

class _Stats(object):
  def __init__(self):
    # The following counters offer an approximation of the cost of the query
    # monitor. They are approximate because they rely on psutil's process wide
    # CPU and IO counters. As such, it's possible for them to get thrown off by
    # some concurrent expensive operation by the agent. Still, they are better
    # than nothing.
    self.total_get_completed_query_calls = 0
    self.total_wallclock_time_seconds = 0.0
    self.total_user_cpu_time_seconds = 0.0
    self.total_system_cpu_time_seconds = 0.0
    # This is basically rchar from /proc/io.
    self.total_logical_bytes_read = 0
    # This is basically read_butes from /proc/io.
    self.total_storage_bytes_read = 0
    # This is the total size in bytes of all the profile data we've returned.
    self.total_completed_queries_size_bytes = 0
    self.total_executing_queries_size_bytes = 0
    # The following are lifetime counts of the number of queries and ids
    # retrieved by this query monitor.
    self.total_completed_queries = 0
    self.total_executing_queries = 0
    self.total_executing_queries_ids = 0
    # The following are lifetime error counts for this query monitor.
    self.total_slow_completed_query_warnings = 0
    self.total_unexpected_start_time_advances = 0
    self.total_get_executing_query_ids_exceptions = 0
    self.total_get_executing_query_profile_exceptions = 0
    self.total_completed_query_limit_reached = 0
    self.total_completed_query_parse_exceptions = 0

  def increment(self, attribute, delta=1):
    v = getattr(self, attribute)
    setattr(self, attribute, v + delta)

  def to_dict(self):
    return dict({
        'total_get_completed_query_calls':
          self.total_get_completed_query_calls,
        'completed_query_monitor_total_wallclock_time_seconds':
          self.total_wallclock_time_seconds,
        'completed_query_monitor_total_user_cpu_time_seconds':
          self.total_user_cpu_time_seconds,
        'completed_query_monitor_total_system_cpu_time_seconds':
          self.total_system_cpu_time_seconds,
        'completed_query_monitor_total_logical_bytes_read':
          self.total_logical_bytes_read,
        'completed_query_monitor_total_storage_bytes_read':
          self.total_storage_bytes_read,
        'total_completed_queries':
          self.total_completed_queries,
        'total_executing_queries':
          self.total_executing_queries,
        'total_executing_queries_ids':
          self.total_executing_queries_ids,
        'total_slow_completed_query_warnings':
          self.total_slow_completed_query_warnings,
        'total_unexpected_start_time_advances':
          self.total_unexpected_start_time_advances,
        'total_get_executing_query_ids_exceptions':
          self.total_get_executing_query_ids_exceptions,
        'total_get_executing_query_profile_exceptions':
          self.total_get_executing_query_profile_exceptions,
        'total_completed_query_limit_reached':
          self.total_completed_query_limit_reached,
        'total_completed_query_parse_exceptions':
          self.total_completed_query_parse_exceptions,
        'total_completed_queries_size_bytes':
          self.total_completed_queries_size_bytes,
        'total_executing_queries_size_bytes':
          self.total_executing_queries_size_bytes
        })

# This class exists to allow for easier testing of the overall Impala Daemon
# monitor. It also handles tracking various query monitoring statistics.
class QueryMonitor(object):
  # We take an optional urlopen_with_timeout function as a parameter to help with
  # testing. If passed it, this function can be used to mock out urllib requests.
  #
  # The completed_query_slow_log_threshold_seconds specifies a threshold for the
  # wallclock time that a call to get_completed_queries can take before we log
  # a warning that it is taking to long.
  def __init__(self,
               urlopen_with_timeout=urlopen_with_timeout,
               completed_query_slow_log_threshold_seconds=1.0):
    self._urlopen_with_timeout = urlopen_with_timeout
    self._completed_query_slow_log_threshold_seconds = \
        completed_query_slow_log_threshold_seconds
    self._stats = _Stats()

  # To enable easier testing.
  def _query_log_dir_exists(self, query_log_dir):
    return os.path.isdir(query_log_dir)

  def get_completed_queries(self,
                            log_path,
                            start_datetime,
                            end_datetime,
                            last_reported_query_id,
                            last_accessed_file_timestamp,
                            limit):
    start = time.time()
    proc = _get_cur_process()
    start_cpu = proc.cpu_times()
    start_io = _safe_get_io_counters()

    next_start_datetime, next_last_file_timestamp, completed_query_profiles = \
        _get_completed_queries(log_path,
                               start_datetime,
                               end_datetime,
                               last_reported_query_id,
                               limit,
                               self._stats,
                               last_accessed_file_timestamp)
    completed_query_count = 0
    if completed_query_profiles is not None:
      completed_query_count = len(completed_query_profiles)
      self._stats.increment('total_completed_queries', completed_query_count)
      for p in completed_query_profiles:
        self._stats.increment('total_completed_queries_size_bytes', len(p[1]))

    end = time.time()
    end_cpu = proc.cpu_times()
    end_io = _safe_get_io_counters()

    delta_wallclock = end - start
    delta_cpu_user = end_cpu.user - start_cpu.user
    delta_cpu_system = end_cpu.system - start_cpu.system
    delta_logical_bytes = end_io.read_count - start_io.read_count
    delta_storage_bytes = end_io.read_bytes - start_io.read_bytes

    if delta_wallclock > self._completed_query_slow_log_threshold_seconds:
      self._stats.increment('total_slow_completed_query_warnings')

    self._stats.increment('total_get_completed_query_calls')
    self._stats.increment('total_wallclock_time_seconds', delta_wallclock)
    self._stats.increment('total_user_cpu_time_seconds', delta_cpu_user)
    self._stats.increment('total_system_cpu_time_seconds', delta_cpu_system)
    self._stats.increment('total_logical_bytes_read', delta_logical_bytes)
    self._stats.increment('total_storage_bytes_read', delta_storage_bytes)

    return next_start_datetime, next_last_file_timestamp, completed_query_profiles

  # Gets a list of IDs of currently executing queries by hitting the Impala Daemon
  # web server at the input url.
  def get_executing_query_ids(self, url, timeout, username=None, password=None):
    if url is None:
      LOG.error("A url parameter is required.")
      return [], False

    try:
      opened_url = self._urlopen_with_timeout(url,
                                              timeout=timeout,
                                              username=username,
                                              password=password)
      query_ids = _parse_executing_query_ids(opened_url)
    except Exception, ex:
      THROTTLED_LOG.exception("Error fetching executing query ids at '%s'" %
                              (url,))
      self._stats.increment('total_get_executing_query_ids_exceptions')
      return [], False

    if query_ids is not None:
      self._stats.increment('total_executing_queries_ids', len(query_ids))

    return query_ids, True

  # Gets a base64 encoded compact thrift format runtime profile from the Impala
  # Daemon web server at the input url for the input query id. This will return
  # None on a timeout, an error or if the profile could not be found.
  def get_executing_query_profile(self, url, query_id, timeout, username=None,
                                  password=None):
    if url is None:
      LOG.error("A url parameter is required.")
      return None, False
    if query_id is None:
      LOG.error("A query_id parameter is required.")
      return None, False
    url_with_parms = "%s?query_id=%s" % (url, query_id)
    try:
      opened_url = self._urlopen_with_timeout(url_with_parms,
                                              timeout=timeout,
                                              username=username,
                                              password=password)
      profile = _parse_executing_query_profile(opened_url)
    except Exception, ex:
      THROTTLED_LOG.exception("Error fetching executing query profile at '%s'" %
                              (url,))
      self._stats.increment('total_get_executing_query_profile_exceptions')
      return None, False

    if profile is not None:
      self._stats.increment('total_executing_queries')
      self._stats.increment('total_executing_queries_size_bytes', len(profile))

    return profile, True

  def get_stats(self):
    return self._stats.to_dict()

# This extension of the QueryMonitor class handles the bookkeeping necessary to
# monitor the stream of queries in an Impala Daemon log file.
class QueryStreamMonitor(QueryMonitor):
  def __init__(self,
               start_datetime=None,
               urlopen_with_timeout=urlopen_with_timeout,
               completed_query_slow_log_threshold_seconds=1.0):
    QueryMonitor.__init__(self,
                          urlopen_with_timeout,
                          completed_query_slow_log_threshold_seconds)
    if start_datetime is None:
      start_datetime = datetime.datetime.utcnow()

    # These members tracks the start time we will use in our next check of the log
    # for completed impala queries.
    self._completed_queries_monitoring_start_datetime = start_datetime
    self._completed_queries_monitoring_last_reported_query_id = None
    # This variable will be passed to QueryMonitor in each pass. Any log file
    # with timestamp older than this variable will be ignored and will not be
    # opened by TimeRangeFilter.
    self._last_accessed_file_timestamp = None

  def get_completed_queries(self,
                            log_file,
                            end_datetime=None,
                            completed_query_report_limit=1000):

    if end_datetime is None:
      end_datetime = datetime.datetime.utcnow()

    next_start_datetime, next_last_file_timestamp, completed_queries = \
      QueryMonitor.get_completed_queries(
        self,
        log_file,
        self._completed_queries_monitoring_start_datetime,
        end_datetime,
        self._completed_queries_monitoring_last_reported_query_id,
        self._last_accessed_file_timestamp,
        completed_query_report_limit)

    if next_start_datetime is None:
      # If we got no time to use for our next monitoring call, just go ahead and
      # use the current time as the new start time. This should only happen if
      # there's funny business going on.
      THROTTLED_LOG.warning("Call to get completed queries did not return a "
                            "new start time. Advancing query monitoring "
                            "period.")
      self._completed_queries_monitoring_start_datetime = \
          datetime.datetime.utcnow()
      self._stats.increment('total_unexpected_start_time_advances')
    else:
      self._completed_queries_monitoring_start_datetime = next_start_datetime
    self._last_accessed_file_timestamp = next_last_file_timestamp

    if len(completed_queries) > 0:
      self._completed_queries_monitoring_last_reported_query_id = \
          completed_queries[-1][0]

    return completed_queries

  def get_stats(self):
    stats = QueryMonitor.get_stats(self)
    stats['completed_queries_monitoring_start_datetime'] = \
        str(self._completed_queries_monitoring_start_datetime)
    stats['completed_queries_monitoring_last_reported_query_id'] = \
        self._completed_queries_monitoring_last_reported_query_id
    return stats

_ROLETYPE = "IMPALAD"
_SERVICETYPE = "IMPALA"

# Section and key names in cloudera-monitor.properties
# Since CM 4.5
_SECTION = 'impalad'
_SERVICE_NAME = 'service_name'
_SERVICE_VERSION = 'service_version'
_ROLE_NAME = 'role_name'
_HOST = 'host'
_LOG_DIR = 'log_dir'
_COLLECT_INTERVAL = 'collect_interval'
_WEBSERVER_PORT = 'impalad_webserver_port'
# Added in CM 4.6.1
_USE_SSL = 'use_ssl'
_WEBSERVER_USERNAME = 'webserver_username'
_WEBSERVER_PASSWORD = 'webserver_password'
# Added in CM 5.2
_SCRATCH_DIRS = 'scratch_dirs'
# Added in CM 5.4
_HEAP_DUMP_DIR = 'oom_heap_dump_dir'
_SERVICE_RELEASE = 'service_release'

# This is not part of the Impala config -- you just configure the log directory,
# so we just hardcode it here.
QUERY_LOG_SUBDIR = "profiles"
QUERY_LOG_FILE_NAME_PREFIX = "impala_profile_log"

# Added in CM 4.6
_EXECUTING_QUERIES_TIMEOUT_SECONDS = 'executing_queries_timeout_seconds'
_QUERY_MONITORING_PERIOD_SECONDS = 'query_monitoring_period_seconds'

class QueryMonitorMetricCollector(WebServerMetricCollector):
  """
  A collector for query profiles and metrics related to that. The only reason
  it extends WebServerMetricCollector is to url open using
  collect_and_parse_and_return().
  """
  def __init__(self, daemon, adapter, query_monitor = None):
    WebServerMetricCollector.__init__(self,
                                      _SECTION,
                                      adapter,
                                      None,
                                      _WEBSERVER_USERNAME,
                                      _WEBSERVER_PASSWORD,
                                      name="QueryMonitorMetricCollector")

    self._daemon = daemon
    self._conf = None

    # Config data
    self._name = None
    self._service_name = None
    self._log_dir = None
    self._username = None
    self._password = None

    self._query_log_dir = None
    self._query_log_file = None
    self._executing_query_ids_url = None
    self._executing_query_profile_url = None
    # This is the timeout we'll use for all communication with the Impala Daemon
    # web server for executing query information. This is ultimately read from
    # the monitor's properties file.
    self._executing_queries_timeout_seconds = 5.0
    # This is the period of the query montioring thread. This is ultimately read
    # from the monitor's properties file.
    self._query_monitoring_period_seconds = 1.0

    # The query_monitor object is just a set of function pointers we use to do
    # Impala query monitoring. It exists to allow for easier testing and for
    # keeping track of some statistics
    if query_monitor is None:
      query_monitor = QueryStreamMonitor()
    self._query_monitor = query_monitor

    # This is a thread from which all query monitring occurs. This thread polls
    # the Impala Daemon web server for executing query information, polls the
    # log file for completed query infomation and then publishes any information
    # it finds to the appropriate firehose.
    self._query_monitoring_thread = None

    # A counter tracking the number of failures when monitoring executing queries.
    self._executing_query_monitoring_failures = 0
    # A counter tracking the number of failures when monitoring completed queries.
    self._completed_query_monitoring_failures = 0

    # We scrape this from the Impala Daemon metrics page and include it in the
    # messages we send with query information.
    self._impala_version = None

  def update_with_conf(self, conf):
    WebServerMetricCollector.update_with_conf(self, conf)

    try:
      service = conf.get(_SECTION, _SERVICE_NAME)
      name = conf.get(_SECTION, _ROLE_NAME)
      log_dir = conf.get(_SECTION, _LOG_DIR)
      host = conf.get(_SECTION, _HOST)
      web_port = conf.get(_SECTION, _WEBSERVER_PORT)
      use_ssl = conf.getboolean_with_default(_SECTION, _USE_SSL, False)
      executing_queries_timeout_seconds = conf.getfloat_with_default(
        _SECTION, _EXECUTING_QUERIES_TIMEOUT_SECONDS, 5.0)
      query_monitoring_period_seconds = conf.getfloat_with_default(
        _SECTION, _QUERY_MONITORING_PERIOD_SECONDS, 1.0)
    except ValueError, ex:
      LOG.exception("Error updating with configurations")
      raise ex
    self._conf = conf

    self._service_name = service
    self._name = name
    self._log_dir = log_dir
    self._username = self._username_value
    self._password = self._password_value

    executing_query_ids_url = \
        self._compute_executing_query_ids_url(host, web_port, use_ssl)
    if executing_query_ids_url is None:
      raise Exception("No executing_query_ids_url")
    executing_query_profile_url = \
        self._compute_executing_query_profile_url(host, web_port, use_ssl)
    if executing_query_profile_url is None:
      raise Exception("No executing_query_profile_url")
    self._set_query_monitoring_thread_period(query_monitoring_period_seconds)
    self._executing_query_ids_url = executing_query_ids_url
    self._executing_query_profile_url = executing_query_profile_url
    self._executing_queries_timeout_seconds = executing_queries_timeout_seconds
    self._query_monitoring_period_seconds = query_monitoring_period_seconds

    self._query_log_dir = os.path.join(self._log_dir,
                                       QUERY_LOG_SUBDIR)
    self._query_log_file = os.path.join(self._query_log_dir,
                                        QUERY_LOG_FILE_NAME_PREFIX)

  def on_start_collecting(self):
    self._start_query_monitoring_thread()

  def on_stop_collecting(self):
    self._stop_query_monitoring_thread()

  def collect_and_parse(self, conf):
    # Metrics are kept already as part of query monitoring loops.
    start = time.time()
    now = time.time()
    self._metrics_tuple = constants.WEB_METRICS_COLLECTION_GOOD, \
                         now, \
                         int((now - start) * 1000)

  def _add_metrics(self, service_version, update, collected_metrics, duration):
    update.add_metric(schema.EXECUTING_QUERY_MONITORING_FAILURE,
                      self._executing_query_monitoring_failures)
    update.add_metric(schema.COMPLETED_QUERY_MONITORING_FAILURE,
                      self._completed_query_monitoring_failures)

  def is_ready_to_report(self, conf, pid):
    return True

  ### Query Monitoring Calls ###

  def _start_query_monitoring_thread(self):
    """Starts the query monitoring thread if necessary."""
    if self._query_monitoring_thread is not None:
      return
    LOG.info("Starting query monitoring thread.")
    self._query_monitoring_thread = StoppableThread(
      "ImpalaDaemonQueryMonitoring",
      self._query_monitoring_period_seconds,
      self._check_for_queries)
    self._query_monitoring_thread.start()

  def _stop_query_monitoring_thread(self):
    """Stops the query monitoring thread if necessary."""
    if self._query_monitoring_thread is None:
      return
    LOG.info("Stopping query monitoring thread.")
    self._query_monitoring_thread.should_stop()
    self._query_monitoring_thread = None

  def _set_query_monitoring_thread_period(self, period):
    query_monitoring_thread = self._query_monitoring_thread
    if query_monitoring_thread is None:
      return
    if period > 0:
      query_monitoring_thread.set_interval(period)
    else:
      # We don't disable the thread for simplicity -- keeping the lifecycle of
      # the thread tied to the lifecycle of the monitor. Instead we adjust the
      # period to just once a minute and make _check_for_queries do nothing.
      query_monitoring_thread.set_interval(60)

  def _get_impala_version(self):
    result, time, duration = \
      WebServerMetricCollector._collect_and_parse_and_return(
              self, self._conf)
    if result is None or isinstance(result, CollectionError):
      # We have already logged...
      return None
    json_parser = result['jsonParser']
    return json_parser.json.get('impala-server.version', None)

  def _check_for_queries(self):
    """
    This is the main method of the query monitoring thread. It checks for
    executing queries first, gets their profiles, and then checks for completed
    queries. If any information is collected, it is sent to HMON immediately.
    """
    if (self._executing_query_ids_url is None or
        self._executing_query_profile_url is None or
        self._query_monitoring_period_seconds <= 0):
      return

    if self._impala_version is None:
      self._impala_version = self._get_impala_version()
    if self._impala_version is None:
      return

    executing_query_ids, monitored_executing_query = \
      self._query_monitor.get_executing_query_ids(
        self._executing_query_ids_url,
        self._executing_queries_timeout_seconds,
        self._username,
        self._password)

    executing_query_profile_map = {}
    for query_id in executing_query_ids:
      profile, monitored_executing_query_profile = \
        self._query_monitor.get_executing_query_profile(
          self._executing_query_profile_url,
          query_id,
          self._executing_queries_timeout_seconds,
          self._username,
          self._password)
      if profile is not None:
        executing_query_profile_map[query_id] = ImpalaRuntimeProfile(profile)
      if not monitored_executing_query_profile:
        monitored_executing_query = False

    completed_query_ids, completed_query_profiles, monitored_completed_query = \
        self._get_completed_query_profiles()

    if not monitored_executing_query:
      self._executing_query_monitoring_failures += 1
    if not monitored_completed_query:
      self._completed_query_monitoring_failures += 1

    # We take a quick pass through the completed profiles to see if any of the
    # completed queries are actually done now. We remove them if they are.
    for query_id in completed_query_ids:
      executing_query_profile_map.pop(query_id, None)

    if (len(executing_query_profile_map) > 0 or
        len(completed_query_profiles) > 0):
      self._daemon.send_impala_query_update(
        ImpalaQueryUpdate(role_name=self._name,
                          service_name=self._service_name,
                          impala_version=self._impala_version,
                          executing_query_profiles=\
                            executing_query_profile_map.values(),
                          completed_query_profiles=\
                            completed_query_profiles))

  ### Completed Query Monitoring Calls ###

  def _get_completed_query_profiles(self):
    """
    Calls into the query monitor to get information on newly completed queries,
    advancing our state tracking what constitutes "newly completed" queries as
    appropriate.

    This function returns two lists in matching order. The first list has the
    query IDs of the queries, the second is the profiles as ImpalaRuntimeProfile
    objects.

    This should only be called by the query monitoring thread.
    """
    query_log_dir = self._query_log_dir
    query_log_file = self._query_log_file
    if query_log_dir is None or query_log_file is None:
      LOG.debug("Unable to fetch completed queries. No log file specified.")
      return [], [], False

    # We check for the presence of the query log directory to detect whether we
    # should even do query monitoring. This directory was introduced in Impala
    # 1.0.1.
    if not self._query_monitor._query_log_dir_exists(query_log_dir):
      return [], [], False

    try:
      completed_queries = \
          self._query_monitor.get_completed_queries(query_log_file)

      completed_query_ids = []
      completed_query_profiles = []
      for query_id, profile in completed_queries:
        completed_query_ids.append(query_id)
        completed_query_profiles.append(ImpalaRuntimeProfile(profile))

      return completed_query_ids, completed_query_profiles, True

    except Exception, ex:
      LOG.exception("Error fetching completed queries from '%s'" %
                    (query_log_file,))
      return [], [], False

  def _compute_executing_query_ids_url(self, host, port, use_ssl):
    if host is None or port is None:
      return None
    protocol = "http"
    if use_ssl:
      protocol = "https"
    return "%s://%s:%s/inflight_query_ids" % (protocol, host, port)

  def _compute_executing_query_profile_url(self, host, port, use_ssl):
    if host is None or port is None:
      return None
    protocol = "http"
    if use_ssl:
      protocol = "https"
    return "%s://%s:%s/query_profile_encoded" % (protocol, host, port)

  def get_stats(self):
    return self._query_monitor.get_stats()
