# Copyright (c) 2011-2012 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import re

from cmf.monitor import constants
from cmf.monitor.json_metrics_utils import ParseError, DictWithAttr

class JsonParserJmx(object):
  def __init__(self, json_dict, key, context={}):
    """
    May raise ParseError if the json structure is not what we expect. See the
    _get_beans function for an explanation of how the context argument is used.
    The key argument should correspond to an entry in the BEANS dictionary in
    this file.
    """
    beans = BEANS[key]
    json_beans = _get_beans(json_dict, beans, context)
    for bean in beans:
      setattr(self, bean.get_alias(), json_beans.get(bean.get_alias()))

def _get_beans(json_dictionary, beans, context):
  """
  Given a Hadoop or Cloudera Manager JMX json dict, returns a map of beans found
  matching the input set of bean objects as a dictionary from bean alias to the
  bean(s) as a dictionary of metric values. May raise ParseError.

  Returns a DictWithAttr.

  The beans argument is a collection of Bean objects that must implement four
  methods: matches(attr, context), get_alias(), is_singleton() and
  get_discriminator_attr().

  The context argument is an arbitrary dictionary passed through to the matches
  method invocation on the beans. It is used to pass configuration information
  from the monitor for which we are collecting into the parsing process. For
  example, we use it to pass port information for a NameNode to distinguish the
  RPC statistics for the client facing server vs. the service facing server.
  For example, consider the following:

     {
    "name" : "Hadoop:service=NameNode,name=RpcDetailedActivityForPort8020",
    "modelerType" : "RpcDetailedActivityForPort8020",
    ...
    "NumOpenConnections" : 5,
    ...
    }, {
    "name" : "Hadoop:service=NameNode,name=RpcDetailedActivityForPort8022",
    "modelerType" : "RpcDetailedActivityForPort8022",
    ...
    "NumOpenConnections" : 5,
    ...
    },

  If we want to distinguish these two beans with different metrics we need some
  way to know with which metrics we ought to associate them. To do this we use
  the PortAwareMetricsBean, leveraging the input context below.
  """
  bean_dict_list = json_dictionary.get('beans')
  if bean_dict_list is None:
    err = "Cannot find 'beans' list in input json"
    raise ParseError(err)

  if not isinstance(bean_dict_list, list):
    err = "Expecting a 'beans' list. Got %r" % (bean_dict_list, )
    raise ParseError(err)

  ret = {}
  for d in map(lambda x: DictWithAttr(x), bean_dict_list):
    for bean in beans:
      attr = bean.get_discriminator_attr()
      discriminator = d.get(attr, None)
      if discriminator == None:
        continue
      matched_obj = bean.matches(discriminator, context)
      if matched_obj:
        additional_attr = bean.get_additional_bean_attr(matched_obj)
        if additional_attr is not None:
          for key in additional_attr:
            d[key] = additional_attr[key]
        alias = bean.get_alias()
        if bean.is_singleton() == True:
          ret[alias] = d
        else:
          if alias in ret.keys():
            s = ret[alias]
            s.append(d)
            ret[alias] = s
          else:
            s = [d]
            ret[alias] = s
        break

  return ret

class Bean:
  """
  Holds metadata used to represent a bean or beans from which we collect metrics.
  """
  def __init__(self, regex, alias, singleton=True, discriminator_attr='name'):
    # A regular expression used to identify the bean(s).
    self._regex = regex
    # The name we'll use internally for the bean(s).
    self._alias = alias
    # Indicates if this represents a singleton bean
    self._singleton = singleton
    # Indicates the attribute that is to be used when attempting to match
    # against this Bean. By default, we use the 'name' attribute.
    # When inspecting Solr JMX MBeans in JSON, we use the 'category' attribute
    # instead as the discriminator or the matcher attribute.
    # For e.g: By using a regular expression for 'CORE' and 'category'
    # as the discriminator_attr, we can successfully extract all CORE
    # MBeans from a JMX JSON snapshot such as:
    # ..
    # {
    #   "name" : "solr/mr_col_shard2_replica1:type=core,id=mr_col_shard2_replica1",
    #   "modelerType" : "org.apache.solr.core.JmxMonitoredMap$SolrDynamicMBean",
    #   "category" : "CORE",
    #   "source" : "$URL$",
    #   "description" : "SolrCore",
    #   "name" : "mr_col_shard2_replica1",
    #   "version" : "1.0",
    #   "coreHashCode" : "756266007",
    #   "coreName" : "mr_col_shard2_replica1",
    #   "startTime" : "Tue Nov 12 13:59:15 PST 2013",
    #   "refCount" : 1,
    #   "indexDir" : "hdfs://abs-1.ent.cloudera.com:8020/solr/mr_col/core_node4/data/index/",
    #   "aliases" : "[mr_col_shard2_replica1]",
    #   "collection" : "mr_col",
    #   "shard" : "shard2"
    # }
    #
    self._discriminator_attr = discriminator_attr

  # Whether the input bean attr matches this bean defition. See _get_beans for
  # an explanation of the context parameter (which this class ignores entirely)
  def matches(self, attr, context):
    return self._regex.match(attr)

  # Whether the input bean attr matches this bean definition. This will only be
  # called by metric-tool and so must work without context information.
  def matches_for_metric_tool(self, attr):
    return self._regex.match(attr)

  def get_alias(self):
    return self._alias

  def is_singleton(self):
    return self._singleton

  def get_discriminator_attr(self):
    return self._discriminator_attr

  def get_additional_bean_attr(self, matched_obj):
    return None

# This can be used to make a PortAwareMetricsBean match any port.
MATCH_ANY_PORT = "match-any-port"

class PortAwareMetricsBean:
  """
  Holds metadata used to represent a bean from which we collect metrics. This is
  distinct from Bean in that it can take a port number from a metrics access
  context to customize its extraction behavior. This bean assumes that the input
  regex has a single extraction pattern.
  """
  def __init__(self, regex, alias, port_name_in_context, singleton=True,
               discriminator_attr='name'):
    # A regular expression used to identify the bean.
    self._regex = regex
    # The name we'll use internally for the bean.
    self._alias = alias
    # The name we use to lookup the port value in the access context.
    self._port_name_in_context = port_name_in_context
    # Indicates if this represents a singleton bean
    self._singleton = singleton
    # Indicates the discriminating attribute to be used when inspecting beans.
    # By default, we use the 'name' attribute. See further details above in
    # the comments assocated with '_discriminator_attr' in the constructor
    # for Bean.
    self._discriminator_attr = discriminator_attr

  # See Bean above. This method looks up the port we are supposed to match from
  # the input context via the name used in construction. That value is compared
  # to the value extracted from the input bean name. If there is a match, or if
  # context told us to MATCH_ANY_PORT we match, otherwise we do not. See the
  # example in _get_beans for more detail.
  def matches(self, attr, context):
    port = context.get(self._port_name_in_context)
    if port is None:
      return False
    m = self._regex.match(attr)
    if m is None:
      return False
    if port == MATCH_ANY_PORT:
      return True
    return port == int(m.group(1))

  def matches_for_metric_tool(self, attr):
    return self._regex.match(attr)

  def get_alias(self):
    return self._alias

  def is_singleton(self):
    return self._singleton

  def get_discriminator_attr(self):
    return self._discriminator_attr

  def get_additional_bean_attr(self, matched_obj):
    return None

class MatchedGroupReplacement:
  """
  Contains information about the possible values for a matched group in a regex
  for a bean.
  """
  def __init__(self, attr_name, replacement_map):

    # The attribute name to add to the matched bean.
    self._attr_name = attr_name

    # A map containing possible replacement for specific matched
    # values. See get_replacement for more details.
    self._replacement_map = replacement_map

  def get_replacement(self, value):
    """
    If the replacement map is not None and there's a replacement
    for this value, returns the replacement. Otherwise returns
    'value'.
    """
    if self._replacement_map is not None and \
       value in self._replacement_map:
         return self._replacement_map[value]
    return value

  def get_attr_name_to_use(self):
    return self._attr_name

class BeanWithAdditionalAttributes:
  """
  Holds metadata used to represent a bean from which we collect metrics. This is
  distinct from Bean in that it takes a set of MatchedGroupReplacement and
  adds all matched replacements as attributes to the bean.
  """
  def __init__(self, regex, alias, singleton=True, discriminator_attr='name',
               groups_replacement_info = {}):
    # A regular expression used to identify the bean.
    self._regex = regex
    # The name we'll use internally for the bean.
    self._alias = alias
    # Indicates if this represents a singleton bean
    self._singleton = singleton
    # Indicates the discriminating attribute to be used when inspecting beans.
    # By default, we use the 'name' attribute. See further details above in
    # the comments assocated with '_discriminator_attr' in the constructor
    # for Bean.
    self._discriminator_attr = discriminator_attr
    # a map between a gropu index and the MatchedGroupReplacement for that
    # group.
    self._groups_replacement_info = groups_replacement_info

  # Whether the input bean attr matches this bean defition. See _get_beans for
  # an explanation of the context parameter (which this class ignores entirely)
  def matches(self, attr, context):
    return self._regex.match(attr)

  # Whether the input bean attr matches this bean definition. This will only be
  # called by metric-tool and so must work without context information.
  def matches_for_metric_tool(self, attr):
    return self._regex.match(attr)

  def get_alias(self):
    return self._alias

  def is_singleton(self):
    return self._singleton

  def get_discriminator_attr(self):
    return self._discriminator_attr

  def get_additional_bean_attr(self, matched_obj):
    if matched_obj.lastindex < 1:
      return None
    ret = {}
    for group_index in range(1, matched_obj.lastindex + 1):
      if not group_index in self._groups_replacement_info:
        continue
      replacement_info = self._groups_replacement_info[group_index]
      ret[replacement_info.get_attr_name_to_use()] = \
          replacement_info.get_replacement(matched_obj.group(group_index))
    return ret

_JAVA_MEMORY_BEAN = Bean(re.compile('java.lang:type=Memory$'), 'java_memory')
_JAVA_DIRECT_MEMORY_BEAN = Bean(re.compile('java.nio:type=BufferPool,name=direct$'),
                                'java_direct_memory')
_JAVA_GC_BEANS = Bean(re.compile('java.lang:type=GarbageCollector'), 'java_gc',
                      singleton=False)
PAUSE_ALIAS = 'pause_detail'
_MGMT_PAUSE_REPLACMENT_GROUP = \
  MatchedGroupReplacement(constants.PAUSE_BEAN_ATTR_NAME,
                          {"pauseTimeMs": constants.PAUSE_TIME_ALIAS,
                          "numPausesDetected": constants.PAUSES_ALIAS})
_MGMT_JVM_PAUSE_MONITOR_BEAN = BeanWithAdditionalAttributes(
    re.compile('com.cloudera.enterprise.debug:type=JvmPauseMonitor,name=(.+)'),
    PAUSE_ALIAS,
    singleton=False,
    groups_replacement_info={1: _MGMT_PAUSE_REPLACMENT_GROUP})
_EVENT_STATS_BEAN = \
    Bean(re.compile('com.cloudera.cmf.event.publish:type=EventStorePublisherWithRetryMXBean'),
         'event_stats')

def _make_jvm_metrics_bean(name):
  return Bean(re.compile('Hadoop:service=%s,name=JvmMetrics' % name), 'jvm')

def _make_ugi_metrics_bean(name):
  return Bean(re.compile('Hadoop:service=%s,name=UgiMetrics' % name), 'ugi')

def _make_rpc_metrics_bean(name):
  return Bean(re.compile('Hadoop:service=%s,name=RpcActivityForPort\d+' % name),
              'rpc')

def _make_detailed_rpc_metrics_bean(name):
  return Bean(re.compile('Hadoop:service=%s,name=RpcDetailedActivityForPort\d+' % name),
              'detailed_rpc')

def _make_port_aware_rpc_metrics_bean(name, alias, port_name_in_context):
  return PortAwareMetricsBean(
    re.compile('Hadoop:service=%s,name=RpcActivityForPort(\d+)' % name),
    alias,
    port_name_in_context)

def _make_port_aware_detailed_rpc_metrics_bean(name, alias, port_name_in_context):
  return PortAwareMetricsBean(
    re.compile('Hadoop:service=%s,name=RpcDetailedActivityForPort(\d+)' % name),
    alias,
    port_name_in_context)

def _make_metrics_control_metrics_bean(name):
  return Bean(re.compile('Hadoop:service=%s,name=MetricsSystem,sub=Control' % name),
              'metrics_control')

def _make_metrics_stats_metrics_bean(name):
  return Bean(re.compile('Hadoop:service=%s,name=MetricsSystem,sub=Stats' % name),
              'metrics_stats')

# Some beans we skip entirely, even though they are Hadoop beans.
# For YARN it's because we haven't gotten to that yet: OPSAPS-13366
# SOLR_SERVER beans are treated seperately in its monitor's
# 'json_parser_jmx' as it involves use of a multi-level scheme
# to determine the beans to ignore.
IGNORED_BEANS = dict(
    NAMENODE=[
        # Ignore the StartupProgress beans because they get set when the role
        # stars, and then don't change, so they don't really fit into our model.
        'Hadoop:service=NameNode,name=StartupProgress',
        ],
    SECONDARYNAMENODE=[
        # See namenode comments for why we ignore this bean
        'Hadoop:service=SecondaryNameNode,name=StartupProgress',
        ],
    DATANODE=[
        # See namenode comments for why we ignore this bean
        'Hadoop:service=DataNode,name=StartupProgress',
        ],
    NODEMANAGER=[
        # This bean appearing in our samples appears to be a bug. We have pinged
        # the MR team about why it's there.
        'Hadoop:service=NodeManager,name=StartupProgress',
        'Hadoop:service=NodeManager,name=ShuffleMetrics',
        ],
    JOURNALNODE=[
        # See namenode comments for why we ignore this bean
        'Hadoop:service=JournalNode,name=StartupProgress',
        # The journal node status is a json structure we don't want to try to
        # parse here
        'Hadoop:service=JournalNode,name=JournalNodeInfo',
        ],
    RESOURCEMANAGER=[
        # The following 2 beans don't really have useful information,
        # so we are choosing to ignore them.
        'Hadoop:service=ResourceManager,name=ClusterMetrics',
        'Hadoop:service=ResourceManager,name=RMNMInfo'
        ],
    JOBHISTORY=[
        # This bean appearing in our samples appears to be a bug. We have pinged
        # the MR team about why it's there.
        'Hadoop:service=JobHistoryServer,name=StartupProgress'
        ],
    REGIONSERVER=[
        # See namenode comments for why we ignore this bean
        'Hadoop:service=HBase,name=StartupProgress'
        ]
    )

# Sometimes we want all beans of a particular type to be ignored.
# For e.g. for MGMT roles, JMX Memory / GarbageCollector metrics are treated
# specially - see json_metrics_utils.add_derived_jvm_metrics_jmx and
# json_metrics_utils.add_derived_gc_metrics_jmx where metrics are derived
# from a composite in the bean (HeapMemoryUsage and NonHeapMemoryUsage
# for JMX Memory) or across all beans of the same type (GarbageCollector)
# In such cases we ignore all beans of that particular type when making
# updates to the schema.
IGNORED_BEAN_TYPES = dict(
    ACTIVITYMONITOR=[
        'java.lang:type=Memory',
        'java.lang:type=GarbageCollector'
        ],
    EVENTSERVER=[
        'java.lang:type=Memory',
        'java.lang:type=GarbageCollector'
        ],
    HOSTMONITOR=[
        'java.lang:type=Memory',
        'java.lang:type=GarbageCollector'
        ],
    NAVIGATOR=[
        'java.lang:type=Memory',
        'java.lang:type=GarbageCollector'
        ],
    SERVICEMONITOR=[
        'java.lang:type=Memory',
        'java.lang:type=GarbageCollector'
        ],
    REPORTSMANAGER=[
        ]
    )
# The NameNode can have two RPC servers. See OPSAPS-14550 for more information.
NAMENODE_CLIENT_RPC_PORT_NAME = 'client_rpc_port'
NAMENODE_SERVICE_RPC_PORT_NAME = 'service_rpc_port'

BEANS = dict(
    NAMENODE=[
        _JAVA_MEMORY_BEAN,
        _make_jvm_metrics_bean('NameNode'),
        _make_port_aware_rpc_metrics_bean('NameNode',
                                          'rpc',
                                          NAMENODE_CLIENT_RPC_PORT_NAME),
        _make_port_aware_detailed_rpc_metrics_bean('NameNode',
                                                   'detailed_rpc',
                                                   NAMENODE_CLIENT_RPC_PORT_NAME),
        _make_port_aware_rpc_metrics_bean('NameNode',
                                          'service_rpc',
                                          NAMENODE_SERVICE_RPC_PORT_NAME),
        _make_port_aware_detailed_rpc_metrics_bean('NameNode',
                                                   'service_detailed_rpc',
                                                   NAMENODE_SERVICE_RPC_PORT_NAME),
        _make_detailed_rpc_metrics_bean('NameNode'),
        _make_ugi_metrics_bean('NameNode'),
        _make_metrics_control_metrics_bean('NameNode'),
        _make_metrics_stats_metrics_bean('NameNode'),
        Bean(re.compile('Hadoop:service=NameNode,name=NameNodeInfo'),
             'namenode_info'),
        Bean(re.compile('Hadoop:service=NameNode,name=NameNodeActivity'),
             'namenode_activity'),
        Bean(re.compile('Hadoop:service=NameNode,name=FSNamesystem$'),
             'fsnamesystem'),
        Bean(re.compile('Hadoop:service=NameNode,name=FSNamesystemState'),
             'fsnamesystem_state'),
        Bean(re.compile('Hadoop:service=NameNode,name=IPCLoggerChannel'),
             'ipc_logger_channel'),
        ],
    SECONDARYNAMENODE=[
        _JAVA_MEMORY_BEAN,
        _make_jvm_metrics_bean('SecondaryNameNode'),
        _make_rpc_metrics_bean('SecondaryNameNode'),
        _make_detailed_rpc_metrics_bean('SecondaryNameNode'),
        _make_ugi_metrics_bean('SecondaryNameNode'),
        _make_metrics_control_metrics_bean('SecondaryNameNode'),
        _make_metrics_stats_metrics_bean('SecondaryNameNode'),
        ],
    DATANODE=[
        _JAVA_MEMORY_BEAN,
        _make_jvm_metrics_bean('DataNode'),
        _make_rpc_metrics_bean('DataNode'),
        _make_detailed_rpc_metrics_bean('DataNode'),
        _make_ugi_metrics_bean('DataNode'),
        _make_metrics_control_metrics_bean('DataNode'),
        _make_metrics_stats_metrics_bean('DataNode'),
        Bean(re.compile('Hadoop:service=DataNode,name=DataNodeActivity'),
             'datanode_activity'),
        Bean(re.compile('Hadoop:service=DataNode,name=DataNodeInfo'),
             'datanode_info'),
        Bean(re.compile('Hadoop:service=DataNode,name=FSDatasetState'),
             'fs_dataset_state'),
        ],
    JOURNALNODE=[
        _JAVA_MEMORY_BEAN,
        _make_jvm_metrics_bean('JournalNode'),
        _make_rpc_metrics_bean('JournalNode'),
        _make_detailed_rpc_metrics_bean('JournalNode'),
        _make_ugi_metrics_bean('JournalNode'),
        _make_metrics_control_metrics_bean('JournalNode'),
        _make_metrics_stats_metrics_bean('JournalNode'),
        Bean(re.compile('Hadoop:service=JournalNode,name=Journal-.+'),
             'journal'),
        ],
    SERVICEMONITOR=[
        _JAVA_MEMORY_BEAN,
        _JAVA_GC_BEANS,
        _EVENT_STATS_BEAN,
        Bean(re.compile("com.cloudera.cmf.event.publish:type=PipelineStageMXBean,stage=roleStageTSStore"),
             'role_stage'),
        _MGMT_JVM_PAUSE_MONITOR_BEAN
        ],
    HOSTMONITOR=[
        _JAVA_MEMORY_BEAN,
        _JAVA_GC_BEANS,
        _EVENT_STATS_BEAN,
        Bean(re.compile("com.cloudera.cmf.event.publish:type=PipelineStageMXBean,stage=HMONHostStageReceiver"),
             'host_stage'),
        _MGMT_JVM_PAUSE_MONITOR_BEAN
        ],
    ACTIVITYMONITOR=[
        _JAVA_MEMORY_BEAN,
        _JAVA_GC_BEANS,
        Bean(re.compile("com.cloudera.cmf.event.publish:type=PipelineStageMXBean,stage=activityTree"),
             'activity_tree_pipeline'),
        Bean(re.compile("com.cloudera.cmf.event.publish:type=PipelineStageMXBean,stage=activityMonitoring"),
             'activity_monitor_pipeline'),
        Bean(re.compile("com.cloudera.enterprise:type=PeriodicEnterpriseServiceMXBean,service=ActivityAndAttemptPurgeService"),
             'activity_monitor_purger'),
        _MGMT_JVM_PAUSE_MONITOR_BEAN
        ],
    EVENTSERVER=[
        _JAVA_MEMORY_BEAN,
        _JAVA_GC_BEANS,
        Bean(re.compile('com.cloudera.cmf.eventcatcher.server:type=EventCatcherServiceMXBean'),
             'event_server'),
        Bean(re.compile("com.cloudera.cmf.event.publish:type=PipelineStageMXBean,stage=tagger-writer"),
             'event_pipeline')
        ],
    REPORTSMANAGER=[
        _JAVA_MEMORY_BEAN,
        _JAVA_GC_BEANS
        ],
    NODEMANAGER=[
        _JAVA_MEMORY_BEAN,
        _make_jvm_metrics_bean('NodeManager'),
        _make_rpc_metrics_bean('NodeManager'),
        _make_detailed_rpc_metrics_bean('NodeManager'),
        _make_ugi_metrics_bean('NodeManager'),
        _make_metrics_control_metrics_bean('NodeManager'),
        _make_metrics_stats_metrics_bean('NodeManager'),
        Bean(re.compile("Hadoop:service=NodeManager,name=NodeManagerMetrics"),
             'nodemanager_metrics')
        ],
    RESOURCEMANAGER=[
        _JAVA_MEMORY_BEAN,
        _make_jvm_metrics_bean('ResourceManager'),
        _make_rpc_metrics_bean('ResourceManager'),
        _make_detailed_rpc_metrics_bean('ResourceManager'),
        _make_ugi_metrics_bean('ResourceManager'),
        _make_metrics_control_metrics_bean('ResourceManager'),
        _make_metrics_stats_metrics_bean('ResourceManager')
        ],
    JOBHISTORY=[
        _JAVA_MEMORY_BEAN,
        _make_jvm_metrics_bean('JobHistoryServer'),
        _make_rpc_metrics_bean('JobHistoryServer'),
        _make_detailed_rpc_metrics_bean('JobHistoryServer'),
        _make_ugi_metrics_bean('JobHistoryServer'),
        _make_metrics_control_metrics_bean('JobHistoryServer'),
        _make_metrics_stats_metrics_bean('JobHistoryServer')
        ],
    NAVIGATOR=[
        _JAVA_MEMORY_BEAN,
        _JAVA_GC_BEANS,
        Bean(re.compile("com.cloudera.navigator:type=NavigatorCountersMXBean"), 'navigator_counters')
        ],
    NAVIGATORMETASERVER=[
        _JAVA_MEMORY_BEAN,
        _JAVA_GC_BEANS,
        Bean(re.compile("com.cloudera.navigator:type=SolrStatsMXBean"),
             'solr_document_count')
        ],
    TELEMETRYPUBLISHER=[
        _JAVA_MEMORY_BEAN,
        _JAVA_GC_BEANS,
        Bean(re.compile("com.cloudera.cdx.extractor.model:type=TelemetryPublisherCountersMXBean,stream=cm-snapshot"),
             'cm_snapshot_tp_counters'),
        Bean(re.compile("com.cloudera.cdx.extractor.model:type=TelemetryPublisherCountersMXBean,stream=YARN-jobs"),
             'yarn_jobs_tp_counters'),
        Bean(re.compile("com.cloudera.cdx.extractor.model:type=TelemetryPublisherCountersMXBean,stream=YARN-jhist"),
             'yarn_jhist_tp_counters'),
        Bean(re.compile("com.cloudera.cdx.extractor.model:type=TelemetryPublisherCountersMXBean,stream=YARN-logs"),
             'yarn_logs_tp_counters'),
        Bean(re.compile("com.cloudera.cdx.extractor.model:type=TelemetryPublisherCountersMXBean,stream=HIVE-query-lineage"),
             'hive_query_lineage_tp_counters'),
        Bean(re.compile("com.cloudera.cdx.extractor.model:type=TelemetryPublisherCountersMXBean,stream=HIVE-query-audits"),
             'hive_query_audits_tp_counters'),
        Bean(re.compile("com.cloudera.cdx.extractor.model:type=TelemetryPublisherCountersMXBean,stream=SPARK_ON_YARN-lineage"),
             'spark_on_yarn_lineage_tp_counters'),
        Bean(re.compile("com.cloudera.cdx.extractor.model:type=TelemetryPublisherCountersMXBean,stream=SPARK_ON_YARN-logs"),
             'spark_on_yarn_logs_tp_counters'),
        Bean(re.compile("com.cloudera.cdx.extractor.model:type=TelemetryPublisherCountersMXBean,stream=SPARK2_ON_YARN-logs"),
             'spark2_on_yarn_logs_tp_counters'),
        Bean(re.compile("com.cloudera.cdx.extractor.model:type=TelemetryPublisherCountersMXBean,stream=SPARK_ON_YARN-event-log"),
             'spark_on_yarn_event_log_tp_counters'),
        Bean(re.compile("com.cloudera.cdx.extractor.model:type=TelemetryPublisherCountersMXBean,stream=SPARK2_ON_YARN-event-log"),
             'spark2_on_yarn_event_log_tp_counters'),
        Bean(re.compile("com.cloudera.cdx.extractor.model:type=TelemetryPublisherCountersMXBean,stream=IMPALA-query-lineage"),
             'impala_query_lineage_tp_counters'),
        Bean(re.compile("com.cloudera.cdx.extractor.model:type=TelemetryPublisherCountersMXBean,stream=IMPALA-query-profile"),
             'impala_query_profile_tp_counters'),
        Bean(re.compile("com.cloudera.cdx.extractor.model:type=TelemetryPublisherCountersMXBean,stream=OOZIE-workflows"),
             'oozie_workflows_tp_counters'),
        Bean(re.compile("com.cloudera.cdx.extractor.model:type=TelemetryPublisherCountersMXBean,stream=HMS-metastore"),
             'hms_metastore_tp_counters')
        ],
    REGIONSERVER=[
        _JAVA_MEMORY_BEAN,
        _make_jvm_metrics_bean('HBase'),
        _make_ugi_metrics_bean('HBase'),
        _make_metrics_control_metrics_bean('HBase'),
        _make_metrics_stats_metrics_bean('HBase'),
        Bean(re.compile("Hadoop:service=HBase,name=RegionServer,sub=Server"),
             'server'),
        Bean(re.compile("Hadoop:service=HBase,name=RegionServer,sub=Regions"),
             'regions'),
        Bean(re.compile("Hadoop:service=HBase,name=RegionServer,sub=Replication"),
             'replication'),
        Bean(re.compile("Hadoop:service=HBase,name=RegionServer,sub=WAL"),
             'wal'),
        Bean(re.compile("Hadoop:service=HBase,name=\S+,sub=IPC"),
             'ipc')
        ],
    MASTER=[
        _JAVA_MEMORY_BEAN,
        _make_jvm_metrics_bean('HBase'),
        _make_ugi_metrics_bean('HBase'),
        _make_metrics_control_metrics_bean('HBase'),
        _make_metrics_stats_metrics_bean('HBase'),
        Bean(re.compile("Hadoop:service=HBase,name=Master,sub=Server"),
             'server'),
        Bean(re.compile("Hadoop:service=HBase,name=Master,sub=FileSystem"),
             'file_system'),
        Bean(re.compile("Hadoop:service=HBase,name=Master,sub=AssignmentManger"),
             'assignment_manager'),
        Bean(re.compile("Hadoop:service=HBase,name=Master,sub=Balancer"),
             'balancer'),
        Bean(re.compile("Hadoop:service=HBase,name=\S+,sub=IPC"),
             'ipc')
        ],
    SOLR_SERVER=[
        _JAVA_MEMORY_BEAN,
        _JAVA_DIRECT_MEMORY_BEAN,
        _JAVA_GC_BEANS,
        Bean(re.compile("CORE"), 'core', singleton=False, discriminator_attr="category"),
        Bean(re.compile("UPDATEHANDLER"), 'updatehandler', singleton=False, discriminator_attr="category"),
        Bean(re.compile("QUERYHANDLER"), 'queryhandler', singleton=False, discriminator_attr="category"),
        Bean(re.compile("CACHE"), 'cache', singleton=False, discriminator_attr="category"),
        ]
    )

