# Copyright (c) 2011 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging
import time

from cmf.monitor import constants
from cmf.monitor import schema
from cmf.monitor.abstract_monitor import CollectionError
from cmf.monitor.abstract_mapreduce_monitor import AbstractMapReduceMonitor
from cmf.monitor.base import DebugStatusReport
from cmf.monitor.jobtracker.json_parser import JsonParser
from cmf.monitor.json_metrics_utils import add_derived_jvm_metrics_metrics, \
    add_metrics_metrics, get_metrics_metric
from cmf.monitor.cgroup_metrics_utils import add_cgroup_metrics
from cmf.monitor.proc_metrics_utils import add_proc_metrics
from cmf.monitor.role_update import RoleUpdate
from cmf.monitor.service_update import ServiceUpdate
from cmf.throttling_logger import ThrottlingLogger

LOG = logging.getLogger('JobTrackerMonitor')
THROTTLED_LOG = ThrottlingLogger(LOG, constants.DEFAULT_MONITOR_LOG_RATE)

_ROLETYPE = 'JOBTRACKER'
_SERVICETYPE = 'MAPREDUCE'

# Section and key names in cloudera-monitor.properties
# Since CM 3.7
_SECTION = 'jobtracker'
_SERVICE_NAME = 'service_name'
_ROLE_NAME = 'role_name'
_HOST = 'host'
_MAPRED_JOB_TRACKER_HTTP_PORT = 'mapred_job_tracker_http_port'
_COLLECT_INTERVAL = 'collect_interval'
# Added in CM 4.0
_SERVICE_VERSION = 'service_version'
# Added in CM 4.1
_LOG_DIR = 'log_dir'
# Added in CM 5.1
_USE_SSL = 'use_ssl'
# Added in CM 5.4
_HEAP_DUMP_DIR = 'oom_heap_dump_dir'

def get_jobtracker_metric(metrics, metric, default=None):
  return get_metrics_metric(metrics, 'mapred', 'jobtracker', metric, default)

class JobTrackerMonitor(AbstractMapReduceMonitor):

  def __init__(self, daemon, registry, cgroup_root=None, cgroup_name=None):
    AbstractMapReduceMonitor.__init__(self,
                                      daemon,
                                      registry,
                                      cgroup_root,
                                      cgroup_name,
                                      _SECTION,
                                      _MAPRED_JOB_TRACKER_HTTP_PORT,
                                      _ROLETYPE)
    self._last_conf_path = None
    self._metrics_tuple = (None, None, None)
    self._last_role_reported_timestamp = None
    self._last_service_reported_timestamp = None
    self._last_collection_ended = None
    self._last_collection_length = None
    self._role_dir_monitors = [ ]
    self._metric_accessors = daemon.metric_accessors

    # Config data
    self._name = None
    self._service_name = None
    self._service_version = None
    self._log_dir = None

  def __str__(self):
    return "JobTrackerMonitor for %s" % (self._name,)

  def _reload_config(self, confparse):
    """Returns True/False for success"""
    try:
      # Extraction
      name = confparse.get(_SECTION, _ROLE_NAME)
      service = confparse.get(_SECTION, _SERVICE_NAME)
      service_version = \
          confparse.getint_with_default(_SECTION, _SERVICE_VERSION,
              constants.SERVICE_VERSION_CDH5)
      host = confparse.get(_SECTION, _HOST)
      log_dir = confparse.get_with_default(_SECTION, _LOG_DIR, None)
      heap_dump_dir = confparse.get_with_default(_SECTION, _HEAP_DUMP_DIR, None)
      port = confparse.getint(_SECTION, _MAPRED_JOB_TRACKER_HTTP_PORT)
      use_ssl = confparse.getboolean_with_default(_SECTION, _USE_SSL, False)
    except ValueError, ex:
      self.record_error(ex)
      LOG.error("Value error in %s: %s" % (confparse.path, ex))
      return False

    metrics_url = self._compute_metrics_url(host, port, use_ssl)
    if metrics_url is None:
      return False

    # Assign
    self._name = name
    self._service_name = service
    self._service_version = service_version
    self._log_dir = log_dir
    self._metrics_url = metrics_url

    # Make monitors for the log directory and the data directories.
    self._role_dir_monitors = [ ]
    if heap_dump_dir is not None:
      self.create_dir_monitor(heap_dump_dir, self._role_dir_monitors)
    if self._log_dir is not None:
      self.create_dir_monitor(self._log_dir, self._role_dir_monitors)

    return True

  def collect(self):
    """Overrides AbstractMonitor. Collect JobTracker metrics. Invoked
       asynchronously."""
    start = time.time()

    # Reload config if necessary
    confparse = self.conf_parser
    if self._last_conf_path != confparse.path:
      if not self._reload_config(confparse):
        return
      self._last_conf_path = confparse.path

    # Collect directory info
    for mon in self._role_dir_monitors:
      mon.collect()

    # If our pid value is None, it means that the associated process is not
    # running. If this is the case, we skip metric collection.
    if self._pid is None:
      self._metrics_tuple = (None, None, None)
      return

    LOG.debug("Fetching JobTracker metrics at '%s'" % (self._metrics_url,))
    self._metrics_tuple = self.collect_metrics_from_url(self._metrics_url)

    self._last_collection_ended = time.time()
    self._last_collection_length = self._last_collection_ended - start

  def parse_metrics(self, json):
    return JsonParser(json, self._service_version)

  def get_role_update(self):
    """Returns a RoleUpdate to be sent to the firehose"""
    update = RoleUpdate(name=self._name,
                        roletype=_ROLETYPE,
                        servicetype=_SERVICETYPE,
                        service_name=self._service_name)

    if self.add_metrics_from_webserver(self._metrics_tuple,
                                       self._last_role_reported_timestamp,
                                       update):
      self._last_role_reported_timestamp = update.timestamp

    add_proc_metrics(update, self._pid, LOG)
    add_cgroup_metrics(update, self._cgroup_root, self._cgroup_name)

    for mon in self._role_dir_monitors:
      update.add_dir_update(mon.get_role_dir_update())

    self.add_unexpected_exit_metrics(update)

    return update

  def get_service_update(self):
    (metrics, timestamp, duration) = self._metrics_tuple
    if metrics is None or isinstance(metrics, CollectionError):
      return None

    # We don't want to double report the same set of metrics.
    if timestamp == self._last_service_reported_timestamp:
      return None
    self._last_service_reported_timestamp = timestamp

    mr_update = ServiceUpdate(name=self._service_name,
                              service_type=_SERVICETYPE,
                              timestamp=timestamp)
    self._add_service_metrics_update(mr_update, metrics)
    return mr_update

  def _add_service_metrics_update(self, update, metrics):
    update.add_metric(schema.MAP_SLOTS,
                      get_jobtracker_metric(metrics, 'map_slots'))
    update.add_metric(schema.MAPS_RUNNING,
                      get_jobtracker_metric(metrics, 'occupied_map_slots'))
    update.add_metric(schema.REDUCE_SLOTS,
                      get_jobtracker_metric(metrics, 'reduce_slots'))
    update.add_metric(schema.REDUCES_RUNNING,
                      get_jobtracker_metric(metrics, 'occupied_reduce_slots'))
    update.add_metric(schema.TRACKERS_BLACKLISTED,
                      get_jobtracker_metric(metrics, 'trackers_blacklisted'))
    update.add_metric(schema.JOBS_COMPLETED,
                      get_jobtracker_metric(metrics, 'jobs_completed'))
    update.add_metric(schema.JOBS_FAILED,
                      get_jobtracker_metric(metrics, 'jobs_failed'))
    update.add_metric(schema.JOBS_KILLED,
                      get_jobtracker_metric(metrics, 'jobs_killed'))
    update.add_metric(schema.JOBS_PREPARING,
                      get_jobtracker_metric(metrics, 'jobs_preparing'))
    update.add_metric(schema.JOBS_RUNNING,
                      get_jobtracker_metric(metrics, 'jobs_running'))
    update.add_metric(schema.JOBS_SUBMITTED,
                      get_jobtracker_metric(metrics, 'jobs_submitted'))
    update.add_metric(schema.MAPS_FAILED,
                      get_jobtracker_metric(metrics, 'maps_failed'))
    update.add_metric(schema.REDUCES_FAILED,
                      get_jobtracker_metric(metrics, 'reduces_failed'))
    # Work-around for MAPREDUCE-1238
    update.add_metric(schema.WAITING_MAPS,
                      max(0, get_jobtracker_metric(metrics, 'waiting_maps', 0)))
    update.add_metric(schema.WAITING_REDUCES,
                      max(0, get_jobtracker_metric(metrics, 'waiting_reduces', 0)))

  def add_webserver_metrics_to_role_update(self, update, metrics):
    if self._service_version == constants.SERVICE_VERSION_CDH5:
      self._add_webserver_metrics_to_role_update_cdh5(update, metrics)
    else:
      THROTTLED_LOG.warn("Unknown jobtracker version: %s" % (self._service_version))

  def _add_webserver_metrics_to_role_update_cdh5(self, update, metrics):
    add_derived_jvm_metrics_metrics(update, metrics)
    add_metrics_metrics(update, metrics, self._metric_accessors, _ROLETYPE, 'cdh5')

  def get_debug_status_report(self):
    return DebugStatusReport(config_items=self.conf_parser.items(_SECTION),
                             errors=(self.last_error,),
                             last_collection_ended=self._last_collection_ended,
                             last_collection_length=self._last_collection_length)

  # For testing.
  def clear_last_reported_timestamp(self):
    super(JobTrackerMonitor, self).clear_last_reported_timestamp()
    self._last_role_reported_timestamp = None
    self._last_service_reported_timestamp = None
