# Copyright (c) 2011-2012 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging

from cmf.monitor import schema
from cmf.monitor import constants
from cmf.throttling_logger import ThrottlingLogger

LOG = logging.getLogger(__name__)
THROTTLED_LOG = ThrottlingLogger(LOG, constants.DEFAULT_MONITOR_LOG_RATE)

_BYTES_PER_KB = 1024
_BYTES_PER_MB = 1048576

# This class encapsulates the information needed to look up a metric from a
# Hadoop metrics style JSON object.
class MetricAccessor:
  def __init__(self, category, section, metric):
    self.category = category
    self.section = section
    self.metric = metric

# This class encapsulates the information needed to look up a metric from a
# JSON metrics object produced by the JMXJsonServlet.
class JmxMetricAccessor:
  def __init__(self, bean_alias, metric, bean_id=None, bean_type=None):
    self.bean_alias = bean_alias
    self.metric = metric
    self.bean_id = bean_id
    self.bean_type = bean_type

# This class encapsulates the information needed to look up a metric from a
# an Impala provided JSON metrics dictionary.
class ImpalaMetricAccessor:
  def __init__(self, metric, stat=None):
    self.metric = metric
    self.stat = stat

# This function builds a dictionary of metric accessors (see above) from the
# metric_source fields in the metrics schema that is structured as followed:
#
# [entity type string][version string][metric id] -> metric accessor
#
# For example, a single entry would look like the following:
#
# ['NAMENODE']['cdh4']['19178'] =
#   MetricAccessor('rpc', 'detailed-metrics' 'blockReceived_avg_time')
def build_metric_accessors(schema):
  accessors = dict()
  for metric in schema.itervalues():
    if metric.get('deprecated', 'false').lower() == 'true':
      continue
    for entity, sources in metric.get('metric_source', {}).iteritems():
      if entity not in accessors:
        accessors[entity] = dict()
      for version, source in sources.iteritems():
        if version not in accessors[entity]:
          accessors[entity][version] = dict()

        # The code below is parsing the dictionary values in the 'metric_source'
        # portion of the metric schema. Examples of those values include:
        #
        # ...
        # "NAMENODE": {
        #  "cdh4": "JMX$$fsnamesystem::Hadoop:service=,name=FSNamesystem::BlocksTotal"
        #  "cdh5": "JMX$$fsnamesystem::Hadoop:service=,name=FSNamesystem::BlocksTotal"
        # }
        # ...
        #
        # Here the METRICS$$ and JMX$$ indicate that the metric is either in the
        # Hadoop metrics or JMXJsonServlet format respectively. These values are
        # checked to determine what type of metric accessor to build and how the
        # remainder of the string should be parsed.
        #
        # The remainder of the METRICS$$ format looks like:
        #
        # rpc.metrics.getProtocolVersion.aboveOneSec._avg_time
        #
        # to parse this, we split on two '.'s. The first part is the category,
        # the second is the section and the remainder is the metric. Note that
        # there are metrics with '.'s in their names.
        #
        # The remainder of the JMX$$ format looks like:
        #
        # fsnamesystem::Hadoop:service=,name=FSNamesystem::BlocksTotal
        #
        # to parse this, we split on two '::'s. The first part is the bean alias,
        # the second is unused and the remainder is the metric. (The second part
        # is actually used, just not here -- it's used as a key into the alias
        # and ignored metrics maps in platform-metrics.py).
        #
        # Solr beans are processed based on their categories. Multiple bean
        # types can potentially be associated with a category. To help associate
        # a metric with the correct bean type, the bean type is included in the
        # metric accessor. The remainder (following JMX$$) of the Solr format
        # looks like:
        #
        # queryhandler::QUERYHANDLER::99thPcRequestTime::org.apache.solr.handler.ReplicationHandler
        #
        # to parse this, we split on three (or four) '::'s. The first part is
        # the bean alias, the second is unused, the 3rd is the metric, the 4th
        # is the 'bean_id' and optional 5th field is the 'bean_type'.
        #
        # The remainder of the IMPALA$$ format is just the key into the Impala
        # JSON metrics dictionary. An example string is:
        #
        # IMPALA$$simple-scheduler.assignments.total
        #
        # Impala also has statistical metrics denoted by the IMPALA$$STATS$$
        # format. For example:
        #
        # IMPALA$$STATS$$statestore-subscriber.heartbeat-interval-time::count
        #
        # This should be accessed by looking up the 'count' key in the metric
        # dictionary under 'statestore-subscriber.heartbeat-interval-time'.
        
        if source.startswith('JMX$$'):
          # We are dealing with a JMX accessor...
          splits = source.split('JMX$$', 1)[1].split('::', 2)
          accessors[entity][version][metric['id']] = \
              JmxMetricAccessor(splits[0], splits[2])
        elif source.startswith('METRICS$$'):
          # We are dealing with a metrics page accessor...
          splits = source.split('METRICS$$', 1)[1].split('.', 2)
          accessors[entity][version][metric['id']] = \
              MetricAccessor(splits[0], splits[1], splits[2])
        elif source.startswith('IMPALA$$STATS$$'):
          # We are dealing with an Impala stat accessor...
          splits = source.split('IMPALA$$STATS$$', 1)[1].split('::', 1)
          accessors[entity][version][metric['id']] = \
              ImpalaMetricAccessor(splits[0], splits[1])
        elif source.startswith('IMPALA$$'):
          # We are dealing with an Impala accessor...
          accessors[entity][version][metric['id']] = \
              ImpalaMetricAccessor(source.split('$$', 1)[1])
        elif source.startswith('SOLR$$'):
          # We are dealing with Solr variation of the JMX accessor
          # All Solr metric accessors have 'bean_id's and some have the
          # optional 'bean_type'. The resulting JmxMetricAccessor
          # reflects this distinction accordingly.
          # For e.g: the Solr accessor for
          # 1. "queryhandler::QUERYHANDLER::requests::org.apache.solr.handler.ReplicationHandler"
          # only has a 'bean_id' of 'org.apache.solr.handler.ReplicationHandler' and no 'bean_type'.
          # 2. "queryhandler::QUERYHANDLER::requests::org.apache.solr.handler.component.SearchHandler::/browse"
          # has a 'bean_id' of 'org.apache.solr.handler.component.SearchHandler' and a
          # 'bean_type' of '/browse'
          splits = source.split('SOLR$$', 1)[1].split('::', 4)
          if len(splits) == 4:
            accessors[entity][version][metric['id']] = \
              JmxMetricAccessor(splits[0], splits[2], splits[3])
          elif len(splits) == 5:
            accessors[entity][version][metric['id']] = \
              JmxMetricAccessor(splits[0], splits[2], splits[3], splits[4])
        elif source.startswith('SOLR7$$'):
          # We are dealing with Solr variation of the mertrics accessor
          # All Solr 7 metric accessors have category or aggregation
          # optional a section.
          # For e.g: the Solr 7 accessor for
          # 1. "SOLR7$$::CORE::QUERY./browse.requests"
          # This has only category and metric, but no section, as the metric is not a structure
          # 2. "SOLR7$$::CORE::QUERY./browse.errors::count"
          # Second part is second and beind defined.
          # 3. "SOLR7$$::solr.jetty::org.eclipse.jetty.server.handler.DefaultHandler.requests::15minRate"
          # Aggregation is not a variable but a literal and will be processed automatically.
          # Variable aggregation will be processed by ReplicaMetricsHandler
          splits = source.split('SOLR7$$', 1)[1].split('::', )
          if len(splits) == 3:
            accessors[entity][version][metric['id']] = \
              MetricAccessor(splits[0], splits[1], splits[2])
          elif len(splits) == 2:
            accessors[entity][version][metric['id']] = \
              MetricAccessor(splits[0], None, splits[1])

  return accessors

# This function uses the input accessors map and adds all available metrics from
# the input metrics object to the input update. The arguments are as follows:
#
# update: a RoleUpdate
# metrics: a JsonParser object built from a Hadoop metrics sytle JSON object
# accessors: a map of metric accessors built by build_metric_accessors
# entity: an entity type string from a metric source, e.g. 'NAMENODE'
# version: a version string from a metric source, e.g. 'cdh4'
def add_metrics_metrics(update, metrics, accessors, entity, version):
  metric_accesors = accessors.get(entity, {}).get(version, {})
  for metric_id, accessor in metric_accesors.iteritems():
    default = None
    handle_duplicates = False
    # "Detailed RPC" metrics for core Hadoop (MR, HDFS) are lazily created. That
    # means that they will not be present in the metrics JSON until the first
    # time that RPC is called. We prefer inserting a 0 to this 'not present'
    # behavior.
    if accessor.category == 'rpc' and accessor.section == 'detailed-metrics':
      default = 0;
    # The task tracker does not reset its RPC metrics when it gets
    # reinitialized (see OPSAPS-3319), so we have multiple entries for
    # each metric. See the comments in get_metrics_metric for details on how
    # we deal with this.
    if entity == 'TASKTRACKER' and accessor.category == 'rpc':
      handle_duplicates = True
    update.add_metric(metric_id,
                      get_metrics_metric(metrics,
                                         accessor.category,
                                         accessor.section,
                                         accessor.metric,
                                         default,
                                         handle_duplicates))

# This function uses the input accessors map and adds all available metrics from
# the input metrics object to the input update. The arguments are as follows:
#
# update: a RoleUpdate
# metrics: a JsonParser object built from a JSON object produced by the
#          JMXJsonServlet
# accessors: a map of metric accessors built by build_metric_accessors
# entity: an entity type string from a metric source, e.g. 'NAMENODE'
# version: a version string from a metric source, e.g. 'cdh4'
def add_metrics_jmx(update, metrics, accessors, entity, version):
  metric_accesors = accessors.get(entity, {}).get(version, {})
  for metric_id, accessor in metric_accesors.iteritems():
    default = None
    # "Detailed RPC" metrics for core Hadoop (MR, HDFS) are lazily created. That
    # means that they will not be present in the metrics JSON until the first
    # time that RPC is called. We prefer inserting a 0 to this 'not present'
    # behavior.
    if (accessor.bean_alias == 'rpc' or
        accessor.bean_alias == 'detailed_rpc' or
        accessor.bean_alias == 'service_rpc' or
        accessor.bean_alias == 'service_detailed_rpc'):
      default = 0

    # Similar to the "Detailed RPC" metrics above, some of the Solr replication
    # related metrics will not be available until after the first time the
    # the replication occurs. Insert 0 for these metrics.
    if (accessor.bean_alias == 'replication'):
      default = 0

    metric_value = get_jmx_metric(metrics,
                                  accessor.bean_alias,
                                  accessor.metric,
                                  default)
    update.add_metric(metric_id, metric_value)

# Utility function to get a metric value from a JsonParser built from a Hadoop
# metrics style JSON object. The JSON structure looks something like:
#
# {
# ...
#   "jvm":{
#      "metrics":[
#         [
#            {
#               "hostName":"centos62-25.ent.cloudera.com",
#               ...
#            },
#            {
#               "gcCount":41066,
#               ...
#            }
#         ]
#      ]
#   },
#   ...
# }
# 
# We refer to the 'jvm' key as the category, the 'metrics' key as section, and
# 'gcCount' as the metric.
#
# Note that this format can be a little tricky. For example, in OPSAPS-3319 we
# have entries for the same metrc in different sections because the tasktracker
# doesn't clear out the old metrics when it gets re-inited. Callers can indicate
# to handle this with the 'handle_duplicates' parameter. This flag works as
# follows:
# 1. For most metrics sum together all the values. Since most metrics are
# counters this just works
# 2. For metrics ending in _avg_time get the first non-zero value. This works
# because those are time varying rate metrics which get reset every minute so
# they will only be non-zero for active metrics (if two happen to be active
# in the same minute then we'll combine them, but something weird was happening
# then anyway, so let's no worry about it)
def get_metrics_metric(metrics, category, section, metric, default,
    handle_duplicates=False):
  if metrics is None or metrics.json is None:
    return default
  category_dict = metrics.json.get(category, None)
  if category_dict is None:
    return default
  section_list = category_dict.get(section, None)
  if section_list is None:
    return default
  total = 0
  for s in section_list:
    for metric_dictionary in s:
      if metric in metric_dictionary:
        metric_value = metric_dictionary.get(metric)
        if handle_duplicates:
          total = total + metric_value
        else: 
          return metric_value
  if handle_duplicates:
    return total
  return default

# Utility function to get a metric value from a JsonParserJmx built from a JSON
# object produed by the JMXJsonServlet.  The JSON structure looks something
# like:
#
# {
# ...
#  }, {
#    "name" : "Hadoop:service=DataNode,name=JvmMetrics",
#    ...
#    "GcCount" : 82938,
#    ...
#  }, {
# ...
# }
# 
# See JsonParserJmx for how the above is parsed to understand the access code
# in this function.
def get_jmx_metric(metrics, bean, metric, default=None):
  if metrics is not None and _is_not_none(metrics, bean):
    return getattr(metrics, bean).get(metric, default)
  else:
    return default

# This is a utility method to add some metrics derived from raw JVM metrics that
# are in the Hadoop metrics style.
def add_derived_jvm_metrics_metrics(update, metrics):
  update.add_metric(schema.JVM_TOTAL_THREADS,
                    get_metrics_metric(metrics,
                                       'jvm',
                                       'metrics',
                                       'threadsBlocked',
                                       0) +
                    get_metrics_metric(metrics,
                                       'jvm',
                                       'metrics',
                                       'threadsNew',
                                       0) +
                    get_metrics_metric(metrics,
                                       'jvm',
                                       'metrics',
                                       'threadsRunnable',
                                       0) +
                    get_metrics_metric(metrics,
                                       'jvm',
                                       'metrics',
                                       'threadsTerminated',
                                       0) +
                    get_metrics_metric(metrics,
                                       'jvm',
                                       'metrics',
                                       'threadsTimedWaiting',
                                       0) +
                    get_metrics_metric(metrics,
                                       'jvm',
                                       'metrics',
                                       'threadsWaiting',
                                       0))

# This is a utility method to add some metrics derived from raw JVM metrics that
# are in the JMX metrics style.
def add_derived_jvm_metrics_jmx(update, metrics, \
                                collect_all_memory_metrics=False, \
                                include_jvm_metrics=True):
  # The JMX version of the JvmMetrics does not cotain the max memory of the JVM (CDH4 only),
  # so we calculate it from a different bean.
  if _is_not_none(metrics, 'java_memory'):
    update.add_metric(schema.JVM_MAX_MEMORY_MB,
                      metrics.java_memory.HeapMemoryUsage.max / float(_BYTES_PER_MB))
    # We collect the rest of the memory metrics only if we are explicitly told to do so.
    if collect_all_memory_metrics:
      update.add_metric(schema.JVM_HEAP_COMMITTED_MB,
                        metrics.java_memory.HeapMemoryUsage.committed / float(_BYTES_PER_MB))
      update.add_metric(schema.JVM_HEAP_USED_MB,
                        metrics.java_memory.HeapMemoryUsage.used / float(_BYTES_PER_MB))
      update.add_metric(schema.JVM_NON_HEAP_COMMITTED_MB,
                        metrics.java_memory.NonHeapMemoryUsage.committed / float(_BYTES_PER_MB))
      update.add_metric(schema.JVM_NON_HEAP_USED_MB,
                        metrics.java_memory.NonHeapMemoryUsage.used / float(_BYTES_PER_MB))
  # Add Direct Memory metrics if present
  if _is_not_none(metrics, 'java_direct_memory'):
    update.add_metric(schema.JVM_DIRECT_MEMORY_BUFFERS,
                      metrics.java_direct_memory.Count)
    update.add_metric(schema.JVM_DIRECT_MEMORY_USED,
                      metrics.java_direct_memory.MemoryUsed)
    update.add_metric(schema.JVM_DIRECT_MEMORY_CAPACITY,
                      metrics.java_direct_memory.TotalCapacity)

  # Some roles do not have a 'jvm' bean (e.g. Solr server). In that case, we
  # skip processing it.
  if include_jvm_metrics and _is_not_none(metrics, 'jvm'):
    update.add_metric(schema.JVM_TOTAL_THREADS,
                      metrics.jvm.get('ThreadsBlocked', 0) +
                      metrics.jvm.get('ThreadsNew', 0) +
                      metrics.jvm.get('ThreadsRunnable', 0) +
                      metrics.jvm.get('ThreadsTerminated', 0) +
                      metrics.jvm.get('ThreadsTimedWaiting', 0) +
                      metrics.jvm.get('ThreadsWaiting', 0))

def prefer_first_attribute(ordered_pairs):
  """ The default behavior when loading JSON is for the last instance
      of a repeated attribute to take precedence. This utility is used
      as hook when loading JSON to eliminate duplicate keys by preferring
      the first instance of the key.
      See CDH-15751 against the Solr server that dumps JSON exhibiting
      duplicate 'name' entries.
  """
  d = {}
  for k, v in ordered_pairs:
    if k in d:
      continue
    d[k] = v
  return d

# This is a utility method to add GC metrics derived from raw JVM metrics using
# the same scheme used by Hadoop style JvmMetrics. See JvmMetrics.java in
# hadoop-common-project/hadoop-common/src/main/java/org/apache/hadoop/metrics2/source or
# hadoop-common-project/hadoop-common/src/main/java/org/apache/hadoop/metrics/jvm.
def add_derived_gc_metrics_jmx(update, metrics):
  if _is_not_none(metrics, 'java_gc'):
    beans = metrics.java_gc
    gcCollectionCount = 0
    gcCollectionTimeMs = 0
    for bean in beans:
      for key in bean.keys():
        if 'CollectionCount' == key:
          gcCollectionCount += bean[key]
        if 'CollectionTime' == key:
          gcCollectionTimeMs += bean[key]
    update.add_metric(schema.JVM_GC_COUNT, gcCollectionCount)
    update.add_metric(schema.JVM_GC_TIME_MS, gcCollectionTimeMs)

def add_pause_metrics(update, metrics):
  if _is_not_none(metrics, 'pause_detail'):
    for bean in metrics.pause_detail:
      if bean[constants.PAUSE_BEAN_ATTR_NAME] == constants.PAUSES_ALIAS:
        update.add_metric(schema.PAUSES, bean["Count"])
      elif bean[constants.PAUSE_BEAN_ATTR_NAME] == constants.PAUSE_TIME_ALIAS:
        update.add_metric(schema.PAUSE_TIME, bean["Count"])
  elif (_is_not_none(metrics, 'jvm') and
        _is_not_none(metrics.jvm, 'GcNumInfoThresholdExceeded') and
        _is_not_none(metrics.jvm, 'GcNumWarnThresholdExceeded')):
    update.add_metric(schema.PAUSES,
                      metrics.jvm.GcNumInfoThresholdExceeded + \
                      metrics.jvm.GcNumWarnThresholdExceeded)

class ParseError(Exception):
  pass

class DictWithAttr(dict):
  """
  A wrapper class to allow attribute access for dictionary-like objects (e.g.
  json objects). Note that this returns None for non-existant attributes rather
  thank throwing a KeyError. Also note that this means hasattr does not work on
  these objects.
  """
  def __init__(self, dictionary):
    try:
      dict.__init__(self, dictionary)
    except:
      raise ParseError("Cannot initialize dictionary with %r" % (dictionary,))

  def __getattr__(self, attr):
    try:
      res = self[attr]
      if isinstance(res, dict):
        # Recursively wrap the result if it's also a dictionary
        return DictWithAttr(res)
      return res
    except KeyError:
      return None

# See OPSAPS-23363 for why we have this helper. Essentially hasattr does not
# work when called on a DictWithAttr object (presumably, though not confirmed
# because of the behavior of it eating KeyErrors). Instead of using hasattr,
# folks should use this function. It calls hasattr (which may or may not work)
# but also goes ahead and calls getattr and checks for null. The reason we do
# both is that sometimes we are called on DictWithAttr and sometimes on other
# objects. Good Times!
def _is_not_none(obj, attr):
  return hasattr(obj, attr) and getattr(obj, attr) is not None
  
