# Copyright (c) 2011 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import ConfigParser
import logging
import os
import time

LOG = logging.getLogger(__name__)

class ConfParser(object):
  """
  Config parser that converts NoOptionError into None.
  """
  def __init__(self, conf_dir, filename):
    """
    ConfParser(conf_dir) -> ConfParser object

    May raise IOError.
    """
    self._conf_file = os.path.join(conf_dir, filename)
    self._cp = ConfigParser.SafeConfigParser()
    if not self._cp.read(self._conf_file):
      raise IOError("Failed to read conf file '%s'" % (self._conf_file,))
    self._timestamp = time.time()

    # Wrapped methods
    self.get = self._error_wrapper(self._cp.get)
    self.getint = self._error_wrapper(self._cp.getint)
    self.getfloat = self._error_wrapper(self._cp.getfloat)
    self.getboolean = self._error_wrapper(self._cp.getboolean)
    self.items = self._error_wrapper(self._cp.items)

  def get_default_section(self):
    """
    get_default_section() -> section name
    Returns the first section name. Or None if empty.
    """
    sections = self._cp.sections()
    return sections and sections[0] or None

  @property
  def path(self):
    return self._conf_file

  @property
  def timestamp(self):
    return self._timestamp

  def _error_wrapper(self, fn):
    """Convert NoOptionError to None"""
    def res(*args, **kwargs):
      try:
        return fn(*args, **kwargs)
      except ConfigParser.NoOptionError:
        LOG.warn("%s has no option '%r'" % (self._conf_file, args))
        return None
    return res

  def getlist(self, section, option, sep=','):
    """Parse an option value as a list. Returns None if not found."""
    res = self.get(section, option)
    if res is None:
      return None
    return res.split(sep)

  def getlist_with_default(self, section, option, default, sep=','):
    """Parse an option value as a list. Returns None if not found."""
    res = self.get_with_default(section, option, None)
    if res is None:
      return default
    return res.split(sep)

  def get_with_default(self, section, option, default):
    try:
      return self._cp.get(section, option)
    except ConfigParser.NoOptionError:
      return default
    except AttributeError:
      return default

  def getint_with_default(self, section, option, default):
    try:
      return self._cp.getint(section, option)
    except ConfigParser.NoOptionError:
      return default

  def getfloat_with_default(self, section, option, default):
    try:
      return self._cp.getfloat(section, option)
    except ConfigParser.NoOptionError:
      return default

  def getboolean_with_default(self, section, option, default):
    try:
      return self._cp.getboolean(section, option)
    except ConfigParser.NoOptionError:
      return default
    
