# Copyright (c) 2011 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging
import os
import psutil
import time

import re

from cmf.monitor import constants
from cmf.monitor import schema
from cmf.throttling_logger import ThrottlingLogger

_CLOCK_TICKS = os.sysconf(os.sysconf_names["SC_CLK_TCK"])

_FD_MAX_RE = re.compile('Max open files\s+\d+\s+(\d+)\s+files')
_SWAP_OUT_RE = re.compile('pswpout (\d+)')

_VMSTAT_PARSE_FAILED = False
_FILENR_PARSE_FAILED = False
_EPSILON=0.0001

LOG = logging.getLogger('ProcMetrics')
THROTTLED_LOG = ThrottlingLogger(LOG, constants.DEFAULT_MONITOR_LOG_RATE)

def get_fd_max(pid, logger=None):
  if pid is None:
    return None
  f = None
  try:
    try:
      f = open('/proc/%d/limits' % pid, 'r')
      for line in f:
        m = _FD_MAX_RE.match(line)
        if m:
          return int(m.groups()[0])
      if logger is not None:
        logger.error("Could not parse limits for for process %d" % pid)
      return None
    except (Exception), e:
      if logger is not None:
        logger.error("Failed to read file descriptor max for process %d: %s" %
                     (pid, e))
      return None
  finally:
    if f is not None:
      f.close()

def get_fd_count(pid, logger=None):
  if pid is None:
    return None
  try:
    return len(os.listdir("/proc/%d/fd/" % pid))
  except (Exception), e:
    if logger is not None:
      logger.error("Failed to get file descriptor count for process %d: %s" %
                   (pid, e))
    return None

def get_host_fd_stats(logger=None):
  """Returns the total number of open file descriptors for the host as well as
     the max number of file descriptors allowed for the host. We collect from
     /proc/sys/fs/file-nr (see Documentation/sysctl/fs.txt for an explanation
     of the file's contents better than the manpage). A sample line looks like:
     10528   0       799198"""
  global _FILENR_PARSE_FAILED
  fd_open = fd_max = None
  f = None
  try:
    try:
      f = open('/proc/sys/fs/file-nr', 'r')
      stats = f.read().split()
      fd_open = int(stats[0]) - int(stats[1])
      fd_max = int(stats[2])
    except (Exception):
      if logger is not None and not _FILENR_PARSE_FAILED:
        logger.exception("Failed to get host fd stats")
        _FILENR_PARSE_FAILED = True
      fd_open = fd_max = None
  finally:
    if f is not None:
      f.close()
  return (fd_open, fd_max)

def get_swap_out(logger=None):
  global _VMSTAT_PARSE_FAILED

  f = None
  try:
    try:
      f = open('/proc/vmstat', 'r')
      for line in f:
        m = _SWAP_OUT_RE.match(line)
        if m:
          _VMSTAT_PARSE_FAILED = False
          return int(m.groups()[0])
      if logger is not None and not _VMSTAT_PARSE_FAILED:
        logger.error("Could not parse pswpout from /proc/vmstat")
        _VMSTAT_PARSE_FAILED = True
        return None
    except (Exception):
      if logger is not None:
        logger.exception("Failed to get pages swapped out")
        _VMSTAT_PARSE_FAILED = True
      return None
  finally:
    if f is not None:
      f.close()

def get_child_cpu_times(pid, logger=None):
  """Returns the amount of time this process's waited for children have been
  scheduled in user and kernel mode. Children are waited for by well behaving
  processes after they exit. This means that this metric will miss two kinds of child
  processes: currently executing ones and ones that weren't waited for correctly.
  It appears that the tasktracker does wait for its children. And grandchildren
  that aren't waited for correctly will be lost, but I believe there's no way
  around this."""
  if pid is None:
    return None
  f = None
  try:
    try:
      f = open("/proc/%s/stat" % pid)
      st = f.read().strip()
      # ignore the first two values ("pid (exe)")
      st = st[st.find(')') + 2:]
      values = st.split(' ')
      cutime = float(values[13]) / _CLOCK_TICKS
      cstime = float(values[14]) / _CLOCK_TICKS
      return (cutime, cstime)
    except (Exception):
      if logger is not None:
        logger.exception("Could not get child cpu times for process %d" % pid)
      return None
  finally:
    if f is not None:
      f.close()

def _get_proc_metrics_map(pid, logger=None):
  metrics_map = _get_proc_metrics(pid, True, logger)
  child_cpu_tuple = get_child_cpu_times(pid, logger)
  if child_cpu_tuple is not None:
    # Note that this isn't really the WITH_DESCENDANTS metric, but it's a
    # convenient place to put the child cpu metrics because this map is only
    # used internally.
    metrics_map[schema.TOTAL_CPU_USER_WITH_DESCENDANTS] = child_cpu_tuple[0]
    metrics_map[schema.TOTAL_CPU_SYSTEM_WITH_DESCENDANTS] = child_cpu_tuple[1]
  return metrics_map

def _get_process_and_descendant_metrics(pid, logger=None):
  """Get the mem_rss, mem_virtual, total_cpu_user, and total_cpu_system for
  a process and all its descendants"""
  # Get the metrics for this process before gettings its children so that we
  # handle the potential race condition consistently. See
  #_get_with_descendants_cpu_metrics for more details
  ret = {}
  if pid is None:
    return ret
  ret[pid] = _get_proc_metrics_map(pid, logger)

  process = psutil.Process(pid)
  # This will return child processes. Parents will appear in the list before
  # children. This is important for making sure we don't double count child
  # metrics.
  child_processes = process.children(recursive=True)
  for child_process in child_processes:
    ret[child_process.pid] = _get_proc_metrics_map(child_process.pid, logger)
  return ret

def add_with_descendants_proc_metrics(update,
                                      pid,
                                      all_previous_metrics,
                                      previous_user_cpu,
                                      previous_system_cpu,
                                      logger=None):
   """Add the with descendants proc metrics for memory and CPU. The caller
   should pass in previous values for their metrics if they have them so that
   we can calculate a more accurate estimate of CPU usage (see below for
   more details. This returns the current values for the metrics.

   We don't collect IO metrics because I'm not aware of any
   reliable way to calculate them"""
   metrics = _get_process_and_descendant_metrics(pid)

   # Add memory first. These are easy because they are a gauge which means
   # we don't care about processes that already completed
   total_rss = 0
   total_virtual = 0
   for metrics_map in metrics.itervalues():
     total_rss += metrics_map.get(schema.MEM_RSS, 0)
     total_virtual += metrics_map.get(schema.MEM_VIRTUAL, 0)

   update.add_metric(schema.MEM_RSS_WITH_DESCENDANTS, total_rss)
   update.add_metric(schema.MEM_VIRTUAL_WITH_DESCENDANTS, total_virtual)

   total_user_cpu, total_system_cpu =\
        _get_with_descendants_cpu_metrics(pid,
                                          metrics,
                                          all_previous_metrics,
                                          previous_user_cpu,
                                          previous_system_cpu,
                                          logger)
   update.add_metric(schema.TOTAL_CPU_USER_WITH_DESCENDANTS, total_user_cpu)
   update.add_metric(schema.TOTAL_CPU_SYSTEM_WITH_DESCENDANTS, total_system_cpu)

   return metrics, total_user_cpu, total_system_cpu

def _get_with_descendants_cpu_metrics(pid,
                                      metrics,
                                      all_previous_metrics,
                                      previous_user_cpu,
                                      previous_system_cpu,
                                      logger=None):
   """This is a helper function to aid testing. The metrics map has all the
   descendants metrics and the metrics for the process itself"""
   # For CPU things are a little more complicated.
   # CPU counters work as follows:
   # - Each process has two counters
   #   - One for the CPU taken by the process itself
   #   - Another for the CPU taken by all its children that have completed and
   #     it has waited on.
   # So to calculate all the CPU used we have to find all the active
   # descendants of the process and sum together both CPU counters for
   # all of them
   #
   # Doing this introduces the following race condition:
   # 1. We calculate cutime for this process (only includes finished processes)
   # 2. Another child finishes.
   # 3. We go through the rest of the processes and calculate their user time
   # and add it to the total
   # In this case the child that finished wouldn't be included anywhere. To
   # handle this case (kind of) we calculate a floor for the new value which
   # assumes a few things to get a reasonable floor for the value.
   # 1. All processes that started and stopped within the
   # collection period used neglible CPU.
   # 2. All processes that ended within the time didn't use much new CPU
   # To do this we add the old value to the CPU used by current processes
   # (either since last time or since they started). This is a less than perfct
   # estimate, but I think it protected against the worst cases and the race
   # condition should be rare. Also, in the long run it will tend towards the
   # true value.

   # Total cpu is the sum of this process' CPU (both it and it's waited
   # upon children) and the CPU of all its descendants (again with waited upon
   # children added in). This value can only undercount the actual total user
   # cpu because parent cpu metrics are looked up before their children so if
   # a child shows up in its parent's total it must have already finished and
   # therefore won't exist when we check its CPU.
   total_user_cpu = 0
   # Floor user cpu is the sum of the old CPU reported and the CPU from the
   # processes started since last time. If the new CPU value is less than this
   # it means we've hit a race condition where we undercounted total CPU, so
   # we should just fall back to this.
   if previous_user_cpu is None:
     floor_user_cpu = 0
   else:
     floor_user_cpu = previous_user_cpu
   total_system_cpu = 0
   if previous_system_cpu is None:
     floor_system_cpu = 0
   else:
     floor_system_cpu = previous_system_cpu
   for pid, metrics_map in metrics.iteritems():
     user_cpu = metrics_map.get(schema.TOTAL_CPU_USER, 0)
     user_cpu_descendants = metrics_map.get(schema.TOTAL_CPU_USER_WITH_DESCENDANTS, 0)
     total_user_cpu += user_cpu + user_cpu_descendants
     system_cpu = metrics_map.get(schema.TOTAL_CPU_SYSTEM, 0)
     system_cpu_descendants = metrics_map.get(schema.TOTAL_CPU_SYSTEM_WITH_DESCENDANTS, 0)
     total_system_cpu += system_cpu + system_cpu_descendants
     if pid in all_previous_metrics:
       # This is a pre-existing process. Add in the difference between its
       # previous CPU and the current value. Don't look at the difference
       # in its descendants CPU because otherwise it could get double counted.
       previous_metrics = all_previous_metrics.get(pid)
       floor_user_cpu += user_cpu - previous_metrics.get(schema.TOTAL_CPU_USER, 0)
       floor_system_cpu += system_cpu - previous_metrics.get(schema.TOTAL_CPU_SYSTEM, 0)
     else:
       # This is a new process so its descendants must be new too, so let's
       # add in both counters
       floor_user_cpu += user_cpu + user_cpu_descendants
       floor_system_cpu += system_cpu + system_cpu_descendants

   # Since it appears that the kernel ocassionally reports non-monitonically
   # increasing CPU counter values let's be robust to that. If the numbers
   # have decreased then log a warning here and let the firehose handle
   # it like it does for the other CPU counter metrics. This does mean that
   # if the counter decreases it may take a little while to get back to the old
   # value, so we will return the same (floor) value multiple times. We believe
   # that on a real-world cluster we'll get back to the floor value quickly
   # so in practice this won't be a big deal.
   if floor_user_cpu > total_user_cpu + _EPSILON:
     if logger is not None:
       THROTTLED_LOG.info("Descendants user CPU lower than expected for process %s:\
           %s, %s", pid, floor_user_cpu, total_user_cpu)
     total_user_cpu = floor_user_cpu
   if floor_system_cpu > total_system_cpu + _EPSILON:
     if logger is not None:
       THROTTLED_LOG.info("Descendants system CPU lower than expected for process %s:\
            %s, %s", pid, floor_system_cpu, total_system_cpu)
     total_system_cpu = floor_system_cpu

   return total_user_cpu, total_system_cpu

def _get_proc_metrics(pid, only_cpu_and_mem, logger):
  """Returns a map from metric_id to value for each of the metrics."""
  ret = {}
  # For the rest of the metrics, we require a pid, so we're done if we were not
  # given one.
  if pid is None:
    return ret

  try:
    if not only_cpu_and_mem:
      fd_max = get_fd_max(pid, logger)
      if fd_max is not None:
        ret[schema.FD_MAX] = fd_max

      fd_count = get_fd_count(pid, logger)
      if fd_count is not None:
        ret[schema.FD_OPEN] = fd_count

    process = psutil.Process(pid)
    cpu_times = process.cpu_times()
    if cpu_times is not None:
      ret[schema.TOTAL_CPU_USER] = cpu_times.user
      ret[schema.TOTAL_CPU_SYSTEM] = cpu_times.system

    if not only_cpu_and_mem:
      try:
        io_counters = process.io_counters()
        ret[schema.READ_BYTES] = io_counters.read_bytes
        ret[schema.WRITE_BYTES] = io_counters.write_bytes
      except psutil.NoSuchProcess:
        # We don't support I/O metrics on linux machines with kernels that don't
        # support I/O Stats (that don't have /proc/<pid>/io), so don't collect
        # the metrics in that case.
        pass

      uptime = time.time() - process.create_time()
      if uptime >= 0:
        ret[schema.UPTIME] = uptime
      elif logger is not None:
        logger.warning("Negative  uptime observed. Not reporting uptime to "
                       "Service Monitor.")

    rss, vms = process.memory_info()[0:2]
    ret[schema.MEM_RSS] = rss
    ret[schema.MEM_VIRTUAL] = vms
    swap = get_process_swap_memory(process)
    if swap is not None:
      ret[schema.MEM_SWAP] = swap
    return ret
  except (Exception), e:
      if logger is not None:
        logger.error("Failed to get process metrics %d: %s" %
                     (pid, e))
      return ret

def get_overcommit_values():
  """Return the system's overcommit memory mode and overcommit ratio from endpoints
  /proc/sys/vm/overcommit_memory and /proc/sys/vm/overcommit_ratio respectively"""
  overcommit_memory = overcommit_ratio = None
  try:
    f = open('/proc/sys/vm/overcommit_memory', 'rb')
    overcommit_memory = int(f.readline())
    if overcommit_memory is None:
      THROTTLED_LOG.warn("'overcommit_memory' couldn't be determined.")
  finally:
    if f is not None:
      f.close()
      f = None
  try:
    f = open('/proc/sys/vm/overcommit_ratio', 'rb')
    overcommit_ratio = int(f.readline())
    if overcommit_ratio is None:
      THROTTLED_LOG.warn("'overcommit_ratio' couldn't be determined.")
  finally:
    if f is not None:
      f.close()

  return overcommit_memory, overcommit_ratio

def get_process_swap_memory(process, get_memory_maps_fn=psutil.Process.memory_maps):
  try:
    path_memory_maps = get_memory_maps_fn(process)
    if 'swap' in path_memory_maps[0]._fields:
      return sum(map(lambda x : x.swap, path_memory_maps))
    else:
      return None
  except (Exception), e:
    THROTTLED_LOG.warn("Failed to get swap memory usage for process %d: %s" %
                       (process.pid, e))
  return None

def add_proc_metrics(update, pid, logger=None):
  metric_map = _get_proc_metrics(pid, False, logger)
  for metric, value in metric_map.items():
    update.add_metric(metric, value)
