# Copyright (c) 2011-2012 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

try:
  from cStringIO import StringIO
except:
  from StringIO import StringIO

import time
from cmf.monitor.firehoses import Firehoses

class RoleUpdate(object):
  """
  This models the RoleUpdate Avro record and gives it a bit more structure.
  """
  def __init__(self, name, roletype, servicetype, service_name, timestamp=None):
    if timestamp is None:
      timestamp = time.time()
    self.timestamp = timestamp
    self._name = name
    self._service_name = service_name
    self._roletype = roletype
    self._servicetype = servicetype
    self._metrics = [ ]
    self._dir_updates = [ ]
    self._flume_updates = [ ]
    self._region_updates = [ ]
    self._regionserver_replication_peer_updates = [ ]
    self._solr_updates = [ ]
    self._additional_role_attributes = None
    self._impala_daemon_update = None
    self._entity_type_updates = None

  @property
  def role_name(self):
    return self._name

  @property
  def service_name(self):
    return self._service_name

  def add_metric(self, metricid, value):
    # A value of None causes avro to blow up when we encode the message, so to
    # be safe, we protect against that here.
    if value is not None:
      self._metrics.append((metricid, value))
    return self

  def add_dir_update(self, role_dir_update):
    if role_dir_update is not None:
      self._dir_updates.append(role_dir_update)
    return self

  def add_flume_update(self, flume_update):
    self._flume_updates.append(flume_update)
    return self

  def set_impala_daemon_update(self, impala_daemon_update):
    self._impala_daemon_update = impala_daemon_update
    return self

  def set_entity_type_updates(self, entity_type_updates):
    self._entity_type_updates = entity_type_updates
    return self

  def add_region_update(self, region_update):
    self._region_updates.append(region_update)
    return self

  def add_regionserver_replication_peer_update(
    self, regionserver_replication_peer_update):
    self._regionserver_replication_peer_updates.append(
      regionserver_replication_peer_update)
    return self

  def add_solr_update(self, solr_update):
    self._solr_updates.append(solr_update)
    return self

  def has_solr_update_for_core(self, core):
    for solr_update in self._solr_updates:
      if core == solr_update.get_replica_name():
        return solr_update
    return None

  def set_additional_role_attributes(self, attributes):
    self._additional_role_attributes = attributes
    return self

  def __str__(self):
    sio = StringIO()
    sio.write(">> %s:" %
              time.strftime("%x %X", time.localtime(self.timestamp)))
    for metricid, val in self._metrics:
      sio.write("\n   %-10d : %s" % (metricid, val))
    for dir_update in self._dir_updates:
      sio.write("\n%s" % (dir_update,))
    for flume_update in self._flume_updates:
      sio.write("\n%s" % (flume_update,))
    for region_update in self._region_updates:
      sio.write("\n%s" % (region_update,))
    for solr_update in self._solr_updates:
      sio.write("\n%s" % (solr_update,))
    for regionserver_replication_update \
          in self._regionserver_replication_peer_updates:
      sio.write("\n%s" % (regionserver_replication_update,))
    return sio.getvalue()

  def to_avro_dict(self):
    """Return a Python dictionary that maps to the Avro struct"""
    impala_daemon_update = None
    if self._impala_daemon_update is not None:
      impala_daemon_update = self._impala_daemon_update.to_avro_dict()
    entity_type_updates = None
    if self._entity_type_updates is not None:
      entity_type_updates = [ e.to_avro_dict()
                              for e
                              in self._entity_type_updates ]
    return dict(ts_secs=long(self.timestamp),
                rolename=self._name,
                roletype=self._roletype,
                serviceType=self._servicetype,
                service_name=self._service_name,
                metrics=[ dict(id=k, value=Firehoses.normalize_metric_value(v))
                          for (k, v) in self._metrics ],
                dirUpdates=[ d.to_avro_dict() for d in self._dir_updates ],
                flumeUpdates=[ f.to_avro_dict() for f in self._flume_updates ],
                regionUpdates=[ r.to_avro_dict() for r in self._region_updates ],
                regionServerReplicationPeerUpdates=[ r.to_avro_dict() for r
                  in self._regionserver_replication_peer_updates ],
                solrUpdates=[ r.to_avro_dict() for r in self._solr_updates ],
                additionalRoleAttributes=self._additional_role_attributes,
                impalaDaemonUpdate=impala_daemon_update,
                entityTypeUpdates=entity_type_updates)
