# Copyright (c) 2013 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging
import os
import simplejson
import time

from cmf.monitor import constants
from cmf.monitor import schema
from cmf.monitor.abstract_solr_monitor import AbstractSolrMonitor
from cmf.monitor.abstract_monitor import CollectionError
from cmf.monitor.base import DebugStatusReport
from cmf.monitor.solrserver.json_parser_jmx import JsonParser
from cmf.monitor.solrserver.json_parser_jmx import IGNORED_SOLR_BEANS
from cmf.monitor.solrserver.json_parser_jmx import extract_solr_bean_id_and_type
from cmf.monitor.solrserver.json_parser_jmx import ignore_solr_metric
from cmf.monitor.solrserver.json_parser_jmx import SOLR_CORE_STATUS_METRICS
from cmf.monitor.solrserver.cdh6.solr7_metrics_handler import Solr7MetricsHandler
from cmf.monitor.solrserver.solr_update import SolrUpdate
from cmf.monitor.json_metrics_utils import add_derived_jvm_metrics_jmx, \
    add_metrics_jmx, add_derived_gc_metrics_jmx
from cmf.monitor.cgroup_metrics_utils import add_cgroup_metrics
from cmf.monitor.proc_metrics_utils import add_proc_metrics
from cmf.monitor.role_update import RoleUpdate
from cmf.throttling_logger import ThrottlingLogger

LOG = logging.getLogger('SolrServerMonitor')
THROTTLED_LOG = ThrottlingLogger(LOG, constants.DEFAULT_MONITOR_LOG_RATE)

_ROLETYPE = 'SOLR_SERVER'
_SERVICETYPE = 'SOLR'

_CORE_STATUS_RELATIVE_URL = "/admin/cores?wt=json&action=STATUS"
_SOLR_KRB5_CC_NAME = "krb5cc_cm_agent_solr"

# Section and key names in cloudera-monitor.properties
# Since CM 5.0
_SECTION = 'solr_server'
_SERVICE_NAME = 'service_name'
_ROLE_NAME = 'role_name'
_HOST = 'host'
_SOLR_SOLRSERVER_HTTP_PORT = 'solr_http_port'
_LOG_DIR = 'log_dir'
_COLLECT_INTERVAL = 'collect_interval'
_SERVICE_VERSION = 'service_version'
# Since CM 5.0.1
_SOLR_KERBEROS_PRINCIPAL = 'solr_kerberos_principal'
# Added in CM 5.4
_HEAP_DUMP_DIR = 'oom_heap_dump_dir'
_USE_SSL = 'use_ssl'
_SOLR_SOLRSERVER_HTTPS_PORT = 'solr_https_port'

class SolrServerMonitor(AbstractSolrMonitor):

  def __init__(self, daemon, registry, cgroup_root=None, cgroup_name=None):
    AbstractSolrMonitor.__init__(self,
                                 daemon,
                                 registry,
                                 cgroup_root,
                                 cgroup_name,
                                 _SECTION,
                                 _SOLR_SOLRSERVER_HTTP_PORT,
                                 _SOLR_SOLRSERVER_HTTPS_PORT,
                                 _ROLETYPE)
    self._last_conf_path = None
    self._metrics_tuple = (None, None, None)
    self._last_reported_web_metrics_timestamp = None
    self._last_reported_core_status_timestamp = None
    self._last_collection_ended = None
    self._last_collection_length = None
    self._role_dir_monitors = [ ]
    self._metric_accessors = daemon.metric_accessors
    # Config data
    self._name = None
    self._service_name = None
    self._log_dir = None
    # Solr core status
    self._core_status_tuple = (None, None, None)
    self._replica_metric_accessors = None
    self._solr7_metrics_handler = Solr7MetricsHandler()

  def __str__(self):
    return "SolrServerMonitor for %s" % (self._name,)

  def _reload_config(self, confparse):
    """Returns True/False for success"""
    try:
      # Extraction
      name = confparse.get(_SECTION, _ROLE_NAME)
      service_name = confparse.get(_SECTION, _SERVICE_NAME)
      service_version = confparse.getint(_SECTION, _SERVICE_VERSION)
      host = confparse.get(_SECTION, _HOST)
      web_http_port = confparse.getint(_SECTION, _SOLR_SOLRSERVER_HTTP_PORT)
      web_https_port = confparse.getint_with_default(_SECTION, _SOLR_SOLRSERVER_HTTPS_PORT,
                                        None)
      log_dir = confparse.get(_SECTION, _LOG_DIR)
      heap_dump_dir = confparse.get_with_default(_SECTION, _HEAP_DUMP_DIR, None)
      # If security is on this will be passed to us in the monitor configuration.
      kerberos_principal = confparse.get_with_default(_SECTION,
                                                      _SOLR_KERBEROS_PRINCIPAL,
                                                      None)
      use_ssl = confparse.getboolean_with_default(_SECTION, _USE_SSL, False)

    except ValueError, ex:
      self.record_error(ex)
      LOG.error("Value error in %s: %s" % (confparse.path, ex))
      return False

    # used by _compute_metrics_url
    self._service_version = service_version

    metrics_url = self._compute_metrics_url(host, web_http_port, web_https_port, use_ssl)
    if metrics_url is None:
      return False


    core_status_url = self._compute_core_status_url(host, web_http_port, web_https_port, use_ssl)
    if core_status_url is None:
      return False

    # Assign
    self._name = name
    self._service_name = service_name
    self._log_dir = log_dir
    self._metrics_url = metrics_url
    self._core_status_url = core_status_url

    # Make monitors for the log directory and the data directories.
    self._role_dir_monitors = [ ]
    if self._log_dir is not None:
      self.create_dir_monitor(self._log_dir, self._role_dir_monitors)
    if heap_dump_dir is not None:
      self.create_dir_monitor(heap_dump_dir, self._role_dir_monitors)

    if not self._service_version in constants.VERSION_STRINGS:
      THROTTLED_LOG.warn("Unknown solrserver version: %s" % (self._service_version))
      return False
    version_string = constants.VERSION_STRINGS[self._service_version]
    if not version_string in self._metric_accessors["SOLR_REPLICA"]:
      THROTTLED_LOG.warn("No metric accessor for version: %s" % (version_string))
      return False

    self._replica_metric_accessors = self._metric_accessors["SOLR_REPLICA"][version_string]
    if not self.set_kerberos_principal_to_use(kerberos_principal,
                                              _SOLR_KRB5_CC_NAME):
      return False
    return True

  def collect(self):
    """Overrides AbstractMonitor. Collect metrics. Invoked asynchronously."""
    start = time.time()

    # Reload config if necessary
    confparse = self.conf_parser
    if self._last_conf_path != confparse.path:
      if not self._reload_config(confparse):
        return
      self._last_conf_path = confparse.path

    # Collect directory info
    for mon in self._role_dir_monitors:
      mon.collect()

    # If our pid value is None, it means that the associated process is not
    # running. If this is the case, we skip metric collection.
    if self._pid is None:
      self._metrics_tuple = (None, None, None)
      return

    ccache_name = self._krb5cc_name
    if ccache_name is not None and not os.path.exists(ccache_name):
      LOG.info(
        "Credential cache %s is missing. Will not attempt to collect "
        "metrics. This is normal if the role has just started." % ccache_name)
      self._metrics_tuple = (None, None, None)
      return

    LOG.debug("Fetching SolrServer metrics at '%s'" % (self._metrics_url,))
    self._metrics_tuple = \
       self.collect_metrics_from_url(self._metrics_url,
                                     use_prefer_first_attribute_hook=True)

    self._last_collection_ended = time.time()
    self._last_collection_length = self._last_collection_ended - start

    self._core_status_tuple = \
       self._collect_core_status_from_url(self._core_status_url)

  def parse_metrics(self, json):
    return JsonParser(json, self._service_version)

  def get_role_update(self):
    """Returns a RoleUpdate to be sent to the firehose"""
    update = RoleUpdate(name=self._name,
                        roletype=_ROLETYPE,
                        servicetype=_SERVICETYPE,
                        service_name=self._service_name)

    if self.add_metrics_from_webserver(self._metrics_tuple,
                                       self._last_reported_web_metrics_timestamp,
                                       update):
      self._last_reported_web_metrics_timestamp = update.timestamp

    add_proc_metrics(update, self._pid, LOG)
    add_cgroup_metrics(update, self._cgroup_root, self._cgroup_name)

    for mon in self._role_dir_monitors:
      update.add_dir_update(mon.get_role_dir_update())

    self.add_unexpected_exit_metrics(update)

    if self.add_metrics_from_core_status_retrieval(self._core_status_tuple,
                                                   self._last_reported_core_status_timestamp,
                                                   update):
      self._last_reported_core_status_timestamp = update.timestamp
    return update

  def add_webserver_metrics_to_role_update(self, update, metrics):
    if self._service_version in constants.VERSION_STRINGS:
      self._add_webserver_metrics_to_role_update_with_version(update, metrics, self._service_version)
    else:
      THROTTLED_LOG.warn("Unknown solrserver version: %s" % (self._service_version))

  def _add_webserver_metrics_to_role_update_with_version(self, update, metrics, version):
    if version == constants.SERVICE_VERSION_CDH5:
      self._add_webserver_metrics_to_role_update_until_cdh5(update, metrics, version)
    else:
      self._add_webserver_metrics_to_role_update_since_cdh6(update, metrics, version)

  def _add_webserver_metrics_to_role_update_until_cdh5(self, update, metrics, version):
    collect_all_memory_metrics = True
    include_jvm_metrics = False
    add_derived_jvm_metrics_jmx(update, metrics, collect_all_memory_metrics, \
                                include_jvm_metrics)
    add_metrics_jmx(update, metrics, self._metric_accessors, _ROLETYPE, constants.VERSION_STRINGS[version])
    add_derived_gc_metrics_jmx(update, metrics)
    self._add_solr_replica_metrics(update, metrics)

  def _add_webserver_metrics_to_role_update_since_cdh6(self, update, metrics, version):
    roleAccessors = self._metric_accessors.get(_ROLETYPE, {}).get(constants.VERSION_STRINGS[version], {})
    self._solr7_metrics_handler.add_jvm_metrics(update, metrics, roleAccessors)
    if self._is_feature_enabled(update, "send_replica_update"):
      self._solr7_metrics_handler.add_replica_metrics(update, metrics, self._replica_metric_accessors)
    return None
  # Retrieves the CORE category MBeans from a Solr server's JMX JSON snapshot.
  # The 'coreHashCode' attribute is used as an identifier of a particular
  # core / replica. All other categories of MBeans applicable to the same
  # core / replica will include this hash code as a reference in their
  # metric snapshots.
  # An example of the JMX JSON snapshot for a core is:
  # ..
  # {
  #   "name" : "solr/mr_col_shard2_replica1:type=core,id=mr_col_shard2_replica1",
  #   "modelerType" : "org.apache.solr.core.JmxMonitoredMap$SolrDynamicMBean",
  #   "category" : "CORE",
  #   "source" : "$URL$",
  #   "description" : "SolrCore",
  #   "name" : "mr_col_shard2_replica1",
  #   "version" : "1.0",
  #   "coreHashCode" : "756266007",
  #   "coreName" : "mr_col_shard2_replica1",
  #   "startTime" : "Tue Nov 12 13:59:15 PST 2013",
  #   "refCount" : 1,
  #   "indexDir" : "hdfs://abs-1.ent.cloudera.com:8020/solr/mr_col/core_node4/data/index/",
  #   "aliases" : "[mr_col_shard2_replica1]",
  #   "collection" : "mr_col",
  #   "shard" : "shard2"
  # }
  #
  def _get_cores(self, metrics, update):
    bean_dict = getattr(metrics, 'core', None)
    if bean_dict is None:
      return
    cores = dict()
    for bean in bean_dict:
      core_name = bean.get('coreName', None)
      if core_name is None:
        continue
      core_hash = bean.get('coreHashCode', None)
      if core_hash is None:
        continue
      cores[core_hash] = self._get_or_create_solr_update(core_name, update)
    return cores

  # Gets an existing SolrUpdate or creates one and adds it
  # to the role update.
  def _get_or_create_solr_update(self, core_name, update):
    solr_update = update.has_solr_update_for_core(core_name)
    if solr_update is None:
      solr_update = SolrUpdate(core_name)
      update.add_solr_update(solr_update)
    return solr_update

  # Metrics from the UPDATEHANDLER MBean category are appended to the
  # SolrUpdates associated with relevant core in this method.
  def _add_update_handler_metrics(self, update, metrics, cores):
    bean_dicts = getattr(metrics, 'updatehandler', None)
    if bean_dicts is None:
      return
    for bean in bean_dicts:
      core_hash = bean.get('coreHashCode', None)
      if core_hash is None or core_hash not in cores:
        continue
      solr_update = cores[core_hash]
      for metric_id, accessor in self._replica_metric_accessors.iteritems():
        if accessor.metric in bean:
          solr_update.add_metric(metric_id, bean[accessor.metric])

  # Metrics from the QUERYHANDLER, CACHE MBean categories are appended to the
  # SolrUpdates associated with the relevant core in this method.
  def _add_handler_metrics(self, handler, update, metrics, cores):
    bean_dicts = getattr(metrics, handler, None)
    if bean_dicts is None:
      return
    ignored_handlers = IGNORED_SOLR_BEANS.get(handler.upper(), [])
    for bean in bean_dicts:
      core_hash = bean.get('coreHashCode', None)
      if core_hash is None or core_hash not in cores:
        continue
      bean_name = bean.get('name', None)
      if bean_name is None:
        continue
      bean_id, bean_type = extract_solr_bean_id_and_type(bean_name, handler.upper())
      if bean_id is None or bean_id in ignored_handlers:
        continue
      solr_update = cores[core_hash]
      for metric_id, accessor in self._replica_metric_accessors.iteritems():
        if bean_id == accessor.bean_id and accessor.metric in bean:
          if bean_type is None or bean_type == accessor.bean_type:
            if ignore_solr_metric(handler.upper(), accessor.bean_id, accessor.metric):
              continue
            solr_update.add_metric(metric_id, bean[accessor.metric])

  def _add_solr_replica_metrics(self, update, metrics):
    if not self._is_feature_enabled(update, "send_replica_update"):
      return
    cores = self._get_cores(metrics, update)
    self._add_update_handler_metrics(update, metrics, cores)
    self._add_handler_metrics('queryhandler', update, metrics, cores)
    self._add_handler_metrics('cache', update, metrics, cores)

  def _compute_core_status_url(self, host, http_port, https_port, use_ssl):
    if use_ssl:
      if host is None or https_port is None:
        return None
      return "https://%s:%s/solr%s" % (host, https_port, _CORE_STATUS_RELATIVE_URL)
    else:
      if host is None or http_port is None:
        return None
    return "http://%s:%s/solr%s" % (host, http_port, _CORE_STATUS_RELATIVE_URL,)

  def _collect_core_status_from_url(self, url, username=None, password=None):
    result = None
    start = time.time()
    try:
      code = constants.SOLR_CORE_STATUS_COLLECTION_COMMUNICATION_FAILURE
      openedUrl = self.urlopen(url, username=username, password=password)
      code = constants.SOLR_CORE_STATUS_COLLECTION_PARSE_FAILURE
      result = simplejson.load(openedUrl)
    except Exception, ex:
      THROTTLED_LOG.exception("Error fetching Solr core status at '%s'" % (url,))
      result = CollectionError(code)
    now = time.time()
    duration = int((now - start) * 1000.0)
    return (result, now, duration)

  def _add_core_status_metrics(self, update, core_status):
    if not self._is_feature_enabled(update, "send_replica_update"):
      return
    status = core_status.get("status", None)
    if status is None:
      THROTTLED_LOG.warn("Cannot find 'status' in Solr Core status json")
      return
    for replica_name in status.keys():
      replica = status.get(replica_name)
      if replica is None:
        THROTTLED_LOG.warn("Cannot find replica info for %s in Solr core status json" % \
                           (replica_name,))
        continue
      index = replica.get("index", None)
      if index is None:
        THROTTLED_LOG.warn("Cannot find 'index' info for %s in Solr replica json" % \
                           (replica_name,))
        continue
      solr_update = self._get_or_create_solr_update(replica_name, update)
      for metric_name, metric_id in SOLR_CORE_STATUS_METRICS.items():
        metric_value = index.get(metric_name, None)
        if metric_value is None:
          THROTTLED_LOG.warn("Metric %s not found in 'index' for replica %s " % \
                             (metric_name, replica_name,))
        else:
          solr_update.add_metric(metric_id, metric_value)

  def add_metrics_from_core_status_retrieval(self,
                                             core_status_tuple,
                                             last_reported_timestamp,
                                             update):
    core_status, timestamp, duration = core_status_tuple
    if core_status is None or timestamp == last_reported_timestamp:
      return False

    update.add_metric(schema.SOLR_CORE_STATUS_COLLECTION_DURATION, duration)
    update.timestamp = timestamp
    if isinstance(core_status, CollectionError):
      update.add_metric(schema.SOLR_CORE_STATUS_COLLECTION_STATUS, core_status.code)
    else:
      update.add_metric(schema.SOLR_CORE_STATUS_COLLECTION_STATUS,
                        constants.SOLR_CORE_STATUS_COLLECTION_GOOD)
      self._add_core_status_metrics(update, core_status)
    return True

  def get_debug_status_report(self):
    return DebugStatusReport(config_items=self.conf_parser.items(_SECTION),
                             errors=(self.last_error,),
                             last_collection_ended=self._last_collection_ended,
                             last_collection_length=self._last_collection_length)
