import logging
import re
from cmf.throttling_logger import ThrottlingLogger
from cmf.monitor import constants
from cmf.monitor.solrserver.solr_update import SolrUpdate
from cmf.monitor import schema
from cmf.monitor.json_metrics_utils import MetricAccessor

LOG = logging.getLogger('solrserver.Solr7MetricsHandler')
THROTTLED_LOG = ThrottlingLogger(LOG, constants.DEFAULT_MONITOR_LOG_RATE)
_BYTES_PER_MB = 1048576

class Solr7MetricsHandler(object):

  # Adds replica metrics to the update object
  # from the solr metrics output by the replica accessors
  def add_replica_metrics(self, update, metrics, accessors):
    coreUpdates = self._get_core_updates(metrics, update);
    solrMetrics = getattr(metrics, "metrics");
    nodeAccessors = {}
    coreAccessors = {}
    for metric_id, accessor in accessors.iteritems():
      if not isinstance(accessor, MetricAccessor):
        LOG.warn("Accessor type of %s (%s) is not valid in Solr7" % (accessor.__class__.__name__, accessor.bean_id))
        continue
      if accessor.category=="CORE":
        coreAccessors[metric_id] = accessor
      else:
        nodeAccessors[metric_id] = accessor

    for core_key, solr_update in coreUpdates.iteritems():
      dictKey = "solr.core." + core_key
      coreMetrics = solrMetrics[dictKey]
      if coreMetrics is None:
        LOG.warn("Metrics for replica '%s' were not found in metrics response" % core_key)
        continue
      self._add_metrics_by(coreMetrics, coreAccessors, solr_update, False)

    self._add_metrics_by(solrMetrics, nodeAccessors, update, True)


  # Adds jvm metrics to the update object
  # from the solr metrics output
  def add_jvm_metrics(self, update, metrics, accessors):
    solrMetrics = getattr(metrics, "metrics");
    self._add_metrics_by(solrMetrics, accessors, update, True)

    solrJvm = solrMetrics['solr.jvm'];
    update.add_metric(schema.JVM_GC_COUNT, self._sum_by_pattern(solrJvm, re.compile('^gc\.[^\.]+\.count$')))
    update.add_metric(schema.JVM_GC_TIME_MS, self._sum_by_pattern(solrJvm, re.compile('^gc\.[^\.]+\.time$')))

    update.add_metric(schema.JVM_MAX_MEMORY_MB,
                      solrJvm['memory.heap.max'] / float(_BYTES_PER_MB))
    update.add_metric(schema.JVM_HEAP_COMMITTED_MB,
                      solrJvm['memory.heap.committed'] / float(_BYTES_PER_MB))
    update.add_metric(schema.JVM_HEAP_USED_MB,
                      solrJvm['memory.heap.used'] / float(_BYTES_PER_MB))
    update.add_metric(schema.JVM_NON_HEAP_COMMITTED_MB,
                      solrJvm['memory.non-heap.committed'] / float(_BYTES_PER_MB))
    update.add_metric(schema.JVM_NON_HEAP_USED_MB,
                      solrJvm['memory.non-heap.used'] / float(_BYTES_PER_MB))

  def _sum_by_pattern(self, map, key_pattern):
    result = None
    for key, val in map.iteritems():
      if key_pattern.match(key) is not None:
        if result is None:
          result = 0
        result += val
    return result

  def _add_metrics_by(self, root, accessors , solr_update, resolveCategory):
    for metric_id, accessor in accessors.iteritems():
      if not isinstance(accessor, MetricAccessor):
        LOG.warn("Accessor type of %s is not valid in Solr7" % accessor.__class__.__name__)
        continue
      if resolveCategory:
        if not accessor.category in root:
          LOG.warn("%s category node was not found" % accessor.category)
          continue
        metricsNode = root[accessor.category]
      else:
        metricsNode = root
      if accessor.section is None:
        section = metricsNode
      else:
        if accessor.section not in metricsNode:
          continue
        section = metricsNode[accessor.section]
      if accessor.metric in section:
        solr_update.add_metric(metric_id, section[accessor.metric])

  def _get_core_updates(self, metrics, update):
    solrMetrics = getattr(metrics, 'metrics', None)
    cores = dict()
    for nodeKey, node in solrMetrics.iteritems():
      if nodeKey[:10] != "solr.core.":
        continue
      core_key = nodeKey[10:]
      components1 = core_key.split(".")
      components = [ '', components1[len(components1) - 2], components1[len(components1) - 1] ]
      components[0] = components1[0]
      i = 1
      while i < len(components1) - 2 :
        components[0] = components[0] + '.' + components1[i]
        i = i + 1
      cores[core_key] = self._get_or_create_solr_update(components, update)
    return cores

    # Gets an existing SolrUpdate or creates one and adds it
    # to the role update.
  def _get_or_create_solr_update(self, components, update):
    core_name = "_".join(components)
    solr_update = update.has_solr_update_for_core(core_name)
    if solr_update is None:
      solr_update = SolrUpdate(components)
      update.add_solr_update(solr_update)
    return solr_update