# Copyright (c) 2013 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import re
from cmf.monitor.jmx_beans import JsonParserJmx
from cmf.monitor.json_metrics_utils import ParseError
from cmf.monitor import schema
from cmf.monitor import constants

# We inspect Solr metrics based on their categories. In a JSON JMX
# dump, these will be present in a bean under the 'category' attribute.
# We aren't yet using the metrics provided by these Solr MBean categories.
# As we start using them, move them out of this list.
IGNORED_SOLR_CATEGORIES = [
  'OTHER',
  'HIGHLIGHTING',
  'QUERYPARSER'
]

# A map of 'id's associated with QUERYHANDLER beans and the
# corresponding prefix attached to their metrics.
# Where an 'id' is insufficient to distinguish between beans,
# a bean type is used. For such 'id's, the value is a map
# keyed by the bean type.
# For example we have multiple types of beans associated with
# the QUERYHANDLER beans of id 'org.apache.solr.handler.component.SearchHandler'
# To distinguish between beans with the following names:
# 1. "solr/mr_collection_shard2_replica1:type=/query,id=org.apache.solr.handler.component.SearchHandler" and
# 2. "solr/mr_collection_shard2_replica1:type=/select,id=org.apache.solr.handler.component.SearchHandler"
# we need to use both the 'id' 'org.apache.solr.handler.component.SearchHandler'
# and the 'type' ('/query' or '/select') to ensure that metrics
# are associated with the correct bean.
QUERYHANDLER_BEAN_TYPES = {
  "Lazy[solr.SearchHandler]" : "lazy_search",
  "Lazy[solr.DocumentAnalysisRequestHandler]" : "lazy_doc_analysis",
  "Lazy[solr.FieldAnalysisRequestHandler]" : "lazy_field_analysis",
  "org.apache.solr.handler.ReplicationHandler" : "replication",
  "org.apache.solr.handler.UpdateRequestHandler" : "update",
  "org.apache.solr.handler.PingRequestHandler" : "ping",
  "org.apache.solr.handler.CSVRequestHandler" : "csv",
  "org.apache.solr.handler.DumpRequestHandler" : "dump",
  "org.apache.solr.handler.JsonUpdateRequestHandler" : "json_update",
  "org.apache.solr.handler.RealTimeGetHandler" : "real_time_get",
  "org.apache.solr.handler.admin.AdminHandlers" : "admin",
  "org.apache.solr.handler.admin.LoggingHandler" : "logging",
  "org.apache.solr.handler.admin.LukeRequestHandler" : "luke",
  "org.apache.solr.handler.admin.PluginInfoHandler" : "plugin",
  "org.apache.solr.handler.admin.PropertiesRequestHandler" : "properties",
  "org.apache.solr.handler.admin.ShowFileRequestHandler" : "show_file",
  "org.apache.solr.handler.admin.SolrInfoMBeanHandler" : "solr_info",
  "org.apache.solr.handler.admin.SystemInfoHandler" : "system_info",
  "org.apache.solr.handler.admin.ThreadDumpHandler" : "thread",
  "org.apache.solr.handler.component.SearchHandler" : {
    "/query" : "query",
    "/select" : "select",
    "/browse" : "browse",
    "org.apache.solr.handler.component.SearchHandler" : "default",
  }
}

# A map of types associated with CACHE beans and the
# corresponding prefix attached to their metrics.
CACHE_BEAN_TYPES = {
  "documentCache" : "document_cache",
  "fieldCache" : "field_cache",
  "filterCache" : "filter_cache",
  "queryResultCache" : "query_result_cache",
  "fieldValueCache" : "field_value_cache"
}

# A list of all Solr categories and their associated bean types.
# Note that a bean type is needed only if we want to distinguish
# between the beans of interest of a particular category.
# For e.g. the QUERYHANDLER category has ReplicationHandler,
# UpdateRequestHandler, PingRequestHandler etc. By associating
# bean types with QUERYHANDLERS, the metrics associated with thse
# beans can be distinguished.
SOLR_BEAN_TYPES = {
  "QUERYHANDLER": QUERYHANDLER_BEAN_TYPES,
  "CACHE": CACHE_BEAN_TYPES,
  "CORE": {},
  "UPDATEHANDLER": {}
}

# We aren't yet using the metrics provided by these QUERYHANDLER
# bean types. As we start using them, move them out of this list.
IGNORED_QUERYHANDLERS = [
  "Lazy[solr.SearchHandler]",
  "Lazy[solr.DocumentAnalysisRequestHandler]",
  "Lazy[solr.FieldAnalysisRequestHandler]",
  "org.apache.solr.handler.DumpRequestHandler",
  "org.apache.solr.handler.admin.AdminHandlers",
  "org.apache.solr.handler.admin.LoggingHandler",
  "org.apache.solr.handler.admin.PluginInfoHandler",
  "org.apache.solr.handler.admin.PropertiesRequestHandler",
  "org.apache.solr.handler.admin.ThreadDumpHandler",
]

# Add ignore caches to this list
IGNORED_CACHES = [
]

# Global map of all the Solr beans that we are not gathering
# metrics from.
IGNORED_SOLR_BEANS = {
  "QUERYHANDLER": IGNORED_QUERYHANDLERS,
  "CACHE": IGNORED_CACHES
}

# Common attributes across most / all Solr beans that are ignored
# These are mainly text fields / provide descriptions of metrics
# rather than actually being metrics.
IGNORED_SOLR_COMMON_METRICS = [
  'category',
  'description',
  'source',
  'version',
  'startTime',
  'indexDir',
  'aliases',
  'collection',
  'shard',
  'searcherName',
  'caching',
  'reader',
  'readerDir',
  'openedAt',
  'registeredAt',
  'coreHashCode',
  'indexVersion',
  'coreName',
  'refCount'
]

# Map of bean types and the associated metrics that are ignored
# for QUERYHANDLERS.
IGNORED_QUERYHANDLER_METRICS = {
  "org.apache.solr.handler.ReplicationHandler" : [
    "indexSize",
    "indexVersion",
    "generation",
    "indexPath",
    "isMaster",
    "isSlave",
    "masterUrl",
    "isPollingDisabled",
    "isReplicating",
    "replicateAfter",
    "replicationEnabled",
    "indexReplicatedAt",
  ]
}

# Map of bean types and the associated metrics that are ignored
# for CACHES.
IGNORED_CACHE_METRICS = {
  "fieldCache" : [
    "insanity_count",
  ]
}

# Global map of all the Solr metrics that are being ignored
# key-ed by the category.
IGNORED_SOLR_METRICS = {
  "QUERYHANDLER": IGNORED_QUERYHANDLER_METRICS,
  "CACHE": IGNORED_CACHE_METRICS
}

SOLR_METRIC_NAME_OVERRIDES = \
    dict({
      # We do not want this to be mistaken for a counter,
      # rename metric to not have 'count'
      'entries_count' : 'entries'
      })

# Extractor used to retrieve the value of the 'id' attribute from a Bean name.
# This value helps distinguish between different kinds of QUERYHANDLER,
# UPDATEHANDLER and CORE beans.
# Given a name of "solr/mr_collection_shard2_replica1:type=\
#                  /query,id=org.apache.solr.handler.component.SearchHandler"
# using the following pattern to match will return:
# 'org.apache.solr.handler.component.SearchHandler'
_BEAN_ID_EXTRACTOR='.*,id=(.*$)'

# Extractor used to retrieve the value of the 'type' attribute from a Bean name.
# This value is used to distinguish between instances of beans that share the
# 'id' for QUERYHANDLER beans and between different kinds of CACHE beans.
# Given a name of "solr/simple_collection_shard2_replica2:type=\
#                  documentCache,id=org.apache.solr.search.LRUCache"
# using the following pattern to match will return: 'documentCache'
_BEAN_TYPE_EXTRACTOR='.*:type=(.*)(,.*$)'

# Solr bean names are derived from JMX ObjectNames and have attributes in them.
# Patterns are used to extract attribute values that help distinguish
# between different kinds of beans within a bean category.
# Map of categories and their associated extractors.
EXTRACTION_PATTERNS = {
  "CORE": _BEAN_ID_EXTRACTOR,
  "UPDATEHANDLER": _BEAN_ID_EXTRACTOR,
  "QUERYHANDLER": {
    'id_pattern' : _BEAN_ID_EXTRACTOR,
    'type_pattern' : _BEAN_TYPE_EXTRACTOR
  },
  "CACHE": _BEAN_TYPE_EXTRACTOR
}

# Extract the Solr bean id and type from the bean name
def extract_solr_bean_id_and_type(name, category):
  if category not in EXTRACTION_PATTERNS:
    return (None, None)

  pattern = EXTRACTION_PATTERNS[category]
  if isinstance(pattern, str):
    id = re.search(pattern, name)
    if id is not None:
      return (id.group(1), None)
  else:
    # We have a map as the pattern
    id_pattern = pattern['id_pattern']
    type_pattern = pattern['type_pattern']
    id = re.search(id_pattern, name)
    if id is None:
      return (None, None)
    id = id.group(1)
    bean_type = QUERYHANDLER_BEAN_TYPES.get(id, None)
    if bean_type is None:
      return (id, None)
    type = None
    if not isinstance(bean_type, str):
      # We are dealing with a QUERYHANDLER for which
      # we need to look at bean types.
      type = re.search(type_pattern, name)
      if type is not None:
        type = type.group(1)
    return (id, type)

def ignore_solr_metric(category, bean_type, metric):
  if metric in IGNORED_SOLR_COMMON_METRICS:
    return True
  if category in IGNORED_SOLR_METRICS:
    ignored_category_metrics = IGNORED_SOLR_METRICS[category]
    ignored_bean_type_metrics = ignored_category_metrics.get(bean_type, None)
    if ignored_bean_type_metrics is not None and metric in ignored_bean_type_metrics:
      return True
  return False

# List of metrics we get from the Solr Core Status API call
SOLR_CORE_STATUS_METRICS = {
  "sizeInBytes" : schema.INDEX_SIZE,
  "numDocs" : schema.NUM_DOCS,
  "maxDoc" : schema.MAX_DOC,
  "deletedDocs" : schema.DELETED_DOCS
}

class JsonParser(JsonParserJmx):
  def __init__(self, json_dict, version):
    """
    May raise ParseError if the json structure is not what we expect.
    """
    if version == constants.SERVICE_VERSION_CDH5: # CDH5
      JsonParserJmx.__init__(self, json_dict, 'SOLR_SERVER')
    else: # CDH6 and above
      self._parse_nodes_cdh6(json_dict)

  def _parse_nodes_cdh6(self, json_dict):
    for key in  ["metrics"]:
      node = json_dict.get(key)
      if node is None:
        raise ParseError("'%s' node was not found on metrics result" % key)
      setattr(self, key, node)
