# Copyright (c) 2014 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import bz2
import logging
import logging.handlers
import os
import re

LOG = logging.getLogger(__name__)

def get_sorted_log_files(directory, filename):
  """ 
  Returns the log files that exist in sorted order. The list is returned in two
  forms, a set of file ids and a set of full filenames. The file list will
  contain an entry for the uncompressed filename if it exists. The id list will
  contain the numerical id for each of the compressed files. For example for
  '/dir/log.2.bz2' it would have a 2.
  """
  pattern = re.compile("^%s\.(\d+)\.bz2$" % re.escape(filename))
  fullname = os.path.join(directory, filename)

  ids = []
  for f in os.listdir(directory):
    match = pattern.match(f)
    if match is not None:
      ids.append(int(match.group(1)))
  ids = sorted(ids)

  files = []
  if os.path.exists(fullname):
    files.append(fullname)
  for i in ids:
    files.append("%s.%d.bz2" % (fullname, i))

  return ids, files

class _CompressingRotatingFileHandler(logging.handlers.RotatingFileHandler):
  """
  This is a customized RotatingFileHandler that does the following:
  #1. bzip2 compresses when rotating files, the actual file being logged to is
      uncompressed, but when it is rotated out the contents are compressed.
  #2. retains a certain amount of data rather than a fixed number of files. When
      files are detailed, the oldest files are deleted first.
  """
  def __init__(self,
               logfile,
               max_uncompressed_file_size_bytes,
               bytes_retained):
    logging.handlers.RotatingFileHandler.__init__(
      self, logfile, maxBytes=max_uncompressed_file_size_bytes)
    self._max_uncompressed_file_size = max_uncompressed_file_size_bytes
    self._bytes_retained = bytes_retained
    self._log_directory, self._filename = os.path.split(logfile)

  def doRollover(self):
    if self.stream:
      self.stream.close()
      self.stream = None

    ids, files = get_sorted_log_files(self._log_directory, self._filename)

    # Next we delete all later files that are over our size limit, if any. We
    # walk the files in order (1, 2, 3 ...) accumulating our total size and
    # delete all those past our size limit.
    size = 0
    if self._bytes_retained > 0:
      for filename in files:
        size = size + os.path.getsize(filename)
        if filename == self.baseFilename:
          continue
        if size > self._bytes_retained:
          LOG.debug("Over data retention limit. Removing '%s'." % filename)
          os.remove(filename)

    # Finally, we perform file rotation for the remaining files.
    if len(ids) > 0:
      for i in range(ids[-1], 0, -1):
        # There's no need to rotate this file if the previous file does not
        # exist.
        if i > 1 and not os.path.exists("%s.%d.bz2" % (self.baseFilename, i - 1)):
          continue
        sfn = "%s.%d.bz2" % (self.baseFilename, i)
        dfn = "%s.%d.bz2" % (self.baseFilename, i + 1)
        if os.path.exists(sfn):
          os.rename(sfn, dfn)
    dfn = self.baseFilename + ".1.bz2"
    if os.path.exists(dfn):
      os.remove(dfn)
    f_in = None
    f_out = None
    try:
      f_in = open(self.baseFilename, 'rb')
      f_out = bz2.BZ2File(dfn, 'wb')
      f_out.writelines(f_in)
    except:
      if not os.path.exists(dfn):
        if os.path.exists(self.baseFilename):
          os.rename(self.baseFilename, dfn)
    if f_in is not None:
      f_in.close()
    if f_out is not None:
      f_out.close()
    if os.path.exists(self.baseFilename):
      os.remove(self.baseFilename)

    self.stream = open(self.baseFilename, 'w')

class CompressingRotatingLogger(object):
  """
  A logger with a customized RotatingFileHandler built to compress while it does
  file rotation and with a configurable size-based data retention policy. See
  the comments in the handler above for more detail.
  """
  def __init__(self,
               logfile,
               logformat,
               dateformat,
               max_uncompressed_file_size_bytes,
               bytes_retained):
    handler = _CompressingRotatingFileHandler(logfile,
                                              max_uncompressed_file_size_bytes,
                                              bytes_retained)
    handler.setFormatter(logging.Formatter(logformat, dateformat))
    self._logger = logging.Logger(logfile)
    self._logger.propagate = False
    self._logger.setLevel(logging.INFO)
    self._logger.addHandler(handler)

  def log(self, message):
    self._logger.info(message)
