# Copyright (c) 2014 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging
import os
import psutil
import subprocess
import threading
import time

from cmf.monitor.constants import DATE_FORMAT
from cmf.monitor import constants
from cmf.monitor.stacks.compressing_rotating_logger import CompressingRotatingLogger
from cmf.monitor.stoppable_thread import StoppableThread
from cmf.throttling_logger import ThrottlingLogger
from cmf.version import VERSION
import cmf.version_detector

LOG = logging.getLogger(__name__)
THROTTLED_LOG = ThrottlingLogger(LOG, constants.DEFAULT_MONITOR_LOG_RATE)

_SECTION = 'stacks'
_ROLENAME = 'rolename'
_METHOD = 'stacks_collection_method'
_ENABLED = 'stacks_collection_enabled'
_LOGFILE = 'stacks_collection_log_file'
_FREQUENCY = 'stacks_collection_frequency'
_SERVLET_URL = 'stacks_collection_servlet_url'
_DATA_RETENTION = 'stacks_collection_data_retention'

# We want the following information in our stacks log messages:
# 1. the timestamp
# 2. the collection method, which is visible in the thread name
# 3. the agent version
# 4. the collection duration (specified in the message below, see OPSAPS-23438)
# 5. the sample itself
# We do all of this once per line since it is simple to do so and it compresses
# very well.
LOG_FORMAT = "[%(asctime)s] %(threadName)s " + VERSION + "%(message)s"

class _AbstractCollector(object):
  """Base class for all stack collectors. This class handles logic common to all
     collectors: setting up the periodic collection thread and the logger. Base
     classes need to implement the collect method. They may also need to compose
     with the refresh method to refresh their own specific state.

     We assume calls to refresh and stop are serialized by the caller."""

  def __init__(self):
    self._stopped = False
    self._thread = None
    self._logger = None
    self._method = None
    self._logfile = None
    self._rolename = None
    self._data_retention = None

  def refresh(self, conf, max_uncompressed_file_size_bytes):
    """Updates the collector based on the input configuration file. This returns
       False if the collector could not be updated and should be stopped."""
    if self._stopped:
      return False

    if ((not self._check_and_set_param(
          conf, _ROLENAME, '_rolename', None, conf.get_with_default)) or
        (not self._check_and_set_param(
          conf, _LOGFILE, '_logfile', None, conf.get_with_default)) or
        (not self._check_and_set_param(
          conf, _METHOD, '_method', None, conf.get_with_default)) or
        (not self._check_and_set_param(
          conf, _DATA_RETENTION, '_data_retention', -1, conf.getint_with_default))):
      return False
    frequency = conf.getfloat_with_default(_SECTION, _FREQUENCY, 5.0)

    if self._thread is None:
      self._logger = CompressingRotatingLogger(self._logfile,
                                               LOG_FORMAT,
                                               DATE_FORMAT,
                                               max_uncompressed_file_size_bytes,
                                               self._data_retention)
      thread_name = "%s-%s-stacks-collector" % (self._rolename, self._method)
      self._thread = StoppableThread(thread_name,
                                     frequency,
                                     self._collect)
      self._thread.start()
    else:
      self._thread.set_interval(frequency)
    return True

  def _check_and_set_param(self, conf, key, attribute, default, lookup_fn):
    value = lookup_fn(_SECTION, key, default)
    if value is None:
      THROTTLED_LOG.error("No %s required setting in %s." % (key, conf.path))
      return False
    current_value = getattr(self, attribute)
    if current_value is not None and current_value != value:
      THROTTLED_LOG.info("%s setting changed from %s to %s in %s. The existing "
                         "monitor will be stopped then a new monitor will be "
                         "started." % (key,
                                       current_value,
                                       value,
                                       conf.path))
      return False
    setattr(self, attribute, value)
    return True

  def stop(self):
    """Stops the collector. Signals any thread to stop, but does not wait."""
    if not self._stopped:
      if self._thread is not None:
        self._thread.should_stop()
      self._stopped = True

  def get_role_name(self):
    return self._rolename

  def _collect(self):
    """To be implemented by the specific stack collectors."""
    pass

  def _log_sample(self, sample, duration):
    """..."""
    self._logger.log(("__COLLECTION_DURATION__ %d __BEGIN_STACKS_SAMPLE__\n" +
                      "%s\n__END_STACKS_SAMPLE__") % (duration, sample))

class _JstackCollector(_AbstractCollector):
  """A stack collector that invokes the jstack utility to collect stacks."""

  def __init__(self):
    _AbstractCollector.__init__(self)
    self._pid = None

  def refresh(self, conf, max_uncompressed_file_size_bytes, pid):
    if self._pid != pid:
      LOG.info("%s pid change from '%s' to '%s'" %
               (conf.get_with_default(_SECTION, _ROLENAME, None),
                str(self._pid),
                str(pid)))
      self._pid = pid
    return _AbstractCollector.refresh(self,
                                      conf,
                                      max_uncompressed_file_size_bytes)

  def _switch_user(self, gid, uid):
    """This function is used as a preexec function in the call to Popen below.
       This should be kept as simple as possible to avoid hitting errors and
       opening us up to http://bugs.python.org/issue16327."""
    os.setgid(gid)
    os.setuid(uid)

  def _collect(self):
    try:
      pid = self._pid
      if pid is None:
        return
      java = cmf.version_detector.VersionDetector.get_java_location()
      if java is None:
        return
      jstack = os.path.join(java, 'bin', 'jstack')
      if not os.path.exists(jstack):
        THROTTLED_LOG.info("Not collecting: '%s' does not exist." % jstack)
        return
      start = time.time()
      process = psutil.Process(pid)
      gid = process.gids()[0]
      uid = process.uids()[0]
      p = subprocess.Popen([jstack, str(pid)],
                           preexec_fn=lambda : self._switch_user(gid, uid),
                           close_fds=True,
                           stdin=None,
                           stdout=subprocess.PIPE,
                           stderr=subprocess.PIPE)
      stdout, stderr = p.communicate()
      if stdout:
        duration = int((time.time() - start) * 1000.0)
        self._log_sample(stdout, duration)
      else:
        THROTTLED_LOG.info("Failed to collect stacks: %s" % stderr)
    except Exception, e:
      THROTTLED_LOG.exception("Failed to collect stacks")

class _ServletCollector(_AbstractCollector):
  """A stack collector that pulls stacks from a URL. This takes a function to do
     the actual urlopen call so that callers can pass in something that deals
     with any authentication that is necessary."""

  def __init__(self, urlopen):
    _AbstractCollector.__init__(self)
    self._urlopen = urlopen
    self._url = None

  def refresh(self, conf, max_uncompressed_file_size_bytes, pid):
    url = conf.get_with_default(_SECTION, _SERVLET_URL, None)
    if url is None:
      THROTTLED_LOG.error("No %s required setting in %s." % (_SERVLET_URL,
                                                             conf.path))
      return False
    if self._url != url:
      LOG.info("%s url change from '%s' to '%s'" % (
          conf.get_with_default(_SECTION, _ROLENAME, None),
          self._url,
          url))
      self._url = url
    return _AbstractCollector.refresh(self,
                                      conf,
                                      max_uncompressed_file_size_bytes)

  def _collect(self):
    try:
      start = time.time()
      sample = self._urlopen(self._url).read()
      duration = int((time.time() - start) * 1000.0)
      self._log_sample(sample, duration)
    except Exception, e:
      THROTTLED_LOG.exception("Failed to collect stacks from '%s'" % self._url)

class StacksCollectionManager(object):
  """Object used to coordinate periodic stacks collection of monitored roles. A
     single StacksCollectionManager exists per-agent. Periodic stacks collection
     for a role is configured through a cloudera-stacks-monitor.propreties file.

     The expectation is that a role monitor will call refresh_collector as part
     of it's periodic reporting cycle. This will start the monitor when needed
     and will deal with changes to the pid, etc. When a role monitor is stopped,
     it must call stop_collector.

     This class is thread safe. We expect calls from different monitoring
     threads."""

  def __init__(self, max_uncompressed_file_size_bytes):
    if max_uncompressed_file_size_bytes is None:
      raise "Invalid arguments: max uncompressed file size bytes required."
    self._collectors = dict()
    self._max_uncompressed_file_size_bytes = max_uncompressed_file_size_bytes
    self._lock = threading.RLock()
    LOG.info("Using max_uncompressed_file_size_bytes: %d" %
             max_uncompressed_file_size_bytes)

  def refresh_collector(self, conf, pid, urlopen):
    self._lock.acquire()
    try:
      if conf is None:
        return
      method = conf.get_with_default(_SECTION, _METHOD, None)
      enabled = conf.get_with_default(_SECTION, _ENABLED, 'false')
      if enabled != 'true' or method is None:
        self.stop_collector(conf)
        return
      collector = self._collectors.get(conf.path, None)
      if collector is None:
        # If an existing collector already exists, do not start a new one. We do
        # not expect this to happen, but could if a previous collector hung and
        # the underlying role was restarted. The helper handles logging a
        # message.
        if not self._validate_no_existing_collector_for_role(conf):
          return
        LOG.info("Starting %s stacks collector for %s." % (method, conf.path))
        if method == 'jstack':
          collector = _JstackCollector()
        elif method == 'servlet':
          collector = _ServletCollector(urlopen)
        else:
          THROTTLED_LOG.error("Unknown collection method %s." % method)
          return
        self._collectors[conf.path] = collector
      if not collector.refresh(conf,
                               self._max_uncompressed_file_size_bytes,
                               pid):
        self.stop_collector(conf)
    finally:
      self._lock.release()

  def stop_collector(self, conf):
    self._lock.acquire()
    try:
      if conf is None:
        return
      collector = self._collectors.get(conf.path, None)
      if collector is None:
        return
      LOG.info("Stopping stacks collector for %s." % conf.path)
      del self._collectors[conf.path]
      collector.stop()
    finally:
      self._lock.release()

  def _validate_no_existing_collector_for_role(self, conf):
    rolename = conf.get_with_default(_SECTION, _ROLENAME, None)
    if rolename is None:
      THROTTLED_LOG.error("%s is missing required %s setting." % (conf.path,
                                                                  _ROLENAME))
      return False
    for path, c in self._collectors.iteritems():
      if c.get_role_name() == rolename:
        # It's not crazy to think this might happen during normal operation from
        # time to time, especially during a role restart. So we'll avoid warning
        # or error. If it does happen we expect that the next refresh will take
        # care of starting the new collector.
        THROTTLED_LOG.info("Collector '%s' already exists for role '%s'. Not "
                           "starting new collector." % (path, rolename))
        return False
    return True

