#!/usr/bin/env python
from __future__ import absolute_import
# (c) Copyright 2011 Cloudera, Inc. All rights reserved.
"""
Code for collecting the CPU utilization of a running process
"""
import time
import cmf.monitor.tasktracker.dataset as dataset
import logging
import psutil

LOG = logging.getLogger(__name__)

# See psutil/_pslinux.py:_status_map
_PROCESS_STATE = {
  psutil.STATUS_RUNNING: "RUNNING",
  psutil.STATUS_SLEEPING: "SLEEPING",
  psutil.STATUS_DISK_SLEEP: "WAITING_FOR_DISK",
  psutil.STATUS_ZOMBIE: "ZOMBIED",
  psutil.STATUS_STOPPED: "STOPPED",
  psutil.STATUS_WAKING: "PAGING",
  psutil.STATUS_TRACING_STOP: "TRACED_STOPPED",
  psutil.STATUS_DEAD: "DEAD"
}

_PROCESS_STATE_DEFAULT = "SLEEPING"

class MockAttemptUpdater(dataset.Updater):
  """ Mock attempt updater """
  def update(self, data):
    data.process_state.setvalue("RUNNING")
    data.cpu_user.setvalue(1.0)
    data.cpu_system.setvalue(1.0)
    data.total_cpu_user.setvalue(1.0)
    data.total_cpu_system.setvalue(1.0)
    data.mem_virtual.setvalue(1)
    data.mem_rss.setvalue(1)
    data.disk_read.setvalue(1.0)
    data.disk_write.setvalue(1.0)

class AttemptUpdater(dataset.Updater):
  """ Updates CPU utilization for attempts """

  def __init__(self, attempt_id, tasktracker):
    super(AttemptUpdater, self).__init__(attempt_id, tasktracker)
    self._attempt_id = attempt_id
    # Last process.get_cpu_times() and time thereof.
    self._last_cpu_times = None
    self._last_cpu_times_check = None
    # psutil.Process object
    self._process = None

  def update(self, data):
    """ Reports the percent CPU user and system for a particular process """
    pid = data.pid.getvalue()

    # We can't do anything until we have a valid pid
    if pid <= 0:
      return

    if self._process is None:
      try:
        self._process = psutil.Process(int(pid))
      except psutil.NoSuchProcess:
        return

    try:
      now = time.time()
      cpu_times = self._process.cpu_times()

      data.total_cpu_user.setvalue(cpu_times.user)
      data.total_cpu_system.setvalue(cpu_times.system)
      data.process_state.setvalue(_PROCESS_STATE.get(self._process.status(), _PROCESS_STATE_DEFAULT))

      if self._last_cpu_times is not None:
        # Compute instantaneous percentages
        data.cpu_user.setvalue( (cpu_times.user - self._last_cpu_times.user) / (now - self._last_cpu_times_check) )
        data.cpu_system.setvalue( (cpu_times.system - self._last_cpu_times.system) / (now - self._last_cpu_times_check) )
      self._last_cpu_times = cpu_times
      self._last_cpu_times_check = now

      rss, virtual = self._process.memory_info()

      # Memory
      data.mem_virtual.setvalue(virtual)
      data.mem_rss.setvalue(rss)

      # IO
      try:
        io_counters = self._process.io_counters()
        data.disk_read.setvalue(io_counters.read_count)
        data.disk_write.setvalue(io_counters.write_count)
      except:
        # Some kernels don't support this; fail gracefully.
        pass

    except psutil.NoSuchProcess:
        # Process finished, nothing more to do.
        pass
