#!/usr/bin/env python
from __future__ import absolute_import
# (c) Copyright 2010 Cloudera, Inc. All rights reserved.
"""
Module for storing and retrieving metrics which are collected by the agent
"""
import time

import cmf.monitor.tasktracker.value as value

# Elements tagged with this ID will not be exported in the JSON schema
ELEMENT_NOT_EXPORTED = 987654321

class Element(object):
  """ All datasets are compromised of an array of elements """

  def __init__(self, elid, name, desc, units, typeclass, typename, choices):
    self.elid = elid
    self.name = name
    self.desc = desc
    self.units = units
    self.typename = typename
    self.typeclass = typeclass
    self.choices = choices

  def __str__(self):
    return "id:%s name:%s description:%s units:%s type:%s" % \
        (self.elid, self.name, self.desc, self.units, self.typename)

class LongElement(Element):
  """ An element that holds a numeric value w/o time-series """

  def __init__(self, elid, name, desc, units=None):
    Element.__init__(self, elid, name, desc, units, value.LongValue, "long", None)

class DoubleElement(Element):
  """ An element that holds a numeric value w/o time-series """

  def __init__(self, elid, name, desc, units=None):
    Element.__init__(self, elid, name, desc, units, value.DoubleValue, "double", None)

class StringElement(Element):
  """ An element that holds a string value """

  def __init__(self, elid, name, desc, units=None):
    Element.__init__(self, elid, name, desc, units, value.StringValue, "string", None)

class GaugeElement(Element):
  """ An element that holds a gauge value """

  def __init__(self, elid, name, desc, units=None):
    Element.__init__(self, elid, name, desc, units, value.GaugeValue, "gauge", None)

class CounterElement(Element):
  """ An element that holds a counter value """

  def __init__(self, elid, name, desc, units=None):
    Element.__init__(self, elid, name, desc, units, value.CounterValue, "counter", None)

class StateElement(Element):
  """ An element that holds a fixed-state value (enum) """

  def __init__(self, elid, name, desc, choices, units=None):
    Element.__init__(self, elid, name, desc, units, value.StateValue, "state", choices)


class Dataset(object):
  """
  A dataset is comprised of a number of elements of potentially mixed type.
  Information about the elements in the dataset (schema) is defined in the
  class variable 'members'. The updaters are defined in the input variable
  'updaters'. A subclass is likely to define both a class specific list of
  members as well as a list of updaters to use. Additionally, the metrics
  module contains code to update the 'members' variable of various Dataset
  subclasses from a schema file. See metrics.import_schema_from_string.
  """
  def __init__(self, name, tasktracker, updaters, collect_interval,
               report_interval):
    self.timestamp = time.time()
    self._name = name
    self._updaters = [updater(name, tasktracker) for updater in updaters]
    # See the class comment above for how self.members is populated.
    for member in self.members:
      m = member.typeclass(member)
      m.set_intervals(collect_interval, report_interval)
      setattr(self, member.name, m)

  def update(self):
    """
    Triggers the dataset to run through all updaters to collect values. This
    also updates the timestamp associated with the dataset.
    """
    for updater in self._updaters:
      updater.update(self)
    self.timestamp = time.time()

  def exported_elements(self):
    """ Returns a generator to walk all the exported members for the dataset """
    for member in self.members:
      if member.elid == ELEMENT_NOT_EXPORTED:
        continue
      yield member

  def dump(self):
    """ Dumps all the key-value pairs in the dataset for debugging """
    info = ""
    for member in self.members:
      info += '%25s (%04s) = %s %s "%s"\n' % (member.name,
                                            member.elid,
                                            getattr(self, member.name),
                                            member.units,
                                            member.desc)
    return info

  @classmethod
  def schema(cls):
    """ Dumps the schema of the dataset for debugging """
    info = ""
    for member in cls.members:
      info += str(member)
    return info

  def __str__(self):
    return "%s(%s)" % (type(self), self.dump())


class Updater(object):
  """ Abstract class which defines the interface for dataset updaters """

  def __init__(self, name, tasktracker):
    self._name = name
    self._tasktracker = tasktracker

  def update(self, data):
    """ All updaters needs to have an update method """
    raise NotImplementedError("Updater needs to implement an update method")
