#!/usr/bin/env python
from __future__ import absolute_import
# (c) Copyright 2010 Cloudera, Inc. All rights reserved.
"""
Module for defining all the metric collected in the agent
"""
import cmf.monitor.tasktracker.cpu as cpu
import cmf.monitor.tasktracker.dataset as dataset
import logging
import os
import simplejson
import cmf.monitor.tasktracker.tasktracker as tasktracker

LOG = logging.getLogger(__name__)
_METRIC_SCHEMA = 'schema.json'
_METRIC_SCHEMA_PATH = os.path.join(os.path.dirname(__file__), _METRIC_SCHEMA)
METRICS = set(['spilled_records', 'process_state', 'reduce_input_records',
               'cpu_user', 'disk_write', 'hdfs_read', 'shuffle_finish_time',
               'attempt_status', 'cpu_system', 'task_type', 'map_output_records',
               'reduce_output_records', 'total_cpu_system', 'disk_read',
               'map_input_records', 'progress', 'map_output_bytes', 'map_input_bytes',
               'sort_finish_time', 'start_time', 'combine_input_records',
               'reduce_input_groups', 'tt_name', 'phase', 'total_cpu_user',
               'reduce_shuffle_bytes', 'combine_output_records', 'finish_time',
               'hdfs_write', 'mem_virtual', 'mem_rss'])

class MissingTypeError(Exception):
  """ Exception thrown when a metric is missing a 'type' """
  pass

class InvalidTypeError(Exception):
  """ Exception thrown when a metric has an illegal value """
  pass

class MissingContextError(Exception):
  """ Exception thrown when a metric is missing a 'context' """
  pass

class InvalidContextError(Exception):
  """ Exception thrown when a metric has an illegal context """
  pass

class MissingSourceError(Exception):
  """ Exception thrown when a metric is missing a 'source' """
  pass

class MissingIdError(Exception):
  """ Exception thrown if a metric is missing an id """
  pass

class DuplicateIdError(Exception):
  """ Exception thrown if there are duplicate metric ids """
  pass

class MissingChoicesError(Exception):
  """ Exception thrown if there is a missing choice """
  pass

class AttemptDataset(dataset.Dataset):
  """ Class that defines the attempt dataset. Members are imported from the schema file. """
  members = [dataset.StringElement(dataset.ELEMENT_NOT_EXPORTED, "job_id",
                                   "The job id for this attempt"),
             dataset.StringElement(dataset.ELEMENT_NOT_EXPORTED, "task_id",
                                   "The task id for this attempt"),
             dataset.LongElement(dataset.ELEMENT_NOT_EXPORTED, "pid",
                                   "pid of the attempt"),
             ]

  def __init__(self, name, tracker, mock, collect_interval, report_interval):
    if mock:
      if tracker.generate_defaults():
        updaters = [tasktracker.MockDefaultsAttemptUpdater,
                    cpu.MockAttemptUpdater]
      else:
        updaters = [tasktracker.MockAttemptUpdater,
                    cpu.MockAttemptUpdater]
    else:
      updaters = [tasktracker.AttemptUpdater,
                  cpu.AttemptUpdater]
    super(AttemptDataset, self).__init__(name, tracker, updaters,
                                         collect_interval,
                                         report_interval)


METRIC_ID = 'id'
METRIC_TYPE = 'type'
METRIC_TYPE_ELEMENTS = { 'string': dataset.StringElement,
                         'long': dataset.LongElement,
                         'double': dataset.DoubleElement,
                         'timestamp': dataset.LongElement,
                         'state': dataset.StateElement,
                         'enum': dataset.StateElement,
                         'gauge': dataset.GaugeElement,
                         'counter': dataset.CounterElement }
METRIC_CONTEXT = 'context'
METRIC_CONTEXT_ATTEMPT = 'attempt'
METRIC_DESCRIPTION = 'desc'
METRIC_UNITS = 'units'
METRIC_CHOICES = 'choices'
METRIC_SOURCE = 'source'
METRIC_SOURCE_AGENT = 'agent'
METRIC_DEPRECATED = 'deprecated'

def _element_type(name, info):
  """ Import a metric definition into the appropriate dataset """
  metric_type = info.get(METRIC_TYPE, None)
  if metric_type is None:
    raise MissingTypeError("Missing %s for metric=%s" % (METRIC_TYPE, name))
  eltype = METRIC_TYPE_ELEMENTS.get(metric_type, None)
  if eltype is None:
    raise InvalidTypeError("Illegal type=%s for metric=%s" % (metric_type, name))
  return eltype

def _element_datasets(name, info):
  """Returns a list of the datasets to add the metric to"""
  contexts = info.get(METRIC_CONTEXT, None)
  if contexts is None:
    raise MissingContextError("Missing %s for metric %s" % (METRIC_CONTEXT, name))
  ret = []
  if METRIC_CONTEXT_ATTEMPT in contexts:
    ret.append(AttemptDataset)
  if len(ret) > 0:
    return ret
  raise InvalidContextError("Illegal contexts for metric=%s" % (name,))

def import_schema_from_string(schema_string, metrics=METRICS):
  """ Import metric schema defintion from a string """
  schema = simplejson.loads(schema_string)
  id_map = {}
  for name in metrics:
    info = schema[name]
    element_datasets = _element_datasets(name, info)
    element_type = _element_type(name, info)
    element_id = info.get(METRIC_ID, None)
    if element_id is None:
      raise MissingIdError("Metric %s is missing an %s" % (name, METRIC_ID))
    name_with_id = id_map.get(element_id, None)
    if name_with_id is not None:
      raise DuplicateIdError("Duplicate id=%s found for both %s and %s",
                        element_id, name, name_with_id)
    id_map[element_id] = name
    description = info.get(METRIC_DESCRIPTION, None)
    units = info.get(METRIC_UNITS, None)
    if element_type == dataset.StateElement:
      choices = info.get(METRIC_CHOICES, None)
      if choices is None:
        raise MissingChoicesError("State metric %s is missing %s" % (name, METRIC_CHOICES))
      element = element_type(element_id, name, description, choices, units)
    else:
      element = element_type(element_id, name, description, units)
    for element_dataset in element_datasets:
      element_dataset.members.append(element)

def import_schema_from_file(schema_file):
  """ Import metric schema definition from specified file """
  LOG.info("Importing tasktracker metric schema from file %s" %
           (_METRIC_SCHEMA_PATH,))
  return import_schema_from_string(open(schema_file, 'r').read())

def import_schema():
  """ Imports the metric schema definition from the default location """
  return import_schema_from_file(_METRIC_SCHEMA_PATH)
