# Copyright (c) 2011-2012 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import
import threading
import time
import logging

from cmf.monitor import constants, schema
from cmf.monitor.abstract_mapreduce_monitor import AbstractMapReduceMonitor
from cmf.monitor.abstract_monitor import AbstractMonitor
from cmf.monitor.base import DebugStatusReport
from cmf.monitor.cgroup_metrics_utils import add_cgroup_metrics
from cmf.monitor.firehoses import Firehoses
from cmf.monitor.json_metrics_utils import add_derived_jvm_metrics_metrics, add_metrics_metrics
from cmf.monitor.proc_metrics_utils import add_proc_metrics, add_with_descendants_proc_metrics
from cmf.monitor.role_update import RoleUpdate
from cmf.monitor.stoppable_thread import StoppableThread
from cmf.monitor.tasktracker.json_parser import JsonParser
from cmf.monitor.tasktracker.metrics import import_schema, AttemptDataset
from cmf.monitor.tasktracker.tasktracker import TaskTracker
from cmf.monitor.tasktracker.tasktracker import UNKNOWN_ENTITY
from cmf.throttling_logger import ThrottlingLogger

LOG = logging.getLogger("TaskTrackerMonitor")
THROTTLED_LOG = ThrottlingLogger(LOG, constants.DEFAULT_MONITOR_LOG_RATE)
_ROLETYPE = "TASKTRACKER"
_SERVICETYPE = "MAPREDUCE"


# Section and key names in cloudera-monitor.properties
# Since CM 3.7
_SECTION = 'tasktracker'
_ROLE_NAME = 'role_name'
_SERVICE_NAME = 'service_name'
_COLLECT_INTERVAL = 'collect_interval'
_TASKTRACKER_INSTRUMENTATION_URL = 'tasktracker_instrumentation_url'
_HOST = 'host'
_TASKTRACKER_HTTP_PORT = 'task_tracker_http_port'
_TASKTRACKER_LOCAL_DIRS = 'task_tracker_mapred_local_dir'
# Added in CM 4.0
_SERVICE_VERSION = 'service_version'
# Added in CM 4.1
_LOG_DIR = 'log_dir'
# Added in CM 5.1
_USE_SSL = 'use_ssl'
# Added in CM 5.4
_HEAP_DUMP_DIR = 'oom_heap_dump_dir'

# This is a hack to avoid having to add a bunch of temporary support plumbing
# while the ArchiveRecord concept is removed. Those records care about the
# reporting interval, while their replacements will not. Instead of plumbing
# through the reporting interval -- and having to deal with adjustments to
# it -- we'll just hardcode a value for use in the various Datasets and records
# for now. Once we've moved away from their use, this will be removed.
_HACK_HARDCODED_REPORT_INTERVAL = 60.0

# This is how frequently we poll for attempt data in seconds.
_ATTEMPT_POLLING_INTERVAL = 5.0


def init_tasktracker_monitor():
  import_schema()


class TaskTrackerMonitor(AbstractMapReduceMonitor):

  def __init__(self, daemon, registry, cgroup_root=None, cgroup_name=None,\
               attempt_monitor=None):
    AbstractMapReduceMonitor.__init__(self,
                                      daemon,
                                      registry,
                                      cgroup_root,
                                      cgroup_name,
                                      _SECTION,
                                      _TASKTRACKER_HTTP_PORT,
                                      _ROLETYPE)
    self._last_conf_path = None
    self._metrics_tuple = (None, None, None)
    self._last_reported_timestamp = None
    self._last_collection_ended = None
    self._last_collection_length = None
    self._tasktracker = None
    self._role_dir_monitors = [ ]
    self._metric_accessors = daemon.metric_accessors

    # When testing we will manually call the attempt collection function. We
    # therefore disable the attempt monitor by putting in a dummy monitor. If
    # we don't do this, we can get occasional and difficult to diagnose test
    # failures.
    self._attempt_monitor = attempt_monitor

    # This is a map of attempt ID -> AttemptDataset. This is only ever accessed
    # from collection thread. It holds information on all active attempts that
    # we are tracking. Once an attempt completes, it is removed from this map.
    # Note that an attempt that runs for longer than our reporting period is in
    # both this map and the _attempts_to_report map.
    self._attempt_datasets = {}

    # This is a map of attempt ID -> AttemptDataset. This is populated from the
    # collection thread and then is read and cleared from the reporting thread.
    # Access to this is only permitted while holding the _report_lock. Note that
    # an attempt that runs for longer than our reporting period is in both this
    # map and the _attempts_to_report map.
    self._attempts_to_report = {}
    self._report_lock = threading.Lock()

    # Config data
    self._name = None
    self._service_name = None
    self._service_version = None
    self._collect_interval = None
    self._instrumentation_url = None
    self._task_tracker_local_dirs = [ ]
    self._log_dir = None
    self._previous_metrics = {}
    self._previous_cpu_user = None
    self._previous_cpu_system = None

  def __str__(self):
    return "TaskTrackerMonitor for %s" % (self._name,)

  def _reload_config(self, confparse):
    """Returns True/False for success"""
    try:
      # Extraction
      name = confparse.get(_SECTION, _ROLE_NAME)
      service = confparse.get(_SECTION, _SERVICE_NAME)
      service_version = \
          confparse.getint_with_default(_SECTION, _SERVICE_VERSION, \
                                        constants.SERVICE_VERSION_CDH5)
      log_dir = confparse.get_with_default(_SECTION, _LOG_DIR, None)
      heap_dump_dir = confparse.get_with_default(_SECTION, _HEAP_DUMP_DIR, None)
      collect_interval = confparse.getfloat(_SECTION, _COLLECT_INTERVAL)
      instrumentation_url = confparse.get(_SECTION,
                                          _TASKTRACKER_INSTRUMENTATION_URL)
      web_host = confparse.get(_SECTION, _HOST)
      web_port = confparse.get(_SECTION, _TASKTRACKER_HTTP_PORT)
      task_tracker_local_dirs = \
        confparse.getlist(_SECTION, _TASKTRACKER_LOCAL_DIRS)
      use_ssl = confparse.getboolean_with_default(_SECTION, _USE_SSL, False)
    except ValueError, ex:
      self.record_error(ex)
      LOG.error("Value error in %s: %s" % (confparse.path, ex))
      return False

    metrics_url = self._compute_metrics_url(web_host, web_port, use_ssl)
    if metrics_url is None:
      return False

    # Assign
    self._name = name
    self._service_name = service
    self._service_version = service_version
    self._log_dir = log_dir
    self._collect_interval = collect_interval
    self._metrics_url = metrics_url
    self._task_tracker_local_dirs = task_tracker_local_dirs

    if self._instrumentation_url != instrumentation_url:
      self._instrumentation_url = instrumentation_url
      self._tasktracker = TaskTracker(instrumentation_url.rstrip('/'),
                                      self.get_task_metrics_timeout_seconds())

    self._role_dir_monitors = []
    if self._log_dir is not None:
      self.create_dir_monitor(self._log_dir, self._role_dir_monitors)
    if heap_dump_dir is not None:
      self.create_dir_monitor(heap_dump_dir, self._role_dir_monitors)
    for directory in self._task_tracker_local_dirs:
      self.create_dir_monitor(directory, self._role_dir_monitors)

    return True

  def ensure_collecting(self):
    AbstractMonitor.ensure_collecting(self)

    if self._attempt_monitor is None:
      self._attempt_monitor = StoppableThread("TaskTrackerAttemptMonitor",
                                              _ATTEMPT_POLLING_INTERVAL,
                                              self.collect_attempt_metrics)
      self._attempt_monitor.start()

  def stop(self):
    AbstractMonitor.stop(self)

    if self._attempt_monitor is not None:
      self._attempt_monitor.should_stop()
      self._attempt_monitor = None

  def collect(self):
    """Overrides AbstractMonitor. Collect TT metrics. Invoked asynchronously."""
    start = time.time()

    # Reload config if necessary
    confparse = self.conf_parser
    if self._last_conf_path != confparse.path:
      if not self._reload_config(confparse):
        return
      self._last_conf_path = confparse.path

    self._collect_hadoop_metrics()

    # Collect directory information
    for dir_monitor in self._role_dir_monitors:
      dir_monitor.collect()

    self._last_collection_ended = time.time()
    self._last_collection_length = self._last_collection_ended - start

  def collect_attempt_metrics(self, mock_tasktracker=None):
    # If our pid value is None, it means that the associated process is not
    # running. If this is the case, we skip metric collection.
    if self._pid is None:
      return

    LOG.debug("<<<<< Updating the attempts dataset >>>>>")

    mock = False
    tasktracker = None
    if mock_tasktracker is not None:
      mock = True
      tasktracker = mock_tasktracker
    else:
      if self._tasktracker is None:
        return
      tasktracker = self._tasktracker

    current_attempts = tasktracker.attempts()
    previous_attempts = self._attempt_datasets.keys()

    # Evict from our cache if it's not in the TT anymore
    for attempt_id in previous_attempts:
      if not attempt_id in current_attempts:
        LOG.debug("--- Removing old attempt %s from cache." % (attempt_id,))
        attempt = self._attempt_datasets[attempt_id]
        if attempt is not None and not self._is_attempt_finished(attempt):
          LOG.info("About to remove %s. Assumed it succeeded." % (attempt_id,))
          attempt.attempt_status.setvalue("SUCCEEDED")
          self._report_attempt(attempt_id, attempt)
        del self._attempt_datasets[attempt_id]

    # Update attempts
    for attempt_id in current_attempts:
      if not attempt_id in self._attempt_datasets:
        LOG.debug("+++ Monitoring new attempt %s." % (attempt_id,))
        attempt = AttemptDataset(attempt_id,
                                 tasktracker,
                                 mock,
                                 _ATTEMPT_POLLING_INTERVAL,
                                 _HACK_HARDCODED_REPORT_INTERVAL)
        self._attempt_datasets[attempt_id] = attempt
      else:
        attempt = self._attempt_datasets[attempt_id]

      # A cached attempt could be None if it's finished.
      if attempt is not None:
        attempt.update()
        # We should only send updates if we know what job we are talking
        # about. Otherwise the update is useless.
        if attempt.job_id.getvalue() != UNKNOWN_ENTITY and  \
           attempt.task_id.getvalue() != UNKNOWN_ENTITY:
          self._report_attempt(attempt_id, attempt)

        # Disable future updates on completed attempts.
        if self._is_attempt_finished(attempt):
          LOG.debug("%s is no longer running" % (attempt_id,))
          del self._attempt_datasets[attempt_id]

  def _is_attempt_finished(self, attempt):
    # The *_UNCLEAN stages are, in theory, supposed to be purely intermediate
    # phases, which always progress to KILLED or FAILED. However, it seems that,
    # for some reason (possibly due to bugs in the TT instrumentation), there is
    # no point at which the TT instrumentation gets a callback after the tasks
    # have progressed from an unclean state to a purely finished state. For that
    # reason, we'll be a little proactive in making the assumption that if it's
    # unclean, it's basically finished.
    return attempt.attempt_status.last_state_is("SUCCEEDED") or \
           attempt.attempt_status.last_state_is("FAILED") or \
           attempt.attempt_status.last_state_is("FAILED_UNCLEAN") or \
           attempt.attempt_status.last_state_is("KILLED") or \
           attempt.attempt_status.last_state_is("KILLED_UNCLEAN")

  def _collect_hadoop_metrics(self):
    # If our pid value is None, it means that the associated process is not
    # running. If this is the case, we skip metric collection.
    if self._pid is None:
      self._metrics_tuple = (None, None, None)
      return

    LOG.debug("Fetching TaskTracker metrics at '%s'" % (self._metrics_url,))
    self._metrics_tuple = self.collect_metrics_from_url(self._metrics_url)

  def parse_metrics(self, json):
    return JsonParser(json, self._service_version)

  def _report_attempt(self, attempt_id, attempt):
    try:
      self._report_lock.acquire()
      self._attempts_to_report[attempt_id] = attempt
    finally:
      self._report_lock.release()

  def get_attempt_updates(self):
    try:
      self._report_lock.acquire()

      if self._attempts_to_report is None:
        return []

      updates = []
      for attempt_id, dataset in self._attempts_to_report.iteritems():
        context = dict(job_id=dataset.job_id.getvalue(),
                       task_id=dataset.task_id.getvalue(),
                       task_attempt_id=attempt_id,
                       service_name=self._service_name)
        metrics = []
        for element in dataset.exported_elements():
          value = getattr(dataset, element.name).getvalue()
          if value is None:
            continue
          if element.typename == 'state':
            if len(value) > 0:
              metrics.append(dict(id=element.elid, value=value[-1]))
          elif self._sendUpdateForElement(element):
            metrics.append(dict(id=element.elid,
                                value=Firehoses.normalize_metric_value(value)))
        updates.append(dict(ts_secs=long(dataset.timestamp),
                            context=context,
                            metrics=metrics))

      # Reset dataset attempts; monitoring thread will re-populate it
      # for next time.
      self._attempts_to_report = {}
      return updates
    finally:
      self._report_lock.release()

  def _sendUpdateForElement(self, element):
    """Determines if we should send this metric to firehose"""
    # CDH4 onwards, we poll JT for cpu and memory metrics
    if element.elid in (schema.TOTAL_CPU_USER, schema.MEM_RSS):
      return False
    return True

  def get_role_update(self):
    """Returns a RoleUpdate to be sent to the firehose"""
    update = RoleUpdate(name=self._name,
                        roletype=_ROLETYPE,
                        servicetype=_SERVICETYPE,
                        service_name=self._service_name)

    if self.add_metrics_from_webserver(self._metrics_tuple,
                                       self._last_reported_timestamp,
                                       update):
      self._last_reported_timestamp = update.timestamp

    add_proc_metrics(update, self._pid, LOG)
    self._previous_metrics, self._previous_cpu_user, self._previous_cpu_system =\
        add_with_descendants_proc_metrics(update,
                                          self._pid,
                                          self._previous_metrics,
                                          self._previous_cpu_user,
                                          self._previous_cpu_system,
                                          LOG)
    add_cgroup_metrics(update, self._cgroup_root, self._cgroup_name)

    for dir_monitor in self._role_dir_monitors:
        update.add_dir_update(dir_monitor.get_role_dir_update())

    self.add_unexpected_exit_metrics(update)

    return update

  def add_webserver_metrics_to_role_update(self, update, metrics):
    if self._service_version == constants.SERVICE_VERSION_CDH5:
      self._add_webserver_metrics_to_role_update_cdh5(update, metrics)
    else:
      THROTTLED_LOG.warn("Unknown tasktracker version: %s" %
          (self._service_version))

  def _add_webserver_metrics_to_role_update_cdh5(self, update, metrics):
    add_derived_jvm_metrics_metrics(update, metrics)
    add_metrics_metrics(update, metrics, self._metric_accessors, _ROLETYPE, 'cdh5')

  def get_debug_status_report(self):
    return DebugStatusReport(config_items=self.conf_parser.items(_SECTION),
                             errors=(self.last_error,),
                             last_collection_ended=self._last_collection_ended,
                             last_collection_length=self._last_collection_length)

  @property
  def instrumentation_url(self):
    return self._instrumentation_url

  @property
  def task_tracker_local_dirs(self):
    return self._task_tracker_local_dirs
