#!/usr/bin/env python
from __future__ import absolute_import
# (c) Copyright 2010-2011 Cloudera, Inc. All rights reserved.
"""
Class for collecting data from a TaskTracker
"""
import sys
import time
import simplejson
import cmf.monitor.tasktracker.dataset as dataset
import logging
from cmf.util.url import urlopen_with_timeout

LOG = logging.getLogger(__name__)

ATTEMPT_PID = 'pid'
ATTEMPT_JOBID = 'job_id'
ATTEMPT_TASKID = 'task_id'
ATTEMPT_STARTTIME = 'start_time'
ATTEMPT_FINISHTIME = 'finish_time'
ATTEMPT_TT_NAME = 'tt_name'
ATTEMPT_TASKTYPE = 'task_type'

# Counters - TODO: get definitive list
ATTEMPT_HDFSREAD = 'HDFS_BYTES_READ'
ATTEMPT_HDFSWRITE = 'HDFS_BYTES_WRITTEN'
ATTEMPT_SPILLED_RECORDS = 'SPILLED_RECORDS'
ATTEMPT_MAP_INPUT_RECORDS = 'MAP_INPUT_RECORDS'
ATTEMPT_MAP_OUTPUT_RECORDS = 'MAP_OUTPUT_RECORDS'
ATTEMPT_COMBINE_INPUT_RECORDS = 'COMBINE_INPUT_RECORDS'
ATTEMPT_COMBINE_OUTPUT_RECORDS = 'COMBINE_OUTPUT_RECORDS'
ATTEMPT_MAP_INPUT_BYTES = 'MAP_INPUT_BYTES'
ATTEMPT_MAP_OUTPUT_BYTES = 'MAP_OUTPUT_BYTES'

ATTEMPT_REDUCE_INPUT_BYTES = 'REDUCE_INPUT_BYTES'
ATTEMPT_REDUCE_OUTPUT_BYTES = 'REDUCE_OUTPUT_BYTES'
ATTEMPT_REDUCE_INPUT_RECORDS = 'REDUCE_INPUT_RECORDS'
ATTEMPT_REDUCE_OUTPUT_RECORDS = 'REDUCE_OUTPUT_RECORDS'
ATTEMPT_REDUCE_SHUFFLE_BYTES = 'REDUCE_SHUFFLE_BYTES'
ATTEMPT_REDUCE_INPUT_GROUPS = 'REDUCE_INPUT_GROUPS'

ATTEMPT_REDUCE_SHUFFLE_FINISH_TIME = 'REDUCE_SHUFFLE_FINISH_TIME'
ATTEMPT_REDUCE_SORT_FINISH_TIME = 'REDUCE_SORT_FINISH_TIME'

# These constants must be kept in sync with the corresponding constants
# and types in activity monitor.
ATTEMPT_STATE = 'state'
ATTEMPT_STATE_DEFAULT = 'CMF_UNKNOWN'
ATTEMPT_PHASE = 'phase'
ATTEMPT_PHASE_DEFAULT = 'STARTING'
ATTEMPT_PROGRESS = 'progress'
UNKNOWN_ENTITY = '__CMF_UNKNOWN_ENTITY__'

class TaskTracker(object):
  """ Class for polling the TaskTracker """

  def __init__(self, url_root, timeout):
    self.url_root = url_root
    self._next_log_time = 0
    self._timeout = timeout

  def _fetch_tt_data(self, path):
    """ Fetches the task attempt data from the TaskTracker """
    # TODO: handle failure correctly...
    full_path = self.url_root + path
    try:
      data = urlopen_with_timeout(full_path, timeout=self._timeout).read()
    except IOError, ioe:
      now = time.time()
      if now >= self._next_log_time:
        self._next_log_time = now + 30 * 60     # Quiet for 30 minutes
        LOG.error("TaskTracker at %s is not responding: %s." %
                  (self.url_root, ioe))
        LOG.info("Further attempts to connect will be made, but no repeat " +
                 "error messages logged until %s." %
                 (time.ctime(self._next_log_time),))
      return None
    try:
      return simplejson.loads(data)
    except ValueError:
      if len(data) != 0:
        LOG.error("Invalid JSON at (%s): '%s'" % (full_path, data))
      return None

  def attempt_info(self, attempt_id):
    """ Fetches information for a specific task attempt """
    return self._fetch_tt_data("/attempt/%s" % (attempt_id, )) or {}

  def attempts(self):
    """ Fetch the array of attempts associated with running jobs """
    return self._fetch_tt_data("/attempt/") or []

  def status(self):
    """ Fetches overall status information e.g. slot information """
    return self._fetch_tt_data("/") or {}

class MockTaskTracker(object):
  """ Simple mock of the TT """

  def __init__(self, job_name, mock_attempt_offset, num_attempts, num_waves, wave_time, generate_defaults):
    self._started = long(time.time())
    self._job_name = job_name
    self._offset = int(mock_attempt_offset)
    self._num_attempts = int(num_attempts)
    self._num_waves = int(num_waves)
    self._wave_time = int(wave_time)
    self._generate_defaults = generate_defaults

  def _get_wave_number(self):
    return int(  1.0 * (long(time.time()) - self._started) / self._wave_time)

  def attempts(self):
    """ Returns an array of attempt ids that are running """
    current_wave_number = self._get_wave_number()
    if current_wave_number == (self._num_waves + 1):
      return []
    elif current_wave_number > self._num_waves:
      sys.exit(0)

    wave_offset = self._offset  + (current_wave_number * self._num_attempts)
    return ["%s_attempt_%d" % (self._job_name, i) for i in range(wave_offset, wave_offset + self._num_attempts)]

  def status(self):
    return None

  def generate_defaults(self):
    return self._generate_defaults

class MockAttemptUpdater(dataset.Updater):
  """ Mock attempt updater for the TT """

  def __init__(self, name, tasktracker):
    super(MockAttemptUpdater, self).__init__(name, tasktracker)
    self._started = long(time.time()) * 1000

  def update(self, data):
    data.job_id.setvalue('mock_job')
    data.task_id.setvalue("mock_task_%s" % (self._name,))
    data.task_type.setvalue("MAP")

    data.start_time.setvalue(self._started)
    data.finish_time.setvalue(0)
    data.pid.setvalue(1)
    data.hdfs_read.setvalue(100)
    data.hdfs_write.setvalue(100)

    data.spilled_records.setvalue(100)
    data.map_input_records.setvalue(100)
    data.map_output_records.setvalue(100)
    data.combine_input_records.setvalue(100)
    data.combine_output_records.setvalue(100)
    data.map_input_bytes.setvalue(100)
    data.map_output_bytes.setvalue(100)

    data.reduce_input_records.setvalue(100)
    data.reduce_output_records.setvalue(100)
    data.reduce_shuffle_bytes.setvalue(100)
    data.reduce_input_groups.setvalue(100)
    data.shuffle_finish_time.setvalue(100)
    data.sort_finish_time.setvalue(100)

    data.attempt_status.setvalue(ATTEMPT_STATE_DEFAULT)
    data.phase.setvalue(ATTEMPT_PHASE_DEFAULT)
    data.progress.setvalue(50)

class MockDefaultsAttemptUpdater(dataset.Updater):
  """ Mock attempt updater for the TT """

  def __init__(self, name, tasktracker):
    super(MockDefaultsAttemptUpdater, self).__init__(name, tasktracker)
    self._started = long(time.time()) * 1000

  def update(self, data):
    data.job_id.setvalue(UNKNOWN_ENTITY)
    data.task_id.setvalue(UNKNOWN_ENTITY)
    data.tt_name.setvalue(UNKNOWN_ENTITY)

    data.start_time.setvalue(-1)
    data.finish_time.setvalue(-1)
    data.pid.setvalue(-1)
    data.hdfs_read.setvalue(0)
    data.hdfs_write.setvalue(0)

    data.spilled_records.setvalue(0)
    data.map_input_records.setvalue(0)
    data.map_output_records.setvalue(0)
    data.combine_input_records.setvalue(0)
    data.combine_output_records.setvalue(0)
    data.map_input_bytes.setvalue(0)
    data.map_output_bytes.setvalue(0)

    data.reduce_input_records.setvalue(0)
    data.reduce_output_records.setvalue(0)
    data.reduce_shuffle_bytes.setvalue(0)
    data.reduce_input_groups.setvalue(0)
    data.shuffle_finish_time.setvalue(0)
    data.sort_finish_time.setvalue(0)

    data.attempt_status.setvalue(ATTEMPT_STATE_DEFAULT)
    data.phase.setvalue(ATTEMPT_PHASE_DEFAULT)
    data.progress.setvalue(0.0)


class AttemptUpdater(dataset.Updater):
  """ Class responsible for updating the attempt metrics from the TT """

  def update(self, data):
    LOG.debug("Updating information for attempt %s" % (self._name,))
    attempt_info = self._tasktracker.attempt_info(self._name)
    data.job_id.setvalue(attempt_info.get(ATTEMPT_JOBID, UNKNOWN_ENTITY))
    data.task_id.setvalue(attempt_info.get(ATTEMPT_TASKID, UNKNOWN_ENTITY))
    data.tt_name.setvalue(attempt_info.get(ATTEMPT_TT_NAME, UNKNOWN_ENTITY))

    tasktype = attempt_info.get(ATTEMPT_TASKTYPE)
    if tasktype is not None:
      data.task_type.setvalue(tasktype)

    data.start_time.setvalue(attempt_info.get(ATTEMPT_STARTTIME, -1))
    data.finish_time.setvalue(attempt_info.get(ATTEMPT_FINISHTIME, -1))
    data.pid.setvalue(attempt_info.get(ATTEMPT_PID, -1))
    data.hdfs_read.setvalue(attempt_info.get(ATTEMPT_HDFSREAD, 0))
    data.hdfs_write.setvalue(attempt_info.get(ATTEMPT_HDFSWRITE, 0))

    if tasktype == "MAP":
      data.map_input_records.setvalue(attempt_info.get(ATTEMPT_MAP_INPUT_RECORDS, 0))
      data.map_output_records.setvalue(attempt_info.get(ATTEMPT_MAP_OUTPUT_RECORDS, 0))
      data.map_input_bytes.setvalue(attempt_info.get(ATTEMPT_MAP_INPUT_BYTES, 0))
      data.map_output_bytes.setvalue(attempt_info.get(ATTEMPT_MAP_OUTPUT_BYTES, 0))

    if tasktype == "REDUCE":
      data.reduce_input_records.setvalue(attempt_info.get(ATTEMPT_REDUCE_INPUT_RECORDS, 0))
      data.reduce_output_records.setvalue(attempt_info.get(ATTEMPT_REDUCE_OUTPUT_RECORDS, 0))
      data.reduce_shuffle_bytes.setvalue(attempt_info.get(ATTEMPT_REDUCE_SHUFFLE_BYTES, 0))
      data.reduce_input_groups.setvalue(attempt_info.get(ATTEMPT_REDUCE_INPUT_GROUPS, 0))
      data.shuffle_finish_time.setvalue(attempt_info.get(ATTEMPT_REDUCE_SHUFFLE_FINISH_TIME, 0))
      data.sort_finish_time.setvalue(attempt_info.get(ATTEMPT_REDUCE_SORT_FINISH_TIME, 0))

    # These metrics are possible for both main types of tasks.
    if tasktype in ("MAP", "REDUCE"):
      data.spilled_records.setvalue(attempt_info.get(ATTEMPT_SPILLED_RECORDS, 0))
      data.combine_input_records.setvalue(attempt_info.get(ATTEMPT_COMBINE_INPUT_RECORDS, 0))
      data.combine_output_records.setvalue(attempt_info.get(ATTEMPT_COMBINE_OUTPUT_RECORDS, 0))

    data.attempt_status.setvalue(attempt_info.get(ATTEMPT_STATE, ATTEMPT_STATE_DEFAULT))
    data.phase.setvalue(attempt_info.get(ATTEMPT_PHASE, ATTEMPT_PHASE_DEFAULT))
    data.progress.setvalue(int(100 * float(attempt_info.get(ATTEMPT_PROGRESS, 0.0))))
