#!/usr/bin/env python
from __future__ import absolute_import
# (c) Copyright 2010 Cloudera, Inc. All rights reserved.
""" A simple pure python implementation of a time-series database """
import time
import math

_MAX32BIT = pow( 2.0, 32)
_MAX64BIT = pow( 2.0, 64)

class Bucket(object):
  """ 
  A bucket holds the min/avg/max/stdev/last/sum of all values stored in it 
  
  See http://www.cs.berkeley.edu/~mhoemmen/cs194/Tutorials/variance.pdf
  for details about how avg and stdev are calculated using a one-pass
  method which isn't prone to cancellation
  """

  def __init__(self):
    """ Constructor """
    self._min_value = None
    self._max_value = None
    self._last_value = None
    self._last_timestamp = None

    # Variable names chosen to match equations in paper above
    # The calculated average
    self._Mk = 0.0
    # Used for calculating the standard variance
    self._Qk = 0.0
    # The number of samples
    self._k = 0

  def __repr__(self):
    return "min(%s)\tavg(%s)\tmax(%s)\tstdev(%s)\tn(%s)\tlast(%s)" % \
     (self.min(), self.avg(), self.max(), self.stdev(), self.samples(),self.last())

  def copy(self):
    """ Returns a new copy of itself """
    new_copy = Bucket()
    new_copy._min_value = self._min_value
    new_copy._max_value = self._max_value
    new_copy._last_value = self._last_value
    new_copy._last_timestamp = self._last_timestamp
    new_copy._Mk = self._Mk
    new_copy._Qk = self._Qk
    new_copy._k = self._k
    return new_copy

  def store(self, value, timestamp):
    """ Method for placing a new value in a bucket """
    if self._min_value is None or value < self._min_value:
      self._min_value = float(value)

    if self._max_value is None or value > self._max_value:
      self._max_value = float(value)

    self._k += 1
    if self._k == 0:
      self._Mk = float(value)
      self._Qk = 0
    else:
      diff = value - self._Mk
      self._Qk += (self._k - 1) * diff * diff / self._k
      self._Mk += (diff/ self._k)
    
    self._last_value = value
    self._last_timestamp = timestamp

  def sum(self):
    """ Return the sum of all values stored in this bucket """
    return self._Mk * self._k

  def min(self):
    """ Return the minimum value for this bucket """
    return self._min_value

  def max(self):
    """ Return the maximum value for this bucket """
    return self._max_value

  def avg(self):
    """ Return the average for this bucket """
    return self._Mk
  
  def stdev(self):
    """ Return the standard deviation for this bucket """
    if self._k > 0:
      return math.sqrt(self._Qk/self._k)
    else:
      return None

  def samples(self):
    """ Return the number of samples in the bucket """
    return self._k

  def last(self):
    """ Returns the last value inserted into the bucket """
    return (self._last_timestamp, self._last_value)

  def clear(self):
    """ Resets all data in a bucket """
    self._k = 0
    self._min_value = None
    self._max_value = None
    self._last_value = None
    self._last_timestamp = None
    self._Mk = 0.0
    self._Qk = 0.0
