#!/usr/bin/env python
from __future__ import absolute_import
# (c) Copyright 2010 Cloudera, Inc. All rights reserved.
"""
Module for storing and retrieving values
"""
import cmf.monitor.tasktracker.timelydb as timelydb
import time
import logging

LOG = logging.getLogger(__name__)

_MAX32BIT = pow( 2.0, 32)
_MAX64BIT = pow( 2.0, 64)

STATE_IDX = 'state_idx'

class Value(object):
  """ Base class for storing and retrieving values """

  def __init__(self, element):
    self._value = None
    self._element = element

  def getvalue(self, clear_on_read=False):
    """ Returns the value """
    return self._value

  def setvalue(self, value, timestamp=None):
    """ Sets the value """
    self._value = value

  def set_intervals(self, collect_interval, report_interval):
    """
    Stores timing information. By default this does nothing, but value types
    that care about the timing information can override this.
    """
    return

class LongValue(Value):
  """ Numerical value without time-series processing """

  def __str__(self):
    return "long(%s)" % (self.getvalue(),)

  def setvalue(self, value):
    """ Set the value to be a long """
    self._value = long(value)

class DoubleValue(Value):
  """
  Numerical value without time-series processing

  floats in python can actually be double precision.
  """

  def __str__(self):
    return "double(%s)" % (self.getvalue(),)

  def setvalue(self, value):
    """ Set the value to be a double """
    self._value = float(value)


class StringValue(Value):
  """ Simple string value """

  def __str__(self):
    return "string(%s)" % (self.getvalue(),)


def stats_to_dict(stats, collect_interval, report_interval, last_sample):
  """ Converts stats to an avro dictionary """
  # TODO: fix this? Assumes time boundary on getvalue()
  (last_update, last_value) = stats.last()
  nsamples = stats.samples()
  end_ts_secs = long(time.time())
  start_ts_secs = end_ts_secs - report_interval
  step_secs = collect_interval
  if nsamples == 0:
    last_update = 0
    last_value = 0
    minimum = 0
    average = 0
    maximum = 0
    deviation = 0
  else:
    minimum = stats.min()
    average = stats.avg()
    maximum = stats.max()
    deviation = stats.stdev()
  return dict(start_ts_secs=long(start_ts_secs),
              end_ts_secs=long(end_ts_secs),
              last_update_ts_secs=long(last_update),
              step_secs=long(step_secs),
              last_value=float(last_value),
              minimum=float(minimum),
              average=float(average),
              maximum=float(maximum),
              deviation=float(deviation),
              nsamples=long(nsamples),
              last_sample=float(last_sample))

class TimelyValue(Value):
  """ Extends the Value class to support TimelyDb storage """

  def __init__(self, element, valuetype):
    Value.__init__(self, element)
    self._stats = timelydb.Bucket()
    self._valuetype = valuetype

  @property
  def stats(self):
    """ Read-only access to the stats """
    return self._stats

  def getvalue(self, clear_on_read=False):
    """ Returns a stats avro dict form """
    if self._stats.samples() == 0:
      return None
    dictobj = stats_to_dict(self.stats,
                            self._collect_interval,
                            self._report_interval,
                            self._last_sample)
    # Clear on read
    if clear_on_read:
      self.stats.clear()
    return dictobj

  def set_intervals(self, collect_interval, report_interval):
    self._collect_interval = collect_interval
    self._report_interval = report_interval

  def __str__(self):
    return "%s(%s)" % (self._valuetype,
                       self.getvalue(clear_on_read=False))

class GaugeValue(TimelyValue):
  """ Gauge time-series value """
  def __init__(self, element):
    TimelyValue.__init__(self, element, "gauge")

  def setvalue(self, value, timestamp=None):
    """ Write the value to a timelydb """
    if timestamp is None:
      timestamp = long(time.time())
    self._last_sample = value
    self.stats.store(value, timestamp)

class CounterValue(TimelyValue):
  """ Counter time-series value """

  def __init__(self, element):
    TimelyValue.__init__(self, element, "counter")
    self._last_value = None
    self._last_timestamp = None

  def setvalue(self, value, timestamp=None):
    """ Set the value for a counter keeping in mind counter wrap """
    if value is None:
      return

    if timestamp is None:
      timestamp = long(time.time())

    if self._last_value is None:
      self._last_value = value
      self._last_timestamp = timestamp
      return

    if timestamp <= self._last_timestamp:
      return

    diff = value - self._last_value
    if diff < 0:
      diff += _MAX32BIT
    if diff < 0:
      diff += _MAX64BIT - _MAX32BIT
    if diff >= 0:
      update_value = diff / (timestamp - self._last_timestamp)

    self.stats.store(update_value, timestamp)

    self._last_value = value
    self._last_sample = value
    self._last_timestamp = timestamp

class StateValue(Value):
  """ Value with fixed states (enum) """

  def __init__(self, element):
    Value.__init__(self, element)
    self._statechanges = []
    self._last_value = None

  def getvalue(self, clear_on_read=True):
    """ Return the state in avro form """
    # TODO: update schema to be a state change array...
    # for now, just send the last one. :(
    value = self._statechanges
    # Clear on read
    if clear_on_read:
      self._statechanges = []
    return value

  def last_state_is(self, state):
    """ Returns True is the last state matches the 'state' param """
    return self._last_value == state

  def setvalue(self, value):
    """ Check that value being set is valid for the given schema """
    self._last_value = value
    state_idx = self._element.choices.index(value)
    self._statechanges.append({'ts_secs' : long(time.time()),
                               STATE_IDX : long(state_idx)})

  def __str__(self):
    return "state(%s)" % (self.getvalue(clear_on_read=False),)
