# Copyright (c) 2011 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging
import threading
import time

LOG = logging.getLogger(__name__)

class WakeableThread(threading.Thread):
  """
  A thread that can be awoken to execute some action.
  
  Note that you are guaranteed at least one call to to the function after each
  call to wake, but there is no guarantee that you will get only one call.
  """
  def __init__(self, name, fn, *args, **kwargs):
    threading.Thread.__init__(self, name=name)
    self.setDaemon(True)
    self._fn = fn
    self._args = args
    self._kwargs = kwargs
    self._should_stop = False
    self._latches = []
    self._timeout = None
    self._cv = threading.Condition()

  def run(self):
    LOG.debug("Starting %s" % (self.getName(),))
    while not self._should_stop:
      self._cv.acquire()
      wait_time = self._timeout
      while len(self._latches) == 0:
        start = time.time()
        self._cv.wait(wait_time)
        if self._should_stop:
          break
        if wait_time is not None:
          wait_time -= time.time() - start
        # We want to break out of the loop even if we have no latches to signal
        # if we have a timeout. We detect this by keeping track of our total
        # wait time and seeing if this is over the timeout value. This will be
        # safe even in the case of spurious wakeups.
        if wait_time is not None and wait_time <= 0.0:
          break
      # We grab the latches that we will signal before we release the lock.
      latches = self._latches
      self._latches = []
      self._cv.release()
      try:
        self._fn(*self._args, **self._kwargs)
      except Exception, _:
        # Only catch Exception, and allow InterruptException through
        LOG.exception("Exception in thread '%s'" % (self.getName(),))

      # We signal only those latches that were present when we last held the
      # lock. This may result in 'extra' function calls from the point of view
      # of some wake caller, but this is preferable to the alternative where a
      # wake caller might see no calls to the function after their call to wake.
      self._signal_latches(latches)

    LOG.debug("Stopping %s" % (self.getName(),))
    self._signal_latches()

  def should_stop(self, latch=None):
    """Instruct the thread to stop when it is next awoken"""
    self._cv.acquire()
    self._should_stop = True
    if latch is not None:
      self._latches.append(latch)
    self._cv.notify()
    self._cv.release()

  def wake(self, latch):
    self._cv.acquire()
    self._latches.append(latch)
    self._cv.notify()
    self._cv.release()

  def _signal_latches(self, latches=None):
    self._cv.acquire()
    if latches is None:
      latches = self._latches
      self._latches = []
    for latch in latches:
      latch.countdown()
    self._cv.release()
      
  def set_name(self, val):
    self.setName(val)

  def set_timeout(self, timeout):
    self._timeout = timeout
