# Copyright (c) 2012 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import
import avro
import logging
import os
import platform
import pyinotify
import re
import shutil
import simplejson as json
import subprocess
import sys
import toposort

import cmf.agent
import cmf.util
from cmf.parcel_id import ParcelId

LOG = logging.getLogger(__name__)

# Keep in sync with master_agent.avdl
ERROR_CODE_FIELD = 'code'
ERROR_MSG_FIELD = 'msg'

UNEXPECTED_ERROR_CODE = '700'
JSON_ERROR_CODE = '701'
NON_COMP_ERROR_CODE = '702'
INCONSISTENT_ERROR_CODE = '703'

DONT_DELETE=".dont_delete"

class Repository(object):
  """
  Class that defines the parcel repository.

  The repository is a map-like object that directly maps
  product names to dictionaries that in turn map product
  versions to parcel manifests.
  """

  CDH_NAME = 'CDH'
  ALTERNATIVES_FILE = 'meta/alternatives.json'
  MANIFEST_FILE = 'meta/parcel.json'
  PERMISSIONS_FILE = 'meta/permissions.json'
  VARLOG_FILE = 'meta/varlog.json'

  def __init__(self, path, package_dir, sudo=None, os_ops=None):
    self.path = path
    self.package_dir = package_dir
    self.products = {}
    self.invalid_products = set()
    self.errors = {}
    self.parcel_added = None
    self.ensure_users_groups_permissions = None
    self.sudo = sudo
    self.osops = os_ops
    self.ignored_parcels = set()

    # There is not enough hate in the world to express how this makes me feel.
    self.is_suse = False
    self.is_suse_11 = False
    if "dist" in dir(platform):
      self.is_suse = platform.dist()[0] == 'SuSE'
      if self.is_suse and "11" in platform.dist()[1]:
        self.is_suse_11 = True

  def set_ensure_users_groups_permissions(self, ensure):
    if self.ensure_users_groups_permissions != ensure:
      ensure_string = ""
      if not ensure:
        ensure_string = "not "
      LOG.info("Agent does " + ensure_string + "create users/groups and apply file permissions")
      self.ensure_users_groups_permissions = ensure

  def _validate(self, parcel):
    '''Helper to aid in unit testing'''
    parcel_info_schema = cmf.agent.MASTER_AGENT_PROTOCOL.types_dict.get("ParcelInfo")
    return avro.io.validate(parcel_info_schema, parcel)

  def refresh(self):
    contents = []
    products = {}
    invalid_products = set()
    errors = {}
    try:
      contents = os.listdir(self.path)
    except:
      LOG.exception("Exception while enumerating parcels")
      return

    for child in contents:
      child_path = os.path.join(self.path, child)

      if child.startswith("."): # Skip ".flood"
        continue
      # Go out of our way to avoid processing symlinks that point
      # to other entries in the repository. This isn't actually
      # fatal, as we derive identity from the parcel metadata, so
      # there's no duplicate; but as we're deliberately introducing
      # active parcel symlinks, let's not have those mistaken for
      # the real thing.
      if os.path.islink(child_path):
        real_path = os.path.realpath(child_path)
        real_root = os.path.realpath(self.path)
        common_prefix = os.path.commonprefix([real_root, real_path])
        if os.path.samefile(self.path, common_prefix):
          LOG.debug("Ignoring symlink '%s' while loading parcel manifests",
                    child)
          continue

      manifest = os.path.join(child_path, self.MANIFEST_FILE)
      LOG.info("Loading parcel manifest for: %s", child)
      try:
        fd = open(manifest)
        try:
          parcel = json.load(fd)
        finally:
          fd.close()
        # Validate against the avro schema up front so that we skip
        # packages that would otherwise cause heartbeating to fail.
        pid = ParcelId.dir(child)
        if self._validate(parcel):
          name = parcel['name']
          version = parcel['version']
        else:
          msg = "Ignoring non-compliant parcel manifest: %s" % (child,)
          LOG.warn(msg)
          # Assume manifest cannot be trusted for name or version
          self._store_error(errors, pid.product, pid.version, NON_COMP_ERROR_CODE, msg)
          invalid_products.add((pid.product, pid.version))
          continue
        if (name, version) in self.ignored_parcels:
          LOG.info("Ignoring parcel files: %s %s", name, version)
          continue
        if name != pid.product or version != pid.version:
          msg = "Ignoring parcel with inconsistent naming. Directory: %s-%s, Manifest: %s-%s" % (pid.product, pid.version, name, version)
          LOG.warn(msg)
          self._store_error(errors, pid.product, pid.version, INCONSISTENT_ERROR_CODE, msg)
          invalid_products.add((pid.product, pid.version))
          continue
      except json.JSONDecodeError, e:
        LOG.exception("Exception while loading parcel manifest: %s", child)
        pid = ParcelId.dir(child)
        self._store_error(errors, pid.product, pid.version, JSON_ERROR_CODE, str(e));
        invalid_products.add((pid.product, pid.version))
        continue
      except Exception, e:
        LOG.exception("Exception while reading parcel: %s", child)
        try:
          pid = ParcelId.dir(child)
          self._store_error(errors, pid.product, pid.version, UNEXPECTED_ERROR_CODE, str(e));
        except Exception, e:
          self._store_error(errors, child, "", UNEXPECTED_ERROR_CODE, str(e));
        # We don't store these exceptions as invalid parcels as we simply don't
        # know enough about what's there to be sure it's a broken parcel vs a
        # stray file that we shouldn't delete out from under the user.
        continue

      if not name in products:
        products[name] = dict()
      products[name][version] = parcel

    new_parcels = []
    for name, versions in products.items():
      if not name in self.products:
        new_parcels.extend(versions.values())
      else:
        for version, parcel in versions.items():
          if not version in self.products[name]:
            new_parcels.append(parcel)

    for parcel in new_parcels:
      if self.ensure_users_groups_permissions:
        self.ensure_users_and_groups(parcel)
        self.ensure_permissions(parcel)
      else:
        LOG.info("Creating users, groups, and modifying " +
                 "permissions for parcels is disabled. Skipping.")
      self.ensure_log_directories(parcel)

    self.products = products
    self.invalid_products = invalid_products
    self.errors = errors

    # callback for new parcels
    if self.parcel_added is not None:
      for parcel in new_parcels:
        self.parcel_added(parcel['name'], parcel['version'])

  def _store_error(self, errors, name, version, error_code, error_msg):
    if not name in errors:
      errors[name] = dict()
    err = dict()
    err[ERROR_CODE_FIELD] = error_code
    err[ERROR_MSG_FIELD] = error_msg
    errors[name][version] = err

  def _get_unversioned_dependencies(self, parcel):
    """Returns a set of all dependencies of this parcel with dependency
       information stripped.
    """
    deps = set()
    if not 'depends' in parcel:
      return deps
    dep_strs = [d.strip() for d in parcel['depends'].split(',')]
    for dep_str in dep_strs:
      # Dependency strings may contain version information, but we don't
      # need it now, so we need only match the name.
      #
      # Some examples:
      # CDH
      # CDH (>> 5.0)
      # Solr (!= 4.3.2)
      m = re.match("(\w+)", dep_str)
      if m:
        deps.add(m.group(1))
    return deps


  def topological_sort(self, parcels):
    """Topologically sort the provided list of parcels in dependency order.

       We assume the presence of a few invariants on the input:
       1. Each product appears at most once in the list of parcels. This
          should be true of any set of activated parcels in a cluster.
       2. All dependencies in the list of parcels are valid. This is
          enforced by the CM server; it won't activate a parcel with unmet
          dependencies.

       Taken together, these invariants allow us to ignore dependency
       version information and sort purely based on products.
    """
    parcels_by_product = dict()
    to_sort = dict()
    for parcel in parcels:
      product = parcel['name']

      assert product not in parcels_by_product, \
        "Duplicate parcel %s found" % (product,)
      parcels_by_product[product] = parcel

      to_sort[product] = self._get_unversioned_dependencies(parcel)

    # Sort and convert the parcel products back into actual parcels.
    #
    # A normal topological sort produces a list of sets. The order of the
    # sets is topological, but the order of set elements doesn't matter. A
    # flattened sort "flattens" the results into a pure list.
    #
    # An exception will be raised if there's a cycle in the graph.
    sorted_parcels = []
    for product in toposort.toposort_flatten(to_sort):
      parcel = parcels_by_product.get(product)
      if parcel:
        # The sort will add missing dependencies to the input before
        # sorting. For example, given input {a:set('b')}, the sort will be
        # done on {a:set('b'), b:set()}.
        #
        # These missing dependencies must be skipped when converting the
        # results; we use parcels_by_product as the source of truth.
        sorted_parcels.append(parcel)
    return sorted_parcels

  def get_parcel_dirname(self, parcel):
    name = parcel['name']
    version = parcel['version']
    return name + "-" + version

  def get_parcel_home(self, parcel):
    return os.path.join(self.path, self.get_parcel_dirname(parcel))

  def get_parcel_path(self, parcel, path):
    parcel_home = self.get_parcel_home(parcel)

    return os.path.join(parcel_home, path)

  def get_script_path(self, parcel, script):
    parcel_home = self.get_parcel_home(parcel)

    if not script in parcel['scripts']:
      return None

    return os.path.join(parcel_home,
                        'meta',
                        parcel['scripts'][script])

  def _get_provided_tags(self, parcel):
    """Get the tags that the specified parcel provides.
       For a 'modern' parcel that specifies its tags, simply return them.
       For legacy parcels, we have a fixed policy to assign tags based on parcel
       name.
    """
    name = parcel['name']
    if 'provides' in parcel:
      provides = parcel['provides']
      LOG.info("Obtained tags %s for parcel %s" % (provides, name))
    else:
      # Hard coded handling of legacy parcels
      if name == 'CDH':
        provides = ['cdh']
      elif name == 'IMPALA':
        provides = ['impala']
      elif name == 'SOLR':
        provides = ['solr', 'flume-plugin', 'hbase-plugin', 'hue-plugin']
      elif name == 'HADOOP_LZO':
        provides = ['hadoop_lzo', 'cdh-plugin', 'impala-plugin', 'solr-plugin']
      else:
        # Assume third party parcels unconditionally apply to all processes
        provides = ['*']
      LOG.info("Assuming tags %s for tag-less parcel %s" % (provides, name))

    return provides

  def prepare_environment(self, parcels, env, required, optional):
    """Prepare the environment with the necessary information
       to allow the service scripts to load the parcels that match the
       required and optional tags.

       Returns a map of the parcels that were ultimately loaded.

       Modifies the passed-in environment to include the relevant parcel
       related variables needed by service scripts.
    """

    LOG.info("prepare_environment begin: %s, %s, %s" % (parcels, required, optional))

    if not parcels:
      LOG.info("No parcels activated for use")
      return {}

    if not required and not optional:
      LOG.info("Service does not request any parcels")
      return {}

    # Establish what parcels we have
    available_parcels, missing_parcels = self.identify_available_parcels(parcels)
    LOG.info("The following requested parcels are not available: %s" % (missing_parcels,))

    used_parcels = []
    satisfied_req_tags = {}
    for name, version in available_parcels.items():
      need_parcel = False
      parcel = self.products[name][version]
      provides = self._get_provided_tags(parcel)

      if '*' in provides:
        need_parcel = True
      else:
        for tag in required:
          if tag in provides:
            if tag in satisfied_req_tags:
              # Overlay parcels may provide the same tags as their dependents.
              other = satisfied_req_tags[tag]
              deps = self._get_unversioned_dependencies(parcel)
              other_deps = self._get_unversioned_dependencies(other)
              if (not name in other_deps and
                  not other['name'] in deps):
                raise ParcelTagOversatisfiedException("Multiple parcels providing required tag: %s (%s, %s)" %
                                                      (tag, name, other['name']))
            satisfied_req_tags[tag] = parcel
            need_parcel = True
            # We keep testing the parcel for required tags in case it provides
            # more than one of them
        for tag in optional:
          if tag in provides:
            need_parcel = True
            # We do not need to bother testing for other optional tags, as the
            # parcel is already selected.
            break

      if need_parcel:
        used_parcels.append(parcel)

    missing_reqs = set(required) - set(satisfied_req_tags)
    if missing_reqs:
      raise ParcelTagUnsatisfiedException("No parcel provided required tags: %s" % (missing_reqs,))

    if not used_parcels:
      return {}

    # Parcels are sorted topologically. This allows the scripts of dependent
    # parcels (those at the end of the list) to override environment
    # variables set by scripts of earlier parcels.
    #
    # In effect, this is a rudimentary "overlay" effect between parcels.
    scripts = []
    dirnames = []
    for parcel in self.topological_sort(used_parcels):
      script_path = self.get_script_path(parcel, 'defines')

      # This is enforced by the parcel validator.
      assert script_path, \
              "Parcel %s has no 'defines' script" % (parcel['name'],)
      scripts.append(script_path)
      dirnames.append(self.get_parcel_dirname(parcel))

    env['SCM_DEFINES_SCRIPTS'] = ':'.join(scripts)
    env['PARCEL_DIRNAMES'] = ':'.join(dirnames)
    env['PARCELS_ROOT'] = self.path

    ret_map = {}
    for parcel in used_parcels:
      ret_map[parcel['name']] = parcel['version']
    LOG.info("prepare_environment end: %s", ret_map)
    return ret_map

  def configure_inotify(self):
    wm = pyinotify.WatchManager()
    handler = RepositoryEvent(wm=wm, repo=self)
    inotify = pyinotify.Notifier(wm, default_proc_fun=handler, timeout=10)
    wm.add_watch(self.path,
                 pyinotify.IN_CREATE | pyinotify.IN_CLOSE_WRITE |
                 pyinotify.IN_MOVED_TO | pyinotify.IN_MOVED_FROM |
                 pyinotify.IN_MODIFY,
                 rec=False, auto_add=False)

    # Add watches for all existing manifest files.
    for parcel in os.listdir(self.path):
      manifest = os.path.join(self.path, parcel, self.MANIFEST_FILE)
      if os.path.isfile(manifest):
        wm.add_watch(manifest, pyinotify.IN_DELETE_SELF)

    inotify.process_events()
    return inotify

  @staticmethod
  def callout(args, env=os.environ.copy()):
    '''Helper to wrap subprocess.call and assist with unit testing'''
    LOG.info("Executing command %s", str(args))
    child = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True, env=env)
    out_str, err_str = child.communicate()
    if child.returncode != 0:
      if len(out_str) > 0:
        LOG.debug("stdout: %s", out_str)
      if len(err_str) > 0:
        LOG.debug("stderr: %s", err_str)
    return child.returncode


  def ensure_users_and_groups(self, parcel):
    """Ensure that a new parcel's users and/or groups are properly configured.
    """

    if not self.running_as_root():
      LOG.info("No users/groups will be created for parcel %s-%s since the agent is not running as root.", parcel['name'], parcel['version'])
      return
    LOG.info("Ensuring users/groups exist for new parcel %s-%s.",
             parcel['name'], parcel['version'])

    groups = parcel['groups']
    users = parcel['users']

    # Add explicit groups and implicit per-user groups.
    for group in groups + users.keys():
      try:
        # groupadd returns '9' if the group already exists.
        ret = self.callout(["/usr/sbin/groupadd", "-r", group])
        if ret != 0 and ret != 9:
          LOG.error("Failed to create group %s for parcel %s-%s: %d",
                    group, parcel['name'], parcel['version'], ret)
      except (TypeError, OSError), e:
        LOG.exception("Failed to create group %s for parcel %s-%s",
                      group, parcel['name'], parcel['version'])

    for user, data in users.items():
      try:
        if self.is_suse and self.is_suse_11:
          umask_arg = '-U'
          umask_param = '022'
        else:
          umask_arg = '-K'
          umask_param = 'UMASK=022'

        useradd_args = [ "/usr/sbin/useradd",
                         "-r", "-m",
                         "-g", user,
                         umask_arg, umask_param,
                         "--home", data['home'],
                         "--comment", data['longname'],
                         "--shell", data['shell'] ]
        if len(data['extra_groups']) > 0:
          useradd_args.extend(["--groups", ",".join(data['extra_groups'])])

        useradd_args.append(user)

        # useradd returns '9' if the user already exists.
        ret = self.callout(useradd_args)
        if ret != 0 and ret != 9:
          LOG.error("Failed to create user %s for parcel %s-%s: %d",
                    user, parcel['name'], parcel['version'], ret)
      except (TypeError, OSError), e:
        LOG.exception("Failed to create user %s for parcel %s-%s",
                      user, parcel['name'], parcel['version'])

  def ensure_permissions(self, parcel):
    """Ensure that a new parcel's files have the right permissions
       applied to them.
    """
    LOG.info("Ensuring correct file permissions for new parcel %s-%s.",
             parcel['name'], parcel['version'])

    permissions_path = self.get_parcel_path(parcel, self.PERMISSIONS_FILE)
    if not os.path.exists(permissions_path):
      return

    permissions = None
    try:
      fd = open(permissions_path)
      try:
        permissions = json.load(fd)
      finally:
        fd.close()
    except:
      LOG.exception("Exception while reading permissions: %s", permissions_path)
      return

    for filename, args in permissions.items():
      path = self.get_parcel_path(parcel, filename)
      try:
        file = cmf.util.validate_and_open_fd(path, self.get_parcel_home(parcel))
        os.close(file)
      except OSError:
        LOG.exception("Error while attempting to modify permissions of file '%s'.", path)
        continue
      except:
        LOG.exception("Rejecting request to modify permissions of file '%s' outside parcel directory.", path)
        continue
      self.apply_permissions(path, args)

  def ensure_log_directories(self, parcel):
    """Ensure that a new parcel's /var/log directories exist and have the
       right permissions applied to them.

       Note that Cloudera produced parcels should not be using this feature.
    """
    varlog_path = self.get_parcel_path(parcel, self.VARLOG_FILE)
    if not os.path.exists(varlog_path):
      return

    LOG.info("Creating directories under /var/log for new parcel %s-%s.",
             parcel['name'], parcel['version'])

    directories = None
    try:
      fd = open(varlog_path)
      try:
        directories = json.load(fd)
      finally:
        fd.close()
    except:
      LOG.exception("Exception while reading /var/log entries: %s", varlog_path)
      return

    for path, args in directories.items():
      try:
        path = os.path.realpath(path)
        if not path.startswith("/var/log"):
          LOG.error("Rejecting request to create log directory outside /var/log: %s", path)
          continue
        if os.path.exists(path):
          # If the target exists, then don't do anything to it, otherwise this
          # could be exploited to modify permissions or ownership of existing
          # directories
          continue
        self.makedirs(path)
      except OSError, e:
        LOG.exception("Exception while creating directory: %s", path)
        continue
      self.apply_permissions(path, args)

  def apply_permissions(self, path, args):

      try:
        # Only change ownership if running as root or if the target owner is
        # is root. In the latter case, we don't change the group since we want
        # that to remain as the agent's non-root group.
        user = args['user']
        group = args['group']
        if not self.running_as_root():
          group = None
        if self.running_as_root() or user == "root":
          LOG.info("chown: %s %s %s", path, user, group)
          self.chown(path, user, group)
        else:
          LOG.info("not changing ownership of %s since we are running as a non-root user", path)

      except (KeyError, OSError), e:
        LOG.exception("Exception while changing owernship of path: %s", path)

      try:
        LOG.info("chmod: %s %s", path, args['permissions'])
        self.chmod(path, args['permissions'])
      except OSError, e:
        LOG.exception("Exception while changing permissions of path: %s", path)

  def chmod(self, path, permissions):
    int(permissions, 8) # validaiton
    args = ["chmod", permissions, path]
    if self.sudo:
      args.insert(0, self.sudo)
    self.callout(args)

  def makedirs(self, *pargs, **kwargs):
    return os.makedirs(*pargs, **kwargs)

  def chown(self, path, user, group):
    user_group = user
    if group:
      user_group += ":%s" % (group,)
    args = ["chown", user_group, path]
    if self.sudo:
      args.insert(0, self.sudo)
    self.callout(args)

  def running_as_root(self):
    return os.getuid() == 0

  def configure_all_symlinks(self, old_symlink_parcels, new_symlink_parcels,
                             manage_old_symlinks, manage_new_symlinks):
    if new_symlink_parcels is not None:
      old = old_symlink_parcels
      for name, version in new_symlink_parcels.items():
        if old is None or name not in old or old[name] != version:
          if manage_new_symlinks:
            LOG.info("Activating system symlinks for parcel %s-%s", name, version)
            self.ensure_alternatives(self[name][version], True)
          self.ensure_active_symlink(self[name][version], True)
        else:
          LOG.debug("System symlinks for parcel %s-%s already configured", name, version)

    if old_symlink_parcels is not None:
      new = new_symlink_parcels
      for name, version in old_symlink_parcels.items():
        if new is None or name not in new or new[name] != version:
          prod = self[name]
          if prod is None:
            LOG.warn("Parcel %s-%s is not in repository", name, version)
            continue

          self.ensure_active_symlink(prod[version], False)
          if manage_old_symlinks:
            LOG.info("Deactivating system symlinks for parcel %s-%s", name, version)
            self.ensure_alternatives(prod[version], False)

  def ensure_alternatives(self, parcel, activate):
    """Ensure that a parcel's alternatives entries are created or removed as
       requested
    """
    if activate:
      action = self.osops.alternatives.activate
      verb = "activate"
    else:
      action = self.osops.alternatives.deactivate
      verb = "deactivate"

    alternatives_path = self.get_parcel_path(parcel, self.ALTERNATIVES_FILE)
    if not os.path.exists(alternatives_path):
      return

    LOG.info("Ensuring alternatives entries are %sd for parcel %s-%s.",
             verb, parcel['name'], parcel['version'])

    try:
      fd = open(alternatives_path)
      try:
        alternatives = json.load(fd)
      finally:
        fd.close()
    except:
      LOG.exception("Exception while reading alternatives: %s", alternatives_path)
      return

    parcel_dir = os.path.join(self.path, self.get_parcel_dirname(parcel))

    for name, args in alternatives.items():
      action(name, args['destination'],
                   os.path.join(parcel_dir, args['source']),
                   args['priority'],
                   args['isDirectory'])

  def ensure_active_symlink(self, parcel, activate):
    """Make sure that the specified parcel's active symlink policy is
       respected. Either create the desired symlink or make sure it
       doesn't exist.

       This symlink is unversioned, so it provides a stable link to
       the currently active version of a product.
    """

    try:
      link_path = os.path.join(self.path, parcel['name'])
      if activate:
        if os.path.islink(link_path):
          os.remove(link_path)
        # Only set the active symlink if the parcel requests it.
        if parcel['setActiveSymlink']:
          # Symlink should be relative
          os.symlink(os.path.basename(self.get_parcel_home(parcel)), link_path)
      elif parcel['setActiveSymlink']:
        # Only remove on deactivation if this parcel wants the symlink when active
        if os.path.exists(link_path):
          linked_path = os.readlink(link_path)
          if linked_path == self.get_parcel_dirname(parcel):
            os.remove(link_path)
    except (TypeError, OSError), e:
      if activate:
        arg = 'create'
      else:
        arg = 'remove'
      LOG.exception("Failed to %s active parcel symlink for parcel %s-%s",
                    arg, parcel['name'], parcel['version'])

  def identify_available_parcels(self, parcels):
    """Check which parcels are available in the repository.

       Returns a tuple of maps (available parcels, missing parcels)
    """

    missing = {}
    available = {}
    for name, version in parcels.items():
      if name not in self.products or version not in self.products[name]:
        missing[name] = version
      else:
        available[name] = version
    return available, missing

  def get_components_for_active_parcel(self, name, version):
    """Return the components of an active parcel in the form used by the version
       detector
    """
    if name not in self.products:
      return []
    if version not in self.products[name]:
      return []
    return self.get_components_for_parcel(self.products[name][version])

  def get_components_for_parcel(self, parcel):
    """Return the components of a parcel in the form used by the version
       detector.
    """
    result = []
    if parcel is None:
      return result

    cdh_version = cdh_release = 'NOT_APPLICABLE'
    if parcel['name'] == self.CDH_NAME:
      cdh_version = self.CDH_NAME + parcel['version'][0]
    # Every component in a parcel inherits the release of the parcel.
    cdh_release = parcel['name'] + ' ' + parcel['version']

    for component in parcel['components']:
      r = dict()
      r['name'] = component['name']

      component_version = component.get('pkg_version')
      if component_version is None:
        component_version = component['version']
      r['component_version'] = component_version
      r['cdh_version'] = cdh_version
      r['cdh_release'] = cdh_release

      component_release = component.get('pkg_release')
      if component_release is None:
        component_release = "NO_VERSION_FILE"
      r['component_release'] = component_release
      r['source'] = "PARCEL"
      r['active'] = True

      result.append(r)
    return result

  def delete(self, product, version):
    """Delete the specified parcel from the repository home. A refresh will be
       synchronously triggered, so the new repo state will be reflected
       immediately.
    """
    parcel = dict()
    parcel['name'] = product
    parcel['version'] = version
    home = self.get_parcel_home(parcel)
    if os.path.exists(home):
      if os.path.exists(os.path.join(home, DONT_DELETE)):
        if (product, version) not in self.ignored_parcels:
          self.ignored_parcels.add((product, version))
          LOG.info("Parcel marked for preservation, not deleting: %s %s", product, version)
          self.refresh()
      else:
        shutil.rmtree(home)

  def keys(self):
    return self.products.keys()

  def items(self):
    return self.products.items()

  def __getitem__(self, key):
    return self.products.__getitem__(key)

class ParcelException(Exception):
  def __init__(self, message):
    self.message = message

  def __str__(self):
    return self.message

class ParcelNotAvailableException(ParcelException):
  pass

class ParcelTagUnsatisfiedException(ParcelException):
  pass

class ParcelTagOversatisfiedException(ParcelException):
  pass

class RepositoryEvent(pyinotify.ProcessEvent):
  def my_init(self, **kargs):
    self.repo = kargs['repo']
    self.wm = kargs['wm']

  def process_default(self, event):
    """
    Process an inotify event. There are two cases we care about:

    - A new directory has been added to the repository. In this case, we
      add a new watch to the new parcel's manifest file, to make sure we
      are notified if it is deleted.
    - A manifest file has been deleted.

    Both cases trigger a refresh of the repository.
    """
    try:
      if event.mask & pyinotify.IN_ISDIR:
        manifest = os.path.join(event.path, event.name, Repository.MANIFEST_FILE)
        if os.path.isfile(manifest):
          self.wm.add_watch(manifest, pyinotify.IN_DELETE_SELF)
        else:
          wd = self.wm.get_wd(manifest)
          if wd is not None:
            self.wm.rm_watch(wd)
      if not event.mask & pyinotify.IN_IGNORED:
        self.repo.refresh()
    except:
      LOG.exception("Exception while refreshing parcel repository.")

if __name__ == '__main__':
  if len(sys.argv) != 2:
    sys.exit(1)

  if not os.path.exists(sys.argv[1]):
    sys.exit(0)

  LOG_FORMAT = '[%(asctime)s] %(process)d %(threadName)s %(module)-12s %(levelname)-8s %(message)s'
  DATE_FORMAT = '%d/%b/%Y %H:%M:%S %z'
  logging.basicConfig(level=logging.DEBUG,
    format=LOG_FORMAT, datefmt=DATE_FORMAT)

  repo = Repository(sys.argv[1], '')
  repo.refresh()

  print "------------------------------------------"

  print "Enumerated components:", repo.keys()
  for name, product in repo.items():
    print "%s:" % name
    for version, parcel in product.items():
      print "\t%s" % version
      for component in parcel['components']:
        print"\t\t%s-%s" % (component['name'], component['version'])
