# Copyright (c) 2012 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging
import shutil
import subprocess
import tempfile

import os

import cmf.agent
import cmf.downloader as downloader
import cmf.util
import cmf.util.diff as diff
from cmf.parcel_id import ParcelId, VALID_EXT
from cmf.throttling_logger import ThrottlingLogger

LOG = logging.getLogger(__name__)
THROTTLED_LOG = ThrottlingLogger(LOG)

PARCEL_CHECKSUM_HEADER = 'CM-Parcel-Checksum'

class ParcelCache(object):
  """
  Class that defines the parcel cache.

  The cache manages packed parcels. It decides when to download, where to
  store, and how to unpack parcels.
  """

  def __init__(self, path, repo, downloader, apply_patch=diff.apply_patch):
    self.path = path
    self.repo = repo
    self.downloader = downloader
    self.apply_patch = apply_patch
    LOG.info("Using %s for parcel cache" % (path))

  def _get_repo_parcels(self):
    """Get the identifying information for all known parcels, including
       invalid parcels that are normally ignored.

       We want to include the invalid parcels so they can be enumerated for
       deletion.
    """
    parcels = set()
    for product in self.repo.products:
      for version in self.repo.products[product]:
        parcels.add(ParcelKey(product, version))
    for product, version in self.repo.invalid_products:
        parcels.add(ParcelKey(product, version))
    return parcels

  def update(self, server_managed_parcels, keep_downloaded_parcels):
    """Update cache based on current list of server managed parcels. May
       download new or delete existing parcels.

       Depending on the value of keep_downloaded_parcels, it may keep or delete
       a downloaded parcel after unpacking.
    """
    managed_parcels = set()
    dirty = False
    # for lookup from ParcelKey back to url
    parcel_url = dict()
    for url in server_managed_parcels:
      pid = ParcelId.url(url)
      key = ParcelKey(pid.product, pid.version)
      managed_parcels.add(key)
      parcel_url[repr(key)] = url

    available_parcels = self._get_repo_parcels()
    for unmanaged in (available_parcels - managed_parcels):
      self.delete_unmanaged(unmanaged.product, unmanaged.version)
    for newly_managed in (managed_parcels - available_parcels):
      url = parcel_url[repr(newly_managed)]
      # we do not skip if already exists because some retry mechanisms rely on
      # repeated, e.g. diff unpacking when base parcel not yet available
      pid = ParcelId.url(url)
      unpacked_dir = os.path.join(self.repo.path, pid.get_parcel_dir())
      # there can be a lag after a parcel is unpacked until it is loaded into
      # inventory and becomes "available_parcels". since our responsibility is
      # only as far as getting the parcel unpacked, we double check here to
      # avoid queuing another job immediately after a prior one finishes.
      if not os.path.exists(unpacked_dir):
        self.queue_parcel_download(url, self.repo.path, keep_downloaded_parcels)

      try:
        self.repo.ignored_parcels.remove((pid.product, pid.version))
        dirty = True
      except KeyError:
        pass


    # abort outdated downloads that are no longer managed
    for dl in self.downloader.get_current_downloads():
      url = dl[downloader.URL_FIELD]
      pid = ParcelId.url(url)
      key = ParcelKey(pid.product, pid.version)
      if key not in managed_parcels:
        self.downloader.abort(url)
    if dirty:
      self.repo.refresh()


  def prune_parcel_download_history(self, product, version):
    """Prune download history of urls that specify the parcel product and
       version.
    """
    def filter(url):
      return ParcelId.url(url).match(product, version)

    self.downloader.prune_history(filter)

  def delete_unmanaged(self, product, version):
    '''
    Delete parcel that is no longer managed by server, from both the cache
    and the repository. Any cached files with matching product and version
    are deleted.
    '''

    LOG.info("Deleting unmanaged parcel %s-%s" % (product, version))

    # Delete from repo first so cache errors don't stop it happening
    self.repo.delete(product, version)

    try:
      cache_files = os.listdir(self.path)
    except:
      LOG.exception("Exception while enumerating caches")
      return
    for cache_file in cache_files:
      filename, ext = os.path.splitext(cache_file)
      if ext is None or len(ext) < 1 or ext[1:] not in VALID_EXT:
        continue
      if ParcelId.filename(cache_file).match(product, version):
        try:
          os.remove(os.path.join(self.path, cache_file))
        except Exception:
          LOG.exception("Exception while removing cache %s" % (cache_file))

  def queue_parcel_download(self, url, unpack_dir, keep_downloaded_parcel):
    """Queue a parcel download with auto unpacking upon completion. The url
       should download a file that ends with either .parcel or .diff. If a
       diff, the base parcel file is expected under the downloader path.
    """
    def cb(url, op):
      if op.state is downloader.ABORTED:
        # clean out the downloaded file and skip the unpacking
        LOG.info("Distribution aborted during download, skipping unpack.")
        if op.path and os.path.exists(op.path):
          os.remove(os.path)
        return
      elif op.state is not downloader.DOWNLOADED:
        LOG.debug("Unable to unpack because download state is %s", str(op.state))
        return

      pid = ParcelId.url(url)

      # If CM has sent a header with the checksum of the parcel we just
      # downloaded, calculate the checksum and raise an error if it does
      # not match.
      parcelHash = op.headers and op.headers.get(PARCEL_CHECKSUM_HEADER, None)
      keep_file = False
      if parcelHash:
        LOG.info('Checking checksum of parcel %s...', pid.get_filename())
        digest = cmf.util.sha1(op.path)
        if parcelHash.lower() != digest:
          LOG.warn('Parcel %s did not match checksum from %s: header %s != calculated %s.',
              pid.get_filename(), url, parcelHash, digest)
          raise Exception, 'Checksum mismatch'
        else:
          keep_file = True
      else:
        # Possible when file already exists on disk
        LOG.warn('No checksum in header, skipping verification')

      unpacked_path = os.path.join(unpack_dir, pid.get_parcel_dir())
      if os.path.exists(unpacked_path):
        LOG.info("%s already exists, skipping unpacking." % (unpacked_path))
        return
      try:
        LOG.info("Unpacking %s into %s" % (op.path, unpack_dir))
        if pid.is_diff:
          base_parcel = os.path.join(self.path, pid.get_base_parcel_filename())
          base_dir = os.path.join(unpack_dir, pid.get_base_parcel_dir())
          if os.path.exists(base_parcel):
            self.apply_patch(base_parcel, op.path, unpack_dir, self.path)
          elif os.path.exists(base_dir):
            LOG.info("No base parcel, but there is a base directory")
            self.apply_patch(base_dir, op.path, unpack_dir, self.path)
          else:
            # base parcel not yet distributed. retry on next hb->queue->cb.
            return
        else:
          unpack_parcel(op.path, unpack_dir, self.path)
          if not keep_downloaded_parcel and op.path:
            os.remove(op.path)
      except Exception, e:
        THROTTLED_LOG.exception("Error while unpacking %s" % (op.path))
        e.keep_file = keep_file
        raise e

    def unpack_parcel(src, dst, tmp):
      # unpack into dir named as <product>-<version>
      if not os.path.exists(dst):
        LOG.info("Making unpack dir %s" % (dst))
        os.makedirs(dst)
      if not os.path.exists(src):
        raise Exception("Src file %s does not exist" % (src))

      priv_tmp = tempfile.mkdtemp(dir=tmp)

      try:
        # See OPSAPS-13299 for why use subprocess instead of tarfile.
        # See OPSAPS-24760 - Use subprocess.Popen instead of subprocess.call to capture tar command output
        proc = subprocess.Popen(["tar", "xf", src, "--no-same-owner", "-C", priv_tmp], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        output = proc.communicate()
        return_code = proc.returncode
        LOG.debug("Untar output for %s to %s:\n stdout: [%s],\n stderr: [%s]" % (src, dst, output[0], output[1]))
        if return_code != 0:
          raise Exception("Untar failed with return code: %d, with tar output:\n stdout: [%s],\n stderr: [%s]" % (return_code, output[0], output[1]))

        dirname = ParcelId.url(src).get_parcel_dir()
        parcel_path = os.path.join(priv_tmp, dirname)
        if not os.path.exists(parcel_path):
          raise Exception("Parcel did not unpack to required directory name: %s" % (dirname))
        shutil.move(parcel_path, os.path.join(dst, dirname))
      finally:
        shutil.rmtree(priv_tmp, True)

      LOG.info("Unpack of parcel %s successful" % (src))

    # ParcelId helps validate url
    ParcelId.url(url)
    self.downloader.queue_download(url, callback=cb)

class ParcelKey(object):
  def __init__(self, product, version):
    self.product = product
    self.version = version

  def __eq__(self, other):
    return self.product == other.product and self.version == other.version

  def __hash__(self):
    return hash(self.product) ^ hash(self.version)

  def __str__(self):
    return "%s-%s" % (self.product, self.version)
