# Copyright (c) 2012 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import
import logging
import os
import re
import urlparse

LOG = logging.getLogger(__name__)

EXT_PARCEL = "parcel"
EXT_DIFF = "diff"
VALID_EXT = [EXT_PARCEL, EXT_DIFF]

SEP = "-"
DOT = "."

class ParcelId(object):
  """
  Uniquely identify a parcel url. All the fields can be extracted from a parcel
  filename. In addition, it supports parsing out the filename and all fields
  from urls. To use http://foo.com/cdh4/rhel5/product1-1.0.0-rhel5.parcel as
  an example, the components would be:
  * filename = "product1-1.0.0-rhel5.parcel"
  * product = "product1"
  * version = "1.0.0"
  * base_version = "1.0.0"
  * distro = "rhel5"
  * ext = "parcel"
  * is_diff = False
  For a diff, e.g. http://foo.com/cdh4/product1-1.0.0-1.1.0-rhel5.diff
  the components would be:
  * filename = "product1-1.0.0-1.1.0-rhel5.diff"
  * product = "product1"
  * version = "1.1.0"
  * base_version = "1.0.0"
  * distro = "rhel5"
  * ext = "diff"
  """
  def __init__(self, is_diff, product, version, base_version, distro):
    self.is_diff = is_diff
    self.product = product
    self.version = version
    self.base_version = base_version
    self.distro = distro
  
  @staticmethod
  def url(url):
    o = urlparse.urlparse(url)
    if len(o) < 3:
      raise Exception("Unexpected urlparse result: %s" % (str(o)))
    filename = os.path.basename(o[2])
    return ParcelId.filename(filename)
  
  @staticmethod
  def filename(filename):
    split = os.path.splitext(filename)
    if split[1] is None:
      raise Exception("Parcel filename must contain extension: %s" % (filename))
    ext = split[1][1:]
    if ext not in VALID_EXT:
      raise Exception("Parcel extension of %s must be one of %s" % (filename, VALID_EXT))
    
    if ext == EXT_PARCEL:
      return ParcelId._parse_parcel(split[0])
    else:
      return ParcelId._parse_diff(split[0])
      
  """
  Determine only product name and version from a parcel directory name.
  """
  @staticmethod
  def dir(dir):
    sep_idx = dir.find(SEP)
    if sep_idx == -1:
      raise Exception("Invalid parcel directory: %s" % (dir))
    product = dir[0:sep_idx]
    version = dir[sep_idx + 1:]
    return ParcelId(False, product, version, version, None)
  
  @staticmethod
  def _parse_parcel(name):
    # This pattern defines the expected form for a regular parcel package.
    #  [product]-[version]-[distro].parcel
    # The group for [version] is greedy so that it captures embedded '-'s.
    # The [product] and [distro] are not allowed to have embedded '-'s.
    parts = re.match(r"^(.*?)-(.*)-(.*?)$", name).groups()

    if len(parts) != 3:
      raise Exception("Invalid parcel filename: %s" % (name))
    return ParcelId(False, parts[0], parts[1], parts[1], parts[2])

  @staticmethod
  def _parse_diff(name):
    # This pattern defines the expected form for a parcel diff package.
    #   [product]-[base_version]-[new_version]-[distro].parcel
    #
    # Note that for diff parcels we require that both the base
    # and the target have the same number of dashes in their
    # versions.
    #
    # The [product] and [distro] are not allowed to have embedded '-'s.
    parts = name.split("-")
    product = parts[0]
    distro = parts[-1]
    versions = parts[1:-1]
    if len(versions) % 2 != 0:
      raise Exception("Invalid diff filename (inconsistent numbers of dashes in versions): %s" % (name))
    base_version = "-".join(versions[0:len(versions)/2])
    new_version = "-".join(versions[len(versions)/2:])
      
    return ParcelId(is_diff=True, product=product, version=new_version, base_version=base_version, distro=distro)

  def get_filename(self):
    if self.is_diff:
      return DOT.join([SEP.join([self.product, self.base_version, self.version,
                                 self.distro]), EXT_DIFF]);
    else:
      return DOT.join([SEP.join([self.product, self.version, self.distro]),
                       EXT_PARCEL]);
  
  def get_base_parcel_filename(self):
    if not self.is_diff:
      return None
    return SEP.join([self.product, self.base_version, self.distro]) + "." + EXT_PARCEL
  
  def get_base_parcel_dir(self):
    if not self.is_diff:
      return None
    return SEP.join([self.product, self.base_version])

  def get_parcel_dir(self):
    return SEP.join([self.product, self.version])
  
  def match(self, product, version):
    return self.product == product and self.version == version 

  def __repr__(self):
    return "[ is_diff: %s, product: %s, version: %s, old_version: %s, distro: %s ]" % (self.is_diff, self.product, self.version, self.base_version, self.distro)
