"""
Copyright 2016, Cloudera Inc.

Process management and tracking.
"""
from __future__ import absolute_import, print_function

# pylint: disable=too-many-lines,too-many-public-methods,too-many-statements
# pylint: disable=too-many-instance-attributes,import-error, fixme
# pylint: disable=line-too-long,bad-continuation,bare-except,broad-except,invalid-name,
# pylint: disable=missing-docstring,too-many-branches,too-many-lines

import grp
import pwd
import sys

import signal
import os
import re
import logging
import time
import subprocess

import psutil
from httplib import BadStatusLine, CannotSendRequest
import simplejson as json

import cmf.audit
import cmf.audit.util
import cmf.kt_renewer

from cmf.common import RetryableProcessException
from cmf.monitor.conf import CLOUDERA_MONITOR_PROPERTIES
from cmf.monitor.monitor_properties import ConfParser
from cmf.template_util import render_text
import cmf.util.os_ops
from cmf.util import (user_group_to_ids, unzip, source, deploy_special_file, fix_hue_home, fix_hcat_home, try_tail,
                      dict_diff, write_conf_file, dumpstacks)
from cmf.util.workqueue import WorkQueue
from cmf.supervisor import SupervisorWrapper


LOG = logging.getLogger(__name__)

# CM status codes are expected to be reported in the format
# "CM_STATUS_CODE=<status code>"
CM_STATUS_CODE_MARKER = "CM_STATUS_CODE="
CM_STATUS_CODE_REGEX = CM_STATUS_CODE_MARKER + "(.*)$"
MAX_CM_STATUS_CODE_STR_LEN = len(CM_STATUS_CODE_MARKER) + 80

# Substitution variable for the configured DSSD device path in /dev
DSSD_DEVICE_PATH = "{{DSSD_DEVICE_PATH}}"

# This matches regex in EnvironmentParamSpec.java and bash.
VALID_ENV_NAME_RE = re.compile("^[a-zA-Z_][a-zA-Z0-9_]*$")

class Process(object):
  # Running processes that exit with a status not listed here will be
  # automatically restarted by supervisor (if configured to do so).
  SUPERVISOR_EXPECTED_EXITCODES = [0, 2]

  # Map of all Process objects.
  Proc_Map = {}
  queue = None # Initialized in issue_updates(), see comments there.

  """
  Manages a supervised or one-off process.
  """

  PROC_JSON = "proc.json"
  CONFIG_ZIP = "config.zip"
  PROCESS_TIMESTAMP = "process_timestamp"
  CONF_DIR_REF = "{{CMF_CONF_DIR}}"
  EXIT_CODE = "exit_code"
  known_resources_files = {}

  osops = None
  cfg = None

  class InvalidProcess(Exception):
    pass

  @classmethod
  def ClassInit(cls, osops, cfg):
    cls.osops = osops
    cls.cfg = cfg

  def __init__(self, cfg, agent, raw, parent=None):
    self.agent = agent
    self.raw = raw.copy()
    self.id = raw["id"]
    self.name = "%s-%s" % (self.id, self.raw["name"])
    self.raw_name = self.raw["name"]
    self.dir = os.path.join(cfg.paths.process_dir, self.name)
    self.conf_path = os.path.join(self.dir, "supervisor.conf")
    self.logdir = os.path.join(self.dir, "logs")
    self.conf_link_name = os.path.join(cfg.paths.supervisor_include_path,
                                       self.name + ".conf")
    self.activated = False
    self.early_error = None
    self.environment = {}
    self.synced = False  # True if/when supervisord state is in sync.
    self.audit_plugin = None
    self.metadata_plugin = None
    self.profile_plugin = None

    # Proxy reported by supervisor may be proc_monitor, so the real Process pid
    # may be different.
    self.proxy_pid = (None, None)

    # list of tuples (bind_address, port)
    self.ports = []


    if parent:
      self.synced = parent.synced
      self.early_error = parent.early_error
      self.private_cgroups = parent.private_cgroups
      self.shared_cgroups = parent.shared_cgroups
      self.parcels_in_use = parent.parcels_in_use
      self.audit_plugin = parent.audit_plugin
      self.metadata_plugin = parent.metadata_plugin
      self.profile_plugin = parent.profile_plugin
      self.monitor = parent.monitor
      self.borrowed_keytab = parent.borrowed_keytab
      self.activated = parent.activated
      self.ports = parent.ports
    else:
      if raw.get('running', True) != True:
        filtered_raw = raw.copy()
        filtered_raw.pop("configuration_data", None)
        filtered_raw.pop("environment", None)
        #LOG.warning("New process is not running: %s", filtered_raw)
      # CGroups that are private to this process.
      self.private_cgroups = []

      # CGroups that may be shared with other processes.
      self.shared_cgroups = []

      self.parcels_in_use = {}

      # TODO: Validate names?
      # TODO: Validate hostname?

      # The path to the keytab file if it exists and was borrowed to be used
      # as a candidate to kinit from. See borrow_keytab for more details.
      self.borrowed_keytab = None

      # self.monitor may be None
      self.monitor = None

    self.config_dictionary = {'verify_cert_file': cfg.verify_cert_file,
                              'verify_cert_dir': cfg.verify_cert_dir,
                              'max_cert_depth': cfg.max_cert_depth,
                              'client_key_file': cfg.client_key_file,
                              'client_cert_file': cfg.client_cert_file,
                              'client_keypw_file': cfg.client_keypw_file,
                              'client_keypw_cmd': cfg.client_keypw_cmd,
                              'conf_dir': self.dir} #location of navigator.client.properties

    self.supervisord_data_path = os.path.join(self.dir, "supervisor_status")
    self.process_info = None
    # A new process won't have this file present, but a pre-existing process
    # (post agent restart) will.
    # For processes that have EXITED and de-registered from supervisord before
    # agent restart, this helps report back correct status to CM server.
    # For active processes, process_info will be updated before sending next heartbeat
    # by fetching latest status from supervisord.
    if os.path.exists(self.supervisord_data_path):
      with open(self.supervisord_data_path) as f:
        self.process_info = json.load(f)

    # audit, metadata and profile plugins needs to be created only for non one of process.
    if "one_off" not in self.raw or not self.raw["one_off"]:
      self.plugin_init()
    else:
      LOG.debug("Skipping Plugin creation for one-off process: %s", self.raw['name'])

    self.Proc_Map[self.id] = self


  def plugin_init(self):
    self.audit_plugin = cmf.audit.util.make_audit_plugin(self.raw['name'], self.config_dictionary)
    self.metadata_plugin = cmf.audit.util.make_metadata_plugin(self.raw['name'], self.config_dictionary)
    self.profile_plugin = cmf.audit.util.make_profile_plugin(self.raw['name'])

  @classmethod
  def GetProcesses(cls, filter_func=None):
    '''
    Return all processes conforming to "filter", or all of them if filter is None.
    :param filter: Callable taking one parameter (process object) return true/false to indicate inclusion
                  of process in return value list.
    :return: List of processes accepted by filter.
    '''
    if filter_func is None:
      return cls.Proc_Map.values()
    out = []
    for proc in cls.Proc_Map.values():
      if filter_func(proc):
        out.append(proc)
    return out

  def is_failed(self):
    return self.early_error is not None

  def update_home_dir(self, user, environment):
    # pylint: disable=no-self-use
    try:
      check_home_dir = "~%s" % (user,)
      expanded_home_dir = os.path.expanduser(check_home_dir)
      if expanded_home_dir != check_home_dir:
        # Expansion successful
        environment['HOME'] = expanded_home_dir
      else:
        LOG.warn("Unable to expand user home for user: %s", user)
    except:
      LOG.exception("Failed to update home dir for user %s", user)

  @property
  def all_cgroups(self):
    """
    Returns a list of all cgroups created for this process.
    """
    return self.private_cgroups + self.shared_cgroups

  @property
  def has_cgroups(self):
    """
    Returns True if this process has any created cgroups, False otherwise.
    """
    return len(self.all_cgroups) > 0

  def _make_monitor_daemon(self):
    if self.monitor is not None:
      return
    if self.agent.monitor_daemon is None:
      LOG.warn("Not creating a monitor for %s: MonitorDaemon not created yet", self.name)
      return

    if self._should_monitor():
      self.monitor = self.agent.monitor_daemon.make_monitor(self.raw['name'],
                                                            self.name,
                                                            self.cfg.paths.cgroups_dir,
                                                            self.audit_plugin)
    else:
      LOG.info("Not creating a monitor for %s: should_monitor returns false", self.name)

  def _get_should_monitor_flag(self):
    monitor_path = os.path.join(self.dir, CLOUDERA_MONITOR_PROPERTIES)
    if not os.path.exists(monitor_path):
      return False
    conf_parser = ConfParser(self.dir, CLOUDERA_MONITOR_PROPERTIES)
    default_section = conf_parser.get_default_section()
    if default_section is None:
      LOG.warn("Default section not found in file %s, process %s", monitor_path, self)
      return False
    else:
      return conf_parser.getboolean_with_default(default_section, "should_monitor", True)

  def _should_monitor(self):
    """
    Helper to determine whether we should montior this process. There are two
    cases to handle: the RECONSTRUCTED path handles the case where we are
    recreating our state at startup from a pre-existing supervisord instance.
    Here, we check whether the process's existing directory has the cloudera
    monitoring properties file. The simpler case simple of dealing with a
    process from the heartbeat simple checks if this is a non-one-off process.
    """

    # Using a default value of True here means that if one off is not set we
    # will not monitor the process. For us to monitor the process, it must be
    # set and set to False.
    # Another condition is that the 'should_monitor' configuration for a service or role
    # is set to be false, which means the user intend not to monitor this
    # role or service. Then we should return false.

    return (self._get_should_monitor_flag() is True) and not self.raw.get("one_off", True)

  def register_error(self, error):
    LOG.debug("Registering error %s for process %s", error, self.name)
    if self.early_error is not None:
      LOG.warn("Overriding error message %s. Did you forget to clear_error first?", self.early_error)
    self.early_error = error

  def clear_error(self):
    LOG.debug("Clearing errors for process %s", self.name)
    self.early_error = None

  def refresh_daemons(self):
    if self.audit_plugin is None:
      self.audit_plugin = cmf.audit.util.make_audit_plugin(self.raw['name'], self.config_dictionary)

    if self.metadata_plugin is None:
      self.metadata_plugin = cmf.audit.util.make_metadata_plugin(self.raw['name'])
    try:
      if self.audit_plugin:
        LOG.info("Begin audit plugin refresh")
        self.audit_plugin.refresh(self.dir)
    except:
      LOG.exception("Error refreshing audit plugin at %s", self.dir)

    try:
      if self.metadata_plugin:
        LOG.info("Begin metadata plugin refresh")
        self.metadata_plugin.refresh(self.dir)
    except:
      LOG.exception("Error refreshing metadata plugin at %s", self.dir)

    try:
      if self.profile_plugin:
        LOG.info("Begin profile plugin refresh")
        self.profile_plugin.refresh(self.dir)
    except:
      LOG.exception("Error refreshing profile plugin at %s", self.dir)

    self._make_monitor_daemon()
    if self.monitor is not None:
      try:
        LOG.info("Begin monitor refresh.")
        self.monitor.refresh(self.dir)
      except:
        LOG.exception("Error refreshing monitor at %s", self.dir)

    LOG.info("Daemon refresh complete for process %s.", self)

  def check_user(self, user):
    # pylint: disable=no-self-use
    try:
      pwd.getpwnam(user)
    except KeyError:
      raise Exception("User [%s] does not exist.", user)

  def check_group(self, group):
    # pylint: disable=no-self-use
    try:
      grp.getgrnam(group)
    except KeyError:
      raise Exception("Group [%s] does not exist.", group)

  def check_user_and_group(self, user, group):
    self.check_user(user)
    self.check_group(group)

  def filter_cgroups(self, paths):
    """
    Returns a new list of cgroup paths that may exclude certain cgroups
    deemed to be inappropriate for this process.
    """
    filtered = list(paths)

    # We retain the memory cgroup if a soft or hard limit was configured.
    remove_memory = True
    for res in self.raw["resources"]:
      mem = res["memory"]
      if mem and (mem["hard_limit"] != -1 or
                  mem["soft_limit"] != -1):
        remove_memory = False
    if remove_memory:
      path = self.agent.cg_manager.find_group_path(paths, "memory", self.name)
      if path:
        filtered.remove(path)

    return filtered

  @classmethod
  def handle_heartbeat(cls, agent, process, raw, force):
    try:
      if process is None:
        process = cls(agent.cfg, agent, raw)
        LOG.info("[%s] Instantiating process", process.name)
      process.update_heartbeat(raw, force)
    except (RetryableProcessException, BadStatusLine, CannotSendRequest):
      if raw is not None:
        filtered_raw = raw.copy()
        filtered_raw.pop("configuration_data", None)
        LOG.exception("Failed to update - Will retry soon: %s", filtered_raw)
    except Exception as exc:
      if raw is not None:
        filtered_raw = raw.copy()
        filtered_raw.pop("configuration_data", None)
        LOG.exception("Failed to update %s", filtered_raw)
      if process:
        process.register_error(exc)

  @classmethod
  def issue_updates(cls, agent, processes):
    results = []

    # This should be statically initialized, but because the process may daemonize, it's not safe
    # to create threads (work queues) in static initialization.  Therefore it's deferred.
    if cls.queue is None:
      cls.queue = WorkQueue()

    for (proc, new_raw, force) in processes:
      results.append(cls.queue.execute(lambda p=proc, n=new_raw, f=force: cls.handle_heartbeat(agent, p, n, f)))
    dump_threshold = 10
    can_dump = True
    wait_start = time.time()
    while len(results) > 0:
      result = results[0]
      result.wait(dump_threshold)
      if result.is_done():
        results.remove(result)
      if can_dump and (time.time()-wait_start > dump_threshold):
        dumpstacks(None, None)
        can_dump = False

  def update_heartbeat(self, proc_data, force_update=False):
    '''
    Handle an update of process heartbeat data.
    If "proc_data" is None, then the process is no longer included in the heartbeat process list, and thus
    is to be destroyed.
    :param proc_data: Raw heartbeat data.
    :param force_update: Assume HB data has changed, even if it hasn't.
    :return: None
    '''
    if self.raw == proc_data and not force_update and self.synced:
      return

    if proc_data:
      d = dict_diff(self.raw, proc_data, key_no_show=set('configuration_data'))
      LOG.info("[%s] Updating process: %s %s", self.name, proc_data['running'], d)
    else:
      LOG.info("[%s] Updating process (remove).", self.name)

    try:
      # Save the old raw data so that we can restore it in case of error.
      old = self.raw
      # We must process the update with self.raw being updated.  But this is dangerously
      # wrong - since now other attributes of "self" may end up being modified even if
      # we roll back the update to "self.raw" in the exception handler.  Historically,
      # we've been able to live with this...
      self.raw = proc_data

      if proc_data is not None and proc_data["running"]:
        self.fs_update()

        # We don't start processes more than once.
        # If the process is stopped - don't start it.
        # But if there is not exit_code file, then it hasn't been started yet and so we
        # can issue a start request.
        if not self.activated and \
            (self.process_info is None or self.is_supervisor_stopped()) and \
            not os.path.exists(os.path.join(self.dir, self.EXIT_CODE)):
          self.supervisor_start()
          self.activated = True
        self.refresh_daemons()
      else:
        self.deactivate()

      if proc_data is None and not self.activated:
        self.remove()
        self.synced = True
      elif proc_data is None:
        # Waiting for deactivation -- still activated

        # We may have issued a process stop request, but it hasn't completed so we don't accept
        # the new state as the current state of the process yet.
        self.raw = old

        # Set synced false to let us go through this logic again and effectively poll
        # for supervisord to tell that the process has stopped.
        self.synced = False
      else:
        self.synced = True


    except Exception:
      self.synced = False
      self.raw = old
      raise


    # Everything is good with the active process.
    self.clear_error()

    if self.raw:
      # Write out the raw process state so that it can be reconstructed.
      # configuration_data is written as config.zip - since it can't be written out as json text
      try:
        f = self.agent.os_ops.open_write_file(os.path.join(self.dir, self.PROC_JSON))
        out = self.raw.copy()
        cfg_data = out.get('configuration_data', None)
        try:
          del out['configuration_data']
        except KeyError:
          pass
        try:
          f.write(json.dumps(out, indent=2))
        finally:
          f.close()

        config_zip = os.path.join(self.dir, self.CONFIG_ZIP)
        # Use os.open/os.fdopen to ensure that we create the file with mode 0600
        f = self.agent.os_ops.open_write_file(config_zip)
        try:
          f.write(cfg_data)
        finally:
          f.close()
      except Exception, e:
        LOG.exception("[%s] Failed to write out saved config: %s", self, e)
    return

  def write_configuration(self, do_filter=False):
    '''
    Update all config files, regardless of the current activation state.
    Either we're writing the files out for the first time prior to activation,
    or doing an update of dynamic config state.

    :param do_filter: True -> update only files identified by "refresh_files".
    :return:
    '''

    if not do_filter:
      file_filter = None
    else:
      file_filter = self.raw['refresh_files']

    LOG.info("[%s] Refreshing process files: %s", self.name, file_filter)
    self.update_process_environment_and_parcels()
    self.update_process_environment_for_cgroups()

    # Write process config files might potentially convert into the raw
    # state in the retry rounds. So we must do it in the first activation!
    self.write_process_conf(file_filter)
    self.borrow_keytab()


    return False

  def supervisor_start(self):

    self.check_port_conflicts()

    # Before doing the real work for the activation, we'd differentiate
    # whether it is the first time activation or not. Based on the current
    # process status if exists, we proceed differently.
    if not os.path.exists(self.conf_link_name):
      os.symlink(self.conf_path, self.conf_link_name)

    self.agent.supervisor_client.start_process(self.name)

    # Reread the managed processes's config and restart processes whose
    # config has been changed.
    self.agent.supervisor_client.reload()

    return False

  def fs_update(self):
    '''
    Update all FS content representing this process.
    Either this will be a partial update of specific config files (specified by
    HB contents), or a full write-out of the configuration if the configuration doesn't exist.

    The presence of the supervisor config file indicates that an initial version of all other
    configs has been written.

    :return:
    '''

    user = self.raw["user"]
    group = self.raw["group"]
    uid, _ = user_group_to_ids(user, group)
    if os.getuid() != 0 and os.getuid() != uid:
      raise Exception("Non-root agent cannot execute process as user '%s'" % user)

    if os.path.exists(self.conf_path) or os.path.isfile(self.conf_link_name):
      # This logic stream protects us from OPSAPS-26645: The core process configuration is
      # only ever written out once.  Subsequently we only update files as per "refresh_files".
      self.write_configuration(True)
      # Re-evaluate any dynamic resources.
      self.evaluate_resources(False)
      return

    # This is the first activation of the process therefore supervisord
    # can't find this process by its name. Proceed normally!
    LOG.info("First time to activate the process [%s].", self.name)

    # We only do a whole bunch of work from scratch for the process in the
    # first activation because some of the role process control scripts
    # will change the process config when to start the process. The
    # consecutive activation of the process shouldn't overwrite the process
    # config to raw state. See OPSAPS-26645 for details.

    # Control group creation is conditioned on the following:
    # 1. Whether RM is enabled. Just because we have mounted cgroup
    #    subsystems doesn't mean we should create new cgroups.
    # 2. Whether the process is one off. The lifecycles of one off
    #    processes are radically different: they may never deactivate.
    if self.agent.rm_enabled and not self.raw["one_off"]:
      all_cgroups = self.agent.cg_manager.get_group_paths(self.name)
      filtered_cgroups = self.filter_cgroups(all_cgroups)

      # Rewrite the cgroup paths as dictionaries so that any additional
      # args can be used by create_group().
      private_paths = [{"path" : p} for p in filtered_cgroups]
      shared_paths = self.get_named_cgroups()
      for entry in private_paths + shared_paths:
        try:
          self.agent.cg_manager.create_group(**entry)
        except:
          # Clean up any created cgroups before reraising.
          for created in self.all_cgroups:
            self.agent.cg_manager.destroy_group(created)
          self.private_cgroups = []
          self.shared_cgroups = []
          raise

        if entry in shared_paths:
          self.shared_cgroups.append(entry["path"])
        else:
          assert entry in private_paths
          self.private_cgroups.append(entry["path"])

    # This is a common problem so we do an explicit check
    if self.cfg.sudo is None:
      self.check_user_and_group(user, group)

    self.agent.mkworkdir(self.dir, user=user, group=group)
    # Create log directory for the process
    self.agent.mkworkdir(self.logdir, user=user, group=group)
    self.write_configuration(False)
    self.evaluate_resources(True)
    self.write_supervisor_conf()

  def write_process_conf(self, namelist=None):
    '''
    Write process-specific configuration files (i.e. those specified in the HB payload).

    :param namelist: List of files to update: None implies all.
    :return:
    '''
    user = self.raw["user"]
    group = self.raw["group"]


    if self.agent.honor_users_and_groups:
      # This is a common problem so we do an explicit check
      self.check_user_and_group(user, group)
      uid, gid = user_group_to_ids(user, group)
    else:
      uid, gid = None, None

    # Even if uid/gid are not None, they are only honored when unzipping
    # new files.
    unzip(self.raw["configuration_data"], self.dir, uid=uid, gid=gid,
          namelist=namelist)

    # Broaden permissions on everything but keytab and tgt files. Files and
    # directories created in unzip have mode 600 (files) and 700
    # (directories). Extend them to 640 (files) and 751 (directories).
    #
    # Note that if an explicit list of names was specified, the changes
    # are restricted to only those files and directories in the list.
    #
    # TODO: Ideally each file's mode would be defined in configuration_data,
    # but the ZIP archive format doesn't support permissions.
    for root, dirs, files in os.walk(self.dir):
      for f in files:
        if not f.endswith(".keytab") and \
           not f.endswith(".tgt") and \
           not f.endswith(".key") and \
           (namelist is None or f in namelist):
          os.chmod(os.path.join(root, f), 0640)

      for d in dirs:
        if namelist is None or d in namelist:
          os.chmod(os.path.join(root, d), 0751)

    # For a supervised process, all the config files in the config dir.
    # However, some files need to be deployed to special locations and
    # need to be owned by a user other than the one running the process.
    # Deploy them now, provided they're in the explicit list of names (if
    # one exists).
    if self.raw["special_file_info"]:
      # We include both the inherited environment (os.environ)
      # and the additional environment.
      env = os.environ.copy()
      env.update(self.environment)
      # To correctly support parcel deployments, we must obtain the parcels'
      # environment variables too. This has to be done through a shell as the
      # parcels provide their environment in the form of a shell script.
      env.update(source(self.resolve_program("common/cloudera-config.sh"),
                        "source_parcel_environment", env))
      for info in self.raw["special_file_info"]:
        try:
          if namelist is None or info["name"] in namelist:
            deploy_special_file(info, self.dir, env,
                                self.cfg.sudo)
        except Exception:
          LOG.exception("Error while deploying: %s.  Environment was: %s.", info, env)

  def check_port_conflicts(self):
    '''
    Check if there are any processes running using the ports this process is expected to use.
    Issue warning messages if found.
    '''

    # ie. check for python 2.7
    if len(self.ports) == 0 or not hasattr(subprocess, "check_output"):
      return
    try:
      args = []
      for x in self.ports:
        # Check port conflicts if port value is not negative and not zero
        # port value can be negative in the configuration if the service is disabled
        if x[1] > 0:
          args += ["sport", "=", str(x[1]), "or"]

      if len(args):
        args = args[:-1]

      cmd = ["ss", "-np", "state", "listening", "("] +  args + [')']

      output = subprocess.check_output(cmd, shell=False)
      pid_match_re = re.compile('pid=([0-9]*),')
      pids = set()
      for line in output.split("\n"):
        m = pid_match_re.search(line)
        if m:
          pids.add(m.group(1))

      for proc in pids:
        cmdline = "<N/A>"
        try:
          with open(os.path.join('/proc/%s/cmdline' % proc)) as f:
            cmdline = f.read()
        except:
          pass
        LOG.warn("[%s] port conflict with pid %s:  %s", self.name, proc,
                 cmdline.replace('\000', ' '))
    except Exception as exc:
      LOG.error("[%s] Failed port check: %s", self.name, exc)

  def borrow_keytab(self):
    """We borrow the process credential for monitoring. OPSAPS-6906."""
    # Always use the HTTP principal.
    #
    # The process's krb principal is hardcoded by CM (always `hdfs', `mapred',
    # etc.). And that krb principal is not the same as the process user, which
    # is configurable. So the conclusion is that we have to do some hardcoding.
    # And that might as well be `HTTP'.
    princ_prefix = "HTTP"

    for root, _, files in os.walk(self.dir):
      for f in files:
        if f.endswith(".keytab"):
          f_path = os.path.join(root, f)

          # Is our principal potentially in there?
          if princ_prefix not in file(f_path).read():
            continue

          cmf.kt_renewer.add_keytab_to_kinit_from(
              f_path, "%s/%s" % (princ_prefix, self.cfg.reported_hostname,))
          self.borrowed_keytab = f_path

  def update_process_environment_for_cgroups(self):
    """
    Adds environment variables for all mounted cgroup
    subsystems.  Environment variables are of the form
    CGROUP_ROOT_SUBSYSTEM and CGROUP_GROUP_SUBSYSTEM, where
    SUBSYSTEM is one of the mounted subsystems, in all caps.

    Also lists the mount points for named cgroups.
    CGROUP_NAMED_TYPES lists the types which have names.  Will not
    be defined if no named cgroup types were created.
    CGROUP_NAMED_<type>_NAME has the name of a particular type.
    CGROUP_NAMED_<type>_PATH has the path of a particular type.
    """
    for subsys, path in self.agent.cg_manager.get_mounts().iteritems():
      self.environment["CGROUP_ROOT_" + subsys.upper()] = path
    for subsys, group in self.agent.cg_manager.get_agent_cgroups().iteritems():
      if isinstance(group, (list, tuple)):
        group = '/'.join(group)
      self.environment["CGROUP_GROUP_" + subsys.upper()] = group
    named_groups = []
    for entry in self.get_named_cgroups():
      gtype = entry["type"]
      named_groups.append(gtype)
      self.environment["CGROUP_NAMED_" + gtype + "_NAME"] = entry["name"]
      self.environment["CGROUP_NAMED_" + gtype + "_PATH"] = entry["path"]
    if named_groups:
      self.environment["CGROUP_NAMED_TYPES"] = " ".join(named_groups)

  def update_process_environment_and_parcels(self):
    """Calculate and store the environment and parcels that should be used for
       this process.
    """
    environment = self.agent.supervisee_environment.copy()
    environment['CONF_DIR'] = self.dir

    parcels_in_use = self.agent.repo.prepare_environment(self.raw['parcels'],
                                                         environment,
                                                         self.raw['required_tags'],
                                                         self.raw['optional_tags'])

    raw_env = self.raw["environment"]
    if 'SET_PYTHON_PATH' in raw_env:
      environment['PYTHONPATH'] = ":".join(sys.path)
    for key, value in raw_env.iteritems():
      value = value.replace("{{AGENT_COMMON_DIR}}", os.path.join(self.cfg.package_dir, "common"))
      value = value.replace(self.CONF_DIR_REF, self.dir)
      environment[key] = value

    # OPSAPS-30610
    # Set home dir based on user, if available.
    self.update_home_dir(self.raw['user'], environment)

    # Hue home moved. If we've got the wrong value, we'll try the CDH4 value
    fix_hue_home(environment)
    # Hcatalog home moved.
    fix_hcat_home(environment)

    self.environment = environment
    self.parcels_in_use = parcels_in_use

  def write_supervisor_conf(self):
    '''
    Write the supervisor config file for the process.
    :return:
    '''
    # find rlimits
    limit_fds = None
    limit_memlock = None
    for res in self.raw["resources"]:
      limits = res["rlimits"]
      if limits:
        LOG.info("reading limits: %s", limits)
        if limits['limit_fds']:
          limit_fds = limits['limit_fds']
        if limits['limit_memlock']:
          limit_memlock = limits['limit_memlock']

    LOG.info("[%s] Launching process. one-off %s, command %s, args %s", self,
      self.raw["one_off"], self.raw["program"], self.raw["arguments"])

    # Prepare arguments to launch program via redactor wrapper script.
    program = os.path.normpath(os.path.join(sys.prefix, "bin", "cm"))
    args = []
    if self.raw["one_off"]:
      args += ["launch", "--one-off", "--"]
    else:
      args += ["launch", "--"]

    args.append(self.resolve_program(self.raw["program"]))
    args.extend(self.raw["arguments"])

    expected_exitcodes = self._get_expected_exitcodes()

    sup_data = {
      "program_name": self.name,
      "command": program,
      "arguments": args,
      "directory": self.dir,
      "environment": self.environment,
      "one_off": self.raw["one_off"],
      "auto_restart": self.raw["auto_restart"],
      "exitcodes": ",".join(str(EC) for EC in expected_exitcodes),
      "limit_fds": limit_fds,
      "limit_memlock": limit_memlock,
    }
    # Newlines will break supervisor configuration, and
    # shouldn't be here.
    assert "\n" not in sup_data["program_name"]
    assert "\n" not in sup_data["command"]
    for k, v in sup_data["environment"].items():
      # Validate that key is valid.
      if not VALID_ENV_NAME_RE.match(k):
        LOG.warning("Removing invalid environment key: %s=%s", k, v)
        del sup_data["environment"][k]

    if self.agent.honor_users_and_groups:
      sup_data["user"] = self.raw["user"]
      sup_data["group"] = self.raw["group"]
      sup_data["extra_groups"] = ','.join(self.raw["extra_groups"])

    if self.has_cgroups:
      # Supervisor will attach the process to these cgroups.
      sup_data["cgroups"] = ",".join(self.private_cgroups)

      # But not these; they're only reflected back when the agent reconnects.
      sup_data["shared_cgroups"] = ','.join(self.shared_cgroups)

    sup_data["start_timeout_seconds"] = self.raw.get("start_timeout_seconds", None)
    if sup_data["start_timeout_seconds"] is None:
      sup_data["start_timeout_seconds"] = 20

    sup_config = render_text("supervisee.conf", data=sup_data)
    # Replace all "%" in the config to prevent attempts at formatting expansions
    # inside supervisord.
    sup_config = sup_config.replace("%", "%%")
    write_conf_file(self.conf_path, sup_config)

  def _get_expected_exitcodes(self):
    # CM server may optionally specify exitcodes in which case with default value
    # of null in avro protocol key will be present with None value.
    process_expected_exitcodes = self.raw.get("expected_exitcodes", None) or []
    return set(Process.SUPERVISOR_EXPECTED_EXITCODES).union(process_expected_exitcodes)

  def resolve_program(self, program):
    """
    Returns path of program relative to package_dir,
    if program is a relative path.
    """
    if program.startswith("/"):
      return program
    else:
      return os.path.join(self.cfg.package_dir, program)

  def stop_monitors(self):
    '''
    Stop all process monitors/daemons (idempotent).
    '''
    LOG.info("[%s] stopping monitors", self)
    try:
      if self.audit_plugin:
        self.audit_plugin.stop_nav_monitoring()
        self.audit_plugin = None
    except:
      LOG.exception("Error stopping audit plugin at %s", self.dir)

    try:
      if self.metadata_plugin:
        self.metadata_plugin.stop_nav_monitoring()
        self.metadata_plugin = None
    except:
      LOG.exception("Error stopping metadata plugin at %s", self.dir)

    try:
      if self.profile_plugin:
        self.profile_plugin.stop_nav_monitoring()
        self.profile_plugin = None
    except:
      LOG.exception("Error stopping profile plugin at %s", self.dir)

    if self.monitor is not None:
      try:
        self.monitor.stop()
        self.monitor = None
      except:
        LOG.exception("Error stopping monitor at %s", self.dir)

  def deactivate(self):
    """
    Process is no longer requested as a running process by CM server,
    or it has stopped on its own and that needs to be reconciled.
    This is asynchronous -- when new supervisor state is received via
    update_supervisor_data, the result of the supervisor operation will
    be recognized and this operation will be completed via deactivate_complete().
    """
    if self.process_info is not None and not self.is_supervisor_stopped() and self.activated:
      LOG.info("[%s] Deactivating process", self)

      try:
        self.agent.supervisor_client.stop_process(self.name)

        data = self.agent.supervisor_client.get_process_info(self.name)
        self.update_supervisor_data(data)
      except Exception as exc:
        LOG.exception("[%s] deactivation fail: %s", self, exc)
        raise
    else:
      LOG.info("[%s] Deactivating process (skipped)", self)
      if self.raw is None or (not self.activated and not self.raw["running"]):
        # Process is not running and according to CM server it should not be running;
        # monitoring not needed now.
        # If self.raw is None, it means that CM doesn't know about the process at all; therefore
        # there should be no monitors for it even if we have yet to shut it down.
        self.stop_monitors()

  def deactivate_complete(self):
    """
    Supervisor data has indicated that the process is no longer running.
    (And it won't run again.)  The process needs to be unregistered
    from the supervisor.
    """
    LOG.info("[%s] Unregistered supervisor process %s", self, self.fixed_supervisor_state())
    needs_reload = self.delete_conf()

    # Ignore the value of 'activated': some processes (such as HBase)
    # won't be deactivated, they'll just be forgotten the next time
    # they start up. Because we unset the cgroup lists, this is safe.
    for path in self.all_cgroups:
      self.agent.cg_manager.destroy_group(path)
    self.private_cgroups = []
    self.shared_cgroups = []

    # We only need to talk to supervisord when
    # we've encountered a substantive change.
    if needs_reload or not self.synced:
      self.agent.supervisor_client.reload()

    if self.borrowed_keytab is not None:
      cmf.kt_renewer.remove_keytab_to_kinit_from(self.borrowed_keytab)
      self.borrowed_keytab = None

    self.activated = False
    if self.raw is None or not self.raw["running"]:
      # Process is not running and according to CM server it should not be running or should not exist;
      # monitoring not needed now.
      self.stop_monitors()

  def remove(self):
    # Server doesn't want to know anything about this anymore, delete it.
    # since the server no longer recognizes this id.
    del Process.Proc_Map[self.id]
    self.mark_orphan(self)

  # Create a marker file in the given process config directory so that later
  # on it will be deleted when it becomes old enough.
  @classmethod
  def mark_orphan(cls, process, now=None):
    """
    Handle orphaned processes on the agent.

    An orphaned process is not listed in the heartbeat response of the CM
    server but is still present in the agent in either the active or failed
    process lists. We need to remove these processes' associated config
    directories 'orphan_process_directory_staleness_threshold' seconds after
    we identify that they've been orphaned. We don't delete them immediately
    because we want to retain temporary access to the logs for debugging
    purposes, even though the process has no visibility in the CM server.

    Note:
    1) Only when the heartbeat response does not list
       a process does it get marked as orphaned.
    2) It's possible that a process is not known to supervisord, so it's not
       in either list (active or failed). We still need to track it and
       delete it when the right time comes.
    3) Some config directories are actually client config deployment
       attempts. They, too, are subject to orphan cleanup; we must ask the
       client config module which directories are active.
    """

    if isinstance(process, Process):
      if process.activated:
        LOG.warning("[%s] Deferring orphaning due to running process.", process)
        return
      process = process.name

    # Check if this orphan process has been marked
    marker = os.path.join(cls.cfg.paths.process_dir, process, cls.PROCESS_TIMESTAMP)
    if now is None:
      now = time.time()

    if not os.path.isfile(marker):
      LOG.info("[%s] Orphaning process", process)
      try:
        f = open(marker, 'w')
        f.write(str(now))
        f.close()
      except:
        LOG.exception("Error creating marker %s", marker)


  def delete_conf(self):
    """
    Deletes the supervisor configuration file for this process. Any errors
    are ignored.

    Returns True/False to indicate if supervisord configuration has changed/needs reloading.
    """
    try:
      os.remove(self.conf_link_name)
      return True
    except:
      LOG.exception("Couldn't remove supervisor include for process %s", self)

    return False

  def get_named_cgroups(self):
    """
    Retrieve any named cgroups from the process resources.
    """
    cgroups = []
    for res in self.raw["resources"]:
      # Support for "named_cpu" resource is legacy, but required so that CM server
      # can avoid staleness flagging when upgrading from pre-6.2.0 to post-6.2.0.
      named_resource = res.get("named_cpu")
      if named_resource:
        for path in self.agent.cg_manager \
                              .get_group_paths(named_resource["name"], "cpu"):
          cgroups.append({
            "name" : named_resource["name"],
            "type" : "cpu",
            "path" : path,
            "user" : named_resource["user"],
            "group" : named_resource["group"],
            "mode" : named_resource["mode"]})
      named_resource = res.get("named_resource")
      if named_resource:
        for path in self.agent.cg_manager \
                              .get_group_paths(named_resource["name"],
                                               named_resource["type"]):
          LOG.info("cgroup %s / %s -> path %s",
                   named_resource["name"], named_resource["type"], path)
          cgroups.append({
              "name" : named_resource["name"],
              "type" : named_resource["type"],
              "path" : path,
              "user" : named_resource["user"],
              "group" : named_resource["group"],
              "mode" : named_resource["mode"]})

    return cgroups

  def _do_named_generic_resource(self, named_resource):
    # Named generic cgroups/resources
    path = self.agent.cg_manager.find_group_path(self.all_cgroups,
                                                 named_resource["type"],
                                                 named_resource["name"])
    if path:
      # We assume here that the optional record with cgroup specific
      # properties is named after the type.  e.g. if
      # named_resource["type"] = "blkio" then the optional resource
      # record would be called "blkio", so we could look it up with
      # the expressions named_resource.get("blkio").
      resource = named_resource.get(named_resource["type"])
      if resource:
        self.agent.cg_manager.configure_group(path, named_resource["type"],
                                              resource)

      # Work around OPSAPS-12606.
      if named_resource["type"] == "cpu":
        self.agent.cg_manager.configure_group(path, "cpu", dict(rt_runtime=1000))

  # This method is legacy support for the previous resrouce
  # specification protocol, which is used by the server to avoid
  # staleness after upgrage of CM.
  def _do_named_cpu_resource(self, cpu_resource):
    # Named cgroups
    path = self.agent.cg_manager.find_group_path(self.all_cgroups,
                                                 "cpu", cpu_resource["name"])
    if path:
      self.agent.cg_manager.configure_group(path, "cpu",
                                            dict(shares=cpu_resource["shares"]))

    # Work around OPSAPS-12606.
    self.agent.cg_manager.configure_group(path, "cpu", dict(rt_runtime=1000))

  def _do_cpu_resource(self, cpu_resource):
    # Set cgroup cpu parameters.
    path = self.agent.cg_manager.find_group_path(self.all_cgroups,
                                                 "cpu", self.name)
    if path:
      self.agent.cg_manager.configure_group(path, "cpu", cpu_resource)

      # Work around OPSAPS-12606.
      self.agent.cg_manager.configure_group(path, "cpu", dict(rt_runtime=1000))

  def _do_io_resource(self, io_resource):

    # Set cgroup blkio parameters.
    path = self.agent.cg_manager.find_group_path(self.all_cgroups,
                                                 "blkio", self.name)
    if path:
      self.agent.cg_manager.configure_group(path, "blkio", io_resource)

  def _do_port_resource(self, port_resource):
    self.ports += [(port_resource["bind_address"], port_resource["port"])]

  def _do_mem_resource(self, mem_resource):
    # Set cgroup memory parameters.
    path = self.agent.cg_manager.find_group_path(self.all_cgroups,
                                                 "memory", self.name)
    if path:
      self.agent.cg_manager.configure_group(path, "memory", mem_resource)

  def _do_install_file_resource(self, file_resource):
    """
    Process a resource that refers to a file that the agent must install.
    The file must be a known resource that is already registered with the agent.
    """
    src = file_resource['source']
    try:
      src = self.known_resources_files[src]
    except KeyError:
      LOG.warning("[%s] Requested resource not present: %s", self.name, src)
      return
    dst = file_resource['path']
    dst = dst.replace(self.CONF_DIR_REF, self.dir)
    LOG.info("[%s] Installing resource: %s -> %s", self.name, src, dst)
    user = file_resource['user']
    group = file_resource['group']
    mode = file_resource['mode']
    self.osops.install_file(dst, src, user, group, mode)


  def _do_dir_contents_resource(self, contents):
    # Set directory contents permissions
    path = contents["path"]
    if path == DSSD_DEVICE_PATH:
      if self.agent.supports_dssd:
        path = self.agent.cached_dssd_device_dir
      else:
        path = None
    user = contents["user"]
    group = contents["group"]
    mode = contents["mode"]
    if path is not None and user is not None and group is not None and mode is not None:
      LOG.info("Chowning contents of %s to %s %s with mode 0%o",
               path, user, group, mode)
      # We don't recurse?  Odd but true.
      self.osops.set_directory_metadata(path, user, group, mode, False)

    else:
      LOG.warning("Ignoring invalid DirectoryContentsResource: %s, %s, %s, 0%o",
                  contents["path"], user, group, mode)

  def _resource_iterator(self, fn):
    """
    Helper function to loop over all resources in a list and present them to a real handler (fn).
    Isolates loop and exception logic in common code.
    """
    def _iter(resources):
      for res in resources:
        try:
          fn(res)
        except Exception, exc:
          # Resource evaluation failures are non-fatal.
          LOG.exception("[%s] Could not evaluate resource %s: %s", self.name, res, exc)
    return _iter

  def _do_directory_resources(self, resources):

    # Directory creation is sorted so that if a process
    # requests both /a/b and /a, then /a gets created with
    # the right permissions, rather than root/root.
    for d in sorted(resources, key=lambda x: x["path"]):
      try:
        return_value = self.osops.mkabsdir(d["path"], user=d["user"], group=d["group"], mode=d["mode"])
        # osops.mkabsdir() returns (=False) without correcting the ownership/mode
        # if the directory path already exists on the system due to
        # commit 1a4d6ad76b4b4280152864f77ade8f04a6ae0d22
        # Hence, the case when an audit directory is created before a
        # service directory and if both require a common dir path, the ownership
        # is not altered to the service's requirement. Which leads to service not
        # starting.
        #
        # For instance, the default location of audit log dir for hdfs
        # '/var/log/hadoop-hdfs/audit' has ownership of hdfs:hdfs
        # and hdfs's log directory '/var/log/hadoop-hdfs' has hdfs:hadoop.
        # When user modifies the audit location to be '/var/new-log/hadoop-hdfs/audit'
        # it gets an ownership of hdfs:hdfs but the path '/var/new-log/hadoop-hdfs', as it
        # didn't existed before, gets an ownership of root:root
        # Later, if the DN, NM log dirs are modified to be at '/var/new-log/hadoop-hdfs'
        # hdfs refuses to start due to incorrect ownership of '/var/new-log/hadoop-hdfs'.
        # Which remains to be the root:root (due to mkabsdir)
        #
        # Therefore, explicitly calling set_path_metadata() to rectify ownership/mode
        # for ONLY those directory resources (json; diff of changes) that were modified by the
        # user via CM WebUI
        if return_value is False:
          self.osops.set_path_metadata(d["path"], d["user"], d["group"], d["mode"])
      except:
        # Resource evaluation failures are non-fatal.
        LOG.exception("Could not evaluate resource %s", d)

  def _get_resource_handler_map(self):
    handlers = {'cpu': self._resource_iterator(self._do_cpu_resource),
                'named_cpu': self._resource_iterator(self._do_named_cpu_resource),
                'io': self._resource_iterator(self._do_io_resource),
                'memory': self._resource_iterator(self._do_mem_resource),
                'install': self._resource_iterator(self._do_install_file_resource),
                'contents': self._resource_iterator(self._do_dir_contents_resource),
                'tcp_listen': self._resource_iterator(self._do_port_resource),
                'directory': self._do_directory_resources,
                'named_resource': self._resource_iterator(self._do_named_generic_resource)}
    return handlers

  def evaluate_resources(self, first_time):
    """
    Evaluate process resources, making system changes as needed. Unknown
    or unused resources are ignored.

    If first_time is set, will evaluate both static and dynamic resources.
    """

    type_map = dict()
    handlers = self._get_resource_handler_map()

    for res in self.raw["resources"]:
      # Skip static resources if it's not the first time.
      if not res["dynamic"] and not first_time:
        continue

      for h in handlers.keys():
        r = res.get(h, None)
        if r:
          type_map.setdefault(h, []).append(r)

    for k, resources in type_map.iteritems():
      handler = handlers[k]
      try:
        LOG.info("[%s] Evaluating resource: %s", self.name, k)

        handler(resources)
      except Exception, exc:
        LOG.exception("Could not evaluate resource type %s %s", k, exc)

  def update_supervisor_data(self, supervisor_data):
    """
    Receive new supervisor data for processes.
    Update the process monitors in response.
    """
    if supervisor_data is None:
      if self.activated:
        self.deactivate_complete()
      return
    self.process_info = supervisor_data
    with self.osops.open_write_file(self.supervisord_data_path) as f:
      json.dump(self.process_info, f)
    pid = None

    if self.process_info.get("statename", None) == "RUNNING":
      pid = int(self.process_info["pid"])

    if not self.proxy_pid or self.proxy_pid[0] != pid:
      try:
        # If the supervisor reported process is the "proc_monitor" then the real
        # pid is the last item on the command line.
        proc = psutil.Process(pid)
        cmdline = proc.cmdline()
        real_pid = pid
        if "proc_monitor" in cmdline:
          real_pid = int(cmdline[cmdline.index('proc_monitor') + 1])
        elif "proc_watcher" in cmdline:
          real_pid = int(cmdline[cmdline.index('proc_watcher') + 1])

        self.proxy_pid = (pid, real_pid)
        pid = real_pid
      except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
        pass
    else:
      pid = self.proxy_pid[1]

    if self.monitor is not None:
      self.monitor.refresh_pid(pid)

    if self.activated and self.is_supervisor_stopped():
      self.deactivate_complete()

  def fixed_supervisor_state(self):
    # Heuristic to protect against OPSAPS-9208.
    #
    # If an auto-restartable process fails with an unexpected exit code,
    # supervisor will transition it to EXITED, and then start it up again
    # (transitioning it to STARTING). We must be careful not to publish
    # the momentary EXITED state; otherwise it might cause an in-flight
    # bring up command to fail.
    #
    # In effect, this heuristic makes automatic restarts more invisible
    # to the system (though not completely; the subsequent STARTING state
    # will still be published).
    #
    # Note that an auto restarted process cannot be one off; this is
    # guaranteed by supervisee.conf.

    state_name = self.process_info.get("statename", "STOPPED")
    exit_code = self.process_info.get("exitstatus", -1)
    if state_name == "EXITED" and \
        self.raw is not None and self.raw.get("auto_restart", False) and \
            exit_code not in self._get_expected_exitcodes():
      state_name = "STARTING"
      self.process_info["statename"] = "STARTING"
    return state_name

  def is_supervisor_stopped(self):
    return self.fixed_supervisor_state() in ["EXITED", "FATAL", "STOPPED"]

  def make_status(self):
    """
    Returns a ProcessStatus object given supervisor_data.
    """

    process_status = dict()
    process_status["run_generation"] = self.raw.get("run_generation", 0)
    process_status["parcels_in_use"] = self.parcels_in_use

    process_stats = dict()
    process_stats["id"] = self.id
    # See OPSAPS-4593. We don't report most stats at the moment, and these are
    # not displayed anywhere in our UI. The schema and code here remain since
    # we plan to add these back shortly.
    process_stats["rss"] = -1
    process_stats["vms"] = -1
    process_stats["cpu_percent"] = -1.0
    process_stats["user_cpu_time"] = -1.0
    process_stats["system_cpu_time"] = -1.0

    process_status["name"] = self.raw_name

    if self.early_error is not None:
      fail_str = "Error found before invoking supervisord: %s" % self.early_error
      process_status["status"] = "FATAL"
      process_status["stdout_tail"] = fail_str
      process_status["stderr_tail"] = fail_str
      process_status["cm_status_code"] = ""
      process_status["exit_code"] = -1
      process_status["pid"] = -1
      process_stats["wallclock_time"] = -1
    elif self.process_info is None:
      # If supervisor doesn't know about the process, it's certainly
      # not running in our view of the world. This is a more useful
      # response in edge cases where the Agent is restarted and forgets
      # about proccesses.
      #
      # However, if we had problems talking to supervisor, respond with
      # an unknown status. This has been observed when the machine is
      # under heavy load and can be safely considered a transient state
      # because if it happens often enough, the agent will have exited.
      process_status["pid"] = -1
      process_status["status"] = "UNKNOWN"
      process_status["stdout_tail"] = "UNKNOWN"
      process_status["stderr_tail"] = "UNKNOWN"
      process_status["cm_status_code"] = ""
      process_status["exit_code"] = -1
      process_stats["wallclock_time"] = -1
    else:
      process_status["pid"] = self.proxy_pid[1] if self.proxy_pid[1] else -1

      process_status["status"] = self.fixed_supervisor_state()
      process_status["stdout_tail"] = try_tail(self.process_info["stdout_logfile"])
      process_status["stderr_tail"], process_status["cm_status_code"] = \
          Process._extract_err_str_and_cm_status_code(try_tail(self.process_info["stderr_logfile"]))
      process_status["exit_code"] = self.process_info.get("exitstatus", -1)
      if self.process_info["stop"] <= self.process_info["start"]:
        process_stats["wallclock_time"] = self.process_info["now"] - self.process_info["start"]
      else:
        process_stats["wallclock_time"] = self.process_info["stop"] - self.process_info["start"]

    if self.raw.get("one_off") and \
       os.path.exists(os.path.join(self.dir, "exit_code")):
      try:
        with open(os.path.join(self.dir, "exit_code")) as f:
          process_status["exit_code"] = int(f.read())
      except:
        pass

    return process_status, process_stats

  @staticmethod
  def _extract_err_str_and_cm_status_code(err_log):
    if err_log is None:
      return "", ""

    search_start = max(0, len(err_log) - MAX_CM_STATUS_CODE_STR_LEN)
    match = re.search(CM_STATUS_CODE_REGEX, err_log[search_start:])
    if match is not None:
      # "- 1" required below to skip the new line character preceding the error
      # code.
      return err_log[:max(0, search_start + match.start() - 1)], match.group(1)
    else:
      return err_log, ""

  def __str__(self):
    return self.name


class BaseReconstructedProcess(Process):
  """
  Base class for a process defined from filesystem contents (as opposed to
  an incoming definition coming from a HB response). The base class is used
  for Process.list_processes()
  """

  # Marker in a raw process object indicating that process
  # was reconstructed from agent data
  RECONSTRUCTED = "__reconstructed__"

  def __init__(self, full_id, existing_process, cfg, agent=None):
    split_id = full_id.split('-', 1)
    process_id = int(split_id[0])

    proc_json = os.path.join(cfg.paths.process_dir, full_id, self.PROC_JSON)
    config_zip = os.path.join(cfg.paths.process_dir, full_id, self.CONFIG_ZIP)
    timestamp = os.path.join(cfg.paths.process_dir, full_id, self.PROCESS_TIMESTAMP)
    if os.path.exists(timestamp):
      raise Process.InvalidProcess()

    try:
      f = open(proc_json)
      try:
        raw = json.loads(f.read())
      finally:
        f.close()
      if os.path.exists(config_zip):
        try:
          f = open(config_zip)
          try:
            raw['configuration_data'] = f.read()
          finally:
            f.close()
        except Exception, e:
          # This isn't fatal -- we still have the process definition.
          LOG.warn("Cannot recreate process configuration data: %s %s", full_id, e)
    except Exception, e:
      LOG.warn("Cannot reinstantiate process: %s", full_id)

      fake_raw = dict()
      fake_raw["id"] = process_id
      fake_raw["name"] = split_id[1]
      fake_raw[self.RECONSTRUCTED] = True

      # We need to deduce the correct activated state from the
      # process state.
      # RUNNING: Obviously, this should be activated
      # STARTING: To get into the STARTING state, it must have been activated
      # BACKOFF: To get into this state, it must have been STARTING
      # FATAL: Less obvious, but this state only occurs when BACKOFF fails
      # EXITED: Least obvious. This occurs when the process terminates itself.
      #         When we see such a process, we don't know if the server is
      #         aware of this state. If we mark it as activated, and the server
      #         thinks its not running, it will ask us to deactivate it, which
      #         is a no-op. If the server thinks it is running, it will not
      #         ask us to change anything. Then it will see the EXITED state
      #         and mark the process as deactivated, and we're back in sync.
      # STOPPING and STOPPED: These map to deactivated.
      # UNKNOWN: Could be anything but let's assume deactivated.
      if not existing_process:
        fake_raw['running'] = 'UNKNOWN'
      else:
        fake_raw['running'] = existing_process['statename'] in ['RUNNING', 'STARTING', 'BACKOFF']
      raw = fake_raw

    Process.__init__(self, cfg, agent, raw)

    if existing_process:
      self.activated = existing_process['statename'] in ['RUNNING', 'STARTING', 'BACKOFF']
    else:
      self.activated = False

  def plugin_init(self):
    pass

  @classmethod
  def rebuild_process_set(cls, cfg, sup_data, agent=None):
    matched = set()

    # Look for all processes defined in the process dir.
    for proc_id in os.listdir(cfg.paths.process_dir):
      if not os.path.isdir(os.path.join(cfg.paths.process_dir, proc_id)):
        continue
      sup_proc = sup_data.get(proc_id, None)
      try:
        cls(proc_id, sup_proc, cfg, agent)
        matched.add(proc_id)
      except Process.InvalidProcess:
        pass
      except ValueError:
        # Will be thrown on a directory that fails to parse as a process name,
        # e.g. a client-config dir
        continue
      except Exception as exc:
        LOG.exception("Invalid process definition: %s: %s", proc_id, exc)
        continue

    # Look for all processes defined by the supervisor
    for proc_id in sup_data.keys():
      if proc_id in matched:
        continue

      try:
        cls(proc_id, sup_data[proc_id], cfg, agent)
      except Process.InvalidProcess:
        pass
      except ValueError:
        # Will be thrown on a directory that fails to parse as a process name,
        # e.g. a client-config dir
        continue
      except Exception as exc:
        LOG.exception("Invalid process definition: %s: %s", proc_id, exc)
        continue

  @classmethod
  def list_processes(cls, cfg):
    """
    List known processes to stdout, meant to execute as a standalone command.
    Includes processes already defined in a running supervisor and processes that exist
    due to directory content that shows the process as existing, or having existed in the past.
    """

    cls.ClassInit(cmf.util.os_ops.OSOps(False), cfg)

    try:
      sup = SupervisorWrapper(cfg)
      sup_data = sup.get_supervisor_data()
    except Exception as exc:
      LOG.warning("Failed supervisor connection (results may be incomplete): %s", exc)
      sup_data = {}

    cls.rebuild_process_set(cfg, sup_data)
    print("CM PID System   Name")
    for proc in Process.GetProcesses():
      proc.update_supervisor_data(sup_data.get(proc.name))
      status, _ = proc.make_status()
      pid = status.pop('pid')
      if pid in [0, -1]:
        pid = status['status']
      else:
        pid = str(pid)
      print("{id:6} {pid:8} {name}".format(id=proc.id, pid=pid, **status))

  def signal_watcher(self, sig):
    """
    Send a signal to the watcher process (if any) for the process represented by this object.
    """
    if self.proxy_pid and self.proxy_pid[0] != self.proxy_pid[1]:
      try:
        # If the supervisor reported process is the "proc_monitor" then the real
        # pid is the last item on the command line.
        proc = psutil.Process(self.proxy_pid[0])
        cmdline = proc.cmdline()

        # cmdline[0] is "python"
        if os.path.split(cmdline[1])[1] == "cm" and cmdline[2] == "proc_watcher":
          LOG.info("Identified watcher: %s", proc.pid)
          proc.send_signal(sig)
      except Exception as exc:
        LOG.error("Failed watcher restart %s: %s", self.name, exc)

  @classmethod
  def signal_watchers(cls, cfg, term=False):
    """
    Identify all Processes in the system and for any watcher process that's running
    send it a SIGUSR1, or if "term", then send a SIGTERM.
    """
    if term:
      sig = signal.SIGTERM
    else:
      sig = signal.SIGUSR1

    try:
      sup = SupervisorWrapper(cfg)
      sup_data = sup.get_supervisor_data()
    except Exception as exc:
      LOG.warning("Failed supervisor connection (results may be incomplete): %s", exc)
      sup_data = {}

    cls.rebuild_process_set(cfg, sup_data)
    for proc in Process.GetProcesses():
      proc.update_supervisor_data(sup_data.get(proc.name, None))
      proc.signal_watcher(sig)


class ReconstructedProcess(BaseReconstructedProcess):
  """
  Derivation of BaseReconstructedProcess to be used to create a process object that
  will be used (i.e. a live process following an agent restart).
  """
  def __init__(self, full_id, existing_process, cfg, agent=None):
    BaseReconstructedProcess.__init__(self, full_id, existing_process, cfg, agent)

    # Ask supervisor about existing cgroups and set up reference counts.
    #
    # Note that it's not safe to actually create the cgruops here as the
    # roots may not yet be mounted.

    # CM5: 'cgroups' lists the cgroups that are wholely owned by this
    # process. Supervisor attached the process to each of them.
    private_cgroups = set()
    if existing_process:
      private_cgroups.update(existing_process.get('cgroups', []))

    # CM5: 'shared_cgroups' lists the named cgroups shared with other
    # processes. Supervisor _did not_ attach the process to them.
    shared_cgroups = set()
    if existing_process:
      shared_cgroups.update(existing_process.get('shared_cgroups', []))

    # CM45: 'has_cgroups' implies that this process has a cgroup for each
    # subsystem. Supervisor attached the process to each of them.
    if existing_process and existing_process.get('has_cgroups'):
      private_cgroups.update(agent.cg_manager.get_group_paths(self.name,
                                                             mounted=False))

    for path in private_cgroups.union(shared_cgroups):
      agent.cg_manager.create_group(path, do_create=False)
    self.private_cgroups = list(private_cgroups)
    self.shared_cgroups = list(shared_cgroups)
    self.__reported = False


  def update_heartbeat(self, proc_data, force_update=False):
    '''
    Since this is a "Reconstructed" process, we can convert it to a real Process, because
    now we have HB data.
    '''
    LOG.info("[%s] Reinstantiating process %s", self.name, self.id)
    proc = Process(self.cfg, self.agent, self.raw, self)
    proc.update_heartbeat(proc_data, True)

