"""
Copyright (c) 2016 Cloudera, Inc. All rights reserved.
"""
from __future__ import absolute_import

# pylint: disable=broad-except,invalid-name,missing-docstring, too-many-statements

# Wrapper script applies redaction engine policy specified in REDACTION_RULES_FILE
# to processes launched by supervisord.
# Redactor runs as parent of real managed process applying the policy as it reads the
# stdout and stderr streams of the process via a pipe.

import os
import sys
import threading
import signal
import time

import fcntl
import errno
import logging
import cmf.util.os_ops
import cmf.redaction
import cmf.util

REDACTION_RULES_FILE = "redaction-rules.json"

EXEC_MARKER = ".exec_marker"

LOG = logging.getLogger()

# Assuming supervisor waits 30 seconds to issue SIGKILL after SIGTERM.
# Thus we have to issue a SIGKILL in 25 seconds before the watcher gets
# a SIGKILL itself.
SIGKILL_WAIT_TIME = 25

os_ops = cmf.util.os_ops.OSOps()

def exec_watch(pid):
  """
  Re-exec this process.  Post-re-exec the process will just be running "run_redaction".
  Ensure that any pipes for redaction survive the exec and are identified in command
  line arguments.
  """

  try:
    cmd = (sys.argv[0], "proc_watcher", str(pid))
    LOG.info("Re-exec watcher: %s", " ".join(cmd))
    os.execv(cmd[0], cmd)
    assert False
  except Exception as err:
    LOG.exception("Failed re-exec: %s", err)

def sighandler(pid, sig, frame):
  """
  SIGTERM signal handler.
  Forward SIGTERM to watched child.
  Upgrade SIGTERM to SIGKILL if not responsive.
  Include special SIGTERM handling for upgrade as necessary.

  """
  # pylint: disable=unused-argument
  ppid = os.getppid()

  if ppid == 1 and os.path.exists("./.upgrade_marker"):
    # Supervisord has died and a .upgrade_marker file exists.  This indicates an upgrade.
    # In this case we are not supposed to die.   The upgrade process will set up a "cm proc_monitor"
    # to connect to this process.

    # Place-holder logic for future upgrade handling code.
    # Upgrade logic will re-exec watcher via SIGUSR1 to upgrade code here.
    # This will make the watcher tolerant of SIGTERMs due to the death of the supervisor.

    # Supervisor keeps std{out,err}.log writable only by root, so this process loses access
    # to those files once supervisord is dead.  We'll make do by writing to alternate files.
    stdout = os_ops.open_write_file("logs/stdout-x.log", mode=0600, append=True)
    stderr = os_ops.open_write_file("logs/stderr-x.log", mode=0600, append=True)

    os.dup2(stdout.fileno(), 1)
    os.dup2(stderr.fileno(), 2)

    sys.stdout = stdout
    sys.stderr = stderr
    exec_watch(pid)
    return
  try:
    LOG.info("Killing with SIGTERM")
    os.kill(pid, sig)

    # Spawn a thread to kill with SIGKILL after a delay.
    # This should be async so as to avoid blocking the signal handling thread.
    def rekill():
      time.sleep(SIGKILL_WAIT_TIME)
      LOG.info("Killing with SIGKILL")
      os.kill(pid, signal.SIGKILL)

    t = threading.Thread(target=rekill)
    t.daemon = True
    t.start()

  except Exception as exc:
    LOG.exception("Kill failure: %s", exc)

def write_exit_code(exit_code, exclusive=False):
  with os_ops.open_write_file("exit_code", exclusive=exclusive) as f:
    f.write("%s" % exit_code)

def watch(pid):
  """
  Watch a CDH role instance.  Intercept and forward signals.
  This code may be "upgraded" via a SIGUSR1 to inject new functionality during agent upgrades.
  """
  # If we get a SIGTERM, pass it on to the child.
  signal.signal(signal.SIGTERM, lambda sig, frame: sighandler(pid, sig, frame))

  # Restart this watcher on SIGUSR1.
  signal.signal(signal.SIGUSR1, lambda sig, frame: exec_watch(pid))

  # Remove the EXEC_MARKER to allow process launching to proceed.
  # Once we reach this point there won't be any implicit wait() calls
  # made behind our back (e.g. via libraries calling subprocess stuff).
  try:
    os.unlink(EXEC_MARKER)
  except OSError as err:
    # EXEC_MARKER won't exist if we re-exec via SIGUSR1.
    if err.errno != errno.ENOENT:
      raise

  # Wait for child death.
  while True:
    try:
      p, status = os.waitpid(pid, 0)
      if p == pid:
        if os.WIFSIGNALED(status):
          # Exit with 1 code to force a restart by supervisord
          write_exit_code(1)
          sys.exit(1)
        else:
          write_exit_code(os.WEXITSTATUS(status))
          sys.exit(os.WEXITSTATUS(status))
    except OSError as err:
      if err.errno == errno.EINTR:
        continue
      if err.errno == errno.ECHILD:
        write_exit_code(1)
        sys.exit(1)
    except Exception:
      continue

def exec_redactor(child_stdout, child_stderr):
  """
  Re-exec process into a redactor.
  The redactor will have file-descriptors passed to it as pipes, identified via
  command line arguments.  These pipes must be marked to not be closed on exec.
  """

  for fd in [0, 1, 2, child_stdout, child_stderr]:
    if fd is not None:
      flags = fcntl.fcntl(fd, fcntl.F_GETFD)
      fcntl.fcntl(fd, fcntl.F_SETFD, flags & ~fcntl.FD_CLOEXEC)

  try:
    cmd = (sys.argv[0], "redactor", "--fds", str(child_stdout), str(child_stderr))
    LOG.info("Re-exec redactor: %s", " ".join(cmd))
    os.execv(cmd[0], cmd)
    assert False
  except Exception as err:
    LOG.exception("Failed re-exec: %s", err)

def launch(one_off, args):
  """
  Launcher process.  Exec'ed by supervisor.
  Will set up a redactor process if stdout/stderr redaction is required.
  Will re-exec itself into a "watcher" -- which will intercept/relay signals
  from the supervisor.  The watcher allows for the CDH role instance process to be
  safely detached from the supervisor when we want to upgrade/restart the supervisor.
  """
  LOG.info("Started launcher: %s", " ".join(args))

  # We don't want agent's path to leak into managed process being launched.
  # Like following "PATH=/usr/lib64/cmf/agent/build/env/bin:/sbin:/usr/sbin:/bin:/usr/bin".
  path = os.environ.get('PATH')
  if path:
    os.environ['PATH'] = cmf.util.remove_agent_path(path)
  # OPSAPS-40013 hack
  try:
    del os.environ['PYTHONHOME']
  except KeyError:
    pass

  # One-off processes cannot be re-executed.  We use the presence of the exit_code file
  # to indicate this.  Write an empty file here.  If it exists an exception will be raised
  # and execution of the process will fail.
  try:
    write_exit_code("", exclusive=one_off)
  except Exception as exc:
    LOG.error("Failed exclusive write of exit code: %s", exc)
    sys.exit(1)

  # Don't bother applying redaction engine
  # if the redaction rules file doesn't exist.
  do_redact = True
  if not os.path.exists(REDACTION_RULES_FILE):
    LOG.error("Redaction rules file doesn't exist, not redacting logs. file: %s, directory: %s",
              REDACTION_RULES_FILE, os.getcwd())
    do_redact = False

  # Write an EXEC_MARKER file.  While this file exists, the actual exec of the
  # target process may not occur. Only when this file is deleted do we know
  # that we have a process in place to perform a wait() call to capture the
  # exit code.  Without doing this, there are various wait() calls made by
  # python libraries (e.g. ctypes calls subprocess.Popen/wait, which
  # then calls os.wait().  These intervening calls to wait() may pick up the
  # exit code of our watched process, and subsequently the wait() call we
  # intend to make in watch() will fail because the process we are wanting
  # to wait for has already been handled by a prior wait() call (which we knew
  # nothing about).
  with os_ops.open_write_file(EXEC_MARKER) as f:
    f.write("")


  # Turn off buffering on stdout.
  sys.stdout = os.fdopen(sys.stdout.fileno(), 'w', 0)
  pipe_stdout = os.pipe()
  pipe_stderr = os.pipe()

  try:
    pid = os.fork()
  except OSError:
    sys.stderr.write("Failed creating child process for redaction. Aborting.\n")
    sys.exit(1)

  if pid == 0:
    # Close read end of the pipe
    os.close(pipe_stdout[0])
    os.close(pipe_stderr[0])
    if do_redact:

      os.dup2(pipe_stdout[1], 1)
      os.dup2(pipe_stderr[1], 2)

    # Wait for EXEC_MARKER to go away.
    while os.path.exists(EXEC_MARKER):
      time.sleep(0.01)

    sys.stdout.flush()
    sys.stderr.flush()
    os.execve(args[0], args, os.environ)
    assert False, "Unexpected return after exec"


  # Close write end of the pipe
  os.close(pipe_stdout[1])
  os.close(pipe_stderr[1])

  if do_redact:
    try:
      redactor_pid = os.fork()
    except OSError as exc:
      LOG.exception("Failed creating child process for redaction. Aborting: %s", exc)
      sys.exit(1)
    if redactor_pid == 0:
      exec_redactor(pipe_stdout[0], pipe_stderr[0])
      # Doesn't return.
      assert False, "Unexpected return after exec"

  os.close(pipe_stdout[0])
  os.close(pipe_stderr[0])

  # This is equivalent to calling "run_redaction", but we'll do it via the re-exec path.
  # This ensure that the re-exec path is always executed.
  exec_watch(pid)
  # Doesn't return
  assert False, "Unexpected return after exec"

def redactor(child_stdout, child_stderr):
  """
  Core logic of the redactor.  Main thread runs stdout redaction and a second thread
  runs stderr redaction.  Continue until we get EOF on both.
  """
  LOG.info("Started redactor")

  # Turn off buffering on stdout.
  sys.stdout = os.fdopen(sys.stdout.fileno(), 'w', 0)

  # Restart this redactor on SIGUSR1.
  signal.signal(signal.SIGUSR1, lambda sig, frame: exec_redactor(child_stdout, child_stderr))

  try:
    redactor_obj = cmf.redaction.RedactionEngine()
    redactor_obj.add_policy_from_file(REDACTION_RULES_FILE)
  except ValueError as ve:
    LOG.exception("Error parsing redaction rules file. %s", ve)
    return
  except Exception as exc:
    LOG.exception("Generic error reading/parsing redaction rules file. %s", exc)
    return


  # Get file interface to the pipe to read line-by-line.
  file_stdout = os.fdopen(child_stdout, 'r')
  file_stderr = os.fdopen(child_stderr, 'r')

  # readline() method blocks when no lines are available
  # and hence it's better to run redactor on stdout and stderr in parallel
  # using threads.

  def redact_file(in_file, redactor_obj, out_file):
    while True:
      out = in_file.readline()
      if not out:
        break
      out = redactor_obj.redact(out)
      out_file.write(out)

  err_thread = threading.Thread(target=redact_file,
                                kwargs={'in_file' : file_stderr,
                                        'redactor_obj' : redactor_obj,
                                        'out_file' : sys.stderr})
  err_thread.start()

  redact_file(file_stdout, redactor_obj, sys.stdout)

  err_thread.join()
  sys.exit(0)
