# Copyright (c) 2015 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import logging

from cmf.throttling_logger import ThrottlingLogger

LOG = logging.getLogger('SafetyValve')
THROTTLED_LOG = ThrottlingLogger(LOG)

class SafetyValve(object):
  """
  A class that handles the agent safety valve configuration param. It knows how
  to process the safety valve configuration parameter value and offers a thread
  safe access to the values.
  Keys and values are stripped of whitespaces.
  """

  _SAFETY_VALVE_KEY = 'host_agent_safety_valve'

  def __init__(self):
    self._safety_valve = dict()
    self._safety_valve_string = None

  def get(self, key, default_value=None):
    return self._safety_valve.get(key, default_value)

  def get_safety_valve_dictionary(self):
    """
    returns a copy of the current safety valve dictionary.
    """
    current_safet_valve = self._safety_valve
    return dict(current_safet_valve)

  def update_safety_valve(self, host_configuration):
    """
    Update the safety valve configuration if needed. If the safety valve has not
    changed this is very fast. If it has changed some string manipulation is
    done but should be very fast as well.
    """
    if host_configuration is None:
      raise Exception("host configuration cannot be None")

    new_safety_valve_string = \
      host_configuration.get(SafetyValve._SAFETY_VALVE_KEY, None)

    if self._safety_valve_string == new_safety_valve_string:
      return

    LOG.info("Updating agent safety valve with %s" % new_safety_valve_string)
    self._safety_valve_string = new_safety_valve_string
    self._update_safety_valve_values()

  def _update_safety_valve_values(self):
    new_safety_valve = dict()
    if self._safety_valve_string is None:
      self._safety_valve = new_safety_valve
      return

    # The safety valve is a new line separated key values pairs.
    try:
      for line in self._safety_valve_string.splitlines():
        parts = line.split("=", 1)
        if len(parts) != 2:
          continue
        key = parts[0].strip()
        value = parts[1]
        if len(key) == 0:
          continue
        new_safety_valve[key] = value
      self._safety_valve = new_safety_valve
    except:
      THROTTLED_LOG.exception("Could not parse safety valve: %s" %
                              (self._safety_valve_string,))

