#!/usr/bin/env python
from __future__ import absolute_import
# Copyright (c) 2011-2012 Cloudera, Inc. All rights reserved.
import cherrypy
import cherrypy._cpserver
import os
import errno
import socket
import logging
import mimetypes
import pwd
import signal
import sys
import cheroot.ssl
import cheroot.makefile
import M2Crypto.SSL
import M2Crypto.threading

import simplejson as json

import cmf.agent
import cmf.logging
import cmf.version
import cmf.log_search
import cmf.log_retrieve
import cmf.fencing_server
import cmf.redaction
import cmf.util.console
import cmf.https


from cmf.agent_log_compressor import compress_agent_logs
from cmf.log_compressor import get_compressed_logs
from cmf.template_util import render_html
from cmf.util import DEFAULT_LOG_FILE_TAIL
from cmf.util import bytes_to_hex
from cmf.util import tail
from cmf.util import validate_and_open
from cherrypy.lib.static import serve_fileobj
from cmf.monitor.impalad.query_monitor import QUERY_LOG_SUBDIR, QUERY_LOG_FILE_NAME_PREFIX
from cmf.monitor.stacks.compressing_rotating_logger import get_sorted_log_files
import requests

LOG = logging.getLogger(__name__)
AGENT_LOG_TYPE = cmf.log_search.SCMAGENT
UTF8 = "utf-8"

TEN_SECONDS_IN_MILLIS=10 * 1000

EXTERNAL_STATUS_SERVER = None

class PutWrapper(object):
  """
  Object to wrap a function that takes a string argument as a handler
  for a CherryPy PUT handler.
  """
  exposed = True
  def __init__(self, action):
    self.action = action

  def PUT(self):
    body = cherrypy.request.body.read()
    return self.action(body)


class StatusServer(object):
  def __init__(self, cfg):
    self.cfg = cfg
    self.host = cherrypy.config.get('server.socket_host')
    self.port = cherrypy.config.get('server.socket_port')
    self.hostname = socket.gethostname()
    self.token = ""
    self.token_path = self.cfg.paths.auth_token

  def _check_token(self):
    """
    Checks that the token given in the request cookie matches the agent's token.

    Throws appropriate error otherwise. Note that since the agent's traffic
    is not encrypted, anyone can sniff the token off the wire...
    """
    cookie = cherrypy.request.cookie
    if "token" not in cookie and self.cfg.optional_token:
      return

    if "token" not in cookie or self.token != cookie["token"].value:
      if "token" in cookie:
        with open(self.token_path) as token_file:
          self.token = token_file.read().strip()
        if self.token == cookie["token"].value:
          return

      raise cherrypy.HTTPError(403, "Missing or invalid token.")


def check_token(fn):
  """
  Ensure request checks the auth token.
  """
  def wrapper(status_server, *pargs, **kwargs):
    status_server._check_token()
    return fn(status_server, *pargs, **kwargs)
  return wrapper

class StatusServerInternal(StatusServer):

  def __init__(self, agent, wakeup_fd, eval_context):
    self.agent = agent
    self.eval_context = eval_context
    self.wakeup_fd = wakeup_fd
    super(StatusServerInternal, self).__init__(agent.cfg)

  def __del__(self):
    if self.wakeup_fd is not None:
      os.close(self.wakeup_fd)
      self.wakeup_fd = None

  @cherrypy.expose
  @check_token
  def python_console(self):
    port = cmf.util.console.ServerHandler.StartOrConnect(self.eval_context, self.cfg)
    return str(port)

  @cherrypy.expose
  @check_token
  def logs(self):
    """
    Returns the last agent.BUFFER_SIZE characters of the agent's log. Newest
    entries are at the bottom.
    """
    cherrypy.response.headers["Content-Type"] = "text/plain"
    buf = cmf.logging.get_log_buffer()
    if buf:
      return str("\n".join(buf))
    else:
      return "No log buffer found."

  @cherrypy.expose
  @check_token
  def firehoses(self):
    """
    Returns the last firehoses reports.
    """
    return render_html("firehoses.html",
        dict(version=cmf.version.VERSION, agent=self.agent))

  @cherrypy.expose
  @check_token
  def metrics(self):
    """
    Returns the agent registered metrics.
    """
    return render_html("metrics.html",
        dict(version=cmf.version.VERSION, agent=self.agent))

  @cherrypy.expose
  @check_token
  def index(self):
    try:
      self.agent.lock.acquire()

      # Set the token in the cookie so that accessing internal pages would work
      # without having to proxy through the CM server
      cherrypy.response.cookie["token"] = self.token
      return render_html("index.html",
          dict(version=cmf.version.VERSION, agent=self.agent))
    finally:
      self.agent.lock.release()

  @cherrypy.expose
  @check_token
  def heartbeat(self):
    """
    Request an explicit heartbeat. The heartbeat is not returned in the body of
    the reply, but rather sent through the normal channels.
    """
    if self.wakeup_fd is None:
      raise cherrypy.HTTPError(503)
    try:
      os.write(self.wakeup_fd, "1")
    except:
      LOG.exception("Failed to write to wakeup pipe.")
      raise cherrypy.HTTPError(500)
    return "OK"

  @cherrypy.expose
  @check_token
  def loglevel(self, **params):
    '''
    Update log levels of loggers as defined in "params".
    - All loggers specified must be valid and already exist.
    - All log levels must be valid log levels or integers.
    :param params: dict of logger_name:loglevel
    :return:
    '''
    root = logging.getLogger()
    apply = {}

    for (k, v) in params.iteritems():
      try:
        v = int(v)
      except ValueError:
        if v in vars(logging) and isinstance(vars(logging)[v], int):
          v = vars(logging)[v]
        else:
          LOG.error("Invalid loglevel: %s", v)

      if not k in root.manager.loggerDict:
        LOG.error("Request to updated undefined logger: %s", k)
        raise cherrypy.HTTPError(404)
      apply[k] = v
    for (k, v) in apply.iteritems():
      logging.getLogger(k).setLevel(v)
      LOG.info("Set loglevel: %s = %s", k, v)


class StatusServerExternal(StatusServer):
  def __init__(self, cfg):
    super(StatusServerExternal, self).__init__(cfg)
    self.redaction_policy = self.cfg.paths.redaction_policy_string

  def reload_config(self):
    try:
      with open(self.redaction_policy) as redaction_policy_file:
        redaction_policy_str = redaction_policy_file.read().strip()
      cmf.redaction.REDACTOR.set_policy_from_string(redaction_policy_str)
      LOG.info("Successfully applied redaction policy from file %s", self.redaction_policy)
    except:
      LOG.exception("Error setting redaction policy")

  def _validate_and_open(self, path, no_links=True,
                         restrict_to_parent_path=True, valid_owners=None,
                         path_blacklist=None):
    """
    Validates and opens a file, providing the agent directory as the parent
    directory if the caller wants to restrict file access with the process
    directory. See util.validate_and_open.
    """
    parent_path = None
    if restrict_to_parent_path:
      parent_path = self.cfg.paths.base_dir

    try:
      return validate_and_open(path, parent_path, no_links, valid_owners,
                               path_blacklist)
    except OSError, e:
      if e.errno == errno.ENOENT:
        raise cherrypy.NotFound(path)
      else:
        raise cherrypy.HTTPError(500, str(e))
    except Exception, e:
      raise cherrypy.HTTPError(500, str(e))

  def _serve_file(self, relativepath):
    """
    Serves a file relative to the agent process folder.

    Note that because of cherrypy, this happens
    to respect HTTP Range headers.
    """
    path = os.path.join(self.cfg.paths.base_dir, "process", relativepath)
    (content_type, _) = mimetypes.guess_type(path)

    f = self._validate_and_open(path)
    return serve_fileobj(f, content_type or "text/plain")

  def _collect_agent_log_files(self):
    """
    Collects all the files from agent log directory.
    """
    log_dir = self.cfg.log_dir
    log_files = set()
    for file in os.listdir(log_dir):
      file_path = os.path.join(log_dir, file)
      if os.path.isfile(file_path):
        log_files.add(file_path)
    return log_files

  @cherrypy.expose
  @check_token
  def process(self, process_name, *args, **kw):
    """
    Handle requests for http://host/process/<process_name>/<op>
    where process_name is the process name and op is the particular operation
      op: view a file from the process_name private store (config directory)
    """
    op = args[0]
    type = args[1]
    args = args[1:]
    if op == "files" and type == "logs":
      return self._serve_file(os.path.join(process_name, *args))
    elif op == "files" and type == "host_statistics":
      return self._serve_file(os.path.join(process_name, args[1]))
    elif op == "files" and type == "inspector":
      return self._serve_file(os.path.join(process_name, "output.json"))
    elif op == "files" and type == "distcp":
      return self._serve_file(os.path.join(process_name, "counters.txt"))
    elif op == "files" and type == "existing_configs":
      return self._serve_file(os.path.join(process_name, "configs.zip"))
    elif op == "files" and type == "diagnostics_text":
      return self._serve_file(os.path.join(process_name, "output"))
    elif op == "files" and type == "diagnostics_binary":
      return self._serve_file(os.path.join(process_name, "output.gz"))
    elif op == "files" and type == "diagnostics_linked_data":
      return self._serve_and_delete_diagnostics_linked_data(process_name)
    else:
      # unknown op
      msg = "Invalid request. Process: %s %s %s" % (process_name, op, args)
      raise cherrypy.HTTPError(404, msg)

  def _serve_and_delete_diagnostics_linked_data(self, process_name):
    """
    Helper function to serve gzipped diagnostic data generated outside the
    process directory. The expectations here are that the script that generated
    the data produces a symlink 'output.gz' to the actual data file. This call
    will read that symlink to find the actual path to the data file. It will
    verify that it is owned by the correct user (determined by looking at the
    owner of the process directory), and that it is not itself a symlink. The
    file will then be served. Once the file is served, the file is deleted. (We
    delete the file to avoid accumulating data on the hosts. In practice this
    data is usually copied to the CM command storage subsystem and so is still
    available to CM if it needs it).
    """

    # We expect the data file to be owned by the same user as the associated
    # process. We get that user by statting the process directory.
    process_directory = os.path.join(self.cfg.paths.base_dir, "process", process_name)
    valid_owners = [pwd.getpwuid(os.stat(process_directory).st_uid)[0]]

    # We expect a symlink called output.gz to the real file in the process
    # directory. We do a readlink to get the actual data filename. This errors
    # out if output.gz is not a symlink, which is desirable. The check we do
    # before the read link call is merely to give a nicer error message in the
    # log. It is the readlink behavior we depend on for correctness.
    symlink = os.path.join(process_directory, "output.gz")
    if not os.path.islink(symlink):
      LOG.error("%s is not a symlink." % symlink)
      raise cherrypy.HTTPError(500)
    data = os.readlink(symlink)

    data_directory, data_filename = os.path.split(data)
    data_directory = os.path.realpath(data_directory)
    data = os.path.join(data_directory, data_filename)

    f = self._validate_and_open(data,
                                restrict_to_parent_path=False,
                                valid_owners=valid_owners)

    try:
      content_type, _ = mimetypes.guess_type(data)
      return serve_fileobj(f, content_type or "text/plain")
    finally:
      os.remove(data)

  def parse_search_logs_args(self,
                             start_time,
                             end_time,
                             log_paths,
                             log_types,
                             role_ids,
                             log_level,
                             search_timeout,
                             query,
                             role_result_limit,
                             total_byte_limit):

    # These are required args and should throw an exception if not specified.
    start_time = int(start_time)
    end_time = int(end_time)
    log_paths = log_paths.split(",")
    log_types = log_types.split(",")
    role_ids = role_ids.split(",")

    if log_level is not None:
      log_level = cmf.log_search.parse_log_level(log_level)

    if search_timeout is not None:
      search_timeout = int(search_timeout)

    if role_result_limit is not None:
      role_result_limit = int(role_result_limit)

    if total_byte_limit is not None:
      total_byte_limit = int(total_byte_limit)

    # No parsing needed for query

    return {
      "start_time": start_time,
      "end_time": end_time,
      "log_paths": log_paths,
      "log_types": log_types,
      "role_ids": role_ids,
      "log_level": log_level,
      "search_timeout": search_timeout,
      "query": query,
      "role_result_limit": role_result_limit,
      "total_byte_limit": total_byte_limit
    }

  @cherrypy.expose
  @check_token
  def estimate_logs(self,
                    start_time,
                    end_time,
                    log_paths=None,
                    log_types=None,
                    estimate_timeout_millis=TEN_SECONDS_IN_MILLIS):
    """
    Calculates the available bytes between the start_time and end_time for
    all files matching the log_paths. It uses log_retrieve.py for calculating
    the size.
    :param start_time: Time at which the log estimate should start
    :param end_time: Time at which the log estimate should end
    :param log_paths: List of paths to logs for which the estimate should be calculated
    :param log_types: List of types of logs, like LOG4J or GLOG, we parse each log type
     differently
    :param estimate_timeout_millis: Time to wait until we cause a timeout for estimating
    """
    if log_paths is None:
      log_paths = ""
    else:
      log_paths += ","

    if log_types is None:
      log_types = ""
    else:
      log_types += ","

    result = self.invoke_estimate_logs(
      estimate_logs_func=cmf.log_retrieve.calculate_log_estimate,
      start_time=start_time,
      end_time=end_time,
      log_paths=log_paths,
      log_types=log_types,
      estimate_timeout_millis=estimate_timeout_millis)

    cherrypy.response.headers["Content-Type"] = "application/json; charset=utf-8"
    return result["data"].encode(UTF8)

  def invoke_estimate_logs(self,
                           estimate_logs_func,
                           start_time,
                           end_time,
                           log_paths,
                           log_types,
                           estimate_timeout_millis):
    try:
      args = self.parse_estimate_logs_args(
        start_time, end_time, log_paths, log_types, estimate_timeout_millis)

      result = estimate_logs_func(
        start_time=args["start_time"],
        end_time=args["end_time"],
        log_paths=args["log_paths"],
        log_types=args["log_types"],
        estimate_timeout_millis=args["estimate_timeout_millis"])

      return result
    except Exception, e:
      LOG.exception("Log estimate failure: %s", e)
      return {
        "data" : json.dumps(dict(
          success=False,
          messages=[e.args[0]]
        ))
      }

  def parse_estimate_logs_args(self,
                               start_time,
                               end_time,
                               log_paths,
                               log_types,
                               estimate_timeout_millis):
    start_time = int(start_time)
    end_time = int(end_time)
    log_paths = log_paths.split(",")
    log_types = log_types.split(",")

    return {
      "start_time": start_time,
      "end_time": end_time,
      "log_paths": log_paths,
      "log_types": log_types,
      "estimate_timeout_millis": estimate_timeout_millis
    }

  def invoke_search_logs(self,
                         search_logs_func,
                         start_time=None,
                         end_time=None,
                         log_paths=None,
                         log_types=None,
                         role_ids=None,
                         log_level=None,
                         search_timeout=None,
                         query=None,
                         role_result_limit=None,
                         total_byte_limit=None,
                         plaintext=False):
    try:
      args = self.parse_search_logs_args(
        start_time, end_time, log_paths, log_types, role_ids, log_level,
        search_timeout, query, role_result_limit, total_byte_limit)

      result = search_logs_func(
        start_time=args["start_time"],
        end_time=args["end_time"],
        log_paths=args["log_paths"],
        log_types=args["log_types"],
        role_ids=args["role_ids"],
        log_level=args["log_level"],
        search_timeout=args["search_timeout"],
        query=args["query"],
        role_result_limit=args["role_result_limit"],
        total_byte_limit=args["total_byte_limit"],
        host=self.hostname,
        port=self.port,
        plaintext=plaintext)

      return result
    except Exception, e:
      LOG.exception("Log search failure: %s", e)
      return {
        "content_type": "application/json",
        "data": json.dumps(dict(
                  success=False,
                  messages=[e.args[0]], # the error message
                  host=self.host,
                  port=self.port))
      }

  @cherrypy.expose
  @check_token
  def search_logs(self,
                  start_time=None,
                  end_time=None,
                  log_paths=None,
                  log_types=None,
                  add_agent_logs=None,
                  role_ids=None,
                  log_level=None,
                  search_timeout=None,
                  query=None,
                  role_result_limit=None,
                  total_byte_limit=None,
                  plaintext=None):

    if self.convert_to_bool(add_agent_logs):
      if log_paths is None:
        log_paths = ""
      else:
        log_paths += ","

      if log_types is None:
        log_types = ""
      else:
        log_types += ","

      if role_ids is None:
        role_ids = ""
      else:
        role_ids += ","

      log_paths += self.cfg.logfile
      log_types += AGENT_LOG_TYPE
      role_ids += "-1"


    result = self.invoke_search_logs(
      search_logs_func=cmf.log_search.search_logs,
      start_time=start_time,
      end_time=end_time,
      log_paths=log_paths,
      log_types=log_types,
      role_ids=role_ids,
      log_level=log_level,
      search_timeout=search_timeout,
      query=query,
      role_result_limit=role_result_limit,
      total_byte_limit=total_byte_limit,
      plaintext=self.convert_to_bool(plaintext))

    cherrypy.response.headers["Content-Type"] = result["content_type"]
    return result["data"].encode(UTF8)

  @cherrypy.expose
  @check_token
  def search_logs_compressed(self,
                             start_time=None,
                             end_time=None,
                             log_paths=None,
                             log_types=None,
                             role_ids=None,
                             log_level=None,
                             search_timeout=None,
                             query=None,
                             role_result_limit=None,
                             total_byte_limit=None):
    result = self.invoke_search_logs(
      search_logs_func=cmf.log_search.search_logs_compressed,
      start_time=start_time,
      end_time=end_time,
      log_paths=log_paths,
      log_types=log_types,
      role_ids=role_ids,
      log_level=log_level,
      search_timeout=search_timeout,
      query=query,
      role_result_limit=role_result_limit,
      total_byte_limit=total_byte_limit)

    f = self._validate_and_open(result["path"], restrict_to_parent_path=False)
    return serve_fileobj(f, result["content_type"])

  @cherrypy.expose
  @check_token
  def retrieve_log_compressed(self,
                              log_path,
                              log_type,
                              end_time,
                              byte_limit,
                              search_timeout_millis=cmf.log_retrieve.TEN_SECONDS_IN_MILLIS,
                              start_time=None,
                              **kwargs_ignored):
    """Gets a compressed portion of plaintext log file at path log_path.
    log_type of the file must be specified as with search_logs.
    If start_time is not specified, the portion of the file to be collected
    will be determined by seeking to end time and reading backwards up to byte_limit bytes.
    However if specified, the portion between start time and end time will be collected and
    will not be limited by max_bytes.
    Byte limit refers to the uncompressed size.
    This function handles rolled over log files."""

    cherrypy.response.headers['Content-type'] = 'application/x-gzip'
    start_time_val = None
    if start_time is not None:
      start_time_val = long(start_time)
    return cmf.log_retrieve.retrieve_log_compressed(
      long(end_time),
      str(log_path),
      str(log_type),
      long(byte_limit),
      cmf.redaction.REDACTOR,
      search_timeout_millis=search_timeout_millis,
      start_time=start_time_val)

  retrieve_log_compressed._cp_config = {'response.stream': True}

  @cherrypy.expose
  @check_token
  def agent_log_with_context(self,
                             offset=None,
                             direction=None,
                             num_events="100"):
    return self.log_with_context(path=self.cfg.logfile,
                                 offset=offset,
                                 direction=direction,
                                 num_events=num_events,
                                 log_type=AGENT_LOG_TYPE);
  @cherrypy.expose
  @check_token
  def log_with_context(self,
                       path=None,
                       offset=None,
                       timestamp=None,
                       direction=None,
                       num_events="100",
                       log_type=None):
    """
    Shows you a single log entry in the context of its sorrounding log entries
    Make a request like so:
      /log_in_context?path=/var/log/hadoop/datanode.log&offset=0&logType=LOG4J
    If the offset is provided, it is used to find the log entry of interest.
    If the offset provided is None, the timestamp is used to find the log entry
    of interest. If both are not provided, the last entry in the log file is
    used.
    """
    try:
      if offset is not None:
        offset = int(offset)
      if timestamp is not None:
        timestamp = int(timestamp)
      num_events = int(num_events)

      result = cmf.log_search.log_with_context(
          path,
          offset,
          timestamp,
          direction,
          num_events,
          log_type,
          host=self.host,
          port=self.port)
      cherrypy.response.headers["Content-Type"] = result["content_type"]
      return result["data"].encode(UTF8)
    except Exception, e:
        cherrypy.response.headers["Content-Type"] = "application/json"
        return json.dumps(dict(
          success=False,
          message=str(e),
          usage=self.log_with_context.__doc__,
          host=self.host,
          port=self.port))

  @cherrypy.expose
  def download_agent_log(self, **kwargs):
    return self.download_log(self.cfg.logfile, **kwargs)

  def is_glog_log(self, log_file):
    return (log_file.startswith("impalad") or
            log_file.startswith("statestored") or
            log_file.startswith("catalogd") or
            log_file.startswith("kudu"))

  def convert_to_bool(self, variable):
    """
    Converts a string version of "true" or "false" to
    a boolean value. Useful for REST boolean arguments.
    """
    if variable is None:
      return False
    return variable == "true"

  @cherrypy.expose
  @check_token
  def download_log(self, path=None, **kwargs):
    """
    This function is used for downloading a log. By default the log is
    returned as a download. The optional args are:
      plaintext:
        Set to true if the log should be returned as plaintext
        string instead of an attachment.
      onlyTail:
        If only the tail of the log is desired, you can pass onlyTail=true.
        By default, the whole log file is returned.
      compress:
        To compress the log file, pass compress=true and the log file
        will be gzipped. By default the log file is not compressed.
    """

    # As agent code runs as root, we don't want to follow all symlinks
    # as somone with lower privilege can create a symlink to a privileged
    # file. We will then end up with privilege escalation security hole
    # where agent running as root will allow download of this file.
    # To prevent this, we allow symlink resolution in directory part
    # of the path and file can be a symlink as long as it is in the same
    # directory. We can then safely open log files with O_NOFOLLOW to
    # prevent exploits.
    (log_dir, log_file) = os.path.split(path)

    #coerce the booleans
    plaintext = self.convert_to_bool(kwargs.get('plaintext', None))
    onlyTail = self.convert_to_bool(kwargs.get('onlyTail', None))
    compress = self.convert_to_bool(kwargs.get('compress', None))

    new_log_file_path = None
    # We need special-casing for glog logs because these logs are accessed
    # through symlinks and validate_and_open() explicitly refuses to
    # dereference symlinks for security reasons. As a result, we need to
    # dereference them up-front and then verify that we haven't been pointed
    # somewhere unreasonable.
    #
    # This second check also means we can't just let the caller tell us that
    # it's a glog log - we need the independent check, in this case based on
    # the file name.
    if self.is_glog_log(log_file):
      new_log_file_path = os.path.realpath(path)
      (log_dir, log_file) = os.path.split(new_log_file_path)
      if not self.is_glog_log(log_file):
        raise cherrypy.HTTPError (403, "path %s is not accessible." % (path,))
    else:
      log_dir = os.path.realpath(log_dir)
      new_log_file_path = os.path.join(log_dir, log_file)

    # Simple blacklist of path prefixes that are not reasonable for log locations
    path_blacklist = [ "/boot/", "/etc/", "/usr/", "/lib/", "/lib64/", "/home/",
                       "/root/", "/sys/", "/proc/", "/dev/", "/bin/", "/sbin/" ]

    fd = self._validate_and_open(new_log_file_path,
                                 restrict_to_parent_path=False,
                                 path_blacklist=path_blacklist)

    if compress:
      if not plaintext:
        cherrypy.response.headers['Content-type'] = 'application/x-gzip'
        cherrypy.response.headers['Content-disposition'] = 'attachment; filename=\"%s.gz\"' % (log_file,)

      statinfo = os.stat(new_log_file_path)
      end_offset =  statinfo.st_size
      start_offset = 0

      if onlyTail:
        # To avoid a partial line in the compressed tail, we do
        # a little gymnastics.
        fd.seek(max(0, end_offset - DEFAULT_LOG_FILE_TAIL))
        fd.readline()
        start_offset = fd.tell()

      return cmf.log_retrieve.build_gzip_generator_single(
        fd,
        start_offset,
        end_offset,
        cmf.redaction.REDACTOR)
    else:
      if onlyTail:
        fd = tail(fd, DEFAULT_LOG_FILE_TAIL)

      if plaintext:
        return serve_fileobj(fd, "text/plain")
      else:
        return serve_fileobj(fd, "application/x-download", "attachment")

  @property
  def base_url(self):
    return "http://%s:%s/" % ("127.0.0.1", self.cfg.agent_internal_port)

  def _redirect_to_agent(self, path):
    """
    Send a request to the real agent on the specified path, making the request look just
    like the request that came in.
    """
    cookies = {'token': self.token}
    return requests.get(os.path.join(self.base_url, path), cookies=cookies).content

  @cherrypy.expose
  @check_token
  def logs(self):
    """
    Returns the last agent.BUFFER_SIZE characters of the agent's log. Newest
    entries are at the bottom.
    """
    cherrypy.response.headers["Content-Type"] = "text/plain"
    return self._redirect_to_agent("logs")

  @cherrypy.expose
  @check_token
  def index(self):
    return self._redirect_to_agent("index")

  @cherrypy.expose
  @check_token
  def firehoses(self):
    """
    Returns the last firehoses reports.
    """
    return self._redirect_to_agent("firehoses")

  @cherrypy.expose
  @check_token
  def heartbeat(self):
    """
    Triggers an immediate agent heartbeat..
    """
    return self._redirect_to_agent("heartbeat")

  @cherrypy.expose
  @check_token
  def metrics(self):
    """
    Returns the agent registered metrics.
    """
    return self._redirect_to_agent("metrics")

  @cherrypy.expose
  @check_token
  def compressed_agent_logs(self, max_bytes, top_level_dir):
    """
    Returns the last uncompressed max_bytes of the agent's logs, compressed in
    a zip file, with all log files contained in the given top_level_dir.
    """
    # In python 2.4 zipfile.py doesn't like unicode strings so we turn the
    # argument into an ascii string.
    top_level_dir = str(top_level_dir)
    log_files = self._collect_agent_log_files()
    LOG.info("Active log files: %s", log_files)
    compressed_log_path = None
    try:
      try:
        max_bytes = int(max_bytes)

        compressed_log_path = compress_agent_logs(log_files,
                                                  max_bytes,
                                                  top_level_dir,
                                                  cmf.redaction.REDACTOR)
        f = self._validate_and_open(compressed_log_path,
                                    restrict_to_parent_path=False)
        return serve_fileobj(f, "application/zip")
      except ValueError:
        msg = "Invalid request. 'max_bytes' must be an integer."
        raise cherrypy.HTTPError(400, msg)
    finally:
      if compressed_log_path:
        os.remove(compressed_log_path)

  @cherrypy.expose
  @check_token
  def compressed_impala_profile_logs(self, max_bytes, impalad_log_directory):
    """
    Returns the last uncompressed max_bytes of an impalad's profile logs as a
    compressed zip file.
    """

    max_bytes = int(max_bytes)
    if max_bytes > self.cfg.impala_profile_bundle_max_bytes:
      LOG.info("Received request for impala profile log bundle with max bytes "
               "of %d. Using this agent's impala_profile_bundle_max_bytes "
               "value of %d instead." %
               (max_bytes, self.cfg.impala_profile_bundle_max_bytes))
      max_bytes = self.cfg.impala_profile_bundle_max_bytes
    impalad_log_directory = str(impalad_log_directory)

    # As agent code runs as root, we don't want to follow all symlinks
    # as somone with lower privilege can create a symlink to a privileged
    # file. We will then end up with privilege escalation security hole
    # where agent running as root will allow download of this file.
    # To prevent this, we allow symlink resolution in directory part
    # of the path but open all actual files with O_NOFOLLOW.

    files = []
    profile_dir = os.path.join(impalad_log_directory, QUERY_LOG_SUBDIR)
    profile_dir = os.path.realpath(profile_dir)
    if os.path.exists(profile_dir):
      for f in os.listdir(profile_dir):
        if not os.path.isfile(os.path.join(profile_dir, f)):
          continue
        if not f.startswith(QUERY_LOG_FILE_NAME_PREFIX):
          continue
        files.append(os.path.join(profile_dir, f))

    # The profile log file names include the time at which they were created.
    # By sorting the list and reversing it we end up taking the most recent
    # profile logs first.
    files.sort()
    files.reverse()

    return self._serve_log_file_bundle(max_bytes, files, impalad_log_directory)

  @cherrypy.expose
  @check_token
  def compressed_stacks_logs(self, max_bytes, stacks_filename):
    """
    Returns the last uncompressed max_bytes of stack logs from the input stacks
    directory.
    """

    max_bytes = int(max_bytes)
    if max_bytes > self.cfg.stacks_log_bundle_max_bytes:
      LOG.info("Received request for a stacks log bundle with max bytes of %d. "
               "Using this agent's stacks_log_bundle_max_bytes value of %d "
               "instead." %
               (max_bytes, self.cfg.stacks_log_bundle_max_bytes))
      max_bytes = self.cfg.stacks_log_bundle_max_bytes
    stacks_filename = str(stacks_filename)

    # As agent code runs as root, we don't want to follow all symlinks
    # as somone with lower privilege can create a symlink to a privileged
    # file. We will then end up with privilege escalation security hole
    # where agent running as root will allow download of this file.
    # To prevent this, we allow symlink resolution in directory part
    # of the path but open all actual files with O_NOFOLLOW.

    directory, filename = os.path.split(stacks_filename)
    directory = os.path.realpath(directory)
    stacks_filename = os.path.join(directory, filename)

    _, files = get_sorted_log_files(directory, filename)

    return self._serve_log_file_bundle(max_bytes, files, directory)

  def _serve_log_file_bundle(self, max_bytes, files, directory):
    bundle_path = None
    file_open_function = \
        lambda f: self._validate_and_open(f, restrict_to_parent_path=False)
    try:
      bundle_path = get_compressed_logs(files,
                                        max_bytes,
                                        directory,
                                        cmf.redaction.REDACTOR,
                                        file_open_function=file_open_function)
      f = self._validate_and_open(bundle_path, restrict_to_parent_path=False)
      return serve_fileobj(f, "application/zip")
    finally:
      if bundle_path:
        os.remove(bundle_path)



def start_status_server(agent, cfg, wakeup_fd):
  """
  Returns the URL of the status server.
  """
  cherrypy.config.update({
    'server.socket_host': "127.0.0.1",
    'server.socket_port': cfg.agent_internal_port,
    'environment': 'embedded',
    # useful for development:
    # 'request.show_tracebacks': True,
  })
  app_conf = {
    '/static': {
      'tools.staticdir.on': True,
      'tools.staticdir.root': os.path.dirname(os.path.abspath(__file__)),
      'tools.staticdir.dir': 'static',
    },
  }
  # The following is much like the following quickstart,
  # but doesn't block.  Instead, we register to shut down
  # the web server at exit time.
  #   cherrypy.quickstart(StatusServer(agent), '/', config=app_conf)
  console = cmf.util.console.EvalContext(agent)
  cherrypy.tree.mount(StatusServerInternal(agent, wakeup_fd, console), "", config=app_conf)
  cherrypy.tree.mount(cmf.fencing_server.FencingServer(agent), "/fence", config=app_conf)

  # The /console/upload path is handled via a PUT handler, and thus needs to be configured
  # with the MethodDispatcher
  console_conf = { "/": {'request.dispatch': cherrypy.dispatch.MethodDispatcher()}}
  cherrypy.tree.mount(PutWrapper(console.eval),"/console/upload", config=console_conf)

  cherrypy.engine.start()
  agent.register_for_shutdown(stop_status_server)
  if (cfg.use_tls
      and cfg.client_cert_file != None
      and cfg.client_key_file != None):
    return "https://%s:%s/" % (cfg.reported_hostname, cfg.agent_httpd_port)
  return "http://%s:%s/" % (cfg.reported_hostname, cfg.agent_httpd_port)

def stop_status_server():
  cherrypy.engine.stop()
  cherrypy.engine.exit()

def gen_ssl_adapter(_ssl_context):
  """
  Generate an SSL Adapter as required by CherryPy.  CherryPy doesn't support M2Crypto as an SSL
  back-end, so we need to provide our own adapter for it.  This one is pretty bare-bones - it only
  supports the limited functionality we need for the status server.

  We need to generate a class dynamically because we need to embed the M2Crypto SSL context in the
  class definition - there's no other way for us to pass it through.
  """
  class SSLAdapter(cheroot.ssl.Adapter):
    ssl_context = _ssl_context
    def __init__(self, certificate, private_key, certificate_chain=None, ciphers=None):
          """Set up certificates, private key ciphers and reset context."""
          self.certificate = certificate
          self.private_key = private_key
          self.certificate_chain = certificate_chain
          self.ciphers = ciphers
          self.context = self.ssl_context

    def get_environ(self):
      return {}

    def bind(self, sock):
      # This is called before doing an accept.  We don't do anything pre-accept.
      return sock

    def wrap(self, sock):
      # This is called after accept().
      # This mirrors logic in M2Crypto.SSL.Connection.accept
      # It's the post-accept logic needed to handshake the SSL connection.
      ssl = M2Crypto.SSL.Connection(self.ssl_context, sock)
      ssl.addr = sock.getpeername()
      ssl.setup_ssl()
      ssl.set_accept_state()
      ssl.accept_ssl()
      check = getattr(self.ssl_context, 'post_connection_check', None)
      if check is not None:
        if not check(ssl.get_peer_cert(), ssl.addr[0]):
                raise M2Crypto.SSL.Checker.SSLVerificationError(
                    'post connection check failed')
      return (ssl, {})

    def makefile(self, sock, mode='r', bufsize=-1):
      """Return socket file object."""

      # Generate a wrapper around IO methods to convert exceptions from M2Crypto.SSL.SSLError
      # to socket.error.  cheroot expects exceptions of the latter type.
      def mk_wrapper(fn):
        def wrapper(*pargs, **kwargs):
          try:
            return fn(*pargs, **kwargs);
          except M2Crypto.SSL.SSLError as exc:
            raise socket.error(*exc.args)
          except Exception as exc:
            LOG.exception("Intercepted error: %s %s", fn.__name__, exc);
            raise
        return wrapper

      obj = cheroot.makefile.MakeFile(sock, mode, bufsize)
      for x in ["send", "recv"]:
        setattr(obj, x, mk_wrapper(getattr(obj, x)))
      return obj

  return SSLAdapter

def conn_check(cert, addr, **kwargs):
  """
  Function to log connection requests and certificate fingerprints.
  """
  if cert is None:
    LOG.warn("Rejected connection from %s", addr)
    return False
  LOG.info("Connection from: %s %s", addr, cert.get_fingerprint())
  return True

def start_external_server(cfg):
  """
  Standalone server.
  Any status data that can be provided outside the agent process (e.g. log search)
  is done by the standalone external server.
  """

  # Restart when the user runs "cm agent restart"
  signal.signal(signal.SIGUSR1, restart_signal)
  signal.signal(signal.SIGHUP, reload_config)

  cherrypy.config.update({
    'global': {
      'engine.autoreload.on': True,
      'checker.on': False,
      'tools.log_headers.on': False,
      # useful for development:
      'request.show_tracebacks': True,
      'request.show_mismatched_params': True,
      'log.screen': True,
    },
    '/static': {
      'tools.staticdir.on': True,
      'tools.staticdir.root': os.path.dirname(os.path.abspath(__file__)),
      'tools.staticdir.dir': 'static',
    }
  }
  )

  global EXTERNAL_STATUS_SERVER
  EXTERNAL_STATUS_SERVER = StatusServerExternal(cfg)
  cherrypy.tree.mount(EXTERNAL_STATUS_SERVER)
  cherrypy.server.unsubscribe()


  if (cfg.use_tls
      and cfg.client_cert_file != None
      and cfg.client_key_file != None):
    M2Crypto.SSL.Connection.DISABLE_TIMEOUT_HACK = True
    M2Crypto.threading.init()
    ssl_context = cmf.https.make_ssl_context(cfg)
    ssl_context.post_connection_check = conn_check
    cheroot.server.ssl_adapters['cm_agent_adapter'] = gen_ssl_adapter(ssl_context)

    ssl_server = cherrypy._cpserver.Server()
    # Tell CherryPy to use the ssl adapter we just registered above.
    ssl_server.ssl_module = 'cm_agent_adapter'
    ssl_server.ssl_certificate = cfg.client_cert_file
    ssl_server.ssl_private_key = cfg.client_key_file
    ssl_server.ssl_certificate_chain = cfg.verify_cert_file

    ssl_server._socket_host = cfg.listening_hostname
    ssl_server.socket_port = cfg.agent_httpd_port
    ssl_server.subscribe()
    LOG.info("Set up TLS service on port: %s", cfg.agent_httpd_port)

  else:
    server = cherrypy._cpserver.Server()
    server._socket_host = cfg.listening_hostname
    server.socket_port = cfg.agent_httpd_port
    server.subscribe()
    LOG.info("Set up non-TLS service on port: %s", cfg.agent_httpd_port)

  cherrypy.engine.start()
  cherrypy.engine.block()

def restart_signal(sig_num, stack_frame):
  '''
  Re-exec the status server.
  '''
  # pylint: disable=unused-argument
  LOG.info("Restarting status server.")
  os.execv(sys.argv[0], sys.argv)

def reload_config(sig_num, stack_frame):
  '''
  Reload config
  '''
  LOG.info("Reloading status server config")
  EXTERNAL_STATUS_SERVER.reload_config()
