# Copyright (c) 2013 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

import fcntl
import logging
import os
import signal
import subprocess
import time

from cmf.throttling_logger import ThrottlingLogger

LOG = logging.getLogger(__name__)
THROTTLED_LOG = ThrottlingLogger(LOG)

def _non_block_read(output):
  """
  Sets a file descriptor to non-blocking mode and
  returns the result of a read from the stream.
  Restores the descriptor to its original status
  """
  fd = output.fileno()
  fl = fcntl.fcntl(fd, fcntl.F_GETFL)
  fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)
  try:
    try:
      return output.read()
    except:
      return ''
  finally:
    fcntl.fcntl(fd, fcntl.F_SETFL, fl)

def subprocess_with_timeout(args, timeout,
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                            env=None, cwd=None):
  """
  Calls subprocess.Popen() and blocks for it to finish, timing out if
  'timeout' seconds elapse.
  """
  kwargs = {
    'args' : args,
    'stdout' : stdout,
    'stderr' : stderr,
    'close_fds' : True,
  }
  if env is not None:
    kwargs['env'] = env
  if cwd is not None:
    kwargs['cwd'] = cwd
  p = subprocess.Popen(**kwargs)
  t_begin = time.time()
  seconds_passed = 0
  stdout_cap = ''
  stderr_cap = ''

  # We spin here (with a 0.1 second granularity)
  # to find when the process has died and to
  # pump its output from it.  If we didn't
  # handle the output, buffered output could
  # cause it to block unnecessarily and not
  # finish.
  #
  # Even if we didn't care about output,
  # using signals (a la solutions at http://stackoverflow.com/questions/1191374/subprocess-with-timeout)
  # is problematic in a context with threads and using
  # select() on the output may be problematic too (what happens
  # when it's closed?).  I've never tried using select() on
  # the /proc/pid directory.
  while p.poll() is None and seconds_passed < timeout:
    # p.std* blocks on read(), which messes up the timeout timer.
    if stdout == subprocess.PIPE:
      stdout_cap += _non_block_read(p.stdout)
    if stderr == subprocess.PIPE:
      stderr_cap += _non_block_read(p.stderr)
    time.sleep(0.1)
    seconds_passed = time.time() - t_begin

  if seconds_passed >= timeout:
    try:
      if stdout == subprocess.PIPE:
        p.stdout.close()
      if stderr == subprocess.PIPE:
        p.stderr.close()
      # use os.kill instead of p.terminate for python 2.4 compatibility
      os.kill(p.pid, signal.SIGTERM)
    except:
      THROTTLED_LOG.exception("Kill subprocess exception with args %s" % args)

    # Make sure it's really dead
    if p.poll() is None:
      time.sleep(1.0)
      os.kill(p.pid, signal.SIGKILL)

    THROTTLED_LOG.exception("Timeout with args %s" % args)
    raise Exception("timeout with args %s" % args)

  if stdout == subprocess.PIPE:
    stdout_cap += _non_block_read(p.stdout)
  if stderr == subprocess.PIPE:
    stderr_cap += _non_block_read(p.stderr)
  return p.returncode, stdout_cap, stderr_cap
