"""
Copyright (c) 2017 Cloudera, Inc. All rights reserved.

Supervisor communications internal API wrapper.
"""
from __future__ import absolute_import

# pylint: disable=broad-except, invalid-name, missing-docstring, unused-argument, import-error, bare-except
import random
import threading
import time
import sys

import os.path
from xmlrpclib import Fault

import psutil
import re

import subprocess
import supervisor.options
import supervisor.supervisorctl

import logging

from supervisor.states import ProcessStates
from supervisor.xmlrpc import Faults, getFaultDescription

from cmf.common import RetryableProcessException
from cmf.template_util import render_text
from cmf.util import synchronized, kill_process, os_ops, write_conf_file
import cmf.version
from cmf.util.synced_action import SyncedAction

LOG = logging.getLogger(__name__)
MAX_RECENT_SUPERVISOR_FAILURES = 5
SUPERVISOR_CONNECT_WAIT_TIME_SECS = 2

# Process name for the supervisord listener.
SUPERVISOR_LISTENER_PROC = "cmflistener"

# Number of seconds to wait for a process to die
PROCESS_DEATH_CHECK_INTERVAL_SECS = 1

# Number of attempts to check if a process we're trying to kill has died
MAX_PROCESS_DEATH_CHECKS = 5

class SupervisorError(Exception):
  pass

def create_supervisord_markers_file(marker_path, os_ops_obj):
  """
  Create supervisor marker file
  """
  try:
    with os_ops_obj.open_write_file(marker_path, mode=664) as handle:
      cm_version_pattern = re.compile("(\\d+)\\.(\\w|\\d+)\\.(\\w|\\d+)")
      LOG.info("CMF version: %s", cmf.version.VERSION)
      cm_version_full = cm_version_pattern.match(cmf.version.VERSION)
      cm_version = cm_version_full.group(0)
      handle.write(cm_version)
  except:
    LOG.exception("Failed to generate the supervisor marker file.")

class SupervisorWrapper(object):
  # pylint: disable=too-many-instance-attributes

  def __init__(self, cfg, os_ops_obj=None):
    """
    Configure the supervisor client.
    We could do the XMLRPC configuration manually, but
    why bother if there's helper code to do it for us!
    """
    # If the supervisor config file is missing, supervisor_options.realize() will
    # force a sys.exit() call, which we want to avoid.
    self.supervisor_dir = cfg.paths.supervisor_dir
    conf_file = os.path.join(self.supervisor_dir, "supervisord.conf")
    if not os.path.exists(conf_file):
      raise SupervisorError("No supervisor config present")
    self.lock = threading.RLock()
    self.os_ops = os_ops_obj if os_ops_obj is not None else os_ops.OSOps(True)
    self.stats = {"latency_sec": 0, "successes": 0, "failures": 0}
    self.__supervisor_options = supervisor.options.ClientOptions()
    self.__supervisor_options.realize(args=["-c", conf_file])
    self.client = self.__supervisor_options.getServerProxy()
    self.ctl = supervisor.supervisorctl.Controller(self.__supervisor_options)
    self.identifier = self.__get_supervisor_identification()
    self.pid = self.__get_supervisor_pid()
    self.version = self.__get_supervisor_version()
    self.psutil_process = psutil.Process(self.pid)
    self.recent_supervisor_failures = 0
    LOG.info("Supervisor version: %s, pid: %s",
             self.client.supervisor.getSupervisorVersion(),
             self.pid)

  def is_running(self):
    # The __eq__ operator and "is_running" method of Process isn't safe in case of a clock
    # change.  The best we can do is just rely on the name and cmdline.  If supervisor really
    # dies, then we lose the socket connection anyways.
    try:
      proc = psutil.Process(self.pid)
      if proc.name != self.psutil_process.name or proc.cmdline() != self.psutil_process.cmdline():
        return False
    except Exception as exc:
      LOG.exception("Failed checking supervisor process: %s", exc)
      return False

    return True

  def get_lock(self):
    """
    Returns the lock for this object
    """
    return self.lock

  @classmethod
  def connect(cls, cfg, os_ops_obj):
    """
    Connect to the supervisor and try to get its process info
    If we fail to connect to the supervisor the agent exits.
    """
    for attempt in range(0, MAX_RECENT_SUPERVISOR_FAILURES):
      try:
        # give the supervisor a chance to start
        time.sleep(SUPERVISOR_CONNECT_WAIT_TIME_SECS)
        LOG.info("Trying to connect to supervisor (Attempt %d)", attempt + 1)
        obj = cls(cfg, os_ops_obj)
      except Exception as exc:
        LOG.exception("Failed! trying again in %s second(s): %s",
                      SUPERVISOR_CONNECT_WAIT_TIME_SECS, exc)
        continue

      LOG.info("Successfully connected to supervisor")
      return obj

    raise SupervisorError("Failed to connect to supervisor")

  @classmethod
  def __gen_supervisor_cfg(cls, cfg, identifier):
    """
    Generate a configuration of the supervisor.
    :return: (args, env) -- command line arguments for the supervisor and the environment to
                            exec it in.
    """
    supervisord_log = os.path.join(cfg.log_dir, "supervisord.log")
    listener_log = os.path.join(cfg.log_dir, "cmf_listener.log")

    #
    # We don't have an existing supervisor, so let's start one.
    #
    # Some services will execute scripts from inside the process directory,
    # so we mount a new tmpfs (without noexec) to guarantee this capability.
    # As doing so requires root privileges, failure is non-fatal.
    #
    # Conditioned on a new supervisor so that no running processes will
    # notice the transition.

    config = os.path.join(cfg.paths.supervisor_dir, "supervisord.conf")
    # First, we create the config
    write_conf_file(config,
                    render_text("supervisord.conf", data={
                        'supervisor_include_path': cfg.paths.supervisor_include_path,
                        'supervisord_httpd_port': cfg.supervisord_httpd_port,
                        'supervisord_username': str(random.randint(sys.maxint / 2, sys.maxint)),
                        'supervisord_password': str(random.randint(sys.maxint / 2, sys.maxint)),
                        'supervisord_logfile': supervisord_log,
                        'identifier': identifier,
                        'supervisord_events': cfg.supervisord_events,
                        'listener_name': SUPERVISOR_LISTENER_PROC,
                        'cmf_python' : sys.executable,
                        'cmf_python_path' : os.path.join(sys.prefix, "bin"),
                        'cmf_agent_path' : os.path.dirname(__file__),
                        'cmf_agent_fifo' : cfg.paths.event_pipe_path,
                        'cmf_listener_log' : listener_log,
                    }))


    def make_supervisor_env():
      """
      Curate the environment for the supervisor.
      """
      # pylint: disable=no-self-use
      res = os.environ.copy()
      # Take out things that shouldn't be there
      for censored in ('KRB5CCNAME',):
        try:
          del res[censored]
        except KeyError:
          pass
      return res

    args = [os.path.join(os.path.dirname(cfg.supervisord_path), "python"), cfg.supervisord_path]
    env = make_supervisor_env()
    return args, env

  @classmethod
  def execute(cls, cfg, identifier):
    """
    Configure and exec() the supervisor from the current process.
    Invoked when agent code is started by systemd to bootstrap the supervisor.
    :return: Doesn't.
    """
    args, env = cls.__gen_supervisor_cfg(cfg, identifier)
    # Create a file to store stdout/err for supervisord
    outfile = open(os.path.join(cfg.log_dir, "supervisord.out"), "w")

    # We are being invoked to start supervisord, not run the agent.
    # Thus we just want to do an exec() and run supervisord in the foreground (not as daemon).
    os.chdir(cfg.paths.supervisor_dir)
    args.append("-n")  # No daemonization - connected to systemd

    # Close stdin, redirect stdout/stderr
    sys.stdout.close()
    sys.stderr.close()
    os.dup2(outfile.fileno(), 1)
    os.dup2(outfile.fileno(), 2)

    sys.stdout = os.fdopen(1, "w")
    sys.stderr = os.fdopen(2, "w")
    sys.stdin.close()

    # Close all file descriptors except stdin, stdout, stderr
    for x in os.listdir("/proc/self/fd"):
      try:
        x = int(x)
        if x <= 2:
          continue
        os.close(x)
      except Exception as e:
        print "Failed closing: %s %s" % (x, e)
    os.execvpe(args[0], args, env)
    # We don't get here: exec a process or raise an exception.

    assert False

  @classmethod
  def start_daemon(cls, cfg, identifier, os_ops_obj):
    """
    # Legacy path to start supervisord from the agent (needed only for non-systemd systems) - RHEL6
    """
    args, env = cls.__gen_supervisor_cfg(cfg, identifier)
    # Create a file to store stdout/err for supervisord
    try:
      # Now start the guy
      with os_ops_obj.open_write_file(os.path.join(cfg.log_dir, "supervisord.out")) as outfile:
        p = subprocess.Popen(
            # Explicitly run supervisord with the virtualenv python.  The "relocatable"
            # env/bin/supervisord inherits site-package via a virtualenv bug/limitation.
            args=args,
            stdin=None,
            stdout=outfile,
            stderr=outfile,
            close_fds=True,
            env=env,
            cwd=cfg.paths.supervisor_dir)
        # Supervisor forks itself off immediately, so we wait on the
        # subprocess to avoid keeping a defunct process around.
        p.wait()
    except:
      LOG.exception("Failed to start supervisord daemon")

  @synchronized
  def load_supervisord_marker_file(self, fresh_start):
    """
    Loads/creates supervisor marker file and return supervisor info for agent heartbeat
    """
    sup_version = self.version
    cm_version = None
    marker_path = os.path.join(self.supervisor_dir, "__STARTING_CM_VERSION__")

    if fresh_start:
      create_supervisord_markers_file(marker_path, self.os_ops)
    elif os.path.exists(marker_path):
      try:
        f = open(marker_path, "r")
        cm_version = f.readline().rstrip("\n")
      except:
        LOG.exception("Failed to read the existing supervisor marker file.")
      finally:
        f.close()

    # Populate the supervisor info for the heartbeat.
    retval = dict()
    retval["component_name"] = "Supervisord"
    retval["sup_version"] = sup_version
    if cm_version is not None:
      retval["sup_version"] += "-" + "cm" + cm_version
    return retval

  def fetch_data_failure(self):
    self.client = self.__supervisor_options.getServerProxy()

  @synchronized
  def get_supervisor_data(self, count_failures=False):
    data = None
    try:
      sup = self.client.supervisor

      supervisor_latency_start = time.time()
      supervisor_procs = sup.getAllProcessInfo()
      supervisor_latency_end = time.time()
      data = dict((v["name"], v) for v in supervisor_procs)

      latency = supervisor_latency_end - supervisor_latency_start
      self.stats["latency_sec"] += latency
      self.stats["successes"] += 1
      # Reset failure count
      self.recent_supervisor_failures = 0

    except Exception as exc:
      self.stats["failures"] += 1
      LOG.exception("Could not contact supervisor: %s", exc)
      # The XML-RPC Proxy tends not to recover from errors, so we refresh it.
      self.fetch_data_failure()
      if count_failures:
        self.recent_supervisor_failures += 1
        if self.recent_supervisor_failures > MAX_RECENT_SUPERVISOR_FAILURES:
          LOG.error("Failed to contact supervisor after %u attempts. Agent " \
                    "will exit.", self.recent_supervisor_failures)
          sys.exit(-1)

    return data

  def reload(self):
    """
    Interface function to reload new supervisord configurations
    """
    #pylint: disable=fixme
    SyncedAction.schedule(self.__supervisor_reload_internal)

  @synchronized
  def __supervisor_reload_internal(self):
    """
    Supervisor's configuration re-reading requires
    both a re-read, and explicit starts.  supervisorctl
    seems to do this for us, whereas the XML-RPC mechanism
    requires manual work.
    """
    LOG.info("Triggering supervisord update.")
    self.ctl.onecmd("update")
    remote_id = self.client.supervisor.getIdentification()
    if self.identifier != remote_id:
      raise Exception("Failed to communicate to supervisor daemon.  Mismatched ids: %s vs %s" % (
          self.identifier, remote_id))

  @synchronized
  def __get_supervisor_version(self):
    """
    Returns the version of the current supervisord
    """
    return self.client.supervisor.getSupervisorVersion()

  @synchronized
  def __get_supervisor_pid(self):
    """
    Returns the pid of the running supervisord
    """
    return self.client.supervisor.getPID()

  @synchronized
  def __get_supervisor_identification(self):
    """
    Returns the identifying string of supervisord
    """
    return self.client.supervisor.getIdentification()

  @synchronized
  def get_process_info(self, name, retries=1):
    """
    Return information about a process managed by supervisord
    """
    while retries >= 0:
      try:
        return self.client.supervisor.getProcessInfo(name)
      except Exception as exc:
        if retries <= 0:
          raise exc
        LOG.warn("Failed while getting process info. Retrying. (%s)", exc)
        self.fetch_data_failure()
        retries -= 1

  @synchronized
  def get_all_process_info(self):
    """
    Return information about all processes managed by supervisord
    """
    return self.client.supervisor.getAllProcessInfo()

  @synchronized
  def start_process(self, name):
    """
    Launches a process via supervisord
    """
    try:
      # Get the current state of the process if exists so to make proper
      # decision how to proceed.
      info = self.get_process_info(name)
      if info['state'] != ProcessStates.RUNNING:
        # This is the retry activation but the previous activation didn't go
        # through completely. Currently the process state is other than
        # 'RUNNING'. We need to explicitly start the process.
        try:
          return self.client.supervisor.startProcess(name)
        except Fault, fault:
          if fault.faultCode != Faults.STILL_RUNNING and fault.faultCode != Faults.ALREADY_STARTED:
            raise RetryableProcessException(fault)
          return True

    except Fault, fault:
      if fault.faultCode != Faults.BAD_NAME:
        # If the supervisor went wrong, we should retry to get the process
        # info in the hope that the supervisor goes back to normal in the
        # retry rounds. It also possible that the supervisor is working but
        # the instruction may just fail for some reason. We'd also retry.
        fault_code = getFaultDescription(fault.faultCode)
        LOG.info("Error in getting process info from supervisord. "
                 "Error Code: [%s]. Need to retry!", fault_code)
        raise RetryableProcessException(fault)
      return Faults.BAD_NAME

  @synchronized
  def stop_process(self, name):
    """
    Stops a process managed by supervisord
    """
    try:
      return self.client.supervisor.stopProcess(name)
    except Fault, fault:
      # NOT_RUNNING: It's already stopped, so why worry
      # BAD_NAME: Supervisord doesn't know what we're on about.
      #           Even if it's running, we can't help.
      if fault.faultCode == Faults.NOT_RUNNING:
        return True
      elif fault.faultCode == Faults.BAD_NAME:
        LOG.info("Supervisord doesn't know about process %s. "
                 "It cannot be deactivated by this Agent, "
                 "assuming it even exists", self)
        return False
      else:
        raise RetryableProcessException(fault)

  @synchronized
  def kill(self):
    """
    Kills supervisord process if no processes are managed
    """
    procs = self.get_supervisor_data()
    if procs is not None and len(list(procs)) > 0:
      LOG.info("%u processes are being managed; Supervisor will continue to run.", len(list(procs)))
      return

    # Now shut it down
    LOG.info("Shutting down supervisord, pid %s", self.pid)
    kill_process(self.pid, MAX_PROCESS_DEATH_CHECKS, PROCESS_DEATH_CHECK_INTERVAL_SECS)
