# Copyright (c) 2012 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import
import asyncore
import errno
import getopt
import logging
import os
import signal
import sys

from supervisor.childutils import EventListenerProtocol

LOG = None

# Cribbed from agent.py.
LOG_FORMAT = '[%(asctime)s] %(process)d %(threadName)s %(module)-12s %(levelname)-8s %(message)s'
DATE_FORMAT = '%d/%b/%Y %H:%M:%S %z'

class Event(object):
  """
  Helper class used to parse an event. Since we use non-blocking I/O in the
  agent case, this class helps keep things in sync across several read() calls
  on the underlying fd. It's more complex than necessary for the blocking case
  used in the event listener, but at least the code is shared.
  """

  def __init__(self):
    self.properties = {}
    self.payload = None
    self.payload_properties = {}
    self._header = ''
    self._payload_len = None

  def feed(self, data):
    """
    Feed data into the event. If the input data has more data than this
    event expects, returns the leftover data.

    @param data: The data to feed.
    @return Data that is not part of this event.
    """
    if self._payload_len is None:
      # Haven't read the whole header yet.
      data = data.split('\n', 1)
      self._header += data[0]
      if len(data) == 2:
        # Header is finished, parse all properties, and grab the
        # payload length.
        data = data[1]
        self._populate(self._header, self.properties)
        self._payload_len = int(self.properties.get('len', 0))
        if self._payload_len:
          self.payload = ''
      else:
        # Header not finished, all data was consumed.
        return ''

    # Consume as much data as needed.
    needed = self.needed_data()
    if not needed:
      return data
    if needed >= len(data):
      self.payload += data
      return ''
    else:
      self.payload += data[:needed]
      return data[needed:]

  def needed_data(self):
    """
    Returns how many bytes the event still expects, or None if the header
    hasn't been read yet.
    """
    if self._payload_len:
      return self._payload_len - len(self.payload)
    return self._payload_len

  def _populate(self, data, property_map):
    """
    Parses the given data as properties into the
    provided property map.
    """
    if data is not None:
      for val in data.split(' '):
        val = val.split(':', 1)
        if len(val) == 2:
          property_map[val[0]] = val[1]
    return property_map

class SupervisordEventMonitor(asyncore.file_dispatcher):
  """
  An asyncore dispatcher that monitors a given file for supervisord events.
  The dispatcher will be notified of read events, and code using this
  monitor can use the consume_events() method to fetch events that have
  been read from supervisord.

  If the supervisord listener is stopped, the dispatcher will reinitialize
  itself.
  """

  def __init__(self, path):
    self.heartbeat = False
    self.path = path

    self.currEvent = Event()
    self.events = [ ]

    # For testing purposes.
    self.bytes_read = 0

    # The superclass will enable non-blocking mode, but if we don't also set
    # it now, open() will block if there's no connected writer.
    self.fifo = os.open(path, os.O_RDONLY | os.O_NONBLOCK)
    asyncore.file_dispatcher.__init__(self, self.fifo)

  def writable(self):
    return False

  def handle_read(self):
    # Read as much data as we can from the file before we're told that
    # the buffer is drained.
    buf = ''
    while True:
      try:
        buf += os.read(self.fifo, 4096)
      except OSError, ose:
        if ose.errno == errno.EWOULDBLOCK:
          break
        elif ose.errno == errno.EINTR:
          continue
        raise

    self.bytes_read += len(buf)
    while buf:
      buf = self.currEvent.feed(buf)
      if self.currEvent.needed_data() == 0:
        self.currEvent._populate(self.currEvent.payload,
                                 self.currEvent.payload_properties)
        # Event is complete, queue it and create a new one.
        LOG.debug('Event ready: %r', self.currEvent.properties)
        self.events.append(self.currEvent)
        self.currEvent = Event()

  def handle_expt(self):
    """
    Handles exception events from asyncore. Given that we're using
    file_dispatcher and the system supports poll(), exceptions will be
    limited to:
    - new asyncore: POLLPRI on the fd
    - old asyncore: POLLHUP, POLLERR, or POLLNVAL on the fd

    We definitely want to reinit for POLLHUP and POLLERR. The rest are
    questionable, but unlikely to happen in our environment.
    """
    LOG.info("Received exception event from asyncore, reopening fifo")
    self._reinit()

  def handle_close(self):
    """
    Handles close events from asyncore. Given that we're using
    file_dispatcher and the system supports poll(), close events will be
    limited to:
    - new asyncore: POLLHUP, POLLERR, or POLLNVAL on the fd,
                    most error conditions
    - old asyncore: never
    """
    LOG.info("Received close event from asyncore, reopening fifo")
    self._reinit()

  def _reinit(self):
    """
    Reinitialize the dispatcher, throwing away any in-flight events.
    """
    self.close()

    # As if this wasn't hard enough, new versions of asyncore os.dup() our
    # fd but older ones won't. So let's try to close it as best we can.
    try:
      os.close(self.fifo)
    except:
      pass

    self.__init__(self.path)

  def consume_events(self):
    """
    Returns the list of ready events. The internal list of events is emptied
    when this method is called.
    """
    ret = self.events
    self.events = []
    return ret


def main():
  """
  The supervisord event listener implementation. Reads events from stdin
  and pipes them to the file given in the command line. Keeps supervisord
  happy by printing the right things to stdout.
  """
  opts, args = getopt.getopt(sys.argv[1:], 'l:')

  logfile = None
  for o, v in opts:
    if o == '-l':
      logfile = v

  if len(args) != 1:
    sys.exit(1)

  if not os.path.exists(args[0]):
    sys.exit(0)

  if logfile:
    logging.basicConfig(filename=logfile, level=logging.INFO,
                        format=LOG_FORMAT, datefmt=DATE_FORMAT)
  LOG = logging.getLogger(__name__)

  LOG.info('Starting event listener as pid %s' % (os.getpid(),))

  # Ignore SIGPIPE from FIFO writes below; we'll use the raised exceptions
  # to handle control flow.
  signal.signal(signal.SIGPIPE, signal.SIG_IGN)

  supervisor = EventListenerProtocol()
  fifo = -1
  try:
    try:
      eof = False
      while not eof:
        supervisor.ready()

        # Read the event.
        evt = Event()
        header = sys.stdin.readline()
        if not header:
          eof = True
          break

        evt.feed(header)
        if evt.needed_data() > 0:
          evt.feed(sys.stdin.read(evt.needed_data()))

        # Open the agent FIFO (if it's not already open).
        if fifo == -1:
          try:
            fifo = os.open(args[0], os.O_WRONLY | os.O_NONBLOCK)
            LOG.info('Opened agent FIFO')
          except OSError, ose:
            if ose.errno == errno.ENXIO:
              LOG.info('Cannot open agent FIFO (agent probably dead), dropping event')
            else:
              raise

        # Write the data we just read to the FIFO if it's open.
        if fifo != -1:
          try:
            os.write(fifo, evt._header)
            os.write(fifo, '\n')
            if evt.payload:
              os.write(fifo, evt.payload)
          except OSError, ose:
            if ose.errno == errno.EPIPE:
              LOG.info('Cannot write to agent FIFO (agent probably dead), closing FIFO and dropping event')

              # Let's reopen the next time around.
              os.close(fifo)
              fifo = -1
            else:
              raise
        supervisor.ok()
    except:
      LOG.exception('Error in supervisord listener loop')
      raise
  finally:
    if fifo != -1:
      os.close(fifo)
      LOG.info('Closed agent FIFO')


if __name__ == '__main__':
  main()
else:
  LOG = logging.getLogger(__name__)
