# Copyright (c) 2011 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import
from mako.lookup import TemplateLookup

import os
"""
Utilities for rendering (with Mako) our templates.

The difference between the render_html and render_text
methods is that the former does HTML escaping by default,
to attempt to prevent against simple attacks.
"""

HTML_LOOKUP = TemplateLookup(
  directories=[os.path.join(os.path.dirname(__file__), "templates")],
  # escape HTML by default; disable with |n
  default_filters=['unicode', 'h'])

TEXT_LOOKUP = TemplateLookup(
  directories=[os.path.join(os.path.dirname(__file__), "templates")],
  default_filters=['unicode'])

def render_html(template_name, data):
  return render(template_name, data, HTML_LOOKUP)

def render_text(template_name, data):
  return render(template_name, data, TEXT_LOOKUP)

def render(template_name, data, lookup):
  """
  Renders template file using Mako with data.
  """
  template = lookup.get_template(template_name)
  return template.render_unicode(**data)

def handle_arguments(arguments):
  """
  Escaping here is challenging.
  We add spaces after newlines to make ConfigParser
  treat it like a continuation.  We also put
  arguments in quotes, but quote quotes in a funny
  way to make the arguments work.  Supervisord
  converts this into argv with shlex.split
  (in supervisor/process.py).

  This is the shlex interaction.
  >> arguments = [ "a c", "d ", "z\ny", 'z"z', "'foo" + '"foo' ]
  >> arguments2 = shlex.split(" ".join([ '"%s"' % x.replace('"', '"\'"\'"') for x in arguments ]))
  >> print arguments == arguments2
  True
  >> print arguments2
  ['a c', 'd ', 'z\ny', 'z"z', '\'foo"foo']

  """
  return " ".join([ '"%s"' % x.replace("\n", "\n ").replace('"', '"\'"\'"') for x in arguments ])

def handle_environment(environment):
  """
  Escaping here is challenging too!
  The option first travels through ConfigParser,
  and then through dict_of_key_value_pairs, which
  uses shlex.

  Newlines can be handled by adding a space after them.  (However, newlines
  followed by spaces will have the spaces collapsed.)

  >> c = ConfigParser.RawConfigParser(); c.readfp(io.BytesIO("[section]\nkey='a\r\n b\n c'\n")); list(shlex.shlex(c.get("section", "key")))
  ["'a\nb\nc'"]

  However, quotes are a lost cause.  See https://github.com/Supervisor/supervisor/issues/328
  and shlex's posix=True flag, which isn't available in all python versions.

  Given all this, we remove single quotes and add a space after newlines.
  """
  return ",".join(["%s='%s'" % (k, v.replace("'", "").replace("\n", "\n ")) for (k,v) in environment.iteritems()])
