# Copyright (c) 2012 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import

"""
A wrapper around Python's logger to provide log throttling
"""

from logging import getLogger
from time import time
from threading import Lock

class ThrottlingLogger(object):
  _MAX_MESSAGES_DEFAULT = 1000
  
  def __init__(self, logger,
               period_sec = 1800,
               max_messages = _MAX_MESSAGES_DEFAULT):
    '''Create a new throttling log wrapper around a Logger object
       The default throttling period is 1800 seconds, or 30 minutes
    '''
    self._logger = logger or getLogger()
    self._period_sec = period_sec
    self._max_messages = max_messages
    self._message_times = dict()
    self._lock = Lock()

  def get_time(self):
    '''Function to get current time, exposed for testing.'''
    return time()

  def _should_log(self, message):
    self._lock.acquire()
    try:
      last_logged = 0
      msg_count = 0
      curr_time = self.get_time()
      expired_time = curr_time - self._period_sec

      if len(self._message_times) >= self._max_messages:
        # If the dictionary is full, remove any messages that are expired
        self._message_times = dict(filter(lambda x: x[1][0] >= expired_time,
                                          self._message_times.items()))
        if len(self._message_times) >= self._max_messages:
          # If no items were cleared, just clear the dictionary
          self._logger.warning("Throttling logger is full; clearing all messages. This may"
                               " indicate the throttling logger is being used incorrectly.")
          self._message_times.clear()

      if self._message_times.has_key(message):
        (last_logged, msg_count) = self._message_times[message]

      if last_logged < expired_time:
        self._message_times[message] = (curr_time, 0)
        if msg_count == 0:
          return message
        else:
          return "(%s skipped) %s" % (msg_count, message)
      else:
        self._message_times[message] = (last_logged, msg_count + 1)
        return None
    finally:
      self._lock.release()

  def debug(self, msg, *args, **kwargs):
    msg = self._should_log(msg)
    if msg:
      self._logger.debug(msg, *args, **kwargs)

  def info(self, msg, *args, **kwargs):
    msg = self._should_log(msg)
    if msg:
      self._logger.info(msg, *args, **kwargs)

  def warn(self, msg, *args, **kwargs):
    self.warning(msg, *args, **kwargs)

  def warning(self, msg, *args, **kwargs):
    msg = self._should_log(msg)
    if msg:
      self._logger.warning(msg, *args, **kwargs)

  def error(self, msg, *args, **kwargs):
    msg = self._should_log(msg)
    if msg:
      self._logger.error(msg, *args, **kwargs)

  def exception(self, msg, *args):
    msg = self._should_log(msg)
    if msg:
      self._logger.exception(msg, *args)

  def critical(self, msg, *args, **kwargs):
    msg = self._should_log(msg)
    if msg:
      self._logger.critical(msg, *args, **kwargs)

  def log(self, level, msg, *args, **kwargs):
    msg = self._should_log(msg)
    if msg:
      self._logger.log(msg, level, *args, **kwargs)
