#!/usr/bin/env python
from __future__ import absolute_import
# Copyright (c) 2014 Cloudera, Inc. All rights reserved.
"""
Functionality for manipulating tmpfs.
"""

import logging
import os
import os.path
import psutil
import subprocess

LOG = logging.getLogger(__name__)

def _mount_tmpfs(path, is_mount, mode=None, sudo=None):
  """
  Mount or unmount a tmpfs filesystem at the desired path, optionally
  with a desired mode. If a sudo command is provided (non-root mode)
  then it will be used to (un)mount and the mount will be configured
  with the non-root user/group as the owner.

  Returns True on success, False on failure.
  """
  if is_mount:
    args = ["mount", "-t", "tmpfs", "cm_processes", path]
    mount_options = []
    if mode:
      mount_options.append("mode=0%o" % (mode,))
    if sudo:
      uid, gid = os.getuid(), os.getgid()
      mount_options.append("uid=%s" % (uid,))
      mount_options.append("gid=%s" % (gid,))
    if len(mount_options) > 0:
      args.extend(["-o", ",".join(mount_options)])
  else:
    args = ["umount", path]

  if sudo:
    args.insert(0, sudo)

  popen = subprocess.Popen(args, stderr=subprocess.PIPE)
  _, stderr = popen.communicate()
  if popen.returncode:
    LOG.warn("Failed to %s tmpfs at %s, rc: %s stderr: %s"
             % (args[0], path, popen.returncode, stderr))
  else:
    LOG.info("Successfully %sed tmpfs at %s" % (args[0], path))

  return popen.returncode == 0

def mount_tmpfs(path, clean_existing=False, sudo=None):
  """
  Mount a tmpfs at some given path. If an existing executable tmpfs is
  already mounted, it will be reused (provided clean_existing is False).

  Returns True if the mount was created (or was reused), False otherwise.
  """

  # Set up or reuse an executable tmpfs mount, if found.
  reuse = False
  for p in psutil.disk_partitions(all=True):
    try:
      # Ordering of this "if" statement is important.  The "samefile" call is invoked only
      # if the fs in question is a tmpfs mount.
      if p.fstype == "tmpfs" and "noexec" not in p.opts and os.path.exists(p.mountpoint) and os.path.samefile(p.mountpoint, path):
        if clean_existing:
          if not _mount_tmpfs(path, False, sudo=sudo):
            # Failed to unmount, reuse.
            reuse = True
        else:
          # Existing tmpfs without noexec, reuse.
          reuse = True
        break
    except Exception, exc:
      # We can't make a determination as to whether this is a new or old mount.
      # It's not safe to proceed, thus failing (and perhaps forcing the user to resolve the error)
      # is the right thing to do.
      LOG.exception("Error while comparing %s to %s: %s", p.mountpoint, path, exc)
      raise
  if reuse:
    LOG.info("Reusing mounted tmpfs at %s" % (path,))
    return True
  else:
    return _mount_tmpfs(path, True, mode=0751, sudo=sudo)
