"""
Copyright (c) 2017 Cloudera, Inc. All rights reserved.

Support for minimally disruptive upgrade from CM5.x to CM6.x.
"""
from __future__ import absolute_import


# pylint: disable=broad-except, invalid-name, missing-docstring, unused-argument, import-error
import ctypes
import logging
import signal
import sys
import time
import psutil

import os.path

import cmf.util.os_ops
from cmf.supervisor import SupervisorWrapper

LOG = logging.getLogger(__name__)

# Assuming supervisor waits 30 seconds to issue SIGKILL after SIGTERM.
# Thus we have to issue a SIGKILL in 25 seconds before proc_monitor gets
# a SIGKILL itself.
SIGKILL_WAIT_TIME = 25

os_ops = cmf.util.os_ops.OSOps(True)

def rewrite_role_supervisor_conf(cfg, proc):
  """
  Given a supervisor process, edit its supervisor.conf fragment, if it exists.
  The new config file will tell supervisod to execute "cm proc_monitor <pid>", instead of the
  real role (that command in turn translates to "monitor_process", below).

  :return: True/False indicating success
  """

  pdir = os.path.join(cfg.paths.process_dir, proc['name'])
  supervisor_conf = os.path.join(pdir, "supervisor.conf")
  conf_link = os.path.join(cfg.paths.supervisor_include_path,
                           proc['name'] + ".conf")
  if not os.path.exists(supervisor_conf):
    # Older CM5 had supervisor conf outside of process directory,
    # in the supervisor conf include directory
    if not os.path.exists(conf_link):
      LOG.info("No configuration for process %s, skipping", proc['name'])
      return False

    # These old configs can be moved to their update location and symlinked
    os_ops.move_file(supervisor_conf, conf_link)
    os_ops.symlink(conf_link, supervisor_conf)

  process = psutil.Process(proc['pid'])
  cmdline = process.cmdline()
  if "proc_watcher" in cmdline:
    index = cmdline.index("proc_watcher")
    pid = cmdline[index + 1]
  elif "proc_monitor" in cmdline:
    index = cmdline.index("proc_monitor")
    pid = cmdline[index + 1]
  else:
    pid = proc['pid']

  def editor(data):
    """
    Replace original process command line by inserting "cm proc_monitor <pid> -- " in front of it.
    """
    data = data.split('\n')
    for (idx, cline) in enumerate(data):
      if not cline.startswith("command="):
        continue

      if cline.find("cm proc_monitor") >= 0:
        # Already done
        break

      new_command = cline.replace("command=",
                                  "command=/opt/cloudera/cm-agent/bin/cm proc_monitor %s -- " % pid)
      data[idx] = new_command
      break

    return "\n".join(data)


  os_ops.install_file(supervisor_conf + ".new", supervisor_conf)
  os_ops.edit_file(supervisor_conf + ".new", editor)

  os_ops.install_file(supervisor_conf + ".old", supervisor_conf)
  os_ops.install_file(supervisor_conf, supervisor_conf + ".new")

  return True

def mark_watcher_upgrade(cfg, proc):
  """
  Create a file ".upgrade_marker" to tell a proc watcher to not die on SIGTERM, as
  they would normally do when supervisord dies (and generates a SIGTERM to the child).
  """
  marker_file = os.path.join(cfg.paths.process_dir, proc['name'], ".upgrade_marker")
  fobj = os_ops.open_write_file(marker_file)
  fobj.close()


def pre_upgrade(cfg):
  """
  Detect a supervisor.
  If we find one, then we have to kill it, while keeping the CDH roles active.
  The challenge here is that the supervisor has been modified to use the "diewithparent" library
  so that it's children will die when it dies.  That's exactly the opposite of what we want.

  To fix that we need to force each child of the supervisor to undo the diewithparent logic,
  by making the opposite "prctl" syscall.  This will be accomplished by ptrace() mechanisms
  (wrapped by the "libptrace_do" library) injecting a prctl() syscall into the running processes.
  """

  upgrade_cm5 = False
  try:
    sup = SupervisorWrapper(cfg)

    cmdline = os.path.join("/proc", str(sup.pid), "cmdline")
    with open(cmdline) as f:
      line = f.read()
      if line.find("build/env/bin/supervisord") >= 0:
        upgrade_cm5 = True

  except Exception as exc:
    LOG.warn("No existing supervisor found: %s", exc)
    return 0

  lib = ctypes.cdll.LoadLibrary(os.path.join(cfg.install_dir,
                                             "lib/libptrace_do.so"))

  def detach_proc_from_supervisor(pid):
    """
    Undo "diewithparent" logic on "pid".
    """
    ctx = lib.ptrace_do_init(pid)
    # 157 = syscall number for prctl
    # 1 = PR_SET_PDEATHSIG
    # 0 = disable
    lib.ptrace_do_syscall(ctx, 157, 1, 0, 0, 0, 0, 0)
    lib.ptrace_do_cleanup(ctx)

  for proc in sup.get_supervisor_data().values():
    if proc['name'] in ['cmflistener', 'flood', 'status_server']:
      continue
    if proc['statename'] == "RUNNING":
      LOG.info("Process: %s %s %s", proc['name'], proc['statename'], proc['pid'])
      if rewrite_role_supervisor_conf(cfg, proc):
        if upgrade_cm5:
          detach_proc_from_supervisor(int(proc['pid']))
        else:
          detach_proc_from_supervisor(int(proc['pid']))
          mark_watcher_upgrade(cfg, proc)


  # Kill supervisord without giving it a chance to shut down anything
  LOG.info("Killing old supervisord (pid %s)", sup.pid)
  os.kill(sup.pid, signal.SIGKILL)

def monitor_process(cfg, pid, command_args):
  """
  Monitor "pid" until it dies.
  Forward SIGTERM signals from us to the process.
  """

  # sigkill_time must not be a simple (int) symbol, because otherwise
  # when it is assigned to inside sighandler a new local symbol is created.
  # This is deficiency fixed in Python 3:
  # https://docs.python.org/3/reference/simple_stmts.html#nonlocal
  sigkill_time = [None]

  def reexec():
    # Process doesn't exist, so exec a new one.

    # The old config refers to CM5 install paths.
    # We have to update the command line and environment variables to the new
    # installation.

    # Append the agent bin directory to the PATH (i.e. /opt/cloudera/cm-agent/bin)
    os.environ['PATH'] = os.environ['PATH'] + ":" + os.path.split(sys.argv[0])[0]
    args = []
    # Update command line
    for x in command_args:
      y = x.replace("/usr/lib/cmf/service", cfg.package_dir)
      if os.path.exists(y):
        x = y
      else:
        y = x.replace("/usr/lib64/cmf/service", cfg.package_dir)
        if os.path.exists(y):
          x = y
      args.append(x)

    # Update environment
    for k in os.environ.keys():
      v = os.environ[k].replace("/usr/lib64/cmf/service", cfg.package_dir)
      v = v.replace("/usr/lib/cmf/service", cfg.package_dir)
      os.environ[k] = v

    # Remove old redactor prefix command.
    args = [arg for arg in args if not arg.endswith("cmf-redactor")]

    # Any processes not launched with the launcher should be updated to use it.
    # This would take effect if/when the CDH role is upgraded.
    if args[0] != sys.argv[0] or args[1] != "launch":
      # Inject a "cm launch" prefix command (which replaces the redactor)
      args = [sys.argv[0], "launch"] + args

    LOG.info("Proc monitor re-exec: %s", " ".join(args))
    os.execvp(args[0], args)

  def sighandler(sig, frame):
    LOG.info("Forwarding signal: %s to %s", sig, pid)
    os.kill(int(pid), sig)

    sigkill_time[0] = time.time() + SIGKILL_WAIT_TIME

  signal.signal(signal.SIGTERM, sighandler)

  proc_path = os.path.join("/proc", str(pid))
  try:
    psutil.Process(pid)
  except (psutil.NoSuchProcess, psutil.ZombieProcess):
    reexec()

  # We're not the parent, so we can't waitpid on an arbitrary process.
  while os.path.exists(proc_path):
    time.sleep(2)
    if sigkill_time[0] and sigkill_time[0] <= time.time():
      os.kill(int(pid), signal.SIGKILL)

  # We don't know the return value of the monitored process.
  # Only a parent or ancestor can ever know that.

  # Use a -1 exit code to trigger supervisord's auto-restart
  sys.exit(-1)
