"""
Copyright (c) 2011-2013 Cloudera, Inc. All rights reserved.

Misc. utilities
"""
from __future__ import absolute_import


# pylint: disable=too-many-lines,too-many-public-methods,too-many-statements
# pylint: disable=too-many-instance-attributes,import-error, fixme
# pylint: disable=line-too-long,bad-continuation,bare-except,broad-except,invalid-name,
# pylint: disable=missing-docstring,too-many-branches,too-many-lines

import codecs
import logging
import os
import pwd
import grp
import uuid
from string import Template
from datetime import datetime
import subprocess
import sys
import stat
import socket
import functools

try:
  from cStringIO import StringIO
except ImportError:
  from StringIO import StringIO
from zipfile import ZipFile

try:
  from hashlib import sha1 as _sha1
except ImportError:
  from sha import new as _sha1

import signal
import time
import errno

LOG = logging.getLogger(__name__)

DEFAULT_LOG_FILE_TAIL = 2048

CDH4_HUE_HOME = '/usr/share/hue'
CDH4_HCAT_HOME = '/usr/lib/hcatalog'

def sha1(path):
  """
  Calculate the SHA-1 checksum of the file at the given path. Returns the
  lower-case hex string representing the checksum.
  """
  digest = _sha1()
  f = open(path, 'r')
  try:
    buf = f.read(4096)
    while buf:
      digest.update(buf)
      buf = f.read(4096)
  finally:
    f.close()
  return digest.hexdigest().lower()


def unzip(zip_data, target_directory, uid=None, gid=None, namelist=None):
  """
  Unzips a zipfile into target_directory, with mode 600 (files) or 700
  (directories). User and group ownership is set for newly created files
  and directories.

  python2.4 has very rudimentary zipfile support;
  in newer versions this is just ZipFile.extract_all.

  namelist parameter takes a list of names to unzip. These must exactly
  match the names, including directories, that appear in the zip file.

  Returns a 2-tuple of the number of files and directories extracted from
  the archive, including those already present on disk.
  """
  # pylint: disable=too-many-locals
  count = 0
  dir_count = 0

  zip_file = None
  if zip_data != "":
    zip_file = ZipFile(StringIO(zip_data), "r")

  if namelist is None:
    if zip_file is not None:
      namelist = zip_file.namelist()
    else:
      namelist = []

  for f in namelist:
    try:
      info = zip_file.getinfo(f)
    except KeyError:
      LOG.warn("Not extracting'%s' because it's not in the archive.", f)
      continue
    target = os.path.normpath(os.path.join(target_directory, f))
    if target_directory != os.path.commonprefix([target_directory, target]):
      LOG.error("Not extracting '%s' because it wishes to escape the target directory.", f)
      continue

    # Create all parent directories to target.
    d = os.path.dirname(target)
    dirs_to_create = []
    while not os.path.exists(d):
      assert target_directory == os.path.commonprefix([d, target_directory])
      dirs_to_create.append(d)
      d = os.path.dirname(d)
    for d in reversed(dirs_to_create):
      os.mkdir(d, stat.S_IRUSR | stat.S_IXUSR | stat.S_IWUSR)
      if uid != None:
        os.chown(d, uid, gid)

    # Create the new file or directory.
    #
    # Existing files are truncated; existing directories are left as-is.
    exist = os.path.exists(target)
    if info.filename.endswith('/'):
      if not exist:
        os.mkdir(target, stat.S_IRUSR | stat.S_IXUSR | stat.S_IWUSR)
      dir_count += 1
    else:
      out = open(target, "w")
      try:
        os.chmod(target, stat.S_IRUSR | stat.S_IWUSR)
        out.write(zip_file.read(f))
      finally:
        out.close()
      count += 1
    if uid != None and not exist:
      os.chown(target, uid, gid)
  LOG.info("Extracted %d files and %d dirs to %s.", count, dir_count, target_directory)

  return (count, dir_count)

def tail(file_or_filename, length):
  """
  Reads last (at most) length of the given file or filename.
  We take some pains to make sure that the first line is after
  a newline, but we don't worry about partial lines at
  the end.
  """
  assert length >= 0
  # pylint: disable=redefined-variable-type
  if hasattr(file_or_filename, "seek"):
    f = codecs.EncodedFile(file_or_filename, "ascii", "ascii", "ignore")
  else:
    f = codecs.open(file_or_filename, 'r', "ascii", "ignore")

  f.seek(0, 2) # 2==os.SEEK_END, which isn't available on py2.4
  f.seek(-1*(min(length, f.tell())), 2)
  now = f.tell()
  t = f.read(length)
  # t is, at this point, the last 'length' of file
  # (or the entire if file, if it's too small)
  if now == 0:
    return t
  else:
    # Ignore a potential partial line
    return t[t.find("\n") + 1:]

def try_tail(filename, length=DEFAULT_LOG_FILE_TAIL):
  """
  If, for whatever reason, the log file is deleted
  or unavailable, we want to degrade gracefully.
  """
  try:
    return tail(filename, length)
  except:
    LOG.debug("Could not tail file: %s", filename)
    return "Failed to read log."

def user_group_to_ids(user, group):
  """Given username and group name, returns ids thereof."""
  return pwd.getpwnam(user).pw_uid, grp.getgrnam(group).gr_gid

def dumpstacks(signal__unused, frame__unused):
  """
  Dump stack trace for all the threads to the LOG.
  """
  # pylint: disable=unused-argument
  import threading
  import traceback
  id2name = dict([(th.ident, th.name) for th in threading.enumerate()])
  code = []
  code.append("\nDumping all Thread Stacks ...")
  # pylint: disable=protected-access
  for threadId, stack in sys._current_frames().items():
    if id2name.get(threadId) is None:
      code.append("\n# Thread: %d" % (threadId, ))
    else:
      code.append("\n# Thread: %s(%d)" % (id2name[threadId], threadId))

    for filename, lineno, name, line in traceback.extract_stack(stack):
      code.append('File: "%s", line %d, in %s' % (filename, lineno, name))
      if line:
        code.append("  %s" % (line.strip()))
  LOG.info("\n".join(code))

def register_sigquit_stack_dump():
  signal.signal(signal.SIGQUIT, dumpstacks)

def delete_quietly(filepath):
  if filepath is not None:
    try:
      os.remove(filepath)
    except:
      pass

def close_quietly(fd):
  if fd is not None:
    try:
      fd.close()
    except:
      pass

def deploy_special_file(special_file_info, src_dir, env_vars, sudo=None):
  '''
  This function applies the instructions given in special_file_info
  by changing the location, owner and permissions of a file present in
  src_dir. Sudo is optionally used to run the commands in the deplyment process.
  All env_vars starting with '$' are replaced by their values in env_vars dict.
  '''

  # Expand source file and check if it's an absolute pathname
  expanded_file = Template(special_file_info["name"]).substitute(env_vars)
  name_is_absolute = os.path.isabs(expanded_file)

  # If source file is not absolute, append to source directory
  if not name_is_absolute:
    orig_file = os.path.join(src_dir, special_file_info["name"])
  else:
    # Source is already a full pathname
    orig_file = expanded_file

  username = special_file_info["user"]
  group = special_file_info["group"]

  chown(username, group, orig_file, sudo)
  chmod(special_file_info["permissions"], orig_file, sudo)

  deploy_dir = Template(special_file_info["deploy_dir"]).substitute(env_vars)

  # create intermediate dirs with 755 perms
  if not os.path.exists(deploy_dir):
    LOG.info("Creating dir before move: %s.", deploy_dir)
    callout(["mkdir", "-p", "-m", "755", deploy_dir], sudo)

  # Get the name of the file.
  if not name_is_absolute:
    dest_file = deploy_dir + "/" + special_file_info["name"]
  else:
    # Get filename from source
    dest_file = deploy_dir + "/" + os.path.basename(expanded_file)

  callout(["cp", "-p", orig_file, dest_file], sudo)
  LOG.info("Copied special file %s to %s.", orig_file, dest_file)
  # shutil does not preserve ownership when moving / copying across file systems
  # copy2 explicitly does not copy ownership ever, so do it again
  # Redo file permissions as well.
  chown(username, group, dest_file, sudo)
  chmod(special_file_info["permissions"], dest_file, sudo)

def chown(username, group, f, sudo):
  callout(["chown", "%s:%s" % (username, group), f], sudo)

def chmod(permissions, f, sudo):
  callout(["chmod", oct(permissions), f], sudo)


def callout(args, sudo=None):
  if sudo:
    args.insert(0, sudo)

  child = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
  out_str, err_str = child.communicate()
  if child.returncode != 0:
    if len(out_str) > 0:
      LOG.warn("stdout: %s", out_str)
    if len(err_str) > 0:
      LOG.warn("stderr: %s", err_str)
    raise Exception("Failed to run command %s" % (args,))

  return child.returncode

def bytes_to_hex(b):
  return ''.join(["%02x" % ord(x) for x in b])

def hex_to_bytes(h):
  val = ''
  while len(h):
    val += chr(int(h[0], 16) * 16 + int(h[1], 16))
    h = h[2:]
  return val

def validate_file_owner_from_stat(fstat, valid_owners):
  """
  Checks if the owner of the file the stat belongs to is in the list of
  valid owners.
  """
  uid = fstat.st_uid
  try:
    user = pwd.getpwuid(uid)[0]
  except KeyError:
    raise Exception("Unable to find username with uid %d" % (uid,))

  if user not in valid_owners:
    raise Exception("Invalid file owner: %s" % (user,))

def validate_file_owner(path, valid_owners):
  """
  Checks if the owner of the file specified by the path is in the list of
  valid owners.
  """
  st = os.stat(path)
  validate_file_owner_from_stat(st, valid_owners)

def validate_and_open_fd(path, parent_path=None, no_links=True,
                         valid_owners=None, path_blacklist=None):
  """
  Returns a file descriptor after performing some validations on the path and file.
  The validations are:
    - The file exists
    - Whether links should be followed (no symlinks, no files with more than
      one hard link).
    - Whether to restrict the path to be within a specified parent directory
    - Restrict the owners of the file to set of valid owners

  If any of these validations fail an exception is thrown.

  The file may also be a directory.
  """
  # Get rid of any up level references ("..")
  path = os.path.normpath(path)
  if parent_path is not None:
    parent_path = os.path.normpath(parent_path)

  # Check file falls within the process directory if needed
  if parent_path is not None:
    if not path.startswith(parent_path + '/'):
      raise Exception("Path is not within process directory: %s" % (path,))

  flags = os.O_RDONLY
  if no_links or parent_path is not None:
    # Don't follow symlinks
    flags |= os.O_NOFOLLOW

  if path_blacklist is not None:
    for prefix in path_blacklist:
      if path.startswith(prefix):
        raise Exception("Path is under blacklisted prefix: %s" % (path,))

  fd = os.open(path, flags)
  try:
    st = os.fstat(fd)

    # If we don't want to follow links, we want to prevent hardlinks. Since
    # anyone can create a hardlink to any file (ubuntu being one exception)
    # the only thing we can do is disallow giving access to files that have
    # more than one hardlink. Note that directories have multiple hardlinks
    # by their nature and can't be hardlinked by users, so skip this test
    # for them.
    if (no_links or parent_path is not None) and not stat.S_ISDIR(st.st_mode):
      if st.st_nlink > 1:
        raise Exception("Path has more than one hardlink: %s" % (path,))

    # Validate owner if needed
    if valid_owners is not None:
      validate_file_owner_from_stat(st, valid_owners)
  except Exception, e:
    os.close(fd)
    raise e

  return fd

def validate_and_open(path, parent_path=None, no_links=True,
                      valid_owners=None, path_blacklist=None):
  """
  Returns a file object after performing some validations on the path and file.

  See {@link validate_and_open_fd} for details on the validations performed.

  If any of these validations fail an exception is thrown.

  While validate_and_open_fd can handle a directory, this method can only
  handle files as python file objects cannot wrap a directory fd.
  """
  return os.fdopen(validate_and_open_fd(path, parent_path, no_links,
                                        valid_owners, path_blacklist))

def source(path, command, caller_env):
  """
  Source a shell script in a subprocess and extract its environment and return
  it as a dict. For our specific purposes we need to support executing a command
  after sourcing the script to correctly populate the environment.
  """
  for k, v in caller_env.iteritems():
    if v is None:
      LOG.warn("None environment value: %s=%s", k, v,)

  pipe = subprocess.Popen(['/bin/bash', '-c', ". %s; %s; env" % (path, command)],
                          stdout=subprocess.PIPE, env=caller_env)
  data = pipe.communicate()[0]

  try:
    return dict((line.split("=", 1) for line in data.splitlines()))
  except Exception, e:
    LOG.exception("Failed to parse environment variables: " + data)
    raise e

def get_host_uuid(lib_dir):
  """
  Retrieve what we consider the host's uuid.

  There's no such thing as an absolute host uuid. There are a couple of credible
  choices - each with their own (dis)advantages.

  See here for a reasonable discussion of the issue:
  http://0pointer.de/blog/projects/ids.html

  For us, the main consideration is between an id that survives motherboard
  changes and one that survives OS data loss . In our context, the former is
  probably more useful, as a wiped machine probably can't replace itself
  seamlessly, while after a motherboard swap, it can.

  However, the dbus uuid is very easy to bake into a VM image - it'll happen
  automatically unless you go out of your way to avoid it - so you pretty
  much have to assume all VMs have the same id - thereby making it useless.
  As a result, we're going to generate our own uuid file for the Agent and
  use that.
  """
  uuid_path = os.path.join(lib_dir, "uuid")
  new_uuid = uuid.uuid4()

  while not os.path.exists(uuid_path) or os.stat(uuid_path).st_size == 0:

    try:
      fd = os.open(uuid_path, os.O_WRONLY | os.O_CREAT | os.O_EXCL, 0644)
      uuid_file = os.fdopen(fd, 'w')
    except OSError, exc:
      if exc.errno != errno.EEXIST:
        raise
      LOG.exception("Failed attempt to create UUID file: %s %s", uuid_path, exc)
      time.sleep(0.1)
      continue

    LOG.info("Generating Agent UUID: '%s'", uuid_path)
    try:
      uuid_file.write(str(new_uuid))
    except Exception, exc:
      LOG.exception("Failed to generate and store Agent UUID: %s", exc)
      try:
        os.unlink(uuid_path)
      except:
        pass
      raise
    finally:
      uuid_file.close()

  try:
    uuid_file = open(uuid_path, 'r')
  except:
    LOG.exception("Unable to open UUID file: '%s'.", uuid_path)
    raise

  try:
    ret = uuid_file.read()
    if "\n" in ret:
      LOG.warning(
        "UUID file (%s) contains a newline (possibly trailing). It is recommended that UUIDs are whitespace-less strings.",
      uuid_path)
    elif "\t" in ret or " " in ret:
      LOG.warning(
        "UUID file (%s) contains whitespace. It is recommended that UUIDs are whitespace-less strings.", uuid_path)
    uuid_mtime = os.path.getmtime(uuid_path)
    LOG.info("Agent UUID file was last modified at %s", datetime.fromtimestamp(uuid_mtime))
    return ret
  finally:
    uuid_file.close()

def fix_hue_home(env, alternate=CDH4_HUE_HOME):
  """
  Check the given dictionary to make sure CDH_HUE_HOME exists, otherwise
  replace it with the alternate, which defaults to the CDH4 hue home
  """
  if os.path.exists(env['CDH_HUE_HOME']):
    return
  env['CDH_HUE_HOME'] = alternate

def fix_hcat_home(env, alternate=CDH4_HCAT_HOME):
  """
  Check the given dictionary to make sure CDH_HCAT_HOME exists, otherwise
  replace it with the alternate, which defaults to the CDH4 hcat home
  """
  if os.path.exists(env['CDH_HCAT_HOME']):
    return
  env['CDH_HCAT_HOME'] = alternate

def catch_all(return_val=None):
  """
  Decorator to catch all uncaught exceptions thrown by function it decorates
  and print the error to the log. Used to make sure that no exception is
  propagated out of a function (which may get uncaught and kill the thread,
  useful when caller is an outside module).
  """
  def decorator(fn):
    def wrapper(self, *args, **kwargs):
      try:
        return fn(self, *args, **kwargs)
      except Exception, ex:
        logging.getLogger(fn.__module__).exception("Unknown exception occurred: %s", ex)
        return return_val

    return wrapper

  return decorator

def synchronized(fn):
  """
  Decorator to provide a java style method level "synchronized" facility.
  The object that uses this decorator should expose a "get_lock" method.
  """
  def new_fn(self, *args, **kwargs):
    self.get_lock().acquire()
    try:
      return fn(self, *args, **kwargs)
    finally:
      self.get_lock().release()
  functools.update_wrapper(new_fn, fn)
  return new_fn

def enum(**enums):
  """
  Function to create a python enumeration class. Create it like:
    Houses = enum(GRYFFINDOR=1, SLYTHERIN=2, HUFFLEPUFF=3, RAVENCLAW=4)
  and use it like:
    potters_house = Houses.GRYFFINDOR
  """
  return type('Enum', (), enums)


def dict_diff(a, b, key_limit=None, key_exclude=set(), key_no_show=set()):
  '''
  Calculate a diff of two dictionaries.  The result is a dictionary where
  each key contains a tuple of the values from a and b if they differ.
  :param a:
  :param b:
  :param key_limit: Limit the result to include only these keys.
  :param key_exclude: Explicitly don't include these keys in the result.
  :param key_no_show: Do not include values of differences for these keys.
  :return:
  '''
  # pylint: disable=dangerous-default-value
  if a is None or b is None:
    return None
  k = set(a.keys())
  k |= set(b.keys())
  if key_limit is None:
    key_limit = k

  k = k - key_exclude
  out = {}
  for key in sorted(k):
    if key not in key_limit:
      continue
    x = a.get(key, None)
    y = b.get(key, None)
    if x != y:
      if key in key_no_show:
        out[key] = None
      else:
        out[key] = (x, y)
  return out

def write_conf_file(path, content):
  '''
  Write "content" to a file named "path"; ensure resulting file is
  only owner-readable.

  :param path:
  :param content:
  :return:
  '''
  try:
    f = open(path, "w")
    os.chmod(path, stat.S_IRUSR | stat.S_IWUSR)
    f.write(content.encode('utf-8'))
  finally:
    f.close()


def kill_process(pid, max_death_checks, death_check_interval):
  """
  Kills the process with the given pid, first by sending SIGTERM and if
  that fails we resort to SIGKILL
  """
  # First let's ask it nicely
  try:
    os.kill(pid, signal.SIGTERM)
  except OSError, error:
    LOG.exception("Failed to kill process with pid %d", pid)
    return

  # Check a few times to see if the process vanished
  for _ in range(0, max_death_checks):
    time.sleep(death_check_interval)
    try:
      LOG.info("waiting for process to terminate...")
      os.kill(pid, 0)
    except OSError, error:
      if error.errno == errno.ESRCH:
        LOG.info("Successfully killed process with pid %d", pid)
        return
    continue

  # The process is not cooperating, really kill it
  LOG.info("Killing process with pid %d", pid)
  os.kill(pid, signal.SIGKILL)


def get_max_int():
  """
  Safely returns largest positive integer supported by Python.
  """
  try:
    max_int = sys.maxint
  except AttributeError:
    # Using sys.maxsize since sys.maxint is deprecated in python 3.0
    max_int = sys.maxsize
  return max_int

def remove_agent_path(path):
  """
  Removes agent's path from supplied colon delimited PATH environment variable.
  :param path: Not none PATH environment variable.
  :return: PATH environment less agent's path.
  """
  p = path.split(":")

  while True:
    try:
      p.remove(os.path.split(sys.argv[0])[0])
    except ValueError:
      break
  return ":".join(p)

_fqdn = None
_ip_address = None
def default_hostname():
  """
  Return the system hostname, to be assumed to be the default host name reported by the agent.
  :return: ( FQDN, ip address)
  """

  # Cache the result
  # pylint: disable=global-statement
  global _fqdn, _ip_address
  if _fqdn is not None and _ip_address is not None:
    return _fqdn, _ip_address
  try:
    _fqdn = socket.getfqdn()
    _ip_address = socket.gethostbyname(_fqdn)
    return _fqdn, _ip_address
  except:
    LOG.exception("Could not determine hostname or ip address; proceeding.")
    return None, None

