"""
Copyright (c) 2017 Cloudera, Inc. All rights reserved.

Installs Auto-TLS certificates on this host and configures agent config to
enable TLS if needed.
"""
from __future__ import absolute_import

import logging
import os
import re
import shutil
from subprocess import Popen, PIPE, STDOUT
import tarfile
import tempfile

# Location of agent config.ini
from cmf.config import DEFAULT_CONFIG_FILE

# Auto-TLS filenames
from cmf.tools.cert import GLOBAL_CA_CERT_FILE_NAME, GLOBAL_TRUSTSTORE_FILE_NAME, \
  IN_CLUSTER_CA_CERT_FILE_NAME, IN_CLUSTER_TRUSTSTORE_FILE_NAME, \
  HOST_KEY_FILE_NAME, HOST_CERT_CHAIN_FILE_NAME, HOST_KEY_CERT_CHAIN_FILE_NAME, \
  HOST_KEY_PW_FILE_NAME, HOST_KEYSTORE_FILE_NAME

# Location of certs installed by Auto-TLS
DEFAULT_CERT_DIR = '/var/lib/cloudera-scm-agent/agent-cert'

# Location of CMCA on CM host
DEFAULT_CMCA_DIR = '/opt/cloudera/CMCA'

# List of configs set by Auto-TLS
AUTO_TLS_CONFIG_SECTION = 'Security'
AUTO_TLS_CONFIGS = {'client_cert_file': DEFAULT_CERT_DIR + '/' + HOST_CERT_CHAIN_FILE_NAME,
                    'verify_cert_file': DEFAULT_CERT_DIR + '/' + IN_CLUSTER_CA_CERT_FILE_NAME,
                    'client_key_file': DEFAULT_CERT_DIR + '/' + HOST_KEY_FILE_NAME,
                    'client_keypw_file': DEFAULT_CERT_DIR + '/' + HOST_KEY_PW_FILE_NAME,
                    'cm_auto_cert_dir': DEFAULT_CERT_DIR}
USE_TLS_REGEX = re.compile(r'^#*\s*(use_tls=).*$')

# Files in cert tar generated by certmanager
AUTO_TLS_FILES = [GLOBAL_CA_CERT_FILE_NAME,
                  GLOBAL_TRUSTSTORE_FILE_NAME,
                  IN_CLUSTER_CA_CERT_FILE_NAME,
                  IN_CLUSTER_TRUSTSTORE_FILE_NAME,
                  HOST_KEY_FILE_NAME,
                  HOST_CERT_CHAIN_FILE_NAME,
                  HOST_KEY_CERT_CHAIN_FILE_NAME,
                  HOST_KEY_PW_FILE_NAME,
                  HOST_KEYSTORE_FILE_NAME]

# This subset of files also exists in DEFAULT_CMCA_DIR
CM_HOST_AUTO_TLS_FILES = [GLOBAL_TRUSTSTORE_FILE_NAME,
                          HOST_KEYSTORE_FILE_NAME]

"""
Catches hex strings in openssl "x509 output"

Example:
            Public Key Algorithm: rsaEncryption
                Public-Key: (2048 bit)
                Modulus:
                    00:c6:7c:29:1e:fc:a8:c1:ef:a4:3a:8a:bb:ea:cb:
                    7e:a4:b3:ba:3b:c5:93:9c:b1:ef:d0:ac:76:00:6a:
                    ...
                    14:55
                Exponent: 65537 (0x10001)
"""
HEX_REGEX = re.compile(r"\n\s*(?:[\da-f][\da-f]:){1,}(?:\n\s*(?:[\da-f][\da-f]:)*)*[\da-f][\da-f]")

LOG = logging.getLogger(__name__)

def install_certs(ctx, certtar):
  # pylint: disable=too-many-branches
  """
  Installs certificates on host and configures agent config if needed
  """

  # Check for conflicting TLS configs
  cfg = ctx.obj.cfg
  found_unexpected_config = False
  found_configs = 0
  if AUTO_TLS_CONFIG_SECTION in cfg.value_map():
    for (key, val) in cfg.value_map()[AUTO_TLS_CONFIG_SECTION].iteritems():
      if key in AUTO_TLS_CONFIGS:
        found_configs = found_configs + 1
        if val != AUTO_TLS_CONFIGS[key]:
          LOG.error("Detected TLS configuration different from expected "
                    "Auto-TLS configuration: %s = %s (expected: %s)",
                    key, val, AUTO_TLS_CONFIGS[key])
          found_unexpected_config = True

  if found_unexpected_config:
    LOG.warning("Not proceeding with cert installation to avoid breakage. To "
                "enable Auto-TLS on this host, first remove all TLS-related "
                "configuration from config.ini")
    return 1

  # If the path already exists, just delete it. CM server has already backed up
  # old host keys/certs.
  if os.path.exists(DEFAULT_CERT_DIR):
    shutil.rmtree(DEFAULT_CERT_DIR)
  os.makedirs(DEFAULT_CERT_DIR, mode=0755)

  with tarfile.open(certtar, mode="r:") as archive:
    for member in archive.getmembers():
      filename = member.name

      # certmanager tars some of these files with a "./" prefix for some reason
      # so we need to strip it
      if filename.startswith('./'):
        filename = filename[2:]

      if filename in AUTO_TLS_FILES:
        LOG.info("Extracting new Auto-TLS file: %s", filename)
        archive.extract(member, path=DEFAULT_CERT_DIR)

        # The CM host has a separate directory which contains the JKS files used
        # by the web UI. We need to update files there too.
        existing_file = os.path.join(DEFAULT_CMCA_DIR, filename)
        if filename in CM_HOST_AUTO_TLS_FILES and os.path.exists(existing_file):
          LOG.info("Extracting new Auto-TLS file (for CM host): %s", filename)
          archive.extract(member, path=DEFAULT_CMCA_DIR)

      elif filename != ".":
        LOG.warning("Unexpected certificate file: %s", filename)

  # Write the necessary configs into the agent's config.ini if needed
  if found_configs < len(AUTO_TLS_CONFIGS):
    LOG.info("Writing Auto-TLS configs to config.ini")
    with open(DEFAULT_CONFIG_FILE) as conf_file:
      with tempfile.NamedTemporaryFile(delete=False) as tmp:
        for line in conf_file.readlines():
          # Flip the use_tls line to on
          line = USE_TLS_REGEX.sub('use_tls=1', line)
          tmp.write(line)

        tmp.write("\n### FOLLOWING SECTION AUTO-GENERATED BY AUTO-TLS ###\n")
        tmp.write("### DO NOT MODIFY ###\n")
        tmp.write("[Security]\n")
        for (key, val) in AUTO_TLS_CONFIGS.iteritems():
          tmp.write("%s=%s\n" % (key, val))
        tmp.write("### PRECEDING SECTION AUTO-GENERATED BY AUTO-TLS ###\n")

    shutil.copystat(DEFAULT_CONFIG_FILE, tmp.name)
    shutil.move(DEFAULT_CONFIG_FILE, DEFAULT_CONFIG_FILE + ".backup")
    shutil.move(tmp.name, DEFAULT_CONFIG_FILE)
  else:
    LOG.info("Auto-TLS configs already present in config.ini")

def dump_cert(ctx):
  """
  Dumps the agent cert for collection by diag bundles
  """
  cfg = ctx.obj.cfg

  print "Use TLS: %s" % cfg.use_tls
  if not cfg.client_cert_file:
    print "No client certificate is configured."
    return

  print "Client certificate path: %s" % cfg.client_cert_file
  if not os.path.exists(cfg.client_cert_file):
    print "Client certificate does not exist at the configured path."
    return

  openssl = Popen(["openssl", "x509", "-in", cfg.client_cert_file, "-text", "-noout"],
                  stdout=PIPE, stderr=STDOUT)
  output = openssl.communicate()[0]

  # Redact hex dumps in output (e.g. modulus, signature) before printing
  print HEX_REGEX.sub(" [BUNDLE-REDACTED]", output)
  if openssl.returncode != 0:
    print "Openssl returned exit code: %s" % openssl.returncode
