# Copyright (c) 2017 Cloudera, Inc. All rights reserved.
"""
Utility functions related to Avro usage
"""
#pylint: disable=too-many-branches,bare-except,line-too-long
#pylint: disable=import-error,dangerous-default-value
from __future__ import absolute_import
import re
import collections
import logging

from avro.protocol import Protocol
from avro.schema import (RecordSchema, UnionSchema, MapSchema, EnumSchema,
                         ArraySchema, PrimitiveSchema)

LOG = logging.getLogger(__name__)

def can_utf8_encode_string(data):
  """
  Checks if a string can be encoded to UTF-8
  :param data: The string to be tested
  :return: True if the string can be encoded, False otherwise
  """
  if isinstance(data, (str, unicode)):
    try:
      data.encode('utf-8')
    except UnicodeDecodeError:
      return False
    return True
  else:
    return False

def get_avro_bytes_fields(schema):
  """
  This method creates a list of dot-separated xpath-type regex testers based off
  Avro schema which tests if any Avro field is of type "bytes".
  e.g. For a record R if a map field M is supposed to be type bytes, this method
  will return a regex tester for 'R.M.*' in the output list
  :param schema: The Avro schema
  :return A list of regex testers for the 'bytes' field xpaths
  """
  retval = []
  if schema is None:
    return retval

  # Perform DFS on the schema and create regex testers for "bytes" fields
  stack = list()
  stack.append((schema, ""))
  while len(stack) > 0:
    top = stack.pop()
    cur_schema = top[0]
    cur_xpath = top[1]
    cur_type = str(type(cur_schema))
    if isinstance(cur_schema, Protocol):
      if cur_xpath != "":
        cur_xpath += "."
      for key, val in cur_schema.types_dict.iteritems():
        stack.append((val, cur_xpath+key))
    elif isinstance(cur_schema, RecordSchema):
      if cur_xpath != "":
        cur_xpath += "."
      for key, val in cur_schema.fields_dict.iteritems():
        stack.append((val.props['type'], cur_xpath+key))
    elif isinstance(cur_schema, UnionSchema):
      for item in cur_schema.schemas:
        stack.append((item, cur_xpath))
    elif isinstance(cur_schema, MapSchema):
      if cur_xpath != "":
        cur_xpath += "."
      # Map keys will be generated at runtime, need a wildcard here
      stack.append((cur_schema.values, cur_xpath+"*"))
    elif isinstance(cur_schema, ArraySchema):
      stack.append((cur_schema.items, cur_xpath))
    elif isinstance(cur_schema, EnumSchema):
      pass
    elif isinstance(cur_schema, PrimitiveSchema):
      if cur_schema.type == "bytes":
        # Desired field type found
        retval.append(re.compile(cur_xpath))
    else:
      LOG.warn("Unknown field type %s found while parsing avro schema", cur_type)
  return retval


def sanitize_unicode(obj, skip_fields=[], xpath=""):
  """
  Attempts to sanitize unicode characters from a list
  :param obj: The object to be sanitized
  :param skip_fields: Fields to skip
  :param xpath: Optional base xpath
  :return: The sanitized object
  """
  #pylint: disable=too-many-return-statements
  if obj is None:
    return None
  if isinstance(obj, (str, unicode)):
    for fld in skip_fields:
      if fld.match(xpath):
        # Matched a field which should not be unicode decoded
        return obj
    if not can_utf8_encode_string(obj):
      try:
        obj = obj.decode('utf-8')
      except:
        pass
    return obj
  if isinstance(obj, dict):
    retdict = dict()
    if xpath != "":
      xpath += "."
    for key, val in obj.iteritems():
      retdict[sanitize_unicode(key)] = sanitize_unicode(val, skip_fields, xpath+key)
    return retdict
  if isinstance(obj, (list, collections.Iterable)):
    retval = []
    for item in obj:
      retval.append(sanitize_unicode(item, skip_fields, xpath))
    return retval
  if isinstance(obj, tuple):
    return tuple(sanitize_unicode(list(obj)), skip_fields, xpath)
  return obj


def generate_avro_sanitizer(schema):
  """
  Returns a unicode sanitizer function for a provided schema
  :param schema: Avro schema to be used
  :return: The sanitizer routine
  """
  skip_fields = get_avro_bytes_fields(schema)

  def validator(obj):
    """
    Return the wrapped sanitized routine
    """
    return sanitize_unicode(obj, skip_fields)

  return validator
