"""
Copyright (c) 2011-2016 Cloudera, Inc. All rights reserved.

Contains classes for managing parsed config files.
"""
from __future__ import absolute_import

# pylint: disable=invalid-name,too-few-public-methods
import logging
import ConfigParser
import collections
import os
import os.path
LOG = logging.getLogger(__name__)

DEFAULT = ConfigParser.DEFAULTSECT

class MissingKeyError(Exception):
  """
  Exception to indicate missing key.
  """
  pass


class ConfigItem(object):
  """
  Descriptor base class for items in the configuration.
  """
  def __init__(self, section, default=None, name=None, log_on_default=None):
    self.section = section
    self.name = name
    self.value = None
    self.default = default
    self.log_on_default = log_on_default

  def retrieve(self, ini):
    """
    Retrieve the value represented by this ConfigItem from the ConfigParser object.
    :param ini: ConfigParser object to search in.
    :return: As above.
    """
    return ini.get(self.section, self.name)

  def __get__(self, instance, owner):
    try:
      # pylint: disable=protected-access
      val = instance._retrieve(self.retrieve)
    except AttributeError:
      if instance is None:
        return self
    except ValueError:
      LOG.error("Invalid value type for [%s].%s", self.section, self.name)
      raise
    except (ConfigParser.Error, MissingKeyError):
      val = self.default
      if callable(val):
        val = val(self, instance)
      if self.log_on_default:
        self.log_on_default("Using default value for '%s.%s': %s", self.section, self.name, val)
    if val is not None:
      setattr(instance, self.name, val)
    return val


class ConfigIntItem(ConfigItem):
  """
  Represents config values that are integers.
  """
  def retrieve(self, ini):
    return ini.getint(self.section, self.name)

class ConfigFloatItem(ConfigItem):
  """
  Represents config values that are floating point numbers.
  """
  def retrieve(self, ini):
    return ini.getfloat(self.section, self.name)


class ConfigBoolItem(ConfigItem):
  """
  Represent config values that are booleans.
  """
  def retrieve(self, ini):
    return True if ini.getboolean(self.section, self.name) else False


class ConfigListItem(ConfigItem):
  """
  Represents config values that are presented as lists parsed from comma-seperated values.
  """
  def retrieve(self, ini):
    return ini.get(self.section, self.name).split(",")


class MetaConfig(type):
  """
  Metaclass to ensure we generate names for ConfigItem object
  based on the field they are assigned to.
  """
  def __init__(cls, name, bases, attrs):
    configs = {}
    for b in bases:
      configs.update(getattr(b, "__config__", {}))
    for (k, v) in attrs.iteritems():
      if isinstance(v, ConfigItem):
        configs[k] = v
        if v.name is None:
          v.name = k
    attrs["__config__"] = configs
    super(MetaConfig, cls).__init__(name, bases, attrs)


class Config(object):
  """
  Object representing user-provided configuration.
  To be used by declaring a sub-class, which has class level members that are defined
  with ConfigItem, so as to specify the schema of the expected configuration values, their
  types and defaults.
  """
  __metaclass__ = MetaConfig
  __config__ = {}

  def __init__(self, configs=None):
    self.ini = []

    # Add names to all unnamed ConfigItem objects.
    if configs is None:
      configs = []
    if not isinstance(configs, (list, tuple)):
      configs = [configs]
    for config in configs:
      ini = ConfigParser.SafeConfigParser()
      if isinstance(config, Config):
        self.ini.extend(config.ini)
        continue
      if isinstance(config, dict):
        for k, v in config.iteritems():
          try:
            section, key = k.split(".", 1)
            if not ini.has_section(section):
              ini.add_section(section)
          except ValueError:
            section = DEFAULT
            key = k

          ini.set(section, key, str(v))
      else:
        try:
          ini.read([config])
        except Exception, e:
          LOG.exception("Failure to read config: %s: %s", config, e)
          raise

      self.ini.append(ini)

  @classmethod
  def make_config(cls, config_location, override, default_section):
    """
    Evaluate config files or config-file locations and configuration overrides provided by
    the user to create an "cls" object.
    :param config_location: Path to config file or directory of .ini files.
    :param override: list of overrides - each is a "k=v" string.
    :return:
    """
    overrides = {}

    if override:
      for x in override:
        try:
          k, v = x.split("=", 1)
        except ValueError:
          raise Exception("Invalid override: %s", x)

        w = k.split(".", 1)
        if len(w) == 1:
          k = ".".join((default_section, k))
        overrides[k] = v

    if config_location is None:
      configs = [overrides]
    elif os.path.isdir(config_location):
      configs = [overrides]
      for ent in sorted(os.listdir(config_location)):
        if os.path.splitext(ent)[1] == ".ini":
          configs.append(os.path.join(config_location, ent))
    else:
      configs = [overrides, config_location]

    return cls(configs)

  def value_map(self):
    """
    Returns a dictionary representing all configuration key/value items,
    in their raw "ConfigParser" form.
    :return:
    """
    out = collections.defaultdict(dict)
    for ini in reversed(self.ini):
      for sect in ini.sections() + [ConfigParser.DEFAULTSECT]:
        # DEFAULT section may not exist.
        try:
          options = ini.options(sect)
        except ConfigParser.NoSectionError:
          continue
        for option in options:
          out[sect][option] = ini.get(sect, option)
    for name in self.__config__.keys():
      item = getattr(self.__class__, name)
      if isinstance(item, ConfigItem):
        val = getattr(self, name)
        out[item.section][item.name] = val
    return out

  def show_config(self):
    """
    Return a string showing the state of the configuration.
    :return:
    """
    out = []
    for (section, values) in self.value_map().iteritems():
      for (k, v) in values.iteritems():
        out.append("%s.%s = %s" % (section, k, v))
    return "\n".join(out)

  def _retrieve(self, ini_fetcher):
    """
    Iterate through all ConfigParser objects and call ini_fetcher with each.
    :param ini_fetcher: Callback to invoke to fetch config values.
    :return: Value of configuration item, or ConfigParser.Error exception.
    """
    err = None
    for x in self.ini:
      try:
        return ini_fetcher(x)
      except ConfigParser.NoOptionError:
        pass
      except ConfigParser.Error, e:
        err = e
        continue
    if err is None:
      raise MissingKeyError()
    # pylint: disable=raising-bad-type
    raise err

class TestConfig(Config):
  """
  Test class that also demonstrates combined usage of Config and Config*Item.
  """

  int_val = ConfigIntItem(DEFAULT, 5)
  float_val = ConfigFloatItem("float_section", 4.0)
  bad_val = ConfigIntItem(DEFAULT, 3)
  bool_val = ConfigBoolItem("bool_section", True)
  list_val = ConfigListItem(DEFAULT, "a,b,c,d")
  no_default = ConfigItem(DEFAULT)
  dynamic = ConfigItem(DEFAULT, lambda cfgItem, cfg: "some default", name="date")


  @classmethod
  def run_test(cls):
    """
    Execute some tests.
    """
    logging.basicConfig(level=logging.INFO)
    import tempfile
    import textwrap

    with tempfile.NamedTemporaryFile() as f1:
      config1 = """
      [DEFAULT]
      int_val = 3
      list_val = x,y,z
      bad_val = "foo"
      [float_section]
      float_val = 3.4

      [bool_section]
      bool_val = 0
      """
      f1.write(textwrap.dedent(config1))
      f1.flush()

      config2 = {"int_val": 4}

      cfg = cls([config2, f1.name])

      # config2 takes precedence
      assert cfg.int_val == 4, cfg.int_val

      # config2 not present, should get a different value
      cfg2 = cls([f1.name])
      assert cfg2.int_val == 3, cfg2.int_val

      assert cfg.no_default is None
      assert cfg.dynamic == "some default"
      assert cfg.list_val == ['x', 'y', 'z']
      assert cfg.float_val == 3.4, cfg.float_val

      assert cfg.bool_val is False, cfg.bool_val
      try:
        _ = cfg.bad_val
        assert False
      except ValueError:
        pass

if __name__ == "__main__":
  TestConfig.run_test()
