"""
Copyright 2016, Cloudera Inc.
"""
from __future__ import absolute_import

# pylint: disable=invalid-name, too-few-public-methods, no-self-use, broad-except
import SocketServer
import code
import socket
import sys
import threading
import logging
import os
import os.path

LOG = logging.getLogger(__name__)

class LineBuffer(object):
  """
  Line buffer to buffer incoming input at line breaks from a specified input source.
  """
  def __init__(self, source):
    self.src = source
    self.buffer = []
    self.partial = None

  def readline(self):
    """
    Return the next line of input.

    Reads from the source input and buffers the input data so as to present
    individual lines of input.
    """
    while len(self.buffer) == 0:
      b = self.src(1024)
      if len(b) == 0:
        break
      lines = b.split("\n")
      if self.partial:
        self.partial += lines[0]
        lines.pop(0)
        self.buffer.append(self.partial)
        self.partial = None

      self.buffer.extend(lines)
      if len(self.buffer) != 0:
        self.partial = self.buffer[-1]
      self.buffer.pop(-1)
    if len(self.buffer) == 0:
      raise EOFError()
    l = self.buffer.pop(0)
    return l


class EvalContext(object):
  """
  Context for evaluation of remote console commands or batch jobs.
  """
  def __init__(self, agent):
    self.agent = agent
    self.locals = {}
    self.out = []
    self.logger = logging.getLogger("dynamic")
    def output(fmt, *pargs):
      """
      Format and save messages to "self.out"
      """
      self.out.append(str(fmt) % pargs)
    self.locals["agent"] = self.agent
    self.locals["output"] = output
    self.locals["LOG"] = self.logger

  def eval(self, body):
    """
    Evaluate a string of a batch job.
    Return a list of strings corresponding to the calls to "output" made in the course
    of execution of the batch job.
    """
    try:
      self.out = []
      # pylint: disable=exec-used
      exec body in globals(), self.locals

      LOG.info("Output is: '%s'", self.out)
      result = ""
      for out in self.out:
        if not out.endswith("\n"):
          out = "%s\n" % out
        result += out
      return result
    except Exception as e:
      LOG.exception("Failed eval: %s", e)
      raise


  def getConsoleClass(self):
    """
    Return a class for a console that will embed the expected "locals" environment.
    """
    _locals = self.locals

    class _Console(code.InteractiveConsole):
      def __init__(self):
        code.InteractiveConsole.__init__(self, _locals)

    return _Console

class ServerHandler(SocketServer.BaseRequestHandler):
  """
  TCP socket handler to handle incoming interactive console sessions.
  The core functionality is in "handle()".
  """
  server = None
  lock = threading.RLock()

  @classmethod
  def StartOrConnect(cls, _context, cfg):
    """
    Start or connect to an existing console session.
    Return the path of the UNIX socket that the session is listening on,
    starting it if necessary.
    """
    path = os.path.join(cfg.paths.base_dir, "console")
    if cls.server is None:
      class _ServerHandler(cls):
        context = _context

      if os.path.exists(path):
        os.unlink(path)
      x = SocketServer.UnixStreamServer(path, _ServerHandler)
      os.chmod(path, 0700)

      thr = threading.Thread(name="console", target=x.serve_forever)
      thr.daemon = True
      thr.start()

      cls.server = x
      LOG.info("Starting console on: %s", path)

    return path

  def handle(self):
    """
    Handle a connection request on a socket.
    """
    def recv_wrapper(*pargs, **kwargs):
      """
      Wrapper for socket recv to ensure we don't drop connection
      due to timeouts.
      """
      while True:
        try:
          data = self.request.recv(*pargs, **kwargs)
          return data
        except socket.timeout:
          pass
    line_buf = LineBuffer(recv_wrapper)
    request = self.request

    # Get the class from which to create a console that has the
    # right execution context embedded in it.
    # pylint: disable=no-member
    consoleClass = self.context.getConsoleClass()

    class Console(consoleClass):
      """
      Extension of 'consoleClass' with appropriate I/O methods.
      """
      # pylint: disable=missing-docstring
      def write(self, data):
        request.sendall(data)

      def flush(self):
        pass

      def raw_input(self, prompt=""):
        if len(prompt):
          request.sendall(prompt)
        data = line_buf.readline()
        return data

    # Run the console
    with self.lock:
      old_stdout = sys.stdout
      try:
        c = Console()
        sys.stdout = c
        c.interact()
      except Exception as e:
        LOG.exception("Console exception: %s %s", type(e), e)
      finally:
        sys.stdout = old_stdout


if __name__ == "__main__":
  SocketServer.TCPServer.allow_reuse_address = True
  srv = SocketServer.TCPServer(("0.0.0.0", 7192), ServerHandler)
  srv.serve_forever()
