"""
Copyright (c) 2012 Cloudera, Inc. All rights reserved.

Utilties for generating diffs.
"""

# pylint: disable=too-many-lines,too-many-public-methods,too-many-statements
# pylint: disable=too-many-instance-attributes,import-error
# pylint: disable=line-too-long,bad-continuation,bare-except,broad-except,invalid-name,
# pylint: disable=missing-docstring,too-many-branches,too-many-lines


from __future__ import absolute_import
import filecmp
import logging
import os
import shutil
import simplejson as json
import subprocess
import sys
import tarfile
import tempfile
import time

from cmf.parcel_id import ParcelId

DELETED = ".deleted"

LOG = logging.getLogger(__name__)

def get_deleted_meta(directory):
  """Get full path of the deleted files meta file, if it existed under the dir
  """
  return os.path.join(directory, DELETED)

def diff(left, right, out):
  """Deep compare the two directories and generate a diff directory containing
     either modified or new files, and a metadata file ".deleted" for files or
     directories that have been deleted.
  """
  assert os.path.exists(left)
  assert os.path.exists(right)
  LOG.info("Generating diff between %s and %s into %s", left, right, out)
  if not os.path.exists(out):
    os.makedirs(out)

  _recursive_diff(filecmp.dircmp(left, right), out)

def _recursive_diff(compare, out):
  # copy new or changed files into diff out
  new_or_changed = []
  for name in compare.right_only:
    LOG.debug("Including right_only %s", name)
    new_or_changed.append(name)
  for name in compare.diff_files:
    LOG.debug("Including diff_files %s", name)
    # note that links whose target file changed but the link itself did not
    # change are considered diff_files. we deal with this during _copy_link by
    # silent overwrite
    new_or_changed.append(name)
  for name in new_or_changed:
    # could be a file or brand new dir
    _cpr(os.path.join(compare.right, name), out)

  # write excludes into delete meta file
  if len(compare.left_only) > 0:
    if not os.path.exists(out):
      os.makedirs(out)
    fp = open(get_deleted_meta(out), 'w')
    try:
      LOG.debug("Writing %s into deleted meta file under %s", compare.left_only, out)
      json.dump(compare.left_only, fp)
    finally:
      fp.close()

  for subdir, subcmp in compare.subdirs.iteritems():
    _recursive_diff(subcmp, os.path.join(out, subdir))

def _cpr(src, dst):
  '''Copy src whether it is a file, link, or dir into dst dir.'''
  # islink must be checked before isfile because they're both true for valid link
  if os.path.islink(src):
    _copy_link(src, dst)
  elif os.path.isfile(src):
    if not os.path.exists(dst):
      os.makedirs(dst)
    shutil.copy2(src, dst)
  else:
    # copytree requires dst to not exist
    dst_dir = os.path.join(dst, os.path.basename(src))
    shutil.copytree(src, dst_dir, symlinks=True)

def _copy_link(src, dst):
  '''Copy either absolute or relative link.'''
  if not os.path.exists(dst):
    os.makedirs(dst)
  linkto = os.readlink(src)
  cwd = os.getcwd()
  # need to change cwd to copy relative links correctly
  os.chdir(dst)
  try:
    LOG.debug("Copying link %s to %s under %s", os.path.basename(src), linkto, dst)
    if os.path.islink(os.path.basename(src)):
      # possible because unchanged links may be included into diff due to target changes
      os.unlink(os.path.basename(src))
    os.symlink(linkto, os.path.basename(src))
  finally:
    # restore cwd
    os.chdir(cwd)

def patch(left, delta, out):
  """Apply the diff directory, generated by the diff function, to the left
     directory and output the patched directory into out. out must either be
     equal to left (in-place patching) or be a non-existent dir.
  """
  assert os.path.exists(left)
  assert os.path.exists(delta)

  LOG.info("Patching %s with diff %s", out, delta)

  if not os.path.exists(out):
    shutil.copytree(left, out, symlinks=True)

  _apply_patch(delta, out)

def _apply_patch(delta, out):
  # Patch immediate directory content recursively
  if not os.path.exists(out):
    # ensures that empty dirs are copied over
    os.mkdir(out)
  for f in os.listdir(delta):
    diff_file = os.path.join(delta, f)
    if diff_file == get_deleted_meta(delta):
      fp = open(diff_file, 'r')
      try:
        deleted = json.load(fp)
        for gone in deleted:
          _rmrf(os.path.join(out, gone))
      finally:
        fp.close()
    elif not os.path.islink(diff_file) and os.path.isdir(diff_file):
      # only recursively apply real dirs
      _apply_patch(diff_file, os.path.join(out, f))
    else:
      _cpr(diff_file, out)

def _rmrf(path):
  # remove a path regardless of whether it's a file or dir or link
  if os.path.islink(path):
    # we want to remove even dead links
    os.unlink(path)

  if not os.path.exists(path):
    return
  elif os.path.isfile(path):
    os.remove(path)
  else:
    shutil.rmtree(path)

def tardiff(left, right, out):
  """Convenient call for working with tarballs. Untar left and right, use the
     only sub dir within each untarred dir to generate a diff, and tar up the
     diff named as <product>-<left-version>-<right-version>-<platform>.diff under out.
  """
  assert os.path.isfile(left)
  assert os.path.isfile(right)
  tmp = tempfile.mkdtemp()
  try:
    left_dir = os.path.join(tmp, 'left')
    os.mkdir(left_dir)
    _untar(left, left_dir)
    left_root = _get_single_subdir(left_dir)

    right_dir = os.path.join(tmp, 'right')
    os.mkdir(right_dir)
    _untar(right, right_dir)
    right_root = _get_single_subdir(right_dir)
    diff_dir = os.path.join(tmp, 'diff')
    diff(os.path.join(left_dir, left_root), os.path.join(right_dir, right_root), diff_dir)
    return _tar(diff_dir, out, _get_diff_filename(left_root, right_root))
  finally:
    shutil.rmtree(tmp, ignore_errors=True)

def _get_diff_filename(left, right):
  lparts = left.split('-', 3)
  rparts = right.split('-', 3)
  assert len(lparts) > 2
  assert len(rparts) > 2
  return "%s-%s-%s-%s.diff" % (lparts[0], lparts[1], rparts[1], rparts[2])

def _get_single_subdir(d):
  dirs = os.listdir(d)
  assert len(dirs) == 1
  return dirs[0]

def _untar(src, dst):
  # See OPSAPS-13299 for why use subprocess instead of tarfile.
  if not os.path.exists(dst):
    os.makedirs(dst)
  return_code = subprocess.call(["tar", "xf", src, "--no-same-owner", "-C", dst])
  if return_code != 0:
    raise Exception("Untar failed with return code %d" % (return_code, ))

def _tar(src, dst, name):
  output = os.path.join(dst, name)
  tar = tarfile.open(output, 'w:gz')
  try:
    for f in os.listdir(src):
      tar.add(os.path.join(src, f), arcname=f)
  except Exception as e:
    LOG.exception("Error while tar %s: %s", src, e)
  finally:
    tar.close()
  return output

def apply_patch(left, delta, out, tmp=None):
  """Convenience call for working with diff parcels.
     If the left is a file, it is untarred, and if it is a directory, it is copied.
     After preparing 'left' then diff, patch left with the diff dir, name it as
     <product>-<new-version> and move patched dir into out.
  """
  assert os.path.isfile(left) or os.path.isdir(left), "%s is not a file or directory" % left
  assert os.path.isfile(delta), "%s is not a file" % delta
  assert delta.endswith(".diff"), "%s must end with .diff" % delta
  patched_name = _get_patched_dirname(os.path.basename(delta))
  patched_path = os.path.join(out, patched_name)
  assert not os.path.exists(patched_path), "Patched dir already exists: %s" % patched_path

  if os.path.exists(out):
    assert os.path.isdir(out), "%s must be a directory" % out
  else:
    os.makedirs(out)

  if tmp is None:
    tmp = tempfile.mkdtemp()
  else:
    assert os.path.exists(tmp)
    # create subdir since it may be a shared tmp dir
    tmp = tempfile.mkdtemp(dir=tmp)
  try:
    left_dir = os.path.join(tmp, 'left')
    os.mkdir(left_dir)
    if os.path.isdir(left):
      shutil.copytree(left, os.path.join(left_dir, os.path.basename(left)), True)
    else:
      _untar(left, left_dir)
    left_root = _get_single_subdir(left_dir)
    left_root_path = os.path.join(left_dir, left_root)

    diff_dir = os.path.join(tmp, 'diff')
    os.mkdir(diff_dir)
    _untar(delta, diff_dir)
    patch(left_root_path, diff_dir, left_root_path)

    # rename patched dir to the final name, e.g. left_root=CDH-4.1.0,
    # diff=CDH-4.1.0-4.2.0.diff, and output dir should be CDH-4.2.0
    patched_path = os.path.join(tmp, patched_name)
    os.rename(left_root_path, patched_path)
    # move the patched and properly named dir to out
    final_dir = os.path.join(out, patched_name)
    shutil.move(patched_path, final_dir)
    return final_dir
  finally:
    shutil.rmtree(tmp, ignore_errors=True)

def _get_patched_dirname(delta):
  parcel_id = ParcelId.filename(delta)
  return parcel_id.get_parcel_dir()

if __name__ == '__main__':
  if len(sys.argv) < 4:
    print "diff.py [diff|tardiff] <left dir> <right dir> <output dir>"
    print "diff.py [patch|tarpatch] <left dir> <diff dir> <output dir>"
    sys.exit(1)

  ops = ['diff', 'patch', 'tardiff', 'tarpatch']
  if sys.argv[1] in ops:
    func = globals()[sys.argv[1]]
  else:
    print "First argument must be one of ", ops
    sys.exit(1)

  LOG_FORMAT = '[%(asctime)s] %(process)d %(threadName)s %(module)-12s %(levelname)-8s %(message)s'
  DATE_FORMAT = '%d/%b/%Y %H:%M:%S %z'
  logging.basicConfig(level=logging.DEBUG,
                      format=LOG_FORMAT, datefmt=DATE_FORMAT)

  LOG.info("Starting %s", sys.argv[1])
  begin = time.clock()
  func(sys.argv[2], sys.argv[3], sys.argv[4])
  end = time.clock()
  LOG.info("Finished %s. Took %f seconds.", sys.argv[1], (end - begin))
