# Copyright 2016, Cloudera Inc.
"""
Provides the @lazy decorator.
"""

#pylint: disable = invalid-name, too-few-public-methods
class lazy(object):
  """
  An alternative implementation of lazy_loaded_property with a couple of differences:
  - Once a lazy loaded attribute is initialized, it cannot be reset or changed.
  - The access path for these attributes is more efficient, since after the first access, the
    attribute is fetched directly via getattr() mechanisms without any special handler code being
    invoked.
  - No shadow lazy attributes are introduced into the object namespace.
  """
  def __init__(self, fn):
    self.fn = fn

  def __get__(self, instance, owner):
    match = None
    for klass in instance.__class__.mro():
      for (k, v) in vars(klass).iteritems():
        if v is self:
          match = k
          break
      if match:
        break
    else:
      if instance is None:
        return self
      assert False, "Could not find attribute match for %s %s %s %s" % (self, self.fn.__name__,
                                                                        instance, owner)
    setattr(instance, match, None)  # Break recursion possibility
    try:
      val = self.fn(instance)
    except:
      delattr(instance, match)
      raise
    setattr(instance, match, val)
    return val

  @classmethod
  def Invalidate(cls, obj, prop_name):
    """
    Reset property "prop_name" and cause it to be re-initialized on the next access.
    """
    delattr(obj, prop_name)
