# Copyright 2017, Cloudera Inc
"""
Abstraction of OS-level operations.
"""

import os
import errno
import shutil

import pwd
import grp
import stat
import logging
import subprocess
from cmf.util.lazy import lazy
from cmf.throttling_logger import ThrottlingLogger

LOG = logging.getLogger(__name__)
THROTTLED_LOG = ThrottlingLogger(LOG)


# pylint: disable=broad-except, invalid-name,missing-docstring, too-many-arguments, unused-argument

class AlternativesManager(object):
  """
  Class to abstract operations on system "alternatives".
  """

  # The alternatives database can be found in different places.
  DB_CANDIDATES = ["/var/lib/alternatives",
                   "/var/lib/dpkg/alternatives",
                   "/var/lib/rpm/alternatives"]
  # update-alternatives can be in different places
  # in different OSes.  We simply look it up at
  # construction and go with it.
  BIN_CANDIDATES = ["/usr/sbin/update-alternatives",
                    "/usr/bin/update-alternatives"]

  ALTERNATIVES_ARGS = []

  def __init__(self, os_ops):
    self.os_ops = os_ops

  # Lazy attributes can be overriden/hard-coded by test code.
  @lazy
  def alt_bin(self):
    for bin_path in self.BIN_CANDIDATES:
      if os.path.exists(bin_path):
        return bin_path
    raise Exception("No update-alternatives command found")

  @lazy
  def db_dir(self):
    for db_path in self.DB_CANDIDATES:
      if os.path.isdir(db_path):
        return db_path
    raise Exception("No alternatives db dir found")


  def pre_update_alternative(self, name):
    """
    Prior to calling alternatives utilities, ensure that there is no empty
    alternatives file that could trip those up.
    """
    p = os.path.join(self.db_dir, name)
    try:
      if os.path.exists(p) and os.stat(p).st_size == 0:
        LOG.warn("Removing empty alternatives file: %s", p)
        os.unlink(p)
    except Exception, exc:
      LOG.exception("Failed validating alternative: %s: %s", p, exc)

  def activate(self, name, dest, source, priority=None, is_dir=False):
    if is_dir:
      self.os_ops.sudo_command("mkdir", "-p", "-m", "0755", os.path.split(dest)[0])

    self.pre_update_alternative(name)
    if not priority:
      priority = 50
    cmd = [self.alt_bin] + self.ALTERNATIVES_ARGS + \
          ["--install", dest, name, source, str(priority)]
    self.os_ops.sudo_command(*cmd)


  def deactivate(self, name, dest, source, priority=None, is_dir=False):
    self.pre_update_alternative(name)
    cmd = [self.alt_bin] + self.ALTERNATIVES_ARGS + \
          ["--remove", name, source]
    if self.os_ops.sudo_command(*cmd) == 0 and dest and is_dir:
      try:
        os.unlink(dest)
      except Exception:
        pass

  def lookup(self, alt_name):
    """
    Uses update-alternatives to find all alternatives for a particular name.
    Returns a list of (path, priority).
    Returns empty list on error.

    """
    cmd = [self.alt_bin] + self.ALTERNATIVES_ARGS + ["--display", alt_name]
    result = []
    try:
      out = self.os_ops.sudo_command_output(*cmd)
    except Exception as exc:
      THROTTLED_LOG.warn("Failed alternatives lookup: %s (%s)", alt_name, exc)
      return result

    for line in out.split('\n'):
      try:
        words = line.rstrip().split(" ")
        if words[0][0] != "/":
          continue
        path = words[0]
        try:
          priority = int(words[-1])
        except ValueError:
          continue
        result.append((path, priority))
      except Exception as exc:
        THROTTLED_LOG.warn("Failed parsing alternatives line: %s %s %s", alt_name, exc, line)
    return result


class OSOps(object):
  """
  Object for interacting with the underlying system.
  Performs system-level operations.
  May/ may not enforce ownership/permissions depending on its configuration.
  """

  # Where to read ephemeral port range
  IP_LOCAL_PORT_RANGE = "/proc/sys/net/ipv4/ip_local_port_range"

  # pylint: disable=no-self-use, invalid-name
  def __init__(self, honor_users_and_groups=True, default_perm_mode=0o755, sudo=None):
    self.honor_users_and_groups = honor_users_and_groups
    self.default_perm_mode = default_perm_mode
    self.sudo = sudo
    self.alternatives = AlternativesManager(self)

  class MetaData(object):
    """
    Representation of file metadata
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, user, group, mode):
      self.user = user
      self.group = group
      self.mode = mode

  def exec_command(self, *command):
    """
    Execute a system command.
    Return the process exit code (or -1 on error).
    """
    try:
      LOG.info("Executing command: %s", command)
      p = subprocess.Popen(command, stdin=None, stdout=None, stderr=None)
      return p.wait()
    except Exception, exc:
      LOG.exception("Failed executing process: '%s': %s", " ".join(command), exc)
      return -1

  def exec_command_output(self, *command):
    """
    Execute a system command.
    Return the process stdout contents if exit code is 0, or raise and exception.
    """
    child = subprocess.Popen(command,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE,
                             close_fds=True)
    stdout, _ = child.communicate()
    if child.returncode != 0:
      raise Exception("Failed to execute command: %s -> %s" % (command, child.returncode))
    return stdout

  def sudo_command(self, *command):
    """
    Same as exec_command(), but uses the configured "sudo" setting.
    """
    if self.sudo:
      return self.exec_command(*([self.sudo] + list(command)))
    return self.exec_command(*command)

  def sudo_command_output(self, *command):
    """
    Same as exec_command_output(), but uses the configured "sudo" setting.
    """
    if self.sudo:
      return self.exec_command_output(*([self.sudo] + list(command)))
    return self.exec_command_output(*command)

  def get_path_metadata(self, path):
    """
    Retrieve "MetaData" for "path"
    """
    file_st = os.stat(path)
    user = pwd.getpwuid(file_st.st_uid)[0]
    group = grp.getgrgid(file_st.st_gid)[0]
    mode = oct(stat.S_IMODE(file_st.st_mode))
    return self.MetaData(user, group, mode)

  def set_path_metadata(self, path, owner=None, group=None, mode=None):
    """
    Change ownership and permissions on a specific file/path.
    """

    if self.honor_users_and_groups and (owner or group):
      if not owner:
        owner = os.stat(path).st_uid
      if not group:
        group = os.stat(path).st_gid
      if isinstance(owner, basestring):
        owner = pwd.getpwnam(owner).pw_uid
      if isinstance(group, basestring):
        group = grp.getgrnam(group).gr_gid
      try:
        os.chown(path, owner, group)
      except Exception as exc:
        owner_name = owner
        group_name = group
        try:
          # pylint: disable=broad-except
          owner_name = pwd.getpwuid(owner)[0]
          group_name = grp.getgrgid(group)[0]
        except Exception:
          pass
        LOG.error("Failed chown: %s %s/%s: %s", path, owner_name, group_name, exc)
        raise
    if mode:
      try:
        os.chmod(path, mode)
      except Exception as exc:
        LOG.error("Failed chmod: %s 0o%o %s", path, mode, exc)
        raise

  def move_file(self, dest,
                source,
                owner=None,
                group=None,
                mode=None):
    """
    Move a file from source to dest.
    """
    # pylint: disable=too-many-arguments
    self.install_file(dest, source, owner, group, mode)
    self.remove_file(source)

  def remove_file(self, path):
    """
    Remove a file.
    """
    os.unlink(path)

  def symlink(self, dest, source):
    """
    Create a symlink.
    :param source:
    :return:
    """
    os.symlink(source, dest)

  def install_file(self, dest,
                   source,
                   owner=None,
                   group=None,
                   mode=None):
    """
    Copy/install a file into a location using specified ownership/permissions,
    to be used if applicable.
    """
    # pylint: disable=too-many-arguments
    shutil.copy2(source, dest)
    self.set_path_metadata(dest, owner, group, mode)

  def file_tree_walk(self, path):
    """
    Iterates over an FS tree from path, presenting all full paths.
    """
    for root, dirs, files in os.walk(path):
      for path in dirs + files:
        yield os.path.join(root, path)
    raise StopIteration()

  def edit_file(self, path, editor):
    """
    Rewrite a file by processing it through an "editor" function that manipulates the file contents.
    """
    # pylint: disable=invalid-name
    fd = os.open(path, os.O_RDWR)
    try:
      # Use os.dup to ensure the original fd stays open even as the file objects get destroyed.
      fobj_read = os.fdopen(os.dup(fd), 'r')
      data = fobj_read.read()

      # Invoke the editor function to update the contents.
      data = editor(data)
      fobj = os.fdopen(os.dup(fd), 'w')

      fobj.truncate(0)
      fobj.seek(os.SEEK_SET, 0)
      fobj.write(data)
      fobj.flush()

    finally:
      try:
        # pylint: disable=broad-except
        os.close(fd)
      except Exception:
        pass

  # pylint: disable=too-many-arguments
  def open_write_file(self, path, owner=None, group=None, mode=None, append=False, exclusive=False):
    """
    Return a file object for a file to be written.
    - File is created if necessary.
    - File is truncated.

    If "mode" is unspecified the file will have 0600 permissions (owner read/write only).

    "exclusive" being true requires the file to be created.  If the file is present an exception
    is propagated from os.open()

    If the file already exists, the effect of this function will be equivalent to the file
    being removed and then re-created with the desired metadata.

    This method is to be used to open files for writing in a manner whereby we maintain control
    of the file permissions explicitly.  And avoid potential races such as the file being created
    with lose permissions, and an undesirable reader opening the file before we get to tighten the
     permissions.

    """
    if mode is None:
      mode = 0o600
    # Initially open the file as owner-accessible only
    write_mode = os.O_APPEND if append else os.O_TRUNC
    if exclusive:
      write_mode |= os.O_EXCL
    fd = os.open(path, os.O_RDWR | os.O_CREAT | write_mode, mode)
    self.set_path_metadata(path, owner, group, mode)
    return os.fdopen(fd, 'w')

  def set_directory_content_metadata(self, path,
                                     owner=None,
                                     group=None,
                                     mode=None,
                                     recurse=False):
    """
    Change ownership/permissions on a directory and its contents.
    """
    # pylint: disable=too-many-arguments
    directories = [path]
    while directories:
      curdir = directories.pop(0)
      for f in os.listdir(curdir):
        p = os.path.join(curdir, f)
        self.set_path_metadata(p, owner, group, mode)
        if os.path.isdir(p) and recurse:
          directories.append(p)

  def makedirs(self, name, mode=0o755):
    """
    This is based on os.makedirs but umask based masking is ignored and mode is set
    explicitly.
    :param name: This name (path) will be created if it does not already exist
    :param mode: All directories created will have the mode set independent of umask
    """
    if os.path.exists(name):
      return
    head, tail = os.path.split(name)
    if (head == "/") and not tail:
      return
    if not tail and head: # path ending with /
      head, tail = os.path.split(head)
    if head and tail and not os.path.exists(head):
      try:
        self.makedirs(head, mode)
      except OSError, e:
        if e.errno != errno.EEXIST:
          raise
      if tail == ".":  # Handle paths that end with "."
        return
    os.mkdir(name, mode)
    os.chmod(name, mode) # Override any masking due to umask

  def mkabsdir(self, path, user=None, group=None, mode=None):
    """
    Make a directory given an absolute path.
    If the directory exists, user,group,mode attributes will not be changed if they
    don't match what is specified by the caller.
    """
    if not os.path.isabs(path):
      raise Exception("Expecting an absolute path, not %s" % path)

    if os.path.isdir(path):
      # Log warnings if user/group/mode are different than what's expected
      mdata = self.get_path_metadata(path)
      if self.honor_users_and_groups:

        if user is not None and user != mdata.user:
          LOG.warning('Expected user %s for %s but was %s', user, path, mdata.user)
        if group is not None and group != mdata.group:
          LOG.warning('Expected group %s for %s but was %s', group, path, mdata.group)

      if mode is not None and oct(mode) != mdata.mode:
        LOG.warning('Expected mode 0o%o for %s but was %s', mode, path, mdata.mode)
      return False

    # Note that makedirs is hierarchical, intermediate directories will be created
    # owned by the user:group of the agent process and mode set to the default_perm_mode.
    # The leaf directory will be created with the ownership:permission that were passed in.
    try:
      self.makedirs(path, self.default_perm_mode)
    except Exception as exc:
      LOG.error("Failed directory creation: %s: %s", path, exc)
      raise
    LOG.info("Created directory %s %s %s 0o%o", path, user, group, mode)
    self.set_path_metadata(path, user, group, mode)
    return True

  def get_ephemeral_port_range(self):
    """
    Return host's ephemeral port range
    """
    f = None
    try:
      f = open(self.IP_LOCAL_PORT_RANGE)
      line = f.readline().rstrip('\n')
      ports = line.split()
      return [long(ports[0]), long(ports[1])]
    except IOError as e:
      LOG.warn("Failed reading ephemeral port range: %s", e)
    finally:
      if f is not None:
        f.close()
    return []
