# Copyright (c) 2017 Cloudera, Inc. All rights reserved.
"""
Implements infrastructure to invoke synchronized actions across multiple threads
"""
from __future__ import absolute_import
import threading
import logging

LOG = logging.getLogger(__name__)

class SyncedAction(object):
  """
  Performs a specified action no more than once per "delay" seconds.
  For any group of threads that calls a given instance of this object, all threads
  that perform this action within "delay" seconds of the first one, will block until
  the first thread has completed the action.
  """
  # pylint: disable=too-few-public-methods
  DEFAULT_DELAY = 2.0
  action_map = dict()
  action_lock = threading.RLock()

  def __init__(self, action, delay=DEFAULT_DELAY):
    self.action = action
    self.delay = delay
    self.seq = 0
    self.done = 0
    self.pending = False
    self.lock = threading.RLock()
    self.cond = None

  def __call__(self):
    do_exec = False
    with self.lock:
      # The first thread in creates the "cond" object - the train that all subsequent
      # threads get to board.
      if self.cond is None:
        self.cond = threading.Condition()
        self.seq += 1
        do_exec = True

      # Threads board the "current train waiting to leave the station".
      cond = self.cond
    seq = self.seq
    LOG.debug("[%s] Exec action: %s %s %s", threading.currentThread().getName(),
              self.action.__name__, self.done, seq)
    while self.done < seq:
      with cond:
        # The leader thread has to wait to obey the specified delay. The passenger/follower
        # threads will keep on waiting until the leader has finished execution
        cond.wait(self.delay)
      LOG.debug("[%s] Done wait: %s %s %s", threading.currentThread().getName(),
                self.action.__name__, do_exec, self.done)

      # "done" holds the "seq" value for the last completed action. The train will only start when
      # current "seq" is a successor to done. The "do_exec" flag ensures the leader can only start
      # the train
      if do_exec and self.done == seq - 1:
        # The train is leaving -- only the threads that are currently aboard will continue to have
        # access to the "cond" object on their stacks.  Any other thread that enters this function
        # after this point will have to make a new train.
        with self.lock:
          # This is locked so as to ensure that self.cond doesn't change during the
          # locked code sequence at the top of this function.
          self.cond = None

        # At this point the train is moving, leaving the station but still in the station.
        # We have to wait until it has left to mark it as such (self.done = seq).
        # That will open the "if" statement gate above and ensures that there are never two
        # actions in flight at the same time.
        self.action()

        self.done = seq
        with cond:
          cond.notify_all()

  @classmethod
  def schedule(cls, action, delay=DEFAULT_DELAY):
    """This is the interface function for scheduling a synced action
    """
    with cls.action_lock:
      synced_action = cls.action_map.get(action.__name__, None)
      if synced_action is None:
        synced_action = SyncedAction(action, delay)
        cls.action_map[action.__name__] = synced_action

    synced_action()

