"""
Copyright (c) 2011-2013 Cloudera, Inc. All rights reserved.

URL manipulation and retrieval utilities.
"""
from __future__ import absolute_import

# pylint: disable=too-many-lines,too-many-public-methods,too-many-statements
# pylint: disable=too-many-instance-attributes,import-error
# pylint: disable=line-too-long,bad-continuation,bare-except,broad-except,invalid-name,
# pylint: disable=missing-docstring,too-many-branches,too-many-lines

import httplib
import logging
import socket
import sys
import time
import urllib2
import cmf.https

from urllib2_kerberos import HTTPKerberosAuthHandler

from M2Crypto import httpslib
from M2Crypto import SSL
from M2Crypto import m2

LOG = logging.getLogger(__name__)

# urlopen_with_timeout.
#
# The optional secure_http_service_name parameter allows callers to connect to
# secure HTTP servers via the urllib2_kerberos library. We have a modified
# version of the HTTPKerberosAuthHandler code which takes the Kerberos service
# name rather than construct the name using the HTTP request host. We always add
# the HTTPKerberosAuthHandler to urllib2 opener handlers because it has no effect
# if security is not actually enabled.
#
# The optional username and pasword parameters similarly handle setting up HTTP
# digest authentication. Again, this has no effect if HTTP digest authentication
# is not in use on the connection.
#
# The cafile, capath and max_cert_depth control the SSL certificate verification
# behavior. https://www.openssl.org/docs/ssl/SSL_CTX_load_verify_locations.html
# explains the semantics of the parameters. Passing none for both means that
# no verification of the server certification (including the server's hostname)
# will be performed.
def urlopen_with_timeout(url,
                         data=None,
                         timeout=None,
                         secure_http_service_name=None,
                         username=None,
                         password=None,
                         cafile=None,
                         capath=None,
                         max_cert_depth=9):
  # pylint: disable=too-many-arguments
  # OPSAPS-41908 - M2Crypto barfs on unicode objects
  if isinstance(url, unicode):
    url = str(url)
  full_url = _get_full_url(url)
  openers = []
  openers.append(_make_https_handler(cafile,
                                     capath,
                                     max_cert_depth))
  if cmf.https.ssl_url_opener and cmf.https.ssl_url_opener.handler:
    openers.append(cmf.https.ssl_url_opener.handler)
  openers.append(HTTPKerberosAuthHandler(secure_http_service_name))

  if username is not None and password is not None:
    openers.append(_make_http_digest_auth_handler(full_url, username, password))

  if sys.version_info < (2, 6):
    # The timeout parameter to urlopen was introduced in Python 2.6.
    # To workaround it in older versions of python, we copy, with
    # minor modification, httplib.HTTPConnection, and hook it all
    # up.
    openers.append(_make_timeout_handler(timeout))
    opener = urllib2.build_opener(*openers)
    return opener.open(url, data)

  opener = urllib2.build_opener(*openers)
  return opener.open(url, data, timeout)

def _get_full_url(url):
  # OPSAPS-41908 - M2Crypto barfs on unicode objects
  if isinstance(url, unicode):
    url = str(url)
    full_url = url
  elif isinstance(url, urllib2.Request):
    full_url = str(url.get_full_url())
  elif isinstance(url, str):
    full_url = url
  else:
    raise Exception("Unexpected URL type: %s" % (type(url),))
  return full_url

def head_request_with_timeout(url,
                              data=None,
                              timeout=None,
                              secure_http_service_name=None,
                              username=None,
                              password=None,
                              cafile=None,
                              capath=None,
                              max_cert_depth=9):
  # pylint: disable=too-many-arguments
  class HeadRequest(urllib2.Request):
    def get_method(self):
      return "HEAD"

  if isinstance(url, urllib2.Request):
    raise Exception("Unsupported url type: urllib2.Request.")
  # OPSAPS-41908 - M2Crypto barfs on unicode objects

  if isinstance(url, unicode):
    url = str(url)

  return urlopen_with_timeout(HeadRequest(url),
                              data,
                              timeout,
                              secure_http_service_name,
                              username,
                              password,
                              cafile,
                              capath,
                              max_cert_depth)

def _make_timeout_handler(timeout):
  # Create these two helper classes fresh each time, since
  # timeout needs to be in the closure.

  class TimeoutHTTPConnection(httplib.HTTPConnection):
    def connect(self):
      """Connect to the host and port specified in __init__."""
      msg = "getaddrinfo returns an empty list"
      for res in socket.getaddrinfo(self.host, self.port, 0,
                                    socket.SOCK_STREAM):
        af, socktype, proto, _, sa = res
        try:
          self.sock = socket.socket(af, socktype, proto)
          if timeout is not None:
            self.sock.settimeout(timeout)
          if self.debuglevel > 0:
            print "connect: (%s, %s)" % (self.host, self.port)
          self.sock.connect(sa)
        except socket.error, msg:
          if self.debuglevel > 0:
            print 'connect fail:', (self.host, self.port)
          if self.sock:
            self.sock.close()
          self.sock = None
          continue
        break
      if not self.sock:
        raise socket.error, msg

  class TimeoutHTTPHandler(urllib2.HTTPHandler):
    http_request = urllib2.AbstractHTTPHandler.do_request_
    def http_open(self, req):
      return self.do_open(TimeoutHTTPConnection, req)
    https_open = http_open

  return TimeoutHTTPHandler

def _make_http_digest_auth_handler(url, username, password):
  password_manager = urllib2.HTTPPasswordMgrWithDefaultRealm()
  password_manager.add_password(None, # realm
                                url,
                                username,
                                password)
  return urllib2.HTTPDigestAuthHandler(password_manager)

def _make_https_handler(cafile=None,
                        capath=None,
                        max_cert_depth=9):
  class HTTPSConnection(httpslib.HTTPSConnection):
    """
    A class that extends the default HTTPSConnection to ensure two things:
    1) Enforce tlsv1 protocol for all ssl connection. Some older pythons
       (e.g., sles11, probably all versions <= 2.6) attempt SSLv23 handshake
       that is rejected by newer web servers. See OPSAPS-32192 for an example.
    2) Force validation if cafile/capath is supplied.
    """
    # pylint: disable=too-few-public-methods

    def __init__(self, host, port=None, **ssl):
      # pylint: disable=unused-argument
      self.sock = None
      # Specifying sslv23 enables the following ssl versions:
      # SSLv3, SSLv23, TLSv1, TLSv1.1, and TLSv1.2. We will explicitly exclude
      # SSLv3 and SSLv2 below. This mimics what is done by create_default_context
      # on newer python versions (python >= 2.7).
      ctx = SSL.Context('sslv23')

      # SSL_OP_ALL turns on all workarounds for known bugs. See
      # https://www.openssl.org/docs/manmaster/ssl/SSL_CTX_set_options.html for
      # a full list of these workarounds. I believe that we don't really need
      # any of these workarounds, but, this is default in later pythons and is
      # future looking.
      ctx.set_options(m2.SSL_OP_ALL | m2.SSL_OP_NO_SSLv2 | m2.SSL_OP_NO_SSLv3)

      if cafile is not None or capath is not None:
        ctx.set_verify(SSL.verify_peer | SSL.verify_fail_if_no_peer_cert,
                       max_cert_depth)
        ctx.load_verify_info(cafile=cafile, capath=capath)
        self._postConnectionCheck = True
      else:
        ctx.set_verify(SSL.verify_none, max_cert_depth)
        self._postConnectionCheck = False
      httpslib.HTTPSConnection.__init__(self, host, port, ssl_context=ctx)

    def connect(self):
      # This is a bit ugly but we need to override the connect method in order
      # to disable hostname verification. This is buried deep inside M2Crypto
      # and the only way to disable it is to disable post connection checks on
      # the socket itself.
      self.sock = SSL.Connection(self.ssl_ctx)
      if self.session:
        self.sock.set_session(self.session)

      if not self._postConnectionCheck:
        self.sock.postConnectionCheck = None

      self.sock.connect((self.host, self.port))

  class HTTPSHandler(urllib2.HTTPSHandler):

    def https_open(self, req):
      return self.do_open(HTTPSConnection, req)

  return HTTPSHandler()

def urlopen_with_retry_on_authentication_errors(function,
                                                retries,
                                                sleeptime):
  # See OPSAPS-28469: we retry on 401 errors on the presumption that we
  # are hitting a race with the kinit from the kt_renewer.
  attempt = 1
  while True:
    try:
      return function()
    except urllib2.HTTPError, err:
      if err.code == 401 and attempt <= retries:
        LOG.exception("Autentication error on attempt %d. Retrying after "
                      "sleeping %f seconds.", attempt, sleeptime)
        time.sleep(sleeptime)
        attempt += 1
      else:
        raise
