# Copyright (c) 2017 Cloudera, Inc. All rights reserved.
"""
Implementation of a work queue
"""
from __future__ import absolute_import
import threading
import logging

import Queue
import time

LOG = logging.getLogger(__name__)

class WorkQueue(object):
  """
  The work queue class abstracts the queue and the workers
  """
  # pylint: disable=too-few-public-methods
  class Result(object):
    """
    Represents a future result of a work queue-performed operation.
    """
    def __init__(self):
      """
      Constructor
      """
      self.value = None
      self.exc = None
      self._cond = threading.Condition()

    def finish(self, val, exc):
      """
      Called by WorkQueue to supply an operation result.
      Wakes up any threads in a wait() call.

      :param val: Return value of operation.
      :param exc: Exception raised by operation. Should be False in case of no exception.
      :return: None.
      """
      with self._cond:
        self.value = val
        self.exc = exc
        self._cond.notify_all()

    def is_done(self):
      """
      Returns whether a worker has completed the task and the result is available. This method is
      not thread-safe. The calling function should ensure required synchronizations
      :return: True, if the task is done
      """
      return self.exc is not None

    def complete(self):
      """
      If the operation for the result has completed return the operation's return value,
      or if there was an exception, re-raise that exception.
      :return:
      """
      # pylint: disable=raising-bad-type
      if self.exc is False:
        return self.value
      if self.exc is not None:
        raise self.exc

    def wait(self, timeout=None):
      """
      Wait until the result is ready; then execute complete().
      :param timeout: Optional timeout.
      :return:
      """
      deadline = time.time() + timeout if timeout is not None else None
      with self._cond:
        while not self.is_done():
          if time.time() >= deadline:
            return None
          wait_interval = max(0.1, deadline - time.time()) if deadline is not None else None
          self._cond.wait(wait_interval)
      return self.complete()

  def __init__(self, num_threads=4):
    self.threads = [threading.Thread(target=self.__run_queue) for _ in range(num_threads)]
    self.queue = Queue.Queue()
    for thr in self.threads:
      thr.daemon = True
      thr.start()

  def execute(self, action):
    """
    Perform an action asynchronously.  Returns a "Result" object that will subsequently
    communicate the result of the asynchronously invoked action.

    :param action: Callable to be executed - takes no arguments.
    :return:
    """
    result = self.Result()
    self.queue.put_nowait((action, result))
    return result

  def __run_queue(self):
    """
    Internal body of thread pool members; pick next item from work queue and execute it.
    """
    # pylint: disable=broad-except
    while True:
      (action, result) = self.queue.get(True)

      try:
        action_result = action()
        result.finish(action_result, False)
      except Exception as exc:
        result.finish(None, exc)

