#!/usr/bin/env python
from __future__ import absolute_import
# Copyright (c) 2012 Cloudera, Inc. All rights reserved.

import glob
import logging
import os
import re
import subprocess

import cmf.version
from cmf.util import fix_hue_home, CDH4_HUE_HOME, fix_hcat_home, CDH4_HCAT_HOME
from cmf.throttling_logger import ThrottlingLogger

# Some components are special and have an env var that doesn't match their
# proper component name
SPECIAL_ENV_TO_COMPONENT_NAME = {'CDH_HCAT_HOME' : 'hcatalog',
                                 'KEYTRUSTEE_KP_HOME' : 'keytrustee_kp',
				 'KEYTRUSTEE_SERVER_HOME' : 'keytrustee_server'}

SPECIAL_SCRIPTS = {'mr1' : '/bin/hadoop',
                   'mr2' : '/bin/mapred',
                   'httpfs' : '/sbin/httpfs.sh',
                   'zookeeper' : '/bin/zkServer.sh',
                   'hue' : '/build/env/bin/hue',
                   'flume' : '/bin/flume-ng',
                   'impala' : '/sbin/impalad',
                   'hcatalog' : '/sbin/webhcat_server.sh',
                   'sqoop2' : '/bin/sqoop.sh',
                   'solr' : '/bin/solrd',
                   'spark' : '/bin/spark-shell',
                   'crunch' : '/crunch-core-*.jar',
                   'hbase_indexer' : '/bin/hbase-indexer',
                   'parquet' : '/parquet-*.jar',
                   'kms' : '/sbin/kms.sh',
                   'kafka' : '/bin/kafka-server-start.sh',
                   'keytrustee_kp' : '/lib/keytrustee*.jar',
                   'keytrustee_server' : '',}

AVAILABLE_CDH_VERSIONS = ['cdh3', 'cdh4', 'cdh5', 'cdh6']

LOG = logging.getLogger(__name__)
THROTTLED_LOG = ThrottlingLogger(LOG)

VERSION_FILE_CDH_KEY='cloudera.cdh.release'
VERSION_FILE_COMPONENT_KEY='cloudera.pkg.version'
VERSION_FILE_PKG_NAME_KEY='cloudera.pkg.name'
VERSION_FILE_RELEASE_KEY='cloudera.pkg.release'

NAME_KEY='name'
SOURCE_KEY='source'
ACTIVE_KEY='active'
CDH_VERSION_KEY='cdh_version'
CDH_RELEASE_KEY='cdh_release'
COMPONENT_VERSION_KEY='component_version'
COMPONENT_RELEASE_KEY='component_release'
COMPONENT_CONFIG_KEY='component_config'
JAVA_HOME="JAVA_HOME"


class VersionDetector(object):

  cached_installed_versions = None

  @classmethod
  def _get_key_values_from_file(cls, version_file):
    """
    Returns a dict where lines in version_file are split as key=value.
    """
    version_dict = dict()
    f = open(version_file)
    contents = f.read().strip().splitlines()
    f.close()
    for line in contents:
      if line.find('=') >= 0:
        kv_array = line.split('=')
        if len(kv_array) == 2:
          version_dict[kv_array[0]] = kv_array[1]
    return version_dict

  @classmethod
  def _set_all_version_fields(cls, version_dict, value):
    version_dict[CDH_VERSION_KEY] = \
      version_dict[CDH_RELEASE_KEY] = \
      version_dict[COMPONENT_RELEASE_KEY] = \
      version_dict[COMPONENT_VERSION_KEY] = value

  @classmethod
  def _get_version_for_cdh_component(cls, cdh_component_home, is_cm=False):
    """
    Returns version for the CDH component.
    """
    version_file = cdh_component_home + "/cloudera/cdh_version.properties"
    if is_cm:
      version_file = cdh_component_home + "/cloudera/cm_version.properties"
    version_dict = dict()
    version_dict[SOURCE_KEY] = "PACKAGE"
    version_dict[ACTIVE_KEY] = True

    if os.path.exists(version_file):
      file_dict = cls._get_key_values_from_file(version_file)
      if VERSION_FILE_PKG_NAME_KEY in file_dict:
        version_dict[NAME_KEY] = file_dict[VERSION_FILE_PKG_NAME_KEY]
      if VERSION_FILE_CDH_KEY in file_dict and \
           VERSION_FILE_RELEASE_KEY in file_dict and \
           VERSION_FILE_COMPONENT_KEY in file_dict:
        version_dict[COMPONENT_VERSION_KEY] = file_dict[VERSION_FILE_COMPONENT_KEY]
        version_dict[COMPONENT_RELEASE_KEY] = file_dict[VERSION_FILE_RELEASE_KEY]
        cdh_release = file_dict[VERSION_FILE_CDH_KEY]
        if cdh_release == 'NA':
          version_dict[CDH_VERSION_KEY] = "NOT_APPLICABLE"
          # For non-CDH component,use pkg.name + pkg.version as release
          version_dict[CDH_RELEASE_KEY] = file_dict[VERSION_FILE_PKG_NAME_KEY] \
              + ' ' + version_dict[COMPONENT_VERSION_KEY]
          return version_dict

        try:
          # The first 4 chars represent CDH version.
          # This is good only till we get to CDH10.
          cdh_version = cdh_release[:4]
          if cdh_version in AVAILABLE_CDH_VERSIONS:
            # Return the correct version
            version_dict[CDH_VERSION_KEY] = cdh_version.upper()
            version_dict[CDH_RELEASE_KEY] = cls._format_cdh_release_string(cdh_release)
            return version_dict
          raise Exception("Unsupported version: " + cdh_version)
        except Exception, e:
          LOG.exception("Error while reading version file: " + version_file \
                        + " value: " + file_dict[VERSION_FILE_CDH_KEY] \
                        + "\nexception: " + str(e))
      # Version file is present but isn't in correct format or has unsupported version
      cls._set_all_version_fields(version_dict, 'BAD_VERSION_FILE')
    else:
      # Version file is not present
      cls._set_all_version_fields(version_dict, 'NO_VERSION_FILE')

    return version_dict

  @classmethod
  def _format_cdh_release_string(cls, release):
    # all cdh_version.properties should contain a cdh release string
    if release.startswith('cdh'):
      return 'cdh ' + release[3:]
    else:
      return release

  @classmethod
  def _check_file_exists_with_pattern(cls, directory, filename_pattern):
    """
    Returns true if a file with the given filename pattern exists in the
    specified directory.
    """
    if not os.path.exists(directory):
      return False
    for f in os.listdir(directory):
      if re.match(filename_pattern, f) is not None:
        return True
    return False

  @classmethod
  def get_component_versions(
    cls,
    supervisee_environment,
    package_dir,
    alternate_hue_home=CDH4_HUE_HOME,
    alternate_hcat_home=CDH4_HCAT_HOME,
    java_home_config=None,
    supervisor_info=None):
    """
    The main loop which goes through all the environment variables of
    type CDH_xxx_HOME and finds out the versions for the corresponding components.
    """
    installed_versions = []

    # Combine the OS environment and the supervisee environment
    all_env = os.environ.copy()
    all_env.update(supervisee_environment)

    # make sure we have the right hue home dir
    fix_hue_home(all_env, alternate_hue_home)
    # make sure we have the right hcat home dir
    fix_hcat_home(all_env, alternate_hcat_home)

    for k, v in all_env.iteritems():
      if k in SPECIAL_ENV_TO_COMPONENT_NAME:
        cdh_component = SPECIAL_ENV_TO_COMPONENT_NAME[k]
      else:
        cdh_match = re.match("CDH_(.+)_HOME", k)
        # ignore any env variable not in CDH_XXX_HOME format
        if cdh_match is None:
          continue

        # Note that component name may be different for
        # packages vs parcels. Parcel component names
        # are looked up directly from parcel metadata.
        cdh_component = cdh_match.group(1).lower()

      cdh_component_script = v + "/bin/" + cdh_component
      if cdh_component in SPECIAL_SCRIPTS:
        cdh_component_script = v + SPECIAL_SCRIPTS[cdh_component]

      version_info = dict()
      # returned if even the script doesn't exist for the component
      cls._set_all_version_fields(version_info, 'NO_SCRIPT')
      version_info[SOURCE_KEY] = "PACKAGE"
      version_info[ACTIVE_KEY] = True

      try:
        if cdh_component == 'hue_plugins':
          # Even though hue-plugins are a CDH component, they share their home with
          # hadoop and we already had a cdh-version.properties for that. So, we have
          # to use cm-version.properties for hue-plugins.
          if cls._check_file_exists_with_pattern(os.path.join(v, "lib"), "hue-plugins(.+).jar"):
            version_info = cls._get_version_for_cdh_component(v, True)
        elif len(glob.glob(cdh_component_script)) > 0:
          # Using glob to allow for wildcards for cases like crunch
          version_info = cls._get_version_for_cdh_component(v)

        if cdh_component == "hcatalog" and version_info[COMPONENT_VERSION_KEY] == 'NO_VERSION_FILE':
          # hcat has no version file in CDH4.2.0. Report the CDH version correctly
          version_info[CDH_VERSION_KEY] = "CDH4"
          version_info[CDH_RELEASE_KEY] = "cdh 4.2.0"
          version_info[COMPONENT_VERSION_KEY] = "0.4.0+215"
          version_info[COMPONENT_RELEASE_KEY] = "1.cdh4.2.0.p0.8"
      except Exception, e:
        LOG.exception("Error while reading version file.")

      # Use cdh_component only if no name.
      if NAME_KEY not in version_info:
        version_info[NAME_KEY] = cdh_component
      installed_versions.append(version_info)

    installed_versions += cls.get_agent_component_versions(
      supervisee_environment,
      package_dir,
      java_home_config,
      supervisor_info)
    cls.cached_installed_versions = installed_versions
    return installed_versions


  @classmethod
  def get_agent_component_versions(
    cls,
    supervisee_environment,
    package_dir,
    java_home_config,
    supervisor_info):
    installed_versions = []

    # Combine the OS environment and the supervisee environment
    all_env = os.environ.copy()
    all_env.update(supervisee_environment)

    # Add in mgmt daemons version
    if "MGMT_HOME" in all_env:
      version_info = dict()
      version_info[NAME_KEY] = "cm_daemons"
      version_info[CDH_VERSION_KEY] = version_info[CDH_RELEASE_KEY] = "NOT_APPLICABLE"
      version_info[COMPONENT_RELEASE_KEY] = "NO_SCRIPT"
      version_info[COMPONENT_VERSION_KEY] = "NO_SCRIPT"
      version_info[SOURCE_KEY] = "PACKAGE"
      version_info[ACTIVE_KEY] = True
      path = ""
      try:
        home = all_env.get("MGMT_HOME")
        path = os.path.join(home, "cloudera/cm_version.properties")
        file_dict = cls._get_key_values_from_file(path)
        version_info[COMPONENT_VERSION_KEY] = file_dict["version"]
        version_info[COMPONENT_RELEASE_KEY] = file_dict[VERSION_FILE_RELEASE_KEY]
        version_info[CDH_RELEASE_KEY] = "CM " + file_dict["version"]
      except Exception, e:
        LOG.exception("Error while reading version file:" + path)
      installed_versions.append(version_info)

    # Supervisord component
    if supervisor_info is not None:
      installed_versions.append(cls._get_supervisord_component(supervisor_info))

    cls._get_java_component_helper(
      all_env,
      installed_versions,
      package_dir,
      java_home_config)

    # Add in the agent version
    version_info = dict()
    version_info[NAME_KEY] = "cm_agent"
    version_info[CDH_VERSION_KEY] = version_info[CDH_RELEASE_KEY] = "NOT_APPLICABLE"
    version_info[COMPONENT_VERSION_KEY] = cmf.version.VERSION
    version_info[COMPONENT_RELEASE_KEY] = cmf.version.RELEASE
    version_info[SOURCE_KEY] = "PACKAGE"
    version_info[ACTIVE_KEY] = True
    installed_versions.append(version_info)
    return installed_versions

  @classmethod
  def _get_supervisord_component(cls, supervisor_info):
    version_info = dict()
    version_info[NAME_KEY] = supervisor_info["component_name"]
    version_info[CDH_VERSION_KEY] = version_info[CDH_RELEASE_KEY] = "NOT_APPLICABLE"
    version_info[COMPONENT_RELEASE_KEY] = "NO_SCRIPT"
    version_info[COMPONENT_VERSION_KEY] = supervisor_info["sup_version"]
    version_info[SOURCE_KEY] = "PACKAGE"
    version_info[ACTIVE_KEY] = True
    return version_info

  @classmethod
  def _get_java_component_helper(
    cls,
    all_env,
    installed_versions,
    package_dir,
    java_home_config):
    identified_java_versions = dict()
    if java_home_config is not None:
      cls.add_java_component(
        identified_java_versions,
        package_dir,
        '',
        java_home_config)
    elif "JAVA_HOME" in all_env:
      THROTTLED_LOG.warn("CMF_AGENT_JAVA_HOME environment variable host override will be deprecated in future. "
                         "JAVA_HOME setting configured from CM server takes precedence over host agent override. "
                         "Configure JAVA_HOME setting from CM server.")
      cls.add_java_component(
        identified_java_versions,
        package_dir,
        '',
        all_env.get("JAVA_HOME"))
    else:
      # Add in java 8 & 11
      cls.add_java_component(
        identified_java_versions,
        package_dir,
        '',
        '')

    # Populate installed_versions from entries in identified_java_versions
    for val in identified_java_versions.itervalues():
      installed_versions.append(val)

  @classmethod
  def add_java_component(
    cls,
    identified_java_versions,
    package_dir,
    version,
    java_home_config):
    version_info = dict()

    version_info[CDH_VERSION_KEY] = version_info[CDH_RELEASE_KEY] = "NOT_APPLICABLE"
    version_info[COMPONENT_RELEASE_KEY] = "NO_SCRIPT"
    version_info[COMPONENT_VERSION_KEY] = "NO_SCRIPT"
    version_info[SOURCE_KEY] = "SYSTEM"
    version_info[ACTIVE_KEY] = True
    try:
      java_info = cls.get_java_version(package_dir, version, java_home_config)
      version_info[COMPONENT_VERSION_KEY] = " ".join(java_info[1:])
      version_info[JAVA_HOME] = java_info[0].replace("JAVA_HOME=","")
      version_info[COMPONENT_CONFIG_KEY] =  { "JAVA_HOME" : version_info[JAVA_HOME] }

      if '' == version or 'misc' == version:
        version_pattern = re.compile(".*\"(\\d+)\\.(\\d+)\\.(\\d+)(_(\\d+))?.*")
        # Example java_info
        # ['JAVA_HOME=/usr/java/default',
        #  'java version "1.8.0_11"',
        #  'Java(TM) SE Runtime Environment (build 1.8.0_11-b12)',
        #  'Java HotSpot(TM) 64-Bit Server VM (build 25.11-b03, mixed mode)']
        #
        # sample text for jdk11:
        #
        # openjdk version "11.0.2" 2019-01-15
        # OpenJDK Runtime Environment 18.9 (build 11.0.2+9)
        # OpenJDK 64-Bit Server VM 18.9 (build 11.0.2+9, mixed mode)
        #
        full_java_version = version_pattern.match(java_info[1])
        # full_java_version would either match "1.8.0_11" or "11.0.2"
        if full_java_version.group(1) == "11":
          # java 11
          version_info[NAME_KEY] = "java" + full_java_version.group(1)
        else:
          # java 8
          version_info[NAME_KEY] = "java" + full_java_version.group(2)
      else:
        version_info[NAME_KEY] = "java" + version

      identified_java_versions[version_info[NAME_KEY]] = version_info
      THROTTLED_LOG.info("Identified java component %s with full version %s for requested version %s."
        % (version_info[NAME_KEY], version_info[COMPONENT_VERSION_KEY], version,))
    except Exception, e:
      LOG.debug("Error while reading Java %s version (might not be an error if this java version does not exist on this host)."
        % (version,))
      pass

  @classmethod
  def get_java_version(cls, package_dir, major_version, java_home_config):
    """Loads a shell script that resolves JAVA_HOME the same way that we do elsewhere"""
    script = os.path.join(package_dir, "mgmt/java_version.sh")
    p = subprocess.Popen([script, major_version, java_home_config],
      close_fds=True, stdin=None, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()
    if stdout:
      return stdout.split("\n")
    else:
      raise Exception("Failed to get Java version: %s" % (repr([stdout, stderr]), ))

  @classmethod
  def get_java_location(cls):

    # Order of preference
    javas = ["java8", "java11"]
    java_homes = {}
    if cls.cached_installed_versions is None:
      # This shouldn't be the case after the first HB
      raise Exception("JAVA components not yet discovered")

    for item in cls.cached_installed_versions:
      if item[NAME_KEY] in javas:
        java_homes[item[NAME_KEY]] = item[JAVA_HOME]
    for java in javas:
      try:
        return java_homes[java]
      except KeyError:
        pass
    return None
