# Copyright (c) 2013 Cloudera, Inc. All rights reserved.

import os
import simplejson

import avro.ipc as ipc
import avro.protocol as protocol
from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
from threading import Thread

PROTOCOL = protocol.parse(
  file(os.path.join(os.path.dirname(__file__), "../../cmf/audit/navigator.avpr")).read())

_event_handler = None


class AuditResponder(ipc.Responder):
  """
  This class responds to audit requests. Currently all requests are successful.
  In future it can be customized to behave differently for different requests.
  """

  def __init__(self):
      ipc.Responder.__init__(self, PROTOCOL)

  def invoke(self, msg, req):
    _event_handler.request_received(req)
    return {"success": True}


class AuditRequestHandler(BaseHTTPRequestHandler):
  """
  This class receives HTTP request from audit clients. It delegates actual
  handling to AuditResponder taking care of HTTP aspect.
  """

  def do_POST(self):
    self.responder = AuditResponder()
    call_request_reader = ipc.FramedReader(self.rfile)
    call_request = call_request_reader.read_framed_message()
    resp_body = self.responder.respond(call_request)
    self.send_response(200)
    self.send_header('Content-Type', 'avro/binary')
    self.end_headers()
    resp_writer = ipc.FramedWriter(self.wfile)
    resp_writer.write_framed_message(resp_body)

class AuditTpRequestHandler(BaseHTTPRequestHandler):
  """
  This class receives HTTP request from audit clients. It delegates actual
  handling to AuditResponder taking care of HTTP aspect.
  """
  def do_POST(self):
    length = int(self.headers.get('content-length'))
    content = self.rfile.read(length)
    req = {'request':{'events': []}}
    content = simplejson.loads(content)
    for entry in content:
      for keys in entry:
        req['request']['serviceName'] = entry[keys]['serviceName']
      req['request']['events'].append(entry)
    _event_handler.request_received(req)
    self.send_response(204)
    self.end_headers()

class ProfileRequestHandler(BaseHTTPRequestHandler):
  """
  This class receives HTTP request from profile clients. It delegates actual
  handling to AuditResponder taking care of HTTP aspect.
  """

  def do_POST(self):
    length = int(self.headers.get('content-length'))
    content = self.rfile.read(length)
    req = {'request':{'events': []}}
    content = simplejson.loads(content)
    for entry in content:
        req['request']['serviceName'] = "IMPALA-1"
        req['request']['events'].append(entry)
    _event_handler.request_received(req)
    self.send_response(204)
    self.end_headers()

class MetadataRequestHandler(BaseHTTPRequestHandler):
  """
  This class receives HTTP request from audit clients. It delegates actual
  handling to AuditResponder taking care of HTTP aspect.
  """

  def do_POST(self):
    length = int(self.headers.get('content-length'))
    content = self.rfile.read(length)
    _event_handler.request_received(self.path, content)
    self.send_response(204)
    self.end_headers()


class NavigatorServer(object):
  """
  This class runs a real server that unit tests can start and stop and use to
  test agent audit/metadata code with a real server. We do have unit tests that
  catch a lot of issues, but we see integration failures when all the code is
  deployed in real system. This server will help catch inter-module/inter-class
  dependency issue using nose tests.
  """

  _server_addr = ('0.0.0.0', 0)  # OS will pick up a free port

  def __init__(self, event_handler, server_type, request_handler=AuditRequestHandler):
    """
    Initializes NavigatorServer object with an event to trigger when a request
    arrives and gets successfully processed.
    """
    global _event_handler
    _event_handler = event_handler

    if server_type == "tp":
      self._request_handler = AuditTpRequestHandler
    elif server_type == "profile":
      self._request_handler = ProfileRequestHandler
    else:
      self._request_handler = request_handler

  def start_server(self):
    """
    Start a new thread for navigator server requests.
    """
    self._server = HTTPServer(self._server_addr, self._request_handler)
    self._navigator_server = Thread(target=self._listen_for_connections,
                                name="MetadataServerTestThread")
    self._navigator_server.setDaemon(True)
    self._navigator_server.start()

  def start_server_on_host_port(self, host, port):
    """
    Start a new thread for navigator server requests.
    """
    self._server_addr = (host, port)
    self._server = HTTPServer(self._server_addr, self._request_handler)
    self._navigator_server = Thread(target=self._listen_for_connections,
                                name="MetadataServerTestThread")
    self._navigator_server.setDaemon(True)
    self._navigator_server.start()

  def _listen_for_connections(self):
    self._server.allow_reuse_address = True
    self._server.serve_forever()

  def stop_server(self):
    """
    'shutdown' method was not available in python 2.4, so catching
    AttributeError and ignoring it.
    """
    try:
      self._server.shutdown()
      self._server.server_close()
    except AttributeError:
      # Ignore it as shutdown method isn't present in python 2.4
      pass

  def get_port(self):
    return self._server.server_port

  def get_host(self):
    return self._server.server_name
