# Copyright (c) 2013 Cloudera, Inc. All rights reserved.
import os
import shutil

from cmf.audit.navigator_logs import NavigatorLogDirectory
from nose.tools import assert_equal, assert_false, assert_true

_AUDIT_LOG_DIR = "/tmp/audit"
_AUDIT_LOG_PREFIX = "impala_audit_event_log_1.0"
_AUDIT_FILE_NAME1 = "impala_audit_event_log_1.0-1377798507857"
_AUDIT_FILE_NAME2 = "impala_audit_event_log_1.0-1377798507877"
_AUDIT_FILE_NAME3 = "impala_audit_event_log_1.0-1377798507897"
_AUDIT_LOG_TEXT = """{"1377775857227":{"query_id":"84bd714e4ef1fa3:75ff54a3b1824295","session_id":"2d4c159702aeebe2:cb5377ebedf6168b","start_time":"2013-08-29 11:30:56.897841000","authorization_failure":false,"status":"","user":"impalauser","impersonator":null,"statement_type":"CREATE_DATABASE","network_address":"::ffff:127.0.0.1:34009","sql_statement":"create DATABASE impalaQueryDB_1377801039","catalog_objects":[{"name":"impalaQueryDB_1377801039","object_type":"DATABASE","privilege":"CREATE"}]}}
{"1377775857472":{"query_id":"70413d07e1ee7ec6:5d1b5f76aee04183","session_id":"2d4c159702aeebe2:cb5377ebedf6168b","start_time":"2013-08-29 11:30:57.441235000","authorization_failure":false,"status":"","user":"impalauser","impersonator":null,"statement_type":"USE","network_address":"::ffff:127.0.0.1:34009","sql_statement":"use impalaQueryDB_1377801039","catalog_objects":[{"name":"impalaQueryDB_1377801039","object_type":"DATABASE","privilege":"ANY"}]}}
{"1377775857516":{"query_id":"d04daaf9e488e953:5cb8bdc461eb9895","session_id":"2d4c159702aeebe2:cb5377ebedf6168b","start_time":"2013-08-29 11:30:57.488402000","authorization_failure":false,"status":"","user":"impalauser","impersonator":null,"statement_type":"DROP_TABLE","network_address":"::ffff:127.0.0.1:34009","sql_statement":"drop TABLE IF EXISTS testaudittable1","catalog_objects":[]}}"""

class TestAuditLogDirectory(object):
  """
  This class tests if audit logs can be tail'ed under various
  conditions:
  1. No log exist
  2. Log currently being tailed disappears
  3. Log gets corrupted audit line
  4. Log gets rolled over
  and more cases as we discover them ...
  """

  def setup(self):
    shutil.rmtree(_AUDIT_LOG_DIR, True)
    os.makedirs(_AUDIT_LOG_DIR)
    self._audit_logs = NavigatorLogDirectory("Impalad1",
                                         _AUDIT_LOG_DIR,
                                         _AUDIT_LOG_PREFIX)

  def teardown(self):
    shutil.rmtree(_AUDIT_LOG_DIR)

  def _create_audit_log_file(self, file_name, content = "Test Audit Log Line"):
    """
    Creates a new audit log file and returns the path to the file.
    """
    audit_file_path = os.path.join(_AUDIT_LOG_DIR, file_name)
    audit_file = open(audit_file_path, "w")
    print >> audit_file, content
    audit_file.close()
    self._audit_logs.append_new_log_to_list(file_name)

    return audit_file_path

  def test_no_audit_log(self):
    """
    Test the case when there is no log in the audit log directory.
    """
    audit_log = self._audit_logs.get_log(None)
    assert_false(audit_log.exists())

  def test_non_existent_audit_log(self):
    """
    Test if a audit log file name that doesn't exist is handled correctly.
    """
    audit_log = self._audit_logs.get_log("nonexistent_file_name")
    assert_false(audit_log.exists())

  def test_first_audit_log(self):
    """
    Test the case when auditing starts for the first time, it picks up the
    first audit log.
    """
    self._create_audit_log_file(_AUDIT_FILE_NAME1)
    self._create_audit_log_file(_AUDIT_FILE_NAME2)
    audit_log = self._audit_logs.get_log(None)
    assert_true(audit_log.exists())
    assert_equal(audit_log.get_file_name(), _AUDIT_FILE_NAME1)

  def test_get_audit_log(self):
    """
    Test the case when more than one audit log is present and we can get the
    one that is current one.
    """
    self._create_audit_log_file(_AUDIT_FILE_NAME1)
    self._create_audit_log_file(_AUDIT_FILE_NAME2)
    self._create_audit_log_file(_AUDIT_FILE_NAME3)
    audit_log = self._audit_logs.get_log(_AUDIT_FILE_NAME2)
    assert_true(audit_log.exists())
    assert_equal(audit_log.get_file_name(), _AUDIT_FILE_NAME2)

  def test_get_next_audit_log(self):
    """
    Test the case when more than one audit log is present and we can get the
    one that is next to current one.
    """
    self._create_audit_log_file(_AUDIT_FILE_NAME1)
    self._create_audit_log_file(_AUDIT_FILE_NAME2)
    self._create_audit_log_file(_AUDIT_FILE_NAME3)
    next_audit_log = self._audit_logs.get_next_log_file_name(_AUDIT_FILE_NAME2)
    assert_equal(next_audit_log, _AUDIT_FILE_NAME3)
    next_audit_log = self._audit_logs.get_next_log_file_name(_AUDIT_FILE_NAME1)
    assert_equal(next_audit_log, _AUDIT_FILE_NAME2)
    # There is no next audit log after last file
    next_audit_log = self._audit_logs.get_next_log_file_name(_AUDIT_FILE_NAME3)
    assert_true(next_audit_log is None)

  def test_get_next_audit_log_first_time(self):
    """
    Test that we can get next log file name when we start first time and the
    current log file in None.
    """
    self._create_audit_log_file(_AUDIT_FILE_NAME1)
    next_audit_log = self._audit_logs.get_next_log_file_name(None)
    assert_equal(next_audit_log, _AUDIT_FILE_NAME1)

  def test_get_next_audit_log_no_audit_log(self):
    """
    Test that a call to get_next_audit_log works when there are no audit logs.
    """
    next_audit_log = self._audit_logs.get_next_log_file_name(None)
    assert_true(next_audit_log is None)

  def test_get_next_audit_log_missing_current_log(self):
    """
    Test that a call to get_next_audit_log works when the current audit log
    was deleted/removed.
    """
    self._create_audit_log_file(_AUDIT_FILE_NAME2)
    self._create_audit_log_file(_AUDIT_FILE_NAME3)
    next_audit_log = self._audit_logs.get_next_log_file_name(_AUDIT_FILE_NAME1)
    assert_equal(next_audit_log, _AUDIT_FILE_NAME2)
    next_audit_log = self._audit_logs.get_next_log_file_name("impala_audit_event_log_1.0-1377798507878")
    assert_equal(next_audit_log, _AUDIT_FILE_NAME3)

  def test_remove_audit_log(self):
    """
    Test log deletion/removal is handled correctly.
    """
    self._create_audit_log_file(_AUDIT_FILE_NAME1)
    self._create_audit_log_file(_AUDIT_FILE_NAME2)
    self._create_audit_log_file(_AUDIT_FILE_NAME3)
    next_audit_log = self._audit_logs.get_next_log_file_name(_AUDIT_FILE_NAME1)
    assert_equal(next_audit_log, _AUDIT_FILE_NAME2)

    # Now get rid of _AUDIT_FILE_NAME2, we should then get _AUDIT_FILE_NAME3
    # as the next file
    self._audit_logs.remove_log_from_list(_AUDIT_FILE_NAME2);
    next_audit_log = self._audit_logs.get_next_log_file_name(_AUDIT_FILE_NAME1)
    assert_equal(next_audit_log, _AUDIT_FILE_NAME3)

  def test_read_audit_log(self):
    """
    Test if we can successfully read content of audit log.
    """
    self._create_audit_log_file(_AUDIT_FILE_NAME1, _AUDIT_LOG_TEXT)
    audit_log = self._audit_logs.get_log(_AUDIT_FILE_NAME1)
    events = [event.values()[0]  for _, _, _, event in audit_log.read_nav_log()]
    assert_equal(len(events), 3)
    for event in events:
      assert_false(event["authorization_failure"])
      assert_equal(event["session_id"], '2d4c159702aeebe2:cb5377ebedf6168b')

  def test_corrupted_audit_log_line(self):
    """
    Test if an audit log line gets partially written/flushed, we are able to
    handle it correctly. Split the line in two write operations and make sure
    we don't miss anything.
    """
    audit_line = """{"1377775857472":{"query_id":"70413d07e1ee7ec6:5d1b5f76aee04183","session_id":"2d4c159702aeebe2:cb5377ebedf6168b","start_time":"2013-08-29 11:30:57.441235000","authorization_failure":false,"status":"","user":"impalauser","impersonator":null,"statement_type":"USE","network_address":"::ffff:127.0.0.1:34009","sql_statement":"use impalaQueryDB_1377801039","catalog_objects":[{"name":"impalaQueryDB_1377801039","object_type":"DATABASE","privilege":"ANY"}]}}"""
    partial_audit_line1 = audit_line[:50]
    partial_audit_line2 = audit_line[50:]
    audit_file_path = self._create_audit_log_file(_AUDIT_FILE_NAME1, _AUDIT_LOG_TEXT)

    # Write partial line
    audit_file = open(audit_file_path, "a+")
    print >> audit_file, partial_audit_line1,
    audit_file.flush()
    os.fsync(audit_file)

    audit_log = self._audit_logs.get_log(_AUDIT_FILE_NAME1)
    events = [event.values()[0]  for _, _, _, event in audit_log.read_nav_log()]
    assert_equal(len(events), 3)
    for event in events:
      assert_false(event["authorization_failure"])
      assert_equal(event["session_id"], '2d4c159702aeebe2:cb5377ebedf6168b')

    # Now write the rest of line and confirm that we can read that
    print >> audit_file, partial_audit_line2
    print >> audit_file, _AUDIT_LOG_TEXT
    audit_file.flush()
    os.fsync(audit_file)
    audit_file.close()

    events = [event.values()[0]  for _, _, _, event in audit_log.read_nav_log()]
    assert_equal(len(events), 4)
    for event in events:
      assert_false(event["authorization_failure"])
      assert_equal(event["session_id"], '2d4c159702aeebe2:cb5377ebedf6168b')

  def test_invalid_audit_log_line(self):
    """
    Test if an audit log line is invalid JSON we abort.
    """
    audit_file_path = self._create_audit_log_file(_AUDIT_FILE_NAME1, _AUDIT_LOG_TEXT)

    # Write an invalid line
    audit_file = open(audit_file_path, "a+")
    print >> audit_file, "Invalid JSON audit line"
    print >> audit_file, _AUDIT_LOG_TEXT
    audit_file.flush()
    os.fsync(audit_file)

    audit_log = self._audit_logs.get_log(_AUDIT_FILE_NAME1)
    events = [event.values()[0]  for _, _, _, event in audit_log.read_nav_log()]
    # There are 7 lines in the file total: 3 valid lines, one invalid, then
    # 3 more valid lines. Since we expect to abort when we reach the invalid
    # line, `events` should only contain the first 3 lines.
    assert_equal(len(events), 3)
    for event in events:
      assert_false(event["authorization_failure"])
      assert_equal(event["session_id"], '2d4c159702aeebe2:cb5377ebedf6168b')

  def test_audit_log_purge(self):
    """
    Test if audit logs more than max_num_audit_log get purged successfully.
    """
    self._create_audit_log_file(_AUDIT_FILE_NAME1)
    self._create_audit_log_file(_AUDIT_FILE_NAME2)
    self._create_audit_log_file(_AUDIT_FILE_NAME3)

    self._audit_logs.purge_old_nav_logs(_AUDIT_FILE_NAME2, 2)
    assert_equal(3, len(os.listdir(_AUDIT_LOG_DIR)))

    self._audit_logs.purge_old_nav_logs(_AUDIT_FILE_NAME2, 1)
    assert_equal(2, len(os.listdir(_AUDIT_LOG_DIR)))

    self._create_audit_log_file(_AUDIT_FILE_NAME1)
    self._audit_logs.purge_old_nav_logs(_AUDIT_FILE_NAME3, 2)
    assert_equal(2, len(os.listdir(_AUDIT_LOG_DIR)))

    self._audit_logs.purge_old_nav_logs(_AUDIT_FILE_NAME3, 1)
    assert_equal(1, len(os.listdir(_AUDIT_LOG_DIR)))

  def test_get_bytes_to_send_empty_dir(self):
    """
    Test that we get correct data when audit log directory contains no audit
    logs.
    """
    none_audit_log = self._audit_logs.get_log(None)
    bytes_to_send = self._audit_logs.get_bytes_to_send(none_audit_log)
    assert_equal(0, bytes_to_send)

  def test_get_bytes_to_send_all_files(self):
    """
    Test that if audit log directory contains multiple files, and we start from
    first file, we get size of all files returned as bytes to send.
    """
    none_audit_log = self._audit_logs.get_log(None)

    self._create_audit_log_file(_AUDIT_FILE_NAME1, _AUDIT_LOG_TEXT)
    self._create_audit_log_file(_AUDIT_FILE_NAME2, _AUDIT_LOG_TEXT)
    self._create_audit_log_file(_AUDIT_FILE_NAME3, _AUDIT_LOG_TEXT)

    # file length has one extra newline character at the very end
    file_len = len(_AUDIT_LOG_TEXT) + 1

    bytes_to_send = self._audit_logs.get_bytes_to_send(none_audit_log)
    assert_equal(file_len * 3, bytes_to_send)

  def test_get_bytes_to_send_for_partially_sent_dir(self):
    """
    Test that if audit log directory contains multiple files, and we have already
    sent some data to server, we get remaining bytes as data yet to be sent
    """
    self._create_audit_log_file(_AUDIT_FILE_NAME1, _AUDIT_LOG_TEXT)
    self._create_audit_log_file(_AUDIT_FILE_NAME2, _AUDIT_LOG_TEXT)
    self._create_audit_log_file(_AUDIT_FILE_NAME3, _AUDIT_LOG_TEXT)

    first_file_offset = 510
    first_audit_log = self._audit_logs.get_log(None)
    first_audit_log.seek(first_file_offset)

    # file length has one extra newline character at the very end
    file_len = len(_AUDIT_LOG_TEXT) + 1

    bytes_to_send = self._audit_logs.get_bytes_to_send(first_audit_log)
    assert_equal(file_len * 3 - first_file_offset, bytes_to_send)
