# Copyright (c) 2013 Cloudera, Inc. All rights reserved.

import cmf.audit
import os
import shutil

from cmf.audit import AUDIT_CONFIG_FILENAME
from cmf.audit.navigator_thread import AuditThread
from cmf.audit.util import make_audit_plugin
from cmf_test.audit.navigator_server import NavigatorServer
from nose.tools import assert_true, assert_equal
from cmf_test.audit.test_util import wait_and_assert_true
from cmf_test.audit.test_util import skiprhel5

from threading import Event

_AUDIT_LOG_DIR = "/tmp/audit_log"
_ROLE_CONFIG_DIR = "/tmp/nav_client_config_dir"
_ROLE_CONFIG_DIR2 = "/tmp/nav_client_config_dir2"
_MAX_WAIT_TIMEOUT_5_MINS = 5*60

_GENERIC_AUDIT_FILE_PREFIX = "GENERIC-1-GENERIC-abee5a758942d6c689c7a5e7dd2da179"
_GENERIC_AUDIT_FILE_NAME = _GENERIC_AUDIT_FILE_PREFIX + "-NavigatorLog4jAppender-1385512800584"
#TODO: test with non-String extra values.
_GENERIC_AUDIT_LOG_TEXT = """{"serviceName":"GENERIC-1","allowed":true,"username":"abcd","eventTime":1385512859079,"ipAddress":"10.20.187.50","operation":"one","impersonator":null,"extraValue1":null,"extraValue2":"value","extraValue3":123}
{"serviceName":"GENERIC-1","allowed":true,"username":"efgh","eventTime":1385512859376,"ipAddress":"10.20.187.50","operation":"three","impersonator":null,"extraValue1":null,"extraValue2":"anotherValue","extraValue3":123}"""

_IMPALA_AUDIT_FILE_NAME = "impala_audit_event_log_1.0-1377798507857"
_IMPALA_AUDIT_LOG_TEXT = """{"1377775857227":{"query_id":"84bd714e4ef1fa3:75ff54a3b1824295","session_id":"2d4c159702aeebe2:cb5377ebedf6168b","start_time":"2013-08-29 11:30:56.897841000","authorization_failure":false,"status":"","user":"impalauser","impersonator":null,"statement_type":"CREATE_DATABASE","network_address":"::ffff:127.0.0.1:34009","sql_statement":"create DATABASE impalaQueryDB_1377801039","catalog_objects":[{"name":"impalaQueryDB_1377801039","object_type":"DATABASE","privilege":"CREATE"}]}}
{"1377775857472":{"query_id":"70413d07e1ee7ec6:5d1b5f76aee04183","session_id":"2d4c159702aeebe2:cb5377ebedf6168b","start_time":"2013-08-29 11:30:57.441235000","authorization_failure":false,"status":"","user":"impalauser","impersonator":null,"statement_type":"USE","network_address":"::ffff:127.0.0.1:34009","sql_statement":"use impalaQueryDB_1377801039","catalog_objects":[{"name":"impalaQueryDB_1377801039","object_type":"DATABASE","privilege":"ANY"}]}}
{"1377775857516":{"query_id":"d04daaf9e488e953:5cb8bdc461eb9895","session_id":"2d4c159702aeebe2:cb5377ebedf6168b","start_time":"2013-08-29 11:30:57.488402000","authorization_failure":false,"status":"","user":"impalauser","impersonator":null,"statement_type":"DROP_TABLE","network_address":"::ffff:127.0.0.1:34009","sql_statement":"drop TABLE IF EXISTS testaudittable1","catalog_objects":[]}}"""

_NN_AUDIT_FILE_NAME = "HDFS-1-NAMENODE-8289bf63ac31fd75423fb00cf25e8e3b-NavigatorLog4jAppender-1385512800584"
_NN_AUDIT_LOG_TEXT = """{"allowed":true,"serviceName":"HDFS-1","username":"hbase","src":"/hbase/hbase.id","eventTime":1385512859079,"ipAddress":"10.20.187.50","operation":"getfileinfo","dest":null,"permissions":null,"impersonator":null,"delegationTokenId":null}
{"allowed":true,"serviceName":"HDFS-1","username":"hbase","src":"/hbase/hbase.version","eventTime":1385512859081,"ipAddress":"10.20.187.50","operation":"getfileinfo","dest":null,"permissions":null,"impersonator":null,"delegationTokenId":null}
{"allowed":true,"serviceName":"HDFS-1","username":"cloudera-scm","src":"/tmp/.cloudera_health_monitoring_canary_files","eventTime":1385512859245,"ipAddress":"10.20.187.50","operation":"getfileinfo","dest":null,"permissions":null,"impersonator":null,"delegationTokenId":null}
{"allowed":true,"serviceName":"HDFS-1","username":"cloudera-scm","src":"/tmp/.cloudera_health_monitoring_canary_files/.canary_file_2013_11_26-16_40_59","eventTime":1385512859291,"ipAddress":"10.20.187.50","operation":"create","dest":null,"permissions":"rw-rw-rw-","impersonator":null,"delegationTokenId":null}
{"allowed":true,"serviceName":"HDFS-1","username":"cloudera-scm","src":"/tmp/.cloudera_health_monitoring_canary_files/.canary_file_2013_11_26-16_40_59","eventTime":1385512859376,"ipAddress":"10.20.187.50","operation":"open","dest":null,"permissions":null,"impersonator":null,"delegationTokenId":null}"""

_CONFIG_DICT = {"verify_cert_file": None, "verify_cert_dir": None, "max_cert_depth": None, "client_key_file": None, "client_cert_file": None, "client_keypw_file": None, "client_keypw_cmd": None}

_IMPALA_NEW_AUDIT_FILE_NAME = "impala_audit_event_log_1.0-137779850786"

class AuditEvents(object):
  """
  Class to hold all the audit events received by the mock audit server.
  """

  def __init__(self, event):
    self._services = set()
    self._events_received = []
    self._event = event
    self._expected_request_count = 1

  def request_received(self, req):
    """
    Called by audit server when it receives an audit request.
    req is the python object for the audit request sent by agent,
    a 'WAERequest' obejct defined in navigator.avdl.
    """
    service_name = req['request']['serviceName']
    events_received = req['request']['events']
    self._services.add(service_name)
    self._events_received.extend(events_received)

    self._expected_request_count -= 1
    if self._expected_request_count == 0:
      self._event.set()

  def clear(self):
    """
    Reinitialize all member variables so that a test can run
    with clean state.
    """
    self._services = set()
    self._events_received = []
    self._expected_request_count = 1

  def set_expected_request_count(self, expected_request_count):
    self._expected_request_count = expected_request_count

  def is_service_present(self, service_name):
    return service_name in self._services

  def get_events_received(self):
    return self._events_received

class TestAuditServer(object):
  """
  This class runs an audit server and tests if whole module works together. This
  is more of integration test for the module than a unit test and helps us
  catch integration issues that individual unit tests fail to catch.
  """
  _port = None
  _host = None
  _event = Event()
  _audit_events = AuditEvents(_event)

  @classmethod
  def setup_class(cls):
    """
    Start audit server on localhost and a free port. Then create a audit config
    file with this information.
    """
    #start nms server
    cls._server = NavigatorServer(cls._audit_events, "nms")
    cls._server.start_server()
    cls._host = cls._server.get_host()
    cls._port = cls._server.get_port()
    #start tp server
    cmf.https.ssl_url_opener = cmf.https.SSLContextOpener(None, None)
    cls._tp_server = NavigatorServer(cls._audit_events, "tp")
    cls._tp_server.start_server()
    cls._tp_host = cls._tp_server.get_host()
    cls._tp_port = cls._tp_server.get_port()

  @classmethod
  def teardown_class(cls):
    """
    Stop audit server.
    """
    cls._server.stop_server()

  def setup(self):
    """
    Setup audit logs and audit client configuration.
    """
    shutil.rmtree(_AUDIT_LOG_DIR, True)
    shutil.rmtree(_ROLE_CONFIG_DIR, True)
    shutil.rmtree(_ROLE_CONFIG_DIR2, True)
    os.makedirs(_AUDIT_LOG_DIR)
    os.makedirs(_ROLE_CONFIG_DIR)
    os.makedirs(_ROLE_CONFIG_DIR2)
    self._audit_events.clear()
    self._stop_audit_thread()

  def teardown(self):
    """
    Cleanup audit logs and client configuration.
    """
    shutil.rmtree(_AUDIT_LOG_DIR)
    shutil.rmtree(_ROLE_CONFIG_DIR)
    shutil.rmtree(_ROLE_CONFIG_DIR2)
    self._stop_audit_thread()

  def _start_audit_thread(self):
    """
    Start audit thread.
    """
    cmf.audit.audit_thread = AuditThread(0, "TestAuditThread")
    cmf.audit.audit_thread._nav_thread.start()

  def _stop_audit_thread(self):
    """
    Stop audit thread if its running.
    """
    if cmf.audit.audit_thread._nav_thread.is_alive():
      cmf.audit.audit_thread.stop()

  def setup_audit_config(self, service_name, service_type, role_name, role_type,
                         role_conf_dir = _ROLE_CONFIG_DIR):
    """
    Creates an audit client configuration file for the specified service and
    rolename.
    """
    conf_file_path = os.path.join(role_conf_dir, AUDIT_CONFIG_FILENAME)
    conf_file = None
    try:
      conf_file = open(conf_file_path, 'w')
      print >> conf_file, "navigator.server.url=http://%s:%d" % (self._host, self._port)
      print >> conf_file, "telemetry_publisher.server.url=http://%s:%d" % (self._tp_host, self._tp_port)
      print >> conf_file, "navigator.client.serviceName=" + service_name
      print >> conf_file, "navigator.client.serviceType=" + service_type
      print >> conf_file, "navigator.client.roleName=" + role_name
      print >> conf_file, "navigator.client.roleType=" + role_type
      print >> conf_file, "navigator.batch.queue_policy=DROP"
      print >> conf_file, "audit_event_log_dir=" + _AUDIT_LOG_DIR
      print >> conf_file, "navigator.client.clusterName=Cluster 1"
      print >> conf_file, "navigator.batch.timeout=0"
      print >> conf_file, "navigator.client.max_num_audit_log=0"
    finally:
      if conf_file is not None:
        conf_file.close()

  def test_send_generic_audit(self):
    """
    Send generic audit events to a real audit server.
    """
    self.setup_audit_config("GENERIC-1", "GENERIC",
      _GENERIC_AUDIT_FILE_PREFIX, "GENERIC")
    _CONFIG_DICT["conf_dir"] = _ROLE_CONFIG_DIR
    audit_plugin = make_audit_plugin("generic-PROCESS", _CONFIG_DICT)
    self._start_audit_thread()
    audit_plugin.refresh(_ROLE_CONFIG_DIR)

    # Now create an audit log file and put few events in it
    # Then check if the events are received correctly
    audit_file_path = os.path.join(_AUDIT_LOG_DIR, _GENERIC_AUDIT_FILE_NAME)
    audit_file = open(audit_file_path, "w")
    print >> audit_file, _GENERIC_AUDIT_LOG_TEXT
    audit_file.close()

    # Wait for event to be sent and received by audit server
    self._event.wait(60)
    self._event.clear()
    self._stop_audit_thread()

    assert_true(self._audit_events.is_service_present("GENERIC-1"))
    assert_equal(4, len(self._audit_events.get_events_received()))

  def test_send_hdfs_audit(self):
    """
    Send hdfs audit events to a real audit server.
    """
    self.setup_audit_config("HDFS-1", "HDFS",
        "HDFS-1-NAMENODE-8289bf63ac31fd75423fb00cf25e8e3b", "NAMENODE")
    _CONFIG_DICT["conf_dir"] = _ROLE_CONFIG_DIR
    audit_plugin = make_audit_plugin("hdfs-NAMENODE", _CONFIG_DICT)
    self._start_audit_thread()
    audit_plugin.refresh(_ROLE_CONFIG_DIR)

    # Now create an audit log file and put few events in it
    # Then check if the events are received correctly
    audit_file_path = os.path.join(_AUDIT_LOG_DIR, _NN_AUDIT_FILE_NAME)
    audit_file = open(audit_file_path, "w")
    print >> audit_file, _NN_AUDIT_LOG_TEXT
    audit_file.close()

    # Wait for event to be sent and received by audit server
    self._event.wait(60)
    self._event.clear()
    self._stop_audit_thread()

    assert_true(self._audit_events.is_service_present("HDFS-1"))
    assert_equal(10, len(self._audit_events.get_events_received()))

  def test_send_impala_audit(self):
    """
    Send impala audit events to a real audit server.
    """
    self.setup_audit_config("IMPALA-1", "IMPALA",
        "IMPALA-1-IMPALAD-abee5a758942d6c689c7a5e7dd2da179", "IMPALAD")
    _CONFIG_DICT["conf_dir"] = _ROLE_CONFIG_DIR
    audit_plugin = make_audit_plugin("impala-IMPALAD", _CONFIG_DICT)
    self._start_audit_thread()
    audit_plugin.refresh(_ROLE_CONFIG_DIR)

    # Now create an audit log file and put few events in it
    # Then check if the events are received correctly
    audit_file_path = os.path.join(_AUDIT_LOG_DIR, _IMPALA_AUDIT_FILE_NAME)
    audit_file = open(audit_file_path, "w")
    print >> audit_file, _IMPALA_AUDIT_LOG_TEXT
    audit_file.close()

    # Wait for event to be sent and received by audit server
    self._event.wait(60)
    self._event.clear()
    self._stop_audit_thread()

    assert_true(self._audit_events.is_service_present("IMPALA-1"))
    assert_equal(6, len(self._audit_events.get_events_received()))

  @skiprhel5
  def test_send_impala_with_tp_on_and_off(self):
    """
    Send impala audit events to a real audit server with off/on telemetry publisher.
    """
    self.setup_audit_config("IMPALA-1", "IMPALA",
        "IMPALA-1-IMPALAD-abee5a758942d6c689c7a5e7dd2da179", "IMPALAD")
    _CONFIG_DICT["conf_dir"] = _ROLE_CONFIG_DIR
    audit_plugin = make_audit_plugin("impala-IMPALAD", _CONFIG_DICT)
    self._start_audit_thread()
    audit_plugin.refresh(_ROLE_CONFIG_DIR)

    #stop TP server, so events should only go to audit server
    self._tp_server.stop_server()

    # Now create an audit log file and put few events in it
    # Then check if the events are received correctly
    audit_file_path = os.path.join(_AUDIT_LOG_DIR, _IMPALA_AUDIT_FILE_NAME)
    audit_file = open(audit_file_path, "w")
    print >> audit_file, _IMPALA_AUDIT_LOG_TEXT
    audit_file.close()

    # Wait for event to be sent and received by audit server
    self._event.wait(60)
    self._event.clear()

    assert_true(self._audit_events.is_service_present("IMPALA-1"))
    assert_equal(3, len(self._audit_events.get_events_received()))
    self._audit_events.clear()

    #restart TP server, now it should get same event as audit server
    self._tp_server.start_server_on_host_port(self._tp_host, self._tp_port)
    # Wait for event to be sent and received by TP server
    self._event.wait(60)
    self._event.clear()
    self._stop_audit_thread()
    assert_true(self._audit_events.is_service_present("IMPALA-1"))
    assert_equal(3, len(self._audit_events.get_events_received()))

  @skiprhel5
  def test_send_impala_with_audit_on_and_off(self):
    """
    Send impala audit events to a off/on audit server and telemetry publisher.
    """
    self.setup_audit_config("IMPALA-1", "IMPALA",
        "IMPALA-1-IMPALAD-abee5a758942d6c689c7a5e7dd2da179", "IMPALAD")
    _CONFIG_DICT["conf_dir"] = _ROLE_CONFIG_DIR
    audit_plugin = make_audit_plugin("impala-IMPALAD", _CONFIG_DICT)
    self._start_audit_thread()
    audit_plugin.refresh(_ROLE_CONFIG_DIR)

    #stop audit server, so events should only go to audit server
    self._server.stop_server()

    # Now create an audit log file and put few events in it
    # Then check if the events are received correctly
    audit_file_path = os.path.join(_AUDIT_LOG_DIR, _IMPALA_AUDIT_FILE_NAME)
    audit_file = open(audit_file_path, "w")
    print >> audit_file, _IMPALA_AUDIT_LOG_TEXT
    audit_file.close()

    # Wait for event to be sent and received by audit server
    self._event.wait(60)
    self._event.clear()

    assert_true(self._audit_events.is_service_present("IMPALA-1"))
    assert_equal(3, len(self._audit_events.get_events_received()))
    self._audit_events.clear()

    #restart TP server, now it should get same event as audit server
    self._server.start_server_on_host_port(self._host, self._port)
    # Wait for event to be sent and received by TP server
    self._event.wait(60)
    self._event.clear()
    self._stop_audit_thread()
    assert_true(self._audit_events.is_service_present("IMPALA-1"))
    assert_equal(3, len(self._audit_events.get_events_received()))

  @skiprhel5
  def test_send_impala_many_files(self):
    """
    Send many impala audit events to a audit server and telemetry publisher.
    """
    self.setup_audit_config("IMPALA-1", "IMPALA",
        "IMPALA-1-IMPALAD-abee5a758942d6c689c7a5e7dd2da179", "IMPALAD")
    _CONFIG_DICT["conf_dir"] = _ROLE_CONFIG_DIR
    audit_plugin = make_audit_plugin("impala-IMPALAD", _CONFIG_DICT)
    self._start_audit_thread()
    audit_plugin.refresh(_ROLE_CONFIG_DIR)

    for i in range(0, 5):
      # Now create an audit log file and put few events in it
      # Then check if the events are received correctly
      audit_file_path = os.path.join(_AUDIT_LOG_DIR, _IMPALA_AUDIT_FILE_NAME+str(i))
      audit_file = open(audit_file_path, "w")
      print >> audit_file, _IMPALA_AUDIT_LOG_TEXT
      audit_file.close()

    # Wait for event to be sent and received by audit server
    self._event.wait(60)
    self._event.clear()
    self._stop_audit_thread()

    assert_true(self._audit_events.is_service_present("IMPALA-1"))
    assert_equal(30, len(self._audit_events.get_events_received()))
    self._audit_events.clear()

  @skiprhel5
  def test_enable_disable_audit(self):
    """
    Test if we can enable and disable auditing for a component successfully.
    OPSAPS-19671 is the motivation for this test. This test will enable
    auditing for multiple components and then toggles auditing for one
    component few times and checks everything continues working.
    """

    ### PART 1: Setup HDFS and Impala audits and confirm that it works
    self._audit_events.set_expected_request_count(2)
    self.setup_audit_config("HDFS-1", "HDFS",
        "HDFS-1-NAMENODE-8289bf63ac31fd75423fb00cf25e8e3b", "NAMENODE")
    self.setup_audit_config("IMPALA-1", "IMPALA",
        "IMPALA-1-IMPALAD-abee5a758942d6c689c7a5e7dd2da179", "IMPALAD",
        role_conf_dir = _ROLE_CONFIG_DIR2)

    _CONFIG_DICT["conf_dir"] = _ROLE_CONFIG_DIR
    audit_plugin1 = make_audit_plugin("hdfs-NAMENODE", _CONFIG_DICT)
    _CONFIG_DICT["conf_dir"] = _ROLE_CONFIG_DIR2
    audit_plugin2 = make_audit_plugin("impala-IMPALAD", _CONFIG_DICT)

    self._start_audit_thread()
    audit_plugin1.refresh(_ROLE_CONFIG_DIR)
    audit_plugin2.refresh(_ROLE_CONFIG_DIR2)

    # Now create impala and hdfs audit log files and put few events
    # in them. Then check if the events are received correctly
    audit_file_path = os.path.join(_AUDIT_LOG_DIR, _NN_AUDIT_FILE_NAME)
    audit_file = open(audit_file_path, "w")
    print >> audit_file, _NN_AUDIT_LOG_TEXT
    audit_file.close()

    audit_file_path = os.path.join(_AUDIT_LOG_DIR, _IMPALA_AUDIT_FILE_NAME)
    audit_file = open(audit_file_path, "w")
    print >> audit_file, _IMPALA_AUDIT_LOG_TEXT
    audit_file.close()

    # Wait for event to be sent and received by audit server
    self._event.wait(60)
    self._event.clear()

    assert_true(self._audit_events.is_service_present("HDFS-1"))
    assert_true(self._audit_events.is_service_present("IMPALA-1"))
    assert_equal(8, len(self._audit_events.get_events_received()))

    self._audit_events.clear()

    ### PART 2: disable hdfs audit and check if impala audits can still be received
    conf_file_path = os.path.join(_ROLE_CONFIG_DIR, AUDIT_CONFIG_FILENAME)
    conf_file = None
    try:
      conf_file = open(conf_file_path, 'w')
      conf_file.truncate()
    finally:
      if conf_file is not None:
        conf_file.close()
    audit_plugin1.refresh(_ROLE_CONFIG_DIR)

    audit_file_path = os.path.join(_AUDIT_LOG_DIR, _IMPALA_AUDIT_FILE_NAME)
    audit_file = open(audit_file_path, "a")
    print >> audit_file, _IMPALA_AUDIT_LOG_TEXT
    audit_file.close()

    # Wait for event to be sent and received by audit server
    self._event.wait(60)
    self._event.clear()

    assert_true(self._audit_events.is_service_present("IMPALA-1"))
    assert_equal(3, len(self._audit_events.get_events_received()))

    self._audit_events.clear()
    self._audit_events.set_expected_request_count(2)

    ### PART 3: enable hdfs audit and check if everything works
    self.setup_audit_config("HDFS-1", "HDFS",
        "HDFS-1-NAMENODE-8289bf63ac31fd75423fb00cf25e8e3b", "NAMENODE")
    audit_plugin1.refresh(_ROLE_CONFIG_DIR)

    audit_file_path = os.path.join(_AUDIT_LOG_DIR, _NN_AUDIT_FILE_NAME)
    audit_file = open(audit_file_path, "a")
    print >> audit_file, _NN_AUDIT_LOG_TEXT
    audit_file.flush()
    os.fsync(audit_file.fileno())
    audit_file.close()

    audit_file_path = os.path.join(_AUDIT_LOG_DIR, _IMPALA_AUDIT_FILE_NAME)
    audit_file = open(audit_file_path, "a")
    print >> audit_file, _IMPALA_AUDIT_LOG_TEXT
    audit_file.flush()
    os.fsync(audit_file.fileno())
    audit_file.close()


    self._event.wait(60)
    self._event.clear()

    wait_and_assert_true(lambda :self._audit_events.is_service_present("HDFS-1"), _MAX_WAIT_TIMEOUT_5_MINS)
    wait_and_assert_true(lambda :self._audit_events.is_service_present("IMPALA-1"), _MAX_WAIT_TIMEOUT_5_MINS)
    assert_equal(8, len(self._audit_events.get_events_received()))

    self._audit_events.clear()
    self._stop_audit_thread()
