# Copyright (c) 2013 Cloudera, Inc. All rights reserved.
import os
import pyinotify
import shutil

from cmf.audit.inotify_event_processor import INotifyEventProcessor
from nose.tools import assert_equal
from pyinotify import WatchManager, ThreadedNotifier
from threading import Event

_AUDIT_LOG_DIR = "/tmp/impala-audit"
_AUDIT_LOG_DIR2 = "/tmp/hdfs-audit"
_AUDIT_FILE_NAME = "impala_audit_event_log_1.0-1377798507897"
_AUDIT_FILE_NAME2 = "HDFS-1-NAMENODE-656d088e0add1f78dc95737f7bd75778-1386200211203"

class MockAuditLogDirectory(object):
  """
  Mock class to handle audit log events. Used in asserting that audit log
  events were generated.
  """
  def __init__(self, dir):
    self._new_file_name, self._deleted_file_name = None, None
    self._event = Event()
    self._dir = dir

  def get_dir(self):
    return self._dir

  def append_new_log_to_list(self, file_name):
    """
    Called when a new log appears in the audit log directory.
    """
    self._new_file_name = file_name
    self._event.set()

  def get_new_file_name(self):
    return self._new_file_name

  def remove_log_from_list(self, file_name):
    """
    Called when a audit log file is removed from audit log directory
    """
    self._deleted_file_name = file_name
    self._event.set()

  def get_deleted_file_name(self):
    return self._deleted_file_name

  def wait_for_inotify_event(self):
    self._event.wait(60) # Wait for at most a minute
    self._event.clear()

class MockAuditPlugin(object):
  """
  Mock class to handle audit plugin methods.
  """
  def __init__(self, audit_log_dir, is_csd=False):
    self._audit_log_dir = audit_log_dir
    self._audit_pipelines = [MockAuditPipeline(audit_log_dir)]
    self.is_csd = is_csd

  def is_csd_process(self):
    return self.is_csd

  def get_nav_pipelines(self):
    return self._audit_pipelines

  def get_nav_conf(self):
    class MockAuditConf(object):
      def __init__(self, audit_log_dir):
        self._audit_log_dir = audit_log_dir
      @property
      def log_dir(self):
        return self._audit_log_dir

    return MockAuditConf(self._audit_log_dir.get_dir())

  nav_conf = property(get_nav_conf, None)

class MockAuditPipeline(object):

  def __init__(self, audit_log_dir):
    self._audit_log_dir = audit_log_dir

  def get_nav_logs_dir(self):
    return self._audit_log_dir

class TestINotifyEventProcessor(object):
  """
  Class to test if the audit log inotify events are handled correctly. As of
  now we register for three events: CREATE, DELETE and MOVED_FROM
  This class if these events are generated and handled correctly.
  """
  def setup(self):
    shutil.rmtree(_AUDIT_LOG_DIR, True)
    shutil.rmtree(_AUDIT_LOG_DIR2, True)
    os.makedirs(_AUDIT_LOG_DIR)
    os.makedirs(_AUDIT_LOG_DIR2)
    self._audit_logs = MockAuditLogDirectory(_AUDIT_LOG_DIR)
    self._audit_logs2 = MockAuditLogDirectory(_AUDIT_LOG_DIR2)
    audit_plugins = [MockAuditPlugin(self._audit_logs),
                     MockAuditPlugin(self._audit_logs2)]
    self._inotify_event_processor = INotifyEventProcessor(nav_plugins=audit_plugins)
    wm = WatchManager()
    self._notifier = ThreadedNotifier(wm, self._inotify_event_processor)
    self._notifier.start()
    wm.add_watch([_AUDIT_LOG_DIR, _AUDIT_LOG_DIR2],
                 pyinotify.IN_CREATE | pyinotify.IN_DELETE | pyinotify.IN_MOVED_FROM)

  def teardown(self):
    shutil.rmtree(_AUDIT_LOG_DIR, True)
    shutil.rmtree(_AUDIT_LOG_DIR2, True)
    self._notifier.stop()

  def _create_audit_log_files(self):
    """
    Create log files in all the audit log directories we are monitoring.
    """
    audit_file_paths = []
    audit_file_paths.append(
      self._create_audit_log_file(_AUDIT_LOG_DIR, _AUDIT_FILE_NAME))
    audit_file_paths.append(
      self._create_audit_log_file(_AUDIT_LOG_DIR2, _AUDIT_FILE_NAME2))

    return audit_file_paths

  def _create_audit_log_file(self, log_dir, log_name):
    """
    Creates a new audit log file and returns the path to the file.
    """
    audit_file_path = os.path.join(log_dir, log_name)
    audit_file = open(audit_file_path, "w")
    print >> audit_file, "Test Message"
    audit_file.close()

    return audit_file_path

  def test_create_audit_log_file(self):
    """
    Test if a new audit log is created we get the notification.
    """
    self._create_audit_log_file(_AUDIT_LOG_DIR, _AUDIT_FILE_NAME)
    self._audit_logs.wait_for_inotify_event()
    assert_equal(_AUDIT_FILE_NAME, self._audit_logs.get_new_file_name())

  def test_create_multiple_audit_log_files(self):
    """
    Test if multiple new audit log is created we get the notification.
    """
    self._create_audit_log_files()
    self._audit_logs.wait_for_inotify_event()
    self._audit_logs2.wait_for_inotify_event()
    assert_equal(_AUDIT_FILE_NAME, self._audit_logs.get_new_file_name())
    assert_equal(_AUDIT_FILE_NAME2, self._audit_logs2.get_new_file_name())

  def test_delete_audit_log_file(self):
    """
    Test if an audit log is deleted we get the notification.
    """
    audit_file_path = self._create_audit_log_file(_AUDIT_LOG_DIR, _AUDIT_FILE_NAME)
    self._audit_logs.wait_for_inotify_event()

    # Now delete the audit log file and check if we get correct event
    os.remove(audit_file_path)
    self._audit_logs.wait_for_inotify_event()
    assert_equal(_AUDIT_FILE_NAME, self._audit_logs.get_deleted_file_name())

  def test_delete_multiple_audit_log_files(self):
    """
    Test if multiple audit log files are deleted we get the notification.
    """
    audit_file_paths = self._create_audit_log_files()
    self._audit_logs.wait_for_inotify_event()
    self._audit_logs2.wait_for_inotify_event()

    # Now delete the audit log file and check if we get correct event
    os.remove(audit_file_paths[0])
    os.remove(audit_file_paths[1])
    self._audit_logs.wait_for_inotify_event()
    self._audit_logs2.wait_for_inotify_event()
    assert_equal(_AUDIT_FILE_NAME, self._audit_logs.get_deleted_file_name())
    assert_equal(_AUDIT_FILE_NAME2, self._audit_logs2.get_deleted_file_name())

  def test_move_audit_log_file(self):
    """
    Test if an audit log is moved we get the notification.
    """
    # Create audit log file
    audit_file_path = self._create_audit_log_file(_AUDIT_LOG_DIR, _AUDIT_FILE_NAME)
    self._audit_logs.wait_for_inotify_event()

    # Now move the audit log file and check if we get correct event
    new_audit_file_path = os.path.join("/tmp", _AUDIT_FILE_NAME)
    shutil.move(audit_file_path, new_audit_file_path)
    self._audit_logs.wait_for_inotify_event()
    assert_equal(_AUDIT_FILE_NAME, self._audit_logs.get_deleted_file_name())
    os.remove(new_audit_file_path)

  def test_move_multiple_audit_log_files(self):
    """
    Test if multiple audit logs are moved we get the notification.
    """
    # Create audit log file
    audit_file_paths = self._create_audit_log_files()
    self._audit_logs.wait_for_inotify_event()
    self._audit_logs2.wait_for_inotify_event()

    # Now move the audit log file and check if we get correct event
    new_audit_file_path1 = os.path.join("/tmp", _AUDIT_FILE_NAME)
    shutil.move(audit_file_paths[0], new_audit_file_path1)
    new_audit_file_path2 = os.path.join("/tmp", _AUDIT_FILE_NAME2)
    shutil.move(audit_file_paths[1], new_audit_file_path2)

    self._audit_logs.wait_for_inotify_event()
    self._audit_logs2.wait_for_inotify_event()
    assert_equal(_AUDIT_FILE_NAME, self._audit_logs.get_deleted_file_name())
    assert_equal(_AUDIT_FILE_NAME2, self._audit_logs2.get_deleted_file_name())
    os.remove(new_audit_file_path1)
    os.remove(new_audit_file_path2)
