# Copyright (c) 2016 Cloudera, Inc. All rights reserved.

import cmf.audit
import simplejson
import os
import shutil
import time

from cmf.audit import METADATA_CONFIG_FILENAME
from cmf.audit.navigator_thread import MetadataThread
from cmf_test.audit.test_util import make_spark_metadata_plugin
from cmf_test.audit.test_util import wait_and_assert_true, wait_and_assert_false
from cmf_test.audit.navigator_server import NavigatorServer, MetadataRequestHandler
from nose.tools import assert_true, assert_equal, assert_false
from threading import Event

_LINEAGE_LOG_DIR = "/tmp/spark_lineage_log"
_ROLE_CONFIG_DIR = "/tmp/nav_client_config_dir"
_MAX_WAIT_TIMEOUT_5_MINS = 5*60

_SPARK_LINEAGE_FILE_NAME = "spark_lineage_log_SampleSparkApp-Exe-1-1478759200510"
_SPARK_LINEAGE_FILE_NAME_2 = "spark_lineage_log_AnotherSparkApp-Exe-1-1478759200510"
_SPARK_LINEAGE_WAL_FILE_NAME = "SampleSparkApp-Exe-1_metadata_wal"
_SPARK_LINEAGE_WAL_FILE_NAME_2 = "AnotherSparkApp-Exe-1_metadata_wal"
_SPARK_LINEAGE_LOG_TEXT = """{"version":"1.0","applicationID":"SampleSparkApp","yarnApplicationId":"SampleSparkApp-Exe-1", "timestamp":"1476233015", "duration":50, "user":"admin", "linkId":123456, "dataQuery":"select j.job_id, s.description, s.salary from sample_07 s join job_data j on (s.code = j.code)", "inputs": [{"dataSourceType":"HIVE", "dataSourceFormat":"", "hiveMetastoreLocation":"thrift://salilsurendran4-1.vpc.cloudera.com:9083","source":"default.job_data", "fields":["job_id"] }, {"dataSourceType":"HIVE", "dataSourceFormat":"", "hiveMetastoreLocation":"thrift://salilsurendran4-1.vpc.cloudera.com:9083","source":"default.sample_07", "fields":["description","salary"]}]}
{"version":"1.0", "applicationID":"SampleSparkApp", "yarnApplicationId":"SampleSparkApp-Exe-1", "timestamp":"1476233080", "duration":50, "user":"admin", "linkId":123456, "dataQuery":"create table new_sample_07 as select job_id, salary, description", "outputs": [{"dataSourceType":"HIVE", "hiveMetastoreLocation":"thrift://salilsurendran4-1.vpc.cloudera.com:9083","destination":"default.new_sample_07", "fields":["job_id","salary","description"]}]}"""
_SPARK_LINEAGE_LOG_TEXT_2 = """{"version":"1.0","applicationID":"AnotherSparkApp","yarnApplicationId":"AnotherSparkApp-Exe-1", "timestamp":"1476233015", "duration":50, "user":"admin", "linkId":123456, "dataQuery":"select j.job_id, s.description, s.salary from sample_07 s join job_data j on (s.code = j.code)", "inputs": [{"dataSourceType":"HIVE", "dataSourceFormat":"", "hiveMetastoreLocation":"thrift://salilsurendran4-1.vpc.cloudera.com:9083","source":"default.job_data", "fields":["job_id"] }, {"dataSourceType":"HIVE", "dataSourceFormat":"", "hiveMetastoreLocation":"thrift://salilsurendran4-1.vpc.cloudera.com:9083","source":"default.sample_07", "fields":["description","salary"]}]}
{"version":"1.0", "applicationID":"AnotherSparkApp", "yarnApplicationId":"AnotherSparkApp-Exe-1", "timestamp":"1476233080", "duration":50, "user":"admin", "linkId":123456, "dataQuery":"create table new_sample_07 as select job_id, salary, description", "outputs": [{"dataSourceType":"HIVE", "hiveMetastoreLocation":"thrift://salilsurendran4-1.vpc.cloudera.com:9083","destination":"default.new_sample_07", "fields":["job_id","salary","description"]}]}"""
_SPARK_LINEAGE_LOG_END_MARKER_TEXT = """{"version":"1.0", "applicationID":"SampleSparkApp", "yarnApplicationId":"SampleSparkApp-Exe-1", "timestamp":"1476233080", "ended":true }"""
_SPARK_LINEAGE_LOG_END_MARKER_TEXT_2 = """{"version":"1.0", "applicationID":"AnotherSparkApp", "yarnApplicationId":"AnotherSparkApp-Exe-1", "timestamp":"1476233080", "ended":true }"""


class MetadataEntries(object):
  """
  Class to hold all the lineage entries received by the mock navigator server.
  """

  def __init__(self, event):
    self._services = set()
    self._events_received = []
    self._event = event
    self._expected_request_count = 1

  def request_received(self, path, content):
    """
    Called by navigator server when it receives a metadata request.
    req is the python object for the request sent by agent.
    """
    path_parts = path.split('/')
    service_name = path_parts[-1]
    events_received = simplejson.loads(content)

    self._services.add(service_name)
    self._events_received.extend(events_received)

    self._expected_request_count -= 1
    if self._expected_request_count == 0:
      self._event.set()

  def clear(self):
    """
    Reinitialize all member variables so that a test can run
    with clean state.
    """
    self._services = set()
    self._events_received = []
    self._expected_request_count = 1

  def set_expected_request_count(self, expected_request_count):
    self._expected_request_count = expected_request_count

  def is_service_present(self, service_name):
    return service_name in self._services

  def get_events_received(self):
    return self._events_received


class TestMetadataServer(object):
  """
  This class runs a navigator server and tests if whole module works together.
  This is more of integration test for the module than a unit test and helps us
  catch integration issues that individual unit tests fail to catch.
  """
  _port = None
  _host = None
  _event = Event()
  _metadata_entries = MetadataEntries(_event)

  @classmethod
  def setup_class(cls):
    """
    Start navigator server on localhost and a free port. Then save this
    information so we can later create a lineage config file with this
    information.
    """
    cmf.https.ssl_url_opener = cmf.https.SSLContextOpener(None, None)
    cls._server = NavigatorServer(cls._metadata_entries, None, MetadataRequestHandler)
    cls._server.start_server()
    cls._host = cls._server.get_host()
    cls._port = cls._server.get_port()

  @classmethod
  def teardown_class(cls):
    """
    Stop navigator server.
    """
    cls._server.stop_server()

  def setup(self):
    """
    Setup lineage logs and lineage client configuration.
    """
    shutil.rmtree(_LINEAGE_LOG_DIR, True)
    shutil.rmtree(_ROLE_CONFIG_DIR, True)
    os.makedirs(_LINEAGE_LOG_DIR)
    os.makedirs(_ROLE_CONFIG_DIR)
    self._metadata_entries.clear()
    self._stop_metadata_thread()

  def teardown(self):
    """
    Cleanup lineage logs and client configuration.
    """
    self._stop_metadata_thread()

  def _start_metadata_thread(self):
    """
    Start audit thread.
    """
    cmf.audit.metadata_thread = MetadataThread(0, "TestMetadataThread")
    cmf.audit.metadata_thread._nav_thread.start()

  def _stop_metadata_thread(self):
    """
    Stop audit thread if its running.
    """
    if cmf.audit.metadata_thread._nav_thread.is_alive():
      cmf.audit.metadata_thread.stop()

  def _wait_for_watch_to_be_set(self, nav_thread, path_to_be_watched):
    wait_and_assert_true(lambda: nav_thread.get_watch_manager() is not None, 360)
    wm = nav_thread.get_watch_manager()
    for i in wm.watches:
      watch = wm.watches[i]
      if watch.path == path_to_be_watched:
        return True

  def _create_lineage_client_config(self, service_name, service_type, role_name,
                                    role_type):
    """
    Write a navigator.lineage.client.properties file.
    """
    conf_file_path = os.path.join(_ROLE_CONFIG_DIR, METADATA_CONFIG_FILENAME)
    conf_file = None
    try:
      conf_file = open(conf_file_path, 'w')
      print >> conf_file, "navigator.server.url=http://%s:%d" % (
        self._host, self._port)
      print >> conf_file, "navigator.client.serviceName=" + service_name
      print >> conf_file, "navigator.client.serviceType=" + service_type
      print >> conf_file, "navigator.client.roleName=" + role_name
      print >> conf_file, "navigator.client.roleType=" + role_type
      print >> conf_file, "navigator.client.clusterName=Cluster 1"
      print >> conf_file, "navigator.server.timeout_seconds=120"
      print >> conf_file, "lineage_event_log_dir=" + _LINEAGE_LOG_DIR
    finally:
      if conf_file is not None:
        conf_file.close()

    return conf_file_path

  def _write_to_lineage_file(self, log_dir, file_name, text):
    log_file_path = os.path.join(log_dir, file_name)
    log_file = open(log_file_path, "a")
    print >> log_file, text
    log_file.close()

  def _add_end_marker(self, log_dir, file_name, text):
    print

  def _set_up_spark_plugin(self):
    self._create_lineage_client_config("SPARK-1", "SPARK",
                                       "SPARK-ROLENAME-abee5a758942d6c689c7a5e7dd2da179", "SPARKD")
    metadata_plugin = make_spark_metadata_plugin()
    self._start_metadata_thread()
    metadata_plugin.refresh(_ROLE_CONFIG_DIR)

    # Wait for watch to be set
    self._wait_for_watch_to_be_set(cmf.audit.metadata_thread, _LINEAGE_LOG_DIR)

    return metadata_plugin

  def _create_a_single_pipeline(self):
    self._metadata_entries.set_expected_request_count(1)
    # Set-up spark plugin and directory watch
    metadata_plugin = self._set_up_spark_plugin()

    # Wait for watch to be set
    self._wait_for_watch_to_be_set(cmf.audit.metadata_thread, _LINEAGE_LOG_DIR)

    # Make sure no pipelines are created at this point
    assert_true(len(metadata_plugin.get_nav_pipelines()) == 0)

    # create a file in the log directory and see if a new pipeline gets created
    self._write_to_lineage_file(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_FILE_NAME, _SPARK_LINEAGE_LOG_TEXT)

    # inotify event will get generated, wait long enough for inotify event to get processed and sent to server
    self._event.wait(30)
    self._event.clear()

    # A new pipeline should be created assert the pipeline is indeed created
    assert_true(len(metadata_plugin.get_nav_pipelines()) == 1)

    # Verify corresponding wal file also gets created
    wait_and_assert_true(lambda: os.path.exists(os.path.join(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_WAL_FILE_NAME)))

    return metadata_plugin

  def test_metadata_csd_plugin_lineage(self):

    metadata_plugin = self._create_a_single_pipeline();

    # assert that exactly 2 events are received
    assert_equal(len(self._metadata_entries.get_events_received()), 2)

    # assert events are correctly received in order, we check that with timestamps
    assert_equal(self._metadata_entries.get_events_received()[0]["timestamp"], str(1476233015))
    assert_equal(self._metadata_entries.get_events_received()[1]["timestamp"], str(1476233080))

    # verify the log file + wal file isn't deleted
    assert_true(os.path.exists(os.path.join(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_FILE_NAME)))
    assert_true(os.path.exists(os.path.join(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_WAL_FILE_NAME)))

    self.teardown()

  def test_metadata_csd_plugin_lineage_multiple_apps(self):

    self._metadata_entries.set_expected_request_count(2)
    # Set-up spark plugin and directory watch
    metadata_plugin = self._set_up_spark_plugin()

    # Wait for watch to be set
    self._wait_for_watch_to_be_set(cmf.audit.metadata_thread, _LINEAGE_LOG_DIR)

    # Make sure no pipelines are created at this point
    assert_true(len(metadata_plugin.get_nav_pipelines()) == 0)

    # create file2 in the log directory
    self._write_to_lineage_file(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_FILE_NAME, _SPARK_LINEAGE_LOG_TEXT)
    self._write_to_lineage_file(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_FILE_NAME_2, _SPARK_LINEAGE_LOG_TEXT)

    # inotify event will get generated, wait long enough for inotify event to get processed and sent to server
    self._event.wait(30)
    self._event.clear()

    assert_true(len(metadata_plugin.get_nav_pipelines()) == 2)
    assert_equal(len(self._metadata_entries.get_events_received()), 4)
    assert_equal(self._metadata_entries.get_events_received()[0]["timestamp"], str(1476233015))
    assert_equal(self._metadata_entries.get_events_received()[1]["timestamp"], str(1476233080))

    # verify the log file + wal file isn't deleted
    assert_true(os.path.exists(os.path.join(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_FILE_NAME)))
    wait_and_assert_true(lambda: os.path.exists(os.path.join(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_WAL_FILE_NAME)), 60)
    assert_true(os.path.exists(os.path.join(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_FILE_NAME_2)))
    wait_and_assert_true(lambda: os.path.exists(os.path.join(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_WAL_FILE_NAME_2)), 60)

    # write the end markers
    self._metadata_entries.set_expected_request_count(2)
    self._write_to_lineage_file(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_FILE_NAME, _SPARK_LINEAGE_LOG_END_MARKER_TEXT)
    self._write_to_lineage_file(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_FILE_NAME_2, _SPARK_LINEAGE_LOG_END_MARKER_TEXT_2)

    # wait for end events to be received
    self._event.wait(30)
    self._event.clear()

    # Make sure pipelines are removed.
    wait_and_assert_true(lambda :len(metadata_plugin.get_nav_pipelines()) == 0)

    # assert that exactly 6 events are received
    assert_equal(len(self._metadata_entries.get_events_received()), 6)
    assert_equal(self._metadata_entries.get_events_received()[4]["ended"], True)
    assert_equal(self._metadata_entries.get_events_received()[5]["ended"], True)

    # verify all the wal and log files are deleted
    wait_and_assert_false(lambda: os.path.exists(os.path.join(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_FILE_NAME)), _MAX_WAIT_TIMEOUT_5_MINS)
    wait_and_assert_false(lambda: os.path.exists(os.path.join(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_WAL_FILE_NAME)), _MAX_WAIT_TIMEOUT_5_MINS)
    wait_and_assert_false(lambda: os.path.exists(os.path.join(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_FILE_NAME_2)), _MAX_WAIT_TIMEOUT_5_MINS)
    wait_and_assert_false(lambda: os.path.exists(os.path.join(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_WAL_FILE_NAME_2)), _MAX_WAIT_TIMEOUT_5_MINS)

    # Verify that the inotify object also removes the nav_log_dirs
    event_processor = cmf.audit.metadata_thread.get_event_processor()
    wait_and_assert_true(lambda: len(event_processor._get_nav_log_dirs(_LINEAGE_LOG_DIR)) == 0)

    self.teardown()

  def test_metadata_csd_plugin_pipeline_deletion(self):

    self._metadata_entries.set_expected_request_count(2)
    # Set-up spark plugin and directory watch
    metadata_plugin = self._set_up_spark_plugin()

    # Wait for watch to be set
    self._wait_for_watch_to_be_set(cmf.audit.metadata_thread, _LINEAGE_LOG_DIR)

    # Make sure no pipelines are created at this point
    assert_true(len(metadata_plugin.get_nav_pipelines()) == 0)

    # create file2 in the log directory
    self._write_to_lineage_file(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_FILE_NAME, _SPARK_LINEAGE_LOG_TEXT)
    self._write_to_lineage_file(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_FILE_NAME_2, _SPARK_LINEAGE_LOG_TEXT)

    # inotify event will get generated, wait long enough for inotify event to get processed and sent to server
    self._event.wait(30)
    self._event.clear()
    assert_true(len(metadata_plugin.get_nav_pipelines()) == 2)
    assert_equal(len(self._metadata_entries.get_events_received()), 4)
    assert_equal(self._metadata_entries.get_events_received()[0]["timestamp"], str(1476233015))
    assert_equal(self._metadata_entries.get_events_received()[1]["timestamp"], str(1476233080))

    # verify the log file + wal file isn't deleted
    assert_true(os.path.exists(os.path.join(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_FILE_NAME)))
    wait_and_assert_true(lambda: os.path.exists(os.path.join(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_WAL_FILE_NAME)), 60)
    assert_true(os.path.exists(os.path.join(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_FILE_NAME_2)))
    wait_and_assert_true(lambda: os.path.exists(os.path.join(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_WAL_FILE_NAME_2)), 60)

    # Verify that the inotify object has two nav_log_dirs objects
    event_processor = cmf.audit.metadata_thread.get_event_processor()
    wait_and_assert_true(lambda: len(event_processor._get_nav_log_dirs(_LINEAGE_LOG_DIR)) == 2)

    # write the end marker for the first pipeline
    self._metadata_entries.set_expected_request_count(1)
    self._write_to_lineage_file(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_FILE_NAME, _SPARK_LINEAGE_LOG_END_MARKER_TEXT)

    # wait for end events to be received
    self._event.wait(30)
    self._event.clear()

    # Make sure pipelines are removed.
    wait_and_assert_true(lambda: len(metadata_plugin.get_nav_pipelines()) == 1)
    for pipeline in metadata_plugin.get_nav_pipelines():
      assert_equal(pipeline.get_name(), "AnotherSparkApp-Exe-1")

    # Verify that the inotify object also removes the correct nav_log_dirs object
    event_processor = cmf.audit.metadata_thread.get_event_processor()
    wait_and_assert_true(lambda: len(event_processor._get_nav_log_dirs(_LINEAGE_LOG_DIR)) == 1)
    nav_log_dirs = event_processor._get_nav_log_dirs(_LINEAGE_LOG_DIR)[0]
    assert_equal(nav_log_dirs.get_pipeline_name(), "AnotherSparkApp-Exe-1")

    # verify the wal and log files are deleted for first pipeline and not for second
    assert_false(os.path.exists(os.path.join(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_FILE_NAME)))
    assert_false(os.path.exists(os.path.join(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_WAL_FILE_NAME)))
    assert_true(os.path.exists(os.path.join(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_FILE_NAME_2)))
    assert_true(os.path.exists(os.path.join(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_WAL_FILE_NAME_2)))

    self.teardown()

  def test_metadata_csd_plugin_delete_log_file(self):

    metadata_plugin = self._create_a_single_pipeline();

    # Now delete the log file
    os.remove(os.path.join(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_FILE_NAME))

    # Assert the log file has been deleted but the wal file hasn't
    wait_and_assert_true(lambda : not os.path.exists(os.path.join(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_FILE_NAME)))
    wait_and_assert_true(lambda: os.path.exists(os.path.join(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_WAL_FILE_NAME)))

    # Now verify the pipeline hasn't been deleted
    wait_and_assert_true(lambda: len(metadata_plugin.get_nav_pipelines()) == 1)

    self.teardown()

  def test_metadata_csd_plugin_delete_wal_file(self):

    self._create_a_single_pipeline();

    # Now delete the wal file
    os.remove(os.path.join(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_WAL_FILE_NAME))

    # Verify that the inotify object also removes the nav_log_dirs
    event_processor = cmf.audit.metadata_thread.get_event_processor()
    wait_and_assert_true(lambda: len(event_processor._get_nav_log_dirs(_LINEAGE_LOG_DIR)) == 0)

    # Assert the log file has been deleted but the wal file hasn't
    wait_and_assert_true(lambda: os.path.exists(os.path.join(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_FILE_NAME)))
    wait_and_assert_true(lambda: not os.path.exists(os.path.join(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_WAL_FILE_NAME)))

    self.teardown()
