# Copyright (c) 2016 Cloudera, Inc. All rights reserved.

import cmf.audit
import simplejson
import os
import shutil
import time

from cmf.audit import METADATA_CONFIG_FILENAME
from cmf.audit.navigator_thread import MetadataThread
from cmf_test.audit.test_util import wait_and_assert_true
from cmf_test.audit.test_util import make_spark_metadata_plugin
from cmf_test.audit.navigator_server import NavigatorServer, MetadataRequestHandler
from nose.tools import assert_true, assert_equal
from threading import Event

_LINEAGE_LOG_DIR = "/tmp/spark_lineage_log"
_ROLE_CONFIG_DIR = "/tmp/nav_client_config_dir"
_SPARK_LINEAGE_FILE_NAME = "spark_lineage_log_SampleSparkApp-Exe-1-1478759200510"
_SPARK_LINEAGE_LOG_TEXT = """{"version":"1.0", \
"applicationID":"Sample_Spark_App",\
"yarnApplicationId":"SampleSparkApp-Exe-1", "timestamp":"1476233015", "duration":50, "user":"admin", "linkId":123456, "dataQuery":"select j.job_id, s.description, s.salary from sample_07 s join job_data j on (s.code = j.code)", "inputs": [{"dataSourceType":"HIVE", "dataSourceFormat":"", "hiveMetastoreLocation":"thrift://salilsurendran4-1.vpc.cloudera.com:9083","source":"default.job_data", "fields":["job_id"] }, {"dataSourceType":"HIVE", "dataSourceFormat":"", "hiveMetastoreLocation":"thrift://salilsurendran4-1.vpc.cloudera.com:9083","source":"default.sample_07", "fields":["description","salary"] } ] }
{"version":"1.0", "applicationID":"Sample_Spark_App", "yarnApplicationId":"SampleSparkApp-Exe-1", "timestamp":"1476233080", "duration":50, "user":"admin", "linkId":123456, "dataQuery":"create table new_sample_07 as select job_id, salary, description", "outputs": [{"dataSourceType":"HIVE", "hiveMetastoreLocation":"thrift://salilsurendran4-1.vpc.cloudera.com:9083","destination":"default.new_sample_07", "fields":["job_id","salary","description"] } ] }"""

class MetadataEntries(object):
  """
  Class to hold all the lineage entries received by the mock navigator server.
  """

  def __init__(self, event):
    self._services = set()
    self._events_received = []
    self._event = event
    self._expected_request_count = 1

  def request_received(self, path, content):
    """
    Called by navigator server when it receives a metadata request.
    req is the python object for the request sent by agent.
    """
    path_parts = path.split('/')
    service_name = path_parts[-1]
    events_received = simplejson.loads(content)

    self._services.add(service_name)
    self._events_received.extend(events_received)

    self._expected_request_count -= 1
    if self._expected_request_count == 0:
      self._event.set()

  def clear(self):
    """
    Reinitialize all member variables so that a test can run
    with clean state.
    """
    self._services = set()
    self._events_received = []
    self._expected_request_count = 1

  def set_expected_request_count(self, expected_request_count):
    self._expected_request_count = expected_request_count

  def is_service_present(self, service_name):
    return service_name in self._services

  def get_events_received(self):
    return self._events_received


class TestMetadataServer(object):
  """
  This class runs a navigator server and tests if whole module works together.
  This is more of integration test for the module than a unit test and helps us
  catch integration issues that individual unit tests fail to catch.
  """
  _port = None
  _host = None
  _event = Event()
  _metadata_entries = MetadataEntries(_event)

  @classmethod
  def setup_class(cls):
    """
    Start navigator server on localhost and a free port. Then save this
    information so we can later create a lineage config file with this
    information.
    """

    cls._server = NavigatorServer(cls._metadata_entries, None, MetadataRequestHandler)
    cls._server.start_server()
    cls._host = cls._server.get_host()
    cls._port = cls._server.get_port()

  @classmethod
  def teardown_class(cls):
    """
    Stop navigator server.
    """
    cls._server.stop_server()

  def setup(self):
    """
    Setup lineage logs and lineage client configuration.
    """
    shutil.rmtree(_LINEAGE_LOG_DIR, True)
    shutil.rmtree(_ROLE_CONFIG_DIR, True)
    os.makedirs(_LINEAGE_LOG_DIR)
    os.makedirs(_ROLE_CONFIG_DIR)
    self._metadata_entries.clear()
    self._stop_metadata_thread()

  def teardown(self):
    """
    Cleanup lineage logs and client configuration.
    """
    shutil.rmtree(_LINEAGE_LOG_DIR)
    shutil.rmtree(_ROLE_CONFIG_DIR)
    self._stop_metadata_thread()

  def _start_metadata_thread(self):
    """
    Start audit thread.
    """
    cmf.audit.metadata_thread = MetadataThread(0, "TestMetadataThread")
    cmf.audit.metadata_thread._nav_thread.start()

  def _stop_metadata_thread(self):
    """
    Stop audit thread if its running.
    """
    if cmf.audit.metadata_thread._nav_thread.is_alive():
      cmf.audit.metadata_thread.stop()

  def _wait_for_watch_to_be_set(self, nav_thread, path_to_be_watched):
    wait_and_assert_true(lambda: nav_thread.get_watch_manager() is not None, 360)
    wm = nav_thread.get_watch_manager()
    for i in wm.watches:
      watch = wm.watches[i]
      if watch.path == path_to_be_watched:
        return True

  def _create_lineage_client_config(self, service_name, service_type, role_name,
      role_type):
    """
    Write a navigator.lineage.client.properties file.
    """
    conf_file_path = os.path.join(_ROLE_CONFIG_DIR, METADATA_CONFIG_FILENAME)
    conf_file = None
    try:
      conf_file = open(conf_file_path, 'w')
      print >> conf_file, "navigator.server.url=http://%s:%d" % (
          self._host, self._port)
      print >> conf_file, "navigator.client.serviceName=" + service_name
      print >> conf_file, "navigator.client.serviceType=" + service_type
      print >> conf_file, "navigator.client.roleName=" + role_name
      print >> conf_file, "navigator.client.roleType=" + role_type
      print >> conf_file, "navigator.client.clusterName=Cluster 1"
      print >> conf_file, "navigator.server.timeout_seconds=120"
      print >> conf_file, "lineage_event_log_dir=" + _LINEAGE_LOG_DIR
    finally:
      if conf_file is not None:
        conf_file.close()

    return conf_file_path

  def test_metadata_csd_plugin_crash_lineage(self):

    # create a file in the log directory first, to simulate an agent crash
    log_file_path = os.path.join(_LINEAGE_LOG_DIR, _SPARK_LINEAGE_FILE_NAME)
    log_file = open(log_file_path, "w")
    print >> log_file, _SPARK_LINEAGE_LOG_TEXT
    log_file.close()

    # create lineage properties file and start mock server
    self._create_lineage_client_config("SPARK-1", "SPARK",
        "SPARK-ROLENAME-abee5a758942d6c689c7a5e7dd2da179", "SPARKD")
    metadata_plugin = make_spark_metadata_plugin()
    self._start_metadata_thread()
    metadata_plugin.refresh(_ROLE_CONFIG_DIR)

    # wait till the watch on the lineage directory is set
    self._wait_for_watch_to_be_set(cmf.audit.metadata_thread, _LINEAGE_LOG_DIR)

    # wait for inotify event to get processed
    self._event.wait(30)
    self._event.clear()

    # make sure pipeline is created at this point
    assert_true(len(metadata_plugin.get_nav_pipelines()) == 1)

    # verify 2 events are received
    assert_true(len(metadata_plugin.get_nav_pipelines()) == 1)

    self._stop_metadata_thread()

