# Copyright (c) 2015 Cloudera, Inc. All rights reserved.

import cmf.audit
import simplejson
import os
import shutil

from cmf.audit import METADATA_CONFIG_FILENAME
from cmf.audit.navigator_thread import MetadataThread
from cmf.audit.util import make_metadata_plugin
from cmf_test.audit.navigator_server import NavigatorServer, MetadataRequestHandler
from cmf_test.audit.test_util import skiprhel5
from nose.tools import assert_true, assert_equal
from threading import Event

_LINEAGE_LOG_DIR = "/tmp/lineage_log"
_ROLE_CONFIG_DIR = "/tmp/nav_client_config_dir"
_ROLE_CONFIG_DIR2 = "/tmp/nav_client_config_dir2"

_IMPALA_LINEAGE_FILE_NAME = "impala_lineage_log_1.0-1377798507857"
_IMPALA_LINEAGE_LOG_TEXT = """{\
"hash": "cde9af1116c8bbfaded7b9654f1f9c4a", \
"timestamp": 1428084071, \
"vertices": [\
{"vertexType": "COLUMN", "vertexId": "code", "id": 0}, \
{"vertexType": "COLUMN", "vertexId": "default.sample_07.code", "id": 1}, \
{"vertexType": "COLUMN", "vertexId": "description", "id": 2}, \
{"vertexType": "COLUMN", "vertexId": "default.sample_07.description", "id": 3}, \
{"vertexType": "COLUMN", "vertexId": "total_emp", "id": 4}, \
{"vertexType": "COLUMN", "vertexId": "default.sample_07.total_emp", "id": 5}, \
{"vertexType": "COLUMN", "vertexId": "salary", "id": 6}, \
{"vertexType": "COLUMN", "vertexId": "default.sample_07.salary", "id": 7}, \
{"vertexType": "COLUMN", "vertexId": "default.sample_08.salary", "id": 8}], \
"edges": [\
{"sources": [1], "targets": [0], "edgeType": "PROJECTION"}, \
{"sources": [3], "targets": [2], "edgeType": "PROJECTION"}, \
{"sources": [5], "targets": [4], "edgeType": "PROJECTION"}, \
{"sources": [7], "targets": [6], "edgeType": "PROJECTION"}, \
{"sources": [7, 8], "targets": [0, 2, 4, 6], "edgeType": "PREDICATE"}], \
"user": "root", \
"queryText": "select * from sample_07 where salary > (select min(salary) from sample_08)"\
}
{"hash": "cde9af1116c8bbfaded7b9654f1f9c4b", "timestamp": 1428084072, "vertices": [{"vertexType": "COLUMN", "vertexId": "code", "id": 0}, {"vertexType": "COLUMN", "vertexId": "default.sample_07.code", "id": 1}], "edges": [{"sources": [1], "targets": [0], "edgeType": "PROJECTION"}], "user": "root", "queryText": "simple fake query"}
{"hash": "cde9af1116c8bbfaded7b9654f1f9c4c", "timestamp": 1428084073, "vertices": [], "edges": [], "user": "root", "queryText": "even more simple fake query 1"}"""

_CONFIG_DICT = {"conf_dir": _ROLE_CONFIG_DIR}

class MetadataEntries(object):
  """
  Class to hold all the lineage entries received by the mock navigator server.
  """

  def __init__(self, event):
    self._services = set()
    self._events_received = []
    self._event = event
    self._expected_request_count = 1

  def request_received(self, path, content):
    """
    Called by navigator server when it receives a metadata request.
    req is the python object for the request sent by agent.
    """
    path_parts = path.split('/')
    service_name = path_parts[-1]
    events_received = simplejson.loads(content)

    self._services.add(service_name)
    self._events_received.extend(events_received)

    self._expected_request_count -= 1
    if self._expected_request_count == 0:
      self._event.set()

  def clear(self):
    """
    Reinitialize all member variables so that a test can run
    with clean state.
    """
    self._services = set()
    self._events_received = []
    self._expected_request_count = 1

  def set_expected_request_count(self, expected_request_count):
    self._expected_request_count = expected_request_count

  def is_service_present(self, service_name):
    return service_name in self._services

  def get_events_received(self):
    return self._events_received


class TestMetadataServer(object):
  """
  This class runs a navigator server and tests if whole module works together.
  This is more of integration test for the module than a unit test and helps us
  catch integration issues that individual unit tests fail to catch.
  """
  _port = None
  _host = None
  _event = Event()
  _metadata_entries = MetadataEntries(_event)

  @classmethod
  def setup_class(cls):
    """
    Start navigator server on localhost and a free port. Then save this
    information so we can later create a lineage config file with this
    information.
    """
    #regular metadata server
    cls._server = NavigatorServer(cls._metadata_entries, None, MetadataRequestHandler)
    cls._server.start_server()
    cls._host = cls._server.get_host()
    cls._port = cls._server.get_port()
    #telemetry publisher metadata server
    cls._tp_server = NavigatorServer(cls._metadata_entries, None, MetadataRequestHandler)
    cls._tp_server.start_server()
    cls._tp_host = cls._tp_server.get_host()
    cls._tp_port = cls._tp_server.get_port()

  @classmethod
  def teardown_class(cls):
    """
    Stop navigator server.
    """
    cls._server.stop_server()

  def setup(self):
    """
    Setup lineage logs and lineage client configuration.
    """
    shutil.rmtree(_LINEAGE_LOG_DIR, True)
    shutil.rmtree(_ROLE_CONFIG_DIR, True)
    shutil.rmtree(_ROLE_CONFIG_DIR2, True)
    os.makedirs(_LINEAGE_LOG_DIR)
    os.makedirs(_ROLE_CONFIG_DIR)
    os.makedirs(_ROLE_CONFIG_DIR2)
    self._metadata_entries.clear()
    self._stop_metadata_thread()

  def teardown(self):
    """
    Cleanup lineage logs and client configuration.
    """
    shutil.rmtree(_LINEAGE_LOG_DIR)
    shutil.rmtree(_ROLE_CONFIG_DIR)
    shutil.rmtree(_ROLE_CONFIG_DIR2)
    self._stop_metadata_thread()

  def _start_metadata_thread(self):
    """
    Start audit thread.
    """
    cmf.audit.metadata_thread = MetadataThread(0, "TestMetadataThread")
    cmf.audit.metadata_thread._nav_thread.start()

  def _stop_metadata_thread(self):
    """
    Stop audit thread if its running.
    """
    if cmf.audit.metadata_thread._nav_thread.is_alive():
      cmf.audit.metadata_thread.stop()

  def _create_lineage_client_config(self, service_name, service_type, role_name,
      role_type):
    """
    Write a navigator.lineage.client.properties file.
    """
    conf_file_path = os.path.join(_ROLE_CONFIG_DIR, METADATA_CONFIG_FILENAME)
    conf_file = None
    try:
      conf_file = open(conf_file_path, 'w')
      print >> conf_file, "navigator.server.url=http://%s:%d" % (
          self._host, self._port)
      print >> conf_file, "telemetry_publisher.server.url=http://%s:%d" % (self._tp_host, self._tp_port)
      print >> conf_file, "navigator.client.serviceName=" + service_name
      print >> conf_file, "navigator.client.serviceType=" + service_type
      print >> conf_file, "navigator.client.roleName=" + role_name
      print >> conf_file, "navigator.client.roleType=" + role_type
      print >> conf_file, "navigator.client.clusterName=Cluster 1"
      print >> conf_file, "navigator.server.timeout_seconds=120"
      print >> conf_file, "lineage_event_log_dir=" + _LINEAGE_LOG_DIR
    finally:
      if conf_file is not None:
        conf_file.close()

    return conf_file_path

  def test_send_impala_lineage(self):
    """
    Send impala lineage entries to a real navigator server.
    """
    self._create_lineage_client_config("IMPALA-1", "IMPALA",
        "IMPALA-1-IMPALAD-abee5a758942d6c689c7a5e7dd2da179", "IMPALAD")
    metadata_plugin = make_metadata_plugin("impala-IMPALAD", _CONFIG_DICT)
    self._start_metadata_thread()
    metadata_plugin.refresh(_ROLE_CONFIG_DIR)

    # Now create a lineage log file and put few events in it
    # Then check if the events are received correctly
    log_file_path = os.path.join(_LINEAGE_LOG_DIR, _IMPALA_LINEAGE_FILE_NAME)
    log_file = open(log_file_path, "w")
    print >> log_file, _IMPALA_LINEAGE_LOG_TEXT
    log_file.close()

    # Wait for event to be sent and received by navigator server
    self._event.wait(60)
    self._event.clear()
    self._stop_metadata_thread()

    assert_true(self._metadata_entries.is_service_present("IMPALA-1"))
    assert_equal(6, len(self._metadata_entries.get_events_received()))

  @skiprhel5
  def test_send_impala_with_tp_on_and_off(self):
    """
    Send impala lineage events to a real nav server with off/on telemetry publisher.
    """
    self._create_lineage_client_config("IMPALA-1", "IMPALA",
        "IMPALA-1-IMPALAD-abee5a758942d6c689c7a5e7dd2da179", "IMPALAD")
    metadata_plugin = make_metadata_plugin("impala-IMPALAD", _CONFIG_DICT)
    self._start_metadata_thread()
    metadata_plugin.refresh(_ROLE_CONFIG_DIR)

    #stop TP server, so events should only go to metadata server
    self._tp_server.stop_server()

    # Now create a lineage log file and put few events in it
    # Then check if the events are received correctly
    log_file_path = os.path.join(_LINEAGE_LOG_DIR, _IMPALA_LINEAGE_FILE_NAME)
    log_file = open(log_file_path, "w")
    print >> log_file, _IMPALA_LINEAGE_LOG_TEXT
    log_file.close()

    # Wait for event to be sent and received by metadata server
    self._event.wait(60)
    self._event.clear()

    assert_true(self._metadata_entries.is_service_present("IMPALA-1"))
    assert_equal(3, len(self._metadata_entries.get_events_received()))
    self._metadata_entries.clear()

    #restart TP server, now it should get same event as metadata server
    self._tp_server.start_server_on_host_port(self._tp_host, self._tp_port)
    # Wait for event to be sent and received by TP server
    self._event.wait(60)
    self._event.clear()
    self._stop_metadata_thread()
    assert_true(self._metadata_entries.is_service_present("IMPALA-1"))
    assert_equal(3, len(self._metadata_entries.get_events_received()))

  @skiprhel5
  def test_send_impala_with_metadata_on_and_off(self):
    """
    Send impala lineage events to a real nav server with off/on telemetry publisher.
    """
    self._create_lineage_client_config("IMPALA-1", "IMPALA",
        "IMPALA-1-IMPALAD-abee5a758942d6c689c7a5e7dd2da179", "IMPALAD")
    metadata_plugin = make_metadata_plugin("impala-IMPALAD", _CONFIG_DICT)
    self._start_metadata_thread()
    metadata_plugin.refresh(_ROLE_CONFIG_DIR)

    #stop TP server, so events should only go to metadata server
    self._server.stop_server()

    # Now create a lineage log file and put few events in it
    # Then check if the events are received correctly
    log_file_path = os.path.join(_LINEAGE_LOG_DIR, _IMPALA_LINEAGE_FILE_NAME)
    log_file = open(log_file_path, "w")
    print >> log_file, _IMPALA_LINEAGE_LOG_TEXT
    log_file.close()

    # Wait for event to be sent and received by metadata server
    self._event.wait(60)
    self._event.clear()

    assert_true(self._metadata_entries.is_service_present("IMPALA-1"))
    assert_equal(3, len(self._metadata_entries.get_events_received()))
    self._metadata_entries.clear()

    #restart TP server, now it should get same event as metadata server
    self._server.start_server_on_host_port(self._host, self._port)
    # Wait for event to be sent and received by TP server
    self._event.wait(60)
    self._event.clear()
    self._stop_metadata_thread()
    assert_true(self._metadata_entries.is_service_present("IMPALA-1"))
    assert_equal(3, len(self._metadata_entries.get_events_received()))

  @skiprhel5
  def test_send_impala_many_files(self):
    """
    Send many impala lineage events to a metadata server and telemetry publisher.
    """
    self._create_lineage_client_config("IMPALA-1", "IMPALA",
        "IMPALA-1-IMPALAD-abee5a758942d6c689c7a5e7dd2da179", "IMPALAD")
    _CONFIG_DICT["conf_dir"] = _ROLE_CONFIG_DIR
    metadata_plugin = make_metadata_plugin("impala-IMPALAD", _CONFIG_DICT)
    self._start_metadata_thread()
    metadata_plugin.refresh(_ROLE_CONFIG_DIR)

    for i in range(0, 5):
      # Now create a lineage log file and put few events in it
      # Then check if the events are received correctly
      log_file_path = os.path.join(_LINEAGE_LOG_DIR, _IMPALA_LINEAGE_FILE_NAME+str(i))
      log_file = open(log_file_path, "w")
      print >> log_file, _IMPALA_LINEAGE_LOG_TEXT
      log_file.close()

    # Wait for event to be sent and received by metadata server
    self._event.wait(60)
    self._event.clear()
    self._stop_metadata_thread()

    assert_true(self._metadata_entries.is_service_present("IMPALA-1"))
    assert_equal(30, len(self._metadata_entries.get_events_received()))
    self._metadata_entries.clear()
