# Copyright (c) 2015 Cloudera, Inc. All rights reserved.
import os
import shutil
import cmf.audit

from cmf.audit import METADATA_CONFIG_FILENAME
from cmf.audit.navigator_conf import DEFAULT_MIN_BATCH_SIZE, DEFAULT_MAX_REQUEST_SIZE
from cmf.audit.navigator_sender import Result, MetadataSender
from cmf.audit.navigator_thread import MetadataThread
from cmf.audit.util import make_metadata_plugin
from nose.tools import assert_equal, nottest
from threading import Event

_LINEAGE_LOG_DIR = "/tmp/lineage_log"
_LINEAGE_LOG_DIR_2 = "/tmp/lineage_log2"
_ROLE_CONFIG_DIR = "/tmp/nav_client_config_dir"
_IMPALA_LINEAGE_FILE_NAME = "impala_lineage_log_1.0-1377798507857"
_IMPALA_LINEAGE_LOG_TEXT = """{\
"hash": "cde9af1116c8bbfaded7b9654f1f9c4a", \
"timestamp": 1428084071, \
"vertices": [\
{"vertexType": "COLUMN", "vertexId": "code", "id": 0}, \
{"vertexType": "COLUMN", "vertexId": "default.sample_07.code", "id": 1}, \
{"vertexType": "COLUMN", "vertexId": "description", "id": 2}, \
{"vertexType": "COLUMN", "vertexId": "default.sample_07.description", "id": 3}, \
{"vertexType": "COLUMN", "vertexId": "total_emp", "id": 4}, \
{"vertexType": "COLUMN", "vertexId": "default.sample_07.total_emp", "id": 5}, \
{"vertexType": "COLUMN", "vertexId": "salary", "id": 6}, \
{"vertexType": "COLUMN", "vertexId": "default.sample_07.salary", "id": 7}, \
{"vertexType": "COLUMN", "vertexId": "default.sample_08.salary", "id": 8}], \
"edges": [\
{"sources": [1], "targets": [0], "edgeType": "PROJECTION"}, \
{"sources": [3], "targets": [2], "edgeType": "PROJECTION"}, \
{"sources": [5], "targets": [4], "edgeType": "PROJECTION"}, \
{"sources": [7], "targets": [6], "edgeType": "PROJECTION"}, \
{"sources": [7, 8], "targets": [0, 2, 4, 6], "edgeType": "PREDICATE"}], \
"user": "root", \
"queryText": "select * from sample_07 where salary > (select min(salary) from sample_08)"\
}
{"hash": "cde9af1116c8bbfaded7b9654f1f9c4b", "timestamp": 1428084072, "vertices": [{"vertexType": "COLUMN", "vertexId": "code", "id": 0}, {"vertexType": "COLUMN", "vertexId": "default.sample_07.code", "id": 1}], "edges": [{"sources": [1], "targets": [0], "edgeType": "PROJECTION"}], "user": "root", "queryText": "simple fake query"}
{"hash": "cde9af1116c8bbfaded7b9654f1f9c4c", "timestamp": 1428084073, "vertices": [], "edges": [], "user": "root", "queryText": "even more simple fake query 1"}"""

class MockMetadataSender(MetadataSender):
  """
  Mock metadata sender object that we will pass to MetadataThread for it to send
  events. Lets us validate events sent by metadata thread.
  """
  def __init__(self, event, successes_to_mock=1, failures_to_mock=0):
    """
    Mock sender to send metadata event to metadata server. It sets the "event"
    when "send_nav_log_entries" method is called. If "failures_to_mock" is set,
    then it fails the request to send "failures_to_mock" number of times
    before succeeding.
    """
    self._entries = []
    self._plugin = None
    self._event = event
    self._failures_to_mock = failures_to_mock
    self._successes_to_mock = successes_to_mock

  def send_nav_log_entries(self, plugin, log_entry_batch):
    """
    Store the arguments and make them available for correctness checking
    """
    self._plugin = plugin
    entries = [entry for (_, _, entry) in log_entry_batch]

    # Invoke the entry parser to makes sure it can parse entries correctly
    self._get_parsed_entries(plugin, log_entry_batch)

    if self._failures_to_mock > 0:
      self._failures_to_mock -= 1
      return Result.FAILED
    else:
      self._entries.extend(entries)
      self._successes_to_mock -= 1
      if self._successes_to_mock < 1:
        # Set the event flag to true so any waiting tasks can free up
        self._event.set()

    return Result.SUCCESS

  def get_event_count(self):
    return len(self._entries)

  def get_event_parser(self):
    if self._plugin is None:
      return None
    else:
      return self._plugin.get_log_entry_parser()

  def assert_exact_num_of_successes(self):
    assert_equal(self._successes_to_mock, 0)

  def assert_exact_num_of_failures(self):
    assert_equal(self._failures_to_mock, 0)

  def get_service_name(self):
    if self._plugin is None:
      return None
    else:
      return self._plugin.get_service_name()

  def get_service_type(self):
    if self._plugin is None:
      return None
    else:
      return self._plugin.get_service_type()


class TestMetadataThread(object):
  """
  Test that main metadata thread can process events for all role types.
  """
  '''
  def setup(self):
    """
    Setup audit logs and audit client configuration.
    """
    shutil.rmtree(_ROLE_CONFIG_DIR, True)
    shutil.rmtree(_LINEAGE_LOG_DIR, True)
    shutil.rmtree(_LINEAGE_LOG_DIR_2, True)
    os.makedirs(_ROLE_CONFIG_DIR)
    os.makedirs(_LINEAGE_LOG_DIR)
    os.makedirs(_LINEAGE_LOG_DIR_2)
    # stop the main audit thread
    cmf.audit.audit_thread.stop()

  def teardown(self):
    """
    Cleanup audit logs and client configuration.
    """
    shutil.rmtree(_ROLE_CONFIG_DIR)
    shutil.rmtree(_LINEAGE_LOG_DIR)
    shutil.rmtree(_LINEAGE_LOG_DIR_2)
  '''

  def _create_lineage_log_file(self, file_name,
      lineage_log_dir=_LINEAGE_LOG_DIR, content="Test Lineage Log Line"):
    """
    Creates a new lineage log file and returns the path to the file.
    """
    lineage_file_path = os.path.join(lineage_log_dir, file_name)
    lineage_file = open(lineage_file_path, "w")
    print >> lineage_file, content
    lineage_file.close()

    return lineage_file_path

  def _create_lineage_client_config(self, service_name, service_type, role_name,
      role_type, batch_size=DEFAULT_MIN_BATCH_SIZE,
      max_request_size=DEFAULT_MAX_REQUEST_SIZE,
      lineage_log_dir=_LINEAGE_LOG_DIR):
    """
    Write a navigator.lineage.client.properties file.
    """
    conf_file_path = os.path.join(_ROLE_CONFIG_DIR, METADATA_CONFIG_FILENAME)
    conf_file = None
    try:
      conf_file = open(conf_file_path, 'w')
      print >> conf_file, "navigator.server.url=" \
                          "http://nightly-1.ent.cloudera.com:7186"
      print >> conf_file, "navigator.client.serviceName=" + service_name
      print >> conf_file, "navigator.client.serviceType=" + service_type
      print >> conf_file, "navigator.client.roleName=" + role_name
      print >> conf_file, "navigator.client.roleType=" + role_type
      print >> conf_file, "navigator.client.clusterName=Cluster 1"
      print >> conf_file, "navigator.server.timeout_seconds=120"
      print >> conf_file, "lineage_event_log_dir=" + lineage_log_dir
      print >> conf_file, "navigator.batch.min_size=" + str(batch_size)
      print >> conf_file, "navigator.client.max_request_size=" + \
                          str(max_request_size)
    finally:
      if conf_file is not None:
        conf_file.close()

    return conf_file_path

  def _check_entries(self, service_name, plugin, lineage_file_name,
      lineage_file_text, num_event,
      event_parser="cmf.audit.generic_metadata_parser",
      expected_parser_name="parse_generic_metadata_event", successes_to_mock=1,
      failures_to_mock=0):
    """
    Helper method to contain all the common code for tests.
    """
    event = Event()
    metadata_sender = MockMetadataSender(event, successes_to_mock, failures_to_mock)
    test_metadata_thread = MetadataThread(0, "TestMetadataThread")
    plugin.refresh(_ROLE_CONFIG_DIR, refresh_thread=lambda _: None)
    plugin.immediate_refresh()

    test_metadata_thread.add_navigator_plugin(plugin)
    test_metadata_thread.refresh_nav_plugin(plugin, metadata_sender)

    # Now create a lineage log file and put few events in it
    # Then check if the events are received correctly
    self._create_lineage_log_file(lineage_file_name, content=lineage_file_text)

    event.wait(60)  # Wait for event to be sent
    test_metadata_thread.stop()
    assert_equal(metadata_sender.get_service_name(), service_name)
    assert_equal(metadata_sender.get_event_count(), num_event)
    assert_equal(metadata_sender.get_event_parser().func_name,
      expected_parser_name)
    assert_equal(metadata_sender.get_event_parser().__module__, event_parser)
    metadata_sender.assert_exact_num_of_successes()
    metadata_sender.assert_exact_num_of_failures()

  @nottest
  def test_impalad_entries(self):
    """
    Test if lineage entries from impalad can be parsed successfully.
    """
    service_name = "IMPALA-1"
    self._create_lineage_client_config(
      service_name, "IMPALA",
      "IMPALA-1-IMPALAD-abee5a758942d6c689c7a5e7dd2da179",
      "IMPALAD")

    self._check_entries(service_name,
      make_metadata_plugin("impala-IMPALAD"),
      _IMPALA_LINEAGE_FILE_NAME,
      _IMPALA_LINEAGE_LOG_TEXT,
      3)

  @nottest
  def test_event_retry(self):
    """
    Test if the sender fails to send message, we retry those till we succeed.
    This test makes sure we don't drop events in case of any failure when
    transmitting them.
    """
    service_name = "IMPALA-1"
    self._create_lineage_client_config(service_name, "IMPALA",
      "IMPALA-1-IMPALAD-abee5a758942d6c689c7a5e7dd2da179",
      "IMPALAD")

    self._check_entries(service_name,
      make_metadata_plugin("impala-IMPALAD"),
      _IMPALA_LINEAGE_FILE_NAME,
      _IMPALA_LINEAGE_LOG_TEXT,
      3,
      failures_to_mock=1)

  @nottest
  def test_max_request_size(self):
    """
    Test if the number of lineage entries exceed max request size, we break it
    into multiple send requests.
    """
    service_name = "IMPALA-1"
    self._create_lineage_client_config(service_name, "IMPALA",
      "IMPALA-1-IMPALAD-abee5a758942d6c689c7a5e7dd2da179",
      "IMPALAD",
      max_request_size=10)

    self._check_entries(service_name,
      make_metadata_plugin("impala-IMPALAD"),
      _IMPALA_LINEAGE_FILE_NAME,
      _IMPALA_LINEAGE_LOG_TEXT,
      3,
      successes_to_mock=3)

  @nottest
  def test_batch_size(self):
    """
    Test if number of events are more than the batch size, we break it into
    multiple requests.
    """
    service_name = "IMPALA-1"
    self._create_lineage_client_config(service_name, "IMPALA",
      "IMPALA-1-IMPALAD-abee5a758942d6c689c7a5e7dd2da179",
      "IMPALAD",
      batch_size=2)

    self._check_entries(service_name,
      make_metadata_plugin("impala-IMPALAD"),
      _IMPALA_LINEAGE_FILE_NAME,
      _IMPALA_LINEAGE_LOG_TEXT,
      3,
      successes_to_mock=2)

  @nottest
  def test_lineage_log_dir_move(self):
    """
    Test when lineage directory moves, we pick up new directory and read logs
    from the new directory.
    OPSAPS-17815 reported this issue, so created test to make sure we don't
    break it in future.
    """
    service_name = "IMPALA-1"
    conf_file_path = self._create_lineage_client_config(service_name, "IMPALA",
      "IMPALA-1-IMPALAD-abee5a758942d6c689c7a5e7dd2da179",
      "IMPALAD")

    event = Event()
    plugin = make_metadata_plugin("impala-IMPALAD")
    metadata_sender = MockMetadataSender(event, 1, 0)

    test_metadata_thread = MetadataThread(0, "TestMetadataThread")

    plugin.refresh(_ROLE_CONFIG_DIR, refresh_thread=lambda _: None)
    plugin.immediate_refresh()

    test_metadata_thread.add_navigator_plugin(plugin)
    test_metadata_thread.refresh_nav_plugin(plugin, metadata_sender)

    # Now create a lineage log file and put few events in it
    self._create_lineage_log_file(_IMPALA_LINEAGE_FILE_NAME,
      content=_IMPALA_LINEAGE_LOG_TEXT)

    event.wait(60)  # Wait for event to be sent
    event.clear()  # reset event so that we can wait again

    # Check that it worked
    assert_equal(metadata_sender.get_service_name(), service_name)
    assert_equal(metadata_sender.get_event_count(), 3)
    metadata_sender.assert_exact_num_of_successes()
    metadata_sender.assert_exact_num_of_failures()

    # Now do this again with a new lineage log directory
    metadata_sender._successes_to_mock += 1
    os.remove(conf_file_path)
    self._create_lineage_client_config(service_name, "IMPALA",
      "IMPALA-1-IMPALAD-abee5a758942d6c689c7a5e7dd2da179",
      "IMPALAD",
      lineage_log_dir=_LINEAGE_LOG_DIR_2)

    plugin.refresh(_ROLE_CONFIG_DIR, refresh_thread=lambda _: None)
    plugin.immediate_refresh()

    test_metadata_thread.add_navigator_plugin(plugin)
    test_metadata_thread.refresh_nav_plugin(plugin, metadata_sender)

    # Now create a lineage log file and put few events in it
    self._create_lineage_log_file(_IMPALA_LINEAGE_FILE_NAME,
      lineage_log_dir=_LINEAGE_LOG_DIR_2,
      content=_IMPALA_LINEAGE_LOG_TEXT)

    event.wait(60)  # Wait for event to be sent
    test_metadata_thread.stop()

    assert_equal(metadata_sender.get_event_count(), 6)
    metadata_sender.assert_exact_num_of_successes()
    metadata_sender.assert_exact_num_of_failures()
