# Copyright (c) 2017 Cloudera, Inc. All rights reserved.

import ConfigParser
import os
import shutil
import StringIO

from cmf.audit.navigator_conf import (PROFILE_CONFIG_FILENAME, ProfileConfig)
from cmf.audit.navigator_logs import NoneNavigatorLogFile, NonexistentNavigatorLogFile
from cmf.audit.util import make_profile_plugin

from nose.tools import assert_equal, assert_false, assert_true

ROLE_CONFIG_DIR = "/tmp/profile_role_config_dir"
PROFILE_LOG_DIR = "/tmp/profile_log"

class TestProfilePlugin(object):
  """
  Class that sets up and cleans up state for all the plugin tests.
  Contains test that are role insensitive.
  """

  def __init__(self,
               service_name = "TestService-1",
               service_type = "TestService",
               role_name = "TestRoleName",
               role_type = "TestRoleType"):
    self._service_name = service_name
    self._service_type = service_type
    self._role_name = role_name
    self._role_type = role_type

  def get_role_name(self):
    return self._role_name

  def setup(self):
    """
    Create profile log, metrics and role config directories
    """
    shutil.rmtree(ROLE_CONFIG_DIR, True)
    shutil.rmtree(PROFILE_LOG_DIR, True)
    os.makedirs(ROLE_CONFIG_DIR)
    os.makedirs(PROFILE_LOG_DIR)
    # Write a navigator.client.properties file.
    conf_file_path = os.path.join(ROLE_CONFIG_DIR, PROFILE_CONFIG_FILENAME)
    conf_file = None
    try:
      conf_file = open(conf_file_path, 'w')
      print >> conf_file, "navigator.server.url=http://host00001:10110"
      print >> conf_file, "navigator.client.serviceName=" + self._service_name
      print >> conf_file, "navigator.client.serviceType=" + self._service_type
      print >> conf_file, "navigator.client.roleName=" + self._role_name
      print >> conf_file, "navigator.client.roleType=" + self._role_type
      print >> conf_file, "navigator.batch.queue_policy=DROP"
      print >> conf_file, "telemetry_publisher.server.url=http://host00001:10110"
      print >> conf_file, "profile_event_log_dir=" + PROFILE_LOG_DIR
      print >> conf_file, "navigator.profile_log_max_file_size=100"
      print >> conf_file, "navigator.client.max_num_profile_log=10"
    finally:
      if conf_file is not None:
        conf_file.close()

    metrics_file_path = os.path.join(PROFILE_LOG_DIR, "metrics")
    metrics_file = None
    try:
      metrics_file = open(metrics_file_path, "w")
      print >> metrics_file, "profile_plugin_parse_errors=2"
      print >> metrics_file, "profile_plugin_filtered=478"
      print >> metrics_file, "profile_plugin_coalesced=13"
      print >> metrics_file, "profile_plugin_exceptions=0"
      print >> metrics_file, "profile_plugin_events=11"
    finally:
      if metrics_file is not None:
        metrics_file.close()

  def teardown(self):
    """
    Cleanup the configuration file.
    """
    shutil.rmtree(ROLE_CONFIG_DIR)
    shutil.rmtree(PROFILE_LOG_DIR)

class TestGenericProfilePlugin(TestProfilePlugin):
  def test_generic_profile_plugin_refresh(self):
    """
    Test if generic profile plugin's refresh method picks up all the properties
    for generic plugin configuration.
    """
    plugin = make_profile_plugin("generic-PROCESS")

    refresh_calls = []
    def mock_refresh_profile_thread(self):
      refresh_calls.append(self)

    plugin.refresh(ROLE_CONFIG_DIR, refresh_thread=mock_refresh_profile_thread)
    plugin.immediate_refresh()

    assert_equal(plugin.nav_conf.log_dir, PROFILE_LOG_DIR)
    assert_equal(1, len(refresh_calls))

    assert_equal(plugin.get_service_name(), self._service_name)
    assert_equal(plugin.get_role_type(), self._role_type)
    assert_equal(plugin.get_ipc_request_name(), "writeGenericProfileEvents")
    assert_true(plugin.nav_conf.has_telemetry_publisher_url())

    from cmf.audit.generic_profile_parser import parse_generic_profile_event
    assert_true(plugin.get_log_entry_parser() is parse_generic_profile_event)

    pipelines = plugin.get_nav_pipelines()
    assert_equal(len(pipelines), 1)

    for pipeline in pipelines:
      assert_equal(pipeline._get_log_prefix(self.get_role_name()),
        self.get_role_name())
      assert_false(pipeline.get_nav_logs_dir() is None)
      assert_true(type(pipeline.get_current_nav_log()) is NoneNavigatorLogFile)

      file_name, offset = pipeline.get_wal().get_file_info()
      assert_true(file_name is None)
      assert_true(offset is None)
      pipeline.switch_log("non-existing-file")
      assert_true(type(pipeline.get_current_nav_log()) is NonexistentNavigatorLogFile)

  def test_disabled_generic_profile_plugin_refresh(self):
    """
    Test if disabled generic profile plugin's refresh method is skipped.
    """
    plugin = make_profile_plugin("generic-PROCESS")

    refresh_calls = []
    def mock_refresh_profile_thread(self):
      refresh_calls.append(self)

    plugin.refresh(ROLE_CONFIG_DIR + 'x', refresh_thread=mock_refresh_profile_thread)

    assert_false(plugin.is_nav_enabled())
    assert_equal(0, len(refresh_calls))

  def test_profile_plugin_with_telemetry_pub(self):
    """
    Test that correct api url is returned when telemetry publisher is present.
    """
    conf_str = '[navigator]\n' + \
    'navigator.server.url=http://nms_host:7186\n' + \
    'telemetry_publisher.server.url=http://tp_host:7187'

    conf_fd = StringIO.StringIO(conf_str)
    config = ConfigParser.SafeConfigParser()
    config.readfp(conf_fd)
    options = {}
    for key_value in config.items("navigator"):
      options["navigator." + key_value[0]] = key_value[1]
    conf = ProfileConfig(options)

    plugin = make_profile_plugin("IMPALA-1")
    plugin.nav_conf = conf

    assert_true(plugin.nav_conf.has_telemetry_publisher_url())
    assert_equal(plugin.get_api_url("cluster", "service"), "/api/v1/impalaProfiles/cluster/service")
