# Copyright (c) 2013-2015 Cloudera, Inc. All rights reserved.
import errno
import os
import shutil
import tempfile
from cmf.audit import wal
from cmf.audit.wal import Wal
from nose.tools import assert_equal, nottest

_IMPALA_AUDIT_WAL_FILE_NAME = "impalad_audit_wal"
_AUDIT_FILE_NAME = "impala_audit_event_log_1.0-1377798507897"
_AUDIT_LOG_TEXT = """{"1377775857227":{"query_id":"84bd714e4ef1fa3:75ff54a3b1824295","session_id":"2d4c159702aeebe2:cb5377ebedf6168b","start_time":"2013-08-29 11:30:56.897841000","authorization_failure":false,"status":"","user":"impalauser","impersonator":null,"statement_type":"CREATE_DATABASE","network_address":"::ffff:127.0.0.1:34009","sql_statement":"create DATABASE impalaQueryDB_1377801039","catalog_objects":[{"name":"impalaQueryDB_1377801039","object_type":"DATABASE","privilege":"CREATE"}]}}
{"1377775857472":{"query_id":"70413d07e1ee7ec6:5d1b5f76aee04183","session_id":"2d4c159702aeebe2:cb5377ebedf6168b","start_time":"2013-08-29 11:30:57.441235000","authorization_failure":false,"status":"","user":"impalauser","impersonator":null,"statement_type":"USE","network_address":"::ffff:127.0.0.1:34009","sql_statement":"use impalaQueryDB_1377801039","catalog_objects":[{"name":"impalaQueryDB_1377801039","object_type":"DATABASE","privilege":"ANY"}]}}
{"1377775857516":{"query_id":"d04daaf9e488e953:5cb8bdc461eb9895","session_id":"2d4c159702aeebe2:cb5377ebedf6168b","start_time":"2013-08-29 11:30:57.488402000","authorization_failure":false,"status":"","user":"impalauser","impersonator":null,"statement_type":"DROP_TABLE","network_address":"::ffff:127.0.0.1:34009","sql_statement":"drop TABLE IF EXISTS testaudittable1","catalog_objects":[]}}"""
_AUDIT_WAL_OFFSET = 300

def remove_quietly(filename):
  """
  Remove a file quietly and ignore error if the file doesn't exist.
  """
  try:
    os.remove(filename)
  except OSError, e:
    if e.errno != errno.ENOENT:
      raise

class TestWalFile(object):
  """
  Class to test handling of various conditions of an Audit log wal file.
  """

  def setup(self):
    """
    Creates a wal file and an audit log file that is used for testing.
    """
    self._audit_dir = tempfile.mkdtemp(suffix='.scm-agent-test')

    wal_file, audit_file = None, None
    try:
      self._audit_file_path = os.path.join(self._audit_dir, _AUDIT_FILE_NAME)
      self._wal_file_path = os.path.join(self._audit_dir, _IMPALA_AUDIT_WAL_FILE_NAME)
      wal_file = open(self._wal_file_path, "w")
      audit_file = open(self._audit_file_path, "w")
      print >> audit_file, _AUDIT_LOG_TEXT,
      print >> wal_file, "%d,%s" % (_AUDIT_WAL_OFFSET, _AUDIT_FILE_NAME)
    finally:
      if wal_file is not None:
        wal_file.close()
      if audit_file is not None:
        audit_file.close()

  def teardown(self):
    """
    Close and remove the wal and the audit log for next test.
    """
    shutil.rmtree(self._audit_dir, ignore_errors=True)

  def test_simple_wal_read(self):
    """
    Simple positive test to check if a valid wal file can be read.
    """
    wal = Wal(self._audit_dir, _IMPALA_AUDIT_WAL_FILE_NAME, "Test")
    file_name, offset = wal.get_file_info()
    assert_equal(file_name, _AUDIT_FILE_NAME)
    assert_equal(offset, _AUDIT_WAL_OFFSET)

  def test_corrupted_wal_line(self):
    """
    Tests if the last line in the wal file is corrupted, then previous wal line
    is read successfully.
    """
    wal_file = open(self._wal_file_path, "a")
    print >> wal_file, "corrupted line",
    wal_file.close()

    # Even with a corrupt line we should get valid info from wal file
    wal = Wal(self._audit_dir, _IMPALA_AUDIT_WAL_FILE_NAME, "Test")
    file_name, offset = wal.get_file_info()
    assert_equal(file_name, _AUDIT_FILE_NAME)
    assert_equal(offset, _AUDIT_WAL_OFFSET)

  def test_only_corrupted_lines(self):
    """
    Test if WAL file contains only corrupted line
    """
    wal_file = open(self._wal_file_path, "w")
    print >> wal_file, "corrupted line",
    wal_file.close()

    # Even with all corrupt lines we should get valid info from wal file
    wal = Wal(self._audit_dir, _IMPALA_AUDIT_WAL_FILE_NAME, "Test")
    file_name, offset = wal.get_file_info()
    assert_equal(file_name, None)
    assert_equal(offset, None)

  def test_empty_wal_file(self):
    """
    Test if an empty wal file is handled correctly.
    """
    wal_file = open(self._wal_file_path, "w")
    wal_file.truncate()
    wal_file.close()

    # Even no content we should get valid info from wal file
    wal = Wal(self._audit_dir, _IMPALA_AUDIT_WAL_FILE_NAME, "Test")
    file_name, offset = wal.get_file_info()
    assert_equal(file_name, None)
    assert_equal(offset, None)

  def test_nonexistent_wal_file(self):
    """
    Test if Wal class can be instantiated successfully even if the wal file
    is not present.
    """
    os.remove(self._wal_file_path)
    wal = Wal(self._audit_dir, _IMPALA_AUDIT_WAL_FILE_NAME, "Test")
    file_name, offset = wal.get_file_info()
    assert_equal(file_name, None)
    assert_equal(offset, None)

  def test_checkpoint(self):
    """
    Test if wal file can be checkpointed successfully.
    """
    checkpoint_filename = "new_audit_file_to_trigger_wal_checkpoint"
    checkpoint_offset = 500

    wal = Wal(self._audit_dir, _IMPALA_AUDIT_WAL_FILE_NAME, "Test")
    wal._checkpoint(checkpoint_filename, checkpoint_offset)
    wal.close()

    # Reopen wal file and make sure we read it correctly
    wal = Wal(self._audit_dir, _IMPALA_AUDIT_WAL_FILE_NAME, "Test")
    file_name, offset = wal.get_file_info()
    assert_equal(file_name, checkpoint_filename)
    assert_equal(offset, checkpoint_offset)

  def test_multiple_wals_for_same_file(self):
    """
    Test if multiple wals can point to the same file without breaking each other.
    """
    wal1 = Wal(self._audit_dir, _IMPALA_AUDIT_WAL_FILE_NAME, "Test1")

    new_offset = 301
    wal1.append(_AUDIT_FILE_NAME, new_offset)

    file_name, offset = wal1.get_file_info()
    assert_equal(file_name, _AUDIT_FILE_NAME)
    assert_equal(offset, new_offset)

    wal2 = Wal(self._audit_dir, _IMPALA_AUDIT_WAL_FILE_NAME, "Test2")

    new_offset = 302
    wal2.append(_AUDIT_FILE_NAME, new_offset)

    # Check that both wal objects show the update. We don't use get_file_info,
    # because that only looks at the Wal instance's member variables. Instead
    # we use _get_init_file_info, since this actually looks at the file content.
    file_name, offset = wal1._get_init_file_info()
    assert_equal(file_name, _AUDIT_FILE_NAME)
    assert_equal(offset, new_offset)

    file_name, offset = wal2._get_init_file_info()
    assert_equal(file_name, _AUDIT_FILE_NAME)
    assert_equal(offset, new_offset)

    # Test checkpointing while two wal objects point to the same file -- we
    # expect that the file will not be checkpointed (i.e. the file won't be
    # replaced with a new file that has only the last entry), because this
    # would cause the other wal object to stop writing properly to the file.
    original_audit_wal_checkpoint_size = wal.AUDIT_WAL_CHECKPOINT_SIZE
    try:
      wal.AUDIT_WAL_CHECKPOINT_SIZE = 1
      new_offset = 303
      wal2._last_checkpoint_time -= wal.CHECKPOINT_INTERVAL_SECONDS
      wal2.append(_AUDIT_FILE_NAME, new_offset)
    finally:
      wal.AUDIT_WAL_CHECKPOINT_SIZE = original_audit_wal_checkpoint_size

    # Read the first line to confirm that we did not checkpoint (if we had
    # checkpointed, this file would have only one line, and that line would be
    # the offset we just set)
    wal2._walfd.seek(0)
    line = wal2._walfd.readline()
    file_name, offset = wal2._get_filename_offset_from_line(line)
    assert_equal(file_name, _AUDIT_FILE_NAME)
    assert_equal(offset, _AUDIT_WAL_OFFSET)

    # Read the last line to confirm that we saved the latest offset even though
    # we didn't checkpoint
    file_name, offset = wal2._get_init_file_info()
    assert_equal(file_name, _AUDIT_FILE_NAME)
    assert_equal(offset, new_offset)

    # Test closing
    wal2.close()
    new_offset = 304
    wal1.append(_AUDIT_FILE_NAME, new_offset)

    file_name, offset = wal1._get_init_file_info()
    assert_equal(file_name, _AUDIT_FILE_NAME)
    assert_equal(offset, new_offset)

    # Test checkpointing now that only one wal object exists -- at this point
    # the checkpoint should succeed.
    new_offset = 305
    wal1._checkpoint(_AUDIT_FILE_NAME, new_offset)

    # Read the first line to confirm that we checkpointed (if we failed to
    # checkpoint, there would be multiple lines in the wal file)
    wal1._walfd.seek(0)
    line = wal1._walfd.readline()
    file_name, offset = wal1._get_filename_offset_from_line(line)
    assert_equal(file_name, _AUDIT_FILE_NAME)
    assert_equal(offset, new_offset)
