#!/usr/bin/env python
# -- coding: utf-8 --

# Copyright (c) 2012 Cloudera, Inc. All rights reserved.
import tempfile
import time
import os
import logging
import simplejson as json
import datetime
import gzip
import zipfile
import random
import string
import StringIO
import tempfile
import time
import shutil
import sys
try:
  from cStringIO import StringIO
except:
  from StringIO import StringIO

from nose.tools import assert_true, assert_equal, assert_false, assert_raises
from nose.plugins.skip import SkipTest
from clusterstats.log.streaming.file import LogFile

import cmf.log_search as ls
import cmf.log_retrieve as lr
import cmf.redaction

TEST_LOGS_DIR = os.path.join(os.path.dirname(__file__), "test-logs")
TEST_LOGS = ["hadoop-cmf-hdfs1-DATANODE-4-dtrejovm2.log.out",
            "hadoop-cmf-hdfs1-DATANODE-6-dtrejovm2.log.out",
            "hadoop-cmf-hdfs1-DATANODE-6-localhost.log.out",
            "hadoop-cmf-hdfs1-NAMENODE-1-dtrejovm2.log.out",
            "hadoop-cmf-hdfs1-NAMENODE-8-dtrejovm2.log.out",
            "hadoop-cmf-hdfs1-NAMENODE-8-localhost.log.out",
            "hadoop-cmf-hdfs1-SECONDARYNAMENODE-3-dtrejovm2.log.out",
            "hadoop-cmf-hdfs1-SECONDARYNAMENODE-5-dtrejovm2.log.out",
            "hadoop-cmf-hdfs1-SECONDARYNAMENODE-5-localhost.log.out",
            "hadoop-hadoop-datanode-localhost.localdomain.log",
            "hadoop-hadoop-namenode-localhost.localdomain.log",
            "hadoop-hadoop-tasktracker-localhost.localdomain.log",
            "hadoop-cmf-hdfs1-DATANODE-4-dtrejovm2.log.out.2011-06-07"
            ]

TYPES_LIST = ["LOG4J"] * len(TEST_LOGS)

HUE_LOG_NAME = "runcpserver.log"

LOGS_LIST = [os.path.join(TEST_LOGS_DIR, log) for log in TEST_LOGS]

HUE_LOG = os.path.join(TEST_LOGS_DIR, HUE_LOG_NAME)

AGENT_LOG_NAME = "cloudera-scm-agent.log"

AGENT_LOG = os.path.join(TEST_LOGS_DIR, AGENT_LOG_NAME)

def test_get_byte_offsets():
  """
  test that byte offsets are correctly found in a log file
  """
  log_path = os.path.join(TEST_LOGS_DIR, "cloudera-scm-agent.log")
  log_type = "SCMAGENT"

  # The file goes from 17/Sep/2012 11:14:48 +0000 to 17/Sep/2012 11:36:48 +0000
  time_string = "%d/%b/%Y %H:%M:%S"
  # These numbers are multiplied by 1000 because esatimate_logs expects the time
  # like a long, not like a float.
  before_start = time.mktime(time.strptime("17/Sep/2012 11:10:00", time_string)) * 1000
  after_end = time.mktime(time.strptime("17/Sep/2012 11:40:00", time_string)) * 1000
  one_third = time.mktime(time.strptime("17/Sep/2012 11:20:00", time_string)) * 1000
  two_thirds = time.mktime(time.strptime("17/Sep/2012 11:23:20", time_string)) * 1000

  # Test that should return entire file
  compressor = lr.LogCompressor(log_path, None, None, 0, 0) # Not actually using state here
  fd = LogFile(log_path)
  start, end = compressor.get_byte_offsets(fd, log_type, None, after_end, 3000)
  assert_equal(0, start)
  assert_equal(2894, end)

  # Test that should return none of file
  start, end = compressor.get_byte_offsets(fd, log_type, None, before_start, 3000)
  assert_equal(0, start)
  assert_equal(0, end)

  # Test that should return end of file
  # 2894 - 500 should be in the middle of a line that ends at 2438
  start, end = compressor.get_byte_offsets(fd, log_type, None, after_end, 500)
  assert_equal(2438, start)
  assert_equal(2894, end)

  # Test that should return beginning of file
  start, end = compressor.get_byte_offsets(fd, log_type, None, one_third, 3000)
  assert_equal(0, start)
  assert_equal(977, end)

  # Test that should return middle of file
  start, end = compressor.get_byte_offsets(fd, log_type, None, two_thirds, 1500)
  assert_equal(977, start)
  assert_equal(2438, end)

def test_get_log_index():
  """
  tests that files will be indexed and therefore sorted correctly
  """
  compressor = lr.LogCompressor("", None, None, 0, 0) # Not actually using state here
  assert_equal(0, compressor.get_log_index("cloudera-scm-agent.log"))
  assert_equal(12, compressor.get_log_index("cloudera-scm-agent.log.12"))

  older_impala = "impalad.nightly-5.ent.cloudera.com.impala.log.INFO.20121130-061408.14555"
  newer_impala = "impalad.nightly-5.ent.cloudera.com.impala.log.INFO.20121130-155821.24443"
  newer_impala2 = "impalad.nightly-5.ent.cloudera.com.impala.log.INFO.20121130-155821.1000"
  # get log index should give the newer file a lower number regardless of pid.
  assert_true(compressor.get_log_index(newer_impala) < compressor.get_log_index(older_impala))
  assert_true(compressor.get_log_index(newer_impala2) < compressor.get_log_index(older_impala))
  assert_equal(compressor.get_log_index(newer_impala), compressor.get_log_index(newer_impala2))


def test_calculate_log_estimate():
    """
    tests for estimating log size
    :return:
    """
    log_paths = [LOGS_LIST[12]]
    log_types = [TYPES_LIST[12]]
    time_string = "%d/%m/%Y %H:%M:%S"

    start_time = time.mktime(time.strptime("07/06/2011 11:56:00", time_string)) * 1000
    end_time = time.mktime(time.strptime("07/06/2011 15:16:00", time_string)) * 1000
    after_end = time.mktime(time.strptime("07/06/2014 15:16:00", time_string)) * 1000
    before_start = time.mktime(time.strptime("07/06/2011 01:00:00", time_string)) * 1000
    start_time_boundary = time.mktime(time.strptime("07/06/2011 09:56:26", time_string)) * 1000
    end_time_boundary = time.mktime(time.strptime("07/06/2011 19:04:57", time_string)) * 1000

    # Both start_time and end_time are available in the file, should return block size
    # of [start_time, end_time]
    actual_result_json = json.loads(json.dumps(
            lr.calculate_log_estimate(start_time, end_time, log_paths, log_types, None)))
    expected_result_json = json.loads(json.dumps({'data': '{"estimate": 8768, "success": true}'}))
    assert_equal(actual_result_json, expected_result_json)

    # Only end_time is available in the file but start_time is not, should return
    # block size [0, end_time]
    actual_result_json = json.loads(json.dumps(
            lr.calculate_log_estimate(before_start, end_time, log_paths, log_types, None)))
    expected_result_json = json.loads(json.dumps({'data': '{"estimate": 9417, "success": true}'}))
    assert_equal(actual_result_json, expected_result_json)

    # Only start_time is available in the file but end_time is not, should return
    # block size [start_time, EOF)
    actual_result_json = json.loads(json.dumps(
            lr.calculate_log_estimate(start_time, after_end, log_paths, log_types, None)))
    expected_result_json = json.loads(json.dumps({'data': '{"estimate": 13802, "success": true}'}))
    assert_equal(actual_result_json, expected_result_json)

    # Both start_time and end_time are greater than the range available in the file,
    # should return block size of 0
    actual_result_json = json.loads(json.dumps(
            lr.calculate_log_estimate(after_end, after_end, log_paths, log_types, None)))
    expected_result_json = json.loads(json.dumps({'data': '{"estimate": 0, "success": true}'}))
    assert_equal(actual_result_json, expected_result_json)

    # Both start_time and end_time are lesser than the range available in the file,
    # should return block size of 0
    actual_result_json = json.loads(json.dumps(
            lr.calculate_log_estimate(before_start, before_start, log_paths, log_types, None)))
    expected_result_json = json.loads(json.dumps({'data': '{"estimate": 0, "success": true}'}))
    assert_equal(actual_result_json, expected_result_json)

    # Both start_time and end_time form the boundaries of the file, should return size of
    # the entire file
    actual_result_json = json.loads(json.dumps(
            lr.calculate_log_estimate(start_time_boundary, end_time_boundary, log_paths, log_types, None)))
    expected_result_json = json.loads(json.dumps({'data': '{"estimate": 14451, "success": true}'}))
    assert_equal(actual_result_json, expected_result_json)

    # Both start_time and end_time are available in the file, but end_time < start_time,
    # should throw an exception
    assert_raises(Exception, lr.calculate_log_estimate, end_time, start_time, log_paths, log_types)

def test_retrieve_log_compressed():
  """
  tests getting parts of plaintext log files.
  """
  log_path = LOGS_LIST[1]
  log_type = TYPES_LIST[1]
  byte_limit = 10000
  # This file ends a bit after 2011-06-03 16:00
  time_string = "%d/%m/%Y %H:%M:%S"
  end_time = time.mktime(time.strptime("03/06/2011 20:00:00", time_string)) * 1000

  gzip_fd = generator_to_fd(lr.retrieve_log_compressed(end_time, log_path, log_type, byte_limit, cmf.redaction.NOP_REDACTOR))

  assert_true(len(gzip_fd.getvalue()) > 0)

  # Unzip and check size
  unzipper = gzip.GzipFile(mode="r", fileobj=gzip_fd)

  unzipped_content = unzipper.read()

  # total file size isn't quite 10000 due to our requirement that whole log events get collected
  assert_equal(len(unzipped_content), 9955)


def test_retrieve_log_compressed_with_symlinks():
  """
  Tests getting parts of multiple plaintext log files, specifically,
  - Collect "agents/cmf/src/cmf_test/test-logs/with-symlinks/hadoop-cmf-hdfs1-DATANODE-4-dtrejovm2.log.out".
  - Collect "agents/cmf/src/cmf_test/test-logs/with-symlinks/hadoop-cmf-hdfs1-DATANODE-6-dtrejovm2.log.out"
  referenced by a symlink.
  - Skip collecting "agents/cmf/src/cmf_test/test-logs/with-symlinks/hadoop-cmf-hdfs1-DATANODE-non-existent-file.log.out"
  referenced by a symlink.
  """
  try:
    test_logs_with_symlinks_dir = os.path.join(TEST_LOGS_DIR, "with-symlinks")
    # The log path is a prefix that will be used as wildcard (i.e. log_path*) to match both
    # the files: hadoop-cmf-hdfs1-DATANODE-4-dtrejovm2.log.out and
    # hadoop-cmf-hdfs1-DATANODE-6-dtrejovm2.log.out
    log_path_prefix = os.path.join(test_logs_with_symlinks_dir, "hadoop-cmf-hdfs1-DATANODE-")
    # Path to a file that exists
    valid_log_path = os.path.join(test_logs_with_symlinks_dir, "hadoop-cmf-hdfs1-DATANODE-6-dtrejovm2.log.out")
    # Path to a file that doesn't exist
    invalid_log_path = os.path.join(test_logs_with_symlinks_dir, "hadoop-cmf-hdfs1-DATANODE-non-existent-file.log.out")
    # Create symlinks
    valid_symlink = os.path.join(test_logs_with_symlinks_dir, "valid_symlink")
    broken_symlink = os.path.join(test_logs_with_symlinks_dir, "broken_symlink")
    os.symlink(valid_log_path, valid_symlink)
    os.symlink(invalid_log_path, broken_symlink)
    # Verify symlinks and their referenced files
    assert_true(os.path.islink(valid_symlink))
    assert_true(os.path.exists(valid_symlink))
    assert_true(os.path.islink(broken_symlink))
    assert_false(os.path.exists(broken_symlink))
    # Collect and verify
    log_type = TYPES_LIST[1]
    byte_limit = 100000
    time_string = "%d/%m/%Y %H:%M:%S"
    end_time = time.mktime(time.strptime("03/07/2011 20:00:00", time_string)) * 1000
    gzip_fd = generator_to_fd(lr.retrieve_log_compressed(end_time, log_path_prefix,
                                                         log_type, byte_limit,
                                                         cmf.redaction.NOP_REDACTOR))
    assert_true(len(gzip_fd.getvalue()) > 0)
    # Unzip and check size
    unzipper = gzip.GzipFile(mode="r", fileobj=gzip_fd)
    unzipped_content = unzipper.read()
    # Total file size isn't quite the same due to our requirement that whole log
    # events gets collected
    assert_equal(len(unzipped_content),
                 # For hadoop-cmf-hdfs1-DATANODE-4-dtrejovm2.log.out
                 8907 +
                 # For hadoop-cmf-hdfs1-DATANODE-6-dtrejovm2.log.out
                 9955)
  finally:
    if os.path.islink(valid_symlink):
      os.remove(valid_symlink)
    if os.path.islink(broken_symlink):
      os.remove(broken_symlink)

def test_retrieve_log_redacted():
  """
  tests the redaction functionality of log retrieval
  """

  log_path = LOGS_LIST[5]
  log_type = TYPES_LIST[5]
  log_path_redacted = log_path.replace(".out", ".redacted")
  byte_limit = 1000000
  # This file ends a bit after 2011-06-03 16:00
  time_string = "%d/%m/%Y %H:%M:%S"
  end_time = time.mktime(time.strptime("03/06/2011 20:00:00", time_string)) * 1000

  redaction_string = json.dumps({
          'version': 1,
          'rules': [
            {
              'description': 'Party like it\'s 1999',
              'search': '2011',
              'replace': '1999',
            },
            {
              'description': 'Blame',
              'search': '[Nn]ame',
              'replace': 'blame',
            },
            {
              'description': 'Yell',
              'search': 'INFO',
              'replace': 'HEY',
            },
            {
              'description': 'No servers',
              'search': 'Server.*',
              'replace': 'Server ---',
            }
          ]
      })

  cmf.redaction.REDACTOR.set_policy_from_string(redaction_string)

  gzip_fd = generator_to_fd(lr.retrieve_log_compressed(end_time, log_path, log_type, byte_limit, cmf.redaction.REDACTOR))

  assert_true(len(gzip_fd.getvalue()) > 0)

  # Unzip and check size
  unzipper = gzip.GzipFile(mode="r", fileobj=gzip_fd)

  unzipped_content = unzipper.read()

  redacted_fd = open(log_path_redacted, 'r')
  try:
    redacted_content = redacted_fd.read()
    assert_true(unzipped_content == redacted_content)
  finally:
    redacted_fd.close()

def test_retrieve_log_compressed_with_start_time():
  """
  tests getting parts of plaintext log file between a time range.
  """
  log_path = LOGS_LIST[12]
  log_type = TYPES_LIST[12]
  byte_limit = 100000000
  time_string = "%d/%m/%Y %H:%M:%S"
  start_time = time.mktime(time.strptime("07/06/2011 11:56:00", time_string)) * 1000
  end_time = time.mktime(time.strptime("07/06/2011 15:16:00", time_string)) * 1000
  gzip_fd = generator_to_fd(lr.retrieve_log_compressed(end_time, log_path, log_type, byte_limit,
                                                       cmf.redaction.NOP_REDACTOR, 3000, start_time))
  assert_true(len(gzip_fd.getvalue()) > 0)
  unzipper = gzip.GzipFile(mode="r", fileobj=gzip_fd)
  unzipped_content = unzipper.read()

  unzipped_content_val = len(unzipped_content)
  # The total bytes of this content is 8768, which is a difference between end_offset and
  # start_offset [i.e. 9417 - 649 = 8768]
  assert_equal(unzipped_content_val, 8768)

def test_retreive_partial_log_event():
  """
  tests getting parts of plaintext log files.
  """
  log_path = LOGS_LIST[2]
  log_type = TYPES_LIST[1]
  byte_limit = 175
  # This file ends a bit after 2011-06-03 16:00
  time_string = "%d/%m/%Y %H:%M:%S"
  end_time = time.mktime(time.strptime("03/06/2011 15:54:11", time_string)) * 1000 + 242
  gzip_fd = generator_to_fd(lr.retrieve_log_compressed(end_time, log_path, log_type, byte_limit, cmf.redaction.NOP_REDACTOR))

  assert_true(len(gzip_fd.getvalue()) > 0)

  # Unzip and check size
  unzipper = gzip.GzipFile(mode="r", fileobj=gzip_fd)

  unzipped_content = unzipper.read()
  print unzipped_content

  assert_equal(len(unzipped_content), 175)

def test_retrieve_log_compressed_for_large_logs():
  """

  Tests getting parts of a large log4j file.
  Set environment variable RANDOM_SEED to use a specific seed for random file
    generation. Leave it unset to use a random seed.
  Set environment variable TEST_LOG_FILE to use a specific log file. Leave it
    unset to run the test against a randomly generated log file.
  The test is very slow and is therefore normally skipped on dev computers and
    runs only within the Jenkins job. It looks at the presence of the environment
    variable JENKINS_URL to identify if it is running on Jenkins.
    To run it locally, set at least one of the three environment variables;
      JENKINS_URL to a dummy value, RANDOM_SEED or TEST_LOG_FILE

  """
  run_test = "JENKINS_URL" in os.environ or \
             "RANDOM_SEED" in os.environ or \
             "TEST_LOG_FILE" in os.environ
  if not run_test:
    print """ Skipping test_retrieve_log_compressed_for_large_logs.
      To run the test set one of the three environment variables
      JENKINS_URL to a dummy value, RANDOM_SEED or TEST_LOG_FILE"""
    return

  seed = os.environ.get('RANDOM_SEED')
  if seed is None:
    seed = random.randint(0, sys.maxint)
  print "test_retrieve_log_compressed_for_large_logs: RND seed = %s" % seed
  random.seed(int(seed))

  test_log_file = os.environ.get('TEST_LOG_FILE')
  log_path = None
  try:
    if test_log_file is None:
      # Create a random log4j file with 5 million lines with only 10 log4j event lines
      print "Generating a test log file..."
      log_path, start_millis, end_millis = _generate_random_log4j_file(5000000, 10)
      log_end_time_millis = random.randint(start_millis, end_millis)
    else:
      print "Retrieving the log file:"+test_log_file
      log_path = test_log_file
      log_end_time_millis = time.time()*1000

    log_type = TYPES_LIST[1]
    byte_limit = 10000000000
    timeout = 120000 #120 seconds in millis
    print "looking for logs till %s (%d millis since epoch) in file %s" % (time.strftime("%Y-%m-%d %T", time.localtime(log_end_time_millis/1000)), log_end_time_millis, log_path)

    start_time = time.time()
    gzip_fd = generator_to_fd(lr.retrieve_log_compressed(log_end_time_millis, log_path, log_type, byte_limit, cmf.redaction.NOP_REDACTOR, timeout))
    time_taken = time.time() - start_time
    print "time taken: %fs" % time_taken
    assert_true(time_taken*1000 <= timeout + 2000, "Log retrieval timed out: To debug, set the environment variable RANDOM_SEED=%s and rerun the test"%seed)
    assert_true(len(gzip_fd.getvalue()) > 0, "Log retrieval failed: To debug, set the environment variable RANDOM_SEED=%s and rerun the test"%seed)

  finally:
    if test_log_file is None:
      os.remove(log_path)


def test_retrieve_log_compressed_for_logs_non_ascii():
  """

  Tests getting parts of a log4j file with non-ascii chars.
  Set environment variable RANDOM_SEED to use a specific seed for random file
    generation. Leave it unset to use a random seed.
  Set environment variable TEST_LOG_FILE to use a specific log file. Leave it
    unset to run the test against a randomly generated log file.

  """
  seed = os.environ.get('RANDOM_SEED')
  if seed is None:
    seed = random.randint(0, sys.maxint)
  print "test_retrieve_log_compressed_for_large_logs: RND seed = %s" % seed
  random.seed(int(seed))

  test_log_file = os.environ.get('TEST_LOG_FILE_ASC')
  log_path = None
  try:
    if test_log_file is None:
      # Create a random log4j file with 10 log4j event lines
      print "Generating a test log file..."
      log_path, start_millis, end_millis = _generate_random_log4j_file(50, 10, True)
      log_end_time_millis = random.randint(start_millis, end_millis)
    else:
      print "Retrieving the log file:"+test_log_file
      log_path = test_log_file
      log_end_time_millis = time.time()*1000

    log_type = TYPES_LIST[1]
    byte_limit = 10000000000
    timeout = 120000 #120 seconds in millis
    print "looking for logs till %s (%d millis since epoch) in file %s" % (time.strftime("%Y-%m-%d %T", time.localtime(log_end_time_millis/1000)), log_end_time_millis, log_path)

    start_time = time.time()
    gzip_fd = generator_to_fd(lr.retrieve_log_compressed(log_end_time_millis, log_path, log_type, byte_limit, cmf.redaction.NOP_REDACTOR, timeout))
    time_taken = time.time() - start_time
    print "time taken: %fs" % time_taken
    assert_true(time_taken*1000 <= timeout + 2000, "Log retrieval timed out: To debug, set the environment variable RANDOM_SEED=%s and rerun the test"%seed)
    assert_true(len(gzip_fd.getvalue()) > 0, "Log retrieval failed: To debug, set the environment variable RANDOM_SEED=%s and rerun the test"%seed)

  finally:
    print "test completed"

def test_retrieve_log_compressed_timeout():
  """

  Tests timeout in retrieve_log_compressed.

  """
  # From an earlier experiment, we have found that a rnd seed of 2545211759945397115
  # generates a file that takes over 2 seconds to search.
  random.seed(2545211759945397115)
  timeout = 500 #500 millis

  log_path = None
  try:
    # Create a random log4j file with 1 million lines with only 10 log4j event lines
    print "Generating a test log file..."
    log_path, start_time, end_time = _generate_random_log4j_file(1000000, 10)
    end_time_millis = random.randint(int(start_time*1000), int(end_time*1000))

    log_type = TYPES_LIST[1]
    byte_limit = 100000000
    print "looking for logs till %s (%d millis since epoch) in file %s" % (time.strftime("%Y-%m-%d %T", time.localtime(end_time_millis/1000)), end_time_millis, log_path)

    start_time = time.time()
    generator_to_fd(lr.retrieve_log_compressed(end_time_millis, log_path, log_type, byte_limit, cmf.redaction.NOP_REDACTOR, timeout))
    time_taken = time.time() - start_time
    print "time taken: %fs" % time_taken
    assert_true(time_taken*1000 <= timeout + 200, "Log retrieval timeout failed")

  finally:
    if log_path is not None:
      os.remove(log_path)

def _generate_random_log4j_file(total_lines, num_events, nonasc=False):
  """
  Generates a random log4j file containing total_lines of random size
  and randomly placed log4j event lines
  """
  log_file = tempfile.NamedTemporaryFile()
  log_file_name = log_file.name
  # Close the temp file and open it as a regular persitent file.
  log_file.close()
  log_file = open(log_file_name, "w")

  avg_timestamp_jump = 5
  sub_second = random.randint(100, 999)
  event_time = time.localtime(random.randint(1418980313, 1438980313))
  end_millis = start_millis = time.mktime(event_time)*1000 + sub_second
  line = "%s,%d INFO org.apache.hive.service.AbstractService: Service:HiveServer2 is inited.\n" % (time.strftime("%Y-%m-%d %T", event_time),sub_second)
  try:
    log_file.write(line)
    for _ in range(total_lines - 1):
      if random.randint(1, int(total_lines/num_events)) == 1:
        event_time = time.localtime(end_millis/1000 + random.random()*2*avg_timestamp_jump)
        sub_second = random.randint(100, 999)
        line = "%s,%d INFO org.apache.hive.service.AbstractService: Service:HiveServer2 is inited.\n" % (time.strftime("%Y-%m-%d %T", event_time),sub_second)
      else:
        line = ''.join(random.choice(string.ascii_letters) for x in range(int(100*random.random())))+"\n"
        end_millis = time.mktime(event_time)*1000 + sub_second
      log_file.write(line)
      if (nonasc == True):
        log_file.write("Dejan Živković Dejan Živković")
  finally:
    log_file.close()
  return log_file.name, start_millis, end_millis

def test_retrieve_rolled_over():
  """
  tests that retrieve logs will handle rolled over log files.
  """
  log_path = os.path.join(TEST_LOGS_DIR, "rolled-over", "cloudera-scm-agent.log")
  log_type = "SCMAGENT"
  byte_limit = 20000
  # The most recent file ends a bit after 2012-09-17 11:30
  time_string = "%d/%m/%Y %H:%M:%S"
  end_time = time.mktime(time.strptime("17/09/2012 20:00:00", time_string)) * 1000
  gzip_fd = generator_to_fd(lr.retrieve_log_compressed(end_time, log_path, log_type, byte_limit, cmf.redaction.NOP_REDACTOR))

  # Unzip and check size
  unzipper = gzip.GzipFile(mode="r", fileobj=gzip_fd)
  unzipped_content = unzipper.read()
  assert_equal(len(unzipped_content), 2894*4)
  # Make sure that the right file went last (.11, not .2)
  assert_true(unzipped_content.startswith("[14/Sep/2012"))

  # test that rolled over logs won't be retrieved if an individual is requested
  log_path = os.path.join(TEST_LOGS_DIR, "rolled-over", "cloudera-scm-agent.log.2")

  gzip_fd = generator_to_fd(lr.retrieve_log_compressed(end_time, log_path, log_type, byte_limit, cmf.redaction.NOP_REDACTOR))

  # Unzip and check size
  unzipper = gzip.GzipFile(mode="r", fileobj=gzip_fd)
  unzipped_content = unzipper.read()
  assert_equal(len(unzipped_content), 2894)

def test_retrieve_timeout():
  """
  tests that log retrieve timeout works
  """
  # Take the whole of the largest file
  gzip_fd = generator_to_fd(lr.retrieve_log_compressed(1353950619000, LOGS_LIST[2], TYPES_LIST[2], 50000000, cmf.redaction.NOP_REDACTOR, 1))

  unzipper = gzip.GzipFile(mode="r", fileobj=gzip_fd)

  unzipped_content = unzipper.read()

  assert_true(len(unzipped_content) < 87870) # total file size

def generator_to_fd(gen):
  """Convert a generator to a fileobj by feeding it into StringIO"""
  result = StringIO()
  for buf in gen:
    result.write(buf)
  result.seek(0)
  return result
