#!/usr/bin/env python
# Copyright (c) 2011 Cloudera, Inc. All rights reserved.
import tempfile
import time
import os
import logging
import simplejson as json
import datetime
import gzip
import zipfile
import StringIO
import tempfile

from nose.tools import assert_true, assert_equal, assert_false, assert_raises
from nose.plugins.skip import SkipTest
from clusterstats.log.streaming.event_filter import LogLevelFilter, HueLogLevelFilter

import cmf.log_search as ls

TEST_LOGS_DIR = os.path.join(os.path.dirname(__file__), "test-logs")
TEST_LOGS = ["hadoop-cmf-hdfs1-DATANODE-4-dtrejovm2.log.out",
            "hadoop-cmf-hdfs1-DATANODE-6-dtrejovm2.log.out",
            "hadoop-cmf-hdfs1-DATANODE-6-localhost.log.out",
            "hadoop-cmf-hdfs1-NAMENODE-1-dtrejovm2.log.out",
            "hadoop-cmf-hdfs1-NAMENODE-8-dtrejovm2.log.out",
            "hadoop-cmf-hdfs1-NAMENODE-8-localhost.log.out",
            "hadoop-cmf-hdfs1-SECONDARYNAMENODE-3-dtrejovm2.log.out",
            "hadoop-cmf-hdfs1-SECONDARYNAMENODE-5-dtrejovm2.log.out",
            "hadoop-cmf-hdfs1-SECONDARYNAMENODE-5-localhost.log.out",
            "hadoop-hadoop-datanode-localhost.localdomain.log",
            "hadoop-hadoop-namenode-localhost.localdomain.log",
            "hadoop-hadoop-tasktracker-localhost.localdomain.log",
            "impalad.INFO"
            ]

TYPES_LIST = ["LOG4J"] * (len(TEST_LOGS) - 1) + ["GLOG"]

ROLE_IDS_LIST = [x for x in range(0, len(TYPES_LIST))]

HUE_LOG_NAME = "runcpserver.log"

LOGS_LIST = [os.path.join(TEST_LOGS_DIR, log) for log in TEST_LOGS]

HUE_LOG = os.path.join(TEST_LOGS_DIR, HUE_LOG_NAME)

AGENT_LOG_NAME = "cloudera-scm-agent.log"

AGENT_LOG = os.path.join(TEST_LOGS_DIR, AGENT_LOG_NAME)

def test_questionable_impala_log():
  """
  This log file has a line like:
  "-INFO0925 13:59:06.818000 main org.ap..."
  Note the dash at the beginning.
  """
  # Impala logs don't have years, so we simply make the
  # date range very wide.
  end_time=convert_timestamp_to_pdt(2387900694000) # date in 2045
  datajson = ls.search_logs(
    host="localhost",
    port=9001,
    log_paths=[ os.path.join(TEST_LOGS_DIR, "impalad.INFO") ],
    log_types=["GLOG"],
    role_ids=[0],
    log_level="INFO",
    role_result_limit=200,
    start_time=convert_timestamp_to_pdt(1304262000000), # "2011-05-01T08:00:00-07:00"
    end_time=end_time)["data"]
  # What's going on is that there are non-Impala logs spliced in here, which is why
  # we avoid the FileWatch line.
  # cat impalad.INFO | grep '^[A-Z]' | grep 0925 | grep -v FileWatch | wc -l
  data = json.loads(datajson)
  assert_equal(59, len(data["events"]))

def test_log_search():
  """
  test unfiltered search on test-logs.
  """

  data = ls.search_logs(
    host="localhost",
    port=9001,
    log_paths=LOGS_LIST,
    log_types=TYPES_LIST,
    role_ids=ROLE_IDS_LIST,
    role_result_limit=200,
    start_time=convert_timestamp_to_pdt(1304262000000), # "2011-05-01T08:00:00-07:00"
    end_time=convert_timestamp_to_pdt(1309791600000))["data"] # "2011-07-04T08:00:00-07:00"

  data = json.loads(data)

  # verified with grep(tm)
  assert_equal(len(data["events"]), 144)

def test_log_search_plaintext():
  """
  test unfiltered search on test-logs, with plaintext output.
  """

  data = ls.search_logs(
    host="localhost",
    port=9001,
    log_paths=LOGS_LIST,
    log_types=TYPES_LIST,
    role_ids=ROLE_IDS_LIST,
    role_result_limit=200,
    start_time=convert_timestamp_to_pdt(1304262000000), # "2011-05-01T08:00:00-07:00"
    end_time=convert_timestamp_to_pdt(1309791600000), # "2011-07-04T08:00:00-07:00"
    plaintext=True)

  assert_equal(len(data["data"].replace("\\n", "\n")), 58567)
  assert_true(data["data"].startswith("2011-06-03 16:33:14,294 WARN org.apache.hadoop.util.PluginDispatcher"))

def test_search_special_chars():
  """
  test search that includes space and special characters.
  """

  data = ls.search_logs(
    host="localhost",
    port=9001,
    log_paths=LOGS_LIST,
    log_types=TYPES_LIST,
    role_ids=ROLE_IDS_LIST,
    role_result_limit=200,
    log_level="INFO",
    query="&",
    start_time=convert_timestamp_to_pdt(1304262000000), # "2011-05-01T08:00:00-07:00"
    end_time=convert_timestamp_to_pdt(1309791600000))["data"] # "2011-07-04T08:00:00-07:00"

  data = json.loads(data)

  # verified with grep(tm)
  assert_equal(len(data["events"]), 51)

  data = ls.search_logs(
    host="localhost",
    port=9001,
    log_paths=LOGS_LIST,
    log_types=TYPES_LIST,
    role_ids=ROLE_IDS_LIST,
    role_result_limit=200,
    log_level="INFO",
    query="Server Handler",
    start_time=convert_timestamp_to_pdt(1304262000000), # "2011-05-01T08:00:00-07:00"
    end_time=convert_timestamp_to_pdt(1309791600000))["data"] # "2011-07-04T08:00:00-07:00"

  data = json.loads(data)

  # verified with grep(tm)
  assert_equal(len(data["events"]), 151)

def test_compressed_log_search():
  """
  test unfiltered search on logs with compression enabled.
  """
  path = ls.search_logs_compressed(
    host="localhost",
    port=9001,
    log_paths=LOGS_LIST,
    log_types=TYPES_LIST,
    role_ids=ROLE_IDS_LIST,
    role_result_limit=200,
    start_time=convert_timestamp_to_pdt(1304262000000), # "2011-05-01T08:00:00-07:00"
    end_time=convert_timestamp_to_pdt(1309791600000))["path"] # "2011-07-04T08:00:00-07:00"

  search_data, events = uncompress_search_results(path)

  assert_equal(len(events), 144)
  assert_equal(search_data["success"], True)
  assert_equal(search_data["host"], "localhost")
  assert_equal(search_data["port"], 9001)

def test_compressed_vs_uncompressed_size():
  """
  test that, given the same number of events, the compressed results will have a smaller data size.
  """
  data = ls.search_logs(
    host="localhost",
    port=9001,
    log_paths=LOGS_LIST,
    log_types=TYPES_LIST,
    role_ids=ROLE_IDS_LIST,
    role_result_limit=200,
    start_time=convert_timestamp_to_pdt(1304262000000), # "2011-05-01T08:00:00-07:00"
    end_time=convert_timestamp_to_pdt(1309791600000))["data"] # "2011-07-04T08:00:00-07:00"
  uncompressed_size = len(data)

  path = ls.search_logs_compressed(
    host="localhost",
    port=9001,
    log_paths=LOGS_LIST,
    log_types=TYPES_LIST,
    role_ids=ROLE_IDS_LIST,
    role_result_limit=200,
    start_time=convert_timestamp_to_pdt(1304262000000), # "2011-05-01T08:00:00-07:00"
    end_time=convert_timestamp_to_pdt(1309791600000))["path"] # "2011-07-04T08:00:00-07:00"
  compressed_size = os.stat(path).st_size

  assert_true(compressed_size < uncompressed_size)

def test_compressed_vs_uncompressed_byte_limit():
  """
  test that, given the same byte limit, search_logs_uncompressed will return more events.
  """
  data = ls.search_logs(
    host="localhost",
    port=9001,
    log_paths=LOGS_LIST,
    log_types=TYPES_LIST,
    role_ids=ROLE_IDS_LIST,
    role_result_limit=10,
    total_byte_limit=5000,
    start_time=convert_timestamp_to_pdt(1304262000000), # "2011-05-01T08:00:00-07:00"
    end_time=convert_timestamp_to_pdt(1309791600000))["data"] # "2011-07-04T08:00:00-07:00"
  events_from_uncompressed = json.loads(data)["events"]

  path = ls.search_logs_compressed(
    host="localhost",
    port=9001,
    log_paths=LOGS_LIST,
    log_types=TYPES_LIST,
    role_ids=ROLE_IDS_LIST,
    role_result_limit=10,
    total_byte_limit=5000,
    start_time=convert_timestamp_to_pdt(1304262000000), # "2011-05-01T08:00:00-07:00"
    end_time=convert_timestamp_to_pdt(1309791600000))["path"] # "2011-07-04T08:00:00-07:00"
  search_data, events_from_compressed = uncompress_search_results(path)

  assert_true(len(events_from_compressed) > len(events_from_uncompressed))

def uncompress_search_results(path):
  fd = open(path)
  try:
    zip_fd = zipfile.ZipFile(fd, "r")

    events_bytes = StringIO.StringIO(zip_fd.read("log_events.json.gz"))
    events_fd = gzip.GzipFile(mode="r", fileobj=events_bytes)

    events = json.load(events_fd)
    search_data = json.loads(zip_fd.read("meta.json"))
    return (search_data, events)
  finally:
    fd.close()

def test_data_limits():
  """
  test that the role_result_limit and byte_limit parameters work correctly.
  """

  # Test role_result_limit only.
  data = ls.search_logs(
    host="localhost",
    port=9001,
    log_paths=LOGS_LIST,
    log_types=TYPES_LIST,
    role_ids=ROLE_IDS_LIST,
    role_result_limit=10,
    start_time=convert_timestamp_to_pdt(1304262000000), # "2011-05-01T08:00:00-07:00"
    end_time=convert_timestamp_to_pdt(1309791600000))["data"] # "2011-07-04T08:00:00-07:00"
  data = json.loads(data)
  assert_equal(len(data["events"]), 56)

  # Test total_byte_limit_only.
  data = ls.search_logs(
    host="localhost",
    port=9001,
    log_paths=LOGS_LIST,
    log_types=TYPES_LIST,
    role_ids=ROLE_IDS_LIST,
    total_byte_limit=5000,
    start_time=convert_timestamp_to_pdt(1304262000000), # "2011-05-01T08:00:00-07:00"
    end_time=convert_timestamp_to_pdt(1309791600000))["data"] # "2011-07-04T08:00:00-07:00"
  data = json.loads(data)
  event_sizes = [ls.get_event_dict_size(event_dict) for event_dict in data["events"]]
  assert_true(sum(event_sizes) > 5000)
  assert_true(sum(event_sizes) - event_sizes[-1] <= 5000)

  # We can't rely on exactly the same number of events being generated across
  # runs on different machines, because the path is included in the events.
  # The byte_limit is being honored, which is good enough. We do need to know how
  # many events this generated to make sure the byte limit sets in earlier than
  # the role result limit in the next test.
  num_byte_limit_events = len(data["events"])

  # Test that the byte_limit is honored, since it sets in first.
  data = ls.search_logs(
    host="localhost",
    port=9001,
    log_paths=LOGS_LIST,
    log_types=TYPES_LIST,
    role_ids=ROLE_IDS_LIST,
    role_result_limit=10,
    total_byte_limit=5000,
    start_time=convert_timestamp_to_pdt(1304262000000), # "2011-05-01T08:00:00-07:00"
    end_time=convert_timestamp_to_pdt(1309791600000))["data"] # "2011-07-04T08:00:00-07:00"
  data = json.loads(data)
  assert_equal(len(data["events"]), num_byte_limit_events)

  # Test that the role_result_limit is honored, since it sets in first.
  data = ls.search_logs(
    host="localhost",
    port=9001,
    log_paths=LOGS_LIST,
    log_types=TYPES_LIST,
    role_ids=ROLE_IDS_LIST,
    role_result_limit=1, # 12 roles, 1 result each, so 12 results
    total_byte_limit=5000,
    start_time=convert_timestamp_to_pdt(1304262000000), # "2011-05-01T08:00:00-07:00"
    end_time=convert_timestamp_to_pdt(1309791600000))["data"] # "2011-07-04T08:00:00-07:00"
  data = json.loads(data)
  assert_equal(len(data["events"]), 12)

def test_response_time_gt_zero():
  ls.ENABLE_COST_INFO = True
  """make sure it takes some time to execute the query"""
  data = ls.search_logs(
    host="localhost",
    port=9001,
    log_paths=LOGS_LIST,
    log_types=TYPES_LIST,
    role_ids=ROLE_IDS_LIST,
    role_result_limit=100,
    start_time=convert_timestamp_to_pdt(1304262000000), # "2011-05-01T08:00:00-07:00"
    end_time=convert_timestamp_to_pdt(1309791600000))["data"] # "2011-07-04T08:00:00-07:00"
  data = json.loads(data)
  assert_true(data["cost"]["time"] > 0, "took <= 0 milliseconds. not OK!")

def test_disable_cost_info():
  ls.ENABLE_COST_INFO = False
  """make sure it takes some time to execute the query"""
  data = ls.search_logs(
    host="localhost",
    port=9001,
    log_paths=LOGS_LIST,
    log_types=TYPES_LIST,
    role_ids=ROLE_IDS_LIST,
    role_result_limit=100,
    start_time=convert_timestamp_to_pdt(1304262000000), # "2011-05-01T08:00:00-07:00"
    end_time=convert_timestamp_to_pdt(1309791600000))["data"] # "2011-07-04T08:00:00-07:00"
  data = json.loads(data)
  assert_false("key" in data, "Cost key does not exist")

def test_time_interval_filter():
  """test time interval filter"""
  data = ls.search_logs(
    host="localhost",
    port=9001,
    log_paths=LOGS_LIST,
    log_types=TYPES_LIST,
    role_ids=ROLE_IDS_LIST,
    role_result_limit=100,
    start_time=convert_timestamp_to_pdt(1307548800000), # "2011-06-08T09:00:00-07:00"
    end_time=convert_timestamp_to_pdt(1307552400000))["data"] # 2011-06-8T10:00:00-07:00"
  data = json.loads(data)
  assert_equal(len(data["events"]), 2)

def test_regex_filter_exact_matching():
  """test queries which rely on regex to filter results"""
  data = ls.search_logs(
    host="localhost",
    port=9001,
    log_paths=LOGS_LIST,
    log_types=TYPES_LIST,
    role_ids=ROLE_IDS_LIST,
    role_result_limit=100,
    start_time=convert_timestamp_to_pdt(1304262000000), # "2011-05-01T08:00:00-07:00"
    end_time=convert_timestamp_to_pdt(1309791600000), # "2011-07-04T08:00:00-07:00"
    query="ExitCodeException")["data"]
  data = json.loads(data)
  assert_equal(len(data["events"]), 6)

def test_regex_filter_case_insensitive_matching():
  """test queries which rely on regex to filter results"""
  data = ls.search_logs(
    host="localhost",
    port=9001,
    log_paths=LOGS_LIST,
    log_types=TYPES_LIST,
    role_ids=ROLE_IDS_LIST,
    role_result_limit=100,
    start_time=convert_timestamp_to_pdt(1304262000000), # "2011-05-01T08:00:00-07:00"
    end_time=convert_timestamp_to_pdt(1309791600000), # "2011-07-04T08:00:00-07:00"
    query="ExItCoDeExCePtIon")["data"]
  data = json.loads(data)
  assert_equal(len(data["events"]), 6)

def test_regex_filter_on_source():
  """
  test queries which rely on regex to filter results, especially on the
  `source` property of the event
  """
  data = ls.search_logs(
    host="localhost",
    port=9001,
    log_paths=LOGS_LIST,
    log_types=TYPES_LIST,
    role_ids=ROLE_IDS_LIST,
    role_result_limit=100,
    start_time=convert_timestamp_to_pdt(1304262000000), # "2011-05-01T08:00:00-07:00"
    end_time=convert_timestamp_to_pdt(1309791600000), # "2011-07-04T08:00:00-07:00"
    query="org.apache.hadoop.hdfs.server.datanode.DataNode")["data"]
  data = json.loads(data)
  assert_equal(len(data["events"]), 20)

def test_regex_escaping():
  """
  test queries which use escaped regex
  """
  EXPECTED_MESSAGE = "org.apache.hadoop.util.Shell$ExitCodeException:"

  # this query has special chars individaully escaped, so it should match the message
  data = ls.search_logs(
    host="localhost",
    port=9001,
    log_paths=LOGS_LIST,
    log_types=TYPES_LIST,
    role_ids=ROLE_IDS_LIST,
    log_level="INFO",
    role_result_limit=100,
    start_time=convert_timestamp_to_pdt(1307516400000), # "2011-06-08T00:00:00-07:00"
    end_time=convert_timestamp_to_pdt(1307599200000), # "2011-06-08T23:00:00-07:00"
    query=r"org.apache.hadoop.util.Shell\$ShellCommandExecutor.execute\(Shell.java:375\)")["data"]
  data = json.loads(data)

  assert_equal(len(data["events"]), 1)
  assert_equal(data["events"][0]["source"], "org.apache.hadoop.net.ScriptBasedMapping")
  assert_equal(data["events"][0]["loglevel"], "WARN")
  assert_equal(data["events"][0]["offset"], 572)
  assert_equal(data["events"][0]["message"][:len(EXPECTED_MESSAGE)], EXPECTED_MESSAGE)

  # this query is escaped by \Q<query>\E , so it should match the message
  data = ls.search_logs(
    host="localhost",
    port=9001,
    log_paths=LOGS_LIST,
    log_types=TYPES_LIST,
    role_ids=ROLE_IDS_LIST,
    log_level="INFO",
    role_result_limit=100,
    start_time=convert_timestamp_to_pdt(1307516400000), # "2011-06-08T00:00:00-07:00"
    end_time=convert_timestamp_to_pdt(1307599200000), # "2011-06-08T23:00:00-07:00"
    query=r"\Qorg.apache.hadoop.util.Shell$ShellCommandExecutor.execute(Shell.java:375)\E")["data"]
  data = json.loads(data)

  assert_equal(len(data["events"]), 1)
  assert_equal(data["events"][0]["source"], "org.apache.hadoop.net.ScriptBasedMapping")
  assert_equal(data["events"][0]["loglevel"], "WARN")
  assert_equal(data["events"][0]["offset"], 572)
  assert_equal(data["events"][0]["message"][:len(EXPECTED_MESSAGE)], EXPECTED_MESSAGE)

  # searching for "\Q\E" should return the same results that we get form not
  # providing a query
  data = ls.search_logs(
    host="localhost",
    port=9001,
    log_paths=LOGS_LIST,
    log_types=TYPES_LIST,
    role_ids=ROLE_IDS_LIST,
    role_result_limit=100,
    query=r"\Q\E",
    start_time=convert_timestamp_to_pdt(1307548800000), # "2011-06-08T09:00:00-07:00"
    end_time=convert_timestamp_to_pdt(1307552400000))["data"] # 2011-06-8T10:00:00-07:00"
  data = json.loads(data)
  assert_equal(len(data["events"]), 2)

  # this query shouldn't match anything as special chars are not escaped
  data = ls.search_logs(
    host="localhost",
    port=9001,
    log_paths=LOGS_LIST,
    log_types=TYPES_LIST,
    role_ids=ROLE_IDS_LIST,
    log_level="INFO",
    role_result_limit=100,
    start_time=convert_timestamp_to_pdt(1307516400000), # "2011-06-08T00:00:00-07:00"
    end_time=convert_timestamp_to_pdt(1307599200000), # "2011-06-08T23:00:00-07:00"
    query=r"org.apache.hadoop.util.Shell$ShellCommandExecutor.execute(Shell.java:375)")["data"]
  data = json.loads(data)

  assert_equal(len(data["events"]), 0)

def test_search_timeout():
  """
  tests that search timeout works
  """
  data = ls.search_logs(
    host="localhost",
    port=9001,
    log_paths=LOGS_LIST,
    log_types=TYPES_LIST,
    role_ids=ROLE_IDS_LIST,
    search_timeout=1,
    role_result_limit=100,
    start_time=convert_timestamp_to_pdt(1304262000000), # "2011-05-01T08:00:00-07:00"
    end_time=convert_timestamp_to_pdt(1309791600000), # "2011-07-04T08:00:00-07:00"
    query="org.apache.hadoop.hdfs.server.datanode.DataNode")["data"]
  data = json.loads(data)

  assert_equal(len(data["events"]), 0)
  assert_equal(data["messages"][0], ls.TIMEOUT_MESSAGE, "timeout message wasn't set")

def test_datetime_from_millis():
  """
  test parse datetime, test throw when bad format, test ignoring UTC

  """
  dt = ls.datetime_from_millis(convert_timestamp_to_pdt(1304262000000)) # "2011-05-01T08:00:00-07:00"
  correct = datetime.datetime(2011, 5, 1, 8, 0)

  assert_equal(dt, correct, "didn't parse correctly")

  assert_raises(Exception, ls.datetime_from_millis, "0000-00-00T00:00:00-07:00")

  assert_raises(Exception, ls.datetime_from_millis, "2011-05-01T08:00:00-07:00XXX")

  assert_raises(Exception, ls.datetime_from_millis, None)

def test_millis_from_datetime():
  """
  Test converting a datetime to milliseconds
  """
  dt = datetime.datetime(2012, 8, 29, 16, 12, 22, 659123)
  assert_equal(1346281942659, ls.millis_from_datetime(dt))

def test_parse_log_level():
  """test that log levels are in VALID_LOG_LEVEL set"""
  assert_equal(ls.parse_log_level("   WARN "), "WARN")
  assert_raises(Exception, ls.parse_log_level, "LOLZ")


def test_validate_log_dir():
  """make sure it raises on nonexistent dirs"""
  assert_equal(ls.validate_log_dir(TEST_LOGS_DIR), TEST_LOGS_DIR)
  assert_raises(Exception, ls.validate_log_dir, TEST_LOGS_DIR + "/unrealfolder")

def test_log_with_context():
  """
  make sure results contain less than max number of events, correct text in
  first log event.
  """
  data = ls.log_with_context(
    os.path.join(TEST_LOGS_DIR,
      "hadoop-cmf-hdfs1-DATANODE-4-dtrejovm2.log.out.2011-06-03"),
    0,
    None,
    None,
    400,
    "LOG4J",
    host="localhost",
    port="9001")["data"]
  data = json.loads(data)
  assert_true(len(data["events"]) <= 400)
  assert_equal(data["events"][0]["message"].index("STARTUP_MSG: \n/****"), 0)

def test_log_with_context2():
  """make sure it can read from middle of file and give context."""
  data = ls.log_with_context(
    os.path.join(TEST_LOGS_DIR,
      "hadoop-cmf-hdfs1-DATANODE-6-localhost.log.out"),
    4000,
    None,
    None,
    400,
    "LOG4J",
    host="localhost",
    port="9001")["data"]
  data = json.loads(data)
  assert_true(len(data["events"]) <= 400)
  assert_equal(data["events"][0]["message"].index("org.apache.hadoop.ipc"), 0)

def test_log_with_context_prev():
  """make sure it can read from middle of file and give context."""
  data = ls.log_with_context(
    os.path.join(TEST_LOGS_DIR,
      "hadoop-cmf-hdfs1-DATANODE-6-localhost.log.out"),
    71045,
    None,
    "previous",
    2,
    "LOG4J",
    host="localhost",
    port="9001")["data"]
  data = json.loads(data)
  assert_true(len(data["events"]) == 2)
  assert_equal(data["events"][0]["offset"], 70986)
  assert_equal(data["events"][0]["message"], "jetty-6.1.26\n")
  assert_equal(data["events"][1]["offset"], 70896)
  assert_equal(data["events"][1]["message"], "Jetty bound to port 50075\n")

def test_log_with_context_next():
  """make sure it can read from middle of file and give context."""
  data = ls.log_with_context(
    os.path.join(TEST_LOGS_DIR,
      "hadoop-cmf-hdfs1-DATANODE-6-localhost.log.out"),
    71045,
    None,
    "next",
    2,
    "LOG4J",
    host="localhost",
    port="9001")["data"]
  data = json.loads(data)
  assert_true(len(data["events"]) == 2)
  assert_equal(data["events"][0]["offset"], 71045)
  assert_equal(data["events"][0]["message"], "Started SelectChannelConnector@localhost:50075\n")
  assert_equal(data["events"][1]["offset"], 71138)
  assert_equal(data["events"][1]["message"], "Initializing JVM Metrics with processName=DataNode, sessionId=null\n")

def test_log_with_context_hue():
  """Test that log context view works with hue logs"""
  data = ls.log_with_context(
    HUE_LOG,
    465,
    None,
    None,
    400,
    "HUE",
    host="localhost",
    port="9001")["data"]
  data = json.loads(data)
  assert_equal(len(data["events"]), 17)
  assert_equal(data["events"][6]["offset"], 465)
  assert_equal(data["events"][6]["message"], "(2488) *** Controller starting at Wed Aug 24 17:21:02 2011\n")

def test_log_with_context_timestamp():
  """Test that log context view works with timestamps instead of offsets"""
  data = ls.log_with_context(
    os.path.join(TEST_LOGS_DIR,
      "hadoop-cmf-hdfs1-DATANODE-6-localhost.log.out"),
    None,
    1307141924512,
    None,
    1,
    "LOG4J",
    host="localhost",
    port="9001")["data"]
  data=json.loads(data)

  assert_equal(len(data['events']), 1)
  assert_equal(data['pivotOffset'], 40650)
  assert_equal(data['events'][0]['offset'], 40650)
  assert_equal(data['events'][0]['message'], "Initializing RPC Metrics with hostName=DataNode, port=50020\n")

def test_hue_log_search():
  data = ls.search_logs(
    host="localhost",
    port=9001,
    log_paths=[HUE_LOG],
    log_types=["HUE"],
    role_ids=[1337],
    log_level="INFO",
    role_result_limit=100,
    start_time=1314231662000, # "2011-08-24T17:21:02-07:00"
    end_time=1314231663000)["data"] # "2011-08-24T17:21:03-07:00"
  data = json.loads(data)

  assert_equal(len(data["events"]), 1)
  assert_equal(data["events"][0]["source"], "spawning_controller")
  assert_equal(data["events"][0]["loglevel"], "INFO")
  assert_equal(data["events"][0]["roleId"], 1337)
  assert_equal(data["events"][0]["offset"], 465)
  assert_equal(data["events"][0]["message"], "(2488) *** Controller starting at Wed Aug 24 17:21:02 2011\n")

def test_log_with_context_scm_agent():
  """Test that log context view works with agent logs"""
  data = ls.log_with_context(
    AGENT_LOG,
    465,
    None,
    None,
    400,
    "SCMAGENT",
    host="localhost",
    port="9001")["data"]
  data = json.loads(data)

  assert_equal(len(data["events"]), 19)
  assert_equal(data["events"][8]["offset"], 977)
  assert_equal(data["events"][8]["message"], "10.20.193.105 - - GET\n")

def test_scm_agent_log_search():
  """Test the log search works for scm agent"""
  data = ls.search_logs(
    host="localhost",
    port=9001,
    log_paths=[AGENT_LOG],
    log_types=["SCMAGENT"],
    role_ids=[-1],
    log_level="INFO",
    role_result_limit=100,
    start_time=1347905688000, # "17/Sep/2012 11:14:48 +0000"
    end_time=1347906305000)["data"] # "17/Sep/2012 11:23:25 +0000"
  data = json.loads(data)

  assert_equal(len(data["events"]), 15)
  assert_equal(data["events"][8]["source"], "10665 CP Server Thread-12 _cplogging")
  assert_equal(data["events"][8]["loglevel"], "INFO")
  assert_equal(data["events"][8]["roleId"], -1)
  assert_equal(data["events"][8]["offset"], 977)
  assert_equal(data["events"][8]["message"], "10.20.193.105 - - GET\n")


def test_log_level_regex():
  level_index = LogLevelFilter.levels.index("WARN")
  regex_string = ls.make_log_level_regex_string(level_index, LogLevelFilter)
  assert_equal(regex_string, "WARN|ERROR|FATAL")

  level_index = LogLevelFilter.levels.index("FATAL")
  regex_string = ls.make_log_level_regex_string(level_index, LogLevelFilter)
  assert_equal(regex_string, "FATAL")

  level_index = HueLogLevelFilter.levels.index("WARNING")
  regex_string = ls.make_log_level_regex_string(level_index, HueLogLevelFilter)
  assert_equal(regex_string, "WARNING|ERROR|CRITICAL")

  level_index = HueLogLevelFilter.levels.index("CRITICAL")
  regex_string = ls.make_log_level_regex_string(level_index, HueLogLevelFilter)
  assert_equal(regex_string, "CRITICAL")

def test_dict2Utf():
  """Test that non-utf data can be filtered out of logs."""
  d = { 1: "Here is a non-utf character \x86. Watch out for it" }
  u = ls.dict2Utf(d)
  assert_true("\x86" in d[1])
  assert_equal(u[1], u"Here is a non-utf character \ufffd. Watch out for it")

def test_non_utf_logs():
  """Test that non-utf data can be returned by agent."""
  data = ls.log_with_context(
    os.path.join(TEST_LOGS_DIR, "utf",
      "non-utf.log.out"),
    0,
    None,
    None,
    400,
    "LOG4J",
    host="localhost",
    port="9001")["data"]
  data = json.loads(data)
  assert_equal(len(data["events"]), 3)
  assert_equal(data["events"][0]["message"], u"Message before binary data\n")
  assert_equal(data["events"][1]["message"], u"Invalid utf-8 character \ufffd\n")
  assert_equal(data["events"][2]["message"], u"Message after binary data\n\n")

def test_utf8_logs():
  """Test that utf8 data can be returned by agent."""
  data = ls.log_with_context(
    os.path.join(TEST_LOGS_DIR, "utf",
      "utf.log.out"),
    0,
    None,
    None,
    400,
    "LOG4J",
    host="localhost",
    port="9001")["data"]
  data = json.loads(data)
  assert_equal(len(data["events"]), 3)
  assert_equal(data["events"][0]["message"], u"Message before utf data\n")
  assert_equal(data["events"][1]["message"], u"Valid utf-8 character \u79c1\n")
  assert_equal(data["events"][2]["message"], u"Message after utf data\n\n")

def convert_timestamp_to_pdt(ts):
  dt = ls.datetime_from_millis(ts)
  utc_delta = datetime.timedelta(seconds=time.altzone)
  utc_dt = dt + utc_delta
  pdt_delta = datetime.timedelta(hours=-7)
  pdt_dt = utc_dt + pdt_delta
  return int(time.mktime(pdt_dt.timetuple()) * 1000)

