#!/usr/bin/env python
# (c) Copyright 2010-2011 Cloudera, Inc. All rights reserved.
"""
Module for testing the dataset module
"""
import unittest
from cmf.monitor.tasktracker import dataset

LONG_TEST_VALUE = 4
STRING_TEST_VALUE = "This is a test"
STATE_TEST_VALUES = ["choice1", "choice2", "choice3" ]
GOOD_STATE_VALUE = STATE_TEST_VALUES[0]
BAD_STATE_VALUE = "boom!"
DATASET_NAME = "Test Dataset"
NUM_TEST_LOOPS = 100
COLLECT_INTERVAL = 5.0
REPORT_INTERVAL = 60.0

class TestUpdater(dataset.Updater):
  """ Test Updater for the DatasetTests class """

  def update(self, data):
    """ Generates values for testing """
    data.long_value.setvalue( LONG_TEST_VALUE )
    data.string_value.setvalue( STRING_TEST_VALUE )
    for i in xrange(1, NUM_TEST_LOOPS):
      data.gauge_value.setvalue( i, i )
      data.counter_value.setvalue( i, i )
    data.state_value.setvalue( GOOD_STATE_VALUE )


class TestDataset(dataset.Dataset):
  """ Dataset test class """
  members = [ dataset.LongElement(1,  "long_value", "Long Test Value"),
              dataset.StringElement(2, "string_value", "String Test Value"),
              dataset.GaugeElement(3, "gauge_value", "Gauge Value"),
              dataset.CounterElement(4, "counter_value", "Counter Value"),
              dataset.StateElement(5, "state_value", "State Value", STATE_TEST_VALUES) ]

  def __init__(self, name):
    updaters = [ TestUpdater ]
    super(TestDataset, self).__init__(name, None, updaters, COLLECT_INTERVAL, REPORT_INTERVAL)

class DatasetTests(unittest.TestCase):

  def setUp(self):
    self._dataset = TestDataset(DATASET_NAME)
    self._dataset.update()

  def test_invalid_choice(self):
    """ Test that we get an exception when we pass an invalid choice """
    self.assertRaises(ValueError, self._dataset.state_value.setvalue, BAD_STATE_VALUE)

  def test_invalid_long(self):
    """ Test that we get an exception when we pass a string in to a long value """
    self.assertRaises(ValueError, self._dataset.long_value.setvalue, "A string is not a long")

  def test_long_getvalue(self):
    """ Test that we get the same long value we set """
    self.assertEqual(LONG_TEST_VALUE, self._dataset.long_value.getvalue())

  def test_string_getvalue(self):
    """ Test that we get the same string value we set """
    self.assertEqual(STRING_TEST_VALUE, self._dataset.string_value.getvalue())

  def test_gauge_min(self):
    """ Test the min calculation for gauges """
    self.assertEqual(1, self._dataset.gauge_value.stats.min())

  def test_gauge_max(self):
    """ Test the max calculation for gauges """
    self.assertEqual(NUM_TEST_LOOPS - 1, self._dataset.gauge_value.stats.max())

  def test_gauge_avg(self):
    """ Test the avg calculation for gauges """
    self.assertEqual(NUM_TEST_LOOPS/2, self._dataset.gauge_value.stats.avg())

  def test_gauge_nsamples(self):
    """ Test the number of samples for gauges """
    self.assertEqual(NUM_TEST_LOOPS - 1, self._dataset.gauge_value.stats.samples())

  def test_gauge_stdev(self):
    """ Test the standard deviation for gauges """
    pass

  def test_gauge_last_sample(self):
    """ Test the last sample for gauges """
    self.assertEqual(NUM_TEST_LOOPS - 1, self._dataset.gauge_value.getvalue()['last_sample'])

  def test_counter_min(self):
    """ Test the min calculation for counters """
    self.assertEqual(1, self._dataset.counter_value.stats.min())

  def test_counter_max(self):
    """ Test the max calculation for counters """
    self.assertEqual(1, self._dataset.counter_value.stats.max())

  def test_counter_avg(self):
    """ Test the avg calculation for counters """
    self.assertEqual(1, self._dataset.counter_value.stats.avg())

  def test_counter_nsamples(self):
    """ Test the number of samples for counters """
    self.assertEqual(NUM_TEST_LOOPS - 2, self._dataset.counter_value.stats.samples())

  def test_counter_stdev(self):
    """ Test the standard deviation for counters """
    self.assertEqual(0, self._dataset.counter_value.stats.stdev())

  def test_counter_last_sample(self):
    """ Test the last sample for gauges """
    self.assertEqual(NUM_TEST_LOOPS - 1, self._dataset.counter_value.getvalue()['last_sample'])

if __name__ == '__main__':
  unittest.main()


