#!/usr/bin/env python
# (c) Copyright 2010-2011 Cloudera, Inc. All rights reserved.
"""
Module for testing the metrics module
"""
import unittest
from nose.tools import assert_true
from cmf.monitor.tasktracker import metrics

class MetricsTests(unittest.TestCase):
  """ unittest TestCase for metrics """

  def setUp(self):
    pass

  def test_illegal_json(self):
    """ Detecting illegal JSON """
    self.assertRaises(ValueError, metrics.import_schema_from_string, "{]")

  def test_duplicate_ids(self):
    """ Detecting duplicate IDs in schema """
    self.assertRaises(metrics.DuplicateIdError, metrics.import_schema_from_string,
      """
      {
              "fake_attempt_metric_1": {
                      "context": ["attempt"],
                      "desc": "Fake attempt metric 1",
                      "id": 1,
                      "type": "long"
              },
              "fake_attempt_metric_2": {
                      "context": ["attempt"],
                      "desc": "Fake attempt metric 2",
                      "id": 1,
                      "type": "long"
              }
      }
      """,
      set(['fake_attempt_metric_1', 'fake_attempt_metric_2']))

  def test_bad_context(self):
    """ Detecting a bad context """
    self.assertRaises(metrics.InvalidContextError, metrics.import_schema_from_string,
      """
      {
              "fake_attempt_metric_1": {
                      "context": ["foobar"],
                      "desc": "Fake attempt metric 1",
                      "id": 1,
                      "type": "long"
              }
      }
      """,
      set(['fake_attempt_metric_1']))

  def test_missing_type(self):
    """ Detecting a missing type """
    self.assertRaises(metrics.MissingTypeError, metrics.import_schema_from_string,
      """
      {
              "fake_attempt_metric_1": {
                      "context": ["attempt"],
                      "desc": "Fake attempt metric 1",
                      "id": 1
              }
      }
      """,
      set(['fake_attempt_metric_1']))


  def test_bad_type(self):
    """ Detecting a bad type """
    self.assertRaises(metrics.InvalidTypeError, metrics.import_schema_from_string,
      """
      {
              "swap_total": {
                      "context": ["attempt"],
                      "desc": "Total Swap",
                      "id": 9,
                      "type": "this_is_a_bad_type",
                      "units": "bytes"
              }
      }
      """,
      set(['swap_total']))

  def test_missing_id(self):
    """ Detecting a metric with a missing ID """
    self.assertRaises(metrics.MissingIdError, metrics.import_schema_from_string,
      """
      {
              "fake_attempt_metric_1": {
                      "context": ["attempt"],
                      "desc": "Fake attempt metric 1",
                      "type": "long"
              }
      }
      """,
      set(['fake_attempt_metric_1']))

  def test_state_missing_choices(self):
    """ Detecting missing choices from a state """
    self.assertRaises(metrics.MissingChoicesError, metrics.import_schema_from_string,
      """
      {
              "state": {
                      "this_should_be_choices_but_its_not": [
                              "RUNNING",
                              "SUCCEEDED",
                              "FAILED",
                              "UNASSIGNED",
                              "KILLED",
                              "COMMIT_PENDING",
                              "FAILED_UNCLEAN",
                              "KILLED_UNCLEAN"
                      ],
                      "context": ["attempt"],
                      "desc": "State of this attempt",
                      "id": 15,
                      "type": "state"
              }
      }
      """,
      set(['state']))

  def test_schema_import(self):
    """ Does the agent actually import the current schema """
    metrics.import_schema()

  def test_correct_context(self):
    """ Does the agent catch context mismatches """
    self.assertRaises(metrics.InvalidContextError, metrics.import_schema_from_string,
      """
      {
          "fake_attempt_metric_1": {
            "context": ["this_is_not_a_valid_agent_context"],
            "desc": "Fake attempt metric 1",
            "id": 1,
            "type": "long",
            "attrs": {}
          }
      }
      """,
      set(['fake_attempt_metric_1']))

  def test_correct_contexts(self):
    """Does the agent correctly assign multiple contexts"""
    metrics.import_schema_from_string(
      """
      {
          "fake_attempt_metric_1": {
            "context": ["attempt"],
            "desc": "Fake attempt metric 1",
            "id": 1,
            "type": "long",
            "attrs": {}
          }
      }
      """,
      set(['fake_attempt_metric_1']))
    assert_true("fake_attempt_metric_1" in \
                  [n.name for n in metrics.AttemptDataset.members])



if __name__ == '__main__':
  unittest.main()
