# Copyright (c) 2011 Cloudera, Inc. All rights reserved.

import os
import shutil
import tempfile
try:
  from cStringIO import StringIO
except:
  from StringIO import StringIO

from cmf.monitor import schema, constants
from cmf.monitor.stoppable_thread import StoppableThread
from cmf.monitor import cgroup_metrics_utils
from cmf.monitor.tasktracker import json_parser
from cmf.monitor.tasktracker.monitor import TaskTrackerMonitor
from cmf.monitor.tasktracker.tasktracker import MockTaskTracker
from cmf.monitor.tasktracker.tasktracker import UNKNOWN_ENTITY
from cmf_test.monitor import make_config_file, TestMonitorDaemon
from cmf_test.monitor.test_cgroup_metrics_utils import write_sample_blkio_stat_files
from cmf_test.monitor.test_cgroup_metrics_utils import write_sample_cpuacct_stat_file
from cmf_test.monitor.test_cgroup_metrics_utils import write_sample_memory_stat_file
from nose.tools import assert_true, assert_equals

class MockTaskTrackerMonitor(TaskTrackerMonitor):
  def __init__(self, daemon, registry, attempt_monitor, string,
               cgroup_root=None, cgroup_name=None):
    TaskTrackerMonitor.__init__(self, daemon, registry, cgroup_root,
                                cgroup_name, attempt_monitor)
    self.set_string(string)
    self.set_throw_error(None)

  def set_string(self, string):
    self._string = string

  def set_throw_error(self, error):
    self._error = error

  def urlopen(self, url, **kwargs):
    if self._error:
      raise self._error
    return StringIO(self._string)

def do_nothing():
  pass

# CM 3.7 config file format.
def _make_cm37_format_config_file(version, port):
  properties = """[tasktracker]
role_name = mapreduce-TT-1
service_name = my_mapreduce
host = localhost
collect_interval = 1.0
tasktracker_instrumentation_url = http://127.0.0.1:%s
task_tracker_http_address = 127.0.0.1
task_tracker_http_port = 50035
task_tracker_mapred_local_dir = /path/to/dir,/another/data/dir
"""
  return  make_config_file(properties % (port,))

# CM 4.0 config file: we added the service version
def _make_cm40_format_config_file(version, port):
  properties = """[tasktracker]
role_name = mapreduce-TT-1
service_name = my_mapreduce
service_version = %s
host = localhost
collect_interval = 1.0
tasktracker_instrumentation_url = http://127.0.0.1:%s
task_tracker_http_address = 127.0.0.1
task_tracker_http_port = 50035
task_tracker_mapred_local_dir = /path/to/dir,/another/data/dir
"""
  return  make_config_file(properties % (version, port,))

# CM 4.1 config file: we added the log directory.
def _make_cm41_format_config_file(version, port):
  properties = """[tasktracker]
role_name = mapreduce-TT-1
service_name = my_mapreduce
service_version = %s
host = localhost
log_dir = /path/to/logs
collect_interval = 1.0
tasktracker_instrumentation_url = http://127.0.0.1:%s
task_tracker_http_address = 127.0.0.1
task_tracker_http_port = 50035
task_tracker_mapred_local_dir = /path/to/dir,/another/data/dir
"""
  return  make_config_file(properties % (version, port,))

def test_tt_lifecycle_cdh5():
  do_tt_lifecycle(constants.SERVICE_VERSION_CDH5,
      _make_cm40_format_config_file, json_parser.SAMPLE, 2)
  do_tt_lifecycle(constants.SERVICE_VERSION_CDH5,
      _make_cm41_format_config_file, json_parser.SAMPLE, 3)

def get_metric_dict(version, sample, cgroup_root=None, cgroup_name=None):
  daemon = TestMonitorDaemon()
  reg = daemon.get_registry_for_testing()
  do_nothing_attempt_monitor = \
    StoppableThread("DoNothingAttemptMonitor", 5.0, do_nothing)
  monitor = MockTaskTrackerMonitor(daemon, reg, do_nothing_attempt_monitor,
                                   sample, cgroup_root=cgroup_root,
                                   cgroup_name=cgroup_name)
  monitor.refresh_pid(os.getpid(), True)

  conf_dir = _make_cm41_format_config_file(version, 1234)
  try:
    # Collect
    monitor.refresh(conf_dir)
    assert_equals(monitor, reg.get_registry().pop())
    monitor.collect()

    # Confirm that metrics are what we expected
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    monitor.stop()
    return metric_dict

  finally:
    shutil.rmtree(conf_dir)

def do_tt_lifecycle(version, make_config_file_function, sample,
                    expected_directory_metrics):
  daemon = TestMonitorDaemon()
  reg = daemon.get_registry_for_testing()
  do_nothing_attempt_monitor = \
    StoppableThread("DoNothingAttemptMonitor", 5.0, do_nothing)
  monitor = MockTaskTrackerMonitor(daemon, reg, do_nothing_attempt_monitor,
                                   sample)
  monitor.refresh_pid(os.getpid(), True)

  # Make two different configs
  conf_dir = make_config_file_function(version, 1234)
  conf_dir2 = make_config_file_function(version, 5678)

  try:
    assert_equals(0, len(reg.get_registry()))

    # Start it
    monitor.refresh(conf_dir)
    assert_equals(monitor, reg.get_registry().pop())
    monitor.collect()
    monitor.collect_attempt_metrics(MockTaskTracker('mock_job1', 0, 16, 1, 300, False))
    assert_true('1234' in monitor.instrumentation_url)

    # Refresh with a different config
    monitor.refresh(conf_dir2)
    assert_equals(monitor, reg.get_registry().pop())
    assert_equals(1, len(reg.get_registry()))

    monitor.collect()
    assert_true('5678' in monitor.instrumentation_url)

    # Test metrics and role update
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    assert_equals('mapreduce-TT-1', role_update['rolename'])
    assert_equals('my_mapreduce', role_update['service_name'])
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']

    assert_true(schema.TOTAL_CPU_USER in metric_dict)
    assert_true(schema.TOTAL_CPU_SYSTEM in metric_dict)
    assert_true(schema.TOTAL_CPU_USER_WITH_DESCENDANTS in metric_dict)
    assert_true(schema.TOTAL_CPU_SYSTEM_WITH_DESCENDANTS in metric_dict)
    assert_equals(constants.WEB_METRICS_COLLECTION_GOOD,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)

    if version == constants.SERVICE_VERSION_CDH5:
      for metric_id in daemon.metric_accessors['TASKTRACKER']['cdh5'].iterkeys():
        assert_true(metric_id in metric_dict)

    dir_metrics = role_update['dirUpdates']
    assert_true(len(dir_metrics) == expected_directory_metrics,
                "We configured %d dirs: %s" % (expected_directory_metrics, dir_metrics))
    assert_true(os.path.ismount(dir_metrics[0]['mountpoint']))
    assert_true(dir_metrics[0]['path'] is not None)

    assert_true(schema.UNEXPECTED_EXITS_COUNT in metric_dict)

    attempt_updates = monitor.get_attempt_updates()
    assert_true(len(attempt_updates) != 0)
    assert_true('metrics' in attempt_updates[0].keys())
    assert_true(len(attempt_updates[0]['metrics']) != 0)
    assert_true('id' in attempt_updates[0]['metrics'][0].keys())
    assert_true(attempt_updates[0]['metrics'][0]['id'] is not None)
    assert_true('value' in attempt_updates[0]['metrics'][0].keys())

    # Test that we do not send updates for values generated
    # by JT poller starting version 4
    for attempt_update in attempt_updates:
      metric_dict = { }
      for metric_obj in attempt_update['metrics']:
        metric_dict[metric_obj['id']] = metric_obj['value']
      assert_true((version >= 4) != (schema.TOTAL_CPU_USER in metric_dict))
      assert_true((version >= 4) != (schema.MEM_RSS in metric_dict))

    # Test that we do not generate any updates with defaults
    monitor.collect_attempt_metrics(MockTaskTracker('mock_job2', 0, 16, 1, 300, True))
    attempt_updates = monitor.get_attempt_updates()
    for attempt_update in attempt_updates:
      assert_true(attempt_update['context']['job_id'] != UNKNOWN_ENTITY)

    # Test when the process is not running
    monitor.refresh_pid(None)
    monitor.collect()
    monitor.collect_attempt_metrics(MockTaskTracker('mock_job1', 0, 16, 1, 300, False))

    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()

    assert_equals('mapreduce-TT-1', role_update['rolename'])
    assert_equals('my_mapreduce', role_update['service_name'])
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']

    dir_metrics = role_update['dirUpdates']
    assert_true(len(dir_metrics) == expected_directory_metrics,
                "We configured %d dirs: %s" % (expected_directory_metrics, dir_metrics))
    assert_true(os.path.ismount(dir_metrics[0]['mountpoint']))
    assert_true(dir_metrics[0]['path'] is not None)

    assert_equals(0, len(monitor.get_attempt_updates()))

    # Test with a collection error
    monitor.refresh_pid(os.getpid(), True)
    monitor.set_throw_error(IOError("BOOM!"))
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_equals(constants.WEB_METRICS_COLLECTION_COMMUNICATION_FAILURE,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)

    # Test with some invalid JSON
    monitor.set_throw_error(None)
    monitor.set_string("THIS IS BOGUS JSON!")
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_equals(constants.WEB_METRICS_COLLECTION_PARSE_FAILURE,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)

    # Stop it
    monitor.stop()
    assert_equals(0, len(reg.get_registry()))
  finally:
    shutil.rmtree(conf_dir)
    shutil.rmtree(conf_dir2)

# Test that the TT metrics properly include the cgroup metrics
def test_cgroup_metrics_collection():
  cgroup_root = tempfile.mkdtemp()
  try:
    cgroup_name = 'mapreduce-TT-1'
    write_sample_blkio_stat_files(cgroup_root, cgroup_name)
    write_sample_memory_stat_file(cgroup_root, cgroup_name)
    write_sample_cpuacct_stat_file(cgroup_root, cgroup_name)

    # We don't care about CDH version or the sample JSON at all
    # so just pick some values here.
    metric_dict = get_metric_dict(constants.SERVICE_VERSION_CDH5,
                                  json_parser.SAMPLE_CDH5,
                                  cgroup_root=cgroup_root,
                                  cgroup_name=cgroup_name)

    # These are the sample values written by the functions above
    assert_equals(3.0, metric_dict.get(schema.CGROUP_MEM_RSS))
    assert_equals(4.0, metric_dict.get(schema.CGROUP_MEM_PAGE_CACHE))
    assert_equals(5.0, metric_dict.get(schema.CGROUP_MEM_SWAP))
    assert_equals(41160.0, metric_dict.get(schema.CGROUP_READ_IOS))
    assert_equals(3.0, metric_dict.get(schema.CGROUP_WRITE_IOS))
    assert_equals(7749218304.0, metric_dict.get(schema.CGROUP_READ_BYTES))
    assert_equals(12288.0, metric_dict.get(schema.CGROUP_WRITE_BYTES))

    expected_user_time = 3000.0 / cgroup_metrics_utils.jiffies_per_second
    expected_system_time = 6000.0 / cgroup_metrics_utils.jiffies_per_second

    assert_equals(expected_user_time, metric_dict.get(schema.CGROUP_TOTAL_CPU_USER))
    assert_equals(expected_system_time, metric_dict.get(schema.CGROUP_TOTAL_CPU_SYSTEM))
  finally:
    shutil.rmtree(cgroup_root)
