# Copyright (c) 2013 Cloudera, Inc. All rights reserved.

import os
import shutil
try:
  from cStringIO import StringIO
except:
  from StringIO import StringIO

from cmf.monitor import constants, schema
from cmf.monitor.generic import impala_adapters
from cmf.monitor.generic import GenericMonitor
from cmf.monitor.generic.utils import VersionUtils, VersionType
from cmf.safety_valve import SafetyValve
from cmf.monitor.generic.impala_adapters import CatalogServerAdapter
from cmf_test.monitor import make_config_file, TestMonitorDaemon

from nose.tools import assert_true, assert_equals, assert_false

class MockOpenUrl(object):
  def __init__(self, response):
    self._response = response
    self._error = None
    self._url = None
    self._username = None
    self._password = None

  def __call__(self, url, username, password):
    self._url = url
    self._username = username
    self._password = password

    if self._error:
      raise self._error
    return StringIO(self._response)

  @property
  def url(self):
    return self._url

  @property
  def username(self):
    return self._username

  @property
  def password(self):
    return self._password

  def set_response(self, response):
    self._response = response

  def set_throw_error(self, error):
    self._error = error

def test_catalogserver_lifecycle_cdh5():
  do_catalogserver_lifecycle("5.4.1")

def do_catalogserver_lifecycle(release):
  version = int(release[0])
  daemon = TestMonitorDaemon()
  reg = daemon.get_registry_for_testing()
  adapter = CatalogServerAdapter(SafetyValve())
  monitor = GenericMonitor("IMPALA", "CATALOGSERVER", daemon, reg, adapter)
  mock_open_url = MockOpenUrl(SAMPLE_CDH5);
  monitor._url_collector._urlopen_callout = mock_open_url
  monitor.refresh_pid(os.getpid(), True)

  properties = """[catalogserver]
role_name = %s
service_name = my_impala
service_version = %d
service_release = %s
host = localhost
webserver_port = 1234
webserver_username = %s
webserver_password = %s
log_dir = /path/to/logs
collect_interval = 1.0
catalogserver_webserver_port = 9091
"""

  # Make two different configs
  conf_dir = make_config_file(properties % ("catalogserver-test", version,
                                            release, "foo", "one"))
  conf_dir2 = make_config_file(properties % ("catalogserver", version,
                                             release, "bar", "two"))
  expected_url = "http://localhost:9091/jsonmetrics"
  if VersionUtils.at_least(release, "5.4", VersionType.MINOR):
    expected_url += "?json"

  try:
    assert_equals(0, len(reg.get_registry()))

    # Start it
    monitor.refresh(conf_dir)
    assert_equals(monitor, reg.get_registry().pop())
    monitor.collect()
    assert_equals(expected_url, mock_open_url.url)
    assert_equals("foo", mock_open_url.username)
    assert_equals("one", mock_open_url.password)
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    assert_equals('catalogserver-test', role_update['rolename'])

    # Refresh with a different config
    monitor.refresh(conf_dir2)
    assert_equals(monitor, reg.get_registry().pop())
    assert_equals(1, len(reg.get_registry()))

    monitor.collect()
    assert_equals(expected_url, mock_open_url.url)
    assert_equals("bar", mock_open_url.username)
    assert_equals("two", mock_open_url.password)

    # Test metrics and role update
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()

    assert_equals('catalogserver', role_update['rolename'])
    assert_equals('my_impala', role_update['service_name'])
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_true(metric_dict[schema.TOTAL_CPU_SYSTEM] >= 0.0)
    assert_true(metric_dict[schema.MEM_RSS] >= 0.0)
    assert_true(metric_dict[schema.MEM_VIRTUAL] >= 0.0)
    assert_equals(constants.WEB_METRICS_COLLECTION_GOOD,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)

    version_string = constants.VERSION_STRINGS[version]
    for metric_id in daemon.metric_accessors['CATALOGSERVER'][version_string].iterkeys():
      assert_equals(type(float(1.0)), type(metric_dict[metric_id]))

    dir_metrics = role_update['dirUpdates']
    assert_equals(1, len(dir_metrics), "We configured 1 dirs: %s" % dir_metrics)
    assert_true(os.path.ismount(dir_metrics[0]['mountpoint']))
    assert_equals('/path/to/logs', dir_metrics[0]['path'])

    assert_true(schema.UNEXPECTED_EXITS_COUNT in metric_dict)

    # Test when the process is not running
    monitor.refresh_pid(None)
    monitor.collect()

    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()

    assert_equals('catalogserver', role_update['rolename'])
    assert_equals('my_impala', role_update['service_name'])
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']

    dir_metrics = role_update['dirUpdates']
    assert_true(len(dir_metrics) == 1, "We configured 1 dirs: %s" % dir_metrics)
    assert_true(os.path.ismount(dir_metrics[0]['mountpoint']))
    assert_equals('/path/to/logs', dir_metrics[0]['path'])

    # Test with a collection error
    monitor.refresh_pid(os.getpid(), True)
    mock_open_url.set_throw_error(IOError("BOOM!"))
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_true(metric_dict[schema.MEM_RSS] >= 0.0)
    assert_true(metric_dict[schema.MEM_VIRTUAL] >= 0.0)
    assert_equals(constants.WEB_METRICS_COLLECTION_COMMUNICATION_FAILURE,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)

    # Test with some invalid JSON
    mock_open_url.set_throw_error(None)
    mock_open_url.set_response("THIS IS BOGUS JSON!")
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_true(metric_dict[schema.MEM_RSS] >= 0.0)
    assert_true(metric_dict[schema.MEM_VIRTUAL] >= 0.0)
    assert_equals(constants.WEB_METRICS_COLLECTION_PARSE_FAILURE,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)

    # Stop it
    monitor.stop()
    assert_equals(0, len(reg.get_registry()))
  finally:
    shutil.rmtree(conf_dir)
    shutil.rmtree(conf_dir2)

SAMPLE_CDH5 = r"""
{
   "catalog.version":"catalogd version 1.2.2-INTERNAL RELEASE (build 4161d79734b48bcbff5b47a7c9e09d206cc4b848)",
   "impala.thrift-server.CatalogService.connections-in-use":0,
   "impala.thrift-server.CatalogService.total-connections":0,
   "statestore-subscriber.connected":1,
   "statestore-subscriber.heartbeat-duration":{
      "count":14437,
      "last":0.000111741,
      "min":3.4552e-05,
      "max":0.00209998,
      "mean":8.22269e-05,
      "stddev":5.79026e-05
   },
   "statestore-subscriber.heartbeat-interval-time":{
      "count":14437,
      "last":0.50049,
      "min":0.499322,
      "max":0.61974,
      "mean":0.500526,
      "stddev":0.00101264
   },
   "statestore-subscriber.last-recovery-duration":0,
   "statestore-subscriber.last-recovery-time":"N/A",
   "statestore-subscriber.statestore.client-cache.clients-in-use":0,
   "statestore-subscriber.statestore.client-cache.total-clients":1,
   "statestore-susbcriber.registration-id":"ab4d20fd9e964412:eba41d77590acfb0",
   "tcmalloc.bytes-in-use":23051464,
   "tcmalloc.pageheap-free-bytes":38633472,
   "tcmalloc.pageheap-unmapped-bytes":0,
   "tcmalloc.physical-bytes-reserved":71663616,
   "tcmalloc.total-bytes-reserved":71663616
}
"""

