# Copyright (c) 2013 Cloudera, Inc. All rights reserved.

from cmf.monitor import cgroup_metrics_utils
from cmf.monitor import schema
from nose.tools import assert_equals, assert_false, assert_true

import os
import shutil
import tempfile

class Update(object):
  def __init__(self):
    self.metrics = []

  def add_metric(self, metric, value):
    self.metrics.append((metric, value))

def test_sc_clk_tck():
  # These are the valid values for jiffies per second according to the
  # Linux kernel documentation
  assert_true(cgroup_metrics_utils.jiffies_per_second in [100, 250, 300, 1000])

def _handle_unsupported_subsystem(cgroup_root, cgroup_name):
  # Let's see what happens if we have some unsupported subsystem
  # but none of the supported ones. This should not do anything.
  unsupported_cgroup_path = os.path.join(cgroup_root, "unsupported", cgroup_name)
  os.makedirs(unsupported_cgroup_path)
  open(os.path.join(unsupported_cgroup_path, "unsupported.stat"), 'w').write('total_rss 100\n')

  update = Update()
  cgroup_metrics_utils.add_cgroup_metrics(update, cgroup_root, cgroup_name)
  assert_equals(0, len(update.metrics))

def write_sample_memory_stat_file(cgroup_root, cgroup_name):
  # Now let's write metrics only for the memory subsystem
  memory_cgroup_path = os.path.join(cgroup_root, "memory", cgroup_name)
  os.makedirs(memory_cgroup_path)
  mem_stat_path = os.path.join(memory_cgroup_path, 'memory.stat')
  mem_stat_file = 'total_rss 1\ntotal_mapped_file 2\ntotal_cache 4\ntotal_swap 5\n'
  open(mem_stat_path, 'w').write(mem_stat_file)
  return mem_stat_path

def _handle_memory_subsystem(cgroup_root, cgroup_name):
  mem_stat_path = write_sample_memory_stat_file(cgroup_root, cgroup_name)

  update = Update()
  cgroup_metrics_utils.add_cgroup_metrics(update, cgroup_root, cgroup_name)
  assert_equals(3, len(update.metrics))
  assert_true((schema.CGROUP_MEM_RSS, 3.0) in update.metrics)
  assert_true((schema.CGROUP_MEM_PAGE_CACHE, 4.0) in update.metrics)
  assert_true((schema.CGROUP_MEM_SWAP, 5.0) in update.metrics)

  # If total_rss or total_mapped_file is missing we shouldn't add a metric for
  # the RSS of the cgroup
  mem_stat_file = 'total_mapped_file 2\ntotal_cache 4\ntotal_swap 5\n'
  open(mem_stat_path, 'w').write(mem_stat_file)

  update = Update()
  cgroup_metrics_utils.add_cgroup_metrics(update, cgroup_root, cgroup_name)
  assert_equals(2, len(update.metrics))
  assert_true((schema.CGROUP_MEM_PAGE_CACHE, 4.0) in update.metrics)
  assert_true((schema.CGROUP_MEM_SWAP, 5.0) in update.metrics)

  # If one of the other lines is missing we should just skip that one
  mem_stat_file = 'total_rss 1\ntotal_mapped_file 2\ntotal_cache 4'
  open(mem_stat_path, 'w').write(mem_stat_file)

  update = Update()
  cgroup_metrics_utils.add_cgroup_metrics(update, cgroup_root, cgroup_name)
  assert_equals(2, len(update.metrics))
  assert_true((schema.CGROUP_MEM_RSS, 3.0) in update.metrics)
  assert_true((schema.CGROUP_MEM_PAGE_CACHE, 4.0) in update.metrics)

  # Remove the memory stat file so that we can test other subsystems in isolation
  os.remove(mem_stat_path)

def write_sample_cpuacct_stat_file(cgroup_root, cgroup_name):
  # Now let's write metrics only for the cpuacct subsystem
  cpuacct_cgroup_path = os.path.join(cgroup_root, "cpuacct", cgroup_name)
  cpuacct_stat_path = os.path.join(cpuacct_cgroup_path, 'cpuacct.stat')
  os.makedirs(cpuacct_cgroup_path)
  # The sample values here are chosen so they will be divisible by any admissible
  # number for jiffies_per_second
  cpuacct_stat_file = 'user 3000\nsystem 6000\n'
  open(cpuacct_stat_path, 'w').write(cpuacct_stat_file)
  return cpuacct_stat_path

def _handle_cpuacct_subsystem(cgroup_root, cgroup_name):
  cpuacct_stat_path = write_sample_cpuacct_stat_file(cgroup_root, cgroup_name)

  update = Update()
  cgroup_metrics_utils.add_cgroup_metrics(update, cgroup_root, cgroup_name)
  assert_equals(2, len(update.metrics))

  jiffies_per_second = os.sysconf(os.sysconf_names['SC_CLK_TCK'])
  user_val = 3000.0 / jiffies_per_second
  system_val = 6000.0 / jiffies_per_second

  assert_true((schema.CGROUP_TOTAL_CPU_USER, user_val) in update.metrics)
  assert_true((schema.CGROUP_TOTAL_CPU_SYSTEM, system_val) in update.metrics)

  # As with the memory subsystem let's be robust to missing entries
  cpuacct_stat_file = 'user 3000\n'
  open(cpuacct_stat_path, 'w').write(cpuacct_stat_file)

  update = Update()
  cgroup_metrics_utils.add_cgroup_metrics(update, cgroup_root, cgroup_name)
  assert_equals(1, len(update.metrics))
  assert_true((schema.CGROUP_TOTAL_CPU_USER, user_val) in update.metrics)

  # Remove the cpuacct stat file so that we can test other subsystems in isolation
  os.remove(cpuacct_stat_path)

def write_sample_blkio_stat_files(cgroup_root, cgroup_name):
  # Write the metrics for the blkio subsystem
  blkio_group_path = os.path.join(cgroup_root, 'blkio', cgroup_name)
  os.makedirs(blkio_group_path)
  # Now let's write some sample data to blkio.io_serviced and blkio.io_service_bytes
  io_serviced_path = os.path.join(blkio_group_path, 'blkio.io_serviced')
  open(io_serviced_path, 'w').write("""252:0 Read 41160
                                       252:0 Write 3
                                       252:0 Sync 41163
                                       252:0 Async 0
                                       252:0 Total 41163
                                       Total 41163""")
  io_service_bytes_path = os.path.join(blkio_group_path, 'blkio.io_service_bytes')
  open(io_service_bytes_path, 'w').write("""252:0 Read 7749218304
                                            252:0 Write 12288
                                            252:0 Sync 7749230592
                                            252:0 Async 0
                                            252:0 Total 7749230592
                                            Total 7749230592""")
  return [io_serviced_path, io_service_bytes_path]

def _handle_blkio_subsystem(cgroup_root, cgroup_name):
  blkio_stat_paths = write_sample_blkio_stat_files(cgroup_root, cgroup_name)

  update = Update()
  cgroup_metrics_utils.add_cgroup_metrics(update, cgroup_root, cgroup_name)
  assert_equals(4, len(update.metrics))
  assert_true((schema.CGROUP_READ_IOS, 41160.0) in update.metrics)
  assert_true((schema.CGROUP_WRITE_IOS, 3.0) in update.metrics)
  assert_true((schema.CGROUP_READ_BYTES, 7749218304.0) in update.metrics)
  assert_true((schema.CGROUP_WRITE_BYTES, 12288.0) in update.metrics)

  # Check robustness in the face of missing entries
  open(blkio_stat_paths[0], 'w').write("252:0 Read 41160\n")
  open(blkio_stat_paths[1], 'w').write("252:0 Write 12288\n")

  update = Update()
  cgroup_metrics_utils.add_cgroup_metrics(update, cgroup_root, cgroup_name)
  assert_equals(2, len(update.metrics))
  assert_true((schema.CGROUP_READ_IOS, 41160.0) in update.metrics)
  assert_true((schema.CGROUP_WRITE_BYTES, 12288.0) in update.metrics)

  for report in blkio_stat_paths:
    os.remove(report)

def test_cgroup_metric_collection():
  update = Update()

  # Passing in None for the cgroup_root bails out
  cgroup_metrics_utils.add_cgroup_metrics(update, None, None)
  assert_equals(0, len(update.metrics))

  # So does passing in a nonexistent cgroup_root
  cgroup_metrics_utils.add_cgroup_metrics(update, "doesnotexist", None)
  assert_equals(0, len(update.metrics))

  cgroup_root = tempfile.mkdtemp()
  cgroup_name = "1-TEST-CGROUP"

  # And also passing in None for the cgroup_name
  cgroup_metrics_utils.add_cgroup_metrics(update, cgroup_root, None)
  assert_equals(0, len(update.metrics))

  _handle_unsupported_subsystem(cgroup_root, cgroup_name)
  _handle_memory_subsystem(cgroup_root, cgroup_name)
  _handle_cpuacct_subsystem(cgroup_root, cgroup_name)
  _handle_blkio_subsystem(cgroup_root, cgroup_name)

  shutil.rmtree(cgroup_root)

def test_stat_file_parsing():
  stat_dict = cgroup_metrics_utils._read_stat_file("/tmp")
  assert_true(stat_dict is None)

  stat_dict = cgroup_metrics_utils._read_stat_file("/doesnotexist")
  assert_true(stat_dict is None)

  temp_stat_file = tempfile.NamedTemporaryFile()
  temp_stat_file.write("a\nb 10\nc 20 30\n\n")
  temp_stat_file.flush()

  stat_dict = cgroup_metrics_utils._read_stat_file(temp_stat_file.name)
  assert_false(stat_dict is None)
  assert_equals(2, len(stat_dict))
  assert_equals('10', stat_dict.get('b'))
  assert_equals('20', stat_dict.get('c'))

def test_blkio_stat_file_parsing():
  stat_dict = cgroup_metrics_utils._read_blkio_stat_file("/tmp")
  assert_true(stat_dict is None)

  stat_dict = cgroup_metrics_utils._read_blkio_stat_file("/doesnotexist")
  assert_true(stat_dict is None)

  temp_stat_file = tempfile.NamedTemporaryFile()
  temp_stat_file.write("""
                          8:0 Read 1
                          8:0 Write 2
                          8:0 Sync
                          8:0 Async 4 ignored
                          8:0 Total 5
                          252:0 Read 6
                          252:0 Write 7
                          252:0 Sync 8
                          252:0 Async 9
                          252:0 Total 10
                          Total ignored""")
  temp_stat_file.flush()

  stat_dict = cgroup_metrics_utils._read_blkio_stat_file(temp_stat_file.name)
  assert_false(stat_dict is None)
  assert_equals(7.0, stat_dict.get('Read'))
  assert_equals(9.0, stat_dict.get('Write'))
  assert_equals(8.0, stat_dict.get('Sync'))
  assert_equals(13.0, stat_dict.get('Async'))
  assert_equals(15.0, stat_dict.get('Total'))
