# Copyright (c) 2017 Cloudera, Inc. All rights reserved.

try:
  from cStringIO import StringIO
except:
  from StringIO import StringIO
import os
import shutil
import simplejson

from cmf.monitor import conf, constants
from cmf.monitor.generic.common_adapter import CommonAdapter
from cmf.monitor.monitor_properties import ConfParser
from cmf.monitor.role_update import RoleUpdate
from cmf.safety_valve import SafetyValve
from cmf_test.monitor import make_config_file

from nose.tools import assert_equals

def init_and_check(
    conf_dir,
    metrics_path,
    service_type,
    role_type,
    expected_url,
    expected_file_path,
    expected_role_entity_type):
  # Init the adapter and test the basic configurations.
  csd_conf_parser = ConfParser(conf_dir, conf.CSD_MONITORING_PROPERTIES)
  adapter = CommonAdapter(service_type, role_type, SafetyValve())
  adapter.read_csd_monitoring_properties(csd_conf_parser)
  assert_equals(expected_url, adapter.get_metrics_url(None))
  assert_equals(expected_file_path, adapter.get_metrics_file(None))
  assert_equals(expected_role_entity_type, adapter.role_entity_type)
  if metrics_path is not None:
    adapter.read_service_defined_metrics(metrics_path)
  return adapter

def check_metrics_equal(update, expected_metrics):
  role_metrics = {}
  for metric in update.to_avro_dict()['metrics']:
    role_metrics[metric['id']] = metric['value']
  assert_equals(role_metrics, expected_metrics)

def test_common_adapter_no_collection():
  # Test common adapter metrics not available
  conf_dir = make_config_file(
"""[common_role]
metrics_available = false
web_metrics_enabled = true
http_url = http://0.0.0.0:30000/metrics
role_entity_type = TEST_SERVICE-COMMON_ROLE
""", filename=conf.CSD_MONITORING_PROPERTIES)
  try:
    csd_conf_parser = ConfParser(conf_dir, conf.CSD_MONITORING_PROPERTIES)
    adapter = CommonAdapter("TEST_SERVICE", "COMMON_ROLE", SafetyValve())
    adapter.read_csd_monitoring_properties(csd_conf_parser)
    assert_equals(None, adapter.get_metrics_url(None))
    assert_equals(None, adapter.get_metrics_file(None))
  finally:
    shutil.rmtree(conf_dir)

def test_common_adapter_web_metrics():
  # Test common adapter for web metrics and simple format
  version = constants.SERVICE_VERSION_CDH5
  conf_dir = make_config_file(
"""[common_role]
metrics_available = true
web_metrics_enabled = true
http_url = http://0.0.0.0:30000/metrics
role_entity_type = TEST_SERVICE-COMMON_ROLE
""", filename=conf.CSD_MONITORING_PROPERTIES)
  metrics_path = os.path.join(conf_dir, conf.SERVICE_METRICS_PROPERTIES)
  file(metrics_path, 'w').write(
"""{
  "TEST_SERVICE-COMMON_ROLE" : {
    "32772" : {
      "source" : "testMetricOne"
    },
    "32773" : {
      "source" : "testMetricTwo::value"
    }
  }
}""")
  sample = r"""
{
  "testMetricOne" : 123,
  "testMetricTwo" : {
    "value" : 567
  }
}
"""
  try:
    service_type = "TEST_SERVICE"
    role_type = "COMMON_ROLE"
    expected_url = "http://0.0.0.0:30000/metrics"
    expected_file_path = None
    expected_role_entity_type = "TEST_SERVICE-COMMON_ROLE"

    adapter = init_and_check(
                conf_dir,
                metrics_path,
                service_type,
                role_type,
                expected_url,
                expected_file_path,
                expected_role_entity_type)

    metrics = adapter.parse_metrics_from_url(None,
                                    simplejson.load(StringIO(sample)))
    update = RoleUpdate(name="whatever",
                        roletype=adapter.role_type,
                        servicetype=adapter.service_type,
                        service_name="whatever")
    adapter.add_webserver_metrics(version,
                                  update,
                                  metrics,
                                  None)

    expected_metrics = {
        32772 : 123,
        32773 : 567
        }
    check_metrics_equal(update, expected_metrics)
  finally:
    shutil.rmtree(conf_dir)

def test_common_adapter_ssl():
  # Test common adapter for https endpoint
  conf_dir = make_config_file(
"""[common_role]
metrics_available = true
web_metrics_enabled = true
http_url = http://0.0.0.0:30000/metrics
https_url = https://0.0.0.0:3000/metrics
ssl_enabled = true
role_entity_type = TEST_SERVICE-COMMON_ROLE
""", filename=conf.CSD_MONITORING_PROPERTIES)
  try:
    csd_conf_parser = ConfParser(conf_dir, conf.CSD_MONITORING_PROPERTIES)
    adapter = CommonAdapter("TEST_SERVICE", "COMMON_ROLE", SafetyValve())
    adapter.read_csd_monitoring_properties(csd_conf_parser)
    assert_equals("https://0.0.0.0:3000/metrics", adapter.get_metrics_url(None))
    assert_equals(None, adapter.get_metrics_file(None))
  finally:
    shutil.rmtree(conf_dir)

def test_common_adapter_file_metrics():
  # Test common adapter for file metrics and JMX bean format
  version = constants.SERVICE_VERSION_CDH5
  conf_dir = make_config_file(
"""[common_role]
metrics_available = true
file_metrics_enabled = true
file_path = /path/to/metrics/file
role_entity_type = TEST_SERVICE-COMMON_ROLE
metrics_format = JMX
""", filename=conf.CSD_MONITORING_PROPERTIES)
  metrics_path = os.path.join(conf_dir, conf.SERVICE_METRICS_PROPERTIES)
  file(metrics_path, 'w').write(
"""{
  "TEST_SERVICE-COMMON_ROLE" : {
    "32772" : {
      "source" : "REGEX$$testMetricOne::value"
    },
    "32773" : {
      "source" : "REGEX$$Catalina:type=RequestProcessor,worker=::value"
    },
    "32774" : {
      "source" : "REGEX$$id1=[0-9]+,id2=[0-9]+::value"
    },
    "32775" : {
      "source" : "REGEX$$java.lang:type=Memory::HeapMemoryUsage::used"
    }
  }
}""")
  sample = r"""
{
  "beans" : [ {
      "name" : "testMetricOne",
      "value" : 123
    }, {
      "name" : "Catalina:type=RequestProcessor,worker=http-1600",
      "value" : 456
    }, {
      "name" : "id1=123,id2=456",
      "value" : 789
    }, {
      "name" : "java.lang:type=Memory",
      "HeapMemoryUsage" : {
        "committed" : 410517504,
        "init" : 428077632,
        "max" : 954728448,
        "used" : 125101192
      }
    }
  ]
}
"""
  try:
    service_type = "TEST_SERVICE"
    role_type = "COMMON_ROLE"
    expected_url = None
    expected_file_path = "/path/to/metrics/file"
    expected_role_entity_type = "TEST_SERVICE-COMMON_ROLE"

    adapter = init_and_check(
                conf_dir,
                metrics_path,
                service_type,
                role_type,
                expected_url,
                expected_file_path,
                expected_role_entity_type)

    adapter.read_service_defined_metrics(metrics_path)

    metrics = adapter.parse_metrics_from_file(None,
                                    simplejson.load(StringIO(sample)))
    update = RoleUpdate(name="whatever",
                        roletype=adapter.role_type,
                        servicetype=adapter.service_type,
                        service_name="whatever")
    adapter.add_sample_file_metrics(version,
                                  update,
                                  metrics,
                                  None)
    expected_metrics = {
        32772 : 123,
        32773 : 456,
        32774 : 789,
        32775 : 125101192
        }
    check_metrics_equal(update, expected_metrics)
  finally:
    shutil.rmtree(conf_dir)
