# Copyright (c) 2012 Cloudera, Inc. All rights reserved.

import os
import shutil
import simplejson

import cmf.monitor.flume.json_parser
from StringIO import StringIO
from cmf.monitor import schema, constants
from cmf.monitor.daemon import MonitorDaemon
from cmf.monitor.flume import FlumeAgentMonitor
from cmf_test.monitor import make_config_file, TestMonitorDaemon

from nose.tools import assert_true, assert_equals, assert_false

class MockedFlumeAgentMonitor(FlumeAgentMonitor):
  def __init__(self, daemon, registry, string):
    FlumeAgentMonitor.__init__(self, daemon, registry)
    self.set_string(string)
    self.set_throw_error(None)

  def set_string(self, string):
    self._string = string

  def set_throw_error(self, error):
    self._error = error

  def urlopen(self, url, **kwargs):
    if self._error:
      raise self._error
    return StringIO(self._string)

# Pre CM 5.0 Beta 2 config file
def _make_pre_cm50_format_config_file(version, role_name):
  properties = """[agent]
service_name = my_flume1
service_version = %d
role_name = %s
host = localhost
log_dir = /path/to/logs
collect_interval = 1.0
"""
  return  make_config_file(properties % (version, role_name,))

# CM 5.0 Beta 2 config file: we added the agent name.
def _make_cm50_format_config_file(version, role_name):
  properties = """[agent]
service_name = my_flume1
service_version = %d
role_name = %s
host = localhost
log_dir = /path/to/logs
collect_interval = 1.0
agent_name = tier1
"""
  return  make_config_file(properties % (version, role_name,))

def test_agent_lifecycle_pre_cdh5():
  do_agent_lifecycle(constants.SERVICE_VERSION_CDH5,
                     _make_pre_cm50_format_config_file,
                     None)

def test_agent_lifecycle_cdh5():
  do_agent_lifecycle(constants.SERVICE_VERSION_CDH5,
                     _make_cm50_format_config_file,
                     'tier1')

def do_agent_lifecycle(version, make_config_file_function, agent_name):

  daemon = TestMonitorDaemon()
  reg = daemon.get_registry_for_testing()
  monitor = MockedFlumeAgentMonitor(daemon, reg, \
      cmf.monitor.flume.json_parser.SAMPLE)
  monitor.refresh_pid(os.getpid(), True)

  # Make two different configs
  conf_dir = make_config_file_function(version, "flume-agent-test")
  conf_dir2 = make_config_file_function(version, "flume-agent")

  try:
    assert_equals(0, len(reg.get_registry()))

    # Start it
    monitor.refresh(conf_dir)
    assert_equals(monitor, reg.get_registry().pop())
    monitor.collect()
    role_update = monitor.get_role_update().to_avro_dict()
    assert_equals('flume-agent-test', role_update['rolename'])
    if agent_name is not None:
      assert_equals(1, len(role_update['additionalRoleAttributes']))
      assert_equals(agent_name,
                    role_update['additionalRoleAttributes']['agentName'])

    # Refresh with a different config
    monitor.refresh(conf_dir2)
    assert_equals(monitor, reg.get_registry().pop())
    assert_equals(1, len(reg.get_registry()))

    monitor.collect()

    # Test metrics and role update
    role_update = monitor.get_role_update().to_avro_dict()

    assert_equals('flume-agent', role_update['rolename'])
    assert_equals('my_flume1', role_update['service_name'])
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_true(metric_dict[schema.TOTAL_CPU_SYSTEM] >= 0.0)

    dir_metrics = role_update['dirUpdates']
    assert_true(len(dir_metrics) == 1, "We configured 1 dirs: %s" % dir_metrics)
    assert_true(os.path.ismount(dir_metrics[0]['mountpoint']))
    assert_equals('/path/to/logs', dir_metrics[0]['path'])
    assert_true(len(role_update["flumeUpdates"]) > 0)

    # Test when the process is not running
    monitor.refresh_pid(None)
    monitor.collect()

    role_update = monitor.get_role_update().to_avro_dict()

    assert_equals('flume-agent', role_update['rolename'])
    assert_equals('my_flume1', role_update['service_name'])
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']

    dir_metrics = role_update['dirUpdates']
    assert_true(len(dir_metrics) == 1, "We configured 1 dirs: %s" % dir_metrics)
    assert_true(os.path.ismount(dir_metrics[0]['mountpoint']))
    assert_equals('/path/to/logs', dir_metrics[0]['path'])

    assert_true(schema.UNEXPECTED_EXITS_COUNT in metric_dict)

    # test safety valve handling
    monitor.refresh_pid(os.getpid(), True)
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    safety_valve_str = "my_flume1_feature_send_flume_update_enabled= FALSE"
    daemon._agent.safety_valve.update_safety_valve(
      dict(host_agent_safety_valve=safety_valve_str))
    role_update = monitor.get_role_update().to_avro_dict()
    assert_equals(0, len(role_update["flumeUpdates"]))

    # Test with a collection error
    monitor.refresh_pid(os.getpid(), True)
    monitor.set_throw_error(IOError("BOOM!"))
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_equals(constants.WEB_METRICS_COLLECTION_COMMUNICATION_FAILURE,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)

    # Test with some invalid JSON
    monitor.set_throw_error(None)
    monitor.set_string("THIS IS BOGUS JSON!")
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_equals(constants.WEB_METRICS_COLLECTION_PARSE_FAILURE,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)

    # Stop it
    monitor.stop()
    assert_equals(0, len(reg.get_registry()))
  finally:
    shutil.rmtree(conf_dir)
    shutil.rmtree(conf_dir2)

OPSAPS_17029 = r"""
{
   "SINK.avro-sink4":{
      "BatchCompleteCount":"74",
      "ConnectionFailedCount":"0",
      "EventDrainAttemptCount":"37000",
      "ConnectionCreatedCount":"1",
      "BatchEmptyCount":"11",
      "Type":"SINK",
      "ConnectionClosedCount":"0",
      "EventDrainSuccessCount":"37000",
      "StopTime":"0",
      "StartTime":"1383774073220",
      "BatchUnderflowCount":"0"
   },
   "SINK.avro-sink3":{
      "BatchCompleteCount":"74",
      "ConnectionFailedCount":"0",
      "EventDrainAttemptCount":"37000",
      "ConnectionCreatedCount":"1",
      "BatchEmptyCount":"11",
      "Type":"SINK",
      "ConnectionClosedCount":"0",
      "EventDrainSuccessCount":"37000",
      "StopTime":"0",
      "StartTime":"1383774073236",
      "BatchUnderflowCount":"0"
   },
   "SINK.avro-sink5":{
      "BatchCompleteCount":"74",
      "ConnectionFailedCount":"0",
      "EventDrainAttemptCount":"37000",
      "ConnectionCreatedCount":"1",
      "BatchEmptyCount":"11",
      "Type":"SINK",
      "ConnectionClosedCount":"0",
      "EventDrainSuccessCount":"37000",
      "StopTime":"0",
      "StartTime":"1383774073223",
      "BatchUnderflowCount":"0"
   },
   "SOURCE.CustomSocketSourceCounter":{
      "QueueSize":"0",
      "OpenConnectionCount":"39",
      "QueueFillPercent":"0.0",
      "Type":"SOURCE",
      "EventAcceptedCount":"373632",
      "StopTime":"0",
      "EventReceivedCount":"373632",
      "StartTime":"1383774073258",
      "QueueCapacity":"100000"
   },
   "SINK.avro-sink1":{
      "BatchCompleteCount":"74",
      "ConnectionFailedCount":"0",
      "EventDrainAttemptCount":"37000",
      "ConnectionCreatedCount":"1",
      "BatchEmptyCount":"11",
      "Type":"SINK",
      "ConnectionClosedCount":"0",
      "EventDrainSuccessCount":"37000",
      "StopTime":"0",
      "StartTime":"1383774073220",
      "BatchUnderflowCount":"0"
   },
   "SINK.avro-sink2":{
      "BatchCompleteCount":"73",
      "ConnectionFailedCount":"0",
      "EventDrainAttemptCount":"36500",
      "ConnectionCreatedCount":"1",
      "BatchEmptyCount":"11",
      "Type":"SINK",
      "ConnectionClosedCount":"0",
      "EventDrainSuccessCount":"36500",
      "StopTime":"0",
      "StartTime":"1383774073245",
      "BatchUnderflowCount":"0"
   },
   "CHANNEL.ch1":{
      "EventPutSuccessCount":"185306",
      "ChannelFillPercentage":"0.0",
      "Type":"CHANNEL",
      "StopTime":"0",
      "EventPutAttemptCount":"185306",
      "ChannelSize":"0",
      "StartTime":"1383774073214",
      "EventTakeSuccessCount":"184500",
      "ChannelCapacity":"2000000",
      "EventTakeAttemptCount":"185366"
   }
}
"""

def test_metrics_parsing():
  def test_parsing(metric_count, config):
    p = cmf.monitor.flume.json_parser.JsonParser(simplejson.load(StringIO(config)))
    count = 0
    for metrics in p.updates.itervalues():
      count = count + len(metrics)
    assert_equals(metric_count, count)

  expected = 0
  for component_metrics in \
        cmf.monitor.flume.json_parser.FLUME_JSON_METRICS.itervalues():
    for metric in component_metrics.itervalues():
      expected = expected + 1
  test_parsing(expected, cmf.monitor.flume.json_parser.SAMPLE)

  # No components; not an error.
  test_parsing(0, "{}")
  test_parsing(0, r"""{"foo": {}}""")
  test_parsing(0, r"""{"foo": {"Type":"SOURCE"}}""")
  test_parsing(0, r"""{"SINK.foo": {"Type":"SOURCE"}}""")
  test_parsing(0, r"""{"SOURCE.": {"Type":"SOURCE"}}""")
  test_parsing(0, r"""{"BAR.foo": {"Type":"BAR"}}""")
  test_parsing(8, r"""{
    "CHANNEL.channel1":{
       "EventPutSuccessCount":"10000",
       "Type":"CHANNEL",
       "StopTime":"0",
       "EventPutAttemptCount":"10011",
       "ChannelSize":"10000",
       "StartTime":"1345147577389",
       "EventTakeSuccessCount":"0",
       "ChannelCapacity":"10000",
       "SomeFutureMetric":"1234",
       "EventTakeAttemptCount":"0"
    }
  }""")
  test_parsing(7, r"""{
    "CHANNEL.channel1":{
       "EventPutSuccessCount":"10000",
       "Type":"CHANNEL",
       "StopTime":"0",
       "EventPutAttemptCount":"10011",
       "ChannelSize":"10000",
       "StartTime":"1345147577389",
       "ChannelCapacity":"10000",
       "EventTakeAttemptCount":"0"
    }
  }""")
  # See the JIRA for more background.
  test_parsing(64, OPSAPS_17029)
