# Copyright (c) 2013 Cloudera, Inc. All rights reserved.

try:
  from cStringIO import StringIO
except:
  from StringIO import StringIO
import os
import shutil
import simplejson
import urllib2

from cmf.monitor import conf, constants, schema
from cmf.monitor.daemon import MonitorDaemon
from cmf.monitor.abstract_monitor import CollectionError
from cmf.monitor.generic import GenericMonitor
from cmf.monitor.generic.hue_adapters import _HueIsAliveCollector
from cmf.monitor.generic.adapter import Adapter
from cmf.monitor.generic.adapter_factory import AdapterFactory
from cmf.monitor.generic.common_adapter import CommonAdapter
from cmf.monitor.generic.kafka_adapters import _TOPICMETRICNAMES_SAMPLE
from cmf.monitor.generic.metric_collectors import WebServerMetricCollector, \
  SampleFileMetricCollector, AbstractMetricCollector
from cmf.monitor.generic.service_defined_metrics import SourceProcessor
from cmf.monitor.generic.service_defined_metrics import ServiceDefinedMetrics
from cmf.monitor.generic.utils import SplittingSourceProcessor, \
    extract_metric_from_json
from cmf.monitor.monitor_properties import ConfParser
from cmf.monitor.role_update import RoleUpdate
from cmf.monitor.proc_metrics_utils import get_process_swap_memory
from cmf.safety_valve import SafetyValve
from cmf_test.monitor import make_config_file, replace_config_file, TestMonitorDaemon, \
  parse_metrics_sample, parse_metrics_for_provided_sample, get_role_update_for_sample, \
  get_role_update_for_provided_sample
from psutil import Process

from mock import Mock

from nose.tools import assert_true, assert_equals, assert_false

_MGMT_SERVICE_VERSION = -1

_METRIC_ACCESSORS = TestMonitorDaemon().metric_accessors

def get_sample_file(filename):
  return open(os.path.join(os.path.dirname(__file__), "../../cmf/monitor/samples/" + filename), 'r').read()

class MockAdapter(Adapter):
  def __init__(self):
    self._metrics_url = "test URL"
    self._metrics_file = None
    self._collector = None
    Adapter.__init__(self, "SOMESERVICE", "SOMEROLE", SafetyValve())

  def get_metrics_url(self, conf):
    return self._metrics_url

  def set_metrics_url(self, metrics_url):
    self._metrics_url = metrics_url

  def parse_metrics_from_url(self, conf, json):
    return []

  def parse_metrics_from_file(self, conf, json):
    extracted_metrics = dict()
    for metric in json.get('metrics', []):
      metricid = metric.get('id', None)
      if metricid is None:
        continue
      value = metric.get('value', None)
      if value is None:
        continue
      extracted_metrics[metricid] = float(value)
    return extracted_metrics

  def add_sample_file_metrics(self,
                              service_version,
                              update,
                              collected_metrics,
                              metric_accessors):
    for metricid, value in collected_metrics.iteritems():
      update.add_metric(metricid, value)

  def add_webserver_metrics(self, conf, update, metrics, accessors):
    pass

  def get_metrics_file(self, conf):
    return self._metrics_file

  def set_metrics_file(self, metrics_file):
    self._metrics_file = metrics_file

  def set_collector(self, collector):
    self._collector = collector

  def get_adapter_specific_collectors(self):
    if self._collector:
      return [self._collector]
    else:
      return []

class MockWebEndPoint(WebServerMetricCollector):

  def __init__(self, section, adapter):
    WebServerMetricCollector.__init__(self, section, adapter, "metric_accessors")
    self._string = "{}"
    self._error = None

  def set_string(self, string):
    self._string = string

  def set_throw_error(self, error):
    self._error = error

  def _open_url(self):
    if self._error:
      raise self._error
    return StringIO(self._string)

class MockSampleFileEndPoint(SampleFileMetricCollector):

  def __init__(self, adapter):
    SampleFileMetricCollector.__init__(self, adapter, "metric_accessors")
    self._error = None
    self._json = None

  def set_throw_error(self, error):
    self._error = error

  def set_json_string(self, json):
    self._json = json

  def _get_json(self, file):
    if self._error is not None:
      raise self._error
    if self._json is not None:
      return StringIO(self._json)
    return SampleFileMetricCollector._get_json(self, file)

  def reset_sample_mtime(self):
    self._last_sample_file_collected_modified_time = 0

class MockGenericMonitor(GenericMonitor):
  def __init__(self, daemon, registry, servicetype, roletype, adapter = MockAdapter()):
    GenericMonitor.__init__(self,
                            servicetype,
                            roletype,
                            daemon,
                            registry,
                            adapter)
    self._url_collector = MockWebEndPoint(roletype.lower(), self._adapter)
    self._sample_file_collector = MockSampleFileEndPoint(self._adapter)

def test_lifecycle():
  daemon = TestMonitorDaemon()
  reg = daemon.get_registry_for_testing()
  adapter = MockAdapter()
  collector = AbstractMetricCollector(adapter)
  collector.on_start_collecting = Mock()
  collector.on_stop_collecting = Mock()
  adapter.set_collector(collector)
  monitor = MockGenericMonitor(daemon, reg, "SOMESERVICE", "SOMEROLE", adapter)

  monitor.ensure_collecting()
  assert_true(collector.on_start_collecting.called)

  monitor.stop()
  assert_true(collector.on_stop_collecting.called)

def test_generic_monitor():
  daemon = TestMonitorDaemon()
  reg = daemon.get_registry_for_testing()
  monitor = MockGenericMonitor(daemon, reg, "SOMESERVICE", "SOMEROLE")
  monitor.refresh_pid(os.getpid(), True)

  # Make two different config files.
  conf_dir = make_config_file(
"""[somerole]
role_name = someservice-somerole-test
service_name = someservice
host = localhost
service_version = 4
collect_interval = 1.0
log_dir = /path/to/logs
""")

  conf_dir2 = make_config_file(
"""[somerole]
role_name = someservice-somerole
service_name = someservice
host = localhost
service_version = 4
collect_interval = 1.0
monitored_directories_with_types = [{"path":"/path/to/mon/dir1","type":"dfs_data_dirs"},{"path":"/path/to/mon/dir2","type": "dfs_data_dirs"}]
log_dir = /path/to/logs
""")

  try:
    assert_equals(0, len(reg.get_registry()))

    # Start it
    monitor.refresh(conf_dir)
    assert_equals(monitor, reg.get_registry().pop())
    monitor.clear_last_reported_timestamp()
    monitor.collect()
    role_update = monitor.get_role_update().to_avro_dict()
    assert_equals('someservice-somerole-test', role_update['rolename'])
    assert_true(len(role_update['dirUpdates']) == 1)
    assert_true(role_update['dirUpdates'][0]['directoryType'] is None)

    # Refresh with a different config
    monitor.refresh(conf_dir2)
    assert_equals(monitor, reg.get_registry().pop())
    assert_equals(1, len(reg.get_registry()))

    # Test metrics and role update
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()

    assert_equals('someservice-somerole', role_update['rolename'])
    assert_equals('someservice', role_update['service_name'])
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']

    # Pick a few to verify
    assert_equals(constants.WEB_METRICS_COLLECTION_GOOD,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)
    assert_true(schema.TOTAL_CPU_USER in metric_dict)
    assert_true(schema.TOTAL_CPU_SYSTEM in metric_dict)
    assert_true(schema.UPTIME in metric_dict)
    if has_swap_memory_field():
      assert_true(schema.MEM_SWAP in metric_dict)
    else:
      assert_true(schema.MEM_SWAP not in metric_dict)

    dir_metrics = role_update['dirUpdates']
    assert_true(len(dir_metrics) == 3, "We configured 3 dirs: %s" % dir_metrics)
    assert_true(os.path.ismount(dir_metrics[0]['mountpoint']))
    assert_equals('/path/to/logs', dir_metrics[0]['path'])
    assert_true(role_update['dirUpdates'][2]['directoryType'] == "dfs_data_dirs")

    assert_true(schema.UNEXPECTED_EXITS_COUNT  in metric_dict)

    # Refresh with an updated config with same file path. The updated config
    # should be picked up because confparse object has a different timestamp.
    replace_config_file(conf_dir2, conf.CLOUDERA_MONITOR_PROPERTIES,
"""[somerole]
role_name = someservice-somerole
service_name = someservice
host = localhost
service_version = 4
collect_interval = 1.0
log_dir = /new/path/to/logs
""")
    monitor.refresh(conf_dir2)
    assert_equals(monitor, reg.get_registry().pop())
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    dir_metrics = role_update['dirUpdates']
    assert_true(len(dir_metrics) == 1, "We configured 1 dirs: %s" % dir_metrics)
    assert_true(os.path.ismount(dir_metrics[0]['mountpoint']))
    assert_equals('/new/path/to/logs', dir_metrics[0]['path'])

    # Test when the process is not running
    monitor.refresh_pid(None)
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()

    assert_equals('someservice-somerole', role_update['rolename'])
    assert_equals('someservice', role_update['service_name'])
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']

    dir_metrics = role_update['dirUpdates']
    assert_true(len(dir_metrics) == 1, "We configured 1 dirs: %s" % dir_metrics)
    assert_true(os.path.ismount(dir_metrics[0]['mountpoint']))
    assert_equals('/new/path/to/logs', dir_metrics[0]['path'])

    # Test with a collection error
    monitor.refresh_pid(os.getpid(), True)
    monitor._url_collector.set_throw_error(IOError("BOOM!"))
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_equals(constants.WEB_METRICS_COLLECTION_COMMUNICATION_FAILURE,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)

    # Test with some invalid JSON
    monitor._url_collector.set_throw_error(None)
    monitor._url_collector.set_string("THIS IS BOGUS JSON!")
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_equals(constants.WEB_METRICS_COLLECTION_PARSE_FAILURE,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)

    monitor.stop()
    assert_equals(0, len(reg.get_registry()))
  finally:
    shutil.rmtree(conf_dir)
    shutil.rmtree(conf_dir2)

def test_adapter_factor():
  assert_false(AdapterFactory().make_adapter("SERVICE",
                                             "ROLE",
                                             SafetyValve()) is None)

def test_generic_sample_file_collection():
  daemon = TestMonitorDaemon()
  reg = daemon.get_registry_for_testing()
  monitor = MockGenericMonitor(daemon, reg, "SOMESERVICE", "SOMEROLE")
  monitor._adapter.set_metrics_url(None)
  monitor.refresh_pid(os.getpid(), True)

  # Make two different config files.
  conf_dir = make_config_file(
"""[somerole]
role_name = someservice-somerole-test
service_name = someservice
host = localhost
service_version = 4
collect_interval = 1.0
log_dir = /path/to/logs
""")

  sample_dir = make_config_file(
"""
{
"metrics": [
  { "id": 66666, "value": -3.14 },
  { "id": 66667, "value": 3.15 }
]
}
""",
filename="sample_file.json"
)

  sample_file = os.path.join(sample_dir, "sample_file.json")
  monitor._adapter.set_metrics_file(sample_file)
  try:
    assert_equals(0, len(reg.get_registry()))

    # Start it
    monitor.refresh(conf_dir)
    assert_equals(monitor, reg.get_registry().pop())

    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    assert_equals('someservice-somerole-test', role_update['rolename'])

    # Test metrics and role update
    monitor._sample_file_collector.reset_sample_mtime()
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()

    assert_equals('someservice-somerole-test', role_update['rolename'])
    assert_equals('someservice', role_update['service_name'])
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']

    # Pick a few to verify
    assert_false(schema.WEB_METRICS_COLLECTION_STATUS in metric_dict)
    assert_false(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)
    assert_true(schema.TOTAL_CPU_USER in metric_dict)
    assert_true(schema.TOTAL_CPU_SYSTEM in metric_dict)
    assert_true(schema.UPTIME in metric_dict)
    if has_swap_memory_field():
      assert_true(schema.MEM_SWAP in metric_dict)
    else:
      assert_true(schema.MEM_SWAP not in metric_dict)
    assert_equals(-3.14, metric_dict.get(66666))
    assert_equals(3.15, metric_dict.get(66667))

    dir_metrics = role_update['dirUpdates']
    assert_true(len(dir_metrics) == 1, "We configured 1 dirs: %s" % dir_metrics)
    assert_true(os.path.ismount(dir_metrics[0]['mountpoint']))
    assert_equals('/path/to/logs', dir_metrics[0]['path'])

    assert_true(schema.UNEXPECTED_EXITS_COUNT  in metric_dict)

    # Test with a collection error
    monitor.refresh_pid(os.getpid(), True)
    monitor._sample_file_collector.set_throw_error(IOError("BOOM!"))
    monitor._sample_file_collector.reset_sample_mtime()
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_false(schema.WEB_METRICS_COLLECTION_STATUS in metric_dict)
    assert_false(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)
    results, timestamp, duration = \
      monitor._sample_file_collector.metrics_tuple
    assert_true(isinstance(results, CollectionError))
    assert_equals(constants.FILE_METRICS_COLLECTION_UNKNOWN,
                  results.code)

    # Test with some invalid JSON
    monitor._sample_file_collector.set_throw_error(None)
    monitor._sample_file_collector.set_json_string("THIS IS BOGUS JSON!")
    monitor._sample_file_collector.reset_sample_mtime()
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_false(schema.WEB_METRICS_COLLECTION_STATUS in metric_dict)
    assert_false(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)
    results, timestamp, duration = \
      monitor._sample_file_collector.metrics_tuple
    assert_true(isinstance(results, CollectionError))
    assert_equals(constants.FILE_METRICS_COLLECTION_PARSE_FAILURE,
                  results.code)

    # Test with two collection end-points.
    monitor._sample_file_collector.set_json_string(None)
    monitor._adapter.set_metrics_url("test URL")
    # "Refresh" the configuration
    monitor._last_conf_timestamp = None
    monitor._sample_file_collector.reset_sample_mtime()
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_equals(-3.14, metric_dict.get(66666))
    assert_equals(3.15, metric_dict.get(66667))
    assert_equals(constants.WEB_METRICS_COLLECTION_GOOD,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)
    monitor.stop()
    assert_equals(0, len(reg.get_registry()))
  finally:
    shutil.rmtree(conf_dir)
    shutil.rmtree(sample_dir)

def test_service_defined_metrics():
  conf_dir = make_config_file(
"""
{
  "ECHO" : {
    "32768" : {
      "source" : ""
    },
    "32769" : {
      "counter" : true,
      "source" : ""
    },
    "32770" : {
      "source" : ""
    }
  },
  "ECHO-ECHO_MASTER_SERVER" : {
    "32770" : {
      "source" : ""
    },
    "32771" : {
      "source" : ""
    },
    "32772" : {
      "counter" : true,
      "source" : ""
    }
  },
  "ECHO-ECHO_WEBSERVER" : {
    "32770" : {
      "source" : ""
    },
    "32773" : {
      "source" : ""
    },
    "32774" : {
      "counter" : true,
      "source" : ""
    }
  },
  "ECHO_ENTITY_TYPE_ONE" : {
    "32770" : {
      "source" : ""
    },
    "32775" : {
      "source" : ""
    },
    "32776" : {
      "counter" : true,
      "source" : ""
    }
  },
  "ECHO_ENTITY_TYPE_TWO" : {
    "32770" : {
      "source" : ""
    },
    "32777" : {
      "source" : ""
    },
    "32778" : {
      "counter" : true,
      "source" : ""
    }
  }
}
""", conf.SERVICE_METRICS_PROPERTIES)
  try:
    service_defined_metrics = ServiceDefinedMetrics(
      os.path.join(conf_dir, conf.SERVICE_METRICS_PROPERTIES),
      SourceProcessor())
    assert_true(service_defined_metrics.is_counter(32778))
    assert_false(service_defined_metrics.is_counter(32773))
    assert_false(service_defined_metrics.is_counter(100))
  finally:
    shutil.rmtree(conf_dir)

def has_swap_memory_field():
  return 'swap' in Process(os.getpid()).memory_maps(grouped=True)[0]._fields

def garbage_memory_maps(process):
  # get_process_swap_memory() should catch exception
  return list(xrange(10))

def error_memory_maps(process):
  raise Exception("BOOM")

def test_proc_swap_memory():
  process = Process(os.getpid())
  # test memory_maps fetcher is robust to 'swap' not being present
  if not has_swap_memory_field():
    assert_true(get_process_swap_memory(process) is None)
  assert_true(get_process_swap_memory(process, garbage_memory_maps) is None)
  assert_true(get_process_swap_memory(process, error_memory_maps) is None)

def test_metric_extraction():
  sample = r"""
{
  "jvm" : {
    "vm" : {
      "name" : "Java HotSpot(TM) 64-Bit Server VM"
    }
  }
}
"""
  json = simplejson.load(StringIO(sample))
  assert_equals(None, extract_metric_from_json(json, ["jvm", "vm", "name"]))

def test_make_common_adapter():
  adapter = AdapterFactory().make_adapter("some-service",
                                          "some-role",
                                          SafetyValve())
  isinstance(adapter, CommonAdapter)

def test_zookeeper_server_adapter():
  conf_dir = make_config_file(
"""[server]
role_name = zookeeper-server-1
service_name = my_zookeeper1
service_version = 4
host = localhost
log_dir = /path/to/logs
data_dir = /path/to/data
data_log_dir = /path/to/data/logs
collect_interval = 1.0
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("ZOOKEEPER",
                                            "SERVER",
                                            SafetyValve())
    assert_true(adapter is not None)
    directories = adapter.get_monitored_directories(conf_parser)
    assert_true(len(directories) == 3)
    assert_true(dict(path='/path/to/logs', type=None) in directories)
    assert_true(dict(path='/path/to/data', type=None) in directories)
    assert_true(dict(path='/path/to/data/logs', type=None) in directories)
  finally:
    shutil.rmtree(conf_dir)

def test_mgmt_activitymonitor_adapter():
  conf_dir = make_config_file(
"""[activitymonitor]
role_name = my_activitymontor
service_name = my_mgmt1
host = localhost
firehose_debug_port = 8088
log_dir = /path/to/logs
collect_interval = 1.0
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("MGMT",
                                            "ACTIVITYMONITOR",
                                            SafetyValve())
    assert_true(adapter is not None)
    directories = adapter.get_monitored_directories(conf_parser)
    assert_true(len(directories) == 1)
    assert_true(dict(path='/path/to/logs', type=None) in directories)
    assert_equals('http://localhost:8088/jmx',
                  adapter.get_metrics_url(conf_parser))
    assert_equals(16, len(parse_metrics_sample(
          adapter, conf_parser, _MGMT_SERVICE_VERSION)))
  finally:
    shutil.rmtree(conf_dir)

def test_mgmt_servicemonitor_adapter():
  conf_dir = make_config_file(
"""[servicemonitor]
role_name = my_servicemonitor
service_name = my_mgmt1
host = localhost
firehose_debug_port = 8086
log_dir = /path/to/logs
ldb_dir = /path/to/ldb
collect_interval = 1.0
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("MGMT",
                                            "SERVICEMONITOR",
                                            SafetyValve())
    assert_true(adapter is not None)
    directories = adapter.get_monitored_directories(conf_parser)
    assert_true(len(directories) == 2)
    assert_true(dict(path='/path/to/logs', type=None) in directories)
    assert_true(dict(path='/path/to/ldb', type=None) in directories)
    assert_equals('http://localhost:8086/jmx',
                  adapter.get_metrics_url(conf_parser))
    assert_equals(17, len(parse_metrics_sample(
          adapter, conf_parser, _MGMT_SERVICE_VERSION)))
    metrics_path = os.path.join(conf_dir, conf.SERVICE_METRICS_PROPERTIES)
    file(metrics_path, 'w').write(
"""{
  "SERVICEMONITOR" : {
    "32838" : {
      "counter" : true,
      "source" : "com.cloudera.cmon.tstore.db:type=TimeSeriesEntityCache,name=entitiesDeleted::Count"
    }
  }
}
""")
    adapter.read_service_defined_metrics(metrics_path)
    assert_equals(18, len(parse_metrics_sample(
          adapter, conf_parser, _MGMT_SERVICE_VERSION)))
  finally:
    shutil.rmtree(conf_dir)

def test_mgmt_hostmonitor_adapter():
  conf_dir = make_config_file(
"""[hostmonitor]
role_name = my_hostmonitor
service_name = my_mgmt1
host = localhost
firehose_debug_port = 8086
log_dir = /path/to/logs
ldb_dir = /path/to/ldb
collect_interval = 1.0
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("MGMT",
                                            "HOSTMONITOR",
                                            SafetyValve())
    assert_true(adapter is not None)
    directories = adapter.get_monitored_directories(conf_parser)
    assert_true(len(directories) == 2)
    assert_true(dict(path='/path/to/logs', type=None) in directories)
    assert_true(dict(path='/path/to/ldb', type=None) in directories)
    assert_equals('http://localhost:8086/jmx',
                  adapter.get_metrics_url(conf_parser))
    assert_equals(17, len(parse_metrics_sample(
          adapter, conf_parser, _MGMT_SERVICE_VERSION)))
    metrics_path = os.path.join(conf_dir, conf.SERVICE_METRICS_PROPERTIES)
    file(metrics_path, 'w').write(
"""{
  "HOSTMONITOR" : {
    "32838" : {
      "counter" : true,
      "source" : "com.cloudera.cmon.tstore.db:type=TimeSeriesEntityCache,name=entitiesDeleted::Count"
    }
  }
}
""")
    adapter.read_service_defined_metrics(metrics_path)
    assert_equals(18, len(parse_metrics_sample(
          adapter, conf_parser, _MGMT_SERVICE_VERSION)))
  finally:
    shutil.rmtree(conf_dir)


def test_mgmt_eventserver_adapter():
  conf_dir = make_config_file(
"""[eventserver]
role_name = my_eventserver
service_name = my_mgmt1
host = localhost
log_dir = /path/to/logs
index_dir = /path/to/index
collect_interval = 1.0
eventserver_debug_port = 8086
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("MGMT",
                                            "EVENTSERVER",
                                            SafetyValve())
    assert_true(adapter is not None)
    directories = adapter.get_monitored_directories(conf_parser)
    assert_true(len(directories) == 2)
    assert_true(dict(path='/path/to/logs', type=None) in directories)
    assert_true(dict(path='/path/to/index', type=None) in directories)
    assert_equals('http://localhost:8086/jmx',
                  adapter.get_metrics_url(conf_parser))
    assert_equals(12, len(parse_metrics_sample(
          adapter, conf_parser, _MGMT_SERVICE_VERSION)))
  finally:
    shutil.rmtree(conf_dir)

def test_mgmt_navigator_adapter():
  conf_dir = make_config_file(
"""[navigator]
role_name = my_navigator
service_name = my_mgmt
host = localhost
collect_interval = 1.0
log_dir = /path/to/logs
navigator_debug_port = 8089
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("MGMT",
                                            "NAVIGATOR",
                                            SafetyValve())
    assert_true(adapter is not None)
    directories = adapter.get_monitored_directories(conf_parser)
    assert_true(len(directories) == 1)
    assert_true(dict(path='/path/to/logs', type=None) in directories)
    assert_equals('http://localhost:8089/jmx',
                  adapter.get_metrics_url(conf_parser))
    assert_equals(12, len(parse_metrics_sample(
          adapter, conf_parser, _MGMT_SERVICE_VERSION)))
  finally:
    shutil.rmtree(conf_dir)

def test_mgmt_nms_adapter():
  conf_dir = make_config_file(
"""[navigatormetaserver]
role_name = my_navigatormetaserver
service_name = my_mgmt
host = localhost
collect_interval = 1.0
log_dir = /path/to/logs
navigator_server_port = 7187
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("MGMT",
                                            "NAVIGATORMETASERVER",
                                            SafetyValve())
    assert_true(adapter is not None)
    directories = adapter.get_monitored_directories(conf_parser)
    assert_true(len(directories) == 1)
    assert_true(dict(path='/path/to/logs', type=None) in directories)
    assert_equals('http://localhost:7187/debug/jmx',
                  adapter.get_metrics_url(conf_parser))
    assert_equals(9, len(parse_metrics_sample(
          adapter, conf_parser, _MGMT_SERVICE_VERSION)))
  finally:
    shutil.rmtree(conf_dir)

def test_mgmt_telemetrypublisher_adapter():
  conf_dir = make_config_file(
"""[telemetrypublisher]
role_name = my_telemetrypublisher
service_name = my_mgmt
host = localhost
collect_interval = 1.0
log_dir = /path/to/logs
telemetry_publisher_debug_port = 10111
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("MGMT",
                                            "TELEMETRYPUBLISHER",
                                            SafetyValve())
    assert_true(adapter is not None)
    directories = adapter.get_monitored_directories(conf_parser)
    assert_true(len(directories) == 1)
    assert_true(dict(path='/path/to/logs', type=None) in directories)
    assert_equals('http://localhost:10111/jmx',
                  adapter.get_metrics_url(conf_parser))
    assert_equals(9, len(parse_metrics_sample(
          adapter, conf_parser, _MGMT_SERVICE_VERSION)))
  finally:
    shutil.rmtree(conf_dir)

def test_mgmt_reportsmanager_adapter():
  conf_dir = make_config_file(
"""[reportsmanager]
role_name = %s
service_name = my_mgmt1
host = localhost
log_dir = /path/to/logs
scratch_dir = /path/to/scratch
collect_interval = 1.0
headlamp_debug_port = 8083
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("MGMT",
                                            "REPORTSMANAGER",
                                            SafetyValve())
    assert_true(adapter is not None)
    directories = adapter.get_monitored_directories(conf_parser)
    assert_true(len(directories) == 2)
    assert_true(dict(path='/path/to/logs', type=None) in directories)
    assert_true(dict(path='/path/to/scratch', type=None) in directories)
    assert_equals('http://localhost:8083/jmx',
                  adapter.get_metrics_url(conf_parser))
    assert_equals(7, len(parse_metrics_sample(
          adapter, conf_parser, _MGMT_SERVICE_VERSION)))
  finally:
    shutil.rmtree(conf_dir)

def get_missing_metrics(metrics, sources):
  msg = "missing metrics:\n"
  for metric_id, source in sources.iteritems():
    if not metric_id in metrics:
      msg += "%s::%s\n" % (metric_id, source)
  return msg

def test_monitored_directory_types_error_handling():
  # test invalid JSON for monitored_directories_with_types
  conf_dir = make_config_file(
"""[datanode]
role_name = hdfs-DN
service_name = my_hdfs
service_version = 5
host = localhost
dfs_data_dirs = /path/to/data/1,/path/to/data/2
log_dir = /path/to/logs
monitored_directories_with_types = {[{"path":"/path/to/mon/dir1","type":"dfs_data_dirs,BOOM}
collect_interval = 1.0
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("HDFS", "DATANODE", SafetyValve())
    assert_true(adapter is not None)
    directories = adapter.get_monitored_directories(conf_parser)
    assert_true(len(directories) == 3)
  finally:
    shutil.rmtree(conf_dir)

def test_yarn_resourcemanager_adapter():
  conf_dir = make_config_file(
"""[resourcemanager]
role_name = my_resourcemanager
service_name = my_yarn
service_version = 4
host = localhost
log_dir = /path/to/logs
collect_interval = 1.0
resourcemanager_webserver_port = 8088
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("YARN",
                                            "RESOURCEMANAGER",
                                            SafetyValve())
    assert_true(adapter is not None)
    directories = adapter.get_monitored_directories(conf_parser)
    assert_true(len(directories) == 1)
    assert_true(dict(path='/path/to/logs', type=None) in directories)
    assert_equals('http://localhost:8088/jmx',
                  adapter.get_metrics_url(conf_parser))
    # Semi-gross, but it keeps the test concise.
    conf_parser._cp.set(adapter.section, "service_version", "5")
    assert_equals(91, len(parse_metrics_sample(
          adapter, conf_parser, constants.SERVICE_VERSION_CDH5)))
  finally:
    shutil.rmtree(conf_dir)

  # test HTTPS URL construction
  conf_dir = make_config_file(
"""[resourcemanager]
role_name = my_resourcemanager
service_name = my_yarn
service_version = 4
host = localhost
log_dir = /path/to/logs
collect_interval = 1.0
resourcemanager_webserver_port = 8088
resourcemanager_webserver_https_port = 8090
use_ssl = true
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("YARN",
                                            "RESOURCEMANAGER",
                                            SafetyValve())
    assert_equals('https://localhost:8090/jmx',
                  adapter.get_metrics_url(conf_parser))
  finally:
    shutil.rmtree(conf_dir)

def test_yarn_jobhistory_adapter():
  conf_dir = make_config_file(
"""[jobhistory]
role_name = my_jobhistory
service_name = my_yarn
service_version = 4
host = localhost
webserver_port = 1234
log_dir = /path/to/logs
collect_interval = 1.0
mapreduce_jobhistory_webapp_address = 8081
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("YARN", "JOBHISTORY", SafetyValve())
    assert_true(adapter is not None)
    directories = adapter.get_monitored_directories(conf_parser)
    assert_true(len(directories) == 1)
    assert_true(dict(path='/path/to/logs', type=None) in directories)
    assert_equals('http://localhost:8081/jmx',
                  adapter.get_metrics_url(conf_parser))
    # Semi-gross, but it keeps the test concise.
    conf_parser._cp.set(adapter.section, "service_version", "5")
    assert_equals(47, len(parse_metrics_sample(
          adapter, conf_parser, constants.SERVICE_VERSION_CDH5)))
  finally:
    shutil.rmtree(conf_dir)

  # test HTTPS URL construction
  conf_dir = make_config_file(
"""[jobhistory]
role_name = my_jobhistory
service_name = my_yarn
service_version = 4
host = localhost
log_dir = /path/to/logs
collect_interval = 1.0
mapreduce_jobhistory_webapp_address = 8081
mapreduce_jobhistory_webapp_https_address = 8082
use_ssl = true
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("YARN", "JOBHISTORY", SafetyValve())
    assert_equals('https://localhost:8082/jmx',
                  adapter.get_metrics_url(conf_parser))
  finally:
    shutil.rmtree(conf_dir)

def test_yarn_nodemanager_adapter():
  conf_dir = make_config_file(
"""[nodemanager]
role_name = my_nodemanager
service_name = my_yarn
service_version = 4
host = localhost
webserver_port = 1234
log_dir = /path/to/logs
collect_interval = 1.0
nodemanager_webserver_port = 8010
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("YARN",
                                            "NODEMANAGER",
                                            SafetyValve())
    assert_true(adapter is not None)
    directories = adapter.get_monitored_directories(conf_parser)
    assert_true(len(directories) == 1)
    assert_true(dict(path='/path/to/logs', type=None) in directories)
    assert_equals('http://localhost:8010/jmx',
                  adapter.get_metrics_url(conf_parser))
    # Semi-gross, but it keeps the test concise.
    conf_parser._cp.set(adapter.section, "service_version", "5")
    assert_equals(62, len(parse_metrics_sample(
          adapter, conf_parser, constants.SERVICE_VERSION_CDH5)))
  finally:
    shutil.rmtree(conf_dir)

  # test HTTPS URL construction
  conf_dir = make_config_file(
"""[nodemanager]
role_name = my_nodemanager
service_name = my_yarn
service_version = 4
host = localhost
log_dir = /path/to/logs
collect_interval = 1.0
nodemanager_webserver_port = 8010
nodemanager_webserver_https_port = 8020
use_ssl = true
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("YARN",
                                            "NODEMANAGER",
                                            SafetyValve())
    assert_equals('https://localhost:8020/jmx',
                  adapter.get_metrics_url(conf_parser))
  finally:
    shutil.rmtree(conf_dir)

def test_kafka_broker_adapter():
  conf_dir = make_config_file(
"""[kafka_broker]
role_name = my_kafka_broker
service_name = my_kafka
service_version = 5
service_release = 5.4.0
host = localhost
log_dir = /path/to/logs
collect_interval = 1.0
""")
  metrics_path = os.path.join(conf_dir, conf.SERVICE_METRICS_PROPERTIES)
  file(metrics_path, 'w').write(
"""{
  "KAFKA-KAFKA_BROKER" : {
    "32772" : {
      "source" : "kafka.server.ReplicaManager::LeaderCount::value"
    }
  },
  "KAFKA_BROKER_TOPIC" : {
    "32773" : {
      "source" : "BytesInPerSec::mean"
    }
  },
  "KAFKA_BROKER_TOPIC_PARTITION" : {
    "60388" : {
      "counter" : true,
      "source" : "BytesInPerSec::count"
    }
  },
  "KAFKA_PRODUCER" : {
    "60398" : {
      "counter" : true,
      "source" : "MessagesInPerSec::count"
    }
  },
  "KAFKA_REPLICA" : {
    "32774" : {
      "source" : "Size::value"
    },
    "32775" : {
      "source" : "NumLogSegments::value"
    },
    "60377" : {
      "source" : "ReplicasCount::value"
    },
    "60378" : {
      "source" : "UnderReplicated::value"
    }
  }
}
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    safety_valve = SafetyValve()
    adapter = AdapterFactory().make_adapter("KAFKA",
                                            "KAFKA_BROKER",
                                            safety_valve)
    assert_true(adapter is not None)
    adapter.read_service_defined_metrics(metrics_path)

    # Older Kafka brokers do not have the topic metric names servlet, so we
    # mimic them returning a 404.
    def _return_topicmetricnames_sample(url):
      raise urllib2.HTTPError(url, 404, "Not found", dict(), None)
    adapter._url_open_function = _return_topicmetricnames_sample

    directories = adapter.get_monitored_directories(conf_parser)
    assert_true(len(directories) == 1)
    assert_true(dict(path='/path/to/logs', type=None) in directories)
    # We have no kafka-monitor.properties file.
    assert_equals(None, adapter.get_metrics_url(conf_parser))

    kafka_properties_path = os.path.join(conf_dir, "kafka-monitoring.properties")
    file(kafka_properties_path, 'w').write(
"""kafka.http.metrics.host=0.0.0.0
kafka.http.metrics.port=24042
monitoring.enabled=false
""")
    # We have a kafka-monitor.properties file, but monitoring is off.
    assert_equals(None, adapter.get_metrics_url(conf_parser))

    file(kafka_properties_path, 'w').write(
"""kafka.http.metrics.host=0.0.0.0
kafka.http.metrics.port=24042
monitoring.enabled=true
""")
    assert_equals('http://0.0.0.0:24042/api/metrics',
                  adapter.get_metrics_url(conf_parser))

    version = constants.SERVICE_VERSION_CDH5
    sample = adapter.get_metrics_sample(version)
    metrics = adapter.parse_metrics_from_url(conf_parser,
                                    simplejson.load(StringIO(sample)))
    update = RoleUpdate(name="whatever",
                        roletype=adapter.role_type,
                        servicetype=adapter.service_type,
                        service_name="whatever")
    adapter.add_webserver_metrics(version,
                                  update,
                                  metrics,
                                  accessors=_METRIC_ACCESSORS)
    role_metrics = { }
    for metric in update.to_avro_dict()['metrics']:
      role_metrics[metric['id']] = metric['value']
    assert_equals(1, len(role_metrics))

    entity_type_updates = update.to_avro_dict()['entityTypeUpdates']
    assert_equals(3, len(entity_type_updates))

    broker_topic_update = entity_type_updates[0]
    assert_equals('KAFKA_BROKER_TOPIC', broker_topic_update['entityType'])
    assert_equals(True, broker_topic_update['deleteOldEntities'])
    assert_equals(True, broker_topic_update['createNewEntities'])

    broker_topic_entities = broker_topic_update['entities']
    assert_equals(5, len(broker_topic_entities))
    for topic in broker_topic_entities:
      assert_equals('my_kafka_broker', topic['attributes']['roleName'])
      assert_equals('my_kafka', topic['attributes']['serviceName'])
      assert_true('kafkaTopicName' in topic['attributes'])
      assert_equals(1, len(topic['metrics']))

    replica_update = entity_type_updates[1]
    assert_equals('KAFKA_REPLICA', replica_update['entityType'])
    assert_equals(True, replica_update['deleteOldEntities'])
    replica_entities = replica_update['entities']
    assert_equals(16, len(replica_entities))
    for replica in replica_entities:
      assert_equals('my_kafka_broker', replica['attributes']['roleName'])
      assert_equals('my_kafka', replica['attributes']['serviceName'])
      assert_true('kafkaTopicName' in replica['attributes'])
      assert_true('kafkaPartitionId' in replica['attributes'])
      assert_equals(2, len(replica['metrics']))

    client_update = entity_type_updates[2]
    assert_equals('KAFKA_PRODUCER', client_update['entityType'])
    assert_equals(False, client_update['deleteOldEntities'])
    client_entities = client_update['entities']
    assert_equals(1, len(client_entities))
    for client in client_entities:
        assert_equals('my_kafka_broker', client['attributes']['roleName'])
        assert_equals('my_kafka', client['attributes']['serviceName'])
        assert_true('kafkaClientId' in client['attributes'])
        assert_true('kafkaTopicName' in client['attributes'])
        assert_true('kafkaPartitionId' in client['attributes'])
        assert_equals(1, len(client['metrics']))

    safety_valve_str = "kafka-1_broker_topic_partition_metrics_for_smm_enabled=true"
    safety_valve.update_safety_valve(dict(host_agent_safety_valve=safety_valve_str))
    version = constants.SERVICE_VERSION_CDH5
    sample = adapter.get_metrics_sample(version)
    metrics = adapter.parse_metrics_from_url(conf_parser,
                                    simplejson.load(StringIO(sample)))
    update = RoleUpdate(name="whatever",
                        roletype=adapter.role_type,
                        servicetype=adapter.service_type,
                        service_name="kafka-1")
    adapter.add_webserver_metrics(version,
                                  update,
                                  metrics,
                                  accessors=_METRIC_ACCESSORS)
    role_metrics = { }
    for metric in update.to_avro_dict()['metrics']:
      role_metrics[metric['id']] = metric['value']
    assert_equals(1, len(role_metrics))

    entity_type_updates = update.to_avro_dict()['entityTypeUpdates']
    assert_equals(4, len(entity_type_updates))

    broker_topic_update = entity_type_updates[0]
    assert_equals('KAFKA_BROKER_TOPIC', broker_topic_update['entityType'])
    assert_equals(True, broker_topic_update['deleteOldEntities'])
    assert_equals(True, broker_topic_update['createNewEntities'])

    broker_topic_entities = broker_topic_update['entities']
    assert_equals(5, len(broker_topic_entities))
    for topic in broker_topic_entities:
      assert_equals('my_kafka_broker', topic['attributes']['roleName'])
      assert_equals('my_kafka', topic['attributes']['serviceName'])
      assert_true('kafkaTopicName' in topic['attributes'])
      assert_equals(1, len(topic['metrics']))

    replica_update = entity_type_updates[1]
    assert_equals('KAFKA_REPLICA', replica_update['entityType'])
    assert_equals(True, replica_update['deleteOldEntities'])
    replica_entities = replica_update['entities']
    assert_equals(16, len(replica_entities))
    for replica in replica_entities:
      assert_equals('my_kafka_broker', replica['attributes']['roleName'])
      assert_equals('my_kafka', replica['attributes']['serviceName'])
      assert_true('kafkaTopicName' in replica['attributes'])
      assert_true('kafkaPartitionId' in replica['attributes'])
      assert_equals(2, len(replica['metrics']))

    broker_topic_partition_update = entity_type_updates[2]
    assert_equals('KAFKA_BROKER_TOPIC_PARTITION', broker_topic_partition_update['entityType'])
    assert_equals(True, broker_topic_partition_update['deleteOldEntities'])
    broker_topic_partition_entities = broker_topic_partition_update['entities']
    assert_equals(1, len(broker_topic_partition_entities))
    for topic in broker_topic_partition_entities:
        assert_equals('my_kafka_broker', topic['attributes']['roleName'])
        assert_equals('my_kafka', topic['attributes']['serviceName'])
        assert_true('kafkaTopicName' in topic['attributes'])
        assert_true('kafkaPartitionId' in topic['attributes'])
        assert_equals(1, len(topic['metrics']))

    client_update = entity_type_updates[3]
    assert_equals('KAFKA_PRODUCER', client_update['entityType'])
    assert_equals(False, client_update['deleteOldEntities'])
    client_entities = client_update['entities']
    assert_equals(1, len(client_entities))
    for client in client_entities:
        assert_equals('my_kafka_broker', client['attributes']['roleName'])
        assert_equals('my_kafka', client['attributes']['serviceName'])
        assert_true('kafkaClientId' in client['attributes'])
        assert_true('kafkaTopicName' in client['attributes'])
        assert_true('kafkaPartitionId' in client['attributes'])
        assert_equals(1, len(client['metrics']))

    safety_valve_str = \
      "kafka-1_broker-1_feature_send_broker_topic_entity_update_enabled= FALSE"
    safety_valve.update_safety_valve(
      dict(host_agent_safety_valve=safety_valve_str))
    version = constants.SERVICE_VERSION_CDH5
    sample = adapter.get_metrics_sample(version)
    metrics = adapter.parse_metrics_from_url(conf_parser,
                                    simplejson.load(StringIO(sample)))
    adapter.add_webserver_metrics(version,
                                  update,
                                  metrics,
                                  accessors=_METRIC_ACCESSORS)
    update = RoleUpdate(name="broker-1",
                        roletype=adapter.role_type,
                        servicetype=adapter.service_type,
                        service_name="kafka-1")
    adapter.add_webserver_metrics(version,
                                  update,
                                  metrics,
                                  accessors=_METRIC_ACCESSORS)
    entity_type_updates = update.to_avro_dict()['entityTypeUpdates']
    assert_equals(2, len(entity_type_updates))
    broker_topic_update = entity_type_updates[0]
    assert_equals('KAFKA_REPLICA', broker_topic_update['entityType'])
    client_update = entity_type_updates[1]
    assert_equals('KAFKA_PRODUCER', client_update['entityType'])

    safety_valve_str = \
      "kafka-1_feature_send_replica_entity_update_enabled= FALSE"
    safety_valve.update_safety_valve(
      dict(host_agent_safety_valve=safety_valve_str))
    version = constants.SERVICE_VERSION_CDH5
    sample = adapter.get_metrics_sample(version)
    metrics = adapter.parse_metrics_from_url(conf_parser,
                                    simplejson.load(StringIO(sample)))
    adapter.add_webserver_metrics(version,
                                  update,
                                  metrics,
                                  accessors=_METRIC_ACCESSORS)
    update = RoleUpdate(name="broker-1",
                        roletype=adapter.role_type,
                        servicetype=adapter.service_type,
                        service_name="kafka-1")
    adapter.add_webserver_metrics(version,
                                  update,
                                  metrics,
                                  accessors=_METRIC_ACCESSORS)
    entity_type_updates = update.to_avro_dict()['entityTypeUpdates']
    assert_equals(2, len(entity_type_updates))
    broker_topic_update = entity_type_updates[0]
    assert_equals('KAFKA_BROKER_TOPIC', broker_topic_update['entityType'])
    client_update = entity_type_updates[1]
    assert_equals('KAFKA_PRODUCER', client_update['entityType'])

    # Use the safety valuve to force use of the metric names servlet. It still
    # throws a 404, so we should get no entity updates.
    safety_valve_str = "kafka_topicmetricnames_policy = force"
    safety_valve.update_safety_valve(
      dict(host_agent_safety_valve=safety_valve_str))

    metrics = adapter.parse_metrics_from_url(conf_parser,
                                    simplejson.load(StringIO(sample)))
    adapter.add_webserver_metrics(version,
                                  update,
                                  metrics,
                                  accessors=_METRIC_ACCESSORS)
    update = RoleUpdate(name="broker-1",
                        roletype=adapter.role_type,
                        servicetype=adapter.service_type,
                        service_name="kafka-1")
    adapter.add_webserver_metrics(version,
                                  update,
                                  metrics,
                                  accessors=_METRIC_ACCESSORS)

    entity_type_updates = update.to_avro_dict()['entityTypeUpdates']
    assert_equals(3, len(entity_type_updates))

    broker_topic_update = entity_type_updates[0]
    assert_equals('KAFKA_BROKER_TOPIC', broker_topic_update['entityType'])
    assert_equals(False, broker_topic_update['deleteOldEntities'])
    broker_topic_entities = broker_topic_update['entities']
    assert_equals(0, len(broker_topic_entities))

    replica_update = entity_type_updates[1]
    assert_equals('KAFKA_REPLICA', replica_update['entityType'])
    assert_equals(False, replica_update['deleteOldEntities'])
    replica_entities = replica_update['entities']
    assert_equals(0, len(replica_entities))

    client_update = entity_type_updates[2]
    assert_equals('KAFKA_PRODUCER', client_update['entityType'])
    assert_equals(False, client_update['deleteOldEntities'])
    client_entities = client_update['entities']
    assert_equals(0, len(client_entities))

    # Let's add a broker ID and make sure it is placed in the update. We have
    # to write an entire new set files to make sure we're 5.5.0.
    safety_valve.update_safety_valve(dict(host_agent_safety_valve=""))
    conf_dir = make_config_file(
"""[kafka_broker]
role_name = my_kafka_broker
service_name = my_kafka
service_version = 5
service_release = 5.5.0
host = localhost
log_dir = /path/to/logs
collect_interval = 1.0
""")
    kafka_properties_path = os.path.join(conf_dir, "kafka-monitoring.properties")
    file(kafka_properties_path, 'w').write(
"""kafka.http.metrics.host=0.0.0.0
kafka.http.metrics.port=24042
monitoring.enabled=true
broker.id=104
""")
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    def _return_topicmetricnames_sample(url):
      return StringIO(_TOPICMETRICNAMES_SAMPLE)
    adapter._url_open_function = _return_topicmetricnames_sample
    metrics = adapter.parse_metrics_from_url(conf_parser,
                                             simplejson.load(StringIO(sample)))
    update = RoleUpdate(name="whatever",
                        roletype=adapter.role_type,
                        servicetype=adapter.service_type,
                        service_name="whatever")
    adapter.add_webserver_metrics(version,
                                  update,
                                  metrics,
                                  accessors=_METRIC_ACCESSORS)
    additional_attributes = update.to_avro_dict()['additionalRoleAttributes']
    assert_equals("104", additional_attributes['kafkaBrokerId'])

    entity_type_updates = update.to_avro_dict()['entityTypeUpdates']
    assert_equals(3, len(entity_type_updates))

    broker_topic_update = entity_type_updates[0]
    assert_equals('KAFKA_BROKER_TOPIC', broker_topic_update['entityType'])
    broker_topic_entities = broker_topic_update['entities']
    assert_equals(5, len(broker_topic_entities))

    replica_update = entity_type_updates[1]
    assert_equals('KAFKA_REPLICA', replica_update['entityType'])
    replica_entities = replica_update['entities']
    assert_equals(16, len(replica_entities))

    client_update = entity_type_updates[2]
    assert_equals('KAFKA_PRODUCER', client_update['entityType'])
    client_entities = client_update['entities']
    assert_equals(1, len(client_entities))

    def _return_topicmetricnames_sample(url):
      raise urllib2.HTTPError(url, 503, "BOOM!", dict(), None)
    adapter._url_open_function = _return_topicmetricnames_sample
    update = RoleUpdate(name="whatever",
                        roletype=adapter.role_type,
                        servicetype=adapter.service_type,
                        service_name="whatever")
    metrics = adapter.parse_metrics_from_url(conf_parser,
                                             simplejson.load(StringIO(sample)))
    adapter.add_webserver_metrics(version,
                                  update,
                                  metrics,
                                  accessors=_METRIC_ACCESSORS)
    additional_attributes = update.to_avro_dict()['additionalRoleAttributes']
    assert_equals("104", additional_attributes['kafkaBrokerId'])
    entity_type_updates = update.to_avro_dict()['entityTypeUpdates']
    assert_equals(3, len(entity_type_updates))

    broker_topic_update = entity_type_updates[0]
    assert_equals('KAFKA_BROKER_TOPIC', broker_topic_update['entityType'])
    assert_equals(False, broker_topic_update['deleteOldEntities'])
    broker_topic_entities = broker_topic_update['entities']
    assert_equals(0, len(broker_topic_entities))

    replica_update = entity_type_updates[1]
    assert_equals('KAFKA_REPLICA', replica_update['entityType'])
    assert_equals(False, broker_topic_update['deleteOldEntities'])
    replica_entities = replica_update['entities']
    assert_equals(0, len(replica_entities))

    client_update = entity_type_updates[2]
    assert_equals('KAFKA_PRODUCER', client_update['entityType'])
    assert_equals(False, client_update['deleteOldEntities'])
    client_entities = client_update['entities']
    assert_equals(0, len(client_entities))

    # Use the safety valuve to skip the metric names servlet. We ought to get
    # updates now.
    safety_valve_str = "kafka_topicmetricnames_policy = skip"
    safety_valve.update_safety_valve(
      dict(host_agent_safety_valve=safety_valve_str))

    metrics = adapter.parse_metrics_from_url(conf_parser,
                                    simplejson.load(StringIO(sample)))
    adapter.add_webserver_metrics(version,
                                  update,
                                  metrics,
                                  accessors=_METRIC_ACCESSORS)
    update = RoleUpdate(name="broker-1",
                        roletype=adapter.role_type,
                        servicetype=adapter.service_type,
                        service_name="kafka-1")
    adapter.add_webserver_metrics(version,
                                  update,
                                  metrics,
                                  accessors=_METRIC_ACCESSORS)

    entity_type_updates = update.to_avro_dict()['entityTypeUpdates']
    assert_equals(3, len(entity_type_updates))

    broker_topic_update = entity_type_updates[0]
    assert_equals('KAFKA_BROKER_TOPIC', broker_topic_update['entityType'])
    assert_equals(True, broker_topic_update['deleteOldEntities'])
    broker_topic_entities = broker_topic_update['entities']
    assert_equals(5, len(broker_topic_entities))

    replica_update = entity_type_updates[1]
    assert_equals('KAFKA_REPLICA', replica_update['entityType'])
    assert_equals(True, replica_update['deleteOldEntities'])
    replica_entities = replica_update['entities']
    assert_equals(16, len(replica_entities))

    client_update = entity_type_updates[2]
    assert_equals('KAFKA_PRODUCER', client_update['entityType'])
    client_entities = client_update['entities']
    assert_equals(1, len(client_entities))

  finally:
    shutil.rmtree(conf_dir)

def test_kudu_master_adapter():
  conf_dir = make_config_file(
"""[kudu_master]
role_name = my_kudu_master
service_name = my_kudu
service_version = 5
host = localhost
log_dir = /path/to/logs
collect_interval = 1.0
""")
  metrics_path = os.path.join(conf_dir, conf.SERVICE_METRICS_PROPERTIES)
  file(metrics_path, 'w').write(
"""{
  "KUDU-KUDU_MASTER" : {
    "32772" : {
      "source" : "generic_heap_size::value"
    }
  },
  "KUDU_REPLICA" : {
    "32773" : {
      "source" : "memrowset_size::value"
    },
    "32774" : {
      "counter" : true,
      "source" : "write_op_duration_no_consistency::total_count"
    }
  }
}
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("KUDU",
                                            "KUDU_MASTER",
                                            SafetyValve())
    assert_true(adapter is not None)
    adapter.read_service_defined_metrics(metrics_path)
    directories = adapter.get_monitored_directories(conf_parser)
    assert_true(len(directories) == 1)
    assert_true(dict(path='/path/to/logs', type=None) in directories)
    # We have no kudu-monitor.properties file.
    assert_equals(None, adapter.get_metrics_url(conf_parser))

    kudu_properties_path = os.path.join(conf_dir, "kudu-monitoring.properties")
    file(kudu_properties_path, 'w').write(
"""
webserver_port=12345
""")
    assert_equals('http://localhost:12345/jsonmetricz',
                  adapter.get_metrics_url(conf_parser))

    version = constants.SERVICE_VERSION_CDH5
    sample = adapter.get_metrics_sample(version)
    metrics = adapter.parse_metrics_from_url(conf_parser,
                                    simplejson.load(StringIO(sample)))
    update = RoleUpdate(name="whatever",
                        roletype=adapter.role_type,
                        servicetype=adapter.service_type,
                        service_name="whatever")
    adapter.add_webserver_metrics(version,
                                  update,
                                  metrics,
                                  accessors=_METRIC_ACCESSORS)
    role_metrics = { }
    for metric in update.to_avro_dict()['metrics']:
      role_metrics[metric['id']] = metric['value']
    assert_equals(1, len(role_metrics))

    entity_type_updates = update.to_avro_dict()['entityTypeUpdates']
    assert_equals(1, len(entity_type_updates))
    replica_update = entity_type_updates[0]
    assert_equals('KUDU_REPLICA', replica_update['entityType'])
    assert_equals(False, replica_update['deleteOldEntities'])
    assert_equals(False, replica_update['createNewEntities'])

    replica_entities = replica_update['entities']
    assert_equals(1, len(replica_entities))
    for replica in replica_entities:
      assert_equals('my_kudu_master', replica['attributes']['roleName'])
      assert_true('kuduTabletId' in replica['attributes'])
      assert_equals(2, len(replica['metrics']))

  finally:
    shutil.rmtree(conf_dir)

def test_kudu_tserver_adapter():
  conf_dir = make_config_file(
"""[kudu_tserver]
role_name = my_kudu_tserver
service_name = my_kudu
service_version = 5
host = localhost
log_dir = /path/to/logs
collect_interval = 1.0
""")
  metrics_path = os.path.join(conf_dir, conf.SERVICE_METRICS_PROPERTIES)
  file(metrics_path, 'w').write(
"""{
  "KUDU-KUDU_TSERVER" : {
    "32772" : {
      "source" : "generic_heap_size::value"
    }
  },
  "KUDU_REPLICA" : {
    "32773" : {
      "source" : "memrowset_size::value"
    }
  }
}
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    safety_valve = SafetyValve()
    adapter = AdapterFactory().make_adapter("KUDU",
                                            "KUDU_TSERVER",
                                            safety_valve)
    assert_true(adapter is not None)
    adapter.read_service_defined_metrics(metrics_path)
    directories = adapter.get_monitored_directories(conf_parser)
    assert_true(len(directories) == 1)
    assert_true(dict(path='/path/to/logs', type=None) in directories)
    # We have no kudu-monitor.properties file.
    assert_equals(None, adapter.get_metrics_url(conf_parser))

    kudu_properties_path = os.path.join(conf_dir, "kudu-monitoring.properties")
    file(kudu_properties_path, 'w').write(
"""
webserver_interface=0.0.0.0
webserver_port=12345
""")
    assert_equals('http://localhost:12345/jsonmetricz',
                  adapter.get_metrics_url(conf_parser))

    version = constants.SERVICE_VERSION_CDH5
    sample = adapter.get_metrics_sample(version)
    metrics = adapter.parse_metrics_from_url(conf_parser,
                                    simplejson.load(StringIO(sample)))
    update = RoleUpdate(name="whatever",
                        roletype=adapter.role_type,
                        servicetype=adapter.service_type,
                        service_name="whatever")
    adapter.add_webserver_metrics(version,
                                  update,
                                  metrics,
                                  accessors=_METRIC_ACCESSORS)
    role_metrics = { }
    for metric in update.to_avro_dict()['metrics']:
      role_metrics[metric['id']] = metric['value']
    assert_equals(1, len(role_metrics))

    entity_type_updates = update.to_avro_dict()['entityTypeUpdates']
    assert_equals(1, len(entity_type_updates))
    replica_update = entity_type_updates[0]
    assert_equals('KUDU_REPLICA', replica_update['entityType'])
    assert_equals(False, replica_update['deleteOldEntities'])
    assert_equals(False, replica_update['createNewEntities'])

    replica_entities = replica_update['entities']
    assert_equals(44, len(replica_entities))
    for replica in replica_entities:
      assert_equals('my_kudu_tserver', replica['attributes']['roleName'])
      assert_true('kuduTabletId' in replica['attributes'])
      assert_equals(1, len(replica['metrics']))

    safety_valve_str = \
      "kudu-1_feature_send_replica_entity_update_enabled= FALSE"
    safety_valve.update_safety_valve(
      dict(host_agent_safety_valve=safety_valve_str))
    adapter.add_webserver_metrics(version,
                                  update,
                                  metrics,
                                  accessors=_METRIC_ACCESSORS)
    update = RoleUpdate(name="tserver-1",
                        roletype=adapter.role_type,
                        servicetype=adapter.service_type,
                        service_name="kudu-1")
    adapter.add_webserver_metrics(version,
                                  update,
                                  metrics,
                                  accessors=_METRIC_ACCESSORS)
    entity_type_updates = update.to_avro_dict()['entityTypeUpdates']
    assert_equals(None, entity_type_updates)

  finally:
    shutil.rmtree(conf_dir)

def test_hive_metastore_adapter():
  sample_file = "/foo/bar/metastore/metrics.log"
  conf_dir = make_config_file(
"""[hivemetastore]
role_name = the_meta_metastore
service_name = hive-1
service_version = 5
host = localhost
log_dir = /path/to/logs
collect_interval = 1.0
hive.service.metrics.file.location=%s
""" % (sample_file,))

  metrics_path = os.path.join(conf_dir, conf.SERVICE_METRICS_PROPERTIES)
  file(metrics_path, 'w').write(
"""{
  "HIVEMETASTORE" : {
    "66666" : {
      "source" : "timers::api_get_all_databases::p99"
    },
    "66667" : {
      "source" : "gauges::memory.non-heap.max::value"
    },
    "66668" : {
      "source" : "gauges::memory.api_foobar.used::value"
    }
  }
}
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("HIVE",
                                            "HIVEMETASTORE",
                                            SafetyValve())
    assert_true(adapter is not None)
    adapter.read_service_defined_metrics(metrics_path)
    directories = adapter.get_monitored_directories(conf_parser)
    assert_true(len(directories) == 1)
    assert_true(dict(path='/path/to/logs', type=None) in directories)
    assert_equals(None, adapter.get_metrics_url(conf_parser))
    assert_equals(sample_file, adapter.get_metrics_file(conf_parser))

    version = constants.SERVICE_VERSION_CDH5
    sample = adapter.get_metrics_sample(version)
    metrics = adapter.parse_metrics_from_file(
      conf_parser,
      simplejson.load(StringIO(sample)))
    assert_true(metrics is not None)
    update = RoleUpdate(name="whatever",
                        roletype=adapter.role_type,
                        servicetype=adapter.service_type,
                        service_name="whatever")
    adapter.add_sample_file_metrics(version,
                                    update,
                                    metrics,
                                    accessors=_METRIC_ACCESSORS)
    role_metrics = { }
    for metric in update.to_avro_dict()['metrics']:
      role_metrics[metric['id']] = metric['value']
    assert_equals(3, len(role_metrics))

    assert_equals(34.852, role_metrics.get(66666))
    assert_equals(136314880, role_metrics.get(66667))
    # This tests that the non-existing metric (see above) does get a default
    # value
    assert_equals(0.0, role_metrics.get(66668))

  finally:
    shutil.rmtree(conf_dir)

def test_hiveserver2_adapter():
  sample_file = "/foo/bar/hs2/metrics.log"
  conf_dir = make_config_file(
"""[hiveserver2]
role_name = hs2_the_third
service_name = hive-1
service_version = 5
host = localhost
log_dir = /path/to/logs
collect_interval = 1.0
hive.service.metrics.file.location=%s
""" % (sample_file,))

  metrics_path = os.path.join(conf_dir, conf.SERVICE_METRICS_PROPERTIES)
  file(metrics_path, 'w').write(
"""{
  "HIVESERVER2" : {
    "66666" : {
      "source" : "timers::api_get_functions::p99"
    },
    "66667" : {
      "source" : "counters::jvm.pause.extraSleepTime::count"
    },
    "66668" : {
      "source" : "counters::jvm.pause.api_extra_foobar::count"
    }
  }
}
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("HIVE",
                                            "HIVESERVER2",
                                            SafetyValve())
    assert_true(adapter is not None)
    adapter.read_service_defined_metrics(metrics_path)
    directories = adapter.get_monitored_directories(conf_parser)
    assert_true(len(directories) == 1)
    assert_true(dict(path='/path/to/logs', type=None) in directories)
    assert_equals(None, adapter.get_metrics_url(conf_parser))
    assert_equals(sample_file, adapter.get_metrics_file(conf_parser))

    version = constants.SERVICE_VERSION_CDH5
    sample = adapter.get_metrics_sample(version)
    metrics = adapter.parse_metrics_from_file(
      conf_parser,
      simplejson.load(StringIO(sample)))
    assert_true(metrics is not None)
    update = RoleUpdate(name="whatever",
                        roletype=adapter.role_type,
                        servicetype=adapter.service_type,
                        service_name="whatever")
    adapter.add_sample_file_metrics(version,
                                    update,
                                    metrics,
                                    accessors=_METRIC_ACCESSORS)
    role_metrics = { }
    for metric in update.to_avro_dict()['metrics']:
      role_metrics[metric['id']] = metric['value']
    assert_equals(3, len(role_metrics))

    assert_equals(124.595, role_metrics.get(66666))
    assert_equals(42, role_metrics.get(66667))
    assert_equals(0.0, role_metrics.get(66668))

  finally:
    shutil.rmtree(conf_dir)

def test_regionserver_adapter():
  conf_dir = make_config_file(
"""[regionserver]
role_name = my_hbase-rs1
service_name = my_hbase
service_version = 5
host = localhost
hbase_regionserver_info_port = 50080
log_dir = /path/to/logs
monitored_directories = /path/to/mon/dir1,/path/to/mon/dir2
collect_interval = 1.0
""")
  try:
    safety_valve = SafetyValve()
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("HBASE",
                                            "REGIONSERVER",
                                            safety_valve)
    assert_true(adapter is not None)
    directories = adapter.get_monitored_directories(conf_parser)
    assert_equals(3,
                  len(directories),
                  "found %d directories: %s" % (len(directories), directories))
    assert_true(dict(path='/path/to/logs', type=None) in directories)
    assert_true(dict(path='/path/to/mon/dir1', type=None) in directories)
    assert_true(dict(path='/path/to/mon/dir2', type=None) in directories)
    assert_equals('http://localhost:50080/jmx',
                  adapter.get_metrics_url(conf_parser))

    role_update = \
      get_role_update_for_sample(adapter,
                                 conf_parser,
                                 constants.SERVICE_VERSION_CDH5).to_avro_dict()
    assert_true(len(role_update['regionUpdates']) >= 1)
    for r in role_update['regionUpdates']:
      region_metrics = { }
      for metric_obj in r['metrics']:
        region_metrics[metric_obj['id']] = metric_obj['value']
      for metric_id, acc in _METRIC_ACCESSORS['HREGION']['cdh5'].iteritems():
        if metric_id not in region_metrics:
          print metric_id
          print acc.bean_alias
        #assert_true(metric_id in region_metrics)

    safety_valve_str = \
      "whatever_service-1_whatever_role-1_feature_send_region_updates_enabled= FALSE"
    safety_valve.update_safety_valve(
      dict(host_agent_safety_valve=safety_valve_str))
    role_update = \
      get_role_update_for_sample(adapter,
                                 conf_parser,
                                 constants.SERVICE_VERSION_CDH5).to_avro_dict()
    assert_equals(0, len(role_update['regionUpdates']))
    safety_valve_str = \
      "whatever_service-1_feature_send_region_updates_enabled= FALSE"
    safety_valve.update_safety_valve(
      dict(host_agent_safety_valve=safety_valve_str))
    role_update = \
      get_role_update_for_sample(adapter,
                                 conf_parser,
                                 constants.SERVICE_VERSION_CDH5).to_avro_dict()
    assert_equals(0, len(role_update['regionUpdates']))
    metric_dict = {}
    for metric in role_update['metrics']:
      metric_dict[metric['id']] = metric['value']
    assert_equals(170, len(metric_dict))
    for metric_id in _METRIC_ACCESSORS['REGIONSERVER']['cdh5'].iterkeys():
      assert_true(metric_id in metric_dict)

    assert_true(len(role_update['regionServerReplicationPeerUpdates']) >= 1)
    for r in role_update['regionServerReplicationPeerUpdates']:
      replication_metrics = { }
      for metric_obj in r['metrics']:
        replication_metrics[metric_obj['id']] = metric_obj['value']
      for metric_id, acc in _METRIC_ACCESSORS\
            ['HBASE_REGIONSERVER_REPLICATION_PEER']['cdh5'].iteritems():
        if metric_id not in replication_metrics:
          print metric_id
          print acc.bean_alias
        assert_true(metric_id in replication_metrics)

    metric_dict = {}
    for metric in role_update['metrics']:
      metric_dict[metric['id']] = metric['value']
    assert_equals(170, len(metric_dict))
    for metric_id in _METRIC_ACCESSORS['REGIONSERVER']['cdh5'].iterkeys():
      assert_true(metric_id in metric_dict)
    safety_valve_str = \
      "whatever_service-1_whatever_role-1_feature_send_replication_metrics_enabled= FALSE"
    safety_valve.update_safety_valve(
      dict(host_agent_safety_valve=safety_valve_str))
    role_update = \
      get_role_update_for_sample(adapter,
                                 conf_parser,
                                 constants.SERVICE_VERSION_CDH5).to_avro_dict()
    assert_equals(0, len(role_update['regionServerReplicationPeerUpdates']))

    metric_dict = {}
    for metric in role_update['metrics']:
      metric_dict[metric['id']] = metric['value']
    assert_equals(170, len(metric_dict))
    for metric_id in _METRIC_ACCESSORS['REGIONSERVER']['cdh5'].iterkeys():
      assert_true(metric_id in metric_dict)
  finally:
    shutil.rmtree(conf_dir)

  # test HTTPS URL construction
  conf_dir = make_config_file(
"""[regionserver]
role_name = my_hbase-rs1
service_name = my_hbase
service_version = 5
host = localhost
hbase_regionserver_info_port = 50099
log_dir = /path/to/logs
monitored_directories = /path/to/mon/dir1,/path/to/mon/dir2
collect_interval = 1.0
use_ssl = true
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("HBASE",
                                            "REGIONSERVER",
                                            SafetyValve())
    assert_equals('https://localhost:50099/jmx',
                  adapter.get_metrics_url(conf_parser))
  finally:
    shutil.rmtree(conf_dir)

  # Test OPSAPS-28171: make sure we can parse IPC metric names in both formats.
  metric_dict = parse_metrics_for_provided_sample(adapter,
                                                  conf_parser,
                                                  constants.SERVICE_VERSION_CDH5,
                                                  get_sample_file('hbase_regionserver_cdh5'))
  assert_equals(170, len(metric_dict))
  for metric_id in _METRIC_ACCESSORS['REGIONSERVER']['cdh5'].iterkeys():
    assert_true(metric_id in metric_dict)
  metric_dict = parse_metrics_for_provided_sample(adapter,
                                                  conf_parser,
                                                  constants.SERVICE_VERSION_CDH5,
                                                  get_sample_file('hbase_regionserver_cdh55'))
  assert_equals(170, len(metric_dict))
  for metric_id in _METRIC_ACCESSORS['REGIONSERVER']['cdh5'].iterkeys():
    assert_true(metric_id in metric_dict)

def test_master_adapter():
  conf_dir = make_config_file(
"""[master]
role_name = my_hbase-master1
service_name = my_hbase
service_version = 5
host = localhost
hbase_master_info_port = 50666
log_dir = /path/to/logs
monitored_directories = /path/to/mon/dir1,/path/to/mon/dir2
collect_interval = 1.0
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("HBASE", "MASTER", SafetyValve())
    assert_true(adapter is not None)
    directories = adapter.get_monitored_directories(conf_parser)
    assert_equals(3,
                  len(directories),
                  "found %d directories: %s" % (len(directories), directories))
    assert_true(dict(path='/path/to/logs', type=None) in directories)
    assert_true(dict(path='/path/to/mon/dir1', type=None) in directories)
    assert_true(dict(path='/path/to/mon/dir2', type=None) in directories)
    assert_equals('http://localhost:50666/jmx',
                  adapter.get_metrics_url(conf_parser))

    metric_dict = parse_metrics_sample(adapter,
                                       conf_parser,
                                       constants.SERVICE_VERSION_CDH5)
    assert_equals(122, len(metric_dict))
    for metric_id in _METRIC_ACCESSORS['MASTER']['cdh5'].iterkeys():
      assert_true(metric_id in metric_dict)
  finally:
    shutil.rmtree(conf_dir)

  # test HTTPS URL construction
  conf_dir = make_config_file(
"""[master]
role_name = my_hbase-master1
service_name = my_hbase
service_version = 5
host = localhost
hbase_master_info_port = 50466
log_dir = /path/to/logs
monitored_directories = /path/to/mon/dir1,/path/to/mon/dir2
collect_interval = 1.0
use_ssl = true
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("HBASE", "MASTER", SafetyValve())
    assert_equals('https://localhost:50466/jmx',
                  adapter.get_metrics_url(conf_parser))
  finally:
    shutil.rmtree(conf_dir)

  # Test OPSAPS-28171: make sure we can parse IPC metric names in both formats.
  metric_dict = parse_metrics_for_provided_sample(adapter,
                                                  conf_parser,
                                                  constants.SERVICE_VERSION_CDH5,
                                                  get_sample_file('hbase_master_cdh5'))
  assert_equals(122, len(metric_dict))
  for metric_id in _METRIC_ACCESSORS['MASTER']['cdh5'].iterkeys():
    assert_true(metric_id in metric_dict)
  metric_dict = parse_metrics_for_provided_sample(adapter,
                                                  conf_parser,
                                                  constants.SERVICE_VERSION_CDH5,
                                                  get_sample_file('hbase_master_cdh55'))
  assert_equals(122, len(metric_dict))
  for metric_id in _METRIC_ACCESSORS['MASTER']['cdh5'].iterkeys():
    assert_true(metric_id in metric_dict)

def test_hue_server_adapter():

  class HueIsAliveCollectorWithBehavior(_HueIsAliveCollector):

    def __init__(self, adapter):
      _HueIsAliveCollector.__init__(self, adapter)
      self.do_raise = False

    def _call_is_alive(self, is_alive_url, timeout=None):
      if self.do_raise:
        raise Exception("Boom!")
      return

  sample_file = "/foo/bar/hue_server/metrics.log"
  conf_dir = make_config_file(
"""[hue_server]
role_name = my_hue-server-1
service_name = my_hue
service_version = 5
service_release = 5.2.0
host = foobar
http_port = 666
log_dir = /path/to/logs
monitored_directories = /path/to/mon/dir1,/path/to/mon/dir2
collect_interval = 1.0
location=%s
""" % sample_file)

  metrics_path = os.path.join(conf_dir, conf.SERVICE_METRICS_PROPERTIES)
  file(metrics_path, 'w').write(
"""{
  "HUE_SERVER" : {
    "66666" : {
      "source" : "desktop.auth.oauth.authentication-time::99_percentile"
    },
    "66667" : {
      "source" : "python.threads.count::value"
    },
    "66668" : {
      "source" : "desktop.requests.exceptions.count::count"
    }
  }
}
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("HUE", "HUE_SERVER", SafetyValve())
    assert_true(adapter is not None)
    assert_equals(1, len(adapter.get_adapter_specific_collectors()))
    assert_false(adapter._is_alive_supported(conf_parser))
    collector = adapter.get_adapter_specific_collectors()[0]
    assert_true(isinstance(collector, AbstractMetricCollector))
    collector.update_with_conf(conf_parser)
    # This is less than optimal but we need to mimic the generic monitor
    # behavior here.
    assert_false(collector.is_enabled())

    adapter.read_service_defined_metrics(metrics_path)
    directories = adapter.get_monitored_directories(conf_parser)
    assert_true(len(directories) == 3)
    assert_true(dict(path='/path/to/logs', type=None) in directories)
    assert_equals(None, adapter.get_metrics_url(conf_parser))
    assert_equals(sample_file, adapter.get_metrics_file(conf_parser))

    version = constants.SERVICE_VERSION_CDH5
    sample = adapter.get_metrics_sample(version)
    metrics = adapter.parse_metrics_from_file(
      conf_parser,
      simplejson.load(StringIO(sample)))
    assert_true(metrics is not None)
    update = RoleUpdate(name="whatever",
                        roletype=adapter.role_type,
                        servicetype=adapter.service_type,
                        service_name="whatever")
    adapter.add_sample_file_metrics(version,
                                    update,
                                    metrics,
                                    accessors=_METRIC_ACCESSORS)
    collector.report_metrics(version, update)
    role_metrics = { }
    for metric in update.to_avro_dict()['metrics']:
      role_metrics[metric['id']] = metric['value']
    assert_equals(3, len(role_metrics))

    assert_equals(1234567, role_metrics.get(66666))
    assert_equals(42.5, role_metrics.get(66667))
    assert_equals(7.0, role_metrics.get(66668))
    assert_true(
      role_metrics.get(schema.WEB_METRICS_COLLECTION_STATUS) is None)
    assert_true(
      role_metrics.get(schema.WEB_METRICS_COLLECTION_DURATION) is None)

  finally:
    shutil.rmtree(conf_dir)

  # Now with a release enabling is alive collection.
  conf_dir = make_config_file(
"""[hue_server]
role_name = my_hue-server-1
service_name = my_hue
service_version = 5
service_release = 5.5.0
host = foobar
http_port = 666
http_host=foobar.com
log_dir = /path/to/logs
monitored_directories = /path/to/mon/dir1,/path/to/mon/dir2
collect_interval = 1.0
location=%s
""" % sample_file)

  metrics_path = os.path.join(conf_dir, conf.SERVICE_METRICS_PROPERTIES)
  file(metrics_path, 'w').write(
"""{
  "HUE_SERVER" : {
    "66666" : {
      "source" : "desktop.auth.oauth.authentication-time::99_percentile"
    },
    "66667" : {
      "source" : "python.threads.count::value"
    },
    "66668" : {
      "source" : "desktop.requests.exceptions.count::count"
    }
  }
}
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("HUE", "HUE_SERVER", SafetyValve())
    assert_true(adapter is not None)

    assert_equals(1, len(adapter.get_adapter_specific_collectors()))
    assert_true(adapter._is_alive_supported(conf_parser))
    # Replace the collector with our end-point mock.
    adapter._is_alive_collector = HueIsAliveCollectorWithBehavior(adapter)
    collector = adapter.get_adapter_specific_collectors()[0]

    assert_equals(None, collector.metrics_uri)
    collector.update_with_conf(conf_parser)
    assert_equals('http://foobar.com:666/desktop/debug/is_alive',
                  collector.metrics_uri)
    # This is less than optimal but we need to mimic the generic monitor
    # behavior here.
    if collector.is_enabled():
      collector.collect_and_parse(conf_parser)

    adapter.read_service_defined_metrics(metrics_path)
    directories = adapter.get_monitored_directories(conf_parser)
    assert_true(len(directories) == 3)
    assert_true(dict(path='/path/to/logs', type=None) in directories)
    assert_equals(None, adapter.get_metrics_url(conf_parser))
    assert_equals(sample_file, adapter.get_metrics_file(conf_parser))

    version = constants.SERVICE_VERSION_CDH5
    sample = adapter.get_metrics_sample(version)
    metrics = adapter.parse_metrics_from_file(
      conf_parser,
      simplejson.load(StringIO(sample)))
    assert_true(metrics is not None)
    update = RoleUpdate(name="whatever",
                        roletype=adapter.role_type,
                        servicetype=adapter.service_type,
                        service_name="whatever")
    adapter.add_sample_file_metrics(version,
                                    update,
                                    metrics,
                                    accessors=_METRIC_ACCESSORS)
    collector.report_metrics(version, update)
    role_metrics = { }
    for metric in update.to_avro_dict()['metrics']:
      role_metrics[metric['id']] = metric['value']
    assert_equals(5, len(role_metrics))

    assert_equals(1234567, role_metrics.get(66666))
    assert_equals(42.5, role_metrics.get(66667))
    assert_equals(7.0, role_metrics.get(66668))
    assert_equals(constants.WEB_METRICS_COLLECTION_GOOD,
                  role_metrics.get(schema.WEB_METRICS_COLLECTION_STATUS))
    assert_true(
      role_metrics.get(schema.WEB_METRICS_COLLECTION_DURATION) is not None)

    collector.do_raise = True
    adapter.get_adapter_specific_collectors()[0].collect_and_parse(conf_parser)
    update = RoleUpdate(name="whatever",
                        roletype=adapter.role_type,
                        servicetype=adapter.service_type,
                        service_name="whatever")
    collector.report_metrics(version, update)
    role_metrics = { }
    for metric in update.to_avro_dict()['metrics']:
      role_metrics[metric['id']] = metric['value']
    assert_equals(2, len(role_metrics))
    assert_equals(constants.WEB_METRICS_COLLECTION_COMMUNICATION_FAILURE,
                  role_metrics.get(schema.WEB_METRICS_COLLECTION_STATUS))
    assert_true(
      role_metrics.get(schema.WEB_METRICS_COLLECTION_DURATION) is not None)

  finally:
    shutil.rmtree(conf_dir)

  # Test ssl/tls
  conf_dir = make_config_file(
"""[hue_server]
role_name = my_hue-server-1
service_name = my_hue
service_version = 5
service_release = 5.5.0
host = foobar
http_port = 666
http_host=foobar.com
ssl_enable=true
log_dir = /path/to/logs
monitored_directories = /path/to/mon/dir1,/path/to/mon/dir2
collect_interval = 1.0
location=%s
""" % sample_file)

  metrics_path = os.path.join(conf_dir, conf.SERVICE_METRICS_PROPERTIES)
  file(metrics_path, 'w').write(
"""{
  "HUE_SERVER" : {
    "66666" : {
      "source" : "desktop.auth.oauth.authentication-time::99_percentile"
    },
    "66667" : {
      "source" : "python.threads.count::value"
    },
    "66668" : {
      "source" : "desktop.requests.exceptions.count::count"
    }
  }
}
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("HUE", "HUE_SERVER", SafetyValve())
    assert_true(adapter is not None)

    assert_equals(1, len(adapter.get_adapter_specific_collectors()))
    assert_true(adapter._is_alive_supported(conf_parser))
    collector = adapter.get_adapter_specific_collectors()[0]

    assert_equals(None, collector.metrics_uri)
    collector.update_with_conf(conf_parser)
    assert_equals('https://foobar.com:666/desktop/debug/is_alive',
                  collector.metrics_uri)
  finally:
    shutil.rmtree(conf_dir)

def test_common_adapter_robustness():
  """
  Test that CommonAdapter can handle missing csd-monitoring.properties.
  """
  conf_dir = make_config_file(
"""[my-role]
role_name = my_role1
service_name = my_service
service_version = 5
host = localhost
log_dir = /path/to/logs
monitored_directories = /path/to/mon/dir1,/path/to/mon/dir2
collect_interval = 1.0
""")

  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    daemon = TestMonitorDaemon()
    reg = daemon.get_registry_for_testing()
    adapter = CommonAdapter("my-service", "my-role", SafetyValve())
    monitor = GenericMonitor("my-service", "my-role", daemon, reg, adapter)

    # Initial state. Conf has not been loaded.
    assert_true(monitor._url_collector is not None)
    assert_true(monitor._sample_file_collector is not None)
    assert_equals(None, monitor._last_conf_path)
    assert_equals(None, monitor._last_conf_timestamp)

    # Load conf.
    # There is cloudera-monitor.properties but no csd-monitoring.properties.
    monitor._conf_parser = conf_parser
    reloaded = monitor._reload_config_if_necessary()

    assert_equals(True, reloaded)
    assert_true(monitor._last_conf_path is not None)
    assert_true(monitor._last_conf_timestamp is not None)
    assert_equals("my_role1", monitor._role_name)
    assert_equals("my_service", monitor._service_name)
    assert_equals(5, monitor._service_version)

    # CommonAdapter expects csd-monitoring.properties for metrics url or file.
    assert_equals(None, adapter.get_metrics_url(conf_parser))
    assert_equals(None, adapter.get_metrics_file(conf_parser))

    # Verify that parsing for directories still works.
    directories = adapter.get_monitored_directories(conf_parser)
    assert_equals(3, len(directories))

  finally:
    shutil.rmtree(conf_dir)

def test_common_adapter():
  """
  Test that CommonAdapter is able to load needed information from properties
  files.
  """
  conf_dir = make_config_file(
"""[my-role]
role_name = my_role1
service_name = my_service
service_version = 5
host = localhost
log_dir = /path/to/logs
monitored_directories = /path/to/mon/dir1,/path/to/mon/dir2
collect_interval = 1.0
""")

  try:
    csd_monitoring_file_path = os.path.join(conf_dir, conf.CSD_MONITORING_PROPERTIES)
    file(csd_monitoring_file_path, 'w').write(
"""[my-role]
metrics_available = true
web_metrics_enabled = true
http_url = http://0.0.0.0:30000/metrics
role_entity_type = TEST_SERVICE-COMMON_ROLE
""")

    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    daemon = TestMonitorDaemon()
    reg = daemon.get_registry_for_testing()
    adapter = CommonAdapter("my-service", "my-role", SafetyValve())
    monitor = GenericMonitor("my-service", "my-role", daemon, reg, adapter)

    # Initial state. Conf has not been loaded.
    assert_true(monitor._url_collector is not None)
    assert_true(monitor._sample_file_collector is not None)
    assert_equals(None, monitor._last_conf_path)
    assert_equals(None, monitor._last_conf_timestamp)

    # Load conf.
    # Both cloudera-monitor.properties and csd-monitoring.properties are in place.
    monitor._conf_parser = conf_parser
    reloaded = monitor._reload_config_if_necessary()

    assert_equals(True, reloaded)
    assert_true(monitor._last_conf_timestamp is not None)
    assert_equals("my_role1", monitor._role_name)
    assert_equals("my_service", monitor._service_name)
    assert_equals(5, monitor._service_version)
    assert_equals("http://0.0.0.0:30000/metrics", adapter.get_metrics_url(conf_parser))
    assert_equals(None, adapter.get_metrics_file(conf_parser))
    directories = adapter.get_monitored_directories(conf_parser)
    assert_equals(3, len(directories))

  finally:
    shutil.rmtree(conf_dir)
