# Copyright (c) 2017 Cloudera, Inc. All rights reserved.

try:
  from cStringIO import StringIO
except:
  from StringIO import StringIO
import os
import shutil
import simplejson
import urllib2

from cmf.monitor import conf, constants, schema
from cmf.monitor.generic.adapter_factory import AdapterFactory
from cmf.monitor.generic.hdfs_adapters import DSSDDATANODE_SERVICE_METRICS
from cmf.monitor.monitor_properties import ConfParser
from cmf.safety_valve import SafetyValve

from cmf_test.monitor import make_config_file, parse_metrics_sample, get_role_update_for_sample, \
  get_role_update_for_provided_sample

from nose.tools import assert_true, assert_equals, assert_false

def test_hdfs_namenode_adapter():
  conf_dir = make_config_file(
"""[namenode]
service_name = my_hdfs
service_version = 5
role_name = hdfs-NN
host = localhost
dfs_http_port = 12345
namenode_port = 8020
dfs_name_dirs = /path/to/dir1,/path/to/dir2
log_dir = /path/to/logs
collect_interval = 1.0
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("HDFS",
                                            "NAMENODE",
                                            SafetyValve())
    assert_true(adapter is not None)
    directories = adapter.get_monitored_directories(conf_parser)
    assert_true(len(directories) == 3)
    assert_true(dict(path='/path/to/logs', type=None) in directories)
    assert_true(dict(path='/path/to/dir1', type=None) in directories)
    assert_true(dict(path='/path/to/dir2', type=None) in directories)
    assert_equals('http://localhost:12345/jmx',
                  adapter.get_metrics_url(conf_parser))
    assert_equals(452, len(parse_metrics_sample(
          adapter, conf_parser, constants.SERVICE_VERSION_CDH5)))
    assert_equals(456, len(parse_metrics_sample(
          adapter, conf_parser, constants.SERVICE_VERSION_CDH6)))
  finally:
    shutil.rmtree(conf_dir)

  # test parsing out the service RPC metrics -- we should have the same number
  # as above due to our default handling of RPC metrics.
  conf_dir = make_config_file(
"""[namenode]
service_name = my_hdfs
service_version = 5
role_name = hdfs-NN
host = localhost
dfs_http_port = 12345
namenode_port = 8020
dfs_name_dirs = /path/to/dir1,/path/to/dir2
log_dir = /path/to/logs
collect_interval = 1.0
dfs_namenode_servicerpc_address = 8022
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("HDFS", "NAMENODE", SafetyValve())
    assert_true(adapter is not None)
    assert_equals(452, len(parse_metrics_sample(
          adapter, conf_parser, constants.SERVICE_VERSION_CDH5)))
    assert_equals(456, len(parse_metrics_sample(
          adapter, conf_parser, constants.SERVICE_VERSION_CDH6)))
  finally:
    shutil.rmtree(conf_dir)

  # test HTTPS URL construction
  conf_dir = make_config_file(
"""[namenode]
service_name = my_hdfs
service_version = 5
role_name = hdfs-NN
host = localhost
dfs_http_port = 12345
use_ssl = true
dfs_https_port = 54321
namenode_port = 8020
dfs_name_dirs = /path/to/dir1,/path/to/dir2
log_dir = /path/to/logs
collect_interval = 1.0
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("HDFS", "NAMENODE", SafetyValve())
    assert_equals('https://localhost:54321/jmx',
                  adapter.get_metrics_url(conf_parser))
  finally:
    shutil.rmtree(conf_dir)

def test_hdfs_secondarynamenode_adapter():
  conf_dir = make_config_file(
"""[secondarynamenode]
service_name = my_hdfs
service_version = 5
role_name = hdfs-SNN
host = localhost
dfs_secondary_http_port = 11000
fs_checkpoint_dir_list = /path/to/dir1,/path/to/dir2
log_dir = /path/to/logs
collect_interval = 1.0
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("HDFS",
                                            "SECONDARYNAMENODE",
                                            SafetyValve())
    assert_true(adapter is not None)
    directories = adapter.get_monitored_directories(conf_parser)
    assert_true(len(directories) == 3)
    assert_true(dict(path='/path/to/logs', type=None) in directories)
    assert_true(dict(path='/path/to/dir1', type=None) in directories)
    assert_true(dict(path='/path/to/dir2', type=None) in directories)
    assert_equals('http://localhost:11000/jmx',
                  adapter.get_metrics_url(conf_parser))
    assert_equals(35, len(parse_metrics_sample(
          adapter, conf_parser, constants.SERVICE_VERSION_CDH5)))
    assert_equals(37, len(parse_metrics_sample(
          adapter, conf_parser, constants.SERVICE_VERSION_CDH6)))
  finally:
    shutil.rmtree(conf_dir)

  # test HTTPS URL construction
  conf_dir = make_config_file(
"""[secondarynamenode]
service_name = my_hdfs
service_version = 5
role_name = hdfs-SNN
host = localhost
dfs_secondary_http_port = 10000
log_dir = /path/to/logs
collect_interval = 1.0
fs_checkpoint_dir_list = /path/to/dir
use_ssl = true
dfs_secondary_https_port = 10200
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("HDFS",
                                            "SECONDARYNAMENODE",
                                            SafetyValve())
    assert_equals('https://localhost:10200/jmx',
                  adapter.get_metrics_url(conf_parser))
  finally:
    shutil.rmtree(conf_dir)

def test_hdfs_journalnode_adapter():
  conf_dir = make_config_file(
"""[journalnode]
role_name = hdfs-JN
service_name = my_hdfs
service_version = 5
host = localhost
dfs_journalnode_http_port = 50075
dfs_journalnode_edits_dir = /path/to/edits/dir
log_dir = /path/to/logs
collect_interval = 1.0
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("HDFS",
                                            "JOURNALNODE",
                                            SafetyValve())
    assert_true(adapter is not None)
    directories = adapter.get_monitored_directories(conf_parser)
    assert_true(len(directories) == 2)
    assert_true(dict(path='/path/to/logs', type=None) in directories)
    assert_true(dict(path='/path/to/edits/dir', type=None) in directories)
    assert_equals('http://localhost:50075/jmx',
                  adapter.get_metrics_url(conf_parser))
    assert_equals(91, len(parse_metrics_sample(
          adapter, conf_parser, constants.SERVICE_VERSION_CDH5)))
    assert_equals(93, len(parse_metrics_sample(
          adapter, conf_parser, constants.SERVICE_VERSION_CDH6)))
  finally:
    shutil.rmtree(conf_dir)

  # test HTTPS URL construction
  conf_dir = make_config_file(
"""[journalnode]
role_name = hdfs-JN
service_name = my_hdfs
service_version = 4
host = localhost
dfs_journalnode_http_port = 50075
dfs_journalnode_https_port = 50080
dfs_journalnode_edits_dir = /path/to/edits/dir
log_dir = /path/to/logs
collect_interval = 1.0
use_ssl = true
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("HDFS",
                                            "JOURNALNODE",
                                            SafetyValve())
    assert_equals('https://localhost:50080/jmx',
                  adapter.get_metrics_url(conf_parser))
  finally:
    shutil.rmtree(conf_dir)

def test_hdfs_datanode_adapter():
  conf_dir = make_config_file(
"""[datanode]
role_name = hdfs-DN
service_name = my_hdfs
service_version = 5
host = localhost
dfs_datanode_http_port = 50080
dfs_data_dirs = /path/to/data/1,/path/to/data/2
log_dir = /path/to/logs
monitored_directories = /path/to/mon/dir1,/path/to/mon/dir2
collect_interval = 1.0
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("HDFS", "DATANODE", SafetyValve())
    assert_true(adapter is not None)
    directories = adapter.get_monitored_directories(conf_parser)
    assert_true(len(directories) == 5)
    assert_true(dict(path='/path/to/logs', type=None) in directories)
    assert_true(dict(path='/path/to/data/1', type=None) in directories)
    assert_true(dict(path='/path/to/data/2', type=None) in directories)
    assert_true(dict(path='/path/to/mon/dir1', type=None) in directories)
    assert_true(dict(path='/path/to/mon/dir2', type=None) in directories)
    assert_equals('http://localhost:50080/jmx',
                  adapter.get_metrics_url(conf_parser))
    assert_equals(113, len(parse_metrics_sample(
          adapter, conf_parser, constants.SERVICE_VERSION_CDH5)))
    assert_equals(123, len(parse_metrics_sample(
          adapter, conf_parser, constants.SERVICE_VERSION_CDH6)))
  finally:
    shutil.rmtree(conf_dir)

  # test HTTPS URL construction
  conf_dir = make_config_file(
"""[datanode]
role_name = hdfs-DN
service_name = my_hdfs
service_version = 5
host = localhost
dfs_datanode_http_port = 50080
dfs_datanode_https_port = 50090
dfs_data_dirs = /path/to/data/1,/path/to/data/2
log_dir = /path/to/logs
collect_interval = 1.0
use_ssl = true
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("HDFS", "DATANODE", SafetyValve())
    assert_equals('https://localhost:50090/jmx',
                  adapter.get_metrics_url(conf_parser))
  finally:
    shutil.rmtree(conf_dir)

def test_hdfs_dssddatanode_adapter():
  conf_dir = make_config_file(
"""[dssddatanode]
role_name = hdfs-DSSDDN
service_name = my_hdfs
service_version = 5
host = localhost
dfs_datanode_http_port = 50080
log_dir = /path/to/logs
monitored_directories = /path/to/mon/dir1,/path/to/mon/dir2
collect_interval = 1.0
""")

  metrics_path = os.path.join(conf_dir, conf.SERVICE_METRICS_PROPERTIES)
  file(metrics_path, 'w').write(DSSDDATANODE_SERVICE_METRICS)

  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("HDFS",
                                            "DSSDDATANODE",
                                            SafetyValve())
    assert_true(adapter is not None)
    adapter.read_service_defined_metrics(metrics_path)
    directories = adapter.get_monitored_directories(conf_parser)
    assert_equals(3, len(directories))
    assert_true(dict(path='/path/to/logs', type=None) in directories)
    assert_true(dict(path='/path/to/mon/dir1', type=None) in directories)
    assert_true(dict(path='/path/to/mon/dir2', type=None) in directories)
    assert_equals('http://localhost:50080/jmx',
                  adapter.get_metrics_url(conf_parser))
    # Semi-gross, but it keeps the test concise.
    conf_parser._cp.set(adapter.section, "service_version", "5")
    # There are 148 metrics currently exposed in the mdl for dssd datanodes.
    # * 14 metrics from the rpc detailed activity bean are missing from the
    #   sample. We expect to get the default, which is 0.
    # * There are 7 metrics collected by smon.
    # Since there are 7 "missing" metrics we expect 148 - 7 metrics, or 139.
    parse_metrics_sample(
          adapter, conf_parser, constants.SERVICE_VERSION_CDH5)
    # Parse the sample again. This tests that we do the copy of the beans
    # metadata.
    metrics = parse_metrics_sample(
          adapter, conf_parser, constants.SERVICE_VERSION_CDH5)
    if len(metrics) != 141:
      msg = get_missing_metrics(
        metrics,
        ServiceDefinedMetrics(
          metrics_path,
          SplittingSourceProcessor('::')).get_sources("DSSDDATANODE"))
      print msg
      assert_true(False)
  finally:
    shutil.rmtree(conf_dir)

  # test HTTPS URL construction
  conf_dir = make_config_file(
"""[dssddatanode]
role_name = hdfs-DN
service_name = my_hdfs
service_version = 5
host = localhost
dfs_datanode_http_port = 50080
dfs_datanode_https_port = 50090
log_dir = /path/to/logs
collect_interval = 1.0
use_ssl = true
""")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    adapter = AdapterFactory().make_adapter("HDFS",
                                            "DSSDDATANODE",
                                            SafetyValve())
    assert_equals('https://localhost:50090/jmx',
                  adapter.get_metrics_url(conf_parser))
  finally:
    shutil.rmtree(conf_dir)
