# Copyright (c) 2011-2013 Cloudera, Inc. All rights reserved.

import array
from multiprocessing import Pipe
from psutil._common import *
import re
import signal
import struct
import tempfile
import time
import os
import os.path
import errno

from nose.plugins.skip import SkipTest
from nose.tools import assert_equals, assert_not_equals, assert_true, assert_false, assert_raises, nottest

import cmf.util
from cmf.monitor import schema, constants
from cmf.monitor.constants import DUPLEX_UNKNOWN
from cmf.monitor.host import HostMonitor
from cmf.monitor.host.tcp_metrics import TcpMetrics
from cmf.monitor.host.filesystem_map import FilesystemMap, disk_usage_wrapper
from cmf.monitor.host.network_interfaces import NetworkInterfaces
from cmf.monitor.host_update import HostUpdate
from cmf.monitor.host.disk_devices import DiskDevices
from cmf.monitor.host.disk_device_utils import calculate_await_times, \
    calculate_service_time, calculate_utilization, calculate_average_request_size, \
    calculate_average_queue_length
from cmf.monitor.host.block_device_stats import BlockDeviceStats
from cmf.monitor.host.filesystem import Filesystem
from cmf.monitor.host.filesystems import Filesystems
from cmf.monitor.host.dns_names import DnsNames
from cmf.monitor.host.ntp_monitor import NTPMonitor
from cmf.monitor.host.entropy_monitor import EntropyMonitor
from cmf.monitor.host.certificate_metrics import CertificateMetrics
from cmf_test.monitor import MockFirehoses, TestMonitorDaemon

SAMPLE_PROC_PARTITIONS = """major minor  #blocks  name

   8        0   62914560 sda
   8        1   54524928 sda1
  20        0       3411 dm-0
 105        0 1003393784 cciss/c1d0
 105        1   86123689 cciss/c1d0p1
"""

# Sample disk devices from the above /proc/partitions file, and used in the below
# /proc/diskstats examples as well. The samples were modified to include very
# large longs (greater than (1 << 63) - 1), see OPSAPS-24118 for more details.
SAMPLE_DISK_DEVICES = ["sda", "sda1", "dm-0", "cciss/c1d0", "cciss/c1d0p1"]

SAMPLE_PROC_DISKSTATS_1 = """   1       0 ram0 0 0 0 0 0 0 0 0 0 0 0
   7       0 loop0 0 0 0 0 0 0 0 0 0 0 0
   8       0 sda 79077 39 1995838 42348 14077 18028 948024 10884 0 31016 52968
   8       1 sda1 78747 8 1992962 42292 14077 18028 948024 10884 0 30996 52904
  20       0 dm-0 2 0 4 0 0 0 0 0 0 0 0
 105       0 cciss/c1d0 66519761  2495668 1078049969 35954112 33291312 17183132 1041809864 121175252        0 12723812 157113448
 105       1 cciss/c1d0p1 66519761  2495668 1078049969 35954112 33291312 17183132 1041809864 121175252        0 12723812 157113448
"""

SAMPLE_PROC_DISKSTATS_2 = """   1       0 ram0 0 0 0 0 0 0 0 0 0 0 0
   7       0 loop0 0 0 0 0 0 0 0 0 0 0 0
   8       0 sda 79077 39 1995838 42348 14077 18028 948024 10884 0 31016 52968
   8       1 sda1 78747 8 1992962 42292 14077 18028 948024 10884 0 30996 52904
  20       0 dm-0 2 0 4 0 0 0 0 0 0 0 0
 105       0 cciss/c1d0 9223970893425514001  2496485 1078307121 35961888 33295409 17187513 1041925392 121177676        0 12727416 157123632
 105       1 cciss/c1d0p1 9223970893425514002  2496485 1078307121 35961888 33295409 17187513 1041925392 121177676        0 12727416 15712363
"""

''' Sample calculations:
await_r = (35961888-35954112) / (66534428-66519761) = 0.5301697688688893
await_w = (121177676-121175252) / (33295409-33291312) = 0.5916524286062973

nr_ios = (66534428-66519761) + (33295409-33291312) = 18764
systm = (12727416-12723812) / nr_ios = 0.192069921126
'''

# Sample filesystems (/proc/filesystems) file content for testing the
# filesystems whitelist (OPSAPS-21931)
SAMPLE_PROC_FILESYSTEMS = """nodev	anon_inodefs
nodev	devpts
	ext3
	ext2
nodev	ecryptfs
	fuseblk
nodev	fuse
"""

SAMPLE_AWAIT_READ_TIME = 0.5301697688688893
SAMPLE_AWAIT_WRITE_TIME = 0.5916524286062973
SAMPLE_SERVICE_TIME = 0.192069921126

FIXTURE_PATH = os.path.join(os.path.dirname(__file__), "../test-fixtures", "host-fixture")

def test_block_device_stats_equals():
  stats1 = BlockDeviceStats([5, 0, 0, 10, 5, 0, 0, 10, 0, 0, 0])
  stats2 = BlockDeviceStats([5, 0, 0, 10, 5, 0, 0, 10, 0, 0, 0])
  assert_false(stats1 is stats2)
  assert_equals(stats1, stats2)
  assert_true(stats1 == stats2)

  stats2 = BlockDeviceStats([100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100])
  assert_false(stats1 is stats2)
  assert_not_equals(stats1, stats2)
  assert_false(stats1 == stats2)

def test_disk_await_calculation():
  stats1 = BlockDeviceStats([5, 0, 0, 10, 5, 0, 0, 10, 0, 0, 0])
  stats2 = BlockDeviceStats([10, 0, 0, 20, 10, 0, 0, 20, 0, 0, 0])
  assert_equals((2, 2, 2), calculate_await_times(stats2, stats1))
  assert_equals((None, None, None), calculate_await_times(stats1, stats2))

def test_disk_svctm_calculation():
  stats1 = BlockDeviceStats([5, 0, 0, 0, 5, 0, 0, 0, 0, 10, 0])
  stats2 = BlockDeviceStats([10, 0, 0, 0, 10, 0, 0, 0, 0, 30, 0])
  assert_equals(2, calculate_service_time(stats2, stats1))
  assert_equals(None, calculate_service_time(stats1, stats2))

def test_disk_utilization_calculation():
  stats1 = BlockDeviceStats([5, 0, 0, 0, 5, 0, 0, 0, 0, 10, 0])
  stats2 = BlockDeviceStats([10, 0, 0, 0, 10, 0, 0, 0, 0, 30, 0])
  assert_equals(0.03333333333333333, calculate_utilization(stats2,
                                                           stats1,
                                                           60000.0))
  assert_equals(None, calculate_utilization(stats1, stats2, 60000.0))

def test_disk_average_queue_length_calculation():
  stats1 = BlockDeviceStats([5, 0, 0, 0, 5, 0, 0, 0, 0, 10, 100.0])
  stats2 = BlockDeviceStats([10, 0, 0, 0, 10, 0, 0, 0, 0, 30, 1000.0])
  assert_equals(0.015, calculate_average_queue_length(stats2,
                                                      stats1,
                                                      60000.0))
  assert_equals(None, calculate_average_queue_length(stats1,
                                                     stats2,
                                                     60000.0))

def test_disk_average_request_size_calculation():
  stats1 = BlockDeviceStats([5, 0, 10, 0, 5, 0, 20, 0, 0, 10, 0])
  stats2 = BlockDeviceStats([10, 0, 100, 0, 10, 0, 200, 0, 0, 30, 0])
  assert_equals(27.0, calculate_average_request_size(stats2,
                                                     stats1))
  assert_equals(None, calculate_average_request_size(stats1,
                                                     stats2))

def test_get_disk_devices():
  test_file = tempfile.NamedTemporaryFile()
  try:
    test_file.write(SAMPLE_PROC_PARTITIONS)
    test_file.seek(0)
    disk_devices = DiskDevices.get_disk_devices(partitions_file=test_file.name)
    assert_equals(SAMPLE_DISK_DEVICES, disk_devices)
  finally:
    test_file.close()

def test_get_diskstats():
  test_file = tempfile.NamedTemporaryFile()
  try:
    test_file.write(SAMPLE_PROC_DISKSTATS_2)
    test_file.seek(0)

    diskstats, diskstats_time = \
        DiskDevices.get_diskstats(SAMPLE_DISK_DEVICES, diskstats_file=test_file.name)
    assert_equals(sorted(SAMPLE_DISK_DEVICES), sorted(diskstats.keys()))

    assert_equals(diskstats['sda'], BlockDeviceStats([79077L,39L,1995838L,42348L,14077L,18028L,948024L,10884L,0L,31016L,52968L]))
    assert_equals(diskstats['sda1'], BlockDeviceStats([78747L,8L,1992962L,42292L,14077L,18028L,948024L,10884L,0L,30996L,52904L]))
    assert_equals(diskstats['dm-0'], BlockDeviceStats([2L,0L,4L,0L,0L,0L,0L,0L,0L,0L,0L]))
    assert_equals(diskstats['cciss/c1d0'], BlockDeviceStats([9223970893425514001L,2496485L,1078307121L,35961888L,33295409L,17187513L,1041925392L,121177676L,0L,12727416L,157123632L]))
    assert_equals(diskstats['cciss/c1d0p1'], BlockDeviceStats([9223970893425514002L,2496485L,1078307121L,35961888L,33295409L,17187513L,1041925392L,121177676L,0L,12727416L,15712363L]))

  finally:
    test_file.close()

def test_disk_build_updates():
  prev = { 'cciss/c1d0': BlockDeviceStats(), 'dm-0': BlockDeviceStats() }
  curr = { 'cciss/c1d0': BlockDeviceStats(), 'dm-0': BlockDeviceStats() }
  updates = DiskDevices.build_updates(curr, prev, 60000.0, re.compile("^$"))
  assert_equals(2, len(updates))

  prev = { 'cciss/c1d0': BlockDeviceStats() }
  curr = { 'cciss/c1d0': BlockDeviceStats(), 'dm-0': BlockDeviceStats() }
  updates = DiskDevices.build_updates(curr, prev, 60000.0, re.compile("^$"))
  assert_equals(1, len(updates))
  assert_equals("cciss/c1d0", updates[0].get_device())

  prev = { 'cciss/c1d0': BlockDeviceStats(), 'dm-0': BlockDeviceStats() }
  curr = { 'cciss/c1d0': BlockDeviceStats(), 'dm-0': BlockDeviceStats() }
  updates = DiskDevices.build_updates(curr, prev, 60000.0, re.compile("^cciss.*$"))
  assert_equals(1, len(updates))
  assert_equals("dm-0", updates[0].get_device())

def test_disk_devices_monitor():
  partitions_file = tempfile.NamedTemporaryFile()
  diskstats_file = tempfile.NamedTemporaryFile()
  try:
    partitions_file.write(SAMPLE_PROC_PARTITIONS)
    partitions_file.seek(0)
    diskstats_file.write(SAMPLE_PROC_DISKSTATS_1)
    diskstats_file.seek(0)

    monitor = DiskDevices(partitions_file=partitions_file.name,
                          diskstats_file=diskstats_file.name)
    # diskstats file has been read and first stats collected
    diskstats_file.truncate()

    diskstats_file.write(SAMPLE_PROC_DISKSTATS_2)
    diskstats_file.seek(0)

    monitor.collect()
    updates = monitor.get_disk_updates()
    assert_equals(5, len(updates))

    # Find the DiskUpdate for 'cciss/c1d0p1' for which we have pre-computed calcuations to check
    update = None
    for u in updates:
      if u.get_device() == 'cciss/c1d0p1':
        update = u
    metric_dict = { }
    for metric_obj in update.to_avro_dict()['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict.has_key(schema.AWAIT_READ_TIME))
    assert_true(metric_dict.has_key(schema.AWAIT_WRITE_TIME))
    assert_true(metric_dict.has_key(schema.SERVICE_TIME))
    assert_true(metric_dict.has_key(schema.READ_BYTES))
    assert_true(metric_dict.has_key(schema.WRITE_BYTES))
    assert_true(metric_dict.has_key(schema.READ_IOS))
    assert_true(metric_dict.has_key(schema.WRITE_IOS))
    assert_true(metric_dict.has_key(schema.AVERAGE_REQUEST_SIZE))
    assert_true(metric_dict[schema.SERVICE_TIME] - SAMPLE_SERVICE_TIME < 0.0001)
    assert_true(metric_dict[schema.AWAIT_READ_TIME] - SAMPLE_AWAIT_READ_TIME < 0.0001)
    assert_true(metric_dict[schema.AWAIT_WRITE_TIME] - SAMPLE_AWAIT_WRITE_TIME < 0.0001)
    assert_equals(metric_dict[schema.READ_BYTES], 552093245952)
    assert_equals(metric_dict[schema.WRITE_BYTES], 533465800704)
    # 598856570738195L is the value % (1 << 63) - 1.
    assert_equals(metric_dict[schema.READ_IOS], 598856570738195L)
    assert_equals(metric_dict[schema.WRITE_IOS], 33295409)
    assert_equals(metric_dict[schema.AVERAGE_REQUEST_SIZE],
                  4.040342324457238e-14)
  finally:
    partitions_file.close()
    diskstats_file.close()

SAMPLE_PROC_TCP = \
"""  sl  local_address rem_address   st tx_queue rx_queue tr tm->when retrnsmt   uid  timeout inode
   0: 00000000:C36E 00000000:0000 01 00000000:00000000 00:00000000 00000000   119        0 288886 1 0000000000000000 300 0 0 2 -1
   1: 0100007F:B4B1 00000000:0000 02 00000000:00000000 00:00000000 00000000   119        0 6560844 1 0000000000000000 300 0 0 2 -1
   2: 0100007F:ED33 00000000:0000 03 00000000:00000000 00:00000000 00000000  1000        0 15170726 1 0000000000000000 300 0 0 2 -1
   3: 0100007F:1F54 00000000:0000 04 00000000:00000000 00:00000000 00000000   118        0 7454264 1 0000000000000000 300 0 0 2 -1
   4: 0100007F:1F54 00000000:0000 05 00000000:00000000 00:00000000 00000000   118        0 7454264 1 0000000000000000 300 0 0 2 -1
1000: 0100007F:1F55 00000000:0000 06 00000000:00000000 00:00000000 00000000   119        0 286285 1 0000000000000000 300 0 0 2 -1"""

SAMPLE_PROC_TCP6 = \
"""  sl  local_address                         remote_address                        st tx_queue rx_queue tr tm->when retrnsmt   uid  timeout inode
   0: 00000000000000000000000000000000:1C0E 00000000000000000000000000000000:0000 07 00000000:00000000 00:00000000 00000000  1000        0 25410663 1 0000000000000000 300 0 0 2 -1
   1: 00000000000000000000000000000000:270E 00000000000000000000000000000000:0000 08 00000000:00000000 00:00000000 00000000   124        0 24776364 1 0000000000000000 300 0 0 2 -1
1000: 00000000000000000000000000000000:270E 00000000000000000000000000000000:0000 09 00000000:00000000 00:00000000 00000000   124        0 24776364 1 0000000000000000 300 0 0 2 -1"""

SAMPLE_PROC_NET_DEV = \
"""Inter-|   Receive                                                |  Transmit
 face |bytes    packets errs drop fifo frame compressed multicast|bytes    packets errs drop fifo colls carrier compressed
    lo: 29959835120 29778554    0    0    0     0          0        0    29959835120     29778554    0    0   0   0      0
  eth0: 9223970893425514000 16000000    0    0    0     0    0   0    22200000000 16900000    0    0    0     0       0          0
"""

PROC_NET_DEV_PATTERN = "^\s*\w+:(\s+\d+){16}\s*$"

# Disabled until this works on Python 2.4
@nottest
def test_proc_net_dev_exists():
  if not sys.platform.startswith("linux"):
    raise SkipTest("Requires linux")

  proc_net_dev_pattern = re.compile(PROC_NET_DEV_PATTERN)
  try:
    dev_lines = open("/proc/net/dev", 'r').readlines()[2:]
    for dev_line in filter(None, [line.strip() for line in dev_lines]):
      assert_true(proc_net_dev_pattern.match(dev_line))
  except IOError:
    assert_true(False)

def test_nic_mode():
  test_file = tempfile.NamedTemporaryFile()
  try:
    test_file.write(SAMPLE_PROC_NET_DEV)
    test_file.seek(0)

    network_interfaces = NetworkInterfaces(proc_net_dev = test_file.name)
    network_interfaces.ioctl_ethtool = _ioctl_ethtool_mock_slow
    network_interfaces.collect()

    nics = network_interfaces.get_nic_updates()
    metric_dict = { }
    for metric_obj in nics[0].to_avro_dict()['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']

    assert_true(metric_dict.has_key(schema.SPEED))

    assert_equals(metric_dict[schema.SPEED], 5)
    assert_equals(nics[0]._duplex_mode, 1)
  finally:
    test_file.close()

def test_ethtool_errors():
  for ethtool_fn in [_ioctl_ethtool_unsupported,
                     _ioctl_ethtool_inval,
                     _ioctl_ethtool_unknown_ioerror,
                     _ioctl_ethtool_unknown_error]:
    test_file = tempfile.NamedTemporaryFile()
    try:
      test_file.write(SAMPLE_PROC_NET_DEV)
      test_file.seek(0)

      network_interfaces = NetworkInterfaces(proc_net_dev = test_file.name)
      network_interfaces.ioctl_ethtool = ethtool_fn
      network_interfaces.collect()

      nics = network_interfaces.get_nic_updates()
      metric_dict = { }
      for metric_obj in nics[0].to_avro_dict()['metrics']:
        metric_dict[metric_obj['id']] = metric_obj['value']

      assert_false(metric_dict.has_key(schema.SPEED))
      assert_equals(nics[0]._duplex_mode, DUPLEX_UNKNOWN)
      # 598856570738193L is the value % (1 << 63) - 1.
      assert_equals(598856570738193L, metric_dict[schema.BYTES_RECEIVE])
      assert_equals(16000000, metric_dict[schema.PACKETS_RECEIVE])
      assert_equals(0, metric_dict[schema.ERRS_RECEIVE])
      assert_equals(0, metric_dict[schema.DROP_RECEIVE])
      assert_equals(0, metric_dict[schema.FRAME_RECEIVE])
      assert_equals(22200000000, metric_dict[schema.BYTES_TRANSMIT])
      assert_equals(16900000, metric_dict[schema.PACKETS_TRANSMIT])
      assert_equals(0, metric_dict[schema.ERRS_TRANSMIT])
      assert_equals(0, metric_dict[schema.DROP_TRANSMIT])
      assert_equals(0, metric_dict[schema.COLLS_TRANSMIT])
    finally:
      test_file.close()

def test_network_interfaces():
  # First test the constructor throws if the file doesn't exist
  try:
    NetworkInterfaces(proc_net_dev="/file/doesnt/exist/orimsurprisedifitdoes12345")
  except IOError, ioex:
    assert_true(ioex.errno == errno.ENOENT)

  test_file = tempfile.NamedTemporaryFile()
  try:
    test_file.write(SAMPLE_PROC_NET_DEV)
    test_file.seek(0)

    network_interfaces = NetworkInterfaces(proc_net_dev = test_file.name)
    network_interfaces.ioctl_ethtool = _ioctl_ethtool_mock
    network_interfaces.collect()

    nics = network_interfaces.get_nic_updates()
    # The number of nics returned is 1 if the loopback device was excluded via the regex
    assert_equals(len(nics), 1)
    avro_nic = nics[0].to_avro_dict()
    assert_equals(avro_nic['iface'], "eth0")

    metric_dict = { }
    for metric_obj in avro_nic['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']

    # 598856570738193L is the value % (1 << 63) - 1.
    assert_equals(598856570738193L, metric_dict[schema.BYTES_RECEIVE])
    assert_equals(16000000, metric_dict[schema.PACKETS_RECEIVE])
    assert_equals(0, metric_dict[schema.ERRS_RECEIVE])
    assert_equals(0, metric_dict[schema.DROP_RECEIVE])
    assert_equals(0, metric_dict[schema.FRAME_RECEIVE])
    assert_equals(22200000000, metric_dict[schema.BYTES_TRANSMIT])
    assert_equals(16900000, metric_dict[schema.PACKETS_TRANSMIT])
    assert_equals(0, metric_dict[schema.ERRS_TRANSMIT])
    assert_equals(0, metric_dict[schema.DROP_TRANSMIT])
    assert_equals(0, metric_dict[schema.COLLS_TRANSMIT])
    assert_equals(1234, metric_dict[schema.SPEED])
    assert_equals(nics[0]._duplex_mode, 1)
  finally:
    test_file.close()

SAMPLE_PROC_ENTROPY ="""2000"""

def test_entropy_file_exist():
  try:
    EntropyMonitor(proc_entropy="/file/doesnt/exist")
  except IOError, ioex:
    assert_true(ioex.errno == errno.ENOENT)

def test_entropy():
  entropy_file = tempfile.NamedTemporaryFile()
  try:
    entropy_file.write(SAMPLE_PROC_ENTROPY)
    entropy_file.seek(0)

    update = HostUpdate("/var/log", "/var/run", "/opt/cloudera/parcel")
    entropy_metric = EntropyMonitor(entropy_file.name)
    entropy_metric.add_entropy_metrics(update)

    avro_dict = update.to_avro_dict()
    metric_dict = {}
    for metric_obj in avro_dict['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']

    assert_equals(metric_dict[schema.AVAILABLE_ENTROPY], 2000)

  finally:
    entropy_file.close()


def test_tcp_connection_updates():
  test_tcp = tempfile.NamedTemporaryFile()
  try:
    test_tcp6 = tempfile.NamedTemporaryFile()
    try:
      test_tcp.write(SAMPLE_PROC_TCP)
      test_tcp.seek(0)
      test_tcp6.write(SAMPLE_PROC_TCP6)
      test_tcp6.seek(0)
      update = HostUpdate("/var/log", "/var/run", "/opt/cloudera/parcel")
      tcp_metrics = TcpMetrics(test_tcp.name, test_tcp6.name)
      tcp_metrics.add_tcp_metrics(update)

      avro_dict = update.to_avro_dict()
      metric_dict = { }
      for metric_obj in avro_dict['metrics']:
        metric_dict[metric_obj['id']] = metric_obj['value']

      assert_equals(metric_dict[schema.TCP_CONNECTION_COUNT_ESTABLISHED], 1)
      assert_equals(metric_dict[schema.TCP_CONNECTION_COUNT_SYN_SENT], 1)
      assert_equals(metric_dict[schema.TCP_CONNECTION_COUNT_SYN_RECV], 1)
      assert_equals(metric_dict[schema.TCP_CONNECTION_COUNT_FIN_WAIT1], 1)
      assert_equals(metric_dict[schema.TCP_CONNECTION_COUNT_FIN_WAIT2], 1)
      assert_equals(metric_dict[schema.TCP_CONNECTION_COUNT_TIME_WAIT], 1)
      assert_equals(metric_dict[schema.TCP_CONNECTION_COUNT_CLOSE], 1)
      assert_equals(metric_dict[schema.TCP_CONNECTION_COUNT_CLOSE_WAIT], 1)
      assert_equals(metric_dict[schema.TCP_CONNECTION_COUNT_LAST_ACK], 1)
      assert_equals(metric_dict[schema.TCP_CONNECTION_COUNT_LISTEN], 0)
      assert_equals(metric_dict[schema.TCP_CONNECTION_COUNT_CLOSING], 0)

    finally:
      test_tcp6.close()
  finally:
    test_tcp.close()

def test_filesystems_filter():
  class MockFilesystemMap:
    def get_map(self):
      return {'/data1': MockFilesystem('/data1'),
              '/data2': MockFilesystem('/data2')}

  class MockFilesystem:
    def __init__(self, mount):
      self._mount = mount
    def collect(self):
      pass
    def get_fs_update(self):
      return self._mount

  filesystems = Filesystems(MockFilesystemMap())
  filesystems.set_fs_exclude_regex('^\/data1$')
  filesystems.collect()
  updates = filesystems.get_fs_updates()
  assert_equals(1, len(updates))
  assert_equals('/data2', updates[0])

def test_filesystems_whitelist():
  class MockFilesystemMap(FilesystemMap):
    def __init__(self,
                 monitored_nodev_fstypes,
                 poll_timeout,
                 join_timeout,
                 tolerance,
                 local_filesystem_whitelist,
                 fstypes_info_proc_file):
      FilesystemMap.__init__(self,
                             monitored_nodev_fstypes,
                             poll_timeout,
                             join_timeout,
                             tolerance,
                             local_filesystem_whitelist,
                             fstypes_info_proc_file)

    def get_all_mounted_partitions(self):
      retlist = []
      retlist.append(sdiskpart('/dev/sda1', '/', 'ext2', 'rw'))
      retlist.append(sdiskpart('/dev/sda2', '/windows', 'vfat', 'rw'))
      retlist.append(sdiskpart('/dev/sda3', '/xyz1', 'tmpfs', 'rw'))
      retlist.append(sdiskpart('/dev/sda4', '/xyz2', 'ext4', 'rw'))
      return retlist

  fd, tempFilesystemsInfoFileName = tempfile.mkstemp()
  try:
    os.write(fd, SAMPLE_PROC_FILESYSTEMS)
    os.close(fd);

    mock_fs_map = MockFilesystemMap(["nfs", "nfs4", "tmpfs"], 2, 0.1, 60,
                         ["ext2", "ext3", "ext4"],
                         tempFilesystemsInfoFileName)
    filesystemsMap = mock_fs_map.get_map()

    assert_true(set(filesystemsMap.keys()) == set(["/xyz1", "/", "/xyz2"]))
  finally:
    os.remove(tempFilesystemsInfoFileName);

def test_filesystem_mount_walk():
  fs_map = FilesystemMap(['aufs', 'overlay'])

  # Note that the path doesn't need to exist
  my_path = "/tmp/foo/bar"
  my_fs = fs_map.get_filesystem_by_path(my_path)

  assert_true(my_fs is not None)
  assert_equals(my_fs, fs_map.get_filesystem_by_path(os.path.dirname(my_path)))

def test_host_update():
  daemon = TestMonitorDaemon()
  daemon.firehoses = MockFirehoses("host_id", "hostname")
  reg = daemon.get_registry_for_testing()
  monitor = HostMonitor(daemon, reg)

  monitor.collect()
  avro = monitor.get_host_update().to_avro_dict()

  # Test fs update
  avro_fs_list = avro['fsUpdates']
  assert_true(len(avro_fs_list) >= 1, "You should have a local filesystem")
  fs1 = avro_fs_list[0]
  assert_true(os.path.isdir(fs1['mountpoint']))
  for metric in fs1['metrics']:
    if metric['id'] == schema.CAPACITY:
      assert_true(metric['value'] > 0)
      break
  else:
    assert_true(False, "Metric list does not have the required metric")

  #Test overcommit mode
  assert_true(avro['overcommitMemory'] != constants.OVERCOMMIT_UNKNOWN)

  # Test that metrics are populated.
  by_id = dict( (k["id"], k["value"]) for k in avro["metrics"] )
  for m in [schema.LOAD_1, schema.LOAD_5, schema.LOAD_15, schema.CPU_PERCENT,
      schema.PHYSICAL_MEMORY_USED, schema.PHYSICAL_MEMORY_CACHED,
      schema.PHYSICAL_MEMORY_BUFFERS, schema.PHYSICAL_MEMORY_TOTAL,
      schema.TOTAL_CPU_USER, schema.TOTAL_CPU_SYSTEM, schema.TOTAL_CPU_IDLE,
      schema.TOTAL_CPU_NICE, schema.TOTAL_CPU_IOWAIT, schema.TOTAL_CPU_IRQ,
      schema.TOTAL_CPU_SOFT_IRQ, schema.SWAP_TOTAL, schema.SWAP_FREE,
      schema.SWAP_USED, schema.CORES, schema.PHYSICAL_MEMORY_MEMFREE,
      schema.PHYSICAL_MEMORY_DIRTY, schema.PHYSICAL_MEMORY_WRITEBACK,
      schema.PHYSICAL_MEMORY_MAPPED, schema.PHYSICAL_MEMORY_DIRTY_RATIO,
      schema.UPTIME, schema.PHYSICAL_MEMORY_COMMIT_LIMIT, schema.OVERCOMMIT_RATIO]:
    assert_true(m in by_id, str(m) + " is missing")
  assert_equals(10000, by_id[schema.HMON_MESSAGE_TRANSMIT_DURATION])
  assert_equals(1000, by_id[schema.HMON_MESSAGE_TRANSMIT_SUCCEEDED])
  assert_equals(100, by_id[schema.HMON_MESSAGE_TRANSMIT_FAILED])
  assert_equals(10, by_id[schema.HMON_MESSAGE_BYTES_SENT])
  assert_equals(20000, by_id[schema.SMON_MESSAGE_TRANSMIT_DURATION])
  assert_equals(2000, by_id[schema.SMON_MESSAGE_TRANSMIT_SUCCEEDED])
  assert_equals(200, by_id[schema.SMON_MESSAGE_TRANSMIT_FAILED])
  assert_equals(20, by_id[schema.SMON_MESSAGE_BYTES_SENT])

def test_parcel_dir_collector():
  daemon = TestMonitorDaemon()
  mock_agent = daemon._agent
  mock_agent.parcel_dir_is_set = False
  reg = daemon.get_registry_for_testing()
  monitor = HostMonitor(daemon, reg)

  monitor.collect()
  avro = monitor.get_host_update().to_avro_dict()
  dir_metrics = avro['dirUpdates']
  assert_true(len(dir_metrics) == 2)

  mock_agent.parcel_dir_is_set = True

  monitor.collect()
  avro = monitor.get_host_update().to_avro_dict()
  dir_metrics = avro['dirUpdates']
  assert_true(len(dir_metrics) == 3)

  monitor.collect()
  avro = monitor.get_host_update().to_avro_dict()
  dir_metrics = avro['dirUpdates']
  assert_true(len(dir_metrics) == 3)

def _ioctl_ethtool_mock(sockfd, iface, ecmd):
  # Return speed=1234 and duplex=1, the speed_hi bits are left as 0
  return array.array('B', struct.pack('12xHB29x', 1234, 1))

def _ioctl_ethtool_unsupported(sockfd, iface, ecmd):
  raise IOError(errno.EOPNOTSUPP, "Test- NIC unsupported!!")

def _ioctl_ethtool_inval(sockfd, iface, ecmd):
  raise IOError(errno.EINVAL, "Test- ethtool invalid arg, handle as unsupported!!")

def _ioctl_ethtool_unknown_ioerror(sockfd, iface, ecmd):
  raise IOError(errno.EIO, "Test- unknown IOError, but don't crash!!")

def _ioctl_ethtool_unknown_error(sockfd, iface, ecmd):
  raise Exception("Test- unknown exception, but don't crash!!")

def _ioctl_ethtool_mock_slow(sockfd, iface, ecmd):
  # Return speed=5 and duplex=1, the speed_hi bits are left as 0
  # Given the speed and duplex mode, the "slow mode" metric should be 1
  return array.array('B', struct.pack('12xHB29x', 5, 1))

def test_filesystem_map_disk_usage_wrapper():
  parent_pipe, child_pipe = Pipe(False)
  try:
    disk_usage_wrapper(child_pipe, ["/", "/tmp"])
    usage = parent_pipe.recv()
    assert_true(usage is not None)
    assert_true('/' in usage)
    assert_true('/tmp' in usage)
    assert_true('total' in usage['/'])
    assert_true('used' in usage['/'])
    assert_true('free' in usage['/'])
  finally:
    parent_pipe.close()
    child_pipe.close()

def test_filesystem_map_get_info_for_scm_heartbeat():
  max_int = cmf.util.get_max_int()
  class MockFilesystemMap(FilesystemMap):
    def __init__(self):
      FilesystemMap.__init__(self)
      self._fs_map = {}
      self._fs_map['/a'] = \
          Filesystem(self, '/a', '/a-partition', 'rw,noexec,nosuid,nodev', False, 'blah')
      self._fs_map['/b'] = \
          Filesystem(self, '/b', '/b-partition', 'rw,noexec,nosuid,nodev', True, 'blah')
      self._fs_map['/c'] = \
          Filesystem(self, '/c', '/c-partition', 'ro,noexec,nosuid,nodev', False, 'blah')
      self._fs_map['/cloud'] = \
        Filesystem(self, '/cloud', '/cloud-partition', 'rw,noexec,nosuid,nodev', True, 'blah')

    def disk_usage(self, mountpoint):
      if mountpoint == '/a':
        return dict(total=10, used=6, free=4)
      if mountpoint == '/b':
        return dict(total=20, used=12, free=8)
      if mountpoint == '/c':
        return dict(total=30, used=18, free=12)
      if mountpoint == '/cloud':
        return dict(total=max_int + 1, used=8, free=max_int - 8)
      raise Exception("Unexpected mountpoint %s!" % mountpoint)

  info = MockFilesystemMap().get_info_for_scm_heartbeat()
  filesystem_details = info['filesystem_info']

  # Read-only mount point '/c' skipped.
  assert_equals(3, len(filesystem_details))
  seen = {}

  for fs in filesystem_details:
    if fs['name'] == '/b-partition':
      assert_equals('/b', fs['mount_point'])
      assert_equals(20, fs['total_space_bytes'])
      assert_equals(True, fs['nodev'])
    if fs['name'] == '/a-partition':
      assert_equals('/a', fs['mount_point'])
      assert_equals(10, fs['total_space_bytes'])
      assert_equals(False, fs['nodev'])
    if fs['name'] == '/cloud-partition':
        assert_equals('/cloud', fs['mount_point'])
        # total_space_bytes capped.
        assert_equals(max_int, fs['total_space_bytes'])
        assert_equals(True, fs['nodev'])
    seen[fs['name']] = True
  assert_equals(3, len(seen))

  free_space_info = info['mountpoint_available_space_bytes']
  assert_equals(3, len(free_space_info))
  assert_equals(4, free_space_info['/a'])
  assert_equals(8, free_space_info['/b'])
  assert_equals(max_int - 8, free_space_info['/cloud'])

def test_filesystem_map():
  info = FilesystemMap().get_info_for_scm_heartbeat()
  assert_true(info is not None)
  filesystem_details = info['filesystem_info']
  assert_true(len(filesystem_details) > 0)

def test_filesystem_map_handles_symlinks():
  fs_map = FilesystemMap(['proc', 'aufs', 'overlay'])
  tmp = fs_map.get_filesystem_by_path('/tmp')
  proc = fs_map.get_filesystem_by_path('/proc')
  assert_true(tmp is not None)
  assert_true(proc is not None)
  assert_true(tmp != proc)
  temp_dir = tempfile.mkdtemp()
  link_path = os.path.join(temp_dir, 'test_link')
  try:
    os.symlink('/proc/cpuinfo', link_path)
    assert_true(fs_map.get_filesystem_by_path(link_path).mountpoint == '/proc')
  finally:
    if link_path is not None and os.path.exists(link_path):
      os.unlink(link_path)
    if temp_dir is not None:
      os.rmdir(temp_dir)

def test_disk_usage_for_nodev_filesystem_bad_function_call():
  def _sleep_function(pipe, paths):
    time.sleep(60.0)

  class MockFilesystemMap(FilesystemMap):
    def __init__(self):
      FilesystemMap.__init__(self,
                             join_timeout=10.0,
                             tolerance=10000000.0)
      self._fs_map = {}
      self._fs_map['/tmp'] = Filesystem(self, '/tmp', '/', 'rw', True, 'blah')

    def get_nodev_disk_usage_function(self):
      return _sleep_function

  fs_map = MockFilesystemMap()
  usage = fs_map.disk_usage('/tmp')
  assert_true(usage is None)
  assert_true(fs_map.nodev_refresh_process is None)

def test_disk_usage_for_nodev_filesystem_really_bad_function_call():
  def _sleep_function_with_signal_handler(pipe, paths):
    def _do_nothing_handler(signum, frame):
      pass
    original_handler = signal.signal(signal.SIGTERM, _do_nothing_handler)
    time.sleep(60.0)
    # The first signal will wake us up. Sleep again.
    time.sleep(60.0)

  class MockFilesystemMap(FilesystemMap):
    def __init__(self):
      FilesystemMap.__init__(self, tolerance=10000000.0)
      self._fs_map = {}
      self._fs_map['/tmp'] = Filesystem(self, '/tmp', '/', 'rw', True, 'blah')
      self.get_nodev_disk_usage_function_calls = 0

    def get_nodev_disk_usage_function(self):
      self.get_nodev_disk_usage_function_calls += 1
      return _sleep_function_with_signal_handler

  fs_map = MockFilesystemMap()
  usage = fs_map.disk_usage('/tmp')
  assert_true(usage is None)
  assert_true(fs_map.nodev_refresh_process is not None)
  assert_equals(1, fs_map.get_nodev_disk_usage_function_calls)
  usage = fs_map.disk_usage('/tmp')
  assert_true(usage is None)
  assert_true(fs_map.nodev_refresh_process is not None)
  assert_equals(1, fs_map.get_nodev_disk_usage_function_calls)

def test_disk_usage_for_nodev_filesystem_cache():
  class MockFilesystemMap(FilesystemMap):
    def __init__(self):
      FilesystemMap.__init__(self,
                             join_timeout=10.0,
                             tolerance=10000000.0)
      self._fs_map = {}
      self._fs_map['/tmp'] = Filesystem(self, '/tmp', '/', 'rw', True, 'blah')
      self.get_nodev_disk_usage_function_calls = 0

    def get_nodev_disk_usage_function(self):
      self.get_nodev_disk_usage_function_calls += 1
      return FilesystemMap.get_nodev_disk_usage_function(self)

  fs_map = MockFilesystemMap()
  assert_equals(0, fs_map.get_nodev_disk_usage_function_calls)
  usage = fs_map.disk_usage('/tmp')
  assert_equals(1, fs_map.get_nodev_disk_usage_function_calls)
  assert_true(usage is not None)
  # Another call should hit the cache.
  usage = fs_map.disk_usage('/tmp')
  assert_equals(1, fs_map.get_nodev_disk_usage_function_calls)
  assert_true(usage is not None)
  # If we drop the tolerance, then we should not hit the cache
  fs_map.tolerance = -1.0
  usage = fs_map.disk_usage('/tmp')
  assert_equals(2, fs_map.get_nodev_disk_usage_function_calls)
  assert_true(usage is not None)

def test_dns_names_parse_resolv_conf():
  # Non-existant resolv.conf
  dns_names = DnsNames(None, None)
  dns_names._read_resolv_conf("")
  assert_equals(None, dns_names._domain)
  assert_equals([], dns_names._search)
  # Empty resolv.conf
  fd, resolv_conf = tempfile.mkstemp("conf")
  os.close(fd)
  dns_names = DnsNames(None, None)
  dns_names._read_resolv_conf(resolv_conf)
  assert_equals(None, dns_names._domain)
  assert_equals([], dns_names._search)
  os.unlink(resolv_conf)
  # Valid resolv.conf
  fd, resolv_conf = tempfile.mkstemp("conf")
  os.write(fd, "domain ca.cloudera.com\n")
  os.write(fd, "#domain bad.cloudera.com\n")
  os.write(fd, "search ca.cloudera.com ent.cloudera.com\n")
  os.write(fd, "doman bad.cloudera.com\n")
  os.write(fd, "sarch bad.cloudera.com\n")
  os.close(fd)
  dns_names = DnsNames(None, None)
  dns_names._read_resolv_conf(resolv_conf)
  assert_equals("ca.cloudera.com", dns_names._domain)
  assert_equals(["ca.cloudera.com", "ent.cloudera.com"], dns_names._search)
  os.unlink(resolv_conf)

def test_dns_namesread_resolv_conf():
  class MockDnsNamesOne(DnsNames):
    def set_count(self):
      self._count = 0

    def _read_resolv_conf_contents(self, resolv_conf):
      self._count = self._count + 1
      if self._count % 2 == 0:
        return ["domain even.cloudera.com"]
      else:
        return ["domain odd.cloudera.com"]

  dns_names = MockDnsNamesOne(None, None)
  dns_names.set_count()
  dns_names._read_resolv_conf("", 100)
  assert_equals("odd.cloudera.com", dns_names._domain)
  dns_names._read_resolv_conf("", 399)
  assert_equals("odd.cloudera.com", dns_names._domain)
  dns_names._read_resolv_conf("", 400)
  assert_equals("even.cloudera.com", dns_names._domain)
  dns_names._read_resolv_conf("", 699)
  assert_equals("even.cloudera.com", dns_names._domain)
  dns_names._read_resolv_conf("", 700)
  assert_equals("odd.cloudera.com", dns_names._domain)
  dns_names._read_resolv_conf("", 999)
  assert_equals("odd.cloudera.com", dns_names._domain)
  dns_names._read_resolv_conf("", 1000)
  assert_equals("even.cloudera.com", dns_names._domain)

def test_dns_names():
  DNS_RESOLUTION_STATUS_ERROR    = 0
  DNS_RESOLUTION_STATUS_MISMATCH = 1
  DNS_RESOLUTION_STATUS_MATCH    = 2
  DNS_RESOLUTION_STATUS_SUBPROCESS_FAILED = 3

  class MockDnsNames(DnsNames):
    def __init__(self,
                 package_dir,
                 agent_jar,
                 poll_timeout=30.0,
                 polling_interval = 60.0):
      DnsNames.__init__(self,
                        package_dir,
                        agent_jar,
                        poll_timeout,
                        polling_interval)
      self._monitor_dns = True

    def _agent_jar_path(self):
      return ""

    def _read_resolv_conf_contents(self, resolv_conf):
      return []

    def java_location(self):
      return 'java'

    def set_monitor_dns_collection(self, monitor_dns):
      self._monitor_dns  = monitor_dns

  class MockMatchingDnsNames(MockDnsNames):
    def _subprocess_with_timeout(self, args, timeout=1):
      return (0,
              '{"status": "0", ' +
              '"ip": "10.10.10.10", ' +
              '"hostname": "foo-1.ent.cloudera.com", ' +
              '"canonicalname": "foo-1.ent.cloudera.com", ' +
              '"localhostDuration": "20", ' +
      	      '"canonicalnameDuration": "1" }', "")

  class MockGoodDnsNamesViaDomain(MockDnsNames):
    def _subprocess_with_timeout(self, args, timeout=1):
      return (0,
              '{"status": "0", ' +
              '"ip": "10.10.10.10", ' +
              '"hostname": "foo-1", ' +
              '"canonicalname": "foo-1.ent.cloudera.com", ' +
              '"localhostDuration": "20", ' +
      	      '"canonicalnameDuration": "1" }', "")

    def _read_resolv_conf_contents(self, resolv_conf):
      return ["domain ent.cloudera.com"]

  class MockGoodDnsNamesViaSearch(MockDnsNames):
    def _subprocess_with_timeout(self, args, timeout=1):
      return (0,
              '{"status": "0", ' +
              '"ip": "10.10.10.10", ' +
              '"hostname": "foo-1", ' +
              '"canonicalname": "foo-1.ent.cloudera.com", ' +
              '"localhostDuration": "20", ' +
      	      '"canonicalnameDuration": "1" }', "")

    def _read_resolv_conf_contents(self, resolv_conf):
      return ["search bad.domain.com ent.cloudera.com"]

  class MockMismatchedDnsNames(MockDnsNames):
    def _subprocess_with_timeout(self, args, timeout=1):
      return (0,
              '{"status": "0", ' +
              '"ip": "10.10.10.10", ' +
              '"hostname": "foo-1.ent.cloudera.com", ' +
              '"canonicalname": "foo-2.ent.cloudera.com", ' +
              '"localhostDuration": "20", ' +
      	      '"canonicalnameDuration": "1" }', "")

  class MockErrorDnsNames(MockDnsNames):
    def _subprocess_with_timeout(self, args, timeout=1):
      return (-1, "", "")

  class MockExceptionDnsNames(MockDnsNames):
    def _subprocess_with_timeout(self, args, timeout=1):
      raise Exception

  dns_names = MockMatchingDnsNames(None, None)

  # Make sure that if the collection is turned off collection_dns_names_metrics
  # is a no-op.
  dns_names.set_monitor_dns_collection(False)
  dns_names.collect_dns_metrics()
  assert_true(dns_names._dns_names is None)
  # Turn it on again.
  dns_names.set_monitor_dns_collection(True)
  dns_names.collect_dns_metrics()
  assert_true(dns_names._dns_names is not None)
  assert_equals(DNS_RESOLUTION_STATUS_MATCH, dns_names._dns_names['status'])
  assert_equals(20, dns_names._dns_names['duration'])
  assert_equals(2, dns_names._dns_names['status'])

  # Make sure that after you get the update the last update becomes null.
  update = HostUpdate("/var/log", "/var/run", "/opt/cloudera/parcel")
  dns_names.add_dns_metrics(update)
  avro_dict = update.to_avro_dict()
  metric_dict = { }
  for metric_obj in avro_dict['metrics']:
    metric_dict[metric_obj['id']] = metric_obj['value']
  assert_true(dns_names._dns_names is None)
  assert_equals(metric_dict[schema.DNS_NAME_RESOLUTION_DURATION], 20)
  assert_equals(metric_dict[schema.DNS_NAME_RESOLUTION_STATUS], 2)

  dns_names = MockGoodDnsNamesViaDomain(None, None)
  dns_names.collect_dns_metrics()
  assert_true(dns_names._dns_names is not None)
  assert_equals(DNS_RESOLUTION_STATUS_MATCH, dns_names._dns_names['status'])
  assert_equals(20, dns_names._dns_names['duration'])
  assert_equals(2, dns_names._dns_names['status'])

  dns_names = MockGoodDnsNamesViaSearch(None, None)
  dns_names.collect_dns_metrics()
  assert_true(dns_names._dns_names is not None)
  assert_equals(DNS_RESOLUTION_STATUS_MATCH, dns_names._dns_names['status'])
  assert_equals(20, dns_names._dns_names['duration'])
  assert_equals(2, dns_names._dns_names['status'])

  dns_names = MockMismatchedDnsNames(None, None)
  dns_names.collect_dns_metrics()
  assert_true(dns_names._dns_names is not None)
  assert_equals(DNS_RESOLUTION_STATUS_MISMATCH, dns_names._dns_names['status'])
  assert_equals(20, dns_names._dns_names['duration'])
  assert_equals(1, dns_names._dns_names['status'])

  dns_names = MockErrorDnsNames(None, None)
  dns_names.collect_dns_metrics()
  assert_true(dns_names._dns_names is not None)
  assert_equals(DNS_RESOLUTION_STATUS_SUBPROCESS_FAILED,
                dns_names._dns_names['status'])

  dns_names = MockExceptionDnsNames(None, None)
  dns_names.collect_dns_metrics()
  assert_true(dns_names._dns_names is not None)
  assert_equals(DNS_RESOLUTION_STATUS_SUBPROCESS_FAILED,
                dns_names._dns_names['status'])

def test_run_in_subprocess_with_timeout():
  dns_names = DnsNames("", "")
  dns_names._subprocess_with_timeout(["sleep", "1"], 10)
  assert_raises(Exception,
                dns_names._subprocess_with_timeout,
                ["sleep", "1"],
                0.5)
  try:
    dns_names._subprocess_with_timeout(["sleep", "10"], 0.5)
    assert_true(False, "timeout failed")
  except Exception, e:
    assert_equals(("timeout with args ['sleep', '10']",), e.args)

_UBUNTU1204_NTPQ = """
         remote           refid      st t when poll reach   delay   offset  jitter
==============================================================================
+172.17.2.3      71.19.144.130    3 u    7   64    3   52.297    0.005   0.245
*172.17.2.4      45.79.111.114    3 u    1   64    7   52.392   -1.245   0.059
"""

_SLES11_NTPQ = """
      remote           refid      st t when poll reach   delay   offset  jitter
==============================================================================
+172.17.2.3      71.19.144.130    3 u   14   64   37   52.394   -0.246  97.122
*172.17.2.4      45.79.111.114    3 u    8   64   37   52.321   -1.031  96.342
"""

_NOT_SYNCHRONIZED_NTPQ = """
         remote           refid      st t when poll reach   delay   offset  jitter
==============================================================================
+172.17.2.3      71.19.144.130    3 u    7   64    3   52.297    0.005   0.245
=172.17.2.4      45.79.111.114    3 u    1   64    7   52.392   -1.245   0.059
"""

def test_parse_ntpdc_output():
  ntp_monitor = NTPMonitor()

  status, offset = ntp_monitor._parse_ntpd_clock_offset(_UBUNTU1204_NTPQ)
  assert_equals(constants.NTP_STATUS_RUNNING, status)
  assert_equals(1.245, offset)

  status, offset = ntp_monitor._parse_ntpd_clock_offset(_SLES11_NTPQ)
  assert_equals(constants.NTP_STATUS_RUNNING, status)
  assert_equals(1.031, offset)

  status, offset = ntp_monitor._parse_ntpd_clock_offset(_NOT_SYNCHRONIZED_NTPQ)
  assert_equals(constants.NTP_STATUS_NOT_SYNCHRONIZED, status)
  assert_equals(None, offset)

  class MockNTPMonitor(NTPMonitor):
    def __init__(self, stderr):
      NTPMonitor.__init__(self)
      self._stderr = stderr
      self._is_chronyd_running = False

    def _subprocess_with_timeout(self, args, timeout):
      return 0, "", self._stderr

  for message in ["ntpq: read: Connection refused",
                  "Name or service not known"]:
    mock_ntp_monitor = MockNTPMonitor(message)
    mock_ntp_monitor.collect()
    assert_equals(constants.NTP_STATUS_UNREACHEABLE,
                  mock_ntp_monitor._ntp_status)
    assert_equals(None, mock_ntp_monitor._clock_offset)


_RHEL7_CHRONYC = """
210 Number of sources = 4
MS Name/IP address         Stratum Poll Reach LastRx Last sample
===============================================================================
^- 108.61.194.85.vultr.com       2   6    17     3  +2753us[+2753us] +/-  198ms
^* 104.156.99.30                 2   6    17     5    -74us[+1020us] +/-   52ms
^- blue.1e400.net                2   6    17     4  -2856us[-2856us] +/-   62ms
^- time.gac.edu                  2   6    17     4   -222us[ -222us] +/-   88ms
"""

_UBUNTU1404_CHRONYC = """
210 Number of sources = 4
MS Name/IP address         Stratum Poll Reach LastRx Last sample
===============================================================================
^- ec2-54-83-7-186.compute-1     2   8    17     1   +386us[-1427us] +/-   90ms
^+ clock.trit.net                2   8    17     2  +1490us[+1157ms] +/-   26ms
^* 131.107.13.100                1   8    17     0  -1166us[-2979us] +/-   16ms
^? time-c.nist.gov               1   8     3     6    -29ms[+1127ms] +/-   74ms
"""


_LOCAL_CLOCK_CHRONYC = """
210 Number of sources = 4
MS Name/IP address         Stratum Poll Reach LastRx Last sample
===============================================================================
^- ec2-54-83-7-186.compute-1     2   8    17     1   +386us[-1427us] +/-   90ms
^+ clock.trit.net                2   8    17     2  +1490us[+1157ms] +/-   26ms
#* 131.107.13.100                1   8    17     0  -1166us[-2979us] +/-   16ms
^? time-c.nist.gov               1   8     3     6    -29ms[+1127ms] +/-   74ms
"""


_NOT_SYNCHRONIZED_CHRONYC = """
210 Number of sources = 4
MS Name/IP address         Stratum Poll Reach LastRx Last sample
===============================================================================
^- ec2-54-83-7-186.compute-1     2   8    17     1   +386us[-1427us] +/-   90ms
^+ clock.trit.net                2   8    17     2  +1490us[+1157ms] +/-   26ms
^- 131.107.13.100                1   8    17     0  -1166us[-2979us] +/-   16ms
^? time-c.nist.gov               1   8     3     6    -29ms[+1127ms] +/-   74ms
"""

def test_parse_chronyc_output():
  ntp_monitor = NTPMonitor()

  status, offset = ntp_monitor._parse_chronyd_clock_offset(_RHEL7_CHRONYC)
  assert_equals(constants.NTP_STATUS_RUNNING, status)
  assert_equals(1.02, offset)

  status, offset = ntp_monitor._parse_chronyd_clock_offset(_UBUNTU1404_CHRONYC)
  assert_equals(constants.NTP_STATUS_RUNNING, status)
  assert_equals(2.979, offset)

  status, offset = ntp_monitor._parse_chronyd_clock_offset(_LOCAL_CLOCK_CHRONYC)
  assert_equals(constants.NTP_STATUS_RUNNING, status)
  assert_equals(2.979, offset)

  status, offset = ntp_monitor._parse_chronyd_clock_offset(_NOT_SYNCHRONIZED_CHRONYC)
  assert_equals(constants.NTP_STATUS_NOT_SYNCHRONIZED, status)
  assert_equals(None, offset)

  class MockNTPMonitor(NTPMonitor):
    def __init__(self):
      NTPMonitor.__init__(self)
      self._is_chronyd_running = True

    def _subprocess_with_timeout(self, args, timeout):
      return 1, "506 Cannot talk to daemon", ""

  mock_ntp_monitor = MockNTPMonitor()
  mock_ntp_monitor.collect()
  assert_equals(constants.NTP_STATUS_UNREACHEABLE,
                mock_ntp_monitor._ntp_status)
  assert_equals(None, mock_ntp_monitor._clock_offset)

class MockCertConfig:
  client_cert_file = ""
  def __init__(self, cert_file):
    self.client_cert_file = cert_file

def test_host_cert_metric_expired():
  expired_cert = os.path.join(FIXTURE_PATH, "expired-cert.pem")
  cert_metric = CertificateMetrics(MockCertConfig(expired_cert))
  update = HostUpdate("/var/log", "/var/run", "/opt/cloudera/parcel")
  cert_metric.add_certificate_metrics(update)
  avro_dict = update.to_avro_dict()
  metric_dict = {}
  for metric_obj in avro_dict['metrics']:
    metric_dict[metric_obj['id']] = metric_obj['value']
  assert_true(metric_dict[schema.AGENT_CERT_EXPIRY] < 0)

def test_host_cert_metric_valid():
  valid_cert = os.path.join(FIXTURE_PATH, "valid-cert.pem")
  cert_metric = CertificateMetrics(MockCertConfig(valid_cert))
  update = HostUpdate("/var/log", "/var/run", "/opt/cloudera/parcel")
  cert_metric.add_certificate_metrics(update)
  avro_dict = update.to_avro_dict()
  metric_dict = {}
  for metric_obj in avro_dict['metrics']:
    metric_dict[metric_obj['id']] = metric_obj['value']
  assert_true(metric_dict[schema.AGENT_CERT_EXPIRY] > 0)

