# Copyright (c) 2012 Cloudera, Inc. All rights reserved.

import datetime
import os
import shutil
try:
  from cStringIO import StringIO
except:
  from StringIO import StringIO
import sys
import time
import zlib

from cmf.monitor import constants, schema, conf
from cmf.monitor.generic import GenericMonitor
from cmf.monitor.generic.impala_adapters import ImpaladAdapter
from cmf.monitor.impalad.query_monitor import QueryMonitor,     \
    _parse_executing_query_ids, _parse_executing_query_profile, \
    QueryStreamMonitor, _utc_from_local_timestamp
from cmf.monitor.monitor_properties import ConfParser
from cmf.safety_valve import SafetyValve
from cmf_test.monitor import make_config_file, replace_config_file, TestMonitorDaemon
from cmf_test.monitor.test_catalogserver import MockOpenUrl
from cmf_test.monitor import parse_metrics_sample
from test_impalad_data import SAMPLE_EXECUTING_QUERY_IDS, \
                              SAMPLE_ENCODED_QUERY_PROFILE

from nose.plugins.skip import SkipTest
from nose.tools import assert_true, assert_equals, assert_false
from thrift.protocol import TCompactProtocol
from thrift.transport import TTransport
from thrift.TSerialization import deserialize

# CM 4.5 config file format: no query monitoring settings
def _make_cm45_format_config_file(version, role_name):
  properties = """[impalad]
role_name = %s
service_name = my_impala
service_version = %d
host = localhost
log_dir = /path/to/logs
collect_interval = 1.0
impalad_webserver_port = 9091
"""
  return  make_config_file(properties % (role_name, version))

# CM 4.6 config file format: we added query monitoring properties
def _make_cm46_format_config_file(version, role_name, period="1.0"):
  properties = """[impalad]
role_name = %s
service_name = my_impala
service_version = %d
host = localhost
log_dir = /path/to/logs
collect_interval = 1.0
impalad_webserver_port = 9091
executing_queries_timeout_seconds = 5.0
query_monitoring_period_seconds = %s
"""
  return  make_config_file(properties % (role_name, version, period,))

properties_cm52 = """[impalad]
role_name = %s
service_name = my_impala
service_version = %d
host = localhost
log_dir = /path/to/logs
collect_interval = 1.0
impalad_webserver_port = 9091
executing_queries_timeout_seconds = 5.0
query_monitoring_period_seconds = %s
scratch_dirs = %s
"""

properties_cm52_scratch_dirs = ["/impala/impalad", "/impala/impalad2"]
# CM 5.2 config file format: we added scratch dirs in impalad
def _make_cm52_format_config_file(version,
                                  role_name,
                                  period="1.0",
                                  scratch_dirs=",".join(properties_cm52_scratch_dirs)):
  return make_config_file(properties_cm52 % (role_name,
                                             version,
                                             period,
                                             scratch_dirs))

class MockQueryMonitor(QueryStreamMonitor):
  def __init__(self):
    QueryStreamMonitor.__init__(self)
    self.completed_queries = []
    self.executing_query_ids = []
    self.executing_query_profiles = {}
    self._monitored_executing_query_ids = True
    self._monitored_executing_query_profile = True
    self._monitored_completed_queries = True

  def get_completed_queries(self,
                            log_path,
                            end_datetime=None,
                            completed_query_report_limit=1000):
    if not self._monitored_completed_queries:
      raise ValueError
    return self.completed_queries

  def get_executing_query_ids(self, url, timeout, username=None, password=None):
    return self.executing_query_ids, self._monitored_executing_query_ids

  def get_executing_query_profile(self, url, query_id, timeout, username=None,
                                  password=None):
    return self.executing_query_profiles.get(query_id, None), \
           self._monitored_executing_query_profile

  def _query_log_dir_exists(self, query_log_dir):
    return True

def test_impalad_lifecycle_cdh5():
  do_impalad_lifecycle(constants.SERVICE_VERSION_CDH5,
                       _make_cm46_format_config_file,
                       ImpaladAdapter.SAMPLE_CDH5)
  do_impalad_lifecycle(constants.SERVICE_VERSION_CDH5,
                       _make_cm52_format_config_file,
                       ImpaladAdapter.SAMPLE_CDH5,
                       properties_cm52_scratch_dirs)

def metrics_to_map(metrics):
  metric_dict = { }
  for metric_obj in metrics:
    metric_dict[metric_obj['id']] = metric_obj['value']
  return metric_dict

def do_impalad_lifecycle(version,
                         make_config_file_function,
                         sample, expect_scratch_dirs = [],
                         tolerate_some_missing_metrics = False):
  daemon = TestMonitorDaemon()
  reg = daemon.get_registry_for_testing()
  query_monitor = MockQueryMonitor()
  adapter = ImpaladAdapter(daemon, SafetyValve(), query_monitor)
  monitor = GenericMonitor("IMPALA", "IMPALAD", daemon, reg, adapter)
  monitor._url_collector._urlopen_callout = MockOpenUrl(sample)
  monitor.refresh_pid(os.getpid(), True)

  # Make two different configs
  conf_dir = make_config_file_function(version, "impalad-test")
  conf_dir2 = make_config_file_function(version, "impalad")

  try:
    assert_equals(0, len(reg.get_registry()))

    # Start it
    monitor.refresh(conf_dir)
    assert_equals(monitor, reg.get_registry().pop())
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    assert_equals('impalad-test', role_update['rolename'])

    # Refresh with a different config
    monitor.refresh(conf_dir2)
    assert_equals(monitor, reg.get_registry().pop())
    assert_equals(1, len(reg.get_registry()))

    monitor.collect()

    # Test metrics and role update
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()

    assert_equals('impalad', role_update['rolename'])
    assert_equals('my_impala', role_update['service_name'])
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_true(metric_dict[schema.TOTAL_CPU_SYSTEM] >= 0.0)
    assert_true(metric_dict[schema.MEM_RSS] >= 0.0)
    assert_true(metric_dict[schema.MEM_VIRTUAL] >= 0.0)
    assert_equals(constants.WEB_METRICS_COLLECTION_GOOD,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)
    metrics_found = 0
    version_string = constants.VERSION_STRINGS[version]
    for metric_id in daemon.metric_accessors['IMPALAD'][version_string].iterkeys():
      if metric_id in metric_dict:
        # See OPSAPS-11201 for more detail about why we perform this check. Also
        # see add_metrics_impala in json_metrics_utils for more detail. This will
        # eventually be removed.
        assert_equals(type(float(1.0)), type(metric_dict[metric_id]))
        metrics_found += 1
      else:
        assert_true(tolerate_some_missing_metrics)
    assert_true(metrics_found > 0)

    dir_metrics = role_update['dirUpdates']
    assert_true(len(dir_metrics) == (len(expect_scratch_dirs) + 1),
        "We configured %d dirs: %s" % ((len(expect_scratch_dirs) + 1) , dir_metrics))
    assert_true(os.path.ismount(dir_metrics[0]['mountpoint']))
    assert_equals('/path/to/logs', dir_metrics[0]['path'])
    for dir_metric in dir_metrics[1:]:
      assert_true(dir_metric['path'] in expect_scratch_dirs)

    assert_true(schema.UNEXPECTED_EXITS_COUNT  in metric_dict)

    impala_daemon_update = role_update['impalaDaemonUpdate']
    assert_equals(1, impala_daemon_update['ready'])

    # Test parsing mdl-defined metrics.
    if version >= constants.SERVICE_VERSION_CDH5:
      metrics_path = os.path.join(conf_dir2, conf.SERVICE_METRICS_PROPERTIES)
      file(metrics_path, 'w').write(
"""{
  "IMPALAD" : {
    "59567" : {
      "source" : "impala.thrift-server.backend.connection-setup-queue-size"
    },
    "59568" : {
      "source" : "resource-broker.request-response-time::count"
    }
  },
  "IMPALA_DAEMON_POOL" : {
    "34922" : {
      "counter" : true,
      "source" : "total-admitted"
    },
    "34923" : {
      "counter" : true,
      "source" : "total-released"
    }
  }
}
""")
      monitor.refresh(conf_dir2)
      monitor.collect()
      monitor.clear_last_reported_timestamp()
      role_update = monitor.get_role_update().to_avro_dict()
      assert_equals('impalad', role_update['rolename'])
      assert_equals('my_impala', role_update['service_name'])
      # Test mdl-defined IMPALAD metric
      metric_dict = { }
      for metric_obj in role_update['metrics']:
        metric_dict[metric_obj['id']] = metric_obj['value']
      # Test a : v
      assert_equals(0, metric_dict[59567])
      # Test a::b : v
      assert_equals(97581, metric_dict[59568])
      # Test mdl-defined IMPALAD POOL metric
      entity_type_updates = { }
      for etu_obj in role_update['entityTypeUpdates']:
        pool_name_to_entities = { }
        for entity_obj in etu_obj['entities']:
          pool_name = entity_obj["attributes"]["poolName"]
          pool_name_to_entities[pool_name] = entity_obj
        entity_type_updates[etu_obj['entityType']] = pool_name_to_entities
      pools = entity_type_updates[ImpaladAdapter._DAEMON_POOL_ENTITY_TYPE]
      assert_true(pools is not None)

      root_admin = pools["root.admin.child"]
      assert_true(root_admin is not None)
      metric_dict = metrics_to_map(root_admin['metrics'])
      assert_equals(2, metric_dict[34922])
      assert_equals(13, metric_dict[34923])

      root_foo = pools["foo"]
      assert_true(root_foo is not None)
      metric_dict = metrics_to_map(root_foo['metrics'])
      assert_equals(5, metric_dict[34922])
      assert_equals(7, metric_dict[34923])

    # Test when the process is not running
    monitor.refresh_pid(None)
    monitor.collect()

    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()

    assert_equals('impalad', role_update['rolename'])
    assert_equals('my_impala', role_update['service_name'])
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']

    dir_metrics = role_update['dirUpdates']
    assert_true(len(dir_metrics) == (len(expect_scratch_dirs) + 1),
        "We configured %d dirs: %s" % ((len(expect_scratch_dirs) + 1), dir_metrics))
    assert_true(os.path.ismount(dir_metrics[0]['mountpoint']))
    assert_equals('/path/to/logs', dir_metrics[0]['path'])
    for dir_metric in dir_metrics[1:]:
      assert_true(dir_metric['path'] in expect_scratch_dirs)

    impala_daemon_update = role_update['impalaDaemonUpdate']
    assert_equals(constants.IMPALA_DAEMON_READY_STATE_UNKNOWN,
                  impala_daemon_update['ready'])

    # Test with a collection error
    monitor.refresh_pid(os.getpid(), True)
    monitor._url_collector._urlopen_callout.set_throw_error(IOError("BOOM!"))
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_true(metric_dict[schema.MEM_RSS] >= 0.0)
    assert_true(metric_dict[schema.MEM_VIRTUAL] >= 0.0)
    assert_equals(constants.WEB_METRICS_COLLECTION_COMMUNICATION_FAILURE,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)

    impala_daemon_update = role_update['impalaDaemonUpdate']
    assert_equals(constants.IMPALA_DAEMON_READY_STATE_UNKNOWN,
                  impala_daemon_update['ready'])

    # Test with some invalid JSON
    monitor._url_collector._urlopen_callout.set_throw_error(None)
    monitor._url_collector._urlopen_callout.set_response("THIS IS BOGUS JSON!")
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_true(metric_dict[schema.MEM_RSS] >= 0.0)
    assert_true(metric_dict[schema.MEM_VIRTUAL] >= 0.0)
    assert_equals(constants.WEB_METRICS_COLLECTION_PARSE_FAILURE,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)

    impala_daemon_update = role_update['impalaDaemonUpdate']
    assert_equals(constants.IMPALA_DAEMON_READY_STATE_UNKNOWN,
                  impala_daemon_update['ready'])

    # Stop it
    monitor.stop()
    assert_equals(0, len(reg.get_registry()))
  finally:
    shutil.rmtree(conf_dir)
    shutil.rmtree(conf_dir2)

def test_query_stream_monitor():
  # We have two version 1.0 log files match the file name prefix below:
  # impala-test-data/query-logs/sample-log_1.0-1411533389294: 128 valid queries
  # impala-test-data/query-logs/sample-log_1.0-1411533389999: 5   valid queries
  path = os.path.join(os.path.dirname(__file__),
                      'impala-test-data/query-logs/sample-log_1.0')
  # Fetch all queries from both version 1.0 files: 128 + 5 = 133
  query_monitor = QueryStreamMonitor(start_datetime=datetime.datetime(2013, 4, 2))
  queries = query_monitor.get_completed_queries(path,
                                                datetime.datetime.utcnow())
  assert_equals(133, len(queries))

  sys.path.append('src/cmf/gen-py')
  from RuntimeProfile.ttypes import TRuntimeProfileTree
  for query_id, profile in queries:
    tree = TRuntimeProfileTree()
    deserialize(tree,
                zlib.decompress(profile),
                protocol_factory=TCompactProtocol.TCompactProtocolFactory())
    assert_true(tree.nodes > 0)

  stats = query_monitor.get_stats()
  assert_true(stats['completed_queries_monitoring_start_datetime'] is not None)
  assert_true(stats['completed_queries_monitoring_last_reported_query_id'] is not None)

  # We have one version 1.1 log file match the file name prefix below:
  # impala-test-data/query-logs/sample-log_1.1-1425476757455: 144 valid queries
  path = os.path.join(os.path.dirname(__file__),
                      'impala-test-data/query-logs/sample-log_1.1')
  query_monitor = QueryStreamMonitor(start_datetime=datetime.datetime(2013, 4, 2))
  queries = query_monitor.get_completed_queries(path,
                                                datetime.datetime.utcnow())
  assert_equals(144, len(queries))

  for query_id, profile in queries:
    tree = TRuntimeProfileTree()
    deserialize(tree,
                zlib.decompress(profile),
                protocol_factory=TCompactProtocol.TCompactProtocolFactory())
    assert_true(tree.nodes > 0)

  stats = query_monitor.get_stats()
  assert_true(stats['completed_queries_monitoring_start_datetime'] is not None)
  assert_true(stats['completed_queries_monitoring_last_reported_query_id'] is not None)

def test_query_stream_monitor_unexpected_start_time_advances():
  path = os.path.join(os.path.dirname(__file__),
                      'impala-test-data/query-logs/non-existant-directory/some.log')
  query_monitor = QueryStreamMonitor(start_datetime=datetime.datetime(2013, 4, 2))
  queries = query_monitor.get_completed_queries(path)
  assert_equals(0, len(queries))
  assert_true(query_monitor._completed_queries_monitoring_start_datetime is not None)

  stats = query_monitor.get_stats()
  assert_equals(1, stats['total_unexpected_start_time_advances'])

def test_completed_query_monitor():
  # We have following log files match the file name prefix below:
  # impala-test-data/query-logs/sample-log_1.0-1411533389294: 128 valid queries
  # impala-test-data/query-logs/sample-log_1.0-1411533389999:   5 valid queries
  # impala-test-data/query-logs/sample-log_1.1-1425476757455: 144 valid queries
  path = os.path.join(os.path.dirname(__file__),
                      'impala-test-data/query-logs/sample-log')
  # First, fetch all queries from both version 1.0 files by specifying an end
  # date time earlier than queries in version 1.1 file.
  query_monitor = QueryMonitor()
  next_start_datetime, next_last_file_timestamp, queries = \
      query_monitor.get_completed_queries(path,
                                          datetime.datetime(2013, 4, 2),
                                          datetime.datetime(2015, 1, 1),
                                          None,
                                          None,
                                          1000)
  assert_equals(133, len(queries))
  assert_equals(next_start_datetime, next_start_datetime)

  sys.path.append('src/cmf/gen-py')
  from RuntimeProfile.ttypes import TRuntimeProfileTree
  for query_id, profile in queries:
    tree = TRuntimeProfileTree()
    deserialize(tree,
                zlib.decompress(profile),
                protocol_factory=TCompactProtocol.TCompactProtocolFactory())
    assert_true(tree.nodes > 0)

  stats = query_monitor.get_stats()
  assert_equals(1, stats['total_get_completed_query_calls'])
  assert_equals(0, stats['total_executing_queries'])
  assert_equals(0, stats['total_executing_queries_ids'])
  assert_equals(133, stats['total_completed_queries'])
  assert_true(stats['total_completed_queries_size_bytes'] > 0)
  assert_equals(0, stats['total_get_executing_query_ids_exceptions'])
  assert_equals(0, stats['total_get_executing_query_profile_exceptions'])

  # Now let's test the slow warning.
  path = os.path.join(os.path.dirname(__file__),
                      'impala-test-data/query-logs/sample-log')
  query_monitor = QueryMonitor(completed_query_slow_log_threshold_seconds=0.0)
  next_start_datetime, next_last_file_timestamp, queries = \
      query_monitor.get_completed_queries(path,
                                          datetime.datetime(2013, 4, 2),
                                          datetime.datetime(2015, 1, 1),
                                          None,
                                          None,
                                          1000)
  assert_equals(133, len(queries))
  assert_equals(next_start_datetime, next_start_datetime)

  stats = query_monitor.get_stats()
  assert_equals(1, stats['total_slow_completed_query_warnings'])

  # Now let's test our reporting limit.
  # Fetch first 9 queries from first file
  path = os.path.join(os.path.dirname(__file__),
                      'impala-test-data/query-logs/sample-log')
  query_monitor = QueryMonitor()
  next_start_datetime, next_last_file_timestamp, queries = \
      query_monitor.get_completed_queries(path,
                                          datetime.datetime(2013, 4, 2),
                                          datetime.datetime(2015, 1, 1),
                                          None,
                                          None,
                                          9)
  assert_equals(9, len(queries))
  stats = query_monitor.get_stats()
  assert_equals(1, stats['total_completed_query_limit_reached'])
  # Fetch 10th-18th queries from first file
  next_start_datetime, next_last_file_timestamp, queries = \
      query_monitor.get_completed_queries(path,
                                          next_start_datetime,
                                          datetime.datetime(2015, 1, 1),
                                          queries[-1][0],
                                          next_last_file_timestamp,
                                          9)
  assert_equals(9, len(queries))
  stats = query_monitor.get_stats()
  assert_equals(2, stats['total_completed_query_limit_reached'])
  # Fetch 19th-27th queries from first file
  next_start_datetime, next_last_file_timestamp, queries = \
      query_monitor.get_completed_queries(path,
                                          next_start_datetime,
                                          datetime.datetime(2015, 1, 1),
                                          queries[-1][0],
                                          next_last_file_timestamp,
                                          9)
  assert_equals(9, len(queries))
  stats = query_monitor.get_stats()
  assert_equals(3, stats['total_completed_query_limit_reached'])
  # Fetch all remaining queries, which are 28th-128th from first file and
  # 1st-5th from second files. Thus 106 in total.
  next_start_datetime, next_last_file_timestamp, queries = \
      query_monitor.get_completed_queries(path,
                                          next_start_datetime,
                                          datetime.datetime(2015, 1, 1),
                                          queries[-1][0],
                                          next_last_file_timestamp,
                                          1000)
  assert_equals(106, len(queries))
  stats = query_monitor.get_stats()
  assert_equals(3, stats['total_completed_query_limit_reached'])

  # Test the case when one pass reads all queries in one file while hitting
  # the limit at the same time and continue from there.
  next_start_datetime, next_last_file_timestamp, queries = \
      query_monitor.get_completed_queries(path,
                                          datetime.datetime(2013, 4, 2),
                                          datetime.datetime(2015, 1, 1),
                                          None,
                                          None,
                                          128)
  assert_equals(128, len(queries))
  next_start_datetime, next_last_file_timestamp, queries = \
      query_monitor.get_completed_queries(path,
                                          next_start_datetime,
                                          datetime.datetime(2015, 1, 1),
                                          queries[-1][0],
                                          next_last_file_timestamp,
                                          6)
  assert_equals(5, len(queries))

  # Test the case when one pass reads all querie in one file without hitting
  # the limit, and continue from there. 1369240416670 below is the last query
  # timestamp in impala-test-data/query-logs/sample-log_1.0-1411533389294
  utc_end_dt = _utc_from_local_timestamp(long(1369240416670) / 1000.0)
  next_start_datetime, next_last_file_timestamp, queries = \
      query_monitor.get_completed_queries(path,
                                          datetime.datetime(2013, 4, 2),
                                          utc_end_dt,
                                          None,
                                          None,
                                          1000)
  assert_equals(128, len(queries))
  next_start_datetime, next_last_file_timestamp, queries = \
      query_monitor.get_completed_queries(path,
                                          next_start_datetime,
                                          datetime.datetime(2015, 1, 1),
                                          queries[-1][0],
                                          next_last_file_timestamp,
                                          6)
  assert_equals(5, len(queries))

def test_completed_query_profile_version_1dot1():
  # impala-test-data/query-logs/sample-log_1.1-1425476757455
  path = os.path.join(os.path.dirname(__file__),
                      'impala-test-data/query-logs/sample-log_1.1')
  query_monitor = QueryMonitor()
  next_start_datetime, next_last_file_timestamp, queries = \
      query_monitor.get_completed_queries(path,
                                          datetime.datetime(2015, 1, 1),
                                          datetime.datetime.utcnow(),
                                          None,
                                          None,
                                          1000)
  assert_equals(144, len(queries))

def test_completed_query_mixed_profile_versions():
  # We have following log files match the file name prefix below:
  # impala-test-data/query-logs/sample-log_1.0-1411533389294: 128 valid queries
  # impala-test-data/query-logs/sample-log_1.0-1411533389999:   5 valid queries
  # impala-test-data/query-logs/sample-log_1.1-1425476757455: 144 valid queries
  path = os.path.join(os.path.dirname(__file__),
                      'impala-test-data/query-logs/sample-log')

  # get all queries from 1.0 and 1.1 version profiles.
  query_monitor = QueryMonitor()
  next_start_datetime, next_last_file_timestamp, queries = \
      query_monitor.get_completed_queries(path,
                                          datetime.datetime(2013, 4, 2),
                                          datetime.datetime.utcnow(),
                                          None,
                                          None,
                                          1000)
  assert_equals(277, len(queries))

  # Test the case when one pass stops at 1st query in 2nd profile without
  # hitting the limit, and continue from there. 1411533393000 below is between
  # 1st and 2nd query timestamp in profile log:
  # impala-test-data/query-logs/sample-log_1.0-1411533389999
  utc_end_dt = _utc_from_local_timestamp(long(1411533393000) / 1000.0)
  query_monitor = QueryMonitor()
  next_start_datetime, next_last_file_timestamp, queries = \
      query_monitor.get_completed_queries(path,
                                          datetime.datetime(2013, 4, 2),
                                          utc_end_dt,
                                          None,
                                          None,
                                          1000)
  assert_equals(129, len(queries))

  # Get 2th-5th queries from 2nd profile(version 1.0) and all 144 queries from
  # 3rd profile(version 1.1)
  next_start_datetime, next_last_file_timestamp, queries = \
      query_monitor.get_completed_queries(path,
                                          next_start_datetime,
                                          datetime.datetime.utcnow(),
                                          queries[-1][0],
                                          next_last_file_timestamp,
                                          1000)
  assert_equals(148, len(queries))

def test_executing_query_id_parsing():
  assert_equals(0, len(_parse_executing_query_ids(StringIO(""))))
  ids = _parse_executing_query_ids(StringIO(SAMPLE_EXECUTING_QUERY_IDS))
  assert_equals(20, len(ids))
  assert_equals("51a22189ccad4e83:b36347e112b4c9dd", ids[0])
  assert_equals("5da636ea6a704539:9d46fa83cc830ff2", ids[1])
  assert_equals("39793344ed9c42e2:8984f8ae9a303ed4", ids[19])

SAMPLE_ENCODED_PROFILE_NOT_FOUND = "Query id 323c59bd14a544f5:a8329df97b197e9a not found."

def test_executing_query_profile_parsing():
  assert_equals(None, _parse_executing_query_profile(StringIO("")))
  assert_equals(None, _parse_executing_query_profile(
    StringIO(SAMPLE_ENCODED_PROFILE_NOT_FOUND)))
  profile = _parse_executing_query_profile(
    StringIO(SAMPLE_ENCODED_QUERY_PROFILE))
  assert_true(profile is not None)
  # Let's try decoding it just to make sure it's what we expect
  #
  # This is less than ideal, but hopefully we'll figure out what we want
  # to do with the generate python and the PYTHONPATH when we write the
  # non-test code.
  sys.path.append('src/cmf/gen-py')
  from RuntimeProfile.ttypes import TRuntimeProfileTree
  tree = TRuntimeProfileTree()
  deserialize(tree,
              zlib.decompress(profile),
              protocol_factory=TCompactProtocol.TCompactProtocolFactory())
  assert_true(tree.nodes > 0)

class MockMonitorDaemon(TestMonitorDaemon):
  def __init__(self):
    TestMonitorDaemon.__init__(self)
    self._last_impala_query_update = None

  def send_impala_query_update(self, impala_query_update):
    self._last_impala_query_update = impala_query_update

def test_query_collection():
  do_query_collection(_make_cm46_format_config_file)

def do_query_collection(make_config_file_function):
  daemon = MockMonitorDaemon()
  reg = daemon.get_registry_for_testing()
  query_monitor = MockQueryMonitor()
  adapter = ImpaladAdapter(daemon, SafetyValve(), query_monitor)
  monitor = GenericMonitor("IMPALA", "IMPALAD", daemon, reg, adapter)
  # To make collector.get_impala_version() happy.
  adapter._query_monitor_collector._urlopen_callout = \
      MockOpenUrl(ImpaladAdapter.SAMPLE_CDH5)
  monitor.refresh_pid(os.getpid(), True)

  conf_dir = make_config_file_function(constants.SERVICE_VERSION_CDH5,
                                       "impalad-test")

  try:
    assert_equals(0, len(reg.get_registry()))

    # We should not blow up even if we're called without the configuration being
    # set.
    # monitor._check_for_queries()
    adapter._query_monitor_collector._check_for_queries()

    assert_true(daemon._last_impala_query_update is None)

    monitor.refresh(conf_dir)
    monitor.collect()

    # We should not have any queries since by default we report nothing from the
    # mock query monitor.
    adapter._query_monitor_collector._check_for_queries()
    assert_true(daemon._last_impala_query_update is None)

    # Now we should have some query information. We should have one executing
    # query and one completed query.
    query_monitor.executing_query_ids = ["test1"]
    query_monitor.executing_query_profiles = \
        { "test1" : "profile1" }
    query_monitor.completed_queries = \
        [ ("test2", "profile2") ]

    adapter._query_monitor_collector._check_for_queries()
    assert_true(daemon._last_impala_query_update is not None)
    update = daemon._last_impala_query_update.to_avro_dict()
    assert_equals('impalad-test', update['roleName'])
    assert_equals('my_impala', update['serviceName'])
    assert_true(update['impalaVersion'].startswith('impalad version'))
    assert_equals(1, len(update['executingQueryProfiles']))
    assert_true(len(update['executingQueryProfiles'][0]
                    ['compressedRuntimeProfile']) > 0)
    assert_equals(1, len(update['completedQueryProfiles']))
    assert_true(len(update['completedQueryProfiles'][0]
                    ['compressedRuntimeProfile']) > 0)

    # Next, let's make sure that we handle the case where we have a query ID but
    # no profile for an executing query. When this happens, we should see the
    # completed queries being returned.
    daemon._last_impala_query_update = None
    query_monitor.executing_query_ids = ["test1"]
    query_monitor.executing_query_profiles = { }
    query_monitor.completed_queries = \
        [ ("test1", "profile1"), ("test3", "profile3") ]

    adapter._query_monitor_collector._check_for_queries()
    assert_true(daemon._last_impala_query_update is not None)
    update = daemon._last_impala_query_update.to_avro_dict()
    assert_equals(0, len(update['executingQueryProfiles']))
    assert_equals(2, len(update['completedQueryProfiles']))

  finally:
    shutil.rmtree(conf_dir)

def test_disabled_query_collection():
  daemon = MockMonitorDaemon()
  reg = daemon.get_registry_for_testing()
  query_monitor = MockQueryMonitor()
  adapter = ImpaladAdapter(daemon, SafetyValve(), query_monitor)
  monitor = GenericMonitor("IMPALA", "IMPALAD", daemon, reg, adapter)
  monitor.refresh_pid(os.getpid(), True)

  conf_dir = _make_cm46_format_config_file(
    constants.SERVICE_VERSION_CDH5, "impalad-test", "0.0")

  try:
    assert_equals(0, len(reg.get_registry()))

    # We should not blow up even if we're called without the configuration being
    # set.
    adapter._query_monitor_collector._check_for_queries()
    assert_true(daemon._last_impala_query_update is None)

    monitor.refresh(conf_dir)
    monitor.collect()

    # We should not have any queries since by default we report nothing from the
    # mock query monitor.
    adapter._query_monitor_collector._check_for_queries()
    assert_true(daemon._last_impala_query_update is None)

    # Add some query information and collect. We should not get any information
    # since the query monitor should be disabled.
    query_monitor.executing_query_ids = ["test1"]
    query_monitor.executing_query_profiles = \
        { "test1" : "profile1" }
    query_monitor.completed_queries = \
        [ ("test2", "profile2") ]

    adapter._query_monitor_collector._check_for_queries()
    assert_true(daemon._last_impala_query_update is None)

  finally:
    shutil.rmtree(conf_dir)

def test_reload_conf_on_conf_timestamp_change():
  daemon = MockMonitorDaemon()
  reg = daemon.get_registry_for_testing()
  query_monitor = MockQueryMonitor()
  adapter = ImpaladAdapter(daemon, SafetyValve(), query_monitor)
  monitor = GenericMonitor("IMPALA", "IMPALAD", daemon, reg, adapter)
  monitor.refresh_pid(os.getpid(), True)

  conf_dir = _make_cm52_format_config_file(
    constants.SERVICE_VERSION_CDH5, "impalad-test", "1.0")

  try:
    assert_equals(0, len(reg.get_registry()))

    monitor.refresh(conf_dir)
    monitor.collect()
    assert_equals(1.0, adapter._query_monitor_collector._query_monitoring_period_seconds)

    new_content = properties_cm52 % ("impalad-test",
                                     constants.SERVICE_VERSION_CDH5,
                                     "2.0",
                                     "/impala/impalad")
    replace_config_file(conf_dir,
                        conf.CLOUDERA_MONITOR_PROPERTIES,
                        new_content)
    monitor.refresh(conf_dir)
    monitor.collect()
    assert_equals(2.0, adapter._query_monitor_collector._query_monitoring_period_seconds)
  finally:
    shutil.rmtree(conf_dir)

class URLMock(object):
  def __init__(self):
    self.data = None
    self.error = None

  def urlopen_with_timeout(self, url, timeout, username=None, password=None):
    if self.error is not None:
      raise self.error
    if self.data is None:
      return None
    return StringIO(self.data)

def test_unexpected_executing_query_ids_format():
  mock = URLMock()
  query_monitor = QueryMonitor(urlopen_with_timeout=mock.urlopen_with_timeout)
  # We should handle None without throwing...
  ids, monitored = query_monitor.get_executing_query_ids('whatever', 1000)
  assert_equals([], ids)
  assert_equals(False, monitored)
  # We should handle an exception without throwing
  mock.error = IOError("BOOM!")
  ids, monitored = query_monitor.get_executing_query_ids('whatever', 1000)
  assert_equals([], ids)
  assert_equals(False, monitored)
  # And let's just double check that good input works nicely
  mock.error = None
  mock.data = SAMPLE_EXECUTING_QUERY_IDS
  ids, monitored = query_monitor.get_executing_query_ids('whatever', 1000)
  assert_equals(20, len(ids))
  assert_equals(True, monitored)
  assert_equals("51a22189ccad4e83:b36347e112b4c9dd", ids[0])
  assert_equals("5da636ea6a704539:9d46fa83cc830ff2", ids[1])
  assert_equals("39793344ed9c42e2:8984f8ae9a303ed4", ids[19])

  stats = query_monitor.get_stats()
  assert_equals(20, stats['total_executing_queries_ids'])
  assert_equals(2, stats['total_get_executing_query_ids_exceptions'])

def test_unexpected_executing_query_profiles_format():
  mock = URLMock()
  query_monitor = QueryMonitor(urlopen_with_timeout=mock.urlopen_with_timeout)
  # We should handle None as a query ID without throwing...
  profile, monitored = query_monitor.get_executing_query_profile('whatever',
                                                                 None,
                                                                 1000)
  assert_equals(None, profile)
  assert_equals(False, monitored)

  # We should handle "" as a query ID without throwing...
  profile, monitored = query_monitor.get_executing_query_profile('whatever',
                                                                 '',
                                                                 1000)
  assert_equals(None, profile)
  assert_equals(False, monitored)

  # We should handle non-base64 garbage as a query ID without throwing...
  profile, monitored = query_monitor.get_executing_query_profile('whatever',
                                                                 'asda!!!',
                                                                 1000)
  assert_equals(None, profile)
  assert_equals(False, monitored)

  # We should handle None without throwing...
  profile, monitored = query_monitor.get_executing_query_profile('whatever',
                                                                 'query-id',
                                                                 1000)
  assert_equals(None, profile)
  assert_equals(False, monitored)

  # We should handle an exception without throwing
  mock.error = IOError("BOOM!")
  profile, monitored = query_monitor.get_executing_query_profile('whatever',
                                                                 'query-id',
                                                                 1000)
  assert_equals(None, profile)
  assert_equals(False, monitored)

  # And let's just double check that good input works nicely
  mock.error = None
  mock.data = SAMPLE_ENCODED_QUERY_PROFILE
  profile, monitored = query_monitor.get_executing_query_profile('whatever',
                                                                 'query-id',
                                                                 1000)
  assert_true(profile is not None)
  assert_equals(True, monitored)

  # Let's try decoding it just to make sure it's what we expect
  #
  # This is less than ideal, but hopefully we'll figure out what we want
  # to do with the generate python and the PYTHONPATH when we write the
  # non-test code.
  sys.path.append('src/cmf/gen-py')
  from RuntimeProfile.ttypes import TRuntimeProfileTree
  tree = TRuntimeProfileTree()
  deserialize(tree,
              zlib.decompress(profile),
              protocol_factory=TCompactProtocol.TCompactProtocolFactory())
  assert_true(tree.nodes > 0)

  stats = query_monitor.get_stats()
  assert_equals(1, stats['total_executing_queries'])
  assert_true(stats['total_executing_queries_size_bytes'] > 0)
  assert_equals(4, stats['total_get_executing_query_profile_exceptions'])

def test_unexpected_completed_format():
  path = os.path.join(os.path.dirname(__file__),
                      'impala-test-data/query-logs/sample-unexpected-format')
  query_monitor = QueryMonitor()
  next_start_datetime, next_last_file_timestamp, queries = \
      query_monitor.get_completed_queries(path,
                                          datetime.datetime(2013, 4, 2),
                                          datetime.datetime(2015, 1, 1),
                                          None,
                                          None,
                                          1000)
  assert_equals(1, len(queries))
  assert_equals(next_start_datetime, next_start_datetime)
  stats = query_monitor.get_stats()
  assert_equals(1, stats['total_completed_query_parse_exceptions'])

  sys.path.append('src/cmf/gen-py')
  from RuntimeProfile.ttypes import TRuntimeProfileTree
  for query_id, profile in queries:
    tree = TRuntimeProfileTree()
    deserialize(tree,
                zlib.decompress(profile),
                protocol_factory=TCompactProtocol.TCompactProtocolFactory())
    assert_true(tree.nodes > 0)

def test_empty_or_non_existant_file():
  path = os.path.join(os.path.dirname(__file__),
                      'impala-test-data/query-logs/non-existant-file.log')
  start_datetime = datetime.datetime(2013, 4, 2)
  query_monitor = QueryStreamMonitor(start_datetime=start_datetime)
  queries = query_monitor.get_completed_queries(path,
                                                datetime.datetime(2015, 1, 1),
                                                1000)

  assert_equals(0, len(queries))
  stats = query_monitor.get_stats()
  assert_equals(0, stats['total_unexpected_start_time_advances'])
  assert_equals(start_datetime,
                query_monitor._completed_queries_monitoring_start_datetime)

def test_query_monitoring_status():
  raise SkipTest
  daemon = MockMonitorDaemon()
  reg = daemon.get_registry_for_testing()
  query_monitor = MockQueryMonitor()
  adapter = ImpaladAdapter(daemon, SafetyValve(), query_monitor)
  monitor = GenericMonitor("IMPALA", "IMPALAD", daemon, reg, adapter)
  adapter._query_monitor_collector._urlopen_callout = \
      MockOpenUrl(ImpaladAdapter.SAMPLE_CDH5)
  monitor.refresh_pid(os.getpid(), True)
  conf_dir = _make_cm52_format_config_file(constants.SERVICE_VERSION_CDH5,
                                           "impalad-test",
                                           "1.0")
  # _error_monitoring_queries should be None initially.
  monitor.refresh(conf_dir)
  assert_equals(0,
                monitor._executing_query_monitoring_failures)
  assert_equals(0,
                monitor._completed_query_monitoring_failures)

  # no errors
  monitor.collect()
  monitor._check_for_queries()
  assert_equals(0,
                monitor._executing_query_monitoring_failures)
  assert_equals(0,
                monitor._completed_query_monitoring_failures)

  query_monitor._monitored_completed_queries = False
  monitor._check_for_queries()
  assert_equals(0,
                monitor._executing_query_monitoring_failures)
  assert_equals(1,
                monitor._completed_query_monitoring_failures)

  query_monitor._monitored_completed_queries = True
  monitor._check_for_queries()
  assert_equals(0,
                monitor._executing_query_monitoring_failures)
  assert_equals(1,
                monitor._completed_query_monitoring_failures)

  query_monitor._monitored_executing_query_ids = False
  monitor._check_for_queries()
  assert_equals(1,
                monitor._executing_query_monitoring_failures)
  assert_equals(1,
                monitor._completed_query_monitoring_failures)

  query_monitor._monitored_executing_query_ids = True
  monitor._check_for_queries()
  assert_equals(1,
                monitor._executing_query_monitoring_failures)
  assert_equals(1,
                monitor._completed_query_monitoring_failures)

  monitor._query_monitor.executing_query_ids = ["test1"]
  query_monitor._monitored_executing_query_profile = False
  monitor._check_for_queries()
  assert_equals(2,
                monitor._executing_query_monitoring_failures)
  assert_equals(1,
                monitor._completed_query_monitoring_failures)
  role_update = monitor.get_role_update().to_avro_dict()
  metric_dict = { }
  for metric_obj in role_update['metrics']:
    metric_dict[metric_obj['id']] = metric_obj['value']
  assert_equals(2, metric_dict[schema.EXECUTING_QUERY_MONITORING_FAILURE])
  assert_equals(1, metric_dict[schema.COMPLETED_QUERY_MONITORING_FAILURE])

def test_pool_update_max():
  daemon = TestMonitorDaemon()
  reg = daemon.get_registry_for_testing()
  safety_valve = SafetyValve()
  adapter = ImpaladAdapter(daemon, safety_valve)
  conf_dir = _make_cm52_format_config_file(constants.SERVICE_VERSION_CDH5,
                                           "my_impalad")
  try:
    conf_parser = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    assert_equals(ImpaladAdapter._POOL_UPDATE_MAX_DEFAULT,
                  adapter._get_pool_update_max(conf_parser))

    safety_valve_str = "my_impala_my_impalad_send_pool_entity_update_max = 5"
    safety_valve.update_safety_valve(
      dict(host_agent_safety_valve=safety_valve_str))
    assert_equals(5, adapter._get_pool_update_max(conf_parser))

  finally:
    shutil.rmtree(conf_dir)
