# Copyright (c) 2011 Cloudera, Inc. All rights reserved.

import os
import shutil
try:
  from cStringIO import StringIO
except:
  from StringIO import StringIO

from cmf.monitor import schema, constants
from cmf.monitor.jobtracker import JobTrackerMonitor, json_parser
from cmf_test.monitor import make_config_file, TestMonitorDaemon

from nose.tools import assert_true, assert_equals

class MockJobTrackerMonitor(JobTrackerMonitor):
  def __init__(self, daemon, registry, string):
    JobTrackerMonitor.__init__(self, daemon, registry)
    self.set_string(string)
    self.set_throw_error(None)

  def set_string(self, string):
    self._string = string

  def set_throw_error(self, error):
    self._error = error

  def urlopen(self, url, **kwargs):
    if self._error:
      raise self._error
    return StringIO(self._string)

# CM 3.7 config file format.
def _make_cm37_format_config_file(version, port):
  properties = """[jobtracker]
role_name = mapred-JOBTRACKER
service_name = my_mapreduce
host = localhost
mapred_job_tracker_http_port = %s
collect_interval = 1.0
"""
  return  make_config_file(properties % (port,))

# CM 4.0 config file: we added the service version
def _make_cm40_format_config_file(version, port):
  properties = """[jobtracker]
role_name = mapred-JOBTRACKER
service_name = my_mapreduce
service_version = %s
host = localhost
mapred_job_tracker_http_port = %s
collect_interval = 1.0
"""
  return  make_config_file(properties % (version, port,))

# CM 4.1 config file: we added the log directory.
def _make_cm41_format_config_file(version, port):
  properties = """[jobtracker]
role_name = mapred-JOBTRACKER
service_name = my_mapreduce
service_version = %s
host = localhost
mapred_job_tracker_http_port = %s
log_dir = /path/to/logs
collect_interval = 1.0
"""
  return  make_config_file(properties % (version, port,))

def test_jt_lifecycle_cdh5():
  do_jt_lifecycle(constants.SERVICE_VERSION_CDH5,
      _make_cm40_format_config_file, json_parser.SAMPLE, False)
  do_jt_lifecycle(constants.SERVICE_VERSION_CDH5,
      _make_cm41_format_config_file, json_parser.SAMPLE, True)

def do_jt_lifecycle(version, make_config_file_function, sample, expect_directory_metrics):
  daemon = TestMonitorDaemon()
  reg = daemon.get_registry_for_testing()
  monitor = MockJobTrackerMonitor(daemon, reg, sample)
  monitor.refresh_pid(os.getpid(), True)

  # Make two different configs
  conf_dir = make_config_file_function(version, 50075)
  conf_dir2 = make_config_file_function(version, 50076)

  try:
    assert_equals(0, len(reg.get_registry()))

    # Start it
    monitor.refresh(conf_dir)
    assert_equals(monitor, reg.get_registry().pop())
    monitor.collect()
    assert_true('50075' in monitor.metrics_url, "Config installed")

    # Refresh with a different config
    monitor.refresh(conf_dir2)
    assert_equals(monitor, reg.get_registry().pop())
    assert_equals(1, len(reg.get_registry()))

    monitor.collect()
    assert_true('50076' in monitor.metrics_url, "Using the new config")

    # Test metrics and role update
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()

    assert_equals('mapred-JOBTRACKER', role_update['rolename'])
    assert_equals('my_mapreduce', role_update['service_name'])
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']

    assert_true(schema.TOTAL_CPU_USER in metric_dict)
    assert_true(schema.TOTAL_CPU_SYSTEM in metric_dict)
    assert_equals(constants.WEB_METRICS_COLLECTION_GOOD,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)

    assert_true(schema.UNEXPECTED_EXITS_COUNT  in metric_dict)

    if version == constants.SERVICE_VERSION_CDH5:
      for metric_id in daemon.metric_accessors['JOBTRACKER']['cdh5'].iterkeys():
        assert_true(metric_id in metric_dict)

    # Test service update
    service_update = monitor.get_service_update().to_avro_dict()
    metric_dict = { }
    for metric_obj in service_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']

    assert_true(schema.JOBS_COMPLETED in metric_dict)
    assert_true(schema.JOBS_FAILED in metric_dict)
    assert_true(schema.JOBS_KILLED in metric_dict)
    assert_true(schema.JOBS_PREPARING in metric_dict)
    assert_true(schema.JOBS_RUNNING in metric_dict)
    assert_true(schema.JOBS_SUBMITTED in metric_dict)
    assert_true(schema.MAPS_RUNNING in metric_dict)
    assert_true(schema.REDUCES_RUNNING in metric_dict)
    assert_true(schema.MAP_SLOTS in metric_dict)
    assert_true(schema.REDUCE_SLOTS in metric_dict)
    assert_true(schema.TRACKERS_BLACKLISTED in metric_dict)
    assert_true(schema.MAPS_FAILED in metric_dict)
    assert_true(schema.REDUCES_FAILED in metric_dict)

    dir_metrics = role_update['dirUpdates']
    if expect_directory_metrics:
      assert_true(len(dir_metrics) == 1, "We configured 1 dirs: %s" % dir_metrics)
      assert_true(os.path.ismount(dir_metrics[0]['mountpoint']))
      assert_equals('/path/to/logs', dir_metrics[0]['path'])
    else:
      assert_true(len(dir_metrics) == 0, "We configured 0 dirs: %s" % dir_metrics)

    # Test when the process is not running
    monitor.refresh_pid(None)
    monitor.collect()

    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()

    assert_equals('mapred-JOBTRACKER', role_update['rolename'])
    assert_equals('my_mapreduce', role_update['service_name'])
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']

    dir_metrics = role_update['dirUpdates']
    if expect_directory_metrics:
      assert_true(len(dir_metrics) == 1, "We configured 1 dirs: %s" % dir_metrics)
      assert_true(os.path.ismount(dir_metrics[0]['mountpoint']))
      assert_equals('/path/to/logs', dir_metrics[0]['path'])
    else:
      assert_true(len(dir_metrics) == 0, "We configured 0 dirs: %s" % dir_metrics)

    assert_equals(None, monitor.get_service_update())

    # Test with a collection error
    monitor.refresh_pid(os.getpid(), True)
    monitor.set_throw_error(IOError("BOOM!"))
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_equals(constants.WEB_METRICS_COLLECTION_COMMUNICATION_FAILURE,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)
    assert_equals(None, monitor.get_service_update())

    # Test with some invalid JSON
    monitor.set_throw_error(None)
    monitor.set_string("THIS IS BOGUS JSON!")
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_equals(constants.WEB_METRICS_COLLECTION_PARSE_FAILURE,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)

    # Stop it
    monitor.stop()
    assert_equals(0, len(reg.get_registry()))
  finally:
    shutil.rmtree(conf_dir)
    shutil.rmtree(conf_dir2)
