# Copyright (c) 2011 Cloudera, Inc. All rights reserved.

import getpass
import logging
import os
import stat
import re
import resource
import shutil
import simplejson
import tempfile
import threading
import time
import urllib2

try:
  from cStringIO import StringIO
except:
  from StringIO import StringIO

from cmf.monitor import constants
from cmf.monitor.generic import GenericMonitor
from cmf.monitor.generic.adapter import Adapter
from cmf.monitor.solrserver import SolrServerMonitor
from cmf.monitor import jmx_beans
from cmf.monitor.daemon import MonitorDaemon
from cmf.monitor.count_down_latch import CountDownLatch
from cmf.monitor.fs_update import FsUpdate
from cmf.monitor.nic_update import NicUpdate
from cmf.monitor.host_update import HostUpdate
from cmf.monitor.host_dir_update import HostDirUpdate
from cmf.monitor.impala_query_update import ImpalaRuntimeProfile, ImpalaQueryUpdate
from cmf.monitor.json_metrics_utils import DictWithAttr
from cmf.monitor.jmx_beans import _get_beans, Bean, MATCH_ANY_PORT, \
    _make_port_aware_rpc_metrics_bean, _make_port_aware_detailed_rpc_metrics_bean, \
    _MGMT_PAUSE_REPLACMENT_GROUP, BeanWithAdditionalAttributes, \
    MatchedGroupReplacement
from cmf.monitor import conf
from cmf.monitor.monitor_properties import ConfParser
from cmf.monitor.proc_metrics_utils import get_fd_count, get_fd_max
from cmf.monitor.proc_metrics_utils import get_swap_out
from cmf.monitor.stoppable_thread import StoppableThread
from cmf.monitor.wakeable_thread import WakeableThread
from cmf.monitor.dir_monitor import DirMonitor
from cmf.monitor.disk_update import DiskUpdate
from cmf.monitor.firehoses import Firehoses
from cmf_test.monitor import make_config_file, MockAgent

from nose.tools import assert_equals, assert_not_equals, assert_raises
from nose.tools import assert_false, assert_true


LOG = logging.getLogger(__name__)

def test_dict_with_attr():
  raw = dict(foo=1, bar=dict(baz=2))

  d = DictWithAttr(raw)
  assert_equals(None, d.hello, "Returns None instead of raising")
  assert_not_equals(None, d.bar)

  # d.bar itself should support getattr
  assert_not_equals(None, d.bar.baz)


_METRICS_SAMPLE = r"""
{
   "dfs":{
      "FSDirectory":[
         [
            {
               "hostName":"centos62-25.ent.cloudera.com",
               "sessionId":""
            },
            {
            }
         ]
      ],
      "FSNamesystem":[
         [
            {
               "hostName":"centos62-25.ent.cloudera.com",
               "sessionId":""
            },
            {
               "BlockCapacity":524288,
               "BlocksTotal":57,
               "CapacityRemainingGB":169,
               "CapacityTotalGB":170,
               "CapacityUsedGB":0,
               "CorruptBlocks":0,
               "ExcessBlocks":0,
               "FilesTotal":41,
               "MissingBlocks":0,
               "PendingDeletionBlocks":0,
               "PendingReplicationBlocks":0,
               "ScheduledReplicationBlocks":0,
               "TotalLoad":28,
               "UnderReplicatedBlocks":0
            }
         ]
      ],
      "namenode":[
         [
            {
               "hostName":"centos62-25.ent.cloudera.com",
               "sessionId":""
            },
            {
               "AddBlockOps":13185,
               "CreateFileOps":13185,
               "DeleteFileOps":13187,
               "FileInfoOps":85,
               "FilesAppended":0,
               "FilesCreated":13190,
               "FilesInGetListingOps":113,
               "FilesRenamed":1,
               "GetBlockLocations":13182,
               "GetListingOps":13331,
               "JournalTransactionsBatchedInSync":1,
               "SafemodeTime":34922,
               "Syncs_avg_time":8,
               "Syncs_num_ops":39566,
               "Transactions_avg_time":0,
               "Transactions_num_ops":52753,
               "blockReport_avg_time":0,
               "blockReport_num_ops":887,
               "fsImageLoadTime":1725
            }
         ]
      ]
   },
   "jvm":{
      "metrics":[
         [
            {
               "hostName":"centos62-25.ent.cloudera.com",
               "processName":"NameNode",
               "sessionId":""
            },
            {
               "gcCount":4278,
               "gcTimeMillis":13335,
               "logError":0,
               "logFatal":0,
               "logInfo":170653,
               "logWarn":0,
               "maxMemoryM":270.25,
               "memHeapCommittedM":22.8125,
               "memHeapUsedM":11.36908,
               "memNonHeapCommittedM":34.1875,
               "memNonHeapUsedM":25.986458,
               "threadsBlocked":0,
               "threadsNew":0,
               "threadsRunnable":7,
               "threadsTerminated":0,
               "threadsTimedWaiting":14,
               "threadsWaiting":38
            }
         ]
      ]
   },
   "rpc":{
      "detailed-metrics":[
         [
            {
               "hostName":"centos62-25.ent.cloudera.com",
               "port":"8020"
            },
            {
               "addBlock_avg_time":3,
               "addBlock_num_ops":13185,
               "blockReceived_avg_time":0,
               "blockReceived_num_ops":39543,
               "blockReport_avg_time":0,
               "blockReport_num_ops":887,
               "blocksBeingWrittenReport_avg_time":0,
               "blocksBeingWrittenReport_num_ops":4,
               "complete_avg_time":7,
               "complete_num_ops":13182,
               "create_avg_time":11,
               "create_num_ops":13185,
               "delete_avg_time":10,
               "delete_num_ops":13187,
               "fsync_avg_time":0,
               "fsync_num_ops":5,
               "getBlockLocations_avg_time":0,
               "getBlockLocations_num_ops":13182,
               "getDatanodeReport_avg_time":18,
               "getDatanodeReport_num_ops":13178,
               "getEditLogSize_avg_time":0,
               "getEditLogSize_num_ops":2643,
               "getFileInfo_avg_time":0,
               "getFileInfo_num_ops":85,
               "getListing_avg_time":1,
               "getListing_num_ops":13331,
               "getProtocolVersion_avg_time":0,
               "getProtocolVersion_num_ops":39551,
               "getStats_avg_time":0,
               "getStats_num_ops":13178,
               "mkdirs_avg_time":0,
               "mkdirs_num_ops":5,
               "register_avg_time":0,
               "register_num_ops":4,
               "rename_avg_time":0,
               "rename_num_ops":1,
               "renewLease_avg_time":0,
               "renewLease_num_ops":118893,
               "rollEditLog_avg_time":0,
               "rollEditLog_num_ops":221,
               "rollFsImage_avg_time":0,
               "rollFsImage_num_ops":221,
               "sendHeartbeat_avg_time":0,
               "sendHeartbeat_num_ops":2113321,
               "setPermission_avg_time":0,
               "setPermission_num_ops":2,
               "setSafeMode_avg_time":0,
               "setSafeMode_num_ops":13179,
               "versionRequest_avg_time":0,
               "versionRequest_num_ops":4
            }
         ]
      ],
      "metrics":[
         [
            {
               "hostName":"centos62-25.ent.cloudera.com",
               "port":"8020"
            },
            {
               "NumOpenConnections":7,
               "ReceivedBytes":762786836,
               "RpcProcessingTime_avg_time":0,
               "RpcProcessingTime_num_ops":2434177,
               "RpcQueueTime_avg_time":0,
               "RpcQueueTime_num_ops":2434177,
               "SentBytes":366694933,
               "callQueueLen":0,
               "rpcAuthenticationFailures":0,
               "rpcAuthenticationSuccesses":0,
               "rpcAuthorizationFailures":0,
               "rpcAuthorizationSuccesses":147986
            }
         ]
      ]
   },
   "ugi":{
      "ugi":[
         [
            {
               "hostName":"centos62-25.ent.cloudera.com"
            },
            {
               "loginFailure_avg_time":0,
               "loginFailure_num_ops":0,
               "loginSuccess_avg_time":0,
               "loginSuccess_num_ops":0
            }
         ]
      ]
   }
}
"""

_METRICS_SAMPLE_JMX = SAMPLE = r"""
{
  "beans" : [ {
    "name" : "java.lang:type=MemoryPool,name=CMS Old Gen",
    "modelerType" : "sun.management.MemoryPoolImpl",
    "Name" : "CMS Old Gen",
    "Type" : "HEAP",
    "Valid" : true,
    "CollectionUsage" : {
      "committed" : 0,
      "init" : 65404928,
      "max" : 85196800,
      "used" : 0
    },
    "CollectionUsageThreshold" : 0,
    "CollectionUsageThresholdCount" : 0,
    "MemoryManagerNames" : [ "ConcurrentMarkSweep" ],
    "PeakUsage" : {
      "committed" : 65404928,
      "init" : 65404928,
      "max" : 85196800,
      "used" : 16997432
    },
    "Usage" : {
      "committed" : 65404928,
      "init" : 65404928,
      "max" : 85196800,
      "used" : 16997432
    },
    "UsageThreshold" : 0,
    "UsageThresholdCount" : 0,
    "CollectionUsageThresholdExceeded" : true,
    "CollectionUsageThresholdSupported" : true,
    "UsageThresholdExceeded" : true,
    "UsageThresholdSupported" : true
  }, {
    "name" : "java.lang:type=Memory",
    "modelerType" : "sun.management.MemoryImpl",
    "Verbose" : false,
    "HeapMemoryUsage" : {
      "committed" : 85000192,
      "init" : 0,
      "max" : 104792064,
      "used" : 19721336
    },
    "NonHeapMemoryUsage" : {
      "committed" : 52822016,
      "init" : 24313856,
      "max" : 136314880,
      "used" : 52529480
    },
    "ObjectPendingFinalizationCount" : 0
  }, {
    "name" : "java.lang:type=MemoryPool,name=Code Cache",
    "modelerType" : "sun.management.MemoryPoolImpl",
    "Name" : "Code Cache",
    "Type" : "NON_HEAP",
    "Valid" : true,
    "CollectionUsage" : null,
    "MemoryManagerNames" : [ "CodeCacheManager" ],
    "PeakUsage" : {
      "committed" : 4521984,
      "init" : 2555904,
      "max" : 50331648,
      "used" : 4434624
    },
    "Usage" : {
      "committed" : 4521984,
      "init" : 2555904,
      "max" : 50331648,
      "used" : 4424576
    },
    "UsageThreshold" : 0,
    "UsageThresholdCount" : 0,
    "CollectionUsageThresholdSupported" : false,
    "UsageThresholdExceeded" : true,
    "UsageThresholdSupported" : true
  }, {
    "name" : "com.cloudera.cmf.event.publish:type=PipelineStageMXBean,stage=roleStageTSStore",
    "modelerType" : "com.cloudera.cmon.pipeline.PipelineStage",
    "StageName" : "roleStageTSStore",
    "EventsDropped" : 3,
    "EventsSkipped" : 0,
    "EventsForwarded" : 0,
    "EventsProcessed" : 39,
    "InputQueueSize" : 0,
    "InputQueueCapacity" : 4096
  }, {
    "name" : "java.lang:type=Runtime",
    "modelerType" : "sun.management.RuntimeImpl",
    "Name" : "17500@bigrig",
    "ClassPath" : "/tmp/agent/process/592-cloudera-mgmt-SERVICEMONITOR:/home/cleroy/.m2/repository/mysql/mysql-connector-java/5.1.15/mysql-connector-java-5.1.15.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-fileupload-1.2.2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jopt-simple-3.2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jersey-server-1.4.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-digester-1.8.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-beanutils-core-1.8.0.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/spring-webmvc-3.0.4.RELEASE.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jsp-api-2.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/aspectjrt-1.6.8.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/ecj-3.5.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/configgen-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/snappy-java-1.0.3.2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/hadoop-core-0.20.2-cdh3u2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jetty-6.1.26.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jetty-server-7.3.1.v20110307.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/servlet-api-2.5-6.1.14.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/postgresql-9.0-801.jdbc4.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jsp-2.1-jetty-6.1.26.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/avro-1.5.4.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/libthrift-0.5.0-cdh.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jetty-servlet-7.3.1.v20110307.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/velocity-1.6.4.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jetty-management-6.1.26.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/javassist-3.8.0.GA.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/ini4j-0.5.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/sshj-0.5.0.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jsp-api-2.1-6.1.14.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jamon-api-2.3.0.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/stax-api-1.0.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/ant-1.6.5.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/aopalliance-1.0.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/hbase-0.90.4-cdh3u2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/slf4j-api-1.6.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/lucene-core-3.1.0.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/async-http-client-1.6.5.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/asm-3.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/hibernate-entitymanager-3.6.2.Final.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/httpclient-4.1.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-math-2.2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/spring-asm-3.0.4.RELEASE.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jamon-runtime-2.3.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jackson-jaxrs-1.5.5.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/spring-web-3.0.4.RELEASE.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/common-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/ehcache-core-2.4.3.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/dom4j-1.6.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jsp-2.1-6.1.14.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/mail-1.4.3.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/camel-spring-2.7.2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/mx4j-3.0.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/guava-r09.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/c3p0-0.9.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/usertype.jodatime-1.8.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/hue-plugins-1.2.0-cdh3u2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/mx4j-tools-3.0.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jackson-core-asl-1.8.3.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jetty-continuation-7.3.1.v20110307.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/servlet-api-2.5.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jta-1.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/protocol-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jettison-1.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-logging-1.1.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/usertype.spi-1.8.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-management-1.0.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/persist-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jaxb-api-2.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/firehose-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/event-query-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jersey-json-1.4.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/alfredo-0.1.4.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/camel-test-2.7.2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/hibernate-jpa-2.0-api-1.0.0.Final.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/aspectjweaver-1.6.8.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/xstream-1.3.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jets3t-0.6.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jsp-api-2.1-glassfish-2.1.v20091210.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/event-common-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/log4j-1.2.16.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/activation-1.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/zookeeper-3.3.3-cdh3u2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/poi-3.7.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jersey-core-1.4.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/spring-beans-3.0.4.RELEASE.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/spring-core-3.0.4.RELEASE.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/hsqldb-1.8.0.10.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jetty-http-7.3.1.v20110307.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-io-2.0.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/spring-security-web-3.0.4.RELEASE.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-httpclient-3.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/descriptors-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/camel-core-2.7.2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jamon-processor-2.3.3.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-beanutils-1.7.0.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/slf4j-log4j12-1.6.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/spring-security-core-3.0.4.RELEASE.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jruby-complete-1.6.0.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/hibernate-c3p0-3.6.2.Final.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jetty-security-7.3.1.v20110307.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/xmlenc-0.52.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jetty-util-6.1.26.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-cli-1.2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/netty-3.2.1.Final.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/avro-ipc-1.5.4.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/model-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/spring-tx-3.0.4.RELEASE.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/spring-security-config-3.0.4.RELEASE.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/alertpublisher-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jetty-io-7.3.1.v20110307.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/httpcore-4.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/protobuf-java-2.3.0.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-configuration-1.6.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-logging-api-1.0.4.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/spring-aop-3.0.4.RELEASE.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jetty-util-7.3.1.v20110307.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/event-publish-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jsr311-api-1.1.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/bcprov-jdk16-1.46.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/cglib-2.2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/spring-context-support-3.0.4.RELEASE.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jsp-2.1-glassfish-2.1.v20091210.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/antlr-2.7.6.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-el-1.0.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/hibernate-core-3.6.2.Final.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/paranamer-2.3.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jackson-mapper-asl-1.8.3.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/hibernate-commons-annotations-3.2.0.Final.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/bcpg-jdk16-1.46.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/camel-velocity-2.7.2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/xpp3_min-1.1.4c.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/eventcatcher-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-lang-2.6.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/server-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/joda-time-1.6.2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-collections-3.2.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/camel-mail-2.7.2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/scm-client-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-net-1.4.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/spring-expression-3.0.4.RELEASE.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jdbcdslog-1.0.6.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-codec-1.4.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jaxb-impl-2.1.12.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/core-3.1.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/servlet-api-2.5-20081211.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/spring-context-3.0.4.RELEASE.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jackson-xc-1.5.5.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/headlamp-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/oro-2.0.8.jar",
    "StartTime" : 1324504780089,
    "BootClassPath" : "/usr/lib/jvm/java-6-sun-1.6.0.26/jre/lib/resources.jar:/usr/lib/jvm/java-6-sun-1.6.0.26/jre/lib/rt.jar:/usr/lib/jvm/java-6-sun-1.6.0.26/jre/lib/sunrsasign.jar:/usr/lib/jvm/java-6-sun-1.6.0.26/jre/lib/jsse.jar:/usr/lib/jvm/java-6-sun-1.6.0.26/jre/lib/jce.jar:/usr/lib/jvm/java-6-sun-1.6.0.26/jre/lib/charsets.jar:/usr/lib/jvm/java-6-sun-1.6.0.26/jre/lib/modules/jdk.boot.jar:/usr/lib/jvm/java-6-sun-1.6.0.26/jre/classes",
    "LibraryPath" : "/usr/lib/jvm/java-6-sun-1.6.0.26/jre/lib/amd64/server:/usr/lib/jvm/java-6-sun-1.6.0.26/jre/lib/amd64:/usr/lib/jvm/java-6-sun-1.6.0.26/jre/../lib/amd64:/usr/java/packages/lib/amd64:/usr/lib64:/lib64:/lib:/usr/lib",
    "VmName" : "Java HotSpot(TM) 64-Bit Server VM",
    "VmVendor" : "Sun Microsystems Inc.",
    "VmVersion" : "20.1-b02",
    "BootClassPathSupported" : true,
    "InputArguments" : [ "-Xmx106420423", "-Xdebug", "-Xrunjdwp:server=y,transport=dt_socket,address=4000,suspend=n", "-XX:+UseConcMarkSweepGC", "-XX:+UseParNewGC", "-Dmgmt.log.file=mgmt-cmf-mgmt1-SERVICEMONITOR-localhost.log.out", "-Djava.awt.headless=true" ],
    "ManagementSpecVersion" : "1.2",
    "SpecName" : "Java Virtual Machine Specification",
    "SpecVendor" : "Sun Microsystems Inc.",
    "SpecVersion" : "1.0",
    "SystemProperties" : [ {
      "key" : "java.ext.dirs",
      "value" : "/usr/lib/jvm/java-6-sun-1.6.0.26/jre/lib/ext:/usr/java/packages/lib/ext"
    }, {
      "key" : "java.vm.specification.vendor",
      "value" : "Sun Microsystems Inc."
    }, {
      "key" : "user.timezone",
      "value" : "America/Los_Angeles"
    }, {
      "key" : "java.vm.vendor",
      "value" : "Sun Microsystems Inc."
    }, {
      "key" : "user.name",
      "value" : "cloudera-scm"
    }, {
      "key" : "java.vm.specification.name",
      "value" : "Java Virtual Machine Specification"
    }, {
      "key" : "user.dir",
      "value" : "/tmp/agent/process/592-cloudera-mgmt-SERVICEMONITOR"
    }, {
      "key" : "user.country",
      "value" : "US"
    }, {
      "key" : "user.language",
      "value" : "en"
    }, {
      "key" : "java.specification.version",
      "value" : "1.6"
    }, {
      "key" : "sun.cpu.endian",
      "value" : "little"
    }, {
      "key" : "java.home",
      "value" : "/usr/lib/jvm/java-6-sun-1.6.0.26/jre"
    }, {
      "key" : "sun.jnu.encoding",
      "value" : "UTF-8"
    }, {
      "key" : "file.separator",
      "value" : "/"
    }, {
      "key" : "java.vendor.url",
      "value" : "http://java.sun.com/"
    }, {
      "key" : "java.awt.graphicsenv",
      "value" : "sun.awt.X11GraphicsEnvironment"
    }, {
      "key" : "os.arch",
      "value" : "amd64"
    }, {
      "key" : "java.io.tmpdir",
      "value" : "/tmp"
    }, {
      "key" : "java.runtime.name",
      "value" : "Java(TM) SE Runtime Environment"
    }, {
      "key" : "java.awt.printerjob",
      "value" : "sun.print.PSPrinterJob"
    }, {
      "key" : "file.encoding",
      "value" : "UTF-8"
    }, {
      "key" : "java.version",
      "value" : "1.6.0_26"
    }, {
      "key" : "java.vendor.url.bug",
      "value" : "http://java.sun.com/cgi-bin/bugreport.cgi"
    }, {
      "key" : "java.vm.specification.version",
      "value" : "1.0"
    }, {
      "key" : "file.encoding.pkg",
      "value" : "sun.io"
    }, {
      "key" : "sun.java.command",
      "value" : "com.cloudera.cmon.firehose.Main --pipeline-type SERVICE_MONITORING"
    }, {
      "key" : "sun.java.launcher",
      "value" : "SUN_STANDARD"
    }, {
      "key" : "path.separator",
      "value" : ":"
    }, {
      "key" : "java.runtime.version",
      "value" : "1.6.0_26-b03"
    }, {
      "key" : "java.class.path",
      "value" : "/tmp/agent/process/592-cloudera-mgmt-SERVICEMONITOR:/home/cleroy/.m2/repository/mysql/mysql-connector-java/5.1.15/mysql-connector-java-5.1.15.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-fileupload-1.2.2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jopt-simple-3.2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jersey-server-1.4.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-digester-1.8.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-beanutils-core-1.8.0.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/spring-webmvc-3.0.4.RELEASE.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jsp-api-2.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/aspectjrt-1.6.8.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/ecj-3.5.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/configgen-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/snappy-java-1.0.3.2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/hadoop-core-0.20.2-cdh3u2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jetty-6.1.26.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jetty-server-7.3.1.v20110307.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/servlet-api-2.5-6.1.14.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/postgresql-9.0-801.jdbc4.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jsp-2.1-jetty-6.1.26.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/avro-1.5.4.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/libthrift-0.5.0-cdh.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jetty-servlet-7.3.1.v20110307.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/velocity-1.6.4.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jetty-management-6.1.26.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/javassist-3.8.0.GA.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/ini4j-0.5.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/sshj-0.5.0.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jsp-api-2.1-6.1.14.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jamon-api-2.3.0.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/stax-api-1.0.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/ant-1.6.5.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/aopalliance-1.0.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/hbase-0.90.4-cdh3u2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/slf4j-api-1.6.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/lucene-core-3.1.0.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/async-http-client-1.6.5.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/asm-3.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/hibernate-entitymanager-3.6.2.Final.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/httpclient-4.1.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-math-2.2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/spring-asm-3.0.4.RELEASE.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jamon-runtime-2.3.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jackson-jaxrs-1.5.5.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/spring-web-3.0.4.RELEASE.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/common-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/ehcache-core-2.4.3.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/dom4j-1.6.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jsp-2.1-6.1.14.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/mail-1.4.3.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/camel-spring-2.7.2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/mx4j-3.0.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/guava-r09.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/c3p0-0.9.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/usertype.jodatime-1.8.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/hue-plugins-1.2.0-cdh3u2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/mx4j-tools-3.0.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jackson-core-asl-1.8.3.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jetty-continuation-7.3.1.v20110307.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/servlet-api-2.5.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jta-1.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/protocol-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jettison-1.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-logging-1.1.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/usertype.spi-1.8.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-management-1.0.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/persist-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jaxb-api-2.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/firehose-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/event-query-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jersey-json-1.4.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/alfredo-0.1.4.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/camel-test-2.7.2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/hibernate-jpa-2.0-api-1.0.0.Final.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/aspectjweaver-1.6.8.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/xstream-1.3.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jets3t-0.6.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jsp-api-2.1-glassfish-2.1.v20091210.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/event-common-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/log4j-1.2.16.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/activation-1.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/zookeeper-3.3.3-cdh3u2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/poi-3.7.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jersey-core-1.4.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/spring-beans-3.0.4.RELEASE.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/spring-core-3.0.4.RELEASE.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/hsqldb-1.8.0.10.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jetty-http-7.3.1.v20110307.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-io-2.0.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/spring-security-web-3.0.4.RELEASE.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-httpclient-3.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/descriptors-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/camel-core-2.7.2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jamon-processor-2.3.3.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-beanutils-1.7.0.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/slf4j-log4j12-1.6.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/spring-security-core-3.0.4.RELEASE.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jruby-complete-1.6.0.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/hibernate-c3p0-3.6.2.Final.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jetty-security-7.3.1.v20110307.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/xmlenc-0.52.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jetty-util-6.1.26.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-cli-1.2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/netty-3.2.1.Final.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/avro-ipc-1.5.4.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/model-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/spring-tx-3.0.4.RELEASE.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/spring-security-config-3.0.4.RELEASE.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/alertpublisher-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jetty-io-7.3.1.v20110307.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/httpcore-4.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/protobuf-java-2.3.0.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-configuration-1.6.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-logging-api-1.0.4.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/spring-aop-3.0.4.RELEASE.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jetty-util-7.3.1.v20110307.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/event-publish-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jsr311-api-1.1.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/bcprov-jdk16-1.46.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/cglib-2.2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/spring-context-support-3.0.4.RELEASE.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jsp-2.1-glassfish-2.1.v20091210.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/antlr-2.7.6.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-el-1.0.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/hibernate-core-3.6.2.Final.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/paranamer-2.3.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jackson-mapper-asl-1.8.3.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/hibernate-commons-annotations-3.2.0.Final.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/bcpg-jdk16-1.46.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/camel-velocity-2.7.2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/xpp3_min-1.1.4c.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/eventcatcher-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-lang-2.6.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/server-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/joda-time-1.6.2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-collections-3.2.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/camel-mail-2.7.2.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/scm-client-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-net-1.4.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/spring-expression-3.0.4.RELEASE.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jdbcdslog-1.0.6.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/commons-codec-1.4.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jaxb-impl-2.1.12.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/core-3.1.1.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/servlet-api-2.5-20081211.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/spring-context-3.0.4.RELEASE.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/jackson-xc-1.5.5.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/headlamp-3.8.0-SNAPSHOT.jar:/home/cleroy/git/starship/cmf/distribution/target/enterprise-distribution-3.8.0-SNAPSHOT-daemons/lib/oro-2.0.8.jar"
    }, {
      "key" : "os.name",
      "value" : "Linux"
    }, {
      "key" : "mgmt.log.file",
      "value" : "mgmt-cmf-mgmt1-SERVICEMONITOR-localhost.log.out"
    }, {
      "key" : "line.separator",
      "value" : "\n"
    }, {
      "key" : "os.version",
      "value" : "2.6.38-11-generic"
    }, {
      "key" : "net.sf.ehcache.skipUpdateCheck",
      "value" : "true"
    }, {
      "key" : "sun.arch.data.model",
      "value" : "64"
    }, {
      "key" : "java.class.version",
      "value" : "50.0"
    }, {
      "key" : "sun.io.unicode.encoding",
      "value" : "UnicodeLittle"
    }, {
      "key" : "java.vendor",
      "value" : "Sun Microsystems Inc."
    }, {
      "key" : "sun.boot.class.path",
      "value" : "/usr/lib/jvm/java-6-sun-1.6.0.26/jre/lib/resources.jar:/usr/lib/jvm/java-6-sun-1.6.0.26/jre/lib/rt.jar:/usr/lib/jvm/java-6-sun-1.6.0.26/jre/lib/sunrsasign.jar:/usr/lib/jvm/java-6-sun-1.6.0.26/jre/lib/jsse.jar:/usr/lib/jvm/java-6-sun-1.6.0.26/jre/lib/jce.jar:/usr/lib/jvm/java-6-sun-1.6.0.26/jre/lib/charsets.jar:/usr/lib/jvm/java-6-sun-1.6.0.26/jre/lib/modules/jdk.boot.jar:/usr/lib/jvm/java-6-sun-1.6.0.26/jre/classes"
    }, {
      "key" : "java.awt.headless",
      "value" : "true"
    }, {
      "key" : "java.vm.info",
      "value" : "mixed mode"
    }, {
      "key" : "javamelody.storage-directory",
      "value" : "javamelody_firehose"
    }, {
      "key" : "java.specification.name",
      "value" : "Java Platform API Specification"
    }, {
      "key" : "java.vm.name",
      "value" : "Java HotSpot(TM) 64-Bit Server VM"
    }, {
      "key" : "java.vm.version",
      "value" : "20.1-b02"
    }, {
      "key" : "sun.boot.library.path",
      "value" : "/usr/lib/jvm/java-6-sun-1.6.0.26/jre/lib/amd64"
    }, {
      "key" : "java.endorsed.dirs",
      "value" : "/usr/lib/jvm/java-6-sun-1.6.0.26/jre/lib/endorsed"
    }, {
      "key" : "sun.os.patch.level",
      "value" : "unknown"
    }, {
      "key" : "sun.cpu.isalist",
      "value" : ""
    }, {
      "key" : "user.home",
      "value" : "/home/cloudera-scm"
    }, {
      "key" : "java.library.path",
      "value" : "/usr/lib/jvm/java-6-sun-1.6.0.26/jre/lib/amd64/server:/usr/lib/jvm/java-6-sun-1.6.0.26/jre/lib/amd64:/usr/lib/jvm/java-6-sun-1.6.0.26/jre/../lib/amd64:/usr/java/packages/lib/amd64:/usr/lib64:/lib64:/lib:/usr/lib"
    }, {
      "key" : "java.specification.vendor",
      "value" : "Sun Microsystems Inc."
    }, {
      "key" : "sun.management.compiler",
      "value" : "HotSpot 64-Bit Tiered Compilers"
    } ],
    "Uptime" : 2366925
  }, {
    "name" : "java.lang:type=ClassLoading",
    "modelerType" : "sun.management.ClassLoadingImpl",
    "LoadedClassCount" : 6396,
    "UnloadedClassCount" : 0,
    "TotalLoadedClassCount" : 6396,
    "Verbose" : false
  }, {
    "name" : "java.lang:type=GarbageCollector,name=ConcurrentMarkSweep",
    "modelerType" : "sun.management.GarbageCollectorImpl",
    "LastGcInfo" : null,
    "CollectionCount" : 0,
    "CollectionTime" : 0,
    "Name" : "ConcurrentMarkSweep",
    "Valid" : true,
    "MemoryPoolNames" : [ "Par Eden Space", "Par Survivor Space", "CMS Old Gen", "CMS Perm Gen" ]
  }, {
    "name" : "java.lang:type=Threading",
    "modelerType" : "sun.management.ThreadImpl",
    "ThreadAllocatedMemoryEnabled" : true,
    "ThreadAllocatedMemorySupported" : true,
    "DaemonThreadCount" : 62,
    "PeakThreadCount" : 73,
    "CurrentThreadCpuTimeSupported" : true,
    "ObjectMonitorUsageSupported" : true,
    "SynchronizerUsageSupported" : true,
    "ThreadContentionMonitoringEnabled" : false,
    "ThreadContentionMonitoringSupported" : true,
    "ThreadCpuTimeEnabled" : true,
    "AllThreadIds" : [ 314, 312, 311, 138, 130, 123, 122, 115, 114, 107, 106, 99, 98, 91, 90, 84, 83, 82, 75, 74, 73, 66, 65, 64, 63, 62, 61, 60, 59, 58, 56, 55, 48, 47, 46, 45, 44, 42, 41, 40, 39, 38, 37, 36, 25, 35, 34, 33, 32, 30, 29, 28, 27, 26, 24, 23, 22, 21, 20, 19, 18, 17, 16, 15, 14, 13, 12, 5, 3, 2 ],
    "CurrentThreadCpuTime" : 10000000,
    "CurrentThreadUserTime" : 10000000,
    "ThreadCount" : 70,
    "TotalStartedThreadCount" : 265,
    "ThreadCpuTimeSupported" : true
  }, {
    "name" : "java.util.logging:type=Logging",
    "modelerType" : "java.util.logging.Logging",
    "LoggerNames" : [ "org.jboss.netty.channel.socket.nio.DefaultNioSocketChannelConfig", "org.jboss.netty.channel.socket.nio.NioServerSocketChannel", "org.jboss.netty.channel.DefaultChannelFuture", "org.jboss.netty.channel.socket.nio.NioServerSocketPipelineSink", "org.jboss.netty.channel.socket.nio.NioClientSocketPipelineSink", "org.jboss.netty.util.ThreadRenamingRunnable", "javax.management.misc", "javax.management", "org.jboss.netty.channel.socket.nio.NioWorker", "org.jboss.netty.channel.StaticChannelPipeline", "org.jboss.netty.channel.CompleteChannelFuture", "org.jboss.netty.channel.SimpleChannelHandler", "sun.awt.AppContext", "org.jboss.netty.channel.socket.nio.SelectorUtil", "org.jboss.netty.channel.DefaultChannelPipeline", "global", "org.jboss.netty.channel.socket.nio.NioProviderMetadata", "org.jboss.netty.channel.SimpleChannelUpstreamHandler", "javax.management.mbeanserver", "org.jboss.netty.channel.socket.nio.NioClientSocketChannel", "" ]
  }, {
    "name" : "com.cloudera.cmf.event.publish:type=EventStorePublisherWithRetryMXBean",
    "modelerType" : "com.cloudera.cmf.event.publish.EventStorePublisherWithRetry",
    "SendQueueSize" : 0,
    "SendQueueMaxSize" : 20000,
    "DropCount" : 0,
    "ExceptionCount" : 0,
    "PushbackCount" : 0,
    "SentCount" : 5
  }, {
    "name" : "java.lang:type=Compilation",
    "modelerType" : "sun.management.CompilationImpl",
    "Name" : "HotSpot 64-Bit Tiered Compilers",
    "CompilationTimeMonitoringSupported" : true,
    "TotalCompilationTime" : 17006
  }, {
    "name" : "com.mchange.v2.c3p0:type=C3P0Registry",
    "modelerType" : "com.mchange.v2.c3p0.management.C3P0RegistryManager",
    "NumPooledDataSources" : 1,
    "NumPoolsAllDataSources" : 1,
    "AllIdentityTokens" : [ "z8kflt8krd9brorhx66s|3f9f76a5", "z8kflt8krd9brorhx66s|1535d18b", "z8kflt8krd9brorhx66s|72bdec44" ],
    "AllIdentityTokenized" : "[com.mchange.v2.c3p0.PoolBackedDataSource@2e8d0265 [ connectionPoolDataSource -> com.mchange.v2.c3p0.WrapperConnectionPoolDataSource@553dbbad [ acquireIncrement -> 3, acquireRetryAttempts -> 30, acquireRetryDelay -> 1000, autoCommitOnClose -> false, automaticTestTable -> null, breakAfterAcquireFailure -> false, checkoutTimeout -> 0, connectionCustomizerClassName -> null, connectionTesterClassName -> com.mchange.v2.c3p0.impl.DefaultConnectionTester, debugUnreturnedConnectionStackTraces -> false, factoryClassLocation -> null, forceIgnoreUnresolvedTransactions -> false, identityToken -> z8kflt8krd9brorhx66s|3f9f76a5, idleConnectionTestPeriod -> 100, initialPoolSize -> 10, maxAdministrativeTaskTime -> 0, maxConnectionAge -> 0, maxIdleTime -> 100, maxIdleTimeExcessConnections -> 0, maxPoolSize -> 100, maxStatements -> 0, maxStatementsPerConnection -> 0, minPoolSize -> 10, nestedDataSource -> com.mchange.v2.c3p0.DriverManagerDataSource@51eca323 [ description -> null, driverClass -> null, factoryClassLocation -> null, identityToken -> z8kflt8krd9brorhx66s|1535d18b, jdbcUrl -> jdbc:mysql://localhost/cmon_db, properties -> {user=******, password=******, autocommit=true, release_mode=auto} ], preferredTestQuery -> null, propertyCycle -> 0, testConnectionOnCheckin -> false, testConnectionOnCheckout -> false, unreturnedConnectionTimeout -> 0, usesTraditionalReflectiveProxies -> false; userOverrides: {} ], dataSourceName -> null, factoryClassLocation -> null, identityToken -> z8kflt8krd9brorhx66s|72bdec44, numHelperThreads -> 3 ], com.mchange.v2.c3p0.DriverManagerDataSource@51eca323 [ description -> null, driverClass -> null, factoryClassLocation -> null, identityToken -> z8kflt8krd9brorhx66s|1535d18b, jdbcUrl -> jdbc:mysql://localhost/cmon_db, properties -> {user=******, password=******, autocommit=true, release_mode=auto} ], com.mchange.v2.c3p0.WrapperConnectionPoolDataSource@553dbbad [ acquireIncrement -> 3, acquireRetryAttempts -> 30, acquireRetryDelay -> 1000, autoCommitOnClose -> false, automaticTestTable -> null, breakAfterAcquireFailure -> false, checkoutTimeout -> 0, connectionCustomizerClassName -> null, connectionTesterClassName -> com.mchange.v2.c3p0.impl.DefaultConnectionTester, debugUnreturnedConnectionStackTraces -> false, factoryClassLocation -> null, forceIgnoreUnresolvedTransactions -> false, identityToken -> z8kflt8krd9brorhx66s|3f9f76a5, idleConnectionTestPeriod -> 100, initialPoolSize -> 10, maxAdministrativeTaskTime -> 0, maxConnectionAge -> 0, maxIdleTime -> 100, maxIdleTimeExcessConnections -> 0, maxPoolSize -> 100, maxStatements -> 0, maxStatementsPerConnection -> 0, minPoolSize -> 10, nestedDataSource -> com.mchange.v2.c3p0.DriverManagerDataSource@51eca323 [ description -> null, driverClass -> null, factoryClassLocation -> null, identityToken -> z8kflt8krd9brorhx66s|1535d18b, jdbcUrl -> jdbc:mysql://localhost/cmon_db, properties -> {user=******, password=******, autocommit=true, release_mode=auto} ], preferredTestQuery -> null, propertyCycle -> 0, testConnectionOnCheckin -> false, testConnectionOnCheckout -> false, unreturnedConnectionTimeout -> 0, usesTraditionalReflectiveProxies -> false; userOverrides: {} ]]",
    "AllPooledDataSources" : "[com.mchange.v2.c3p0.PoolBackedDataSource@2e8d0265 [ connectionPoolDataSource -> com.mchange.v2.c3p0.WrapperConnectionPoolDataSource@553dbbad [ acquireIncrement -> 3, acquireRetryAttempts -> 30, acquireRetryDelay -> 1000, autoCommitOnClose -> false, automaticTestTable -> null, breakAfterAcquireFailure -> false, checkoutTimeout -> 0, connectionCustomizerClassName -> null, connectionTesterClassName -> com.mchange.v2.c3p0.impl.DefaultConnectionTester, debugUnreturnedConnectionStackTraces -> false, factoryClassLocation -> null, forceIgnoreUnresolvedTransactions -> false, identityToken -> z8kflt8krd9brorhx66s|3f9f76a5, idleConnectionTestPeriod -> 100, initialPoolSize -> 10, maxAdministrativeTaskTime -> 0, maxConnectionAge -> 0, maxIdleTime -> 100, maxIdleTimeExcessConnections -> 0, maxPoolSize -> 100, maxStatements -> 0, maxStatementsPerConnection -> 0, minPoolSize -> 10, nestedDataSource -> com.mchange.v2.c3p0.DriverManagerDataSource@51eca323 [ description -> null, driverClass -> null, factoryClassLocation -> null, identityToken -> z8kflt8krd9brorhx66s|1535d18b, jdbcUrl -> jdbc:mysql://localhost/cmon_db, properties -> {user=******, password=******, autocommit=true, release_mode=auto} ], preferredTestQuery -> null, propertyCycle -> 0, testConnectionOnCheckin -> false, testConnectionOnCheckout -> false, unreturnedConnectionTimeout -> 0, usesTraditionalReflectiveProxies -> false; userOverrides: {} ], dataSourceName -> null, factoryClassLocation -> null, identityToken -> z8kflt8krd9brorhx66s|72bdec44, numHelperThreads -> 3 ]]",
    "AllIdentityTokenCount" : 3,
    "AllIdentityTokenizedCount" : 3,
    "AllPooledDataSourcesCount" : 1,
    "AllIdentityTokenizedStringified" : [ "com.mchange.v2.c3p0.PoolBackedDataSource@2e8d0265 [ connectionPoolDataSource -> com.mchange.v2.c3p0.WrapperConnectionPoolDataSource@553dbbad [ acquireIncrement -> 3, acquireRetryAttempts -> 30, acquireRetryDelay -> 1000, autoCommitOnClose -> false, automaticTestTable -> null, breakAfterAcquireFailure -> false, checkoutTimeout -> 0, connectionCustomizerClassName -> null, connectionTesterClassName -> com.mchange.v2.c3p0.impl.DefaultConnectionTester, debugUnreturnedConnectionStackTraces -> false, factoryClassLocation -> null, forceIgnoreUnresolvedTransactions -> false, identityToken -> z8kflt8krd9brorhx66s|3f9f76a5, idleConnectionTestPeriod -> 100, initialPoolSize -> 10, maxAdministrativeTaskTime -> 0, maxConnectionAge -> 0, maxIdleTime -> 100, maxIdleTimeExcessConnections -> 0, maxPoolSize -> 100, maxStatements -> 0, maxStatementsPerConnection -> 0, minPoolSize -> 10, nestedDataSource -> com.mchange.v2.c3p0.DriverManagerDataSource@51eca323 [ description -> null, driverClass -> null, factoryClassLocation -> null, identityToken -> z8kflt8krd9brorhx66s|1535d18b, jdbcUrl -> jdbc:mysql://localhost/cmon_db, properties -> {user=******, password=******, autocommit=true, release_mode=auto} ], preferredTestQuery -> null, propertyCycle -> 0, testConnectionOnCheckin -> false, testConnectionOnCheckout -> false, unreturnedConnectionTimeout -> 0, usesTraditionalReflectiveProxies -> false; userOverrides: {} ], dataSourceName -> null, factoryClassLocation -> null, identityToken -> z8kflt8krd9brorhx66s|72bdec44, numHelperThreads -> 3 ]", "com.mchange.v2.c3p0.DriverManagerDataSource@51eca323 [ description -> null, driverClass -> null, factoryClassLocation -> null, identityToken -> z8kflt8krd9brorhx66s|1535d18b, jdbcUrl -> jdbc:mysql://localhost/cmon_db, properties -> {user=******, password=******, autocommit=true, release_mode=auto} ]", "com.mchange.v2.c3p0.WrapperConnectionPoolDataSource@553dbbad [ acquireIncrement -> 3, acquireRetryAttempts -> 30, acquireRetryDelay -> 1000, autoCommitOnClose -> false, automaticTestTable -> null, breakAfterAcquireFailure -> false, checkoutTimeout -> 0, connectionCustomizerClassName -> null, connectionTesterClassName -> com.mchange.v2.c3p0.impl.DefaultConnectionTester, debugUnreturnedConnectionStackTraces -> false, factoryClassLocation -> null, forceIgnoreUnresolvedTransactions -> false, identityToken -> z8kflt8krd9brorhx66s|3f9f76a5, idleConnectionTestPeriod -> 100, initialPoolSize -> 10, maxAdministrativeTaskTime -> 0, maxConnectionAge -> 0, maxIdleTime -> 100, maxIdleTimeExcessConnections -> 0, maxPoolSize -> 100, maxStatements -> 0, maxStatementsPerConnection -> 0, minPoolSize -> 10, nestedDataSource -> com.mchange.v2.c3p0.DriverManagerDataSource@51eca323 [ description -> null, driverClass -> null, factoryClassLocation -> null, identityToken -> z8kflt8krd9brorhx66s|1535d18b, jdbcUrl -> jdbc:mysql://localhost/cmon_db, properties -> {user=******, password=******, autocommit=true, release_mode=auto} ], preferredTestQuery -> null, propertyCycle -> 0, testConnectionOnCheckin -> false, testConnectionOnCheckout -> false, unreturnedConnectionTimeout -> 0, usesTraditionalReflectiveProxies -> false; userOverrides: {} ]" ],
    "AllPooledDataSourcesStringified" : [ "com.mchange.v2.c3p0.PoolBackedDataSource@2e8d0265 [ connectionPoolDataSource -> com.mchange.v2.c3p0.WrapperConnectionPoolDataSource@553dbbad [ acquireIncrement -> 3, acquireRetryAttempts -> 30, acquireRetryDelay -> 1000, autoCommitOnClose -> false, automaticTestTable -> null, breakAfterAcquireFailure -> false, checkoutTimeout -> 0, connectionCustomizerClassName -> null, connectionTesterClassName -> com.mchange.v2.c3p0.impl.DefaultConnectionTester, debugUnreturnedConnectionStackTraces -> false, factoryClassLocation -> null, forceIgnoreUnresolvedTransactions -> false, identityToken -> z8kflt8krd9brorhx66s|3f9f76a5, idleConnectionTestPeriod -> 100, initialPoolSize -> 10, maxAdministrativeTaskTime -> 0, maxConnectionAge -> 0, maxIdleTime -> 100, maxIdleTimeExcessConnections -> 0, maxPoolSize -> 100, maxStatements -> 0, maxStatementsPerConnection -> 0, minPoolSize -> 10, nestedDataSource -> com.mchange.v2.c3p0.DriverManagerDataSource@51eca323 [ description -> null, driverClass -> null, factoryClassLocation -> null, identityToken -> z8kflt8krd9brorhx66s|1535d18b, jdbcUrl -> jdbc:mysql://localhost/cmon_db, properties -> {user=******, password=******, autocommit=true, release_mode=auto} ], preferredTestQuery -> null, propertyCycle -> 0, testConnectionOnCheckin -> false, testConnectionOnCheckout -> false, unreturnedConnectionTimeout -> 0, usesTraditionalReflectiveProxies -> false; userOverrides: {} ], dataSourceName -> null, factoryClassLocation -> null, identityToken -> z8kflt8krd9brorhx66s|72bdec44, numHelperThreads -> 3 ]" ],
    "C3p0Version" : "0.9.1"
  }, {
    "name" : "java.lang:type=GarbageCollector,name=ParNew",
    "modelerType" : "sun.management.GarbageCollectorImpl",
    "LastGcInfo" : {
      "GcThreadCount" : 11,
      "duration" : 3,
      "endTime" : 2352996,
      "id" : 45,
      "memoryUsageAfterGc" : [ {
        "key" : "CMS Perm Gen",
        "value" : {
          "committed" : 48300032,
          "init" : 21757952,
          "max" : 85983232,
          "used" : 48104904
        }
      }, {
        "key" : "Code Cache",
        "value" : {
          "committed" : 4521984,
          "init" : 2555904,
          "max" : 50331648,
          "used" : 4424576
        }
      }, {
        "key" : "CMS Old Gen",
        "value" : {
          "committed" : 65404928,
          "init" : 65404928,
          "max" : 85196800,
          "used" : 16997432
        }
      }, {
        "key" : "Par Eden Space",
        "value" : {
          "committed" : 17432576,
          "init" : 17432576,
          "max" : 17432576,
          "used" : 0
        }
      }, {
        "key" : "Par Survivor Space",
        "value" : {
          "committed" : 2162688,
          "init" : 2162688,
          "max" : 2162688,
          "used" : 1937760
        }
      } ],
      "memoryUsageBeforeGc" : [ {
        "key" : "CMS Perm Gen",
        "value" : {
          "committed" : 48300032,
          "init" : 21757952,
          "max" : 85983232,
          "used" : 48104904
        }
      }, {
        "key" : "Code Cache",
        "value" : {
          "committed" : 4521984,
          "init" : 2555904,
          "max" : 50331648,
          "used" : 4424576
        }
      }, {
        "key" : "CMS Old Gen",
        "value" : {
          "committed" : 65404928,
          "init" : 65404928,
          "max" : 85196800,
          "used" : 16997432
        }
      }, {
        "key" : "Par Eden Space",
        "value" : {
          "committed" : 17432576,
          "init" : 17432576,
          "max" : 17432576,
          "used" : 17432576
        }
      }, {
        "key" : "Par Survivor Space",
        "value" : {
          "committed" : 2162688,
          "init" : 2162688,
          "max" : 2162688,
          "used" : 1819536
        }
      } ],
      "startTime" : 2352993
    },
    "CollectionCount" : 45,
    "CollectionTime" : 386,
    "Name" : "ParNew",
    "Valid" : true,
    "MemoryPoolNames" : [ "Par Eden Space", "Par Survivor Space" ]
  }, {
    "name" : "java.lang:type=MemoryPool,name=Par Eden Space",
    "modelerType" : "sun.management.MemoryPoolImpl",
    "Name" : "Par Eden Space",
    "Type" : "HEAP",
    "Valid" : true,
    "CollectionUsage" : {
      "committed" : 17432576,
      "init" : 17432576,
      "max" : 17432576,
      "used" : 0
    },
    "CollectionUsageThreshold" : 0,
    "CollectionUsageThresholdCount" : 0,
    "MemoryManagerNames" : [ "ConcurrentMarkSweep", "ParNew" ],
    "PeakUsage" : {
      "committed" : 17432576,
      "init" : 17432576,
      "max" : 17432576,
      "used" : 17432576
    },
    "Usage" : {
      "committed" : 17432576,
      "init" : 17432576,
      "max" : 17432576,
      "used" : 1620544
    },
    "CollectionUsageThresholdExceeded" : true,
    "CollectionUsageThresholdSupported" : true,
    "UsageThresholdSupported" : false
  }, {
    "name" : "com.sun.management:type=HotSpotDiagnostic",
    "modelerType" : "sun.management.HotSpotDiagnostic",
    "DiagnosticOptions" : [ {
      "name" : "HeapDumpBeforeFullGC",
      "origin" : "DEFAULT",
      "value" : "false",
      "writeable" : true
    }, {
      "name" : "HeapDumpAfterFullGC",
      "origin" : "DEFAULT",
      "value" : "false",
      "writeable" : true
    }, {
      "name" : "HeapDumpOnOutOfMemoryError",
      "origin" : "DEFAULT",
      "value" : "false",
      "writeable" : true
    }, {
      "name" : "HeapDumpPath",
      "origin" : "DEFAULT",
      "value" : "",
      "writeable" : true
    }, {
      "name" : "CMSAbortablePrecleanWaitMillis",
      "origin" : "DEFAULT",
      "value" : "100",
      "writeable" : true
    }, {
      "name" : "CMSWaitDuration",
      "origin" : "DEFAULT",
      "value" : "2000",
      "writeable" : true
    }, {
      "name" : "PrintGC",
      "origin" : "DEFAULT",
      "value" : "false",
      "writeable" : true
    }, {
      "name" : "PrintGCDetails",
      "origin" : "DEFAULT",
      "value" : "false",
      "writeable" : true
    }, {
      "name" : "PrintGCDateStamps",
      "origin" : "DEFAULT",
      "value" : "false",
      "writeable" : true
    }, {
      "name" : "PrintGCTimeStamps",
      "origin" : "DEFAULT",
      "value" : "false",
      "writeable" : true
    }, {
      "name" : "PrintClassHistogramBeforeFullGC",
      "origin" : "DEFAULT",
      "value" : "false",
      "writeable" : true
    }, {
      "name" : "PrintClassHistogramAfterFullGC",
      "origin" : "DEFAULT",
      "value" : "false",
      "writeable" : true
    }, {
      "name" : "PrintClassHistogram",
      "origin" : "DEFAULT",
      "value" : "false",
      "writeable" : true
    }, {
      "name" : "PrintConcurrentLocks",
      "origin" : "DEFAULT",
      "value" : "false",
      "writeable" : true
    } ]
  }, {
    "name" : "com.mchange.v2.c3p0:type=PooledDataSource[z8kflt8krd9brorhx66s|72bdec44]",
    "modelerType" : "com.mchange.v2.c3p0.management.DynamicPooledDataSourceManagerMBean",
    "overrideDefaultPassword" : null,
    "statementCacheNumConnectionsWithCachedStatementsAllUsers" : 0,
    "dataSourceName" : "z8kflt8krd9brorhx66s|72bdec44",
    "threadPoolNumIdleThreads" : 3,
    "acquireRetryAttempts" : 30,
    "autoCommitOnClose" : false,
    "identityToken" : "z8kflt8krd9brorhx66s|72bdec44",
    "forceIgnoreUnresolvedTransactions" : false,
    "startTimeMillisDefaultUser" : 1324504784485,
    "breakAfterAcquireFailure" : false,
    "maxPoolSize" : 100,
    "idleConnectionTestPeriod" : 100,
    "user" : "cmon",
    "numIdleConnectionsDefaultUser" : 10,
    "numFailedCheckoutsDefaultUser" : 0,
    "numConnections" : 10,
    "numBusyConnections" : 0,
    "connectionCustomizerClassName" : null,
    "automaticTestTable" : null,
    "numFailedIdleTestsDefaultUser" : 2,
    "statementCacheNumStatementsDefaultUser" : 0,
    "effectivePropertyCycleDefaultUser" : 25.0,
    "maxIdleTime" : 100,
    "userOverrides" : "{}",
    "numConnectionsAllUsers" : 10,
    "numBusyConnectionsDefaultUser" : 0,
    "maxIdleTimeExcessConnections" : 0,
    "checkoutTimeout" : 0,
    "allUsers" : "[cmon]",
    "statementCacheNumConnectionsWithCachedStatementsDefaultUser" : 0,
    "usesTraditionalReflectiveProxies" : false,
    "maxAdministrativeTaskTime" : 0,
    "userOverridesAsString" : null,
    "driverClass" : null,
    "connectionTesterClassName" : "com.mchange.v2.c3p0.impl.DefaultConnectionTester",
    "maxStatements" : 0,
    "statementCacheNumCheckedOutDefaultUser" : 0,
    "statementCacheNumStatementsAllUsers" : 0,
    "numUnclosedOrphanedConnectionsDefaultUser" : 0,
    "debugUnreturnedConnectionStackTraces" : false,
    "testConnectionOnCheckout" : false,
    "numConnectionsDefaultUser" : 10,
    "acquireRetryDelay" : 1000,
    "overrideDefaultUser" : null,
    "threadPoolNumTasksPending" : 0,
    "threadPoolSize" : 3,
    "unreturnedConnectionTimeout" : 0,
    "acquireIncrement" : 3,
    "numThreadsAwaitingCheckoutDefaultUser" : 0,
    "threadPoolNumActiveThreads" : 0,
    "minPoolSize" : 10,
    "maxStatementsPerConnection" : 0,
    "statementCacheNumCheckedOutStatementsAllUsers" : 0,
    "testConnectionOnCheckin" : false,
    "numUserPools" : 1,
    "numFailedCheckinsDefaultUser" : 0,
    "numUnclosedOrphanedConnectionsAllUsers" : 0,
    "initialPoolSize" : 10,
    "upTimeMillisDefaultUser" : 2362540,
    "numUnclosedOrphanedConnections" : 0,
    "properties" : "{user=******, password=******, autocommit=true, release_mode=auto}",
    "loginTimeout" : 0,
    "numIdleConnectionsAllUsers" : 10,
    "maxConnectionAge" : 0,
    "factoryClassLocation" : null,
    "jdbcUrl" : "jdbc:mysql://localhost/cmon_db",
    "propertyCycle" : 0,
    "numBusyConnectionsAllUsers" : 0,
    "numIdleConnections" : 10,
    "numHelperThreads" : 3,
    "preferredTestQuery" : null,
    "description" : null
  }, {
    "name" : "java.lang:type=MemoryPool,name=CMS Perm Gen",
    "modelerType" : "sun.management.MemoryPoolImpl",
    "Name" : "CMS Perm Gen",
    "Type" : "NON_HEAP",
    "Valid" : true,
    "CollectionUsage" : {
      "committed" : 0,
      "init" : 21757952,
      "max" : 85983232,
      "used" : 0
    },
    "CollectionUsageThreshold" : 0,
    "CollectionUsageThresholdCount" : 0,
    "MemoryManagerNames" : [ "ConcurrentMarkSweep" ],
    "PeakUsage" : {
      "committed" : 48300032,
      "init" : 21757952,
      "max" : 85983232,
      "used" : 48104904
    },
    "Usage" : {
      "committed" : 48300032,
      "init" : 21757952,
      "max" : 85983232,
      "used" : 48104904
    },
    "UsageThreshold" : 0,
    "UsageThresholdCount" : 0,
    "CollectionUsageThresholdExceeded" : true,
    "CollectionUsageThresholdSupported" : true,
    "UsageThresholdExceeded" : true,
    "UsageThresholdSupported" : true
  }, {
    "name" : "java.lang:type=OperatingSystem",
    "modelerType" : "com.sun.management.UnixOperatingSystem",
    "MaxFileDescriptorCount" : 4096,
    "OpenFileDescriptorCount" : 193,
    "CommittedVirtualMemorySize" : 630296576,
    "FreePhysicalMemorySize" : 2182844416,
    "FreeSwapSpaceSize" : 8516530176,
    "ProcessCpuTime" : 45080000000,
    "TotalPhysicalMemorySize" : 8318742528,
    "TotalSwapSpaceSize" : 8516530176,
    "Name" : "Linux",
    "Version" : "2.6.38-11-generic",
    "AvailableProcessors" : 8,
    "Arch" : "amd64",
    "SystemLoadAverage" : 0.0
  }, {
    "name" : "java.lang:type=MemoryPool,name=Par Survivor Space",
    "modelerType" : "sun.management.MemoryPoolImpl",
    "Name" : "Par Survivor Space",
    "Type" : "HEAP",
    "Valid" : true,
    "CollectionUsage" : {
      "committed" : 2162688,
      "init" : 2162688,
      "max" : 2162688,
      "used" : 1937760
    },
    "CollectionUsageThreshold" : 0,
    "CollectionUsageThresholdCount" : 0,
    "MemoryManagerNames" : [ "ConcurrentMarkSweep", "ParNew" ],
    "PeakUsage" : {
      "committed" : 2162688,
      "init" : 2162688,
      "max" : 2162688,
      "used" : 2162688
    },
    "Usage" : {
      "committed" : 2162688,
      "init" : 2162688,
      "max" : 2162688,
      "used" : 1937760
    },
    "CollectionUsageThresholdExceeded" : true,
    "CollectionUsageThresholdSupported" : true,
    "UsageThresholdSupported" : false
  }, {
    "name" : "JMImplementation:type=MBeanServerDelegate",
    "modelerType" : "javax.management.MBeanServerDelegate",
    "MBeanServerId" : "bigrig_1324504784117",
    "SpecificationName" : "Java Management Extensions",
    "SpecificationVersion" : "1.4",
    "SpecificationVendor" : "Sun Microsystems",
    "ImplementationName" : "JMX",
    "ImplementationVersion" : "1.6.0_26-b03",
    "ImplementationVendor" : "Sun Microsystems"
  }, {
    "name" : "java.lang:type=MemoryManager,name=CodeCacheManager",
    "modelerType" : "sun.management.MemoryManagerImpl",
    "Name" : "CodeCacheManager",
    "Valid" : true,
    "MemoryPoolNames" : [ "Code Cache" ]
  }, {
    "name" : "com.cloudera.enterprise.debug:type=JvmPauseMonitor,name=numPausesDetected",
    "modelerType" : "com.yammer.metrics.reporting.JmxReporter$Counter",
    "Count" : 1
  }, {
    "name" : "com.cloudera.enterprise.debug:type=JvmPauseMonitor,name=pauseTimeMs",
    "modelerType" : "com.yammer.metrics.reporting.JmxReporter$Counter",
    "Count" : 30508
  } ]
}
"""

_METRICS_SAMPLE_FOR_PORTS_TESTS_JMX = SAMPLE = r"""
{
  "beans" : [ {
    "name" : "java.lang:type=Memory",
    "modelerType" : "sun.management.MemoryImpl",
    "Verbose" : false,
    "HeapMemoryUsage" : {
      "committed" : 85000192,
      "init" : 0,
      "max" : 173998080,
      "used" : 44075280
    },
    "NonHeapMemoryUsage" : {
      "committed" : 91553792,
      "init" : 24313856,
      "max" : 136314880,
      "used" : 59269168
    },
    "ObjectPendingFinalizationCount" : 0
  }, {
    "name" : "Hadoop:service=NameNode,name=RpcActivityForPort8022",
    "modelerType" : "RpcActivityForPort8022",
    "tag.port" : "8022",
    "tag.Context" : "rpc",
    "tag.Hostname" : "cleroy-8-2.ent.cloudera.com",
    "ReceivedBytes" : 399393598,
    "SentBytes" : 35307754,
    "RpcQueueTimeNumOps" : 2149407,
    "RpcQueueTimeAvgTime" : 0.17391304347826095,
    "RpcProcessingTimeNumOps" : 2149407,
    "RpcProcessingTimeAvgTime" : 0.06521739130434782,
    "RpcAuthenticationFailures" : 0,
    "RpcAuthenticationSuccesses" : 0,
    "RpcAuthorizationFailures" : 0,
    "RpcAuthorizationSuccesses" : 5203,
    "NumOpenConnections" : 5,
    "CallQueueLength" : 0
  }, {
    "name" : "Hadoop:service=NameNode,name=RpcDetailedActivityForPort8022",
    "modelerType" : "RpcDetailedActivityForPort8022",
    "tag.port" : "8022",
    "tag.Context" : "rpcdetailed",
    "tag.Hostname" : "cleroy-8-2.ent.cloudera.com",
    "SendHeartbeatNumOps" : 833806,
    "SendHeartbeatAvgTime" : 0.05000000000000001,
    "VersionRequestNumOps" : 4,
    "VersionRequestAvgTime" : 2.0,
    "RegisterDatanodeNumOps" : 4,
    "RegisterDatanodeAvgTime" : 4.5,
    "BlockReportNumOps" : 120,
    "BlockReportAvgTime" : 1.0,
    "GetServiceStatusNumOps" : 624000,
    "GetServiceStatusAvgTime" : 0.07692307692307693,
    "MonitorHealthNumOps" : 623999,
    "MonitorHealthAvgTime" : 0.07692307692307693,
    "TransitionToActiveNumOps" : 10,
    "TransitionToActiveAvgTime" : 2.0,
    "BlockReceivedAndDeletedNumOps" : 62280,
    "BlockReceivedAndDeletedAvgTime" : 0.16666666666666669,
    "RollEditLogNumOps" : 5183,
    "RollEditLogAvgTime" : 534.0,
    "TransitionToStandbyNumOps" : 1,
    "TransitionToStandbyAvgTime" : 678.0
  }, {
    "name" : "Hadoop:service=NameNode,name=RpcActivityForPort8020",
    "modelerType" : "RpcActivityForPort8020",
    "tag.port" : "8020",
    "tag.Context" : "rpc",
    "tag.Hostname" : "cleroy-8-2.ent.cloudera.com",
    "ReceivedBytes" : 30386109,
    "SentBytes" : 19643942,
    "RpcQueueTimeNumOps" : 176282,
    "RpcQueueTimeAvgTime" : 0.5,
    "RpcProcessingTimeNumOps" : 176282,
    "RpcProcessingTimeAvgTime" : 1.0,
    "RpcAuthenticationFailures" : 0,
    "RpcAuthenticationSuccesses" : 0,
    "RpcAuthorizationFailures" : 0,
    "RpcAuthorizationSuccesses" : 96665,
    "NumOpenConnections" : 1,
    "CallQueueLength" : 0
  }, {
    "name" : "Hadoop:service=NameNode,name=RpcDetailedActivityForPort8020",
    "modelerType" : "RpcDetailedActivityForPort8020",
    "tag.port" : "8020",
    "tag.Context" : "rpcdetailed",
    "tag.Hostname" : "cleroy-8-2.ent.cloudera.com",
    "GetListingNumOps" : 24753,
    "GetListingAvgTime" : 2.0,
    "GetFileInfoNumOps" : 17783,
    "GetFileInfoAvgTime" : 2.0,
    "SetSafeModeNumOps" : 1,
    "SetSafeModeAvgTime" : 0.0,
    "GetBlockLocationsNumOps" : 10435,
    "GetBlockLocationsAvgTime" : 0.0,
    "GetServerDefaultsNumOps" : 10386,
    "GetServerDefaultsAvgTime" : 0.0,
    "DeleteNumOps" : 10402,
    "DeleteAvgTime" : 25.0,
    "MkdirsNumOps" : 23,
    "MkdirsAvgTime" : 0.5,
    "CreateNumOps" : 10383,
    "CreateAvgTime" : 87.0,
    "AddBlockNumOps" : 10380,
    "AddBlockAvgTime" : 25.0,
    "FsyncNumOps" : 1,
    "FsyncAvgTime" : 40.0,
    "RenewLeaseNumOps" : 71332,
    "RenewLeaseAvgTime" : 0.0,
    "SetPermissionNumOps" : 4,
    "SetPermissionAvgTime" : 22.0,
    "CompleteNumOps" : 10382,
    "CompleteAvgTime" : 31.0,
    "RenameNumOps" : 14,
    "RenameAvgTime" : 38.5,
    "RecoverLeaseNumOps" : 3,
    "RecoverLeaseAvgTime" : 60.0
  }, {
   "name" : "JMImplementation:type=MBeanServerDelegate",
    "modelerType" : "javax.management.MBeanServerDelegate",
    "MBeanServerId" : "cleroy-8-2.ent.cloudera.com_1373587048827",
    "SpecificationName" : "Java Management Extensions",
    "SpecificationVersion" : "1.4",
    "SpecificationVendor" : "Sun Microsystems",
    "ImplementationName" : "JMX",
    "ImplementationVersion" : "1.6.0_31-b04",
    "ImplementationVendor" : "Sun Microsystems"
  } ]
}
"""

def test_get_beans():
  raw = simplejson.loads(_METRICS_SAMPLE_JMX)

  beans = _get_beans(raw, set([Bean(re.compile('does not exist'), 'alias')]), {})
  assert_equals(0, len(beans.keys()))

  additional_bean = Bean(re.compile('java.lang:type=MemoryPool,name=CMS Old Gen'),
                         'test-alias')

  beans = _get_beans(raw, set([jmx_beans._EVENT_STATS_BEAN, additional_bean]), {})
  assert_equals(5, beans[jmx_beans._EVENT_STATS_BEAN.get_alias()].SentCount)
  assert_true(beans[additional_bean.get_alias()] is not None)

def test_get_additional_attr_beans():
  raw = simplejson.loads(_METRICS_SAMPLE_JMX)

  beans = _get_beans(raw, set([jmx_beans._MGMT_JVM_PAUSE_MONITOR_BEAN]), {})
  pause_beans = beans[jmx_beans._MGMT_JVM_PAUSE_MONITOR_BEAN.get_alias()]
  _verify_pause_beans(pause_beans, False)
  # Test that if we use a bogus group that does not match anything we don't blow
  # up.
  bad_group_bean = BeanWithAdditionalAttributes(
      re.compile('com.cloudera.enterprise.debug:type=JvmPauseMonitor,name=(.+)'),
      'test_pause',
      singleton=False,
      groups_replacement_info={5: _MGMT_PAUSE_REPLACMENT_GROUP})
  beans = _get_beans(raw, set([bad_group_bean]), {})
  pause_beans = beans['test_pause']
  _verify_pause_beans(pause_beans, True)
  bad_group_bean = BeanWithAdditionalAttributes(
      re.compile('com.cloudera.enterprise.debug:type=JvmPauseMonitor'),
      'test_pause',
      singleton=False,
      groups_replacement_info={5: _MGMT_PAUSE_REPLACMENT_GROUP})

  beans = _get_beans(raw, set([bad_group_bean]), {})
  pause_beans = beans['test_pause']
  _verify_pause_beans(pause_beans, True)

def test_get_additional_attr_beans_no_replacements():
  raw = simplejson.loads(_METRICS_SAMPLE_JMX)
  no_replacement_group_bean = BeanWithAdditionalAttributes(
      re.compile('com.cloudera.enterprise.debug:type=JvmPauseMonitor,name=(.+)'),
      'test_pause',
      singleton=False,
      groups_replacement_info={1: MatchedGroupReplacement("test_attr", {})})
  beans = _get_beans(raw, set([no_replacement_group_bean]), {})
  pause_beans = beans['test_pause']
  assert_equals(2, len(pause_beans))
  for bean in pause_beans:
    assert_true("test_attr" in bean)
    attr = bean["test_attr"]
    assert_true(attr == "numPausesDetected" or
                attr == "pauseTimeMs")

def _verify_pause_beans(pause_beans, is_discrimiator_attr_none):
  assert_equals(2, len(pause_beans))
  for bean in pause_beans:
    if is_discrimiator_attr_none == False:
      assert_true(constants.PAUSE_BEAN_ATTR_NAME in bean)
      attr = bean[constants.PAUSE_BEAN_ATTR_NAME]
      assert_true(attr == constants.PAUSE_TIME_ALIAS or
                  attr == constants.PAUSES_ALIAS)
    else:
      assert_false(constants.PAUSE_BEAN_ATTR_NAME in bean)

def test_get_beans_port_awareness():
  raw = simplejson.loads(_METRICS_SAMPLE_FOR_PORTS_TESTS_JMX)

  to_match = [_make_port_aware_rpc_metrics_bean('NameNode',
                                                'rpc',
                                                'client-port'),
              _make_port_aware_detailed_rpc_metrics_bean('NameNode',
                                                         'detailed',
                                                         'client-port'),
              _make_port_aware_rpc_metrics_bean('NameNode',
                                                'service-rpc',
                                                'service-port'),
              _make_port_aware_detailed_rpc_metrics_bean('NameNode',
                                                         'service-detailed',
                                                         'service-port')]



  context = dict()
  context['client-port'] = 8020
  beans = _get_beans(raw, to_match, context)

  assert_equals(2, len(beans.keys()))
  assert_equals(96665, beans['rpc'].RpcAuthorizationSuccesses)
  assert_equals(24753, beans['detailed'].GetListingNumOps)

  context = dict()
  context['client-port'] = MATCH_ANY_PORT
  beans = _get_beans(raw, to_match, context)
  assert_equals(2, len(beans.keys()))
  assert_equals(96665, beans['rpc'].RpcAuthorizationSuccesses)
  assert_equals(24753, beans['detailed'].GetListingNumOps)

  context = dict()
  context['client-port'] = 8019
  beans = _get_beans(raw, to_match, context)
  assert_equals(0, len(beans.keys()))

  context = dict()
  context['client-port'] = 8020
  context['service-port'] = 8022
  beans = _get_beans(raw, to_match, context)
  assert_equals(4, len(beans.keys()))
  assert_equals(96665, beans['rpc'].RpcAuthorizationSuccesses)
  assert_equals(24753, beans['detailed'].GetListingNumOps)
  assert_equals(5203, beans['service-rpc'].RpcAuthorizationSuccesses)
  assert_equals(120, beans['service-detailed'].BlockReportNumOps)

def test_conf_parse():
  text = """[section_a]
uk = england,ireland,scotland
china = beijing
is_country = true
north_america = 3
"""
  conf_dir = make_config_file(text)
  try:
    cp = ConfParser(conf_dir, conf.CLOUDERA_MONITOR_PROPERTIES)
    assert_equals('beijing', cp.get('section_a', 'china'))
    assert_equals(True, cp.getboolean('section_a', 'is_country'))
    assert_equals(3, cp.getint('section_a', 'north_america'))
    assert_equals(['england', 'ireland', 'scotland'],
                  cp.getlist('section_a', 'uk'))
    assert_equals(None, cp.get('section_a', 'NOT THERE'))
  finally:
    shutil.rmtree(conf_dir)


def test_stoppable_thread():
  TIMEOUT = 30
  STOP_TIMEOUT = 5

  class Action(object):
    result = 0
    event = threading.Event()
    # The lock protects `result' and `event'
    lock = threading.Lock()

    @staticmethod
    def reset():
      Action.lock.acquire()
      Action.result = 0
      Action.event.clear()
      Action.lock.release()

    @staticmethod
    def do_sum(first, second):
      LOG.error("do_sum: %r, %s" % (first, second))
      Action.lock.acquire()
      Action.result = first + second
      Action.event.set()
      Action.lock.release()
      raise IOError("Because I want to fail")

  sthread = StoppableThread("Test", 0.1, Action.do_sum, 1, second=2)
  sthread.start()

  # Loop twice to make sure that it's not a one-shot event,
  # despite do_sum() raising an error.
  for _ in range(2):
    # Wait for a summation to happen
    Action.reset()
    Action.event.wait(TIMEOUT)
    # Assuming that Action.result is assigned atomically (which is true for int
    # type), we can read it without locking.
    assert_equals(3, Action.result)

  # It can be stopped properly
  sthread.should_stop()
  time.sleep(STOP_TIMEOUT)
  assert_false(sthread.isAlive())

def test_wakeable_thread():
  TIMEOUT = 2

  class Action(object):
    result = 0
    cv = threading.Condition()

    @staticmethod
    def do_sum(first, second):
      LOG.error("do_sum: %r, %s" % (first, second))
      Action.cv.acquire()
      Action.result = Action.result + first + second
      Action.cv.release()
      raise IOError("Because I want to fail")

  wthread1 = WakeableThread("Test", Action.do_sum, 1, second=2)
  wthread1.start()
  wthread2 = WakeableThread("Test", Action.do_sum, 2, second=6)
  wthread2.start()

  # Loop twice to make sure that it's not a one-shot event,
  # despite do_sum() raising an error.
  for _ in range(2):
    Action.result = 0
    latch = CountDownLatch(2)
    wthread1.wake(latch)
    wthread2.wake(latch)
    latch.await()
    assert_equals(11, Action.result)

  # It can be stopped properly
  latch = CountDownLatch(2)
  wthread1.should_stop(latch)
  wthread2.should_stop(latch)
  latch.await()
  time.sleep(TIMEOUT)
  assert_false(wthread1.isAlive())
  assert_false(wthread2.isAlive())

def test_wakeable_thread_timeout():
  TIMEOUT = 2

  class Action(object):
    @staticmethod
    def do(latch):
      latch.countdown()

  latch = CountDownLatch(1)
  wthread = WakeableThread("Test", Action.do, latch)
  wthread.set_timeout(0.1)
  wthread.start()
  latch.await()

  # It can be stopped properly
  latch = CountDownLatch(1)
  wthread.should_stop(latch)
  latch.await()
  time.sleep(TIMEOUT)
  assert_false(wthread.isAlive())

def test_wakeable_concurrent_wakes():
  TIMEOUT = 2

  class Action(object):
    calls = 0
    # The thread will signal this latch when it's entered its function.
    signal_latch = CountDownLatch(1)
    # The thread will wait for us to signal this latch to finish. This lets us
    # control when we add our second wake call.
    wait_latch = CountDownLatch(1)

    @staticmethod
    def do():
      Action.signal_latch.countdown()
      Action.wait_latch.await()
      Action.calls = Action.calls + 1

  wthread = WakeableThread("Test", Action.do)
  wthread.start()
  latch1 = CountDownLatch(1)
  latch2 = CountDownLatch(1)
  # Wake up the thread.
  wthread.wake(latch1)
  # Wait for it to get into the function. Then reset the latch.
  Action.signal_latch.await()
  Action.signal_latch = CountDownLatch(1)
  # Make a second wake call before letting the thread out of the function.
  wthread.wake(latch2)
  # Let the thread out of the function. We lave the latch signalled so that the
  # thread can make it through both calls.
  assert_equals(0, Action.calls)
  Action.wait_latch.countdown()
  latch1.await()
  latch2.await()
  assert_equals(2, Action.calls)

  # It can be stopped properly
  latch = CountDownLatch(1)
  wthread.should_stop(latch)
  latch.await()
  time.sleep(TIMEOUT)
  assert_false(wthread.isAlive())

def negative_test_count_down_latch():
  latch = CountDownLatch(1)
  latch.await(0.5)

  latch.countdown()
  assert_raises(Exception, latch.countdown)

def test_fs_update():
  fs_update = FsUpdate('/home', 'ro', '/dev/sda1', 'ext3').add_metric(metricid=99,
                                                              value=42)
  avro = fs_update.to_avro_dict()
  assert_equals(avro,
                dict(metrics=[{'id': 99, 'value': 42}],
                     mountpoint='/home',
                     mountOptions='ro',
                     partition='/dev/sda1',
                     fstype='ext3'))

def test_nic_update():
  nic_update = NicUpdate('eth0').add_metric(metricid=99, value=42)
  avro = nic_update.to_avro_dict()
  assert_equals(avro,
                dict(metrics=[{'id': 99, 'value': 42}],
                     iface='eth0',
                     duplexMode=constants.DUPLEX_UNKNOWN))

def test_disk_update():
  disk_update = DiskUpdate('/dev/sda1', True).add_metric(metricid=99,
                                                         value=42)
  avro = disk_update.to_avro_dict()
  assert_equals(avro,
                dict(metrics=[{'id': 99, 'value': 42}],
                     device='/dev/sda1',
                     partition=True))

def test_dir_monitor():
  i, path = tempfile.mkstemp()
  dir_monitor = DirMonitor(None, path)
  os.chmod(path, stat.S_IRWXG | stat.S_IRWXU | stat.S_IROTH | stat.S_IXOTH)
  perms, owner, grp = dir_monitor._get_directory_attributes()
  assert_equals(str(perms), '0775')
  assert_equals(owner, getpass.getuser())

def test_host_update():
  host_update = HostUpdate("/var/log", "/var/run", "/opt/cloudera/parcel").\
      add_metric(metricid=1, value=1).\
      add_fs_update(FsUpdate('/vol1', 'rw', '/dev/sda1', 'ext3').\
                    add_metric(metricid=1, value=1)).\
      add_nic_update(NicUpdate('eth0').
                     add_metric(metricid=1, value=1)).\
      add_disk_update(DiskUpdate('/dev/sda1', True).
                      add_metric(metricid=1, value=1)).\
      add_dir_update(
    HostDirUpdate('/dir/1', '/var', 'rw', '/dev/sda', None, None, None, None).
                     add_metric(metricid=2, value=2)).\
      add_dir_update(
    HostDirUpdate('/dir/2', '/var', 'ro', '/dev/sda1', None, None, None, None).
                     add_metric(metricid=2, value=2))

  host_update2 = HostUpdate("/var/log", "/var/run", "/opt/cloudera/parcel").\
      add_metric(metricid=2, value=2).\
      add_fs_update(FsUpdate('/vol2', 'ro', '/dev/sda2', 'ext3').
                    add_metric(metricid=2, value=2)).\
      add_nic_update(NicUpdate('eth1').
                     add_metric(metricid=2, value=2)).\
      add_disk_update(DiskUpdate('/dev/sda2', True).
                      add_metric(metricid=2, value=2)).\
      add_dir_update(
    HostDirUpdate('/dir/3', '/var', 'rw', '/dev/sda', None, None, None, None).
                     add_metric(metricid=3, value=3)).\
      add_dir_update(
    HostDirUpdate('/dir/4', '/var', 'ro', '/dev/sda1', None, None, None, None).
                     add_metric(metricid=3, value=3))

  # Check that we merge correctly
  host_update.merge(host_update2)
  dic = host_update.to_avro_dict()
  assert_equals(dic['fsUpdates'],
                [ dict(mountpoint='/vol1',
                       mountOptions='rw',
                       metrics=[{'id': 1, 'value': 1}],
                       partition='/dev/sda1',
                       fstype='ext3'),
                  dict(mountpoint='/vol2',
                       mountOptions='ro',
                       metrics=[{'id': 2, 'value': 2}],
                       partition='/dev/sda2',
                       fstype='ext3')
                ])
  assert_equals(dic['nicUpdates'],
                [ dict(iface='eth0',
                       metrics=[{'id': 1, 'value': 1}],
                       duplexMode=constants.DUPLEX_UNKNOWN),
                  dict(iface='eth1',
                       metrics=[{'id': 2, 'value': 2}],
                       duplexMode=constants.DUPLEX_UNKNOWN)
                ])
  assert_equals(dic['diskUpdates'],
                [ dict(device='/dev/sda1',
                       partition=True,
                       metrics=[{'id': 1, 'value': 1}]),
                  dict(device='/dev/sda2',
                       partition=True,
                       metrics=[{'id': 2, 'value': 2}])
                ])
  assert_equals(dic['dirUpdates'],
                [ dict(path='/dir/1',
                       mountOptions='rw',
                       mountpoint='/var',
                       partition='/dev/sda',
                       metrics=[{'id': 2, 'value': 2}],
                       permissions=None,
                       owner=None,
                       group=None,
                       fstype=None),
                  dict(path='/dir/2',
                       mountOptions='ro',
                       mountpoint='/var',
                       partition='/dev/sda1',
                       metrics=[{'id': 2, 'value': 2}],
                       permissions=None,
                       owner=None,
                       group=None,
                       fstype=None),
                  dict(path='/dir/3',
                       mountOptions='rw',
                       mountpoint='/var',
                       partition='/dev/sda',
                       metrics=[{'id': 3, 'value': 3}],
                       permissions=None,
                       owner=None,
                       group=None,
                       fstype=None),
                  dict(path='/dir/4',
                       mountOptions='ro',
                       mountpoint='/var',
                       partition='/dev/sda1',
                       metrics=[{'id': 3, 'value': 3}],
                       permissions=None,
                       owner=None,
                       group=None,
                       fstype=None),
                ])
  assert_equals(dic['metrics'], [{'id': 1, 'value': 1},
                                 {'id': 2, 'value': 2}])

def test_impala_query_update():
  executing_profile_one = ImpalaRuntimeProfile("blah1")
  executing_profile_two = ImpalaRuntimeProfile("blah2")
  completed_profile_one = ImpalaRuntimeProfile("blah3")
  completed_profile_two = ImpalaRuntimeProfile("blah4")
  impala_query_update = \
      ImpalaQueryUpdate(role_name="role",
                        service_name="service",
                        impala_version="version",
                        executing_query_profiles=[executing_profile_one,
                                                  executing_profile_two],
                        completed_query_profiles=[completed_profile_one,
                                                  completed_profile_two])

  dic = impala_query_update.to_avro_dict()

  assert_equals("role", dic["roleName"])
  assert_equals("service", dic["serviceName"])
  assert_equals("version", dic["impalaVersion"])
  assert_equals(2, len(dic["executingQueryProfiles"]))
  assert_equals(2, len(dic["completedQueryProfiles"]))
  assert_equals("blah1", dic["executingQueryProfiles"][0]["compressedRuntimeProfile"])
  assert_equals("blah2", dic["executingQueryProfiles"][1]["compressedRuntimeProfile"])
  assert_equals("blah3", dic["completedQueryProfiles"][0]["compressedRuntimeProfile"])
  assert_equals("blah4", dic["completedQueryProfiles"][1]["compressedRuntimeProfile"])

def test_proc_metrics():
  pid = os.getpid()
  assert_true(get_fd_count(pid) > 0)
  assert_equals(get_fd_max(pid), resource.getrlimit(resource.RLIMIT_NOFILE)[1])
  swap_out = get_swap_out()
  assert_true(swap_out is not None)

def test_firehoses():
  firehoses = Firehoses("hostid", "hostname")
  assert_equals(0, len(firehoses.get_debug_status_reports()))
  assert_equals(None, firehoses.get_last_amon_message())
  assert_equals(None, firehoses.get_last_smon_message())
  assert_equals(None, firehoses.get_last_hmon_message())
  assert_equals(5.0, firehoses.get_reporting_interval_seconds())

  # Let's make sure these don't blow up even if no connection information has
  # been sent.
  firehoses.send_amon_update(attempt_updates=[])
  firehoses.send_smon_update(service_updates=[],
                             role_updates=[],
                             impala_query_updates=[])
  firehoses.send_hmon_update(host_update=None)

  firehoses.update_firehose_connection_info([{ 'rolename': 'amon-name',
                                               'roletype': 'ACTIVITYMONITOR',
                                               'address': '127.0.0.1',
                                               'port': 50000,
                                               'report_interval': 1.7}])
  assert_equals(1.7, firehoses.get_reporting_interval_seconds())
  # The actual update is processed on the first message attempt, so do that.
  firehoses.send_amon_update(attempt_updates=[])
  firehoses.send_smon_update(service_updates=[],
                             role_updates=[],
                             impala_query_updates=[])
  firehoses.send_hmon_update(host_update=None)
  assert_true(firehoses._amon is not None)
  assert_true(firehoses._hmon is None)
  assert_true(firehoses._smon is None)


  firehoses.update_firehose_connection_info([{ 'rolename': 'smon-name',
                                               'roletype': 'SERVICEMONITOR',
                                               'address': '127.0.0.1',
                                               'port': 50001,
                                               'report_interval': 10.5},
                                             { 'rolename': 'hmon-name',
                                               'roletype': 'HOSTMONITOR',
                                               'address': '127.0.0.1',
                                               'port': 50002,
                                               'report_interval': 3.4}])
  assert_equals(3.4, firehoses.get_reporting_interval_seconds())
  # The actual update is processed on the first message attempt, so do that.
  firehoses.send_amon_update(attempt_updates=[])
  firehoses.send_smon_update(service_updates=[],
                             role_updates=[],
                             impala_query_updates=[])
  firehoses.send_hmon_update(host_update=None)
  assert_true(firehoses._amon is None)
  assert_true(firehoses._hmon is not None)
  assert_true(firehoses._smon is not None)

def test_metric_normalization():
  value = (1 << 63) + 5
  # We treat numbers larger than (1 << 63) - 1 as "counter resets".
  assert_equals(6, Firehoses.normalize_metric_value(value))

  value = (1 << 63) - 2
  assert_equals(value, Firehoses.normalize_metric_value(value))

  value = "string value"
  assert_equals(value, Firehoses.normalize_metric_value(value))

  value = -1
  assert_equals(value, Firehoses.normalize_metric_value(value))

  value = -(1 << 63) - 5;
  assert_equals(-5, Firehoses.normalize_metric_value(value))

  value = -(1 << 63) + 1;
  assert_equals(value, Firehoses.normalize_metric_value(value))

  value = 0
  assert_equals(value, Firehoses.normalize_metric_value(value))

def test_collection_robustness():
  class MockMonitor(object):
    def __init__(self, throw):
      self._throw = throw

    def get_role_update(self):
      if self._throw:
        raise IOError("BOOM!")
      else:
        return "role update"

    def get_service_update(self):
      if self._throw:
        raise IOError("BOOM!")
      else:
        return "service update"

    def get_host_update(self):
      if self._throw:
        raise IOError("BOOM!")
      else:
        return "host update"

  daemon = MonitorDaemon("host", "host", agent=MockAgent())
  monitor = MockMonitor(False)

  assert_equals("role update", daemon._safe_get_role_update(monitor))
  assert_equals("service update", daemon._safe_get_service_update(monitor))
  assert_equals("host update", daemon._safe_get_host_update(monitor))

  monitor = MockMonitor(True)

  assert_equals(None, daemon._safe_get_role_update(monitor))
  assert_equals(None, daemon._safe_get_service_update(monitor))
  assert_equals(None, daemon._safe_get_host_update(monitor))

def test_invalid_process_name():
  # Test that we don't blow up with an invalid process name
  daemon = MonitorDaemon("host", "host", agent=MockAgent())
  assert_true(daemon.make_monitor("HBaseShutdown", None, None) is None)
  assert_true(daemon.make_monitor("name-too-long4u", None, None) is None)
  assert_true(daemon.make_monitor("", None, None) is None)

def test_make_monitor_with_overrides():
  processes = [
    'mapreduce-JOBTRACKER',
    'mapreduce-TASKTRACKER',
    'flume-AGENT',
    'impala-IMPALAD',
    'solr-SOLR_SERVER',
    'cloudera-mgmt-ALERTPUBLISHER',
    'yarn-JOBHISTORY']

  safety_valve = \
"""
use_generic_monitor_for_mapreduce_jobtracker = true
use_generic_monitor_for_mapreduce_tasktracker = true
use_generic_monitor_for_flume_agent = true
use_generic_monitor_for_solr_solr_server= tRue
use_generic_monitor_for_mgmt_alertpublisher = true
use_generic_monitor_for_yarn_jobhistory= true
"""
  agent = MockAgent()
  daemon = MonitorDaemon("host", "host", agent=agent)
  agent.safety_valve.update_safety_valve(
    dict(host_agent_safety_valve=safety_valve))

  for process in processes:
    monitor = daemon.make_monitor(process, process, process)
    assert_true(isinstance(monitor, GenericMonitor))
    assert_true(isinstance(monitor._adapter, Adapter))

  safety_valve = \
"""
use_generic_monitor_for_maprEduce_jobtracker = true
use_generic_monitor_for_mapreduce_tAsktracker = true
use_generic_monitor_for_flUme_agent = true
use_generic_monitor_for_soLr_solr_server=false
use_generic_monitor_for_MGMT_alertpublisher = true
use_generic_monitor_for_yarn_JOBHISTORY= false
"""
  agent.safety_valve.update_safety_valve(
    dict(host_agent_safety_valve=safety_valve))
  monitor = daemon.make_monitor("solr-SOLR_SERVER",
                                "solr-SOLR_SERVER",
                                "solr-SOLR_SERVER")
  assert_true(isinstance(monitor, SolrServerMonitor))
  monitor = daemon.make_monitor("yarn-JOBHISTORY",
                                "yarn-JOBHISTORY",
                                "yarn-JOBHISTORY")
  # Since the keys are case sensitive we should ignore the yarn_JOBHISTORY key.
  assert_true(isinstance(monitor, GenericMonitor))
  assert_true(isinstance(monitor._adapter, Adapter))

def test_urllib_retry_on_auth_error():
  global urlopen_calls

  safety_valve = """
urlopen_retries_on_auth_errors=9
urlopen_sleep_seconds_on_auth_errors=0.0001
"""
  agent = MockAgent()
  daemon = MonitorDaemon("host", "host", agent=agent)
  agent.safety_valve.update_safety_valve(
    dict(host_agent_safety_valve=safety_valve))
  monitor = daemon.make_monitor("yarn-JOBHISTORY",
                                "yarn-JOBHISTORY",
                                "yarn-JOBHISTORY")
  urlopen_calls = 0
  def test_urlopen_with_timeout_1(url,
                                  data=None,
                                  timeout=None,
                                  secure_http_service_name=None,
                                  username=None,
                                  password=None,
                                  cafile=None,
                                  capath=None,
                                  max_cert_depth=9):
    global urlopen_calls
    urlopen_calls += 1
    raise urllib2.HTTPError(url, 401, "BOOM!", None, None)

  monitor._urlopen_with_timeout = test_urlopen_with_timeout_1
  assert_raises(urllib2.HTTPError, monitor.urlopen, "testurl")
  assert_equals(10, urlopen_calls)

  def test_urlopen_with_timeout_2(url,
                                  data=None,
                                  timeout=None,
                                  secure_http_service_name=None,
                                  username=None,
                                  password=None,
                                  cafile=None,
                                  capath=None,
                                  max_cert_depth=9):
    global urlopen_calls
    urlopen_calls += 1
    return StringIO("GOOD!")

  agent.safety_valve.update_safety_valve(
    dict(host_agent_safety_valve=""))
  monitor._urlopen_with_timeout = test_urlopen_with_timeout_2
  assert_equals("GOOD!", monitor.urlopen("testurl").read())
  assert_equals(11, urlopen_calls)
