# Copyright (c) 2015 Cloudera, Inc. All rights reserved.
import os
import shutil
import time

try:
  from cStringIO import StringIO
except:
  from StringIO import StringIO

from cmf.monitor.abstract_monitor import CollectionError
from cmf.monitor import constants
from cmf.monitor.generic.adapter import Adapter
from cmf.monitor.generic.metric_collectors import SampleFileMetricCollector
from cmf.monitor.role_update import RoleUpdate
from cmf.safety_valve import SafetyValve
from cmf_test.monitor import make_config_file

from nose.tools import assert_true, assert_equals, assert_false

class MockAdapter(Adapter):
  def __init__(self):
    Adapter.__init__(self, "SOMESERVICE", "SOMEROLE", SafetyValve())
    self._metrics_file = None

  def get_metrics_file(self, conf):
    return self._metrics_file

  def set_metrics_file(self, metrics_file):
    self._metrics_file = metrics_file

  def parse_metrics_from_file(self, conf, json):
    extracted_metrics = dict()
    for metric in json.get('metrics', []):
      metricid = metric.get('id', None)
      if metricid is None:
        continue
      value = metric.get('value', None)
      if value is None:
        continue
      extracted_metrics[metricid] = float(value)
    return extracted_metrics

  def add_sample_file_metrics(self,
                              service_version,
                              update,
                              collected_metrics,
                              metric_accessors):
    for metricid, value in collected_metrics.iteritems():
      update.add_metric(metricid, value)

class MockSampleFileEndPoint(SampleFileMetricCollector):

  def __init__(self, adapter):
    SampleFileMetricCollector.__init__(self, adapter, "metric_accessors")
    self._error = None
    self._json = None

  def set_throw_error(self, error):
    self._error = error

  def set_json_string(self, json):
    self._json = json

  def set_process_start_time(self, process_start_time):
    self._process_start_time = process_start_time

  def _get_json(self, file):
    if self._error is not None:
      raise self._error
    if self._json is not None:
      return StringIO(self._json)
    return SampleFileMetricCollector._get_json(self, file)

  def reset_sample_mtime(self):
    self._last_sample_file_collected_modified_time = 0

def test_ready_to_report():
  conf_dir = make_config_file(
"""[somerole]
role_name = someservice-somerole-test
service_name = someservice
host = localhost
service_version = 4
collect_interval = 1.0
log_dir = /path/to/logs
""")

  sample_dir = make_config_file(
"""
{
"metrics": [
  { "id": 66666, "value": -3.14 },
  { "id": 66667, "value": 3.15 }
]
}
""",
filename="sample_file.json"
)

  try:
    pid = os.getpid()
    collector = MockSampleFileEndPoint(MockAdapter())
    is_ready_to_report = collector.is_ready_to_report("UNUSED_CONF", pid)
    assert_equals(None, is_ready_to_report)

    the_far_future = time.time() + 500
    collector.set_process_start_time(the_far_future)
    collector._adapter.set_metrics_file(
      os.path.join(sample_dir, "sample_file.json"))
    is_ready_to_report = collector.is_ready_to_report("UNUSED_CONF", pid)
    assert_false(is_ready_to_report)

    collector.set_process_start_time(time.time() - 500)
    is_ready_to_report = collector.is_ready_to_report("UNUSED_CONF", pid)
    assert_true(is_ready_to_report)
  finally:
    shutil.rmtree(conf_dir)
    shutil.rmtree(sample_dir)

def test_collection_timestamp():
  conf_dir = make_config_file(
"""[somerole]
role_name = someservice-somerole-test
service_name = someservice
host = localhost
service_version = 4
collect_interval = 1.0
log_dir = /path/to/logs
""")

  sample_dir = make_config_file(
"""
{
"metrics": [
  { "id": 66666, "value": -3.14 },
  { "id": 66667, "value": 3.15 }
]
}
""",
filename="sample_file.json"
)

  try:
    collector = MockSampleFileEndPoint(MockAdapter())
    collector._adapter.set_metrics_file(
      os.path.join(sample_dir, "sample_file.json"))
    collector.update_with_conf("UNUSED_CONF")
    update = RoleUpdate(name="ROLE_NAME",
                        roletype="ROLE_TYPE",
                        servicetype="SERVICE_TYPE",
                        service_name="SERVICE_NAME")
    collector.reset_sample_mtime()
    collector.collect_and_parse("UNUSED_CONF")
    has_something_to_report = \
      collector.report_metrics("SERVICE_VERSION", update)
    assert_true(has_something_to_report)

    collector.collect_and_parse("UNUSED_CONF")
    has_something_to_report = \
      collector.report_metrics("SERVICE_VERSION", update)
    assert_false(has_something_to_report)

    collector.reset_sample_mtime()
    collector.collect_and_parse("UNUSED_CONF")
    has_something_to_report = \
      collector.report_metrics("SERVICE_VERSION", update)
    assert_true(has_something_to_report)

    # make sure we have nothing to report.
    collector.collect_and_parse("UNUSED_CONF")
    has_something_to_report = \
      collector.report_metrics("SERVICE_VERSION", update)
    assert_false(has_something_to_report)
    # now "simulate" an IO error. This should trigger a report.
    collector.set_throw_error(Exception("Boom!"))
    collector.reset_sample_mtime()
    collector.collect_and_parse("UNUSED_CONF")
    results, timestamp, duration = collector.metrics_tuple
    assert_true(isinstance(results, CollectionError))
    assert_equals(constants.FILE_METRICS_COLLECTION_UNKNOWN,
                  results.code)
    assert_true(timestamp > collector._last_sample_file_collected_modified_time)

    # "remove" the error. Sinced we touched the file in the previous section to
    # ensure that we try to read from the file we should now report the metrics.
    collector.set_throw_error(None)
    collector.collect_and_parse("UNUSED_CONF")
    has_something_to_report = \
      collector.report_metrics("SERVICE_VERSION", update)
    assert_true(has_something_to_report)

    # now "simulate" a parse failure. This should trigger a report.
    collector.set_json_string("BOGUS_JSON")
    collector.reset_sample_mtime()
    collector.collect_and_parse("UNUSED_CONF")
    results, timestamp, duration = collector.metrics_tuple
    assert_true(isinstance(results, CollectionError))
    assert_equals(constants.FILE_METRICS_COLLECTION_PARSE_FAILURE,
                  results.code)
    assert_true(timestamp > collector._last_sample_file_collected_modified_time)

    # now make sure that we get stale result if the file has not changed since
    # last time.
    collector.set_json_string(None)
    # this sleep is okay as in real life collection is done every 60 seconds. We
    # want to make sure the timestamp in the result is updated.
    collector.collect_and_parse("UNUSED_CONF")
    collector.clear_last_reported_timestamp()
    has_something_to_report = \
      collector.report_metrics("SERVICE_VERSION", update)
    assert_true(has_something_to_report)

    collector.collect_and_parse("UNUSED_CONF")
    results, timestamp, duration = collector.metrics_tuple
    assert_true(isinstance(results, CollectionError))
    assert_equals(constants.FILE_METRICS_COLLECTION_STALE_SAMPLE,
                  results.code)
    assert_true(timestamp > collector._last_sample_file_collected_modified_time)
  finally:
    shutil.rmtree(conf_dir)
    shutil.rmtree(sample_dir)
