# Copyright (c) 2013 Cloudera, Inc. All rights reserved.

import logging
import os
import shutil
import tempfile

try:
  from cStringIO import StringIO
except:
  from StringIO import StringIO

from cmf.monitor import schema
from cmf.monitor import constants
from cmf.monitor.solrserver import SolrServerMonitor
from solrserver.sample_data import SAMPLE_CDH5
from solrserver.sample_data import SAMPLE_CDH6
from solrserver.sample_data import SAMPLE_CDH6_G1GC
from solrserver.sample_data import SAMPLE_CDH6_INCOMPLETE
from cmf.monitor.solrserver.json_parser_jmx import SOLR_BEAN_TYPES
from cmf.monitor.solrserver.json_parser_jmx import IGNORED_SOLR_BEANS
from cmf.monitor.solrserver.json_parser_jmx import SOLR_CORE_STATUS_METRICS
from cmf.monitor.json_metrics_utils import MetricAccessor
from cmf_test.monitor import make_config_file, TestMonitorDaemon

from nose.tools import assert_true, assert_equals, assert_false


LOG = logging.getLogger(__name__)

class MockSolrServerMonitor(SolrServerMonitor):
  def __init__(self, daemon, registry, jmx_string, core_status_string):
    SolrServerMonitor.__init__(self, daemon, registry)
    self.set_jmx_string(jmx_string)
    self.set_core_status_string(core_status_string)
    self.set_throw_error(None)

  def set_jmx_string(self, jmx_string):
    self._jmx_string = jmx_string

  def set_core_status_string(self, core_status_string):
    self._core_status_string = core_status_string

  def set_throw_error(self, error):
    self._error = error

  def urlopen(self, url, **kwargs):
    if self._error:
      raise self._error
    if 'jmx' in url or 'metrics' in url:
      return StringIO(self._jmx_string)
    else:
      return StringIO(self._core_status_string)

  def set_ccache_name(self, ccache_name):
    self._krb5cc_name = ccache_name

# CM 5.0.0 config file
def _make_cm50_format_config_file(version, solr_http_port):
  properties = """[solr_server]
role_name = solr-SOLR_SERVER
service_name = my_solr
service_version = %s
host = localhost
solr_http_port = %s
log_dir = /path/to/logs
collect_interval = 1.0
"""
  return  make_config_file(properties % (version, solr_http_port,))

def test_solr_server_lifecycle_cdh5():
  do_solr_server_lifecycle(5, _make_cm50_format_config_file, SAMPLE_CDH5, SAMPLE_CORE_STATUS)

def test_solr_server_lifecycle_cdh6():
  do_solr_server_lifecycle(6, _make_cm50_format_config_file, SAMPLE_CDH6, SAMPLE_CORE_STATUS)

def test_solr_server_lifecycle_cdh6_g1gc():
  do_solr_server_lifecycle_gc_only(6, _make_cm50_format_config_file, SAMPLE_CDH6_G1GC, SAMPLE_CORE_STATUS, True)

def test_solr_server_lifecycle_cdh6_nogc():
  do_solr_server_lifecycle_gc_only(6, _make_cm50_format_config_file, SAMPLE_CDH6_INCOMPLETE, SAMPLE_CORE_STATUS, False)

def test_no_collection_if_ccache_is_not_there():
  daemon = TestMonitorDaemon()
  reg = daemon.get_registry_for_testing()
  monitor = MockSolrServerMonitor(daemon, reg, SAMPLE_CDH5, SAMPLE_CORE_STATUS)
  monitor.refresh_pid(os.getpid(), True)
  conf_dir = _make_cm50_format_config_file(5, 50075)
  # Start it
  monitor.refresh(conf_dir)
  ccache_file_name = tempfile.mktemp()
  # so since the ccache_file_name is not supposed to exists we should not collect.
  monitor.set_ccache_name(ccache_file_name)
  monitor.collect()
  assert_equals((None, None, None), monitor._metrics_tuple)

  ccache_file_name = tempfile.mkstemp()[1]
  try:
    monitor.set_ccache_name(ccache_file_name)
    monitor.collect()
    result, timestamp, duration = monitor._metrics_tuple
    assert_true(result is not None)
    assert_true(timestamp is not None)
    assert_true(duration is not None)
  finally:
    os.unlink(ccache_file_name)

def get_metrics(handler_metric_ids, solr_replica_updates):
  metrics_found = []
  for replica_update in solr_replica_updates:
    replica_metrics = replica_update['metrics']
    for replica in replica_metrics:
      if replica['id'] in handler_metric_ids:
        metrics_found.append(replica['id'])
  return metrics_found

def is_update_handler_metric(accessor, handler, version):
  if version == 5:
    return accessor.bean_alias == handler
  else:
    if not isinstance(accessor, MetricAccessor):
      return False
    section = accessor.section if accessor.section is not None else accessor.metric
    return "UPDATE./" in section

def verify_update_handler_metrics_present(metric_accessors,
                                          solr_replica_updates, version):
  handler = 'updatehandler'
  handler_metric_ids = []

  for metric_id, accessor in metric_accessors.iteritems():
    if is_update_handler_metric(accessor, handler, version):
      handler_metric_ids.append(metric_id)

  metrics_found = get_metrics(handler_metric_ids, solr_replica_updates)
  assert_true(len(metrics_found) > 0,
              "%s metrics not found in solr update" % (handler,))

def verify_handler_metrics_present_cdh5(metric_accessors,
                                   solr_replica_updates,
                                   handler):
  query_handlers = SOLR_BEAN_TYPES[handler.upper()]
  for query_handler in query_handlers:
    ignored_handlers = IGNORED_SOLR_BEANS.get(handler.upper(), None)
    if ignored_handlers is not None and query_handler in ignored_handlers:
      continue
    handler_metric_ids = []
    for metric_id, accessor in metric_accessors.iteritems():
      if accessor.bean_alias == handler and accessor.bean_id == query_handler:
        handler_metric_ids.append(metric_id)
      else:
        # Beans that need both id and type for identity
        spl = accessor.bean_id.split('::', 1)
        if len(spl) > 1:
          acc_bean_id = spl[0]
          acc_bean_type = spl[1]
          if accessor.bean_alias == handler and acc_bean_id == query_handler:
            handler_metric_ids.append(metric_id)
    metrics_found = get_metrics(handler_metric_ids, solr_replica_updates)
    assert_true(len(metrics_found) > 0,
                "%s metrics for %s not found in solr update" %
                (query_handler, handler,))

def verify_handler_metrics_present_cdh6(metric_accessors,
                                   solr_replica_updates,
                                   handler, expectedCount):
  handler_metric_ids = []
  for metric_id, accessor in metric_accessors.iteritems():
    if not isinstance(accessor, MetricAccessor): continue
    section = accessor.section if accessor.section is not None else accessor.metric
    if not section.startswith(handler): continue
    handler_metric_ids.append(metric_id)

  metrics_found = get_metrics(handler_metric_ids, solr_replica_updates)
  assert_true(len(metrics_found) == expectedCount,
              "metrics for %s not found in solr update" %
              (handler,))

def do_solr_server_lifecycle_gc_only(version, make_config_file_function, \
                                     jmx_json_sample, core_status_json_sample, expect_data):

  daemon = TestMonitorDaemon()
  agent = daemon._agent
  reg = daemon.get_registry_for_testing()
  monitor = MockSolrServerMonitor(daemon, reg, jmx_json_sample,
                                  core_status_json_sample)
  monitor.refresh_pid(os.getpid(), True)

  # Make two different configs
  conf_dir = make_config_file_function(version, 50075)
  conf_dir2 = make_config_file_function(version, 50076)

  try:
    assert_equals(0, len(reg.get_registry()))

    # Start it
    monitor.refresh(conf_dir)
    assert_equals(monitor, reg.get_registry().pop())
    monitor.collect()
    assert_true('50075' in monitor.metrics_url, "Config installed")

    # Refresh with a different config
    monitor.refresh(conf_dir2)
    assert_equals(monitor, reg.get_registry().pop())
    assert_equals(1, len(reg.get_registry()))

    monitor.collect()
    assert_true('50076' in monitor.metrics_url, "Using the new config")

    # Test metrics and role update
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()

    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']

    # Verify that GC metrics are included
    if expect_data:
      assert_true(schema.JVM_GC_COUNT in metric_dict)
      assert_true(schema.JVM_GC_TIME_MS in metric_dict)
    else:
      assert_false(schema.JVM_GC_COUNT in metric_dict)
      assert_false(schema.JVM_GC_TIME_MS in metric_dict)

    # Test when the process is not running
    monitor.refresh_pid(None)
    monitor.collect()

    # Stop it
    monitor.stop()
  finally:
    shutil.rmtree(conf_dir)
    shutil.rmtree(conf_dir2)


def do_solr_server_lifecycle(version, make_config_file_function, \
                             jmx_json_sample, core_status_json_sample):

  daemon = TestMonitorDaemon()
  agent = daemon._agent
  reg = daemon.get_registry_for_testing()
  monitor = MockSolrServerMonitor(daemon, reg, jmx_json_sample,
                                  core_status_json_sample)
  monitor.refresh_pid(os.getpid(), True)

  # Make two different configs
  conf_dir = make_config_file_function(version, 50075)
  conf_dir2 = make_config_file_function(version, 50076)

  try:
    assert_equals(0, len(reg.get_registry()))

    # Start it
    monitor.refresh(conf_dir)
    assert_equals(monitor, reg.get_registry().pop())
    monitor.collect()
    assert_true('50075' in monitor.metrics_url, "Config installed")

    # Refresh with a different config
    monitor.refresh(conf_dir2)
    assert_equals(monitor, reg.get_registry().pop())
    assert_equals(1, len(reg.get_registry()))

    monitor.collect()
    assert_true('50076' in monitor.metrics_url, "Using the new config")

    # Test metrics and role update
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()

    assert_equals('solr-SOLR_SERVER', role_update['rolename'])
    assert_equals('my_solr', role_update['service_name'])
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']

    assert_true(schema.TOTAL_CPU_USER in metric_dict)
    assert_true(schema.TOTAL_CPU_SYSTEM in metric_dict)
    assert_equals(constants.WEB_METRICS_COLLECTION_GOOD,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)

    # Verify that memory metrics are included
    assert_true(schema.JVM_HEAP_COMMITTED_MB in metric_dict)
    assert_true(schema.JVM_HEAP_USED_MB in metric_dict)
    assert_true(schema.JVM_NON_HEAP_COMMITTED_MB in metric_dict)
    assert_true(schema.JVM_NON_HEAP_USED_MB in metric_dict)

    # Verify that GC metrics are included
    assert_true(schema.JVM_GC_COUNT in metric_dict)
    assert_true(schema.JVM_GC_TIME_MS in metric_dict)

    # Verify that Solr core status collection metrics are included
    assert_equals(constants.SOLR_CORE_STATUS_COLLECTION_GOOD,
                  metric_dict[schema.SOLR_CORE_STATUS_COLLECTION_STATUS])
    assert_true(schema.SOLR_CORE_STATUS_COLLECTION_DURATION in metric_dict)

    # Verify that direct memory metrics are included
    assert_true(schema.JVM_DIRECT_MEMORY_BUFFERS in metric_dict)
    assert_true(schema.JVM_DIRECT_MEMORY_USED in metric_dict)
    assert_true(schema.JVM_DIRECT_MEMORY_CAPACITY in metric_dict)

    if version == 5:
      version_string = 'cdh5'
    if version == 6:
      version_string = 'cdh6'

    # Test fs update
    dir_metrics = role_update['dirUpdates']
    assert_true(len(dir_metrics) == 1, "We configured 1 dir: %s" % dir_metrics)
    assert_true(os.path.ismount(dir_metrics[0]['mountpoint']))
    assert_equals('/path/to/logs', dir_metrics[0]['path'])

    # Test Solr updates
    solr_replica_updates = role_update['solrUpdates']
    assert_true(len(solr_replica_updates) == 3, \
                "We configured 3 replica updates but have %d instead" % \
                (len(solr_replica_updates),))

    metric_accessors = daemon.metric_accessors.get("SOLR_REPLICA", {}).get(version_string, {})
    # Verify UPDATEHANDLER metrics are present
    verify_update_handler_metrics_present(metric_accessors, solr_replica_updates, version)
    # Verify QUERYHANDLER metrics are present
    if version == 5:
      verify_handler_metrics_present_cdh5(metric_accessors, solr_replica_updates,
                                     'queryhandler')
      verify_handler_metrics_present_cdh5(metric_accessors, solr_replica_updates,
                                     'cache')
    else:
      verify_handler_metrics_present_cdh6(metric_accessors, solr_replica_updates,
                                          'QUERY', 108)
      verify_handler_metrics_present_cdh6(metric_accessors, solr_replica_updates,
                                          'CACHE', 98)

    # Verify we have all expected metrics from the Core status API call as well as
    # those from JMX
    metric_ids = metric_accessors.keys()
    core_status_metric_ids = SOLR_CORE_STATUS_METRICS.values()
    for replica_update in solr_replica_updates:
      replica_metrics = replica_update['metrics']
      core_status_metrics_seen = set([])
      for metric in replica_metrics:
        jmx_metric = False
        if metric['id'] in core_status_metric_ids:
          core_status_metrics_seen.add(metric['id'])
        else:
          jmx_metric = metric['id'] in metric_ids
        assert_true(jmx_metric or metric['id'] in core_status_metrics_seen, \
                    "Metric id %s is not expected in a solr replica update" % \
                    (metric['id'],))
      for metric in core_status_metric_ids:
        assert_true(metric in core_status_metrics_seen,
                    "Core status metric %d missing from Solr update" % \
                    (metric,))


    # Test safety_valve handling
    monitor.refresh_pid(os.getpid(), True)
    monitor.collect()
    safety_valve_str = \
      "my_solr_solr-solr_server_feature_send_replica_update_enabled= FALSE"
    agent.safety_valve.update_safety_valve(
      dict(host_agent_safety_valve=safety_valve_str))
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    solr_replica_updates = role_update['solrUpdates']
    assert_equals(0, len(solr_replica_updates))

    # Test when the process is not running
    monitor.refresh_pid(None)
    monitor.collect()

    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()

    assert_equals('solr-SOLR_SERVER', role_update['rolename'])
    assert_equals('my_solr', role_update['service_name'])
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']

    dir_metrics = role_update['dirUpdates']
    assert_true(len(dir_metrics) == 1, "We configured 1 dir: %s" % dir_metrics)
    assert_true(os.path.ismount(dir_metrics[0]['mountpoint']))
    assert_equals('/path/to/logs', dir_metrics[0]['path'])

    # Test with a collection error
    monitor.refresh_pid(os.getpid(), True)
    monitor.set_throw_error(IOError("BOOM!"))
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_equals(constants.WEB_METRICS_COLLECTION_COMMUNICATION_FAILURE,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)

    # Test with some invalid JSON
    monitor.set_throw_error(None)
    monitor.set_jmx_string("THIS IS BOGUS JSON!")
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_equals(constants.WEB_METRICS_COLLECTION_PARSE_FAILURE,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)

    # Test with a sample with invalid Solr Beans
    # Should not throw any exceptions / fail.
    # - missing 'name' for the different categoris
    # - missing 'coreHashCode' for the different categoris
    # - missing 'category'
    monitor.set_throw_error(None)
    monitor.set_jmx_string(SOLR_SAMPLE_WITH_BOGUS_BEANS)
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)

    # Stop it
    monitor.stop()
    assert_equals(0, len(reg.get_registry()))
  finally:
    shutil.rmtree(conf_dir)
    shutil.rmtree(conf_dir2)

# Solr SAMPLE with some bogus beans
SOLR_SAMPLE_WITH_BOGUS_BEANS = r"""
{
  "beans" : [ {
    "name" : "solr/mr_collection_shard2_replica1:type=core,id=mr_collection_shard2_replica1",
    "modelerType" : "org.apache.solr.core.JmxMonitoredMap$SolrDynamicMBean",
    "category" : "CORE",
    "source" : "$URL$",
    "description" : "SolrCore",
    "name" : "mr_collection_shard2_replica1",
    "version" : "1.0",
    "coreHashCode" : "756266007",
    "coreName" : "mr_collection_shard2_replica1",
    "startTime" : "Tue Nov 12 13:59:15 PST 2013",
    "refCount" : 1,
    "indexDir" : "hdfs://czhuub12-1.ent.cloudera.com:8020/solr/mr_collection/core_node4/data/index/",
    "aliases" : "[mr_collection_shard2_replica1]",
    "collection" : "mr_collection",
    "shard" : "shard2"
  }, {
    "bogusName": "** No coreHashCode UpdateHandler **",
    "name" : "No coreHashCode bean",
    "modelerType" : "org.apache.solr.core.JmxMonitoredMap$SolrDynamicMBean",
    "category" : "UPDATEHANDLER",
    "source" : "$URL$",
    "description" : "Update handler that efficiently directly updates the on-disk main lucene index",
    "version" : "1.0",
    "commits" : 0,
    "autocommits" : 0,
    "optimizes" : 0,
    "rollbacks" : 0,
    "expungeDeletes" : 0,
    "docsPending" : 0,
    "adds" : 0,
    "deletesById" : 0,
    "deletesByQuery" : 0,
    "errors" : 0,
    "cumulative_adds" : 0,
    "cumulative_deletesById" : 0,
    "cumulative_deletesByQuery" : 0,
    "cumulative_errors" : 0
  }, {
    "bogusName": "** No name UpdateHandler **",
    "coreHashCode" : "756266007",
    "modelerType" : "org.apache.solr.core.JmxMonitoredMap$SolrDynamicMBean",
    "category" : "UPDATEHANDLER",
    "source" : "$URL$",
    "description" : "Update handler that efficiently directly updates the on-disk main lucene index",
    "version" : "1.0",
    "commits" : 0,
    "autocommits" : 0,
    "optimizes" : 0,
    "rollbacks" : 0,
    "expungeDeletes" : 0,
    "docsPending" : 0,
    "adds" : 0,
    "deletesById" : 0,
    "deletesByQuery" : 0,
    "errors" : 0,
    "cumulative_adds" : 0,
    "cumulative_deletesById" : 0,
    "cumulative_deletesByQuery" : 0,
    "cumulative_errors" : 0
  }, {
    "bogusName" : "** No coreHashCode QueryHandler **",
    "name" : "solr/MissingCoreBogusQueryHandler,id=org.apache.solr.handler.component.SearchHandler",
    "modelerType" : "org.apache.solr.core.JmxMonitoredMap$SolrDynamicMBean",
    "category" : "QUERYHANDLER",
    "source" : "$URL$",
    "description" : "Search using components: query,facet,mlt,highlight,stats,spellcheck,debug,",
    "version" : "4.4.0-search-1.0.0",
    "handlerStart" : 1384293532267,
    "requests" : 0,
    "errors" : 0,
    "timeouts" : 0,
    "totalTime" : 0.0,
    "avgRequestsPerSecond" : 0.0,
    "5minRateReqsPerSecond" : 0.0,
    "15minRateReqsPerSecond" : 0.0,
    "avgTimePerRequest" : 0.0,
    "medianRequestTime" : 0.0,
    "75thPcRequestTime" : 0.0,
    "95thPcRequestTime" : 0.0,
    "99thPcRequestTime" : 0.0,
    "999thPcRequestTime" : 0.0
  }, {
    "bogusName" : "** No name QueryHandler **",
    "coreHashCode" : "756266007",
    "modelerType" : "org.apache.solr.core.JmxMonitoredMap$SolrDynamicMBean",
    "category" : "QUERYHANDLER",
    "source" : "$URL$",
    "description" : "Search using components: query,facet,mlt,highlight,stats,spellcheck,debug,",
    "version" : "4.4.0-search-1.0.0",
    "handlerStart" : 1384293532267,
    "requests" : 0,
    "errors" : 0,
    "timeouts" : 0,
    "totalTime" : 0.0,
    "avgRequestsPerSecond" : 0.0,
    "5minRateReqsPerSecond" : 0.0,
    "15minRateReqsPerSecond" : 0.0,
    "avgTimePerRequest" : 0.0,
    "medianRequestTime" : 0.0,
    "75thPcRequestTime" : 0.0,
    "95thPcRequestTime" : 0.0,
    "99thPcRequestTime" : 0.0,
    "999thPcRequestTime" : 0.0
  }, {
    "bogusName" : "** No category bean **",
    "coreHashCode" : "756266007",
    "modelerType" : "org.apache.solr.core.JmxMonitoredMap$SolrDynamicMBean",
    "source" : "$URL$",
    "description" : "Search using components: query,facet,mlt,highlight,stats,spellcheck,debug,",
    "version" : "4.4.0-search-1.0.0",
    "handlerStart" : 1384293532267,
    "requests" : 0,
    "errors" : 0,
    "timeouts" : 0,
    "totalTime" : 0.0,
    "avgRequestsPerSecond" : 0.0,
    "5minRateReqsPerSecond" : 0.0,
    "15minRateReqsPerSecond" : 0.0,
    "avgTimePerRequest" : 0.0,
    "medianRequestTime" : 0.0,
    "75thPcRequestTime" : 0.0,
    "95thPcRequestTime" : 0.0,
    "99thPcRequestTime" : 0.0,
    "999thPcRequestTime" : 0.0
  } ]
}
"""

# Sample JSON response used for tests based on a request to retrieve
# the status of a Solr Core using the Solr REST API.
# The URL target for this request is:
#   'http://<host>:8983/solr/admin/cores?wt=json&action=STATUS'
# Note that this sample should be retrieved from the same Solr server
# as the one used in SAMPLE_CDH5
#
SAMPLE_CORE_STATUS = r"""
{
  "responseHeader": {
    "status": 0,
    "QTime": 58
  },
  "defaultCoreName": "collection1",
  "initFailures": {},
  "status": {
    "hbase_collection_shard2_replica1": {
      "name": "hbase_collection_shard2_replica1",
      "isDefaultCore": false,
      "instanceDir": "/var/lib/solr/hbase_collection_shard2_replica1/",
      "dataDir": "hdfs://abs-1.ent.cloudera.com:8020/solr/hbase_collection/core_node3/data/",
      "config": "solrconfig.xml",
      "schema": "schema.xml",
      "startTime": "2013-12-17T19:29:31.145Z",
      "uptime": 168003957,
      "index": {
        "numDocs": 0,
        "maxDoc": 0,
        "deletedDocs": 0,
        "version": 1,
        "segmentCount": 0,
        "current": true,
        "hasDeletions": false,
        "directory": "org.apache.lucene.store.NRTCachingDirectory:NRTCachingDirectory(org.apache.solr.store.hdfs.HdfsDirectory@2a193b49 lockFactory=org.apache.solr.store.hdfs.HdfsLockFactory@4cf689c6; maxCacheMB=192.0 maxMergeSizeMB=16.0)",
        "userData": {},
        "sizeInBytes": 45,
        "size": "45 bytes"
      }
    },
    "mr_collection_shard2_replica1": {
      "name": "mr_collection_shard2_replica1",
      "isDefaultCore": false,
      "instanceDir": "/var/lib/solr/mr_collection_shard2_replica1/",
      "dataDir": "hdfs://abs-1.ent.cloudera.com:8020/solr/mr_collection/core_node2/data/",
      "config": "solrconfig.xml",
      "schema": "schema.xml",
      "startTime": "2013-12-17T19:29:31.143Z",
      "uptime": 168003985,
      "index": {
        "numDocs": 1022,
        "maxDoc": 1022,
        "deletedDocs": 0,
        "version": 3,
        "segmentCount": 1,
        "current": true,
        "hasDeletions": false,
        "directory": "org.apache.lucene.store.NRTCachingDirectory:NRTCachingDirectory(org.apache.solr.store.hdfs.HdfsDirectory@7dbef6b3 lockFactory=org.apache.solr.store.hdfs.HdfsLockFactory@1e4c301d; maxCacheMB=192.0 maxMergeSizeMB=16.0)",
        "userData": {
          "commitTimeMSec": "1386896828942"
        },
        "lastModified": "2013-12-13T01:07:08.942Z",
        "sizeInBytes": 130991,
        "size": "127.92 KB"
      }
    },
    "simple_collection_shard2_replica2": {
      "name": "simple_collection_shard2_replica2",
      "isDefaultCore": false,
      "instanceDir": "/var/lib/solr/simple_collection_shard2_replica2/",
      "dataDir": "hdfs://abs-1.ent.cloudera.com:8020/solr/simple_collection/core_node1/data/",
      "config": "solrconfig.xml",
      "schema": "schema.xml",
      "startTime": "2013-12-17T19:29:31.144Z",
      "uptime": 168004007,
      "index": {
        "numDocs": 0,
        "maxDoc": 0,
        "deletedDocs": 0,
        "version": 1,
        "segmentCount": 0,
        "current": true,
        "hasDeletions": false,
        "directory": "org.apache.lucene.store.NRTCachingDirectory:NRTCachingDirectory(org.apache.solr.store.hdfs.HdfsDirectory@5e834b36 lockFactory=org.apache.solr.store.hdfs.HdfsLockFactory@580ab3c7; maxCacheMB=192.0 maxMergeSizeMB=16.0)",
        "userData": {},
        "sizeInBytes": 145,
        "size": "145 bytes"
      }
    }
  }
}
"""
