# Copyright (c) 2012 Cloudera, Inc. All rights reserved.

import os
import shutil
try:
  from cStringIO import StringIO
except:
  from StringIO import StringIO

from cmf.monitor import constants, schema
from cmf.monitor.generic import GenericMonitor
from cmf.monitor.generic.impala_adapters import _ImpalaAdapter
from cmf.monitor.generic.impala_adapters import StateStoreAdapter
from cmf.safety_valve import SafetyValve
from cmf_test.monitor import make_config_file, MockAgent, TestMonitorDaemon

from nose.tools import assert_true, assert_equals, assert_false

SERVICE_RELEASE_CDH52 = "5.2.0"
SERVICE_RELEASE_CDH54 = "5.4.0"
SERVICE_RELEASE_CDH55 = "5.5.0"

class MockOpenUrl(object):
  def __init__(self, response):
    self._response = response
    self._error = None
    self._url = None
    self._username = None
    self._password = None

  def __call__(self, url, username, password):
    self._url = url
    self._username = username
    self._password = password

    if self._error:
      raise self._error
    return StringIO(self._response)

  @property
  def url(self):
    return self._url

  @property
  def username(self):
    return self._username

  @property
  def password(self):
    return self._password

  def set_response(self, response):
    self._response = response

  def set_throw_error(self, error):
    self._error = error

def _make_cdh5_format_config_file(version, role_name, release=None):
    properties = """[statestore]
role_name = %s
service_name = my_impala
service_version = %d
host = localhost
webserver_port = 1234
log_dir = /path/to/logs
collect_interval = 1.0
statestore_webserver_port = 9091
"""
    return make_config_file(properties % (role_name,
                                          version))

def _make_cdh52_format_config_file(version, role_name, release):
    properties = """[statestore]
role_name = %s
service_name = my_impala
service_version = %d
service_release = %s
host = localhost
webserver_port = 1234
log_dir = /path/to/logs
collect_interval = 1.0
statestore_webserver_port = 9091
"""
    return make_config_file(properties % (role_name,
                                          version,
                                          release))

def _make_cdh54_format_config_file(version, role_name, release):
    properties = """[statestore]
role_name = %s
service_name = my_impala
service_version = %d
service_release = %s
host = localhost
webserver_port = 1234
log_dir = /path/to/logs
collect_interval = 1.0
statestore_webserver_port = 9091
"""
    return make_config_file(properties % (role_name,
                                          version,
                                          release))

def _make_cdh55_format_config_file(version, role_name, release):
    properties = """[statestore]
role_name = %s
service_name = my_impala
service_version = %d
service_release = %s
host = localhost
webserver_port = 1234
log_dir = /path/to/logs
collect_interval = 1.0
statestore_webserver_port = 9091
"""
    return make_config_file(properties % (role_name,
                                          version,
                                          release))

def _make_garbage_format_config_file(version, role_name, release):
    properties = """[statestore]
role_name = %s
service_name = my_impala
service_version = %d
service_release = %s
host = localhost
webserver_port = 1234
log_dir = /path/to/logs
collect_interval = 1.0
statestore_webserver_port = 9091
"""
    return make_config_file(properties % (role_name,
                                          version,
                                          release))

def test_state_store_lifecycle_cdh5():
  do_state_store_lifecycle(constants.SERVICE_VERSION_CDH5,
                           _make_cdh5_format_config_file,
                           SAMPLE_CDH5)

def test_state_store_lifecycle_cdh52():
  do_state_store_lifecycle(constants.SERVICE_VERSION_CDH5,
                           _make_cdh52_format_config_file,
                           SAMPLE_CDH5,
                           SERVICE_RELEASE_CDH52)

def test_state_store_lifecycle_cdh54_new_url():
  do_state_store_lifecycle(constants.SERVICE_VERSION_CDH5,
                           _make_cdh54_format_config_file,
                           SAMPLE_CDH54,
                           SERVICE_RELEASE_CDH54,
                           True)

def test_state_store_lifecycle_cdh55():
  do_state_store_lifecycle(constants.SERVICE_VERSION_CDH5,
                           _make_cdh54_format_config_file,
                           SAMPLE_CDH54,
                           SERVICE_RELEASE_CDH55,
                           True)

def test_state_store_lifecycle_garbage_version():
  do_state_store_lifecycle(constants.SERVICE_VERSION_CDH5,
                           _make_garbage_format_config_file,
                           SAMPLE_CDH5,
                           "zxd41")

def do_state_store_lifecycle(version,
                             make_config_file_function,
                             json_sample,
                             service_release=None,
                             use_new_url=False):
  daemon = TestMonitorDaemon()
  reg = daemon.get_registry_for_testing()
  adapter = StateStoreAdapter(SafetyValve())
  monitor = GenericMonitor("IMPALA", "STATESTORE", daemon, reg, adapter)
  mock_open_url = MockOpenUrl(json_sample);
  monitor._url_collector._urlopen_callout = mock_open_url

  monitor.refresh_pid(os.getpid(), True)

  # Make two different configs
  conf_dir = make_config_file_function(version, "statestore-test", service_release)
  conf_dir2 = make_config_file_function(version, "statestore", service_release)

  try:
    assert_equals(0, len(reg.get_registry()))

    # Start it
    monitor.refresh(conf_dir)
    assert_equals(monitor, reg.get_registry().pop())
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    assert_equals('statestore-test', role_update['rolename'])
    if use_new_url:
      assert_true("?json" in monitor._url_collector._metrics_uri)

    # Test with bogus release formats
    assert_false(adapter._use_new_metrics_url("2.x.4"))
    assert_false(adapter._use_new_metrics_url("a.4"))
    assert_false(adapter._use_new_metrics_url("5"))
    assert_false(adapter._use_new_metrics_url(""))

    # Refresh with a different config
    monitor.refresh(conf_dir2)
    assert_equals(monitor, reg.get_registry().pop())
    assert_equals(1, len(reg.get_registry()))
    monitor.collect()

    # Test metrics and role update
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()

    assert_equals('statestore', role_update['rolename'])
    assert_equals('my_impala', role_update['service_name'])
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_true(metric_dict[schema.TOTAL_CPU_SYSTEM] >= 0.0)
    assert_true(metric_dict[schema.MEM_RSS] >= 0.0)
    assert_true(metric_dict[schema.MEM_VIRTUAL] >= 0.0)
    assert_equals(constants.WEB_METRICS_COLLECTION_GOOD,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)

    version_string = constants.VERSION_STRINGS[version]
    for metric_id in daemon.metric_accessors['STATESTORE'][version_string].iterkeys():
      assert_equals(type(float(1.0)), type(metric_dict[metric_id]))

    dir_metrics = role_update['dirUpdates']
    assert_true(len(dir_metrics) == 1, "We configured 1 dirs: %s" % dir_metrics)
    assert_true(os.path.ismount(dir_metrics[0]['mountpoint']))
    assert_equals('/path/to/logs', dir_metrics[0]['path'])

    assert_true(schema.UNEXPECTED_EXITS_COUNT  in metric_dict)

    # Test when the process is not running
    monitor.refresh_pid(None)
    monitor.collect()

    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()

    assert_equals('statestore', role_update['rolename'])
    assert_equals('my_impala', role_update['service_name'])
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']

    dir_metrics = role_update['dirUpdates']
    assert_true(len(dir_metrics) == 1, "We configured 1 dirs: %s" % dir_metrics)
    assert_true(os.path.ismount(dir_metrics[0]['mountpoint']))
    assert_equals('/path/to/logs', dir_metrics[0]['path'])

    # Test with a collection error
    monitor.refresh_pid(os.getpid(), True)
    mock_open_url.set_throw_error(IOError("BOOM!"))
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_true(metric_dict[schema.MEM_RSS] >= 0.0)
    assert_true(metric_dict[schema.MEM_VIRTUAL] >= 0.0)
    assert_equals(constants.WEB_METRICS_COLLECTION_COMMUNICATION_FAILURE,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)

    # Test with some invalid JSON
    mock_open_url.set_throw_error(None)
    mock_open_url.set_response("THIS IS BOGUS JSON!")
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_true(metric_dict[schema.MEM_RSS] >= 0.0)
    assert_true(metric_dict[schema.MEM_VIRTUAL] >= 0.0)
    assert_equals(constants.WEB_METRICS_COLLECTION_PARSE_FAILURE,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)

    # Stop it
    monitor.stop()
    assert_equals(0, len(reg.get_registry()))

  finally:
    shutil.rmtree(conf_dir)
    shutil.rmtree(conf_dir2)

SAMPLE_CDH5 = r"""
{
   "impala.thrift-server.StatestoreService.connections-in-use":5,
   "impala.thrift-server.StatestoreService.total-connections":5,
   "statestore.live-backends":5,
   "statestore.live-backends.list":[
      "nightly-5.ent.cloudera.com:22000",
      "nightly-5.ent.cloudera.com:26000",
      "nightly-6.ent.cloudera.com:22000",
      "nightly-7.ent.cloudera.com:22000",
      "nightly-8.ent.cloudera.com:22000"
   ],
   "statestore.total-key-size-bytes":185,
   "statestore.total-topic-size-bytes":661,
   "statestore.total-value-size-bytes":476,
   "statestore.version":"statestored version 1.2.2-INTERNAL RELEASE (build 4161d79734b48bcbff5b47a7c9e09d206cc4b848)",
   "subscriber.client-cache.clients-in-use":0,
   "subscriber.client-cache.total-clients":5,
   "tcmalloc.bytes-in-use":1259984,
   "tcmalloc.pageheap-free-bytes":876544,
   "tcmalloc.pageheap-unmapped-bytes":0,
   "tcmalloc.physical-bytes-reserved":3145728,
   "tcmalloc.total-bytes-reserved":3145728,
   "thread-manager.running-threads":16,
   "thread-manager.total-threads-created":16
}
"""

SAMPLE_CDH54 = r"""
{
    "__common__": {
        "process-name": "statestored",
        "navbar": [
            {
                "link": "/",
                "title": "/"
            },
            {
                "link": "/logs",
                "title": "/logs"
            },
            {
                "link": "/memz",
                "title": "/memz"
            },
            {
                "link": "/metrics",
                "title": "/metrics"
            },
            {
                "link": "/rpcz",
                "title": "/rpcz"
            },
            {
                "link": "/subscribers",
                "title": "/subscribers"
            },
            {
                "link": "/threadz",
                "title": "/threadz"
            },
            {
                "link": "/topics",
                "title": "/topics"
            },
            {
                "link": "/varz",
                "title": "/varz"
            }
        ]
    },
    "impala.thrift-server.StatestoreService.connections-in-use": 4,
    "impala.thrift-server.StatestoreService.total-connections": 4,
    "rpc-method.statestore.StatestoreService.RegisterSubscriber.call_duration": {
        "count": 4,
        "last": 0,
        "min": 0,
        "max": 1,
        "mean": 0.25,
        "stddev": 0.433013
    },
    "statestore.keep-alive-durations": {
        "count": 151360,
        "last": 0.000732935,
        "min": 0,
        "max": 0.522203,
        "mean": 0.00054856,
        "stddev": 0.00142062
    },
    "statestore.live-backends": 4,
    "statestore.live-backends.list": [
        "catalog-server@nightly-1.ent.cloudera.com:26000",
        "impalad@nightly-2.ent.cloudera.com:22000",
        "impalad@nightly-3.ent.cloudera.com:22000",
        "impalad@nightly-4.ent.cloudera.com:22000"
    ],
    "statestore.topic-update-durations": {
        "count": 75705,
        "last": 0.0007875,
        "min": 0,
        "max": 0.445605,
        "mean": 0.00064599,
        "stddev": 0.0027781
    },
    "statestore.total-key-size-bytes": 223,
    "statestore.total-topic-size-bytes": 4587,
    "statestore.total-value-size-bytes": 4364,
    "statestore.version": "statestored version 2.2.0-cdh5-INTERNAL RELEASE (build bc14b72450a2093d40fff156918048d3865c379c)",
    "subscriber.client-cache.clients-in-use": 0,
    "subscriber.client-cache.total-clients": 8,
    "tcmalloc.bytes-in-use": 333960,
    "tcmalloc.pageheap-free-bytes": 0,
    "tcmalloc.pageheap-unmapped-bytes": 1015808,
    "tcmalloc.physical-bytes-reserved": 1081344,
    "tcmalloc.total-bytes-reserved": 2097152,
    "thread-manager.running-threads": 25,
    "thread-manager.total-threads-created": 25
}
"""

