# Copyright (c) 2012 Cloudera, Inc. All rights reserved.
import os
import random
import shutil
import sys
import tempfile

from nose.tools import assert_equal
from cmf.supervisor_listener import Event, SupervisordEventMonitor

def create_events():
  seed = random.randint(0, sys.maxint)
  print "create_events: RNG seed = %d" % seed
  random.seed(seed)

  cnt = random.randint(10, 100)
  events = [ ]

  # Add a malformed event (no len, bad key:val pairs). Code should parse
  # it as a single event with no payload.
  events.append("bad_event foo:bar baz\n")

  # Add a random number of randomly-sized events.
  for i in range(cnt):
    length = random.randint(32, 512)
    data = "event:%d len:%d\n" % (i, length)
    data += ("0" * length)
    events.append(data)

  # Add an event with a property payload
  data = "event:real len:44\n"
  data += "from_state:RUNNING processname:foo pid:28968"
  events.append(data)

  return events


def test_parser():
  """
  Test supervisor_listener.Event's data parser.

  Feed random-sized events in random-sized chunks to the event
  parsing code and make sure events are correctly parsed.
  """
  events = create_events()
  data = "".join(events)
  curr = Event()
  count = 0

  while data:
    chunk_len = min(random.randint(16, 128), len(data))
    rem = curr.feed(data[:chunk_len])
    if curr.needed_data() == 0:
      count += 1
      curr = Event()
    data = data[chunk_len - len(rem):]

  assert_equal(len(events), count)


def test_io():
  """
  Test supervisor_listener.SupervisordEventMonitor I/O handling.

  Feed random-sized events in random-sized chunks to the supervisord
  monitor through a named pipe.
  """
  dirPath = tempfile.mkdtemp()
  fifo = -1
  monitor = None
  try:
    fifoPath = os.path.join(dirPath, 'fifo')
    os.mkfifo(fifoPath, 0600)

    monitor = SupervisordEventMonitor(fifoPath)
    events = create_events()
    data = "".join(events)

    written_bytes = 0
    fifo = os.open(fifoPath, os.O_WRONLY | os.O_SYNC)
    while data:
      chunk_len = min(random.randint(16, 128), len(data))
      os.write(fifo, data[:chunk_len])
      data = data[chunk_len:]

      written_bytes += chunk_len
      monitor.handle_read()

    # Make sure all data is really read.
    while monitor.bytes_read < written_bytes:
      monitor.handle_read()

    consumed_events = monitor.consume_events()
    assert_equal(len(events), len(consumed_events))

    # The last consumed event - should be the one created with a
    # property payload
    last = consumed_events[len(consumed_events) - 1]
    assert last is not None
    assert last.payload_properties["from_state"] == "RUNNING"
    assert last.payload_properties["processname"] == "foo"
    assert last.payload_properties["pid"] == "28968"
  finally:
    if fifo >= 0:
      os.close(fifo)
    if monitor:
      monitor.close()
    shutil.rmtree(dirPath)
