#!/usr/bin/env python
# Copyright (c) 2013 Cloudera, Inc. All rights reserved.

import os
import shutil
import subprocess
import stat
import tempfile
import time

from mock import patch, call, Mock, MagicMock

from nose.tools import assert_equal
from nose.tools import assert_false
from nose.tools import assert_raises
from nose.tools import assert_true

from cmf.cgroups import SPECIAL_MEMORY_VALUES, CGroupManager

# A CGroupManager that doesn't mount anything.
class TestCGroupManager(CGroupManager):
  ORPHANED_CGROUPS_REMOVAL_WAIT_TIME_SECS = 1
  EPSILON_SLEEP_SECS = 0.1
  def __init__(self, path, subsystems=[], sudo=None):
    CGroupManager.__init__(self, path, sudo=sudo)
    # Make copies so that the caller can safely mutate 'subsystems'.
    self._capable_subsystems = list(subsystems)
    self._mounted_subsystems = list(subsystems)
    self._existing_subsystems = []
    self.umounted_paths = []
    self._agent_cgroups = {}
    for s in subsystems:
      self._agent_cgroups[s] = ""

  def _probe(self):
    return None

  def _do_umount(self, path):
    self.umounted_paths.append(path)
    return True

# A TestCgroupManager that tracks paths actually created and destroyed.
class TrackingTestCGroupManager(TestCGroupManager):
  def __init__(self, path, subsystems=[]):
    TestCGroupManager.__init__(self, path, subsystems)
    self.destroyed = []
    self.created = []

  def _create_group(self, path, user=None, group=None, mode=None, name=None, type=None):
    self.created.append(path)

  def _destroy_group(self, path, defer=True):
    self.destroyed.append(path)

@patch.object(TestCGroupManager, "_do_mount")
def test_enable(mount_method):
  """Test the enable() method"""
  mount_method.return_value = True

  cgm = TestCGroupManager("/blah", ["foo", "bar", "baz"])
  cgm._mounted_subsystems = []
  cgm._existing_subsystems = { "baz": "/sys/fs/cgroup/baz" }

  cgm.enable()

  assert_equal(["foo", "bar"], cgm._mounted_subsystems)
  assert_equal(2, mount_method.call_count)

@patch.object(TestCGroupManager, "_do_mount")
def test_enable_with_nothing_to_do(mount_method):
  """Test the enable() method when nothing needs to be mounted"""
  mount_method.return_value = True

  cgm = TestCGroupManager("/blah", ["foo", "bar", "baz"])
  cgm._mounted_subsystems = ["foo", "bar"]
  cgm._existing_subsystems = { "baz": "/sys/fs/cgroup/baz" }

  cgm.enable()

  assert_equal(0, mount_method.call_count)

@patch.object(TestCGroupManager, "_find_existing_subsystems")
def test_disable(find_method):
  """Test the disable() method"""
  find_method.return_value = { "cpu": "/blah/cpu", "blkio": "/blah/blkio",
                               "devices": "/sys/fs/cgroup/foobish" }
  cgm = TestCGroupManager("/blah", ["cpu", "blkio", "memory"])
  cgm.disable()
  assert_equal(["/blah/blkio", "/blah/cpu"], cgm.umounted_paths)

def test_disable_with_nothing_to_do():
  """Test the disable() method when nothing needs to be unmounted"""
  cgm = TestCGroupManager("/blah", ["cpu", "blkio", "memory"])
  cgm._mounted_subsystems = []
  cgm.disable()
  assert_equal([], cgm.umounted_paths)

def test_memory_conversions():
  """Test cgroup<->CM memory value conversions"""
  cgm = TestCGroupManager("doesn't matter")

  for special_value in SPECIAL_MEMORY_VALUES:
    assert_equal(-1, cgm._memory_cgroup_to_cm(special_value))
  assert_equal(0, cgm._memory_cgroup_to_cm(0))
  assert_equal(1177, cgm._memory_cgroup_to_cm(1234567890))
  assert_raises(AssertionError, cgm._memory_cgroup_to_cm, -1)

  assert_equal(-1, cgm._memory_cm_to_cgroup(-1))
  assert_equal(0, cgm._memory_cm_to_cgroup(0))
  assert_equal(1073741824, cgm._memory_cm_to_cgroup(1024))

def test_get_group_paths():
  """Test the get_group_paths function"""
  tmpdir = tempfile.mkdtemp()
  try:
    mounted_subsystems = ["a", "b", "c"]
    name = "foo"
    cgm = TestCGroupManager(tmpdir, mounted_subsystems)
    cgm._existing_subsystems = { "e": "/bar/baz/e", "f": "/baz/bar/f"}
    cgm._capable_subsystems.extend(["d", "e", "f"]) # but not mounted
    cgm._agent_cgroups.update({ "d": "cgroup1", "e": "cgroup1", "f": "cgroup1"})

    assert_equal(set([os.path.join(subsys, "cgroup1", name) for subsys in cgm._existing_subsystems.values()] +
                     [os.path.join(tmpdir, subsys, name) for subsys in mounted_subsystems]),
                 set(cgm.get_group_paths(name)))
    assert_equal(set([os.path.join(subsys, "cgroup1", name) for subsys in cgm._existing_subsystems.values()] +
                     [os.path.join(tmpdir, subsys, name) for subsys in mounted_subsystems + ["d/cgroup1"]]),
                 set(cgm.get_group_paths(name, mounted=False)))
    assert_equal([os.path.join(tmpdir, "a", name)],
                 cgm.get_group_paths(name, subsystem="a"))
    assert_equal([],
                 cgm.get_group_paths(name, subsystem="d"))
    assert_equal([os.path.join(tmpdir, "d", "cgroup1", name)],
                 cgm.get_group_paths(name, subsystem="d", mounted=False))
    assert_equal([os.path.join("/bar/baz/e", "cgroup1", "foo")],
                 cgm.get_group_paths(name, subsystem="e"))
    assert_equal([os.path.join("/bar/baz/e", "cgroup1", "foo")],
                 cgm.get_group_paths(name, subsystem="e", mounted=False))
    assert_raises(AssertionError, cgm.get_group_paths, "/stuff")
  finally:
    shutil.rmtree(tmpdir)

def test_find_group_path():
  """Test the find_group_path function"""
  tmpdir = tempfile.mkdtemp()
  try:
    mounted_subsystems = ["a", "c"]
    cgm = TestCGroupManager(tmpdir, mounted_subsystems)
    cgm._agent_cgroups = { "a": "cgroup1", "c": "cgroup2"}
    name = "foo"
    paths = [os.path.join(tmpdir, subsys, cgm._agent_cgroups[subsys], name) for subsys in mounted_subsystems]

    assert_equal(os.path.join(tmpdir, "a", "cgroup1", name),
                 cgm.find_group_path(paths, "a", name))
    assert_equal(None, cgm.find_group_path(paths, "a", "oops"))
    assert_equal(None, cgm.find_group_path(paths, "b", name))
    assert_equal(None, cgm.find_group_path(paths, "b", "oops"))
  finally:
    shutil.rmtree(tmpdir)


def test_get_mounts():
  """Test the get_mounts function"""
  tmpdir = tempfile.mkdtemp()
  try:
    mounted_subsystems = ["a", "b", "c"]
    cgm = TestCGroupManager(tmpdir, mounted_subsystems)
    cgm._existing_subsystems = { "d": "/foo/bar/d", "e": "/bar/foo/e" }

    # No dict comprehensions in python 2.4.
    results = dict()
    for subsys in mounted_subsystems:
      results[subsys] = os.path.join(tmpdir, subsys)
    results.update(cgm._existing_subsystems)

    assert_equal(results, cgm.get_mounts())
  finally:
    shutil.rmtree(tmpdir)

def test_has_at_least_one_group():
  """Test the has_at_least_one_group function"""
  cgm = TestCGroupManager("doesn't matter")
  assert_false(cgm.has_at_least_one_group())

  cgm._ref_counts["path"] = 1
  assert_true(cgm.has_at_least_one_group())

  del cgm._ref_counts["path"]
  assert_false(cgm.has_at_least_one_group())

  cgm._orphaned_groups.append("path")
  assert_true(cgm.has_at_least_one_group())

  cgm._orphaned_groups.remove("path")
  assert_false(cgm.has_at_least_one_group())

def test_destroy_orphaned_groups():
  """Test the destroy_orphaned_groups function"""

  cgm = TrackingTestCGroupManager("doesn't matter")

  cgm.destroy_orphaned_groups()
  assert_equal([], cgm.destroyed)
  assert_equal([], cgm._orphaned_groups)

  cgm._defer_group_removal("/a")
  cgm._defer_group_removal("/b")
  # There is a wait time before orphaned cgroups
  # can be removed.
  cgm.destroy_orphaned_groups()
  assert_equal([], cgm.destroyed)
  assert_equal(["/a", "/b"], cgm._orphaned_groups)

  sleep_secs = cgm.ORPHANED_CGROUPS_REMOVAL_WAIT_TIME_SECS +\
               cgm.EPSILON_SLEEP_SECS

  time.sleep(sleep_secs)
  cgm.destroy_orphaned_groups()
  assert_equal(["/a", "/b"], cgm.destroyed)
  assert_equal([], cgm._orphaned_groups)

  cgm._defer_group_removal("/c")
  time.sleep(sleep_secs)
  cgm.destroy_orphaned_groups()
  assert_equal(["/a", "/b", "/c"], cgm.destroyed)
  assert_equal([], cgm._orphaned_groups)

def test_actual_create_group():
  """Test cgroup creation mechanics"""
  tmpdir = tempfile.mkdtemp()
  try:
    cgm = TestCGroupManager(tmpdir)

    p1 = os.path.join(tmpdir, "p1")
    cgm._create_group(p1)
    assert_true(os.path.exists(p1))

    cgm._create_group(p1) # no-op
    assert_true(os.path.exists(p1))

    # Can't reliably test user/group functionality, but can test mode.
    p2 = os.path.join(tmpdir, "p2")
    cgm._create_group(p2, mode=0777)
    assert_true(os.path.exists(p2))
    assert_equal(0777, stat.S_IMODE(os.stat(p2).st_mode))
  finally:
    shutil.rmtree(tmpdir)

def test_actual_destroy_group():
  """Test cgroup deletion mechanics"""
  tmpdir = tempfile.mkdtemp()
  try:
    cgm = TestCGroupManager(tmpdir)

    # Create a/b/c, with no permissions to delete c (or even list b).
    a = os.path.join(tmpdir, "a")
    b = os.path.join(a, "b")
    c = os.path.join(b, "c")
    os.makedirs(c)
    os.chmod(b, 0000)
    try:
      assert_equal([], cgm._orphaned_groups)
      cgm._destroy_group(a) # will fail to delete a/b, but will delete c.
      assert_true(os.path.exists(a))
      assert_true(os.path.exists(b))
      assert_false(os.path.exists(c))

      assert_equal([], cgm._orphaned_groups)
      cgm._destroy_group(a, defer=True) # fails again, orphans
      assert_true(os.path.exists(a))
      assert_true(os.path.exists(b))
      assert_false(os.path.exists(c))
      assert_equal([a], cgm._orphaned_groups)
      cgm._orphaned_groups.remove(a)
    finally:
      # So that we can delete 'a'.
      os.chmod(b, 0777)

    # Now we can delete 'a', deleting all subdirectories too. Doing it
    # again is a no-op.
    for _ in xrange(2):
      cgm._destroy_group(a)
      assert_equal([], cgm._orphaned_groups)
      assert_false(os.path.exists(a))

  finally:
    shutil.rmtree(tmpdir)

def test_cgroup_ref_counts():
  """Test cgroup referencing counting"""
  tmpdir = tempfile.mkdtemp()
  try:
    p1 = os.path.join(tmpdir, "subsys", "n1")
    p2 = os.path.join(tmpdir, "subsys", "n2")
    cgm = TrackingTestCGroupManager(tmpdir, ["subsys"])
    assert_equal({}, cgm._ref_counts)
    assert_equal([], cgm.created)
    assert_equal([], cgm.destroyed)

    cgm.create_group(p1)
    assert_equal({p1 : 1}, cgm._ref_counts)
    assert_equal([p1], cgm.created)
    assert_equal([], cgm.destroyed)

    cgm.create_group(p2)
    assert_equal({p1 : 1, p2 : 1}, cgm._ref_counts)
    assert_equal([p1, p2], cgm.created)
    assert_equal([], cgm.destroyed)

    cgm.create_group(p1)
    assert_equal({p1 : 2, p2 : 1}, cgm._ref_counts)
    assert_equal([p1, p2], cgm.created)
    assert_equal([], cgm.destroyed)

    cgm.destroy_group(p1)
    assert_equal({p1 : 1, p2 : 1}, cgm._ref_counts)
    assert_equal([p1, p2], cgm.created)
    assert_equal([], cgm.destroyed)

    cgm.destroy_group(p1)
    assert_equal({p2 : 1}, cgm._ref_counts)
    assert_equal([p1, p2], cgm.created)
    assert_equal([p1], cgm.destroyed)

    cgm.destroy_group(p2)
    assert_equal({}, cgm._ref_counts)
    assert_equal([p1, p2], cgm.created)
    assert_equal([p1, p2], cgm.destroyed)

    assert_raises(AssertionError, cgm.destroy_group, p1)
    assert_raises(Exception, cgm.create_group, "/does/not/exist")
  finally:
    shutil.rmtree(tmpdir)

@patch("grp.getgrgid")
@patch("pwd.getpwuid")
@patch("os.getgid")
@patch("os.getuid")
@patch("subprocess.Popen")
def test_mount(popen, getuid, getgid, getpwuid, getgrgid):
  """Test cgroup fs mounting with and without sudo"""

  # Test with no sudo
  popen_success = Mock()
  popen_success.communicate.return_value = (None, None)
  popen_success.returncode = 0
  popen_failure = Mock()
  popen_failure.communicate.return_value = (None, None)
  popen_failure.returncode = 1
  popen.side_effect = [popen_success, popen_success, popen_success, popen_success, popen_failure]
  getuid.return_value = 101
  getgid.return_value = 102
  getpwuid.return_value = ["user"]
  getgrgid.return_value = ["group"]
  cgm = TestCGroupManager("/blah", ["subsys"])
  cgm._do_umount = Mock()
  calls = []

  cgm._do_mount("subsys", "/blah")
  assert_true(popen.called)
  calls.append(call(["mount", "-t", "cgroup", "-o", "subsys", "cm_cgroups", "/blah"], stderr=subprocess.PIPE))
  popen.assert_has_calls(calls)
  assert_equal(0, len(cgm.umounted_paths))

  # Test with sudo
  cgm = TestCGroupManager("/blah", ["subsys"], sudo="sudo")
  cgm._do_mount("subsys", "/blah")
  assert_true(popen.called)
  calls.append(call(["sudo", "mount", "-t", "cgroup", "-o", "subsys", "cm_cgroups", "/blah"], stderr=subprocess.PIPE))
  calls.append(call(["sudo", "chown", "-R", "user:group", "/blah"], stderr=subprocess.PIPE))
  popen.assert_has_calls(calls)

  # Test with sudo, but the chown will fail so we should unmount
  cgm = TestCGroupManager("/blah", ["subsys"], sudo="sudo")
  cgm._do_mount("subsys", "/blah")
  assert_true(popen.called)
  calls.append(call(["sudo", "mount", "-t", "cgroup", "-o", "subsys", "cm_cgroups", "/blah"], stderr=subprocess.PIPE))
  calls.append(call(["sudo", "chown", "-R", "user:group", "/blah"], stderr=subprocess.PIPE))
  popen.assert_has_calls(calls)
  assert_equal(1, len(cgm.umounted_paths))
  assert_equal("/blah", cgm.umounted_paths[0])

@patch("__builtin__.open")
def test_find_existing_subsystems(mock_open):
  """Test finding existing subsystems"""
  # Mocking iterators properly requires a MagicMock.
  # http://programmaticallyspeaking.com/mocking-__iter__-with-a-magic-mock-in-python.html
  file_mock = MagicMock()
  file_mock.__iter__.return_value = ["cgroup /sys/fs/cgroup/foo cgroup a 0 0\n",
                                     "cgroup /sys/fs/cgroup/bar cgroup b 0 0\n",
                                     "cgroup /sys/fs/cgroup/baz cgroup c 0 0\n"]
  mock_open.return_value = file_mock

  cgm = TestCGroupManager("/blah")
  mounts = cgm._find_existing_subsystems(["a", "b", "d"])

  expected = { "a": "/sys/fs/cgroup/foo", "b": "/sys/fs/cgroup/bar"}
  assert_equal(expected, mounts)

def test_check_path():
  """Test the logic that checks if a cgroup path is valid"""
  cgm = TestCGroupManager("/blah", ["foo", "bar"])
  cgm._existing_subsystems = { "baz": "/sys/fs/cgroup/baz"}
  cgm._check_path("/blah/foo/cgroup1")
  cgm._check_path("/blah/bar/cgroup2")
  cgm._check_path("/sys/fs/cgroup/baz/cgroup3")
  assert_raises(Exception, cgm._check_path, "/invalid/path/to/cgroup")

@patch("os.path.exists")
@patch("__builtin__.open")
def test_find_agent_cgroups(mock_open, mock_exists):
  """Test finding the Agent's cgroups"""
  file_mock = MagicMock()
  file_mock.__iter__.return_value = ["10:memory:/user.slice/user-1000.slice/session-c1.scope\n",
                                     "9:freezer:/user.slice/user-1000.slice/session-c1.scope\n",
                                     "8:cpuset:/user.slice/user-1000.slice/session-c1.scope\n",
                                     "7:blkio:/user.slice/user-1000.slice/session-c1.scope\n",
                                     "6:perf_event:/user.slice/user-1000.slice/session-c1.scope\n",
                                     "5:cpu,cpuacct:/user.slice/user-1000.slice/session-c1.scope\n",
                                     "4:net_cls,net_prio:/user.slice/user-1000.slice/session-c1.scope\n",
                                     "3:devices:/user.slice/user-1000.slice/session-c1.scope\n",
                                     "2:hugetlb:/user.slice/user-1000.slice/session-c1.scope\n",
                                     "1:name=systemd:/user.slice/user-1000.slice/session-c1.scope\n" ]
  mock_open.return_value = file_mock

  mock_exists.return_value = True

  cgm = TestCGroupManager("/blah")
  cgroups = cgm._find_agent_cgroups(["cpu", "cpuacct", "memory", "blkio"])
  expected = { "cpu": "user.slice/user-1000.slice/session-c1.scope",
               "cpuacct": "user.slice/user-1000.slice/session-c1.scope",
               "memory": "user.slice/user-1000.slice/session-c1.scope",
               "blkio": "user.slice/user-1000.slice/session-c1.scope" }
  assert_equal(expected, cgroups)

  mock_exists.return_value = False
  cgroups = cgm._find_agent_cgroups(["cpu", "cpuacct", "memory", "blkio"])
  assert_equal({}, cgroups)

@patch("os.path.exists")
@patch("__builtin__.open")
def test_find_agent_cgroups_override(mock_open, mock_exists):
  """Test finding the Agent's cgroups with overrides"""
  file_mock = MagicMock()
  file_mock.__iter__.return_value = ["10:memory:/user.slice/user-1000.slice/session-c1.scope\n",
                                     "9:freezer:/user.slice/user-1000.slice/session-c1.scope\n",
                                     "8:cpuset:/user.slice/user-1000.slice/session-c1.scope\n",
                                     "7:blkio:/user.slice/user-1000.slice/session-c1.scope\n",
                                     "6:perf_event:/user.slice/user-1000.slice/session-c1.scope\n",
                                     "5:cpu,cpuacct:/user.slice/user-1000.slice/session-c1.scope\n",
                                     "4:net_cls,net_prio:/user.slice/user-1000.slice/session-c1.scope\n",
                                     "3:devices:/user.slice/user-1000.slice/session-c1.scope\n",
                                     "2:hugetlb:/user.slice/user-1000.slice/session-c1.scope\n",
                                     "1:name=systemd:/user.slice/user-1000.slice/session-c1.scope\n" ]
  mock_open.return_value = file_mock

  mock_exists.return_value = True

  cgm = TestCGroupManager("/blah")

  cgroup_paths_overrides = dict()
  cgroup_paths_overrides["cpu"] = "/override"
  cgroup_paths_overrides["cpuacct"] = "/override"
  cgroup_paths_overrides["memory"] = None

  cgroups = cgm._find_agent_cgroups(["cpu", "cpuacct", "memory", "blkio"], cgroup_paths_overrides)
  expected = { "cpu": "override",
               "cpuacct": "override",
               "memory": "user.slice/user-1000.slice/session-c1.scope",
               "blkio": "user.slice/user-1000.slice/session-c1.scope" }
  assert_equal(expected, cgroups)

  mock_exists.return_value = False
  cgroups = cgm._find_agent_cgroups(["cpu", "cpuacct", "memory", "blkio"])
  assert_equal({}, cgroups)
