# Copyright (c) 2014 Cloudera, Inc. All rights reserved.

import logging
import getpass
import os
import shutil
import simplejson as json
import stat
import tempfile

from StringIO import StringIO
from mock import MagicMock
from nose.tools import assert_equal
from nose.tools import assert_false
from nose.tools import assert_not_equal
from nose.tools import assert_raises
from nose.tools import assert_true
from zipfile import ZipFile

from cmf.client_configs import GENERATION_FILENAME
from cmf.client_configs import METADATA_FILENAME
from cmf.client_configs import NO_GENERATION_FOUND
from cmf.client_configs import ClientConfigKey
from cmf.client_configs import ClientConfigValue
from cmf.client_configs import ClientConfigs
from cmf.util.os_ops import OSOps

logging.basicConfig(level=logging.DEBUG)

FIXTURE_PATH = os.path.join(os.path.dirname(__file__),
                            "test-fixtures",
                            "client-configs-fixture")

class MockAgent(object):
  """
  Mocks the agent class.
  Only contains very basic features required for the test.
  """
  def __init__(self):
    self.hostname = "hostname"
    self.directories = []
    self.os_ops = OSOps(False)

  def mkabsdir(self, **kwargs):
    """"""
    self.directories.append(Directory(kwargs['path'], kwargs['user'],
                                kwargs['group'], kwargs['mode']))

class Directory(object):
  def __init__(self, path=None, user=None, group=None, mode=None):
    self.path = path
    self.user = user
    self.group = group
    self.mode = mode

class TestClientConfigs(ClientConfigs):
  """
  Test class for client config which uses mock agent.
  """
  def __init__(self):
    self.mock_agent = MockAgent()
    ClientConfigs.__init__(self, agent=self.mock_agent)

def test_cc_keys_values():
  k1 = ClientConfigKey("foo", "/foo/d1")
  k2 = ClientConfigKey("foo", "/foo/d2")
  k3 = ClientConfigKey("bar", "/bar/d2")
  k4 = ClientConfigKey("", "/bar/d2")
  k5 = ClientConfigKey("", "/bar/d2") # same as k4
  assert_equal(k4, k5)
  all_keys = set([k1, k2, k3, k4, k5])
  assert_equal(4, len(all_keys))
  assert_true(k1 in all_keys)
  assert_true(k2 in all_keys)
  assert_true(k3 in all_keys)
  assert_true(k4 in all_keys)

  assert_equal("foo", k1.name)
  assert_equal("/foo/d1", k1.path)

  v1 = ClientConfigValue(50, 0)
  v2 = ClientConfigValue(90, 1)
  v3 = ClientConfigValue(90, 7)

  assert_equal(50, v1.priority)
  assert_equal(None, v1.stdout_tail)
  assert_equal(None, v1.stderr_tail)
  assert_equal(None, v1.exit_code)
  assert_equal(None, v1.parcels_in_use)

  def unhashability():
    { v1 : "foo" }

  assert_raises(TypeError, unhashability)

  all_entries = { k1 : v1, k2 : v2, k3 : v3 }
  assert_equal(3, len(all_entries))

  # Test data loaded from disk.
  temp_deploy_dir = tempfile.mkdtemp()
  temp_conf_dir = tempfile.mkdtemp()

  try:
    # Create fake metadata file.
    deploy_result = open(os.path.join(temp_deploy_dir, METADATA_FILENAME), "w")
    try:
      deploy_result_metadata = {}
      deploy_result_metadata["exit_code"] = 123
      json.dump(deploy_result_metadata, deploy_result, indent=2, sort_keys=True)
    finally:
      deploy_result.close()

    pm = open(os.path.join(temp_conf_dir, METADATA_FILENAME), "w")
    try:
      json.dump({
        "parcels_in_use" : { "foo" : "bar" }
      }, pm, indent=2, sort_keys=True)
    finally:
      pm.close()

    # Create fake logs.
    os.mkdir(os.path.join(temp_deploy_dir, "logs"))
    for log in ["stdout.log", "stderr.log"]:
      f = open(os.path.join(temp_deploy_dir, "logs", log), "w")
      try:
        f.write(log)
      finally:
        f.close()

    v1.load(temp_deploy_dir, temp_conf_dir)
    assert_equal("stdout.log", v1.stdout_tail)
    assert_equal("stderr.log", v1.stderr_tail)
    assert_equal(123, v1.exit_code)
    assert_equal({ "foo" : "bar" }, v1.parcels_in_use)
  finally:
    shutil.rmtree(temp_deploy_dir)
    shutil.rmtree(temp_conf_dir)

def test_create_deploy_dirs():
  def run_test(client_configs, name, path, generation, expected_deploy):
    """
    Tests the effects of _create_deploy_dirs().
    """
    deploy, logs = client_configs._create_and_backup_deploy_dirs(name, path, generation)
    assert_equal(os.path.join(root, expected_deploy), deploy)
    assert_equal(os.path.join(root, expected_deploy, "logs"), logs)
    assert_true(os.path.isdir(deploy))
    assert_true(os.path.isdir(logs))

  root = tempfile.mkdtemp()
  try:
    client_configs = ClientConfigs(agent=MockAgent(),
                                   deploy_dir=root)

    # First run.
    run_test(client_configs, "foo", "bar", 0, "ccdeploy_foo_bar_0")

    # Run it again, testing that existing files in the directory are pruned.
    test_filename = os.path.join(root, "ccdeploy_foo_bar_0", "file")
    open(test_filename, "w").close()
    assert_true(os.path.exists(test_filename))
    run_test(client_configs, "foo", "bar", 0, "ccdeploy_foo_bar_0")
    assert_false(os.path.exists(test_filename))

    # Run with a different generation. Should create new directories.
    run_test(client_configs, "foo", "bar", 1, "ccdeploy_foo_bar_1")

    # Run with a name similar to what we'll find in production.
    run_test(client_configs,
             "hadoop-conf",
             "/etc/hadoop/conf.cloudera.yarn1", 0,
             "ccdeploy_hadoop-conf_etchadoopconf.cloudera.yarn1_0")

  finally:
    shutil.rmtree(root)

def test_adapt_cc_to_env():
  client_configs = ClientConfigs(agent=MockAgent())
  template = {"foo" : "bar",
              "UPDATE_ALTERNATIVES" : client_configs.bin_path}

  # Simple test. Also ensure that 'template' wasn't modified.
  cc = {}
  expected_env = template.copy()
  assert_equal(( expected_env, {} ),
               client_configs._adapt_cc_to_env(template, cc))
  assert_equal(expected_env, template)

  # Complex test: generation is dropped and values are stringified.
  cc = {
    "generation" : 1,
    "cdh_version" : 5,
    "directory_name" : "hadoop-conf"
  }
  expected_env = {
    "foo" : "bar",
    "CDH_VERSION" : "5",
    "DIRECTORY_NAME" : "hadoop-conf",
    "UPDATE_ALTERNATIVES" : client_configs.bin_path
  }
  assert_equal( ( expected_env, {} ),
                client_configs._adapt_cc_to_env(template, cc))

  # Test special handling of runner_args.
  cc = {
    "runner_args" : ["1", "2", "3"]
  }
  expected_env = {
    "UPDATE_ALTERNATIVES" : client_configs.bin_path,
    "RUNNER_ARGS" : "1 2 3"
  }
  assert_equal( ( expected_env, {} ),
                client_configs._adapt_cc_to_env({}, cc))

  # Test special handling of additional_env.
  cc = {
    "additional_env" : {
      "k1" : "v1",
      "k2" : "v2",
      "k3_numeric_val" : 6
    }
  }
  expected_env = {
    "UPDATE_ALTERNATIVES" : client_configs.bin_path,
    "k1" : "v1",
    "k2" : "v2",
    "k3_numeric_val" : "6"
  }
  assert_equal( ( expected_env, {} ),
                client_configs._adapt_cc_to_env({}, cc))

  # Test parcel handling (without a real parcel repo).
  cc = {
    "parcels" : { "foo" : "v1", "bar" : "v2" },
    "required_tags" : [],
    "optional_tags" : []
  }
  client_configs.parcel_repo = MagicMock()
  client_configs.parcel_repo.prepare_environment = MagicMock(
    return_value=set(["foo", "v1"])
  )
  assert_equal(
     ( {"UPDATE_ALTERNATIVES" : client_configs.bin_path},
       set(["foo", "v1"]) ),
     client_configs._adapt_cc_to_env({}, cc))
  assert_equal(1, client_configs.parcel_repo.prepare_environment.call_count)

def test_deploy_client_config():
  root = tempfile.mkdtemp()
  try:
    deploy_env = {
      # Required for cmf.util.fix_hue_home
      "CDH_HUE_HOME" : "/does/not/exist",
      # Required for cmf.util.fix_hcat_home
      "CDH_HCAT_HOME" : "/does/not/exist"
    }
    client_configs = ClientConfigs(agent=MockAgent(),
                                   deploy_dir=root,
                                   deploy_env=deploy_env)
    cc = {
      "alt_name" : "foo",
      "path" : root,
      "generation" : 1,
      "directory_name" : "foo-conf"
    }

    # This is an empty zip archive with a single foo-conf subdirectory, so
    # we can effectively test the generation writing logic.
    f = open(os.path.join(FIXTURE_PATH, "foo-clientconfig.zip"), "r")
    try:
      cc["data"] = f.read()
    finally:
      f.close()

    deploy_dir = client_configs._generate_deploy_subdir(cc["alt_name"],
                                                        cc["path"],
                                                        cc["generation"])

    # Inject a testing script to use instead of deploy-cc.sh.
    test_script = os.path.join(root, "test.sh")
    f = open(test_script, "w")
    f.write(("""
#!/bin/bash

function assert_env_var {
  local KEY=$1
  local EXPECTED_VAL=$2
  eval local ACTUAL_VAL=\$"$KEY"
  if [[ "$ACTUAL_VAL" != "$EXPECTED_VAL" ]]; then
    echo "Bad value for variable $KEY (was \'$ACTUAL_VAL\' and not \'$EXPECTED_VAL\')"
    exit 1
  fi
}

# Test the values of some expected environment variables.
assert_env_var CONF_DIR %(conf_dir)s
assert_env_var DEST_PATH %(root)s
assert_env_var DIRECTORY_NAME %(directory_name)s

# Test the generation file contents.
GEN_FILENAME="$DIRECTORY_NAME/%(gen_filename)s"
GEN=$(cat $GEN_FILENAME)
EXPECTED_GEN=%(expected_gen)s
if [[ "$GEN" != "$EXPECTED_GEN" ]]; then
  echo "Bad value for generation file $GEN_FILENAME (was \'$GEN\' and not \'$EXPECTED_GEN\')"
  exit 1
fi

# Test the permissions of the unpacked configuration data.
ACCESS=$(stat -c %%a $DIRECTORY_NAME)
EXPECTED_ACCESS=751
if [[ "$ACCESS" != $EXPECTED_ACCESS ]]; then
    echo "Bad permissions for directory $DIRECTORY_NAME (was \'$ACCESS\' and not $EXPECTED_ACCESS)"
  exit 1
fi
for FILE in $DIRECTORY_NAME/*; do
  ACCESS=$(stat -c %%a $FILE)
  EXPECTED_ACCESS=664
  if [[ "$ACCESS" != $EXPECTED_ACCESS ]]; then
    echo "Bad permissions for file $FILE (was \'$ACCESS\' and not $EXPECTED_ACCESS)"
    exit 1
  fi
done

# Write to stdout and stderr; they will be tested later.
echo to_stdout
echo to_stderr 1>&2
""" % {
  "conf_dir" : deploy_dir,
  "root" : root,
  "directory_name" : cc["directory_name"],
  "gen_filename" : GENERATION_FILENAME,
  "expected_gen" : cc["generation"]
}).strip())
    f.close()
    os.chmod(test_script, stat.S_IRWXU)
    client_configs.deploy_script = test_script

    # Simplifies testing permissions.
    old_umask = os.umask(0002)
    try:
      client_configs._deploy_client_config(cc)
    finally:
      os.umask(old_umask)

    # A phony ClientConfigValue in which we load the results.
    ccv = ClientConfigValue(None, None)
    ccv.load(deploy_dir, os.path.join(deploy_dir, 'foo-conf'))
    assert_equal(0, ccv.exit_code, "stdout: %s\nstderr: %s" % (ccv.stdout_tail,
                                                               ccv.stderr_tail))
    assert_equal("to_stdout\n", ccv.stdout_tail)
    assert_equal("to_stderr\n", ccv.stderr_tail)
    assert_equal({}, ccv.parcels_in_use)
  finally:
    shutil.rmtree(root)

def test_read_write_generations():
  root = tempfile.mkdtemp()
  try:
    client_configs = ClientConfigs(agent=MockAgent())

    # Should fail silently and return NO_GENERATION_FOUND.
    assert_equal(NO_GENERATION_FOUND, client_configs._read_generation(root))

    client_configs._write_generation(root, 0)
    assert_equal(0, client_configs._read_generation(root))

    client_configs._write_generation(root, 5)
    assert_equal(5, client_configs._read_generation(root))

    # Shouldn't raise an exception.
    client_configs._write_generation("/does/not/exist", 1)
  finally:
    shutil.rmtree(root)

def create_fully_loaded(kwargs=None):
  """
  Returns a ClientConfigs instance that's "fully loaded" with CM-managed
  client configs found in fixtures.
  """
  # Read the fixtures into lookup_results.
  lookup_results = {}
  for root, _, fnames in os.walk(FIXTURE_PATH):
    for fname in fnames:
      full_fname = os.path.join(root, fname)
      f = open(full_fname, "r")
      lookup_results[fname] = f.read()
      f.close()

  def side_effect(*args, **kwargs):
    # Laste element of "args" is the alternative name.
    return lookup_results[args[-1]]

  if not kwargs:
    kwargs = {}

  agent = MockAgent()
  agent.os_ops.alternatives.ALTERNATIVES_ARGS = ["--admindir", FIXTURE_PATH]
  agent.os_ops.alternatives.db_dir = FIXTURE_PATH
  agent.os_ops.exec_command_output = MagicMock(side_effect=side_effect)
  # Reload the deployed client configs from disk to test persistence.
  client_configs = ClientConfigs(agent=agent, **kwargs)
  client_configs.load()
  return client_configs

def test_load_and_parse():
  k1 = ClientConfigKey("hadoop-conf", "/etc/hadoop/conf.cloudera.HDFS-1")
  k2 = ClientConfigKey("hadoop-conf", "/etc/hadoop/conf.cloudera.MAPREDUCE-1")
  k3 = ClientConfigKey("hadoop-conf", "/etc/hadoop/conf.cloudera.YARN-1")
  k4 = ClientConfigKey("hbase-conf", "/etc/hbase/conf.cloudera.HBASE-1")
  k5 = ClientConfigKey("spark-conf", "/etc/spark/conf.cloudera.SPARK_ON_YARN-1")
  v1 = ClientConfigValue(90, NO_GENERATION_FOUND)
  v2 = ClientConfigValue(91, NO_GENERATION_FOUND)
  v3 = ClientConfigValue(92, NO_GENERATION_FOUND)
  v4 = ClientConfigValue(90, NO_GENERATION_FOUND)
  v5 = ClientConfigValue(51, NO_GENERATION_FOUND)
  expected_deployed = { k1 : v1, k2 : v2, k3 : v3, k4 : v4, k5 : v5 }

  # Read the client configs through load().
  client_configs = create_fully_loaded()
  assert_equal(expected_deployed, client_configs.deployed)
  assert_equal(expected_deployed, client_configs.results)

  # Now do it again by parsing individual alternatives.
  expected_parsed = {
    "hadoop-conf" : { k1 : v1, k2 : v2, k3 : v3 },
    "hbase-conf" : { k4 : v4 },
    "spark-conf" : { k5 : v5 },
  }

  osops = OSOps()
  osops.alternatives.ALTERNATIVES_ARGS = ["--altdir", FIXTURE_PATH]
  for root, _, fnames in os.walk(FIXTURE_PATH):
    for fname in fnames:
      assert_equal(expected_parsed.pop(fname, {}), client_configs._lookup_alternatives(fname))
  assert_true(len(expected_parsed) == 0)

def test_rectify():

  def run_test(client_configs, cc, expected_deploy, expected_deregister):
    """
    Tests the effects of rectify() using mocks.
    """

    pm_dir = tempfile.mkdtemp()
    try:
      # Create a fake metadata file to inject a successful exit code
      # into each deployment attempt.
      pm = open(os.path.join(pm_dir, METADATA_FILENAME), "w")
      try:
        deploy_result_metadata = {}
        deploy_result_metadata["exit_code"] = 0
        json.dump(deploy_result_metadata, pm, indent=2, sort_keys=True)
      finally:
        pm.close()

      client_configs._deploy_client_config = MagicMock(return_value=pm_dir)
      client_configs._deregister_alternative = MagicMock()
      client_configs.rectify(cc)
      assert_equal(expected_deploy,
                   client_configs._deploy_client_config.call_count)
      assert_equal(expected_deregister,
                   client_configs._deregister_alternative.call_count)
    finally:
      shutil.rmtree(pm_dir)

  client_configs = ClientConfigs(agent=MockAgent())

  cc = [{
    "path" : "/etc/hive/conf.cloudera.hive",
    "generation" : 0,
    "alt_name" : "hive-conf",
  }, {
    "path": "/etc/hbase/conf.cloudera.hbase2",
    "generation" : 0,
    "alt_name" : "hbase-conf",
  }, {
    "path": "/etc/hadoop/conf.cloudera.yarn",
    "generation" : 0,
    "alt_name" : "hadoop-conf",
  }, {
    "path": "/etc/hadoop/conf.cloudera.hdfs",
    "generation" : 0,
    "alt_name" : "hdfs-conf",
  }, {
    "path": "/etc/hbase/conf.cloudera.hbase",
    "generation" : NO_GENERATION_FOUND, # will not be deregistered
    "alt_name" : "hbase-conf",
  }, {
    "path": "/etc",
    "generation" : 0,
    "alt_name" : "" # empty alt_name; it's a cluster config
  }]

  # Add default values for fields that aren"t relevant to this test.
  for entry in cc:
    entry.update({
      "rolename" : "",
      "alt_link" : "",
      "data" : "",
      "cdh_version" : -1,
      "directory_name" : "",
      "priority" : -1,
      "parcels" : {},
      "required_tags" : [],
      "optional_tags" : [],
      "additional_env" : {},
      "runner_program" : "",
      "runner_args" : []
    })

  # First run: we deploy all client configs.
  run_test(client_configs, cc, len(cc), 0)

  # Same input; nothing should happen.
  run_test(client_configs, cc, 0, 0)

  # Two entries were updated.
  cc[1]["generation"] += 1
  cc[3]["generation"] += 1
  run_test(client_configs, cc, 2, 2)

  # Three entries were deleted, but only one was deregistered. Why?
  # 1. The last entry was a cluster config.
  # 2. The second to last entry had no generation.
  saved = cc.pop()
  cc.pop()
  cc.pop()
  run_test(client_configs, cc, 0, 1)

  # Add back one entry.
  saved["generation"] += 1
  cc.append(saved)
  run_test(client_configs, cc, 1, 0)

def test_rectify_no_input():
  client_configs = ClientConfigs(agent=MockAgent())
  client_configs._deploy_client_config = MagicMock()
  client_configs._deregister_alternative = MagicMock()
  client_configs.rectify(None)
  client_configs.rectify([])
  assert_equal(0, client_configs._deploy_client_config.call_count)
  assert_equal(0, client_configs._deregister_alternative.call_count)

def test_get_status():
  client_configs = ClientConfigs(agent=MockAgent())
  assert_equal([], client_configs.get_status())

  # This can simulate the case where the deployment directory was wiped out and
  # the agent construct the client config knowledge base from the relics of
  # the previous agent. Hence we have k1:v1 and k2:v2. But since the previous
  # deployment directory was lost but we know the previous deployment has been
  # successfully deployed (deducted from the current alternative database), we
  # should expect the deployment status exit code to be zero.
  k1 = ClientConfigKey("foo", "/p1/p2/p3")
  k2 = ClientConfigKey("bar", "/p1/p2/p3")
  v1 = ClientConfigValue(100, NO_GENERATION_FOUND)
  v2 = ClientConfigValue(100, 3)
  client_configs.results = { k1 : v1, k2 : v2 }
  expected_statuses = [{
    'generation' : 3,
    'path' : '/p1/p2/p3',
    'alt_name' : 'bar',
    'exit_code' : 0,
    'parcels_in_use' : {},
    'stdout_tail' : '',
    'stderr_tail' : ''
  }, {
    'generation' : NO_GENERATION_FOUND,
    'path' : '/p1/p2/p3',
    'alt_name' : 'foo',
    'exit_code' : 0,
    'parcels_in_use' : {},
    'stdout_tail' : '',
    'stderr_tail' : ''
  }]
  # Sort the output to ensure that self.deployed iteration order doesn't
  # affect the ordering of the results.
  assert_equal(expected_statuses, sorted(client_configs.get_status()))

  # Inject deployment results and test again.
  v1.exit_code = 0
  v1.stdout_tail = "OK out"
  v1.stderr_tail = "OK err"
  v1.parcels_in_use = set(["foo", "v1"])
  v2.exit_code = 143
  v2.stdout_tail = "ERROR out"
  v2.stderr_tail = "ERROR err"
  v2.parcels_in_use = {}
  expected_statuses[1]["exit_code"] = v1.exit_code
  expected_statuses[1]["stdout_tail"] = v1.stdout_tail
  expected_statuses[1]["stderr_tail"] = v1.stderr_tail
  expected_statuses[1]["parcels_in_use"] = v1.parcels_in_use
  expected_statuses[0]["exit_code"] = v2.exit_code
  expected_statuses[0]["stdout_tail"] = v2.stdout_tail
  expected_statuses[0]["stderr_tail"] = v2.stderr_tail
  assert_equal(expected_statuses, sorted(client_configs.get_status()))

def test_get_active_deployments():
  client_configs = ClientConfigs(agent=MockAgent(),
                                 deploy_dir="dragons")
  assert_equal([], client_configs.get_active_deployments())
  k1 = ClientConfigKey("foo", "/p1/p2/p3")
  k2 = ClientConfigKey("bar", "/p1/p2/p3")
  v1 = ClientConfigValue(100, NO_GENERATION_FOUND)
  v2 = ClientConfigValue(100, 3)
  client_configs.results = { k1 : v1, k2 : v2 }
  assert_equal(["ccdeploy_foo_p1p2p3_" + str(NO_GENERATION_FOUND),
                "ccdeploy_bar_p1p2p3_3"],
               client_configs.get_active_deployments())


def test_backup_log_files():
  try:
    root = tempfile.mkdtemp()
    deploy_env = {
      # Required for cmf.util.fix_hue_home
      "CDH_HUE_HOME" : "/does/not/exist",
      # Required for cmf.util.fix_hcat_home
      "CDH_HCAT_HOME" : "/does/not/exist"
    }
    client_configs = ClientConfigs(agent=MockAgent(),
                                   deploy_dir=root,
                                   deploy_env=deploy_env)
    cc = {
      "alt_name" : "foo",
      "path" : root,
      "generation" : 1,
      "directory_name" : "foo-conf"
    }

    # This is an empty zip archive with a single foo-conf subdirectory, so
    # we can effectively test the generation writing logic.
    f = open(os.path.join(FIXTURE_PATH, "foo-clientconfig.zip"), "r")
    try:
      cc["data"] = f.read()
    finally:
      f.close()

    deploy_dir = client_configs._generate_deploy_subdir(cc["alt_name"],
                                                        cc["path"],
                                                        cc["generation"])

    # Inject a testing script to use instead of deploy-cc.sh.
    test_script = os.path.join(root, "test.sh")
    f = open(test_script, "w")
    f.write(("""
#!/bin/bash

function assert_env_var {
  local KEY=$1
  local EXPECTED_VAL=$2
  eval local ACTUAL_VAL=\$"$KEY"
  if [[ "$ACTUAL_VAL" != "$EXPECTED_VAL" ]]; then
    echo "Bad value for variable $KEY (was \'$ACTUAL_VAL\' and not \'$EXPECTED_VAL\')"
    exit 1
  fi
}

# Test the values of some expected environment variables.
assert_env_var CONF_DIR %(conf_dir)s
assert_env_var DEST_PATH %(root)s
assert_env_var DIRECTORY_NAME %(directory_name)s

# Test the generation file contents.
GEN_FILENAME="$DIRECTORY_NAME/%(gen_filename)s"
GEN=$(cat $GEN_FILENAME)
EXPECTED_GEN=%(expected_gen)s
if [[ "$GEN" != "$EXPECTED_GEN" ]]; then
  echo "Bad value for generation file $GEN_FILENAME (was \'$GEN\' and not \'$EXPECTED_GEN\')"
  exit 1
fi

# Test the permissions of the unpacked configuration data.
ACCESS=$(stat -c %%a $DIRECTORY_NAME)
EXPECTED_ACCESS=751
if [[ "$ACCESS" != $EXPECTED_ACCESS ]]; then
    echo "Bad permissions for directory $DIRECTORY_NAME (was \'$ACCESS\' and not $EXPECTED_ACCESS)"
  exit 1
fi
for FILE in $DIRECTORY_NAME/*; do
  ACCESS=$(stat -c %%a $FILE)
  EXPECTED_ACCESS=664
  if [[ "$ACCESS" != $EXPECTED_ACCESS ]]; then
    echo "Bad permissions for file $FILE (was \'$ACCESS\' and not $EXPECTED_ACCESS)"
    exit 1
  fi
done

# Write to stdout and stderr; they will be tested later.
echo to_stdout
echo to_stderr 1>&2
""" % {
  "conf_dir" : deploy_dir,
  "root" : root,
  "directory_name" : cc["directory_name"],
  "gen_filename" : GENERATION_FILENAME,
  "expected_gen" : cc["generation"]
}).strip())
    f.close()
    os.chmod(test_script, stat.S_IRWXU)
    client_configs.deploy_script = test_script
    old_umask = os.umask(0002)

    client_configs._deploy_client_config(cc)
    base_parts = ["ccdeploy"]
    base_parts.append("foo")
    base_parts.append(root)
    base_parts.append(str(1))
    base_dir = os.path.join(root, "_".join(base_parts))
    log_subdir = os.path.join(deploy_dir, "logs")
    stdout_log = os.path.join(log_subdir, "stdout.log")
    stderr_log = os.path.join(log_subdir, "stderr.log")
    try:
      f = open(stdout_log, 'a+')
      f.write("hello")
    finally:
      f.close()
    try:
      f = open(stderr_log, 'a+')
      f.write("world")
    finally:
      f.close()

    # Deploy the same client config again. The previous log files set
    # should exist as backup.
    client_configs._deploy_client_config(cc)
    assert_true(os.path.exists(stdout_log))
    try:
      f = open(stdout_log, 'r')
      assert_equal("to_stdout\n", f.read())
    finally:
      f.close()
    assert_true(os.path.exists(stderr_log))
    try:
      f = open(stderr_log, 'r')
      assert_equal("to_stderr\n", f.read())
    finally:
      f.close()
    assert_true(os.path.exists(stderr_log + ".bak"))
    try:
      f = open(stdout_log + ".bak", 'r')
      assert_equal("to_stdout\nhello", f.read())
    finally:
      f.close()
    try:
      f = open(stderr_log + ".bak", 'r')
      assert_equal("to_stderr\nworld", f.read())
    finally:
      f.close()
  finally:
    shutil.rmtree(root)

def test_end_to_end():
  root = tempfile.mkdtemp()

  try:
    client_configs = create_fully_loaded({
      'deploy_dir' : root,
      'deploy_env' : {
        # Required for cmf.util.fix_hue_home
        "CDH_HUE_HOME" : "/does/not/exist",
        # Required for cmf.util.fix_hcat_home
        "CDH_HCAT_HOME" : "/does/not/exist"
      },
      'package_dir' : os.path.join(os.path.dirname(__file__),
                                   "..", "..", "service"),
    })

    # Inject a fake generation into the loaded client configs so that
    # rectify() deletes them properly.
    fake_gen = 0
    for value in client_configs.deployed.itervalues():
      value.generation = fake_gen

    ccs = [{
      "alt_link" : "/etc/hadoop/conf",
      "alt_name" : "hadoop-conf",
      "generation" : fake_gen,
      "path" : "/etc/hadoop/conf.cloudera.HDFS-1",
      "priority" : 90,
    }, {
      "alt_link" : "/etc/hadoop/conf",
      "alt_name" : "hadoop-conf",
      "generation" : fake_gen,
      "path" : "/etc/hadoop/conf.cloudera.MAPREDUCE-1",
      "priority" : 91,
    }, {
      "alt_link" : "/etc/hadoop/conf",
      "alt_name" : "hadoop-conf",
      "generation" : fake_gen,
      "path" : "/etc/hadoop/conf.cloudera.YARN-1",
      "priority" : 92,
    }, {
      "alt_link" : "/etc/hbase/conf",
      "alt_name" : "hbase-conf",
      "generation" : fake_gen,
      "path": "/etc/hbase/conf.cloudera.HBASE-1",
      "priority" : 90,
    }, {
      "alt_link" : "/etc/spark/conf",
      "alt_name" : "spark-conf",
      "generation" : fake_gen,
      "path": "/etc/spark/conf.cloudera.SPARK_ON_YARN-1",
      "priority" : 51,
    }]

    # Add default values for fields that aren"t relevant to this test.
    for entry in ccs:
      entry.update({
        "rolename" : "",
        "data" : "",
        "cdh_version" : -1,
        "directory_name" : "",
        "parcels" : {},
        "required_tags" : [],
        "optional_tags" : [],
        "additional_env" : {},
        "runner_program" : "",
        "runner_args" : []
      })

    # First call should be a no-op.
    non_mocked_deploy = client_configs._deploy_client_config
    client_configs._deploy_client_config = MagicMock()
    client_configs._deregister_alternative = MagicMock()
    client_configs.rectify(ccs)
    assert_equal(0, client_configs._deploy_client_config.call_count)
    assert_equal(0, client_configs._deregister_alternative.call_count)

    # Drop all client configs.
    num_ccs = len(ccs)
    ccs = []
    client_configs.rectify(ccs)
    assert_equal(0, client_configs._deploy_client_config.call_count)
    assert_equal(num_ccs, client_configs._deregister_alternative.call_count)

    # Make a fake /var/lib/alternatives.
    scratch_db = os.path.join(root, "var", "lib", "alternatives")
    os.makedirs(scratch_db)

    # Make a fake /etc/alternatives.
    scratch_alts = os.path.join(root, "etc", "alternatives")
    os.makedirs(scratch_alts)

    # Load a real client config archive.
    f = open(os.path.join(FIXTURE_PATH, "mapreduce-1-clientconfig.zip"), "r")
    try:
      configuration_data = f.read()
    finally:
      f.close()

    alt_link = os.path.join(root, "etc", "hadoop", "conf")
    path = os.path.join(root, "etc", "hadoop", "conf.cloudera.MAPREDUCE-1")

    # Do a real deploy for one client config.
    #
    # We inject several environment variable so that deploy-cc.sh doesn't
    # step all over the actual system:
    # 1. TEST_USER to avoid chowning something to root.
    # 2. TEST_UA_FLAGS to avoid modifying the real alternatives database.
    # 3. PATH so that it can find update-alternatives no matter where it is.
    cc = {
      "alt_link" : alt_link,
      "alt_name" : "hadoop-conf",
      "data" : configuration_data,
      "generation" : fake_gen,
      "path": path,
      "priority" : 90,
      "rolename" : "",
      "cdh_version" : 5,
      "directory_name" : "hadoop-conf",
      "parcels" : {},
      "required_tags" : [],
      "optional_tags" : [],
      "additional_env" : {
        "TEST_USER" : getpass.getuser(),
        "TEST_UA_FLAGS" : "--admindir %s --altdir %s" % (scratch_db,
                                                         scratch_alts),
        "PATH" : ":".join(["/usr/sbin", os.getenv("PATH")])
      },
      "runner_program" : "",
      "runner_args" : []
    }
    ccs = [cc]
    client_configs._deploy_client_config = non_mocked_deploy
    client_configs.rectify(ccs)

    # Check the status.
    status = client_configs.get_status()
    assert_equal(1, len(status))
    assert_equal(cc["alt_name"], status[0]["alt_name"])
    assert_equal(cc["generation"], status[0]["generation"])
    assert_equal(cc["path"], status[0]["path"])
    stdout = status[0]["stdout_tail"]
    stderr = status[0]["stderr_tail"]
    assert_not_equal("", stdout)
    assert_not_equal("", stderr)
    assert_equal(0, status[0]["exit_code"],
                 "stdout: %s\nstderr: %s" % (stdout, stderr))
    assert_equal({}, status[0]["parcels_in_use"])

    # Check the various changes to disk.

    assert_true(os.path.islink(alt_link))
    assert_equal(os.path.realpath(path), os.path.realpath(alt_link))

    assert_true(os.path.isdir(path))
    deployed_files = set(os.listdir(path))
    zipped_files = set(os.path.basename(f)
                       for f in ZipFile(StringIO(configuration_data)).namelist())
    assert_true(zipped_files.issubset(deployed_files))
    assert_true(GENERATION_FILENAME in deployed_files)

    new_alt = os.path.join(scratch_alts, cc["alt_name"])
    assert_true(os.path.islink(new_alt))
    assert_equal(os.path.realpath(path), os.path.realpath(new_alt))

    new_db_entry = os.path.join(scratch_db, cc["alt_name"])
    assert_true(os.path.isfile(new_db_entry))
    f = open(new_db_entry, "r")
    try:
      # Cope with inconsistent trailing newlines.
      new_db_entry_contents = f.read().strip()
    finally:
      f.close()

    assert_equal("auto\n%s\n\n%s\n%d" % (alt_link, path, cc["priority"]),
                 new_db_entry_contents)

    agent = MockAgent()
    agent.os_ops.alternatives.ALTERNATIVES_ARGS = ["--altdir", scratch_alts, "--admindir", scratch_db]
    agent.os_ops.alternatives.db_dir = scratch_db
    # Reload the deployed client configs from disk to test persistence.
    client_configs = ClientConfigs(agent=agent, deploy_dir=root)
    client_configs.load()
    statuses = client_configs.get_status()
    assert_equal(1, len(statuses))
    status = statuses[0]
    assert_equal(cc["alt_name"], status["alt_name"])
    assert_equal(cc["path"], status["path"])
    assert_equal(cc["generation"], status["generation"])
    assert_true(len(status["stdout_tail"]) > 50)
    assert_true(len(status["stderr_tail"]) > 50)
    assert_equal(0, status["exit_code"])
    assert_equal({}, status["parcels_in_use"])

    # Ensure that this deployment is active.
    assert_equal([os.path.basename(
      client_configs._generate_deploy_subdir(cc["alt_name"],
                                             cc["path"],
                                             cc["generation"]))],
                  client_configs.get_active_deployments())
  finally:
    shutil.rmtree(root)

def test_resources_dirs():
  try:
    client_configs = TestClientConfigs()
    user, group = "test_user", "test_group"
    dir1 = tempfile.mkdtemp()
    dir1_mode = 0700
    cc = {
      "alt_name" : "foo",
      "path" : dir1,
      "generation" : 1,
      "directory_name" : "foo-conf",
      "resources" : [{
        "directory" : {
          "path" : dir1,
          "user" : user,
          "group" : group,
          "mode" : dir1_mode,
          "bytes_free_warning_threshhold_bytes" : 0
        }}
      ]
    }

    # Create resources
    client_configs.evaluate_resources(cc["resources"])
    # Validate resources
    assert_equal(1, len(client_configs.mock_agent.directories))
    actual_path = client_configs.mock_agent.directories[0]
    assert_equal(dir1, actual_path.path)
    assert_equal(user, actual_path.user)
    assert_equal(group, actual_path.group)
    assert_equal(dir1_mode, actual_path.mode)

    # Now create a new directory with different permission level
    dir2_mode = 0500
    dir2 = os.path.join(dir1, "resource2")
    cc["resources"][0]["directory"]["path"] = dir2
    cc["resources"][0]["directory"]["mode"] = dir2_mode
    client_configs.evaluate_resources(cc["resources"])

    # Validate leaf directory
    assert_equal(2, len(client_configs.mock_agent.directories))
    actual_leaf_path = client_configs.mock_agent.directories[1]
    assert_equal(dir2, actual_leaf_path.path)
    assert_equal(user, actual_leaf_path.user)
    assert_equal(group, actual_leaf_path.group)
    assert_equal(dir2_mode, actual_leaf_path.mode)

  finally:
    shutil.rmtree(dir1)