#!/usr/bin/env python
# Copyright (c) 2012 Cloudera, Inc. All rights reserved.

import os
import random
import shutil
import sys
import tempfile
import simplejson as json
from datetime import datetime
import logging

from nose.tools import assert_true, assert_equal, assert_false, assert_raises, assert_not_equal

from cmf.util import unzip
from collect_host_stats import archive_directory

from clusterstats.client.stats import StatChainRunner, CommandStat, StreamingCommandStat, ProcStdoutStderrStat, HOST_COMMAND_TERMINATED
from clusterstats.client.fs_daos import MachineCollectionFsDao

LOG = logging.getLogger(__name__)
FIXTURE_PATH = os.path.join(os.path.dirname(__file__), "test-fixtures", "collect-host-stats-fixture")

def test_archive_dir():
  """tests that directories are properly archived"""
  tmpdir = None
  try:
    tmpdir = tempfile.mkdtemp()
    archive_path = os.path.join(tmpdir, "collect-host-stats-fixture.zip")
    archive_directory(FIXTURE_PATH, archive_path)
    fd = open(archive_path)
    try:
      unzip(fd.read(), tmpdir)
    finally:
      fd.close()

    extracted_dir_path = os.path.join(tmpdir, os.path.basename(FIXTURE_PATH))
    assert_true(os.path.exists(os.path.join(extracted_dir_path)))
    assert_dirs_equal(FIXTURE_PATH, extracted_dir_path)
  finally:
    shutil.rmtree(tmpdir)

def test_charsets():
  """tests that machine stats collector handles non-unicode characters"""
  tmpdir = None
  try:
    tmpdir = tempfile.mkdtemp()
    extract_dir = os.path.join(tmpdir, "extract")
    os.mkdir(extract_dir)
    machine_collection_dao = MachineCollectionFsDao(extract_dir)
    def persist_sys_stat(stat_key, stat_result):
      machine_collection_dao.put_sys_stat(
          stat_key, stat_result.get_result_data(),
          stat_result.get_error_data(), stat_result.get_other())

    create_file_with_content(os.path.join(tmpdir, "file1"), "Watch out - \xef")
    stat_runner = StatChainRunner.get_builder(). \
                    set_persister(persist_sys_stat). \
                    add("cat", CommandStat(["cat", os.path.join(tmpdir, "file1")])). \
                    build()
    stat_runner.run()
    assert_true(os.path.exists(os.path.join(extract_dir, "sysstats", "cat_stdout")))
  finally:
    shutil.rmtree(tmpdir)

def test_performance():
  """tests that performance file will be generated"""
  tmpdir = None
  try:
    tmpdir = tempfile.mkdtemp()
    extract_dir = os.path.join(tmpdir, "extract")
    os.mkdir(extract_dir)
    machine_collection_dao = MachineCollectionFsDao(extract_dir)
    machine_collection_dao.initialize_performance()
    def persist_sys_stat(stat_key, stat_result):
      machine_collection_dao.put_sys_stat(
          stat_key, stat_result.get_result_data(),
          stat_result.get_error_data(), stat_result.get_other())
      machine_collection_dao.update_performance(stat_key)

    create_file_with_content(os.path.join(tmpdir, "file1"), "A File")
    stat_runner = StatChainRunner.get_builder(). \
                    set_persister(persist_sys_stat). \
                    add("cat", CommandStat(["cat", os.path.join(tmpdir, "file1")])). \
                    build()
    stat_runner.run()
    machine_collection_dao.complete_performance()
    performance_file_name = os.path.join(extract_dir, "sysstats", "performance.json")
    assert_true(os.path.exists(performance_file_name))
    perf_file = open(performance_file_name, 'r')
    try:
      perf_data = json.load(perf_file)
      assert_equal(3, len(perf_data))
      assert_equal("start", perf_data[0]['command'])
      assert_equal("cat", perf_data[1]['command'])
      assert_equal("end", perf_data[2]['command'])
    finally:
      perf_file.close()
  finally:
    shutil.rmtree(tmpdir)

def test_proc_stdout_stderr_handle_overflow():
  """tests ProcStdoutStderrStat._handle_overflow"""
  stat = ProcStdoutStderrStat()

  main_list = [1, 2, 3, 4]
  overflow_list = [5, 6, 7, 8, 9, 10]

  # No overflow.
  ml, ol = stat._handle_overflow(main_list, overflow_list, 4)
  assert_equal(main_list, ml)
  assert_equal(overflow_list, ol)

  # Overflow, with overflow list supplied.
  ml, ol = stat._handle_overflow(main_list, overflow_list, 3)
  assert_equal([1, 2, 3], ml)
  assert_equal([4, 5, 6, 7, 8, 9, 10], ol)

  # Overflow, no overflow list supplied.
  ml, ol = stat._handle_overflow(main_list, None, 1)
  assert_equal([1], ml)
  assert_equal(None, ol)

class TestStreamingCommandStat(StreamingCommandStat):
  def __init__(self, num_stdout_lines, num_stderr_lines):
    # Run python against this script as a subprocess.
    super(TestStreamingCommandStat, self).__init__(
        [sys.executable, os.path.realpath(__file__),
        str(num_stdout_lines), str(num_stderr_lines)])

    self.actual_stdout_lines = 0
    self.actual_stderr_lines = 0

  def _get_expected_line(self, line, ch):
    expected_len = int(line[0:2])
    return line[0:2] + ch * expected_len + "\n"

  def _process_stdout_line(self, line):
    assert_equal(self._get_expected_line(line, "o"), line)
    self.actual_stdout_lines += 1
    return "O"

  def _process_stderr_line(self, line):
    assert_equal(self._get_expected_line(line, "e"), line)
    self.actual_stderr_lines += 1
    return "E"

def test_streaming_command_stat():
  """tests basic functionality of StreamingCommandStat"""
  num_stdout = 10000
  num_stderr = 7000
  stat = TestStreamingCommandStat(num_stdout, num_stderr)
  result = stat.collect(None, None, None)
  assert_equal("O" * num_stdout, result.get_result_data())
  assert_equal("E" * num_stderr, result.get_error_data())
  assert_equal(num_stdout, stat.actual_stdout_lines)
  assert_equal(num_stderr, stat.actual_stderr_lines)

def _test_streaming_command_stat_proc(num_stdout_lines, num_stderr_lines):
  def _build_line(ch):
    line_len = random.randint(50, 99)
    return str(line_len) + ch * line_len + "\n"

  while num_stdout_lines > 0 or num_stderr_lines > 0:
    if num_stdout_lines > 0:
      sys.stdout.write(_build_line("o"))
      num_stdout_lines -= 1
      # Flush stdout once in a while to exercise interleaved stdout/stderr
      # processing (stderr is unbuffered).
      if random.uniform(0, 1.0) < 0.05:
        sys.stdout.flush()

    if num_stderr_lines > 0:
      sys.stderr.write(_build_line("e"))
      num_stderr_lines -= 1

def test_timeout_default():
  """tests host stats with default 60s timeout"""
  tmpdir = None
  try:
    tmpdir = _test_timeout(None, CommandStat, "test_timeout_default")
    sysstats = os.path.join(tmpdir, "sysstats")
    assert_stat(sysstats, "really_slow_runner", True)
    assert_stat(sysstats, "fast_runner", False)
    assert_stat(sysstats, "slow_runner", False)
  finally:
    if tmpdir is not None:
      shutil.rmtree(tmpdir)

def test_timeout_30s():
  """tests host stats with 30s timeout"""
  tmpdir = None
  try:
    tmpdir = _test_timeout(30, CommandStat, "test_timeout_30s")
    sysstats = os.path.join(tmpdir, "sysstats")
    assert_stat(sysstats, "really_slow_runner", True)
    assert_stat(sysstats, "fast_runner", False)
    assert_stat(sysstats, "slow_runner", True)
  finally:
    if tmpdir is not None:
      shutil.rmtree(tmpdir)

def test_streaming_timeout_default():
  """tests streaming host stats with default 60s timeout"""
  tmpdir = None
  try:
    tmpdir = _test_timeout(None, StreamingCommandStat, "test_streaming_timeout_default")
    sysstats = os.path.join(tmpdir, "sysstats")
    assert_stat(sysstats, "really_slow_runner", True)
    assert_stat(sysstats, "fast_runner", False)
    assert_stat(sysstats, "slow_runner", False)
  finally:
    if tmpdir is not None:
      shutil.rmtree(tmpdir)

def test_streaming_timeout_30s():
  """tests streaming host stats with 30s timeout"""
  tmpdir = None
  try:
    tmpdir = _test_timeout(30, StreamingCommandStat, "test_streaming_timeout_30s")
    sysstats = os.path.join(tmpdir, "sysstats")
    assert_stat(sysstats, "really_slow_runner", True)
    assert_stat(sysstats, "fast_runner", False)
    assert_stat(sysstats, "slow_runner", True)
  finally:
    if tmpdir is not None:
      shutil.rmtree(tmpdir)

def _test_timeout(timeout, cls, test):
  tmpdir = tempfile.mkdtemp()
  machine_collection_dao = MachineCollectionFsDao(tmpdir)
  machine_collection_dao.initialize_performance()
  def persist_sys_stat(stat_key, stat_result):
    machine_collection_dao.put_sys_stat(
        stat_key, stat_result.get_result_data(),
        stat_result.get_error_data(), stat_result.get_other())
    machine_collection_dao.update_performance(stat_key)
  stat_runner = StatChainRunner.get_builder(). \
                  set_persister(persist_sys_stat). \
                  set_timeout(timeout). \
                  add("really_slow_runner",  cls(["bash", "-c", "sleep 90; echo Done."])). \
                  add("fast_runner",  cls(["bash", "-c", "sleep 1; echo Done."])). \
                  add("slow_runner",  cls(["bash", "-c", "sleep 50; echo Done."])). \
                  add("kinda_slow_runner",  cls(["python", "-c", "import time; time.sleep(40); print('Done.')"])). \
                  build()
  start_time = datetime.now()
  LOG.info("[%s]: %s Start collecting" % (start_time, test))
  stat_runner.run()
  machine_collection_dao.put_sys_stats_run(stat_runner.get_run_order())
  machine_collection_dao.complete_performance()
  end_time = datetime.now()
  LOG.info("[%s]: %s Finished collecting" % (end_time, test))
  total_time = end_time - start_time
  LOG.info("%s Total time: %s" % (test, str(total_time)))
  return tmpdir

def assert_stat(sysstats, stat, terminated):
  assert_true(os.path.exists(os.path.join(sysstats, stat + "_stdout")))
  assert_true(os.path.exists(os.path.join(sysstats, stat + "_stderr")))
  assert_true(os.path.exists(os.path.join(sysstats, stat + "_retcode")))
  if terminated is False:
    assert_stdout_ends_with(os.path.join(sysstats, stat + "_stdout"), "Done.")
  else: 
    assert_stdout_ends_with(os.path.join(sysstats, stat + "_stdout"), HOST_COMMAND_TERMINATED)
  assert_retcode(os.path.join(sysstats, stat + "_retcode"), terminated) 

def assert_stdout_ends_with(stdout, expected):
  fin = open(stdout, 'r')
  try:
    line = fin.readlines()[-1]
  finally:
    fin.close()
  line = line.rstrip('\n')
  assert_equal(expected, line)

def assert_retcode(retcode, terminated):
  fin = open(retcode, 'r')
  try:
    line = fin.readlines()[-1]
  finally:
    fin.close()
  if terminated is True:
    # SIGTERM
    assert_equal("-15", line)
  else:
    assert_equal("0", line)

def create_file_with_content(path, content):
  handle = open(path, "w")
  try:
    handle.write(content)
  finally:
    handle.close()

def assert_dirs_equal(expected_dir, actual_dir):
  def store_visit_calls(arg, dirname, filenames):
    visit_calls_dict, dir_path_dirname = arg
    dirname = dirname[len(dir_path_dirname) + 1:]
    visit_calls_dict[dirname] = sorted(filenames)

  expected_calls = {}
  expected_dir_path_dirname = os.path.dirname(expected_dir)
  os.path.walk(expected_dir, store_visit_calls, (expected_calls, expected_dir_path_dirname))

  actual_calls = {}
  actual_dir_path_dirname = os.path.dirname(actual_dir)
  os.path.walk(actual_dir, store_visit_calls, (actual_calls, actual_dir_path_dirname))

  assert_equal(expected_calls, actual_calls)

if __name__== "__main__":
  assert_equal(3, len(sys.argv))
  _test_streaming_command_stat_proc(int(sys.argv[1]), int(sys.argv[2]))
