#!/usr/bin/env python
# Copyright (c) 2012 Cloudera, Inc. All rights reserved.

import filecmp
import shutil
import tarfile
import tempfile

import os
import simplejson as json
from nose.tools import assert_true, assert_equal, assert_false, with_setup, raises

from cmf.util import diff

LEFT_ONLY = "left-only"
RIGHT_ONLY = "right-only"
COMMON = "common"
DIFFERENT = "different"

NESTED_DIR = "nested"
EMPTY_DIR = "empty"
RIGHT_ONLY_EMPTY_DIR = "right-only-empty"
DELETED_DIR = "deleted"

LINKED = "linked"
LINK = "link"
CHANGED_LINKED = "actualfilechanged"
UNCHANGED_LINK = "linknochange"
DANGLING_LINK = "dangling"
DELETED_DANGLING_LINK = "badlink"
ABS_LINK = "etc"

tmp = None
left = None
right = None

def _write_file(path, content):
  file = open(path, 'w')
  file.write(content)
  file.close()

def _create_file(dir, name, content=None):
  path = os.path.join(dir, name)
  if content is None:
    content = name
  _write_file(path, content)

def _create_files(left, right):
  """Populate left/right dirs with standard set of files
  """
  os.makedirs(left)
  _create_file(left, "left-only")
  _create_file(left, "common")
  _create_file(left, "different", "left")

  os.makedirs(right)
  _create_file(right, "right-only")
  _create_file(right, "common")
  _create_file(right, "different", "right")

def before():
  global tmp
  global left
  global right
  tmp = tempfile.mkdtemp()
  left = os.path.join(tmp, "left")
  right = os.path.join(tmp, "right")

  # create standard dir with left-only, right-only, common, and different files
  _create_files(left, right)
  # create another nested pair of dirs
  left_nested = os.path.join(left, NESTED_DIR)
  right_nested = os.path.join(right, NESTED_DIR)
  _create_files(left_nested, right_nested)
  # create empty dir
  os.mkdir(os.path.join(left, EMPTY_DIR))
  os.mkdir(os.path.join(right, EMPTY_DIR))
  # create right-only empty dir
  os.mkdir(os.path.join(right, RIGHT_ONLY_EMPTY_DIR))
  # create deleted (left-only) dir
  deleted_dir = os.path.join(left, DELETED_DIR)
  os.mkdir(deleted_dir)
  _create_file(deleted_dir, "somefile")
  # create links only for left
  cwd = os.getcwd()
  _create_file(left, CHANGED_LINKED, "original")
  os.chdir(left) # for link to be relative
  try:
    os.symlink(CHANGED_LINKED, UNCHANGED_LINK)
    os.symlink(DELETED_DANGLING_LINK + "-void", DELETED_DANGLING_LINK)
  finally:
    os.chdir(cwd)
  # create links only for right
  _create_file(right, CHANGED_LINKED, "changed")
  _create_file(right, LINKED)
  cwd = os.getcwd()
  os.chdir(right) # for link to be relative
  try:
    os.symlink(LINKED, LINK)
    os.symlink(CHANGED_LINKED, UNCHANGED_LINK)
    os.symlink(DANGLING_LINK + "-1.0.0", DANGLING_LINK)
    os.symlink("/etc", ABS_LINK)
  finally:
    os.chdir(cwd)

def after():
  global tmp
  shutil.rmtree(tmp, ignore_errors=True)

@with_setup(before, after)
def test_diff():
  out = os.path.join(tmp, "out")
  diff.diff(left, right, out)
  _verify_diff(out)

def _verify_diff(out):
  _assert_test_dir(out)
  _assert_test_dir(os.path.join(out, NESTED_DIR))
  # identical empty dir shouldn't show up in diff out
  assert_false(os.path.exists(os.path.join(out, EMPTY_DIR)))
  # new empty dirs should show up
  assert_true(os.path.exists(os.path.join(out, RIGHT_ONLY_EMPTY_DIR)))
  # deleted dir should show up in meta
  deleted_file = diff.get_deleted_meta(out)
  assert_true(DELETED_DIR in _load_json_file(deleted_file))
  # verify links
  assert_true(os.path.exists(os.path.join(out, LINKED)))
  assert_true(os.path.islink(os.path.join(out, LINK)))
  assert_equal(LINKED, os.readlink(os.path.join(out, LINK)))
  # unchanged link also included in diff because its target changed
  assert_true(os.path.islink(os.path.join(out, UNCHANGED_LINK)))
  assert_true(os.path.exists(os.path.join(out, CHANGED_LINKED)))
  assert_true(os.path.islink(os.path.join(out, DANGLING_LINK)))
  assert_false(os.path.islink(os.path.join(out, DELETED_DANGLING_LINK)))
  assert_true(os.path.islink(os.path.join(out, ABS_LINK)))
  assert_equal("/etc", os.readlink(os.path.join(out, ABS_LINK)))

def _assert_test_dir(dir):
  assert_false(os.path.exists(os.path.join(dir, LEFT_ONLY)))
  assert_true(os.path.exists(os.path.join(dir, RIGHT_ONLY)))
  assert_false(os.path.exists(os.path.join(dir, COMMON)))
  assert_true(os.path.exists(os.path.join(dir, DIFFERENT)))
  deleted_meta = diff.get_deleted_meta(dir)
  assert_true(os.path.exists(deleted_meta))
  deleted = _load_json_file(deleted_meta)
  assert_true(LEFT_ONLY in deleted)

def _load_json_file(path):
  fp = open(path, 'r')
  try:
    des = json.load(fp)
    return des
  finally:
    fp.close()

@with_setup(before, after)
def test_patch():
  out = os.path.join(tmp, "out")
  diff.diff(left, right, out)

  patched = os.path.join(tmp, "patched")
  diff.patch(left, out, patched)
  # patched should completely match right
  _assert_dir_equal(right, patched)

def _assert_dir_equal(left, right):
  """Recursively assert that the contents of the directories are equal.
  """
  cmp = filecmp.dircmp(left, right)
  assert_equal(0, len(cmp.left_only))
  assert_equal(0, len(cmp.right_only))
  for file in cmp.common_files:
    assert_true(filecmp.cmp(os.path.join(left, file), os.path.join(right, file), shallow=False))

PARCEL1 = "ABC-1.0.0-rhel5"
PARCEL2 = "ABC-1.0.1-rhel5"
PARCEL2_DIR = PARCEL2[0:PARCEL2.rfind('-')]

parcel1_tar = None
parcel2_tar = None

def before_tar():
  before()
  # rename dirs to proper parcel names
  global tmp
  global left
  global right
  global parcel1_tar
  global parcel2_tar

  parcel1_tar = _tar_with_dir(left, PARCEL1, tmp)
  parcel2_tar = _tar_with_dir(right, PARCEL2, tmp)

def _tar_with_dir(src, dir_name, dst):
  filename = dir_name + ".tar.gz"
  output = os.path.join(dst, filename)
  tar = tarfile.open(output, 'w:gz')
  try:
    tar.add(src, arcname=dir_name)
    tar.close()
  except Exception, e:
    tar.close()
  return output

@with_setup(before_tar, after)
def test_tardiff():
  tar = diff.tardiff(parcel1_tar, parcel2_tar, tmp)
  assert_true(os.path.exists(tar))
  assert_equal("ABC-1.0.0-1.0.1-rhel5.diff", os.path.basename(tar))
  out = os.path.join(tmp, "diff")
  diff._untar(tar, out)
  _verify_diff(out)
  return tar

@with_setup(before_tar, after)
def test_tarpatch():
  difftar = test_tardiff()
  patched_out = tempfile.mkdtemp(dir=tmp)
  try:
    patched = diff.apply_patch(parcel1_tar, difftar, patched_out)
    _assert_dir_equal(right, patched)
    assert_equal(PARCEL2_DIR, os.path.basename(patched))
  finally:
    shutil.rmtree(patched_out, ignore_errors=True)

@with_setup(before_tar, after)
@raises(AssertionError)
def test_tarpatch_output_already_exists():
  difftar = test_tardiff()
  os.mkdir(os.path.join(tmp, PARCEL2_DIR))
  diff.apply_patch(parcel1_tar, difftar, tmp)

@with_setup(before_tar, after)
def test_dirpatch():
  difftar = test_tardiff()
  patched_out = tempfile.mkdtemp(dir=tmp)
  try:
    patched = diff.apply_patch(left, difftar, patched_out)
    _assert_dir_equal(right, patched)
    assert_equal(PARCEL2_DIR, os.path.basename(patched))
  finally:
    shutil.rmtree(patched_out, ignore_errors=True)

@with_setup(before_tar, after)
@raises(AssertionError)
def test_dirpatch_output_already_exists():
  difftar = test_tardiff()
  os.mkdir(os.path.join(tmp, PARCEL2_DIR))
  diff.apply_patch(left, difftar, tmp)
