#!/usr/bin/env python
# Copyright (c) 2012 Cloudera, Inc. All rights reserved.

import os
import shutil
import StringIO
import tarfile
import tempfile
import urllib2
import urlparse

from nose.tools import assert_true, assert_equal, assert_false, assert_raises, raises
from mock import ANY
from mock import call
from mock import Mock

from cmf import downloader
from cmf.downloader import Download
from cmf.parcel_id import ParcelId

FIXTURE_PATH = os.path.join(os.path.dirname(__file__), "test-fixtures", "parcel-repository-fixture")

PARCEL_FILENAME = "awesome-1.0-rhel.parcel"
PARCEL_URL = "https://archive.cloudera.com/" + PARCEL_FILENAME
DIFF_URL = "https://archive.cloudera.com/awesome-1.0-1.1-rhel.diff"

class HttpResponseMock:
  def __init__(self, body, status, headers):
    self.headers = headers
    self.body = StringIO.StringIO(body)
    self.status = status

  def info(self):
    return self.headers

  def read(self, length):
    return self.body.read(length)

def test_download_state_immutable_once_terminal():
  dl = downloader.Download(PARCEL_URL)
  assert_equal(downloader.PENDING, dl.state)
  dl.state = downloader.EXCEPTION
  assert_equal(downloader.EXCEPTION, dl.state)
  dl.state = downloader.PENDING
  # cannot be changed once terminal
  assert_equal(downloader.EXCEPTION, dl.state)
  
def test_http_error():
  '''Test that downloader handles http error'''

  code = 404
  class HttpErrorURLopener(downloader.URLopener):
    def fetch_url(self, request, filename=None):
      raise urllib2.HTTPError(
        request.get_full_url(),
        code,
        "Not found",
        dict(), None)

  try:
    tmp = tempfile.mkdtemp()
    dw = downloader.Downloader(tmp, 1, HttpErrorURLopener())
    dw.queue_download(PARCEL_URL)
    dw.wait()
    op = dw.ops[PARCEL_URL]
    dw.cleanUp()
    assert_equal(code, op.code)
    assert_equal(downloader.EXCEPTION, op.state)
    assert_equal(0, op.downloaded_bytes)
    assert_true(op.exception_msg is not None)
  finally:
    shutil.rmtree(tmp)

def test_exception():
  '''Test that downloader handles exception and clear out temporary file'''

  class ExceptionURLopener(downloader.URLopener):
    def fetch_url(self, request):
      raise IOError()

  tmp = tempfile.mkdtemp()
  try:
    dw = downloader.Downloader(tmp, 1, ExceptionURLopener())
    dw.queue_download(PARCEL_URL)
    dw.wait()
    op = dw.ops[PARCEL_URL]
    dw.cleanUp()
    assert_equal(downloader.UNEXPECTED_ERROR_CODE, op.code)
    assert_equal(downloader.EXCEPTION, op.state)
    assert_equal(0, op.downloaded_bytes)
    assert_true(op.exception_msg is not None)
    assert_false(os.path.exists(os.path.join(tmp, PARCEL_FILENAME)))
  finally:
    shutil.rmtree(tmp)

class FileURLopener(downloader.URLopener):
  def fetch_url(self, request):
    content = "foobar"
    headers = {'content-length': str(len(content)), }
    response = HttpResponseMock(
      status="200",
      body=content,
      headers=headers,
    )
    return response

# This class is used to emulate the situation where there is partially
# downloaded file. The URLOpener will read the HTTP response and append the
# content into the local download file.
class ResumableFileURLOpener(downloader.URLopener):
  def fetch_url_url(self, request):
    content = "bar"
    headers = {'content-length': str(len(content)), }
    response = HttpResponseMock(
      status="200",
      body=content,
      headers=headers,
    )
    return response

def _download_file(create_leftover=False):
  tmp = tempfile.mkdtemp()
  try:
    try:
      if create_leftover:
        dw = downloader.Downloader(tmp, 1, ResumableFileURLOpener())
        path = dw._get_downloaded_path(PARCEL_URL)
        file = open(path + downloader.TEMP_FILE_SUFFIX, 'w')
        file.write("foo")
        file.close()
      else:
        dw = downloader.Downloader(tmp, 1, FileURLopener())
      dw.queue_download(PARCEL_URL)
      dw.wait()
      op = dw.ops[PARCEL_URL]
      dw.cleanUp()
      assert_equal(200, op.code)
      assert_equal(downloader.COMPLETE, op.state)
      assert_equal(6, op.total_bytes)
      assert_equal(6, op.downloaded_bytes)
      assert_equal(None, op.exception_msg)

      # Verify file
      filename = dw._get_downloaded_path(PARCEL_URL)
      assert_true(os.path.exists(filename))
      file = open(filename, 'r')
      assert_equal("foobar", file.readline())
      file.close()
    except Exception, e:
      pass
  finally:
    shutil.rmtree(tmp)

def test_without_leftover_file():
  '''Test that downloader handles downloaded files'''
  _download_file()

def test_with_leftover_file():
  '''Test that downloader handles existing temp file'''
  _download_file(True)

def test_existing_file():
  tmp = tempfile.mkdtemp()
  try:
    dw = downloader.Downloader(tmp, 1, FileURLopener())
    # Create file even before download
    path = dw._get_downloaded_path(PARCEL_URL)
    file = open(path, 'w')
    file.write("boo")
    file.close()
    lmt = os.path.getmtime(path)
    
    cb = Mock()
    assert_true(dw.queue_download(PARCEL_URL, cb))
    dw.wait()
    assert_true(os.path.exists(path))
    assert_equal(lmt, os.path.getmtime(path))
    assert_true(cb.called)
    assert_equal(downloader.COMPLETE, cb.call_args[0][1].state)
    assert_equal(200, cb.call_args[0][1].code)
  finally:  
    shutil.rmtree(tmp)

def test_existing_file_with_overwrite():
  tmp = tempfile.mkdtemp()
  try:
    dw = downloader.Downloader(tmp, 1, FileURLopener())
    # Create file even before download
    path = dw._get_downloaded_path(PARCEL_URL)
    file = open(path, 'w')
    file.write("boo")
    file.close()
    
    assert_true(dw.queue_download(PARCEL_URL, overwrite=True))
    dw.wait()
    assert_true(os.path.exists(path))
    file = open(path, 'r')
    assert_equal("foobar", file.readline())
    file.close()
  finally:
    shutil.rmtree(tmp)

def test_callback():
  '''Test that callback is called after download completes'''
  _callback(downloader.COMPLETE)

def test_callback_exception():
  '''Test that callback throwing exception is handled'''
  _callback(downloader.EXCEPTION, True)

def test_callback_exception_leave_file():
  '''Test that callback throwing exception can leave file undeleted'''
  _callback(downloader.EXCEPTION, True, True)

def _callback(expected_state, exception=False, keep_file=False):
  cb = Mock()
  if exception:
    cb.side_effect = Exception("Boo!")
    if keep_file:
      cb.side_effect.keep_file = True

  tmp = tempfile.mkdtemp()
  dw = downloader.Downloader(tmp, 1, FileURLopener())
  rm = Mock()
  dw.set_remove_callout(rm)
  dw.queue_download('http://www.foobar.com', cb)
  dw.wait()
  op = dw.ops['http://www.foobar.com']
  dw.cleanUp();
  assert_equal(expected_state, op.state)
  cb.assert_called_once_with('http://www.foobar.com', op)
  if exception:
    if keep_file:
      assert_false(rm.called)
    else:
      rm.assert_called_once_with(op.path)
  shutil.rmtree(tmp)

def test_abort():
  tmp = tempfile.mkdtemp()
  opener = FileURLopener()
  dw = downloader.Downloader(tmp, 1, opener)

  dw._set_exception = Mock()
  dw.progress = Mock()
  try:
    try:
      url = "http://www.foobar.com"
      filename = tmp + dw._get_downloaded_path(url)
      tempname = filename + downloader.TEMP_FILE_SUFFIX
      assert_false(os.path.exists(tempname))
      dw.ops[url] = Download(url)
      dw.abort(url)
      op = dw.ops[url]
      dw._download(op)
    except Exception, e:
      pass
  finally:
    dw.cleanUp()
    shutil.rmtree(tmp)
    assert_equal(downloader.ABORTED, op.state)
    assert_false(dw._set_exception.called)
    assert_false(dw.progress.called)
    assert_false(os.path.exists(tempname))
    assert_equal(0, op.downloaded_bytes)

def test_prune_history():
  tmp = tempfile.mkdtemp()
  dw = downloader.Downloader(tmp, 1, Mock())
  try:
    url1 = "http://www.foo.com/foo-0.5.0-rhel.parcel"
    dw.ops[url1] = downloader.Download(url1)
    url2 = "http://www.foo.com/bar-1.0.0-rhel.parcel"
    dw.ops[url2] = downloader.Download(url2)
    dw.ops[url2].state = downloader.DOWNLOADED
    url3 = "http://www.foo.com/bar-1.0.0-1.1.0-rhel.diff"
    dw.ops[url3] = downloader.Download(url3)
    url4 = "http://www.foo.com/bar-1.1.0-rhel.parcel"
    dw.ops[url4] = downloader.Download(url4)
    dw.ops[url4].state = downloader.COMPLETE
    assert_equal(4, len(dw.ops))
    
    def filter(url):
      return url.endswith("parcel")
    dw.prune_history(filter)
    # url1 and 2 are active so cannot be pruned
    assert_equal(3, len(dw.ops))
    assert_true(url3 in dw.ops)
  finally:
    dw.wait()
    dw.cleanUp()
    shutil.rmtree(tmp)

def test_double_queue_download():
  '''Test that active downloads for same url are prevented'''
  tmp = tempfile.mkdtemp()
  dw = downloader.Downloader(tmp, 1, Mock())
  dw.threadpool.putRequest = Mock()
  parcel_url = "https://archive.cloudera.com/foo-bar-baz-boo"
  # no ongoing download
  assert_true(dw.queue_download(parcel_url))
  assert_true(dw.threadpool.putRequest.assert_called)
  dw.threadpool.putRequest.reset_mock()

  # PENDING
  dw.ops[parcel_url] = downloader.Download(parcel_url)
  assert_false(dw.queue_download(parcel_url))
  assert_false(dw.threadpool.putRequest.called)
  dw.threadpool.putRequest.reset_mock()
  
  # DOWNLOADING
  dw.ops[parcel_url].state = downloader.DOWNLOADING
  assert_false(dw.queue_download(parcel_url))
  assert_false(dw.threadpool.putRequest.called)
  dw.threadpool.putRequest.reset_mock()

  # DOWNLOADED
  dw.ops[parcel_url].state = downloader.DOWNLOADED
  assert_false(dw.queue_download(parcel_url))
  assert_false(dw.threadpool.putRequest.called)
  dw.threadpool.putRequest.reset_mock()
  
  # EXCEPTION
  dw.ops[parcel_url].state = downloader.EXCEPTION
  assert_true(dw.queue_download(parcel_url))
  assert_true(dw.threadpool.putRequest.called)
  dw.threadpool.putRequest.reset_mock()
  
  # ABORTED
  dw.ops[parcel_url].state = downloader.ABORTED
  assert_true(dw.queue_download(parcel_url))
  assert_true(dw.threadpool.putRequest.called)
  dw.threadpool.putRequest.reset_mock()

  dw.cleanUp();
  shutil.rmtree(tmp)

def test_download_end_time():
  '''Test download end time'''
  url = "http://www.foo.com/hmm"
  op = downloader.Download(url)
  op.total_bytes = 100
  op.downloaded_bytes = 79
  op.code = 0

  assert_true(op.start_time is not None)
  assert_true(op.download_end_time is None)
  assert_true(op.end_time is None)

  op.state = downloader.DOWNLOADED
  assert_true(op.download_end_time is not None)
  assert_true(op.end_time is None)

  op.state = downloader.COMPLETE
  assert_true(op.download_end_time is not None)
  assert_true(op.end_time is not None)

def test_repr():
  '''Test download repr'''
  url = "http://www.foo.com/hmm"
  op = downloader.Download(url)
  op.state = downloader.EXCEPTION
  op.total_bytes = 100
  op.downloaded_bytes = 79
  op.code = 0
  str(op)

def test_get_current_downloads():
  '''Test that downloader reports current downloads'''
  tmp = tempfile.mkdtemp()
  dw = downloader.Downloader(tmp, 1, FileURLopener())
  try:
    complete_url = "http://www.foo.com/bar"
    op = downloader.Download(complete_url)
    op.state = downloader.DOWNLOADED
    op.total_bytes = 12345
    op.downloaded_bytes = 123
    op.code = 200
    dw.ops[complete_url] = op
    dls = dw.get_current_downloads()
    assert_equal(1, len(dls))
    assert_equal(complete_url, dls[0][downloader.URL_FIELD])
    assert_equal(12345, dls[0][downloader.TOTAL_BYTES_FIELD])
    assert_equal(123, dls[0][downloader.DOWNLOADED_BYTES_FIELD])
    assert_equal(None, dls[0][downloader.ERROR_FIELD])
    dw.ops.clear()
    
    running_url = "http://www.foo.com/hmm"
    op = downloader.Download(running_url)
    op.state = downloader.DOWNLOADING
    op.total_bytes = 100
    op.downloaded_bytes = 79
    op.code = 0
    dw.ops[running_url] = op
    dls = dw.get_current_downloads()
    assert_equal(1, len(dls))
    assert_equal(running_url, dls[0][downloader.URL_FIELD])
    assert_equal(100, dls[0][downloader.TOTAL_BYTES_FIELD])
    assert_equal(79, dls[0][downloader.DOWNLOADED_BYTES_FIELD])
    assert_equal(None, dls[0][downloader.ERROR_FIELD])
    dw.ops.clear()
  
    error_url = "http://www.foo.com/ehh"
    op = downloader.Download(error_url)
    op.state = downloader.EXCEPTION
    op.exception_msg = 'foobar'
    op.total_bytes = 47
    op.downloaded_bytes = 7
    op.code = 206
    dw.ops[error_url] = op
    dls = dw.get_current_downloads()
    assert_equal(1, len(dls))
    assert_equal(error_url, dls[0][downloader.URL_FIELD])
    assert_equal(47, dls[0][downloader.TOTAL_BYTES_FIELD])
    assert_equal(7, dls[0][downloader.DOWNLOADED_BYTES_FIELD])
    assert_true(dls[0][downloader.START_TIME_FIELD] is not None)
    assert_true(dls[0][downloader.END_TIME_FIELD] is None)
    assert_equal({ downloader.ERROR_CODE_FIELD:str(op.code), 
                   downloader.ERROR_MSG_FIELD:op.exception_msg },
                 dls[0][downloader.ERROR_FIELD])
    dw.ops.clear()
  finally:
    dw.wait()
    dw.cleanUp()
    shutil.rmtree(tmp)
