#!/usr/bin/env python
# Copyright (c) 2015 Cloudera, Inc. All rights reserved.

import os
import stat

from mock import call, patch, Mock
from nose.tools import assert_true, assert_equal, assert_false

from cmf.agent import Agent
from cmf.safety_valve import SafetyValve


from os.path import isdir as test_dssd_detection_real_isdir
from os import listdir as test_dssd_detection_real_listdir
from os import stat as test_dssd_detection_real_stat

@patch("os.path.isdir")
@patch("os.listdir")
@patch("os.stat")
def test_dssd_detection(mock_stat, mock_listdir, mock_isdir):
  """
  Test detection of DSSD support on host
  """
  dssd_dev_dir = "/dev/dssd/vpci/client"
  dssd_dev_files = ["23:0", "24:0"]
  dssd_dev_paths = [os.path.join(dssd_dev_dir, f) for f in dssd_dev_files]
  bogus_dssd_dev_dir = "/dev/dssd/vpci/client/bogus"

  def do_mock_isdir(path):
    if path == dssd_dev_dir:
      return True
    elif path == bogus_dssd_dev_dir:
      return False
    else:
      return test_dssd_detection_real_isdir(path)

  def do_mock_listdir(path):
    if path == dssd_dev_dir:
      return dssd_dev_files
    elif path == bogus_dssd_dev_dir:
      assert_true(False, "os.listdir should not have been called for %s" % (path,))
    else:
      return test_dssd_detection_real_listdir(path)

  def do_mock_stat(path):
    if path in dssd_dev_paths:
      return Mock(st_mode=stat.S_IFCHR)
    elif os.path.dirname(path) == bogus_dssd_dev_dir:
      assert_true(False, "os.stat should not have been called for %s" % (path,))
    else:
      return test_dssd_detection_real_stat(path)

  mock_isdir.side_effect = do_mock_isdir
  mock_listdir.side_effect = do_mock_listdir
  mock_stat.side_effect = do_mock_stat
  dssd_dev_dir_key = "dssd_vpci_device_dir"

  def make_safety_valve(dev_dir):
    if dev_dir is None:
      return {SafetyValve._SAFETY_VALVE_KEY: ""}
    else:
      return {SafetyValve._SAFETY_VALVE_KEY: "%s=%s" % (dssd_dev_dir_key, dev_dir)}

  def assert_no_call(mock, *args):
    assert_false(call(*args) in mock.call_args_list)

  def reset_mocks():
    mock_isdir.reset_mock()
    mock_listdir.reset_mock()
    mock_stat.reset_mock()

  agent = Agent()

  # Safety valve not set: DSSD device detected in default device dir.
  assert_true(agent.supports_dssd)
  assert_equal(dssd_dev_dir, agent.cached_dssd_device_dir)
  mock_isdir.assert_any_call(dssd_dev_dir)

  # Safety valve set to default dir: uses cached result.
  reset_mocks()
  agent._update_safety_valve(make_safety_valve(dssd_dev_dir))
  assert_true(agent.supports_dssd)
  assert_equal(dssd_dev_dir, agent.cached_dssd_device_dir)
  assert_no_call(mock_isdir, dssd_dev_dir)

  # Whitespace around safety valve value is not significant.
  reset_mocks()
  agent._update_safety_valve(make_safety_valve("  %s  " % (dssd_dev_dir,)))
  assert_true(agent.supports_dssd)
  assert_equal(dssd_dev_dir, agent.cached_dssd_device_dir)
  assert_no_call(mock_isdir, dssd_dev_dir)

  # Safety valve set to invalid device dir: no DSSD devices detected.
  reset_mocks()
  agent._update_safety_valve(make_safety_valve(bogus_dssd_dev_dir))
  assert_false(agent.supports_dssd)
  assert_equal(bogus_dssd_dev_dir, agent.cached_dssd_device_dir)
  mock_isdir.assert_any_call(bogus_dssd_dev_dir)

  # Setting safety valve to empty reinstates default device dir.
  reset_mocks()
  agent._update_safety_valve(make_safety_valve(None))
  assert_true(agent.supports_dssd)
  assert_equal(dssd_dev_dir, agent.cached_dssd_device_dir)
  mock_isdir.assert_any_call(dssd_dev_dir)
