# Copyright (c) 2018 Cloudera, Inc. All rights reserved.

from nose.tools import assert_equal, assert_almost_equal, assert_true, assert_false

import host_perf_diag
import bandwidth_server_start_stop
import subprocess

def test_make_chunks():
  def helper_test_chunks(orig_list, chunk_size, expected_list):
    assert_equal(expected_list, host_perf_diag.make_chunks(orig_list, chunk_size))
  helper_test_chunks([1, 2, 3, 4, 5, 6], 3, [[1, 2, 3], [4, 5, 6]])
  helper_test_chunks([1, 2, 3, 4, 5], 3, [[1, 2, 3], [4, 5]])
  helper_test_chunks([1, 2, 3], 3, [[1, 2, 3]])
  helper_test_chunks([], 2, [[]])

def test_parse_xmit_stats_line():
  # Tuples of input line and expected packet loss output
  test_input = (("2 packets transmitted, 2 received, 30.0% packet loss, time 1001ms", 30.0),
                ("2 packets transmitted, 2 received, 30.59% packet loss, time 1001ms", 30.59),
                ("2 packets transmitted, 2 received, 0% packet loss, time 1001ms", 0),
                ("2 packets transmitted, 2 received, 100% packet loss, time 1001ms", 100))

  for test_case in test_input:
    input_line = test_case[0]
    expected_output = test_case[1]
    # using almost_equal for floating point output values
    assert_almost_equal(expected_output, host_perf_diag.parse_xmit_stats_line(input_line))


def test_parse_timing_stats_line():
  test_input = (
    ("rtt min/avg/max/mdev = 9.796/9.799/9.803/10.099 ms", 9.796, 9.799, 9.803, 10.099),
    ("rtt min/avg/max/mdev = 0/9.799/9.803/.099 ms", 0, 9.799, 9.803, 0.099),
    ("rtt min/avg/max/mdev = 19.796/0/109.803/0.099 ms", 19.796, 0, 109.803, 0.099),
    ("rtt min/avg/max/mdev = 9.796/0/9.803/0 ms", 9.796, 0, 9.803, 0))

  for test_case in test_input:
    input_line = test_case[0]
    actual_output = host_perf_diag.parse_timing_stats_line(input_line)
    # Using almost_equal for floating point output values
    # and can't directly compare dictionaries for almost_equal() with py2.6 on RHEL6.
    assert_almost_equal(test_case[1], actual_output["minLatencyMillisecs"])
    assert_almost_equal(test_case[2], actual_output["avgLatencyMillisecs"])
    assert_almost_equal(test_case[3], actual_output["maxLatencyMillisecs"])
    assert_almost_equal(test_case[4], actual_output["mdevLatencyMillisecs"])

def test_run_ping() :
  # Using public DNS servers which are expected to be available.
  # Update following list and expected_output in sync.
  hosts = ["8.8.8.8", "1.1.1.1", "foobar.foo"]
  expected_output = {hosts[0]: True, hosts[1]: True, hosts[2]: False}

  host_chunks = [hosts]
  (host_stat_list, success) = host_perf_diag.run_pings(host_chunks, count=10,
                                                       packet_size=56,
                                                       timeout_secs=10)
  assert_true(success)
  assert_equal(len(hosts), len(host_stat_list))
  for host_stat in host_stat_list:
    host = host_stat["hostName"]
    assert_equal(expected_output[host], host_stat["success"])

def test_run_bandwidth_diagnostics():
  host_stat_list = [ {"hostName" : "foo", "success" : True} ]
  bw_process = "../../src/cmf_test/mock_perf.py"
  host_perf_diag.run_bandwidth_diagnostics(host_stat_list, 10, bw_process, skip_ld_preload=True)
  assert_almost_equal(host_stat_list[0]["bwSentBitsPerSecond"], 5.02185e+08),
  assert_almost_equal(host_stat_list[0]["bwReceivedBitsPerSecond"], 5.01663e+08)


def test_kill_child_processes():
  pmap = {}
  for i in range(5):
    pmap[i] = subprocess.Popen(["sleep", "10"], stdout=subprocess.PIPE)
  host_perf_diag.kill_child_processes_on_timeout(pmap, True)
  for proc in pmap.itervalues():
    assert_true(proc.poll() is not None)

def test_bandwidth_server_stop():
  subprocess.Popen(["sleep", "10"])
  assert_true(bandwidth_server_start_stop.stop_iperf_process(proc_name="sleep"))

def test_bandwidth_server_stop_proc_not_found():
  assert_false(bandwidth_server_start_stop.stop_iperf_process(proc_name="non-existent-process"))

def test_bandwidth_server_start_proc_killing_previous():
  subprocess.Popen(["sleep", "10"])
  assert_true(
    bandwidth_server_start_stop.start_iperf_process(proc_name="sleep",
                                                    skip_ld_preload=True,
                                                    args="10",
                                                    binary_name="sleep"))
