#!/usr/bin/env python
# Copyright (c) 2013 Cloudera, Inc. All rights reserved.

import logging
import os
import shutil
import tempfile
import time

from nose.tools import assert_true, assert_equals, assert_false

from cmf import agent, kt_renewer

LOG = logging.getLogger(__name__)

class TestKtRenewer(object):
  """
  We put in a fake `kinit' binary, which audits the invocation info to a file.
  Then we verify that the calls are made correctly.

  The fake kinit takes special "keytab" values and behaves differently.
  """

  @classmethod
  def setup_class(cls):
    cls.kinit = kt_renewer.KINIT_BIN
    cls.kdestroy = kt_renewer.KDESTROY_BIN
    kt_renewer.KINIT_BIN = os.path.join(
        os.path.dirname(__file__),
        'test-fixtures',
        'source-fixture',
        'fake_kinit.sh')

    kt_renewer.KDESTROY_BIN = os.path.join(
        os.path.dirname(__file__),
        'test-fixtures',
        'source-fixture',
        'fake_kdestroy.sh')

    cls.renewal_period = kt_renewer._RENEWAL_PERIOD
    kt_renewer._RENEWAL_PERIOD = 0.1


  @classmethod
  def teardown_class(cls):
    kt_renewer.KINIT_BIN = cls.kinit
    kt_renewer.KDESTROY_BIN = cls.kdestroy
    kt_renewer._RENEWAL_PERIOD = cls.renewal_period


  def setup(self):
    """Make a new audit file for every test method"""
    self.audit_filepath = tempfile.mkstemp()[1]


  def teardown(self):
    kt_renewer.stop_cred_manager()
    os.unlink(self.audit_filepath)


  @property
  def audit_file(self):
    if os.path.isfile(self.audit_filepath):
      return file(self.audit_filepath).read().rstrip()
    return ""


  @property
  def princ(self):
    return self.audit_filepath


  def _wait_for_kinit(self, n_runs=1):
    """
    Wait for cred manager to run at least n_runs number of times,
    in order to guarantees that all those keytabs have been consumed.
    The testcase feeds keytabs to the cred manager. It's set up
    so that each run would process at least one keytab.
    """
    start = kt_renewer._get_cred_manager().num_runs
    ntrials = n_runs * 10
    for i in xrange(ntrials):
      updated = kt_renewer._get_cred_manager().num_runs
      if updated > start + n_runs:
        break
      time.sleep(kt_renewer._RENEWAL_PERIOD)


  def test_kinit(self):
    assert_true(kt_renewer._get_cred_manager().last_candidate_used is None,
                "last cadidate was initialized to %s" %
                (kt_renewer._get_cred_manager().last_candidate_used,))
    assert_false(kt_renewer._CANDIDATES_LAST_KINIT_TIMESTAMP_KEY in
                 kt_renewer._get_cred_manager()._last_kinit)
    kt_renewer.add_keytab_to_kinit_from('valid', self.princ)
    self._wait_for_kinit()

    assert_true('valid 0' in self.audit_file,
                "Called kinit; returned 0")
    assert_equals(('valid', self.princ),
                  kt_renewer._get_cred_manager().last_candidate_used,
                  'unexpected last candidate used: %s' %
                  (kt_renewer._get_cred_manager().last_candidate_used,))
    assert_true(kt_renewer._CANDIDATES_LAST_KINIT_TIMESTAMP_KEY in
                kt_renewer._get_cred_manager()._last_kinit)


  def test_invalid_ticket(self):
    assert_true(kt_renewer._get_cred_manager().last_candidate_used is None,
                "last cadidate was initialized to %s" %
                (kt_renewer._get_cred_manager().last_candidate_used,))
    assert_false(kt_renewer._CANDIDATES_LAST_KINIT_TIMESTAMP_KEY in
                 kt_renewer._get_cred_manager()._last_kinit)
    kt_renewer.add_keytab_to_kinit_from('invalid', self.princ)
    self._wait_for_kinit(2)

    assert_true('invalid 1' in self.audit_file,
                "Called kinit; returned 1")
    # We expect the renewer to retry.
    assert_true(self.audit_file.count('invalid') >= 2,
                  "Invalid ticket didn't retry: %s (%d)" %
                  (self.audit_file, self.audit_file.count('invalid')))
    # But we should not cache any ticket as we did not successfully kinit.
    assert_true(kt_renewer._get_cred_manager().last_candidate_used is None,
                "last cadidate was initialized to %s" %
                (kt_renewer._get_cred_manager().last_candidate_used,))
    assert_false(kt_renewer._CANDIDATES_LAST_KINIT_TIMESTAMP_KEY in
                 kt_renewer._get_cred_manager()._last_kinit)


  def test_kdc_offline(self):
    kt_renewer.add_keytab_to_kinit_from('kdc-offline', self.princ)
    self._wait_for_kinit()

    assert_true(self.audit_file.count('kdc-offline 1') > 1,
                "Despite error return, we kept using that ticket")
    # We should not cache any ticket as we did not successfully kinit.
    assert_true(kt_renewer._get_cred_manager().last_candidate_used is None,
                "last cadidate was initialized to %s" %
                (kt_renewer._get_cred_manager().last_candidate_used,))


  def test_invalid_tickets(self):
    """
    test that we tried to kinit with all invalid tickets.
    """
    kt_renewer._get_cred_manager()._async_kinit_multi_for_test(
        ('valid-once', self.princ),
        ('invalid', self.princ))
    self._wait_for_kinit(3)

    assert_equals(1, self.audit_file.count('valid-once 0'),
                  "First time was valid: %s" % (self.audit_file,))
    assert_true(self.audit_file.count('valid-once 1') > 1,
                "Second time was invalid; no more retry %s" %
                (self.audit_file,))
    assert_true(self.audit_file.count('invalid 1') > 1,
                "We tried the invalid ticket just once %s" %
                (self.audit_file,))
    # But we should cache the ticket that we successfully kinit-ed from once,
    # even though we failed to kinit from it later.
    assert_equals(('valid-once', self.princ),
                  kt_renewer._get_cred_manager().last_candidate_used,
                  'unexpected last candidate used: %s' %
                  (kt_renewer._get_cred_manager().last_candidate_used,))

  def test_keytab_lifecycle(self):
    kt_renewer.add_keytab_to_kinit_from('valid-for-awhile', self.princ)
    self._wait_for_kinit(3)
    assert_true('valid-for-awhile' in self.audit_file,
                "Called kinit; returned 0")
    assert_true(self.audit_file.count('valid-for-awhile') > 3,
                "valid ticket was not retried %s" %
                (self.audit_file,))
    assert_equals(('valid-for-awhile', self.princ),
                  kt_renewer._get_cred_manager().last_candidate_used,
                  'unexpected last candidate used: %s' %
                  (kt_renewer._get_cred_manager().last_candidate_used,))
    kt_renewer.remove_keytab_to_kinit_from('valid-for-awhile')
    self._wait_for_kinit()
    # We should reset the last successful keytab after we realize that it is no
    # longer valid
    assert_true(kt_renewer._get_cred_manager().last_candidate_used is None,
                "last cadidate was initialized to %s" %
                (kt_renewer._get_cred_manager().last_candidate_used,))
    # add a new one keytab, see that we try to kinit from it.
    kt_renewer.add_keytab_to_kinit_from('new-valid', self.princ)
    self._wait_for_kinit(3)
    assert_true(self.audit_file.count('new-valid') > 3,
                "valid ticket was not retried %s" %
                (self.audit_file,))
    assert_equals(('new-valid', self.princ),
                  kt_renewer._get_cred_manager().last_candidate_used,
                  'unexpected last candidate used: %s' %
                  (kt_renewer._get_cred_manager().last_candidate_used,))

  def test_stop_after_valid_found(self):
    kt_renewer.add_keytab_to_kinit_from('valid-forever', self.princ)
    self._wait_for_kinit(3)
    assert_true('valid-forever' in self.audit_file,
                "Called kinit; returned 0")
    assert_true(self.audit_file.count('valid-forever') > 3,
                "valid ticket was not retried %s" %
                (self.audit_file,))
    assert_equals(('valid-forever', self.princ),
                  kt_renewer._get_cred_manager().last_candidate_used,
                  'unexpected last candidate used: %s' %
                  (kt_renewer._get_cred_manager().last_candidate_used,))
    # add more valid tickets. We should not try to kinit them though.
    kt_renewer.add_keytab_to_kinit_from('valid-forever-1', self.princ)
    kt_renewer.add_keytab_to_kinit_from('valid-forever-2', self.princ)
    self._wait_for_kinit(3)
    assert_false('valid-forever-1' in self.audit_file,
                 "Called kinit; returned 0")
    assert_false('valid-forever-2' in self.audit_file,
                 "Called kinit; returned 0")
    assert_true(self.audit_file.count('valid-forever') > 6,
                "valid ticket was not retried %s" %
                (self.audit_file,))
    assert_equals(('valid-forever', self.princ),
                  kt_renewer._get_cred_manager().last_candidate_used,
                  'unexpected last candidate used: %s' %
                  (kt_renewer._get_cred_manager().last_candidate_used,))

  def test_async(self):
    start = time.time()
    kt_renewer.add_keytab_to_kinit_from('invalid-block', self.princ)
    end = time.time()

    assert_true(end - start < 1.0, "We didn't get blocked by the kinit call")

  def test_find_keytab_for_principal(self):
    principal = "test_principal"
    not_our_principal = "somebody_else"
    tmp = tempfile.mkdtemp()
    try:
      # create three files, two keytabs and one non keytab.
      fd1, non_keytab = tempfile.mkstemp(dir=tmp)
      fd2, keytab_with_princ = tempfile.mkstemp(suffix= ".keytab", dir=tmp)
      fd3, keytab_wo_princ = tempfile.mkstemp(suffix= ".keytab", dir=tmp)
      os.write(fd1, str(principal))
      os.write(fd2, str(principal))
      os.write(fd3, str(not_our_principal))
      keytab_found = kt_renewer.find_keytab_for_principal(tmp, principal)
      assert_equals(keytab_with_princ, keytab_found, "Found wrong file")
    finally:
      shutil.rmtree(tmp)

  def test_add_remove_specific(self):
    cc_name = tempfile.mkstemp()[1]
    another_princ = None
    try:
      another_princ = tempfile.mkstemp()[1]
      assert_false("valid-keytab-1" in
                   kt_renewer._get_cred_manager()._last_kinit)
      assert_false("valid-keytab-2" in
                   kt_renewer._get_cred_manager()._last_kinit)
      kt_renewer.add_specific_principal_to_kinit("valid-keytab-1",
                                                 self.princ,
                                                 cc_name)
      assert_true("valid-keytab-1" in
                  kt_renewer._get_cred_manager()._last_kinit)
      assert_false("valid-keytab-2" in
                   kt_renewer._get_cred_manager()._last_kinit)
      # make sure that we can add two solr servers using the same principal
      # but running in a different directory. This can happen in the very
      # unlikely case that the user defined two solr services in the same
      # cluster and this node has two solr servers one for each service. They
      # can both use the same principal, but will have different keytabs (one
      # in each solr server process directory). We want to make sure that we
      # kinit both.
      kt_renewer.add_specific_principal_to_kinit("valid-keytab-2",
                                                 another_princ,
                                                 cc_name)
      assert_true("valid-keytab-1" in
                  kt_renewer._get_cred_manager()._last_kinit)
      assert_true("valid-keytab-2" in
                  kt_renewer._get_cred_manager()._last_kinit)
      # We only need to wait once as the entries in the specifics "cache" are
      # always kinit-ed, on every run.
      self._wait_for_kinit()
      assert_true('valid-keytab-1 0' in self.audit_file,
                  "Called kinit; but failed")
      assert_true(cc_name in self.audit_file,
                  "Called kinit but cc_name not present: %s" % self.audit_file)
      another_princ_content = file(another_princ).read().rstrip()
      assert_true('valid-keytab-2 0' in another_princ_content,
                  "Called kinit; but failed")
      assert_true(cc_name in another_princ_content,
                  "Called kinit but cc_name not present: %s" %
                  another_princ_content)
      kt_renewer.remove_specific_principal_to_kinit(self.princ,
                                                    "valid-keytab-1",
                                                    destroy_credentials=True)
      assert_false("valid-keytab-1" in
                   kt_renewer._get_cred_manager()._last_kinit)
      assert_true("valid-keytab-2" in
                  kt_renewer._get_cred_manager()._last_kinit)
      kdestroy_output = file(cc_name).read().rstrip()
      assert_true(cc_name in kdestroy_output, "Kdestroy was not called")
      self._wait_for_kinit()
      # read the "audit" file again
      another_princ_content2 = file(another_princ).read().rstrip()
      assert_true(len(another_princ_content2) > len(another_princ_content),
                  "We expected to kinit another_princ again")
      # we should have kinit another_princ again.
      assert_true(another_princ_content2.endswith(
                  'valid-keytab-2 0 %s' % cc_name),
                  "expected kinit again with the same cc_name: %s" %
                  another_princ_content2)
    finally:
      os.unlink(cc_name)
      if another_princ is not None:
        os.unlink(another_princ)

  def test_set_agent_wide_ccache_name(self):
    # this will initialize the credential manager.
    kt_renewer._get_cred_manager()

    cc_directory = tempfile.gettempdir()
    ccache_file_name = os.path.join(cc_directory,
                                    kt_renewer._AGENT_CCACHE_FILE_NAME)
    assert_equals(ccache_file_name,
                  kt_renewer._get_cred_manager()._agent_ccache_name)
    assert_equals(ccache_file_name, os.environ.get('KRB5CCNAME'))
    kt_renewer.add_keytab_to_kinit_from('valid-for-awhile', self.princ)

    self._wait_for_kinit(3)
    assert_true('valid-for-awhile' in self.audit_file,
                "Called kinit; returned 0")
    assert_true(self.audit_file.count('valid-for-awhile') > 3,
                "valid ticket was not retried %s" %
                (self.audit_file,))
    assert_equals(('valid-for-awhile', self.princ),
                  kt_renewer._get_cred_manager().last_candidate_used,
                  'unexpected last candidate used: %s' %
                  (kt_renewer._get_cred_manager().last_candidate_used,))
    assert_true(ccache_file_name in self.audit_file)

    cc_directory = tempfile.mkdtemp()
    ccache_file_name = os.path.join(cc_directory,
                                    kt_renewer._AGENT_CCACHE_FILE_NAME)
    assert_false(ccache_file_name in self.audit_file)
    try:
      kt_renewer.set_agent_wide_credential_cache_directory(cc_directory)
      assert_equals(ccache_file_name,
                    kt_renewer._get_cred_manager()._agent_ccache_name)
      assert_equals(ccache_file_name, os.environ.get('KRB5CCNAME'))
      self._wait_for_kinit()
      assert_true(ccache_file_name in self.audit_file)
    finally:
      shutil.rmtree(cc_directory)

  def test_renewal_times(self):
    old_renewal = kt_renewer._RENEWAL_PERIOD
    specifics_audit_file = tempfile.mkstemp()[1]
    try:
      # 1000 seconds, eternity in terms of our test.
      kt_renewer._RENEWAL_PERIOD = 1000
      assert_true(kt_renewer._get_cred_manager().last_candidate_used is None,
                  "last cadidate was initialized to %s" %
                  (kt_renewer._get_cred_manager().last_candidate_used,))
      assert_false(kt_renewer._CANDIDATES_LAST_KINIT_TIMESTAMP_KEY in
                   kt_renewer._get_cred_manager()._last_kinit)
      # Let's wait for the initial loop to finish.
      self._wait_for_one_kinit(0)
      # The kt_renewer should be notified and kinit the candidate once.
      start = kt_renewer._get_cred_manager().num_runs
      kt_renewer.add_keytab_to_kinit_from('valid', self.princ)
      self._wait_for_one_kinit(start)
      assert_equals(start + 1, kt_renewer._get_cred_manager().num_runs,
                    "kt_renewal run more than expected")
      audit_file_size = len(self.audit_file)
      assert_true('valid 0' in self.audit_file,
                  "Called kinit; returned 0: %s" % self.audit_file)
      assert_equals(1, self.audit_file.count('valid 0'),
                    "Expected 'valid' keytab to be kinit once")
      assert_equals(('valid', self.princ),
                    kt_renewer._get_cred_manager().last_candidate_used,
                    'unexpected last candidate used: %s' %
                    (kt_renewer._get_cred_manager().last_candidate_used,))
      assert_true(kt_renewer._CANDIDATES_LAST_KINIT_TIMESTAMP_KEY in
                  kt_renewer._get_cred_manager()._last_kinit)
      # add another candidate, we should not kinit again.
      start = kt_renewer._get_cred_manager().num_runs
      kt_renewer.add_keytab_to_kinit_from('valid-too', self.princ)
      self._wait_for_one_kinit(start)
      assert_equals(start + 1, kt_renewer._get_cred_manager().num_runs,
                    "kt_renewal run more than expected")
      assert_equals(audit_file_size, len(self.audit_file),
                    "kinit called again: %s" % self.audit_file)
      # add a specific, kinit should be called for it, but not for the
      # candidates.
      start = kt_renewer._get_cred_manager().num_runs
      kt_renewer.add_specific_principal_to_kinit('specific-1',
                                                 specifics_audit_file,
                                                 'specific-1-cc-name')
      self._wait_for_one_kinit(start)
      assert_equals(start + 1, kt_renewer._get_cred_manager().num_runs,
                    "kt_renewal run more than expected")
      assert_equals(audit_file_size, len(self.audit_file),
                    "kinit called again: %s" % self.audit_file)
      specifics_file = file(specifics_audit_file).read().rstrip()
      assert_equals(1, specifics_file.count('specific-1 0'),
                    "Expected 'specific-1' keytab to be kinit once")
      assert_true('specific-1' in
                  kt_renewer._get_cred_manager()._last_kinit)
      # add another specific, kinit should be called for it and only it.
      start = kt_renewer._get_cred_manager().num_runs
      kt_renewer.add_specific_principal_to_kinit('specific-2',
                                                 specifics_audit_file,
                                                 'specific-2-cc-name')
      self._wait_for_one_kinit(start)
      assert_equals(start + 1, kt_renewer._get_cred_manager().num_runs,
                    "kt_renewal run more than expected")
      assert_equals(audit_file_size, len(self.audit_file),
                    "kinit called again: %s" % self.audit_file)
      specifics_file = file(specifics_audit_file).read().rstrip()
      assert_equals(1, specifics_file.count('specific-1 0'),
                    "Expected 'specific-1' keytab to be kinit once")
      assert_equals(1, specifics_file.count('specific-2 0'),
                    "Expected 'specific-2' keytab to be kinit once")
      assert_true('specific-2' in
                  kt_renewer._get_cred_manager()._last_kinit)
      # add a candidate, nothing should change.
      start = kt_renewer._get_cred_manager().num_runs
      kt_renewer.add_keytab_to_kinit_from('valid-too-too', self.princ)
      self._wait_for_one_kinit(start)
      assert_equals(start + 1, kt_renewer._get_cred_manager().num_runs,
                    "kt_renewal run more than expected")
      assert_equals(audit_file_size, len(self.audit_file),
                    "kinit called again: %s" % self.audit_file)
      specifics_file = file(specifics_audit_file).read().rstrip()
      assert_equals(1, specifics_file.count('specific-1 0'),
                    "Expected 'specific-1' keytab to be kinit once")
      assert_equals(1, specifics_file.count('specific-2 0'),
                    "Expected 'specific-2' keytab to be kinit once")
    finally:
      kt_renewer._RENEWAL_PERIOD = old_renewal
      os.unlink(specifics_audit_file)


  def _wait_for_one_kinit(self, start):
    for i in xrange(10):
      updated = kt_renewer._get_cred_manager().num_runs
      if updated > start:
        break
      time.sleep(0.1)