#!/usr/bin/env python
# Copyright (c) 2012 Cloudera, Inc. All rights reserved.
import simplejson as json
import time

from cmf.log_filterer import FilterConfiguration, CategoryRule
from nose.tools import assert_true, assert_equal, assert_false, assert_raises

def test_never_log():
  rule = CategoryRule("match", 0, 100, [], {})

  assert_true(rule.check_line("match") is None)

def test_always_log():
  rule = CategoryRule("match", -1, 100, [], {})

  assert_true(rule.check_line("match") is not None)
  assert_true(rule.check_line("match") is not None)

def _base_config_setup_json(content, rate, period, group0 = None, group1 = None,
  attribute_key = None, attribute_value = None):
  base_config = "{\"path\" : "
  base_config +=       "{ \"rules\" :"
  base_config +=         " [{\"content\" : \"" + content + "\""
  base_config +=         ", \"rate\" : \"" + rate + "\""
  base_config +=         ", \"period\" : \"" + period + "\""
  if group0 is not None:
    base_config +=         ", \"group0\" : \"" + group0 + "\""
  if group1 is not None:
    base_config +=         ", \"group1\" : \"" + group1 + "\""
  if attribute_key is not None:
    base_config +=         ", \"attribute:" + attribute_key + "\" : \""
    base_config +=          attribute_value + "\""
  base_config +=      "}]"
  base_config += "}}"
  return json.loads(base_config)

def test_bad_rate_configuration():
  filter_config = FilterConfiguration(_base_config_setup_json("match", "abc", "2"))
  assert_equal(0, len(filter_config.file_rules))

def test_bad_period_configuration():
  filter_config = FilterConfiguration(_base_config_setup_json("match", "2", "0"))
  assert_equal(0, len(filter_config.file_rules))

def test_no_logs_and_rules_configuration():
   config = json.loads("{}")
   filter_config = FilterConfiguration(config)
   assert_equal(0, len(filter_config.file_rules))

def test_regex_groups():
  filter_config = FilterConfiguration(_base_config_setup_json("match (\\\w+) (\\\d+)",
      "-1", "0", "word", "digit"))
  attrs = filter_config.match("path", "match testWord 1234")

  assert_equal(2, len(attrs))
  assert_equal(attrs['word'], "testWord")
  assert_equal(attrs['digit'], "1234")

def test_attributes():
  filter_config = FilterConfiguration(_base_config_setup_json("match", "-1", "0",
      None, None, "myKey", "myValue"))
  attrs = filter_config.match("path", "match")

  assert_equal(1, len(attrs))
  assert_equal(attrs['myKey'], "myValue")

def test_attributes_and_group():
  filter_config = FilterConfiguration(_base_config_setup_json("match (\\\w+)",
      "-1", "0", "word", None, "myKey", "myValue"))
  attrs = filter_config.match("path", "match testWord")

  assert_equal(2, len(attrs))
  assert_equal(attrs['myKey'], "myValue")
  assert_equal(attrs['word'], "testWord")


def test_log_rules_json():
  """ Test rate limiting end to end so we make sure we parse the json correctly"""

  config = _base_config_setup_json("match", "1", "1")

  filter_config = FilterConfiguration(config)

  assert_true(filter_config.file_rules['path'][0].period_secs == 60)
  filter_config.file_rules['path'][0].period_secs = 1

  assert_true(filter_config.match("path", "match") is not None)

  # If the first check is just before the 1 minute border then
  # it could succeed the second time, in which case we need to make sure
  # it fails the time after that
  accept_next = filter_config.match("path", "match") is not None
  if accept_next:
    assert_true(filter_config.match("path", "match") is None)

  time.sleep(1.1)

  assert_true(filter_config.match("path", "match") is not None)

