# Copyright (c) 2012 Cloudera, Inc. All rights reserved.
from cmf import log_tailer
import tempfile
import os
import test_log_filterer
import time
import shutil

from nose.tools import assert_true, assert_equal, assert_false, assert_raises

class CallbackProxy(object):
  def __init__(self):
     self.count = 0

  def send_event(self, line, attributes):
    self.count = self.count + 1
    return True

  def set_event_server_config(self, server, port):
    pass

  def shutdown(self):
    pass

def base_setup():
  tmpdir = tempfile.mkdtemp()
  filepath = tmpdir + "/test.txt"

  config = get_basic_config(filepath)

  fd = open(filepath, 'w+')
  callback_proxy = CallbackProxy()
  tailer = log_tailer.LogTailer(None, None, callback_proxy)

  tailer.set_config(config)

  return fd, tailer, callback_proxy

def base_teardown(fd, tailer):
  tailer.shutdown()
  fd.close()
  shutil.rmtree(os.path.dirname(fd.name))

def get_basic_config(filepath):
  config = "{\"" + filepath + "\" : "
  config += "{ \"rules\" : [{\"content\" : \"str\", \"rate\" : \"-1\""
  config += ", \"period\" : \"5\"}]"
  config += "}}"

  return config

def test_single_file_regular_expression():
  fd, tailer, callback_proxy = base_setup()

  try:
    # Write something that SHOULD NOT match the regex
    fd.write("a test\n")
    fd.flush()
    tailer.process_events()
    assert_equal(0, callback_proxy.count)

    # Write something that SHOULD match the regex
    fd.write("a strtest\n")
    fd.flush()
    tailer.process_events()

    assert_equal(1, callback_proxy.count)

    # Write multiple lines at once
    fd.writelines(["strtest1\n", "notmatch\n", "strtest2\n"])
    fd.flush()
    tailer.process_events()
    assert_equal(3, callback_proxy.count)
  finally:
    base_teardown(fd, tailer)

def test_multiple_files_regular_expression():
  tmpdir = tempfile.mkdtemp()
  try:
    filepath1 = tmpdir + "/test1.txt"
    filepath2 = tmpdir + "/test2.txt"

    fd1 = open(filepath1, 'w+')
    fd2 = open(filepath2, 'w+')

    config = "{ \"" + filepath1 + "\" : "
    config +=  "{\"rules\" : "
    config +=   "[{\"content\" : \"aa\", \"rate\" : \"-1\", \"period\" : \"5\"}]},"
    config += "\"" + filepath2 + "\" :"
    config +=  "{\"rules\" : "
    config +=   "[{\"content\" : \"aa\", \"rate\" : \"-1\", \"period\" : \"5\"}]}}"

    callback_proxy = CallbackProxy()
    tailer = log_tailer.LogTailer(None, None, callback_proxy)
    tailer.set_config(config)

    fd1.write("aa\n")
    fd1.flush()
    fd2.write("aa\n")
    fd2.flush()

    tailer.process_events()

    assert_equal(2, callback_proxy.count)
    config = "\"logsandrules\" : [{ \"filepath\":\"" + tmpdir + "/test.txt\","
    config += "\"content\" : \"str\"}]"
  finally:
    tailer.shutdown()
    fd1.close()
    fd2.close()
    shutil.rmtree(tmpdir)

def test_partial_line_write():
  fd, tailer, callback_proxy = base_setup()

  try:
    # Write something that SHOULD NOT match the regex
    fd.write("st")
    fd.flush()
    tailer.process_events()
    assert_equal(0, callback_proxy.count)

    fd.write("r\n")
    fd.flush()
    tailer.process_events()
    assert_equal(1, callback_proxy.count)
  finally:
    base_teardown(fd, tailer)

def test_partial_line_multiline_write():
  fd, tailer, callback_proxy = base_setup()

  try:
    fd.write("str\ns")
    fd.flush()
    tailer.process_events()
    assert_equal(1, callback_proxy.count)

    fd.write("t")
    fd.flush()
    tailer.process_events()
    assert_equal(1, callback_proxy.count)

    fd.write("r\n")
    fd.flush()
    tailer.process_events()
    assert_equal(2, callback_proxy.count)
  finally:
    base_teardown(fd, tailer)

def test_no_failure_on_nonexistent_file_path():
  # Bad file path
  tmpdir = tempfile.mkdtemp()
  try:
    filepath = tmpdir + "/nonexistentfile.txt"
    config = get_basic_config(filepath)

    callback_proxy = CallbackProxy()
    tailer = log_tailer.LogTailer(None, None, callback_proxy)
    tailer.set_config(config)

    # Bad directory path
    config = get_basic_config("/nonexistent/test.txt")
    tailer.set_config(config)
  finally:
    tailer.shutdown()
    shutil.rmtree(tmpdir)

def test_log_rotation():
  tmpdir = tempfile.mkdtemp()
  try:
    filepath = tmpdir + "/rotate.txt"
    config = get_basic_config(filepath)

    fd = open(filepath, 'w+')

    callback_proxy = CallbackProxy()
    tailer = log_tailer.LogTailer(None, None, callback_proxy)
    tailer.set_config(config)

    fd.write("str\n")
    new_path = tmpdir + "/rotate2.txt"
    os.rename(filepath, new_path)

    fd2 = open(filepath, 'w+')
    fd2.flush()
    fd2.write("str\n")
    fd.flush()
    fd2.flush()

    tailer.process_events()
    assert_equal(2, callback_proxy.count)
  finally:
    fd2.close()
    tailer.shutdown()
    fd.close()
    shutil.rmtree(tmpdir)

def test_move_file_to_location():
  tmpdir = tempfile.mkdtemp()
  try:
    start_filepath = tmpdir + "/test.txt"
    end_filepath = tmpdir + "/test2.txt"
    fd = open(start_filepath, 'w+')
    fd.write("str\n")
    fd.flush()

    config = get_basic_config(end_filepath)
    callback_proxy = CallbackProxy()
    tailer = log_tailer.LogTailer(None, None, callback_proxy)
    tailer.set_config(config)

    os.rename(start_filepath, end_filepath)

    tailer.process_events()

    fd.write("str\n")
    fd.flush()

    tailer.process_events()
    assert_equal(1, callback_proxy.count)
  finally:
    tailer.shutdown()
    fd.close()
    shutil.rmtree(tmpdir)

def test_delete_file():
  tmpdir = tempfile.mkdtemp()
  try:
    filepath = tmpdir + "/test.txt"

    config = get_basic_config(filepath)
    callback_proxy = CallbackProxy()
    tailer = log_tailer.LogTailer(None, None, callback_proxy)
    tailer.set_config(config)
    fd = open(filepath, 'w+')
    tailer.process_events()
    fd.write("str\n")
    fd.flush()

    os.remove(filepath)
    tailer.process_events()
    assert_equal(1, callback_proxy.count)
  finally:
    fd.close()
    tailer.shutdown()
    shutil.rmtree(tmpdir)

def test_dynamically_change_config_file_path():
  fd, tailer, callback_proxy = base_setup()

  try:
    fd.write("str\n")
    fd.flush()
    tailer.process_events()
    assert_equal(1, callback_proxy.count)

    tmpdir = tempfile.mkdtemp()
    filepath2 = tmpdir + "/other.txt"
    fd2 = open(filepath2, 'w+')
    config = get_basic_config(filepath2)
    tailer.set_config(config)

    fd.write("str\n")
    fd.flush()
    tailer.process_events()
    assert_equal(1, callback_proxy.count)

    fd2.write("str\n")
    fd2.flush()
    tailer.process_events()
    assert_equal(2, callback_proxy.count)
  finally:
    fd2.close()
    shutil.rmtree(tmpdir)
    base_teardown(fd, tailer)

def test_dynamically_change_config_regex():
  tmpdir = tempfile.mkdtemp()
  try:
    filepath = tmpdir + "/regEx.txt"
    fd = open(filepath, 'w+')
    config = "{\"" + filepath + "\" : "
    config +=   "{\"rules\" : [{\"content\" : \"regExp\", \"rate\" : \"-1\","
    config +=   "\"period\" : \"5\"}]"
    config += "}}"

    callback_proxy = CallbackProxy()
    tailer = log_tailer.LogTailer(None, None, callback_proxy)
    tailer.set_config(config)

    new_config = config.replace("regExp", "newExp")
    tailer.set_config(new_config)

    fd.write("regExp\n")
    fd.flush()
    tailer.process_events()
    assert_equal(0, callback_proxy.count)

    fd.write("newExp\n")
    fd.flush()
    tailer.process_events()
    assert_equal(1, callback_proxy.count)
  finally:
    tailer.shutdown()
    fd.close()
    shutil.rmtree(tmpdir)

def test_dynamically_change_config_to_same():
  tmpdir = tempfile.mkdtemp()
  try:
    filepath = tmpdir + "/regEx.txt"
    fd = open(filepath, 'w+')
    config = get_basic_config(filepath)

    callback_proxy = CallbackProxy()
    tailer = log_tailer.LogTailer(None, None, callback_proxy)
    tailer.set_config(config)

    fd.write("str\n")
    fd.flush()

    tailer.set_config(config)

    tailer.process_events()
    assert_equal(1, callback_proxy.count)
  finally:
    tailer.shutdown()
    fd.close()
    shutil.rmtree(tmpdir)

def test_short_lived_file():
  tmpdir = tempfile.mkdtemp()
  try:
    filepath = tmpdir + "/test.txt"

    config = get_basic_config(filepath)
    callback_proxy = CallbackProxy()
    tailer = log_tailer.LogTailer(None, None, callback_proxy)
    tailer.set_config(config)

    fd = open(filepath, 'w+')
    os.remove(filepath)

    tailer.process_events()
  finally:
    fd.close()
    tailer.shutdown()
    shutil.rmtree(tmpdir)

def test_delete_directory():
  tmpdir = tempfile.mkdtemp()
  try:
    filepath = tmpdir + "/test.txt"
    fd = open(filepath, 'w+')

    config = get_basic_config(filepath)
    callback_proxy = CallbackProxy()
    tailer = log_tailer.LogTailer(None, None, callback_proxy)
    tailer.set_config(config)

    fd.write("str\n")
    fd.flush()
    os.remove(filepath)
    os.removedirs(tmpdir)

    tailer.process_events()
    assert_equal(1, callback_proxy.count)

    # Note that we don't support auto-looking for newly created files
    # without resetting the config
    directory, path = os.path.split(filepath)
    os.mkdir(directory)
    fd2 = open(filepath, 'w+')
    fd2.write("test\n")
    fd2.flush()
    tailer.set_config(config)

    fd2.write("str\n")
    fd2.flush()
    tailer.process_events()
    assert_equal(2, callback_proxy.count)
  finally:
    fd2.close()
    fd.close()
    tailer.shutdown()
    shutil.rmtree(tmpdir)

def test_invalid_json():

  # Make sure it doesn't throw an exception
  bad_config = "{[}"
  callback_proxy = CallbackProxy()
  tailer = log_tailer.LogTailer(None, None, callback_proxy)
  try:
    tailer.set_config(bad_config)
  finally:
    tailer.shutdown()
