# Copyright 2017, Cloudera Inc.
import uuid

# Workaround for issues in RHEL 7 containers.
import socket
socket.has_ipv6 = False

from test.test_support import run_unittest
import unittest

from nose.tools import assert_equal
from nose.tools import assert_false
from nose.tools import assert_true
from mock import Mock
from mock import patch

from cmf.util.os_ops import OSOps
import os.path
import logging
import tempfile


LOG = logging.getLogger(__name__)

def make_obj(**kwargs):
  return type("anon",(object,), kwargs)()

class BaseTests(unittest.TestCase):
  mode_755 = 0o755
  mode_123 = 0o123
  mode_775 = 0o775

  @patch("shutil.copy2")
  def test_file_copy(self, copy):
    ops = OSOps(True)
    ops.set_path_metadata = Mock()
    copy.return_value = None

    assert_equal(ops.install_file("/tmp/foo", "/tmp/bar"), None)
    copy.assert_called_once_with("/tmp/bar", "/tmp/foo")
    ops.set_path_metadata.assert_called_once_with("/tmp/foo", None, None, None)

  @patch("pwd.getpwnam")
  @patch("grp.getgrnam")
  @patch("os.chown")
  @patch("os.chmod")
  @patch("os.stat")
  def test_set_metadata(self, stat, chmod, chown, getgrnam, getpwnam):
    ops = OSOps(True)
    stat.return_value = make_obj(st_uid=12, st_gid=34)
    getpwnam.return_value = make_obj(pw_uid=13)
    getgrnam.return_value = make_obj(gr_gid=35)

    assert_equal(ops.set_path_metadata("/tmp/foo", mode=self.mode_755), None)
    chmod.assert_called_once_with("/tmp/foo", self.mode_755)
    chmod.reset_mock()

    # Only owner set -- group taken from existing file (35)
    assert_equal(ops.set_path_metadata("/tmp/foo", mode=self.mode_755, owner="foo"), None)
    getpwnam.assert_called_once_with("foo")
    chmod.assert_called_once_with("/tmp/foo", self.mode_755)
    chown.assert_called_once_with("/tmp/foo",13, 34)
    chmod.reset_mock()
    chown.reset_mock()

    # Only group set -- owner take from existing file (12)
    assert_equal(ops.set_path_metadata("/tmp/foo", mode=self.mode_755, group="foo"), None)
    getgrnam.assert_called_once_with("foo")
    chown.assert_called_once_with("/tmp/foo", 12, 35)
    chmod.assert_called_once_with("/tmp/foo", self.mode_755)

    chown.reset_mock()
    chmod.reset_mock()

  def test_file_editor(self):
    ops = OSOps(True)

    content1 = "foo foo"
    content2 = "bar"
    fileobj = tempfile.NamedTemporaryFile()
    try:
      fileobj.write(content1)
      fileobj.flush()

      def editor1(data):
        self.assertEqual(data, content1, "Original file content corrupted: %s" % data)
        return content2
      ops.edit_file(fileobj.name, editor1)

      def editor2(data):
        # Validates that editor1 worked correctly
        self.assertEqual(data, content2, "Altered file content corrupted: '%s' '%s'" % (data, content2))
        return content2

      ops.edit_file(fileobj.name, editor2)

    finally:
      fileobj.close()

  @patch("os.fdopen")
  @patch("os.open")
  def test_open_write_file(self, opener, fdopener):
    """
    Test that "open_write_file" sets file metadata.
    That method otherwise relies heavily on system libraries, and we're not trying to test those.
    """
    ops = OSOps(True)
    ops.set_path_metadata = Mock()

    ops.open_write_file("path","owner", "group", self.mode_123)

    ops.set_path_metadata.assert_called_once_with("path", "owner", "group", self.mode_123)

  @patch("os.path.isdir")
  @patch("os.listdir")
  def test_set_directory_content_metadata(self, listdir, isdir):

    ops = OSOps(True)
    contents = { "dirbase": {"f1":None, "f2":None, "d1": {"f1": None}}}
    def _listdir(path):
      words = path.strip("/").split('/')
      x = contents
      for w in words:
        if w:
          x = x[w]
      out = x.keys()
      return out

    def _isdir(path):
      return os.path.split(path)[1][0] == 'd'

    listdir.side_effect = _listdir
    isdir.side_effect = _isdir


    ops.set_path_metadata = Mock()

    ops.set_directory_content_metadata("/", "owner", "group",self.mode_123, True)

    call_paths = set(["/dirbase","/dirbase/f1", "/dirbase/f2", "/dirbase/d1", "/dirbase/d1/f1"])
    paths = set()
    for c in ops.set_path_metadata.call_args_list:
      paths.add(c[0][0])
      assert_equal(c[0][1:4], ("owner", "group", self.mode_123))
    assert_equal(call_paths, paths, "Wrong call paths")

  @patch("os.path.isdir")
  def test_mkabsdir(self, isdir):
    """
    mkabsdir() should check for directory existence and create directories.
    """
    non_default_mode = self.mode_775
    ops = OSOps(True, non_default_mode) #Initialize with a value different from default for mode
    ops.set_path_metadata = Mock()
    ops.makedirs = Mock()

    isdir.return_value = False

    ops.mkabsdir("/foo", 13, 34, self.mode_123)

    ops.makedirs.assert_called_once_with("/foo", non_default_mode)
    ops.set_path_metadata.assert_called_once_with("/foo", 13, 34, self.mode_123)

    # Errors in directory creation fail with an exception
    ops.makedirs.reset_mock()
    ops.makedirs.side_effect = Exception("/foo")
    try:
      ops.mkabsdir("/foo", 13, 34, self.mode_123)
    except Exception, e:
      assert_equal(str(e), "/foo")
    ops.makedirs.assert_called_once_with("/foo", non_default_mode)

  @patch("os.path.exists")
  @patch("os.chmod")
  @patch("os.mkdir")
  def test_makedirs(self, mock_mkdir, mock_chmod, mock_exists):
    """
    Tests the makedir call in OSOps.
    """

    ops = OSOps(True)

    assert ops.default_perm_mode == self.mode_755
    non_default_mode = self.mode_775
    assert ops.default_perm_mode != non_default_mode

    # Test /
    path_exists = {"/": True}
    def path_exists_effect(name):
      return path_exists[name]
    mock_exists.side_effect=path_exists_effect

    ops.makedirs("/", non_default_mode)

    assert mock_mkdir.call_count == 0
    assert mock_chmod.call_count == 0
    mock_mkdir.reset_mock()
    mock_chmod.reset_mock()

    # Test /dir

    name = "/" + str(uuid.uuid4())

    path_exists = {"/": True, name: False}
    def path_exists_effect(name):
      return path_exists[name]
    mock_exists.side_effect=path_exists_effect

    ops.makedirs(name, non_default_mode)

    mock_mkdir.assert_called_with(name, non_default_mode)
    mock_chmod.assert_called_with(name, non_default_mode)
    assert mock_mkdir.call_count == 1
    assert mock_chmod.call_count == 1

    mock_mkdir.reset_mock()
    mock_chmod.reset_mock()

    # Test /dir1/dir2

    name_dir1 = "/" + str(uuid.uuid4())
    name_dir2 = name_dir1 + "/" + str(uuid.uuid4())

    path_exists = {"/": True, name_dir1: False, name_dir2: False}
    def path_exists_effect(name):
      return path_exists[name]
    mock_exists.side_effect=path_exists_effect

    ops.makedirs(name_dir2, non_default_mode)

    mock_mkdir.assert_any_call(name_dir1, non_default_mode)
    mock_mkdir.assert_any_call(name_dir2, non_default_mode)
    mock_chmod.assert_any_call(name_dir1, non_default_mode)
    mock_chmod.assert_any_call(name_dir2, non_default_mode)

    assert mock_mkdir.call_count == 2
    assert mock_chmod.call_count == 2

    mock_mkdir.reset_mock()
    mock_chmod.reset_mock()

    # test /dir1/dir2/.

    name_dir1 = "/" + str(uuid.uuid4())
    name_dir2 = name_dir1 + "/" + str(uuid.uuid4())
    name_dir3 = name_dir2 + "/."

    path_exists = {"/": True, name_dir1: False, name_dir2: False, name_dir3: False}
    def path_exists_effect(name):
      return path_exists[name]
    mock_exists.side_effect=path_exists_effect

    ops.makedirs(name_dir3, non_default_mode)

    mock_mkdir.assert_any_call(name_dir1, non_default_mode)
    mock_mkdir.assert_any_call(name_dir2, non_default_mode)
    mock_chmod.assert_any_call(name_dir1, non_default_mode)
    mock_chmod.assert_any_call(name_dir2, non_default_mode)
    assert mock_mkdir.call_count == 2
    assert mock_chmod.call_count == 2

    mock_mkdir.reset_mock()
    mock_chmod.reset_mock()

    # test /dir1/dir2/dir3/

    name_dir1 = "/" + str(uuid.uuid4())
    name_dir2 = name_dir1 + "/" + str(uuid.uuid4())
    name_dir3 = name_dir2 + "/" + str(uuid.uuid4()) + "/"

    path_exists = {"/": True, name_dir1: False, name_dir2: False, name_dir3: False}
    def path_exists_effect(name):
      return path_exists[name]
    mock_exists.side_effect=path_exists_effect

    ops.makedirs(name_dir3, non_default_mode)

    mock_mkdir.assert_any_call(name_dir1, non_default_mode)
    mock_mkdir.assert_any_call(name_dir2, non_default_mode)
    mock_mkdir.assert_any_call(name_dir3, non_default_mode)
    mock_chmod.assert_any_call(name_dir1, non_default_mode)
    mock_chmod.assert_any_call(name_dir2, non_default_mode)
    mock_chmod.assert_any_call(name_dir3, non_default_mode)
    assert mock_mkdir.call_count == 3
    assert mock_chmod.call_count == 3

    mock_mkdir.reset_mock()
    mock_chmod.reset_mock()

    # test /<exists>

    name_dir1 = "/" + str(uuid.uuid4())

    path_exists = {name_dir1: True, "/": True}
    def path_exists_effect(name):
      return path_exists[name]
    mock_exists.side_effect=path_exists_effect

    ops.makedirs(name_dir1, non_default_mode)

    assert mock_mkdir.call_count == 0
    assert mock_chmod.call_count == 0

    mock_mkdir.reset_mock()
    mock_chmod.reset_mock()

    # test /<exists>/<exists>/.

    name_dir1 = "/" + str(uuid.uuid4())
    name_dir2 = name_dir1 + "/" + str(uuid.uuid4()) + "/."

    path_exists = {name_dir1: True, name_dir2: True}
    def path_exists_effect(name):
      return path_exists[name]

    mock_exists.side_effect=path_exists_effect

    ops.makedirs(name_dir1, non_default_mode)

    assert mock_mkdir.call_count == 0
    assert mock_chmod.call_count == 0

    mock_mkdir.reset_mock()
    mock_chmod.reset_mock()

    # test /<exists>/<rand>

    name_dir1 = "/" + str(uuid.uuid4())
    name_dir2 = name_dir1 + '/'+ str(uuid.uuid4())

    path_exists = {name_dir1: True, name_dir2: False}
    def path_exists_effect(name):
      return path_exists[name]
    mock_exists.side_effect=path_exists_effect

    ops.makedirs(name_dir2, non_default_mode)

    mock_mkdir.assert_any_call(name_dir2, non_default_mode)
    mock_chmod.assert_any_call(name_dir2, non_default_mode)
    assert mock_mkdir.call_count == 1
    assert mock_chmod.call_count == 1

    mock_mkdir.reset_mock()
    mock_chmod.reset_mock()

    # test /<exists>/<rand>/<rand>/.

    name_dir1 = "/" + str(uuid.uuid4())
    name_dir2 = name_dir1 + '/'+ str(uuid.uuid4())
    name_dir3 = name_dir2 + '/'+ str(uuid.uuid4())

    path_exists = {name_dir1: True, name_dir2: False, name_dir3: False}
    def path_exists_effect(name):
      try:
        return path_exists[name]
      except KeyError, e:
        return False

    mock_exists.side_effect=path_exists_effect

    ops.makedirs(name_dir3 + "/.", non_default_mode)

    mock_mkdir.assert_any_call(name_dir2, non_default_mode)
    mock_mkdir.assert_any_call(name_dir3, non_default_mode)
    mock_chmod.assert_any_call(name_dir2, non_default_mode)
    mock_chmod.assert_any_call(name_dir3, non_default_mode)
    assert mock_mkdir.call_count == 2
    assert mock_chmod.call_count == 2
    mock_mkdir.stop()
    mock_chmod.stop()
    mock_exists.stop()

  @patch("os.path.isdir")
  def test_mkabsdir_not_abs_path(self, isdir):
    """
    mkabsdir() called with a non-absolute path should fail.
    """
    ops = OSOps(True)
    ops.set_path_metadata = Mock()
    isdir.return_value = False

    try:
      ops.mkabsdir("foo")
      assert_false("Expected an exception.")
    except Exception:
      pass

  @patch("os.walk")
  def test_find_paths(self, walker):

    paths = { "/a": ("/a", ["d1", "d2"], ["f1", "f2"]),
              "/a/d1" : ("/a/d1", [], ["f3"]),
              "/a/d2" : ("/a/d2", [], [])}

    def _walker(path):
      for x in paths.values():
        yield x
      raise StopIteration()

    walker.side_effect = _walker
    ops = OSOps(True)
    found = set()
    for x in ops.file_tree_walk("/a"):
      found.add(x)

    expected = set(["/a/d1", "/a/d2", "/a/f1", "/a/f2", "/a/d1/f3"])
    self.assertEqual(expected, found,
                     "Mismatched file set: expected:%s result:%s" % (expected, found))

  def test_get_ephemeral_ports(self):
    """
    Test if ephemeral ports are returned.
    """
    ops = OSOps(True)
    ports = ops.get_ephemeral_port_range()
    assert_true(len(ports) == 2)
    assert_true(ports[0] is not None)
    assert_true(ports[1] is not None)
    assert_true(ports[1] > ports[0])

  def test_get_ephemeral_ports_override(self):
    """
    Test no ephemeral ports are returned if file does not exist.
    """
    class OSOpsTest(OSOps):
      def __init__(self):
        OSOps.__init__(self, True)
        self.IP_LOCAL_PORT_RANGE = "/doesnotexist"
      
    ops = OSOpsTest()
    ports = ops.get_ephemeral_port_range()
    assert_true(len(ports) == 0)

  @patch("subprocess.Popen")
  def test_exec_commands(self, mock_popen):
    """
    Test that exec_command and sudo_command call subprocess.Popen as expected.
    """
    ops = OSOps(True, sudo="mocksudo")
    ops.exec_command("true", "foo")
    mock_popen.assert_called_with(("true", "foo"), stdin=None, stderr=None, stdout=None)
    mock_popen.reset_mock()

    ops.sudo_command("true", "foo")
    mock_popen.assert_called_with(("mocksudo", "true", "foo"), stdin=None, stderr=None, stdout=None)
    mock_popen.reset_mock()

  @patch("os.rename")
  @patch("os.stat")
  def test_alternatives(self, mock_stat, mock_rename):
    """
    Test that alternatives activation and deactivation result in "update-alternatives"
    invocations as expected.
    """

    ops = OSOps(True, sudo="fake_sudo")
    ops.alternatives.db_dir = "/tmp"
    ops.alternatives.alt_bin = "/fake-update-alternatives"

    ops.alternatives.DB_CANDIDATES = ["/tmp"]
    ops.exec_command = Mock()
    ops.sudo_command = Mock()
    ops.alternatives.activate("alt_name", "alt_dest", "alt_source", 88, False)
    ops.exec_command.assert_not_called()
    ops.sudo_command.assert_called_with(ops.alternatives.alt_bin,
                                        "--install", "alt_dest", "alt_name", "alt_source", "88")
    ops.sudo_command.reset_mock()

    ops.alternatives.deactivate("alt_name", "alt_dest", "alt_source", 88, False)
    ops.exec_command.assert_not_called()
    ops.sudo_command.assert_called_with(ops.alternatives.alt_bin,
                                        "--remove", "alt_name", "alt_source")


def main():
  run_unittest(BaseTests)

if __name__ == "__main__":
  logging.basicConfig(level=logging.INFO)
  main()
