#!/usr/bin/env python
# Copyright (c) 2012 Cloudera, Inc. All rights reserved.

import os
import shutil
import simplejson as json
import tempfile

from sets import Set

from mock import Mock
from nose.tools import assert_true, assert_equal, assert_false, assert_raises, raises

import cmf.parcel as parcel

from cmf.parcel import ParcelNotAvailableException, ParcelTagUnsatisfiedException, ParcelTagOversatisfiedException
import test_deep_equality

FIXTURE_PATH = os.path.join(os.path.dirname(__file__), "test-fixtures", "parcel-repository-fixture")
FIXTURE_SYMLINK_PATH = os.path.join(os.path.dirname(__file__), "test-fixtures-symlink", "parcel-repository-fixture")
AGENT_PACKAGE_PATH = '/cmf/agent/services' # Doesn't need to be a real path

def mock_call(arg1, arg2=None, arg3=None):
  return 0

class MockableRepository(parcel.Repository):
  def set_callout(self, func):
    '''Used by tests to mock invoking subprocesses'''
    setattr(self, "callout", func)

  def set_chown(self, func):
    '''Used by tests to mock chown'''
    setattr(self, "chown", func)

  def set_chmod(self, func):
    '''Used by tests to mock chmod'''
    setattr(self, "chmod", func)

  def set_running_as_root(self, func):
    '''Used by tests to mock running_as_root'''
    setattr(self, "running_as_root", func)

  def set_makedirs(self, func):
    '''Used by tests to mock makedirs'''
    setattr(self, "makedirs", func)

  def set_parcel_added(self, func):
    """Callback when a new parcel has been added to the repository. The func
       will be invoked with the parcel name and version as arguments.
    """
    self.parcel_added = func

def make_new_repository(path=FIXTURE_PATH, package_dir=AGENT_PACKAGE_PATH):
  osops = Mock()

  repo = MockableRepository(path, package_dir, os_ops = osops)
  repo.set_callout(mock_call)
  repo.set_chown(mock_call)
  repo.set_chmod(mock_call)
  repo.set_makedirs(mock_call)
  repo.refresh()
  return repo

def test_parcel_repository():
  """Test that parcel enumeration returns expected contents"""

  ref = {
      'product1':
            { '1.0.0': {'components': [{'name': 'component1',
                                        'version': '1.0.0'},
                                       {'name': 'component2',
                                        'version': '0.6.0'},
                                       {'name': 'component3',
                                        'version': '4.5.0'}],
                        'packages': {},
                        'groups': ['group1'],
                        'maxPrevVersion': '2.0.0',
                        'minPrevVersion': '1.0.0',
                        'name': 'product1',
                        'provides': [ 'product1', 'foo-plugin', 'bar-plugin' ],
                        'setActiveSymlink': True,
                        'scripts': {'defines': 'defines.sh',
                                    'alternatives': 'product1_alternatives.sh'},
                        'users': {'user1': {'longname': 'User One',
                                            'home': '/tmp/user1',
                                            'shell': '/bin/false',
                                            'extra_groups': [ 'group1' ]}},
                        'version': '1.0.0'},
              '1.1.0': {'components': [{'name': 'component1',
                                        'version': '1.1.0'},
                                       {'name': 'component2',
                                        'version': '0.6.1'},
                                       {'name': 'component3',
                                        'version': '4.6.0'}],
                        'packages': {},
                        'groups': [],
                        'maxPrevVersion': '2.0.0',
                        'minPrevVersion': '1.0.0',
                        'name': 'product1',
                        'setActiveSymlink': True,
                        'scripts': {'defines': 'defines.sh'},
                        'users': {},
                        'version': '1.1.0'}},
      'product1_overlay':
            { '1.0.0': {'components': [{'name': 'component1',
                                        'version': '1.0.0'},
                                       {'name': 'component2',
                                        'version': '0.6.0'},
                                       {'name': 'component3',
                                        'version': '4.5.0'}],
                        'depends': "product1 (= 1.0.0)",
                        'packages': {},
                        'groups': [],
                        'maxPrevVersion': '2.0.0',
                        'minPrevVersion': '1.0.0',
                        'name': 'product1_overlay',
                        'provides': [ 'product1' ],
                        'setActiveSymlink': True,
                        'scripts': {'defines': 'defines.sh' },
                        'users': {},
                        'version': '1.0.0'}},
      'product2':
            { '2.0.0': {'components': [],
                        'packages': {},
                        'groups': [],
                        'maxPrevVersion': '3.0.0',
                        'minPrevVersion': '2.0.0',
                        'name': 'product2',
                        'provides': [ 'product2', 'foo-plugin' ],
                        'setActiveSymlink': True,
                        'scripts': {},
                        'users': {},
                        'version': '2.0.0'},
              '2.5.0': {'components': [],
                        'packages': {},
                        'groups': ['group2'],
                        'name': 'product2',
                        'provides': [ 'product2', 'foo-plugin' ],
                        'setActiveSymlink': False,
                        'scripts': {'defines': 'more-defines.sh'},
                        'users': {'user2': {'longname': 'User Two',
                                            'home': '/tmp/user2',
                                            'shell': '/bin/false',
                                            'extra_groups': []}},
                        'version': '2.5.0'},
              '3.0.0': {'components': [],
                        'packages': {},
                        'groups': [],
                        'name': 'product2',
                        'provides': [ 'product2', 'foo-plugin' ],
                        'setActiveSymlink': True,
                        'scripts': {},
                        'users': {},
                        'version': '3.0.0'}}}

  ref_invalid = Set([ ('product1', '0.9.0'),
                      ('product1', '0.9.1'),
                      ('product1', '0.9.2') ])

  repo = make_new_repository(path=FIXTURE_PATH)
  test_deep_equality.deep_eq(repo.products, ref, _assert=True)
  test_deep_equality.deep_eq(repo.invalid_products, ref_invalid, _assert=True)

  repo = make_new_repository(path=FIXTURE_SYMLINK_PATH)
  test_deep_equality.deep_eq(repo.products, ref, _assert=True)
  test_deep_equality.deep_eq(repo.invalid_products, ref_invalid, _assert=True)

def check_parcel_results(available_parcels, expected_scripts, expected_parcels, required_tags, optional_tags):
  repo = make_new_repository()

  env = {}
  selected = repo.prepare_environment(available_parcels, env, required_tags, optional_tags)

  if not expected_parcels:
    assert_false('SCM_DEFINES_SCRIPTS' in env)
    assert_false('PARCELS_ROOT' in env)
  else:
    test_deep_equality.deep_eq(env['SCM_DEFINES_SCRIPTS'], ':'.join(expected_scripts), _assert=True)
    assert_equal(env['PARCELS_ROOT'], FIXTURE_PATH)
    for script in expected_scripts:
      assert_true(os.path.exists(script))
  assert_equal(selected, expected_parcels)

def test_prepare_environment():
  """Test that parcel environment variables are correctly injected"""

  parcels = { 'product1': '1.0.0', 'product2': '2.5.0' }
  required = ['product1', 'product2']
  optional = []
  scripts = [ os.path.join(FIXTURE_PATH, 'product1-1.0.0', 'meta/defines.sh'),
              os.path.join(FIXTURE_PATH, 'product2-2.5.0', 'meta/more-defines.sh')]

  check_parcel_results(parcels, scripts, parcels, required, optional)

def test_prepare_environment_optional_parcels():
  """Test that optional parcels are selected"""

  parcels = { 'product1': '1.0.0', 'product2': '2.5.0' }
  required = ['product1']
  optional = ['product2']
  scripts = [ os.path.join(FIXTURE_PATH, 'product1-1.0.0', 'meta/defines.sh'),
              os.path.join(FIXTURE_PATH, 'product2-2.5.0', 'meta/more-defines.sh')]

  check_parcel_results(parcels, scripts, parcels, required, optional)

def test_prepare_environment_multiple_optional_parcels():
  """Test that multiple parcels for an optional tag are selected"""

  parcels = { 'product1': '1.0.0', 'product2': '2.5.0' }
  required = []
  optional = ['foo-plugin']
  scripts = [ os.path.join(FIXTURE_PATH, 'product1-1.0.0', 'meta/defines.sh'),
              os.path.join(FIXTURE_PATH, 'product2-2.5.0', 'meta/more-defines.sh')]

  check_parcel_results(parcels, scripts, parcels, required, optional)

def test_prepare_environment_missing_optional_parcels():
  """Test that optional parcels are ignored if missing"""

  parcels = { 'product1': '1.0.0' }
  required = ['product1']
  optional = ['product2']
  scripts = [ os.path.join(FIXTURE_PATH, 'product1-1.0.0', 'meta/defines.sh')]

  check_parcel_results(parcels, scripts, parcels, required, optional)

def test_prepare_environment_ignore_irrelevant_parcels():
  """Test that irrelevant parcels are ignored"""

  parcels = { 'product1': '1.0.0', 'product2': '2.5.0' }
  expected = { 'product1': '1.0.0' }
  required = ['product1']
  optional = []
  scripts = [ os.path.join(FIXTURE_PATH, 'product1-1.0.0', 'meta/defines.sh')]

  check_parcel_results(parcels, scripts, expected, required, optional)

@raises(ParcelTagUnsatisfiedException)
def test_prepare_environment_missing_required_parcels():
  """Test that missing required parcels produces an exception"""

  parcels = { 'product1': '1.0.0', 'product2': '2.5.0' }
  expected = {}
  required = ['product1', 'product3']
  optional = ['product2']
  scripts = []

  check_parcel_results(parcels, scripts, expected, required, optional)

@raises(ParcelTagOversatisfiedException)
def test_prepare_environment_too_many_required_parcels():
  """Test that too many required parcels produces an exception"""

  parcels = { 'product1': '1.0.0', 'product2': '2.5.0' }
  expected = {}
  required = ['foo-plugin']
  optional = []
  scripts = []

  check_parcel_results(parcels, scripts, expected, required, optional)

def test_prepare_environment_multiple_required_tags_one_parcels():
  """Test that one parcel providing multiple required tags is ok"""

  parcels = { 'product1': '1.0.0' }
  required = ['product1', 'foo-plugin', 'bar-plugin']
  optional = []
  scripts = [ os.path.join(FIXTURE_PATH, 'product1-1.0.0', 'meta/defines.sh')]

  check_parcel_results(parcels, scripts, parcels, required, optional)

@raises(ParcelTagUnsatisfiedException)
def test_prepare_environment_badparcel():
  """Test that parcel environment checks to see if parcels exist
     before setting up the environment
  """
  repo = make_new_repository()

  env = {}
  parcels = { 'product1': '1.0.0', 'product2': '2.4.0' }
  repo.prepare_environment(parcels, env, ['product2'], [])

def test_prepare_environment_no_parcels():
  """Test that everything is ok if no parcels are selected"""

  parcels = {}
  required = ['foo']
  optional = ['bar']
  scripts = []

  check_parcel_results(parcels, scripts, parcels, required, optional)

def test_prepare_environment_some_parcels_no_tags():
  """Test that no parcels are selected if no tags are specified"""

  parcels = { 'product1': '1.0.0', 'product2': '2.5.0' }
  required = []
  optional = []
  scripts = []

  check_parcel_results(parcels, scripts, {}, required, optional)

def test_prepare_environment_wildcard_parcel_no_tags():
  """Test that wildcard parcels are not selected if no tags are specified"""

  parcels = { 'product1': '1.1.0', 'product2': '2.5.0' }
  required = []
  optional = []
  scripts = []

  check_parcel_results(parcels, scripts, {}, required, optional)

def test_prepare_environment_no_parcels_no_tags():
  """Test that everything is ok if no parcels or tags are specified"""

  parcels = {}
  required = []
  optional = []
  scripts = []

  check_parcel_results(parcels, scripts, parcels, required, optional)

def test_prepare_environment_overlay_parcels():
  """Test that we don't raise an oversatisfied exception when using overlay parcels"""

  parcels = { 'product1': '1.0.0', 'product1_overlay': '1.0.0' }
  required = ["product1"]
  scripts = [ os.path.join(FIXTURE_PATH, 'product1-1.0.0', 'meta/defines.sh'),
              os.path.join(FIXTURE_PATH, 'product1_overlay-1.0.0', 'meta/defines.sh') ]

  check_parcel_results(parcels, scripts, parcels, required, [])

def test_parcel_inotify():
  """Test that inotify is picking up changes in the parcel directory"""

  global count
  count = 0

  def count_callouts(args):
    global count
    count += 1
    return 0

  def running_as_root():
    return True

  try:
    tmpdir = tempfile.mkdtemp()

    repo = MockableRepository(tmpdir, AGENT_PACKAGE_PATH)
    repo.set_callout(count_callouts)
    repo.set_running_as_root(running_as_root)
    repo.set_ensure_users_groups_permissions(True)
    repo.refresh()
    assert_equal(0, len(repo.products))

    inotify = repo.configure_inotify()

    _copy_parcel(tmpdir, "product1-1.0.0")
    _iterate_inotify(inotify, repo, 1, 1, 0)
    assert_equal(3, count)

    _copy_parcel(tmpdir, "product1-1.1.0")
    _iterate_inotify(inotify, repo, 1, 2, 0)
    assert_equal(7, count)

    _copy_parcel(tmpdir, "product2-2.0.0")
    _iterate_inotify(inotify, repo, 2, 2, 1)
    assert_equal(7, count)

    _copy_parcel(tmpdir, "product2-2.5.0")
    _iterate_inotify(inotify, repo, 2, 2, 2)
    assert_equal(10, count)

    _copy_parcel(tmpdir, "product2-3.0.0")
    _iterate_inotify(inotify, repo, 2, 2, 3)
    assert_equal(10, count)

    shutil.rmtree(os.path.join(tmpdir, "product1-1.0.0"))
    _iterate_inotify(inotify, repo, 2, 1, 3)
    assert_equal(10, count)

    shutil.rmtree(os.path.join(tmpdir, "product1-1.1.0"))
    _iterate_inotify(inotify, repo, 1, 0, 3)
    assert_equal(10, count)

    shutil.rmtree(os.path.join(tmpdir, "product2-2.0.0"))
    _iterate_inotify(inotify, repo, 1, 0, 2)
    assert_equal(10, count)

    shutil.rmtree(os.path.join(tmpdir, "product2-2.5.0"))
    _iterate_inotify(inotify, repo, 1, 0, 1)
    assert_equal(10, count)

    shutil.rmtree(os.path.join(tmpdir, "product2-3.0.0"))
    _iterate_inotify(inotify, repo, 0, 0, 0)
    assert_equal(10, count)

  finally:
    inotify.stop()
    shutil.rmtree(tmpdir)

def test_parcels_without_users_groups_permissions():
  """Test that users/groups/permissions application can be skipped"""

  global count
  count = 0

  def count_callouts(args):
    global count
    count += 1
    return 0

  try:
    tmpdir = tempfile.mkdtemp()

    repo = MockableRepository(tmpdir, AGENT_PACKAGE_PATH)
    repo.set_callout(count_callouts)
    repo.set_ensure_users_groups_permissions(False)
    repo.refresh()
    assert_equal(0, len(repo.products))

    inotify = repo.configure_inotify()

    _copy_parcel(tmpdir, "product1-1.0.0")
    _iterate_inotify(inotify, repo, 1, 1, 0)
    assert_equal(0, count)

    _copy_parcel(tmpdir, "product1-1.1.0")
    _iterate_inotify(inotify, repo, 1, 2, 0)
    assert_equal(0, count)

    _copy_parcel(tmpdir, "product2-2.0.0")
    _iterate_inotify(inotify, repo, 2, 2, 1)
    assert_equal(0, count)

    _copy_parcel(tmpdir, "product2-2.5.0")
    _iterate_inotify(inotify, repo, 2, 2, 2)
    assert_equal(0, count)

    _copy_parcel(tmpdir, "product2-3.0.0")
    _iterate_inotify(inotify, repo, 2, 2, 3)
    assert_equal(0, count)

    shutil.rmtree(os.path.join(tmpdir, "product1-1.0.0"))
    _iterate_inotify(inotify, repo, 2, 1, 3)
    assert_equal(0, count)

    shutil.rmtree(os.path.join(tmpdir, "product1-1.1.0"))
    _iterate_inotify(inotify, repo, 1, 0, 3)
    assert_equal(0, count)

    shutil.rmtree(os.path.join(tmpdir, "product2-2.0.0"))
    _iterate_inotify(inotify, repo, 1, 0, 2)
    assert_equal(0, count)

    shutil.rmtree(os.path.join(tmpdir, "product2-2.5.0"))
    _iterate_inotify(inotify, repo, 1, 0, 1)
    assert_equal(0, count)

    shutil.rmtree(os.path.join(tmpdir, "product2-3.0.0"))
    _iterate_inotify(inotify, repo, 0, 0, 0)
    assert_equal(0, count)

  finally:
    inotify.stop()
    shutil.rmtree(tmpdir)

def test_ensure_users_and_groups():
  '''Test that new parcel activation commands are correctly composed'''

  repo = make_new_repository()

  global count
  count = 0

  global running_as_root
  running_as_root = True

  def mock_running_as_root():
    return running_as_root

  repo.set_running_as_root(mock_running_as_root)

  if repo.is_suse:
    umask_arg = '-U'
    umask_param = '022'
  else:
    umask_arg = '-K'
    umask_param = 'UMASK=022'

  def validate_product1(args):
    expected_user1 = [
      '/usr/sbin/useradd',
      '-r', '-m',
      '-g', 'user1',
      umask_arg, umask_param,
      '--home', '/tmp/user1',
      '--comment', 'User One',
      '--shell', '/bin/false',
      '--groups', 'group1',
      'user1',
    ]

    expected_user1group = [
      '/usr/sbin/groupadd',
      '-r',
      'user1',
    ]

    expected_group1 = [
      '/usr/sbin/groupadd',
      '-r',
      'group1',
    ]

    if args[0] == '/usr/sbin/useradd':
      test_deep_equality.deep_eq(expected_user1, args, _assert=True)
    elif args[0] == '/usr/sbin/groupadd' and args[2] == 'user1':
      test_deep_equality.deep_eq(expected_user1group, args, _assert=True)
    elif args[0] == '/usr/sbin/groupadd' and args[2] == 'group1':
      test_deep_equality.deep_eq(expected_group1, args, _assert=True)
    else:
      assert 0, "Unknown Command"
    global count
    count += 1
    return 0

  repo.set_callout(validate_product1)
  repo.ensure_users_and_groups(repo['product1']['1.0.0'])
  assert_equal(3, count)

  def validate_product2(args):
    expected_user2 = [
      '/usr/sbin/useradd',
      '-r', '-m',
      '-g', 'user2',
      umask_arg, umask_param,
      '--home', '/tmp/user2',
      '--comment', 'User Two',
      '--shell', '/bin/false',
      'user2',
    ]

    expected_user2group = [
      '/usr/sbin/groupadd',
      '-r',
      'user2',
    ]

    expected_group2 = [
      '/usr/sbin/groupadd',
      '-r',
      'group2',
    ]

    if args[0] == '/usr/sbin/useradd':
      test_deep_equality.deep_eq(expected_user2, args, _assert=True)
    elif args[0] == '/usr/sbin/groupadd' and args[2] == 'user2':
      test_deep_equality.deep_eq(expected_user2group, args, _assert=True)
    elif args[0] == '/usr/sbin/groupadd' and args[2] == 'group2':
      test_deep_equality.deep_eq(expected_group2, args, _assert=True)
    else:
      assert 0, "Unknown Command"
    global count
    count += 1
    return 9

  repo.set_callout(validate_product2)
  repo.ensure_users_and_groups(repo['product2']['2.5.0'])
  assert_equal(6, count)

  running_as_root = False
  repo.ensure_users_and_groups(repo['product2']['2.5.0'])
  # No users/groups added in non-root configuration
  assert_equal(6, count)

def test_ensure_permissions():
  '''Test that file ownership/permission arguments are correctly parsed'''

  repo = make_new_repository()

  global yarn_path
  yarn_path = os.path.join(FIXTURE_PATH,
                           "product1-1.1.0",
                           "lib/hadoop-yarn/bin/container-executor")
  global mapred_path
  mapred_path = os.path.join(FIXTURE_PATH,
                             "product1-1.1.0",
                             "lib/hadoop-0.20-mapreduce/sbin/Linux-amd64-64/task-controller")

  global chown_count
  chown_count = 0

  global chmod_count
  chmod_count = 0

  global running_as_root
  running_as_root = True

  def validate_chown(path, user, group):
    global yarn_path
    global mapred_path
    global chown_count
    global running_as_root

    if path == yarn_path or path == mapred_path:
      assert_equal("root", user)
      if running_as_root:
        assert_equal("root", group)
      else:
        assert_equal(None, group)

    else:
      assert 0, "Unknown Path: %s" % path
    chown_count += 1

  def validate_chmod(path, perms):
    global yarn_path
    global mapred_path
    global chmod_count

    if path == yarn_path:
      assert_equal("6050", perms)
    elif path == mapred_path:
      assert_equal("2754", perms)
    else:
      assert 0, "Unknown Path %s" % path
    chmod_count += 1

  def mock_running_as_root():
    return running_as_root

  repo.set_chown(validate_chown)
  repo.set_chmod(validate_chmod)
  repo.set_running_as_root(mock_running_as_root)

  repo.ensure_permissions(repo['product1']['1.0.0'])
  assert_equal(0, chown_count)
  assert_equal(0, chmod_count)

  repo.ensure_permissions(repo['product1']['1.1.0'])
  assert_equal(2, chown_count)
  assert_equal(2, chmod_count)

  running_as_root = False
  repo.ensure_permissions(repo['product1']['1.1.0'])
  assert_equal(4, chown_count)
  assert_equal(4, chmod_count)

def test_ensure_varlog():
  '''Test that /var/log directory ownership/permission arguments are correctly parsed'''

  repo = make_new_repository()

  global foo_path
  foo_path = "/var/log/product2/foo"

  global bar_path
  bar_path = "/var/log/product2/bar"

  global chown_count
  chown_count = 0

  global chmod_count
  chmod_count = 0

  global makedirs_count
  makedirs_count = 0

  def validate_chown(path, user, group):
    global foo_path
    global bar_path
    global chown_count

    if path == foo_path or path == bar_path:
      assert_equal("root", user)
      assert_equal("root", group)
    else:
      assert 0, "Unknown Path: %s" % path
    chown_count += 1

  def validate_chmod(path, perms):
    global foo_path
    global bar_path
    global chmod_count

    if path == foo_path:
      assert_equal("0777", perms)
    elif path == bar_path:
      assert_equal("0750", perms)
    else:
      assert 0, "Unknown Path %s" % path
    chmod_count += 1

  def validate_makedirs(path):
    global foo_path
    global bar_path
    global makedirs_count

    if path != foo_path and path != bar_path:
      assert 0, "Unknown Path %s" % path
    makedirs_count += 1

  def mock_running_as_root():
    return True

  repo.set_running_as_root(mock_running_as_root)
  repo.set_chown(validate_chown)
  repo.set_chmod(validate_chmod)
  repo.set_makedirs(validate_makedirs)

  repo.ensure_log_directories(repo['product1']['1.0.0'])
  assert_equal(0, chown_count)
  assert_equal(0, chmod_count)
  assert_equal(0, makedirs_count)

  repo.ensure_log_directories(repo['product2']['2.0.0'])
  assert_equal(2, chown_count)
  assert_equal(2, chmod_count)
  assert_equal(2, makedirs_count)
  # the json has a third entry for /var/log/syslog which will be ignored as
  # that already exists

def test_configure_all_symlinks():
  '''Test logic for detecting newly added/removed symlink parcels'''

  repo = make_new_repository()
  # We're not testing the active parcel symlink, and don't want to modify
  # the checked-in repository
  repo.ensure_active_symlink = Mock()

  def validate_not_called(args):
    assert 0, "Alternatives script should never have been called: %s" % str(args)

  repo.set_callout(validate_not_called)

  repo.configure_all_symlinks(None, None, True, True)
  repo.configure_all_symlinks({}, None, True, True)
  repo.configure_all_symlinks(None, {}, True, True)
  repo.configure_all_symlinks({}, {}, True, True)
  repo.configure_all_symlinks({ 'foo': '1.0' }, { 'foo': '1.0' }, True, True)
  repo.configure_all_symlinks({ 'foo': '1.0', 'bar': '2.0' },
                              { 'foo': '1.0', 'bar': '2.0' }, True, True)

  def validate_activate(args):
    expected_activate = [
      repo.get_script_path(repo['product1']['1.0.0'], 'alternatives'),
      'activate',
    ]

    test_deep_equality.deep_eq(expected_activate, args, _assert=True)

  repo.set_callout(validate_activate)

  repo.configure_all_symlinks(None, { 'product1': '1.0.0' }, True, True)
  repo.configure_all_symlinks({}, { 'product1': '1.0.0' }, True, True)
  repo.configure_all_symlinks({ 'product2': '2.0.0' },
                              { 'product2': '2.0.0', 'product1': '1.0.0' },
                              True, True)

  def validate_deactivate(args):
    expected_deactivate = [
      repo.get_script_path(repo['product1']['1.0.0'], 'alternatives'),
      'deactivate',
    ]

    test_deep_equality.deep_eq(expected_deactivate, args, _assert=True)

  repo.set_callout(validate_deactivate)

  repo.configure_all_symlinks({ 'product1': '1.0.0' }, None, True, True)
  repo.configure_all_symlinks({ 'product1': '1.0.0' }, {}, True, True)
  repo.configure_all_symlinks({ 'product2': '2.0.0', 'product1': '1.0.0' },
                              { 'product2': '2.0.0' }, True, True)

  def validate_both(args):
    expected_activate = [
      repo.get_script_path(repo['product1']['1.0.0'], 'alternatives'),
      'activate',
    ]

    expected_deactivate = [
      repo.get_script_path(repo['product2']['2.0.0'], 'alternatives'),
      'deactivate',
    ]

    if 'activate' in args:
      test_deep_equality.deep_eq(expected_activate, args, _assert=True)
    else:
      test_deep_equality.deep_eq(expected_deactivate, args, _assert=True)

  repo.set_callout(validate_both)

  repo.configure_all_symlinks({ 'product2': '2.0.0' },
                              { 'product1': '1.0.0' }, True, True)

def test_configure_symlinks():
  '''Test that new parcel symlink commands are correctly composed
     Note that this tests configure_symlinks through the fallback from
     ensure_alternatives - so we're also testing that the fallback works.
  '''

  repo = make_new_repository()
  # We're not testing the active parcel symlink; it should not be called.
  repo.ensure_active_symlink = Mock()

  def validate_activate(args, env):
    assert_equal(repo.path, env['PARCELS_ROOT'])

    expected_activate = [
      repo.get_script_path(repo['product1']['1.0.0'], 'alternatives'),
      'activate',
    ]

    test_deep_equality.deep_eq(expected_activate, args, _assert=True)
    return 0

  repo.set_callout(validate_activate)
  repo.ensure_alternatives(repo['product1']['1.0.0'], True)
  repo.ensure_active_symlink.assert_called_never()

  def validate_deactivate(args, env):
    assert_equal(repo.path, env['PARCELS_ROOT'])

    expected_activate = [
      repo.get_script_path(repo['product1']['1.0.0'], 'alternatives'),
      'deactivate',
    ]

    test_deep_equality.deep_eq(expected_activate, args, _assert=True)
    return 0

  repo.set_callout(validate_deactivate)
  repo.ensure_active_symlink.reset_mock()
  repo.ensure_alternatives(repo['product1']['1.0.0'], False)
  repo.ensure_active_symlink.assert_called_never()

def test_ensure_alternatives():
  '''Test that the declarative alternatives commands are correctly composed'''
  repo = make_new_repository()
  # We're not testing the active parcel symlink; it should not be called.
  repo.ensure_active_symlink = Mock()
  expected_activate = [
    'beeline',
    '/usr/bin/beeline',
    os.path.join(repo.path, 'product2-3.0.0', 'bin/beeline'),
    10,
    False
  ]

  repo.ensure_alternatives(repo['product2']['3.0.0'], True)
  repo.osops.alternatives.activate.assert_called_with(*expected_activate)
  repo.osops.alternatives.deactivate.assert_called_never()
  repo.ensure_active_symlink.assert_called_never()
  expected_deactivate = [
    'beeline',
    '/usr/bin/beeline',
    os.path.join(repo.path, 'product2-3.0.0', 'bin/beeline'),
    10,
    False
  ]

  repo.ensure_active_symlink.reset_mock()
  repo.ensure_alternatives(repo['product2']['3.0.0'], False)
  repo.osops.alternatives.deactivate.assert_called_with(*expected_deactivate)
  repo.osops.alternatives.activate.assert_called_never()
  repo.ensure_active_symlink.assert_called_never()

def _copy_parcel(tmpdir, name):
  shutil.copytree(os.path.join(FIXTURE_PATH, name),
                  os.path.join(tmpdir, name))

def _iterate_inotify(inotify, repo, count, p1count, p2count):
  assert_true(inotify.check_events())
  inotify.read_events()
  inotify.process_events()
  assert_equal(count, len(repo.products))
  if p1count > 0:
    assert_equal(p1count, len(repo['product1']))
  else:
    assert_false('product1' in repo.products)
  if p2count > 0:
    assert_equal(p2count, len(repo['product2']))
  else:
    assert_false('product2' in repo.products)

def test_avro_validation():
  '''Test that a non-compliant parcel.json fails validation'''

  repo = MockableRepository(FIXTURE_PATH, AGENT_PACKAGE_PATH)
  manifest = json.load(open(os.path.join(FIXTURE_PATH, 'product1-0.9.0', 'meta', 'parcel.json')))
  assert_false(repo._validate(manifest))

  manifest = json.load(open(os.path.join(FIXTURE_PATH, 'product1-1.0.0', 'meta', 'parcel.json')))
  assert_true(repo._validate(manifest))

def test_identify_available_parcels():
  '''Test identify_available_parcels'''

  try:
    tmpdir = tempfile.mkdtemp()
    tmprepo = os.path.join(tmpdir, os.path.basename(FIXTURE_PATH))

    shutil.copytree(FIXTURE_PATH, tmprepo, symlinks=True)

    repo = make_new_repository(tmprepo)

    active_parcels = { 'product1': '1.0.0',
                       'product2': '3.0.0', }

    available, missing = repo.identify_available_parcels(active_parcels)
    assert_equal(0, len(missing))

    # Missing parcel
    active_parcels = { 'product1': '1.0.0',
                       'product5': '3.0.0', }

    available, missing = repo.identify_available_parcels(active_parcels)
    assert_equal(1, len(available))
    assert_equal(1, len(missing))

    # Missing parcel version
    active_parcels = { 'product1': '1.0.0',
                       'product2': '5.0.0', }

    available, missing = repo.identify_available_parcels(active_parcels)
    assert_equal(1, len(available))
    assert_equal(1, len(missing))

  finally:
    shutil.rmtree(tmpdir)


def test_ensure_active_symlink():
  '''Test ensure_active_symlink'''

  try:
    tmpdir = tempfile.mkdtemp()
    tmprepo = os.path.join(tmpdir, os.path.basename(FIXTURE_PATH))

    shutil.copytree(FIXTURE_PATH, tmprepo, symlinks=True)

    repo = make_new_repository(tmprepo)

    product1_link = os.path.join(tmprepo, "product1")
    product2_link = os.path.join(tmprepo, "product2")

    assert_true(os.path.islink(product1_link))
    assert_equal("product1-1.1.0", os.readlink(product1_link))
    assert_true(os.path.islink(product2_link))
    assert_equal("product2-2.5.0", os.readlink(product2_link))

    repo.ensure_active_symlink(repo['product1']['1.0.0'], True)
    repo.ensure_active_symlink(repo['product2']['3.0.0'], True)

    assert_true(os.path.islink(product1_link))
    assert_equal("product1-1.0.0", os.readlink(product1_link))
    assert_true(os.path.islink(product2_link))
    assert_equal("product2-3.0.0", os.readlink(product2_link))

    repo.ensure_active_symlink(repo['product1']['1.1.0'], True)
    repo.ensure_active_symlink(repo['product2']['2.0.0'], True)

    assert_true(os.path.islink(product1_link))
    assert_equal("product1-1.1.0", os.readlink(product1_link))
    assert_true(os.path.islink(product2_link))
    assert_equal("product2-2.0.0", os.readlink(product2_link))

    os.remove(product1_link)
    os.remove(product2_link)
    assert_false(os.path.exists(product1_link))
    assert_false(os.path.exists(product2_link))

    repo.ensure_active_symlink(repo['product1']['1.1.0'], True)
    repo.ensure_active_symlink(repo['product2']['2.0.0'], True)

    assert_true(os.path.islink(product1_link))
    assert_equal("product1-1.1.0", os.readlink(product1_link))
    assert_true(os.path.islink(product2_link))
    assert_equal("product2-2.0.0", os.readlink(product2_link))

    # Test that a directory masquerading as a symlink doesn't get
    # clobbered by a new symlink
    os.remove(product1_link)
    os.mkdir(product1_link)
    assert_false(os.path.islink(product1_link))

    repo.ensure_active_symlink(repo['product1']['1.1.0'], True)

    assert_false(os.path.islink(product1_link))

    # Test that a directory that mysteriously replaced a symlink
    # doesn't get deleted when deactivating the symlink
    repo.ensure_active_symlink(repo['product1']['1.1.0'], False)

    assert_false(os.path.islink(product1_link))
    assert_true(os.path.isdir(product1_link))

    os.rmdir(product1_link)

    # Test that our other symlink is still around after all that.
    assert_true(os.path.islink(product2_link))
    assert_equal("product2-2.0.0", os.readlink(product2_link))

    # Test that activating a parcel which doesn't want a symlink
    # results in there being no symlink
    repo.ensure_active_symlink(repo['product1']['1.1.0'], True)
    repo.ensure_active_symlink(repo['product2']['2.5.0'], True)

    assert_true(os.path.islink(product1_link))
    assert_equal("product1-1.1.0", os.readlink(product1_link))
    assert_false(os.path.exists(product2_link))

  finally:
    shutil.rmtree(tmpdir)

def test_delete_parcel():
  '''Test delete parcel'''

  tmpdir = tempfile.mkdtemp()
  try:
    tmprepo = os.path.join(tmpdir, os.path.basename(FIXTURE_PATH))
    shutil.copytree(FIXTURE_PATH, tmprepo, symlinks=True)

    repo = MockableRepository(tmprepo, AGENT_PACKAGE_PATH)
    parcel_home = 'product2-2.0.0'
    assert_true(os.path.exists(os.path.join(tmprepo, parcel_home)))
    repo.delete('product2', '2.0.0')
    assert_false(os.path.exists(os.path.join(tmprepo, parcel_home)))
  finally:
    shutil.rmtree(tmpdir)

def test_dont_delete_parcel():
  '''Test non-removal of files of parcel marked with .dont_delete'''

  tmpdir = tempfile.mkdtemp()
  try:
    tmprepo = os.path.join(tmpdir, os.path.basename(FIXTURE_PATH))
    shutil.copytree(FIXTURE_PATH, tmprepo, symlinks=True)

    repo = MockableRepository(tmprepo, AGENT_PACKAGE_PATH)
    parcel_home = 'product2-2.0.0'
    assert_true(os.path.exists(os.path.join(tmprepo, parcel_home)))
    marker_file = os.path.join(tmprepo, parcel_home, parcel.DONT_DELETE)
    open(marker_file, "w").close()
    repo.delete('product2', '2.0.0')
    # The parcel contents should still be there...
    assert_true(os.path.exists(os.path.join(tmprepo, parcel_home)))
    os.unlink(marker_file)

    repo.delete('product2', '2.0.0')
    assert_false(os.path.exists(os.path.join(tmprepo, parcel_home)))

  finally:
    shutil.rmtree(tmpdir)

def test_parcel_added():
  '''Test parcel_added callout'''

  tmpdir = tempfile.mkdtemp()
  try:
    tmprepo = os.path.join(tmpdir, os.path.basename(FIXTURE_PATH))
    os.mkdir(tmprepo)
    repo = MockableRepository(tmprepo, AGENT_PACKAGE_PATH)
    repo.set_callout(mock_call)
    repo.set_chown(mock_call)
    repo.set_chmod(mock_call)
    repo.set_makedirs(mock_call)
    repo.set_parcel_added(Mock())

    # empty repo
    repo.refresh()
    assert_false(repo.parcel_added.called)

    # copy fixture over and refresh
    os.rmdir(tmprepo)
    shutil.copytree(FIXTURE_PATH, tmprepo, symlinks=True)
    repo.refresh()
    assert_equal(6, repo.parcel_added.call_count)
    repo.parcel_added.assert_any_called('product1', '1.0.0')
    repo.parcel_added.assert_any_called('product1', '1.1.0')
    repo.parcel_added.assert_any_called('prodoct1_overlay', '1.0.0')
    repo.parcel_added.assert_any_called('product2', '2.0.0')
    repo.parcel_added.assert_any_called('product2', '2.5.0')
    repo.parcel_added.assert_any_called('product2', '3.0.0')
  finally:
    shutil.rmtree(tmpdir)

def test_errors():
  '''Test that parcel loading errors are tracked.'''

  tmpdir = tempfile.mkdtemp()
  try:
    tmprepo = os.path.join(tmpdir, os.path.basename(FIXTURE_PATH))
    shutil.copytree(FIXTURE_PATH, tmprepo, symlinks=True)

    # Mess up product1 1.0.0's manifest.
    manifest = os.path.join(tmprepo, "product1-1.0.0/meta/parcel.json")
    file = open(manifest, 'w')
    file.write("foobar")
    file.close()

    # Random garbage
    os.mkdir(os.path.join(tmprepo, "lost+found"))

    repo = make_new_repository(path=tmprepo)

    # Invalid JSON.
    parcel_err = repo.errors['product1']['1.0.0']
    assert_equal(parcel.JSON_ERROR_CODE, parcel_err[parcel.ERROR_CODE_FIELD])

    # Non-compliant manifest.
    parcel_err = repo.errors['product1']['0.9.0']
    assert_equal(parcel.NON_COMP_ERROR_CODE, parcel_err[parcel.ERROR_CODE_FIELD])

    # Unrelated directory
    parcel_err = repo.errors['lost+found']['']
    assert_equal(parcel.UNEXPECTED_ERROR_CODE, parcel_err[parcel.ERROR_CODE_FIELD])
  finally:
    shutil.rmtree(tmpdir)

def test_get_component_info():
  '''Test component info for version detection'''

  repo = make_new_repository()

  prod1_components = repo.get_components_for_active_parcel('product1', '1.1.0')
  assert_equal(3, len(prod1_components))
  for component in prod1_components:
    assert_equal('NOT_APPLICABLE', component['cdh_version'])
    assert_equal('product1 1.1.0', component['cdh_release'])

  assert_equal(0, len(repo.get_components_for_active_parcel('product2', '2.5.0')))

def test_topological_sort():
  '''Test parcel topological sorting.'''
  repo = make_new_repository()

  def all_perms(elements):
    '''Generate all permutations of a list of elements.

       From http://stackoverflow.com/a/104436.
    '''
    if len(elements) <= 1:
      yield elements
    else:
      for perm in all_perms(elements[1:]):
        for i in range(len(elements)):
          # nb elements[0:1] works in both string and list contexts
          yield perm[:i] + elements[0:1] + perm[i:]

  def run_all_perms(expected, parcels):
    '''Sort and test all possibly permutations of input. Because the
       topological sort is stable, every permutation should yield the exact
       same output.'''
    for perm in all_perms(parcels):
      assert_equal(expected, repo.topological_sort(perm))

  # Boring edge cases.
  assert_equal([], repo.topological_sort([]))
  p = {'name' : 'asdf'}
  assert_equal([p], repo.topological_sort([p]))

  # Unsorted input, no dependencies.
  pa = {'name' : 'a'}
  pb = {'name' : 'b'}
  pc = {'name' : 'c'}
  run_all_perms([pa, pb, pc], [pa, pb, pc])

  # a->c and b->c.
  pa = {'name' : 'a', 'depends' : 'c'}
  pb = {'name' : 'b', 'depends' : 'c'}
  pc = {'name' : 'c'}
  run_all_perms([pc, pa, pb], [pa, pb, pc])

  # a->b->c.
  pa = {'name' : 'a', 'depends' : 'b'}
  pb = {'name' : 'b', 'depends' : 'c'}
  pc = {'name' : 'c'}
  run_all_perms([pc, pb, pa], [pa, pb, pc])

  # a->b and a->c.
  pa = {'name' : 'a', 'depends' : 'b, c'}
  pb = {'name' : 'b'}
  pc = {'name' : 'c'}
  run_all_perms([pb, pc, pa], [pa, pb, pc])

  # a->b and b->c and b->d. With some extraneous versioning information.
  pa = {'name' : 'a', 'depends' : 'b (>= 5.0)'}
  pb = {'name' : 'b', 'depends' : 'd (!= 3)'}
  pc = {'name' : 'c'}
  pd = {'name' : 'd'}
  run_all_perms([pc, pd, pb, pa], [pa, pb, pc, pd])

  # a->b, with b showing up multiple times.
  pa = {'name' : 'a', 'depends' : 'b (>= 4.1), b (<< 5.0)' }
  pb = {'name' : 'b'}
  run_all_perms([pb, pa], [pa, pb])

  # Dependency cycle.
  pa = {'name' : 'a', 'depends' : 'b' }
  pb = {'name' : 'b', 'depends' : 'a' }
  assert_raises(ValueError, run_all_perms, None, [pa, pb])

  # Missing dependencies.
  pa = {'name' : 'a', 'depends' : 'b' }
  run_all_perms([pa], [pa])

  # Duplicated parcels.
  pa = {'name' : 'a' }
  pb = {'name' : 'a' }
  assert_raises(AssertionError, run_all_perms, None, [pa, pb])

def test_get_unversioned_dependencies():
  '''Test parsing of (unversioned) dependencies from parcel metadata.'''
  repo = make_new_repository()

  p = dict()
  assert_equal(set(), repo._get_unversioned_dependencies(p))

  p = dict(depends="")
  assert_equal(set(), repo._get_unversioned_dependencies(p))

  for dep in ["CDH",
              "CDH (<< 5.3.0)",
              "CDH (!= 5.3.0), CDH (!= 5.2.0)",
              "CDH followed by garbage",
              ",,CDH,,"]:
    p = dict(depends=dep)
    assert_equal(set(["CDH"]), repo._get_unversioned_dependencies(p))

  p = dict(depends="CDH, AWESOME, OTHER_GUY")
  assert_equal(set(["CDH", "AWESOME", "OTHER_GUY"]),
               repo._get_unversioned_dependencies(p))

  p = dict(depends=",CDH,,OTHER_GUY,,")
  assert_equal(set(["CDH", "OTHER_GUY"]),
               repo._get_unversioned_dependencies(p))
