#!/usr/bin/env python
# Copyright (c) 2012 Cloudera, Inc. All rights reserved.

import os
import shutil
import simplejson as json
import tarfile
import tempfile

from sets import Set

from nose.tools import assert_true, assert_equal, assert_false, assert_raises
from mock import Mock

import cmf.util
from cmf import downloader, parcel, parcel_cache
from cmf.parcel_id import ParcelId

FIXTURE_PATH = os.path.join(os.path.dirname(__file__), "test-fixtures", "parcel-repository-fixture")
AGENT_PACKAGE_PATH = '/cmf/agent/services' # Doesn't need to be a real path

def test_update_delete_unmanaged_parcel():
  '''Test that parcels no longer managed by server are deleted'''

  repo = parcel.Repository(FIXTURE_PATH, AGENT_PACKAGE_PATH)
  products = {}
  products['product1'] = dict()
  products['product1']['1.0.0'] = None
  products['product1']['2.0.0'] = None
  repo.products = products

  dl = Mock()
  dl.get_current_downloads.return_value = []
  cache = parcel_cache.ParcelCache(None, repo, dl)
  cache.delete_unmanaged = Mock()
  managed_parcels = [
    'http://foo.bar/baz/product1-1.0.0-rhel.parcel',
    'http://foo.bar/baz/product1-2.1.0-rhel.parcel',
  ]
  cache.update(managed_parcels, True)
  cache.delete_unmanaged.assert_called_once_with('product1', '2.0.0')

def test_update_delete_invalid_parcel():
  '''Test that invalid parcels no longer managed by server are deleted'''

  repo = parcel.Repository(FIXTURE_PATH, AGENT_PACKAGE_PATH)
  products = {}
  products['product1'] = dict()
  products['product1']['1.0.0'] = None
  products['product1']['2.0.0'] = None
  repo.products = products
  invalid_products = Set()
  invalid_products.add(('product1', '0.9.0'))
  repo.invalid_products = invalid_products

  dl = Mock()
  dl.get_current_downloads.return_value = []
  cache = parcel_cache.ParcelCache(None, repo, dl)
  cache.delete_unmanaged = Mock()
  managed_parcels = [
    'http://foo.bar/baz/product1-0.9.0-el6.parcel',
    'http://foo.bar/baz/product1-1.0.0-el6.parcel',
    'http://foo.bar/baz/product1-2.0.0-el6.parcel',
  ]
  cache.update(managed_parcels, False)
  assert_false(cache.delete_unmanaged.called, False)

  managed_parcels = [
    'http://foo.bar/baz/product1-1.0.0-el6.parcel',
    'http://foo.bar/baz/product1-2.0.0-el6.parcel',
  ]
  cache.update(managed_parcels, True)
  cache.delete_unmanaged.assert_called_once_with('product1', '0.9.0')

def test_update_download_managed_parcel():
  '''Test that new parcels managed by server are downloaded'''

  tmp = tempfile.mkdtemp()
  try:
    repo = parcel.Repository(tmp, AGENT_PACKAGE_PATH)
    products = {}
    products['product1'] = dict()
    products['product1']['1.0.0'] = None
    products['product1']['2.0.0'] = None
    repo.products = products
  
    dl = Mock()
    dl.get_current_downloads.return_value = []
    cache = parcel_cache.ParcelCache(None, repo, dl)
    cache.queue_parcel_download = Mock()
    managed_parcels = [
      'http://foo.bar/baz/product1-1.0.0-rhel.parcel',
      'http://foo.bar/baz/product1-2.0.0-rhel.parcel',
      'http://foo.bar/baz/product1-2.1.0-rhel.parcel',
      'http://foo.bar/baz/product2-1.0.0-rhel.parcel',
    ]
    cache.update(managed_parcels, True)
    cache.queue_parcel_download.assert_any_call('http://foo.bar/baz/product2-1.0.0-rhel.parcel', tmp, True)
    cache.queue_parcel_download.assert_any_call('http://foo.bar/baz/product1-2.1.0-rhel.parcel', tmp, True)
    cache.queue_parcel_download.reset_mock()
    
    # Fake unpacked dir for 2-1.0.0, so download should not be queued to allow
    # time for parcel to be loaded.
    pid = ParcelId.url(managed_parcels[3])
    product2_1_dir = os.path.join(tmp, pid.get_parcel_dir())
    os.mkdir(product2_1_dir)
    cache.update(managed_parcels, False)
    cache.queue_parcel_download.assert_called_once_with('http://foo.bar/baz/product1-2.1.0-rhel.parcel', tmp, False)
  finally:  
    shutil.rmtree(tmp)

def test_update_abort_downloads_of_unmanaged_parcels():
  '''Test that downloads of unmanaged parcels are aborted -- cancel scenario'''

  repo = parcel.Repository(FIXTURE_PATH, AGENT_PACKAGE_PATH)
  products = {}
  repo.products = products
  url = 'http://foo.bar/baz/product1-1.0.0-rhel.parcel'
  dl = Mock()
  dl_entry = dict()
  dl_entry[downloader.URL_FIELD] = url
  dl_entry[downloader.TOTAL_BYTES_FIELD] = 100
  dl_entry[downloader.DOWNLOADED_BYTES_FIELD] = 57
  dl.get_current_downloads.return_value = [dl_entry]
  cache = parcel_cache.ParcelCache(None, repo, dl)

  cache.update([], True)
  dl.abort.assert_called_with(url)

def test_delete_unmanaged():
  '''Test that managed parcels are deleted from both cache and repo'''

  tmp = tempfile.mkdtemp()

  # create some fake parcels in cache
  _create_parcel(tmp, 'product1-1.0.0-rhel.parcel')
  _create_parcel(tmp, 'product2-1.0.0-rhel.parcel')
  _create_parcel(tmp, 'product2-1.0.0-1.1.0-rhel.diff')
  _create_parcel(tmp, 'product2-1.0.0-foo.parcel')
  _create_parcel(tmp, 'product2-1.1.0-rhel.parcel')

  repo = Mock()
  cache = parcel_cache.ParcelCache(tmp, repo, Mock())
  cache.delete_unmanaged('product2', '1.1.0')
  assert_true(os.path.exists(os.path.join(tmp, 'product1-1.0.0-rhel.parcel')))
  assert_true(os.path.exists(os.path.join(tmp, 'product2-1.0.0-rhel.parcel')))
  assert_false(os.path.exists(os.path.join(tmp, 'product2-1.0.0-1.1.0-rhel.diff')))
  assert_true(os.path.exists(os.path.join(tmp, 'product2-1.0.0-foo.parcel')))
  assert_false(os.path.exists(os.path.join(tmp, 'product2-1.1.0-rhel.parcel')))
  repo.delete.assert_called_once_with('product2', '1.1.0')

  shutil.rmtree(tmp)

def _create_parcel(dir, filename):
  path = os.path.join(dir, filename)
  f = open(path, 'w')
  f.close()

def _tar_parcel(dir, output):
  ''' Tars everything under dir into a tarball with basename as archive name'''
  tar = tarfile.open(output, 'w')
  for file in os.listdir(dir):
    tar.add(os.path.join(dir, file), os.path.join(os.path.basename(dir), file))
  tar.close()

def test_queue_parcel_download_keep():
  '''Test that parcel downloads unpack as a part of callback (with cache retention)'''
  _queue_parcel_download(True)

def test_queue_parcel_download_no_keep():
  '''Test that parcel downloads unpack as a part of callback (without cache retention)'''
  _queue_parcel_download(False)

def _queue_parcel_download(keep_parcel_in_cache):
  # We actually test untar with a real tarball.

  product_dir = "product1-1.0.0"
  filename = "%s-deb.parcel" % (product_dir)
  parcel_fixture_dir = os.path.join(FIXTURE_PATH, product_dir)

  tmp = tempfile.mkdtemp()
  try:
    unpack_dir = os.path.join(tmp, 'unpack')
    os.mkdir(unpack_dir)

    class MockDownloader(object):

      def __init__(self, correct_checksum=True):
        self.correct_checksum = correct_checksum

      def queue_download(self, url, callback=None, overwrite=False, proceed=False):
        # tars parcel fixture into dir as if it's downloaded
        purl = ParcelId.url(url)
        tar_name = os.path.join(tmp, purl.get_filename())
        _tar_parcel(parcel_fixture_dir, tar_name)
        op = downloader.Download(url)
        op.state = downloader.DOWNLOADED
        op.path = tar_name
        if self.correct_checksum:
          op.headers = { parcel_cache.PARCEL_CHECKSUM_HEADER: cmf.util.sha1(tar_name) }
        else:
          op.headers = { parcel_cache.PARCEL_CHECKSUM_HEADER: 'garbage' }
        callback(url, op)
        return True

    patch = Mock()
    repo = Mock()
    cache = parcel_cache.ParcelCache(tmp, repo, MockDownloader(), patch)
    parcel_url = "https://archive.cloudera.com/%s" % (filename)
    cache.queue_parcel_download(parcel_url, unpack_dir, keep_parcel_in_cache)
    compare_dir(parcel_fixture_dir, os.path.join(unpack_dir, product_dir))

    # Make sure an exception is thrown if checksums don't match.
    cache = parcel_cache.ParcelCache(tmp, repo, MockDownloader(False), patch)
    assert_raises(Exception, cache.queue_parcel_download, parcel_url, unpack_dir)

    assert_equal(keep_parcel_in_cache, os.path.exists(os.path.join(tmp, filename)))

  finally:
    shutil.rmtree(tmp)

def test_queue_diff_download():
  '''Test that diff downloads unpack and patch'''

  base_filename = "product1-1.0.0-deb.parcel"
  diff_filename = "product1-1.0.0-1.0.1-deb.diff"

  tmp = tempfile.mkdtemp()
  try:
    _create_parcel(tmp, base_filename)
    unpack_dir = os.path.join(tmp, 'unpack')

    class MockDownloader(object):
      def queue_download(self, url, callback=None, overwrite=False, proceed=False):
        op = downloader.Download(url)
        op.state = downloader.DOWNLOADED
        op.path = os.path.join(tmp, diff_filename)
        callback(url, op)
        return True

    patch = Mock()
    repo = Mock()
    cache = parcel_cache.ParcelCache(tmp, repo, MockDownloader(), patch)
    url = "https://archive.cloudera.com/%s" % (diff_filename)
    cache.queue_parcel_download(url, unpack_dir, True)
    patch.assert_called_once_with(os.path.join(tmp, base_filename),
                                  os.path.join(tmp, diff_filename),
                                  unpack_dir,
                                  cache.path)
  finally:
    shutil.rmtree(tmp)

def compare_dir(this, that):
  '''Compare all files within and recursively compare subdirs'''
  assert_true(os.path.exists(this))
  assert_true(os.path.exists(that))
  for file in os.listdir(this):
    path = os.path.join(this, file)
    if os.path.isdir(path):
      compare_dir(path, os.path.join(that, file))
    else:
      compare_file(path, os.path.join(that, file))

def compare_file(this, that):
  '''Only compare file size'''
  assert_true(os.path.exists(this))
  assert_true(os.path.exists(that))
  this_stat = os.stat(this)
  that_stat = os.stat(that)
  assert_equal(this_stat.st_size, that_stat.st_size)
