# Copyright 2017, Cloudera Inc.

from test.test_support import run_unittest
import unittest

from cmf.process import Process
from nose.tools import assert_equal, assert_true
from mock import Mock
from mock import patch
from cmf.util.os_ops import OSOps

class ProcessTests(unittest.TestCase):


  def test_install_file_resource(self):
    class TestProcess(Process):
      def __init__(self):
        pass
      dir = "/some/dir"
      name = "test_process"
      osops = Mock()
      known_resources_files = {"{{CMF_CONF_DIR}}/passwd": "/etc/passwd"}
      raw = {"resources":
        [{"dynamic": "false",
          "install":{
            "source": "{{CMF_CONF_DIR}}/passwd",
            "path": "/tmp/foo",
            "user": "someuser",
            "group": "somegroup",
            "mode": 0600,
        }}]}
    test_proc = TestProcess()
    test_proc.evaluate_resources(True)
    test_proc.osops.install_file.assert_called_once_with('/tmp/foo',
                                                         '/etc/passwd',
                                                         'someuser',
                                                         'somegroup',
                                                         0600)

  def test_get_named_cgroups(self):
    """
    Check that all named cgroups are extracted from heartbeat response
    message.
    """
    class TestProcess(Process):
      def __init__(self):
        class FakeAgent:
          def __init__(self):
            self.cg_manager = Mock()
        self.agent = FakeAgent()
      raw = {"resources":
             [{"dynamic": "false",
               "named_cpu": {
                 "name": "my_proc1",
                 "user": "user1",
                 "group": "group1",
                 "mode": 0600,
                 }
               },
              {"dynamic": "false",
               "named_resource": {
                 "type": "cpu",
                 "name": "my_proc2",
                 "user": "user2",
                 "group": "group2",
                 "mode": 0600,
                 }
               },
              {"dynamic": "false",
               "named_resource": {
                 "type": "devices",
                 "name": "my_proc3",
                 "user": "user3",
                 "group": "group3",
                 "mode": 0600,
                 }
              }]}
    test_proc = TestProcess()
    test_proc.agent.cg_manager.get_group_paths.return_value = ["/some/path"]

    named_cgroups = test_proc.get_named_cgroups()
    assert_equal(len(named_cgroups), 3,
                 "Unexpected number of cgroups created: %d" % len(named_cgroups))
    checkusers = []
    for c in named_cgroups:
      checkusers.append(c["user"])
    assert_true("user1" in checkusers, "cgroups named_cpu not accepted.")
    assert_true("user2" in checkusers, "cgroups type 'cpu' not accepted.")
    assert_true("user3" in checkusers, "cgroups type 'devices' not accepted.")

  @patch("cmf.audit.util.make_audit_plugin")
  @patch("cmf.audit.util.make_metadata_plugin")
  def test_make_status(self, metadata_plugin, audit_plugin):
    """
    Ensure that Process.make_status will complete even without complete data.
    """

    metadata_plugin.return_value = None
    audit_plugin.return_value = None
    class TestProcess(Process):
      def __init__(self):
        agent = Mock()
        agent.configure_mock(**{"cfg.paths.process_dir": "/foo",
                                "cfg.paths.supervisor_include_path": "/super"})
        super(TestProcess, self).__init__(agent.cfg, agent, {"id": 0, "name": "foo"}, None)
    test_proc = TestProcess()
    test_proc.make_status()

  def test_expected_exit_codes(self):
    """
    Ensure that Process._get_expected_exitcodes() works fine even when exit code is not
    supplied by CM server
    """
    class TestProcess(Process):
      def __init__(self, exit_codes):
        self.raw = dict()
        self.raw["expected_exitcodes"] = exit_codes

    test_proc = TestProcess(None)
    actual_exit_codes = test_proc._get_expected_exitcodes()
    expected_exit_codes = set(Process.SUPERVISOR_EXPECTED_EXITCODES)
    assert_equal(expected_exit_codes, actual_exit_codes,
                 "Exit codes don't match when expected_exitcodes field is None")

    test_proc = TestProcess([1])
    actual_exit_codes = test_proc._get_expected_exitcodes()
    expected_exit_codes.add(1)
    assert_equal(expected_exit_codes, actual_exit_codes,
                 "Exit codes don't match when expected_exitcodes fields is not None")


def test_main():
  run_unittest(ProcessTests)

if __name__ == "__main__":
  test_main()
