#!/usr/bin/env python
# Copyright (c) 2014 Cloudera, Inc. All rights reserved.
# For testing the old orphan process config directory cleanup.

import os
import shutil
import tempfile
import time

import cmf.config
from cmf.agent import Agent
from cmf.process import Process
from mock import MagicMock
from nose.tools import assert_equal
from nose.tools import assert_false
from nose.tools import assert_true
import cmf.main
sub_dir_name_one = "foobar"
process_dir_name_one = "%s-%s" % (1, sub_dir_name_one)
sub_dir_name_two = "hello"
process_dir_name_two = "%s-%s" % (2, sub_dir_name_two)
now = time.time()
timestamp_offset = 180 # 3 minutes

# Create the vanilla test environment
def setup_vanilla_test_environment():
  # Create the Agent instant
  agent = Agent(cmf.config.AgentConfig())
  Process.ClassInit(agent.os_ops, agent.cfg)
  agent.client_configs = MagicMock()
  agent.client_configs.get_active_deployments = MagicMock(return_value=[])
  agent.cfg.paths.process_dir = tempfile.mkdtemp()
  agent.cfg.paths.supervisor_include_path = tempfile.mkdtemp()
  agent.cfg.orphan_process_directory_staleness_threshold = timestamp_offset

  # Create process directories
  conf_dir_foobar = os.path.join(agent.cfg.paths.process_dir,
                                 process_dir_name_one)
  os.mkdir(conf_dir_foobar, 0755)

  conf_dir_hello = os.path.join(agent.cfg.paths.process_dir,
                                process_dir_name_two)
  os.mkdir(conf_dir_hello, 0755)

  # Create files under the subdirectories
  for i in range(5):
    file_foobar = os.path.join(conf_dir_foobar, str(i))
    open(file_foobar, 'a').close()

  for i in range(2):
    file_hello = os.path.join(conf_dir_hello, str(i))
    open(file_hello, 'a').close()

  return agent


def cleanup_test_environment(agent):
  # Cleanup the test environment
  Process.Proc_Map.clear()
  shutil.rmtree(agent.cfg.paths.process_dir)
  shutil.rmtree(agent.cfg.paths.supervisor_include_path)

cleanup_test_environment.__test__ = False


def get_marker(dir, agent):
  conf_dir_path = os.path.join(agent.cfg.paths.process_dir, dir)
  marker = os.path.join(conf_dir_path, "process_timestamp")
  return marker

# Test case 1
# No marker in the subdirectory. A marker will be added with the timestamp at
# 'now'
def test_write_marker_to_config_dir():
  agent = setup_vanilla_test_environment()
  try:

    marker = get_marker(process_dir_name_one, agent)
    has_marker = os.path.isfile(marker)
    assert_false(has_marker)

    Process.mark_orphan(process_dir_name_one, now)

    has_marker = os.path.isfile(marker)
    assert_true(has_marker)
  finally:
    cleanup_test_environment(agent)


# Test case 2
# There is an existing marker in the subdirectory. The marker will not be
# modified.
def test_write_marker_twice():
  agent = setup_vanilla_test_environment()
  try:
    Process.mark_orphan(process_dir_name_one, now)

    marker = get_marker(process_dir_name_one, agent)

    has_marker = os.path.isfile(marker)
    assert_true(has_marker)

    Process.mark_orphan(process_dir_name_one, now)

    # Check the timestamp of the marker has not been changed
    file = open(marker, 'r')
    marker_time = float(file.readline())
    file.close()

    assert_equal(round(abs(marker_time - now)), 0)
  finally:
    cleanup_test_environment(agent)


# Test case 3
# Base on the initial timestamp of the marker file and check the staleness.
# Since it was modified to 'now', the marker is not stale. Modify the timestamp
# of the marker so that it makes the conf subdirectory is old enough
def test_conf_dir_staleness():
  agent = setup_vanilla_test_environment()
  try:
    Process.mark_orphan(process_dir_name_one, now)

    marker = get_marker(process_dir_name_one, agent)
    has_marker = os.path.isfile(marker)
    assert_true(has_marker)

    conf_dir_path = os.path.join(agent.cfg.paths.process_dir,
                                 process_dir_name_one)
    is_stale = agent.check_process_config_dir_staleness(now, conf_dir_path)
    assert_false(is_stale)

    new_time = now + timestamp_offset + 2
    is_stale = agent.check_process_config_dir_staleness(new_time, conf_dir_path)
    assert_true(is_stale)
  finally:
    cleanup_test_environment(agent)


# Test case 4
# Given a directory without a marker, it should not be regarded to be stale.
def test_conf_dir_not_stale():
  agent = setup_vanilla_test_environment()
  try:

    conf_dir_path = os.path.join(agent.cfg.paths.process_dir,
                                 process_dir_name_two)
    is_stale = agent.check_process_config_dir_staleness(now, conf_dir_path)
    assert_false(is_stale)
  finally:
    cleanup_test_environment(agent)


# Test case 6
# Given one process are not in both active and failed process list, mark orphan
# processes API will mark for it.
def test_mark_orphan_processes_have_marker():
  agent = setup_vanilla_test_environment()
  try:

    fake_raw = dict()
    fake_raw["id"] = 2
    fake_raw["name"] = sub_dir_name_two
    failed_process = Process(agent.cfg, agent, fake_raw)
    failed_process.early_error = Exception("Bogus error")

    failed_process.update_heartbeat(None, False)

    # Find the number of markers
    markers = [file for _, _, files in os.walk(agent.cfg.paths.process_dir)
      for file in files
      if file == "process_timestamp"
    ]
    assert_equal(len(markers), 1)
  finally:
    cleanup_test_environment(agent)


# Test case 7
# Given that there is a marker in a orphan process directory, if the timestamp
# of the marker is older than the threshold, the orphan process directory will
# be removed.
def test_delete_obsolete_orphan_conf_dir():
  agent = setup_vanilla_test_environment()
  try:
    Process.mark_orphan(process_dir_name_one, now)
    marker = get_marker(process_dir_name_one, agent)

    has_marker = os.path.isfile(marker)
    assert_true(has_marker)

    new_time = now + timestamp_offset + 2
    conf_dir_path = os.path.join(agent.cfg.paths.process_dir,
                                 process_dir_name_one)

    is_stale = agent.check_process_config_dir_staleness(new_time, conf_dir_path)
    assert_true(is_stale)

    agent.remove_obsolete_orphan_conf_dirs(new_time)

    conf_dirs = os.listdir(agent.cfg.paths.process_dir)
    assert_equal(len(conf_dirs), 1) # 'hello' dir is still present
  finally:
    cleanup_test_environment(agent)
