# Copyright (c) 2014 Cloudera, Inc. All rights reserved.

import os
import re
import random
import shutil
import stat
import string
import tempfile

from nose.tools import assert_true, assert_false

from cmf.monitor.stacks.compressing_rotating_logger import CompressingRotatingLogger
from cmf.monitor.stacks.stacks_collection_manager import LOG_FORMAT

def _make_logger(logfile, max_uncompressed_file_size_bytes, bytes_retained):
  return CompressingRotatingLogger(logfile,
                                   LOG_FORMAT,
                                   '%d/%b/%Y %H:%M:%S %z',
                                   max_uncompressed_file_size_bytes,
                                   bytes_retained)

def _check_retained_file_size(logfile, bytes_retained_limit):
  """
  Get size of all retained compressed rotating files and make sure it is under
  the limit. Returns a tuple(size, count).
  """

  if logfile is None:
    return 0, 0

  log_directory, filename = os.path.split(logfile)
  pattern = re.compile("^%s\.(\d+)\.bz2$" % re.escape(filename))
  size = 0
  count = 0

  try:
    for f in os.listdir(log_directory):
      match = pattern.match(f)
      if match is not None:
        size = size + os.path.getsize(os.path.join(log_directory, f))
        count = count + 1
    if bytes_retained_limit is not None and bytes_retained_limit > 0:
      assert_true(
        size <= bytes_retained_limit,
        ("size %d of files over retention limit %d" % (size, bytes_retained_limit)))
  except OSError:
    # directory not exists.
    pass
  return size, count

def test_logger():
  random.seed(12345)
  message = ''.join(random.choice(string.ascii_uppercase) for i in range(1024))
  tmpdir = tempfile.mkdtemp()
  try:
      logfile = os.path.join(tmpdir, "stacks.log")
      logger = _make_logger(logfile, 1024, 4096)
      logger.log("test the first message!")
      assert_true(os.path.exists(logfile))
      for i in xrange(100):
        logger.log(message)
      assert_true(os.path.exists(logfile))
      size, count = _check_retained_file_size(logfile, 4096)
      assert_true(count > 1)
  finally:
    shutil.rmtree(tmpdir)

def test_logger_sorting():
  random.seed(12345)
  message = ''.join(random.choice(string.ascii_uppercase) for i in range(1024))
  tmpdir = tempfile.mkdtemp()
  try:
      logfile = os.path.join(tmpdir, "stacks.log")
      logger = _make_logger(logfile, 1024, 5 * 4096)
      logger.log("test the first message!")
      assert_true(os.path.exists(logfile))
      for i in xrange(100):
        logger.log(message)
      assert_true(os.path.exists(logfile))
      size, count = _check_retained_file_size(logfile, 5 * 4096)
      assert_true(count > 10)
  finally:
    shutil.rmtree(tmpdir)

def test_logger_no_retention_limit():
  random.seed(12345)
  message = ''.join(random.choice(string.ascii_uppercase) for i in range(1024))
  tmpdir = tempfile.mkdtemp()
  try:
      logfile = os.path.join(tmpdir, "stacks.log")
      logger = _make_logger(logfile, 1024, 0)
      logger.log("test the first message!")
      assert_true(os.path.exists(logfile))

      for i in xrange(100):
        logger.log(message)
      assert_true(os.path.exists(logfile))
      size_1, count_1 = _check_retained_file_size(logfile, 0)
      assert_true(count_1 > 1)

      for i in xrange(100):
        logger.log(message)
      assert_true(os.path.exists(logfile))
      size_2, count_2 = _check_retained_file_size(logfile, 0)
      assert_true(count_2 > count_1)
      assert_true(size_2 > size_1)

      for i in xrange(100):
        logger.log(message)
      assert_true(os.path.exists(logfile))
      size_3, count_3 = _check_retained_file_size(logfile, 0)
      assert_true(count_3 > count_2)
      assert_true(size_3 > size_2)

  finally:
    shutil.rmtree(tmpdir)

def test_logger_decrease_retention_limit():
  random.seed(12345)
  message = ''.join(random.choice(string.ascii_uppercase) for i in range(1024))
  tmpdir = tempfile.mkdtemp()
  try:
      logfile = os.path.join(tmpdir, "stacks.log")
      logger = _make_logger(logfile, 1024, 0)
      logger.log("test the first message!")
      assert_true(os.path.exists(logfile))
      for i in xrange(200):
        logger.log(message)
      assert_true(os.path.exists(logfile))
      size_before_decrease, count_1 = _check_retained_file_size(logfile, 0)
      new_size_limit = size_before_decrease / 2

      # Now use a data retention setting and log enough to kick in rollover. It
      # should delete the files beyond the specified limit(half of original size).
      logfile = os.path.join(tmpdir, "stacks.log")
      logger = _make_logger(logfile, 1024, new_size_limit)
      logger.log("test the first message!")
      assert_true(os.path.exists(logfile))
      for i in xrange(100):
        logger.log(message)
      assert_true(os.path.exists(logfile))
      actual_size, count_2 = _check_retained_file_size(logfile, new_size_limit)
      assert_true(count_2 < count_1)
      assert_true(actual_size < size_before_decrease)
  finally:
    shutil.rmtree(tmpdir)

def test_logger_leaves_unrelated_files():
  random.seed(12345)
  message = ''.join(random.choice(string.ascii_uppercase) for i in range(1024))
  tmpdir = tempfile.mkdtemp()
  try:
      logfile1 = os.path.join(tmpdir, "stacks-1.log")
      logger = _make_logger(logfile1, 1024, 4096)
      logger.log("test the first message!")
      assert_true(os.path.exists(logfile1))
      for i in xrange(100):
        logger.log(message)
      assert_true(os.path.exists(logfile1))
      size_1, count_1 = _check_retained_file_size(logfile1, 4096)

      # Make a second logger that uses a different filename. This should leave
      # the files created by the logger above in place.
      logfile2 = os.path.join(tmpdir, "stacks-2.log")
      logger = _make_logger(logfile2, 1024, 4096)
      logger.log("test the first message!")
      assert_true(os.path.exists(logfile2))
      for i in xrange(100):
        logger.log(message)
      assert_true(os.path.exists(logfile1))
      size_2, count_2 = _check_retained_file_size(logfile1, 4096)
      assert_true(size_1 == size_2)
      assert_true(count_1 == count_2)
      assert_true(os.path.exists(logfile2))
      _check_retained_file_size(logfile2, 4096)
  finally:
    shutil.rmtree(tmpdir)

def test_logger_deals_with_missing_files():
  tmpdir = tempfile.mkdtemp()
  random.seed(12345)
  message = ''.join(random.choice(string.ascii_uppercase) for i in range(1024))
  try:
      logfile = os.path.join(tmpdir, "stacks.log")
      logger = _make_logger(logfile, 1024, 2 * 4096)
      logger.log("test the first message!")
      assert_true(os.path.exists(logfile))
      for i in xrange(100):
        logger.log(message)
      assert_true(os.path.exists(logfile))
      # 2 * 4096 bytes is sufficient to generate xxx.2.bz2 and xxx.3.bz2 files.
      assert_true(os.path.exists(logfile + ".2.bz2"))
      assert_true(os.path.exists(logfile + ".3.bz2"))

      # Now remove a few of the intermediate files and do some more logging. The
      # logger should continue to function.
      os.remove(logfile + ".2.bz2")
      os.remove(logfile + ".3.bz2")

      # 100 messages is sufficient to trigger two rollovers.
      for i in xrange(100):
        logger.log(message)
      assert_true(os.path.exists(logfile))
      assert_true(os.path.exists(logfile + ".1.bz2"))
      assert_true(os.path.exists(logfile + ".2.bz2"))
      assert_true(os.path.exists(logfile + ".3.bz2"))
      _check_retained_file_size(logfile, 2 * 4096)

  finally:
    shutil.rmtree(tmpdir)


