# Copyright (c) 2013 Cloudera, Inc. All rights reserved.
from cmf.subprocess_timeout import subprocess_with_timeout

from nose.tools import assert_equals
from nose.tools import assert_false
from nose.tools import assert_not_equals
from nose.tools import assert_raises
from nose.tools import assert_true
from nose.tools import nottest

import logging
import os
import random
import shutil
import tempfile
import time

def test_timed_out_process():
  try:
    subprocess_with_timeout(["/bin/sleep", "100"], 0.5)
    assert_false(True, "Should have failed.")
  except Exception, e:
    assert_true("timeout with args" in str(e))

def test_normal_process_output():
  assert_equals( (0, "puppies\n", ""), subprocess_with_timeout(["/bin/echo", "puppies"], 100))

def test_normal_process_errcode():
  assert_equals( (3, "", ""), subprocess_with_timeout(["/bin/bash", "-c", "exit 3"], 100))

def test_recalcitrant_process():
  unique = str(random.randint(10000000,99999999))
  try:
    # Sets up trap handlers that do little for signals 1 to 64.
    # If this turns out to fail on various Unix-es, setting the trap
    # just on SIGTERM should be good enough, but this is more complete.
    cmd = ["/bin/bash", "-c", "for i in $(seq 1 64); do trap \"echo $i\" $i; done; while true; do :; done #hogwash" + unique]
    subprocess_with_timeout(cmd, 0.1)
    assert_false(True, "Should have failed.")
  except Exception, e:
    assert_true("timeout with args" in str(e))

  # We need to be double-sure that the process is dead.  Let's make sure.
  # Note: when editing this test, be sure to break the code
  # and make sure the test triggers; it's finnicky.
  # We try this for 10 seconds or so, since some machines may be slow.
  for _ in range(100):
    bad = False
    time.sleep(0.1)
    for pid in os.listdir("/proc"):
      try:
        cmdline = file("/proc/" + pid + "/cmdline").read()
      except:
        continue
      if "#hogwash" + unique in cmdline:
        bad = True
    if not bad:
      # Our bad process is gone; thank goodness.
      return
  assert_false(True)

def test_process_that_closes_fds():
  cmd = ["/bin/bash", "-c", "echo hi; exec 0>&-; exec 1>&-; exec 2>&-; sleep 1"]
  assert_equals( (0, "hi\n", ""),  subprocess_with_timeout(cmd, 2.0) )

def test_process_that_closes_fds_with_timeout():
  cmd = ["/bin/bash", "-c", "exec 0>&-; exec 1>&-; exec 2>&-; sleep 1000"]
  try:
    subprocess_with_timeout(cmd, 2.0)
    assert_false(True, "Should have failed.")
  except Exception, e:
    assert_true("timeout with args" in str(e))

def test_cwd():
  test_dir = os.path.realpath(tempfile.mkdtemp())
  try:
    cmd = [ "/bin/bash", "-c", "[[ $PWD == %s ]]" % (test_dir,) ]
    logging.info(repr(cmd))
    retcode, stdout, stderr = subprocess_with_timeout(cmd, 2.0, cwd=test_dir)
    assert_equals(0, retcode, "stdout: %s\nstderr: %s" % (stdout, stderr))
  finally:
    shutil.rmtree(test_dir)

def test_env():
  foo = "ASDF"
  env = { 'FOO' : foo }
  cmd = [ "/bin/bash", "-c", "[[ $FOO == %s ]]" % (foo,) ]
  retcode, stdout, stderr = subprocess_with_timeout(cmd, 2.0, env=env)
  assert_equals(0, retcode, "stdout: %s\nstderr: %s" % (stdout, stderr))

def test_stdout_stderr_files():
  test_dir = tempfile.mkdtemp()
  try:
    cmd = [ "/bin/bash", "-c", "echo to_stdout; echo to_stderr 1>&2" ]
    stdout_filename = os.path.join(test_dir, "stdout")
    stderr_filename = os.path.join(test_dir, "stderr")

    open_stdout = None
    open_stderr = None
    try:
      # Open stdout/stderr files for writing.
      open_stdout = open(stdout_filename, "w")
      open_stderr = open(stderr_filename, "w")
      assert_equals( (0, "", ""),
                     subprocess_with_timeout(cmd, 2.0,
                                             stdout=open_stdout,
                                             stderr=open_stderr))
      open_stdout.close()
      open_stdout = None
      open_stderr.close()
      open_stderr = None

      # Now reopen them for reading.
      open_stdout = open(stdout_filename, "r")
      open_stderr = open(stderr_filename, "r")
      assert_equals("to_stdout\n", open_stdout.read())
      assert_equals("to_stderr\n", open_stderr.read())
    finally:
      if open_stdout:
        open_stdout.close()
      if open_stderr:
        open_stderr.close()
  finally:
    shutil.rmtree(test_dir)
