#!/usr/bin/env python
# Copyright (c) 2017 Cloudera, Inc. All rights reserved.

import time
from datetime import datetime
import threading
import logging
from cmf.util.synced_action import SyncedAction

from nose.tools import assert_false

LOG = logging.getLogger(__name__)

class ItemHolder(object):
    """
    ItemHolder stores a increasing number in its "item" attribute - a list.
    Every time "action" is called, the "item" attribute is reset and the value in the
    old list is incremented.  The newly incremented value is visible only to code that
    acquired a reference to "item" prior to the invocation of "action()".

    When "item" is reset, all sleepers on "cond" are woken and "action()" sleeps for one
    second before updating the old "item" and completing.

    The purpose here is to allow us to invoke action() via "synced_action" and be able to
    determine, by the caller, which invocation of "action()" it actually witnessed.
    """
    def __init__(self):
        self.item = []
        self.count = 0
        self.started = 0
        self.cond = threading.Condition()

    def action(self):
        """
        This is the synchronized action in this test. Only 1 thread should be able to invoke this action in an allowed
        time slice (2sec). The action updates the item list to its last internal epoch which is compared by all threads
        against the expected result
        """
        self.started += 1
        old_item = self.item
        self.item = []
        LOG.info("[%s]: [%s] Start action: %s", datetime.now(), threading.currentThread().getName(), self.started)
        with self.cond:
            self.cond.notify_all()
        self.count = self.started
        old_item.append(self.count)

        LOG.info("[%s]: [%s] Run action: %s", datetime.now(), threading.currentThread().getName(), self.count)

def perform(item, synced_action, expected_results, fail):
    """
    All threads execute this method. Only one of the threads will be able to execute the synced action in a 2sec time
    slice on everybody's behalf. When the action completes, the result of the action should be same for all threads
    that boarded the same train. The test verifies the post-action result matches the expected result for all threads
    """
    for x, result in enumerate(expected_results):
        val = item.item
        seq = item.count
        SyncedAction.schedule(synced_action)
        LOG.info("[%s]: [%s] result %s: %s", datetime.now(), threading.currentThread().getName(), x, val)
        if val[0] != result:
            fail[0] = True
            LOG.error("[%s]: [%s] bad result %s: %s != %s", datetime.now(), threading.currentThread().getName(), x, val[0], result)

def test_synced_actions():
    logging.basicConfig(level=logging.INFO)
    fail = [False]
    item = ItemHolder()
    synced_action = item.action
    threads = []
    for x in range(0,4):
        # All threads here should catch invocation #1 and #2 of synced_action
        t = threading.Thread(name="thread%s" % x, target=perform, args=(item, synced_action,[1, 2], fail))
        t.start()
        threads.append(t)

    # The MainThread will wait until we know that ItemHolder.action() is executing - by waiting on
    # item.cond. The intention here is to miss the first train and ensure that the behavior is consistent at time window
    while item.started < 1:
        started = item.started
        with item.cond:
            item.cond.wait()

    # Due to the wait above, this invocation of "perform" can't catch invocation 1.
    # The first invocation of synced_action() will be invocation #2.
    perform(item, synced_action, [2,3], fail)
    for t in threads:
        t.join()
    assert_false(fail[0])
