#!/usr/bin/env python
# Copyright (c) 2014 Cloudera, Inc. All rights reserved.

from nose.tools import assert_equal
from nose.tools import assert_false
from nose.tools import assert_true
from mock import call
from mock import patch
from subprocess import PIPE

from cmf import tmpfs

class Partition:
  def __init__(self, mountpoint, fstype, opts):
    self.mountpoint = mountpoint
    self.fstype = fstype
    self.opts = opts

def fake_samefile(a, b):
  return a == b

def fake_exists(a):
  return True

def fake_does_not_exist(a):
  return False

@patch("cmf.tmpfs._mount_tmpfs")
@patch("cmf.tmpfs.psutil.disk_partitions")
@patch("os.path.samefile", fake_samefile)
@patch("os.path.exists", fake_does_not_exist)
def test_invalid_mountpoint(disk_partitions, inner_mount):
  # No partitions: mount is created.
  disk_partitions.return_value = [Partition("/1", "tmpfs", "")]
  assert_true(tmpfs.mount_tmpfs("/1"))
  inner_mount.assert_called_once_with("/1", True, mode=0751, sudo=None)
  inner_mount.reset_mock()

@patch("cmf.tmpfs._mount_tmpfs")
@patch("cmf.tmpfs.psutil.disk_partitions")
@patch("os.path.samefile", fake_samefile)
@patch("os.path.exists", fake_exists)
def test_partition_layouts(disk_partitions, inner_mount):
  # No partitions: mount is created.
  disk_partitions.return_value = []
  assert_true(tmpfs.mount_tmpfs("/1"))
  inner_mount.assert_called_once_with("/1", True, mode=0751, sudo=None)
  inner_mount.reset_mock()

  # Existing partition on wrong path: mount is created.
  disk_partitions.return_value = [ Partition("/bar", "tmpfs", "")]
  assert_true(tmpfs.mount_tmpfs("/2"))
  inner_mount.assert_called_once_with("/2", True, mode=0751, sudo=None)
  inner_mount.reset_mock()

  # Existing partition with wrong fstype: mount is created.
  disk_partitions.return_value = [ Partition("/3", "ext4", "")]
  assert_true(tmpfs.mount_tmpfs("/3"))
  inner_mount.assert_called_once_with("/3", True, mode=0751, sudo=None)
  inner_mount.reset_mock()

  # Existing partition with noexec: mount is created.
  disk_partitions.return_value = [ Partition("/4", "tmpfs", "a,noexec,c")]
  assert_true(tmpfs.mount_tmpfs("/4"))
  inner_mount.assert_called_once_with("/4", True, mode=0751, sudo=None)
  inner_mount.reset_mock()

  # Existing partition is good: mount is reused.
  disk_partitions.return_value = [ Partition("/5", "tmpfs", "a,b,c") ]
  assert_true(tmpfs.mount_tmpfs("/5"))
  assert_false(inner_mount.called)
  inner_mount.reset_mock()

  # Some existing partition is good: mount is reused.
  disk_partitions.return_value = [ Partition("/baz", "iso9660", ""),
                                   Partition("/6", "tmpfs", "a,b,c"),
                                   Partition("/bar", "ext4", "") ]
  assert_true(tmpfs.mount_tmpfs("/6"))
  assert_false(inner_mount.called)
  inner_mount.reset_mock()

  # Existing partition is good but clean_existing=True: mount is unmounted and created.
  disk_partitions.return_value = [ Partition("/7", "tmpfs", "a,b,c") ]
  assert_true(tmpfs.mount_tmpfs("/7", clean_existing=True))
  assert_equal([ call("/7", False, sudo=None),
                 call("/7", True, mode=0751, sudo=None) ],
               inner_mount.call_args_list)

@patch("cmf.tmpfs._mount_tmpfs")
@patch("cmf.tmpfs.psutil.disk_partitions")
@patch("os.path.samefile", fake_samefile)
@patch("os.path.exists", fake_exists)
def test_inner_mount_return_values(disk_partitions, inner_mount):
  # Mocked so that inner_mount fails the first time and succeeds subseqent times.
  inner_mount.side_effect = lambda *args, **kwargs: (inner_mount.call_count > 1)

  # Unmount fails so we reused the mount.
  disk_partitions.return_value = [ Partition("/1", "tmpfs", "a,b,c") ]
  assert_true(tmpfs.mount_tmpfs("/1", clean_existing=True))
  inner_mount.assert_called_once_with("/1", False, sudo=None)
  inner_mount.reset_mock()

  # Mount fails. Returns false.
  inner_mount.return_value = False
  inner_mount.side_effect = None
  disk_partitions.return_value = []
  assert_false(tmpfs.mount_tmpfs("/2"))
  inner_mount.assert_called_once_with("/2", True, mode=0751, sudo=None)

@patch("os.getgid")
@patch("os.getuid")
@patch("cmf.tmpfs.subprocess.Popen")
@patch("os.path.samefile", fake_samefile)
@patch("os.path.exists", fake_exists)
def test_inner_mount_args(popen, getuid, getgid):
  popen.return_value.communicate.return_value = (None, None)
  popen.return_value.returncode = 0
  getuid.return_value = 101
  getgid.return_value = 102

  # Mount with mode.
  assert_true(tmpfs._mount_tmpfs("/1", True, mode=0751))
  popen.assert_called_once_with([ "mount", "-t", "tmpfs", "cm_processes", "/1", "-o", "mode=0751" ],
                                stderr=PIPE)
  popen.reset_mock()

# Mount with sudo and mode.
  assert_true(tmpfs._mount_tmpfs("/1", True, mode=0751, sudo="sudo"))
  popen.assert_called_once_with([ "sudo", "mount", "-t", "tmpfs", "cm_processes", "/1", "-o", "mode=0751,uid=101,gid=102" ],
                                stderr=PIPE)
  popen.reset_mock()

  # Mount without mode.
  assert_true(tmpfs._mount_tmpfs("/1", True))
  popen.assert_called_once_with([ "mount", "-t", "tmpfs", "cm_processes", "/1" ],
                                stderr=PIPE)
  popen.reset_mock()

  # Mount with sudo without mode
  assert_true(tmpfs._mount_tmpfs("/1", True, sudo="sudo"))
  popen.assert_called_once_with([ "sudo", "mount", "-t", "tmpfs", "cm_processes", "/1", "-o", "uid=101,gid=102" ],
                                stderr=PIPE)
  popen.reset_mock()

# Umount.
  assert_true(tmpfs._mount_tmpfs("/1", False))
  popen.assert_called_once_with([ "umount", "/1" ], stderr=PIPE)
  popen.reset_mock()

  # Some kind of failure.
  popen.return_value.returncode = 123
  assert_false(tmpfs._mount_tmpfs("/1", True))
