#!/usr/bin/env python
# encoding=utf8
# Copyright (c) 2013 Cloudera, Inc. All rights reserved.

import cmf.util
import os
import shutil
import tempfile
import stat
import avro

from StringIO import StringIO
from cmf.util import catch_all
from cmf.util import unzip
from nose.tools import assert_true, assert_equal, raises, nottest, assert_false
from mock import patch
from zipfile import ZipFile
import cmf.util.avro

def test_sha1():
  DATA = "this is test data"
  SHA1 = "f3c351aca845352503cc693e50ccbcab83609175" # echo -n "this is test data" | sha1sum

  fd, path = tempfile.mkstemp()
  os.write(fd, DATA)
  os.close(fd)

  assert_equal(SHA1, cmf.util.sha1(path))
  os.unlink(path)

class MockLog(object):
  def __init__(self, expected):
    self.expected = expected

  def info(self, fmt, *pargs):
    pass

  def warn(self, text, *pargs):
    text = text % pargs
    assert text.startswith(self.expected)

  def exception(self, text, *pargs):
    text = text % pargs
    assert text.startswith(self.expected)

def mock_uuid():
  return "DEADBEEF"

@patch('uuid.uuid4', mock_uuid)
def test_get_host_uuid_new_success():
  tmpdir = tempfile.mkdtemp()
  try:
    uuid = cmf.util.get_host_uuid(tmpdir)
    assert_equal("DEADBEEF", uuid)
  finally:
    shutil.rmtree(tmpdir)

def test_get_host_uuid_existing_success():
  tmpdir = tempfile.mkdtemp()
  try:
    uuid_path = os.path.join(tmpdir, "uuid")
    uuid_file = open(uuid_path, 'w')
    try:
      uuid_file.write("DEADBEEF")
    finally:
      uuid_file.close()

    uuid = cmf.util.get_host_uuid(tmpdir)
    assert_equal("DEADBEEF", uuid)
  finally:
    shutil.rmtree(tmpdir)

@raises(OSError)
@patch.object(cmf.util, 'LOG', MockLog("Failed attempt to create UUID file:"))
def test_get_host_uuid_new_failure():
  tmpdir = tempfile.mkdtemp()
  try:
    os.chmod(tmpdir, 0644)

    uuid = cmf.util.get_host_uuid(tmpdir)
    assert_equal(None, uuid)
  finally:
    shutil.rmtree(tmpdir)

@raises(IOError)
@patch.object(cmf.util, 'LOG', MockLog("Unable to open UUID file"))
def test_get_host_uuid_existing_failure():
  tmpdir = tempfile.mkdtemp()
  try:
    uuid_path = os.path.join(tmpdir, "uuid")
    uuid_file = open(uuid_path, 'w')
    try:
      uuid_file.write("DEADBEEF")
    finally:
      uuid_file.close()
    os.chmod(uuid_path, 0000)

    uuid = cmf.util.get_host_uuid(tmpdir)
    assert_equal(None, uuid)
  finally:
    shutil.rmtree(tmpdir)

class TestCatchAll(object):
  """
  Test catch_all decorator in the cmf.util file.
  """

  @catch_all()
  def i_succeed(self):
    return "success"

  @catch_all()
  def i_will_throw(self):
    assert(False) # This will raise AssertionError

  @catch_all("failed")
  def i_will_throw_and_return_failed(self):
    assert(False) # This will raise AssertionError

  def test_catch_exception(self):
    """
    Check that catch_all decorator is able to catch any exception that we
    throw.
    """
    # call the function that throws, test will pass if exception is chewed
    ret = self.i_will_throw()
    assert_equal(ret, None)

  def test_catch_exception_return_success(self):
    """
    Test if the catch_all handler returns proper value when no exception is
    thrown.
    """
    assert_equal(self.i_succeed(), "success")

  def test_catch_exception_return_fail(self):
    """
    Test if the catch_all handler returns proper value when an exception is
    thrown.
    """
    ret = self.i_will_throw_and_return_failed()
    assert_equal(ret, "failed")

def test_zip():
  """
  Tests cmf.util.zip with various archives.
  """

  def load_file(filename):
    """
    Returns the contents of filename.
    """
    f = open(filename, "r")
    try:
      return f.read()
    finally:
      f.close()

  fixture_dir = os.path.join(os.path.dirname(__file__),
                             "test-fixtures",
                             "util-fixture")

  test_dir = tempfile.mkdtemp()
  try:
    fixtures = {
      "mapreduce-1-clientconfig.zip" : (8, 0),
      "empty.zip" : (0, 0),
      "only_one_directory.zip" : (0, 1),
      "one_dot_dot_file.zip" : (0, 0),
      "three_nested_directories.zip" : (0, 3),
    }
    for fixture, counts in fixtures.iteritems():
      data = load_file(os.path.join(fixture_dir, fixture))
      scratch_dir = os.path.join(test_dir, fixture)
      os.mkdir(scratch_dir)

      # Extract twice; the counts should still be the same.
      for _ in xrange(2):
        file_count, dir_count = unzip(data, scratch_dir)
        assert_equal(file_count, counts[0])
        assert_equal(dir_count, counts[1])

      # Ensure the proper files/directories were created.
      for name in ZipFile(StringIO(data)).namelist():
        if ".." in name:
          continue

        full_path = os.path.join(scratch_dir, name)
        assert_true(os.path.exists(full_path))
        if name.endswith('/'):
          assert_true(os.path.isdir(full_path))
        else:
          assert_true(os.path.isfile(full_path))

  finally:
    shutil.rmtree(test_dir)


def test_34952_monkey_patch():
  """
  Test the monkey patch fix for OPSAPS-34952: A request for a URL resulting in a 404
  should not leak file descriptors.
  :return:
  """
  import urllib2
  import cmf.https
  fd_count = len(os.listdir("/proc/self/fd"))
  try:
    urllib2.urlopen("http://www.cloudera.com/fooobar")
  except urllib2.HTTPError, e:
    code = 404
    try:
      code = e.getcode()
    except Exception:
      pass
    assert_equal(404, code, "Expected 404 error.")
    pass
  new_fd_count = len(os.listdir("/proc/self/fd"))
  assert_equal(fd_count, new_fd_count, "urllib2 leaked file descriptors")


def test_conf_file():
  content = "content"
  tmp_name = "/tmp/tmpfile.%s" % os.getpid()
  try:
    assert_false(os.path.exists(tmp_name), "Temp file should not exist before test")
    cmf.util.write_conf_file(tmp_name, content)

    assert_true(os.path.exists(tmp_name), "Temp file does not exist after creation.")
    assert_true(os.stat(tmp_name).st_mode == (stat.S_IRUSR | stat.S_IWUSR | stat.S_IFREG),
                "Wrong file mode: %o" % os.stat(tmp_name).st_mode)

  finally:
    if os.path.exists(tmp_name):
      os.unlink(tmp_name)

def test_unicode_write_and_read():
  """
  Test which writes and reads back a unicode string
  """
  import codecs

  str1 = u'hello world'
  f = open('test', 'w')
  f.write(str1.encode('utf-8'))
  f.close()

  f = open('test', 'r')
  str2 = f.read().decode('utf-8')
  f.close()

  assert_equal(str1, str2, "unicode strings mismatch")

def test_unicode_sanitization():
  """
  Test which exercises unicode sanitization for list and dict
  """
  obj1 = {"attr_┬": "value_"}
  obj2 = ["entry┬", "entry"]
  obj1 = cmf.util.avro.sanitize_unicode(obj1)
  obj2 = cmf.util.avro.sanitize_unicode(obj2)
  for k,v in obj1.iteritems():
    assert_equal(cmf.util.avro.can_utf8_encode_string(k), True, "Unicode character not sanitized")
    assert_equal(cmf.util.avro.can_utf8_encode_string(v), True, "Unicode character not sanitized")
  for entry in obj2:
    assert_equal(cmf.util.avro.can_utf8_encode_string(entry), True, "Unicode character not sanitized")

def test_unicode_bytes_sanitization():
  """
  Test which verifies that 'bytes' fields are not sanitized
  """
  schema = '{' \
           '"namespace": "com.cloudera", ' \
           '"type": "record", ' \
           '"name": "TestMessage", ' \
           '"fields": [{' \
              '"name": "name", ' \
              '"type": "string"' \
           '}, {' \
              '"name": "buffer", ' \
              '"type": "bytes"' \
           '}, {' \
              '"name": "unionFld", ' \
              '"type": ["string", "bytes"]' \
           '}, {' \
              '"name": "bytesMap", ' \
              '"type": {' \
                '"type" : "map", ' \
                '"values" : "bytes"' \
                '}' \
           '}, {"name": "stringMap", ' \
              '"type": {' \
                '"type" : "map", ' \
                '"values" : "string"' \
                '}' \
           '}, {' \
              '"name": "bytesArray", ' \
              '"type": {' \
                '"type" : "array", ' \
                '"items" : "bytes"' \
                '}' \
           '}]}'
  parsed_schema = avro.schema.parse(schema)
  obj = {
    "name": "àaßb",
    "buffer": "àaßb",
    "unionFld": "àaßb",
    "bytesMap": {
      "key": "àaßb"
    }, "stringMap": {
      "key": "àaßb"
    }, "bytesArray": ["àaßb"]
  }
  sanitizer = cmf.util.avro.generate_avro_sanitizer(parsed_schema)
  sanitized_obj = sanitizer(obj)
  is_equal = sanitized_obj["name"]==sanitized_obj["buffer"]
  assert_equal(is_equal, False, "Bytes field were sanitized")
  assert_equal(sanitized_obj["unionFld"], obj["unionFld"], "Bytes field in union was sanitized")
  assert_equal(sanitized_obj["bytesArray"][0], obj["bytesArray"][0], "Bytes field in array was sanitized")
  assert_equal(sanitized_obj["bytesMap"]["key"]==sanitized_obj["stringMap"]["key"], False,
               "Bytes field in map was sanitized")
