#!/usr/bin/env python
# Copyright (c) 2012 Cloudera, Inc. All rights reserved.

import os
import shutil

from nose.tools import assert_true, assert_equal, assert_false, assert_raises
from cmf.version_detector import VersionDetector, SPECIAL_SCRIPTS
import cmf.version

def test_version_detection():
  """
  Test all possible cases for version detection
  1. Component with correct script and version files => correct CDH version
  2. Component with correct script but wrong version file => 'BAD_VERSION_FILE'
  3. Component with correct script but missing version file => 'NO_VERSION_FILE'
  4. Component with missing script => 'NO_SCRIPT'
  5. Hue component where dir exists => cdh5 version
  6. Hue component where dir does not exist => cdh4 version
  7. Hcat component where dir exists => cdh5 version
  8. Hcat component where dir does not exist => cdh4 version
  9. Component where script file has wildcard in it
  """

  # prepare the directory structure for the test
  test_dir_name = os.path.dirname(__file__) + "/version_detection_test"
  os.mkdir(test_dir_name)

  test_env = {"CDH_CDH3TEST_HOME" : test_dir_name + "/cdh3test",
              "CDH_CDH4TEST_HOME" : test_dir_name + "/cdh4test",
              "CDH_CDH5TEST_HOME" : test_dir_name + "/cdh5test",
              "CDH_CDH6TEST_HOME" : test_dir_name + "/cdh6test",
              "CDH_CDH7TEST_HOME" : test_dir_name + "/cdh7test",
              "CDH_NONCDH_HOME" : test_dir_name + "/noncdh", # e.g. Impala
              "CDH_HUE_HOME"      : test_dir_name + "/hue",
              "CDH_HCAT_HOME"     : test_dir_name + "/hive-hcatalog",
              "CDH_PKGNAME_HOME" : test_dir_name + "/not_home_env_name",
              "CDH_WRONGVERSION1_HOME" : test_dir_name + "/wrongversion1",
              "CDH_WRONGVERSION2_HOME" : test_dir_name + "/wrongversion2",
              "CDH_NOVERSION_HOME" : test_dir_name + "/noversion",
              "CDH_NOSCRIPT_HOME" : test_dir_name + "/noscript",
              "WRONG_ENV_VAR" : test_dir_name + "/ignore",
              "CDH_CRUNCH_HOME" : test_dir_name + "/crunch"}

  try:
    for key, value in test_env.iteritems():
      component_name = value.split("/")[-1]
      component_dir = value
      os.mkdir(component_dir)
      os.mkdir(component_dir + "/bin")
      os.mkdir(component_dir + "/cloudera")

      # Create bin and version files for test components.
      #
      # Note: Hue and Hcat use different homes in CDH4 and 5, and are tested
      # using the following testcases:
      # (1) Setup homes for both CDH versions and verify that the right (CDH5)
      #     home is chosen.
      # (2) Remove CDH5 home, and verify that CDH4 home is chosen.
      if key != 'CDH_NOSCRIPT_HOME':
        relpath = "/bin/" + component_name
        # Setup CDH5 bin script for Hue and Hcat
        if key == 'CDH_HUE_HOME':
          relpath = SPECIAL_SCRIPTS['hue']
          os.makedirs(component_dir + relpath[:-3])
        if key == 'CDH_HCAT_HOME':
          relpath = SPECIAL_SCRIPTS['hcatalog']
          os.makedirs(component_dir + relpath[:-3])
        if key == 'CDH_PKGNAME_HOME':
          relpath = "/bin/pkgname"
        if key == 'CDH_CRUNCH_HOME' :
          relpath = "/crunch-core-some_extra_path.jar"
        fd = open(component_dir + relpath, "w")
        fd.write("dummy_script")
        fd.close()

      if key != 'CDH_NOVERSION_HOME':
        fd = open(component_dir + "/cloudera/cdh_version.properties", "w")
        if key == 'CDH_WRONGVERSION2_HOME':
          fd.write("blah")
        else:
          release = component_name
          # Setup CDH5 version file for Hue and Hcat
          if key == 'CDH_HUE_HOME':
            release = 'cdh5.0.0'
          if key == 'CDH_HCAT_HOME':
            release = 'cdh5.0.0'
          if key == 'CDH_CRUNCH_HOME':
            release = 'cdh5.0.0'

          if key == 'CDH_NONCDH_HOME':
            fd.write("cloudera.cdh.release=NA\ncloudera.pkg.name=" + release + "\ncloudera.pkg.version=0.5\ncloudera.pkg.release=1.p0")
          elif key == 'CDH_PKGNAME_HOME':
            fd.write("cloudera.cdh.release=NA\ncloudera.pkg.name=" + component_name + "\ncloudera.pkg.version=3.1\ncloudera.pkg.release=2.p1")
          else:
            fd.write("cloudera.cdh.release=" + release + "\ncloudera.pkg.version=pkg\ncloudera.pkg.release=3.p2")
        fd.close()

    # Create CDH4 Hue bin and version files
    cdh4_hue_home = test_dir_name + "/cdh4hue"
    cdh4_hue_bin = cdh4_hue_home + SPECIAL_SCRIPTS['hue']
    os.makedirs(cdh4_hue_bin[:-3])
    fd = open(cdh4_hue_bin, "w")
    fd.write("dummy_script")
    fd.close()
    os.makedirs(cdh4_hue_home + "/cloudera")
    fd = open(cdh4_hue_home + "/cloudera/cdh_version.properties", "w")
    fd.write("cloudera.cdh.release=cdh4.0.0\ncloudera.pkg.version=pkg\ncloudera.pkg.release=4.p3")
    fd.close()

    # Create CDH4 Hcat bin and version files
    cdh4_hcat_home = test_dir_name + "/cdh4hcat"
    cdh4_hcat_bin = cdh4_hcat_home + SPECIAL_SCRIPTS['hcatalog']
    os.makedirs(cdh4_hcat_bin[:-3])
    fd = open(cdh4_hcat_bin, "w")
    fd.write("dummy_script")
    fd.close()
    os.makedirs(cdh4_hcat_home + "/cloudera")
    fd = open(cdh4_hcat_home + "/cloudera/cdh_version.properties", "w")
    fd.write("cloudera.cdh.release=cdh4.0.0\ncloudera.pkg.version=pkg\ncloudera.pkg.release=5.p4")
    fd.close()

    # Create Supervisor component
    supervisor_info = dict()
    supervisor_info["component_name"] = "Supervisord"
    supervisor_info["sup_version"] = "3.0a12-cm5.4.3"

    actual_versions = VersionDetector.get_component_versions(test_env, "/", alternate_hue_home=cdh4_hue_home, alternate_hcat_home=cdh4_hcat_home, supervisor_info=supervisor_info)

    expected_versions = [ {'name' : 'cdh3test', 'active' : True, 'source': 'PACKAGE', 'cdh_version' : 'CDH3', "cdh_release": "cdh 3test", 'component_version' : 'pkg', 'component_release': '3.p2'},
                          {'name' : 'cdh4test', 'active' : True, 'source': 'PACKAGE', 'cdh_version' : 'CDH4', "cdh_release": "cdh 4test", 'component_version' : 'pkg', 'component_release': '3.p2'},
                          {'name' : 'cdh5test', 'active' : True, 'source': 'PACKAGE', 'cdh_version' : 'CDH5', "cdh_release": "cdh 5test", 'component_version' : 'pkg', 'component_release': '3.p2'},
                          {'name' : 'cdh6test', 'active' : True, 'source': 'PACKAGE', 'cdh_version' : 'CDH6', "cdh_release": "cdh 6test", 'component_version' : 'pkg', 'component_release': '3.p2'},
                          {'name' : 'cdh7test', 'active' : True, 'source': 'PACKAGE', 'cdh_version' : 'BAD_VERSION_FILE', "cdh_release": "BAD_VERSION_FILE", 'component_version' : 'BAD_VERSION_FILE', 'component_release': 'BAD_VERSION_FILE'},
                          {'name' : 'noncdh', 'active' : True, 'source': 'PACKAGE', 'cdh_version' : 'NOT_APPLICABLE', "cdh_release": "noncdh 0.5", 'component_version' : '0.5', 'component_release': '1.p0'},
                          {'name' : 'hue', 'active' : True, 'source': 'PACKAGE', 'cdh_version' : 'CDH5', "cdh_release": "cdh 5.0.0", 'component_version' : 'pkg', 'component_release': '3.p2'},
                          {'name' : 'crunch', 'active' : True, 'source': 'PACKAGE', 'cdh_version' : 'CDH5', "cdh_release": "cdh 5.0.0", 'component_version' : 'pkg', 'component_release': '3.p2'},
                          {'name' : 'hcatalog', 'active' : True, 'source': 'PACKAGE', 'cdh_version' : 'CDH5', "cdh_release": "cdh 5.0.0", 'component_version' : 'pkg', 'component_release': '3.p2'},
                          {'name' : 'wrongversion1', 'active' : True, 'source': 'PACKAGE', 'cdh_version' : 'BAD_VERSION_FILE', "cdh_release": "BAD_VERSION_FILE", 'component_version' : 'BAD_VERSION_FILE', 'component_release': 'BAD_VERSION_FILE'},
                          {'name' : 'wrongversion2', 'active' : True, 'source': 'PACKAGE', 'cdh_version' : 'BAD_VERSION_FILE', "cdh_release": "BAD_VERSION_FILE", 'component_version' : 'BAD_VERSION_FILE', 'component_release': 'BAD_VERSION_FILE'},
                          {'name' : 'not_home_env_name', 'active' : True, 'source': 'PACKAGE', 'cdh_version' : 'NOT_APPLICABLE', "cdh_release": "not_home_env_name 3.1", 'component_version' : '3.1', 'component_release': '2.p1'},
                          {'name' : 'noversion', 'active' : True, 'source': 'PACKAGE', 'cdh_version' : 'NO_VERSION_FILE', "cdh_release": "NO_VERSION_FILE", 'component_version' : 'NO_VERSION_FILE', 'component_release' : 'NO_VERSION_FILE'},
                          {'name' : 'noscript', 'active' : True, 'source': 'PACKAGE', 'cdh_version' : 'NO_SCRIPT', "cdh_release": "NO_SCRIPT", 'component_version' : 'NO_SCRIPT', 'component_release' : 'NO_SCRIPT'},
                          {'name': 'cm_agent', 'active' : True, 'source': 'PACKAGE', 'cdh_version': 'NOT_APPLICABLE', "cdh_release": "NOT_APPLICABLE", 'component_version': cmf.version.VERSION, 'component_release': cmf.version.RELEASE},
                          {'name': 'Supervisord', 'active' : True, 'source': 'PACKAGE', 'cdh_version': 'NOT_APPLICABLE', "cdh_release": "NOT_APPLICABLE", 'component_version': "3.0a12-cm5.4.3", 'component_release': 'NO_SCRIPT'}]

    # get all elements that are in one set but not another
    white_list = set(["cm_daemons"])
    extra_versions = set([elem["name"] for elem in expected_versions]).symmetric_difference(
                     set([elem["name"] for elem in actual_versions])) - white_list
    assert_equal(len(extra_versions), 0, msg=extra_versions)
    for component in actual_versions:
      if component["name"] not in white_list:
        assert_true(component in expected_versions, msg="Missing component: " + str(component))

    # Remove CDH5 hue dir, verify fallback
    shutil.rmtree(test_env['CDH_HUE_HOME'])
    actual_versions = VersionDetector.get_component_versions(test_env, "/", alternate_hue_home=cdh4_hue_home)
    assert_true({'name' : 'hue', 'active': True, 'source': 'PACKAGE', 'cdh_version' : 'CDH4', "cdh_release": "cdh 4.0.0", 'component_version' : 'pkg', 'component_release': '4.p3'} in actual_versions, actual_versions)
    # Remove CDH5 hcat dir, verify fallback
    shutil.rmtree(test_env['CDH_HCAT_HOME'])
    actual_versions = VersionDetector.get_component_versions(test_env, "/", alternate_hcat_home=cdh4_hcat_home)
    assert_true({'name' : 'hcatalog', 'active': True, 'source': 'PACKAGE', 'cdh_version' : 'CDH4', "cdh_release": "cdh 4.0.0", 'component_version' : 'pkg', 'component_release': '5.p4'} in actual_versions, actual_versions)
  finally:
    # Delete the directory structure created for test
    shutil.rmtree(test_dir_name)

def test_version_detection_no_hue_plugins_dir():
  test_env = {"CDH_HUE_PLUGINS_HOME" : "/foo/bar/hue_plugins_test", "CDH_HUE_HOME" : "/blah", "CDH_HCAT_HOME": "/hcat"}
  hue_component = {'name': 'hue_plugins', 'cdh_version': 'NO_SCRIPT', "cdh_release": "NO_SCRIPT", 'component_version': 'NO_SCRIPT', 'component_release': 'NO_SCRIPT','active': True, 'source': 'PACKAGE'}
  assert_true(hue_component in VersionDetector.get_component_versions(test_env, "/"))
