#!/usr/bin/env python
# Copyright (c) 2011 Cloudera, Inc. All rights reserved.

import SimpleHTTPServer
import SocketServer
import StringIO
import argparse
import logging
import pwd
import shutil
import signal
import socket
import subprocess
import tempfile
import threading
import time
import urllib2
import zipfile

import avro.protocol
import os
from M2Crypto.m2 import m2crypto_scm_patchlevel
from avro.io import validate
from avro.schema import Schema
from mock import patch, call, Mock
from nose.plugins.skip import SkipTest
from nose.tools import assert_true, assert_equal, assert_false, assert_raises
from urllib2_kerberos import HTTPKerberosAuthHandler

import cmf.logging
import cmf.util.config
from cmf.agent import Agent, CM_STATUS_CODES_ENV_VAR, add_cm_status_codes_to_env_dict
from cmf.monitor.host import FilesystemMap
from cmf.process import Process
from cmf.template_util import render_text
from cmf.throttling_logger import ThrottlingLogger
from cmf.util import (tail, unzip, deploy_special_file, validate_and_open, validate_and_open_fd,
                      source, kill_process, remove_agent_path)
from cmf.util.url import urlopen_with_timeout, _get_full_url
import sys

# These are sparse unit tests of the agent code. To run them:
# cd <path_to_tests.py>/../..
# mvn test

STATUS_CODES_PROTOCOL = avro.protocol.parse(
    file(os.path.join(os.path.dirname(__file__), "../cmf/status_codes.avpr")).read())
CM_STATUS_CODE_SCHEMA = STATUS_CODES_PROTOCOL.types_dict.get("CMStatusCode")

class FakeAgentEnv(object):
  agent_class = Agent

  def __init__(self, **kwargs):
    cmf.logging.setup_initial_logging("test")
    self.base_dir = None
    self.kwargs = kwargs

  def __enter__(self):
    # clear logging state so that agent can init
    root = logging.getLogger()
    map(root.removeHandler, root.handlers[:])
    map(root.removeFilter, root.filters[:])

    self.base_dir = tempfile.mkdtemp()
    agent_log_dir = os.path.join(self.base_dir, 'log')
    agent_pkg_dir = os.path.join(self.base_dir, 'pkg')
    agent_lib_dir = os.path.join(self.base_dir, 'lib')
    agent_run_dir = os.path.join(self.base_dir, 'run')
    agent_log_file = os.path.join(agent_log_dir, "cloudera-scm-agent.log")
    os.makedirs(agent_lib_dir)
    os.makedirs(agent_log_dir)
    os.makedirs(agent_pkg_dir)
    os.makedirs(agent_run_dir)
    overrides = ["enable_supervisord_events=False",
                 "server_port=7182",
                 "server_host=server.something.com",
                 "package_dir=%s" % agent_pkg_dir,
                 "agent_dir=%s" % agent_run_dir,
                 "lib_dir=%s" % agent_lib_dir,
                 "log_dir=%s" % agent_log_dir,
                 "log_file=%s" % agent_log_file ] + \
      [ "%s=%s" % (k, v) for k,v in self.kwargs.iteritems() ]
    import cmf.main
    cfg = cmf.main.AgentConfig.make_config(None, overrides, cmf.main.AgentConfig.GENERAL)
    args = argparse.Namespace()
    args.override = overrides
    args.cfg = cfg
    args.environment = None
    args.clear_agent_dir = False
    args.daemon = False
    agent = self.agent_class(cfg)
    agent.eval_config()
    agent.configure_service()
    return agent


  def __exit__(self, exc_type, exc_val, exc_tb):
    shutil.rmtree(self.base_dir)



def test_config():
  cmf.util.config.TestConfig.run_test()

def test_agent():
  raise SkipTest
  agent = Agent()

  # We don't configure the agent using command-line
  # flags, but instead configure manually
  package_dir = os.path.join(os.path.dirname(__file__),
                             "..",
                             "..",
                             "..",
                             "service")
  supervisord_path = agent._default_supervisord_path()
  with FakeAgentEnv(package_dir=package_dir, supervisord_path=supervisord_path) as agent:
    # Start the supervisor
    agent.start_supervisor()
    assert_true(os.path.isdir(agent.supervisor_dir),
      "Supervisor working dir should have been created.")
    assert_true(os.path.isfile(os.path.join(agent.supervisor_dir,
                                            "supervisord.conf")),
      "Supervisor configuration should exist.")
    assert_equal(None, agent.supervisor_process.poll(),
      "Supervisor should not have exited.")

    # Unfortunately, the supervisor client (XMLRPC)
    # doesn't know how to reconnect, it seems, so
    # we need a sleep here.  This is a bit worrisome.
    time.sleep(1.0)
    assert_equal("3.0",
                 _try_a_while(agent.supervisor_client.client.supervisor.getAPIVersion),
                 "Should be able to communicate with supervisor by API call.")

    # Prepare a heartbeat
    h = agent.prepare_heartbeat()
    assert_equal(0,
                 len(h["marching_order_status"]),
                 "No supervision should be going on quite yet.")

    # Let's introduce some marching orders
    response = dict(marching_orders=[make_hello_order(1, 9002),
                                     make_hello_order(2, 9003)])
    agent.handle_heartbeat_response(response)

    assert_equal(set([1,2]), set(agent.active_marching_orders.keys()))

    # Let's check their heartbeats
    def n_beats(n):
      def f():
        h = agent.prepare_heartbeat()
        logging.info(h)
        assert_equal(n, len(h["marching_order_status"]))
        return h
      return f
    h = _try_a_while(n_beats(2))
    assert_equal(1, len(h["marching_order_status"][0]["supervisee_status"]))
    assert_equal(1, len(h["marching_order_status"][1]["supervisee_status"]))

    # And now let's remove 1, and add 2 new ones
    response = dict(marching_orders=[make_hello_order(2, 9003),
                                     make_hello_order(3, 9004),
                                     make_hello_order(4, 9005)])
    agent.handle_heartbeat_response(response)
    h = _try_a_while(n_beats(3))
    agent.stop()

def _try_a_while(f, timeout=3.0, sleep=0.1):
  """
  Helper function to repeatedly try a call
  until it doesn't throw an exception.
  """
  end = time.time() + timeout
  tries = 0
  last_e = None
  while time.time() < end:
    try:
      return f()
    except Exception, e:
      last_e = e
      time.sleep(sleep)
      tries += 1

  raise Exception("Could not run code after %d tries.  Last exception: %s" %
                  (tries, e))


def make_hello_order(id, port):
  """
  Makes a marching order for a "hello world" service.
  """
  return {
      "id": id,
      "host_name": "localhost",
      "service_name": "test",
      "role_name": "test_role",
      "user": "test_user",
      "group": "test_group",
      # This is "hello.conf" -> "9876\n" in a zip (-0 compressed) file.
      "configuration_data": make_hello_config(port),
      "supervisees": [
        {
          "name": "hello",
          "program": "hello/hello.py",
          "arguments": ["world"],
          "status_links": { "status": "http://localhost:%s/status" % (port,) }
        }
      ],
      "commands": []
    }

def make_hello_config(port):
  io = StringIO.StringIO()
  z = zipfile.ZipFile(io, "w")
  z.writestr("hello.conf", str(port) + "\n")
  z.close()
  return io.getvalue()

def test_tail():
  io = StringIO.StringIO()
  io.write("0" * 3)
  io.write("1" * 3)
  io.seek(0)

  assert_equal("", tail(io, 0))
  assert_equal("0111", tail(io, 4))
  assert_equal("000111", tail(io, 500))

  io = StringIO.StringIO()
  LINE = "potential partial line\nfull line\nanother\n"
  io.write("potential partial line\nfull line\nanother\n")
  io.seek(0)

  assert_equal(LINE, tail(io, 500))
  assert_equal("another\n", tail(io, len("another\n") + 4))

def test_filesystem_info():
  """
  Test filesystem information parsing
  """
  fs_info = FilesystemMap().get_info_for_scm_heartbeat()

  # Check that the expected keys exist in the return dictionary.
  assert_true("filesystem_info" in fs_info.keys())
  assert_true("mountpoint_available_space_bytes" in fs_info.keys())

  partition_info = fs_info["filesystem_info"]
  mountpoint_free_space = fs_info["mountpoint_available_space_bytes"]

  # For each partition we have we should have a corresponding mount point
  # information.
  for partition in partition_info:
      assert_true(partition["mount_point"] in mountpoint_free_space.keys())
      assert_true(mountpoint_free_space[partition["mount_point"]] <
                  partition["total_space_bytes"])

  assert_equal(len(partition_info), len(mountpoint_free_space.keys()))

def test_unzip():
  tmpdir = tempfile.mkdtemp('test_cmf_unzip')
  zipfilename = os.path.join(tmpdir, "x.zip")
  z = zipfile.ZipFile(zipfilename, "w")
  # Tests nested directories.
  z.writestr("a/b/c/foo", "hi there")
  z.close()
  unzip(file(zipfilename).read(), target_directory=tmpdir)
  assert_equal("hi there",
    file(os.path.join(tmpdir, "a/b/c/foo")).read())
  path = tmpdir

  path = os.path.join(path, "a")
  assert_equal(int(0700), os.stat(path).st_mode & int(0777))

  path = os.path.join(path, "b")
  assert_equal(int(0700), os.stat(path).st_mode & int(0777))

  path = os.path.join(path, "c")
  assert_equal(int(0700), os.stat(path).st_mode & int(0777))

  path = os.path.join(path, "foo")
  assert_equal(int(0600), os.stat(path).st_mode & int(0777))

  # Tests that empty data doesn't do anything unexpected.
  pre_unzip = sorted(os.listdir(tmpdir))
  unzip("", target_directory=tmpdir)
  assert_equal(pre_unzip, sorted(os.listdir(tmpdir)))



def test_avro():
  # Avro has a bug out of the box where it considers ints and longs to be valid
  # as doubles. This causes no end of problems for us, as we use unions that have
  # both long and double elements. See http://review.sf.cloudera.com/r/8674/ for
  # more information.
  #
  # These tests below validate that the avro we're using contains a patch that
  # addresses this problem.
  assert_true(validate(Schema('long'), 0L))
  assert_false(validate(Schema('double'), 0L))
  assert_false(validate(Schema('long'), 0.0))
  assert_true(validate(Schema('double'), 0.0))

def test_m2crypto():
  # We've patched the M2Crypto we use to work around OPSAPS-4306 and OPSAPS-3343.
  # The tests below are intended to catch that the patch is present.
  assert_equal(1, m2crypto_scm_patchlevel,
               "M2Crypto missing patch to work around OPSAPS-4306.")


def test_process_kill():
  """
  Test the kill_process method which is used as fallback if the supervisor did
  not shutdown cleanly via RPC.
  """

  with FakeAgentEnv() as agent:
    # Killing a non existing process shouldn't cause problems
    kill_process(123456, 2, 10)

    # Now let's spawn a process that would die nicely
    test_process_path = os.path.join(os.path.dirname(__file__), "test_victim_process.py")
    sp = launch_and_kill_process(agent, [test_process_path, "nice_process"])

    # -15 -> killed with signal SIGTERM
    assert_equal(-1 * signal.SIGTERM, sp.returncode)

    # Now let's spawn a process that would ignore SIGTERM
    sp = launch_and_kill_process(agent, [test_process_path, "bad_process"])
    assert_equal(-1 * signal.SIGKILL, sp.returncode)

def launch_and_kill_process(agent, args):
  """
  Helper function for test_process_kill that launches a process with the given
  args, waits for it to be ready, kills it and waits for it to get reaped by
  the reaper thread.
  The reaseon for having a concurrent thread on wait is because kill_process()
  waits after sending a SIGTERM for the pid to vanish which won't happen until
  the process is joined.
  The function returns Popep object for the process so that the test can check
  its return code.
  """
  sp = subprocess.Popen(
      args,
      stdout=subprocess.PIPE)

  # wait for process to install its signal handler
  child_message = sp.stdout.readline()
  assert_equal("ready\n", child_message)

  # Reaper thread that reaps child processes. It is needed because a pid
  # does not vanish until it's joined by it's parent process.
  reaper = threading.Thread(target=sp.wait)
  reaper.daemon = True
  reaper.start()

  kill_process(sp.pid, 2, 10)
  reaper.join(10)

  return sp

def _get_file_mode(file_name):
  '''
  Returns file's mode as a string in octal format.
  Only the last 4 digits are relevant for the deploy test.
  '''
  return str(oct(os.stat(file_name).st_mode))[-4:]

@patch("cmf.util.chown")
def test_deploy_special_file(chown):
  '''
  Test that files with special deploy instructions are moved correctly.
  '''
  test_dir = 'test_dir'
  os.mkdir(test_dir)

  try:
    # Create the directory structure and files for test
    src_dir = test_dir + '/src'
    dst_dir = test_dir + '/dst'
    dst_dir2 = test_dir + '/dst2'
    dst_dir3 = test_dir + '/dst3'
    dst_dir4 = test_dir + '/dst4'
    dst_dir5 = test_dir + '/dst5'
    dst_dir6 = test_dir + '/dst6'
    dst_dir7 = test_dir + '/dst7'
    os.mkdir(src_dir)
    os.mkdir(dst_dir)
    os.mkdir(dst_dir2)
    os.mkdir(dst_dir3)
    os.mkdir(dst_dir4)
    os.mkdir(dst_dir5)
    os.mkdir(dst_dir6)
    os.mkdir(dst_dir7)

    fd = open(src_dir + "/file1", "w")
    fd.write("blah")
    fd.close()
    os.chmod(src_dir + "/file1", 0666)
    fd = open(src_dir + "/file2", "w")
    fd.write("blah")
    fd.close()
    os.chmod(src_dir + "/file2", 0666)
    fd = open(src_dir + "/file3", "w")
    fd.write("blah")
    fd.close()
    os.chmod(src_dir + "/file3", 0666)

    # Check that files are at the right place
    assert_true(os.path.exists(src_dir + "/file1"))
    assert_equal('0666', _get_file_mode(src_dir + "/file1"))
    assert_true(os.path.exists(src_dir + "/file2"))
    assert_equal('0666', _get_file_mode(src_dir + "/file2"))
    assert_true(os.path.exists(src_dir + "/file3"))
    assert_true(not os.listdir(dst_dir))
    assert_true(not os.listdir(dst_dir2))

    chown_calls = []
    # Deploy file1
    deploy_info = { 'name' : 'file1', 'deploy_dir' : dst_dir , 'permissions' : 0444, "user" : "foo", "group" : "bar"}
    deploy_special_file(deploy_info, src_dir, dict(), None)
    chown_calls.append(call("foo", "bar", src_dir + "/file1", None))
    chown_calls.append(call("foo", "bar", dst_dir + "/file1", None))
    chown.assert_has_calls(chown_calls)
    assert_true(os.path.exists(src_dir + "/file1"))
    assert_true(os.path.exists(src_dir + "/file2"))
    assert_equal('0666', _get_file_mode(src_dir + "/file2"))
    assert_true(os.path.exists(src_dir + "/file3"))
    assert_true(os.path.exists(dst_dir + "/file1"))
    assert_equal('0444', _get_file_mode(dst_dir + "/file1"))
    assert_true(not os.listdir(dst_dir2))

    # Deploy file2 using env var in the deploy_dir
    deploy_info = { 'name' : 'file2', 'deploy_dir' : '$TEST_DIR/dst2' , 'permissions' : 0440, "user" : "foo", "group" : "bar"}
    deploy_special_file(deploy_info, src_dir, {'TEST_DIR' : test_dir}, None)
    chown_calls.append(call("foo", "bar", src_dir + "/file2", None))
    chown_calls.append(call("foo", "bar", test_dir + "/dst2/file2", None))
    chown.assert_has_calls(chown_calls)
    assert_true(os.path.exists(src_dir + "/file1"))
    assert_true(os.path.exists(src_dir + "/file2"))
    assert_true(os.path.exists(src_dir + "/file3"))
    assert_true(os.path.exists(dst_dir + "/file1"))
    assert_equal('0444', _get_file_mode(dst_dir + "/file1"))
    assert_true(os.path.exists(dst_dir2 + "/file2"))
    assert_equal('0440', _get_file_mode(dst_dir2 + "/file2"))
    assert_equal(4, chown.call_count)

    # Deploying file3 with a bad path should fail
    deploy_info = { 'name' : 'file3', 'deploy_dir' : dst_dir + "blah/", "user" : "foo", "group" : "bar" }
    assert_raises(Exception, deploy_special_file, deploy_info, src_dir, dict(), None)

    # Deploying file3 on top of a file should fail
    deploy_info = { 'name' : 'file3', 'deploy_dir' : dst_dir + "/file1", "user" : "foo", "group" : "bar" }
    assert_raises(Exception, deploy_special_file, deploy_info, src_dir, dict(), None)

    # Deploying file3 with a symlink should work
    os.symlink(test_dir + "/dst", "dst_sym")
    chown.reset_mock()
    chown_calls = []
    try:
      deploy_info = { 'name' : 'file3', 'deploy_dir' : 'dst_sym' , 'permissions' : 0442, "user" : "foo", "group" : "bar"}
      deploy_special_file(deploy_info, src_dir, dict(), None)
      chown_calls.append(call("foo", "bar", src_dir + "/file3", None))
      chown_calls.append(call("foo", "bar", "dst_sym/file3", None))
      chown.assert_has_calls(chown_calls)
      assert_true(os.path.exists(src_dir + "/file1"))
      assert_true(os.path.exists(src_dir + "/file2"))
      assert_true(os.path.exists(src_dir + "/file3"))
      assert_true(os.path.exists(dst_dir + "/file1"))
      assert_equal('0444', _get_file_mode(dst_dir + "/file1"))
      assert_true(os.path.exists(dst_dir2 + "/file2"))
      assert_equal('0440', _get_file_mode(dst_dir2 + "/file2"))
      assert_true(os.path.exists(dst_dir + "/file3"))
      assert_equal('0442', _get_file_mode(dst_dir + "/file3"))
    finally:
      os.unlink('dst_sym')

    # Deploy file using absolute path
    chown.reset_mock()
    chown_calls = []
    try:
      deploy_info = { 'name' : os.path.abspath(src_dir + "/file1"), 'deploy_dir' : dst_dir3 , 'permissions' : 0444, "user" : "foo", "group" : "bar"}
      deploy_special_file(deploy_info, None, dict(), None)
      chown_calls.append(call("foo", "bar", os.path.abspath(src_dir + "/file1"),  None))
      chown_calls.append(call("foo", "bar", dst_dir3 + "/file1", None))
      chown.assert_has_calls(chown_calls)
      assert_true(os.path.exists(src_dir + "/file1"))
      assert_true(os.path.exists(dst_dir3 + "/file1"))
      assert_equal('0444', _get_file_mode(dst_dir3 + "/file1"))
    finally:
      os.unlink(dst_dir3 + "/file1")

    # Deploy file using absolute path and deploy_dir is absolute
    chown.reset_mock()
    chown_calls = []
    try:
      deploy_info = { 'name' : os.path.abspath(src_dir + "/file1"), 'deploy_dir' : os.path.abspath(dst_dir5) , 'permissions' : 0444, "user" : "foo", "group" : "bar"}
      deploy_special_file(deploy_info, None, dict(), None)
      chown_calls.append(call("foo", "bar", os.path.abspath(src_dir + "/file1"),  None))
      chown_calls.append(call("foo", "bar", os.path.abspath(dst_dir5 + "/file1"), None))
      chown.assert_has_calls(chown_calls)
      assert_true(os.path.exists(src_dir + "/file1"))
      assert_true(os.path.exists(dst_dir5 + "/file1"))
      assert_equal('0444', _get_file_mode(dst_dir5 + "/file1"))
    finally:
      os.unlink(dst_dir5 + "/file1")

    # Deploy file using absolute path and environment variable in source
    chown.reset_mock()
    chown_calls = []
    try:
      deploy_info = { 'name' : "$TEST_DIR/file1", 'deploy_dir' : dst_dir4 , 'permissions' : 0444, "user" : "foo", "group" : "bar"}
      deploy_special_file(deploy_info, None, {'TEST_DIR' : os.path.abspath(src_dir)}, None)
      chown_calls.append(call("foo", "bar", os.path.abspath(src_dir + "/file1"), None))
      chown_calls.append(call("foo", "bar", dst_dir4 + "/file1", None))
      chown.assert_has_calls(chown_calls)
      assert_true(os.path.exists(src_dir + "/file1"))
      assert_true(os.path.exists(dst_dir4 + "/file1"))
      assert_equal('0444', _get_file_mode(dst_dir4 + "/file1"))
    finally:
      os.unlink(dst_dir4 + "/file1")

    # Deploy file using absolute path and environment variable in source and environment variable in dest
    chown.reset_mock()
    chown_calls = []
    try:
      deploy_info = { 'name' : "$TEST_DIR/file1", 'deploy_dir' : "$DST_DIR6" , 'permissions' : 0444, "user" : "foo", "group" : "bar"}
      deploy_special_file(deploy_info, None, {'TEST_DIR' : os.path.abspath(src_dir), 'DST_DIR6' : dst_dir6 }, None)
      chown_calls.append(call("foo", "bar", os.path.abspath(src_dir + "/file1"), None))
      chown_calls.append(call("foo", "bar", dst_dir6 + "/file1", None))
      chown.assert_has_calls(chown_calls)
      assert_true(os.path.exists(src_dir + "/file1"))
      assert_true(os.path.exists(dst_dir6 + "/file1"))
      assert_equal('0444', _get_file_mode(dst_dir6 + "/file1"))
    finally:
      os.unlink(dst_dir6 + "/file1")

    # Deploy file using absolute path and environment variable in source and environment variable in dest as absolute path
    chown.reset_mock()
    chown_calls = []
    try:
      deploy_info = { 'name' : "$TEST_DIR/file1", 'deploy_dir' : "$DST_DIR7" , 'permissions' : 0444, "user" : "foo", "group" : "bar"}
      deploy_special_file(deploy_info, None, {'TEST_DIR' : os.path.abspath(src_dir), 'DST_DIR7' : os.path.abspath(dst_dir7) }, None)
      chown_calls.append(call("foo", "bar", os.path.abspath(src_dir + "/file1"), None))
      chown_calls.append(call("foo", "bar", os.path.abspath(dst_dir7 + "/file1"), None))
      chown.assert_has_calls(chown_calls)
      assert_true(os.path.exists(src_dir + "/file1"))
      assert_true(os.path.exists(dst_dir7 + "/file1"))
      assert_equal('0444', _get_file_mode(dst_dir7 + "/file1"))
    finally:
      os.unlink(dst_dir7 + "/file1")
  finally:
    shutil.rmtree(test_dir)

def test_urlopen_with_timeout_timeout():
  s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
  s.bind( ('', 0) )
  port = s.getsockname()[1]
  s.listen(2) # backlog conections

  now = time.time()
  try:
    urlopen_with_timeout("http://localhost:%s/" % (port,), timeout=0.1)
    assert_true(False, "Should have timed out.")
  except:
    # Note that different versions throw different exceptions here.
    pass
  elapsed = time.time() - now
  assert_true(elapsed >= 0.09, msg="Elapsed: %s" % (elapsed,))
  assert_true(elapsed < 20)

  now = time.time()
  try:
    urlopen_with_timeout("http://localhost:%s/" % (port,), timeout=1)
    assert_true(False, "Should have timed out.")
  except:
    pass
  elapsed = time.time() - now
  assert_true(elapsed >= 0.9, msg="Elapsed: %s" % (elapsed,))
  assert_true(elapsed < 20)

  s.close()

def test_urlopen_with_timeout_successful():
  # Start an HTTP server in a separate thread.
  server = SocketServer.TCPServer(('', 0), SimpleHTTPServer.SimpleHTTPRequestHandler)
  port = server.socket.getsockname()[1]
  # Serve just one request.  server.shutdown() not available in py2.4
  server_thread = threading.Thread(target=server.handle_request)
  server_thread.start()

  try:
    # Make sure a simple get works.
    data = urlopen_with_timeout("http://localhost:%s/" % (port,), timeout=300).read()
  finally:
    server_thread.join()

def test_urllib2_kerberos_changes():
  # Start an HTTP server in a separate thread.
  server = SocketServer.TCPServer(('', 0), SimpleHTTPServer.SimpleHTTPRequestHandler)
  port = server.socket.getsockname()[1]
  # Serve just one request.  server.shutdown() not available in py2.4
  server_thread = threading.Thread(target=server.handle_request)
  server_thread.start()

  try:
    opener = urllib2.build_opener()
    # Make sure specifying the new param to urllib2_kerberos doesn't blow up
    # The service name doesn't matter since it won't exist, we want to still be
    # able to fetch the url regardless of whether or not the connection is actually
    # secure. It's too hard to also unit test whether or not the actual security
    # works because we need a real kdc. We test security via systest.
    opener.add_handler(HTTPKerberosAuthHandler("dummy-service-name"))
    data = opener.open("http://localhost:%s/" % (port,)).read()
  finally:
    server_thread.join()

def test_urllib2_kerberos_401error():
  # The urllib2_kerberos library handles the 401 error to try to establish a secure
  # connection. This checks that a 401 error that isn't actually from a secure HTTP
  # server doesn't blow up. Basically normal 401 errors shouldn't change anything.
  class HTTPRequestHandlerWithError(SimpleHTTPServer.SimpleHTTPRequestHandler):
    def do_GET(self):
      self.send_error(401)

  # Start an HTTP server in a separate thread.
  server = SocketServer.TCPServer(('', 0), HTTPRequestHandlerWithError)
  port = server.socket.getsockname()[1]
  # Serve just one request.  server.shutdown() not available in py2.4
  server_thread = threading.Thread(target=server.handle_request)
  server_thread.start()

  got_expected_error = False
  try:
    try:
      opener = urllib2.build_opener()
      opener.add_handler(HTTPKerberosAuthHandler("dummy-service-name"))
      data = opener.open("http://localhost:%s/" % (port,)).read()
    except urllib2.HTTPError, ex:
      got_expected_error = True
  finally:
    server_thread.join()
  assert_true(got_expected_error, "Didn't get expected HTTPError")

def test_get_full_url():
  url = "test"
  full_url = _get_full_url(url)
  assert_equal("test", full_url)

  try:
    full_url = _get_full_url(None)
    raise Exception
  except:
    pass

def test_throttling_logger_throttle():
  class MockLogger(object):
    def __init__(self):
      self.msg_count = 0
      self.last_msg = None

    def get_count(self):
      return self.msg_count

    def get_last_msg(self):
      return self.last_msg

    # warning may be called by throttling logger if the dictionary is full
    def warning(self, msg, *args, **kwargs):
      pass

    def info(self, msg, *args, **kwargs):
      self.last_msg = msg
      self.msg_count += 1

  mock_logger = MockLogger()
  throttle_logger = ThrottlingLogger(mock_logger)

  # write the same message for 3 seconds
  start = time.time()
  while time.time() - start < 1:
    throttle_logger.info("test")

  # count should be 1
  assert_equal(1, mock_logger.get_count())
  assert_equal("test", mock_logger.get_last_msg())

  # set the throttle period to 1 sec to make sure more messages get logged
  mock_logger = MockLogger()
  throttle_logger = ThrottlingLogger(mock_logger, 1)
  start = time.time()
  count = 0
  while time.time() - start < 2.0001:
    count += 1
    throttle_logger.info("test")
  assert_true(mock_logger.get_count() >= 2)
  assert_true(mock_logger.get_last_msg().endswith(" skipped) test"))

  # Test that the map gets cleared
  mock_logger = MockLogger()
  throttle_logger = ThrottlingLogger(mock_logger, 20)
  start = time.time()
  for i in range(0,1000):
    throttle_logger.info("%s" % i)
  assert_equal(1000, mock_logger.get_count())
  throttle_logger.get_time = lambda: 20.01
  for i in range(1000,2000):
    throttle_logger.info("%s" % i)
  assert_equal(2000, mock_logger.get_count())

def test_throttling_logger_wrapper():
  log = logging.Logger("test.logger.1", level=logging.DEBUG)
  strio = StringIO.StringIO()
  handler = logging.StreamHandler(strio)
  log.addHandler(handler)

  # try each of the logging functions without args
  log.debug("debug")
  log.info("info")
  log.warn("warn")
  log.warning("warning")
  log.error("error")
  log.critical("critical")
  log.exception("exception")

  # try each of the logging functions with args
  log.debug("debug %s", 1)
  log.info("info %s", 2)
  log.warn("warn %s", 3)
  log.warning("warning %s", 4)
  log.error("error %s", 5)
  log.critical("critical %s", 6)
  log.exception("exception %s", 7)

  expected_log = '''debug
info
warn
warning
error
critical
exception
None
debug 1
info 2
warn 3
warning 4
error 5
critical 6
exception 7
None
'''

  strio.flush()
  assert_equal(expected_log, strio.getvalue())

def test_validate_and_open():
  test_dir = tempfile.mkdtemp()
  try:
    # Setup some directories, files and symlinks
    dir1 = os.path.join(test_dir, 'dir1')
    dir2 = os.path.join(test_dir, 'dir2')
    file1 = os.path.join(dir1, "file1")
    file2 = os.path.join(dir2, "file2")

    os.mkdir(dir1)
    os.mkdir(dir2)
    fd1 = open(file1, "w")
    fd2 = open(file2, "w")

    f2_sym = "f2_sym"
    os.symlink(file2, f2_sym)

    # Invalid file owner
    assert_raises(Exception, validate_and_open, file1, valid_owners=["fake_user"])

    # File not under specified parent
    assert_raises(Exception, validate_and_open, file2, parent_path=dir1, valid_owners=None)

    # Don't follow symlinks
    assert_raises(Exception, validate_and_open, f2_sym, parent_path=dir1, valid_owners=None)

    # Most restricted case
    current_user = pwd.getpwuid(os.getuid())[0]
    fd3 = validate_and_open(file1, parent_path=dir1, valid_owners=[current_user])
    assert_true(os.path.sameopenfile(fd3.fileno(), fd1.fileno()))

    # Having more than one hardlink should fail the no_links check
    os.link(file1, os.path.join(test_dir, "f1_hardlink"))
    assert_raises(Exception, validate_and_open, file1, parent_path=dir1, valid_owners=[current_user])

    # But hardlinks are allowed for directories
    fdd = os.open(dir1, os.O_RDONLY)
    realfd = validate_and_open_fd(dir1, valid_owners=[current_user])
    assert_equal(os.fstat(realfd), os.fstat(fdd))
    os.close(fdd)

    # Allow symlinks if specified
    fd5 = validate_and_open(os.path.abspath(f2_sym), no_links=False, valid_owners=None)
    assert_true(os.path.sameopenfile(fd5.fileno(), fd2.fileno()))

    # Reject blacklisted prefixes
    assert_raises(Exception, validate_and_open, "/etc/passwd", path_blacklist=["/etc"])

    # Access allowed without blacklist
    fd6 = validate_and_open("/etc/passwd")

  finally:
    if os.path.exists(f2_sym):
      os.unlink(f2_sym)
    fd6.close()
    fd5.close()
    os.close(realfd)
    fd3.close()
    fd2.close()
    fd1.close()
    shutil.rmtree(test_dir)

def test__extract_err_str_and_cm_status_code():
  err_str, status_code = Process._extract_err_str_and_cm_status_code("CM_STATUS_CODE=HDFS_ERR")
  assert_equal(status_code, "HDFS_ERR")
  assert_equal(err_str, "")

  err_str, status_code = Process._extract_err_str_and_cm_status_code('''abcd
CM_STATUS_CODE=HDFS_ERR''')
  assert_equal(status_code, "HDFS_ERR")
  assert_equal(err_str, "abcd")

  err_str, status_code = Process._extract_err_str_and_cm_status_code("")
  assert_equal(status_code, "")
  assert_equal(err_str, "")

  err_str, status_code = Process._extract_err_str_and_cm_status_code(None)
  assert_equal(status_code, "")
  assert_equal(err_str, "")

def test__add_cm_status_codes_to_env_dict():
  # Check that the CM status codes are added correctly to env.
  env = dict()
  add_cm_status_codes_to_env_dict(env)
  assert_equal(env[CM_STATUS_CODES_ENV_VAR], get_cm_status_codes())

def get_cm_status_codes():
  return str.join(" ", CM_STATUS_CODE_SCHEMA.symbols)

# This is not a unit-test for any specific function, but rather validates the
# CM status related helper functions in the service shell script
# cloudera_config.sh and ensures that they output status codes in the format
# expected by the agent.
def test_validate_service_script_cm_status_utils():
  codes = CM_STATUS_CODE_SCHEMA.symbols;

  script_file = os.path.join(os.path.dirname(__file__),
                             "..",
                             "..",
                             "service",
                             "common",
                             "cloudera-config.sh")

  common_script = CM_STATUS_CODES_ENV_VAR + '="' + get_cm_status_codes() + '";'
  common_script += ". %s; set +xv; echo abcd 1>&2;" % script_file

  # Verify that the service shell script utility function cm_error_and_exit
  # outputs the status code in the correct format and returns the correct exit
  # code.
  proc = subprocess.Popen("bash", shell=True, stdin=subprocess.PIPE, stderr=subprocess.PIPE)
  test_script = common_script + "cm_error_and_exit %s" % codes[1]
  stdout, stderr = proc.communicate(test_script)
  assert_equal(proc.returncode, 1)
  err_str, status_code = Process._extract_err_str_and_cm_status_code(stderr)
  assert_equal(status_code, codes[1])
  assert_equal(err_str[len(err_str) - 4:], "abcd")

  # Verify that the service shell script utility function cm_success_and_exit
  # outputs the status code in the correct format and returns the correct exit
  # code.
  proc = subprocess.Popen("bash", shell=True, stdin=subprocess.PIPE, stderr=subprocess.PIPE)
  test_script = common_script + "cm_success_and_exit %s" % codes[1]
  stdout, stderr = proc.communicate(test_script)
  assert_equal(proc.returncode, 0)
  err_str, status_code = Process._extract_err_str_and_cm_status_code(stderr)
  assert_equal(status_code, codes[1])
  assert_equal(err_str[len(err_str) - 4:], "abcd")

def test_source():
  FIXTURE_PATH = os.path.join(os.path.dirname(__file__), "test-fixtures", "source-fixture")

  assert_true('FOO' not in os.environ)
  assert_true('BAR' not in os.environ)
  new_env = source(os.path.join(FIXTURE_PATH, "test-source.sh"), 'bar', os.environ)
  assert_equal(new_env['FOO'], 'foo')
  assert_equal(new_env['BAR'], 'bar')

def test_supervisee_conf():
  # supervisord's parsing code requires this to be a real directory.
  tmpdir = tempfile.mkdtemp()
  os.makedirs(os.path.join(tmpdir, "logs"))

  executable = "/bin/ls"
  # Note that multiple newlines will get collapsed into one.
  # (which is broken).
  args = [ "yes", '"no', "\'\n_newline" ]
  envs = {
    'A': "13",
    'B': "contains,comma",
    'C': "contains\nnewline",
    'D': 'contains"doublequote',
    # These two fail!
    'E': "single'quote",
    'F': "newline\n withspace"
  }
  bad_keys = ('E', 'F')
  data = {
    "program_name": "x",
    "command": executable,
    "arguments": args,
    "directory": tmpdir,
    "environment": envs,
    "one_off": True,
    "auto_restart": True,
    "exitcodes": "1,2,3",
    "limit_fds": 13,
    "limit_memolock": 123,
    "extra_groups": "daemon,adm"
  }

  sup_config = render_text("supervisee.conf",
    data=data)

  # Now let's parse the configuration just like supervisor
  # would.
  from supervisor.options import ServerOptions
  from supervisor.process import Subprocess
  server_options = ServerOptions()
  # we need convert sup_config from unicode to ascii str since
  # render_text() returns unicode string
  parsed = server_options.read_config(StringIO.StringIO(str("[supervisord]\n" + sup_config)))
  process = parsed.process_group_configs[0].process_configs[0]

  # Simple one that should have no trouble.
  assert_equal(13, process.limit_fds)

  # The execv arguments have tricky escaping code.
  assert_equal( (executable, [executable] + args),
    Subprocess(process).get_execv_args() )

  process_env = process.environment
  # Sadly, supervisor's parsing here can't
  # handle being genuinely isomorphic.
  for k in bad_keys:
    del process_env[k]
    del envs[k]
  assert_equal(envs, process_env)


def test_reported_hostname():
  """
  Tests if agent heartbeat contains the reported_hostname setting through the command line options
  """

  agent_reported_hostname = "foo"
  with FakeAgentEnv(reported_hostname=agent_reported_hostname) as agent:
    agent.supervisor_check = Mock()
    agent.get_supervisor_data = Mock(side_effect=[[],])
    # Prepare a heartbeat
    h = agent.prepare_heartbeat()
    assert_equal(h["status"]["host"]["host_name"], agent.cfg.reported_hostname)
    assert_equal(agent.cfg.reported_hostname, agent_reported_hostname)

def test_path_cleanup():
  """
  Tests cmf.util.remove_agent_path
  """
  base = os.path.split(sys.argv[0])[0]
  path = ":".join([base, base, "good", base])
  out = remove_agent_path(path)
  assert_equal(out, "good")


def test_agent_hb_sequence():
  """
  Test for correct handling of initial HBs and that agent ignores HBs with invalid data.
  This test is focused on driving the Agent.__issue_heartbeat method.
  :return:
  """

  def run_test(data):
    # Define a closure extension class - so we don't have to add attributes to Agent.
    class HBTestAgent(Agent):

      def prepare_heartbeat(self):
        data.hb_id += 1
        return data.hb_id - 1

      def send_heartbeat(self, heartbeat):
        # Generate a HB based on "hb_sequence" - which will be a valid or invalid
        # HB response, based on the sequence
        assert_equal(data.last_sent_hb, heartbeat, "Wrong HB sequence")
        data.last_sent_hb = heartbeat + 1

        if data.hb_sequence[heartbeat]:
          return {"data": {"id": heartbeat , "cm_guid": 1}}
        return {"data": None}

      def _init_after_first_heartbeat_response(self, response_data):
        # This gets called only on the first HB that has response data
        assert_false(data.done_init, "Repeated post-first-HB init")
        assert_equal(response_data['id'], data.fails, "Failed to ignore initial invalid HBs")
        self.cm_guid = 1

      def handle_heartbeat_response(self, response):
        # This gets called only with non-None response data
        assert_true(response is not None, "Failed to ignore invalide HB response")
        return

    # Initialize data in the closure.
    data.hb_id = 0
    data.last_sent_hb = 0
    data.done_init = False

    # Count failure at the start - identifies id of first good HB
    data.fails = 0
    for x in data.hb_sequence:
      if x is None:
        data.fails += 1
      else:
        break

    class FakeHBAgentEnv(FakeAgentEnv):
      agent_class = HBTestAgent

    with FakeHBAgentEnv() as agent:
      for x in data.hb_sequence:
        agent._Agent__issue_heartbeat()

  # First HB is failed
  class FirstFailure(object):
    hb_sequence = [None, 1, 1, 1, 1]

  # First 2 HB are failed
  class First2Failure(object):
    hb_sequence = [None, None, 1, 1, 1, 1]

  # No failed HB's
  class NoFailure(object):
    hb_sequence = [1, 1, 1, 1, 1]

  # Some failures late in the HB sequence.
  class LateFailure(object):
    hb_sequence = [1, 1, 1, None, 1, None, None, 1]

  run_test(FirstFailure())
  run_test(First2Failure())
  run_test(NoFailure())
  run_test(LateFailure())


