# Copyright (c) 2015 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import, print_function

import logging
import sys
try:
  import simplejson as json
except ImportError:
  import json

from flood import TORRENT_SCHEMA
from flood.torrent import TorrentItem
from flood.util.async import AsyncExec
from flood.util.cmd import command, CommandContext, chained_command
from flood.util.config import Param
from flood.util.meta import GenericPrinter
from flood.util.misc import expandRangeList
from flood.util.rpc import resolveHostPort, RPCClient
from flood.util.error import Error

log = logging.getLogger()


def printTable(headers, values):
  print(" ".join(["%-8s" %x for x in headers]))
  for k in sorted(values.keys()):
    vals = values[k]
    print(" ".join(["%-8s" % x for x in [k] + vals]))


class TorrentRPCClient(RPCClient):
  '''
  RPCClient class conforming to the TORRENT_SCHEMA.
  Note that the class has implicitly generated methods matching
  the signatures of the TORRENT_SCHEMA AVRO schema.
  '''

  SCHEMA = TORRENT_SCHEMA
  DEFAULT_RPC_PORT = 6665

  @command("Define a flood torrent cluster membership set.",
           hostSpec = Param("Spec of hosts to add"))
  def setCluster(self, hostSpec):
    hosts = expandRangeList(hostSpec)
    for host in hosts:
      hp = resolveHostPort(host, 6666)
      self.AddDHTNode(hp.host, hp.port)
      print("Defined node: %s %s" %(hp.host, hp.port))

  def consume(self, prefixObj, args):
    result = super(TorrentRPCClient, self).consume(prefixObj, args)
    if result is not None:
      GenericPrinter(result)


  def _status(self, torrent):
    '''
    Return the interpreted json document from the AVRO GetTorrentStatus RPC.
    :param torrent: Name of torrent to query.
    :return: As above.
    '''
    result = self.GetTorrentStatus(torrent)
    doc = json.loads(str(result['jsonDoc']))
    return doc

  def _statusField(self, torrent, **fieldMap):
    '''
    Return an object that has attributes matching the keys of "fieldMap".
    The values of "fieldMap" correspond to the key names (append-joined with ".")
    needed to traverse the json document returned in GetTorrentStatus.
    Those in turn are effectively defined in "Torrent.getStatus()".

    :param torrent:  Name of torrent to query.
    :param fieldMap: See above description.
    :return: As above.
    '''
    doc = self._status(torrent)
    class Response(object):
      def __init__(self, fieldMap):
        self.__fieldMap = fieldMap

      def __str__(self):
        return "[" +  " ".join(["%s=%s" % (k, getattr(self, k)) for k in self.__fieldMap]) + "]"

    # Create a sub-class of "Response" that has attributes corresponding to
    # the key names in fieldMap.
    out = type("ResponseX", (Response,), dict([(x, None) for x in fieldMap.keys()]))(fieldMap)
    for (k, v) in fieldMap.items():
      words = v.split(".")
      d = doc
      for word in words:
        try:
          d = d[word]
        except KeyError:
          raise Error("Missing response: %s", k)
      setattr(out, k, d)

    return out

  def stat(self, torrent):
    '''
    Return a simple structure containing the most useful stats about a torrent.
    :param torrent: Name of torrent to query.
    :return:
    '''
    return self._statusField(torrent,
                             totalSize='stats.total_wanted',
                             totalDownloaded='stats.total_wanted_done',
                             upload = 'stats.total_payload_upload',
                             state = 'stats.state',
                             progress = 'stats.progress_ppm',
                             seed = 'urlSeed',
                             location = 'location')

  def progress(self, torrent):
    data = self._statusField(torrent, progress='stats.progress_ppm')
    return (100 * float(data.progress)/ 1000000)


class RPCCallContext(CommandContext):
  def __init__(self, resolved, name, parent):
    CommandContext.__init__(self, name, parent)
  def lookup(self, name):
    return TorrentRPCClient(expandRangeList(name)[0])

class CLIContext(CommandContext):
  '''
  Top-level command line context interpreter.
  All "command" annotated methods are available as CLI sub-commands.
  '''

  @command("Create a .torrent file for a specified file.",
           target = Param("Path to file to create torrent for."),
           url_seed = Param("URL to torrent seed."))
  def mkTorrent(self, target, url_seed=None):
    TorrentItem.Generate(target, url_seed=url_seed)

  @chained_command("Perform RPC call against specified server",
                   server = Param("Host:port of server to contact"))
  def rpcCall(self, server):
    return TorrentRPCClient(server)


  @command("Show progress of torrents amongst several nodes.",
           nodes = Param("Set of nodes to query"),
           torrent = Param("Name of torrent to query"))
  def torrentStatus(self, nodes, torrent):

    table = {}


    class AsyncStatus(AsyncExec):
      '''
      Mechanism for fetching status for a torrent from hosts in a
      group of hosts.
      '''
      def run(self, idx, host):
        rpc = TorrentRPCClient(host)
        result = rpc.stat(torrent)
        table[idx+1] = ["%2.2f%%" % (100 * float(result.progress)/1000000),
                        "%2.2f%%" % (100 * float(result.upload)/result.totalSize),
                        result.seed]

      def complete(self, idx, host, result, exc):
        if exc is not None:
          log.error("Failed host %s %s", idx+1, exc)


    hosts = expandRangeList(nodes)
    AsyncStatus().execute(hosts)
    printTable(("host", "download", "upload", "seed"), table)




def main():
  logging.basicConfig()
  logging.getLogger().setLevel(logging.INFO)
  return CLIContext().execute(sys.argv[1:])

if __name__ == "__main__":
  sys.exit(main())
