# Copyright (c) 2015 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import, print_function

import logging
import shutil

try:
  import libtorrent
except ImportError:
  libtorrent = None
import os
import os.path
import requests
import threading
import stat

from flood import TorrentError
import flood.torrent as torrent
from flood.util.error import Error

log = logging.getLogger(__file__)

class Depot(object):
  torrentSuffix = ".torrent"
  def __init__(self, path, expiration_time_sec=0, session=None):
    """
    Initialize a depot - manager object for torrents stored on the filesystem.
    :param path: Location of the depot.
    :param expiration_time_sec: Torrents that have not been modified in more than
           this many seconds may be removed. 0 value disables this behavior.
    :param session: libtorrent session for this Depot.
    :return:
    """
    self.path = path
    if not os.path.isdir(path):
      log.warn("Attempting to create depot directory: %s", path)
      os.makedirs(path)

    self.torrents = {}
    self.expiration_time_sec = expiration_time_sec
    self.session = session
    if expiration_time_sec != 0:
      log.info("Expiring torrents after %s seconds.", expiration_time_sec)
      threading.Timer(expiration_time_sec, self.expiration_check).start()
    self.refresh()

  def expiration_check(self):
    """
    Remove any torrents that have expired (have not changed in "expiration_time_sec"
    seconds.
    :return:
    """
    try:
      log.info("Running expiration check.")
      for torrent in self.torrents.values():
        if torrent.is_expired(self.expiration_time_sec):
          log.info("Remove expired torrent: %s", torrent.tFile)
          torrent.remove(self.session)
    except Exception, e:
      log.info("Failed expiration check", e)
    threading.Timer(self.expiration_time_sec, self.expiration_check).start()

  def importTorrent(self, torrentUrl):
    """
    Register a URL identifying a torrent to be downloaded and shared to peers.

    :param torrentUrl: Specifies URL of torrent file to be fetched.
    :return:
    """
    torrentName = os.path.split(torrentUrl)[1]
    if not torrentName.endswith(self.torrentSuffix):
      torrentName += self.torrentSuffix
    tFile = os.path.join(self.path, torrentName)
    if torrentUrl.startswith("file://"):
      path = torrentUrl[7:]

      # Require all path elements to be world readable if we are root - we don't know who's asking for this.
      p = path
      while p != "/" and os.getuid() == 0:
        if (os.stat(p).st_mode & stat.S_IROTH) == 0:
          raise Error("Not world readable path element: %s", p)
        p = os.path.split(p)[0]

      shutil.copy(path, tFile)
    else:
      req = requests.get(torrentUrl)
      if req.status_code != requests.codes.ok:
        raise Error("Failed to fetch torrent: %s: %s", torrentUrl, req.status_code)

      with open(tFile, 'wb') as fd:
        for chunk in req.iter_content(16<<10):
          fd.write(chunk)

    try:
      t = torrent.TorrentItem(tFile, self)
    except RuntimeError:
      raise TorrentError("Invalid torrent: %s", torrentName)
    torrentName = t.info.name()
    torrentDir = os.path.join(self.path, torrentName)
    if os.path.isdir(torrentDir):
      raise TorrentError("Already present torrent: %s", torrentName)
    os.makedirs(torrentDir)

    self.torrents[torrentName] = t
    return t

  def refresh(self):
    torrents = {}
    for ent in os.listdir(self.path):
      if ent.endswith(self.torrentSuffix):
        try:
          t = torrent.TorrentItem(os.path.join(self.path, ent), self)
        except RuntimeError:
          log.error("Invalid torrent: %s", ent)
          continue
        torrents[t.info.name()] = t
    self.torrents = torrents

  def getTorrent(self, torrent):
    if isinstance(torrent, libtorrent.torrent_handle):
      torrent = torrent.name()
    return self.torrents[torrent]
