# Copyright (c) 2015 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import, print_function

import logging
import logging.handlers
import sys
import threading
import subprocess
import re

import argparse
import avro.protocol
import os.path

try:
  import libtorrent
except ImportError:
  libtorrent = None


from flood.depot import Depot
from flood.util.rpc import rpc, RPCServer
from flood.torrent import TorrentItem
from flood.util.config import ConfigValue, IntegerConfigValue, Config
from flood.util.error import Error
from flood import TORRENT_SCHEMA, TorrentError

MAX_LOG_SIZE_BYTES = (1<<20)  # 1MB
LOG_BACKUP_COUNT = 10
LOG_FORMAT = '[%(asctime)s] %(process)d %(threadName)s %(module)-12s %(levelname)-8s %(message)s'
DATE_FORMAT = '%d/%b/%Y %H:%M:%S %z'

logging.basicConfig(format=LOG_FORMAT)
logging.getLogger().setLevel(logging.INFO)
log = logging.getLogger()


class DepotConfig(ConfigValue):
  name = "depotPath"
  default = os.getcwd()

class RPCBindAddrConfig(ConfigValue):
  name = "rpcBindAddr"
  default = "127.0.0.1"
  description = "Address to bind RPC server socket to."

class RPCPortConfig(IntegerConfigValue):
  name = "rpcPort"
  description = "TCP Port for RPC service."
  default = 7190

class TorrentPortConfig(IntegerConfigValue):
  name = "torrentPort"
  description = "TCP Port for torrent traffic"
  default = 7191

class DepotExpiration(IntegerConfigValue):
  name = "depotExpiration"
  description = "Number of seconds to keep idle torrent files (0 disables)."
  default = 0

class LogFileName(ConfigValue):
  name = "logFileName"
  description = "Name of log file."
  default = "cloudera-flood.log"

class GlobalCommand(RPCServer):
  AVRO_SCHEMA = TORRENT_SCHEMA
  def __init__(self, *pargs, **kwargs):
    self.session = None
    self.depot = None
    self.peerNodes = set()

  def handleAlerts(self):
    '''
    Thread for picking up and logging "alerts" from libtorrent.
    These alerts are usually logged and saved to torrent-specific
    alert logs that can be fetched via GetTorrentStatus().
    :return:
    '''
    while True:
      level = logging.DEBUG
      self.session.wait_for_alert(100)
      alert = self.session.pop_alert()
      if alert is None:
        continue


      alertMap = {libtorrent.peer_alert: logging.DEBUG,
                  libtorrent.dht_get_peers_alert: logging.DEBUG,
                  libtorrent.incoming_connection_alert: logging.INFO}
      match = False
      for (klass, lvl) in alertMap.iteritems():
        if isinstance(alert, klass) and lvl >= level:
          level = lvl
          match = True
      if match == False:
        level = logging.INFO

      if isinstance(alert, libtorrent.stats_alert):
        for x in alert.transferred:
          if x != 0:
            level = logging.INFO
            break
        else:
          continue

      if isinstance(alert, libtorrent.torrent_deleted_alert):

        hval = str(alert.info_hash)
        try:
          t = TorrentItem.HASH_MAP[hval]
        except KeyError:
          continue
        t.addAlert(alert)
        continue

      if isinstance(alert, libtorrent.torrent_alert):
        if alert.handle.is_valid():
          t = self.depot.getTorrent(alert.handle)
          t.addAlert(alert)
      log.log(level, "%s %s", alert.what(), alert.message())

  def config_logging(self, filename):
    '''
    Configure Flood daemon logging to the specified file.
    A rotating file handler will be instantiated to use this file
    (and thus corresponding "rolled" files).

    :param filename:
    :return:
    '''

    formatter = logging.Formatter(LOG_FORMAT, DATE_FORMAT)
    logLevel = logging.INFO
    root_logger = logging.getLogger()
    assert 1 == len(root_logger.handlers)
    original_handler = root_logger.handlers[0]

    if filename == "-" or filename is None:
      return

    try:
      rh = logging.handlers.RotatingFileHandler(filename,
                                                maxBytes=MAX_LOG_SIZE_BYTES,
                                                backupCount=LOG_BACKUP_COUNT)
      rh.setFormatter(formatter)
      rh.setLevel(logLevel)
      logging.info("Logging to " + filename)
      root_logger.addHandler(rh)
      root_logger.removeHandler(original_handler)
    except:
      logging.exception("Could not open log file; writing to stderr.")

  def check_port_conflicts(self, port_arg):
    '''
    Exit if there are any processes running using the port this process is expected to use.
    '''

    port_in_use = False
    port = str(port_arg)
    # ie. check for python 2.7
    if len(port) == 0 or not hasattr(subprocess, "check_output"):
      return
    try:
      args = []
      args += ["sport", "=", port]
      cmd = ["ss", "-np", "state", "listening", "("] +  args + [')']

      output = subprocess.check_output(cmd, shell=False)
      pid_match_re = re.compile('pid=([0-9]*),')
      pids = set()
      for line in output.split("\n"):
        m = pid_match_re.search(line)
        if m:
          pids.add(m.group(1))

      for proc in pids:
        cmdline = "<N/A>"
        try:
          with open(os.path.join('/proc/%s/cmdline' % proc)) as f:
            cmdline = f.read()
        except:
          pass
        log.error("Port conflict with pid %s:  %s", proc,
                 cmdline.replace('\000', ' '))
    except Exception as exc:
      log.error("Failed port check: %s", exc)
    if port_in_use:
      log.error("Exiting...")
      sys.exit(1)

  def run(self):
    '''
    Runs the flood torrent service.
    '''
    parser = argparse.ArgumentParser(sys.argv[0],
                                     description = __doc__)
    cfg = Config()
    cfg.readConf("flood.conf")
    cfg.configArgParser(parser)

    out = parser.parse_args(sys.argv[1:])

    self.config_logging(cfg.logFileName)

    self.check_port_conflicts(cfg.torrentPort)
    self.session = libtorrent.session(flags = libtorrent.session_flags_t.add_default_plugins)
    self.depot = Depot(cfg.depotPath, cfg.depotExpiration, self.session)

    # Disable use of libtorrent ssl ports
    settings = self.session.get_settings()
    settings['ssl_listen'] = 0
    self.session.set_settings(settings)

    mask = (1<<32)-1
    mask &= ~libtorrent.alert.category_t.progress_notification
    mask &= ~libtorrent.alert.category_t.port_mapping_notification
    mask &= ~libtorrent.alert.category_t.performance_warning
    self.session.set_alert_mask(mask)

    self.session.listen_on(cfg.torrentPort, cfg.torrentPort) # Use only the specified port.
    self.session.start_dht()
    self.session.stop_lsd()
    log.info("Listen port: %s", self.session.listen_port())
    torrents = self.depot.torrents.values()

    last = None
    threads = []
    for fn in [ self.handleAlerts,
                lambda: self.runRPCServer(cfg.rpcBindAddr, cfg.rpcPort)]:
      if last is None:
        last = fn
        continue
      thr = threading.Thread(target=last)
      thr.setDaemon(True)
      thr.start()
      threads.append(thr)
      last = fn

    for t in torrents:
      t.activate(self.session)
      log.info("Activated: torrent %s", t.info.name())

    last()


  @rpc("GetTorrentList")
  def getTorrentList(self):
    return self.depot.torrents.keys()

  @rpc("AddDHTNode")
  def addDHTNode(self, host, port):
    '''
    Add a node that flood should communicate with.
    This node will be treated as a potential peer for all
    current and future torrents.
    :param host:
    :param port:
    :return:
    '''

    host = str(host)
    if not (host, port) in self.peerNodes:
      self.peerNodes.add((host, port))
      try:
        self.session.add_dht_node((host, port))
      except Exception, e:
        log.exception("Failed add_dht_node: %s", e)
        self.session.add_dht_router((host, port))
      for t in self.depot.torrents.values():
        t.handle.connect_peer((host, port), 4)

  @rpc("AddTorrent")
  def addTorrent(self, torrentUrl):

    log.info("Adding torrent: %s", torrentUrl)
    torrent = self.depot.importTorrent(torrentUrl)
    torrent.activate(self.session, peerNodes=self.peerNodes)
    return torrent.info.name()

  @rpc("RemoveTorrent")
  def removeTorrent(self, torrentName):
    try:
      torrent = self.depot.torrents[torrentName]
    except KeyError:
      raise Error("No such torrent: %s", torrentName)
    torrent.remove(self.session)

  @rpc("AddTorrentSeed")
  def addTorrentSeed(self, torrentName, seedUrl):
    try:
      torrent = self.depot.torrents[torrentName]
    except KeyError:
      raise Error("No such torrent: %s", torrentName)
    seedUrl = str(seedUrl)
    log.info("Adding torrent seed: %s %d '%s'", type(seedUrl), len(seedUrl), seedUrl)
    torrent.addSeed(seedUrl)

  @rpc("GetTorrentStatus")
  def getTorrentStatus(self, torrentName):
    try:
      torrent = self.depot.torrents[torrentName]
    except KeyError:
      raise Error("No such torrent: %s", torrentName)
    return torrent.getStatus()


def main():
  return GlobalCommand().run()


if __name__ == "__main__":
  main()
