# Copyright (c) 2015 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import, print_function

import datetime
import hashlib
import logging

try:
  import libtorrent
except ImportError:
  libtorrent = None

import time
import os
from flood.util.lazy import lazy

from flood.util.meta import GenericObjectGenerator
try:
  import simplejson as json
except ImportError:
  import json

log = logging.getLogger("torrent")

class TorrentItem(object):
  TORRENT_SUFFIX = ".torrent"
  HASH_MAP = {}

  @classmethod
  def Generate(cls, path, url_seed=None, nodes=None):
    dPath, name = os.path.split(path)
    st = os.stat(path)
    fs = libtorrent.file_storage()

    fs.add_file(name, st.st_size)

    creator = libtorrent.create_torrent(fs)
    hashes = fileSHA1(path, creator.piece_length())
    for (idx, h) in enumerate(hashes):
      creator.set_hash(idx, h)

    if url_seed:
      creator.add_url_seed(url_seed)

    if nodes:
      for node in nodes:
        creator.add_node(*node)

    desc = creator.generate()

    tFile = os.path.join(path + cls.TORRENT_SUFFIX)
    with open(tFile, "wb") as f:
      f.write(libtorrent.bencode(desc))


  MAX_ALERTS = 16
  DIAG_FILE = "diag.json"
  def __init__(self, torrentFile, depot):
    self.tFile = torrentFile
    self.depot = depot
    with open(torrentFile, "rb") as f:
      data = libtorrent.bdecode(f.read())
    self.info = libtorrent.torrent_info(data)
    self.handle = None
    self.alerts = []
    self.location = os.path.join(self.depot.path, self.info.name())
    self.diag_file = os.path.join(self.location, self.DIAG_FILE)

  @lazy
  def name(self):
    return self.info.name()

  def addAlert(self, alert):
    self.alerts.append((datetime.datetime.now(), alert))
    if len(self.alerts) > self.MAX_ALERTS:
      self.alerts.pop(0)

    if self.diag_file:
      with open(self.diag_file, 'w') as f:
        f.write(self.getStatus()['jsonDoc'])

    if isinstance(alert, libtorrent.torrent_deleted_alert):
      self._removed()


  def is_expired(self, expiration_sec):
    """
    Return true is no file associated with the torrent has been modified
    in the past "expiration_sec"
    :param expiration_sec: Period of inactivity to identify expired torrents.
    :return:
    """
    begin = time.time() - expiration_sec
    if os.stat(self.tFile).st_mtime > begin:
      return False

    tor_dir = os.path.splitext(self.tFile)[0]
    for f in os.listdir(tor_dir):
      if self.DIAG_FILE == f:
        continue
      if os.stat(os.path.join(tor_dir, f)).st_mtime > begin:
        return False
    return True


  def getStatus(self):

    out = {'torrentName': self.info.name(),
           'location' : self.location,
           'urlSeed': self.handle.url_seeds(),
           'alertLog': ["%s %s: %s" % (x[0], x[1].what(), x[1].message()) for x in self.alerts],
           'stats': GenericObjectGenerator(self.handle.status())}

    return {'jsonDoc': json.dumps(out, indent=2, encoding='latin-1')}


  def activate(self, tSession, peerNodes=set()):
    if self.handle:
      return self.handle

    params = {'save_path': self.location + '/',
              'storage_mode': libtorrent.storage_mode_t.storage_mode_sparse}

    nodes = self.info.nodes()
    dhtNodes = params.setdefault('dht_nodes_name', [])
    if nodes and len(nodes):
      dhtNodes.extend(self.info.nodes())
    dhtNodes.extend(list(peerNodes))
    dhtNodes.append(('127.0.0.1', tSession.listen_port()))
    params['dht_nodes_name'] = dhtNodes
    params['ti'] = self.info
    params['max_uploads'] = 1
    log.info("Save torrent: %s", params)
    self.handle = tSession.add_torrent(params)


    h = str(self.handle.info_hash())
    self.hashVal = h
    log.info("[%s] activating: %s", self.info.name(), h)
    self.HASH_MAP[h] = self
    for peer in peerNodes:
      self.handle.connect_peer(peer, 4)
    return self.handle

  def remove(self, session):
    session.remove_torrent(self.handle, 1) # delete files

  def _removed(self):
    '''
    Called upon notification that a torrent has been unregistered from libtorrent.
    :return:
    '''
    try:
      os.unlink(self.tFile)
    except OSError, e:
      log.info("Failed to remove torrent: %s %s", self.tFile, e)
    try:
      os.unlink(self.diag_file)
    except OSError, e:
      log.info("Failed to remove torrent diagnostics file: %s %s", self.diag_file, e)

    try:
      os.rmdir(os.path.splitext(self.tFile)[0])
    except OSError, e:
      log.info("Failed to remove torrent dir: %s %s",
               os.path.splitext(self.tFile)[0], e)
    log.info("[%s] Removed torrent", self.name)
    del self.HASH_MAP[self.hashVal]
    del self.depot.torrents[self.name]


  def addSeed(self, url):
    log.info("Adding seed url: %s", url)
    self.handle.add_url_seed(url)


def fileSHA1(path, pieceSize):
  pieceHashes = []
  with open(path) as f:
    while True:
      s = hashlib.sha1()
      d = f.read(pieceSize)
      if (len(d) == 0):
        break
      s.update(d)
      pieceHashes.append(s.digest())
  return pieceHashes