# Copyright (c) 2015 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import, print_function

import inspect
import logging

from flood.util.error import Error
from flood.util.lazy import lazy

from flood.util.config import Param, VarArg
from flood.util.meta import MethodTracker

log = logging.getLogger("cmd")

class OutputSink(object):
  '''
  Sink for output to be printed that maintains indentation
  context.   Use "with" to increase indentation.

  All output directed to the "output" method will be printed
  to stdout using the current level of tab indentation.

  '''
  def __init__(self, depth=0):
    self.depth = depth

  def __enter__(self):
    return self.__class__(self.depth+1)

  def __exit__(self, exc_type, exc_val, exc_tb):
    pass

  def output(self, fmt, *args):
    print((self.depth * "\t") + (fmt % args))

class Command(object):
  '''
  An instantiated command object, wrapping a function to be bound
  to a method of a class.
  This object type is created via an @command annotation of a
  function in a class declaration.
  '''
  def __init__(self, desc, fn, kwargs, chained=False):
    '''

    :param desc: A description of the wrapped function.
    :param fn:   The wrapped function.
    :param kwargs: A dict-map of ConfigValue object describing the
                   arguments to "fn"
    :return:
    '''
    self.desc = desc
    self.fn = fn
    self.chained = chained
    try:
      self.name = kwargs['name']
    except KeyError:
      try:
        self.name = fn.__name__
      except AttributeError:
        self.name = None
    self.kwargs = kwargs
    self.binding = None

  @lazy
  def argSpec(self):
    '''
    "argspec" of function to execute the command, as per inspect.getargspec.
    :return:
    '''
    return inspect.getargspec(self.fn)

  @lazy
  def vararg(self):
    '''
    ConfigValue defining an "vararg"/"*pargs" argument to the command.
    :return:
    '''
    argSpec = self.argSpec
    if argSpec.varargs:
      try:
        return self.kwargs[argSpec.varargs](name=argSpec.varargs)
      except KeyError:
        return Param("Trailing parameters %s" % argSpec.varargs,
                     VarArg)(name=argSpec.varargs)
    return None

  @lazy
  def argMap(self):
    '''
    List of arguments to the command, where each argument is
    represented by a ConfigValue object.

    :return: A list of argument descriptors of this command
             (using ConfigValue objects).
    '''
    argMap = []
    argSpec = self.argSpec
    for idx, arg in enumerate(argSpec.args):
      if arg == 'self':
        continue
      if arg in self.kwargs:
        argMap.append(self.kwargs[arg](name=arg))
      else:
        argMap.append(Param("Parameter %s" % arg)(name=arg))


    return argMap

  def recordName(self, name):
    self.name = name
    return name

  def help(self, sink):
    '''
    Generate help output for the command to the specified output sink.
    :param sink:
    :return:
    '''
    args = []
    for x in self.argMap:
      args.append("<%s>" % x.name)
    sink.output("%s %s", self.name, " ".join(args))
    sink.output(" - " + self.desc)
    with sink as inner:
      for x in self.argMap:
        inner.output("%s - %s" % (x.name, x.description))
    sink.output("")

  def __call__(self, *args, **kwargs):
    return self.fn.__get__(self.binding)(*args, **kwargs)

  def __get__(self, instance, owner=None):
    '''
    Return a "BoundObj" that emulates and instancemethod type object
    but still delegates all attribute access back to this object.
    This allows the method to be used via a direct call:  a.b(....)
    or access to "Command" attributes:  a.b.consume(....)

    This has to be done with an intermediate object so as to be
    thread-safe - the binding of this object to "instance" must be
    represented only on the stack.

    Class-level access is not supported.

    :param instance:
    :param owner:
    :return:
    '''
    class BoundObj(object):
      def __init__(self, base, binding):
        self.base = base
        self.binding = binding

      def __getattr__(self, item):
        return getattr(self.base, item)

      def __call__(self, *args, **kwargs):
        return self.fn.__get__(self.binding)(*args, **kwargs)

    return BoundObj(self, instance)

  def consume(self, prefixObj, args):
    '''
    Called in the process of command word consumption.

    :param prefixObj: The list of previously resolved objects
                      (i.e. to the left)
    :param args: All unresolved command arguments (i.e. to the right).
    :return:
    '''
    values = []
    varArg = self.vararg
    for arg in self.argMap:
      if len(args) == 0:
        raise Error("No value for argument '%s'", arg.name)
      values.append(arg.assign(args[0]))
      args.pop(0)
    if len(args) > 0 and self.chained == False:
      if varArg is None:
        raise Error("Unrecognized arguments: %s", " ".join(args))
      for a in args:
        varArg.assign(a)
      values.extend(varArg.value)

    retval = self.fn.__get__(prefixObj[-1])(*values)
    # A "chained" command will return a CommandContext (or more generically
    # something conformant to the "consume" interface and will be able to
    # consume more arguments.
    try:
      consumer = retval.consume
    except AttributeError:
      return retval
    try:
      return consumer(prefixObj + [self], args)
    except Exception, e:
      raise


class command(object):
  '''
  Decorator class for methods of "CommandContext" types that are to be
  bound to CLI sub-commands.

  The presence of the decorator implies that the method name is
  a valid sub-command.

  A description of the sub-command can be specified as the first
  positional argument to the decorator.
  Keyword arguments with ConfigValue values, can be specified to define
  the types of argument values allowed/expected.  The keyword names
  should match the names of the parameters of the method (with the
  omission of "self").  The ConfigValue object may perform some
  type-specific transformation of the argument value between the
  command line tet and the actual value passed to the method
  invocation.
  '''

  chained = False
  def __init__(self, desc="Description", **kwargs):
    self.fn = None
    self.desc = desc
    self.kwargs = kwargs

  def __call__(self, fn):
    self.fn = fn
    return Command(self.desc, self.fn, self.kwargs, self.chained)


class chained_command(command):
  chained = True


class delegate_command(Command):
  '''
  A delegate_command specifies a type of CommandContext to which
  command processing will be delegated to:

  class Foo(CommandContext):
      subcmd = delegate_command("Bar Commands", Bar)

  When processing a "Foo" command and the next argument is "subcmd",
  an object of type "Bar" will be instantiated and command processing
  will be delegated to it.

  This is primarily usefull to create sub-command contexts for the
  sake of splitting groups of sub-commands based on their area of functionality.
  '''
  def __init__(self, __description=None, delegate_type=None, **kwargs):
    self.delegate_type = delegate_type
    Command.__init__(self, __description, None, kwargs)

  @lazy
  def argMap(self):
    '''
    Doesn't take any arguments....
    :return:
    '''
    return []

  def consume(self, prefixObj, args):
    try:
      parent = prefixObj[-2]
    except IndexError:
      parent = None

    try:
      name = args[0]
    except IndexError:
      name = None
    resolver = self.delegate_type(prefixObj[-1], name, parent)
    return resolver.consume(prefixObj, args)

def cli_bind(target_fn, description, presenter = None):
  cmd = command(description)
  def __wrapper(self):
    out = getattr(self.target, target_fn)()
    if presenter:
      presenter(out)
    else:
      return out
  return cmd(__wrapper)


class CommandConfigurator(MethodTracker):
  decorator = Command
  trackerAttr = '_commands'

  @classmethod
  def newClass(mcs, kls):
    x = getattr(kls, "wrapper_type", None)
    if x:
      kls.ProxyMap[x] = kls


class CommandContext(object):
  '''
  Base class for an object that has "@command" methods.
  '''
  __metaclass__ = CommandConfigurator
  ProxyMap = {}

  def __init__(self, name=None, parent=None):
    if name is None:
      name = self.__class__.__name__
    self.name = name
    self.inventory = {}
    self.parent = parent

  def nameChain(self):
    if self.parent:
      return self.parent.nameChain() + "." + self.name
    return self.name

  def help(self, objPrefix, outSink=None):
    '''
    Generate help output for this command.

    :param objPrefix: Chain of prefix objects to this command context.
    :param outSink:   Object that takes the output (performs correct indentation).
    :return:
    '''
    if outSink is None:
      outSink = OutputSink()
    nameChain = []
    for x in objPrefix:
      nameChain.append(x.name)
    nameChain.append(self.name)
    outSink.output("Command context: %s" % ".".join(nameChain))

    outSink.output("Sub-commands: %s" %
                   " ".join([x.name for x in self._commands.values()]))
    outSink.output("")
    for x in self._commands.values():
      x.help(outSink)
    for x in self.lookupAll():
      outSink.output("\t%s" % x)

  def lookup(self, name):
    inv = self.inventory
    try:
      return inv[name]
    except KeyError:
      match = None
      for x in inv:
        if x.lower().startswith(name.lower()):
          if match:
            raise Error("Ambiguous prefix: %s (%s, %s)", name, match, x)
          match = x
  def lookupAll(self):
    return self.inventory


  def consume(self, objPrefix, args):
    '''
    Process this object in the course of parsing the command line,
    left to right.

    :param objPrefix:
    :param args:
    :return:
    '''
    try:
      if len(args) == 0 or "--help" in args:
        self.help(objPrefix)
        return
      log.debug("Consuming: %s", args[0])

      name = args[0]
      if name in self._commands:
        meth = getattr(self, name)
        return meth.consume(objPrefix + [self], args[1:])

      if name.startswith(":"):
        name = name[1:]

      try:
        obj = self.lookup(name)
      except KeyError:
        raise Error("No command: %s:%s", self.nameChain(),name)
      try:
        proxyType = self.ProxyMap[type(obj)]
        obj = proxyType(obj, name, self)
      except KeyError:
        pass

      return obj.consume(objPrefix + [self], args[1:])
    except Exception, e:
      log.error("Failed: %s %s: %s", "\\".join([x.name for x in objPrefix]), args, e, exc_info=True)
      raise


  def execute(self, args):
    try:
      self.consume([], args)
      return 0
    except Exception, e:
      log.error("Failed command: %s", e)
      return -1


def main():
  class TestContext2(CommandContext):
    '''
    Test Context 2
    This is TestContext2
    '''
    @command("test method2")
    def testMethod2(self, arg2):
      print("Got arg 2: %s" % arg2)
      return "foo"

  class TestContext1(CommandContext):
    '''
    Test Context 1
    This is Testcontext 1
    '''
    def __init__(self, *pargs, **kwargs):
      CommandContext.__init__(self, *pargs, **kwargs)
      self.inventory['foo'] = TestContext2("foo2", self)

    @command("test method1")
    def testMethod1(self, arg1):
      print("Got arg: %s" % arg1)
      return TestContext2(None, self)

    @command("test method 2",
             arg1 = Param("Argument # 1"))
    def testMethod2(self, arg1):
      return arg1 + "foo"

  logging.basicConfig()
  logging.getLogger().setLevel(logging.DEBUG)


  tc1 = TestContext1()
  tc1.help([])
  tc1.consume([], ["foo", "--help"])
  #tc1.consume([], ["testMethod1", "foo", "testMethod2", "bar"])
  tc1.consume([], ["foo", "testMethod2", "bar"])
  tc1.consume([], ["foo", "testMethod2", "--help"])


if __name__ == "__main__":
  main()