# Copyright (c) 2015 Cloudera, Inc. All rights reserved.

from __future__ import absolute_import, print_function

import argparse
import logging

LOG = logging.getLogger("config")
from flood.util.meta import MetaSubclassDetector

class MetaConfigValue(MetaSubclassDetector):
  '''
  Metaclass for detecting and registering sub-types
  of ConfigValue.
  All not "unregistered" sub-types will be tracked
  in the "ConfigMap" attribute.
  '''
  @classmethod
  def newClass(mcs, kls):
    try:
      if not getattr(kls, "unregistered", False):
        kls.ConfigMap[kls.name] = kls
    except AttributeError:
      pass


class ConfigValue(object):
  '''
  Base class for objects representing a config item, which
  is a setting that the user may specify, potentially via
  a config file or command line.

  The ConfigValue object provides for an API to describe
  configuration items and for performing validation/conversion
  of values being assigned ot them.
  '''
  __metaclass__ = MetaConfigValue
  __subclasses__ = []

  ConfigMap = {}
  default = None
  positional = False
  description = "Undescribed"

  def __init__(self, positional=None, name=None, required=False):
    self.value = None
    if positional is not None:
      self.positional = positional
    if name is not None:
      self.name = name
    self.required = False

  def describe(self):
    return "%s (default:%s)" % (self.name, self.default)

  def assign(self, value):
    self.value = value
    return value

  @classmethod
  def configParser(cls, cfg, parser):
    if cls.positional:
      parser.add_argument(cls.name,
                          metavar=cls.name,
                          type=lambda val: cfg.assign(cls.name, val),
                          help = cls.description,
                          nargs = 1)
    else:
      parser.add_argument("--" + cls.name,
                          metavar=cls.name,
                          type=lambda val: cfg.assign(cls.name, val),
                          help = cls.description,
                          dest = cls.name,
                          nargs = 1)

class ArrayValue(ConfigValue):
  '''
  A ConfigValue that is an array/list of values.
  '''
  default = None
  def assign(self, value):
    if self.value is None:
      self.value = []
    if value is not None:
      self.value.append(value)


class VarArg(ArrayValue):
  '''
  A special ArrayValue for handling trailing positional parameters to
  functions/commands
  '''
  def configParser(self, parser):
    parser.add_argument(self.name, type=self.assign,
                        nargs="...")

def Param(desc, configClass=ConfigValue, defVal=None):
  '''
  Factory to create ConfigValue objects in case a command decorator
  does not include a description of a parameter.
  :param desc:
  :param configClass:
  :param defVal:
  :return:
  '''
  class ConfigItem(configClass):
    unregistered = True
    description = desc
    default = defVal
  return ConfigItem

class IntegerConfigValue(ConfigValue):
  '''
  A ConfigValue that should be an integer.
  '''
  def assign(self, value):
    try:
      return super(IntegerConfigValue, self).assign(int(value))
    except ValueError:
      raise argparse.ArgumentTypeError("Not an integer value for: %s" % self.name)

class Config(object):
  '''
  A Configuration object that can take values from a config file
  or command line arguments.

  A user defines the types of config options that are available
  by defining sub-classes of ConfigValue.  This class will map
  ConfigValue types to argparse command line arguments, or
  permit equivalent configuration to be read from a config file.
  (The latter is not yet implemented.)

  '''

  def __init__(self, valueMap={}, base=None):
    self.base = base
    self.values = {}
    for (k, v) in valueMap.iteritems():
      if k in ConfigValue.ConfigMap:
        self.assign(k, v)
        cfgVal = ConfigValue.ConfigMap[k]()
      else:
        self.values[k] = v

  def has(self, name):
    try:
      self.lookup(name)
      return True
    except KeyError:
      return False

  def lookup(self, name):
    try:
      return self.values[name]
    except KeyError:
      pass
    try:
      if self.base:
        return self.base.lookup(name)
    except KeyError:
      pass

    if name in ConfigValue.ConfigMap:
      return ConfigValue.ConfigMap[name].default
    raise

  def __getattr__(self, item):
    try:
      return self.lookup(item)
    except KeyError:
      raise AttributeError("No config value for: %s", item)

  def readConf(self, configFileName):
    try:
      with open(configFileName, "r") as f:
        for line in f:
          words = [x.strip() for x in line.split("=", 1)]
          if len(words) == 2:
            self.assign(words[0], words[1])
    except Exception, e:
      LOG.error("Failed to read config file %s: %s", configFileName, e)

  def assign(self, name, value):
    cfgType = ConfigValue.ConfigMap[name]
    self.values[name] = cfgType(name=name).assign(value)

  def configArgParser(self, parser):
    for item in ConfigValue.ConfigMap.values():
      item.configParser(self, parser)
