# Copyright (c) 2015 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import, print_function

import datetime
import logging

try:
  import libtorrent
except ImportError:
  libtorrent = None


log = logging.getLogger("meta")
class MetaSubclassDetector(type):
  '''
  Base class for meta-classes to detect and track sub-classes.
  '''
  def __new__(mcs, name, bases, attrs):
    mcs.preNewClass(name, bases, attrs)
    c = super(MetaSubclassDetector, mcs).__new__(mcs, name, bases, attrs)
    mcs.newClass(c)
    return c

  @classmethod
  def newClass(mcs, kls):
    pass

  @classmethod
  def preNewClass(mcs, name, bases, attrs):
    pass

class MethodTracker(MetaSubclassDetector):
  '''
  Metaclass for detecting @command decorated methods.
  All such methods in a class will be registered in
  a map by name (in all lower-case).

  The type of the decorator to be detected is specified
  by the "decorator" attribute of the meta-class.
  The "trackerAttr" attribute defined the name of the
  dictionary of the tracked class members.
  '''
  @classmethod
  def preNewClass(mcs, name, bases, attr):
    if attr.get("__metaclass__", None) == mcs.__class__:
      return

    methods = {}

    for x in bases:
      methods.update(getattr(x, mcs.trackerAttr, {}))

    for (k, v) in attr.items():
      if isinstance(v, mcs.decorator):
        origName = k
        k = v.recordName(k) # record original name
        methods[k] = v
        attr[origName] = v

    attr[mcs.trackerAttr] = methods

def GenericPrinter(obj, prefix=""):
  if isinstance(obj, dict) and len(obj) == 1 and 'p' in obj:
    obj = obj['p']

  if isinstance(obj, dict):
    if len(prefix):
      prefix += "."
    for x in obj:
      GenericPrinter(obj[x], prefix + x)
    return

  if isinstance(obj, (list, tuple)):
    for (i, x) in enumerate(obj):
      GenericPrinter(obj[i], "%s[%s]" % (prefix, i))
    return

  print(prefix + " = " + str(obj))

class BadObj(Exception):
  pass




def GenericObjectGenerator(obj):
  '''
  Take an object and convert it into a dict-tree suitable
  for conversion to json.
  :param obj:
  :param schema:
  :return:
  '''

  if isinstance(obj, (int, long)):
    if hasattr(type(obj), "values"):
      return str(obj)
    return obj
  if isinstance(obj, float):
    return obj
  if isinstance(obj, (str, unicode)):
    return str(obj)
  if isinstance(obj, (list, tuple)):
    return [GenericObjectGenerator(x) for x in obj]
  if isinstance(obj, datetime.timedelta):
    return str(obj)
  if isinstance(obj, libtorrent.big_number):
    return str(obj)
  if isinstance(obj, type):
    raise BadObj()
  if callable(obj):
    log.info("Bad object: %s", obj)
    raise BadObj()

  try:
    out = {}
    for k in dir(obj):
      if k.startswith("__"):
        continue
      try:
        try:
          val = getattr(obj, k)
        except TypeError:
          continue
        out[k] = GenericObjectGenerator(getattr(obj, k))

      except BadObj:
        continue
    return out
  except RuntimeError:
    import pdb
    pdb.set_trace()


