# Copyright (c) 2015 Cloudera, Inc. All rights reserved.
from __future__ import absolute_import, print_function


import collections
import inspect
import socket
from BaseHTTPServer import HTTPServer, BaseHTTPRequestHandler

import avro.ipc
import logging

from flood.util.cmd import Command, CommandContext, command
from flood.util.lazy import lazy
from flood.util.meta import MethodTracker

log = logging.getLogger("rpc")

class HostPort(collections.namedtuple("HostPort",("host", "port"))):
  def __str__(self):
    return "%s:%s" % (self.host, self.port)


def resolveHostPort(spec, defaultPort=None):
  w = spec.split(":", 1)
  if len(w) == 1:
    port = defaultPort
  else:
    try:
      port = int(w[1])
    except ValueError:
      log.error("Invalid port: %s" % w[1])
  return HostPort(socket.gethostbyname(w[0]), port)


class RPCMethod(object):
  '''
  Object that presents itself as a method attached to an RPCClient
  type.  The method is synthesized from the AVRO schema of an
  RPC function.
  '''

  class RPCCommand(Command):
    def __init__(self, rpc, *pargs, **kwargs):
      self._rpc = rpc
      Command.__init__(self, "RPC Method %s" % rpc.schema.name,
                       rpc.getCallable(), {})

    @lazy
    def argSpec(self):
      return self._rpc.argSpec

  def __init__(self, methodSchema):
    self.schema = methodSchema
    self.arguments = []
    for n in self.schema.request.props['fields']:
      self.arguments.append(n.name)

  def getCommand(self):
    return self.RPCCommand(self)

  @lazy
  def argSpec(self):
    return inspect.ArgSpec([ x.name for x in self.schema.request.props['fields']],
                           None, None, None)

  def getCallable(self):
    return lambda *pargs, **kwargs: self._invoke(*pargs, **kwargs)

  def _invoke(self, rpcClient, *pargs, **kwargs):
    msg = {}
    x = 0
    for (x, n) in enumerate(self.schema.request.props['fields']):
      msg[n.name] = pargs[x]
      if n.type.fullname == "int":
          msg[n.name] = int(msg[n.name])
    return rpcClient.requestor.request(self.schema.name, msg)


class MetaSchemaInterpreter(CommandContext.__metaclass__):
  '''
  Metaclass for synthesizing RPCMethod objects based on
  an AVRO schema associated with an RPCClient sub-class.
  '''

  @classmethod
  def preNewClass(mcs, name, bases, attrs):
    try:
      assert('SCHEMA' in attrs)
      if attrs['SCHEMA'] is None:
        return
      METHODS = {}
      for (k, v) in attrs['SCHEMA'].messages.iteritems():
        meth = RPCMethod(v)
        attrs[k] = meth.getCommand()
        METHODS[k] = meth
      attrs['METHODS'] = METHODS
    finally:
      super(MetaSchemaInterpreter, mcs).preNewClass(name, bases, attrs)


class RPCClient(CommandContext):
  '''
  Base class for RPC client objects.
  A sub-class of this type is expected to set a class attribute
  "SCHEMA" to identify the AVRO schema it is using.  All functions
  in the schema will have methods corresponding to them generated in
  the sub-class (using RPCMethod)
  '''

  SCHEMA = None
  METHODS = {}
  DEFAULT_RPC_PORT = 1234
  __metaclass__ = MetaSchemaInterpreter

  def __init__(self, server, idx=None):
    if not isinstance(server, HostPort):
      server = resolveHostPort(server, self.DEFAULT_RPC_PORT)
    self.server = server
    self.idx = idx
    super(RPCClient, self).__init__(str(server), None)

  @property
  def transceiver(self):
    return avro.ipc.HTTPTransceiver(self.server.host, self.server.port)

  @property
  def requestor(self):
    return avro.ipc.Requestor(self.SCHEMA, self.transceiver)


class rpc(command):
  def __init__(self, rpcName=None):
    self.fn = None
    self.name = None
    self._rpcName = rpcName

  def __get__(self, instance, owner):
    return self.fn.__get__(instance, owner)

  def __call__(self, fn):
    self.fn = fn
    return self

  def recordName(self, name):
    self.name = name
    if self._rpcName:
      return self._rpcName
    return self.name


class AvroMapper(MethodTracker):
  trackerAttr = "rpcSet"
  decorator = rpc


class RPCServer(object):
  __metaclass__ = AvroMapper
  AVRO_SCHEMA = None
  def executeRPC(self, msg, req):
    rpcCall = msg.name

    try:
      methodDecorator = self.rpcSet[rpcCall]
    except KeyError:
      raise avro.schema.AvroException("Unexpected call: %s" % rpcCall)

    fn = getattr(self, methodDecorator.name)

    argSpec = inspect.getargspec(fn)
    args = []
    for x in argSpec.args[1:]:
      args.append(req.get(x))

    try:
      return fn(*args)
    except Exception, exc:
      log.error("Failed handling RPC: %s %s", rpcCall, exc, exc_info=True)
      raise


  def runRPCServer(self, bindAddr, rpcPort):
    srv = HTTPServer((bindAddr, rpcPort), RPCHTTPHandlerFactory(self.executeRPC, self.AVRO_SCHEMA))
    srv.allow_reuse_address = True
    srv.serve_forever()


class RPCResponder(avro.ipc.Responder):
  def __init__(self, executor, schema):
    avro.ipc.Responder.__init__(self, schema)
    self.executor = executor

  def invoke(self, msg, req):
    return self.executor(msg, req)


def RPCHTTPHandlerFactory(commandContext, schema):
  class RPCHTTPHandler(BaseHTTPRequestHandler):
    def do_POST(self):
      responder = RPCResponder(commandContext, schema)
      reader = avro.ipc.FramedReader(self.rfile)
      req = reader.read_framed_message()

      resp = responder.respond(req)
      self.send_response(200)
      self.send_header('Content-Type', 'avro/binary')
      self.end_headers()
      writer = avro.ipc.FramedWriter(self.wfile)
      writer.write_framed_message(resp)

  return RPCHTTPHandler