#! /usr/bin/env bash
# Copyright (c) 2018 Cloudera, Inc. All rights reserved.
#
# The httpd load balancer startup script. We have this one common
# script for the httpd for different services, and it sources an
# appropriate service-specific script for service-specific
# configuration.
#

# Each platform uses a different name for HTTPD
HTTPD_CANDIDATES=(
    # CentOS path
    "/usr/sbin/httpd"

    # Debian path
    "/usr/sbin/apache2"

    # SLES path
    "/usr/sbin/httpd2"
)

# Each platform stores its HTTPD modules in a separate location.
HTTPD_MODULE_DIR_CANDIDATES=(
    # CentOS paths
    "/usr/lib64/httpd/modules"
    "/usr/lib/httpd/modules"

    # Debian Paths
    "/usr/lib64/apache2/modules"
    "/usr/lib/apache2/modules"

    # SLES Paths - Following SLES's pattern of using the preforked modules
    "/usr/lib64/apache2-prefork"
    "/usr/lib64/apache2"
)

find_path() {
    for candidate in $@; do
        if [ -e $candidate ]; then
            echo $candidate
            return 0
        fi
    done

    return 1
}

# Search-replace {{$1}} with ${$1} in conf files. Ex: {{FOO}} with ${FOO}
replace_httpd_conf() {
    VAR=$1
    perl -pi -e "s#\{\{${VAR}}}#${!VAR}#" ${CLOUDERA_HTTPD_CONF_DIR}/*.conf
}

# Set CLOUDERA_HTTPD_EXECUTABLE_PATH
set_httpd_exec_path () {
    if [ -z "${CLOUDERA_HTTPD_EXECUTABLE_PATH}" ]; then
        CLOUDERA_HTTPD_EXECUTABLE_PATH=`find_path ${HTTPD_CANDIDATES[@]}`
        if [ "$?" != 0 ]; then
            cat 1>&2 <<EOF
Failed to find the Apache HTTPD executable.
Tried ${HTTPD_CANDIDATES[@]}

Please confirm that Apache HTTPD is properly installed on the Load
Balancer host.  If Apache HTTPD is installed in a non-standard
location, set the 'CLOUDERA_HTTPD_EXECUTABLE_PATH' environment
variable in the Load Balancer Environment Advanced Configuration
Snippet (Safety Valve)
EOF
            exit 1
        fi
    fi

    if [ ! -x "${CLOUDERA_HTTPD_EXECUTABLE_PATH}" ]; then
        cat 1>&2 <<EOF
${CLOUDERA_HTTPD_EXECUTABLE_PATH} executable does not exist or is not
executable.  Check the 'Apache HTTPD Executable Path' configuration
property.
EOF
        exit 1
    fi
}

set_httpd_module_dir() {
    if [ -z "${CLOUDERA_HTTPD_MODULE_DIR}" ]; then
        CLOUDERA_HTTPD_MODULE_DIR=`find_path ${HTTPD_MODULE_DIR_CANDIDATES[@]}`
        if [ "$?" != 0 ]; then
            cat 1>&2 <<EOF
Failed to find the Apache HTTPD module paths.
Tried ${HTTPD_MODULE_DIR_CANDIDATES[@]}

Please confirm that Apache HTTPD is properly installed on the Load
Balancer host.  If Apache HTTPD is installed in a non-standard
location, set the 'CLOUDERA_HTTPD_MODULE_DIR' environment variable in
the Load Balancer Environment Advanced Configuration Snippet (Safety
Valve).
EOF
            exit 1
        fi
    fi

    if [ ! -d "${CLOUDERA_HTTPD_MODULE_DIR}" ]; then
        cat 1>&2 <<EOF
${CLOUDERA_HTTPD_MODULE_DIR} does not exist, is not a directory, or is
not readable.  Check the 'Apache HTTPD Module Path' configuration
property.
EOF
        exit 1
    fi
}

check_httpd_log_dir() {
    if [ -z "${CLOUDERA_HTTPD_LOG_DIR}" ]; then
        cat 1>&2 <<EOF
CLOUDERA_HTTPD_LOG_DIR environment variable is not defined.  please
check the 'Load Balancer Log Directory' configuration variable.
EOF
        exit 1
    fi
}

# Set CONF_DIR as the result of a refresh command
# $1 is the pid of the currently running httpd.
set_httpd_refresh_conf_dir() {
    pid="$1"

    if [ "${pid}" = "" ]; then
        die "httpd process pid was not specified."
    fi

    CONF_DIR=$(xargs -n 1 -0 < /proc/$pid/environ | sed -n 's/^CONF_DIR=\(.*\)/\1/p')
    if [ -n "${CONF_DIR}" ]; then
        if [ ! -d "${CONF_DIR}" ]; then
            die "httpd process dir ${CONF_DIR} does not exist."
        fi
    else
        die "httpd process pid mismatch, ${pid} does not contain CONF_DIR variable in environment."
    fi
}

make_run_dir() {
    if [[ ! -d "${CLOUDERA_HTTPD_CONF_DIR}/run" ]]; then
        mkdir -p "${CLOUDERA_HTTPD_CONF_DIR}/run"
    fi
}

# Calls above functions to set various paths and directories common to
# httpd use.
set_httpd_dirs() {
    export CLOUDERA_HTTPD_CONF_DIR="${CONF_DIR}"
    export CLOUDERA_HTTPD_CONF="${CLOUDERA_HTTPD_CONF_DIR}/httpd.conf"

    set_httpd_exec_path
    set_httpd_module_dir
    check_httpd_log_dir
    make_run_dir

    echo "CLOUDERA_HTTPD_EXECUTABLE_PATH=${CLOUDERA_HTTPD_EXECUTABLE_PATH}"
    echo "CLOUDERA_HTTPD_MODULE_DIR=${CLOUDERA_HTTPD_MODULE_DIR}"
    echo "CLOUDERA_HTTPD_CONF_DIR=${CLOUDERA_HTTPD_CONF_DIR}"
    echo "CLOUDERA_HTTPD_LOG_DIR=${CLOUDERA_HTTPD_LOG_DIR}"
    echo "CLOUDERA_HTTPD_USE_SSL=${CLOUDERA_HTTPD_USE_SSL}"
}

# Search-replace httpd directories in config files
config_httpd_dirs() {
    replace_httpd_conf "CLOUDERA_HTTPD_CONF_DIR"
    replace_httpd_conf "CLOUDERA_HTTPD_LOG_DIR"
    replace_httpd_conf "CLOUDERA_HTTPD_MODULE_DIR"
    replace_conf_dir
}

check_httpd_conf() {
    "${CLOUDERA_HTTPD_EXECUTABLE_PATH}" -f "${CLOUDERA_HTTPD_CONF}" -t
    if [ "$?" != 0 ]; then
        die "${CLOUDERA_HTTPD_CONF} is invalid."
    fi
}

# Send a SIGUSR1 to the currently running httpd.
# The pid is in $1
refresh_httpd() {
    pid="$1"
    if [ "${pid}" = "" ]; then
        die "httpd process pid was not specified"
    fi

    exec kill -USR1 ${pid}
}

# Start a fresh httpd
start_httpd() {
    ARGS=""
    if [ -n "${CLOUDERA_HTTPD_USE_SSL}" ]; then
        ARGS="-DCLOUDERA_HTTPD_USE_SSL"
    fi

    # Run httpd in the foreground so the agent can observe if httpd crashes.
    exec "${CLOUDERA_HTTPD_EXECUTABLE_PATH}" -f "${CLOUDERA_HTTPD_CONF}" -DFOREGROUND ${ARGS}
}

#
# Execution starts here
#

set -x

# Time marker for both stderr and stdout
date; date 1>&2

# More conservative umask for logs, indexes, and such.
umask 027

cloudera_config=`dirname $0`
. ${cloudera_config}/cloudera-config.sh

# $1 is the service that this load balancer is a part of. That
# service's directory should contain a file by the name of "httpd.sh"
# that we can source. The name itself should only consist of letters
# in order to prevent any strange redirection attacks.

if [ "$#" -lt 1 ]; then
    die "Service name required as first argument"
fi
SERVICENAME=$1
shift

if ! echo ${SERVICENAME} | grep -q '^[A-Za-z0-9]*$'; then
    die "SERVICENAME contains non-standard characters"
fi

SERVICEFILE="${cloudera_config}/../${SERVICENAME}/httpd.sh"
if [ ! -f "${SERVICEFILE}" ]; then
    die "Unable to locate ${SERVICEFILE}"
fi

# Finally, source the file with service-specific actions
. "${SERVICEFILE}"

# Load the parcel environment
source_parcel_environment

# Chmod scripts so they can run
make_scripts_executable

if [ "$1" = "refresh" ]; then
    # The pid is in $2 (we hope)
    set_httpd_refresh_conf_dir "$2"
fi

# Set and then configure httpd-related directories.
set_httpd_dirs
config_httpd_dirs

# Call out to the script with actions specific to this type of service
service_specific_actions

# Make sure the config is valid
check_httpd_conf

if [ "$1" = "refresh" ]; then
    refresh_httpd "$2"
else
    start_httpd
fi
