#!/usr/bin/python

# Copyright (c) 2019 Cloudera, Inc. All rights reserved.

import argparse
import logging
import os
import signal
import subprocess
import sys
import time

IPERF_PROC_NAME = "iperf3"
IPERF_PROC_BINARY = "./" + IPERF_PROC_NAME
IPERF_LIB = "./libiperf.so.0"

def check_lib_exists_and_preload(dir_path, skip_ld_preload=False):
  if not skip_ld_preload:
    lib_path = os.path.join(dir_path, IPERF_LIB)
    exec_path = os.path.join(dir_path, IPERF_PROC_NAME)
    if not os.path.isfile(lib_path):
      logging.error("Library %s not found", lib_path)
      return False
    if not os.path.isfile(exec_path):
      logging.error("Executable %s not found", exec_path)
      return False
    os.environ["LD_PRELOAD"] = IPERF_LIB
  return True

def stop_iperf_process(proc_name=IPERF_PROC_NAME, preemptive_stop=False):
  proc = subprocess.Popen(["pidof", "-x", os.path.split(proc_name)[1]], stdout=subprocess.PIPE,
                          close_fds=True)
  out = proc.communicate()[0]
  retcode = proc.poll()
  if retcode:
    if not preemptive_stop:
      logging.error("Failed stopping %s process, return code: %d, msg: %s", proc_name,
                    retcode, "process not found" if retcode == 1 else "")
    return False

  pid = int(out)
  os.kill(pid, signal.SIGTERM)
  logging.info("Successfully terminated %s %s process with pid %d",
               "previously running" if preemptive_stop else "", proc_name, pid)
  return True


def start_iperf_process(proc_name=IPERF_PROC_NAME, skip_ld_preload=False,
                        binary_name=IPERF_PROC_BINARY, args="-s"):
  # iperf3 binary and shared library are installed in same directory as this python script
  dir_path = os.path.dirname(os.path.realpath(__file__))
  if not check_lib_exists_and_preload(dir_path, skip_ld_preload):
    return False

  # Helps stop previously running process which may have been leftover after abort or such.
  try:
    stop_iperf_process(proc_name=proc_name, preemptive_stop=True)
  except:
    logging.info("No previously running %s process or failure to stop it", proc_name)
    pass

  # We've done the validations above for existence of executable and library.
  # If an exception is thrown below on starting iperf3 server, let the exception
  # be raised and this one-off process will exit with non-zero exit code leading
  # to failure of command.
  cmd = binary_name
  proc = subprocess.Popen([cmd, args], env=os.environ, cwd=dir_path,
                          stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

  # If there is some problem during iperf3 server startup due to port bind issue or
  # such it'll exit quickly. We expect this process to keep runnning.
  # Hence wait for sometime and ensure the process is still running.
  time.sleep(2)
  return_code = proc.poll()
  if return_code:
    out, err = proc.communicate()
    logging.error("Failed starting %s process. returncode: %s", proc_name, return_code)
    logging.error("stderr: %s", err)
    logging.error("stdout: %s", out)
    return False
  logging.info("Successfully started %s process with pid %d", proc_name, proc.pid)
  return True


if __name__ == "__main__":
  logging.basicConfig(stream=sys.stdout, level=logging.INFO,
                      format='%(asctime)s %(levelname)s %(message)s ')
  parser = argparse.ArgumentParser(description="Start/Stop " + IPERF_PROC_NAME + " server")
  parser.add_argument("operation", help="Start or Stop " + IPERF_PROC_NAME + " server")
  args = parser.parse_args()

  if args.operation == "start":
    success = start_iperf_process()
  elif args.operation == "stop":
    success = stop_iperf_process()
  else:
    raise ValueError("Unknown operation '" + args.operation + "'")
  if not success:
    exit(1)
