#!/usr/bin/python

# Copyright (c) 2018 Cloudera, Inc. All rights reserved.

import argparse
import json
import logging
import os
import pprint
import re
import subprocess
import sys
import time

import bandwidth_server_start_stop

DEFAULT_CHUNK_SIZE = 50
DEFAULT_TIMEOUT_EPSILON_SECS = 2
# Packet interval below 200 millisecs is not allowed for non-root users.
MIN_PACKET_INTERVAL_SECS = 0.2

IPERF_PROC_NAME = "iperf3"
IPERF_LIB = "./libiperf.so.0"

def read_input(file_name):
  # if the file doesn't exist open will throw exception and process will exit with error code 1
  with open(file_name, 'r') as f:
    input_json = json.load(f)
  return input_json

# Break input_list into sublists each with size chunk_size
def make_chunks(input_list, chunk_size):
  if len(input_list) == 0:
    return [[]]
  chunks = []
  for ii in range(0, len(input_list), chunk_size):
    chunks.append(input_list[ii:ii + chunk_size])
  return chunks

def get_failure_host_stat(host_name):
  host_out = {}
  host_out["hostName"] = host_name
  host_out["success"] = False
  host_out["packetLossPercent"] = 100
  host_out["minLatencyMillisecs"] = -1
  host_out["avgLatencyMillisecs"] = -1
  host_out["maxLatencyMillisecs"] = -1
  host_out["mdevLatencyMillisecs"] = 0

  return host_out

def parse_xmit_stats_line(xmit_stats_line):
  # Sample xmit_stats_line: 2 packets transmitted, 2 received, 30.0% packet loss, time 1001ms
  xmit_stats = xmit_stats_line.split(",")
  packet_loss = float(xmit_stats[2].split("%")[0])

  return packet_loss

def parse_timing_stats_line(timing_stats_line):
  # Sample timing_stats_line: rtt min/avg/max/mdev = 9.796/9.799/9.803/0.099 ms
  timing_stats = timing_stats_line.split("=")[1].split("/")

  ping_min = float(timing_stats[0])
  ping_avg = float(timing_stats[1])
  ping_max = float(timing_stats[2])
  # Look for floating point or decimal number, removing 'ms'
  ping_mdev = re.search(r"[0-9]*(\.[0-9]+)?", timing_stats[3])
  ping_mdev = float(ping_mdev.group(0))

  timing_stats = {}
  timing_stats["minLatencyMillisecs"] = ping_min
  timing_stats["avgLatencyMillisecs"] = ping_avg
  timing_stats["maxLatencyMillisecs"] = ping_max
  timing_stats["mdevLatencyMillisecs"] = ping_mdev
  return timing_stats

def kill_child_processes_on_timeout(pmap, rethrow_exception=False):
  for host, proc in pmap.items():
    if proc.poll() is not None:
      continue
    try:
      proc.stdout.close()
      proc.terminate()
      if proc.poll() is None:
        time.sleep(0.1)
        proc.kill()
    except:
      logging.warning("Error terminating child process for host %s on timeout" % host)
      if rethrow_exception:
        raise

def fetch_ping_output(pmap, timeout_secs):
  host_out_list = []
  time_to_exit = time.time() + int(timeout_secs) + DEFAULT_TIMEOUT_EPSILON_SECS
  while pmap:
    if time.time() > time_to_exit:
      # Each ping command is run with deadline. So each process is expected to finish
      # within supplied timeout. This additional check guards against ping misbehaving
      # and ensuring one-off process does exit.
      kill_child_processes_on_timeout(pmap)
      raise RuntimeError("Timeout error while fetching ping output")
    for host, proc in pmap.items():
      # Check whether ping finished.

      if proc.poll() is not None:
        if proc.returncode == 0:
          output_lines = proc.communicate()[0].splitlines()
          packet_loss = parse_xmit_stats_line(output_lines[-2])
          timing_stats = parse_timing_stats_line(output_lines[-1])
          logging.info("Ping test host %s stats: loss: %.3f%% min: %.3f, avg: %.3f, max: %.3f, mdev: %.3f",
                       host, packet_loss, timing_stats["minLatencyMillisecs"],
                       timing_stats["avgLatencyMillisecs"], timing_stats["maxLatencyMillisecs"],
                       timing_stats["mdevLatencyMillisecs"])
          host_out = {}
          host_out["hostName"] = host
          host_out["success"] = True
          host_out["packetLossPercent"] = packet_loss
          host_out.update(timing_stats)
          host_out_list.append(host_out)
        else:
          # Invalid hostnames and such errors would all fall here.
          logging.error("Failed pinging host %s" % host)
          host_out_list.append(get_failure_host_stat(host))
        # remove host from the process list
        del pmap[host]
    # prevent busy looping
    time.sleep(1)

  return host_out_list

def persist_output(host_stat_list, out_file_name):
  result = {}
  result["targetHostsNetStats"] = host_stat_list
  logging.debug("Final result:\n %s", pprint.pformat(result))
  with open(out_file_name, "w") as write_file:
    json.dump(result, write_file, indent=4)

def run_pings(host_chunks, count, packet_size, timeout_secs):
  success = True
  batch_timeout_count = 0
  host_stat_list = []
  for host_batch in host_chunks:
    pmap = {}
    # Create processes for invoking ping command in parallel
    for host in host_batch:
      # Using lowest possible interval of 200 millisecs instead of waiting for default
      # interval of 1 sec to reduce overall time for latency test.
      # Adaptive ping interval (-A) is not available on all platforms we support.
      pmap[host] = subprocess.Popen(['ping', '-i', str(MIN_PACKET_INTERVAL_SECS), '-n',
                                     '-c', str(count), '-w', str(timeout_secs),
                                     '-s', str(packet_size), host],
                                     stdout=subprocess.PIPE, close_fds=True)
    try:
      host_stat_batch_list = fetch_ping_output(pmap, timeout_secs)
      host_stat_list.extend(host_stat_batch_list)
    except RuntimeError, rte:
      batch_timeout_count = batch_timeout_count + 1

  if batch_timeout_count != 0:
    logging.error("Timeout hit by %d batches of hosts" % batch_timeout_count)
    success = False

  return (host_stat_list, success)

def populate_bandwidth_failure_host_stat(host_stat):
  # Results of bandwidth test does not impact "success" status
  host_stat["bwSentBitsPerSecond"] = -1
  host_stat["bwReceivedBitsPerSecond"] = -1

def run_bandwidth_diagnostics(host_stat_list, timeout_secs, bw_process=IPERF_PROC_NAME,
                              skip_ld_preload=False):
  # iperf3 binary and shared library are installed in same directory as this python script
  dir_path = os.path.dirname(os.path.realpath(__file__))
  if not bandwidth_server_start_stop.check_lib_exists_and_preload(dir_path, skip_ld_preload):
    return False

  cmd = "./" + bw_process
  for host_stat in host_stat_list:
    hostname = host_stat["hostName"]
    if not host_stat["success"]:
      logging.info("Skipping bandwidth test for host %s due to failed latency test",
                   hostname)
      continue

    # subprocess.check_output would be ideal but it's not available on python 2.6
    proc = subprocess.Popen([cmd, "-c", hostname, "-t", str(timeout_secs), "-J", "-R"],
                            stdout=subprocess.PIPE, close_fds=True, env=os.environ,
                            cwd=dir_path)
    out_str = proc.communicate()[0]
    retcode = proc.poll()
    if retcode:
      error_msg = json.loads(out_str)["error"]
      logging.error("Bandwidth test host: %s error: %s", hostname, error_msg)
      populate_bandwidth_failure_host_stat(host_stat)
      continue

    bw_out = json.loads(out_str)
    host_stat["bwSentBitsPerSecond"] = bw_out["end"]["sum_sent"]["bits_per_second"]
    host_stat["bwReceivedBitsPerSecond"] = bw_out["end"]["sum_received"]["bits_per_second"]
    logging.info("Bandwidth test host: %s, send rate: %d, receive rate: %d", hostname,
                 host_stat["bwSentBitsPerSecond"], host_stat["bwReceivedBitsPerSecond"])

  return True


if __name__ == "__main__":
  logging.basicConfig(stream=sys.stdout, level=logging.INFO,
                      format='%(asctime)s %(levelname)s %(message)s ')
  parser = argparse.ArgumentParser(description="Run performance diagnostics")
  parser.add_argument("input_file_name", help="Input json file name")
  parser.add_argument("output_file_name", help="Output json file name")
  args = parser.parse_args()

  input_json = read_input(args.input_file_name)
  # Splitting list of hosts into smaller chunks to limit number
  # of processes spawned in parallel
  host_chunks = make_chunks(input_json["hostNames"], DEFAULT_CHUNK_SIZE)
  count = input_json["pingCount"]
  packet_size = input_json["pingPacketSizeBytes"]
  timeout_secs = input_json["pingTimeoutSecs"]

  host_stat_list, success = run_pings(host_chunks, count=count, packet_size=packet_size,
                                      timeout_secs=timeout_secs)

  # In case CM agent is upgraded but CM server is running older version then
  # bandwidth related request params may not be present in input json.
  try:
    if input_json["runBandwidthDiagnostics"]:
      success = run_bandwidth_diagnostics(host_stat_list, input_json["bandwidthTimeoutSecs"])
  except KeyError, ke:
    logging.warning("Bandwidth related request params not found in input json, %s", ke)

  persist_output(host_stat_list, args.output_file_name)
  if not success:
    exit(1)
