#!/bin/bash

# Copyright (c) 2013 Cloudera, Inc. All rights reserved.

# Time marker for both stderr and stdout
date; date 1>&2

cloudera_config=`dirname $0`
cloudera_config=`cd "$cloudera_config/../common"; pwd`
. ${cloudera_config}/cloudera-config.sh

# Load the parcel environment
source_parcel_environment

# Attempt to find Java and place it in JAVA_HOME. The catalina.sh script expects
# to find it there.
locate_cdh_java_home

set -x

get_gc_args() {
  # Solr can supply JVM GC args here [ applicable for CDH >=6.3.0 ]

  # Uncomment below to override GC arguments provided by cloudera-config.sh:
  ##  JAVA8_GC_TUNING_ARGS="GC argument for JAVA8"
  ##  JAVA11_GC_TUNING_ARGS="GC argument for JAVA8"
  # function from cloudera-config.sh provides BASIC_GC_TUNING_ARGS based on java version
  set_basic_gc_tuning_args_based_on_java_version
}

echo "using $CDH_VERSION as CDH_VERSION"

if [ "$CDH_VERSION" -ge "5" ]; then
  # OPSAPS-15997 - Disable sourcing default env
  export BIGTOP_DEFAULTS_DIR=""
fi

if [ "$CDH_VERSION" -ge "6" ]; then
  deploy_name="jetty-deployment"
else
  deploy_name="tomcat-deployment"
fi

# Search-replace {{CMF_CONF_DIR}} in files
replace_conf_dir

JAAS_OPT=""
if [ -n "$SECURE_ZNODE" ]; then
JAAS_OPT="--jaas $SOLR_AUTHENTICATION_JAAS_CONF"
fi

# OPSAPS-27158 - ZKCLI_TMPDIR variable is used to prepare java.io.tmpdir option in zkcli
# Directly set the ZKCLI_JVM_FLAGS instead of setting ZKCLI_TMPDIR since the CDH5.4.x version
# does not have relevant changes in solrctl
ZKCLI_TMPDIR=$CONF_DIR/temp
mkdir  $ZKCLI_TMPDIR
export ZKCLI_JVM_FLAGS="-Djava.io.tmpdir=$ZKCLI_TMPDIR $ZKCLI_JVM_FLAGS"
DEBUG=${DEBUG:-"--debug --trace"}

DEFAULT_HDFS_DIR_PERMISSIONS="0755"

# This method has does the following:
# 1. If SECURE_ZNODE env variable is present, prepare ZKCLI_JVM_FLAGS with
# 'java.security.auth.login.config' and 'zkACLProvider'.
#
# 2. If SOLR_KERBEROS_ENABLED is set to true, we acquire kerberos ticket with appropriately using
# the correct keytabs for C5 and C6+ clusters. This is required to perform any HDFS operation(s)
# like creating/deleting a directory, but, not required for any operations pertaining to ZK.
function initialize_security() {
  if [ -n "$SECURE_ZNODE" ]; then
    export ZKCLI_JVM_FLAGS="-Djava.security.auth.login.config=$SOLR_AUTHENTICATION_JAAS_CONF -DzkACLProvider=$SOLR_ZKACL_PROVIDER $ZKCLI_JVM_FLAGS"
  fi
  if [ "$SOLR_KERBEROS_ENABLED" = "true" ]; then
    if [ $CDH_VERSION -ge 6 ]; then
      acquire_kerberos_tgt "$SOLR_KERBEROS_KEYTAB" "$SOLR_KERBEROS_PRINCIPAL" "true"
    else
      acquire_kerberos_tgt "$SOLR_KERBEROS_KEYTAB" "$SOLR_KERBEROS_PRINCIPAL"
    fi
  fi
}

if [ "solrinit" = "$1" ]; then
  # CDH-12906 - solrctl expects SOLR_HOME to point to solr libraries
  # instead of /var/lib/solr
  export SOLR_HOME=$CDH_SOLR_HOME
  exec $CDH_SOLR_HOME/bin/solrctl.sh --zk $SOLR_ZK_ENSEMBLE $JAAS_OPT init
fi

SOLR_TEMP=$CONF_DIR/temp
if [ $CDH_VERSION -lt 6 ]; then
  # Prior to C6, solr uses tomcat/catalina and needs additional environment variables to be set.
  if [ $CDH_VERSION -ge 5 ]; then
    if [ -z $SCM_DEFINES_SCRIPTS ]; then
      TOMCAT_CONF=/etc/solr/$SOLR_SERVER_DIR
    else
      TOMCAT_CONF=$CDH_SOLR_HOME/../../etc/solr/$SOLR_SERVER_DIR
    fi
    export CATALINA_BASE=/var/lib/solr/${deploy_name}
    SOLR_PLUGIN_DIR=${SOLR_PLUGIN_DIR:-/var/lib/solr/lib}

    if [ "stop" != "$1" ]; then
      env TOMCAT_CONF=${TOMCAT_CONF} TOMCAT_DEPLOYMENT=${CATALINA_BASE} \
	  SOLR_HOME=${CDH_SOLR_HOME} SOLR_PLUGIN_DIR=${SOLR_PLUGIN_DIR} \
	  bash ${CDH_SOLR_HOME}/${deploy_name}.sh
    fi
  else
    export CATALINA_BASE=$CDH_SOLR_HOME/server
  fi

  CATALINA_OPTS=$(replace_pid $CATALINA_OPTS)
  export CATALINA_HOME=$TOMCAT_HOME
  export CATALINA_BASE=$CDH_SOLR_HOME/server
  export CATALINA_TMPDIR=$SOLR_TEMP
  export CATALINA_OUT=$SOLR_LOG/solr.out
fi

# SOLR tunes SOLR_SERVER role only via JAVA_OPTS
get_gc_args
SOLR_GC_ARGS="$BASIC_GC_TUNING_ARGS"

# If safety valve ENV value is in-use, then user provided GC args are used
[[ ! -z $SOLR_JVM_GC_ARGS_SAFETY_VALVE ]] && SOLR_GC_ARGS="$SOLR_JVM_GC_ARGS_SAFETY_VALVE"

# Now, replace the final GC args within the respective OPTs args
# NOTE: the below replacement only works if CDH >= 6.3.0 [ i.e the version (& future versions) that supports java 11 for CDH ]
export JAVA_OPTS=$(replace_gc_args "$JAVA_OPTS" "$SOLR_GC_ARGS")

export SOLR_RUN=$CONF_DIR
export SOLR_LOG4J_CONFIG=$CONF_DIR/log4j.properties
export SOLR_HDFS_CONFIG=$CONF_DIR/hadoop-conf
export SENTRY_CONF_DIR=$CONF_DIR/sentry-conf
export HDFS_BIN=$CDH_HADOOP_HOME/../../bin/hdfs

if [ "$SOLR_SENTRY_ENABLED" = "true" ]; then
  export SOLR_AUTHORIZATION_SENTRY_SITE=$CONF_DIR/sentry-site.xml
fi
if [ "$SOLR_SENTRY_SERVICE_ENABLED" = "true" ]; then
  export SOLR_AUTHORIZATION_SENTRY_SITE=$CONF_DIR/sentry-conf/sentry-site.xml
fi

if [ "stop" = "$1" ]; then
  exec $CDH_SOLR_HOME/bin/solrd "$@"
fi

# Execute HDFS command using hadoop configuration
#
# Arguments:
#   $1 - Parameters for hdfs command
function run_hdfs_cmd {
  "$HDFS_BIN" --config "$SOLR_HDFS_CONFIG" "$@"
}

# Checks if the given HDFS directory path is valid and exists.
#
# Arguments:
#   $1 - The HDFS directory path.
function check_hdfs_directory_exists {
  # Checks if the given directory path in $1 is empty or not.
  if [[ -z "$1" ]]; then
    echo "The HDFS directory path is not specified. Exiting."
    exit 1
  fi
  eval run_hdfs_cmd dfs -test -d "$1"
  if [ "$?" != 0 ]; then
    echo "Missing HDFS directory: $1. Exiting."
    exit 1
  fi
}

# Checks if the given local directory path is valid and exists.
#
# Arguments:
#   $1 - The local directory path.
function check_local_directory_exists {
  # Checks if the given directory path in $1 is empty or not.
  if [[ -z "$1" ]]; then
    echo "The directory path is not specified. Exiting."
    exit 1
  fi
  if [ ! -d "$1" ]; then
    echo "Missing directory: $1. Exiting."
    exit 1
  fi
}

# Creates a temporary scratch directory under the current running
# directory and copies the contents from the given local directory
# on the local file system.
#
# Note: This method to copy the files from the known metadata
# directory to the process directory just to safeguard against any
# modification/corruption of files in the metadata directory by the
# invoked process.
#
# Arguments:
#   $1 - The path to the directory on local file system.
#   $2 - The name of the scratch directory to be created.
# Returns:
# The created scratch directory path.
function copy_metadata {
  check_local_directory_exists "$2"
  local scratch_dir="$CONF_DIR/$1"
  mkdir "$scratch_dir"
  chmod 0755 "$scratch_dir"
  cp -r ${2}/* "$scratch_dir"
  echo "$scratch_dir"
}

# Creates the directory in HDFS specified in $1
#
# Arguments:
#   $1 - The directory to be created
#   $2 - The permissions for the created directory
function create_hdfs_directory {
  run_hdfs_cmd dfs -test -d "$1"
  if [ "$?" != 0 ]; then
    echo "Creating directory: $1"
    run_hdfs_cmd dfs -mkdir "$1"
    if [ "$?" != 0 ]; then
      echo "ERROR: Failed to create HDFS directory: $1. Exiting."
      exit 1
    fi
    if [ ! -z $2 ]; then
      run_hdfs_cmd dfs -chmod $2 "$1"
      if [ "$?" != 0 ]; then
        echo "ERROR: Failed to update permissions for HDFS directory: $1."
        exit 1
      fi
    fi
  else
    echo "Directory found: $1. Skipping directory creation."
  fi
}

# Deletes the HDFS directory specified in $1
#
# Arguments:
#   $1 - The directory to be deleted.
function delete_hdfs_directory {
  run_hdfs_cmd dfs -test -d "$1"
  if [ "$?" == 0 ]; then
    echo "Deleting directory: $1"
    run_hdfs_cmd dfs -rm -r -skipTrash "$1"
  fi
}

# Constructs a string of elements of the given array with the given
# delimiter.
#
# Note:
# 1. The callers should make sure that the given array ($2) is not empty.
# 2. A trailing delimiter is always included in the final string.
#
# Arguments:
#   $1 - The delimiter.
#   $2 - The target array.
# Returns:
# The constructed string.
function join_array_with_delimiter {
  local DELIMITER="$1"
  local IFS="$1"
  shift
  echo "$*$DELIMITER"
}

if [ "upgrade" = "$1" ]; then
  if [[ -z "$CMF_SERVER_ROOT" ]]; then
    SOLR_UPGRADE_PATH="$CLOUDERA_DIR/solr-upgrade/"
  else
    SOLR_UPGRADE_PATH="$CMF_SERVER_ROOT/solr-upgrade/"
  fi
  initialize_security
  if [ "backup" = "$2" ]; then
    if [ "local" = "$3" ]; then
      echo "Input arguments: $1/$2 $3 local-fs-dir=$4 backup-dir=$5 role-id=$6"
      check_hdfs_directory_exists "$5"
      check_local_directory_exists "$4"
      pattern="*_shard*_replica*"
      readarray result_array < <(find "${4}" -type d -name "${pattern}")
      if [ ${#result_array[@]} -gt 0 ]; then
        result_array=$(join_array_with_delimiter , ${result_array[@]})
        # Check if we have a sub-directory for this role. If not, create one.
        create_hdfs_directory "${5}/localfs_backup" ${DEFAULT_HDFS_DIR_PERMISSIONS}
        target_dir="${5}/localfs_backup/${6}"
        create_hdfs_directory "${target_dir}" ${DEFAULT_HDFS_DIR_PERMISSIONS}
        # Copy all directories matching the pattern and copy to the backup location with
        # overwrite.
        echo "Moving directories in $4 matching pattern $pattern to $target_dir"
        run_hdfs_cmd dfs -copyFromLocal -f `eval 'echo {'$result_array'}'` "${target_dir}"
        if [ "$?" != 0 ]; then
          echo "ERROR: Failed to backup local directory $4 to ${target_dir}"
          exit 1
        else
          echo "Successfully performed backup of local directory $4 to ${target_dir}"
          exit 0
        fi
      else
        echo "Skipping backup. Found no matching directories with pattern $pattern in local data directory: $4"
        exit 0
      fi
    elif [ "zk" = $3 ]; then
      echo "Input arguments: $1/$2 $3 backup-dir=$4"
      check_hdfs_directory_exists "$4"
      # Backup Zookeeper state.
      create_hdfs_directory "${4}/zk_backup" ${DEFAULT_HDFS_DIR_PERMISSIONS}
      echo "Performing Zookeeper state backup to ${4}/zk_backup"
      local_work_dir="$CONF_DIR/work_dir"
      mkdir "$local_work_dir"
      "$SOLR_UPGRADE_PATH/solr-upgrade.sh" $DEBUG --zk $SOLR_ZK_ENSEMBLE download-metadata -d "$local_work_dir"
      if [ "$?" != 0 ]; then
        echo "ERROR: Failed to download metadata to $local_work_dir."
        exit 1
      fi
      run_hdfs_cmd dfs -copyFromLocal -f ${local_work_dir}/* "${4}/zk_backup"
      if [ "$?" != 0 ]; then
        echo "ERROR: Failed to perform Zookeeper state backup to ${4}/zk_backup"
        exit 1
      fi
      echo "Successfully performed Zookeeper state backup to ${4}/zk_backup"
      exit 0
    fi
  elif [ "reinitialize" = "$2" ]; then
    if [ "local" = "$3" ]; then
      echo "Input arguments: $1/$2 $3 local-fs-dir=$4 role-id=$5"
      check_local_directory_exists "$4"
      echo "Deleting directory contents: $4"
      exec rm -rf $4/*
    elif [ "data" = $3 ]; then
      echo "Input arguments: $1/$2 $3, hdfs-data-dir=$4"
      # Reinitialize HDFS data directory state.
      check_hdfs_directory_exists "$4"
      target_dir="${4}/.cmf-c6-upgrade-backup"
      readarray result_array < <(run_hdfs_cmd dfs -ls -C ${4} | grep -v "${target_dir}")
      if [ ${#result_array[@]} -gt 0 ]; then
        result_array=$(join_array_with_delimiter , ${result_array[@]})
        echo "Deleting data in $4"
        run_hdfs_cmd dfs -rmr -skipTrash `eval 'echo {'$result_array'}'`
        if [ "$?" != 0 ]; then
          echo "ERROR: Failed to reinitialize HDFS data directory: $4."
          exit 1
        fi
        echo "Successfully reinitialized HDFS data directory: $4"
      else
        echo "Skipping reinitialize. Found empty HDFS data directory: $4"
      fi
      # Zookeeper delete state.
      echo "Deleting Zookeeper state."
      exec "${CDH_SOLR_HOME}/bin/zkcli.sh" -zkhost "$SOLR_ZK_ENSEMBLE" -cmd clear /
    fi
  if [[ -z "$3" ]]; then
    echo "Metadata directory is required to perform validation."
    exit 1
  fi
  elif [ "validate-metadata" = "$2" ]; then
    scratch_dir=$(copy_metadata "validate.metadata.temp" "$3")
    exec "${SOLR_UPGRADE_PATH}/solr-upgrade.sh" $DEBUG --zk $SOLR_ZK_ENSEMBLE validate-metadata -c "$scratch_dir"
  elif [ "bootstrap-config" = "$2" ]; then
    scratch_dir=$(copy_metadata "bootstrap.config.temp" "$3")
    exec "${SOLR_UPGRADE_PATH}/solr-upgrade.sh" $DEBUG --zk $SOLR_ZK_ENSEMBLE bootstrap-config -c "$scratch_dir"
  elif [ "bootstrap-collections" = "$2" ]; then
    scratch_dir=$(copy_metadata "bootstrap.collections.temp" "$3")
    check_hdfs_directory_exists "$4"
    solr_data_dir="$4"
    local_work_dir="$CONF_DIR/work_dir"
    mkdir "$local_work_dir"
    hdfs_work_dir="/tmp/.cmf-c6-solr-bootstrap-collections-$(date +%s)"
    create_hdfs_directory ${hdfs_work_dir}
    "${SOLR_UPGRADE_PATH}/solr-upgrade.sh" $DEBUG --zk $SOLR_ZK_ENSEMBLE bootstrap-collections -c ${scratch_dir} -d ${local_work_dir} -h ${hdfs_work_dir} -s ${solr_data_dir}
    exit_status=$?
    delete_hdfs_directory ${hdfs_work_dir}
    exit ${exit_status}
  elif [ "migrate-sentry-privileges" = "$2" ]; then
    echo "Input arguments: upgrade=$1, migrate-sentry-privileges=$2, release=$3"
    exec "${CDH_SOLR_HOME}/bin/sentryMigrationTool" -c "${SENTRY_CONF_DIR}/sentry-site.xml" -s "$3"
  else
    echo "ERROR: Exiting due to invalid operation: $2."
    exit 1
  fi
fi

if [ "$SYNC_CONFIG_TO_ZK" = "true" ]; then
  SOLR_HOME_TMP=$SOLR_HOME
  #Required due to CDH-12906
  export SOLR_HOME=$CDH_SOLR_HOME

  if [ "$SOLR_SSL_ENABLED" = "true" ]; then
    eval $SOLR_HOME/bin/solrctl.sh --zk $SOLR_ZK_ENSEMBLE $JAAS_OPT cluster --set-property urlScheme https
  else
    eval $SOLR_HOME/bin/solrctl.sh --zk $SOLR_ZK_ENSEMBLE $JAAS_OPT cluster --set-property urlScheme http
  fi

  #Fail the script if unable to set the cluster property.
  if [ 0 == $? ]; then
    echo "Successfully configured urlScheme property for Solr cluster in Zookeeper"
  else
    echo "ERROR: Failed to configure urlScheme property for Solr cluster in Zookeeper"
    exit 1
  fi

  if [ $CDH_VERSION -ge 6 ]; then
    export SOLR_SEC_CONFIG_FILE="${SOLR_HOME}/clusterconfig/hadoop_multi_scheme_config.json"

    if $SOLR_HOME/bin/zksynctool.sh; then
      echo "Successfully configured security for Solr cluster in Zookeeper"
    else
      echo "ERROR: Failed to configure security for Solr cluster in Zookeeper"
      exit 1
    fi

    # set this very high to keep Solr's forced shutdown from interfering with CM's forced shutdown (value in seconds)
    export SOLR_STOP_WAIT=${SOLR_STOP_WAIT:-2592000}
  fi

  #Required due to CDH-27531
  export SOLR_HOME=$SOLR_HOME_TMP
fi

# Create temp dir
if [ $CDH_VERSION -lt 6 ]; then
  mkdir -p $SOLR_TEMP
fi

exec $CDH_SOLR_HOME/bin/solrd run

