#!/usr/lib64/cmf/agent/build/env/bin/python
# Copyright (c) 2014 Cloudera, Inc. All rights reserved.

"""
Dump MR2 job info into a local directory. This script should run as one of the
Hadoop superuser (e.g. `yarn').

Usage:
  collect_jobs_stats.py --app_ids <comma separated list of ids>
                        --output_dir <dir>

Environment:
  HADOOP_CONF_DIR       Directory containing yarn-site.xml and mapred-site.xml.
                        This can also be specified with --hadoop_conf_dir.

Known issue(s):
  - Does not work with SSL enabled.

The output structure looks like:

    outdir/
        command.log
        layout_version.txt
        <app_id>/
            aggregated_log/
                <contents copied from HDFS aggregated log dir>
                <contents copied from Spark driver log dir> (optional)
                spark_event_logs/ (optional)
                    <contents copied from Spark event log dir> (optional)
            app_attempts.json
            app_info.json
            job_attempts.json
            job_conf.json
            job_counters.json
            job_info.json
            job_tasks.json
        <app_id>/
            ...
"""

import logging
import subprocess
import sys
import time
import urllib
import urllib2
import urlparse
from optparse import OptionParser

import os
import re

try:
  import json
except ImportError:
  import simplejson as json

from cmf.util.url import urlopen_with_timeout
from confparse import ConfParse

LAYOUT_VERSION = "1"

LOG_LEVEL = logging.INFO
LOG_FORMAT = "%(asctime)s %(levelname)-5s %(message)s"
LOG = logging.getLogger()


URL_TIMEOUT = 60                # seconds
HDFS_XFER_TIMEOUT = 10 * 60     # seconds

class MRJobInfoGatherer(object):

  def __init__(self, app_ids, output_dir, hadoop_conf_dir):
    self._app_ids = app_ids
    self._output_dir = output_dir
    self._hadoop_conf_dir = hadoop_conf_dir

    self._yarn_conf = ConfParse(file(os.path.join(hadoop_conf_dir, 'yarn-site.xml')))
    self._mapred_conf = ConfParse(file(os.path.join(hadoop_conf_dir, 'mapred-site.xml')))
    self._rm_addresses = get_rm_webapp_addresses(self._yarn_conf)
    self._jhs_addresses = get_jhs_webapp_addresses(self._mapred_conf)
    # Need to refresh yarn conf with the real thing we get from RM
    self._yarn_conf = self._load_yarn_conf()

  def _load_yarn_conf(self):
    """Load the YARN config from the RM itself"""
    data = self._read_rm_url('/conf')
    return ConfParse(data)

  def _read_rm_url(self, path):
    """Return data read from an RM web url. Not thread-safe."""
    res = None
    active_rm = None

    for rm_addr in self._rm_addresses:
      url = urlparse.urljoin(rm_addr, path)
      # In case of a HTTPS request, the parsed url might contain dirty
      # unicode character(s), say of the format - u'https://hostname'.
      # Converting it to a string gives us the right format
      # - 'https://hostname'
      url = str(url)
      try:
        res = urlopen_with_timeout(url, timeout=URL_TIMEOUT).read()
        if not res.startswith("This is standby RM."):
          active_rm = rm_addr
          break
      except urllib2.HTTPError, ex:
        LOG.error("Fail to fetch '%s': %s" % (url, ex))
        raise ex
      except urllib2.URLError, ex:
        LOG.error("Fail to connect to url '%s': %s" % (url, ex))
        continue
    else:
      LOG.error("Failed to find an active RM")
      raise Exception("Failed to find an active RM")

    # Move the active one to the front
    if active_rm is not None and self._rm_addresses[0] != active_rm:
      self._rm_addresses.remove(active_rm)
      self._rm_addresses.insert(0, active_rm)

    return res

  def _dump_spark_event_logs(self, app_id, spark_hs_address):
    """Read spark event logs from spark history server and dump them."""

    url = urlparse.urljoin(spark_hs_address, '/api/v1/applications/%s/logs' % (app_id,))
    try:
      urllib.urlretrieve(url, self._get_output_dir(app_id, 'spark_event_logs.zip'))
    except Exception, ex:
      LOG.error("Fail to retrieve spark event logs url '%s': %s" % (url, ex))
      raise ex

  def _read_jhs_url(self, path):
    """Return data read from a JHS web url."""
    res = None

    for jhs_address in self._jhs_addresses:
      url = urlparse.urljoin(jhs_address, path)
      # In case of a HTTPS request, the parsed url might contain dirty
      # unicode character(s), say of the format - u'https://hostname'.
      # Converting it to a string gives us the right format
      # - 'https://hostname'
      url = str(url)
      try:
        res = urlopen_with_timeout(url, timeout=URL_TIMEOUT).read()
        break
      except urllib2.HTTPError, ex:
        LOG.error("Fail to fetch '%s': %s" % (url, ex))
        raise ex
      except urllib2.URLError, ex:
        LOG.error("Fail to connect to url '%s': %s" % (url, ex))
        continue
    else:
      LOG.error("Failed to connect to the JHS from addresses " + str(self._jhs_addresses))
      raise Exception("Failed to connect to JHS")

    return res

  def _get_output_dir(self, *suffix):
    return os.path.join(self._output_dir, *suffix)

  def _get_redirected_url(self, url):
    opener = urllib2.build_opener(urllib2.HTTPRedirectHandler)
    request = opener.open(url)
    return request.url

  def _get_spark_hs_address(self, tracking_url):
    redirected_url = self._get_redirected_url(tracking_url)
    match = re.search('(?P<spark_hs_address>.*)/history/(.*)', redirected_url)
    if match is None:
      LOG.error("Failed to get spark history server address from tracking url : " + redirected_url)
      raise Exception("Failed to get spark history server address")
    else:
      return match.group('spark_hs_address')

  def gather(self):
    # Write format version
    dump_data(self._get_output_dir('layout_version.txt'), LAYOUT_VERSION)

    try:
      for app_id in self._app_ids:
        self.do_one_app(app_id, os.path.join(self._output_dir, app_id))
    except Exception:
      LOG.exception("Failed to get debug info for application %s" % (app_id,))

  def do_one_app(self, app_id, output_dir):
    LOG.info("Gathering debug info for %s" % (app_id,))

    os.makedirs(output_dir)

    user = None
    is_spark_app = False
    tracking_url = None

    try:
      app_info = self._read_rm_url('/ws/v1/cluster/apps/%s' % (app_id,))
      user = json.loads(app_info)['app']['user']
      dump_data(self._get_output_dir(app_id, 'app_info.json'), app_info)

      app_attempts = self._read_rm_url('/ws/v1/cluster/apps/%s/appattempts' % (app_id,))
      dump_data(self._get_output_dir(app_id, 'app_attempts.json'), app_attempts)

      applicationType = json.loads(app_info)['app']['applicationType']
      if applicationType == 'SPARK':
        is_spark_app = True
        tracking_url = json.loads(app_info)['app']['trackingUrl']

    except:
      LOG.exception("Resource Manager doesn't contain %s" % (app_id,))

    job_id = app_id.replace('application', 'job')
    try:
      job_info = self._read_jhs_url('/ws/v1/history/mapreduce/jobs/%s' % (job_id,))
      if user is None:
        user = json.loads(job_info)['job']['user']
      dump_data(self._get_output_dir(app_id, 'job_info.json'), job_info)

      job_conf = self._read_jhs_url('/ws/v1/history/mapreduce/jobs/%s/conf' % (job_id,))
      dump_data(self._get_output_dir(app_id, 'job_conf.json'), job_conf)

      job_counters = self._read_jhs_url('/ws/v1/history/mapreduce/jobs/%s/counters' % (job_id,))
      dump_data(self._get_output_dir(app_id, 'job_counters.json'), job_counters)

      tasks = self._read_jhs_url('/ws/v1/history/mapreduce/jobs/%s/tasks' % (job_id,))
      dump_data(self._get_output_dir(app_id, 'job_tasks.json'), tasks)

      job_attempts = self._read_jhs_url('/ws/v1/history/mapreduce/jobs/%s/jobattempts' % (job_id,))
      dump_data(self._get_output_dir(app_id, 'job_attempts.json'), job_attempts)

    except:
      LOG.exception("Job History Server doesn't contain %s" % (job_id,))

    if user is not None:
      try:
        self.dump_job_log(job_id, user, output_dir, is_spark_app, tracking_url)
      except:
        LOG.exception("Yarn application logs not found on HDFS for %s" % (job_id,))

  def dump_job_log(self, job_id, user, outdir, is_spark_app, tracking_url):
    """Get the YARN app log from HDFS aggregated directory"""
    app_id = job_id.replace("job_", "application_")
    yarn_log_dir = '%s/%s/%s/%s' % (
        self._yarn_conf['yarn.nodemanager.remote-app-log-dir'],
        user,
        self._yarn_conf['yarn.nodemanager.remote-app-log-dir-suffix'],
        app_id)

    dst_dir = os.path.join(outdir, 'aggregated_logs')
    os.mkdir(dst_dir)

    try:
      cmdv = [ 'hdfs', '--config', self._hadoop_conf_dir, 'dfs', '-get', yarn_log_dir, dst_dir ]
      self._download_hdfs_logs(cmdv, yarn_log_dir)
    except Exception, ex:
      LOG.error(ex)

    """Get the Spark driver and event logs from configured dfs directory"""
    if is_spark_app:
      spark_conf = None
      downloaded_event_logs = False
      if os.environ.get('CDH_SPARK_HOME'):
        spark_conf = os.environ.get('CDH_SPARK_HOME') + "/conf/spark-defaults.conf"
      elif os.environ.get('SPARK_CONF_DIR'):
        spark_conf = os.environ.get('SPARK_CONF_DIR') + "/spark-defaults.conf"
      else:
        LOG.warn("Unable to find spark-defaults.conf file")
      if spark_conf:
        LOG.info("Found spark conf file: '%s'" % spark_conf)
        driver_log_dir = None
        event_log_dir = None
        try:
          with open(spark_conf, "r") as f:
            for line in f:
              line = line.strip()
              if line.startswith('spark.driver.log.dfsDir'):
                driver_log_dir = self._get_prop_val(line)
              elif line.startswith('spark.eventLog.dir'):
                event_log_dir = self._get_prop_val(line)
        except Exception, ex:
          LOG.error("Failed to download spark logs: %s" % ex)
        if driver_log_dir is not None:
          try:
            driver_log_file = driver_log_dir + app_id + "_driver.log"
            cmd_args = ['hdfs', '--config', self._hadoop_conf_dir, 'dfs', '-get', driver_log_file, dst_dir ]
            self._download_hdfs_logs(cmd_args, driver_log_file)
          except Exception, ex:
            LOG.error("Failed to download spark driver logs: %s" % ex)
        if event_log_dir is not None:
          try:
            event_log_file = event_log_dir + app_id + "*"
            event_dir = os.path.join(dst_dir, 'spark_event_logs')
            os.mkdir(event_dir)
            cmd_args = ['hdfs', '--config', self._hadoop_conf_dir, 'dfs', '-get', event_log_file, event_dir ]
            self._download_hdfs_logs(cmd_args, event_log_file)
            downloaded_event_logs = True
          except Exception, ex:
            LOG.error("Failed to download spark event logs: %s" % ex)
      if not downloaded_event_logs and tracking_url is not None:
        LOG.info("Trying to download spark event logs from SHS..")
        try:
          self._dump_spark_event_logs(app_id, self._get_spark_hs_address(tracking_url))
        except Exception, ex:
          LOG.error("Failed to download spark event logs: %s" % ex)

  def _get_prop_val(self, prop):
    prop_value = None
    if '=' in prop:
      prop_value = prop.split("=")[1].split("\n")[0].split(",")[0]
    else:
      prop_value = prop.split(" ")[1].split("\n")[0].split(",")[0]
    if prop_value is not None:
      prop_value = prop_value.strip()
      if not prop_value.endswith('/'):
        prop_value = prop_value + "/"
    return prop_value

  def _download_hdfs_logs(self, cmdv, dir_name):
    LOG.info("Downloading logs from HDFS: '%s'" % (' '.join(cmdv),))
    popen = subprocess.Popen(
        cmdv,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        close_fds=True)
    deadline = time.time() + HDFS_XFER_TIMEOUT
    while time.time() < deadline:
      rc = popen.poll()
      stdout = popen.stdout.read()
      if stdout:
        LOG.info(stdout)
      if rc is not None:
        if rc != 0:
          raise Exception("Failed to download logs from HDFS '%s'" % (dir_name,))
        break
      time.sleep(0.1)
    else:
      LOG.error("Timed out while downloading logs from HDFS '%s'" % (dir_name,))
      try:
        popen.kill()
      except:
        pass


def dump_data(dst_path, data):
  """Write data to a destination file"""
  try:
    try:
      fd = file(dst_path, 'w')
      fd.write(data)
    finally:
      fd.close()
  except Exception:
    LOG.exception("Failed to write to target file '%s'" % (dst_path,))


def get_http_url(hostport):
  return 'http://%s' % hostport

def add_https_url_if_set(res, conf, key):
  if key in conf:
    res.append('https://%s' % conf.get(key))

def get_rm_webapp_addresses(yarn_conf):
  """Return a list of RM webapp addresses"""
  res = []
  webapp_conf_key = 'yarn.resourcemanager.webapp.address'
  webapp_ssl_conf_key = 'yarn.resourcemanager.webapp.https.address'
  if yarn_conf.get('yarn.resourcemanager.ha.enabled', False):
    # RM HA
    rm_ids = yarn_conf.get('yarn.resourcemanager.ha.rm-ids').split(',')
    for rm_id in rm_ids:
      res.append(get_http_url(yarn_conf[webapp_conf_key + '.' + rm_id]))
      add_https_url_if_set(res, yarn_conf, webapp_ssl_conf_key + '.' + rm_id)
  else:
    res.append(get_http_url(yarn_conf[webapp_conf_key]))
    add_https_url_if_set(res, yarn_conf, webapp_ssl_conf_key)
  return res

def get_jhs_webapp_addresses(mapred_conf):
  res = []
  jhs_webapp_conf_key = 'mapreduce.jobhistory.webapp.address'
  jhs_webapp_ssl_conf_key = 'mapreduce.jobhistory.webapp.https.address'

  res.append(get_http_url(mapred_conf[jhs_webapp_conf_key]))
  add_https_url_if_set(res, mapred_conf, jhs_webapp_ssl_conf_key)

  return res

def main(args):
  logging.basicConfig(level=LOG_LEVEL, format=LOG_FORMAT, stream=sys.stderr)

  parser = OptionParser()
  parser.add_option('--app_ids', dest='app_ids',
      help='Comma separate list of application IDs'),
  parser.add_option('--output_dir', dest='output_dir',
      help='Local directory to put application debug info'),
  parser.add_option('--hadoop_conf_dir', dest='hadoop_conf_dir',
      help='Path to Hadoop conf dir'),

  options, args = parser.parse_args(args)

  app_ids = options.app_ids
  if app_ids is None:
    raise ValueError('Missing required argument --app_ids')
  output_dir = options.output_dir
  if output_dir is None:
    raise ValueError('Missing required argument --output_dir')
  hadoop_conf_dir = options.hadoop_conf_dir
  if not hadoop_conf_dir:
    hadoop_conf_dir = os.environ.get('HADOOP_CONF_DIR')
    if not hadoop_conf_dir:
      raise ValueError('Cannot locate Hadoop conf dir')

  os.makedirs(output_dir)

  # Additional log file
  logfile = os.path.join(output_dir, 'command.log')
  log_handler = logging.FileHandler(logfile)
  log_handler.setLevel(LOG_LEVEL)
  log_handler.setFormatter(logging.Formatter(fmt=LOG_FORMAT))
  logging.getLogger().addHandler(log_handler)

  gatherer = MRJobInfoGatherer(app_ids.split(','), output_dir, hadoop_conf_dir)
  gatherer.gather()


if __name__ == '__main__':
  sys.exit(main(sys.argv))
