#!/bin/bash

# Copyright (c) 2012 Cloudera, Inc. All rights reserved.

# Time marker for both stderr and stdout
date; date 1>&2

cloudera_config=`dirname $0`
cloudera_config=`cd "$cloudera_config/../common"; pwd`
. ${cloudera_config}/cloudera-config.sh

# Load the parcel environment
source_parcel_environment

# attempt to find java
locate_cdh_java_home

# simulate /etc/default/hadoop
. ${cloudera_config}/cdh-default-hadoop

echo "using $JAVA_HOME as JAVA_HOME"
echo "using $CDH_VERSION as CDH_VERSION"
echo "using $CDH_YARN_HOME as CDH_YARN_HOME"
echo "using $CDH_MR2_HOME as CDH_MR2_HOME"
echo "using $CONF_DIR as CONF_DIR"

export YARN_CONF_DIR=$CONF_DIR

YARN=$CDH_YARN_HOME/bin/yarn
YSITE="$CONF_DIR/yarn-site.xml"

set_hadoop_classpath

# debug
set -x

# Search-replace {{CMF_CONF_DIR}} in files
replace_conf_dir

# Update path to cgroups prefix in the yarn config
perl -pi -e "s#\{\{CGROUP_GROUP_CPU}}#$CGROUP_GROUP_CPU#g" "$YSITE"

# Check if we are setting up cgroups mount paths.
CGROUP_MOUNT_PROP="yarn.nodemanager.linux-container-executor.cgroups.mount-path"
if grep -q "$CGROUP_MOUNT_PROP" "$YSITE"; then
  echo "Detected cgroups is enabled."
  # Cgroups is on, so create links to named cgroup resources from yarn-ce directory.
  # CGROUP_NAMED_TYPES is set by CM agent.
  CGROUP_LINK_ROOT=$(grep -A 1 "$CGROUP_MOUNT_PROP" "$YSITE" |\
                            grep value | sed -e 's_</*value>__g' -e 's_ __g')
  if [ -d "$CGROUP_LINK_ROOT" ] && [ -n "$CGROUP_NAMED_TYPES" ]; then
    for GTYPE in $CGROUP_NAMED_TYPES; do
      NAME_VAR="CGROUP_NAMED_${GTYPE}_NAME"
      PATH_VAR="CGROUP_NAMED_${GTYPE}_PATH"
      echo "linking cgroup ${GTYPE}/${!NAME_VAR} to ${!PATH_VAR}"
      [ ! -e "$CGROUP_LINK_ROOT/$GTYPE" ] && mkdir "$CGROUP_LINK_ROOT/$GTYPE"
      if [ -e "$CGROUP_LINK_ROOT/$GTYPE/${!NAME_VAR}" ] || \
         [ -h "$CGROUP_LINK_ROOT/$GTYPE/${!NAME_VAR}" ]; then
          rm -f "$CGROUP_LINK_ROOT/$GTYPE/${!NAME_VAR}"
      fi
      ln -s `readlink -m "${!PATH_VAR}"` "$CGROUP_LINK_ROOT/$GTYPE/${!NAME_VAR}"
    done
  fi
fi

# replace {{PID}} in HeapDumpPath option
export YARN_NODEMANAGER_OPTS=$(replace_pid $YARN_NODEMANAGER_OPTS)
export YARN_RESOURCEMANAGER_OPTS=$(replace_pid $YARN_RESOURCEMANAGER_OPTS)
export HADOOP_JOB_HISTORYSERVER_OPTS=$(replace_pid $HADOOP_JOB_HISTORYSERVER_OPTS)

get_gc_args() {
  # yarn can supply JVM GC args here [ applicable for CDH >=6.3.0 ]

  # Uncomment below to override GC arguments provided by cloudera-config.sh:
  ##  JAVA8_GC_TUNING_ARGS="GC argument for JAVA8"
  ##  JAVA11_GC_TUNING_ARGS="GC argument for JAVA8"
  # function from cloudera-config.sh provides BASIC_GC_TUNING_ARGS based on java version
  set_basic_gc_tuning_args_based_on_java_version
}

# yarn tunes nodemanager, resourcemanager and jobhistoryserver roles with the
# same GC parameters.
get_gc_args
YARN_NODEMANAGER_GC_ARGS="$BASIC_GC_TUNING_ARGS"
YARN_RESOURCEMANAGER_GC_ARGS="$BASIC_GC_TUNING_ARGS"
JOB_HISTORYSERVER_GC_ARGS="$BASIC_GC_TUNING_ARGS"

# If safety valve ENV value is in-use, then user provided GC args are used
[[ ! -z $NODEMANAGER_JVM_GC_ARGS_SAFETY_VALVE ]] && YARN_NODEMANAGER_GC_ARGS="$NODEMANAGER_JVM_GC_ARGS_SAFETY_VALVE"
[[ ! -z $RESOURCEMANAGER_JVM_GC_ARGS_SAFETY_VALVE ]] && YARN_RESOURCEMANAGER_GC_ARGS="$RESOURCEMANAGER_JVM_GC_ARGS_SAFETY_VALVE"
[[ ! -z $JOB_HISTORYSERVER_JVM_GC_ARGS_SAFETY_VALVE ]] && JOB_HISTORYSERVER_GC_ARGS="$JOB_HISTORYSERVER_JVM_GC_ARGS_SAFETY_VALVE"

# Now, replace the final GC args within the respective OPTs args
# NOTE: the below replacement only works if CDH >= 6.3.0 [ i.e the version (& future versions) that supports java 11 for CDH ]
export YARN_NODEMANAGER_OPTS=$(replace_gc_args "$YARN_NODEMANAGER_OPTS" "$YARN_NODEMANAGER_GC_ARGS")
export YARN_RESOURCEMANAGER_OPTS=$(replace_gc_args "$YARN_RESOURCEMANAGER_OPTS" "$YARN_RESOURCEMANAGER_GC_ARGS")
export HADOOP_JOB_HISTORYSERVER_OPTS=$(replace_gc_args "$HADOOP_JOB_HISTORYSERVER_OPTS" "$JOB_HISTORYSERVER_GC_ARGS")

# Make sure topology.py is executable
make_scripts_executable

# Disable IPv6.
export YARN_OPTS="-Djava.net.preferIPv4Stack=true $YARN_OPTS"

acquire_kerberos_tgt yarn.keytab

if [ $1 = "RefreshQueuesAndNodes" ]; then
  # If RM is in HA mode, we need to issue command only to active RM
  if [ -n $2 ] && [ $2 = "true" ]; then
    # find if RM is in standby mode
    HASTATE="$($YARN --config $CONF_DIR rmadmin -getServiceState $3)"
    if [[ $HASTATE =~ "standby" ]]; then
      # Standby RM have no state to refresh, exit successfully
      cm_success_and_exit YARN_RM_IN_STANDBY_MODE
    fi
  fi

  $YARN --config $CONF_DIR rmadmin -refreshQueues
  RET=$?
  if [ $RET -eq 0 ]; then
    echo "RefreshQueues completed successfully"
    if [ -n $4 ] && [ $4 != "unused" ]; then
      $YARN --config $CONF_DIR rmadmin -refreshNodes -g $4 -client
    else
      $YARN --config $CONF_DIR rmadmin -refreshNodes
    fi
    RET=$?
    if [ $RET -eq 0 ]; then
      echo "RefreshNodes completed successfully"
    else
      echo "RefreshNodes failed"
    fi
  else
    echo "RefreshQueues failed. Skipping RefreshNodes."
  fi
  exit $RET
elif [ $1 = "historyserver" ]; then
  # historyserver is actually a MR2 daemon.
  MAPRED=$HADOOP_MAPRED_HOME/bin/mapred
  exec $MAPRED "$@"
elif [ $1 = "application-diagnostic-data" ]; then
  set -e
  PYTHON_CMD=$CMF_PACKAGE_DIR/../bin/python
  COLLECT_APPLICATION_DATA="$(dirname $0)/../support/collect_jobs_stats/collect_jobs_stats.py"
  PROCESS_DIR_NAME=$(basename $(pwd))
  DIAGNOSTICS_DUMP_DIR="$HEAP_DUMP_DIR/$PROCESS_DIR_NAME"
  COLLECT_APP_DATA_ARGS="--app_ids $3 --output_dir $DIAGNOSTICS_DUMP_DIR/diagnostics --hadoop_conf_dir $CONF_DIR"
  $PYTHON_CMD $COLLECT_APPLICATION_DATA $COLLECT_APP_DATA_ARGS
  tar -czf $DIAGNOSTICS_DUMP_DIR.tar.gz -C $DIAGNOSTICS_DUMP_DIR diagnostics
  rm -rf $DIAGNOSTICS_DUMP_DIR
  ln -s $DIAGNOSTICS_DUMP_DIR.tar.gz output.gz
elif [ $1 = "install-mr-framework" ]; then
  acquire_kerberos_tgt yarn.keytab
  # Expected args:
  # $2 = HDFS framework path, including "#mr-framework" suffix
  # $3 = HDFS defaultFS URI
  MR_FRAMEWORK_PATH="$2"
  DEFAULT_FS="$3"

  if [ -n "$UPLOAD_TIMEOUT" ]; then
    REPLICATION_ARGS="-timeout $UPLOAD_TIMEOUT $REPLICATION_ARGS"
  fi
  if [ -n "$UPLOAD_ACCEPTABLE_REPLICATION" ]; then
    REPLICATION_ARGS="-acceptableReplication $UPLOAD_ACCEPTABLE_REPLICATION $REPLICATION_ARGS"
  fi
  if [ -n "$UPLOAD_FINAL_REPLICATION" ]; then
    REPLICATION_ARGS="-finalReplication $UPLOAD_FINAL_REPLICATION $REPLICATION_ARGS"
  fi
  if [ -n "$UPLOAD_INITIAL_REPLICATION" ]; then
    REPLICATION_ARGS="-initialReplication $UPLOAD_INITIAL_REPLICATION $REPLICATION_ARGS"
  fi
  : "${BLACKLIST:=-blacklist .*hadoop-yarn-server-applicationhistoryservice.*\.jar,.*hadoop-yarn-server-resourcemanager.*\.jar,.*hadoop-yarn-server-router.*\.jar,.*hadoop-yarn-server-sharedcachemanager.*\.jar,.*hadoop-yarn-server-timeline-pluginstorage.*\.jar,.*hadoop-yarn-server-timelineservice.*\.jar,.*hadoop-yarn-server-timelineservice-hbase.*\.jar,.*javax.ws.rs-api-.*\.jar,.*jersey-.*-2.*\.jar}"
  MAPRED_BIN=$HADOOP_MAPRED_HOME/bin/mapred
  $MAPRED_BIN --config $CONF_DIR frameworkuploader -fs $DEFAULT_FS -target $MR_FRAMEWORK_PATH -nosymlink -whitelist '.*\.jar' $BLACKLIST $REPLICATION_ARGS
else
  if [ $1 = "nodemanager" ]; then
    # In secure mode, container are launched as user who submitted the job and
    # needs read access to the config files. We don't make everything world-readable
    # because this directory also contains files critical from security standpoint.
    chmod go+r $CONF_DIR/yarn-site.xml
    chmod go+r $CONF_DIR/hdfs-site.xml
    chmod go+r $CONF_DIR/core-site.xml
    chmod go+r $CONF_DIR/mapred-site.xml
    chmod go+r $CONF_DIR/ssl-client.xml
    chmod go+r $CONF_DIR/hadoop-policy.xml

    # Below logic is a quick fix to OPSAPS-28378, see the jira for details.
    if [ "$IS_KERBERIZED" = "true" ]; then
      for nm_local_dir in "$NM_LOCAL_DIRS"; do
        for dir in "$nm_local_dir"/usercache/*; do
          if [ -d "$dir" ]; then
            base=$(basename $dir)
            owner=$(stat -c %U $dir)
            if [ "$owner" != "$base" ]; then
              echo "directory owner doesn't match directory name, will try to chown"
              if chown -R $base $dir ; then
                echo "chown succeeded"
              else
                echo "chown failed, will try to remove the directory"
                if rm -rf "$dir" ; then
                  echo "rm succeeded"
                else
                  echo "rm failed"
                fi
              fi
            fi
          fi
        done
      done
    fi
  fi
  exec $YARN "$@"
fi
